      program p3dr
C     Parallel 3D Cartesian Reconstruction program
C                                                     MAIN PROGRAM: P3DR
C **********************************************************************
C *                                                                    *
C * Main routine: P3DR in file aamain.f                                *
C * Author: Robert E. Lynch (rel@cs.purdue.edu) and                    *
C *         Hong Lin (linh@cs.purdue.edu)                              *
C *         Yongchang Ji (yji@cs.ucf.edu)                              *
C * Date:   1997 -- 2002                                               *
C * Function: Main program for cartesian reconstruction                *
C * Algorithm: P3DR uses a set of pixel frames of projections of       *
C *   an object to reconstruct the 3D density map of the object.       *
C *   Firstly, a 2D Fourier transform is carried out on each           *
C *   pixel frame; then the 2D DFT is used in estimating 3D DFT        *
C *   of the object; finally, a reverse Fourier transform is           *
C *   carried out on the 3D DFT to get the density values of the       *
C *   object.                                                          *
C *      P3DR reads in orientations and pixel frames from two          *
C *      files with extension dat and pif/mrc and carries out          *
C *      reconstruction based on these data.                           *
C * Modification history:                                              *
C *   - Robert E. Lynch, Version 0.0, 1997                             *
C *   - Hong Lin, Version 0.1, Apr.20, 1998                            *
C *   - Hong Lin, Version 0.6, Jul.6, 1998                             *
C *   - Hong Lin, Version 1.0, Aug.18, 1998                            *
C *   - Hong Lin, Version 2.0, Sep.29, 1999.                           *
C *   - Hong Lin, Version 2.9, June 17, 2000                           *
C *   - R.E. Lynch, Aug 2000                                           *
C *        Program name set to P3DR                                    *
C *        Moved print of time and memory-use to subroutine MEM_TIME   *
C *        from main routine --- THIS DOESN"T WORK !!!!!!              *
C *        The setting of the value of  diameter  was removed --- the  *
C *          number of angstroms per pixels is not known until the     *
C *          header of the orientation file is read.                   *
C *        Used new variable  interpolants_size  to store the size     *
C *          of the array  interpolants  which is passed to several    *
C *          routines.  Simplified the routine  intlz_interp           *
C *                                                                    *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *
C *                                                                    *
C **********************************************************************

C confine_3D -- When confine_3D is .true., the edge-length of 3D
C   DFT is set to the same as the edge-length of pixel frames. 
C 
      implicit none

      character(len=256) :: infile

#ifdef MPI
      include 'mpif.h'
#endif

      include '../include/sizes.inc'
      include '../include/allprogs.inc'
      include '../include/info.inc'
      include '../include/com_01.inc'
      include '../include/mem_time.inc'
#include "./map_type.inc"

      logical :: confine_3D = .false.

      integer :: mcw, inunit, err_code, alloc_stat
      integer first_iq, last_iq, max_iq_slab,
     +     first_z,last_z, first_hk,last_hk, max_hk_slab, num_hkl,
     +     den_min_x, den_max_x, den_min_y, den_max_y, 
     +     den_min_z, den_max_z, 
     +     min_xyz, max_xyz, hkl_max, max_z_slab,
     +     first_rr_out, last_rr_out, max_rr_out_slab,
     +     first_u, last_u, max_u_slab,
     +     first_y, last_y, max_y_slab
      integer*8 mem_size
      integer i, j, k, id, nptles
      integer nrffts_1_row, nrffts_1_col, nrffts_1_omega,
     $     nrffts_2_row, nrffts_2_col, nrffts_2_omega

      logical all_same_size

C     Variables used for noise suppression
      integer nfsc
      real convert_to_invang
      real, allocatable :: fsc_freq(:), fsc_value(:)

      double precision elapsed_time
      
      integer clock1, clock2

      character(len=18) :: subname = 'P3DR main routine'
      
      real map_min, map_max, map_min_local, map_max_local, res
      real enlargefact
      real fmin, fmax, favg, fstd
      real :: imgprjstd = 1.

      integer, allocatable   :: indx(:)
      integer, allocatable   :: count(:), zeros(:)
      integer, allocatable   :: countl(:), zerosl(:)
      real, allocatable      :: pixels(:), x(:), y(:),mag(:)
      real, allocatable      :: theta(:), phi(:), omega(:)
      real, allocatable      :: delf_maj_ptle(:), delf_min_ptle(:)
      real, allocatable      :: ang_maj_ptle(:)
      real, allocatable      :: f_vru(:,:,:)
      real, allocatable      :: f_vru_tmp(:)
      real, allocatable      :: F_XK_dZ(:,:,:)
      real, allocatable      :: rho(:,:,:), density_map(:,:,:)
      real, allocatable      :: ctf_val(:,:), abs_ctf(:,:)
      complex, allocatable   :: f_hkl(:,:,:)
      FMAP_TYPE, allocatable :: interpolants(:,:,:,:)
      FMAP_TYPE, allocatable :: f_hkl_ref(:,:,:,:)

      integer actual_images ! Number of images with particles

      call system_clock(clock1)

#ifdef MPI
      mcw = mpi_comm_world
      call MPI_INIT (mpi_err)
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_INIT')
      call MPI_COMM_RANK (MCW, iam, mpi_err)
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_COMM_RANK')
      call MPI_COMM_SIZE (MCW, num_nodes, mpi_err)
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_COMM_SIZE')
      if (num_nodes > max_numnodes) then
         write(*, 9001) num_nodes, max_numnodes
 9001    format(//' Requested number of nodes = NUM_NODES    = ',i10/
     +        ' Available number of nodes = MAX_NUMNODES = ',i10/)
         call error_stop(progname, subname,  
     +        'NUM_NODES requested exceeds MAX_NUMNODES')
      endif
#else
      iam = 0
      num_nodes = 1
#endif
      
      progname = 'P3DR'
      
      if( iam == 0 ) then
         write(*,*) 'P3DR: Parallel 3D reconstruction'
#ifdef MPI
         write(*,*) 'Parallel version run on ', num_nodes , ' CPU(s)'
#else
         write(*,*) 'Serial version run on 1 CPU'
#endif
         write(*,*) 'For doc see: http://cryoem.ucsd.edu/programs.shtm'
         write(*,*) 
         
         call parse_cmd_line(infile)
         
         if(len_trim(infile) == 0) then
            inunit = 5
            call info(inunit)
         else
            inunit=17
            open(file=infile, unit=inunit)
            call info(inunit)
            close(inunit)
         endif

         nptles = sum(nptle_per_img)
         if(nptles == 0) then
            call error_stop('P3DR', 'main program', 
     +           'All particle parameter files are empty')
         endif
         
         write(*,*)
         write(*,*) 'Number of parameter files = ', nparam
         write(*,*) 'Number of image files     = ', nimg
         write(*,*) 'Total number of images: ', nptles
         write(*,*)
         
         k = 0
         do i=1,nparam
            do j=1,nimg_per_param(i)
               k = k + 1
               write(*,*) 'param #',i,'orient file: ',
     $              trim(adjustl(param_fnames(i)))
               write(*,*) 'image #',k,'image  file: ',
     $              trim(adjustl(img_fnames(k)))
               write(*,*) 'views = ',  nids(k)
               write(*,*) 'numcols = ',npix_col,'numrows = ',npix_row
               write(*,*)'Ptle records = ', nptle_per_img(k)
               write(*,*)
            enddo
         enddo
      
         call flush_stdout()   
      endif

C     Broadcast number of parameter files, images, particles,
C     images per parameter file, and particles per image
#ifdef MPI
      call mpi_bcast(nparam, 1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nimg,   1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nptles, 1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nimg_per_param, nparam, mpi_integer, 
     $     0, mcw, mpi_err)
      call mpi_bcast(nptle_per_img, nimg, mpi_integer, 
     $     0, mcw, mpi_err)
      call mpi_bcast(first_index, nimg, mpi_integer, 
     $     0, mcw, mpi_err)
#endif

#ifdef MPI
C     Broadcast parameters from root to all other processes
      call bcast_parameters 
      call mpi_bcast(oddeven, 1, mpi_integer, 0, mcw, mpi_err)
#endif
      
C     Process FSC file
      nfsc = 0
      if(iam == 0) then
         if(fsc_file_name(1:1) == ' ') then
            write(*,*)
            write(*,*) 'No FSC file provided - no noise suppression'
            write(*,*)
         else
            write(*,*)
            write(*,*) 'Noise suppression w/ FSC file = ',
     $           trim(fsc_file_name)
            write(*,*)
            open(unit=21, file=fsc_file_name)
            read(21,*) nfsc
         endif
         call flush_stdout()
      endif

#ifdef MPI
      call mpi_bcast(nfsc, 1, mpi_integer, 0, mcw, mpi_err)
#endif      

      allocate(fsc_freq(nfsc), fsc_value(nfsc))      
      if(nfsc > 0) then
         if(iam == 0) then
            do i=1,nfsc
               read(21,*) fsc_freq(i), fsc_value(i)
            enddo
            close(21)
            write(*,*) 'Number of FSC values = ', nfsc
            write(*,*) 'i   spatial freq      FSC'
            do i=1,nfsc
               write(*,'(i3,4x,f7.4,6x,f7.4)') 
     $             i, fsc_freq(i), fsc_value(i)
            enddo
            write(*,*)
            call flush_stdout()
         endif
      
#ifdef MPI      
         call mpi_bcast(fsc_freq,  nfsc, mpi_real, 0, mcw, mpi_err)
         call mpi_bcast(fsc_value, nfsc, mpi_real, 0, mcw, mpi_err)
#endif
      endif

      call intlz_params(hkl_max, num_hkl, min_xyz, max_xyz, 
     +     nrffta_row, nrffta_col, nrffta_omega, nfc_2d_row, 
     +     nfc_2d_col, nrffts_1_row, nrffts_1_col, 
     +     nrffts_1_omega, nrffts_2_row, nrffts_2_col, 
     +     nrffts_2_omega, den_min_x, den_max_x, den_min_y, 
     +     den_max_y, den_min_z, den_max_z,
     +     confine_3D)
      
      if( iam == 0 ) then
         print *, ' zero_fill after  call of INTLZ_PARAMS =',zero_fill
         print *, ' nrffta_row after call of INTLZ_PARAMS =',nrffta_row
      endif
      
      call intlz_arrays(hkl_max, min_xyz, max_xyz, min_z, max_z,
     +     first_hk, last_hk, max_hk_slab, first_iq, last_iq,
     +     max_iq_slab, first_rr_out, last_rr_out, max_rr_out_slab,
     +     first_u, last_u, max_u_slab,
     +     first_y, last_y, max_y_slab, first_z, last_z, max_z_slab,
     +     npix_row)      

      
C     **********************************************************************
C     *             2D DFT and estimating 3D DFT from 2D DFT               *
C     **********************************************************************
      
      allocate(ctf_val(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col), 
     +     abs_ctf(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col),
     +     interpolants(3,-hkl_max:hkl_max,-hkl_max:hkl_max,0:hkl_max),
     +     stat = alloc_stat)
      if(alloc_stat > 0) then
         call error_stop (progname, subname,  
     +        'Memory allocation ctf_val, abs_ctf, interpolants failed')
      endif 
      
C     Zero interpolants array
      interpolants = 0.0        ! Fortran 90 array syntax

#ifdef MPI
      call mpi_bcast(param_fnames, filename_len*pfilesmx, mpi_character,
     $     0, mcw, mpi_err)
      call mpi_bcast(img_fnames, filename_len*pfilesmx, mpi_character,
     $     0, mcw, mpi_err)
#endif

      
C     Allocate global arrays for particle parameters
      allocate(indx(nptles), theta(nptles), phi(nptles), omega(nptles),
     $     x(nptles), y(nptles), mag(nptles))
      
      if(per_ptle_ctf == 1) then
         allocate(delf_maj_ptle(nptles), delf_min_ptle(nptles),
     $        ang_maj_ptle(nptles))
      endif

C     Get particle parameters
      if(iam == 0) then
         if(per_ptle_ctf == 1) then
            call get_ptle_ioomctf(nparam, filename_len, param_fnames,
     $           oddeven, nptles, nbox, indx, theta, phi, omega, x, y,
     $           mag, delf_maj_ptle, delf_min_ptle, ang_maj_ptle)
         else                  
            call get_ptle_ioom(nparam, filename_len, param_fnames,
     $           oddeven, nptles, nbox, indx, theta, phi, omega, x, y,
     $           mag)
         endif
      endif

C     Distribute particle information to all other processes
#ifdef MPI
      call mpi_bcast(indx,  nptles, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(theta, nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(phi,   nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(omega, nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(x,     nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(y,     nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(mag,   nptles, mpi_real,    0, mcw, mpi_err)
      if(per_ptle_ctf == 1) then
         call mpi_bcast(delf_maj_ptle, nptles, mpi_real, 0,mcw,mpi_err)
         call mpi_bcast(delf_min_ptle, nptles, mpi_real, 0,mcw,mpi_err)
         call mpi_bcast(ang_maj_ptle,  nptles, mpi_real, 0,mcw,mpi_err)
      endif
      call mpi_barrier(mcw, mpi_err)
#endif

C allocate particle buffer only once if size does not change
      all_same_size = .false.
      if(minval(nbox(1:nimg))==maxval(nbox(1:nimg))) then
         all_same_size = .true.
      endif
      if(all_same_size) then
         npix_row = nbox(1)
         npix_col = nbox(1)
         allocate(pixels(npix_col*npix_row))
      endif
      if(iam == 0) then
         write(*,*) 
         write(*,*) '---------------------------------------------'
         write(*,*) 
         call flush_stdout()
      endif

      id = 0
      actual_images = 0
      do i = 1, nimg
         
         if(iam == 0) then
            write(*,*) 'Processing image # ', i
            write(*,*) 'image  file: ',
     $           trim(adjustl(img_fnames(i)))
            call flush_stdout()
         endif

         npix_row         =  nbox(i)
         npix_col         =  nbox(i)
         angstrom_per_pix =  pixsizs(i)
         
         j = first_index(i)
         if(j <= 0) cycle       ! skipping micrographs w/o images
         actual_images = actual_images + 1
         if(.not. all_same_size) then
            allocate(pixels(npix_col*npix_row))
         endif
         call gen_read_pixel_frame(img_fnames(i), pix_file_unit,
     $        j, npix_col, npix_row, 
     $        npix_col, npix_row, 1, pixels, err_code)

         call frame_stats(npix_row, npix_col, pixels,
     $        fmin, fmax, favg, fstd)
         
         if(actual_images == 1) then
            imgprjstd   = fstd
            enlargefact = 1. 
         else 
            enlargefact = imgprjstd / fstd
         endif
         
C     Initialize CTF arrays
         ctf_val = 1.0          ! Fortran 90 array syntax
         abs_ctf = 1.0          ! Fortran 90 array syntax
         
C     Calculate CTF corrections
         if(ctfmode > 0 .and. ctfmode <= 4)  then
            if(per_ptle_ctf == 0) then
               call ctf_para(ctf_val, abs_ctf, 
     $              0, nfc_2d_row, -nfc_2d_col, nfc_2d_col, .true.,
     $              ctfmode, filter, tempfac, ctf_ff1, ctf_ff2,
     $              pixsizs(i), units, volts(i), 
     $              delf_majs(i), delf_mins(i), 
     $              ang_majs(i), ampfacs(i),
     $              cs_coeffs(i), enlargefact, res_max, res_min)
            endif
         endif
         
         do j=1, nptle_per_img(i)
            id = id + 1
            
            if(iam == mod(id-1,num_nodes)) then

               err_code = 0

               call gen_read_pixel_frame(img_fnames(i), pix_file_unit,
     $              indx(id), npix_col, npix_row, 
     $              npix_col, npix_row, 1, pixels, err_code)

               if(err_code > 0) then
                  write(*,*) ' --------- WARNING ---------'
                  write(*,*) 'Invalid image'
                  write(*,*) 'Skipping image number ', indx(id)
                  write(*,*) 'Micrograph ', trim(img_fnames(i))
                  write(*,*) 
                  cycle
               endif

               if(per_ptle_ctf == 1) then 
                  call ctf_para(ctf_val, abs_ctf, 
     $                 0, nfc_2d_row, -nfc_2d_col, nfc_2d_col, .true., 
     $                 ctfmode, filter, tempfac, ctf_ff1, ctf_ff2, 
     $                 pixsizs(i), units, volts(i), 
     $                 delf_maj_ptle(id), delf_min_ptle(id), 
     $                 ang_maj_ptle(id), ampfacs(i), 
     $                 cs_coeffs(i), enlargefact, res_max, res_min) 
               endif            
               
               call cmpt_intrps(hkl_max, interpolants,
     $              pixels, npix_col*npix_row,
     $              ctf_val, abs_ctf, theta(id), phi(id), omega(id),
     $              x(id), y(id), mag(id))
            endif
         enddo
         
         if(.not. all_same_size) then
            deallocate(pixels)
         endif
      enddo

      if(all_same_size) then
         deallocate(pixels)
      endif
      if(iam == 0) then
         write(*,*)
         write(*,*) '----- Done processing images -----'
         write(*,*)
         call flush_stdout()
      endif
      
      deallocate(indx, theta, phi, omega, x, y, mag) 

      if(per_ptle_ctf == 1) then
         deallocate(delf_maj_ptle, delf_min_ptle, ang_maj_ptle)
      endif

C     Apply 32 symmetry operations to interpolants array
      if(symm_code == 532) then
         call apply32(interpolants, hkl_max)
      endif

      deallocate(ctf_val, abs_ctf)

C     Apply noise suppression algorithm
C     Rosenthal and Henderson, JMB 333 721-745 (2003)
      if (nfsc > 0) then
         if(iam == 0 .and. .not. all_same_size) then
            write(*,*) 'WARNING: noise suppression is not correct!'
            write(*,*) ' -- It is calibrated on last pixel size --'
         endif
         convert_to_invang = (mapmagfactor * nfc_2d_row) /
     $     (nrffta_row * angstrom_per_pix * hkl_max)
         call noise_suppression(interpolants, hkl_max, 
     $        convert_to_invang, nfsc, fsc_freq, fsc_value)
      endif

C     **********************************************************************
C     *        Combining estimates of 3D DFT along three directions        *
C     *        arrange the data before 3D DFT                              * 
C     **********************************************************************
C GCA: the next two MPI blocks are inter-related     
#ifdef MPI
      allocate(f_hkl_ref(3, -hkl_max:hkl_max, 
     +     -hkl_max:hkl_max, first_iq:last_iq), stat=alloc_stat)
      if (alloc_stat > 0 ) call error_stop(progname, subname,
     +     'Error in allocating memory for F_HKL_ref')

      call exch_intp(hkl_max, interpolants, f_hkl_ref)
      deallocate (interpolants)
      allocate(countl(0:hkl_max), zerosl(0:hkl_max))

#endif

      allocate(f_hkl(-hkl_max:hkl_max, -hkl_max:hkl_max,
     +     first_iq:last_iq), stat = alloc_stat)
      if (alloc_stat > 0 ) call error_stop(progname, subname,
     +     'Error in allocating memory for F_HKL')

      allocate(count(0:hkl_max), zeros(0:hkl_max))

#ifdef MPI
      call realtocomplx(hkl_max, first_iq, last_iq, f_hkl_ref, 
     $     countl, zerosl, f_hkl)
      deallocate(f_hkl_ref)

      call mpi_reduce(countl(0), count(0), hkl_max+1, MPI_INTEGER,
     +     MPI_SUM, 0, mcw, mpi_err)
      call mpi_reduce(zerosl(0), zeros(0), hkl_max+1, MPI_INTEGER,
     +     MPI_SUM, 0, mcw, mpi_err)
      deallocate(countl, zerosl)

#else
      call realtocomplx(hkl_max, first_iq, last_iq, interpolants, 
     $     count, zeros, f_hkl)
      deallocate (interpolants)
#endif

      if(iam == 0) then
         write(*,*)
         write(*,*) 'Statistics for empty voxels in reciprocal space'
         write(*,*) 'count   zeros     res'
         write(*,*) '------  ------  --------'
         do i = 1, hkl_max
            res = real(nrffta_row)*angstrom_per_pix/i
            if(res > (res_max+0.001)) then
               write(*,'(i6,2x,i6,2x,f8.2)') count(i), zeros(i), res
            endif
         enddo
         write(*,*)
      endif

      deallocate(count, zeros)

C     ----- This is the end of combining 3D DFT along 3 directions -----
      
C     ********************************************************************
C     *                    2D Fourier synthesis                          *
C     ********************************************************************
      
      
C     Allocate memory
      allocate(f_vru(min_xyz:max_xyz, first_rr_out:last_rr_out,
     +     min_xyz:max_xyz), stat=alloc_stat)
      if (alloc_stat > 0) call error_stop(progname, subname, 
     +     'Error in allocating memory for F_VRU')
      
C     FFTSYNTH_1
C     **********
      call fftsynth_1_m_slab(hkl_max,
     +     first_iq, last_iq, first_rr_out, last_rr_out,
     +     f_hkl, nrffts_1_row, nrffts_1_col, nrffts_1_omega, 
     +     min_x, max_x, min_z, max_z, 
     +     min_xyz, max_xyz, f_vru, mem_fftsynth_1_m )
      
      deallocate(f_hkl)

C     ------------ This is the end of 2D Fourier synthesis -----------------
      
C     **********************************************************************
C     *           Exchanging data for 1D Fourier synthesis                 *
C     **********************************************************************
      
C GCA: the next two MPI blocks are inter-related     
#ifdef MPI
C     Allocate storage for subroutine EXCHANGE_2
      mem_size = int8(-min_xyz+max_xyz+1) * (-min_xyz+max_xyz+1) *
     +     (-frst_lst_U(1,iam)+frst_lst_U(2,iam)+1)
      allocate(F_VRU_TMP(mem_size), stat = alloc_stat)
      if (alloc_stat > 0) call error_stop(progname,
     +     'exchange_2_m', 'Error in allocating memory for F_VRU_TMP!')
      
      call exchange_2_slab( min_xyz, max_xyz, F_VRU,
     +     nrffts_1_row, F_VRU_TMP )

      deallocate(F_VRU)
#endif
      
      allocate(f_xk_dz(min_x:max_x, nrffts_1_row,
     +     frst_lst_u(1,iam):frst_lst_u(2,iam)), stat=alloc_stat)
      if(alloc_stat > 0 ) call error_stop (progname, subname,  
     +     ' Memory allocation failed')
      
#ifdef MPI
      call rearrange_2_slab(min_xyz, max_xyz, nrffts_1_row,
     +     F_VRU_TMP, min_x, max_x, F_XK_dZ )

C     Free memory used in 2D Fourier synthesis.
      deallocate(F_VRU_TMP)
#else
      call rearrange_2_slab(min_xyz, max_xyz, nrffts_1_row,
     +     F_VRU, min_x, max_x, F_XK_dZ )
      deallocate(F_VRU)
#endif
      
C     ----- This is the end of exchanging data for 1D Fourier synthesis ----
C     ------ Before move onto the next step, synchronize all the nodes -----
      
C     **********************************************************************
C     *                        1D Fourier synthesis                        *
C     **********************************************************************
      
      
C     Allocate storage for 1D Fourier synthesis.
      allocate(rho(min_x:max_x, min_y:max_y, first_z:last_z),
     +     stat = alloc_stat)
      if (alloc_stat > 0) call error_stop(progname, subname, 
     +     'Error in allocating memory for RHO')
      
C     CALL FFTSYNTH_2
C     ***************
      call fftsynth_2_m_slab( nrffts_1_row, min_x, max_x, min_y, 
     +     max_y, min_z, max_z, F_XK_dZ, nrffts_2_row, 
     +     nrffts_2_col, nrffts_2_omega, rho, 
     +     mem_fftsynth_2_m )
      
C     Free memory used in 2D Fourier synthesis.
      deallocate(F_XK_dZ)
      
C     -------------- This is the end of 1D Fourier synthesis --------------
      
C     *********************************************************************
C     clear the density map and output density map
C     *********************************************************************
      
C     according to the set_indice, node 1 alway has the most largest slab,
      
      allocate(density_map(den_min_x:den_max_x, den_min_y:den_max_y,
     $     max_z_slab), stat = alloc_stat)
      if (alloc_stat > 0) call error_stop(progname,
     +     subname,  'Error in allocating density_map')
      
      call density_clear(rho, min_x, max_x, min_y, max_y,
     +     den_min_x, den_max_x,
     +     frst_lst_z(1,iam), frst_lst_z(2,iam),
     +     density_map, map_min_local, map_max_local, apo_border)
      
C     Get the global maximum and minimum density over entire map
      
#ifdef MPI
      call mpi_reduce(map_min_local, map_min, 1, MPI_REAL, MPI_MIN,
     +     0, mcw, mpi_err)
      call mpi_reduce(map_max_local, map_max, 1, MPI_REAL, MPI_MAX,
     +     0, mcw, mpi_err)
      if( mpi_err .ne. 0 ) call error_stop(progname,'density_clear',
     +     'Error in calling MPI_BCAST background')
#else
      map_min = map_min_local
      map_max = map_max_local
#endif   
            
      deallocate(rho)

#ifdef MPI
      mcw = mpi_comm_world
      call mpi_bcast(map_min, 1, mpi_real, 0, mcw, mpi_err)
      call mpi_bcast(map_max, 1, mpi_real, 0, mcw, mpi_err)
#endif

      if(iam == 0) then
         write(*,*) 'Map statistics'
         write(*,1122) map_min, map_max, map_max-map_min
 1122    format('max/min/max-min: ',f10.3,1x,f10.3,1x,f10.3)
      endif


C     Write out real-space map
      call output_density( density_map, den_min_x,
     +     den_max_x, max_z_slab, MAP_MIN, MAP_MAX )
      
C     Free memory for storing the density values.
      
      deallocate(density_map)
      deallocate(fsc_freq, fsc_value)

C ------- This the end of writing density values into MAP file --------
C ----- Before reporting the timing the memory allocating mesage, -----
C ------------------- synchronize all the nodes. ----------------------

#ifdef MPI
      call mpi_barrier(mcw, mpi_err)
#endif
      
      call system_clock(clock2)
      time = elapsed_time(clock1, clock2)
      
      if ( iam == 0 ) then
         write(*,*) 'Total run time = ', time
      endif

#ifdef MPI
      call MPI_FINALIZE ( mpi_err )
#endif

      end program p3dr


      function locate(xx,n,x)
      implicit none
      integer, intent(in) :: n
      real, intent(in) :: xx(n)
      real, intent(in) :: x
      integer :: locate

C     Given an array xx(1:n), and given a value x, returns a value j
C     such that x is between xx(j) and xx(j+1). xx must be monotonic,
C     either increasing or decreasing. j=0 or j=n is returned to
C     indicate that x is out of range. (Adapted from Numerical Recipes
C     in Fortran 90

      integer :: jl, jm, ju
      logical :: ascnd
      ascnd = (xx(n) >= xx(1))
      jl=0
      ju=n+1
      do
         if (ju-jl <= 1) exit
         jm=(ju+jl)/2
         if (ascnd .eqv. (x >= xx(jm))) then
            jl=jm
         else
            ju=jm
         endif
      enddo
      if (x == xx(1)) then
         locate=1
      else if (x == xx(n)) then
         locate=n-1
      else
         locate=jl
      endif
      end function locate
