      subroutine apply32(interpolants,hkl_max)

C     This routine is used to apply the 32 symmetry operations to the
C     matrix generated by interpolating the 2D DFTs of the particle
C     images into the 3D DFT. For the case of icosahedral symmetry
C     (532), each particle image and its fivefold symmetry-related
C     equivalents are first interpolated into the 3D DFT before calling
C     apply32. The advantage of this approach is that the total number
C     of image interpolations can be reduced by a factor of twelve.
C
C     Using our standard view of the icosahedron, where three mutually
C     perpendicular twofold axes are aligned with the xyz coordinate
C     axes, the twofold and threefold symmetry operations transform grid
C     points into other grid points. This allows the symmetry operations
C     to be applied to the 3D DFT without having to perform trilinear
C     interpolations.
C
C     C2x F(x,y,z) --> F(x,-y,-z)
C     C2y F(x,y,z) --> F(-x,y,-z)
C     C2z F(x,y,z) --> F(-x,-y,z)
C     C3  F(x,y,z) --> F(y,z,x) --> F(z,x,y)
C
C     The value of a function at (x,y,z) can be obtained as follows
C
C     F(x,y,z) = F(x,y,z) + F(x,-y,-z) + F(-x,y,-z) + F(-x,-y,z) +
C                F(y,z,x) + F(y,-z,-x) + F(-y,z,-x) + F(-y,-z,x) +
C                F(z,x,y) + F(z,-x,-y) + F(-z,x,-y) + F(-z,-x,y)
C
C     The Friedel condition, F(-k) = F(k)*, is used so that we only need
C     to store half of the 3D DFT. The formula given above becomes
C
C     F(x,y,z) = F(x,y,z) + F(x,-y,z)* + F(-x,y,z)* + F(-x,-y,z) +
C                F(y,z,x) + F(y,-z,x)* + F(-y,z,x)* + F(-y,-z,x) +
C                F(z,x,y) + F(z,-x,y)* + F(-z,x,y)* + F(-z,-x,y)
C
C     The interpolants array at each grid point contains three terms:
C     the CTF weighting, the real component of the DFT, and the
C     imaginary component of the DFT. Applying the Friedel condition
C     results in a sign change only for the contributions to the
C     imaginary components of the DFT.

      implicit none
#include "./map_type.inc"
      integer, intent(in) :: hkl_max
      FMAP_TYPE, intent(inout) ::
     +     interpolants(3,-hkl_max:hkl_max,-hkl_max:hkl_max,0:hkl_max)

      FMAP_TYPE, allocatable :: temp(:,:,:,:)
      integer alloc_stat, i, j, k

      allocate(temp(3, 0:hkl_max, 0:hkl_max, 0:hkl_max),
     +     stat = alloc_stat)

C     Apply 32 symmetry ops to get values of arrays in 1st octant (+,+,+)
      do k = 0, hkl_max
         do j = 0, hkl_max
            do i = 0, hkl_max

               temp(1,i,j,k) =  ! CTF weighting
     +              interpolants(1,i,j,k) +
     +              interpolants(1,i,-j,k) +
     +              interpolants(1,-i,j,k) +
     +              interpolants(1,-i,-j,k) +
     +              interpolants(1,j,k,i) +
     +              interpolants(1,j,-k,i) +
     +              interpolants(1,-j,k,i) +
     +              interpolants(1,-j,-k,i) +
     +              interpolants(1,k,i,j) +
     +              interpolants(1,k,-i,j) +
     +              interpolants(1,-k,i,j) +
     +              interpolants(1,-k,-i,j)

               temp(2,i,j,k) =  ! Real component of DFT
     +              interpolants(2,i,j,k) +
     +              interpolants(2,i,-j,k) +
     +              interpolants(2,-i,j,k) +
     +              interpolants(2,-i,-j,k) +
     +              interpolants(2,j,k,i) +
     +              interpolants(2,j,-k,i) +
     +              interpolants(2,-j,k,i) +
     +              interpolants(2,-j,-k,i) +
     +              interpolants(2,k,i,j) +
     +              interpolants(2,k,-i,j) +
     +              interpolants(2,-k,i,j) +
     +              interpolants(2,-k,-i,j)

               temp(3,i,j,k) = ! Imaginary component of DFT
     +              interpolants(3,i,j,k) +
     +              (-interpolants(3,i,-j,k)) +
     +              (-interpolants(3,-i,j,k)) +
     +              interpolants(3,-i,-j,k) +
     +              interpolants(3,j,k,i) +
     +              (-interpolants(3,j,-k,i)) +
     +              (-interpolants(3,-j,k,i)) +
     +              interpolants(3,-j,-k,i) +
     +              interpolants(3,k,i,j) +
     +              (-interpolants(3,k,-i,j)) +
     +              (-interpolants(3,-k,i,j)) +
     +              interpolants(3,-k,-i,j)

            enddo
         enddo
      enddo

C     Fill in the four octants of interpolants that lie above xy-plane
      do k = 0, hkl_max
         do j = 0, hkl_max
            do i = 0, hkl_max
               interpolants(1,i,j,k) = temp(1,i,j,k)
               interpolants(2,i,j,k) = temp(2,i,j,k)
               interpolants(3,i,j,k) = temp(3,i,j,k)

               interpolants(1,-i,-j,k) = temp(1,i,j,k)
               interpolants(2,-i,-j,k) = temp(2,i,j,k)
               interpolants(3,-i,-j,k) = temp(3,i,j,k)

               interpolants(1,-i,j,k) = temp(1,i,j,k)
               interpolants(2,-i,j,k) = temp(2,i,j,k)
               interpolants(3,-i,j,k) = -temp(3,i,j,k)

               interpolants(1,i,-j,k) = temp(1,i,j,k)
               interpolants(2,i,-j,k) = temp(2,i,j,k)
               interpolants(3,i,-j,k) = -temp(3,i,j,k)
            enddo
         enddo
      enddo

      deallocate(temp)

      return
      end subroutine apply32
