      subroutine exchange_2_slab(min_xyz, max_xyz, F_VRU,
     +     rrnum, F_VRU_TMP )
C                                                             EXCHANGE_2
C **********************************************************************
C *                                                               *
C * Subroutine: exchange_2_slab                                   *
C * Author: Robert E. Lynch (rel@cs.purdue.edu)                   *
C * Date:   1997                                                  *
C * Used in: exchange_2_m_slab -- The the nearest outer           *
C *            subroutine allocating subsidiary send and receive  *
C *            buffers for the call of MPI routines.              *
C * Function: Exchanging data for 1-D Fourier synthesis           *
C * Usage:    call exchange_2_slab( min_xyz, max_xyz, F_VRU,      *
C *                            num_hkl, F_VRU_TMP )               *
C * Input:                                                        *
C *   min_xyz (Integer) -- The minimum index in X, Y and Z axes   *
C *                        of the real space.                     *
C *   max_xyz (Integer) -- The maximum index in X, Y and Z axes   *
C *                        of the real space.                     *
C *   num_hkl (Integer) -- The number of coordinates in H, K or L *
C *                        axe of the reciprocal space.           *
C *   F_VRU (Real) -- The result of 2D Fourier synthesis.         *
C * Output:                                                       *
C *   F_VRU_TMP (Real) -- The exchanged result of 2D Fourier      *
C *                       synthesis for the current node to keep. *
C *                                                               *
C * Algorithm:                                                    *
C *   Exchange portions of the Fourier coefficients stored in     *
C *   F_VRU to other nodes.  After transfer, the data             *
C *   is stored in array F_VRU_TMP.                               * 
C *                                                               *
C * Modification history:                                         *
C *     - Hong Lin, Version 2.0, Dec.10, 1999.                    *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *

C *                                                               *
C *****************************************************************
C Arguments ------------------------------------------------------------

C  MIN_XYZ,MAX_XYZ  INTEGER*4; minimum and maximum of X, Y, Z

C  F_VRU(MIN_XYZ:MAX_XYZ, FRST_LST_RR_OUT(1,IAM):FRST_LST_RR_OUT(2,IAM),
C       MIN_XYZ:MAX_XYZ)  REAL*4; send buffer for ALL-TO-ALL
C               exchange:   F_VRU( X, K, Z )

C Output ---------------------------------------------------------------
C F_VRU_TMP(  ( - MIN_XYZ + MAX_XYZ + 1) * NUM_HKL
C            * ( - FRST_LST_U(1,IAM) + FRST_LST_U(2:IAM) + 1 ) )
C               REAL*4; recveive buffer for ALL-T0-ALL exchange:
C                        F_VRU_TMP( X, K, delta_Z )
C                  >>BUT<< F_VRU_TMP does not have a uniform multi-
C                  dimensional array; the array is filled as:
C                  F_VRU_TMP( MIN_XYZ:MAX_XYZ,
C                              FRST_LST_RR_OUT(1,1):FRST_LST_RR_OUT(2,1)
C                              FRST_LST_U(1,IAM):FRST_LST_U(2:IAM)
C                                   followed by
C                              MIN_XYZ:MAX_XYZ,
C                              FRST_LST_RR_OUT(1,2):FRST_LST_RR_OUT(2,2)
C                              FRST_LST_U(1,IAM):FRST_LST_U(2:IAM)
C                                   etc.
C                  and the value of
C                     FRST_LST_RR_OUT(1,N):FRST_LST_RR_OUT(2,N)
C                  is different for different values of node number N.
C                  All of them are used so that the array F_VRU_TMP
C                  contains values for all RR: 1: NUM_HKL
C              Consequently another routine must rearrange the values
C              stored in F_VRU_TMP
C ----------------------------------------------------------------------
C
      implicit none

#ifdef MPI
      include 'mpif.h'
#endif

      include '../include/allprogs.inc'
      
      integer, intent(in) :: min_xyz, max_xyz, rrnum
      real, intent(in) :: F_VRU( min_xyz:max_xyz,
     $     frst_lst_RR_out(1,iam):frst_lst_RR_out(2,iam),
     $     min_xyz:max_xyz )

      real , intent(out) :: F_VRU_TMP( min_xyz:max_xyz, 1:rrnum,
     $     frst_lst_U(1,iam):frst_lst_U(2,iam))

      integer*8 num_reals

#ifdef MPI
      integer node, send_offset, recv_offset, send_factor, recv_factor
#endif

      character(len=15) :: subname = 'EXCHANGE_2_SLAB'

C     External
C     BLAS routines; typically these are optimized code from vendor.
C     See modern books on numerical linear algebra or the syntax given
C     in E. Anderson, et al., LAPACK User's Guide, 2nd Edit., Siam
C     Publications, 1996.

      external scopyi8

C     Syntax:
C       CALL SCOPYI8( N, X, INCX, Y, INCY )
C         Copy vector X to vector Y
C             N    INTEGER*8: number of entries of X to be copied
C             X    REAL*4:    vector to be copied to Y; X must have
C                     length at least N*INCX
C             INCX INTEGER*4: stride of X
C             Y    REAL*4:    vector copied from X; Y must have length
C                     at least N*INCY
C             INCY INTEGER*4: stride of Y
C       The procedure is equivalent to
C            JY = 1
C            DO JX = 1, ( (N-1)*INCX + 1 ), INCX
C              Y(JY) = X(JX)
C              JY = JY + INCY
C            ENDDO
C ----------------------------------------------------------------------
C
      if( num_nodes == 1 ) then
C
C       There is only a single node in operation.  Move data from
C       F_VRU to F_VRU_TMP.

         num_reals =   int8( - min_xyz + max_xyz  + 1 )
     +        * ( - frst_lst_RR_out(1,iam) + frst_lst_RR_out(2,iam)+1)
     +        * ( -frst_lst_U(1,iam) + frst_lst_U(2,iam) + 1 )
         
         call scopyi8( num_reals, F_VRU, 1, F_VRU_TMP, 1 )
         
      else

C       There are more than one node in operation.  Distribute data
C       to other nodes by using ALL-TO-ALL exchange.
 
C       The array F_VRU for the case of two nodes is indicated below:

C                     Node 0                          Node 1

C                     cols_1                          cols_1
C              ---------------------            ---------------------
C              |        |          |            |        |          |
C       rows_1 |    A   |    B     |            |        |          |
C              |        |          |            |        |          |
C              |        |          |     rows_2 |   C    |    D     |
C              ---------------------            |        |          |
C                                               |        |          |
C              | cols_2 |  cols_3  |            |        |          |
C                                                --------------------

C                                               | cols_2 |  cols_3  |

C       where  cols_1 .eq. MIN_XYZ;MAX_XYZ (columns for Z when IA = 1)
C              rows_1 .eq. SEND_FACTOR when IAM .eq. 0 (see below)
C              rows_2 .eq. SEND_FACTOR when IAM .eq. 1 (see below)
C              cols_2 .eq. frst_lst_U_out(1,0):frst_lst_U_out(2,0)
C              cols_3 .eq. frst_lst_U_out(1,1):frst_lst_U_out(2,1)

C       After the ALL-TO-ALL exchange, F_VRU_TMP is indicated below

C                     Node 0                          Node 1

C                     cols_2                          cols_3
C                   ----------                    ------------
C                   |        |                    |          |
C            rows_1 |    A   |             rows_1 |    B     |
C                   |        |                    |          |
C                   |        |                    |          |
C                   ----------                    ------------
C                   |        |                    |          |
C                   |        |                    |          |
C                   |        |                    |          |
C            rows_2 |    C   |             rows_2 |    D     |
C                   |        |                    |          |
C                   |        |                    |          |
C                   |        |                    |          |
C                   ----------                    ------------
 
#ifdef MPI
         send_offset = 0
         recv_offset = 0

         send_factor = ( - min_xyz + max_xyz  + 1 )
     +        * (-frst_lst_RR_out(1,iam)+frst_lst_RR_out(2,iam)+1)
         recv_factor = ( - min_xyz + max_xyz  + 1 )
     +        * (-frst_lst_U(1,iam)+frst_lst_U(2,iam)+1)

         do node = 0, num_nodes-1
            scounts(node) = send_factor
     +           * (-frst_lst_U(1,node) + frst_lst_U(2,node) + 1)
            sdispls(node) = send_offset
            send_offset   = send_offset + scounts(node)
            
            rcounts(node) = recv_factor
     +           * (- frst_lst_RR_out(1,node)
     +           + frst_lst_RR_out(2,node) + 1)
            rdispls(node) = recv_offset
            recv_offset   = recv_offset + rcounts(node)
         enddo
         
C     Send data from F_VRU; receive data in F_VRU_TMP
         call mpi_alltoallv (F_VRU, scounts, sdispls,
     +        mpi_real, F_VRU_TMP, rcounts, rdispls, mpi_real,
     +        mpi_comm_world, mpi_err)

         if( mpi_err .ne. 0 ) then
            write(*, 9000) progname, subname
 9000       format(/' ** ERROR ** in program    ',a/
     +           '                subroutine ',a/
     +           ' MPI Error in MPI_ALLTOALLV'/
     +           '>>>>>>>>>> INSERT INFORMATIVE MESSAGE <<<<<<<<<<<')
            call error_stop(progname,subname,
     +           ' Read END-OF-FILE on User''s Control-Input-File')
         endif
#endif

      endif

      return
      end subroutine exchange_2_slab
