      subroutine fftsynth_1_m_slab(hkl_max, 
     +     first_iq, last_iq, first_rr_out, last_rr_out,
     +     f_hkl, nrffts_1_row, nrffts_1_col, nrffts_1_omega, 
     +     min_x, max_x, min_z, max_z, 
     +     min_xyz, max_xyz, f_vru, mem_total )

C                                                             FFTSYNTH_1
C **********************************************************************
C *                                                               *
C * Subroutine: fftsynth_1_slab                                   *
C * Author: Robert E. Lynch (rel@cs.purdue.edu)                   *
C * Date:   1997                                                  *
C * Used in: fftsynth_1_m_slab -- The control subroutine for 2D   *
C *          Fourier synthesis                                    *
C * Function: The driver of the 2-Dimensional Fourier synthesis,  *
C *           The actual 2-D Fourier synthesis is done by         *
C *           subroutine "fftsynth_1a".                           *
C * Usage:                                                        *
C *   call fftsynth_1_slab(hkl_max, first_iq,                     *
C *          last_iq, f_hkl, nrffts_1_row, nrffts_1_col           *
C *          rffts_1_even, rffts_1_odd, rffts_1_trns,             *
C *          nrffts_1_omega, rffts_1_omega, min_x, max_x,         *
C *          min_z, max_z, min_xyz, max_xyz, F_VU, F_VRU )        *
C * Input:                                                        *
C *   f_hkl (Complex) -- The 3D Fourier coefficients estimated    *
C *                      from 2D DFT                              *
C *   hkl_max (Integer) -- The maximum index of HKL space.        *
C *   first_iq (Integer) -- The starting index of R in this node. *
C *   last_iq (Integer) -- The ending index of R in this node.    *
C *   rffts_1_even (Real) -- Working sites of Fourier synthesis.  *
C *     The real part of the coefficients.                        *
C *   rffts_1_odd (Real) -- Working sites of Fourier synthesis.   *
C *     The imaginary part of the coefficients.                   *
C *   nrffts_1_row (Integer) -- The limits of row number of array *
C *                   "rffts_1_even", "rffts_1_odd" and the       *
C *                   column number of "rffts_1_trns".            *
C *   nrffts_1_col (Integer) -- The limits of row number of array *
C *                   "rffts_1_even", "rffts_1_odd" and the row   *
C *                   number of "rffts_1_trns".                   *
C *   rffts_1_trns (Real) -- Transposed array of "rffts_1".       *
C *   rffts_1_omega (Real) -- The working site for calling LAPACK *
C *                    routine.                                   *
C *   nrffts_1_omega (Integer) -- The limits of size of array     *
C *                     "rffts_1_omega.                           *
C *   min_x (Integer) -- The minimum index in X axe of the real   *
C *                      space.                                   *
C *   min_z (Integer) -- The minimum index in Z axe of the real   *
C *                      space.                                   *
C *   max_x (Integer) -- The maximum index in X axe of the real   *
C *                      space.                                   *
C *   max_z (Integer) -- The maximum index in Z axe of the real   *
C *                      space.                                   *
C *   min_xyz (Integer) -- The minimum index in X, Y and Z axes   *
C *   max_xyz (Integer) -- The maximum index in X, Y and Z axes   *
C *   F_VU (Real) -- The auxiliary array for 2D Fourier synthesis *
C * Output:                                                       *
C *   F_VRU (Real) -- The result of 2-D Fourier synthesis         *
C *                                                               *
C * Algorithm:                                                    *
C *   INPUT:                                                      *
C *     Fourier coefficients are stored in arrays                 *
C *               RFFTS_1_EVEN and RFFTS_1_ODD:                   *
C *                                                               *
C *       IA .eq. 1           IA .eq. 2              IA .eq. 3    *
C *     P=H, Q=L, R=K        P=K, Q=H, R=L        P=L, Q=K, R=H   *
C *                                                               *
C *       -------- L           -------- H           -------- K    *  
C *       |                    |                    |             *
C *       |                    |                    |             *
C *       |                    |                    |             *
C *       |                    |                    |             *
C *       H                    K                    L             *
C *                                                               *
C *   OUTPUT:                                                     *
C *     Fourier coefficients are stored in array                  *
C *              F_VRU( V, R, IA, U )                             *
C *                                                               *
C *       IA .eq. 1           IA .eq. 2              IA .eq. 3    *
C *     V=X, R=K, U=Z        V=X, R=L, U=Y        V=Y, R=H, U=Z   *
C *                                                               *
C *       |                    |                    |             *
C *       |                    |                    |             *
C *       -------- K           -------- L           -------- H    *  
C *       |                    |                    |             *
C *       |                    |                    |             *
C *       X                    X                    Y             *
C *                                                               * 
C *    Where for IA .eq. 1, MIN_X      .le. X .le. MAX_X          *
C *                         KMIN(IAM)  .le. K .le. KMAX(IAM)      *
C *                         MIN_Z      .le. Z .le. MAX_Z          *
C *                                                               *
C *       where KMIN(IAM) .eq. FIRST_RR_OUT             *
C *       where KMAX(IAM) .eq. LAST_RR_OUT             *
C *    Thus all Fourier coefficients with all variables H and L   *
C *    have been synthesised to a function with variables X and Z.*
C *    To complete the synthesis, the coeffients with variable K  *
C *    have to be synthesised to a function with variable Y.      *
C *    Similarly for cases IA .eq. 2 and IA .eq. 3.               *
C *                                                               *
C *  For a uniform 3-d cubic grid and a real valued function      *
C *  f(u,v,w) with u = 0:1:(N-1), v = 0:1:(N-1), w = 0:1:(N-1),   *
C *  the Fourier coefficients are given by F(h,k,l) where         *
C *                                                               *
C *    F(p,q,r) = N^(-3/2) [ sum sum sum f(u,v,w)                 *
C *                 exp(-2pi i (up+vq+wr)/N ) ]                   *
C *                                                               *
C *  where the summations are for u,v,w from 0 to N-1.            *
C *                                                               *
C *  With the abbreviations                                       *
C *     c(t) = cos( 2pi t / N ),    s(t) = sin( 2pi t / N )       *
C *  the sum on the right size is equal to                        *
C *                                                               *
C *     sum sum sum (    [ + f(u,v,w) c(up) c(vq) c(wr)           *
C *                        - f(u,v,w) s(up) s(vq) c(wr)           *
C *                        - f(u,v,w) s(up) c(vq) s(wr)           *
C *                        - f(u,v,w) c(up) s(vq) s(wr) ]         *
C *                  - i [ - f(u,v,w) s(up) s(vq) s(wr)           *
C *                        + f(u,v,w) s(up) c(vq) c(wr)           *
C *                        + f(u,v,w) c(up) s(vq) c(wr)           *
C *                        + f(u,v,w) c(up) c(vq) s(wr) ]         *
C *                 ) = RF - i IF                                 *
C *                                                               *
C *  With the abbreviation                                        *
C *      f(u,v,w) c(up) c(vq) c(wr) = ccc(u,v,w)                  *
C *  and other similar abbreviations, this sum is equal to        *
C *                                                               *
C *         [ ccc( p, q, r) - ssc( p, q, r) -                     *
C *           scs( p, q, r) - css( p, q, r) ]                     *
C *      -i [-sss( p, q, r) + scc( p, q, r) +                     *
C *           csc( p, q, r) + ccs( p, q, r) ]                     *
C *                                                               *
C *  If  f  were complex, instead of real, then there would be    *
C *  twice as many terms in this sum.                             *
C *                                                               *
C *  One needs only (N^3)/2 Fourier coefficients to determine the *
C *  (N^3) values of the real valued function  f  by Fourier      *
C *  synthesis. We have the values of  F(p,q,r)  for              *
C *     p = N/2 :1: N/2    q = N/2:1:N/2    r = 0 :1: N/2         *
C *  then the values of the real part, RF, of F/N^(-3/2) are:     *
C *                                                               * 
C *  RF( p, q, r)= ccc( p, q, r)-ssc( p, q, r)-                   *
C *                scs( p, q, r)-css( p, q, r)                    *
C *  RF(-p, q, r)= ccc( p, q, r)+ssc( p, q, r)+                   *
C *                scs( p, q, r)-css( p, q, r)                    *
C *  RF( p,-q, r)= ccc( p, q, r)+ssc( p, q, r)-                   *
C *                scs( p, q, r)+css( p, q, r)                    *
C *  RF(-p,-q, r)= ccc( p, q, r)-ssc( p, q, r)+                   *
C *                scs( p, q, r)+css( p, q, r)                    *
C *                                                               *
C *  Then we can compute (neglecting the unessential factor of    *
C *  1/4)                                                         *
C *                                                               *
C *  ccc(p,q,r) = RF( p, q, r)+RF(-p, q, r)+                      *
C *               RF( p,-q, r)+RF(-p,-q, r)                       *
C *  ssc(p,q,r) =-RF( p, q, r)+RF(-p, q, r)+                      *
C *               RF( p,-q, r)-RF(-p,-q, r)                       *
C *  scs(p,q,r) =-RF( p, q, r)+RF(-p, q, r)-                      *
C *               RF( p,-q, r)+RF(-p,-q, r)                       *
C *  css(p,q,r) =-RF( p, q, r)-RF(-p, q, r)+                      *
C *               RF( p,-q, r)+RF(-p,-q, r)                       *
C *                                                               *
C *  Similarly, the imaginary part, IF, of conj(F) yields         *
C *                                                               *
C *  IF( p, q, r)=-sss( p, q, r)+scc( p, q, r)+                   *
C *                csc( p, q, r)+ccs( p, q, r)                    *
C *  IF(-p, q, r)=+sss( p, q, r)-scc( p, q, r)+                   *
C *                csc( p, q, r)+ccs( p, q, r)                    *
C *  IF( p,-q, r)=+sss( p, q, r)+scc( p, q, r)-                   *
C *                csc( p, q, r)+ccs( p, q, r)                    *
C *  IF(-p,-q, r)=-sss( p, q, r)-scc( p, q, r)-                   *
C *                csc( p, q, r)+ccs( p, q, r)                    *
C *                                                               *
C *  We get (except for an unessential factor of 1/4)             *
C *                                                               *
C *  sss(p,q,r) =-IF( p, q, r)+IF(-p, q, r)+                      *
C *               IF( p,-q, r)-IF(-p,-q, r)                       *
C *  scc(p,q,r) =+IF( p, q, r)-IF(-p, q, r)+                      *
C *               IF( p,-q, r)-IF(-p,-q, r)                       *
C *  csc(p,q,r) =+IF( p, q, r)+IF(-p, q, r)-                      *
C *               IF( p,-q, r)-IF(-p,-q, r)                       *
C *  ccs(p,q,r) =+IF( p, q, r)+IF(-p, q, r)+                      *
C *               IF( p,-q, r)+IF(-p,-q, r)                       *
C *                                                               *
C *  Thus, the values of  ccc,ssc,...,ccs  can be computed for    *
C *                                                               *
C *       p,q = -N/2, ..., -1, 0, 1, ..., N/2                     *
C *                                                               *
C *  and those values of  r  between 0 and N/2 which are on the   *
C *  current node.                                                *
C *                                                               *
C *  When the synthesis has been carried out:  p --> u  and       *
C *  q --> v and the result is stored in F_UV for a specifice     *
C *  orientation of the axes.  After all three axes have been     *
C *  processed, the final output from this routine is stored in   *
C *  F_UVR, with contains the syntheses for a pair of axes for    *
C *  each of the three orientation axes. The values stored differ *
C *  from the actual syntheses by a constant factor.              *
C *                                                               *
C * Modification history:                                         *
C *     - Hong Lin, Version 1.0, Aug.18, 1998                     *
C *     - Hong Lin, Version 2.0, Oct. 5, 1999                     *
C *       A major change is made in the structure of data by      *
C *       Professor Robert Lynch.                                 *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *

C *                                                               *
C *****************************************************************
C    Input  ------------------------------------------------------------
C HKL_MAX  INTEGER*4; minimum and maximum values of H, K, L.
 
C FIRST_R,LAST_R   INTEGER*4; first and last values of K

C f_hkl( -HKL_MAX:HKL_MAX, -HKL_MAX:HKL_MAX, FIRST_R:LAST_R)
C                  COMPLEX*4; complex Fourier coefficients:
C                      f_hkl( H, L, K )

C NRFFTS_1_ROW,NRFFTS_1_COL  INTEGER*4; NRFFTS_1_ROW .eq. NRFFTS_1_COL;
C                  dimensions of 2D square arrays RFFTS_1_EVEN,
C                  RFFTS_1_ODD, RFFTS_1_TRNS which are work arrays to
C                  carry out the Fourier syntheses.

C RFFTS_1_EVEN(1:NRFFTS_1_ROW, 1:NRFFTS_1_COL),
C RFFTS_1_ODD (1:NRFFTS_1_ROW, 1:NRFFTS_1_COL),
C RFFTS_1_TRNS(1:NRFFTS_1_ROW, 1:NRFFTS_1_COL)  REAL*4; square array
C                  used to carry out the Fourier syntheses.
 
C NRFFTS_1_OMEGA   INTERGER*4; dimension of array for parameters for
C                  the Fourier syntheses; NRFFTS_1_OMEGA must be
C                  .ge. NRFFTS_1_ROW + 15.
 
C RFFTS_1_OMEGA    READ*4; array of parameters for Fourier syntheses;
C                  values set in  FFTSYNTH_1A.
 
C MIN_X,MAX_X, MIN_Z,MAX_Z  INTEGER*4; minimum and maximum
C                  values of X, Z, in final output of program.

C MIN_XYZ,MAX_XYZ  INTEGER*4; minimum and maximum of {X,Y,Z}.

C   Output;   ----------------------------------------------------------
C F_VRU(MIN_XYZ:MAX_XYZ, FIRST_RR_OUT:LAST_RR_OUT,
C       MIN_XYZ:MAX_XYZ)  REAL*4; output values: synthesis
C               of two of the three variables (the third is transformed
C               in FFTSYNTH_2):
C                     F_VRU( X, K, 1, Z )
C ----------------------------------------------------------------------

      implicit none
      include '../include/allprogs.inc'

C     Subroutine arguments
      integer hkl_max, first_iq, last_iq,
     +     first_rr_out, last_rr_out,
     +     nrffts_1_row, nrffts_1_col, nrffts_1_omega,
     +     min_x, max_x, min_z, max_z, 
     +     min_xyz, max_xyz
      
      complex f_hkl(-hkl_max:hkl_max, -hkl_max:hkl_max,
     +     first_iq:last_iq)
      
      real F_VRU( min_xyz:max_xyz, first_rr_out:last_rr_out,
     +     min_xyz:max_xyz )
      double precision mem_total
      
C     Local variables
      real, allocatable :: rffts_1_even(:,:),
     +     rffts_1_odd (:,:),
     +     rffts_1_trns(:,:),
     +     rffts_1_omega(:),
     +     F_VU(:,:)

      integer first_u,last_u, first_v,last_v, alloc_stat,
     +        P,P2,P2_p1, Q,Q2,Q2_p1, R,R_store_even,R_store_odd,
     +        x, z

      logical :: R_not_zero, firsttime = .true.
 
      real    RF_pp, RF_mp, RF_pm, RF_mm, RF_0p, RF_0m, RF_p0, RF_m0,
     +        IF_pp, IF_mp, IF_pm, IF_mm, IF_0p, IF_0m, IF_p0, IF_m0

      character(len=17) :: subname = 'fftsynth_1_m_slab'

      allocate(rffts_1_even(nrffts_1_row, nrffts_1_col),
     +     rffts_1_odd(nrffts_1_row, nrffts_1_col),
     +     rffts_1_trns(nrffts_1_row, nrffts_1_col),
     +     rffts_1_omega(nrffts_1_omega),
     +     F_VU(min_xyz:max_xyz, min_xyz:max_xyz),
     +     stat = alloc_stat)

      if (alloc_stat .gt. 0) then
         call error_stop(progname,
     +        'fftsynth_1_m', 
     +        'Error in allocating memory for F_VU!')
      endif

      mem_total = 3*4*nrffts_1_row*nrffts_1_col + 4*nrffts_1_omega +
     +     4*(max_xyz-min_xyz+1)*(max_xyz-min_xyz+1)

      if( firsttime ) then
 
C       INPUT PASSES ALL TESTS
C       ****************************************************************

C       Initialize the parameters used in the FFTs.  The parameters are
C       initialized only the first time FFTSYNTH_1 is called or when
C       the number of rows or columns of RFFTS changes.
 
C                                                            CALL VRFFTI
C                                                            ***********
C       Initialize the FFT parameters used to transform rows of RFFTS
C       each row has length NRFFTS_COL
         call vrffti( nrffts_1_col, rffts_1_omega )

      endif

      firsttime = .false.
 
      first_u = min_x
      last_u  = max_x
      
      first_v = min_z
      last_v  = max_z
 
C     SEE NOTE #1 (below) FOR FORMULAE PERTINENT TO THE CODE BELOW
Cji because fft analysis translate (r1, r2, r3, ..., rn)
Cji into (R1,R2,I3,R4,I5,..., Rn)
Cji while fft synth do the reverse.
 
      do R = first_iq, last_iq
 
C >>>>> TEST FOR FIRST_R EVEN OR ODD ! ! ! ! ! ! <<<<<<<<<<<<<<<<<<<<<<<
C       There has to be a change in the following when more than
C       one node is being used because  FIRST_R can be odd
C >>>>>>>>>>>>>>>>>>>>>>>>>>  <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
 
         R_store_even = 2 * R
         R_store_odd  = R_store_even + 1
         if( R .eq. 0 ) then
C         Only an even part is computed (the even part is a coefficient
C         of a cosine term, an odd part is a coefficient of a sine term)
            R_store_even = 1
            R_not_zero   = .false.
         elseif ( R .eq. hkl_max ) then
Cji       last section should only have even part. same as the first section.
            R_not_zero   = .false.
         else
C         Both an even part and an odd part are computed.  Even and odd
C         coefficients are computed for each R except for R .eq. 0.
            R_not_zero   = .true.
         endif
 
C >>>>>>>>>> The loop-structure can be simplified.  The `DO IA ... '
C            probably should be outside the `DO R ...'
C            We might eliminate 4 of the six arrays
C              RFFTS_1_EVEN( :, :, 1:3), RFFTS_1_ODD( :, :, 1:3)
C >>>>>>>>>>>>>>>>>>>>>>>>>>  <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

C         P .eq. H, V .eq. L, R .eq. K 

C         Case #1:  P and Q positive
         do Q = 1, hkl_max-1
            Q2    = 2*Q
            Q2_p1 = Q2 + 1
            do P = 1, hkl_max-1
               P2 = 2*P
               P2_p1 = P2 + 1
C             RF denotes the real part of the transform, IF denotes the
C             complex part.  The subscipts, pp, mp, ... refer to
C             value of P,Q repectively.   `p' denote positive, `m'
C             denotes negative; e.g., RF_pp = realpart( F( +P, +Q, R) ),
C                 RF_mp  = realpart( F( -P, +Q, R) ), etc.
               RF_pp = real(  f_hkl( P, Q, R ) )
               RF_mp = real(  f_hkl(-P, Q, R ) )
               RF_pm = real(  f_hkl( P,-Q, R ) )
               RF_mm = real(  f_hkl(-P,-Q, R ) )
               IF_pp = aimag( f_hkl( P, Q, R ) )
               IF_mp = aimag( f_hkl(-P, Q, R ) )
               IF_pm = aimag( f_hkl( P,-Q, R ) )
               IF_mm = aimag( f_hkl(-P,-Q, R ) )

C               Set values in an even (cosine) R-plane 
C                   COS(P)COS(Q)COS(R) 
               rffts_1_even(P2,    Q2) =(  RF_pp+RF_mp+RF_pm+RF_mm)/4
C                   SIN(P)SIN(Q)COS(R) 
               rffts_1_even(P2_p1, Q2_p1) =( -RF_pp+RF_mp+RF_pm-RF_mm)/4
C                   SIN(P)COS(Q)COS(R) 
               rffts_1_even(P2_p1, Q2) =(  IF_pp-IF_mp+IF_pm-IF_mm)/4
C                   COS(P)SIN(Q)COS(R) 
               rffts_1_even(P2,    Q2_p1) =(  IF_pp+IF_mp-IF_pm-IF_mm)/4
               if( R_not_zero ) then
C               Set values in an odd (sine) R-plane 
C                   SIN(P)COS(Q)SIN(R) 
                  rffts_1_odd( P2_p1, Q2) =( -RF_pp+RF_mp-RF_pm+RF_mm)/4
C                   COS(P)SIN(Q)SIN(R) 
                  rffts_1_odd( P2, Q2_p1) =( -RF_pp-RF_mp+RF_pm+RF_mm)/4
C                   SIN(P)SIN(Q)SIN(R) 
                  rffts_1_odd( P2_p1,Q2_p1)=(-IF_pp+IF_mp+IF_pm-IF_mm)/4
C                   COS(P)COS(Q)SIN(R) 
                  rffts_1_odd( P2, Q2) =(  IF_pp+IF_mp+IF_pm+IF_mm)/4
               endif

            enddo
         enddo

C     Case #2:  P zero and hkl_max, Q positive
         do Q = 1, hkl_max-1
            Q2    = 2*Q
            Q2_p1 = 2*Q + 1
Cji   P=zero 
            RF_0p = real(  f_hkl( 0, Q, R ) )
            RF_0m = real(  f_hkl( 0,-Q, R ) )
            IF_0p = aimag( f_hkl( 0, Q, R ) )
            IF_0m = aimag( f_hkl( 0,-Q, R ) )
C     Set values in an even (cosine) Q-plane 
C     COS(0)COS(Q)COS(R) 
            rffts_1_even( 1, Q2   ) = 1/2. * (   RF_0p + RF_0m )
C     COS(0)SIN(Q)COS(R) 
            rffts_1_even( 1, Q2_p1) = 1/2. * (   IF_0p - IF_0m )
            if( R_not_zero ) then
C     Set values in an odd (sine) Q-plane 
C     COS(0)SIN(Q)SIN(R) 
               rffts_1_odd(  1, Q2_p1) = 1/2. * ( - RF_0p + RF_0m )
C     COS(0)COS(Q)SIN(R) 
               rffts_1_odd(  1, Q2   ) = 1/2. * (   IF_0p + IF_0m )
            endif
            
Cji   P=hkl_max  
            P2 = 2*hkl_max
            RF_0p = real(  f_hkl( hkl_max, Q, R ) )
            RF_0m = real(  f_hkl( hkl_max,-Q, R ) )
            IF_0p = aimag( f_hkl( hkl_max, Q, R ) )
            IF_0m = aimag( f_hkl( hkl_max,-Q, R ) )
            rffts_1_even( P2, Q2   ) = 1/2. * (  RF_0p + RF_0m )
            rffts_1_even( P2, Q2_p1) = 1/2. * (  IF_0p - IF_0m )
            if( R_not_zero ) then
               rffts_1_odd(  P2, Q2_p1) = 1/2. * ( -RF_0p + RF_0m )
               rffts_1_odd(  P2, Q2   ) = 1/2. * (  IF_0p + IF_0m )
            endif
         enddo
 
C         Case #3:  P positive, Q zero and hkl_max
         do P = 1, hkl_max-1
Cji Q=zero
            P2    = 2*P
            P2_p1 = 2*P + 1
            RF_p0 = real(  f_hkl( P, 0, R ) )
            RF_m0 = real(  f_hkl(-P, 0, R ) )
            IF_p0 = aimag( f_hkl( P, 0, R ) )
            IF_m0 = aimag( f_hkl(-P, 0, R ) )
C     Set values in an even (cosine) h-plane 
C     COS(P)COS(0)COS(R) 
            rffts_1_even( P2,    1) = 1/2. * (   RF_p0 + RF_m0 )
C     SIN(P)COS(0)COS(R) 
            rffts_1_even( P2_p1, 1) = 1/2. * (   IF_p0 - IF_m0 )
            if( R_not_zero ) then
C     Set values in an odd (sine) h-plane 
C     SIN(P)COS(0)SIN(R) 
               rffts_1_odd(  P2_p1, 1) = 1/2. * ( - RF_p0 + RF_m0 )
C     COS(P)COS(0)SIN(R) 
               rffts_1_odd(  P2,    1) = 1/2. * (   IF_p0 + IF_m0 )
            endif
Cji Q=hkl_max
            Q2    = 2 * hkl_max
            RF_p0 = real(  f_hkl( P, hkl_max, R ) )
            RF_m0 = real(  f_hkl(-P, hkl_max, R ) )
            IF_p0 = aimag( f_hkl( P, hkl_max, R ) )
            IF_m0 = aimag( f_hkl(-P, hkl_max, R ) )
            rffts_1_even( P2,    Q2) = 1/2. * (  RF_p0 + RF_m0 )
            rffts_1_even( P2_p1, Q2) = 1/2. * (  IF_p0 - IF_m0 )
            if( R_not_zero ) then
               rffts_1_odd(  P2_p1, Q2) = 1/2. * ( -RF_p0 + RF_m0 )
               rffts_1_odd(  P2,    Q2) = 1/2. * (  IF_p0 + IF_m0 )
            endif
         enddo

C     Last Case, #4:  P zero and hkl_max, Q zero and hkl_max
C     Set values in an even (cosine) R-plane 
C     COS(0)COS(0)COS(R) 
         rffts_1_even( 1, 1) = 4/4. * real(  f_hkl( 0, 0, R ) )
         rffts_1_even( 1, 2*hkl_max) =
     +        4/4. * real(  f_hkl( 0, hkl_max, R ) )
         rffts_1_even( 2*hkl_max, 1) = 
     +        4/4. * real(  f_hkl( hkl_max, 0, R ) )
         rffts_1_even( 2*hkl_max, 2*hkl_max) = 
     +        4/4. * real(  f_hkl( hkl_max, hkl_max, R ) )
         if( R_not_zero ) then
C     Set values in an odd (sine) R-plane 
C     COS(0)COS(0)SIN(R) 
            rffts_1_odd(  1, 1) = 4/4. * aimag( f_hkl( 0, 0, R ) )
            rffts_1_odd( 1, 2*hkl_max) =
     +           4/4. * aimag(  f_hkl( 0, hkl_max, R ) )
            rffts_1_odd( 2*hkl_max, 1) = 
     +           4/4. * aimag(  f_hkl( hkl_max, 0, R ) )
            rffts_1_odd( 2*hkl_max, 2*hkl_max) = 
     +           4/4. * aimag(  f_hkl( hkl_max, hkl_max, R ) )
         endif
         
C     Fourier coefficients are stored in arrays
C     RFFTS_1_EVEN and RFFTS_1_ODD:

C     P=H, Q=L, R=K   
 
C            -------- L        
C            |         
C            |      
C            |    
C            | 
C            H

C         U .eq. X, V .eq. Z
 
         call fftsynth_1a( nrffts_1_row, nrffts_1_col,
     +        rffts_1_even, rffts_1_trns, nrffts_1_omega,
     +        rffts_1_omega, first_u,last_u, first_v,last_v,
     +        min_xyz,max_xyz, F_VU )
         
         do z = min_z, max_z
            do x = min_x, max_x
               F_VRU( x, R_store_even, z ) = F_VU( z, x )
            enddo
         enddo
         
         if( R_not_zero ) then
C           Have an R-odd plane to store
 
            if( R_store_odd .gt. last_rr_out ) then
               write(*, 9040) progname, subname,
     +              R, first_iq, last_iq,
     +              R_store_even, R_store_odd, R_store_odd,
     +              first_rr_out, last_rr_out
 9040          format(/' ** ERROR ** in program    ',a/
     +              '                subroutine ',a/
     +              ' Index out-of-range'/
     +              ' R, first_iq, last_iq        =', 3i5/
     +              ' R_store_even, R_store_odd =', 2i5//
     +              ' R_store_odd =', i5,' must be between',
     +              ' first_rr_out =',i5/ ' and ',
     +              ' last_rr_out =',i5 )
               call error_stop(progname,subname,
     +              ' Error in input to subroutine')
            endif
 
            call fftsynth_1a( nrffts_1_row, nrffts_1_col,
     +           rffts_1_odd, rffts_1_trns, nrffts_1_omega,
     +           rffts_1_omega, first_u,last_u, first_v,last_v,
     +           min_xyz,max_xyz, F_VU )
            
            do z = min_z, max_z
               do x = min_x, max_x
                  F_VRU( x, R_store_odd, z ) = F_VU( z, x )
               enddo
            enddo
            
         endif
      enddo
      
C     OUTPUT:
C       Fourier coefficients are stored in array
C               F_VRU( V, R, U )

C          V=X, R=K, U=Z     
 
C            |             
C            |          
C            -------- K      
C            |        
C            |      
C            X   
 
C      Where MIN_X      .le. X .le. MAX_X       
C            KMIN(IAM)  .le. K .le. KMAX(IAM)     
C            MIN_Z      .le. Z .le. MAX_Z       

C         where KMIN(IAM) .eq. FIRST_RR_OUT
C         where KMAX(IAM) .eq. LAST_RR_OUT
C      Thus all Fourier coefficients with all variables H and L have
C      been synthesised to a function with variables X and Z.  To
C      complete the synthesis, the coeffients with variable K have
C      to be synthesised to a function with variable Y.
       
      deallocate(rffts_1_even, rffts_1_odd, rffts_1_trns,
     +     rffts_1_omega, F_VU)
      
      return
      end subroutine fftsynth_1_m_slab
