      subroutine fftsynth_1a(nrffts_row, nrffts_col, 
     +     rffts, rffts_trns, nrffts_omega, 
     +     rffts_omega, first_u, last_u, 
     +     first_v, last_v, min_xyz, max_xyz, 
     +     f_out )
C                                                            FFTSYNTH_1A
C **********************************************************************
C *                                                               *
C * Subroutine: fftsynth_1a                                       *
C * Author: Robert E. Lynch (rel@cs..purdue.edu)                  *
C * Date:   1997                                                  *
C * Used in: fftsynth_1 -- The driver subroutine for 2-D Fourier  *
C *                        synthesis                              *
C * Function: 2-Dimensional Fourier synthesis, driven by          *
C *           subroutine "fftsynth_1".                            *
C * Usage:                                                        *
C *   call fftsynth_1a( nrffts_row, nrffts_col, rffts,            *
C *          rffts_trns, nrffts_omega, rffts_omega, first_u,      *
C *          last_u, first_v, last_v, min_xyz, max_xyz, f_out )   *
C * Input:                                                        *
C *   first_u (Integer) -- The starting index of U in this node.  *
C *   last_u (Integer) -- The ending index of U in this node.     *
C *   first_v (Integer) -- The starting index of V in this node.  *
C *   last_v (Integer) -- The ending index of V in this node.     *
C *   rffts (Real) -- Working sites of Fourier synthesis.         *
C *   nrffts_row (Integer) -- The limits of row number of array   *
C *     "rffts" and the column number of "rffts_trns".            *
C *   nrffts_col (Integer) -- The limits of row number of array   *
C *     "rffts" and the row number of "rffts_trns".               *
C *   rffts_trns (Real) -- Transposed array of "rffts".           *
C *   rffts_omega (Real) -- The working site for calling LAPACK   *
C *                         routine.                              *
C *   nrffts_omega (Integer) -- The limits of size of array       *
C *                             "rffts_omega.                     *
C *   min_xyz (Integer) -- The minimum index in X, Y and Z axes   *
C *   max_xyz (Integer) -- The maximum index in X, Y and Z axes   *
C * Output:                                                       *
C *   F_OUT (Real) -- The result of 2-D Fourier synthesis         *
C *                                                               *
C * Algorithm:                                                    *
C *   Carry out the Fourier synthesis with respect to two         *
C *   directions for a particular plane of Fourier coefficients.  *
C *   In practice, it will do the following (one in each call):   *
C *                                                               *
C *   1. Carry out the Fourier synthesis with respect to L --> Z  *
C *      and H --> X for an `even' K-plane, keep only those       *
C *      values that have indices                                 *
C *          MIN_X_OUT:MAX_X_OUT     MIN_Z_OUT:MAX_Z_OUT          *
C *                                                               *
C *   2. Carry out the Fourier synthesis with respect to L --> Z  *
C *      and H --> X for an `odd' K-plane, keep only those values *
C *      that have indices                                        *
C *          MIN_X_OUT:MAX_X_OUT     MIN_Z_OUT:MAX_Z_OUT          *
C *                                                               *
C *   3. Carry out the Fourier synthesis with respect to L --> Z  *
C *      and K --> Y for an `even' H-plane, keep only those       *
C *      values that have indices                                 *
C *          MIN_Y_OUT:MAX_Y_OUT     MIN_Z_OUT:MAX_Z_OUT          *
C *                                                               *
C *   4. Carry out the Fourier synthesis with respect to L --> Z  *
C *      and K --> Y for an `odd' H-plane, keep only those values *
C *      that have indices                                        *
C *          MIN_Y_OUT:MAX_Y_OUT     MIN_Z_OUT:MAX_Z_OUT          *
C *                                                               *
C * Modification history:                                         *
C *     - Hong Lin, Version 1.0, Aug.18, 1998                     *
C *     - Hong Lin, Version 2.0, Oct. 5, 1999                     *
C *       Changes were made by Professor Robert Lynch in the      *
C *       structure of the data.                                  *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *

C *                                                               *
C *****************************************************************
C Carry out a 2D FFT synthesis (back transform) of the Fourier
C coefficients in RFFTS.  Store the results in array F_OUT

C Input:  --------------------------------------------------------------

C NRFFTS_ROW, NRFFTS_COL  INTEGER*4;  NRFFTS_ROW .eq. NRFFTS_COL, 
C             dimensions of square arrays RFFTS and RFFTS_TRNS.

C RFFTS(1:NRFFTS_ROW,1:NRFFTS_COL)  REAL*4; array of Fouier coefficients
C             to be synthesized.
 
C RFFTS_TRNS(1:NRFFTS_ROW,1:NRFFTS_COL)  REAL*4; work array.

C NRFFTS_OMEGA    INTEGER*4; dimension of array RFFTS_OMEGA; this must
C             be .GE. NRFFTS_ROW + 15.
 
C RFFTS_OMEGA     REAL*4; parameters used in the FFTs.  Values must
C             be stored in RFFTS_OMEGA by the routine which calls
C             FFTSYNTH_1A by calling VRFFTI:
C                  CALL VRFFTI( NRFFTS_COL, RFFTS_OMEGA )
C             Because the arrays RFFTS and RFFTS_TRNS are square with
C             the same dimensions, the same parameters can be used in
C             the two calls of VRFFTB below.
 
C FIRST_U, LAST_U  INTEGER*4; first and last U indices for output array.

C FIRST_V, LAST_V  INTEGER*4; first and last V indices for output array.

C Output: --------------------------------------------------------------
 
C F_OUT( FIRST_V:LAST_V, FIRST_U:LAST_U) REAL*4; array containing the
C             synthesized values after having transformed the values
C             in the array RFFTS.
C ----------------------------------------------------------------------
      implicit none
      include '../include/allprogs.inc'

C     Subroutine arguments
      integer nrffts_row, nrffts_col, nrffts_omega, first_u, last_u,
     +     first_v, last_v, min_xyz, max_xyz

      real    rffts( nrffts_row, nrffts_col ),
     +     rffts_trns( nrffts_col, nrffts_row ),
     +     rffts_omega( nrffts_omega),
     +     f_out( min_xyz:max_xyz, min_xyz:max_xyz )

C     Local variables
      integer last_row, u_col_put, u_col_get, v_col_get, v_row_put
      logical :: error, firsttime = .true.
      character(len=11) :: subname = 'FFTSYNTH_1A'

C     Local debug variables
      integer iirr, iicc

C External
C     BLAS routines; typically these are optimized code from vendor.
C     See modern books on numerical linear algbra or the syntax given
C     in E. Anderson, et al., LAPACK User's Guide, 2nd Edit., Siam
C     Publications, 1996.

      external scopy
 
C     Syntax:
C       CALL SCOPY( N, X, INCX, Y, INCY )
C         Copy vector X to vector Y
C             N    is the number of entries of the vector X to be copied
C             X    is the vector to be copied to Y
C             INCX is the stride of X
C             Y    is the vector copied from X
C             INCY is the stride of Y
C       The procedure is equivalent to
C            JY = 1
C            DO JX = 1, ( (N-1)*INCX + 1 ), INCX
C              Y(JY) = X(JX)
C              JY = JY + INCY
C            ENDDO

C     REAL DFFT synthesis routine: Sweet, R.A., et al., VFFTPK, A
C     Vectorized Pacakage of Fortran Subprograms for the Fast Fourier
C     Transform of Multiple Real Sequences, 1990; obtainable by e-mail,
C     ftp, or www from netlib@ornl.gov
C
      external vrfftb
C ----------------------------------------------------------------------
 
      if( firsttime ) then
C
C       ****************************************************************
C       CHECK INPUT: This is done only the first time FFTSYNTH_1A is
C       called or when the number of rows or columns of RFFTS,... change

         error = .false.

         if( ( nrffts_row .ne. nrffts_col       ) ) then
            error = .true.
            print 9000, subname,  nrffts_row, nrffts_col
 9000       format(//' Error in SUBPROUTINE ',a/
     +           ' Arrays RFFTS and RFFTS_TRNS must be square and the',
     +           ' same size'/
     +           5x,' NRFFTS_ROW = ', i10, ', NRFFTS_COL = ', i10)
         endif

         if( nrffts_omega .lt. nrffts_col + 15 ) then
            error = .true.
            print 9010,  subname, nrffts_omega, nrffts_col+15
 9010       format(//' Error in SUBPROUTINE ',a/
     +           ' Array RFFTS_OMEGA must be at least as large as',
     +           ' NRFFTS_COL + 15:'/ 5x, ' NRFFTS_OMEGA = ', i10,
     +           ', NRFFTS_COL + 15 = ', i10 ) 
         endif
 
         if( ( (-first_u+last_u+1) .gt. nrffts_row ) .or.
     +        ( (-first_v+last_v+1) .gt. nrffts_row )      ) then
            error = .true.
            print 9020, subname, first_u,last_u, (-first_u+last_u+1),
     +           nrffts_row, first_v,last_v, (-first_v+last_v+1), 
     +           nrffts_row
 9020       format(//' Error in SUBPROUTINE ',a/
     +           ' FIRST_U =',i5, ', LAST_U =',i5/
     +           '   (-FIRST_U+LAST_U+1) =',i5,
     +           ' must be .LE. NRFFTS_ROW =', i5/
     +           ' FIRST_V =',i5, ', LAST_V =',i5/
     +           '   (-FIRST_V+LAST_V+1) =',i5,
     +           ' must be .LE. NRFFTS_ROW =', i5)
         endif
         
         if( ( first_u .lt. min_xyz ) .or. 
     +        ( max_xyz .lt. last_u ) .or.
     +        ( first_v .lt. min_xyz) .or. 
     +        ( max_xyz .lt. last_v ) ) then
            error = .true.
            print 9030, subname, first_u,first_v, min_xyz,
     +           last_u,last_v, max_xyz
 9030       format(//' Error in SUBPROUTINE ',a/
     +           ' FIRST_U =',i5,' and FIRST_V =', i5,
     +           ' must both be .GE. MIN_XYZ =',i5/
     +           ' LAST_U =',i5,' and LAST_V =', i5,
     +           ' must both be .LE. MAX_XYZ =',i5)
         endif
         
         if( error ) then
            write(*, 9040) progname, subname
 9040       format(/' ** ERROR ** in program    ',a/
     +           '                subroutine ',a/
     +           ' Error in input to subroutine')
            call error_stop(progname,subname,
     +           ' Error in input to subroutine')
         endif
 
C       INPUT PASSES ALL TESTS
C       ****************************************************************
         firsttime = .false.

      endif
 
C ----------------------------------------------------------------------
C STEP #1:  Transform the Fourier coefficients in the rows of the square
C    array RFFTS and store the results back in RFFTS.
 
C                RFFTS                      RFFTS
C             ------------- Q            ------------- V
C             |         |                |         |
C             |         |                |         |
C             |         |       -->      |         |
C             |         |                |         |
C             |         |                |         |
C             -----------                -----------
C             |                          |
C              P                          P 
C                                                            CALL VRFFTB
C                                                            ***********
      call vrfftb(nrffts_row, nrffts_col, rffts, rffts_trns,
     +     nrffts_row, rffts_omega ) 
C
C     Each row of RFFTS now contains values of a periodic
C     function, period NRFFTS_COL.
 
C     The array column index Q and the real space coordinate V are
C     related according to

C       col Q = 1, 2, 3, ...          2*HKL_MAX, 2*HKL_MAX+1 = NRFFS_COL
C     index V = 0, 1, 2, ...              -2,        -1
C ----------------------------------------------------------------------
C STEP #2:  Store the transpose of RFFTS in the square array RFFTS_TRNS,
C     but discard those rows which are not needed for the output.
C     Keep those rows with the V coordinates
C                    FIRST_V .le. V .le. LAST_V
C     Order the rows according the V coordinates in increasing order.
C     The relationship between row numbers and coordinates V is:
C       row =     1,        2,     ..., ROW_LAST-1,  ROW_LAST, 
C        V  =  FIRST_V, FIRST_V+1, ...,  LAST_V-1,   LAST_V
 
C                   RFFTS                       RFFTS_TRNS
C              ------------- V        FIRST_V   ------------- P
C              |         |                      |         |
C              |         |            V = 0     |---------|
C              |         |       -->            |         |
C              |         |            LAST_V    |---------|
C              |         |                      |  --0--  |
C              -----------                      -----------
C              |                                |
C              P                                V 
      
C     Use the BLAS routine SCOPY to transpose those columns of RFFTS
C     which correspond to values of V which are included in the output.
C     Rearrange the order of the V's so that instead of having 
C        V = 0, 1, ..., V_MAX, -V_MAX, -V_MAX+1, ..., -1 
C    we have
C        V = FIRST_V, ..., LAST_V
      
C     SCOPY syntax:
C       SCOPY(vector_length,input_vector,stride,output_vector,stride)
 
      v_row_put = 0
      if( first_v .lt. 0 ) then
         do v_col_get = nrffts_col + first_v + 1, nrffts_col
            v_row_put = v_row_put + 1
            call scopy( nrffts_row, rffts(1,v_col_get),       1,
     +           rffts_trns(v_row_put,1), nrffts_row)
         enddo
      endif
 
      if( last_v .gt. -1 ) then
         do v_col_get = 1, last_v + 1
            v_row_put = v_row_put + 1
            call scopy( nrffts_row, rffts(1,v_col_get),       1,
     +           rffts_trns(v_row_put,1), nrffts_row)
         enddo
      endif

      last_row = v_row_put

C ----------------------------------------------------------------------
C STEP #3:  Carry out a real FFT synthesis of each nonzero row of
C     RFFTS_TRNS.  The Fourier (column) index P becomes the real space
C     (column) index U.  The row indices 1:LAST_ROW corresponds to
C     V = FIRST_V:LAST_V
 
C     Then the rows of RFFTS_TRNS are transformed.  The relation between
C     the column index P and the coordinate U is:

C         col P  = 1, 2, 3, ...      2*HKL_MAX, 2*HKL_MAX+1 = NRFFS_COL
C       index U  = 0, 1, 2, ...          -2,        -1
      
C                     RFFTS_TRNS                 RFFTS_TRNS
C          FIRST_V   ------------- P            ------------- U
C                    |         |                |         |
C            V = 0   |---------|                |---------|
C                    |         |       -->      |         |
C          LAST_V    |---------|                |---------|
C                    |  --0--  |                |  --0--  |
C          NRFFS_ROW -----------                -----------
C                    |         | NRFFS_COL      |
C                    V                          V

C                                                            CALL VRFFTB
C                                                            ***********
      call vrfftb(last_row, nrffts_row, rffts_trns, rffts,
     +     nrffts_col, rffts_omega)
C
C ----------------------------------------------------------------------
C STEP #4:  Fill the output array with the transformed rows with indices
C                FIRST_U .le. U .le. LAST_U
      
C                        RFFTS_TRNS                     F_OUT     
C     1       FIRST_V   ------------- U       FIRST_V ---------
C                       |         |                   |   |   |
C             V = 0     |---------|                   |---x---|--- U
C                       |         |      -->          |   |   |
C LAST_ROW    LAST_V    |---------|           LAST_V  ----|----
C                       |  --0--  |                       |    
C             NRFFS_ROW -----------                       V         
C                       |         |                FIRST_U   LAST_U
C                       |         | 
C                       V         NRFFTS_COL
 
      do iicc = min_xyz, max_xyz
         do iirr = min_xyz,max_xyz
            f_out(iirr,iicc) = 0.
         enddo
      enddo

      u_col_put = first_u - 1
      if( first_u .lt. 0 ) then
         do u_col_get = nrffts_col + first_u + 1, nrffts_col
            u_col_put = u_col_put + 1
            call scopy( last_row, rffts_trns(  1, u_col_get),  1,
     +           f_out( first_v, u_col_put ), 1 )
         enddo
      endif
      
      if( -1 .lt. last_u ) then
         do u_col_get = 1, last_u + 1
            u_col_put = u_col_put + 1
            call scopy( last_row, rffts_trns( 1, u_col_get),   1,
     +           f_out( first_v, u_col_put ), 1 )
         enddo
      endif

      return
      end subroutine fftsynth_1a
