      subroutine fftsynth_2_m_slab(rrnum, min_x, max_x, 
     +     min_y, max_y, min_z, max_z, F_XK_dZ, 
     +     nrffts_2_row, nrffts_2_col, nrffts_2_omega, 
     +     rho, mem_total )

C     FFTSYNTH_2_M_SLAB
C     *****************************************************************
C     * Subroutine: fftsynth_2_slab                                   *
C     * Author: Robert E. Lynch (rel@cs.purdue.edu)                   *
C     * Date:   1997                                                  *
C     * Used in: fftsynth_2_m_slab -- The nearest outer subroutine    *
C     *            for allocating memory and doing 1-D Fourier        *
C     *            synthesis (by calling this subroutine).            *
C     * Function: 1-Dimensional Fourier synthesis                     *
C     * Usage:                                                        *
C     *   fftsynth_2_slab( num_hkl, min_x, max_x, min_y, max_y,       *
C     *     min_z, max_z, F_XK_dZ, nrffts_2_row, nrffts_2_col,        *
C     *     rffts_2, rffts_2_trns, nrffts_2_omega, rffts_2_omega,     *
C     *     rho )                                                     *
C     * Input:                                                        *
C     *   F_XK_dZ (Real) -- The output of 2-D Fourier synthesis       *
C     *   min_x (Integer) -- The minimum index in X axe of the real   *
C     *                      space.                                   *
C     *   min_y (Integer) -- The minimum index in Y axe of the real   *
C     *                      space.                                   *
C     *   min_z (Integer) -- The minimum index in Z axe of the real   *
C     *                      space.                                   *
C     *   max_x (Integer) -- The maximum index in X axe of the real   *
C     *                      space.                                   *
C     *   max_y (Integer) -- The maximum index in Y axe of the real   *
C     *                      space.                                   *
C     *   max_z (Integer) -- The maximum index in Z axe of the real   *
C     *                      space.                                   *
C     *   num_hkl (Integer) -- The number of coordinates in H, K or L *
C     *                        axe of the reciprocal space.           *
C     *   rffts_2 (Real) -- Working sites of Fourier synthesis.       *
C     *   rffts_2_trns (Real) -- The Transposed array of "rffts_2".   *
C     *   rffts_2_omega (Real) -- The working site for calling LAPACK *
C     *                           routine.                            *
C     *   nrffts_2_row, nrffts_2_col (Integer) -- The limits of size  *
C     *                                           of array "rffts_2". *
C     *   nrffts_2_omega (Integer) -- The limits of size of array     *
C     *                               "rffts_2_omega".                *
C     * Output:                                                       *
C     *   rho (Real) -- The final calculated density.                 *
C     *                                                               *
C     * Algorithm:                                                    *
C     *   Algorithm used here is very similar to that used in 2-D     *
C     *   FFT synthesis, except that the synthesis is only done along *
C     *   one direction in this subroutine.                           *
C     *                                                               *
C     *   For each plane perpendicular to Z axis, do the following:   *
C     *                                                               *
C     *   Firstly, transfer the plane F_XK_dZ to RFFTS.               *
C     *   Then, carry out a real FFT synthesis (back transform) of    *
C     *   each of the rows of RFFTS.  The Fourier (column) index K    *
C     *   becomes the spatial (column) index Y.                       *
C     *                                                               *
C     *   FFT synthesis is carried out by calling LAPACK routine      *
C     *   "vrfftb"                                                    *
C     *                                                               *
C     * Modification history:                                         *
C     *     - Hong Lin, Version 1.0, Aug.18, 1998                     *
C     *     - Hong Lin, Version 2.0, Oct. 6, 1999                     *
C     *       Changed the data structure so that the input is         *
C     *       separated into three arrays F_XK_dZ, F_YH_dZ, and       *
C     *       F_XL_dY and output is put in RHO and RHO_XZ.            *
C     *     - Hong Lin, Version 2.0, Dec.10, 1999                     *
C     *       Remove data structure F_YH_dZ and F_XL_dY.              *
C     *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002        *
C     *     Merged fftsynth_2_m_slab and fftsynth_2_slab RSS 5/9/06   *
C     *                                                               *
C     *****************************************************************
C     From COMMMON blocks ---------------------------------------------
C     IAM        INTEGER*4; node number; in a COMMON in ALLPROGS.inc
C      
C     NUM_NODES  INTEGER*4; number of nodes; in a COMMON in
C     ALLPROGS.inc
C      
C     FRST_LST_U(1:2,0:(NUM_NODES-1) INTEGER*4; first and last values
C     of U:    Z
C     in a COMMON in SMALL_ARRAYS.inc
C     
C     FRST_LST_Z(1,IAM):FRST_LST_Z(2,IAM) INTEGER*4; first and last values
C     of Z; in a COMMON in SMALL_ARRAYS.inc
C     
C     Arguments --------------------------------------------------------
C     NUM_HKL     INTEGER*4; number of H,K,L: 2*HKL_MAX + 1
C     
C     MIN_X,MAX_X, MIN_Y,MAX_Y, MIN_Z,MAX_Z  INTEGER*4; first and last
C     values of spatial coordinates X, Y, and Z.
C     
C     F_XK_dZ( MIN_X:MAX_X, 1:NUM_HKL, FRST_LST_U(1,IAM):FRST_LST_U(2,IAM))
C     REAL*4; values having X transformed from H, some of
C     Z transfomred from L, and K-Fourier coefficients
C     
C     NRFFTS_2_ROW,NRFFTS_2_COL  INTEGER*4; dimensions of 2D arrays RFFTS_2
C     and RFFTS_2_TRNS
C     
C     RFFTS_2(1:NRFFTS_1_ROW, 1:NRFFTS_1_COL)  REAL*4; a work array to carry
C     out the Fourier synthesis.
C     
C     RFFTS_2_TRNS(1:NRFFTS_1_ROW, 1:NRFFTS_1_COL)  REAL*4; a work array to
C     carry out the Fourier syntheses.
C     
C     NRFFTS_2_OMEGA   INTERGER*4; dimension of array for parameters for
C     the Fourier synthesis; NRFFTS_2_OMEGA must be
C     .ge. NRFFTS_2_ROW + 15.
C     
C     -- Output ------------------------------------------------------------
C     RHO( XMIN:XMAX, YMIN:YMAX, FRST_LST_Z(1,IAM):FRST_LST_Z(2,IAM) )
C     REAL*4; part of the final output.
C     ----------------------------------------------------------------------
C     External
C     BLAS routines; typically these are optimized code from vendor.
C     See modern books on numerical linear algbra or the syntax given
C     in E. Anderson, et al., LAPACK User's Guide, 2nd Edit., Siam
C     Publications, 1996.
C     
C     CALL SCOPY( N, X, INCX, Y, INCY )
C     Copy vector X to vector Y
C     N    INTEGER*4: number of entries of X to be copied
C     X    REAL*4:    vector to be copied to Y; X must have
C     length at least N*INCX
C     INCX INTEGER*4: stride of X
C     Y    REAL*4:    vector copied from X; Y must have length
C     at least N*INCY
C     INCY INTEGER*4: stride of Y
C     The procedure is equivalent to
C     JY = 1
C     DO JX = 1, ( (N-1)*INCX + 1 ), INCX
C       Y(JY) = X(JX)
C       JY = JY + INCY
C     ENDDO
C     
C     REAL DFFT synthesis routines: Sweet, R.A., et al., VFFTPK, A
C     Vectorized Pacakage of Fortran Subprograms for the Fast Fourier
C     Transform of Multiple Real Sequences, 1990; obtainable by e-mail,
C     ftp, or www from netlib@ornl.gov
C     
C     Syntax:  CALL VRFFTB( N, L, R, RT, RDIM, OMEGA )
C     CALL VRFFTI( L, OMEGA )
C     where:  N     is the number of sequences transformed
C     L     is the length of each sequence being transformed
C     R     is a REAL*4 2D array: R(  RDIM, at least L )
C     containing the N sequences (rows) of length L
C     (columns) to be transformed
C     RT    is a REAL*4 2D array: RT( RDIM, at least L )
C     RDIM  is the row dimension of R and RT
C     OMEGA is a REAL*4 array with dimension at least L + 15
C     VRFFTI must be called before VRFFTB is called; VRFFTI does not
C     have to be called between succesive calles of VRFFTB unless the
C     value of L changes.  After VRFFTB has been executed, the
C     synthesised sequences are contained in the first N rows and first
C     L columns of R
C     ----------------------------------------------------------------------
      
      implicit none
      include '../include/allprogs.inc'
      
C     Subroutine arguments
      integer rrnum, min_x,max_x, min_y,max_y, min_z,max_z,
     +     nrffts_2_row, nrffts_2_col, nrffts_2_omega
      
      real F_XK_dZ(min_x:max_x, rrnum,
     +     frst_lst_u(1,IAM):frst_lst_u(2,IAM))
      real rho(min_x:max_x, min_y:max_y,
     +     frst_lst_z(1,IAM):frst_lst_z(2,IAM))
      
      double precision mem_total
      
C     Local variables
      real, allocatable ::  rffts_2(:,:), rffts_2_trns(:,:), 
     +     rffts_2_omega(:)
      
      integer num_x, num_trns, alloc_stat,
     +     RR, u, z, y_col_get,y_col_put
C     --------------------------------------------------------
      
      allocate(rffts_2(nrffts_2_row, nrffts_2_col),
     +     rffts_2_trns(nrffts_2_row, nrffts_2_col),
     +     rffts_2_omega(nrffts_2_omega),
     +     stat = alloc_stat)
      
      if (alloc_stat .gt. 0) then
         call error_stop(progname,
     +        'fftsynth_2_m', 
     +        'Error in allocating memory for rffts_2, etc.')
      endif
      
      mem_total = 2*4*nrffts_2_row*nrffts_2_col + 4*nrffts_2_omega
      
C     Initialize the FFT parameters used to transform rows of RFFTS
C     each row has length NRFFT_COL
C     CALL VRFFTI
C     ***********
      call vrffti( rrnum, rffts_2_omega )
      
      num_x     = - min_x + max_x + 1
      num_trns  = num_x
      
      do U = frst_lst_u(1,iam), frst_lst_u(2,iam)
         do RR = 1, rrnum
            call scopy( num_x, F_XK_dZ( min_x, RR, U ), 1,
     +           rffts_2( 1, RR ),        1 )
         enddo
         
C     Synthesis all remaining Fourier coefficients:
C     (A)   F_XK_dZ --> F_XY_dZ
C     (B)   F_XL_dY --> F_XZ_dY
C     (C)   F_YH_dZ --> F_YX_dZ
         
         call vrfftb( num_trns, rrnum, rffts_2, rffts_2_trns,
     +        nrffts_2_row, rffts_2_omega     ) 
         
         if( ( min_z .le. U ) .and. ( U .le. max_z ) ) then
            
C     Move (A) to the 3-dimensional output array RHO
            
            z = U
            y_col_put = min_y - 1
            if( min_y .lt. 0 ) then
               do y_col_get = nrffts_2_col + min_y + 1, nrffts_2_col
                  y_col_put = y_col_put + 1
                  call scopy( num_x, rffts_2( 1, y_col_get), 1,
     +                 rho( min_x, y_col_put, z ), 1 )
               enddo
            endif
            
            if( -1 .lt. max_y ) then
               do y_col_get = 1, max_y + 1
                  y_col_put = y_col_put + 1
                  call scopy( num_x, rffts_2( 1, y_col_get), 1,
     +                 rho( min_x, y_col_put, z ), 1 ) 
               enddo
            endif
         endif
      enddo
      
      deallocate(rffts_2, rffts_2_trns, rffts_2_omega)
      
      return
      end subroutine fftsynth_2_m_slab
