      subroutine matrixentries_slab(nfc_2d_row, nfc_2d_col, fc_2d,
     $     hkl_max, interpolants, rot_mat, ruv_mat_1, ruv_mat_2,
     $     ruv_sup_1, ruv_sup_2, num_annulus, abs_ctf, imgmagfactor,
     $     mapmagfactor)

C                                                     MATRIXENTRIES_SLAB
C **********************************************************************
C *                                                                    *
C * Subroutine: matrixentries_slab                                     *
C * Author: Hong Lin (linh@cs.purdue.edu)                              *
C * Date:   1997                                                       *
C * Used in: cmpt_intrps -- The subroutine for Fourier transform and   *
C *                         interpolation                              *
C * Function: Interpolation in "piecewise constant" method.            *
C * Usage:                                                             *
C *   call matrixentries_slab( nfc_2d_row, nfc_2d_col, fc_2d,          *
C *          hkl_max, interpolants, rot_mat, ruv_mat_1, ruv_mat_2,     *
C *          ruv_sup_1, ruv_sup_2 )                                    *
C * Input:                                                             *
C *   nfc_2d_row (Integer) -- Limits on the number of rows of the output     *
C *                           array FC_2D.                                   *
C *   nfc_2d_col (Integer) -- Limits on the number of columns of the output  *
C *                           array FC_2D.                                   *
C *   fc_2d (Complex) -- Array with NFC_2D_ROW .eq. NFC-2D_COL, containing   *
C *                      the Fourier coefficients.                           *
C *   hkl_max (Integer) -- The maximum index of HKL space.                   *
C *   rot_mat (Real) -- The matrix for coordinate conversion between XYZ and *
C *                     PQR space.                                           *
C *   ruv_mat_1, ruv_mat_2, ruv_sup_1, ruv_sup_2 (Real) -- For axis IA, the  *
C *     matrix-vector product RUV_MAT(:,:,IA) * [IP,IQ]^T is equal to        *
C *     [R_PRIME,U_PRIME,V_PRIME]^T which gives the coordinates              *
C *     (R_PRIME,IP,IQ) of a point in the 3D standard space of Fourier       *
C *     coefficients and the corresponding point (U_PRIME,V_PRIME,-0.5) and  *
C *     (U_PRIME,V_PRIME,0.5) in the plane for Fouier coefficients of the    *
C *     pixel values.                                                        *
C *                                                                          *
C * Algorithm:                                                               *
C *   This routine computes the information necessary to combine to          *
C *   determine the 3D Fourier coefficients on a cubic mesh, FC(H,K,L),      *
C *   H_MIN:H_MAX, K_MIN:K_MAX, L_MIN:L_MAX.                                 *
C *                                                                          *
C *   The input includes the 2D Fourier coefficients, FC_2D(U,V),            *
C *   U_MIN:U_MAX, V_MIN:V_MAX, of the transform of the projection of the    *
C *   particle.  The plane of the projection has a general orientation with  *
C *   respect to the axes used for FC(H,K,L).  If the axis H is not parallel *
C *   to the plane, then for each pair, (K,L), the intersection of the line  *
C *   through (K,L) and parallel to the H-axis intersects the plane at       *
C *   (U',V') at (H',K,L).   Similarly, we can obtain 2 other intersections. *
C *   In this subroutine, "piecewise constant" method is used, which means   *
C *   that the width of the 2D DFT is 1 (instead of 0 in MATRIXENTRIES).     *
C *   Therefore, it is not necessary to obtain 3 intersections. Instead, we  *
C *   one direction to which the plane of 2D DFT is not parallel and         *
C *   calculate the grid points which are embraced by the 2D DFT slab. We    *
C *   consider the value at the nearest (U,V,W) point in 2D DFT slab as one  *
C *   of the estiamtes for the 3D point. Note that -0.5<=W<=0.5 and the      *
C *   value at (U,V,W) equals to FC_2D(U,V). Then, we average all estimates  *
C *   for the point in 3D HKL space from different pixel frames. As one node *
C *   only processes a portion of the pixel frames, we keep the number of    *
C *   estimates got in this node for each point and do a final average in    *
C *   subroutine SOLVE_BY_INTERP_SLAB.                                 *
C *                                                                    *
C * Modification history:                                              *
C *   - Hong Lin, Version 2.0, Oct. 7, 1999                            *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *

C *                                                                    *
C **********************************************************************
C Arguments  -----------------------------------------------------------
C
C Input:   -----------------------------------------------------------
 
C  NFC_2D_ROW = NFC_2D_COL  INTEGER*4; dimensions of rectangular
C       array FC_2D.
 
C  FC_2D( 0:NFC_2D_ROW, -NFC_2D_COL:NFC_2D_COL)  COMPLEX*4; array of
C       complex Fourier coefficients of pixel values (after zero fill).
 
C  HKL_MAX  INTEGER*4;  lower and upper limits of the indices
C       H,K,L of the output Fouier coefficients.
 
C  RUV_MAT( 3, 2, 3 )  REAL*4;  for axis IA, the matrix-vector product
C       RUV_MAT(:,:,IA) * [IP,IQ]^T is equal to
C       [R_PRIME,U_PRIME,V_PRIME]^T which gives the coordinates
C       (R_PRIME,IP,IQ) of a point in the 3D standard space of
C       Fourier coefficients and the corresponding point
C       (U_PRIME,V_PRIME,0) in the plane for Fouier coefficients of
C       the pixel values.

C Output:  -----------------------------------------------------------
 
C INTERPOLANTS( 3, NUM_INTRP_EQNS, -HKL_MAX:HKL_MAX, 0:HKL_MAX)
C       REAL*4;  The values INTERPOLANTS(I,N,IA,IP,IQ) are:
C         I .eq. 1,2,3 for the Nth set of values R_PRIME, real( FC ),
C            and aimag( FC ), respectively, where R_PRIME is the
C            point of intersection of the line  (r,IP,IQ)  in the 3D
C            standard space of Fourier coefficients and a point in
C            the plane of Fourier coefficients of the pixel values
C            for the axis-system IA
C ----------------------------------------------------------------------
C
      implicit none
#include "./map_type.inc"

      integer, intent(in) :: nfc_2d_row, nfc_2d_col, hkl_max

      real, intent(in) :: abs_ctf(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col),
     $     ruv_mat_1(3,2,3), ruv_mat_2(3,2,3), ruv_sup_1(3,3),
     $     ruv_sup_2(3,3), rot_mat(3,3), num_annulus, imgmagfactor,
     $     mapmagfactor
             
      complex, intent(in) :: fc_2d(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col)
      
      FMAP_TYPE, intent(inout) ::
     +     interpolants(3,-hkl_max:hkl_max,-hkl_max:hkl_max,0:hkl_max)


      complex fc_intrp, fc_ne, fc_nw, fc_sw, fc_se
      
      integer ia, ip, iq, iu_minus, iu_plus, iv_minus, iv_plus,
     $     r_prime_up, r_prime_dn, case, ih, ik, il, ip_start, iq_start,
     $     ia_start, ip_end, iq_end, ia_end
      
      real pi, uv_max, r_prime_1, r_prime_2, r_prime_start, r_prime_end,
     $     u_prime, v_prime, del_u_minus, del_u_plus, del_v_minus,
     $     del_v_plus, mag_annulus_sqr, mag_annulus, map_ratio,
     $     fft_ratio, w_tmp, ct_ne, ct_nw, ct_sw, ct_se, ctf_intrp
      
      logical :: accurate = .false.

C Note: accurate = .true. -- Using Fourier transform theory to
C                            calculate the interpolant;
C       accurate = .false. -- Using bilinear interpolation.

      character(len=18) :: subname = 'MATRIXENTRIES_SLAB'
      

C     NFC_2D_ROW .eq. NFC_2D_COL .eq. - NFC_2D_COL

      pi = acos(-1.0)

C     Scale num_annulus according to the mapmagfactor and ratio of HKL
C     and 2D dimension

      map_ratio        = mapmagfactor * nfc_2d_row / hkl_max
      uv_max           = real(nfc_2d_col)
      mag_annulus      = num_annulus / map_ratio
      mag_annulus_sqr  = mag_annulus * mag_annulus

C     Calculate the correction that must be applied to the interpolated
C     FT value due to the image mag factor, map mag factor, and ratio of
C     3D and 2D dimension. NOTE - Y Ji says that there are still
C     problems with this ratio because exp(1/N)

      fft_ratio = (mapmagfactor**3) / (imgmagfactor**2)
     +     * nfc_2d_row / hkl_max

      ip_start = -hkl_max
      ip_end   = hkl_max
      iq_start = -hkl_max
      iq_end   = hkl_max
      ia_start = -hkl_max
      ia_end   = hkl_max

C     Decide which index among H, K and L will be used as R_PRIME.
      if ( abs(rot_mat(3,1)) >= abs(rot_mat(3,2)) .and.
     +     abs(rot_mat(3,1)) >= abs(rot_mat(3,3))) then
         case = 1
         iq_start = 0
      elseif (abs(rot_mat(3,2)) >= abs(rot_mat(3,1)) .and.
     +        abs(rot_mat(3,2)) >= abs(rot_mat(3,3))) then
         case = 2
         ia_start = 0
      elseif (abs(rot_mat(3,3)) >= abs(rot_mat(3,1)) .and.
     +        abs(rot_mat(3,3)) >= abs(rot_mat(3,2))) then
         case = 3
         ip_start = 0
      else
         case = 0
         call error_stop('P3DR', subname, 
     +        'Error: cannot determine which case should be used!')
      endif
      
      ip_start = max( ip_start, int(-mag_annulus - 1.) )
      ip_end   = min( ip_end, int(mag_annulus + 1.) )
      iq_start = max( iq_start, int(-mag_annulus - 1.))
      iq_end   = min( iq_end, int(mag_annulus + 1.))
      ia_start = max( ia_start, int(-mag_annulus - 1.))
      ia_end   = min( ia_end, int(mag_annulus + 1.))

      do ip = ip_start,ip_end 
C       IP .eq. L, H, K for axis IA .eq. 1, 2, 3, respectively
C       When the array  RUV_P_PRIME(1:3,1:3) is filled, its three
C       row will hold the contributions to R', U', V' with the first
C       column having the contribution for IA = 1 for axis = H, the
C       second for IA = 2 for axis K, and the third for IA = 3 for L.
C     
         do iq = iq_start,iq_end 
C         IQ .eq. K, L, H for axis IA .eq. 1, 2, 3, respectively
C Calculate the start and end index of the R_PRIME's of those grid points
C which are in the slab.
C
            r_prime_start = ruv_mat_1(1,1,case)*real(ip) + 
     +           ruv_mat_1(1,2,case)*real(iq) + 
     +           ruv_sup_1(1,case)
            r_prime_end   = ruv_mat_2(1,1,case)*real(ip) +
     +           ruv_mat_2(1,2,case)*real(iq) + 
     +           ruv_sup_2(1,case)

            r_prime_1     = min(r_prime_start,r_prime_end)
            r_prime_2     = max(r_prime_start,r_prime_end)

            if (r_prime_1 <= 0) then
               r_prime_dn  = int(r_prime_1)
            else
               r_prime_dn  = int(r_prime_1+1.0)
            endif
            if (r_prime_2 < 0) then
               r_prime_up  = int(r_prime_2-1.0)
            else
               r_prime_up  = int(r_prime_2)
            endif
            r_prime_dn    = max(ia_start, r_prime_dn)
            r_prime_up    = min(ia_end,  r_prime_up)

C     Now each column of the array RUV_PQ_PRIME contains the values
C     of R', U', V' due to contributions from both IP and IQ; the
C     first column has values for IA .eq. 1 for H, second column for
C     IA .eq. 2 for K, and third column for IA .eq. 3 for L
 
            do ia = r_prime_dn, r_prime_up
 
C     The components RUV_PQ_PRIME( 1:3, ia ) are in units of the
C     the space of 3d Fourier coefficients whose synthesis gives
C     the output electron density.  The maximum grid point is
C     HKL_MAX.  The maximum grid point in the 2D Fourier
C     coefficients of the pixel values is NFC_2D_COL. The edge
C     lengths of the box and square containing the points are
C     equal.  The conversition factor is for the gird spacings is
C     RAT_FRQ_3D_TO_2D = NFC_2D_COL/HKL_MAX

               if (case == 1) then
                  ih = ia
                  ik = iq
                  il = ip
               elseif (case ==2) then
                  ih = ip
                  ik = ia
                  il = iq
               else
                  ih = iq
                  ik = ip
                  il = ia
               endif
               
C     Get the location of the point where we will calculate the
C     interpolant. Scale by map_ratio/imgmagfactor to account for
C     difference of particle image size from standard

               u_prime = rot_mat(1,1)*real(ih) + rot_mat(1,2)*real(ik)
     +              + rot_mat(1,3)*real(il)
               v_prime = rot_mat(2,1)*real(ih) + rot_mat(2,2)*real(ik)
     +              + rot_mat(2,3)*real(il)
               u_prime = u_prime * map_ratio / imgmagfactor
               v_prime = v_prime * map_ratio / imgmagfactor
               
               w_tmp   = rot_mat(3,1)*real(ih) + rot_mat(3,2)*real(ik)
     +              + rot_mat(3,3)*real(il)
               
C     Terminate execution if |w_tmp| >1
               if (abs(w_tmp).gt.1.0) then
                  print *, ' '
                  print *, 'r_prime_start, r_prime_end,',
     +                 ' r_prime_dn, r_prime_up =',
     +                 r_prime_start, r_prime_end, 
     +                 r_prime_dn, r_prime_up
                  print *, 'ih, ik, il, u_prime, v_prime, w_tmp =',
     +                 ih, ik, il, u_prime, v_prime, w_tmp
                  call error_stop('P3DR', subname,
     +                 'Error: w value out of range!')
               endif

C     Continue with interpolation calculation only if u_prime, v_prime
C     and (ih**2+ik**2+il**2) are within specified limits

               if(  abs(u_prime) <= uv_max .and.
     $              abs(v_prime) <= uv_max .and.
     $              real(ih**2+ik**2+il**2) <= mag_annulus_sqr ) then
                  

C     Get data needed for the bilinear interpolation:
C     (1) Coordinates of four neighboring points
C     (2) Distances from (u'v') to neighboring points
                  
                  iu_minus = floor(u_prime)
                  iu_plus = 1 + iu_minus
                  del_u_minus = u_prime - iu_minus
                  del_u_plus = 1.0 - del_u_minus

                  iv_minus = floor(v_prime)
                  iv_plus = 1 + iv_minus
                  del_v_minus = v_prime - iv_minus
                  del_v_plus = 1.0 - del_v_minus
                  
C     Make sure that coordinates of neighboring points are within the
C     array bounds

                  if(  (iu_plus  < -nfc_2d_row)  .or.
     $                 (iu_minus >  nfc_2d_row)  .or.
     $                 (iv_plus  < -nfc_2d_col)  .or.
     $                 (iv_minus >  nfc_2d_col)) then
                     cycle
                  endif
                  
C     Get the elements of arrays fc_2d and abs_ctf at the four
C     surrounding points. Use the Friedel condition (F(-k) = F(k)*) to
C     get values of array elements where the first index is negative.

                  if( iu_minus < 0 ) then
                     fc_nw = conjg( fc_2d( - iu_minus, - iv_minus ) )
                     fc_ne = conjg( fc_2d( - iu_minus, - iv_plus  ) )
                     ct_nw = abs_ctf( - iu_minus, - iv_minus )
                     ct_ne = abs_ctf( - iu_minus, - iv_plus  )
                  else
                     fc_nw = fc_2d( iu_minus, iv_minus )
                     fc_ne = fc_2d( iu_minus, iv_plus  )
                     ct_nw = abs_ctf( iu_minus, iv_minus )
                     ct_ne = abs_ctf( iu_minus, iv_plus  )
                  endif
                  
                  if( iu_plus < 0 ) then
                     fc_sw = conjg( fc_2d( - iu_plus,  - iv_minus ) )
                     fc_se = conjg( fc_2d( - iu_plus,  - iv_plus  ) )
                     ct_sw = abs_ctf( - iu_plus,  - iv_minus )
                     ct_se = abs_ctf( - iu_plus,  - iv_plus  )
                  else
                     fc_sw = fc_2d( iu_plus,  iv_minus )
                     fc_se = fc_2d( iu_plus,  iv_plus  )
                     ct_sw = abs_ctf( iu_plus,  iv_minus )
                     ct_se = abs_ctf( iu_plus,  iv_plus  )
                  endif
                  
C     If doing an 'accurate' interpolation, update del_u_minus,
C     del_u_plus, del_v_minus, and del_v_plus

                  if (accurate) then
                     if (del_u_minus < 0.00001) then
                        del_u_minus = 1.0
                     else
                        del_u_minus = sin(pi*del_u_minus)/
     $                       (pi*del_u_minus)
                     endif
                     if (del_u_plus < 0.00001) then
                        del_u_plus = 1.0
                     else
                        del_u_plus = sin(pi*del_u_plus)/
     $                       (pi*del_u_plus)
                     endif
                     if (del_v_minus < 0.00001) then
                        del_v_minus = 1.0
                     else
                        del_v_minus = sin(pi*del_v_minus)/
     $                       (pi*del_v_minus)
                     endif
                     if (del_v_plus < 0.00001) then
                        del_v_plus = 1.0
                     else
                        del_v_plus = sin(pi*del_v_plus)/
     $                       (pi*del_v_plus)
                     endif
                  endif
                  
C     Get interpolated values of abs_ctf and fc_2d. Multiply
C     interpolated FT by fft_ratio to account for image magnification
C     factor, map magnification factor, and ratio of 3D and 2D
C     dimensions. Correct interpolated abs_ctf by mutliplying by
C     (1-abs(w_tmp))
                  
                  fc_intrp = 
     +                 ((fc_nw*del_v_plus+fc_ne*del_v_minus)*del_u_plus
     +                 +(fc_sw*del_v_plus+fc_se*del_v_minus)*
     +                 del_u_minus)
     +                 * fft_ratio
                  ctf_intrp = 
     +                 ((ct_nw*del_v_plus+ct_ne*del_v_minus)*del_u_plus
     +                 +(ct_sw*del_v_plus+ct_se*del_v_minus)*
     +                 del_u_minus)
     +                 * ( 1.0 - abs(w_tmp))
                  
                  interpolants(1,ih,il,ik)=interpolants(1,ih,il,ik) + 
     +                 ctf_intrp
                  interpolants(2,ih,il,ik)=interpolants(2,ih,il,ik) +
     +                 real(fc_intrp) * ctf_intrp
                  interpolants(3,ih,il,ik)=interpolants(3,ih,il,ik) +
     +                 aimag(fc_intrp) * ctf_intrp


               endif
            enddo
         enddo
      enddo
      
      return
      end subroutine matrixentries_slab
