      subroutine rottrans_1(rotmat, ruvmat, ruvsup, w )
C                                                             ROTTRANS_1
C **********************************************************************
C *                                                                    *
C * Subroutine: rottrans_1                                             *
C * Author: Hong Lin (linh@cs.purdue.edu)                              *
C * Date:   Jul.27,1999                                                *
C * Used in: cmpt_intrps.f -- The subroutine which does the 2d FFT and *
C *                           computes information for interpolation.  *
C * Function: Construct a matrix RUVMAT which takes a pair of points   *
C *       on the coordinate axes the H,K,L-space and finds R',U',V',   *
C *       where R' is the point on the third axis in H,K,L-space so    *
C *       that the point is on the transformed pixel-value plane,      *
C *       oriented in space according to theta, phi, omega.  Using     *
C *       RM for RUVMAT:                                               *
C *                                                                    *
C * For IA .eq. 1:                                                     *
C *             [ RM(1,1,1) RM(1,2,1) ] * [L]   [ H_prime ]            *
C *             [ RM(2,1,1) RM(2,2,1) ]   [K] = [ U_prime ]            *
C *             [ RM(3,1,1) RM(3,2,1) ]         [ V_prime ]            *
C *   and (H_prime,K,L) is the same point as (U_prime,V_prime,W=w)     *
C *                                                                    *
C * For IA .eq. 2:                                                     *
C *             [ RM(1,1,2) RM(1,2,2) ] * [H]   [ K_prime ]            *
C *             [ RM(2,1,2) RM(2,2,2) ]   [L] = [ U_prime ]            *
C *             [ RM(3,1,2) RM(3,2,2) ]         [ V_prime ]            *
C *   and (H,K_prime,L) is the same point as (U_prime,V_prime,W=w)     *
C *                                                                    *
C * For IA .eq. 3:                                                     *
C *             [ RM(1,1,3) RM(1,2,3) ] * [K]   [ L_prime ]            *
C *             [ RM(2,1,3) RM(2,2,3) ]   [H] = [ U_prime ]            *
C *             [ RM(3,1,3) RM(3,2,3) ]         [ V_prime ]            *
C *   and (H,K,L_prime) is the same point as (U_prime,V_prime,W=w)     *
C *                                                                    *
C * Usage:                                                             *
C *   call rottrans_1( rotmat, ruvmat, ruvsup, w )                     *
C * Input:                                                             *
C *   rotmat(1:3,1:3) (Real) -- A matrix which maps a point (H,K,L)^T  *
C *                             to a point (U,V,W)^T.                  *
C *   w (Real) -- The offset of W for the plane normal to the          *
C *               orientation                                          *
C * Output:                                                            *
C *   ruvmat, ruvsup (Real) -- RUVMAT and RUVSUP transforms a grid     *
C *     point (IP,IQ) in the 3-d Fourier space to a point              *
C *                 (R_prime,U_prime,V_prime)                          *
C *     where R_prime is the third coordinate in the 3-d Fourier       *
C *     transform space, and (U_prime,V_prime) is the point in the     *
C *     plane of Fourier Transforms of the pixels.                     *
C *                                                                    *
C * Algorithm:                                                         *
C *                                                                    *
C * With R denoting the input matrix ROTMAX, we have                   *
C *                                                                    *
C *   [ R(1,1)  R(1,2)  R(1,3) ] [ H ]   [ U ]     [1]     [0]     [0] *
C *   [ R(2,1)  R(2,2)  R(2,3) ] [ K ] = [ V ] = U [0] + V [1] + W [0] *
C *   [ R(3,1)  R(3,2)  R(3,3) ] [ L ]   [ W ]     [0]     [0]     [1] *
C *                                                                    *
C * The pixel values are in the (U,V)-plane, W = w. This               *
C * gives the restriction:                                             *
C *                                                                    *
C *   H * R(3,1) + K * R(3,2) + L * R(3,3) = w                         *
C *                                                                    *
C * If | R(3,1) | > COSMIN, we solve for H in terms of K and L and     *
C * if | R(3,1) | <= COSMIN, then this case is skipped.  Similarly, we *
C * solve for K or for L provided | R(3,2) | > COSMIN or | R(3,3) | >  *
C * COSMIN, respectively.  The value of COSMIN is chosen arbitrarily   *
C * to be 0.001,                                                       *
C *                                                                    *
C * We have the three cases:                                           *
C *                                                                    *
C * (1) H = w/R(3,1)            - K * R(3,2)/R(3,1) - L * R(3,3)/R(3,1)*
C *                                                                    *
C * (2) K = w/R(3,2) - H * R(3,1)/R(3,2)            - L * R(3,3)/R(3,2)*
C *                                                                    *
C * (3) L = w/R(3,3) - H * R(3,1)/R(3,3) - K * R(3,2)/R(3,3)           *
C *                                                                    *
C * CASE 1:  If | R(3,1) | > COSMIN, we pick a pair of integers IK and *
C * IL for K and L, and determine the (nonintegral) value R_prime of H.*
C * We then obtain the (nonintegral) values U_prime and V_prime of U   *
C * and V.  We take IL be first and IK second:                         *
C *                                                                    *
C * R_prime = w/R(3,1) - IL * R(3,3)/R(3,1) - IK * R(3,2)/R(3,1)       *
C *                                                                    *
C * U_prime =   IL [ - R(1,1)*R(3,3)/R(3,1) +  R(1,3)]                 *
C *           + IK [ - R(1,1)*R(3,2)/R(3,1) +  R(1,2)]                 *
C *           + w*R(1,1)/R(3,1)                                        *
C *                                                                    *
C * V_prime =   IL [ - R(2,1)*R(3,3)/R(3,1) +  R(2,3)]                 *
C *           + IK [ - R(2,1)*R(3,2)/R(3,1) +  R(2,2)]                 *
C *           + w*R(2,1)/R(3,1)                                        *
C *                                                                    *
C * These give the entries in the output array.  In matrix form this   *
C * is:                                                                *
C *                                                                    *
C *     [ R_prime ]   [ RUVMAT(1,1,1) RUVMAT(1,2,1) ] [ IL ]           *
C *     [ U_prime ] = [ RUVMAT(2,1,1) RUVMAT(2,2,1) ] [ IK ] +         *
C *     [ V_prime ]   [ RUVMAT(3,1,1) RUVMAT(3,2,1) ]                  *
C *                   [ RUVSUP(1,1) ]                                  *
C *                   [ RUVSUP(2,1) ]                                  *
C *                   [ RUVSUP(3,1) ]                                  *
C *                                                                    *
C * The third index of RUVMAT, 1, denotes that R_prime is along the    *
C * H-axis                                                             *
C *                                                                    *
C * CASE 2:  Similarly, when  | R(3,2) | > COSMIN, we solve for K.     *
C * Call this R_prime and take integers IH first and IL second:        *
C *                                                                    *
C * R_prime = w/R(3,2) - IH * R(3,1)/R(3,2) - IL * R(3,3)/R(3,2)       *
C *                                                                    *
C * U_prime =   IH [ - R(1,2)*R(3,1)/R(3,2) +  R(1,1)]                 *
C *           + IL [ - R(1,2)*R(3,3)/R(3,2) +  R(1,3)]                 *
C *           + w*R(1,2)/R(3,2)                                        *
C *                                                                    *
C * V_prime =   IH [ - R(2,2)*R(3,1)/R(3,2) +  R(2,1)]                 *
C *           + IL [ - R(2,2)*R(3,3)/R(3,2) +  R(2,3)]                 *
C *           + w*R(2,2)/R(3,2)                                        *
C *                                                                    *
C * In matrix form this is:                                            *
C *                                                                    *
C *     [ R_prime ]   [ RUVMAT(1,1,2) RUVMAT(1,2,2) ] [ IH ]           *
C *     [ U_prime ] = [ RUVMAT(2,1,2) RUVMAT(2,2,2) ] [ IL ] +         *
C *     [ V_prime ]   [ RUVMAT(3,1,2) RUVMAT(3,2,2) ]                  *
C *                   [ RUVSUP(1,2) ]                                  *
C *                   [ RUVSUP(2,2) ]                                  *
C *                   [ RUVSUP(3,2) ]                                  *
C *                                                                    *
C * The third index of RUVMAT, 2, denotes that R_prime is along the    *
C * K-axis.                                                            *
C *                                                                    *
C * CASE 3:  Similarly, when  | R(3,3) | > COSMIN, we solve for L.     *
C *  Call this R_prime and take integers IK first and IH second:       *
C *                                                                    *
C * R_prime = w/R(3,3) - IK * R(3,2)/R(3,3) - IH * R(3,1)/R(3,3)       *
C *                                                                    *
C * U_prime =   IK [ - R(1,3)*R(3,2)/R(3,3) +  R(1,2)]                 *
C *           + IH [ - R(1,3)*R(3,1)/R(3,3) +  R(1,1)]                 *
C *           + w*R(1,3)/R(3,3)                                        *
C *                                                                    *
C * V_prime =   IK [ - R(2,3)*R(3,2)/R(3,3) +  R(2,2)]                 *
C *           + IH [ - R(2,3)*R(3,1)/R(3,3) +  R(2,1)]                 *
C *           + w*R(2,3)/R(3,3)                                        *
C *                                                                    *
C * In matrix form this is:                                            *
C *                                                                    *
C *     [ R_prime ]   [ RUVMAT(1,1,3) RUVMAT(1,2,3) ] [ IK ]           *
C *     [ U_prime ] = [ RUVMAT(2,1,3) RUVMAT(2,2,3) ] [ IH ] +         *
C *     [ V_prime ]   [ RUVMAT(3,1,3) RUVMAT(3,2,3) ]                  *
C *                   [ RUVSUP(1,3) ]                                  *
C *                   [ RUVSUP(2,3) ]                                  *
C *                   [ RUVSUP(3,3) ]                                  *
C *                                                                    *
C * The third index of RUVMAT, 3, denotes that R_prime is along        *
C * the L-axis.                                                        *
C *                                                                    *
C * Modification history:                                              *
C *   - Hong Lin, Version 2.0, Oct. 8, 1999                            *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *

C *                                                                    *
C **********************************************************************
      implicit none

C     Subroutine arguments
      real     rotmat(3,3), ruvmat(3,2,3), ruvsup(3,3), w

C     Local variables
      integer iaxis, irow
      real    cfl_dv_cfh, cfk_dv_cfh, cfh_dv_cfk, cfl_dv_cfk,
     +        cfk_dv_cfl, cfh_dv_cfl, cosmin

C     Lower limit on cosine arbitrarily set to 0.001
      parameter ( cosmin = 0.001 )

C ----------------------------------------------------------------------
C     Initialize matrix for failure --- make entries of RUVMAT so large
C     that computed vector components will be outside the FFT image of
C     the pixels and the Fourier coefficient array in the standard space

      do iaxis = 1, 3
         do irow = 1, 3
            ruvmat(irow,1,iaxis) = 100000.
            ruvmat(irow,2,iaxis) = 100000000000.
         enddo
      enddo

C     CASE 1:  r is a point on the H-axis
C     The angle between the normal to the plane of pixel values
C     and the H-axis is sufficiently large.
      if( abs(rotmat(3,1)) .gt. cosmin ) then
         cfl_dv_cfh    = - rotmat(3,3) / rotmat(3,1)
         cfk_dv_cfh    = - rotmat(3,2) / rotmat(3,1)
         ruvmat(1,1,1) = cfl_dv_cfh
         ruvmat(2,1,1) = cfl_dv_cfh * rotmat(1,1) + rotmat(1,3)
         ruvmat(3,1,1) = cfl_dv_cfh * rotmat(2,1) + rotmat(2,3)
         ruvmat(1,2,1) = cfk_dv_cfh
         ruvmat(2,2,1) = cfk_dv_cfh * rotmat(1,1) + rotmat(1,2)
         ruvmat(3,2,1) = cfk_dv_cfh * rotmat(2,1) + rotmat(2,2)
         ruvsup(1,1)   = w / rotmat(3,1)
         ruvsup(2,1)   = w * rotmat(1,1) / rotmat(3,1)
         ruvsup(3,1)   = w * rotmat(2,1) / rotmat(3,1)
      endif

C     CASE 2:  r is a point on the K-axis
C     The angle between the normal to the plane of pixel values
C     and the K-axis is sufficiently large.
      if( abs(rotmat(3,2)) .gt. cosmin ) then
         cfh_dv_cfk    = - rotmat(3,1) / rotmat(3,2)
         cfl_dv_cfk    = - rotmat(3,3) / rotmat(3,2)
         ruvmat(1,1,2) = cfh_dv_cfk
         ruvmat(2,1,2) = cfh_dv_cfk * rotmat(1,2) + rotmat(1,1)
         ruvmat(3,1,2) = cfh_dv_cfk * rotmat(2,2) + rotmat(2,1)
         ruvmat(1,2,2) = cfl_dv_cfk
         ruvmat(2,2,2) = cfl_dv_cfk * rotmat(1,2) + rotmat(1,3)
         ruvmat(3,2,2) = cfl_dv_cfk * rotmat(2,2) + rotmat(2,3)
         ruvsup(1,2)   = w / rotmat(3,2)
         ruvsup(2,2)   = w * rotmat(1,2) / rotmat(3,2)
         ruvsup(3,2)   = w * rotmat(2,2) / rotmat(3,2)
      endif
      
C     CASE 3:  r is a point on the L-axis
C     The angle between the normal to the plane of pixel values
C     and the L-axis is sufficiently large.
      if( abs(rotmat(3,3)) .gt. cosmin ) then
        cfk_dv_cfl = - rotmat(3,2) / rotmat(3,3)
        cfh_dv_cfl = - rotmat(3,1) / rotmat(3,3)
        ruvmat(1,1,3) = cfk_dv_cfl
        ruvmat(2,1,3) = cfk_dv_cfl * rotmat(1,3) + rotmat(1,2)
        ruvmat(3,1,3) = cfk_dv_cfl * rotmat(2,3) + rotmat(2,2)
        ruvmat(1,2,3) = cfh_dv_cfl
        ruvmat(2,2,3) = cfh_dv_cfl * rotmat(1,3) + rotmat(1,1)
        ruvmat(3,2,3) = cfh_dv_cfl * rotmat(2,3) + rotmat(2,1)
        ruvsup(1,3)   = w / rotmat(3,3)
        ruvsup(2,3)   = w * rotmat(1,3) / rotmat(3,3)
        ruvsup(3,3)   = w * rotmat(2,3) / rotmat(3,3)
      endif
 
      return
      end subroutine rottrans_1
