      program pcut
      implicit none

      character(len=256) :: infile

#ifdef MPI
      include 'mpif.h'
#endif

      include '../include/allprogs.inc'

      character(len=4) :: subname = '    '
      character(len=256) :: in_3d_name, out_file_name

      real, allocatable :: map3d(:,:,:)
      
      integer in_rad, out_rad, cut_step, max_slab, alloc_stat,
     +     ierr, npix_col, npix_row, npix_sec, mcw
      real cut_weight, x1, x2, x3, x4, angstrom_per_pix

      double precision mem_total, mem_peak
      double precision time, t_init, t_interp
      double precision elapsed_time

      integer clock1, clock2, clock3

      progname = 'pcut '

#ifdef MPI
      call mpi_init ( mpi_err )
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'error in mpi_init')
      call mpi_comm_rank ( mpi_comm_world, iam, mpi_err )
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'error in mpi_comm_rank')
      call mpi_comm_size ( mpi_comm_world, num_nodes, mpi_err )
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'error in mpi_comm_size')
      if (num_nodes > max_numnodes) then
         write(*, 9001) num_nodes, max_numnodes
 9001    format(//' requested number of nodes = num_nodes    = ',i10/
     +        ' available number of nodes = max_numnodes = ',i10/)
         call error_stop(progname, subname,
     +        'num_nodes requested exceeds max_numnodes')
      endif
#else
      iam = 0
      num_nodes = 1
#endif

      call system_clock(clock1)
      
      if (iam == 0) then
         write(*,*) 'PCUT: Parallel map cutting program'
         write(*,*) 'Parallel version run on ', num_nodes , ' CPUs'
         write(*,*) 'For doc see: http://cryoem.ucsd.edu/programs.shtm'
         write(*,*) 

         call parse_cmd_line(infile)

         if(len_trim(infile) == 0) then
            read (*,'(a)') in_3d_name
            read (*,'(a)') out_file_name
            read (*,*) x1, x2, x3, x4
         else
            open(file=infile, unit=17)
            read (17,'(a)') in_3d_name
            read (17,'(a)') out_file_name
            read (17,*) x1, x2, x3, x4
            close(17)
         endif

         write(*,'(2a)') 'input map filename: ', in_3d_name

         call get_map_dimensions(in_3d_name, npix_col, npix_row, 
     $        npix_sec, angstrom_per_pix, ierr)

         write(*,*) 'pcut c/r/s ', npix_col, npix_row, npix_sec

         if(ierr > 0) then
            call error_stop('PCUT', 'PCUT',
     +           'Error in 3D map mode, it should be 20 or 40')
         endif

         write(*,'(2a)') 'output map filename: ', out_file_name

C     Read in parameters as reals and convert to integers. This make the
C     input format a little more flexible since we avoid potential
C     runtime errors if integer values are specified with a decimal
C     point (e.g. 50.0 instead of 50)

         in_rad     = int(x1)
         out_rad    = int(x2)
         cut_step   = int(x3)
         cut_weight = x4
         
         write(*,*) 'inner radius, outer radius, cut step, cut weight:'
     $        ,in_rad, out_rad, cut_step, cut_weight
      endif

#ifdef MPI
      mcw = mpi_comm_world
      call mpi_bcast(npix_col,     1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(npix_row,     1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(in_rad,       1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(out_rad,      1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(cut_step,     1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(cut_weight,   1, mpi_real,    0, mcw, mpi_err)
      call mpi_barrier(mpi_comm_world, mpi_err)
#endif

      call system_clock(clock2)
      t_init = elapsed_time(clock1, clock2)

      call set_indices(num_nodes, 1, npix_row, frst_lst_y, max_slab)

      allocate(map3d(npix_col, npix_row, max_slab),
     +     stat = alloc_stat)

      if (alloc_stat > 0) then
         call error_stop (progname, subname,
     +     '3d density map memory allocation failed')
      else
         mem_total = mem_total + 4*npix_col*npix_row*max_slab
         mem_peak  = max(mem_peak, mem_total)
      endif

      call cut_map(map3d, max_slab, in_rad, out_rad, cut_step,
     $     cut_weight, npix_col, npix_row, in_3d_name, out_file_name)

      deallocate(map3d)
      mem_total = mem_total - 4*npix_col*npix_row*max_slab

#ifdef MPI
      call mpi_barrier(mpi_comm_world, mpi_err)
#endif

      call system_clock(clock3)
      t_interp = elapsed_time(clock2, clock3)

      if (mem_total .ne. 0) print *,'mem_total=',mem_total,
     +     'at the end, mem_total should be 0!!!'

      if (iam == 0) then
         time = t_init + t_interp

         write(*,40) iam, t_init, t_interp, time
 40      format (/ 'node ', i3,
     +        ' time used in initialization                   =',
     +        f15.4, ' ' / 8x,
     +        ' time used in read map, convert and write map  =',
     +        f15.4, ' ' / 8x,
     +        ' total time used                               =',
     +        f15.4)

         write(*,60) iam, mem_peak
 60      format(/ 'node ', i3, ' memory allocation in bytes:', /
     +        8x, ' peak memory allocated   =', f18.0)

         write(*,*) '**********finished nicely**********'
      endif

#ifdef MPI
      call mpi_finalize ( mpi_err )
#endif

      end program pcut
