      subroutine cut_map(map3d, max_slab, in_rad, out_rad, cut_step,
     $     cut_weight, npix_col, npix_row, in_3d_name, out_file_name)

C     Read in a 3D map, apply the following transformation, and write
C     out the new map
C     
C                           r < (in_rad - cut_step)  --> map3d = 0
C     (in_rad - cut_step) < r < in_rad               --> map3d = map3d*G(d)
C                  in_rad < r < out_rad              --> map3d unchanged
C                 out_rad < r < (out_rad + cut_step) --> map3d = map3d*G(d)
C    (out_rad + cut_step) < r                        --> map3d = 0
C
C     in_rad and out_rad specify the radii of the annulus containing the
C     ordered portion of the map, cut_step is the distance over which to
C     apply a Gaussian falloff, d is the distance between the location
C     of the map point and either the inner radius (r < in_rad) or the
C     other radius (r > out_rad), and G(d) is the Gaussian function
C     given by
C
C     G(d) = exp(-d**2 / (-cut_step**2/log(cut_weight)))

      implicit none
      
#ifdef MPI
      include 'mpif.h'
#endif

      include '../include/allprogs.inc'
      INCLUDE '../include/pifhead.inc'
      
      type (PIFHEAD_T) maphead

      integer, intent(in) :: in_rad, out_rad, cut_step, max_slab
      integer, intent(in) :: npix_col, npix_row
      real, intent(in)    ::  cut_weight
      character(len=256), intent(in) :: in_3d_name, out_file_name
      character(len=1)    :: rw(2)
      real, intent(out)   ::  map3D(npix_col, npix_row, max_slab)

#ifdef MPI      
      integer stat(mpi_status_size)
#endif

      integer i, j, f_z, l_z, ncol, nrow, nsec, x, y, z, sec, outmode
      character(len=7) :: subname = 'Cut_Map'
      real    radius, dist, mapprjbig
      double precision factor

C     --- Copy header information from current map to cut map ---
      if(iam == 0) then

         maphead%lunum  = 1     ! Read global header from input map
         rw(1) = 'r'
         rw(2) = achar(0)
         call pif_open(in_3d_name, maphead%lunum, rw)
         call pif_read_gh(maphead)
         
         maphead%lunum  = 8     ! Write global header to output map
         rw(1) = 'w'
         rw(2) = achar(0)
         call pif_open(out_file_name, maphead%lunum, rw)
         call pif_write_gh(maphead)
         
         maphead%lunum  = 1     ! Read data header for input map
         maphead%imgnum = 1
         call pif_read_dh(maphead)
         call pif_close(1)
         
         ncol = maphead%numcols
         nrow = maphead%numrows
         nsec = maphead%numsecs
         outmode = maphead%mode

         if(ncol .ne. npix_col .or. 
     $      nrow .ne. npix_row .or. 
     $      nsec .ne. npix_col) then
            call error_stop(progname,subname,
     +           'Density map size not consistent with img size')
         endif
         maphead%lunum  = 8     ! Write data header for output map
         maphead%imgnum = 1
         call pif_write_dh(maphead)
      endif

C     --- Read map and distribute chunks across processes ---
      call read_map(in_3d_name, map3D, npix_col, npix_row,
     $     max_slab, mapprjbig)

C     --- Apply transformation to map ---
      f_z = frst_lst_y(1,iam)
      l_z = frst_lst_y(2,iam)
      factor = - cut_step**2 / log(cut_weight)

      do z = 1, l_z - f_z + 1
         do y = 1, npix_row
            do x = 1, npix_col 
               radius = (x - (npix_col+1)/2)**2 + 
     +              (y - (npix_row+1)/2)**2
     +              + (z + f_z - 1 - (npix_col+1)/2)**2
               radius = sqrt(radius)
               if (radius > real(out_rad+cut_step)) then
                  map3D(x,y,z) = 0.0
               elseif (radius < real(in_rad-cut_step)) then
                  map3D(x,y,z) = 0.0
               elseif (radius > real(out_rad)) then
                  dist = radius - out_rad
                  map3D(x,y,z) = map3D(x,y,z)*real(exp(-dist**2/factor))
               elseif (radius < real(in_rad)) then
                  dist = in_rad - radius
                  map3D(x,y,z) = map3D(x,y,z)*real(exp(-dist**2/factor))
               endif
            enddo
         enddo
      enddo
      
C     --- Write out cut map ---
      if (iam == 0) then
         maphead%lunum  = 8
         sec = 0
         do i = 1, l_z - f_z + 1
            call pif_write_map_sections(maphead, map3D(1,1,i), 
     $           outmode, ncol, nrow, nsec, sec, 1)
            sec = sec +1
         enddo
         
         do i = 1, num_nodes-1
            f_z = frst_lst_y(1,i)
            l_z = frst_lst_y(2,i)

#ifdef MPI
            call mpi_recv(map3D(1,1,1), (l_z - f_z + 1)* (ncol * nrow),
     +           mpi_real, i, i, mpi_comm_world, stat, mpi_err)
#endif

            do j = 1, l_z - f_z + 1
               call pif_write_map_sections(maphead, map3D(1,1,j), 
     $              outmode, ncol, nrow, nsec, sec, 1)
               sec = sec +1
            enddo
         enddo
         call pif_close(maphead%lunum)
         
      else

#ifdef MPI
         call mpi_send(map3D(1,1,1),
     +        (l_z - f_z + 1) * (npix_col*npix_row),
     +        mpi_real, 0, iam, mpi_comm_world, mpi_err )
         if (mpi_err .ne. 0) then
            call error_stop(progname, subname,
     +           'Terminated because of  mpi_send  error')
         endif
#endif

      endif
      
      return
      end subroutine cut_map
