C     ************************************************************************
C     *                                                                      *
C     *                 *****     SUBROUTINE GET_BESTMAG     *****           *
C     *                                                                      *
C     *      Compares an image against a model 2D projection over a range of *
C     *    magnifications.  Returns the MAG factor with the best PRJ/IMG     *
C     *    correlation.  Correlation done in real space.                     *
C     *                                                                      *
C     *      Modified from IMG_CMP_PRJ, July 1996, by D. Belnap              *
C     *    MODIFIED  8-JAN-1998 BY T. BAKER (Changed from SCALE to MAG)      *
C     *    LAST MODIFIED 22-APR-1998 BY T. BAKER (Renamed from PFTCC_GETMAG).*
C     ************************************************************************

      SUBROUTINE GET_BESTMAG (IMG, PRJ, TMP, FFT, NCOL, NROW, 
     1 IDIM1, IDIM2, RADLO, RADHI, RESMIN, RESMAX, X_CEN, Y_CEN, 
     2 OME_BEST, MAG_LO, MAG_NUM, MAG_STEP, BEST_MAG, BEST_CCOEF)

      IMPLICIT NONE

      INTEGER*4 I, J, K, NCOL, NROW, IDIM1, IDIM2, MAG_NUM
      real*4    IMG(NCOL,NROW)
      REAL*4    TOT, PRJ(NCOL,NROW), TMP(NCOL,NROW), A, B
      complex   FFT(IDIM1,IDIM2)
      REAL*4    MAG, MAG_STEP, MAG_LO, X_CEN, Y_CEN, RESMIN, RESMAX
      REAL*4    CCOEF, BEST_CCOEF, BEST_MAG, OME_BEST
      REAL*4    RADLO, RADHI, RADDSQ, RADLOSQ, RADHISQ, YSQ
      REAL*8    SUMA, SUMB, SUMASQ, SUMBSQ, SUMAB

C  --------------------------------------------------------------

C  *******************************
C  *    Initialize parameters    *
C  *******************************
      RADLOSQ    = RADLO*RADLO
      RADHISQ    = RADHI*RADHI
      BEST_CCOEF = 0.0
      BEST_MAG   = 1.0
C     **************************************************************
C     *  Filter in reciprocal space if RESMIN or RESMAX are > 0.0  *
C     **************************************************************

      IF (RESMIN.GT.0.0 .OR. RESMAX.GT.0.0) THEN
         CALL MAP_FFT_FILL_PPFT(IMG, NCOL, NROW, FFT, IDIM1, IDIM2)
         CALL FFT_2D (FFT, IDIM1, IDIM2)
         IF (RESMAX.GT.0.0) THEN
            CALL FFT_LOP (FFT, IDIM1, IDIM2, RESMAX, 0.0, .TRUE.)
         ENDIF
         IF (RESMIN.GT.0.0) THEN
            CALL FFT_HIP (FFT, IDIM1, IDIM2, RESMIN, 0.0, .TRUE.)
         ENDIF
         CALL FFT_2D_BACK (FFT, IDIM1, IDIM2) ! Back-transform
         CALL FFT_MAP_FILL_PPFT (FFT, IDIM1, IDIM2, IMG, NCOL, NROW)

         CALL MAP_FFT_FILL_PPFT(PRJ, NCOL, NROW, FFT, IDIM1, IDIM2)
         CALL FFT_2D (FFT, IDIM1, IDIM2)
         IF (RESMAX.GT.0.0) THEN
            CALL FFT_LOP (FFT, IDIM1, IDIM2, RESMAX, 0.0, .TRUE.)
         ENDIF
         IF (RESMIN.GT.0.0) THEN
            CALL FFT_HIP (FFT, IDIM1, IDIM2, RESMIN, 0.0, .TRUE.)
         ENDIF
         CALL FFT_2D_BACK (FFT, IDIM1, IDIM2) ! Back-transform
         CALL FFT_MAP_FILL_PPFT (FFT, IDIM1, IDIM2, PRJ, NCOL, NROW)
      ENDIF

C     ***************************************************************
C     *    LOOP THROUGH SPECIFIED MAGNIFICATION SCALE FACTOR RANGE  *
C     ***************************************************************
      MAG = MAG_LO
      DO I = 1,MAG_NUM          ! INITIALIZE PARAMETERS FOR EACH LOOP
         SUMA    = 0.0
         SUMB    = 0.0
         SUMAB   = 0.0
         SUMASQ  = 0.0
         SUMBSQ  = 0.0
         TOT     = 0.0
         DO K = 1,NROW          ! Load PRJ ("Gold standard") into TMP
            DO J = 1,NCOL
               TMP(J,K) = PRJ(J,K)
            ENDDO
         ENDDO
C     *******************************************************
C     *  Scale projection and move it to X,Y origin of IMG  *
C     *******************************************************
         CALL INTERP2D (TMP, NCOL, NROW, X_CEN, Y_CEN,
     1        FLOAT(NCOL-1)/2.0, FLOAT(NROW-1)/2.0, OME_BEST, MAG, 0.0)
C     **********************************************************
C     *  Loop through IMG and PRJ, get values for coefficient  *
C     *  calculation, J and K must be within RADLO and RADHI   *
C     **********************************************************
         DO K = 1,NROW
            YSQ = (FLOAT(K)-Y_CEN+1.0)**2
            DO J = 1,NCOL
               RADDSQ = (FLOAT(J)-X_CEN+1.0)**2 + YSQ
               IF(RADDSQ.GE.RADLOSQ .AND. RADDSQ.LE.RADHISQ) THEN
                  A      = IMG(J,K)
                  B      = TMP(J,K)
                  SUMA   = SUMA + A
                  SUMB   = SUMB + B
                  SUMAB  = SUMAB + A*B
                  SUMASQ = SUMASQ + A*A
                  SUMBSQ = SUMBSQ + B*B
                  TOT    = TOT + 1.0
               ENDIF
            ENDDO
         ENDDO
C     *************************************
C     *  Compute correlation coefficient  *
C     *************************************
         IF(TOT.EQ.0. .OR. (SUMASQ-SUMA*SUMA).EQ.0. .OR.
     1        (SUMBSQ-SUMB*SUMB).EQ.0.) THEN
            CCOEF = 1.0E-20
         ELSE
            CCOEF = REAL((SUMAB - (SUMA*SUMB)/TOT) /
     1           (SQRT(SUMASQ - (SUMA*SUMA)/TOT)*
     2           SQRT(SUMBSQ - (SUMB*SUMB)/TOT)))
         ENDIF
         IF(ABS(CCOEF).GT.ABS(BEST_CCOEF)) THEN ! Test for best CCOEF
            BEST_CCOEF = CCOEF
            BEST_MAG   = MAG
         ENDIF
         MAG = MAG + MAG_STEP   ! Increment MAG i DO loop
      ENDDO

      RETURN
      END
