      SUBROUTINE GET_PHIOMEGA (R4_PIMG, POLAR_PRJ, NROT, NANNULI, 
     $     DEL_ANG, OME_BEST, ANN_LO, ANN_HI, FLIP, CCPRJ_BEST)

C     Find the best value of omega and sign of phi for image
C     based on the real space correlation coefficient. Before this
C     routine is called, it's assumed that theta and absolute value
C     of phi have already been determined.
C
C     May 1992, Purdue University VAX 8550, T Baker
C     Modified 29-APR-1998, T Baker (Renamed from PFTCC_GETPO)
C     Modified 15-AUG-2006, R Sinkovits, overhaul for improved performance

      IMPLICIT NONE

      INTEGER*4, INTENT(IN) :: NROT, NANNULI, ANN_LO, ANN_HI
      REAL*4, INTENT(IN)    :: DEL_ANG
      REAL*4, INTENT(IN)    :: POLAR_PRJ(NROT,NANNULI)
      REAL*4, INTENT(IN)    :: R4_PIMG(NROT,NANNULI)

      REAL*4, INTENT(OUT)   :: OME_BEST, CCPRJ_BEST
      LOGICAL, INTENT(OUT)  :: FLIP

      INTEGER*4 I, II_POS, II_NEG, J, IXOFF
      REAL*4    A, B, B_POS, B_NEG, SUMA, SUMB, SUMAB_POS, SUMAB_NEG
      REAL*4    SUMASQ, SUMBSQ, DIFFERENCE
      REAL*4    CCPRJ, CCPRJ_POS, CCPRJ_NEG, TOT
      REAL*4    OMEGA_POS, OMEGA_NEG, ACC_POS, ACC_NEG

      CCPRJ_POS = -HUGE(CCPRJ_POS)
      CCPRJ_NEG = -HUGE(CCPRJ_NEG)
      OMEGA_POS = 360.
      OMEGA_NEG = 360.
      SUMA       = 0.0
      SUMB       = 0.0
      SUMASQ     = 0.0
      SUMBSQ     = 0.0
      TOT        = 0.0

C     Precalculate the invariants
      DO J = ANN_LO,ANN_HI
         DO I = 1,NROT
            A = R4_PIMG(I,J+1)*J
            B = POLAR_PRJ(I,J+1)*J
            SUMA = SUMA + A
            SUMB = SUMB + B
            SUMASQ = SUMASQ + A*A
            SUMBSQ = SUMBSQ + B*B
            TOT  = TOT + 1.0
         ENDDO
      ENDDO

C     Calculate the CC for each omega at +/- phi
      DO IXOFF = 0,NROT-1
         SUMAB_POS  = 0.0
         SUMAB_NEG  = 0.0

         DO J = ANN_LO,ANN_HI
            ACC_POS = 0.0
            ACC_NEG = 0.0
            DO I = 1,NROT
               II_POS     = I + IXOFF
               IF(II_POS > NROT) II_POS = II_POS - NROT
               II_NEG     = NROT + 1 - I - IXOFF
               IF(II_NEG <= 0) II_NEG = II_NEG + NROT

               A       = R4_PIMG(I,J+1)
               B_POS   = POLAR_PRJ(II_POS,J+1)
               B_NEG   = POLAR_PRJ(II_NEG,J+1)
               ACC_POS = ACC_POS  + A*B_POS
               ACC_NEG = ACC_NEG  + A*B_NEG
            ENDDO
            SUMAB_POS = SUMAB_POS + ACC_POS*J*J
            SUMAB_NEG = SUMAB_NEG + ACC_NEG*J*J
         ENDDO
         
         CCPRJ = (SUMAB_POS - (SUMA*SUMB)/TOT) /
     1        (SQRT(SUMASQ - (SUMA*SUMA)/TOT)*
     2        SQRT(SUMBSQ - (SUMB*SUMB)/TOT))
         IF(CCPRJ > CCPRJ_POS) THEN
            CCPRJ_POS = CCPRJ
            OMEGA_POS = 360.0 - DEL_ANG*IXOFF
         ENDIF
         
         CCPRJ = (SUMAB_NEG - (SUMA*SUMB)/TOT) /
     1        (SQRT(SUMASQ - (SUMA*SUMA)/TOT)*
     2        SQRT(SUMBSQ - (SUMB*SUMB)/TOT))
         IF(CCPRJ > CCPRJ_NEG) THEN
            CCPRJ_NEG = CCPRJ
            OMEGA_NEG = 360.0 - DEL_ANG*IXOFF
         ENDIF
         
      ENDDO
      
C     Determine which choice for (omega, sign(phi)) results in best
C     match (highest correlation coefficient). Note that in order to
C     get agreement between different machines, the map is flipped only
C     if the difference between the best CCs is greater than a 
C     specified tolerance.
      
      difference = ABS(ccprj_neg - ccprj_pos)
      if(difference > 0.0001) then
         IF(CCPRJ_NEG > CCPRJ_POS) THEN
            FLIP       = .TRUE.
            OME_BEST   = OMEGA_NEG
            CCPRJ_BEST = CCPRJ_NEG
         ELSE
            FLIP       = .FALSE.
            OME_BEST   = OMEGA_POS
            CCPRJ_BEST = CCPRJ_POS
         ENDIF
      ELSE
         FLIP       = .FALSE.
         OME_BEST   = OMEGA_POS
         CCPRJ_BEST = CCPRJ_POS
      ENDIF
      
      RETURN
      END SUBROUTINE GET_PHIOMEGA
