      subroutine get_phiomega_quick(r4_pimg, polar_prj, nrot, nannuli,
     $     del_ang, ome_best, ann_lo, ann_hi,  flip, ccprj_best)

c     find the best value of omega and sign of phi for image
c     based on the real space correlation coefficient. before this
c     routine is called, it's assumed that theta and absolute value
c     of phi have already been determined.
c
c     may 1992, purdue university vax 8550, t baker
c     modified 29-apr-1998, t baker (renamed from pftcc_getpo)
c     modified 15-aug-2006, r sinkovits, overhaul for improved performance

      implicit none

      integer, intent(in) :: nrot, nannuli, ann_lo, ann_hi
      real, intent(in)    :: del_ang
      real, intent(in)    :: polar_prj(nrot,nannuli)
      real, intent(in)    :: r4_pimg(nrot,nannuli)

      real, intent(out)   :: ome_best, ccprj_best
      logical, intent(out)  :: flip

      integer i, j, k, ixoff, ixoff_temp, step
      integer ii_pos, ii_neg, istart, iend
      real sumab_pos, sumab_neg, acc_pos, acc_neg
      real ccprj, a, b, b_pos, b_neg
      real suma, sumb, sumasq, sumbsq, tot, denom
      real omega_pos_best, omega_neg_best, difference

      real ccprj_pos_best, ccprj_neg_best
      real ccprj_pos(0:nrot-1), ccprj_neg(0:nrot-1)
      integer index_pos(0:nrot-1), index_neg(0:nrot-1)

      ccprj_pos = 0.0    !F90 array syntax
      ccprj_neg = 0.0    !F90 array syntax
      
      suma       = 0.0
      sumb       = 0.0
      sumasq     = 0.0
      sumbsq     = 0.0
      tot        = 0.0

      omega_pos_best = 0.0
      omega_neg_best = 0.0
      ccprj_pos_best = 0.0
      ccprj_neg_best = 0.0

c     Precalculate the invariants that are used by all CCs
      do j = ann_lo,ann_hi
         do i = 1,nrot
            a = r4_pimg(i,j+1)*j
            b = polar_prj(i,j+1)*j
            suma = suma + a
            sumb = sumb + b
            sumasq = sumasq + a*a
            sumbsq = sumbsq + b*b
            tot  = tot + 1.0
         enddo
      enddo

      denom = sqrt(sumasq - (suma*suma)/tot) * 
     $     sqrt(sumbsq - (sumb*sumb)/tot)

C     Decide on step size (nrot will always be a power of two)
      if(nrot < 512)   step = 1
      if(nrot == 512)  step = 2
      if(nrot == 1024) step = 4
      if(nrot == 2048) step = 8
      if(nrot == 4096) step = 16

c     calculate CC for +/- phi using coarse step over omega
      do ixoff = 0,nrot-1, step
         sumab_pos  = 0.0
         sumab_neg  = 0.0

         do j = ann_lo,ann_hi
            acc_pos = 0.0
            acc_neg = 0.0
            do i = 1,nrot-ixoff
               ii_pos  = i + ixoff
               ii_neg  = nrot + 1 - ii_pos
               a       = r4_pimg(i,j+1)
               b_pos   = polar_prj(ii_pos,j+1)
               b_neg   = polar_prj(ii_neg,j+1)
               acc_pos = acc_pos  + a*b_pos
               acc_neg = acc_neg  + a*b_neg
            enddo
            do i = nrot-ixoff+1,nrot
               ii_pos  = i + ixoff - nrot
               ii_neg  = nrot + 1 - ii_pos
               a       = r4_pimg(i,j+1)
               b_pos   = polar_prj(ii_pos,j+1)
               b_neg   = polar_prj(ii_neg,j+1)
               acc_pos = acc_pos  + a*b_pos
               acc_neg = acc_neg  + a*b_neg
            enddo
            sumab_pos = sumab_pos + acc_pos*j*j
            sumab_neg = sumab_neg + acc_neg*j*j
         enddo
         
         if(denom > 0) then
            ccprj_pos(ixoff) = (sumab_pos - (suma*sumb)/tot) / denom
            ccprj_neg(ixoff) = (sumab_neg - (suma*sumb)/tot) / denom
         else
            ccprj_pos(ixoff) = 0.0
            ccprj_neg(ixoff) = 0.0
         endif

      enddo

C     Get bext indices for positive and negative phi
      call index_sort(nrot, ccprj_pos, index_pos)
      call index_sort(nrot, ccprj_neg, index_neg)

C     Subtract 1 from index values to convert ranks to offsets
      index_pos = index_pos - 1 ! F90 array syntax
      index_neg = index_neg - 1 ! F90 array syntax
      
      ccprj_pos_best = 0.0
      ccprj_neg_best = 0.0

C     Look for best omega in vicinity of 3 top scoring +|phi| matches
      do k = nrot-1, nrot-3, -1
         istart = index_pos(k) - step + 1
         iend   = index_pos(k) + step - 1
         do ixoff = istart, iend
            if(ixoff < 0) then
               ixoff_temp = ixoff + nrot
            else if (ixoff >= nrot) then
               ixoff_temp = ixoff - nrot
            else
               ixoff_temp = ixoff
            endif
            sumab_pos  = 0.0
            do j = ann_lo,ann_hi
               acc_pos = 0.0
               do i = 1,nrot
                  ii_pos     = i + ixoff_temp
                  if(ii_pos > nrot) ii_pos = ii_pos - nrot
                  a       = r4_pimg(i,j+1)
                  b_pos   = polar_prj(ii_pos,j+1)
                  acc_pos = acc_pos  + a*b_pos
               enddo
               sumab_pos = sumab_pos + acc_pos*j*j
            enddo

            if(denom > 0) then
               ccprj = (sumab_pos - (suma*sumb)/tot) / denom
            else
               ccprj = 0.0
            endif

            if(ccprj > ccprj_pos_best) then
               ccprj_pos_best = ccprj
               omega_pos_best = 360.0 - del_ang*ixoff_temp
            endif
         enddo
      enddo

C     Look for best omega in vicinity of top scoring -|phi| matches
      do k = nrot-1, nrot-3, -1
         istart = index_neg(k) - step + 1
         iend   = index_neg(k) + step - 1
         do ixoff = istart, iend
            if(ixoff < 0) then
               ixoff_temp = ixoff + nrot
            else if (ixoff >= nrot) then
               ixoff_temp = ixoff - nrot
            else
               ixoff_temp = ixoff
            endif
            sumab_neg  = 0.0
            do j = ann_lo,ann_hi
               acc_neg = 0.0
               do i = 1,nrot
                  ii_neg     = nrot + 1 - i - ixoff_temp
                  if(ii_neg <= 0) ii_neg = ii_neg + nrot
                  a       = r4_pimg(i,j+1)
                  b_neg   = polar_prj(ii_neg,j+1)
                  acc_neg = acc_neg  + a*b_neg
               enddo
               sumab_neg = sumab_neg + acc_neg*j*j
            enddo

            if(denom > 0.0) then
               ccprj = (sumab_neg - (suma*sumb)/tot) / denom
            else
               ccprj = 0.0
            endif

            if(ccprj > ccprj_neg_best) then
               ccprj_neg_best = ccprj
               omega_neg_best = 360.0 - del_ang*ixoff_temp
            endif
         enddo
      enddo

      
c     Determine which choice for (omega, sign(phi)) results in best
c     match (highest correlation coefficient). Note that in order to
c     get agreement between different machines, the map is flipped only
c     if the difference between the best ccs is greater than a 
c     specified tolerance.
      
      difference = abs(ccprj_neg_best - ccprj_pos_best)
      if(difference > 0.0001) then
         if(ccprj_neg_best > ccprj_pos_best) then
            flip       = .true.
            ome_best   = omega_neg_best
            ccprj_best = ccprj_neg_best
         else
            flip       = .false.
            ome_best   = omega_pos_best
            ccprj_best = ccprj_pos_best
         endif
      else
         flip       = .false.
         ome_best   = omega_pos_best
         ccprj_best = ccprj_pos_best
      endif
      
      return
      end subroutine get_phiomega_quick
