      program pftsearch

C     *************************************************************************
C     *                                                                       *
C     *                     *****     pftsearch.f     *****                   *
C     *                                                                       *
C     * MODEL-BASED PROCEDURE TO DETERMINE ORIGINS, ORIENTATIONS, AND         *
C     *  MAG SCALE FACTORS OF IMAGED PARTICLES.  IT CAN BE USED TO FIND OR    *
C     *  REFINE THESE PARAMETERS.                                             *
C     *                                                                       *
C     * PROJECT CUBIC (NCOL=NROW=NSEC) 3D MAP (DEFAULT ASSUME 532 SYMMETRY    *
C     *  IN REGULARLY SPACED THETA,PHI ORIENTATIONS FOR A USER-DEFINED        *
C     *  REGION OF 3D SPACE (DEFAULT IS ONE-HALF OF THE ASYMMETRIC UNIT).     *
C     *  THESE CARTESIAN PROJECTIONS ARE THEN CONVERTED TO POLAR MAPS AND     *
C     *  POLAR FFTS.                                                          *
C     *                                                                       *
C     * LOGICAL UNITS:                                                        *
C     *           1 = REFERENCE 3D MAP (READ IF MODE.GT.0)                    *
C     *           2 = INPUT PARAMETERS DATA FILE(s)                           *
C     *           7 = INPUT RAW IMAGE DATA                                    *
C     *           8 = STATS OUTPUT FILE (EMPFT.RES1)                          *
C     *           9 = STATS OUTPUT FILE (EMPFT.RADS)                          *
C     *          10 = STATS OUTPUT FILE (EMPFT.RES2)                          *
C     *          11 = PARAMETER OUTPUT DATA FILES (EMPFT.NEW_DAT)             *
C     *                                                                       *
C     * May 2003 - program inception. Program modified from old pft.f         *
C     *                                                                       *
C     *************************************************************************

      implicit none

      include '../include/sizes.inc'
      include '../include/info.inc'

      character(len=256) :: infile

#ifdef MPI
      include 'mpif.h'
      integer :: mcw = mpi_comm_world
#endif

      include '../include/allprogs.inc'

      common/pft_com/ mode, bin, verbose, mag_cen, mag_step, 
     $     mag_num, mag_norm, xcen_model, ycen_model, sym, dthe, 
     $     dphi_90, pftrad_lo, pftrad_hi, jcut, pixres_min, pixres_max,
     $     ann_lo, ann_hi, varopt, sigcut, idim1,
     $     idim2, radius, del_ang, nview_mod,
     $     quick_omega

C     ---------------------------------------------
      include '../include/pifhead.inc'
      type (pifhead_t) maphead, imghead

      integer sendcount, basecount, remainder
      integer recvcounts(128), displs(128)
      integer prj_cnt, nprjs
      double precision elapsed_time
      double precision t_proj, t_total, t_oo
      integer clock1, clock2, clock3, clock4
      real*4, allocatable :: mod_ts(:), mod_ps(:)
      real*4, allocatable :: prjs(:,:,:), prj(:,:)
      real*4, allocatable :: map3d(:,:,:), map_slice(:,:)
      real*4  mod_om, the1, the2, phi1, tempval
      integer*4 id, k
C     ---------------------------------------------

      integer nptles, i, j, file
      integer packradius(pfilesmx)

      integer inunit
      integer*4 BIN, VERBOSE, MODE, COMPARE_MODE, QUICK_OMEGA
      integer*4 SYM, NCOL, NROW, NSEC, NROT, NANNULI, IDIM1, IDIM2
      integer*4 NVIEW_MOD, NPIX, JCUT, ANN_LO, ANN_HI
      integer*4 NCOLBIN, NROWBIN, NSECBIN
      integer*4 MAG_NUM, MAG_NORM, alloc_stat
      real*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
      real*4    XCEN_MODEL, YCEN_MODEL, FILT_FAC(10)
      real*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
      real*4    RADIUS, SIGCUT, PI
      character(len=filename_len) :: map_filename
      character(len=1)     :: rw(2)
      logical :: varopt, all_empty=.false.

      integer :: kxyz(3), mxyz(3), jmode
      real*4 :: dmin, dmax, dmean
      
      integer, allocatable :: indx(:)
      real, allocatable    :: theta(:), phi(:), omega(:)
      real, allocatable    :: x(:), y(:), mag(:)
      real, allocatable    :: delf_maj_ptle(:), delf_min_ptle(:)
      real, allocatable    :: ang_maj_ptle(:)
      real, allocatable    :: pft_ccs(:), prj_ccs(:), cmp_ccs(:)

#ifdef MPI
      call mpi_init(mpi_err)
      if(mpi_err .ne. 0) then
         call error_stop('pftsearch', 'main', 'Err in mpi_init')
      endif
      call mpi_comm_rank(mcw, iam, mpi_err)
      if(mpi_err .ne. 0) then
         call error_stop('pftsearch', 'main', 'Err in mpi_comm_rank')
      endif
      call mpi_comm_size(mcw, num_nodes, mpi_err)
      if(mpi_err .ne. 0) then
         call error_stop('pftsearch', 'main', 'Err in mpi_comm_size')
      endif
#else
      iam = 0
      num_nodes = 1
#endif

      call system_clock(clock1)
      pi = acos(-1.0)

c     ------------------------ MASTER -------------------------------
      if(iam == 0) then
         write(*,*) 'PFTsearch: Model-based PFT analysis'
         write(*,*) 'Parallel version run on ', num_nodes , ' CPUs'
         write(*,*) 'For doc see: http://cryoem.ucsd.edu/programs.shtm'
         write(*,*) 

         oddeven = 0
         call parse_cmd_line(infile)

C     Get user input
         if(len_trim(infile) == 0) then
            inunit = 5
            call key_info (ncol, nrow, nsec, ncolbin, nrowbin, nsecbin,
     $           nrot, nannuli, npix, compare_mode, filt_fac,
     $           map_filename, inunit)
         else
            inunit=17
            open(file=infile, unit=inunit)
            call key_info (ncol, nrow, nsec, ncolbin, nrowbin, nsecbin,
     $           nrot, nannuli, npix, compare_mode, filt_fac,
     $           map_filename, inunit)
            close(17)
         endif    

C     Determine numbers of particles in parameter files
         nptles = sum(nptle_per_img)

         if(nptles == 0)  then
            if(per_ptle_ctf == 1) then
               call error_stop('PPFT', 'main program', 
     +              'Empty ptle files incompatable with per_ptle_ctf')
            endif

            write(*,*)
            write(*,*) '***** NOTE *****'
            write(*,*) 'All particle parameter files are empty'
            write(*,*) 'Basing particle counts on boxed image files'

            all_empty = .true.
            do file=1,nimg
               if(index(img_fnames(file),'.pif') > 0) then
                  rw(1) = 'r'
                  rw(2) = achar(0)
                  call pif_open(img_fnames(file), pix_file_unit, rw)
                  imghead%lunum = pix_file_unit
                  call pif_read_gh(imghead)
                  imghead%imgnum = 1
                  call pif_read_dh(imghead)
                  call pif_close(pix_file_unit)
                  nptle_per_img(file) = imghead%nimg
                  packradius(file) = imghead%packRadius
               else if (index(img_fnames(file),'.mrc') > 0) then
                  call imopen(pix_file_unit, img_fnames(file), 'ro')
                  call irdhdr(pix_file_unit, kxyz, mxyz, jmode, 
     $                 dmin, dmax, dmean)
                  call imclose(pix_file_unit)
                  nptle_per_img(file) = kxyz(3)
                  packradius(file) = int(kxyz(1)/2)
               endif
            enddo
            nptles = sum(nptle_per_img)
         endif

         write(*,*)
         write(*,*) 'Number of parameter files = ', nparam
         write(*,*) 'Number of image files     = ', nimg
         write(*,*) 'Total number of images: ', nptles
         write(*,*)
         
         k = 0
         do i=1,nparam
            do j=1,nimg_per_param(i)
               k = k + 1
               write(*,*) 'param #',i,'orient file: ',
     $              trim(adjustl(param_fnames(i)))
               write(*,*) 'image #',k,'image  file: ',
     $              trim(adjustl(img_fnames(k)))
               write(*,*) 'views = ',  nids(k)
               write(*,*) 'numcols = ',npix_col,'numrows = ',npix_row
               write(*,*)'Ptle records = ', nptle_per_img(k)
               write(*,*)
            enddo
         enddo

      endif

C     Distribute data to all processes
#ifdef MPI
      call mpi_bcast(pixsizs,   pfilesmx, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(volts,     pfilesmx, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(ampfacs,   pfilesmx, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(cs_coeffs, pfilesmx, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(delf_majs, pfilesmx, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(delf_mins, pfilesmx, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(ang_majs,  pfilesmx, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(ncol,             1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(nrow,             1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(nsec,             1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(ncolbin,          1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(nrowbin,          1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(nsecbin,          1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(nrot,             1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(nannuli,          1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(npix,             1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(compare_mode,     1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(filt_fac,        10, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(mode,             1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(bin,              1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(verbose,          1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(mag_cen,          1, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(mag_step,         1, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(mag_num,          1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(mag_norm,         1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(xcen_model,       1, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(ycen_model,       1, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(sym,              1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(dthe,             1, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(dphi_90,          1, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(pftrad_lo,        1, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(pftrad_hi,        1, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(jcut,             1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(pixres_min,       1, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(pixres_max,       1, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(ann_lo,           1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(ann_hi,           1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(varopt,           1, mpi_logical,  0, mcw, mpi_err)
      call mpi_bcast(sigcut,           1, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(res_min,          1, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(res_max,          1, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(idim1,            1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(idim2,            1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(radius,           1, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(del_ang,          1, mpi_real,     0, mcw, mpi_err)
      call mpi_bcast(nview_mod,        1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(quick_omega,      1, mpi_integer,  0, mcw, mpi_err)
      call mpi_bcast(per_ptle_ctf,     1, mpi_integer,  0, mcw, mpi_err)

      call mpi_bcast(param_fnames, filename_len*pfilesmx, mpi_character,
     $     0, mcw, mpi_err)
      call mpi_bcast(img_fnames, filename_len*pfilesmx, mpi_character,
     $     0, mcw, mpi_err)
      call mpi_bcast(map_filename, filename_len, mpi_character, 
     $     0, mcw, mpi_err)

C     Broadcast number of parameter files, images, particles,
C     images per parameter file, and particles per image

      call mpi_bcast(nparam, 1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nimg,   1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nptles, 1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nimg_per_param, nparam, mpi_integer, 
     $     0, mcw, mpi_err)
      call mpi_bcast(nptle_per_img, nimg, mpi_integer, 
     $     0, mcw, mpi_err)
      call mpi_bcast(first_index, nimg, mpi_integer, 
     $     0, mcw, mpi_err)
#endif


C     Allocate global and local arrays for particle parameters
      allocate(indx(nptles), theta(nptles), phi(nptles), omega(nptles),
     $     x(nptles), y(nptles), mag(nptles), 
     $     pft_ccs(nptles), prj_ccs(nptles), cmp_ccs(nptles))
      
      allocate(delf_maj_ptle(nptles), delf_min_ptle(nptles),
     $     ang_maj_ptle(nptles))

C     Get particle parameters
      if(iam == 0) then
         oddeven = 0
         nbox = 1 ! (Fortran 90 array syntax)

         if(all_empty) then
            id = 0
            do file=1,nimg
               do j=1,nptle_per_img(file)
                  id = id + 1
                  indx(id) = j
                  x = packradius(file)
                  y = packradius(file)
                  mag = 1.0
               enddo
            enddo
         else
            if(per_ptle_ctf == 1) then
               call get_ptle_ioomctf(nparam, filename_len, param_fnames,
     $              oddeven, nptles, nbox,indx, theta, phi, omega, x, y,
     $              mag, delf_maj_ptle,delf_min_ptle, ang_maj_ptle)
            else                  
               call get_ptle_ioom(nparam, filename_len, param_fnames,
     $              oddeven, nptles, nbox, indx, theta, phi, omega, 
     $              x, y, mag)
            endif

         endif
      endif

C     Distribute particle information to all other processes
#ifdef MPI
      call mpi_bcast(indx,  nptles, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(theta, nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(phi,   nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(omega, nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(x,     nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(y,     nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(mag,   nptles, mpi_real,    0, mcw, mpi_err)
      if(per_ptle_ctf == 1) then
         call mpi_bcast(delf_maj_ptle, nptles, mpi_real, 0,mcw,mpi_err)
         call mpi_bcast(delf_min_ptle, nptles, mpi_real, 0,mcw,mpi_err)
         call mpi_bcast(ang_maj_ptle,  nptles, mpi_real, 0,mcw,mpi_err)
      endif
#endif

C     ----------------------------------------------------
      if(iam == 0) then
         tempval =  (ncolbin*nrowbin*nview_mod)
         tempval =  (4*tempval)
         if(tempval .gt. 2.0**32) then
            print *, '*************************************************'
            print *, 'warning - each process is asking for > 4 gb'
            print *, 'you''re asking for too much memory.'
            print *, '   number of cols   - ', ncolbin
            print *, '   number of rows   - ', nrowbin
            print *, '   number of views  - ', nview_mod
            print *, '   == ', tempval, ' total bytes requested.'
            print *, '          (4 bytes per pixel) '
         endif
      endif

      allocate(mod_ts(nview_mod), mod_ps(nview_mod),
     $     prj(ncolbin,nrowbin), stat = alloc_stat)
      if(alloc_stat .gt. 0) then
        print *, '*************************************************'
        print *, '* getmem1 memory allocation error: exit program *'
        call tsbend
      endif

C     Determine range of values for theta and phi
      call get_t1t2p1 (sym, the1, the2, phi1)

C     Get (theta,phi) pairs that cover the asymmetric units
      call calc_mod_tps (mod_ts, mod_ps, the1, the2, phi1, dthe, 
     $     dphi_90, nview_mod, sym)


C     Calculate projections
      call system_clock(clock2)
      
c     pif_read_wmapi2 not filling in edges of map3d for bin=2
c     initialize to zero to make sure that we don't get nan's
      if(iam == 0) then
         
         allocate(map3d(ncolbin,nrowbin,nsecbin), stat = alloc_stat)
         if(alloc_stat > 0) then
            print *, '**********************************************'
            print *, '* not enough memory for 3d map: exit program *'
            call tsbend
         endif
         
         map3d = 0.0            ! f90 array syntax
         
         maphead%lunum = 1
         rw(1) = 'r'
         rw(2) = achar(0)
         call pif_open (map_filename, maphead%lunum, rw) 
         call pif_read_gh (maphead)
         if(maphead%mode == 20) then
            maphead%imgnum = 1  ! load up "img" #1
            call pif_read_dh(maphead)
            call pif_read_wmapi2(maphead, map3d, ncolbin, nrowbin,
     $           nsecbin)
         else
            print *, 'invalid pif mode for 3d map'
            call tsbend
         endif
         call pif_close(maphead%lunum)

      endif

      basecount = nview_mod/num_nodes
      remainder = nview_mod - basecount*num_nodes
      
      if(iam >= (num_nodes-remainder)) then
         sendcount = basecount + 1
      else
         sendcount = basecount
      endif
      
      recvcounts = 0
      displs = 0
      do i=0,num_nodes-1
         if(i >= (num_nodes-remainder)) then
            recvcounts(i+1) = basecount + 1
         else
            recvcounts(i+1) = basecount
         endif
      enddo
      displs(1) = 0
      do i=1,num_nodes-1
         displs(i+1) = displs(i) + recvcounts(i)
      enddo
      
      allocate(prjs(ncolbin,nrowbin,nview_mod))
      allocate(map_slice(ncolbin,nrowbin))
      prjs = 0.0
      mod_om = 90.0
      
      do i=1,nsecbin
         if(iam == 0) then
            map_slice(:,:) = map3d(:,:,i)
         endif
#ifdef MPI
         call mpi_bcast(map_slice, ncolbin*nrowbin, mpi_real, 0,
     $        mcw, mpi_err)
#endif
         do id=1,sendcount
            k = displs(iam+1) + id
            call map_prj_slice(prjs(1,1,k), map_slice, 
     $           ncolbin, nrowbin, nsecbin, 
     $           mod_ts(k), mod_ps(k), mod_om, i)
         enddo
      enddo
      
      deallocate(map_slice)
      if(iam == 0) then
         deallocate(map3d)
      endif

c     The following block of code distributes each process's projections
c     to all other processes using a series of nview_mod broadcast
c     operations. This is probably a little slower than the original
c     method of using a single call to mpi_allgatherv, but avoids the
c     problems associated with mpi buffer limits.

#ifdef MPI
      do i = 0, num_nodes-1
         nprjs = sendcount
         call mpi_bcast(nprjs, 1, mpi_integer, i, mcw, mpi_err)
         do prj_cnt = 1, nprjs
            call mpi_bcast(prjs(1,1,displs(i+1)+prj_cnt),
     $           ncolbin*nrowbin, mpi_real, i, mcw, mpi_err)
         enddo
      enddo
#endif

      call system_clock(clock3)
      
c      if(iam == 0) then
c         write(*,1000)
c         do i=1, nview_mod
c            write(*,1010) i, mod_ts(i), mod_ps(i)
c         enddo
c      endif

c 1000 format('  view # theta    phi',/,21('-'),/)
c 1010 format('+',i5,f8.2,f7.2)


c  ***********************************
c  *  determine theta,phi,omega,x,y  *
c  ***********************************

      call global_cc (bin, nrot, nannuli, nview_mod, idim1, idim2,
     $     pixres_min, pixres_max, ann_lo, ann_hi, jcut, radius,
     $     sym, verbose, mag_cen, mag_step, mag_num, mod_ts, mod_ps,
     $     prjs, varopt, ncol, nrow, ncolbin, nrowbin, nptles,
     $     indx, theta, phi, omega, x, y, mag,
     $     delf_maj_ptle, delf_min_ptle, ang_maj_ptle,
     $     pft_ccs, prj_ccs, cmp_ccs, compare_mode, filt_fac, sigcut,
     $     xcen_model, ycen_model, pftrad_lo, pftrad_hi, quick_omega)
      
      deallocate(prjs, mod_ts, mod_ps, prj)

      call system_clock(clock4) 

      if(iam == 0) then

C     Calculate average CCs and MAG factor
         call calc_avgs(nptles, mag, pft_ccs, prj_ccs, cmp_ccs)
         if(verbose >= 1) then
            write(*,*) '=========================================='
            write(*,*) ' Files EMPFT.RADS, EMPFT.RES1, EMPFT.RES2 '
            write(*,*) '  written to default directory.'
            write(*,*) '=========================================='
            close(8)
            close(9)
            close(10)
         endif

         t_proj = elapsed_time(clock2, clock3)
         t_total = elapsed_time(clock1, clock4)
         t_oo = t_total - t_proj

         write(*,*)
         write(*,*) 'Time calculating projections         = ', t_proj
         write(*,*) 'Time determining origins and orients = ', t_oo
         write(*,*) 'Total PPFT run time                  = ', t_total
         write(*,*)

         write(*,*) 'PFT finished.'
      endif

      deallocate(indx, theta, phi, omega, x, y, mag, pft_ccs,
     $     prj_ccs, cmp_ccs)

      deallocate(delf_maj_ptle, delf_min_ptle, ang_maj_ptle)

#ifdef MPI
      call mpi_finalize(mpi_err)
#endif

      end program pftsearch
