#! /usr/bin/perl
use warnings;
use strict;
use File::Copy;
use FindBin qw($Bin);
use lib "$Bin/modules";
use lib "$Bin/";
use Cwd;
use version;
use get_info;

my $version = get_version();
my ($a, $b);

my @cmdline = @ARGV;
if(scalar(@cmdline) == 0) {

print << "EOF";

autopp $version - automatic data pre-processing program

A) Create particle parameter files
B) Create boxed image files from micrographs & box coordinates
C) Scale and/or shift box coordinates in BCRD files
D) Interpolate PIF micrographs to a new pixel size
E) Convert TIFF files to PIF format
F) Remove blemishes, correct linear gradient, and normalize boxed images
G) Resize boxes in BCRD files
I) Replace string in text files
J) Scale and/or shift origins in particle parameter files
K) Recenter boxes: update bcrd and particle parameter files
M) Autobox micrographs with robem
O) Apodize boxed images
Q) Quit
R) Convert PIF files to MRC format
T) Truncate PIF maps
W) Mask PIF map using one or more PDB files
X) Convert DM3 micrographs to PIF format
Y) Convert MRC files (only 2D images) to PIF format
Z) Convert PIF micrographs to CCP4 format
1) Rename files in a consistent manner
2) Generates tiles from micrographs as boxed image files (NEW in ver. 4.01.13)
3) Estimate defocus levels of PIF or MRC micrographs using ctffind3
4) Generate particle parameter files from ctffind3 output

UNIX wildcards and full/relative pathnames can be used for options that
operate on sets of files. For example, the following are all valid

   file1 file2 file3*
   dir1/file1 dir2/file[2-7].*

New files and directories always written in the directory where autopp
is launched, regardless of path information provided for files

autopp can be run with ALL arguments (no defaults) on command
line. Single quotes group files and escape wildcards. For example

   autopp F 'a*box.pif b*box.pif' n

EOF

} # End printing usage information

print "Enter your choice: ";
my $choice = get_autopp_arg(0);
$choice =~ tr/a-z/A-Z/;

SWITCH: {
    if ($choice eq 'A') { optionA(); last SWITCH; }
    if ($choice eq 'B') { optionB(); last SWITCH; }
    if ($choice eq 'C') { optionC(); last SWITCH; }
    if ($choice eq 'D') { optionD(); last SWITCH; }
    if ($choice eq 'E') { optionE(); last SWITCH; }
    if ($choice eq 'F') { optionF(); last SWITCH; }
    if ($choice eq 'G') { optionG(); last SWITCH; }
    if ($choice eq 'I') { optionI(); last SWITCH; }
    if ($choice eq 'J') { optionJ(); last SWITCH; }
    if ($choice eq 'K') { optionK(); last SWITCH; }
    if ($choice eq 'M') { optionM(); last SWITCH; }
    if ($choice eq 'O') { optionO(); last SWITCH; }
    if ($choice eq 'Q') { optionQ(); last SWITCH; }
    if ($choice eq 'R') { optionR(); last SWITCH; }
    if ($choice eq 'T') { optionT(); last SWITCH; }
    if ($choice eq 'W') { optionW(); last SWITCH; }
    if ($choice eq 'X') { optionX(); last SWITCH; }
    if ($choice eq 'Y') { optionY(); last SWITCH; }
    if ($choice eq 'Z') { optionZ(); last SWITCH; }
    if ($choice eq '1') { option1(); last SWITCH; }
    if ($choice eq '2') { option2(); last SWITCH; }
    if ($choice eq '3') { option3(); last SWITCH; }
    if ($choice eq '4') { option4(); last SWITCH; }

    print "Invalid choice $choice - Please try again\n";
    exit;		
}

print "\n--- autopp complete---\n\n";

# --------------------------------------------------------


sub optionA {

    print << "EOF";

Generate particle parameter files from a text file containing image
names and defocus information. Text file has format:

  boxed_image_filename  defmajor [defminor] [angle]

Parameter filename will be based on image name. For example:

  ../pif/file1_box_fix_norm.pif --> file1.dat_000

The image name will be reproduced exactly as written as the first line
of the parameter file. For stigmated images only one defocus parameter
is required; for astigmatic images all three paramaters are needed.

EOF

    print "Enter name of file containing filenames and defocus info: ";
    my $txtfile = get_autopp_arg(1);

    print "Enter pixel size in angstroms: ";
    my $pixel = get_autopp_arg(2);

    print "Enter accelerating voltage in KV [300]: ";
    my $kv = get_autopp_arg(3);
    if(!$kv) {
	$kv = 300.0;
    }
    my $v = $kv * 1000;

    print "Enter amplitude contrast [0.07]: ";
    my $famp = get_autopp_arg(4);
    if(!$famp) {
	$famp = 0.07;
    }

    print "Enter spherical aberration (Cs) in mm [2.0]: ";
    my $cs = get_autopp_arg(5);
    if(!$cs) {
	$cs = 2.0;
    }

    print "\n";

    open(TXT, $txtfile);
    while(<TXT>) {
	my ($pifname, $defmajor, $defminor, $angle);
	s/^\s+//;
	s/\s+$//;
	my @p = split;
	if(scalar(@p) == 2) {
	    $pifname  = $p[0];
	    $defmajor = $p[1];
	    $defminor = $p[1];
	    $angle    = 0.0;
	} elsif (scalar(@p) == 4) {
	    $pifname  = $p[0];
	    $defmajor = $p[1];
	    $defminor = $p[2];
	    $angle    = $p[3];
	} else {
	    print "line $. has wrong number of fields: $_\n";
	    next;
	}

	my $datfile = $pifname;
	$datfile =~ s/_box//;
	$datfile =~ s/_fix//;
	$datfile =~ s/_norm//;
	$datfile =~ s/_rebox//;
	$datfile =~ s/_apod//;
	$datfile =~ s/.*\///;
	$datfile =~ s/\.pif$/.dat_000/;

	print "Generating file: $datfile\n";

	open(DAT, ">$datfile");
	print DAT "$pifname\n";
	print DAT "$pixel, 1, $v, $famp, $defmajor, $defminor, $angle, $cs\n";
	close(DAT);
    }
    close(TXT);

    unless(defined $ARGV[5]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp a $txtfile $pixel $kv $famp $cs\n";
    }

}



sub optionB {

    print << "EOF";

Generate boxed image files from box coordinates and PIF micrographs.

The user specifies a list of BCRD files and a list of PIF files. For
every PIF file that contains the same base name as the BCRD file, a
new boxed image file will be generated

For example

   file1.bcrd + file1_nov12.pif --> file1_nov12_box.pif
   file2.bcrd + file2_nov12.pif --> file2_nov12_box.pif

The boxed images will not be apodized and existing files will be overwritten

EOF

    my $flag = 0;
    print "Enter list of BCRD files to be processed: ";
    my $blist = get_autopp_arg(1);

    # Expand $list into array of files
    my @bfiles = glob("$blist");
    
    # Test to make sure that files were found
    if(@bfiles <= 0) {
	print "No BCRD files were found - exiting program\n";
	exit;
    }

    print "Enter list of micrographs (scanned images) to be considered: ";
    my $plist = get_autopp_arg(2);

    # Expand $list into array of files
    my @pfiles = glob("$plist");
    
    # Test to make sure that files were found
    if(@pfiles <= 0) {
	print "No particle parameter files were found - exiting program\n";
	exit;
    }

    print "Enter box radius: ";
    my $boxrad = get_autopp_arg(3);

    print "\n";

    my %bhash;

    foreach my $bcrdfile (@bfiles) {
	my $bcrdbase = $bcrdfile;
	$bcrdbase =~ s/.*\///;
	$bcrdbase =~ s/\.bcrd//;
	my @list = ();

	foreach my $piffile (@pfiles) {
	    if($piffile =~ /$bcrdbase/) {
		push(@list, $piffile);
	    }
	}

	if(scalar(@list) > 1) {
	    $flag = 1;
	    print "Warning: Found multiple PIF files for $bcrdfile\n";
	    my $lenshort = 10000;
	    foreach my $x (@list) {
		print "   $x\n";
		if(length($x) < $lenshort) {
		    $bhash{$bcrdfile} = $x;
		    $lenshort = length($x);
		}
	    }
	    print "   Using PIF file with shortest name: $bhash{$bcrdfile}\n";
	}

	if(scalar(@list) == 1) {
	    $bhash{$bcrdfile} = $list[0];
	} elsif (scalar(@list) == 0) {
	    print "Warning: No corresponding PIF file for $bcrdfile\n";
	}
    }

    print "\n";

    foreach my $bcrdfile (sort keys %bhash) {
	my $piffile = $bhash{$bcrdfile};
	
	my $boxfile = $piffile;
	$boxfile =~ s/.*\///;
	$boxfile =~ s/\.pif/\_box\.pif/i;
	if(-e $boxfile) {
	    unlink($boxfile);
	}
	    
	print "Boxing $piffile using $bcrdfile\n";

	my $robcom = "robem -nogui -pre $piffile -box $bcrdfile ";
	$robcom .= "gaussian 0 $boxrad 0 -outfile $boxfile";
	print "$robcom\n";
	my $tmp = `$robcom`;		
    }

    if($flag > 0) {
	print "\n";
	print "WARNING - One or more bcrd files had possible matches\n";
	print "with multiple pif files. Please check autopp output to\n";
	print "make sure that bcrd/pif files were paired correctly\n";
    }

    unless(defined $ARGV[3]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp b \'$blist\' \'$plist\' $boxrad\n";
    }

}



sub optionC {

    print << "EOF";

Scale and/or shift the box coordinates in a set of BCRD files. If
scaling coordinates, box dimensions are also scaled. This option is
particularly useful if boxing had originally been done on binned data.

Note that if both shifting and scaling coordinates, shifting is done
BEFORE scaling.

New files will be written to a new directory named ADJBCRD

EOF

    print "Enter list of BCRD files to be processed: ";
    my $list = get_autopp_arg(1);

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No boxed image files were found - exiting program\n";
	exit;
    }

    print "Enter the scale factor [1]: ";
    my $scale = get_autopp_arg(2);
    if(!$scale) {
	$scale = 1;
    }

    print "Enter the x-shift in pixels [0]: ";
    my $xshift = get_autopp_arg(3);
    if(!$xshift) {
	$xshift = 0;
    }
    $xshift = sprintf ("%d", $xshift);

    print "Enter the y-shift in pixels [0]: ";
    my $yshift = get_autopp_arg(4);
    if(!$yshift) {
	$yshift = 0;
    }
    $yshift = sprintf ("%d", $yshift);

    print "\n";

    unless(-e "ADJBCRD") {
	mkdir ("ADJBCRD", 0755) || die "Cannot make the directory ADJBCRD\n";
    }

    foreach my $bcrdfile (@files) {
	open(IN, $bcrdfile);
	my $outfile = $bcrdfile;
	$outfile =~ s/.*\///;
	open(OUT, ">ADJBCRD/$outfile");
	print "Processing $bcrdfile\n";

	while(<IN>) {
	    my $line = $_;
	    $line =~ s/^\s+//;
	    $line =~ s/\s+$//;

	    if($. <= 3) {
		print OUT "$line\n";
		next;
	    }

	    if($. == 4 || $. == 5) {
		$line *= $scale;
		$line = sprintf ("%d", $line);
		print OUT "$line\n";
		next;
	    }
	    
	    my ($x, $y) = split(/\s+/, $line);
	    $x += $xshift;
	    $y += $yshift;
	    $x *= $scale;
	    $y *= $scale;
	    $x = sprintf ("%d", $x);
	    $y = sprintf ("%d", $y);
	    print OUT "$x $y\n";
	}
	
	close(IN);
	close(OUT);
    }

    unless(defined $ARGV[4]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp c \'$list\' $scale $xshift $yshift\n";
    }

}



sub optionD {
    
    print << "EOF";
    
Interpolate one or more PIF files to a new pixel size

The new PIF files will have the same name as the original files except
with the string _interp added before the .pif extension.

IMPORTANT - this option currently works correctly only when the new pixel
size is larger than the old pixel size.

EOF

    print "Enter list of PIF files to be interpolated: ";
    my $list = get_autopp_arg(1);

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No TIFF files were found - exiting program\n";
	exit;
    }

    print "Enter original pixel size: ";
    my $porig = get_autopp_arg(2);

    print "Enter new pixel size: ";
    my $pnew = get_autopp_arg(3);

    print "\n";

    foreach my $piffile (@files) {
	my $outfile = $piffile;
	$outfile =~ s/.*\///;
	$outfile =~ s/\.pif/_interp.pif/;
	print "Interpolating image $piffile\n";
	my $robcom = "robem -nogui -pre $piffile -interp $porig $pnew";
	$robcom .= " -outfile $outfile";
	print "$robcom\n";
	my $tmp = `$robcom`;
    }

    unless(defined $ARGV[3]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp d \'$list\' $porig $pnew\n";
    }

}



sub optionE {
    
    print << "EOF";
    
Convert one or more TIFF files to PIF format

The new PIF files will have the same name as the corresponding TIFF
files except with the .tif[f] extension changed to .pif

EOF

    print "Enter list of TIFF files to be processed: ";
    my $list = get_autopp_arg(1);
    $list =~ s/\s+$//;

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No TIFF files were found - exiting program\n";
	exit;
    }

    print "Enter binning factor [1]: ";
    my $bin = get_autopp_arg(2);
    if(!$bin) {
	$bin = 1;
    }

    print "Invert contrast (1=invert, 0=no invert)? [0]: ";
    my $invert = get_autopp_arg(3);
    if(!$invert) {
	$invert = 0;
    }

    print "\n";

    foreach my $tiffile (@files) {
	my $piffile = $tiffile;
	$piffile =~ s/.*\///;
	$piffile =~ s/\.tif{1,2}/.pif/;
	print "Converting $tiffile to $piffile\n";
	my $cmd = "tiff2pif $tiffile $piffile $bin $invert";
	print "$cmd\n";
	my $tmp = `$cmd`;
    }

    unless(defined $ARGV[3]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp e \'$list\' $bin $invert\n";
    }

}



sub optionF {

    print << "EOF";

Apply blemish removal, linear gradient removal, and normalization to a
set of boxed image files. Normalization parameters are typically
derived from statistics over full set of images.

The new boxed image file will have same name as original files except
with _fix_norm appended to file prefix. For example

   file1_box.pif --> file1_box_fix_norm.pif

EOF

    print "Enter list of boxed image files to be processed: ";
    my $list = get_autopp_arg(1);

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No boxed image files were found - exiting program\n";
	exit;
    }

    print "\n";
    print "User will provide mean and stddev for normalization y/n ?\n";
    print "Strongly recommend default behavior [n]: ";

    my $userspec = get_autopp_arg(2);
    if(!$userspec) {
	$userspec = 'n';
    }
    
    print "\n";

    my $normval = 0.0;
    my $stdval  = 0.0;

    if($userspec =~ /^y/i) {
	$normval = get_autopp_arg(3);
	$stdval  = get_autopp_arg(4);
    } else {
	unlink('trashnorm.txt');
	foreach my $boxfile (@files) {
	    my $robcom = "robem -nogui -pre $boxfile -DUMPNORM trashnorm.txt";
	    print "$robcom\n";
	    my $tmp = `$robcom`;
	}
	
	open(IN, "<trashnorm.txt");
	my @lines = <IN>;
	close IN;
	
	for (my $i=0; $i<scalar(@lines); $i++) {
	    if($lines[$i] =~ /mean\[(.*)\]\s+stDev\[(.*)\]/i) {
		my $thisnorm = $1;
		my $thisstd = $2;
		$normval += $thisnorm;
		$stdval += $thisstd;	
	    }	
	}
	
	$normval = $normval/scalar(@lines);
	$stdval = $stdval/scalar(@lines);
    }

    print "\n";
    print "normalization value = $normval\n";
    print "standard deviation  = $stdval\n";
    print "\n";

    foreach my $boxfile (@files) {
	my $fixfile = $boxfile;
	$fixfile =~ s/.*\///;
	$fixfile =~ s/\.pif/\_fix\.pif/i;
	my $normfile = $fixfile;
	$normfile =~ s/\.pif/\_norm\.pif/i;

	my $robcom = "robem -nogui -pre $boxfile -rmblem 0.0";
	$robcom .= " -outfile $fixfile";
	print "$robcom\n";
	my $tmp = `$robcom`;
	$robcom = "robem -nogui -pre $fixfile -lingrad -norm ";
	$robcom .= "$normval $stdval -outfile $normfile";
	print "$robcom\n";
	$tmp = `$robcom`;
	
	if(-e $fixfile) {
	    unlink($fixfile);
	}
    }

    unlink('trashnorm.txt');

    if($userspec =~ /y/) {
	unless(defined $ARGV[4]) {
	    print "\n";
	    print "----- Command line equivalent -----\n";
	    print "autopp f \'$list\' $userspec $normval $stdval\n";
	}
    } else {
	unless(defined $ARGV[2]) {
	    print "\n";
	    print "----- Command line equivalent -----\n";
	    print "autopp f \'$list\' $userspec\n";
	}
    }
    
}



sub optionG {

    print << "EOF";

Resize the boxes in one or more BCRD files.

New files will be written to a new directory named ADJBCRD

EOF

    print "Enter list of BCRD files to be processed: ";
    my $list = get_autopp_arg(1);

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No boxed image files were found - exiting program\n";
	exit;
    }

    print "Enter new box radius: ";
    my $boxrad = get_autopp_arg(2);
    $boxrad = sprintf ("%d", $boxrad);

    print "\n";

    unless(-e "ADJBCRD") {
	mkdir ("ADJBCRD", 0755) || die "Cannot make the directory ADJBCRD\n";
    }

    foreach my $bcrdfile (@files) {
	open(IN, $bcrdfile);
	my $outfile = $bcrdfile;
	$outfile =~ s/.*\///;
	open(OUT, ">ADJBCRD/$outfile");
	print "Processing $bcrdfile\n";

	while(<IN>) {
	    if($. == 4 || $. == 5) {
		print OUT "$boxrad\n";
	    } else {
		print OUT;
	    }
	}
	
	close(IN);
	close(OUT);
    }

    unless(defined $ARGV[2]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp g \'$list\' $boxrad\n";
    }

}



sub optionI {

    print << "EOF";
    
Do  a global replace of a string in a set of text files.

Original files will be overwritten - if in doubt create a backup. You
must have both read and write permissions for the files that you want
to modify. All files that are not ordinary text files will be ignored.

EOF
    
    print "Enter list of text files to be processed: ";
    my $list = get_autopp_arg(1);

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No text files were found - exiting program\n";
	exit;
    }

    print "Enter string to be replaced: ";
    my $oldstr = get_autopp_arg(2);

    print "Enter replacement string: ";
    my $newstr = get_autopp_arg(3);

    print "\n";
    
# Escape special characters (.*?[]) in $oldstr    
    $oldstr =~ s/\./\\./g;
    $oldstr =~ s/\+/\\+/g;
    $oldstr =~ s/\?/\\?/g;
    $oldstr =~ s/\[/\\[/g;
    $oldstr =~ s/\]/\\]/g;

    foreach my $file (@files) {
	if(! -r $file) {
	    print "$file is not readable by user or group\n";
	    next;
	} elsif (! -w $file) {
	    print "$file is not writeable by user or group\n";
	    next;
	} elsif (! -T $file) {
	    print "$file is not a plain text file\n";
	    next;
	} else {
	    print "Modifying $file\n";
	}
	
	# Read all records in file into array @lines
	open(FH, "+<$file");
	my @lines = <FH>;
	
	# Truncate file
	seek(FH, 0, 0);
	truncate(FH, 0);
	
	# Write modified records to file
	foreach my $line (@lines) {
	    $line =~ s/$oldstr/$newstr/;
	    print FH $line;
	}
	close(FH);
    }

    unless(defined $ARGV[3]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp i \'$list\' $oldstr $newstr\n";
    }

}



sub optionJ {

    print << "EOF";

Scale and/or shift the origins in a set of particle parameter
files. This option is particularly useful if a reconstruction had been
started with binned data and will be continued with unbinned data.

Note that if both shifting and scaling coordinates, shifting is done
BEFORE scaling.

New files will be written to a new directory named ADJDAT

EOF

    print "Enter list of particle parameter files to be processed: ";
    my $list = get_autopp_arg(1);

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No boxed image files were found - exiting program\n";
	exit;
    }

    print "Enter the scale factor [1]: ";
    my $scale = get_autopp_arg(2);
    if(!$scale) {
	$scale = 1;
    }

    print "Enter the x-shift in pixels [0]: ";
    my $xshift = get_autopp_arg(3);
    if(!$xshift) {
	$xshift = 0;
    }

    print "Enter the y-shift in pixels [0]: ";
    my $yshift = get_autopp_arg(4);
    if(!$yshift) {
	$yshift = 0;
    }

    print "\n";

    unless(-e "ADJDAT") {
	mkdir ("ADJDAT", 0755) || die "Cannot make the directory ADJDAT\n";
    }

    foreach my $datfile (@files) {
	open(IN, $datfile);
	my $outfile = $datfile;
	$outfile =~ s/.*\///;
	open(OUT, ">ADJDAT/$outfile");
	print "Processing file $datfile\n";
	
	while(<IN>) {
	    my $line = $_;
	    $line =~ s/^\s+//;
	    $line =~ s/\s+$//;

	    if($. <= 2) {
		print OUT "$line\n";
		next;
	    }

	    my @p = split(/,/, $line);
	    $p[4] += $xshift;
	    $p[5] += $yshift;
	    $p[4] *= $scale;
	    $p[5] *= $scale;
	    $p[4] = sprintf ("%.3f", $p[4]);
	    $p[5] = sprintf ("%.3f", $p[5]);
	    $line = join(', ', @p);
	    
	    print OUT "$line\n";
	}
	
	close(IN);
	close(OUT);
    }

    unless(defined $ARGV[4]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp j \'$list\' $scale $xshift $yshift\n";
    }

}



sub optionK {

    print << "EOF";

Use origins from particle parameter files to recenter box coordinates
in BCRD files and generate new particle parameter files (DAT files)
with origins at box center.

The user specifies a list of BCRD files and a list of DAT files. For
every DAT file that contains the same base name as the BCRD file, a
new pair of files will be generated. Depending on option chosen, new
files will be moved to NEW[BCRD|DAT] or old files will be moved to
OLD[BCRD|DAT] directories.

NOTE 1: The records in the particle parameter file must correspond to
the records in the bcrd file. Any deletion or reordering of the
records will cause this option to either fail or produce incorrect
results.

NOTE 2: This option will not work for concatenated particle parameter
files. Each particle parameter file must correspond to a single
micrograph.

NOTE 3: in case of disagreement between number and/or content of bcrd
and dat files, options 1 and 2 will provide conceptually different results.
With option 1, the new directories NEWBCRD and NEWDAT will only contain
the files for which  the correction was possible.
With option2, the original data will be all present in the OLDBCRD and
OLDDAT directories, but in the original directories there will a mixture 
of corrected and not corrected files: in this way you can rebox the particles
whose coordinates could be recentered, but you will keep also the stacks whose
center could not be refined, to use together for further processing.

--------------------------------------------------------

EOF

    my $flag = 0;
    print "Enter list of BCRD files to be processed: ";
    my $blist = get_autopp_arg(1);
    $blist =~ s/\s+$//;

    # Expand $list into array of files
    my @bfiles = glob("$blist");
    
    # Test to make sure that files were found
    if(@bfiles <= 0) {
	print "No BCRD files were found - exiting program\n";
	exit;
    }

    print "Enter list of particle parameter files to be considered: ";
    my $plist = get_autopp_arg(2);

    # Expand $list into array of files
    my @pfiles = glob("$plist");
    
    # Test to make sure that files were found
    if(@pfiles <= 0) {
	print "No particle parameter files were found - exiting program\n";
	exit;
    }

    print "Enter box radius: ";
    my $boxrad = get_autopp_arg(3);

    print "Choose 1 = new files -------> NEWBCRD NEWDAT\n";
    print "       2 = original files --> OLDBCRD OLDDAT [1]: ";
    my $choice = get_autopp_arg(4);
    if(!$choice) {
	$choice = 1;
    }

    print "\n";

    # make output directories
    if($choice == 1) {
	unless(-e "NEWBCRD") {
	    mkdir("NEWBCRD", 0755) || die "Can't make the directory NEWBCRD\n";
	}
	
	unless(-e "NEWDAT") {
	    mkdir("NEWDAT", 0755) || die "Can't make the directory NEWDAT\n";
	}
    } else {
	unless(-e "OLDBCRD") {
	    mkdir("OLDBCRD", 0755) || die "Can't make the directory OLDBCRD\n";
	}
	
	unless(-e "OLDDAT") {
	    mkdir("OLDDAT", 0755) || die "Can't make the directory OLDDAT\n";
	}
    }

    # as a safety measure, copy all the selected  dat files tothe OLDDAT directory
    if($choice == 2) {
        foreach my $datfile (@pfiles) {
	        my $olddat = $datfile;
	        $olddat =~ s/.*\///;
	        copy("$datfile","OLDDAT/$olddat") or die "Copy failed: $!";
        }
    }
    
    my %bhash;

    foreach my $bcrdfile (@bfiles) {
	my $bcrdbase = $bcrdfile;
	$bcrdbase =~ s/.*\///;
	$bcrdbase =~ s/\.bcrd//;
	my @list = ();

	foreach my $datfile (@pfiles) {
	    if($datfile =~ /$bcrdbase/) {
		push(@list, $datfile);
	    }
	}

	if(scalar(@list) > 1) {
	    $flag = 1;
	    print "Warning: Found multiple dat files for $bcrdfile\n";
	    my $lenshort = 10000;
	    foreach my $x (@list) {
		print "   $x\n";
		if(length($x) < $lenshort) {
		    $bhash{$bcrdfile} = $x;
		    $lenshort = length($x);
		}
	    }
	    print "   Using dat file with shortest name: $bhash{$bcrdfile}\n";
	}

	if(scalar(@list) == 1) {
	    $bhash{$bcrdfile} = $list[0];
	} elsif (scalar(@list) == 0) {
	    print "WARNING: No corresponding dat file for $bcrdfile\n";
        # make sure that in the OLD directories all files are stored
	    if($choice == 2) {
	        my $oldbcrd = $bcrdfile;
	        $oldbcrd =~ s/.*\///;
	        copy("$bcrdfile","OLDBCRD/$oldbcrd") or die "Copy failed: $!";
	    }
	}
    }

    print "\n";

    foreach my $bcrdfile (sort keys %bhash) {
	my $datfile = $bhash{$bcrdfile};

	print "Processing $bcrdfile\n";
	print "           $datfile\n";

	# Load records of bcrd and dat files into arrays
	open(OLDB, $bcrdfile);
	my @bcrdarr = <OLDB>;
	close(OLDB);

	open(OLDD, $datfile);
	my @datarr  = <OLDD>;
	close(OLDD);

	# Make sure that bcrd and dat files are compatible
	if(scalar(@datarr) - 2 != scalar(@bcrdarr) - 5) {
	    print "           WARNING: Record counts do not agree - SKIPPING\n\n";
        # make sure that in the OLD directory all bcrd files are stored
        # dat files have been already preemptively copied
	    if($choice == 2) {
	        my $oldbcrd = $bcrdfile;
	        $oldbcrd =~ s/.*\///;
	        copy("$bcrdfile","OLDBCRD/$oldbcrd") or die "Copy failed: $!";
	    }
	    next;
	}

	my ($newbcrdfile, $newdatfile);
	if($choice == 1) {
	    $newbcrdfile = $bcrdfile;
	    $newbcrdfile =~ s/.*\///;
	    $newbcrdfile = "NEWBCRD/$newbcrdfile";

	    $newdatfile = $datfile;
	    $newdatfile =~ s/.*\///;
	    $newdatfile = "NEWDAT/$newdatfile";
	} else {
	    my $oldbcrd = $bcrdfile;
	    $oldbcrd =~ s/.*\///;
	    open(OLDB, ">OLDBCRD/$oldbcrd");
	    print OLDB @bcrdarr;
	    close(OLDB);

	    my $olddat = $datfile;
	    $olddat =~ s/.*\///;
	    open(OLDD, ">OLDDAT/${olddat}");
	    print OLDD @datarr;
	    close(OLDD);

	    $newbcrdfile = $bcrdfile;
	    $newdatfile  = $datfile;
	}

	open(NEWB, ">$newbcrdfile");
	open(NEWD, ">$newdatfile");

	for (my $i=0; $i<5; $i++) {
	    print NEWB "$bcrdarr[$i]";
	}
	
	print NEWD "$datarr[0]";
	print NEWD "$datarr[1]";

	for (my $j=2; $j<scalar(@datarr); $j++) {
	    # Get original box coordinates from bcrd file
	    $bcrdarr[$j+3] =~ s/^\s+//;
	    $bcrdarr[$j+3] =~ s/\s+$//;
	    my ($xcord, $ycord) = split /\s+/, $bcrdarr[$j+3];

	    # Get particle origins from particle parameter files
	    $datarr[$j] =~ s/^\s+//;
	    $datarr[$j] =~ s/\s+$//;
	    $datarr[$j] =~ s/\,/ /g;
	    my @p = split /\s+/, $datarr[$j];
	    
	    # Truncate origins to integer values
	    my $xcent_int = int $p[4];
	    my $ycent_int = int $p[5];
	    
	    # Calculate new bcrd coordinates
	    $xcord = $xcord + ($xcent_int - $boxrad);
	    $ycord = $ycord - ($ycent_int - $boxrad);

	    # Update particle origins
	    $p[4] = $boxrad + ($p[4] - $xcent_int);
	    $p[5] = $boxrad + ($p[5] - $ycent_int);
	    

	    # Write out new dat file 
	    printf NEWD "%5d, %8.3f, %8.3f, %8.3f, " ,$p[0],$p[1],$p[2],$p[3];
	    printf NEWD "%7.3f, %7.3f, ", $p[4],$p[5];
	    printf NEWD "%5.3f, %5.3f, %5.3f, %5.3f", $p[6],$p[7],$p[8],$p[9];
	    for (my $i=10; $i<20; $i++) {
		if(defined $p[$i]) {
		    print NEWD ", $p[$i]";
		}
	    }
	    print NEWD "\n";

	    # Write out new bcrd file
	    print NEWB "$xcord $ycord\n";
	}

	close(NEWD);
	close(NEWB);
    }

    if($flag > 0) {
	print "\n";
	print "WARNING - One or more bcrd files had possible matches\n";
	print "with multiple dat files. Please check autopp output to\n";
	print "make sure that bcrd/dat files were paired correctly\n";
    }


    unless(defined $ARGV[4]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp k \'$blist\' \'$plist\' $boxrad $choice\n";
    }

}



sub optionM {

    print << "EOF";

Autobox PIF micrographs using robem.

Box coordinate file will have same name as micrograph except with .pif
extension replaced by .bcrd

EOF

    print "Enter list of micrographs to be processed: ";
    my $list = get_autopp_arg(1);

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No BCRD files were found - exiting program\n";
	exit;
    }

    print "Enter emma radius: ";
    my $emma = get_autopp_arg(2);

    print "Enter box radius: ";
    my $box = get_autopp_arg(3);

    print "Enter feather radius: ";
    my $feather = get_autopp_arg(4);

    print "Enter thinning layers: ";
    my $thin = get_autopp_arg(5);

    print "Enter dilation layers: ";
    my $dilat = get_autopp_arg(6);

    print "Enter number of refinement interations: ";
    my $iter = get_autopp_arg(7);

    print "Enter threshold for discarding particles (0=keep all; 1=discard all)\n: ";
    my $thresh = get_autopp_arg(8);

    print "\n";

    foreach my $piffile (@files) {
	my $tmplink = $piffile;
	$tmplink =~ s/.*\///;
	unless (-e $tmplink) {
	    # Kludge - setting symlink since we can't control
	    # name of output bcrd file. Ensures output written to cwd
	    symlink $piffile, $tmplink;
	    my $robcom = "robem -nogui -pre $tmplink -autobox gaussian ";
	    $robcom .= "$emma $box $feather $thin $iter $dilat $thresh";
	    print "$robcom\n";
	    my $tmp = `$robcom`;
	    unlink($tmplink);
	} else {
	    my $robcom = "robem -nogui -pre $piffile -autobox gaussian ";
	    $robcom .= "$emma $box $feather $thin $iter $dilat $thresh";
	    print "$robcom\n";
	    my $tmp = `$robcom`;
	}
	    
    }

    unless(defined $ARGV[7]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp m \'$list\' $emma $box $feather $thin $dilat $iter $thresh\n";
    }

}



sub optionO {

    print << "EOF";

Apodize a set of boxed images. The box radius will be extracted from
the file header and the user will be asked to specify the width of the
zone over which the Gaussian falloff is applied.

The name of the apodized image file will be the same as the original
file except with _apod appended to the file prefix. For example

   file1_box.pif --> file1_box_apod.pif

EOF

    print "Enter list of boxed image files to be apodized: ";
    my $list = get_autopp_arg(1);

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No boxed image files were found - exiting program\n";
	exit;
    }

    print "Enter the width of the apodization zone [6]: ";
    my $apodwidth = get_autopp_arg(2);
    if(!$apodwidth) {
	$apodwidth = 6;
    }

    print "\n";

    foreach my $boxfile (@files) {
	my $outrad = get_boxrad($boxfile);
	my $inrad = $outrad - $apodwidth;

	my $outfile = $boxfile;
	$outfile =~ s/.*\///;
	$outfile =~ s/\.pif/\_apod\.pif/;

	my $robcom = "robem -nogui -pre $boxfile -rebox gaussian ";
	$robcom .= "$inrad $outrad -outfile $outfile";

	print "$robcom\n";
	my $tmp = `$robcom`;
    }
    
    unless(defined $ARGV[2]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp o \'$list\' $apodwidth\n";
    }

}



sub optionQ {
    print "\nExiting autopp\n";
}



sub optionR {
    
    print << "EOF";
    
Convert one or more PIF files to MRC format. 

The new MRC files will have the same name as the corresponding PIF
files except with the .pif extension changed to .mrc

EOF

    print "Enter list of PIF files to be processed: ";
    my $list = get_autopp_arg(1);

    print"\n";

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No PIF files were found - exiting program\n";
	exit;
    }

    foreach my $piffile (@files) {
	my $mrcfile = $piffile;
	$mrcfile =~ s/.*\///;
	$mrcfile =~ s/\.pif$/.mrc/;
	print "Converting $piffile to $mrcfile\n";
	print "pif2mrc $piffile $mrcfile\n";
	my $tmp = `pif2mrc $piffile $mrcfile`;
    }

    unless(defined $ARGV[1]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp r \'$list\'\n";
    }

}



sub optionT {

    print << "EOF";
    
Truncate one or more PIF maps.

The new PIF map will have the same name as the original map except
with the string -cut added before the .pif extension

EOF

    print "Enter list of PIF files to be processed: ";
    my $list = get_autopp_arg(1);

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No PIF files were found - exiting program\n";
	exit;
    }

    print "Enter starting x coordinate: ";
    my $begx = get_autopp_arg(2);

    print "Enter ending x coordinate: ";
    my $endx = get_autopp_arg(3);

    print "Enter starting y coordinate: ";
    my $begy = get_autopp_arg(4);

    print "Enter ending y coordinate: ";
    my $endy = get_autopp_arg(5);

    print "Enter starting z coordinate: ";
    my $begz = get_autopp_arg(6);

    print "Enter ending z coordinate: ";
    my $endz = get_autopp_arg(7);

    print "\n";
    
    foreach my $infile (@files) {
	my $outfile = $infile;
	$outfile =~ s/.*\///;
	$outfile =~ s/\.pif/-cut.pif/;
	print "Truncating $infile\n";
	my $robcom = "robem -nogui -pre $infile -TRUNC $begx $endx ";
	$robcom .= "$begy $endy $begz $endz -OUTFILE $outfile";
	print "$robcom\n";
	my $tmp = `$robcom`;
    }

    unless(defined $ARGV[7]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp t \'$list\' $begx $endx $begy $endy $begz $endz\n";
    }

}



sub optionW {
    
    print << "EOF";
    
Mask a PIF map using one or more PDB files

EOF

    print "Enter list of PDB files to be used for masking: ";
    my $list = get_autopp_arg(1);

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No TIFF files were found - exiting program\n";
	exit;
    }

    print "Enter border size: ";
    my $border = get_autopp_arg(2);

    print "Enter PIF map: ";
    my $piffile = get_autopp_arg(3);

    print "Enter pixel size: ";
    my $pixel = get_autopp_arg(4);

    print "Enter 1 to mask, 2 to delete density: ";
    my $opt = get_autopp_arg(5);

    print "\n";

    foreach my $pdbfile (@files) {
	print "Masking $piffile using $pdbfile\n";
	my $outfile = $pdbfile;
	$outfile =~ s/.*\///;
	$outfile =~ s/\.pdb$/.pif/;
	my $cmd = "pifmapmask $border $piffile $outfile $pixel $pdbfile $opt";
	print "$cmd\n";
	my $tmp = `$cmd`;
    }

    unless(defined $ARGV[5]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp w \'$list\' $border $piffile $pixel $opt\n";
    }

}



sub optionX {
    
    print << "EOF";
    
Convert one or more DM3 micrograps to PIF format.

The new PIF files will have the same name as the corresponding DM3
files except with the .dm3 extension changed to .pif

EOF

    print "Enter list of DM3 files to be processed: ";
    my $list = get_autopp_arg(1);

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No DM3 files were found - exiting program\n";
	exit;
    }

    print "Enter bin factor: ";
    my $bin = get_autopp_arg(2);

    print"\n";

    foreach my $dm3file (@files) {
	my $piffile = $dm3file;
	$piffile =~ s/.*\///;
	$piffile =~ s/\.dm3$/.pif/;
	print "Convert $dm3file to $piffile\n";
	print "dm32pif $dm3file $piffile $bin\n";
	my $tmp = `dm32pif $dm3file $piffile $bin`;
    }

    unless(defined $ARGV[1]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp x \'$list\' $bin\n";
    }

}



sub optionY {
    
    print << "EOF";
    
Convert one or more MRC files to PIF format.
It works only on 2D images, that is either micrographs or stacks of particles.
If more than one image is in the file, the size is forced to be odd.

The new PIF files will have the same name as the corresponding MRC
files except with the .mrc extension changed to .pif

EOF

    print "Enter list of MRC files to be processed: ";
    my $list = get_autopp_arg(1);

    print"\n";

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No MRC files were found - exiting program\n";
	exit;
    }

    foreach my $mrcfile (@files) {
	my $piffile = $mrcfile;
	$piffile =~ s/.*\///;
	$piffile =~ s/\.mrc$/.pif/;
	print "Convert $mrcfile to $piffile\n";
	print "echo '2D' | mrc2pif $mrcfile $piffile\n";
	my $tmp = `echo '2D' | mrc2pif $mrcfile $piffile`;
    }

    unless(defined $ARGV[1]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp y \'$list\'\n";
    }

}



sub optionZ {
    
    print << "EOF";
    
Convert one or more PIF maps to CCP4 format.

The new CCP4 files will have the same name as the corresponding PIF
files except with the .pif extension changed to .ccp4

EOF

    print "Enter list of PIF files to be processed: ";
    my $list = get_autopp_arg(1);

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No PIF files were found - exiting program\n";
	exit;
    }

    print "Enter pixel size in Angstroms [1.0]: ";
    my $pixel = get_autopp_arg(2);
    if(!$pixel) {
	$pixel = 1.0;
    }

    print "Invert contrast y/n? [y]: ";
    my $invert = get_autopp_arg(3);
    $invert =~ s/y.*/y/i;
    if(!$invert) {
	$invert = 'y';
    }

    print "\n";

    foreach my $piffile (@files) {
	my $ccp4file = $piffile;
	$ccp4file =~ s/.*\///;
	$ccp4file =~ s/\.pif$/.ccp4/;
	print "Converting $piffile to $ccp4file\n";
	print "pif2ccp4 $piffile $ccp4file $pixel $invert\n";
	my $tmp = `pif2ccp4 $piffile $ccp4file $pixel $invert`;
    }

    unless(defined $ARGV[3]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp z \'$list\' $pixel $invert\n";
    }

}



sub option1 {

    print << "EOF";

Rename a set of files in a consistent manner. The user will be
prompted to provide the portion of the filename that should be
replaced. For example

  oldstr = ABC
  newstr = DEF

  fileABC_1 --> fileDEF_1
  fileABC_2 --> fileDEF_2

If in doubt, be sure to make a backup of your files first.

EOF

    print "Enter list of files to be renamed: ";
    my $list = get_autopp_arg(1);
    
    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No files were found - exiting program\n";
	exit;
    }
    
    print "Enter old string to be replaced in file names: ";
    my $oldstr = get_autopp_arg(2);
    
    print "Enter replacement for old string: ";
    my $newstr = get_autopp_arg(3);

    printf "\n";

    foreach my $oldfile (@files) {
	my $newfile = $oldfile;
	$newfile =~ s/$oldstr/$newstr/;
	printf "Renaming $oldfile to $newfile\n";
	rename $oldfile, $newfile;
    }

    unless(defined $ARGV[3]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp 1 \'$list\' $oldstr $newstr\n";
    }


}



sub option2 {
    
    print << "EOF";

Generate boxed image files that are tiles covering a PIF 
micrograph. For 4kx4k or smaller images, overlapping boxes are
used. For larger images, tiled boxes are used.

Boxed image files will have same name as micrographs except with .pif
extension replaced by _tile.pif.

The boxed images will not be apodized and existing files will be overwritten.

EOF

    print "Enter list of PIF files to be processed: ";
    my $list = get_autopp_arg(1);

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
		print "No files were found - exiting program\n";
		exit;
    }
    
    print "Enter box radius in pixels [192]: ";
    my $boxrad = get_autopp_arg(2);
    if(!$boxrad) {
		$boxrad = 192;
    }
    
    my $boxdim = 2*$boxrad + 1;

    print "\n";
    
    foreach my $micrograph (@files) {
		my $bcrdfile = $micrograph;
		$bcrdfile =~ s/.*\///;
		$bcrdfile =~ s/\.pif$/\_tile\.bcrd/i;
		my ($nx, $ny) = get_dimensions($micrograph);

		my $step;
		if($nx <= 4096 && $ny <= 4096) {
		    $step = $boxrad;
		} else {
			$step = $boxdim;
		}
	    
		my $nxmax = int(($nx - 20)/$step) - 1;
		my $nymax = int(($ny - 20)/$step) - 1;

		my @coor = ();
		for (my $i=1; $i<=$nxmax; $i++) {
		    for (my $j=1; $j<=$nymax; $j++) {
				my $x = 10 + $step*$i;
				my $y = 10 + $step*$j;
				my $xy = "$x $y";
				push(@coor, $xy);
			}
		}

		my $boxcount = scalar(@coor);

		open(BCRD, ">$bcrdfile");
		print BCRD "BCRD V2.00\n";
		print BCRD "0\n";
		print BCRD "$boxcount\n";
		print BCRD "$boxrad\n";
		print BCRD "$boxrad\n";
		foreach my $xy (@coor) {
	    	print BCRD "$xy\n";
		}
		close(BCRD);
	
		my $boxfile = $micrograph;
		$boxfile =~ s/.*\///;
		$boxfile =~ s/\.pif/\_tile\.pif/i;
		if(-e $boxfile) {
	    	unlink($boxfile);
		}
	    
		my $robcom = "robem -nogui -pre $micrograph -box $bcrdfile ";
		$robcom .= "gaussian 0 $boxrad 0 -outfile $boxfile";
		print "$robcom\n";
		my $tmp = `$robcom`;
		unlink($bcrdfile);		
	
	}

    unless(defined $ARGV[2]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp 2 \'$list\' $boxrad\n";
    }

}



sub option3 {
    
    print << "EOF";
    
Run the ctffind3 program on a set of micrographs.

Files with the .pif extension will automatically be converted to the
MRC format before attempting to run ctffind3

EOF

    my $currentdir = getcwd();
    chdir($currentdir);

    print "Enter list of micrographs to be processed: ";
    my $list = get_autopp_arg(1);

    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No micrographs were found - exiting program\n";
	exit;
    }
    
    print "Enter number of processors for parallel run [1]: ";
    my $ncpus = get_autopp_arg(2);
    if(!$ncpus) {
	$ncpus=1;
    }
    $ENV{"NCPUS"} = $ncpus;

    print "Enter the spherical aberration (Cs) in mm [2.0]: ";
    my $cs = get_autopp_arg(3);
    if(!$cs) {
	$cs = '2.0';
    }
    
    print "Enter the accelerating voltage in KV [300.0]: ";
    my $kv = get_autopp_arg(4);
    if(!$kv) {
	$kv = 300.0;
    }
    my $v = $kv * 1000;

    print "Enter minimum value for defocus search in um [0.9]: ";
    my $dfmin = get_autopp_arg(5);
    if(!$dfmin) {
	$dfmin = 0.9;
    }

    print "Enter maximum value for defocus search in um [4.0]: ";
    my $dfmax = get_autopp_arg(6);
    if(!$dfmax) {
	$dfmax = 4.0;
    }

    print "Enter step size for defocus search in um [0.05]: ";
    my $dfstep = get_autopp_arg(7);
    if(!$dfstep) {
	$dfstep = 0.05;
    }

    print "Enter the amplitude contrast percentage [0.07]: ";
    my $Famp = get_autopp_arg(8);
    if(!$Famp) {
	$Famp = 0.07;
    }

    my ($mag, $scan);
    
    printf "\n  If pixel size set to zero or not provided, you will\n";
    printf "  be prompted to enter magnification and scan step size\n\n";
    printf "Enter pixel size in Angstroms [0]: ";
    my $pixel = get_autopp_arg(9);
    if(!$pixel) {
	$pixel = "0";
    }

    print "Do you want to keep the file with the power spectrum? (0=no; 1=yes) [0]: ";
    my $keeppow = get_autopp_arg(10);
    if(!$keeppow) {
	$keeppow = 0;
    }

    my $astr = "autopp 3 \'$list\' $ncpus $cs $kv $dfmin $dfmax ";
    $astr .= "$dfstep $Famp $pixel $keeppow";

    $dfmin *= 10000;
    $dfmax *= 10000;
    $dfstep *= 10000;

    if($pixel) {
	$mag = 33000;
	$scan = ($pixel * $mag) / 10000;
    } else {
	print "Enter the magnification [33000.00]: ";
	$mag = get_autopp_arg(10);
	if(!$mag) {
	    $mag = 33000.0;
	}
	
	print "Enter the scan step size in um [7.0]: ";
	$scan = get_autopp_arg(11);
	if(!$scan) {
	    $scan = 7.0;
	}

	$astr .= " $mag $scan";

	$pixel = $scan*10000 / $mag ;
	$pixel = sprintf ("%.3f", $pixel);
    }

    my $pave = 2;
    my $box = 256;
    my $resmin = 100.0;
    my $resmax = $pixel * 2;
    
    foreach my $imagefile (@files) {
        print "\nProcessing $imagefile\n";

	my $numimg = get_numimages($imagefile);
	if($numimg > 1) {
	    print "  #images ($numimg) is greater than one - cttfind 3 only\n";
	    print "  operates on micrographs, not boxed image files\n";
	    next;
	}

	my $outfile = $imagefile;
	my $logfile = $imagefile;
	$outfile =~ s/.*\///;
	$logfile =~ s/.*\///;

	$outfile =~ s/\.pif$/.pow/;
	$outfile =~ s/\.mrc$/.pow/;
	$logfile =~ s/\.pif$/.ctf/;
	$logfile =~ s/\.mrc$/.ctf/;

	my $mrcfile = $imagefile;

	if($imagefile =~ /\.pif$/) {
	    $mrcfile =~ s/\.pif/.mrc/;
	    if(-e $mrcfile) {
		print "Corresponding MRC file already exists for $imagefile\n";
	    } else {
		print "Converting $imagefile to MRC\n";
		$mrcfile =~ s/.*\///;
		my $tmpcom = `pif2mrc $imagefile $mrcfile`;
	    }
	}
	
	my $exefile = "ctf.com";
	open RUN, ">$exefile";
	print "Running ctffind3_a3dm_mp on $mrcfile\n";
	print RUN "#!/bin/bash\n";
	print RUN "ctffind3_a3dm_mp \<\< eof \> $logfile\n";
	print RUN "$mrcfile\n";
	print RUN "$outfile\n";
	print RUN "$cs, $kv, $Famp, $mag, $scan, $pave\n";
	print RUN "$box, $resmin, $resmax, $dfmin, $dfmax, $dfstep, 0.0, 0.0\n";
	print RUN "eof\n";
	close RUN;
	chmod 0755, $exefile;
	system("${currentdir}/${exefile}");
	if ($keeppow == 0){
	    unlink($outfile);
	}
	unlink($exefile);
	$mrcfile = $imagefile;
	$mrcfile =~ s/\.pif/.mrc/;
	$mrcfile =~ s/.*\///;
	if(-e $mrcfile) {
	    unlink($mrcfile);
	}
    } 
    
    unless(defined $ARGV[10]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "$astr\n";
    }

}



sub option4 {
    
    print << "EOF";

Generate a set of empty (header lines only) particle parameter files
from ctffind3 output files. Normally used with option3.

EOF

    print "Enter list of ctffind3 files to be processed: ";
    my $list = get_autopp_arg(1);
    
    # Expand $list into array of files
    my @files = glob("$list");
    
    # Test to make sure that files were found
    if(@files <= 0) {
	print "No files were found - exiting program\n";
	exit;
    }
    
    print "Generate parameter files forcing dfmaj=dfmin (y/n)? [n]: ";
    my $stig = get_autopp_arg(2);
    if(!$stig) {
	$stig='n';
    }
    $stig =~ s/\s//g;
    if($stig =~ /^y/i) {
	$stig = 'y';
    } else {
	$stig = 'n';
    }

    print "Warn me when estimated defocus levels below this value [0]: ";
    my $dflow = get_autopp_arg(3);
    if(!$dflow) {
	$dflow = 0.0;
    }

    print "Warn me when estimated defocus levels above this value [10]: ";
    my $dfhigh = get_autopp_arg(4);
    if(!$dfhigh) {
	$dfhigh = 10.0;
    }

    print "Warn me when ratio defocus major/minor above this value [1.1]: ";
    my $astigthresh = get_autopp_arg(5);
    if(!$astigthresh) {
	$astigthresh = 1.1;
    }
    
    print "Skip generation of parameter file if ctf estimates are beyond given ranges (y/n) [n]: ";
    my $skipoutofrange = get_autopp_arg(6);
    if(!$skipoutofrange) {
	$skipoutofrange = 'n';
    }
    $skipoutofrange =~ s/\s//g;
    if($skipoutofrange =~ /^y/i) {
	$skipoutofrange = 'y';
    } else {
	$skipoutofrange = 'n';
    }

    print "\n";

    my ($dfmajor, $dfminor, $ang, $pixel);
    my ($cs, $v, $kv, $Famp, $mag, $scan);

    foreach my $ctffile (@files) {

	my $datfile = $ctffile;
	$datfile =~ s/.*\///;
	$datfile =~ s/\.ctf/.dat_000/;
	$datfile =~ s/\.log/.dat_000/;
	my $imagefile = $ctffile;
	$imagefile =~ s/.*\///;
	$imagefile =~ s/\.ctf/.pif/;
	$imagefile =~ s/\.log/.pif/;
        print "Generating $datfile from $ctffile\n";

	my $nextline = 0;

	open(CTF, "$ctffile");
	while(<CTF>) {
	    if($nextline) {
		s/^\s+//;
		s/\s+$//;
		($cs, $kv, $Famp, $mag, $scan) = split;
		$v = $kv*1000;
		$pixel = $scan*10000 / $mag ;
		$pixel = sprintf ("%.3f", $pixel);
		$nextline = 0;
	    }

	    if (/CS\[mm\]/) {
		$nextline = 1;
	    }
	    
	    if(/Final Values/) {
		s/^\s+//;
		s/\s+$//;
		($dfmajor, $dfminor, $ang) = split;
		$dfmajor /= 10000;
		$dfminor /= 10000;
		$ang = -$ang;
		last;
	    }
	}
	close(CTF);
	
	my $dfavg = ($dfmajor + $dfminor)/2;
	
	my $wflag = 0;
	if($dfmajor < $dflow) {
	    print "  WARNING: defocus major ($dfmajor) lower than expected\n";
	    $wflag = 1;
	} elsif ($dfmajor < 0) {
	    print "  WARNING: defocus major ($dfmajor) negative\n";
	    $wflag = 1;
	} elsif($dfmajor > $dfhigh) {
	    print "  WARNING: defocus major ($dfmajor) higher than expected\n";
	    $wflag = 1;
	}

	if($dfminor < $dflow) {
	    print "  WARNING: defocus minor ($dfminor) lower than expected\n";
	    $wflag = 1;
	} elsif ($dfminor < 0) {
	    print "  WARNING: defocus minor ($dfminor) negative\n";
	    $wflag = 1;
	} elsif($dfminor > $dfhigh) {
	    print "  WARNING: defocus minor ($dfminor) higher than expected\n";
	    $wflag = 1;
	}

	if($dfminor != 0.0 && abs($dfmajor/$dfminor) > $astigthresh) {
	    print "  WARNING: defocus major/minor ";
	    print "abs($dfmajor/$dfminor) higher than expected\n";
	    $wflag = 1;
	}

	if($dfmajor != 0.0 && abs($dfminor/$dfmajor) > $astigthresh) {
	    print "  WARNING: defocus major/minor ";
	    print "abs($dfminor/$dfmajor) higher than expected\n";
	    $wflag = 1;
	}

	if($wflag) {
	    print "\n";
	}

	if($wflag && $skipoutofrange =~ /y/) {
	    print "#### Skipping generation of parameter file ####\n";
	    next;
	}

	open(DFH, ">$datfile");
	print DFH "../pif/${imagefile}\n";
	if($stig =~ /y/) {
	    print DFH "$pixel, 1, $v, $Famp, $dfavg, $dfavg, 0.0, $cs\n";
	} else {
	    print DFH "$pixel, 1, $v, $Famp, $dfmajor, $dfminor, $ang, $cs\n";
	}
	close(DFH);
    } 

    print "\n------------------------------------------------------------\n\n";
    print "Be sure that first line of each dat file points to a valid\n";
    print "boxed image file. By default we assume that the image file\n";
    print "will be located at ../pif relative to dat file and have same\n";
    print "prefix. If necessary, can fix using option I. For example\n";
    print "\n";
    print "     autopp i '*.dat_000' .pif _box_fix_norm_apod.pif\n";
    print "\n------------------------------------------------------------\n";
    
    unless(defined $ARGV[5]) {
	print "\n";
	print "----- Command line equivalent -----\n";
	print "autopp 4 \'$list\' $stig $dflow $dfhigh $astigthresh $skipoutofrange\n";
    }
    
}



sub get_autopp_arg {

    my ($argcnt) = @_;

    # Get argument for autopp from @cmdline if $cmdline[$argcnt] is
    # defined or from stdin otherwise. Note that this function makes
    # it possible to run autopp with a partial list of arguments. In
    # either case, strip trailing whitespace before returning.
    #
    # Input arguments:
    #   $argcnt - index for @cmdline being tested
    #
    # Return values:
    #   $str - string read from $cmdline[$argcnt] or stdin

    my $tmpstr;
    if(defined $cmdline[$argcnt]) {
	$tmpstr = $cmdline[$argcnt];
	print "$tmpstr\n";
    } else {
	$tmpstr = <STDIN>;
    }
    $tmpstr =~ s/\s+$//;

    return($tmpstr);
}
