#!/usr/bin/perl
use strict;
use warnings;

my ($h, $k) = @ARGV;

if(scalar(@ARGV) == 0) {
    print "\n";
    print "    Usage: capsid_generator h k\n";
    print "           capsid_generator --help\n";
    print "\n";
    print "    h k\n";
    print "    Icosahedral lattice parameters\n";
    print "\n";
    print "    --help\n";
    print "    print detailed instructions on using capsid_generator\n";
    print "    and Chimera to generate figures showing the locations of\n";
    print "    lattice points (capsomers)\n";
    exit(0);
}

if($h eq '--help') {
print << "EOF";

These instructions explain how to use capsid_generator.pl and Chimera
to generate figures showing the locations of capsomers in an
icosahedral lattice. Note that Chimera is not strictly required and
any visualization program that handles PDB files with non-standard
residues should be sufficient. Chimera instructions are provided since
we are most familiar with this program.

---------- Step 1: Generate PDB files ----------

Run capsid_geneartor.pl from the command line with two integer
arguments specifying the h and k parameters of the icosahedral
lattice. For example

  % capsid_generator.pl 7 5

For lattices that possess a handedness, the order of h and k is
important. If the lattice has an odd T numbers, two PDB files with
names of the following form will be generated.

  hH_kK_class[1|2|3].pdb

The classes (1, 2, or 3) refer to the classes defined in the following
paper and reflect that fact that an icosahedral lattice can be
decomposed into trisymmetrons and pentasymmetrons in two ways.

Sinkovits, R. S., and T. S. Baker (2010). A tale of two symmetrons:
Rules for construction of icosahedral capsids from trisymmetrons and
pentasymmetrons. J. Struct. Bio. 170:109-116.

For lattices with even T number, a single file of the following form
will be generated.

  hH_kK_class0.pdb


---------- Step 2: Generate image from PDB file ----------

After opening the PDB file in Chimera

(1) Render the atoms as spheres

  Actions > Atoms/Bonds > Sphere


(2) Set the VdW radius to an appropriate value. This generally has to
be done by trial and error to get the capsomers to touch without
overlapping too much.

  Favorites > Command Line
  On the command line enter: vdwdefine value


(3) Option 1: Assign same color to all capsomers

  Select > Residue > all nonstandard
  Actions > Color > (choose a color)


(3) Option 2: Color by type

The Residue menu will list 15 different types of nonstandard residues
for lattices with T=odd. For T-even, residues are restricted to types
GEN and VER.


  GEN = interior capsomsers in trisymmetrons
  TSE = trisymmetron edge capsomers
  VER = vertex (5-fold) capsomers
  PSA = capsomers in 1st pentasymmetron
  PSB = capsomers in 2nd pentasymmetron
       ...
  PSL = capsomers in 12th pentasymmetron

Select residue type and color as described above. For example

  Select > Residue > GEN
  Actions > Color > (choose a color)

EOF

exit(0)
}


my ($big, $small);
if($h >= $k) {
    $big = $h;
    $small = $k;
} else {
    $big = $k;
    $small = $h;
}    
    
if($small%2 == 0 && $big%2 == 0) { # Class 0
    my $flip;
    if($h > $k) {
	$flip = 1;
    } else {
	$flip = 0;
    }

    my $class = 0;
    my $psedge = 0;
    build_lattice($small, $big, $class, $psedge, $flip);
}

if($small%2 == 1) { # Class I
    my $flip;
    if($h > $k) {
	$flip = 1;
    } else {
	$flip = 0;
    }

    my $class = 1;
    my $psedge = ($small+1)/2;				   
    build_lattice($small, $big, $class, $psedge, $flip);
}

if($big%2 == 1) { # Class II
    my $flip;
    if($h > $k) {
	$flip = 1;
    } else {
	$flip = 0;
    }

    my $class = 2;
    my $psedge = ($big+1)/2;				   
    build_lattice($small, $big, $class, $psedge, $flip);
}


if(($small + $big)%2 == 1) { # Class III
    my $flip;
    if($h > $k) {
	$flip = 1;
    } else {
	$flip = 0;
    }

    my $class = 3;
    my $psedge = ($small+$big+1)/2;				   
    build_lattice($small, $big, $class, $psedge, $flip);
}



# ------------------------------------------------------------------


sub build_lattice {
    my ($h, $k, $class, $psedge, $flip) = @_;

    # This function builds a right-handed icosahedral lattice. It is
    # assumed that the condition $h <= $k holds. Left-handed lattices
    # are constructed by first building the corresponding right-handed
    # lattice and then applying the inversion operation to the
    # capsomer coordinates.

    my $outfile;
    if($flip) {
	$outfile = "h${k}_k${h}_class${class}.pdb";
    } else {
	$outfile = "h${h}_k${k}_class${class}.pdb";
    }

    open(FH, ">$outfile");

    # Distances from center of icos to symmetry points when the
    # edge length (distance between adjacent 5-folds) is 100
    my $r2fold = 80.902;
    my $r3fold = 75.576;
    my $r5fold = 95.106;

    # Define center of triangular face when left corner at origin
    my $xcen = 50.0;
    my $ycen = 100*sqrt(3)/6;

    # T-number and vectors for steps along h and k directions
    # defined for edge length equal to 1
    my $T = $h*$h + $k*$k + $h*$k;
    my $hx = (2*$k + $h) / (2.0*$T);
    my $hy = (sqrt(3.0)*$h) / (2.0*$T);
    my $kx = (2*$h + $k) / (2.0*$T);
    my $ky = -(sqrt(3.0)*$k) / (2.0*$T);
    
    # Set tolerance for calculations to avoid testing floating
    # point values for equality
    my $tol = 0.00001;
    
    # Define angle between 2-fold and 3-fold
    my $pi = atan2(1,1)*4;
    my $theta = (20.905/360)*2*$pi;
    my $ctheta = cos($theta);
    my $stheta = sin($theta);
    my $ttheta = $stheta/$ctheta;
    
    # Define locations of capsomers in ASU at non-symmetry points
    my (@xc, @yc, @zc, @ctype, @ispent, @istriedge);
    my $m=0;
    for(my $i=-($h+$k); $i<($h+$k); $i++) {
	for(my $j=-($h+$k); $j<($h+$k); $j++) {
	    
	    # Skip capsomser at 5-fold
	    if($i==0 && $j==0) {next;}
	    
	    # Skipping capsomer at 2-folds
	    if($i == $h/2 && $j == $k/2) {next;}
	    
	    # Skip capsomer at 3-fold. Using knowledge that capsomer
	    # located at 3-fold iff T divisible by three
	    if($T%3==0 && $i==($h+2*$k)/3 && $j==($h-$k)/3) {next;}
	    
	    my $x = $hx*$i + $kx*$j;
	    my $y = $hy*$i + $ky*$j;
	    my $z = 0.0;
	    
	    # Skip over capsomers outside of bounding box
	    if($x < -$tol || $x > 1+$tol ||
	       $y < -$tol || $y > sqrt(3)/2 + $tol) {
		next;
	    }
	    
	    # Skip capsomers above 60 deg line relative to (0,0)
	    if (atan2($y,$x) > $pi/6 + $tol) {
		next;
	    }
	    
	    # Skip capsomers AT or below 120 deg line relative to (1,0)
	    # avoids double counting when applying rotation operations
	    if (atan2($y,$x-1) < 5*$pi/6 + $tol) {
		next;
	    }
	    
	    # Skip capsomers along x-axis with x>1/2
	    # avoids double counting when applying rotation operations
	    if(abs($y) < $tol && $x > 0.5-$tol) {
		next;
	    }

	    # Identify capsomers that belong to
	    # pentasymmetrons. Setting ispent to 2 indicates that
	    # capsomer will be transformed from location near 5-fold
	    # at (1,0) to 5-fold at origin
	    if($i<$psedge && ($i+$j) < $psedge && $class > 0) {
		$ispent[$m] = 1;
	    } elsif (abs($i-$k) + abs($j-$h) < $psedge && $class > 0) {
		$ispent[$m] = 2;
	    } else {
		$ispent[$m] = 0;
	    }	    

	    # Identify capsomers on edges of trisymmetrons for classes
	    # 1 and 2. For Class 0 (even T number), there are no
	    # solutions involving just TS and PS. For class 3, the TS
	    # are isolated from each other and there is no need to
	    # mark the boundaries.
	    
	    $istriedge[$m] = 0; # Default - not on TS edge

	    if($class == 1) {
		if(($j<= $psedge && ($i+$j) <= $psedge) ||
		   abs($i-$k) + abs($j-$h) <= $psedge) {
		    unless($ispent[$m]) {
			$istriedge[$m] = 1;
		    }
		}
		if($j == $h-$psedge || $j == $h-$psedge+1) {
		    unless($ispent[$m]) {
			$istriedge[$m] = 1;
		    }
		}
	    }

	    if($class == 2) {
		if(($i <= $psedge && ($i+$j) <= $psedge) ||
		   abs($i-$k) + abs($j-$h) <= $psedge) {
		    unless($ispent[$m]) {
			$istriedge[$m] = 1;
		    }
		}
		if($i == $k-$psedge || $i == $k-$psedge+1) {
		    unless($ispent[$m]) {
			$istriedge[$m] = 1;
		    }
		}
	    }

	    # Scale (x,y) by 100 and set origin a distance $r3fold in
	    # front of 3-fold axis
	    $xc[$m] = 100*$x - $xcen;
	    $yc[$m] = 100*$y - $ycen;
	    $zc[$m] = $r3fold;
	    
	    $m++;
	}
    }
    
    my $count = 1;
    my $one = 1.0;
    
# Define pentasymmetron types
    my @pst_right =
	('PSA', 'PSL', 'PSK', 'PSI', 'PSC', 'PSD', 'PSA', 'PSH', 'PSI', 'PSL',
	 'PSE', 'PSB', 'PSA', 'PSF', 'PSL', 'PSH', 'PSG', 'PSC', 'PSA', 'PSK',
	 'PSH', 'PSF', 'PSD', 'PSE', 'PSA', 'PSI', 'PSF', 'PSK', 'PSB', 'PSG',
	 'PSJ', 'PSD', 'PSC', 'PSB', 'PSK', 'PSL', 'PSJ', 'PSG', 'PSB', 'PSD',
	 'PSF', 'PSI', 'PSJ', 'PSE', 'PSD', 'PSG', 'PSH', 'PSK', 'PSJ', 'PSC',
	 'PSG', 'PSE', 'PSL', 'PSF', 'PSJ', 'PSB', 'PSE', 'PSC', 'PSI', 'PSH');
    my @pst_left =
	('PSJ', 'PSC', 'PSD', 'PSB', 'PSL', 'PSK', 'PSJ', 'PSE', 'PSB', 'PSC',
	 'PSH', 'PSI', 'PSJ', 'PSG', 'PSC', 'PSE', 'PSF', 'PSL', 'PSJ', 'PSD',
	 'PSE', 'PSG', 'PSK', 'PSH', 'PSJ', 'PSB', 'PSG', 'PSD', 'PSI', 'PSF',
	 'PSA', 'PSK', 'PSL', 'PSI', 'PSD', 'PSC', 'PSA', 'PSF', 'PSI', 'PSK',
	 'PSG', 'PSB', 'PSA', 'PSH', 'PSK', 'PSF', 'PSE', 'PSD', 'PSA', 'PSL',
	 'PSF', 'PSH', 'PSC', 'PSG', 'PSA', 'PSI', 'PSH', 'PSL', 'PSB', 'PSE');    
    my @pst;
    if($flip) {
	@pst = @pst_left;
    } else {
	@pst = @pst_right;
    }


    # ----- Define locations of capsomers at 2-folds -----
    my (@c2x, @c2y, @c2z);
    ($c2x[0],  $c2y[0],  $c2z[0])  = (  0.000,   0.000,  80.902);
    ($c2x[1],  $c2y[1],  $c2z[1])  = (  0.000,   0.000, -80.902);
    ($c2x[2],  $c2y[2],  $c2z[2])  = (  0.000,  80.902,   0.000);
    ($c2x[3],  $c2y[3],  $c2z[3])  = (  0.000, -80.902,   0.000);
    ($c2x[4],  $c2y[4],  $c2z[4])  = ( 25.000,  40.451,  65.451);
    ($c2x[5],  $c2y[5],  $c2z[5])  = ( 25.000,  40.451, -65.451);
    ($c2x[6],  $c2y[6],  $c2z[6])  = ( 25.000, -40.451,  65.451);
    ($c2x[7],  $c2y[7],  $c2z[7])  = ( 25.000, -40.451, -65.451);
    ($c2x[8],  $c2y[8],  $c2z[8])  = ( 40.451,  65.451,  25.000);
    ($c2x[9],  $c2y[9],  $c2z[9])  = ( 40.451,  65.451, -25.000);
    ($c2x[10],  $c2y[10],  $c2z[10])  = ( 40.451, -65.451,  25.000);
    ($c2x[11],  $c2y[11],  $c2z[11])  = ( 40.451, -65.451, -25.000);
    ($c2x[12],  $c2y[12],  $c2z[12])  = ( 65.451,  25.000,  40.451);
    ($c2x[13],  $c2y[13],  $c2z[13])  = ( 65.451,  25.000, -40.451);
    ($c2x[14],  $c2y[14],  $c2z[14])  = ( 65.451, -25.000,  40.451);
    ($c2x[15],  $c2y[15],  $c2z[15])  = ( 65.451, -25.000, -40.451);
    ($c2x[16],  $c2y[16],  $c2z[16])  = ( 80.902,   0.000,   0.000);
    ($c2x[17],  $c2y[17],  $c2z[17])  = (-25.000,  40.451,  65.451);
    ($c2x[18],  $c2y[18],  $c2z[18])  = (-25.000,  40.451, -65.451);
    ($c2x[19],  $c2y[19],  $c2z[19])  = (-25.000, -40.451,  65.451);
    ($c2x[20],  $c2y[20],  $c2z[20])  = (-25.000, -40.451, -65.451);
    ($c2x[21],  $c2y[21],  $c2z[21])  = (-40.451,  65.451,  25.000);
    ($c2x[22],  $c2y[22],  $c2z[22])  = (-40.451,  65.451, -25.000);
    ($c2x[23],  $c2y[23],  $c2z[23])  = (-40.451, -65.451,  25.000);
    ($c2x[24],  $c2y[24],  $c2z[24])  = (-40.451, -65.451, -25.000);
    ($c2x[25],  $c2y[25],  $c2z[25])  = (-65.451,  25.000,  40.451);
    ($c2x[26],  $c2y[26],  $c2z[26])  = (-65.451,  25.000, -40.451);
    ($c2x[27],  $c2y[27],  $c2z[27])  = (-65.451, -25.000,  40.451);
    ($c2x[28],  $c2y[28],  $c2z[28])  = (-65.451, -25.000, -40.451);
    ($c2x[29],  $c2y[29],  $c2z[29])  = (-80.902,   0.000,   0.000);
    
    # ----- Define locations of capsomers at 3-folds -----
    my (@c3x, @c3y, @c3z);
    ($c3x[0],  $c3y[0],  $c3z[0])  = (  0.000,  26.967,  70.601);
    ($c3x[1],  $c3y[1],  $c3z[1])  = (  0.000,  26.967, -70.601);
    ($c3x[2],  $c3y[2],  $c3z[2])  = (  0.000, -26.967,  70.601);
    ($c3x[3],  $c3y[3],  $c3z[3])  = (  0.000, -26.967, -70.601);
    ($c3x[4],  $c3y[4],  $c3z[4])  = ( 26.967,  70.601,   0.000);
    ($c3x[5],  $c3y[5],  $c3z[5])  = ( 26.967, -70.601,   0.000);
    ($c3x[6],  $c3y[6],  $c3z[6])  = ( 43.634,  43.634,  43.634);
    ($c3x[7],  $c3y[7],  $c3z[7])  = ( 43.634,  43.634, -43.634);
    ($c3x[8],  $c3y[8],  $c3z[8])  = ( 43.634, -43.634,  43.634);
    ($c3x[9],  $c3y[9],  $c3z[9])  = ( 43.634, -43.634, -43.634);
    ($c3x[10], $c3y[10], $c3z[10]) = ( 70.601,   0.000,  26.967);
    ($c3x[11], $c3y[11], $c3z[11]) = ( 70.601,   0.000, -26.967);
    ($c3x[12], $c3y[12], $c3z[12]) = (-26.967,  70.601,   0.000);
    ($c3x[13], $c3y[13], $c3z[13]) = (-26.967, -70.601,   0.000);
    ($c3x[14], $c3y[14], $c3z[14]) = (-43.634,  43.634,  43.634);
    ($c3x[15], $c3y[15], $c3z[15]) = (-43.634,  43.634, -43.634);
    ($c3x[16], $c3y[16], $c3z[16]) = (-43.634, -43.634,  43.634);
    ($c3x[17], $c3y[17], $c3z[17]) = (-43.634, -43.634, -43.634);
    ($c3x[18], $c3y[18], $c3z[18]) = (-70.601,   0.000,  26.967);
    ($c3x[19], $c3y[19], $c3z[19]) = (-70.601,   0.000, -26.967);
    
    # ----- Define locations of capsomers at 5-folds -----
    my (@c5x, @c5y, @c5z);
    ($c5x[0],  $c5y[0],  $c5z[0])  = (0.000,    80.902,  50.000);
    ($c5x[1],  $c5y[1],  $c5z[1])  = (0.000,    80.902, -50.000);
    ($c5x[2],  $c5y[2],  $c5z[2])  = (0.000,   -80.902,  50.000);
    ($c5x[3],  $c5y[3],  $c5z[3])  = (0.000,   -80.902, -50.000);
    ($c5x[4],  $c5y[4],  $c5z[4])  = (50.000,    0.000,  80.902);
    ($c5x[5],  $c5y[5],  $c5z[5])  = (50.000,    0.000, -80.902);
    ($c5x[6],  $c5y[6],  $c5z[6])  = (80.902,   50.000,   0.000);
    ($c5x[7],  $c5y[7],  $c5z[7])  = (80.902,  -50.000,   0.000);
    ($c5x[8],  $c5y[8],  $c5z[8])  = (-50.000,   0.000,  80.902);
    ($c5x[9],  $c5y[9],  $c5z[9])  = (-50.000,   0.000, -80.902);
    ($c5x[10], $c5y[10], $c5z[10]) = (-80.902 , 50.000,   0.000);
    ($c5x[11], $c5y[11], $c5z[11]) = (-80.902, -50.000,   0.000);
    
    # Print remarks
    printf FH "REMARK 100  Non-standard amino acids used as follows\n";
    printf FH "REMARK 200  PSA-PSL pentasymmetron capsomers not at 5fold\n";
    printf FH "REMARK 300  TSE - capsomers on edge of trisymmetrons\n";
    printf FH "REMARK 400  GEN - other trisymmetron capsomers\n";
    printf FH "REMARK 500  VER - capsomers at 5fold vertices\n";

    # Print capsomers at 2-folds if h and k are both even
    if($h%2 == 0 && $k%2 == 0) {
	for(my $j=0; $j<30; $j++) {
	    if($flip) { # Invert coordinates for left handed lattice
		$c2x[$j] *= -1;
		$c2y[$j] *= -1; 
		$c2z[$j] *= -1;
	    }

	    my $ttype = "GEN";
	    printf FH "ATOM  ";
	    printf FH "%5d  CA  %s  %4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n",
	    $count, $ttype, $one, $c2x[$j], $c2y[$j], $c2z[$j],
	    $one, $one;
	    $count++;
	}
    }
    
    # Print capsomers at 3-folds if T divisible by 3
    if($T%3 == 0) {
	for(my $j=0; $j<20; $j++) {
	    if($flip) { # Invert coordinates for left handed lattice
		$c3x[$j] *= -1;
		$c3y[$j] *= -1; 
		$c3z[$j] *= -1;
	    }
	    
	    my $ttype = "GEN";
	    printf FH "ATOM  ";
	    printf FH "%5d  CA  %s  %4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n",
	    $count, $ttype, $one, $c3x[$j], $c3y[$j], $c3z[$j],
	    $one, $one;
	    $count++;
	}
    }
    
    # Always print capsomers at 5-folds
    for(my $j=0; $j<12; $j++) {
	if($flip) { # Invert coordinates for left handed lattice
	    $c5x[$j] *= -1;
	    $c5y[$j] *= -1; 
	    $c5z[$j] *= -1;
	}
	
	my $ttype = "VER";
	printf FH "ATOM  ";
	printf FH "%5d  CA  %s  %4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n",
	$count, $ttype, $one, $c5x[$j], $c5y[$j], $c5z[$j],
	$one, $one;
	$count++;
    }
    
    

    # ----- Handling capsomers at non-symmetry points -----
    my @rot;
    my $n = genrot(1,1,1,\@rot);
    
    my @xrot;
    $xrot[0][0] =  1; $xrot[0][1] =        0;  $xrot[0][2] =       0;
    $xrot[1][0] =  0; $xrot[1][1] =  $ctheta;  $xrot[1][2] = $stheta;
    $xrot[2][0] =  0; $xrot[2][1] = -$stheta;  $xrot[2][2] = $ctheta;
    
    my ($xt, $yt, $zt, $ttype, @vecin, @vectemp, @vecout);
    
    for (my $j=0; $j<$n; $j++) {
	for (my $i=0; $i<scalar(@xc); $i++) {

	    # Rotating coordinates of capsomers so that 2-fold
	    # is at (x,y,z) = (0,0,$r2fold)
	    @vecin = ($xc[$i], $yc[$i], $zc[$i]);
	    matvec(\@xrot, \@vecin, \@vectemp, 3);
	    
	    # Transforming coordinates of PS capsomers near "right" 
	    # 5-fold. Ensures that all PS capsomers in ASU are clustered
	    # about the same 5-fold axis
	    if($ispent[$i] == 2) {
		$vectemp[0] = -$vectemp[0];
		$vectemp[1] = -$vectemp[1];
	    }
	    
	    # Apply icosahedral rotation matrices
	    matvec(\@{$rot[$j]}, \@vectemp, \@vecout, 3);
	    ($xt, $yt, $zt) = @vecout;
	    
	    # Set "residue" type for capsomers
	    if($ispent[$i]) {
		$ttype = $pst[$j];
	    } elsif ($istriedge[$i]) {
		$ttype = "TSE";
	    } else {
		$ttype = "GEN";
	    }

	    # Invert coordinates for left handed lattice
	    if($flip) {
		$xt *= -1;
		$yt *= -1; 
		$zt *= -1;
	    }
	    
	    printf FH "ATOM  ";
	    printf FH "%5d  CA  $ttype  %4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n",
	    $count, $one, $xt, $yt, $zt,
	    $one, $one;
	    $count++;
	    
	}
    }    
    close(FH);
}

# --------------------------------------------------------------------

sub matmul {
    my ($aref, $bref, $cref, $n) = @_;

    for(my $i=0; $i<$n ; $i++) {
	for(my $j=0; $j<$n ; $j++) {
	    $$cref[$i][$j] = 0.0;
	    for(my $k=0; $k<$n ; $k++) {
		$$cref[$i][$j] += $$aref[$i][$k]*$$bref[$k][$j];
	    }
	}
    }

}    


sub matvec {
    my ($aref, $bref, $cref, $n) = @_;

    for(my $i=0; $i<$n ; $i++) {
	$$cref[$i] = 0.0;
	for(my $k=0; $k<$n ; $k++) {
	    $$cref[$i] += $$aref[$i][$k]*$$bref[$k];
	}
    }

}    
	

sub genrot {

    my ($use2, $use3, $use5, $rotref) = @_;

    my (@r2x, @r2y, @s3, @s5, @p3, @p3inv, @p5, @p5inv, @rot);

    for(my $i=0; $i<3; $i++) {
	for(my $j=0; $j<3; $j++) {
	    $p3[$i][$j] = 0.0;
	    $p3inv[$i][$j] = 0.0;
	    $p5[$i][$j] = 0.0;
	    $p5inv[$i][$j] = 0.0;
	    for(my $k=0; $k<2; $k++) {
		$r2x[$k][$i][$j] = 0.0;
		$r2y[$k][$i][$j] = 0.0;
	    }
	    for(my $k=0; $k<3; $k++) {
		$s3[$k][$i][$j] = 0.0;
	    }
	    for(my $k=0; $k<5; $k++) {
		$s5[$k][$i][$j] = 0.0;
	    }
	}
    }

    my $pi = atan2(1,1)*4;
    my $tpiov3 = $pi/1.5;
    my $tpiov5 = $pi/2.5;
    
    my $root5 = sqrt(5.0);
         
    my $xx3  = $root5/6.0;
    my $cos3 = sqrt(0.5+$xx3);
    my $sin3 = sqrt(0.5-$xx3);
    
    my $xx5  = 0.5/$root5;
    my $cos5 = sqrt(0.5+$xx5);
    my $sin5 = sqrt(0.5-$xx5);
         
#     Generate y-dyad matrices
    for(my $k=0; $k<3; $k++) {
	$r2y[0][$k][$k] = 1.0;
    }
    $r2y[1][0][0] = -1.0;
    $r2y[1][1][1] =  1.0;
    $r2y[1][2][2] = -1.0;
    
#     Generate 3-fold matrices
    my $theta = -$tpiov3;
    for(my $k=0; $k<3; $k++) {
	$theta = $theta + $tpiov3;
	my $cosine = cos($theta);
	my $sine = sin($theta);
	$s3[$k][0][0] = 1.0;
	$s3[$k][1][1] = $cosine;
	$s3[$k][2][1] = $sine;
	$s3[$k][1][2] = -$sine;
	$s3[$k][2][2] = $cosine;
    }
         
#     Generate 5-fold matrices
    $theta = -$tpiov5;
    for(my $k=0; $k<5; $k++) {
	$theta  = $theta+$tpiov5;
	my $cosine = cos($theta);
	my $sine   = sin($theta);
	$s5[$k][0][0] = $cosine;
	$s5[$k][1][0] = $sine;
	$s5[$k][0][1] = -$sine;
	$s5[$k][1][1] = $cosine;
	$s5[$k][2][2] = 1.0;
    }
    
#     Generate x-dyad matrices
    for(my $k=0; $k<3; $k++) {
	$r2x[0][$k][$k] = 1.0;
    }
    $r2x[1][0][0] = 1.0;
    $r2x[1][1][1] = -1.0;
    $r2x[1][2][2] = -1.0;
	
#     Generate  xz-threefold transition matrices
    $p3[0][0] = $cos3;
    $p3[2][0] = $sin3;
    $p3[1][1] = 1.0;
    $p3[0][2] = -$sin3;
    $p3[2][2] = $cos3;
    $p3inv[0][0] = $cos3;
    $p3inv[2][0] = -$sin3;
    $p3inv[1][1] = 1.0;
    $p3inv[0][2] = $sin3;
    $p3inv[2][2] = $cos3;
         
#     Generate  xz-fivefold transition matrices
    $p5[0][0] = $cos5;
    $p5[2][0] = $sin5;
    $p5[1][1] = 1.0;
    $p5[0][2] = -$sin5;
    $p5[2][2] = $cos5;
    $p5inv[0][0] = $cos5;
    $p5inv[2][0] = -$sin5;
    $p5inv[1][1] = 1.0;
    $p5inv[0][2] = $sin5;
    $p5inv[2][2] = $cos5;
         
#     Generate icosahedral rotation matrices. Note use of Fortran 90
#     array syntax and matmul intrinsic function. The rotation matrices
#     are calculated as the product 
#
#     r2y x (p5 x s5 x p5inv) x r2x x (p3 x s3 x p3inv)
#     
#     r2x and r2y are rotations of (0,180) about x- and y-axes
#     s3 is a rotation of (0,120,240) about x-axis
#     s5 is a rotation of (0,72,144,216,288) about z-axis
#
#     p3[5] rotates coordinates so that the 3[5]-fold symmetry axis is
#     aligned with the x[z]-axis. p3[5]inv rotates back to the
#     orthonormal frame.

    my (@tmat1, @tmat2, @tmat3, @tmat4, @tmat5, @tmat6);

    my $l2_max = 1;
    my $j2_max = 1;
    my $i3_max = 1;
    my $k5_max = 1;

    if($use2) {
	$l2_max = 2;
	$j2_max = 2;
    }

    if($use3) {
	$i3_max = 3;
    }

    if($use5) {
	$k5_max = 5;
    }

    my $n = 0;
    for(my $l2=0; $l2<$l2_max; $l2++) {
	matmul(\@{$r2y[$l2]}, \@p5, \@tmat1, 3);
	for(my $k5=0; $k5<$k5_max; $k5++) {
	    matmul(\@tmat1, \@{$s5[$k5]}, \@tmat2, 3);
	    matmul(\@tmat2, \@p5inv, \@tmat3, 3);
	    for(my $j2=0; $j2<$j2_max; $j2++) {
		matmul(\@tmat3, \@{$r2x[$j2]}, \@tmat4, 3);
		matmul(\@tmat4, \@p3, \@tmat5, 3);
		for(my $i3=0; $i3<$i3_max; $i3++) {
		    matmul(\@tmat5, \@{$s3[$i3]}, \@tmat6, 3);
		    matmul(\@tmat6, \@p3inv, \@rot, 3);
		    
		    for(my $i=0; $i<3; $i++) {
			for(my $j=0; $j<3; $j++) {
			    $$rotref[$n][$i][$j] = $rot[$i][$j];
			}
		    }
		    $n = $n+1;
		}
	    }
	}
    }

    return($n);
}
