use strict;
use warnings;

sub get_ann_lo_hi {

    my ($file, $pftrad_lo, $pftrad_step, $boxrad, $bin_factor) = @_;

    # Purpose: Determine the inner and outer radii of the ordered
    # region of the structure. The "rads" file generated by PFTsearch,
    # which contains the correlation coefficient (CC) as a function of
    # radius, is analyzed to identify the location of a fixed size
    # window over which the sum of the CC is a maximum. The window is
    # set to be 1/3 of the box radius and the lower and upper
    # boundaries of the window correspond to the inner and outer radii
    # of the annulus.
    #
    # In order to minimize impact of noise, all operations are done on
    # smoothed data where the value of the CC at each radii is
    # replaced by the three point average of the data point and its
    # neighbors.
    #
    # One additional correction is made to the locations of the window
    # boundaries. If the value of the CC at the boundary is greater
    # than 1/2 of the peak CC, the location of the boundary is moved
    # "out" until the value of the CC is below 1/2 of the peak height
    # or the end of the data set is reached.
    #
    # Note that the PFTsearch output does not actually contain radii,
    # but rather indices. These indicies can be converted to radii (in
    # units of pixels) using the following formula
    #
    #       r = (pftrad_lo  +  index * pftrad_step) * bin_factor
    #
    # Input arguments:
    #   $file        - Name of file containing CC data
    #   $pftrad_lo   - Lower limit for radius at which CC is calculated
    #   $pftrad_step - Radius increment
    #   $boxrad      - Radius of image box
    #
    # Return values:
    #   $ann_lo - lower annulus radius
    #   $ann_hi - upper annulus radius
    
    # Parse the input file and load array of CC's
    open(RADS, $file) || die 
	"Unable to open file $file in get_ann_lo_hi\n";
    my @cc = ();
    my $reached_data = 0;
    while(<RADS>) {
	# Look for the start of average CC data
	unless($reached_data) {
	    if(/--------/) {
		$reached_data = 1;
	    }
	    next;
	}
	
	# Get rid of leading and trailing whitespace
	s/^\s+//;
	s/\s+$//;
	
	my (undef, undef, $cc) = split(/\s+/);
	push(@cc, $cc);
    }
    close(RADS);

    # Set the size of the sliding window to 1/3 of the box
    # radius. Since the CC's in the PFT output are written at
    # intervals of pftrad_step*bin_factor, the number of contributions
    # to the window is equal to (boxrad)/(3*pftrad_step*bin_factor)

    my $window = $boxrad / (3.0 * $pftrad_step * $bin_factor);
    $window = sprintf("%0.f", $window); # Round to nearest integer

    # Calculate smoothed CC
    my $ncc = @cc;
    my @smoothed_cc = ();
    $smoothed_cc[0] = $cc[0];
    $smoothed_cc[$ncc-1] = $cc[$ncc-1];
    for (my $i=1; $i<$ncc-1; $i++) {
	my $smoothed_cc = ($cc[$i-1]+$cc[$i]+$cc[$i+1])/3;
	$smoothed_cc[$i] = $smoothed_cc;
    }

    # Identify peak in smoothed CC
    my ($peakval, $peakloc) = (0.0, 0);
    for (my $i=0; $i<$ncc; $i++) {
	if($smoothed_cc[$i] > $peakval) {
	    $peakval = $smoothed_cc[$i];
	    $peakloc = $i;
	}
    }

    # Determine the location of the window that maximizes the sum of
    # the correlation coefficients

    my $sum_cc_best = 0;
    my ($ann_lo, $ann_hi);
    for (my $low=0; $low<$ncc-$window; $low++) {
	my $sum_cc = 0;
	for (my $i=$low; $i<$low+$window; $i++) {
	    $sum_cc += $smoothed_cc[$i];
	}
	if($sum_cc > $sum_cc_best) {
	    $sum_cc_best = $sum_cc;
	    $ann_lo = $low;
	    $ann_hi = $low + $window - 1;
	}
	my $high = $low+$window-1;
    }

    # If the value of the smoothed CC at the window boundary is
    # greater than one-half of the smoothed CC peak height, move the
    # boundaries outward (i.e. push the upper limit to larger radii,
    # lower limit to lower radii)

    while(($ann_hi < $ncc-1) && $smoothed_cc[$ann_hi]*2 >= $peakval) {
	++$ann_hi;
    }

    while(($ann_lo > 0) && $smoothed_cc[$ann_lo]*2 >= $peakval) {
	--$ann_lo;
    }

    # Convert $ann_lo and $ann_hi from indices of the @cc array to radii
    $ann_lo = ($pftrad_lo + $pftrad_step*$ann_lo) * $bin_factor;
    $ann_hi = ($pftrad_lo + $pftrad_step*$ann_hi) * $bin_factor;

    return($ann_lo, $ann_hi);
}

sub combine_ppft {

    # Purpose: auxillary program for combining and averaging the rads
    # files from multiple runs of PPFT. Only used when PPFT
    # calculation are broken up into runs over individual files. 
    #
    # Input arguments:
    #   $avg_file - averaged PPFT rads file
    #   @ind_files - individual PPFT rads files
    # 
    # Return values:
    #   None

    # NOTE - as of 9/5/2006, options for calling this routine are no
    # longer included in auto3dem.pl

    my ($avg_file, @ind_files) = @_;
    
    my $nlocal;
    my $ntotal = 0;
    my (@ccu, @ccf);
    
    foreach my $file (@ind_files) {
	
	my $indx = 0;
	my $reached_data = 0;
	open(FH, $file);
	while(<FH>) {
	    
	    # Look for number of images
	    if(/images/i) {
		s/[\sa-zA-Z]//g;
		$nlocal = $_;
		$ntotal += $nlocal;
	    }
	    
	    # Look for the start of average CC data
	    unless($reached_data) {
		if(/--------/) {
		    $reached_data = 1;
		}
		next;
	    }
	    
	    # Get rid of leading and trailing whitespace
	    s/^\s+//;
	    s/\s+$//;
	    
	    my (undef, $ccu, $ccf) = split(/\s+/);
	    $ccu *= $nlocal;
	    $ccf *= $nlocal;
	    $ccu[$indx] += $ccu;
	    $ccf[$indx] += $ccf;
	    $indx++;
	}
	close(FH);
    }
    
    my $nrows = @ccu;
    for (my $i=0; $i<$nrows; $i++) {
	$ccu[$i] /= $ntotal;
	$ccf[$i] /= $ntotal;
    }
  
    open(AVG, ">$avg_file");
    print AVG "---------------------------\n";
    for (my $i=0; $i<$nrows; $i++) {
	printf AVG "%d\t%0.3f\t%0.3f\n", $i, $ccu[$i], $ccf[$i];
    }
    close(AVG);

    return;
}
1;
