use strict;
use warnings;
use lib do { __FILE__ =~ m|^(.*?)[^/]*$|; "$1"; };

sub get_paramfile_info {

    # Purpose: Get number of particles, number of micrographs, and
    # defocus range in particle parameter file. Works for all
    # parameter files including those which contain data for multiple
    # micrographs
    #
    # Input arguments:
    #   $ptlefile - name of parameter file
    #
    # Return values:
    #   $nptle  - number of particle records
    #   $nmicro - number of micrographs
    #   $defmin - minimum defocus value
    #   $defmax - maximum defocus value

    my($ptlefile) = @_;

    my $nptle = 0;
    my $nmicro = 0;
    my $defmin = 99.99;
    my $defmax = -99.99;
    my $header_flag;

    open(PTLE, $ptlefile) || die
	"Unable to open file $ptlefile for reading in count_ptles\n";

    while(<PTLE>) {

	# Save the record and strip off trailing whitespace
	my $record = $_;
	$record =~ s/\s+//g;

	unless($record =~ /\w/) {
	    next; # skipping blank lines
	}

	if($record =~ /pif/i || $_ =~ /mrc/i) {
	    $header_flag = 1;
	    $nmicro++;
	    next;
	}
	if($header_flag) {
	    $header_flag = 0;
	    my (undef, undef, undef, undef, $def) = split(/,/, $record);
	    if($def < $defmin) {
		$defmin = $def;
	    }
	    if($def > $defmax) {
		$defmax = $def;
	    }
	    next;
	}

	$nptle++;
    }

    close(PTLE);
    return($nptle, $nmicro, $defmin, $defmax);
}

sub get_pixel_size {

    # Purpose: extract the pixel size from the second line of a file
    # containing particle orientation and origin information
    #
    # Input arguments:
    #   $filename - name of data file
    #
    # Return values:
    #   $pixel_size - pixel size

    my ($filename) = @_;
    open(DAT, $filename) || 
	die "Unable to open file $filename in subroutine get_pixel_size";

    <DAT>;
    my $line = <DAT>;
    my ($pixel_size) = split(/,/, $line);
    $pixel_size =~ s/\s+//g;

    close(DAT);
    return $pixel_size;
}

sub get_img_name {

    # Purpose: get the name of the image file from the first line of a
    # file containing particle orientation and origin information
    #
    # Input arguments:
    #   $filename - name of data file
    #
    # Return values:
    #   $img_file - name of image file

    my ($filename) = @_;
    open(DAT, $filename) || 
	die "Unable to open file $filename in subroutine get_img_name";

    my $img_name = <DAT>;
    $img_name =~ s/\s+//g;

    close(DAT);
    return $img_name;
}

sub get_numimages {
    
    # Purpose: get the number of images contained in a PIF or MRC file
    #
    # Input arguments:
    #   $filename - name of PIF or MRC file
    #
    # Return values:
    #   $numimages - number of images in PIF or MRC file

    my ($filename) = @_;

    my $filename_nopath = $filename;
    $filename_nopath =~ s/.*\///;

    # Try first to open file with full path name. If unsuccessful,
    # then try opening again with path information stripped. Having
    # both options allows more flexibility in how image name is
    # specified in particle parameter file.

    open(FH, $filename) || open(FH, $filename_nopath) || die 
	"Unable to open file $filename or $filename_nopath in
	subroutine get_numimages\n";
    binmode FH;

    my $numimages;
    if ($filename =~ /\.pif\s*$/) {

	# Determine the endianness of the machine that created this
	# file based on the 4 bytes offset 28 bytes from the start of
	# the file. According to the PIF standard, a value of zero
	# corresponds to little endian, while a value of 1 corresponds
	# to big endian. Since we don't know the endianness of the
	# file ahead of time, the 4 bytes will arbitrarily be
	# interpreted as a big-endian long int and we'll test whether
	# or not the result is non-zero

	my $bigend;
	seek(FH, 28, 0);
	read FH, $bigend, 4;
	$bigend = unpack "N", $bigend; # # Big endian interpretation
	
	# numimages variable is offset 24 bytes from start of file
	seek(FH, 24, 0);
	read FH, $numimages, 4;
	
	# Interpret as big- or little-endian based on endianness of file
	if($bigend) {
	    $numimages = unpack "N", $numimages;
	} else {
	    $numimages = unpack "V", $numimages;
	}
    } elsif ($filename =~ /\.mrc\s*$/) {

	# The MRC format does not contain an endianness flag, but we
	# should be able to figure out the endianness on the basis of
	# whether or not the dimensions of the data set have
	# reasonable values (greater than zero, but not too big)

	my ($a, $b, $c, $nx, $ny, $nz);
	
	read FH, $a, 4;
	read FH, $b, 4;
	read FH, $c, 4;
	
	# Make guess that data is little endian
	$nx = unpack "V", $a;
	$ny = unpack "V", $b;
	$nz = unpack "V", $c;
	
	# If results don't make sense, then interpret as big endian
	if($nx <= 0 || $nx > 10000 || 
	   $ny <= 0 || $ny > 10000 || 
	   $nz <= 0 || $nz > 10000) {
	    $nx = unpack "N", $a;
	    $ny = unpack "N", $b;
	    $nz = unpack "N", $c;
	}

	$numimages = $nz;
    }

    close(FH);
    return($numimages);
}

sub get_dimensions {
    
    # Purpose: get the dimensions of a PIF or MRC image
    #
    # Input arguments:
    #   $filename - name of PIF or MRC file
    #
    # Return values:
    #   $nx, $ny, $nz - x, y, and z dimensions of PIF or MRC image

    my ($filename) = @_;

    my $filename_nopath = $filename;
    $filename_nopath =~ s/.*\///;

    # Try first to open file with full path name. If unsuccessful,
    # then try opening again with path information stripped. Having
    # both options allows more flexibility in how image name is
    # specified in particle parameter file.

    open(FH, $filename) || open(FH, $filename_nopath) || die 
	"Unable to open file $filename or $filename_nopath in
	subroutine get_numimages\n";
    binmode FH;

    my ($nx, $ny, $nz);
    if ($filename =~ /\.pif\s*$/) {

	# Determine the endianness of the machine that created this
	# file based on the 4 bytes offset 28 bytes from the start of
	# the file. According to the PIF standard, a value of zero
	# corresponds to little endian, while a value of 1 corresponds
	# to big endian. Since we don't know the endianness of the
	# file ahead of time, the 4 bytes will arbitrarily be
	# interpreted as a big-endian long int and we'll test whether
	# or not the result is non-zero

	my $bigend;
	seek(FH, 28, 0);
	read FH, $bigend, 4;
	$bigend = unpack "N", $bigend; # # Big endian interpretation
	
	# x, y, and z dimensions are offset 68, 72, and 76 bytes from start
	seek(FH, 68, 0);
	read FH, $nx, 4;
	read FH, $ny, 4;
	read FH, $nz, 4;
	
	# Interpret as big- or little-endian based on endianness of file
	if($bigend) {
	    $nx = unpack "N", $nx;
	    $ny = unpack "N", $ny;
	    $nz = unpack "N", $nz;
	} else {
	    $nx = unpack "V", $nx;
	    $ny = unpack "V", $ny;
	    $nz = unpack "V", $nz;
	}
    } elsif ($filename =~ /\.mrc\s*$/) {

	# The MRC format does not contain an endianness flag, but we
	# should be able to figure out the endianness on the basis of
	# whether or not the dimensions of the data set have
	# reasonable values (greater than zero, but not too big)

	my ($a, $b, $c);
	
	read FH, $a, 4;
	read FH, $b, 4;
	read FH, $c, 4;
	
	# Make guess that data is little endian
	$nx = unpack "V", $a;
	$ny = unpack "V", $b;
	$nz = unpack "V", $c;
	
	# If results don't make sense, then interpret as big endian
	if($nx <= 0 || $nx > 10000 || 
	   $ny <= 0 || $ny > 10000 || 
	   $nz <= 0 || $nz > 10000) {
	    $nx = unpack "N", $a;
	    $ny = unpack "N", $b;
	    $nz = unpack "N", $c;
	}
    }

    close(FH);
    return($nx, $ny, $nz);
}

sub get_boxrad {
    
    # Purpose: get the box size (radius) from a PIF or MRC file
    #
    # Input arguments:
    #   $filename - name of PIF or MRC file
    #
    # Return values:
    #   $boxrad - box size (radius)
    #
    # Note - the variable $boxrad corresponds to the field packRadius
    # in the PIF format specification. We'll assume that the MRC
    # images are square and use the first word, corresponding to NX

    my ($filename) = @_;

    my $filename_nopath = $filename;
    $filename_nopath =~ s/.*\///;

    # Try first to open file with full path name. If unsuccessful,
    # then try opening again with path information stripped. Having
    # both options allows more flexibility in how image name is
    # specified in particle parameter file.

    open(FH, $filename) || open(FH, $filename_nopath) || die 
	"Unable to open file $filename or $filename_nopath in
	subroutine get_boxrad\n";
    binmode FH;

    my $boxrad;
    if ($filename =~ /\.pif\s*$/) {

	# Determine the endianness of the machine that created this
	# file based on the 4 bytes offset 28 bytes from the start of
	# the file. According to the PIF standard, a value of zero
	# corresponds to little endian, while a value of 1 corresponds
	# to big endian. Since we don't know the endianness of the
	# file ahead of time, the 4 bytes will arbitrarily interpreted
	# as a big-endian long int and we'll test whether or not the
	# result is non-zero
	
	my ($bigend, $mode);
	seek(FH, 28, 0);
	read FH, $bigend, 4;
	$bigend = unpack "N", $bigend; # Big endian interpretation

	# Determine the mode for the EM data type based on the 4 bytes
	# offset 80 bytes from the start of the file. Intepret as big-
	# or little-endian 32-bit int based on endianness of file.

	seek(FH, 80, 0);
	read FH, $mode, 4;
	if($bigend) {
	    $mode = unpack "N", $mode; # Big endian interpretation
	} else {
	    $mode = unpack "V", $mode; # Little endian interpretation
	}
	
	# If the mode is equal to 97, then we have depthcued data. In
	# this case, box radius (packRadius) is offset 2068 bytes from
	# start of file. Otherwise, the offset is 532 bytes.

	if($mode == 97) {
	    seek(FH, 2068, 0);
	    read FH, $boxrad, 4;
	} else {
	    seek(FH, 532, 0);
	    read FH, $boxrad, 4;
	}    
	
	# Interpret as big- or little-endian based on endianness of file
	
	if($bigend) {
	    $boxrad = unpack "N", $boxrad;
	} else {
	    $boxrad = unpack "V", $boxrad;
	}
    } elsif ($filename =~ /\.mrc\s*$/) {
	
	# The MRC format does not contain an endianness flag, but we
	# should be able to figure out the endianness on the basis of
	# whether or not the dimensions of the data set have
	# reasonable values (greater than zero, but not too big)
	
	my ($a, $b, $c, $nx, $ny, $nz);
	
	read FH, $a, 4;
	read FH, $b, 4;
	read FH, $c, 4;
	
	# Make guess that data is little endian
	$nx = unpack "V", $a;
	$ny = unpack "V", $b;
	$nz = unpack "V", $c;
	
	# If results don't make sense, then interpret as big endian
	if($nx <= 0 || $nx > 10000 || 
	   $ny <= 0 || $ny > 10000 || 
	   $nz <= 0 || $nz > 10000) {
	    $nx = unpack "N", $a;
	    $ny = unpack "N", $b;
	    $nz = unpack "N", $c;
	}
	
	# This gives the correct boxrad whether $nx even or odd
	$boxrad = int($nx/2);
    }
    
    close(FH);
    return($boxrad);
}
1;
