use strict;
use warnings;
use lib do { __FILE__ =~ m|^(.*?)[^/]*$|; "$1"; };

use Cwd;

sub init_params {

    # Purpose: set the input parameters for auto3dem. Defaults are
    # assigned for some parameters, but these can be overwritten using
    # data in the input file.
    #
    # In order to make it easier to keep track of variables,
    # parameters are collected into hashes - one describing the
    # overall flow of auto3dem and one for each of the main parallel
    # applications.
    # 
    # The "program" records of the input file have the format
    #
    #        ident  key  value
    #
    # Where the ident = (auto|po2r|p3dr|pcut|psf|ppft|pctfr)
    # "auto" specifies control parameters that direct the overall flow
    # of auto3dem
    #
    # The "data" records of the input file have the format
    #
    #       data  datafile
    #
    # Input arguments:
    #   $param_file - input file
    #   $auto_ref   - reference to hash of control parameters
    #   $po2r_ref   - reference to hash of PO2R input parameters
    #   $p3dr_ref   - reference to hash of P3DR input parameters
    #   $pcut_ref   - reference to hash of PCUT input parameters
    #   $psf_ref    - reference to hash of PSF input parameters
    #   $ppft_ref   - reference to hash of PSF input parameters
    #   $pctfr_ref  - reference to hash of PCTFR input parameters
    #   $data_ref   - reference to array of data files
    #
    # Return values:
    #   None

    my ($param_file, $auto_ref, $po2r_ref, $p3dr_ref,
	$pcut_ref, $psf_ref, $ppft_ref, $pctfr_ref, $data_ref) = @_;

    open(IN, $param_file) || die
	"Unable to open file $param_file";

    # Set default control parameters
    $$auto_ref{fsc_hithresh} = 0.5; # High FSC threshold
    $$auto_ref{fsc_lothresh} = 0.3; # Low FSC threshold
    $$auto_ref{res_adj}   = 0.01;   # Tweak for resolution adjustment
    $$auto_ref{gauss_adj} = 0.01;   # Tweak for gaussian adjustment
    $$auto_ref{have_map}  = 1;      # Have starting map
    $$auto_ref{new_ptles} = 0;      # Don't refine ptles rel to existing map
    $$auto_ref{freeze_annulus} = 0; # Don't freeze radii of annulus
    $$auto_ref{freeze_res}     = 0; # Don't freeze resolutions
    $$auto_ref{delete_maps}    = 1; # Delete intermediate maps
    $$auto_ref{restart} = 0;        # Not a restart run
    $$auto_ref{iter_start} = 1;     # Normally start at iteration 1
    $$auto_ref{term_search} = 0;    # Don't terminate search mode runs   
    $$auto_ref{term_refine} = 0;    # Don't terminate refine mode runs
    $$auto_ref{switch_mode} = 1 ;   # Allow switch from search to refine
    $$auto_ref{bin_reduce}  = 0;    # Auto reduction of bin factor
    $$auto_ref{refine_ctf}  = 0;    # Refine CTF
    $$auto_ref{quit_early}  = 0;    # Quit early if PSF res_max reached
    $$auto_ref{rundir} = 'dat';     # Look for data in dat
    $$auto_ref{noctf} = 0;          # Allow CTF corrections to be applied
    $$auto_ref{per_ptle_ctf} = 0;   # Disable per particle CTF
    $$auto_ref{noise_suppression} = 0; # Disable noise suppression
    $$auto_ref{generate_map} = 'yes';  # Generate reconstruction after alignment
                                       # accepted strings: 'yes', 'no', 'only'
    $$auto_ref{map_suffix} = 'none';   # Suffix to append to map name
                                       # used if generate_map is set to 'only'
                                       # 'none' is a magic word to indicate no appending 
    $$auto_ref{adapt_angle} = 'yes';   # Determine angle from particle radius and resolution
    $$auto_ref{estimate_res} =  1;     # Do resolution estimation
    $$auto_ref{rmc} =  0;              # Specify whether random model computation

    $$auto_ref{theta_reject_lower} =  360;  # Accepting all orientations
    $$auto_ref{theta_reject_upper} = -360;  # Accepting all orientations
    $$auto_ref{phi_reject_lower}   =  360;  # Accepting all orientations
    $$auto_ref{phi_reject_upper}   = -360;  # Accepting all orientations
    $$auto_ref{omega1}             = 0;     # Accepting all orientations
    $$auto_ref{omega1_tol}         = 360;   # Accepting all orientations
    $$auto_ref{omega2}             = 0;     # Accepting all orientations
    $$auto_ref{omega2_tol}         = 0;     # Accepting all orientations
    $$auto_ref{nselect_offset}     = 0;     # Single selection criterion
    $$auto_ref{select_delta}       = 0;     # Zero perturbation of criteria
    $$auto_ref{global_select}      = 1;     # Get select crit from all files

    $$auto_ref{hollow_map}         = 0;     # Don't hollow map
    $$auto_ref{hollow_cut_step}    = 12;    # Number of steps used in masking
    $$auto_ref{hollow_cut_weight}  = 0.001; # Weight used in masking
    $$auto_ref{hollow_in_rad}      = 0;     # Dummy value for hollow map r_in
    $$auto_ref{hollow_out_rad}     = 0;     # Dummy value for hollow map r_out

    # Set default program parameters
    $$psf_ref{res_step} = 50;              # Number of steps for FSC curve
    $$psf_ref{res_min} = 60;               # Lower res limit for FSC curve

    $$po2r_ref{symm_code}        = 532;    # Icosahedral symmetry
    $$po2r_ref{mode}             = 'local';# refinement mode
    $$po2r_ref{gangle}           = 2;      # Step size for global refinement (degrees)
    $$po2r_ref{dcenter}          = 0.1;    # Step size for origin refinement
    $$po2r_ref{nangle}           = 4;      # Number of angle steps in each dir
    $$po2r_ref{ncenter}          = 4;      # Number of origin steps in each dir
    $$po2r_ref{tempfac}          = 0;      # Temperature factor
    $$po2r_ref{zero_fill}        = 1;      # Zero fill
    $$po2r_ref{ctf_ff1}          = 1.0;    # CTF factor 1
    $$po2r_ref{ctf_ff2}          = 0.1;    # CTF factor 2
    $$po2r_ref{ctfmode}          = 1;      # CTF mode
    $$po2r_ref{filter}           = 1;      # Filter mode
    $$po2r_ref{funcmode}         = 1;      # Funcmode
    $$po2r_ref{funcweight}       = 1.;     # Funcweight
    $$po2r_ref{quick_search}     = 1;      # Enable quick search
    $$po2r_ref{per_ptle_ctf}     = 0;      # Disable per particle CTF
    $$po2r_ref{handtest}         = 1;      # Enable testing for handflip
    $$po2r_ref{ticos_equiv}      = 0;      # Disable test over 532 equivs (obsolete)
    $$po2r_ref{global_por}       = 0;      # Disable global search (obsolete)
    $$po2r_ref{nmagref}          = 10;     # Number of magnification steps in each dir
                                           # 0 = disabled; >1 = exclusive refinement (older approach)
    $$po2r_ref{magref_step}      = 0.005;  # Step size for magnification refinement
    $$po2r_ref{magref_reset}     = 1;      # Reset relative magnification to 1 before refinement
    $$po2r_ref{magref_calibrate} = 0;      # Magnification is calibrated against gold standard
                                           # 0 = values are rescaled to average of 1.
                                           # 1 = values are kept as estimated

    $$pctfr_ref{dangle}        = 0.05;     # Step size for CTF refinement
    $$pctfr_ref{nangle}        = 4;        # Number of steps in each dir
    $$pctfr_ref{tempfac}       = 0;        # Temperature factor
    $$pctfr_ref{zero_fill}     = 1;        # Zero fill
    $$pctfr_ref{ctf_ff1}       = 0.5;      # CTF factor 1
    $$pctfr_ref{ctf_ff2}       = 0.1;      # CTF factor 2
    $$pctfr_ref{ctfmode}       = 1;        # CTF mode
    $$pctfr_ref{filter}        = 1;        # Filter mode
    $$pctfr_ref{funcmode}      = 3;        # Funcmode (phase method)
    $$pctfr_ref{funcweight}    = 0;        # Funcweight

    $$p3dr_ref{symm_code}     = 532;       # Icosahedral symmetry
    $$p3dr_ref{tempfac}       = 0;         # temperature factor
    $$p3dr_ref{zero_fill}     = 1;         # Zero fill
    $$p3dr_ref{ctf_ff1}       = 1.0;       # 1st CTF factor
    $$p3dr_ref{ctf_ff2}       = 0.1;       # 2nd CTF factor
    $$p3dr_ref{ctfmode}       = 1;         # CTF mode
    $$p3dr_ref{filter}        = 1;         # Filter mode
    $$p3dr_ref{magfactor}     = 1.0;       # Magnification factor
    $$p3dr_ref{map_dim}       = 0;         # Map dimension
    $$p3dr_ref{apo_border}    = 12;        # Default apodization border
    $$p3dr_ref{per_ptle_ctf}  = 0;         # Disable per particle CTF

    $$ppft_ref{symm_code}    = 532;        # Icosahedral symmetry
    $$ppft_ref{bin_factor}   = 1;          # No binning
    $$ppft_ref{prj_filename} = "pft.prjs"; # Projections file name
    $$ppft_ref{pft_filename} = "pft.pfts"; # Polar FT file name

    $$ppft_ref{ctfmode}            = 3;    # CTF mode (divide images by CTF)
    $$ppft_ref{filter_factor_1}    = 0.1;  # CTF filter factor
    $$ppft_ref{input_mode}         = 2;    # Input mode
    $$ppft_ref{jcut}               = 1;    # Minimum order Bessel function
    $$ppft_ref{mag_cen}            = 1.0;  # Midpoint for mag scale search
    $$ppft_ref{mag_norm}           = 1;    # Flag for normalizing mag
    $$ppft_ref{mag_num}            = 1;    # Extent of mag scale search
    $$ppft_ref{mag_step}           = 0.0;  # Grid sise of mag scale search
    $$ppft_ref{pftrad_step}        = 1.0;  # Radial step size in CC calculation
    $$ppft_ref{pftrad_lo}          = 1.0;  # Lower radii for CC calculation
    $$ppft_ref{sigcut}             = 0.0;  # Threshold for variance mask
    $$ppft_ref{temperature_factor} = 0.0;  # Temperature factor
    $$ppft_ref{verbose}            = -1;   # Verbosity
    $$ppft_ref{quick_omega}        = 1;    # Quick omega search
    $$ppft_ref{per_ptle_ctf}       = 0;    # Disable per particle CTF

    $$pcut_ref{cut_step} = 12;             # Number of steps used in masking
    $$pcut_ref{cut_weight} = 0.001;        # Weight used in masking

    # Set default executable names
    $$pctfr_ref{bin} = "PCTFR";
    $$p3dr_ref{bin}  = "P3DR";
    $$po2r_ref{bin}  = "POR";
    $$pcut_ref{bin}  = "PCUT";
    $$psf_ref{bin}   = "PSF";
    $$ppft_ref{bin}  = "PPFT";

    # Set defaults for max number of CPUs when running programs
    $$p3dr_ref{max_cpu}  = 256;
    $$pctfr_ref{max_cpu} = 256;
    $$po2r_ref{max_cpu}  = 256;
    $$ppft_ref{max_cpu}  = 256;
    $$pcut_ref{max_cpu}  = 8;
    $$psf_ref{max_cpu}   = 8;

    # Parse input file and set input parameters
    while (<IN>) {
	# Get rid of comments ("#" to end of line)
	s/#.*//;

	# Get rid of leading and trailing whitespace
	s/^\s+//;
	s/\s+$//;

	# Skip empty lines
	unless(/\w/) {
	    next;
	}

	#Parse line on whitespace
	my $value;
	my ($ident, $key, @remainder) = split(/\s+/, $_);
	if($ident eq 'auto' && $key eq 'recipient') {
	    $value = join(' ', @remainder);
	} else {
	    $value = $remainder[0];
	}

	# Convert all identifiers to lower case; convert keys to
	# lowercase for lines that don't specify data files
	$ident =~ tr/A-Z/a-z/;
	unless($ident eq 'data') {
	    $key  =~ tr/A-Z/a-z/;
	}

	# Compatibility feature for older versions of auto3dem. 

	# Input for PO2R can be specified using 'por' or 'po2r'
	if($ident eq 'por') {
	    $ident = 'po2r';
	}
	# fsc_lothresh can also be specified as fsc_locut
	if($key eq 'fsc_locut') {
	    $key = 'fsc_lothresh';
	}
	# fsc_hithresh can also be specified as fsc_hicut
	if($key eq 'fsc_hicut') {
	    $key = 'fsc_hithresh';
	}

	# ctfmode and ctf_mode both accepted for po2r, p3dr, pctfr,
	# and ppft. (ctf_mode changed to ctfmode)
	if($key eq 'ctf_mode') {
	    $key = 'ctfmode';
	}

	# ppft symmetry accepts both symmetry and symm_code
	# (symmetry changed to symm_code)
	if($key eq 'symmetry') {
	    $key = 'symm_code';
	}

	# Skip lines that don't have required number of fields, that
	# have invalid keys, or that don't correspond to valid choices
	# of data or program
	if($ident eq 'auto'  ||
	   $ident eq 'po2r'  ||
	   $ident eq 'pctfr' ||
	   $ident eq 'p3dr'  ||
	   $ident eq 'pcut'  ||
	   $ident eq 'psf'   ||
	   $ident eq 'ppft') {
	    unless(defined $key) {
		print "\n";
		print "*** Warning - parameter file line $.\n";
		print "missing key for '$ident'\n";
		next;
	    }
	    unless(defined $value) {
		print "\n";
		print "*** Warning - parameter file line $.\n";
		print "missing value for '$ident $key'\n";
		next;
	    }	    
	    if(deprecated_key($ident, $key)) {
		print "\n";
		print "*** Warning - parameter file line $.\n";
		print "'$key' no longer used and is ignored for '$ident', ";
		print "unless it can be converted to an equivalent parameter\n";
		next;
	    }
	    unless(valid_key($ident, $key)) {
		print "\n";
		print "*** Warning - parameter file line $.\n";
		print "'$key' not valid key for '$ident'\n";
		next;
	    }
	} elsif ($ident eq 'all') {
	    print "\n";
	    print "*** Warning - parameter file line $.\n";
	    print "'all' identifier no longer used and is being ignored\n";
	    print "Use 'auto symm_code n' to set symmetry for all programs\n";
	    next;
	} elsif ($ident eq 'data') {
	    unless(defined $key) {
		print "\n";
		print "*** Warning - parameter file line $.\n";
		print "missing file on 'data' line\n";
		next;
	    }
	} else {
	    print "\n";
	    print "*** Warning - parameter file line $.\n";
	    print "'$ident' not program valid choice\n";
	    next;
	}

	if ($ident eq 'po2r')  {
	    $$po2r_ref{$key}       = $value;
	} elsif ($ident eq 'pctfr') {
	    $$pctfr_ref{$key}      = $value;
	} elsif ($ident eq 'p3dr')  {
	    $$p3dr_ref{$key}      = $value;
	} elsif ($ident eq 'pcut')  {
	    $$pcut_ref{$key}      = $value;
	} elsif ($ident eq 'psf')  {
	    $$psf_ref{$key}       = $value;
	} elsif ($ident eq 'ppft') {
	    $$ppft_ref{$key}      = $value;
	} elsif ($ident eq 'auto') {
	    $$auto_ref{$key}      = $value;
	} elsif ($ident eq 'data') {
	    push(@{$data_ref}, $key);
	}
    }
    close(IN);

    # Tests to make sure that input data is complete and consistent
    my $err = 0;

    unless(defined $$auto_ref{mode} && 
	   ($$auto_ref{mode} eq 'search' || $$auto_ref{mode} eq 'refine')){
	print "\n*** Error message\n";
	print "Valid mode (search or refine) needs to be specified\n";
	$err = 1;
    }

    unless($$auto_ref{generate_map} eq 'yes' || $$auto_ref{generate_map} eq 'no' ||
            $$auto_ref{generate_map} eq 'only' ){
	print "\n*** Error message\n";
	print "Valid generate_map (yes, no or only) needs to be specified\n";
	$err = 1;
    }

    unless($$auto_ref{adapt_angle} eq 'yes' || $$auto_ref{adapt_angle} eq 'no' ){
	print "\n*** Error message\n";
	print "Valid adapt_angle (yes or no) needs to be specified\n";
	$err = 1;
    }

   if($$auto_ref{new_ptles}) {
	unless(defined $$auto_ref{start_map}) {
	    print "\n*** Error message\n";
	    print "Starting map not specified\n";
	    print "  If 'new_ptles' flag is set, need to specify map\n";
	    $err = 1;
	}
    }

    if($$auto_ref{have_map}) {
	unless(defined $$auto_ref{start_map}) {
	    print "\n*** Error message\n";
	    print "Starting map not specified\n";
	    print "  If 'have_map' flag is set, need to specify map\n";
	    $err = 1;
	}
    }

    if($$auto_ref{hollow_map}) {
	unless(defined $$auto_ref{hollow_in_rad}) {
	    print "\n*** Error message\n";
	    print "Need to set hollow_in_rad\n";
	    $err = 1;
	}
	unless(defined $$auto_ref{hollow_out_rad}) {
	    print "\n*** Error message\n";
	    print "Need to set hollow_out_rad\n";
	    $err = 1;
	}
	if (defined $$auto_ref{hollow_in_rad} &&
	    defined $$auto_ref{hollow_out_rad} &&
	    $$auto_ref{hollow_in_rad} >= $$auto_ref{hollow_out_rad}) {
	    print "\n*** Error message\n";
	    print "hollow_out_rad ($$auto_ref{hollow_out_rad}) ";
	    print "must be greater than ";
	    print "hollow_in_rad ($$auto_ref{hollow_in_rad})\n";
	    $err = 1;
	}
    }	    

    unless(defined $$p3dr_ref{res_min}) {
	print "\n*** Error message\n";
	print "Must specify res_min for p3dr\n";
	$err = 1;
    }

    unless(@{$data_ref} > 0) {
	print "\n*** Error message\n";
	print "Must specify at least one data file\n";
	$err = 1;
    }

    unless(defined $$auto_ref{niter} && $$auto_ref{niter} >= 0) {
	print "\n*** Error message\n";
	print "Number of iterations (auto niter) must be at least zero\n";
	$err = 1;
    }

    # obsolete parameters
    # global refinement, magnification refinement and
    # test over equivalent icosahedral are incompatible with each other
    if($$po2r_ref{ticos_equiv} && $$po2r_ref{global_por}) {
	print "\n*** Error message\n";
	print "ticos_equiv and global_por incompatible for PO2R\n";
	$err = 1;
    }
    if($$po2r_ref{ticos_equiv} && $$po2r_ref{nmagref}) {
	print "\n*** Error message\n";
	print "ticos_equiv and magref incompatible for PO2R\n";
	$err = 1;
    }
    if($$po2r_ref{nmagref} && $$po2r_ref{global_por}) {
	print "\n*** Error message\n";
	print "magref and global_por incompatible for PO2R\n";
	$err = 1;
    }
    if($$po2r_ref{ticos_equiv}) {
	print "\n*** Warning message\n";
	print "PO2R - ticos_equiv is obsolete: use mode ticos_equiv instead\n";
    }
    if($$po2r_ref{global_por}) {
	print "\n*** Warning message\n";
	print "PO2R - global_por is obsolete: use mode global instead\n";
    }

    unless(defined $$po2r_ref{mode}) {
	if ($$po2r_ref{ticos_equiv}) {
	    $$po2r_ref{mode} = 'ticos_equiv';
	} elsif ($$po2r_ref{global_por}) {
	    $$po2r_ref{mode} = 'global';
	} elsif ($$po2r_ref{nmagref}) {
	    $$po2r_ref{mode} = 'mag';
	} else {
	    $$po2r_ref{mode} = 'local';
	}
    }

    unless(defined $$po2r_ref{mode} && 
	   ($$po2r_ref{mode} eq 'local' || $$po2r_ref{mode} eq 'global' ||
	   $$po2r_ref{mode} eq 'mag' || $$po2r_ref{mode} eq 'ticos_equiv')){
	print "\n*** Error message\n";
	print "Valid refinement mode (local or global or mag or ticos_equiv) needs to be specified\n";
	$err = 1;
    }

    if($err) {
	die "Errors found in auto3dem input file - terminating program\n";
    }

    # If auto noctf is set, set ctfmode to zero for all programs
    if($$auto_ref{noctf} == 1) {
	$$p3dr_ref{ctfmode}  = 0;
	$$po2r_ref{ctfmode}  = 0;
	$$pctfr_ref{ctfmode} = 0;
	$$ppft_ref{ctfmode}  = 0;
    }

    # If auto symm_code is set, override symm_code for all programs
    if(defined $$auto_ref{symm_code}) {
	$$po2r_ref{symm_code}  = $$auto_ref{symm_code};
	$$p3dr_ref{symm_code}  = $$auto_ref{symm_code};
	$$ppft_ref{symm_code}  = $$auto_ref{symm_code};
    }

    # Set PPFT angular step size based on symmetry
    # (Currently handles C1-C9 and 532)
    unless(defined $$ppft_ref{delta_theta}) {
	if ($$ppft_ref{symm_code} == 1) {
	    $$ppft_ref{delta_theta} = 4.0;
	} elsif ($$ppft_ref{symm_code} <= 3) {
	    $$ppft_ref{delta_theta} = 2.5;
	} elsif ($$ppft_ref{symm_code} <= 5) {
	    $$ppft_ref{delta_theta} = 2.0;
	} elsif ($$ppft_ref{symm_code} <= 8) {
	    $$ppft_ref{delta_theta} = 1.5;
	} elsif ($$ppft_ref{symm_code} <= 9) {
	    $$ppft_ref{delta_theta} = 1.25;
	} else {
	    $$ppft_ref{delta_theta} = 0.5;
	}
    }
	    
    # If auto per_ptle_ctf is set, enable feature for P3DR, PO2R, and PPFT
    if($$auto_ref{per_ptle_ctf} == 1) {
	$$p3dr_ref{per_ptle_ctf} = 1;
	$$po2r_ref{per_ptle_ctf} = 1;
	$$ppft_ref{per_ptle_ctf} = 1;
    }

    # If not specified, set outfile to name of current directory
    unless(defined $$auto_ref{outfile}) {
	$$auto_ref{outfile} = getcwd();
	$$auto_ref{outfile} =~ s/.*\///;
    }

    # Calculate resolution of P3DR at upper limit of Gaussian falloff
    unless(defined $$p3dr_ref{res_max}) {
	$$p3dr_ref{res_max} = 1/(1/$$p3dr_ref{res_min}+$$auto_ref{gauss_adj});
	$$p3dr_ref{res_max} = sprintf("%.2f", $$p3dr_ref{res_max});
    }

    # Set the upper resolution limit for the FSC curve
    unless(defined $$psf_ref{res_max}) {
	$$psf_ref{res_max} = $$p3dr_ref{res_min};
    }

    # Set the upper resolution limit for image/map comparison in PO2R
    unless(defined $$po2r_ref{res_max}) {
	$$po2r_ref{res_max} = 1/(1/$$p3dr_ref{res_min}-$$auto_ref{res_adj});
	$$po2r_ref{res_max} = sprintf("%.2f", $$po2r_ref{res_max});
    }

    # Set the upper resolution limit for image/map comparison in PCTFR
    unless(defined $$pctfr_ref{res_max}) {
	$$pctfr_ref{res_max} = 1/(1/$$p3dr_ref{res_min}-$$auto_ref{res_adj});
	$$pctfr_ref{res_max} = sprintf("%.2f", $$pctfr_ref{res_max});
    }

    # Set the upper resolution limit for image/map comparison in PPFT
    unless(defined $$ppft_ref{resolution_high}) {
	$$ppft_ref{resolution_high} = 1/
	    (1/$$p3dr_ref{res_min}-$$auto_ref{res_adj});
	$$ppft_ref{resolution_high} = 
	    sprintf("%.2f", $$ppft_ref{resolution_high});
    }

    # Set default angular step size for PO2R to 1/2 that used for PPFT
    unless(defined $$po2r_ref{dangle}) {
	if(defined $$ppft_ref{delta_theta}) {
	    $$po2r_ref{dangle} = 0.5 * $$ppft_ref{delta_theta};
	}
    }
    
    # Make sure that: 0 <= omega1, omega2 <= 360
    #               : omega1_tol, omega2_tol >= 0
    while($$auto_ref{omega1} > 360) {$$auto_ref{omega1} -= 360;}
    while($$auto_ref{omega2} > 360) {$$auto_ref{omega2} -= 360;}
    $$auto_ref{omega1_tol} = abs($$auto_ref{omega1_tol});
    $$auto_ref{omega2_tol} = abs($$auto_ref{omega2_tol});

    # Expand wildcards in names of data files in directory
    # $$auto_ref{rundir}. All UNIX wildcards are allowed.

    my @expanded_list = ();
    foreach my $data (@{$data_ref}) {
	push(@expanded_list, glob("$$auto_ref{rundir}/$data"));
    }
    @{$data_ref} = @expanded_list;
    foreach my $data (@{$data_ref}) {
	$data =~ s/.*\///;
	$data =~ s/\s+//g;
    }
    
    # If wildcards are used, make sure that they expand to at least
    # one data file
    unless(@{$data_ref} > 0) {
	print "Wildcards in data lines fail to expand to at least one file\n";
	die "Errors found in auto3dem input file\n";
    }

}

sub valid_key{

    # Purpose: determine whether or not a key is considered to be
    # valid for a given hash
    #
    # Input parameters
    #   $ident - name of program (auto, p3dr, po2r, etc.)
    #   $key  - key being tested
    #
    # Return values
    #   1 if key is valid, 0 otherwise

    my ($ident, $key) = @_;

    my %hash;

    $hash{po2r} = {symm_code => 1, bin => 1, res_max => 1, res_min => 1,
		   zero_fill => 1, ctfmode => 1, filter => 1,
		   tempfac => 1, ctf_ff1 => 1, ctf_ff2 => 1,
		   dangle => 1, nangle => 1, mode => 1, gangle => 1,
		   funcmode => 1, funcweight => 1, dcenter => 1,
		   ncenter => 1, max_cpu => 1, quick_search => 1,
		   per_ptle_ctf => 1, handtest => 1, ticos_equiv => 1,
		   global_por => 1, nmagref =>1, magref_step =>1,
		   magref_reset =>1, magref_calibrate =>1};

    $hash{pctfr} = {bin => 1, res_max => 1, res_min => 1,
		   zero_fill => 1, ctfmode => 1, filter => 1,
		   tempfac => 1, ctf_ff1 => 1, ctf_ff2 => 1,
		   dangle => 1, nangle => 1,
		   funcmode => 1, funcweight => 1, max_cpu => 1};
    
    $hash{p3dr} = {symm_code => 1, bin => 1, res_max => 1, res_min =>
		   1, zero_fill => 1, ctfmode => 1, filter => 1,
		   tempfac => 1, ctf_ff1 => 1, ctf_ff2 => 1,
		   max_cpu => 1,
		   magfactor => 1, map_dim => 1, apo_border => 1,
		   per_ptle_ctf => 1};
    
    $hash{pcut} = {bin => 1, in_rad => 1, out_rad =>
		   1, cut_step => 1, cut_weight => 1, max_cpu => 1};
    
    $hash{psf} = {bin => 1, res_max => 1, res_min =>
		      1, res_step => 1, max_cpu => 1};
    
    $hash{ppft} = {symm_code => 1, bin => 1, input_mode => 1, 
		   bin_factor => 1, delta_theta => 1, ctfmode => 1,
		   verbose => 1, filter_factor_1 => 1, pftrad_lo => 1,
		   pftrad_hi => 1, pftrad_step => 1, annulus_low => 1,
		   annulus_high => 1, temperature_factor => 1,
		   resolution_low => 1, resolution_high => 1, jcut =>
		   1, sigcut => 1, mag_cen => 1, mag_step => 1,
		   mag_num => 1, mag_norm => 1, model_filename => 1,
		   prj_filename => 1, pft_filename => 1, quick_omega => 1,
		   per_ptle_ctf => 1, max_cpu => 1};

    $hash{auto} = {symm_code => 1,
		   mode => 1,
		   niter => 1, 
		   have_map => 1, 
		   start_map => 1,
		   rundir => 1, 
		   outfile => 1, 
		   delete_maps => 1, 
		   freeze_annulus => 1,
		   freeze_res => 1,
		   fsc_hithresh => 1, 
		   fsc_lothresh => 1, 
		   res_adj => 1, 
		   gauss_adj => 1,
		   theta_reject_lower => 1,
		   theta_reject_upper => 1,
		   phi_reject_lower => 1,
		   phi_reject_upper => 1,
		   omega1 => 1, omega1_tol => 1,
		   omega2 => 1, omega2_tol => 1,
		   box_center_offset => 1,
		   global_select => 1,
		   score_fraction => 1,
		   pft_cc_fraction => 1,
		   prj_cc_fraction => 1,
		   cmp_cc_fraction => 1,
		   score_nstd => 1,
		   pft_cc_nstd => 1,
		   prj_cc_nstd => 1,
		   cmp_cc_nstd => 1,
		   nselect_offset => 1,
		   select_delta => 1,
		   new_ptles => 1,
		   hollow_map => 1,
		   hollow_in_rad => 1,
		   hollow_out_rad => 1,
		   hollow_cut_step => 1,
		   hollow_cut_weight => 1,
		   estimate_res => 1,
		   restart => 1,
		   boxrad => 1,
		   partrad => 1,
	       iter_start => 1,
		   term_search => 1,
		   term_refine => 1,
		   switch_mode => 1,
	       recipient => 1,
           refine_ctf => 1,
	       quit_early => 1,
		   noctf => 1,
		   bin_reduce => 1,
		   per_ptle_ctf => 1,
		   noise_suppression => 1,
		   generate_map => 1,
		   map_suffix => 1,
		   adapt_angle => 1,
	       rmc => 1};
    
    $hash{all}  = {symm_code    => 1}; 
    
    if(defined $hash{$ident}{$key} || $ident eq 'data') {
	return 1;
    } else {
	return 0;
    }
}

sub deprecated_key{

    # Purpose: determine whether or not a key is considered to be
    # deprecated for a given hash
    #
    # Input parameters
    #   $ident - name of program (auto, p3dr, po2r, etc.)
    #   $key  - key being tested
    #
    # Return values
    #   1 if key is valid, 0 otherwise

    my ($ident, $key) = @_;

    my %hash;

    $hash{auto} = {dryrun => 1, 
	           split_ppft => 1,
		   final_map => 1
		   };
    $hash{p3dr}  = {view_perread => 1};
    $hash{po2r}  = {view_perread => 1,
                    ticos_equiv =>1,
                    global_por =>1,
                    };
    $hash{pcut}  = {symm_code => 1};
    $hash{psf}   = {symm_code => 1};
    $hash{pctfr} = {symm_code => 1};

    if(defined $hash{$ident}{$key} || $ident eq 'data') {
	return 1;
    } else {
	return 0;
    }
}
1;
