use strict;
use warnings;
use lib do { __FILE__ =~ m|^(.*?)[^/]*$|; "$1"; };

sub make_program_input {

    # Purpose: generate the input file for the programs PO2R, P3DR,
    # PSF, PPFT, and P3DR
    # 
    # Input arguments:
    #   $input_file - name of input file
    #   $program - name of program (po2r, p3dr, psf, ppft, or pcut)
    #   $param_ref - reference to hash of input parameters
    #   $datafiles_ref - reference to array of data files
    #
    # Return values:
    #   None

    my ($input_file, $program, $param_ref, $datafiles_ref) = @_;
    $program =~ tr/a-z/A-Z/; # Accept upper or lower case

    unless(defined $program) {
	print "Error creating input file - program not defined\n";
	exit(0);
    }

    open(INP, ">$input_file") || die "Unable to open file $input_file\n";

    # Write out input parameters
    if($program eq 'PO2R') {
	print INP "Mode ";
	print INP "$$param_ref{mode}\n";
	print INP "SYMM ";
	print INP "$$param_ref{symm_code}\n";
	print INP "Resmax ";
	print INP "$$param_ref{res_max}, ";
	print INP "$$param_ref{res_min}\n";
	print INP "Zero_fill ";
	print INP "$$param_ref{zero_fill}\n";
	print INP "Ctfmode ";
	print INP "$$param_ref{ctfmode}, ";
	print INP "$$param_ref{filter}, ";
	print INP "$$param_ref{tempfac}, ";
	print INP "$$param_ref{ctf_ff1}, ";
	print INP "$$param_ref{ctf_ff2}\n";
	print INP "In3dfilename ";
	print INP "$$param_ref{in_3d_name}\n";
	print INP "Glob_anglestep ";
	print INP "$$param_ref{gangle}\n";
	print INP "Dangle ";
	print INP "$$param_ref{dangle}, ";
	print INP "$$param_ref{nangle}, ";
	print INP "$$param_ref{funcmode}, ";
	print INP "$$param_ref{funcweight}\n";
	print INP "Dcenter ";
	print INP "$$param_ref{dcenter}, ";
	print INP "$$param_ref{ncenter}\n";
	print INP "Quick_search ";
	print INP "$$param_ref{quick_search}\n";
	print INP "per_ptle_ctf ";
	print INP "$$param_ref{per_ptle_ctf}\n";
	print INP "handtest ";
	print INP "$$param_ref{handtest}\n";
	print INP "magref ";
	print INP "$$param_ref{nmagref}, ";
	print INP "$$param_ref{magref_step}, ";
	print INP "$$param_ref{magref_reset}, ";
	print INP "$$param_ref{magref_calibrate}\n";
    } elsif ($program eq 'PCTFR') {
	print INP "Resmax ";
	print INP "$$param_ref{res_max}, ";
	print INP "$$param_ref{res_min}\n";
	print INP "Zero_fill ";
	print INP "$$param_ref{zero_fill}\n";
	print INP "Ctfmode ";
	print INP "$$param_ref{ctfmode}, ";
	print INP "$$param_ref{filter}, ";
	print INP "$$param_ref{tempfac}, ";
	print INP "$$param_ref{ctf_ff1}, ";
	print INP "$$param_ref{ctf_ff2}\n";
	print INP "In3dfilename ";
	print INP "$$param_ref{in_3d_name}\n";
	print INP "Dangle ";
	print INP "$$param_ref{dangle}, ";
	print INP "$$param_ref{nangle}, ";
	print INP "$$param_ref{funcmode}, ";
	print INP "$$param_ref{funcweight}\n";
    } elsif ($program eq 'P3DR') {
	print INP "SYMM ";
	print INP "$$param_ref{symm_code}\n";
	print INP "Resmax ";
	print INP "$$param_ref{res_max}, ";
	print INP "$$param_ref{res_min}\n";
	print INP "Zero_fill ";
	print INP "$$param_ref{zero_fill}\n";
	print INP "Ctfmode ";
	print INP "$$param_ref{ctfmode}, ";
	print INP "$$param_ref{filter}, ";
	print INP "$$param_ref{tempfac}, ";
	print INP "$$param_ref{ctf_ff1}, ";
	print INP "$$param_ref{ctf_ff2}\n";
	print INP "Oddeven ";
	print INP "$$param_ref{oddeven}\n";
	print INP "outfilename ";
	print INP "$$param_ref{out_file_name}\n";
	print INP "Mapsize ";
	print INP "$$param_ref{magfactor}, ";
	print INP "$$param_ref{map_dim}\n";
	print INP "Apo_border ";
	print INP "$$param_ref{apo_border}\n";
	print INP "per_ptle_ctf ";
	print INP "$$param_ref{per_ptle_ctf}\n";
	if(defined $$param_ref{fsc_file_name}) {
	    print INP "fscfilename ";
	    print INP "$$param_ref{fsc_file_name}\n";
	}
    } elsif ($program eq 'PSF') {
	print INP "$$param_ref{fileeven}\n";
	print INP "$$param_ref{fileodd}\n";
	print INP "$$param_ref{res_max}, ";
	print INP "$$param_ref{res_min}, ";
	print INP "$$param_ref{res_step}, ";
	print INP "$$param_ref{pixel_size}\n";
    } elsif ($program eq 'PCUT') {
	print INP "$$param_ref{in_3d_name}\n";
	print INP "$$param_ref{out_file_name}\n";
	print INP "$$param_ref{in_rad}, ";
	print INP "$$param_ref{out_rad}, ";
	print INP "$$param_ref{cut_step}, ";
	print INP "$$param_ref{cut_weight}\n";
    } elsif ($program eq 'PPFT') {
	print INP "input_mode         = $$param_ref{input_mode}\n";
	print INP "quick_omega        = $$param_ref{quick_omega}\n";
	print INP "bin_factor	      = $$param_ref{bin_factor}\n";
	print INP "symmetry	      = $$param_ref{symm_code}\n";
	print INP "delta_theta	      = $$param_ref{delta_theta}\n";
	print INP "ctf_mode	      = $$param_ref{ctfmode}\n";
	print INP "verbose            = $$param_ref{verbose}\n";
	print INP "filter_factor_1    = $$param_ref{filter_factor_1}\n";
	print INP "pftrad_lo	      = $$param_ref{pftrad_lo}\n";
	print INP "pftrad_hi	      = $$param_ref{pftrad_hi}\n";
	print INP "pftrad_step	      = $$param_ref{pftrad_step}\n";
	print INP "annulus_low	      = $$param_ref{annulus_low}\n";
	print INP "annulus_high	      = $$param_ref{annulus_high}\n";
	print INP "temperature_factor = $$param_ref{temperature_factor}\n";
	print INP "resolution_low     = $$param_ref{resolution_low}\n";
	print INP "resolution_high    = $$param_ref{resolution_high}\n";
	print INP "jcut		      = $$param_ref{jcut}\n";
	print INP "sigcut             = $$param_ref{sigcut}\n";
	print INP "mag_cen            = $$param_ref{mag_cen}\n";
	print INP "mag_step           = $$param_ref{mag_step}\n";
	print INP "mag_num            = $$param_ref{mag_num}\n";
	print INP "mag_norm           = $$param_ref{mag_norm}\n";
	print INP "model_filename     = $$param_ref{model_filename}\n";
	print INP "prj_filename       = $$param_ref{prj_filename}\n";
	print INP "pft_filename       = $$param_ref{pft_filename}\n";
	print INP "pftrads_filename   = $$param_ref{pftrads_filename}\n";
	print INP "pftres1_filename   = $$param_ref{pftres1_filename}\n";
	print INP "pftres2_filename   = $$param_ref{pftres2_filename}\n";
	print INP "per_ptle_ctf       = $$param_ref{per_ptle_ctf}\n";
    } else {
	print "Error creating input file - ";
	print "$program not a valid choice\n";
	exit(0);
    }

    # Write out list of data files (PO2R, P3DR, PCTFR, and PPFT only)
    if($program eq 'PO2R'  ||
       $program eq 'P3DR'  ||
       $program eq 'PCTFR') {
	print INP "Inputparameterfiles\n";
    }
    if($program eq 'PPFT') {
	print INP "END_OF_KEYS\n";
    }
    if($program eq 'PO2R'  ||
       $program eq 'P3DR'  ||
       $program eq 'PPFT'  ||
       $program eq 'PCTFR') {
	for my $datafile (@{$datafiles_ref}) {
	    print INP "$datafile\n";
	}
    }

    close(INP);
}
1;
