use strict;
use warnings;
use lib do { __FILE__ =~ m|^(.*?)[^/]*$|; "$1"; };

sub psf_parse {

    # Purpose: processes the output file from program PSF to determine
    # the resolutions where the Fourier shell coefficient (FSC) drops
    # below two specified thresholds. These are used to determine the
    # resolution at this stage of reconstruction and the suggested
    # high resolution to use in the next iteration. Since the FSC is
    # calculated at discrete points, we err on the side of caution and
    # choose the highest resolution BEFORE the cutoff is reached.
    #
    # Input parameters:
    #   $fsc_hithresh - upper FSC threshold
    #   $fsc_lothresh - lower FSC threshold
    #   $infile    - output file from PSF
    #
    # Return values:
    #   $res_hithresh - resolution where FSC drops below $fsc_hithresh
    #   $res_lothresh - resolution where FSC drops below $fsc_lothresh
    #   $reached_hithresh - set to 1 if FSC drops below $fsc_hithresh,
    #   $reached_lothresh - set to 1 if FSC drops below $fsc_lothresh,
    #   \@spatial_freq  - reference to array of spatial frequencies
    #   \@fsc - reference to array of FSCs

    my ($fsc_hithresh, $fsc_lothresh, $infile) = @_;

    my $date = localtime();

    print OUT "Parsing PSF output file\n";
    print OUT "   Date = $date\n";

    # Define flags
    my $reached_hithresh = 0;  # FSC has not reached fsc_hithresh
    my $reached_lothresh = 0;  # FSC has not reached fsc_lothresh

    # Define resolutions
    my $res_hithresh; # Resolution at the high FSC threshold
    my $res_lothresh; # Resolution at the low FSC threshold

    # Define variable holding resolution from previous record
    my $prev_res = 1000; # Initialize to ridiculously high value

    # Initialize accumulator for sum over FSC
    my $fsc_sum = 0.0;

    # Initialize arrays
    my @spatial_freq = ();
    my @fsc = ();
    
    open(IN, $infile);
    while(<IN>) {
	
	# Get rid of leading and trailing whitespace
	s/^\s+//;
	s/\s+$//;
	
	# Determine whether or not we reached the resolution data
	# Lines should split on whitespace into 8 numeric fields
	my ($bin, $res, $obs, $r_sf, $ccf, $fsc,
	    $r_phase, $phase_diff, $extra) = split(/\s+/, $_); 
	
	unless(defined $phase_diff && $bin =~ /^\d+$/) {
	    next; # Not in data section - not enough fields
	}
	unless($bin =~ /^\d+$/) {
	    next; # Not in data section - first field not numeric
	}
	if(defined $extra) {
	    next; # Not in data section - too many fields
	}

	# Accumulate arrays containing spatial frequency and fsc
	my $spatial_freq = 1.0/$res;
	push(@spatial_freq, $spatial_freq);
	push(@fsc, $fsc);

	# Accumulate sum over FSC
	$fsc_sum += $fsc;

	# Identify the point where CC drops below high cutoff
	unless($reached_hithresh) {
	    if($fsc < $fsc_hithresh) {
	        if(scalar(@fsc)>1) {
		        $res_hithresh = 1.0/($spatial_freq[-2]+($fsc_hithresh-$fsc[-2])*($spatial_freq[-1]-$spatial_freq[-2])/($fsc[-1]-$fsc[-2]));
            } else {
        		$res_hithresh = $prev_res;
    	    }
		    $reached_hithresh = 1;
	    }
	}
	
	# Identify the point where CC drops below low cutoff
	unless($reached_lothresh) {
	    if($fsc < $fsc_lothresh) {
	        if(scalar(@fsc)>1) {
		        $res_lothresh = 1.0/($spatial_freq[-2]+($fsc_lothresh-$fsc[-2])*($spatial_freq[-1]-$spatial_freq[-2])/($fsc[-1]-$fsc[-2]));
            } else {
		        $res_lothresh = $prev_res;
    	    }
     		$reached_lothresh = 1;
	    }
	}

	# Save the resolution from this record to use in next iteration
	$prev_res = $res;
    }

    # If CC never drops below $fsc_lothresh, set $res_lothresh to the
    # highest resolution evaluated in PSF. Do same in the event that
    # CC never drops below $fsc_hithresh.
    unless($reached_lothresh) {
	print OUT "  Note: FSC never drops below $fsc_lothresh\n";
	print OUT "        Using highest resolution evaluated in PSF\n\n";
	$res_lothresh = $prev_res;
    }
    unless($reached_hithresh) {
	print OUT "  Note: FSC never drops below $fsc_hithresh\n";
	print OUT "        Using highest resolution evaluated in PSF\n\n";
	$res_hithresh = $prev_res;
    }
    
    print OUT " Resolution where FSC drops below $fsc_hithresh: ";
    print OUT "$res_hithresh\n";

    print OUT " Resolution where FSC drops below $fsc_lothresh: ";
    print OUT "$res_lothresh\n";

    print OUT "   Sum over FSC: $fsc_sum\n";
    print OUT "\n";
    
    close(IN);

    return($res_hithresh, $res_lothresh, $reached_hithresh,
	   $reached_lothresh, \@spatial_freq, \@fsc);
} 
1;
