#!/usr/bin/perl
use strict;
use warnings;
use lib do { __FILE__ =~ m|^(.*?)[^/]*$|; "$1"; };

sub rad_profile_parse {

    # Purpose: parse a radial density profile file generated by the
    # program imgstats and determine reasonable values for the inner
    # and outer radii of the capsid. Note that records starting with
    # "#" are skipped and the data lines are always assumed to have
    # the following format, where the additional fields may be subject
    # to change
    #
    # r  n  avg  std  [additional fields]
    #
    # r    - radius of annulus (pixels)
    # n    - number of voxels in radial shell
    # avg  - average voxel value
    # std  - standard deviation of voxel values
    #
    # Input parameters:
    #   $infile  - file containing radial density profile
    #
    # Return values:
    #   $rad_in  - inner capsid radius
    #   $rad_out - outer capsid radius

    my ($infile) = @_;

    my $date = localtime();

    my @avg = ();
    my @avg3pt = ();
    my @std = ();
    my @ravgstd = ();
    my @nstd = ();
    my ($ravgstd_maxval, $ravgstd_maxloc, $count) = (0, 0, 0);
    my ($nstd_maxval, $nstd_maxloc) = (0, 0);
    my ($rad_in, $rad_out);

    # Loop through records and find maximum value of r*avg*std and sqrt(n)*std

    open(INSTAT, $infile);
    while(<INSTAT>) {
	if (/\s*#/) {
	    next;
	}

	s/^\s+//;
	my (undef, $nvxl, $avg, $std) = split(/\s+/);
	unless(defined $std) {
	    next;
	}
	
	my $ravgstd = $count*$avg*$std;
	my $nstd = sqrt($nvxl)*$std;
	
	push(@avg, $avg);
	push(@std, $std);
	push(@ravgstd, $ravgstd);
	push(@nstd, $nstd);

	if($ravgstd > $ravgstd_maxval) {
	    $ravgstd_maxval = $ravgstd;
	    $ravgstd_maxloc = $count;
	}

	$count++;
    }
    
    # Get moving 3-point average of avg annulus value
    for (my $i=1; $i<$count-1; $i++) {
	$avg3pt[$i] = ($avg[$i-1] + $avg[$i] + $avg[$i+1])/3.0;
    }

    # Estimating outer capsid radius: Starting from exterior of
    # particle, look for first point where avg*std exceeds 1/10th
    # maximum value.

    for (my $i=$count-1; $i>$ravgstd_maxloc; $i--) {
	if ($ravgstd[$i] > $ravgstd_maxval*0.1) {
	    $rad_out = $i;
	    last;
	}
    }

    # RSS kludge - scale rad_out by 5%
    $rad_out *= 1.05;

    my $rad_in_1 = 1;
    my $rad_in_2 = 1;

    # First estimate of inner capsid radius: Start from maxloc and
    # search inward until average drops to avg[maxloc]/2

    for (my $i=$ravgstd_maxloc; $i>0; $i--) {
	if ($avg[$i] < $avg[$ravgstd_maxloc]/2) {
	    $rad_in_1 = $i;
	    last;
	}
    }

    # Second estimate of inner capsid radius: Start from maxloc and
    # search inward until dip found in 3-point average of avg annulus
    # pixel value

    my $reached_peak = 0;
    for (my $i=$ravgstd_maxloc; $i>1; $i--) {
	if (!$reached_peak && $avg3pt[$i-1] < $avg3pt[$i]) {
	    $reached_peak = 1;
	}
	if($reached_peak && $avg3pt[$i-1] > $avg3pt[$i]) {
	    $rad_in_2 = $i;
	    last;
	}
    }

    if($rad_in_1 > $rad_in_2) {
	$rad_in = $rad_in_1;
    } else {
	$rad_in = $rad_in_2;
    }

	
    # Find the maximum sqrt(n)*stad_dev, starting from rad_in
    # This is the new capsid score
    for (my $i=$rad_in; $i<$count-1; $i++) {
	   if ($nstd[$i] > $nstd_maxval) {
	     $nstd_maxval = $nstd[$i];
	     $nstd_maxloc = $i;
	   }
    }

    print OUT "Parsing imgstats radial density file\n";
    print OUT "   reached_peak = $reached_peak\n";
    print OUT "   rad_in_1  = $rad_in_1\n";
    print OUT "   rad_in_2  = $rad_in_2\n";
    print OUT "   rad_in   = $rad_in\n";
    print OUT "   rad_out  = $rad_out\n";
    print OUT "\n";

    close(INSTAT);

# previous metric
#    return($rad_in, $rad_out, $ravgstd_maxval);
    return($rad_in, $rad_out, $ravgstd_maxloc, $nstd_maxval);
}
1;
