use strict;
use warnings;
use lib do { __FILE__ =~ m|^(.*?)[^/]*$|; "$1"; };

use mode;

sub run_mpi_prog {

    # Purpose: runs program using mpirun. Since we are using the perl
    # system() operator, the program will wait for the job to finish
    # before proceeding.
    #
    # Input arguments
    #   $prog     - name of executable
    #   $infile   - input file
    #   $ncpu     - number of CPUs to use
    #   $outfile  - output file (redirection of stdout)
    #   $nodefile - file containing list of nodes
    #
    # Return values:
    #   none

    my ($launch_cmd, $prog, $infile, $ncpu, $outfile) = @_;

    # Following required to make sure that executable is found
    $prog = `which $prog`;
    chomp($prog);

    my $short_prog = $prog;
    $short_prog =~ s/^.*\///;

    my $time_start = time();

    # Construct full command for launching job and substitute actual
    # number of CPUs for the placeholder string "NCPU"
    my $cmd = "$launch_cmd $prog < /dev/null $infile > $outfile";
    $cmd =~ s/NCPU/$ncpu/;

    # Exit if input file does not exist or is not readable
    unless(-e $infile && -r $infile) {
	print OUT "Unable to launch job\n$cmd\n";
	print OUT "Input file $infile does not exist or is not readable\n";
    }

    # Remove output file if it already exists
    if(-e $outfile) {
	unlink $outfile;
    }

    # Launch job and exit if it does not run properly to completion
    system($cmd);
    my $exit_value  = $? >> 8;
    if($exit_value) {
	print "\n\n----- Exiting AUTO3DEM -----\n\n";
	print "Abnormal termination by following program\n";
	print "$prog\n\n";
	print "Full command executed by auto3dem\n";
	print "$cmd\n\n";
	print "Exit value = $exit_value\n\n";
	print "----- Exiting AUTO3DEM -----\n\n";

	print OUT "\n\n----- Exiting AUTO3DEM -----\n\n";
	print OUT "Abnormal termination by following program\n";
	print OUT "$prog\n\n";
	print OUT "Full command executed by auto3dem\n";
	print OUT "$cmd\n\n";
	print OUT "Exit value = $exit_value\n\n";
	print OUT "----- Exiting AUTO3DEM -----\n\n";

	exit(0)
    }

    # Depending on how parallel jobs are launched, the return value
    # from system() may not be accesible to auto3dem. Parse output
    # file to look for sign of abnormal termination

    open(OUTFILE, $outfile);
    while(<OUTFILE>) {
	if(/mpi.*abort/i || /abnormal program termination/i ||
	   /abort of all ranks/i || /collective abort/i) {

	    print "\n\n----- Exiting AUTO3DEM -----\n\n";
	    print "Abnormal termination by following program\n";
	    print "$prog\n\n";
	    print "Full command executed by auto3dem\n";
	    print "$cmd\n\n";
	    print "----- Exiting AUTO3DEM -----\n\n";
	    
	    print OUT "\n\n----- Exiting AUTO3DEM -----\n\n";
	    print OUT "Abnormal termination by following program\n";
	    print OUT "$prog\n\n";
	    print OUT "Full command executed by auto3dem\n";
	    print OUT "$cmd\n\n";
	    print OUT "----- Exiting AUTO3DEM -----\n\n";
	    
	    exit(0);
	}
    }
    close(OUTFILE);

    my $time_stop = time();
    my $time_elapsed = $time_stop - $time_start;

    print OUT "$cmd\n";
    print OUT "Running $short_prog on $ncpu CPUs\n";
    print OUT "   input file  = $infile\n";
    print OUT "   output file = $outfile\n";
    print OUT "   run time    = $time_elapsed seconds\n";
    print OUT "\n";
}
1;
