use strict;
use warnings;
use lib do { __FILE__ =~ m|^(.*?)[^/]*$|; "$1"; };

sub radii_check {

    # Purpose: Make sure that radii defined for PCUT and PPFT are
    # properly orderd (e.g. inner radius < outer radius) and do not
    # exceed the box radius.
    #
    # Input argument
    #   $auto_ref   - reference to hash of control parameters
    #   $pcut_ref   - reference to hash of PCUT input parameters
    #   $ppft_ref   - reference to hash of PSF input parameters
    #
    # Return value
    #   \@warnings - reference to array of warnings
    #   \@errors   - reference to array of errors
    
    my ($auto_ref, $pcut_ref, $ppft_ref) = @_;
    
    my @warnings = ();
    my @errors = ();

    if($$pcut_ref{in_rad} > $$auto_ref{boxrad}) {
	my $s = "PCUT in_rad ($$pcut_ref{in_rad})  >  ";
	$s .= "box radius ($$auto_ref{boxrad})";
	push(@errors, $s);
    }
    
    if($$pcut_ref{out_rad} > $$auto_ref{boxrad}) {
	my $s = "PCUT out_rad ($$pcut_ref{out_rad})  >  ";
	$s .= "box radius ($$auto_ref{boxrad})";
	push(@errors, $s);
    }
    
    if($$pcut_ref{in_rad} >= $$pcut_ref{out_rad}) {
	my $s = "PCUT in_rad ($$pcut_ref{in_rad})  >=  ";
	$s .= "PCUT out_rad ($$pcut_ref{out_rad})";
	push(@errors, $s);
    }
    
    if($$ppft_ref{annulus_low} > $$auto_ref{boxrad}) {
	my $s = "PPFT annulus_low ($$ppft_ref{annulus_low})  >  ";
	$s .= "box radius ($$auto_ref{boxrad})";
	push(@errors, $s);
    }
    
    if($$ppft_ref{annulus_high} > $$auto_ref{boxrad}) {
	my $s = "PPFT annulus_high ($$ppft_ref{annulus_high})  >  ";
	$s .= "box radius ($$auto_ref{boxrad})";
	push(@errors, $s);
    }
    
    if($$ppft_ref{annulus_low} >= $$ppft_ref{annulus_high}) {
	my $s = "PPFT annulus_low ($$ppft_ref{annulus_low})  >=  ";
	$s .= "PPFT annulus_high ($$ppft_ref{annulus_high})";
	push(@errors, $s);
    }
    
    if($$ppft_ref{pftrad_hi} > $$auto_ref{boxrad}) {
	my $s = "PPFT pftrad_hi ($$ppft_ref{pftrad_hi})  >  ";
	$s .= "box radius ($$auto_ref{boxrad})";
	push(@errors, $s);
    }
    
    return(\@warnings, \@errors);
}


sub param_file_check {
    
    # Purpose: perform quick checks on the particle parameter files to
    # make sure that header information is correct.
    #
    # Input argument
    #   $file - name of particle parameter file
    #   $noctf - flag to indicate that no CTF corrections used
    #
    # Return value
    #   \@warnings - reference to array of warnings
    #   \@errors   - reference to array of errors
    
    my ($file, $noctf) = @_;
    
    my @warnings = ();
    my @errors = ();
    
    open(FL, $file);
    my $header1 = <FL>;
    my $header2 = <FL>;
    close(FL);
    
    #### Test 1st header line
    
    # Make sure that 1st line is not blank
    unless(defined $header1 && $header1 =~ /\w/) {
	push(@errors, "Missing 1st line");
	return(\@warnings, \@errors);
    }
    
    # Make sure that 2nd line is not blank
    unless(defined $header2 && $header2 =~ /\w/) {
	push(@errors, "Missing 2nd line");
	return(\@warnings, \@errors);
    }
    
    #### Test the 2nd header line
    $header2 =~ s/^\s+//;  # Get rid of leading space
    $header2 =~ s/,/ /g;   # Convert commas to whitespace
    
    my ($pixsiz, $units, $volts, $amp_fac, $focus_maj, $focus_min,
	$ang_maj, $cs_coeff) = split(/\s+/, $header2);
    
    # Make sure that all required fields are present
    unless(defined $pixsiz    &&  defined $units     && 
	   defined $volts     &&  defined $amp_fac   && 
	   defined $focus_maj &&  defined $focus_min && 
	   defined $ang_maj   &&  defined $cs_coeff) {
	push(@errors, "Missing fields in 2nd line");
	return(\@warnings, \@errors);
    }

    # Test the individual fields in the 2nd line and accumulate error
    # and warning messages
    
    # Tests on pixel size
    if(isdecimal($pixsiz)) {
	if($pixsiz < 0.3 || $pixsiz > 15.0) {
	    push(@errors, "Pixel size must be in range 0.3-15.0 Angstroms");
	} elsif ($pixsiz > 10.0) {
	    push(@warnings, "Pixel size suspect, greater than 10.0 Angstroms");
	} 
    } else {
	push(@errors, "Pixel size is not a numerical value");
    }
    
    # Tests on units
    if(isdecimal($units)) {
	unless($units == 1 || $units == 2) {
	    push(@errors, "Units must be equal to 1 or 2");
	} 
    } else {
	push(@errors, "Units is not a numerical value");
    }
       
    # Tests on voltage
    if(isdecimal($volts)) {
	if($volts <= 50000) {
	    push(@errors, "Voltage must be greater than 50,000");
	}
    } else {
	push(@errors, "Voltage is not a numerical value");
    }
       
    # Tests on F_amp
    if(isdecimal($amp_fac)) {
	if($amp_fac <= 0.0 || $amp_fac > 1.0) {
	    push(@errors, "Amp fraction must be in range 0 < Famp <= 1");
	}
    } else {
	push(@errors, "Amp factor is not a numerical value");
    }

    # Perfom tests on defocus values only if CTF corrections are used
    unless($noctf) {
	# Tests on focus major
	if(isdecimal($focus_maj)) {
	    if($focus_maj < 0 || $focus_maj > 15.0) {
		push(@errors, 
		     "Focus major must be in range 0-15 um");
	    }
	} else {
	    push(@errors, "Focus major is not a numerical value");
	} 
	
	# Tests on focus minor
	if(isdecimal($focus_min)) {
	    if($focus_min < 0 || $focus_min > 15.0) {
		push(@errors, 
		     "Focus major must be in range 0-15 um");
	    } 
	} else {
	    push(@errors, "Focus major is not a numerical value");
	} 
    }

    # Tests on angle major
    unless(isdecimal($ang_maj)) {
	push(@errors, "Angle major is not a numerical value");
    }
	
    # Tests on Cs coefficient
    if(isdecimal($cs_coeff)) {
	if($cs_coeff < 0.4 || $cs_coeff > 10.0) {
	    push(@errors, "Cs coeeficient must be in range 0.4-10.0 mm");
	}
    } else {
	push(@errors, "Cs coefficient is not a numerical value");
    }

    return(\@warnings, \@errors)
}


sub isdecimal {

    # Purpose: test a string to see if it is a valid decimal
    # number. This is probably not the most efficient implementation,
    # but should be very robust since it allows for integers and
    # floating point numbers with digits before decimal, after
    # decimal, or both. Also allows for leading signs and trailing
    # scientific notation.
    #
    # After stripping off leading sign and trailing scientific
    # notation, strings that contain anything other than the digits
    # 0-9 and a single optional decimal point or that start with more
    # than one leading zero will be rejected.
    #
    # Returns true value (1) if string is a valid representation of a
    # decimal number and false (0) otherwise.

    # NOTE - I can't believe that this is not a standard Perl function
    # since Perl needs to convert strings to numbers all the time. I
    # can't find this functionality in CPAN either, but I probably
    # just didn't look in the right place.

    my ($x) = @_;

    # Trim off trailing exponential notation
    $x =~ s/[dDeE][+-]?\d+$//;

    # Trim off leading sign
    $x =~ s/^[+-]//;

    # Remaining string must be of one of the following forms, where x
    # represensts a string of digits
    # (1) x  (2) x.  (3) .x  (4) x.x
    unless($x =~ /^\d+$/   || $x =~ /^\d+\.$/ ||
	   $x =~ /^\.\d+$/ || $x =~ /^\d+\.\d+$/) {
	return(0);
    }

    # Test to make sure that we don't have more than one leading zero
    # before the decimal point. (e.g. 00.01) 
    $x =~ s/\..*//; 
    if($x =~/^00/) { 
	return(0);
    }

    # If we made it this far, must have a valid decimal number
    return(1);
}
1;
