use strict;
use warnings;
use lib do { __FILE__ =~ m|^(.*?)[^/]*$|; "$1"; };

sub get_stable_ptles {

    my ($ptlefiles_ref) = @_;
    my @selectfiles = ();
    my $total_select = 0;
    my $header_flag = 0;

    foreach my $ptlefile (@{$ptlefiles_ref}) {
	my $selectfile = "${ptlefile}_sel";
	open(PTLE, $ptlefile);
	open(SEL, ">$selectfile");
	while(<PTLE>) {
	    my $origline = $_;
	    my $modline = $origline;
	    $modline =~ s/^\s+//;
	    $modline =~ s/\s+$//;

	    unless($origline =~ /\w/) {
		next; # skipping blank lines
	    }

	    if($_ =~ /pif/i || $_ =~ /mrc/i) {
		print SEL "$origline";
		$header_flag = 1;
	    }
	    elsif($header_flag) {
		print SEL "$origline";
		$header_flag = 0;
	    } else {
		$modline =~ s/,/ /g;
		my ($indx, $theta, $phi, $omega, $x, $y,
		    $mag, $score1, $score2, $score3, 
		    $t1, $t2, $t3, $t4, $t5) = split(/\s+/, $modline);

		if(defined $t4 && defined $t5 &&
		   $t4 =~ /^[a-z]$/i && $t5 =~ /^[a-z]$/i) {
		    if($t4 =~ /S/i && $t5 =~ /S/i) {
			print SEL "$origline";
			$total_select++;
		    } 
		}
		elsif (defined $t1 && defined $t2 &&
		   $t1 =~ /^[a-z]$/i && $t2 =~ /^[a-z]$/i) {
		    if($t1 =~ /S/i && $t2 =~ /S/i) {
			print SEL "$origline";
			$total_select++;
		    } 
		}
		else {
		    print SEL "$origline";
		    $total_select++;
		}

	    }
	}
	close(PTLE);
	close(SEL);

	push(@selectfiles, $selectfile);
    }
    
    return ($total_select, @selectfiles);
}

sub gen_filtered_ptles {

    # Purpose: generate filtered particle parameter files by applying
    # verious selection criteria based on orientations, origins, and
    # values of scores (PO2R output) or correlation coefficients
    # (PFTsearch output)
    # 
    # Input arguments:
    #   $ptlefiles_ref - reference to array of ptle files
    #   $auto_ref - reference to AUTO3DEM parameter hash
    #   $nshift - number of deltas to shift cutoff by
    #
    # Return values:
    #   $total_select - total number of particles selected
    #   $select_ref   - reference to array of selected particle files
    #   $reject_ref   - reference to array of rejected particle files

    my ($ptlefiles_ref, $auto_ref, $nshift) = @_;

    my ($score_cutoff, $pft_cc_cutoff, 
	$prj_cc_cutoff, $cmp_cc_cutoff,
	$pftsearch_gen);
    my $select_ref;
    my $reject_ref;
    my $total_select = 0;

    # If global_select option is chosen, determine score/CC cutoffs on
    # the basis of statistics collected across all files
    if($$auto_ref{global_select}) {
	($score_cutoff, $pft_cc_cutoff, 
	 $prj_cc_cutoff, $cmp_cc_cutoff,
	 $pftsearch_gen) = 
	     get_cutoffs($auto_ref, $nshift, @{$ptlefiles_ref});	
    }
    
    # Loop over the parameter files
    foreach my $ptlefile (@{$ptlefiles_ref}) {
	unless(-r $ptlefile) {
	    print OUT "NOTE - unable to open $ptlefile for filtering\n";
	    print OUT "\n";
	    next;
	}

	# If global_select option is not chosen, determine score/CC
	# cutoffs on the basis of statistics collected individually
	# from each file
	unless($$auto_ref{global_select}) {
	    ($score_cutoff, $pft_cc_cutoff, 
	     $prj_cc_cutoff, $cmp_cc_cutoff,
	     $pftsearch_gen) = 
		 get_cutoffs($auto_ref, $nshift, $ptlefile);
	}

	# Apply selection criteria and generate filtered file
	my $selectfile = "${ptlefile}_sel";
	my $rejectfile = "${ptlefile}_rej";
	my $nselect = select_ptles($ptlefile, $selectfile, 
				   $rejectfile, $auto_ref, 
				   $score_cutoff, $pft_cc_cutoff,
				   $prj_cc_cutoff, $cmp_cc_cutoff,
				   $pftsearch_gen);
	push(@{$select_ref}, $selectfile);
	push(@{$reject_ref}, $rejectfile);
	$total_select += $nselect;
    }
    
    return ($total_select, $select_ref, $reject_ref);
}


sub get_cutoffs {

    # Purpose: determine the cutoffs to be used on the PO2R score or
    # PFTsearch correlation coefficients during particle
    # selection. Also calculates and prints the statistics for the
    # score or coefficients
    #
    # When using CC or score cutoffs, can specify:
    # (1) Fraction of particles to be selected
    # (2) Number of particles to be selected
    # (3) A cutoff of (avg + N * std_dev) to be used
    #
    # The selection of particles is tweaked by nshift*delta, with the
    # usage dependent on the context
    #
    # (1) Fraction of ptles - shifts by the fraction selected
    # (2) Number of ptles - shifts by number selected
    # (3) Number of stddevs - shifts by N*std_dev
    #
    # Input arguments:
    #   $auto_ref - reference to AUTO3DEM parameter hash
    #   $nshift - number of deltas to shift cutoff by
    #   @ptlefiles - array of particle parameter files
    #
    # Return values:
    #   $score_cutoff - score cutoff for PO2R generated file
    #   $pft_cc_cutoff - PFT CC cutoff for PPFT generated file 
    #	$prj_cc_cutoff - PRJ CC cutoff for PPFT generated file 
    #   $cmp_cc_cutoff - CMP CC cutoff for PPFT generated file 
    #   $pftsearch_gen - flag set to 1 if file was generated by pftsearch

    my($auto_ref, $nshift, @ptlefiles) = @_;

    my ($pft_cc_sum,  $prj_cc_sum,  $cmp_cc_sum)  = (0, 0, 0);
    my ($pft_cc2_sum, $prj_cc2_sum, $cmp_cc2_sum) = (0, 0, 0);
    my ($pft_cc_avg, $prj_cc_avg, $cmp_cc_avg)    = (0, 0, 0);
    my ($pft_cc_std, $prj_cc_std, $cmp_cc_std)    = (0, 0, 0);
    my ($score_cutoff, $pft_cc_cutoff, $prj_cc_cutoff, $cmp_cc_cutoff);
    my @pft_cc = ();
    my @prj_cc = ();
    my @cmp_cc = ();
    my @score  = ();
    my $nptles = 0;
    my $header_flag = 0;

    # Calculate sums and sums of squares for score/CCs. Also, fill
    # arrays containing scores/CCs

    foreach my $ptlefile (@ptlefiles) {
	open(PTLE, $ptlefile) || die
	    "Unable to open file $ptlefile for reading in select_ptles\n";

	while(<PTLE>) {

	    unless($_ =~ /\w/) {
		next; # skipping blank lines
	    }

	    if($_ =~ /pif/i || $_ =~ /mrc/i) {
		$header_flag = 1;
		next;
	    }
	    if($header_flag) {
		$header_flag = 0;
		next;
	    }
	    
	    # Get rid of whitespace and split line into fields
	    s/\s+//g;
	    my ($id, $theta, $phi, $omega, 
		$xcent, $ycent, $mag, $pft_cc,
		$prj_cc, $cmp_cc) = split(/,/);
	    
	    # Calculate sums and sum of squares
	    unless($pft_cc =~ /\d/) {
		$pft_cc = 0;
	    }
	    unless($prj_cc =~ /\d/) {
		$prj_cc = 0;
	    }
	    unless($cmp_cc =~ /\d/) {
		$cmp_cc = 0;
	    }

	    push(@pft_cc, $pft_cc);
	    push(@prj_cc, $prj_cc);
	    push(@cmp_cc, $cmp_cc);

	    $pft_cc_sum  += $pft_cc;
	    $pft_cc2_sum += $pft_cc*$pft_cc;
	    $prj_cc_sum  += $prj_cc;
	    $prj_cc2_sum += $prj_cc*$prj_cc;
	    $cmp_cc_sum  += $cmp_cc;
	    $cmp_cc2_sum += $cmp_cc*$cmp_cc;
	    $nptles++;
	}
	close(PTLE);
    }

    # Calculate averages and standard deviations of CC's
    if($nptles > 0) {
	$pft_cc_avg = $pft_cc_sum/$nptles;
	$prj_cc_avg = $prj_cc_sum/$nptles;
	$cmp_cc_avg = $cmp_cc_sum/$nptles;
    }

    if($nptles > 1 && ($pft_cc2_sum - $nptles*$pft_cc_avg**2) > 0) {
	$pft_cc_std = sqrt(($pft_cc2_sum - $nptles*$pft_cc_avg**2)/
			   ($nptles-1));
    }
    
    if($nptles > 1 && ($prj_cc2_sum - $nptles*$prj_cc_avg**2) > 0) {
	$prj_cc_std = sqrt(($prj_cc2_sum - $nptles*$prj_cc_avg**2)/
			   ($nptles-1));
    }
    
    if($nptles > 1 && ($cmp_cc2_sum - $nptles*$cmp_cc_avg**2) > 0) {
	$cmp_cc_std = sqrt(($cmp_cc2_sum - $nptles*$cmp_cc_avg**2)/
			   ($nptles-1));
    }
    
    # Statistics for PO2R score are the same as those for pft_cc
    my ($score_avg, $score_std) = ($pft_cc_avg, $pft_cc_std);

    # Sort score/CC arrays by numerical value. Note that for the PO2R
    # score, lower is better hence forward rather than reverse sort
    @score  = sort numerically @pft_cc;
    @pft_cc = reverse sort numerically @pft_cc;
    @prj_cc = reverse sort numerically @prj_cc;
    @cmp_cc = reverse sort numerically @cmp_cc;

    # Determine the score/CC cutoffs

    if(defined $$auto_ref{score_fraction}) {
	my $nselect;
	if($$auto_ref{score_fraction} <= 1) {
	    my $fraction = $$auto_ref{score_fraction} + 
		$nshift*$$auto_ref{select_delta};
	    if($fraction > 1) {
		$fraction = 1;
	    } elsif ($fraction < 0) {
		$fraction = 0;
	    }
	    $nselect = $nptles * $fraction;
	} else {
	    $nselect = $$auto_ref{score_fraction} +
		$nshift*$$auto_ref{select_delta};
	    if($nselect > $nptles) {
		$nselect = $nptles;
	    }
	}
	$nselect = sprintf("%.0f", $nselect);
	$score_cutoff = $score[$nselect - 1];
    }

    if(defined $$auto_ref{pft_cc_fraction}) {
	my $nselect;
	if($$auto_ref{pft_cc_fraction} <= 1) {
	    my $fraction = $$auto_ref{pft_cc_fraction} + 
		$nshift*$$auto_ref{select_delta};
	    if($fraction > 1) {
		$fraction = 1;
	    } elsif ($fraction < 0) {
		$fraction = 0;
	    }
	    $nselect = $nptles * $fraction;
	} else {
	    $nselect = $$auto_ref{pft_cc_fraction} +
		$nshift*$$auto_ref{select_delta};
	    if($nselect > $nptles) {
		$nselect = $nptles;
	    }
	}
	$nselect = sprintf("%.0f", $nselect);
	$pft_cc_cutoff = $pft_cc[$nselect - 1];
    }

    if(defined $$auto_ref{prj_cc_fraction}) {
	my $nselect;
	if($$auto_ref{prj_cc_fraction} <= 1) {
	    my $fraction = $$auto_ref{prj_cc_fraction} + 
		$nshift*$$auto_ref{select_delta};
	    if($fraction > 1) {
		$fraction = 1;
	    } elsif ($fraction < 0) {
		$fraction = 0;
	    }
	    $nselect = $nptles * $fraction;
	} else {
	    $nselect = $$auto_ref{prj_cc_fraction} +
		$nshift*$$auto_ref{select_delta};
	    if($nselect > $nptles) {
		$nselect = $nptles;
	    }
	}
	$nselect = sprintf("%.0f", $nselect);
	$prj_cc_cutoff = $prj_cc[$nselect - 1];
    }

    if(defined $$auto_ref{cmp_cc_fraction}) {
	my $nselect;
	if($$auto_ref{cmp_cc_fraction} <= 1) {
	    my $fraction = $$auto_ref{cmp_cc_fraction} + 
		$nshift*$$auto_ref{select_delta};
	    if($fraction > 1) {
		$fraction = 1;
	    } elsif ($fraction < 0) {
		$fraction = 0;
	    }
	    $nselect = $nptles * $fraction;
	} else {
	    $nselect = $$auto_ref{cmp_cc_fraction} +
		$nshift*$$auto_ref{select_delta};
	    if($nselect > $nptles) {
		$nselect = $nptles;
	    }
	}
	$nselect = sprintf("%.0f", $nselect);
	$cmp_cc_cutoff = $cmp_cc[$nselect - 1];
    }

    ### Selection schemes based on number of standard deviations from
    ### average override specification of fractions of particles.

    if(defined $$auto_ref{score_nstd}) {
	$score_cutoff = $score_avg + 
	    ($$auto_ref{score_nstd}+$$auto_ref{select_delta})*$score_std;
    }

    if(defined $$auto_ref{pft_cc_nstd}) {
	$pft_cc_cutoff = $pft_cc_avg + 
	    ($$auto_ref{pft_cc_nstd}+$$auto_ref{select_delta})*$pft_cc_std;
    }

    if(defined $$auto_ref{prj_cc_nstd}) {
	$prj_cc_cutoff = $prj_cc_avg + 
	    ($$auto_ref{prj_cc_nstd}+$$auto_ref{select_delta})*$prj_cc_std;
    }

    if(defined $$auto_ref{cmp_cc_nstd}) {
	$cmp_cc_cutoff = $cmp_cc_avg + 
	    ($$auto_ref{cmp_cc_nstd}+$$auto_ref{select_delta})*$cmp_cc_std;
    }

    # Determine whether or not files were generated by PFTsearch or
    # PO2R. For PO2R files, pft_cc_avg and cmp_cc_avg are zero

    my $pftsearch_gen;
    if($prj_cc_avg > 0.0) { #RSS
	$pftsearch_gen = 1;
    } else {
	$pftsearch_gen = 0;
    }

    if($$auto_ref{global_select}) {
	print OUT "Statistics across all files ";
    } else {
	print OUT "Statistics for file $ptlefiles[0]\n";
    }
    print OUT "(selection criterion = $nshift)\n";

    if($pftsearch_gen) {
	print OUT "   pft_cc avg                $pft_cc_avg\n";
	print OUT "   pft_cc stddev             $pft_cc_std\n";
	print OUT "   prj_cc avg                $prj_cc_avg\n";
	print OUT "   prj_cc stddev             $prj_cc_std\n";
	print OUT "   cmp_cc avg                $cmp_cc_avg\n";
	print OUT "   cmp_cc stddev             $cmp_cc_std\n";
    } else {
	print OUT "   score avg                 $pft_cc_avg\n";
	print OUT "   score stddev              $pft_cc_std\n";
    }
    
    if(defined $score_cutoff) {
	print OUT "   score_cutoff              $score_cutoff\n";
    }
    if(defined $pft_cc_cutoff) {
	print OUT "   pft_cc_cutoff             $pft_cc_cutoff\n";
    } 
    if(defined $prj_cc_cutoff) {
	print OUT "   prj_cc_cutoff             $prj_cc_cutoff\n";
    }
    if(defined $cmp_cc_cutoff) {
	print OUT "   cmp_cc_cutoff             $cmp_cc_cutoff\n";
    }
    print OUT "\n";
    
    return($score_cutoff, $pft_cc_cutoff,
	   $prj_cc_cutoff, $cmp_cc_cutoff,
	   $pftsearch_gen);
}

sub select_ptles {

    # Purpose: select particles from a parameter file on the basis of
    # orientation (theta and phi), distance between center of particle
    # and center of box, and correlation coefficients (SEARCH mode) or
    # score (REFINE mode). These filters can be combined.
    #
    # Input arguments:
    #   $ptlefile - name of parameter file
    #   $selectfile - name of new file of selected particles
    #   $rejectfile - name of new file of rejected particles
    #   $auto_ref - reference to AUTO3DEM parameter hash
    #   $score_cutoff - score cutoff for PO2R generated file
    #   $pft_cc_cutoff - PFT CC cutoff for PPFT generated file 
    #	$prj_cc_cutoff - PRJ CC cutoff for PPFT generated file 
    #   $cmp_cc_cutoff - CMP CC cutoff for PPFT generated file 
    #   $pftsearch_gen - flag set to 1 if file was generated by pftsearch
    #
    # Return values:
    #   $nselect - number of particles selected

    my($ptlefile, $selectfile, $rejectfile, $auto_ref,
       $score_cutoff, $pft_cc_cutoff,
       $prj_cc_cutoff, $cmp_cc_cutoff,
       $pftsearch_gen) = @_;

    my $nselect = 0;
    my $header_flag = 0;

    open(PTLE, $ptlefile) || die
	"Unable to open file $ptlefile for reading in select_ptles\n";
    open(SEL, ">$selectfile") || die
	"Unable to open $selectfile for writing in select_ptles\n";
    open(REJ, ">$rejectfile") || die
	"Unable to open $rejectfile for writing in select_ptles\n";

    while(<PTLE>) {

	# Save the record
	my $record = $_;

	unless($record =~ /\w/) {
	    next; # skipping blank lines
	}

	if($record =~ /pif/i || $_ =~ /mrc/i) {
	    print SEL $record;
	    print REJ $record;
	    $header_flag = 1;
	    next;
	}
	if($header_flag) {
	    print SEL $record;
	    print REJ $record;
	    $header_flag = 0;
	    next;
	}

	# Get rid of whitespace and split line into fields
	s/\s+//g;
	my ($id, $theta, $phi, $omega, 
	    $xcent, $ycent, $mag, $pft_cc,
	    $prj_cc, $cmp_cc) = split(/,/);

	### Test phi
	if($phi >= $$auto_ref{phi_reject_lower} &&
	   $phi <= $$auto_ref{phi_reject_upper}) {
	    print REJ $record;
	    next; # Particle failed test
	}
	
	### Test theta
	if($theta >= $$auto_ref{theta_reject_lower} &&
	   $theta <= $$auto_ref{theta_reject_upper}) {
	    print REJ $record;
	    next; # Particle failed test
	}
    
	### Test omega
	# First make sure that 0 <= omega <= 360
	while($omega > 360) {$omega -= 360;}
	while($omega <   0) {$omega += 360;}

	my $omega_test1a = abs($omega - $$auto_ref{omega1});
	my $omega_test1b = abs($omega - 360 - $$auto_ref{omega1});
	my $omega_test1c = abs($omega + 360 - $$auto_ref{omega1});
	my $omega_test2a = abs($omega - $$auto_ref{omega2});
	my $omega_test2b = abs($omega - 360 - $$auto_ref{omega2});
	my $omega_test2c = abs($omega + 360 - $$auto_ref{omega2});
	unless($omega_test1a < $$auto_ref{omega1_tol} ||
	       $omega_test1b < $$auto_ref{omega1_tol} ||
	       $omega_test1c < $$auto_ref{omega1_tol} ||
	       $omega_test2a < $$auto_ref{omega2_tol} ||
	       $omega_test2b < $$auto_ref{omega2_tol} ||
	       $omega_test2c < $$auto_ref{omega2_tol}) {
	    print REJ $record;
	    next; # Particle failed test
	}

	### Test center of particle relative to box center
	if($xcent < $$auto_ref{boxrad}-$$auto_ref{box_center_offset} ||
	   $xcent > $$auto_ref{boxrad}+$$auto_ref{box_center_offset} ||
	   $ycent < $$auto_ref{boxrad}-$$auto_ref{box_center_offset} ||
	   $ycent > $$auto_ref{boxrad}+$$auto_ref{box_center_offset}) {
	    print REJ $record;
	    next; # Particle failed test
	}

	### Test pft_cc for PFTsearch generated files
	if(defined $pft_cc_cutoff && $pftsearch_gen && 
	   $pft_cc < $pft_cc_cutoff) {
	    print REJ $record;
	    next; # Particle failed test
	}

	### Test prj_cc for PFTsearch generated files
	if(defined $prj_cc_cutoff && $pftsearch_gen && 
	   $prj_cc < $prj_cc_cutoff) {
	    print REJ $record;
	    next; # Particle failed test
	}

	### Test cmp_cc for PFTsearch generated files
	if(defined $cmp_cc_cutoff && $pftsearch_gen && 
	   $cmp_cc < $cmp_cc_cutoff) {
	    print REJ $record;
	    next; # Particle failed test
	}

	### Test score for PO2R generated files
	if(defined $score_cutoff && !$pftsearch_gen &&
	   $pft_cc > $score_cutoff) {
	    print REJ $record;
	    next; # Particle failed test
	}
	
	$nselect++;
	print SEL $record;
    }

    close(PTLE);
    close(SEL);
    close(REJ);

    return($nselect);
}

sub numerically {
    $a <=> $b;
}
1;
