use strict;
use warnings;
use lib do { __FILE__ =~ m|^(.*?)[^/]*$|; "$1"; };

sub update_res {

    # Purpose: update the resolutions used by programs PO2R, PSF,
    # PPFT, and P3DR based on the FSC curve and values of adjustment
    # factors. If $reached_lothresh is false, shift toward higher
    # resolutions.
    #
    # Input arguments:
    #   $po2r_ref         - reference to hash of PO2R input parameters
    #   $p3dr_ref         - reference to hash of P3DR input parameters
    #   $psf_ref          - reference to hash of PSF input parameters
    #   $ppft_ref         - reference to hash of PPFT input parameters
    #   $pctfr_ref        - reference to hash of PCTFR input parameters
    #   $auto_ref         - reference to hash of control parameters
    #   $pixel_size       - pixel size from images
    #   $res_lothresh     - resolution where FSC curve crosses fsc_lothresh
    #   $reached_lothresh - true if FSC curve dropped below fsc_lothresh
    #
    # Return values:
    #   None

    my ($po2r_ref, $p3dr_ref, $psf_ref, $ppft_ref, $pctfr_ref,
	$auto_ref, $pixel_size, $res_lothresh, $reached_lothresh) = @_;

    my $date = localtime();
    print OUT "Updating resolutions\n";
    print OUT "   Date = $date\n";

    # Calculate resolutions
    # PO2R res_max > P3DR res_min > P3DR res_max
    # PPFT resolution_high = PO2R res_max = PCTFR res_max
    # PSF res_max = P3DR res_min

    # First calculate P3DR res_max. If $reached_lothresh is false, then a
    # larger step is taken toward higher resolutions.

    if($reached_lothresh) {
	$$p3dr_ref{res_max} = 1/(1/$res_lothresh + $$auto_ref{gauss_adj} +
				 $$auto_ref{res_adj});
    } else {
	$$p3dr_ref{res_max} = 1/(1/$res_lothresh + $$auto_ref{gauss_adj} +
				 $$auto_ref{res_adj} + $$auto_ref{res_adj});
    }

    # Make sure that the Nyquist limit is not exceeded. If it is, set
    # P3DR res_max to just above the Nyquist limit. Note the use of
    # the temporary variable to store the rounded version of
    # res_max. Avoids quirky behavior when working very close to the
    # Nyquist limit.

    my $temp_res_max = sprintf("%.2f", $$p3dr_ref{res_max});
    if($temp_res_max <= 2.0*$pixel_size) {
	$$p3dr_ref{res_max} = 2.0*$pixel_size + 0.05;
    }

    # Work backwards from P3DR res_max to calculate other resolutions

    $$p3dr_ref{res_min} = 1/(1/$$p3dr_ref{res_max} - $$auto_ref{gauss_adj});
    $$po2r_ref{res_max} = 1/(1/$$p3dr_ref{res_min} - $$auto_ref{res_adj});
    $$ppft_ref{resolution_high} = $$po2r_ref{res_max};
    $$pctfr_ref{res_max} = $$po2r_ref{res_max};
    $$psf_ref{res_max} = $$p3dr_ref{res_min};

    # Round off resolutions to two decimal places

    $$p3dr_ref{res_max} = sprintf("%.2f", $$p3dr_ref{res_max});
    $$p3dr_ref{res_min} = sprintf("%.2f", $$p3dr_ref{res_min});
    $$po2r_ref{res_max} = sprintf("%.2f", $$po2r_ref{res_max});
    $$ppft_ref{resolution_high} = sprintf("%.2f", $$ppft_ref{resolution_high});
    $$pctfr_ref{res_max} = sprintf("%.2f", $$pctfr_ref{res_max});
    $$psf_ref{res_max} = sprintf("%.2f", $$psf_ref{res_max});

    print OUT "   (PO2R res_max, PPFT resolution_high)";
    print OUT " set to $$ppft_ref{resolution_high}\n";
    print OUT "   (P3DR res_min, PSF res_max)         ";
    print OUT " set to $$psf_ref{res_max}\n";
    print OUT "   (P3DR res_max)                      ";
    print OUT " set to $$p3dr_ref{res_max}\n";

    print OUT "\n";

    return;
}
1;
