use strict;
use warnings;
use lib do { __FILE__ =~ m|^(.*?)[^/]*$|; "$1"; };

use init_params;

sub write_params {

    # Purpose: write out parameter file
    #
    # Input arguments:
    #   $param_file - input file
    #   $auto_ref   - reference to hash of control parameters
    #   $po2r_ref   - reference to hash of PO2R input parameters
    #   $p3dr_ref   - reference to hash of P3DR input parameters
    #   $pcut_ref   - reference to hash of PCUT input parameters
    #   $psf_ref    - reference to hash of PSF input parameters
    #   $ppft_ref   - reference to hash of PSF input parameters
    #   $pctfr_ref  - reference to hash of PCTFR input parameters
    #   $data_ref   - reference to array of data files
    #
    # Return values:
    #   1 if file successfully written, 0 otherwise

    my ($param_file, $auto_ref, $po2r_ref, $p3dr_ref,
	$pcut_ref, $psf_ref, $ppft_ref, $pctfr_ref, $data_ref) = @_;

    unless (open(FH, ">$param_file")) {
	print "Unable to write file $param_file";
	return(0);
    }

    print FH "# General auto3dem control parameters\n";
    if(valid_key('auto', 'symm_code') && defined $$auto_ref{symm_code}) {
	print FH "auto\tsymm_code\t$$auto_ref{symm_code} ";
	print FH "# overrides symm_code for all programs\n";
    }
    print FH "auto\tmode\t$$auto_ref{mode}\n";
    print FH "auto\trundir\t$$auto_ref{rundir}\n";
    print FH "auto\toutfile\t$$auto_ref{outfile}\n";
    print FH "auto\tdelete_maps\t$$auto_ref{delete_maps}\n";
    print FH "auto\tbox_center_offset\t$$auto_ref{box_center_offset}\n";
    print FH "auto\tnew_ptles\t$$auto_ref{new_ptles}\n";
    print FH "auto\testimate_res\t$$auto_ref{estimate_res}\n";
    print FH "auto\tterm_search\t$$auto_ref{term_search}\n";
    print FH "auto\tterm_refine\t$$auto_ref{term_refine}\n";
    print FH "auto\tswitch_mode\t$$auto_ref{switch_mode}\n";
    print FH "auto\tbin_reduce\t$$auto_ref{bin_reduce}\n";
    print FH "auto\trefine_ctf\t$$auto_ref{refine_ctf}\n";
    print FH "auto\tquit_early\t$$auto_ref{quit_early}\n";
    print FH "auto\tnoctf\t$$auto_ref{noctf}\n";
    print FH "auto\tper_ptle_ctf\t$$auto_ref{per_ptle_ctf}\n";
    print FH "auto\tnoise_suppression\t$$auto_ref{noise_suppression}\n";
    print FH "auto\trmc\t$$auto_ref{rmc}\n";
    print FH "\n";

    if(valid_key('auto', 'recipient') && defined $$auto_ref{recipient}) {
	print FH "# email notification\n";
	print FH "auto\trecipient\t$$auto_ref{recipient}\n";
	print FH "\n";
    }

    if(valid_key('auto', 'boxrad')) {
	print FH "# Box radius - only needs to be specified manually ";
	print FH "if missing from PIF header\n";
	print FH "auto\tboxrad\t$$auto_ref{boxrad}\n";
    }
    print FH "\n";

    print FH "# Particle radius - only used for adaptive angle\n";
    print FH "#  can be specified at start, but at each iter is derived from radial profile\n";
    print FH "auto\tpartrad\t$$auto_ref{partrad}\n";
    print FH "\n";

    print FH "# 'Freeze' parameters - set to 1 to keep constant\n";
    foreach my $key (sort keys %{$auto_ref}) {
	if(valid_key('auto', $key) && $key =~ /freeze/) {
	    print FH "auto\t$key\t$$auto_ref{$key}\n";
	}
    }
    print FH "\n";

    print FH "# Map specification parameters\n";
    foreach my $key (sort keys %{$auto_ref}) {
	if(valid_key('auto', $key) &&
	   ($key =~ /have_map/ || $key =~ /start_map/)) {
	    print FH "auto\t$key\t$$auto_ref{$key}\n";
	}
    }
    print FH "\n";

    print FH "# Map reconstruction mode\n";
    print FH "#  accepted values: yes, no or only\n";
    print FH "auto\tgenerate_map\t$$auto_ref{generate_map}\n";
    print FH "#  none for not appending anything to the file name\n";
    print FH "auto\tmap_suffix\t$$auto_ref{map_suffix}\n";
    print FH "\n";

    print FH "# Adaptive angle (accepted: yes or no)\n";
    print FH "auto\tadapt_angle\t$$auto_ref{adapt_angle}\n";
    print FH "\n";

    print FH "# Iteration parameters\n";
    foreach my $key (sort keys %{$auto_ref}) {
	if(valid_key('auto', $key) &&
	   ($key =~ /niter/ || $key =~ /iter_start/ || $key =~ /restart/)) {
	    print FH "auto\t$key\t$$auto_ref{$key}\n";
	}
    }
    print FH "\n";

    print FH "# Particle selection criteria involving scores or CCs\n";
    foreach my $key (sort keys %{$auto_ref}) {
	if(valid_key('auto', $key) && 
	   ($key =~ /_fraction/ || $key =~ /_nstd/ || $key =~ /select/)) {
	    print FH "auto\t$key\t$$auto_ref{$key}\n";
	}
    }
    print FH "\n";

    print FH "# Hollow map parameters (ignored if hollow_map = 0)\n";
    foreach my $key (sort keys %{$auto_ref}) {
	if(valid_key('auto', $key) && $key =~ /hollow/) {
	    print FH "auto\t$key\t$$auto_ref{$key}\n";
	}
    }
    print FH "\n";

    print FH "# Parameters for rejecting particle based on theta and phi\n";
    print FH "# Leave (lower,upper) = (360,-360) to keep all orientations\n";
    foreach my $key (sort keys %{$auto_ref}) {
	if(valid_key('auto', $key) && $key =~ /_reject_/) {
	    print FH "auto\t$key\t$$auto_ref{$key}\n";
	}
    }
    print FH "\n";

    print FH "# Parameters for rejecting particle based on omega\n";
    print FH "# Leave omega*_tol equal to 360 to keep all omega\n";
    print FH "auto\tomega1\t$$auto_ref{omega1}\n";
    print FH "auto\tomega1_tol\t$$auto_ref{omega1_tol}\n";
    print FH "auto\tomega2\t$$auto_ref{omega2}\n";
    print FH "auto\tomega2_tol\t$$auto_ref{omega2_tol}\n";
    print FH "\n";

    print FH "# FSC curve parameters\n";
    foreach my $key (sort keys %{$auto_ref}) {
	if(valid_key('auto', $key) && $key =~ /fsc/) {
	    print FH "auto\t$key\t$$auto_ref{$key}\n";
	}
    }
    print FH "\n";

    print FH "# Resolution calculation parameters\n";
    foreach my $key (sort keys %{$auto_ref}) {
	if(valid_key('auto', $key) &&
	   ($key =~ /res_adj/ || $key =~ /gauss_adj/)) {
	    print FH "auto\t$key\t$$auto_ref{$key}\n";
	}
    }
    print FH "\n";


# --------------------------------------------------------

    foreach my $key (sort keys %{$po2r_ref}) {
	if(valid_key('po2r', $key) && 
	   not deprecated_key('po2r', $key)) {
#	   not ($key =~ /ticos_equiv/ || $key =~ /global_por/)) {
	    print FH "po2r\t$key\t$$po2r_ref{$key}\n";
	}
    }
    print FH "\n";

    foreach my $key (sort keys %{$pctfr_ref}) {
	if(valid_key('pctfr', $key)) {
	    print FH "pctfr\t$key\t$$pctfr_ref{$key}\n";
	}
    }
    print FH "\n";

    foreach my $key (sort keys %{$p3dr_ref}) {
	if(valid_key('p3dr', $key)) {
	    print FH "p3dr\t$key\t$$p3dr_ref{$key}\n";
	}
    }
    print FH "\n";

    foreach my $key (sort keys %{$pcut_ref}) {
	if(valid_key('pcut', $key)) {
	    print FH "pcut\t$key\t$$pcut_ref{$key}\n";
	}
    }
    print FH "\n";

    foreach my $key (sort keys %{$psf_ref}) {
	if(valid_key('psf', $key)) {
	    print FH "psf\t$key\t$$psf_ref{$key}\n";
	}
    }
    print FH "\n";

    foreach my $key (sort keys %{$ppft_ref}) {
	if(valid_key('ppft', $key)) {
	    print FH "ppft\t$key\t$$ppft_ref{$key}\n";
	}
    }
    print FH "\n";

    foreach my $file (sort @{$data_ref}) {
	print FH "data\t$file\n";
    }

    close(FH);
    return(1);
}
1;

	




