#!/usr/bin/perl
use strict;
use warnings;

# Return infomation from DM3 file header along with pixel size derived
# from header data. See important notes in get_pixel_size() regarding
# microscope specific quantities.

my $infile = $ARGV[0];

open(FH, $infile);
binmode FH;

my $filesize = -s $infile;
my $buffer;
read(FH, $buffer, $filesize);

my ($loc, $offset, $type);
my ($voltage, $cs, $mag, $cam_pix);

print "\nNOTE - pixel values are specific for calibrations performed on\n";
print "Baker lab microscopes. You will need to edit the get_pixel_size()\n";
print "function if parsing DM3 files collected elsewhere\n\n";

# Extracting "Voltage"
$loc = index($buffer, "Voltage");
$offset = $loc + 7 + 8;
seek(FH, $offset, 0);
read(FH, $type, 4);
$type = unpack "N", $type;
if($type == 6) {
    read(FH, $voltage, 4);
    $voltage = unpack "f", $voltage;
} else {
    read(FH, $voltage, 8);
    $voltage = unpack "d", $voltage;
}
print "voltage       = $voltage\n";

# Extracting "Indicated Magnification"
$loc = index($buffer, "Indicated Magnification");
$offset = $loc + 23 + 8;
seek(FH, $offset, 0);
read(FH, $type, 4);
$type = unpack "N", $type;
if($type == 6) {
    read(FH, $mag, 4);
    $mag = unpack "f", $mag;
} else {
    read(FH, $mag, 8);
    $mag = unpack "d", $mag;
}
print "magnification = $mag\n";

# Extracting "Cs(mm)"
$loc = index($buffer, "Cs(mm)");
$offset = $loc + 6 + 8;
seek(FH, $offset, 0);
read(FH, $type, 4);
$type = unpack "N", $type;
if($type == 6) {
    read(FH, $cs, 4);
    $cs = unpack "f", $cs;
} else {
    read(FH, $cs, 8);
    $cs = unpack "d", $cs;
}
print "Cs            = $cs\n";

# Extracting "Pixel Size (um)"
$loc = index($buffer, "Camera Number");
$loc = index($buffer, "Pixel Size (um)", $loc);
$offset = $loc + 15 + 32;
seek(FH, $offset, 0);
read(FH, $type, 4);
$type = unpack "N", $type;
if($type == 6) {
    read(FH, $cam_pix, 4);
    $cam_pix = unpack "f", $cam_pix;
} else {
    read(FH, $cam_pix, 8);
    $cam_pix = unpack "d", $cam_pix;
}
print "Camera pixel  = $cam_pix\n";

close(FH);

my $pixel = get_pixel_size($voltage, $mag, $cam_pix);
print "pixel         = $pixel\n";


sub get_pixel_size {

    # Return the CCD pixel size based on the nominal magnification,
    # voltage, and camera physical pixel size.
    #
    # IMPORTANT - These values are microscope specific and depend on
    # calibrations performed by FEI at time of microscope
    # installation. Also, we are using camera pixel size to
    # distinguish microscopes. This may not be valid in the future if
    # both scopes end up having cameras with the same pixel sizes.

    my ($voltage, $mag, $cam_pix) = @_;
    my $pixel;

    my %polara_200;                my %polara_300;		
    $polara_200{2300}   = 47.77;   $polara_300{2300}   = 49.87;
    $polara_200{3100}   = 36.27;   $polara_300{3100}   = 37.27;
    $polara_200{3900}   = 29.07;   $polara_300{3900}   = 29.72;
    $polara_200{4700}   = 24.73;   $polara_300{4700}   = 25.19;
    $polara_200{9400}   = 12.47;   $polara_300{9400}   = 12.77;
    $polara_200{12000}  = 9.814;   $polara_300{12000}  = 10.07;
    $polara_200{15500}  = 7.437;   $polara_300{15500}  = 7.655;
    $polara_200{20000}  = 5.834;   $polara_300{20000}  = 5.977;
    $polara_200{23000}  = 4.894;   $polara_300{23000}  = 5.031;
    $polara_200{31000}  = 3.639;   $polara_300{31000}  = 3.780;
    $polara_200{39000}  = 2.917;   $polara_300{39000}  = 3.010;
    $polara_200{59000}  = 1.883;   $polara_300{59000}  = 1.960;
    $polara_200{78000}  = 1.439;   $polara_300{78000}  = 1.498;
    $polara_200{93000}  = 1.206;   $polara_300{93000}  = 1.255;
    $polara_200{115000} = 0.972;   $polara_300{115000} = 0.979;
    $polara_200{155000} = 0.724;   $polara_300{155000} = 0.755;

    my %sphera_200;
    $sphera_200{1700}   = 60.82;
    $sphera_200{2500}   = 40.98;
    $sphera_200{3500}   = 29.79;
    $sphera_200{5000}   = 20.51;
    $sphera_200{6500}   = 15.97;
    $sphera_200{7800}   = 13.27;
    $sphera_200{9600}   = 10.65;
    $sphera_200{11500}  = 9.036;
    $sphera_200{14500}  = 7.040;
    $sphera_200{19000}  = 5.130;
    $sphera_200{25000}  = 4.123;
    $sphera_200{29000}  = 3.475;
    $sphera_200{50000}  = 2.033;
    $sphera_200{62000}  = 1.646;
    $sphera_200{80000}  = 1.278;
    $sphera_200{100000} = 1.026;
    $sphera_200{150000} = 0.691;
    $sphera_200{280000} = 0.287;
    $sphera_200{490000} = 0.191;

    if ($voltage == 200000 && $cam_pix == 15) {
	$pixel = $polara_200{$mag};
    } elsif ($voltage == 300000 && $cam_pix == 15) {
	$pixel = $polara_300{$mag};
    } elsif ($voltage == 200000 && $cam_pix == 14) {
	$pixel = $sphera_200{$mag};
    } else {
	$pixel = "undefined";
    }

    return $pixel;
}
