#!/usr/bin/perl
use warnings;
use strict;
use FindBin qw($Bin);
use lib "$Bin/modules";

use Cwd;
use Math::Trig;
use get_info;
use version;

use pbs_config;

# setup_rmc.pl is used to setup the random model calculation (RMC). It
# does not actually launch any of the parallel image processing codes,
# but rather generates the directory structure, auto3dem parameter
# files, random particle orientations, and command script required to
# carry out the RMC

# Input: All input optional and specified by flags. Defaults values
# shown in [ ] following descriptions
#
# -boxrad=n       image box size [obtained from image file]
# -fsc_nbins      number of bins for FSC calculation [50]
# -fsc_res_min    minimum resolution (A) for FSC calculation [60.0]
# -fsc_res_max    maximum resolution (A) for FSC calculation [30.0]
# -list=name      name of file listing particle parameter files [list]
# -multi          run multiple RMCs in parallel
# -ncpu=n         number of CPUs [4]
# -nimages=n      number of images to use in constructing model [150]
# -nmodels=n      number of random models [10]
# -noctf          turn off CTF corrections
# -nodefile=name  name of file containing list of compute nodes
# -res            highest resolution used in constructing map
# -rmax           maximum capsid radius
# -rmin           minimum capsid radius
# -rundir=name    directory containing particle parameter files [dat]
# -seed=n         seed for random number generator. Use an integer for
#                 reproducible result, -seed=time for seed based on
#                 seconds since 1/1/1970 [1]
# -symm_code      symmetry of reconstruction [532]
# -trad           traditional RMC calculation (best map chosen using FSC)
#
# ----- Directory structure for RMC -----
#
# The user specifies the directory containing the particle image and
# parameter files. Within this directory, subdirectories are then
# created for each random model: RMC1, RMC2, etc. Within the
# subdirectories new parameter files are written which contain the
# random particle orientations.
#
# ----- Random orientation generation -----
#
# First generate list of (theta,phi) values evenly covering 1/2 of the
# asymmetric unit (ASU), typically done using spacing of 0.5
# degrees. Then, for each particle, randomly choose
#
# (1) (theta,|phi|) from set of orientations
# (2) Sign of phi
# (3) Omega in the range 0 to 360
# 
# The new particle parameter files contain the same header information
# as the original files, but with random orientations and particle
# centers initialized to lie at the center of the box.

# ----- auto3dem parameter files -----
#
# For each random model, an auto3dem parameter file is
# created. Default values are heavily relied upon and only a minimal
# set of parameters are explicitly set. These include the name of the
# final map, the run directory, resolutions, and annulus sizes. The
# initial choices for a number of these parameters depend explicitly
# on the box and pixel sizes of the images.

# ----- Command file -----
#
# The RMC can be broken down into a set of independent auto3dem runs
# each using a different set of random particle orientations. This
# script generates a command file that launches the individual
# auto3dem runs and parses the results to determine which random model
# results in the best resolution.

my $version = get_version();
my $version_string = "------ SETUP_RMC version $version ------";

# Save the starting directory
my $startdir = getcwd();

# Get input from command line and use to overwrite default values

# Print usage info if no arguments provided
if(@ARGV == 0 || $ARGV[0] =~ /^-help/i) {
print << "EOF";
$version_string

NAME
  setup_rmc - script for setting up random model calculation 

EXAMPLE 
  setup_rmc -ncpu 4 -seed 123 -list listfile
  setup_rmc -usedefaults

DESCRIPTION
  All input specified using the syntax -key value. Defaults values are
  shown in brackets following descriptions. To use all default values,
  use the -usedefaults flag. 'n' = integer, 'f' = float, 's' = string

  -boxrad n       image box size [obtained from image file]
  -fsc_nbins n    number of bins for FSC calculation [50]
  -fsc_res_min f  minimum resolution (A) for FSC calculation [60]
  -fsc_res_max f  maximum resolution (A) for FSC calculation [30]
  -list s         file or expression listing particle parameter files 
                  If not specified, then setup_rmc looks for 
                  (1) file named 'list' or (2) *00n files in rundir
  -multi          run multiple RMCs in parallel
  -ncpu n         number of CPUs for each RMC computation[4]
  -nimages n      number of images to use in constructing model [150]
  -nmodels n      number of random models [10]
  -noctf          turn off CTF corrections
  -nodefile s     file containing list of compute nodes
  -res f          highest resolution used in constructing map
  -rmax n         maximum capsid radius
  -rmin n         minimum capsid radius
  -rundir s       directory containing particle parameter files [dat]
  -seed n         seed for random number generator. Use an integer for
                  reproducible result or -seed=time for seed based on
                  seconds since 1/1/1970 [1]
  -symm_code n    symmetry of reconstruction [532]
  -trad           traditional RMC calculation (best map chosen using FSC)

NOTES 
  Nodefile specification depends on your computing environment and in
  many cases it can be omitted. For systems running the PBS scheduler,
  the nodefile is automatically obtained from the environment assuming
  that "#PBS -V" was used in the batch script. Keys are case
  insensitive (e.g. -ncpu, -Ncpu, -NCPU all OK), but string values are
  case sensitive. You will rarely need to set boxrad since it is
  normally determined from image file.

*** Continue running setup_rmc using all default values? ***
Enter [y]es to continue
EOF

    my $continue = <>;
    unless($continue =~ /y/i || $continue =~ /yes/i) {
	exit(0);
    }
}

print "--- setup_rmc starting ---\n\n";

if(defined $ARGV[0] && 
   ($ARGV[0] =~ /^-usedefaults$/i || $ARGV[0] =~/^-help$/i)  ) {
    @ARGV = ();
}

# Process command line args so that we can split on ' -'
my $command_line = join ' ', @ARGV;
$command_line =~ s/^\s+//;   # Get rid of leading whitespace
$command_line =~ s/\s+$//;   # Get rid of trailing whitespace
$command_line =~ s/=/ /g;    # Change '=' to ' '
$command_line =~ s/^-//;     # Get rid of leading '-'
@ARGV = split(/\s+-/, $command_line);

my %valid = (boxrad => 1, list => 1, fsc_res_min => 1, fsc_res_max => 1,
	     fsc_nbins => 1, multi => 1, ncpu => 1, nimages => 1,
	     nmodels => 1, noctf => 1, nodefile => 1, rundir => 1,
	     res => 1, rmax => 1, rmin => 1, trad => 1,
	     seed => 1, help => 1, usedefaults => 1, symm_code => 1);

my %cmdline_arg;
foreach my $arg (@ARGV) {
    my($key, $value) = split(/\s+/, $arg);
    $key =~ tr/A-Z/a-z/;
    unless($valid{$key}) {
	print "\n--- Error ---\n";
	print "  -$key is invalid keyword. Must be one of the following\n\n";
	foreach my $key (sort keys %valid) {
	    print "  -$key\n";
	}
	print "\nType setup_rmc without any arguments to see usage info\n";
	exit(0);
    }
    $cmdline_arg{$key} = $value;
}

my ($boxrad_cmdline, $list, $fsc_res_min, $fsc_res_max,
    $fsc_nbins, $multi, $ncpu, $nimages, $nmodels, $nodefile,
    $res, $rmax, $rmin, $trad, $rundir, $seed, $symm_code);

print "setup_rmc parameters:\n";

if(defined $cmdline_arg{boxrad}) {
    $boxrad_cmdline = $cmdline_arg{boxrad};
    print "  User specified boxrad = $boxrad_cmdline\n";
}

if(defined $cmdline_arg{rmax}) {
    $rmax = $cmdline_arg{rmax};
    print "  User specified max capsid radius = $rmax\n";
}

if(defined $cmdline_arg{rmin}) {
    $rmin = $cmdline_arg{rmin};
    print "  User specified min capsid radius = $rmin\n";
}

if(defined $cmdline_arg{rundir}) {
    $rundir = $cmdline_arg{rundir};
    print "  User specified rundir = $rundir\n";
} else {
    $rundir = 'dat';
    print "  Default rundir = $rundir\n";
}

if(defined $cmdline_arg{res}) {
    $res = $cmdline_arg{res};
    print "  User specified res = $res\n";
} else {
    $res = 29.0;
    print "  Default maximum map resolution = $res\n";
}

if(defined $cmdline_arg{list}) {
    $list = $cmdline_arg{list};
    print "  User specified list = $list\n";
} else {
    print "  list not specified - setup_rmc will first check for\n";
    print "    file 'list', then look for files $rundir/*00n\n";
}

if(defined $cmdline_arg{ncpu}) {
    $ncpu = $cmdline_arg{ncpu};
    print "  User specified ncpu = $ncpu\n";
} else {
    $ncpu = 4;
    print "  Default ncpu = $ncpu\n";
}

if(defined $cmdline_arg{fsc_nbins}) {
    $fsc_nbins = $cmdline_arg{fsc_nbins};
    print "  User specified fsc_nbins = $fsc_nbins\n";
} else {
    $fsc_nbins = 50;
    print "  Default fsc_nbins = 50\n";
}

if(defined $cmdline_arg{fsc_res_min}) {
    $fsc_res_min = $cmdline_arg{fsc_res_min};
    print "  User specified fsc_res_min = $fsc_res_min\n";
} else {
    $fsc_res_min = 60.0;
    print "  Default fsc_res_min = 60.0\n";
}

if(defined $cmdline_arg{fsc_res_max}) {
    $fsc_res_max = $cmdline_arg{fsc_res_max};
    print "  User specified fsc_res_max = $fsc_res_max\n";
} else {
    $fsc_res_max = 30.0;
    print "  Default fsc_res_max = 30.0\n";
}

if(defined $cmdline_arg{nimages}) {
    $nimages = $cmdline_arg{nimages};
    print "  User specified nimages = $nimages\n";
} else {
    $nimages = 150;
    print "  Default nimages = $nimages\n";
}

if(defined $cmdline_arg{nmodels}) {
    $nmodels = $cmdline_arg{nmodels};
    print "  User specified nmodels = $nmodels\n";
} else {
    $nmodels = 10;
    print "  Default nmodels = $nmodels\n";
}

if(defined $cmdline_arg{nodefile}) {
    $nodefile = $cmdline_arg{nodefile};
    print "  User specified nodefile = $nodefile\n";
    unless($nodefile =~ /^\//) {
	$nodefile = "$startdir/$nodefile";
    }
}

if(defined $cmdline_arg{symm_code}) {
    $symm_code = $cmdline_arg{symm_code};
    print "  User specified symm_code = $symm_code\n";
} else {
    $symm_code = 532;
    print "  Default symm_code = $symm_code\n";
}

if(defined $cmdline_arg{seed}) {
    $seed = $cmdline_arg{seed};
    if($seed =~ /time/i) {
	$seed = time();
	print "  User specified seed based on time = $seed\n";
    } elsif ($seed =~ /^\d+$/) {
	print "  User specified seed = $seed\n";
    } else {
	print "\n--- Error ---\n";
	print "  Invalid seed specified (seed = $seed)\n";
	print "  Seed must be integer or the string 'time'\n";
	exit(0);
    }
} else {
    $seed = 1;
    print "  Default seed = $seed\n";
}

if(exists $cmdline_arg{multi}) {
    $multi = 1;
} else {
    $multi = 0;
}

if(exists $cmdline_arg{trad}) {
    $trad = 1;
} else {
    $trad = 0;
}

# Initialize random number generator
srand($seed);

# Fill arrays of views (theta, phi) in the ASU
my $dang = 0.5;
my @theta = ();
my @phi = ();
get_all_views($dang, \@theta, \@phi);
my $nviews = @theta;

my %ptlefile;
my %head;
my %numimages;
my %boxrad;
my %pixel_size;
my %defocus;

# Get particle parameter files from $list. Note the $list can be
# either a file or an expression. If $list was not defined, then try
# defaults (1) look for file named 'list' or (2) look for *00n files
# in the $rundir directory. Expand wildcards as necessary.

my @list = ();
if(defined $list) {
    if(-r $list) { # Option 1 - $list is a readable file
	open(LIST, "$list");
	while(<LIST>) {
	    my $record = $_;
	    $record =~ s/#.*$//g; # Get rid of comments
	    unless(defined $record && $record =~ /\w/) {
		next;
	    }
	    $record =~ s/\s+//g;
	    push(@list, glob("${rundir}/$record"));
	}
	close(LIST);
    } else { # Option 2 - $list is an expression
	$list =~ s/^.*\///; # Strip off path information
	$list =~ s/\s+//g;  # Get rid of whitespace
	@list = glob("${rundir}/$list");
    }
} else {
    if(-r 'list') { # Option 3 - $list not specified, try to open 'list'
	open(LIST, 'list');
	while(<LIST>) {
	    my $record = $_;
	    $record =~ s/#.*$//g; # Get rid of comments
	    unless(defined $record && $record =~ /\w/) {
		next;
	    }
	    $record =~ s/\s+//g;
	    push(@list, glob("${rundir}/$record"));
	}
	close(LIST);
    } else { # Option 4 - $list not specified, look for files in $rundir
	for (my $i=0; $i<1000; $i++) {
	    my $suffix;
	    if($i < 10) {
		$suffix = "00${i}";
	    } elsif ($i < 100) {
		$suffix = "0${i}";
	    } else {
		$suffix = "${i}";
	    }
	    @list = glob("${rundir}/*${suffix}");
	    if(@list > 0) {
		last;
	    }
	}
    }
}

# Get rid of path information and whitespace in particle parameter files
foreach my $ptlefile (@list) {
    $ptlefile =~ s/.*\///;
    $ptlefile =~ s/\s+//g;
}

# Make sure that we have at least one file in @list
unless(@list > 0) {
    print "\n--- Error ---\n";
    print "  No particle parameter files found\n";
    if(defined $list) {
	print "  $list is empty or doesn't expand to at least one file\n";
    } else {
	print "  File 'list' does not exist and cannot find files\n";
	print "  $rundir/*000, $rundir/*001, etc.\n";
    }
    exit(0);
}

# Make sure that all files in @list are readable
foreach my $ptlefile (@list) {
    unless(-r "${rundir}/${ptlefile}") {
	print "\n--- Error ---\n";
	print "  File ${rundir}/${ptlefile} is unreadable\n";
	exit(0);
    }
}

print "\nParameter files in $rundir considered for model construction:\n";
foreach my $ptlefile (@list) {
    print "  $ptlefile\n";
}
print "\n";

# For each image file listed in each particle parameter file, get the
# name the PIF/MRC image file, header information, number of images, box
# radius, and pixel size.

chdir($rundir);
foreach my $ptlefile (@list) {
    open(PTLE, $ptlefile);
    my $header_flag = 0;
    my $img;
    while(<PTLE>) {
	unless(/\w/) {
	    next; # skipping  blank lines
	}
	s/\s+//g; # get rid of whitespace

	if($_ =~ /pif/i || $_ =~ /mrc/i) {
	    $img = $_;
	    $header_flag = 1;
	} elsif ($header_flag) {
	    $head{$img} = $_;
	    my ($pixel_size, undef, undef, undef, 
	     $defocus_maj, $defocus_min) = split(/,/, $head{$img});
	    $defocus{$img} = ($defocus_maj + $defocus_min)/2.0;
	    $numimages{$img} = get_numimages($img);
	    $boxrad{$img} = get_boxrad($img);
	    $pixel_size{$img} = $pixel_size;
	    $ptlefile{$img} = $ptlefile;
	    $header_flag = 0;
	}
    }
    close(PTLE);
}

chdir($startdir);

# Use a total of $nimages particles to construct the random
# model. Sort the micrographs by defocus and select particles starting
# with the highest defocus micrographs until the limit is reached. The
# array @imgs will contain the list of micrographs to be used and the
# corresponding values of the hash %numimages are reset to the
# required number of images from each micrograph.

my @imgs = ();
my $running_total = 0;
foreach my $img (sort { $defocus{$b} <=> $defocus{$a} } keys %defocus) {
    push(@imgs, $img);
    if($running_total + $numimages{$img} >= $nimages) {
	$numimages{$img} = $nimages - $running_total;
	last;
    }
    $running_total += $numimages{$img};
}



# Use first file to get boxrad and pixelsize
my $boxrad_from_img = $boxrad{$imgs[0]};
my $pixel_size = $pixel_size{$imgs[0]};
my $boxrad;

if($boxrad_from_img == 0) {
    undef $boxrad_from_img;
}

# The following code is used to allow for the specification of the
# box radius in either the PIF/MRC file or the auto3dem parameter
# file. Necessary because some PIF/MRC files are missing the box
# radius

# Case 1 - box radius not set in either location
unless(defined $boxrad_cmdline || defined $boxrad_from_img) {
    print "Box radius not specified in PIF/MRC file or command line\n";
    print "Either fix $imgs[0] or manually specify box radius\n";
    die "Terminating execution\n";
}

# Case 2 - box radius specified in both locations
if(defined $boxrad_cmdline && defined $boxrad_from_img) {
    $boxrad = $boxrad_cmdline;
    print "Box radius specified in both PIF/MRC file and command line\n";
    if($boxrad_cmdline == $boxrad_from_img) {
	print "Box radius ($boxrad) specified consistently\n";
    } else {
	print "\nWarning boxrad from PIF/MRC file ($boxrad_from_img) ";
	print "not equal to boxrad from command line ";
	print "($boxrad_cmdline)\n";
	print "Value in parameter file overides PIF/MRC file\n";
    }
}

# Case 3 - box radius specified only in PIF/MRC file
if(defined $boxrad_from_img && !defined $boxrad_cmdline) {
    $boxrad = $boxrad_from_img;
    print "Box radius from PIF/MRC file = $boxrad\n";
}

# Case 4 - box radius specified only in command line
if(!defined $boxrad_from_img && defined $boxrad_cmdline) {
    $boxrad = $boxrad_cmdline;
    print "Box radius from command_line = $boxrad\n";
}

# Create directory structure, set symbolic links, generate files
# containing random particle orientations, write auto3dem parameter
# files, and create command file for running RMC

if($multi) {
    my ($pbs_batchname, $pbs_nodespec, $pbs_rmctime, 
	$pbs_fulltime, $pbs_tasks) = get_pbs_config();

# Eventually number of nodes/procs needs to be automatically obtained
# from the system, and compared with number of cpu/RMC requested
    $pbs_tasks = $ncpu;
    $pbs_nodespec = "nodes=1:ppn=$ncpu";

    open(MULTI, ">RMC_run");
    
    print MULTI "#!/bin/bash\n\n";
    print MULTI "rm -f RMC_DONE\n";
    print MULTI "JOBLIST=''\n";
    print MULTI "rm -rf RMC_temp\n";
    print MULTI "mkdir RMC_temp\n\n";

    print MULTI "# Launch jobs for random model calculations\n";
    print MULTI "for (( i = 1 ; i <= $nmodels ; i++ ))\n";
    print MULTI "do\n";
    print MULTI "  SHOWQ=`qsub RMC\$[i]_job";
    if(defined $nodefile) {
	print MULTI " -nodefile=$nodefile`\n";
    } else {
	print MULTI "`\n";
    }
    print MULTI "  JOBID=\$(echo \$SHOWQ | sed -e \"s/\\..*//\") \n";
    print MULTI "  JOBLIST=\$JOBLIST:\$JOBID\n";
    print MULTI "done\n\n";
    
    print MULTI "# Construct a list of jobs for later use\n";
    print MULTI "JOBLIST=\$(echo \$JOBLIST | sed -e \"s/^://\")\n";
    print MULTI "NOCOLON=\$(echo \$JOBLIST | sed -e \"s/:/ /g\")\n";
    print MULTI "echo \"RMC batch job IDs = \$NOCOLON\"\n\n";
    
    print MULTI "# Launch job that cleans up after RMC runs and finds best model. The\n";
    print MULTI "# depend clause guarantees that this job runs after all RMC\n";
    print MULTI "# calculations have been completed.\n\n";
    
    print MULTI "SHOWQ=`qsub -W depend=afterany:\$JOBLIST RMC_final`\n";
    print MULTI "JOBID=\$(echo \$SHOWQ | sed -e \"s/\\..*//\")\n";
    print MULTI "echo \"Find best model job ID = \$JOBID\"\n\n";
    
# GCA: rmc_setup can not launch auto3dem after determining the model 
#    print MULTI "# Launch job that runs full calculation\n\n";
#    
#    print MULTI "SHOWQ=`qsub -W depend=afterany:\$JOBID auto3dem_job`\n";
#    print MULTI "JOBID=\$(echo \$SHOWQ | sed -e \"s/\\..*//\")\n";
#    print MULTI "echo \"Find best model job ID = \$JOBID\"\n\n";
    
# GCA: does this check mechanism work?
    print MULTI "# Start loop and exit if only if one of the following conditions met\n";
    print MULTI "# (1) An RMC calculation finishes early\n";
    print MULTI "# (2) All RMC calculations run to completion\n\n";
    
    print MULTI "FLAG=1\n";
    print MULTI "while [ \"\$FLAG\" = \"1\" ]\n";
    print MULTI "do\n";
    print MULTI "  sleep 1\n\n";

    print MULTI "  # Check if any RMC jobs are still running\n";
    print MULTI "  FLAG=0\n";
    print MULTI "  QSOUT=`qstat \$NOCOLON 2> /dev/null`\n";
    print MULTI "  for VAR in \$QSOUT ;\n";
    print MULTI "    do\n";
    print MULTI "    if [ \"\$VAR\" = \"R\" ] ; then\n";
    print MULTI "      FLAG=1\n";
    print MULTI "    fi\n";
    print MULTI "    if [ \"\$VAR\" = \"S\" ] ; then\n";
    print MULTI "      FLAG=1\n";
    print MULTI "    fi\n";
    print MULTI "    if [ \"\$VAR\" = \"H\" ] ; then\n";
    print MULTI "      FLAG=1\n";
    print MULTI "    fi\n";
    print MULTI "    if [ \"\$VAR\" = \"Q\" ] ; then\n";
    print MULTI "      FLAG=1\n";
    print MULTI "    fi\n";
    print MULTI "  done\n\n";

    print MULTI "  # Check if any RMC job finished early\n";
    print MULTI "  if [ -f RMC_DONE ] ; then\n";
    print MULTI "      rm -f RMC_DONE\n";
    print MULTI "      echo \"Killing jobs \$NOCOLON\"\n";
    print MULTI "      qdel \$NOCOLON 2> /dev/null\n";
    print MULTI "      break\n";
    print MULTI "  fi\n\n";
  
    print MULTI "done\n\n";

    print MULTI "rm -f RMC_DONE\n";
    close(MULTI);
    chmod 0744, "RMC_run";

    for(my $i=1; $i<=$nmodels; $i++) {
	open(RMCJOB, ">RMC${i}_job");
	print RMCJOB "#!/bin/csh\n";
	print RMCJOB "#PBS -q $pbs_batchname\n";
	print RMCJOB "#PBS -N RMC${i}_job\n";
	print RMCJOB "#PBS -l $pbs_nodespec\n";
	print RMCJOB "#PBS -l walltime=$pbs_rmctime\n";
	print RMCJOB "#PBS -o RMC${i}_job.out\n";
	print RMCJOB "#PBS -e RMC${i}_job.err\n";
	print RMCJOB "#PBS -V\n\n";
	print RMCJOB "cd $startdir\n";
	print RMCJOB "auto3dem -ncpu $pbs_tasks -input RMC${i}_master\n";
	close(RMCJOB);
    }
    
    my $base = $startdir;
    $base =~ s/^.*\///;

    open(FULLJOB, ">auto3dem_job");
    print FULLJOB "#!/bin/csh\n";
    print FULLJOB "#PBS -q $pbs_batchname\n";
    print FULLJOB "#PBS -N full_job\n";
    print FULLJOB "#PBS -l $pbs_nodespec\n";
    print FULLJOB "#PBS -l walltime=$pbs_fulltime\n";
    print FULLJOB "#PBS -o $startdir/${base}.out\n";
    print FULLJOB "#PBS -e $startdir/${base}.err\n";
    print FULLJOB "#PBS -V\n\n";
    print FULLJOB "cd $startdir\n";
    print FULLJOB "auto3dem -ncpu $pbs_tasks -input ${startdir}/${base}_master\n";
    close(FULLJOB);
    
    open(RMCFIN, ">RMC_final");
    print RMCFIN "#!/bin/csh\n";
    print RMCFIN "#PBS -q $pbs_batchname\n";
    print RMCFIN "#PBS -N RMC_final\n";
    print RMCFIN "#PBS -l nodes=1:ppn=1\n";
    print RMCFIN "#PBS -l walltime=0:02:00\n";
    print RMCFIN "#PBS -o $startdir/RMC_temp/RMC_final.out\n";
    print RMCFIN "#PBS -e $startdir/RMC_temp/RMC_final.err\n";
    print RMCFIN "#PBS -V\n\n";
    print RMCFIN "cd $startdir\n\n"; 

    if ($trad) {
	print RMCFIN "findbest RMC_logfile_list $rundir fsc\n";
    } else {
	print RMCFIN "findbest RMC_logfile_list $rundir imgstats\n";
    }
    print RMCFIN "remove_useless.pl $rundir $nmodels\n";
    print RMCFIN "mv RMC*_master RMC_temp\n";
    print RMCFIN "mv RMC*_RESTARTS RMC_temp\n";
    print RMCFIN "mv RMC*_summary RMC_temp\n";
    print RMCFIN "mv RMC*_log RMC_temp\n";
    print RMCFIN "mv RMC*_continue RMC_temp\n";
    print RMCFIN "mv RMC_logfile_list RMC_temp\n";
    print RMCFIN "mv RMC_run RMC_temp\n";
    print RMCFIN "mv RMC*restart* RMC_temp\n";
    print RMCFIN "mv RMC*_job* RMC_temp\n";
    print RMCFIN "mv RMC_final* RMC_temp\n";
    print RMCFIN "mv dat/RMC* RMC_temp\n";
    close(RMCFIN);
    chmod 0744, "RMC_final";    
} else {
    open(CMD, ">RMC_run");
    
    print CMD "#!/bin/sh\n\n";
    print CMD "rm -f RMC_DONE\n";
    print CMD "i=1\n";
    print CMD "while [ \$i -le $nmodels ]\n"; 
    print CMD "do\n";
    print CMD "  if [ -f RMC_DONE ] ; then\n";
    print CMD "    rm -f RMC_DONE\n";
    print CMD "    break\n";
    print CMD "  fi  \n";
    print CMD "  auto3dem -ncpu $ncpu -input RMC\${i}_master";
    if(defined $nodefile) {
	print CMD "-nodefile=$nodefile\n";
    } else {
	print CMD "\n";
    }
    print CMD "  i=\$(( \$i + 1 ))\n";
    print CMD "done";
    print CMD "\n";
    
    if ($trad) {
	print CMD "findbest RMC_logfile_list $rundir fsc\n";
    } else {
	print CMD "findbest RMC_logfile_list $rundir imgstats\n";
    }
    print CMD "remove_useless.pl $rundir $nmodels\n";
    print CMD "rm -rf RMC_temp\n";
    print CMD "mkdir RMC_temp\n";
    print CMD "mv RMC*_master RMC_temp\n";
    print CMD "mv RMC*_RESTARTS RMC_temp\n";
    print CMD "mv RMC*_summary RMC_temp\n";
    print CMD "mv RMC*_log RMC_temp\n";
    print CMD "mv RMC*_continue RMC_temp\n";
    print CMD "mv RMC_logfile_list RMC_temp\n";
    print CMD "mv RMC_run RMC_temp\n";
    print CMD "mv $rundir/RMC* RMC_temp\n";
    close(CMD);
}    

open(LOGLIST, ">RMC_logfile_list");

# Set parameters used for random model calculations
my $p3dr_res_min = $res + 0.5;
my $p3dr_res_max = $res;
my $ppft_resolution_low = $boxrad*2.0*$pixel_size/5.0;
my $ppft_resolution_high = $res + 1.0;
my $pftrad_hi = $boxrad;
my ($inner_limit, $outer_limit);
if(defined $rmin) {
    $inner_limit = $rmin;
} else {
    $inner_limit  = sprintf "%0.0f", $boxrad/3.0;
}
if(defined $rmax) {
    $outer_limit = $rmax;
} else {
    $outer_limit = sprintf "%0.0f", $boxrad - 10.0;
}

# Write out auto3dem master files for random model calculations
for (my $i=1; $i<=$nmodels; $i++) {
    print LOGLIST "RMC${i}_log\n";

    open(AUTO, ">RMC${i}_master");

    print AUTO "# Parameter file for random model $i\n\n";
    if ($trad) {
	print AUTO "auto rmc           1\n"; 
    } else {
	print AUTO "auto rmc           2\n"; 
    }
    print AUTO "auto rundir        ${rundir}/RMC${i}\n";
    print AUTO "auto mode          search\n";
    print AUTO "auto symm_code     $symm_code\n";
    print AUTO "auto switch_mode   0 # Don't allow switch to refine mode\n";
    print AUTO "auto niter         10\n";
    print AUTO "auto have_map      0\n";
    print AUTO "auto outfile       RMC${i}\n";

    if(defined $rmin || defined $rmax || $trad) {
	print AUTO "auto freeze_annulus  1\n";
    }

    print AUTO "auto freeze_res    1\n";
    print AUTO "auto boxrad        $boxrad\n";
    print AUTO "auto quit_early    1 # Quit if fcs_hithresh not reached\n";
    print AUTO "auto noise_suppression 0\n";
    print AUTO "auto adapt_angle no\n";
    unless ($trad) {
	print AUTO "auto estimate_res  0 # Don't estimate resolution\n";
    }

###    Following may be implemented later (currently placeholder)
###    print AUTO "auto term_search   1 # Allow early termination in search mode\n";
   
    if(exists $cmdline_arg{noctf}) {
	print AUTO "auto noctf           1\n";
    }
    print AUTO "\n";

    print AUTO "p3dr res_min         $p3dr_res_min\n";
    print AUTO "p3dr res_max         $p3dr_res_max\n";
    print AUTO "\n";

    print AUTO "psf  res_min         $fsc_res_min\n";
    print AUTO "psf  res_max         $fsc_res_max\n";
    print AUTO "psf  res_step        $fsc_nbins\n";
    print AUTO "\n";

    print AUTO "pcut in_rad          $inner_limit\n";
    print AUTO "pcut out_rad         $outer_limit\n";
    print AUTO "\n";

    unless ($trad) {
	print AUTO "auto hollow_cut_step 6\n";
	print AUTO "auto hollow_cut_weight 0.001\n";
	print AUTO "auto hollow_map 1\n";
	print AUTO "auto hollow_in_rad 0\n";
	print AUTO "auto hollow_out_rad $outer_limit\n";
    }

    print AUTO "ppft annulus_low     $inner_limit\n";
    print AUTO "ppft annulus_high    $outer_limit\n";
    print AUTO "ppft delta_theta     1.0\n";
    print AUTO "ppft resolution_low  $ppft_resolution_low\n";
    print AUTO "ppft resolution_high $ppft_resolution_high\n";
    print AUTO "ppft pftrad_hi       $pftrad_hi\n";

    if($pixel_size <= 5.0) {
	print AUTO "ppft bin_factor      2\n";
    }

    print AUTO "\n";

    my %already_used;
    foreach my $img (@imgs) {
	unless (exists $already_used{$ptlefile{$img}}) {
	    print AUTO "data $ptlefile{$img}\n";
	    $already_used{$ptlefile{$img}} = 1;
	}
    }
    %already_used = ();
    close(AUTO);

    # Create directory for random model calculation then write out
    # particle parameter files containing random orientations. If
    # relative path name was provided to images, then prefix with
    # "../" since we're going one directory down

    my $rmcdir = "${rundir}/RMC${i}";
    mkdir $rmcdir;
    chdir($rmcdir);
    my $append;
    foreach my $img (@imgs) {
	my $image_file = $img;
	unless ($image_file =~ /^\//) {
	    $image_file = "../${image_file}";
	}
	if (exists $already_used{$ptlefile{$img}}) {
	    $append = 1;
	} else {
	    $append = 0;
	    $already_used{$ptlefile{$img}} = 1;
	}
	make_random_file($image_file, $ptlefile{$img}, $append,
			 $head{$img}, $numimages{$img}, $boxrad);
    }
    chdir($startdir);
}
close(LOGLIST);

# Generate auto3dem parameter file to continue reconstruction
my $auto_master = "${startdir}_master";
$auto_master =~ s/^.*\///;
open(AUTO, ">$auto_master");

print AUTO << "EOF";
# This auto3dem input file was generated automatically by
# setup_rmc. Reasonable defaults have been set, but you may
# want to manually edit this file to override these defaults
#
# Note that binning in the context of this file refers to the
# automatic processing of data by program PPFT and is independent
# of any binning done during preprocessing. Refine mode calculations
# are always done with full resolution (unbinned) data.

EOF

print AUTO "auto    mode        search\n";
print AUTO "auto    niter       10\n";
print AUTO "auto    symm_code   $symm_code\n";
print AUTO "auto    start_map   rmc.pif\n";
print AUTO "auto    rundir      $rundir\n";
if(exists $cmdline_arg{noctf}) {
    print AUTO "auto    noctf       1\n";
}
print AUTO "p3dr    res_min     25.0\n";

if ($pixel_size < 4.0) {
    print AUTO "\n";
    print AUTO "# Start w/ binned data. Comment following line to disable\n";
    print AUTO "ppft bin_factor 2\n";
} else {
    print AUTO "\n";
    print AUTO "# Start w/ unbinned data. Uncomment following line to enable\n";
    print AUTO "# ppft bin_factor 2\n";
}

print AUTO << "EOF";

# Uncomment "auto bin_reduce" line to allow transition in search mode
# from binned to unbinned data. Otherwise go directly from search mode
# (binned) to refine mode. This option has no effect if already using
# unbinned data. Keep disabled for faster reconstruction. Enabling may
# sometimes (but not often) improve results.

# auto bin_reduce 1

# If the inner/outer radii of the ordered region of the particle are
# known (for example, by examining cross-section of map), uncomment
# the "auto freeze_annulus" line and set the parameters in the
# following lines to the desired value. This often results in a higher
# quality reconstruction.

# auto freeze_annulus 1  # (keep following parameters fixed)
# ppft annulus_low    n  # (n = inner radius w/ protrusions)
# ppft annulus_high   n  # (n = outer radius w/ protrusions)
# pcut in_rad         n  # (n = inner radius w/o protrusions)
# pcut out_rad        n  # (n = outer radius w/o protrusions)

# ----------------------------------------------------------

# List of data files given below based on input to setup_rmc.
# It is your responsibility to make sure list is complete.

EOF

foreach my $ptlefile (@list) {
    print AUTO "data    $ptlefile\n";
}
close(AUTO);

# Generate cleanup command
open(CLEAN, ">RMC_cleanup");
print CLEAN "rm -rf RMC_temp RMC_cleanup\n";
close(CLEAN);

# Make run and cleanup scripts executable
chmod 0744, "RMC_cleanup", "RMC_run";

print "\nInstructions:\n";
print "  Run RMC_run to construct starting model\n";
print "  Run RMC_cleanup to remove temporary files ";
print "after calculations complete\n";
print "  Starting model will be found at ${rundir}/rmc.pif\n";
print "  auto3dem input file $auto_master automatically generated\n";
print "\n";
print "--- setup_rmc done ---\n";
# --------------------------------------------------------------

sub make_random_file {

    # Purpose: Create a parameter file corresponding to a PIF/MRC
    # file. Particle orientations are chosen randomly within the
    # asymmetric unit and coordinates of origin are set equal to the
    # box radius.
    #
    # Input arguments:
    #   $img - name of PIF/MRC file
    #   $ptlefile - name of particle parameter file
    #   $append - if true, append to ptlefile
    #   $head - header line (voltage, defocus, etc.)
    #   $numimages - number of images in PIF/MRC file
    #   $boxrad - box radius

    my ($img, $ptlefile, $append, $head, $numimages, $boxrad) = @_;

    if($append) {
	open(PTLE, ">>$ptlefile");
    } else {
	open(PTLE, ">$ptlefile");
    }

    print PTLE "$img\n";
    print PTLE "$head\n";

    for (my $i=1; $i<=$numimages; $i++) {

	# Choose random theta, |phi| in ASU
	my $j = int(rand($nviews));
	my $theta = $theta[$j];
	my $phi = $phi[$j];
    
	# Randomly assign sign to phi
	if(int(rand(2)) == 1) {
	    $phi = -$phi;
	}

	# Random select omega in range 0 to 360
	my $omega = rand(360.0);
	$omega = sprintf "%0.2f", $omega;

	printf PTLE "%5d, %7.2f, %7.2f, %7.2f, " ,
	$i, $theta, $phi, $omega;

	printf PTLE "%6.2f, %6.2f, ",
	$boxrad, $boxrad;

	printf PTLE "%5.3f, %5.3f, %5.3f, %5.3f\n",
	1.0, 0.0, 0.0, 0.0;
    }
    close(PTLE);
}

sub get_all_views {

    # Purpose: Generate a set of orientations (theta, phi) that evenly
    # cover the icosahedral asymmetric unit. For each value of theta,
    # values of phi are chosen that span the interval
    # 0<=phi<=phimax. In order to evenly sample points on a sphere, a
    # spacing $dang/sin(theta) is used. The maximum value of phi is
    # calculated by a call to subroutine get_phimax()
    #
    # Input parameters: 
    #   $dang - angular spacing for theta
    #   $theta_ref - reference to array of theta's (changed on output) 
    #   $phi_ref - reference to array of phi's (changed on output)
    #
    # Return values:
    #   None

    my ($dang, $theta_ref, $phi_ref) = @_;

    # The first orientation is always chosen to be (69.09, 0)
    my $theta = 69.09;
    my $phi = 0.0;

    push(@{$theta_ref}, $theta);
    push(@{$phi_ref}, $phi);

    # Subsequent orientations start at 69.0 + $dang
    $theta = 69.0 + $dang;

    while ($theta <= 90.0) {
	my $dphi = $dang/sin($theta/180.0*acos(-1.0));
	my $phimax = get_phimax($theta);
	$phi = 0.0;
	while ($phi <= $phimax) {
	    my $theta_rnd = sprintf "%0.2f", $theta;
	    my $phi_rnd   = sprintf "%0.2f", $phi;
	    push(@{$theta_ref}, $theta_rnd);
	    push(@{$phi_ref}, $phi_rnd);
	    $phi = $phi + $dphi;	
	}		
	$theta = $theta + $dang;
    } 
}

sub get_phimax {

    # Purpose: Calculate the maximum allowable value of the angle phi
    # in the icosahedral asymmetic unit for a given value of
    # theta. Calculation is based on spherical triognometry relations
    # described in CRC Standard Mathematical Tables
    # Input parameter:
    #   $theta - inclination angle
    #
    # Return value:
    #   $phimax - maximum allowable azimuthal angle

    my ($theta) = @_;

    my $small_b = 69.09484255211070;
    my $c       = 31.71747441146101;
    my $deg_to_rad = acos(-1.0)/180.0;
    my $rad_to_deg = 1.0/$deg_to_rad;
    my $ninety = 90.0;

    my $temp = (1/tan($c*$deg_to_rad/2))  * 
	(sin(($ninety-$small_b)*$deg_to_rad/2)/ 
	 sin(($ninety+$small_b)*$deg_to_rad/2)) ;
    my $a_minus_b = 2*$rad_to_deg*atan($temp);
    $temp = (1/tan($c*$deg_to_rad/2))  * 
	(cos(($ninety-$small_b)*$deg_to_rad/2)/ 
	 cos(($ninety+$small_b)*$deg_to_rad/2)) ;
    my $a_plus_b = 2*$rad_to_deg*atan($temp) ;
    my $b = ($a_plus_b - $a_minus_b) / 2;
    my $a = $a_minus_b + $b;
    
    $temp    = sin($a*$deg_to_rad)*sin($small_b*$deg_to_rad) / 
	sin($theta*$deg_to_rad) ;
    $b       = $rad_to_deg*asin($temp);
    $temp    = sin(($theta-$small_b)*$deg_to_rad/2)  / 
	( sin(($theta+$small_b)*$deg_to_rad/2) *  
          tan(($a-$b)*$deg_to_rad/2) );
    my $phimax  = 2*$rad_to_deg*atan($temp) ;

    return($phimax);
}
