#include "image_types.h"
#include "imageconvert.h"

Micrograph * Micrograph_new() {

  Micrograph * mg = (Micrograph *) malloc(sizeof(Micrograph));

  if (mg==NULL) {
      fprintf(stderr,"Micrograph_new() - Can not allocate memory for the micrograph!\n");
      return mg;     
  }

  memset(mg->filename,'\0',1024);
  memset(mg->input_format,'\0',4);
  memset(mg->output_format,'\0',4);

  mg->header = Micrograph_header_new();
  mg->data = NULL;
  
  return mg;
}

void Micrograph_delete( Micrograph *mg){
  if(mg->header) free(mg->header);
  if(mg->data) free(mg->data);

  free(mg);
}

Micrograph_header * Micrograph_header_new() {

  Micrograph_header * mgh = (Micrograph_header *) malloc(sizeof(Micrograph_header));

  if (mgh==NULL) {
      fprintf(stderr,"Micrograph_header_new() - Can not allocate memory for the micrograph header!\n");
      return mgh;     
  }

  mgh->nx = -1;
  mgh->ny = -1;
  mgh->dx = -1.;
  mgh->dy = -1.;
  
  return mgh;
}

void Micrograph_header_delete( Micrograph_header *mgh){
  if(mgh) free(mgh);
}

int Micrograph_read ( Micrograph *mg, int bin ){

  int ret = 0;
  if ( strlen(mg->input_format) == 0 ) {
      fprintf(stderr,"ERROR: read format not specified\n");
      return(1);
  }      
  if ( strcmp(mg->input_format,"dm3") == 0  ){
      printf("Reading dm3 micrograph\n");
      DM3Micrograph *dm3mg = DM3Micrograph_new();
      strcpy(dm3mg->filename,mg->filename);
      DM3Micrograph_read(dm3mg);
      DM3Micrograph_export(dm3mg, mg);
  } else {
      printf("ERROR: reading format %s not implemented!\n",mg->input_format);
      return(1);
  }

  if (bin>1) Micrograph_bin(mg, bin);

  return ret;
}

int Micrograph_write( Micrograph *mg, char *fname ){

  int ret = 0;
  if ( strlen(mg->output_format) == 0 ) {
      printf("ERROR: write format not specified\n");
      return 1;
  }      
  if ( strcmp(mg->output_format,"pif") == 0  ){
      printf("Writing pif micrograph\n");
      PIFMicrograph *pifmg = PIFMicrograph_new();
      PIFMicrograph_import(mg, pifmg);
      strcpy(pifmg->filename,fname);
      PIFMicrograph_write(pifmg);
  } else {
      printf("ERROR: writing in format %s not implemented!\n",mg->output_format);
      ret = 1;
  }

  return ret;
}

int type2bytes (int t) {
    int nb = 0;
    switch ( t ) {
        case UBYTE:
        case BYTE: nb = 1; break;
        case USHORT:
        case SHORT: nb = 2; break;
        case UINT:
        case INT:
        case FLOAT: nb = 4; break;
        case DOUBLE: nb = 8; break; 
        default:
            fprintf(stderr, "Error(type2bytes): Can not interpret data type %d!\n", t);
    }
    return(nb);
}


int Micrograph_bin( Micrograph *mg, int bin ) {

  Micrograph_header * mgh = mg->header;

  if ( mgh->nx < bin || mgh->ny < bin) {
    fprintf(stderr,"WARNING: bin size is larger than linear size of micrograph: binning skipped\n");
    return(1);
  }

  if ( bin < 2 ) return(1);

  int nnx = mgh->nx/bin;
  if ( (float)mgh->nx/bin > nnx ) nnx +=1;
  int nny = mgh->ny/bin;
  if ( (float)mgh->ny/bin > nny ) nny +=1;
  
	
  unsigned long   newdatasize = nnx*nny*type2bytes(mgh->data_type);
  char * bindata = (char *) calloc(newdatasize, 1);

  char*       bdd = (char *) bindata;
  char*       bd  = (char *) mg->data;
  short*      sdd = (short *) bindata;
  short*      sd  = (short *) mg->data;
  int*        idd = (int *) bindata;
  int*        id  = (int *) mg->data;
  float*      fdd = (float *) bindata;
  float*      fd  = (float *) mg->data;
  

  long x,y,i, xx, yy, ii;
  
  for ( y=0; y<mgh->ny; y++ ) {
      yy = y/bin;
      for ( x=0; x<mgh->nx; x++ ) {
          xx = x/bin;
          i = y*mgh->nx + x;
          ii = yy*nnx + xx;
          
          switch ( mgh->data_type ) {
              case BYTE:
                  bdd[ii] += bd[i]; break;
              case SHORT:
                  sdd[ii] += sd[i]; break;
              case INT:
                  idd[ii] += id[i]; break;
              case FLOAT:
                  fdd[ii] += fd[i]; break;
              default:
                  break;
          }         
      }
  }
  
  for ( yy=0; y<nny; yy++ ) {
      for ( xx=0; xx<nnx; xx++ ) {
          ii = yy*nnx + xx;
          switch ( mgh->data_type ) {
              case BYTE:
                  bdd[ii] = (char) ((float) bdd[ii] / bin); break;
              case SHORT:
                  sdd[ii] = (short) ((float) sdd[ii] / bin); break;
              case INT:
                  idd[ii] = (int) ((float) idd[ii] / bin); break;
              case FLOAT:
                  fdd[ii] = fdd[ii] / bin; break;
              default:
                  break;
          }         
      }
  }
  
  //unsigned long   datasize = mgh->nx*mgh->ny*type2bytes(mgh->data_type);
  free(mg->data);
  mg->data = bindata;
  mgh->nx = nnx;
  mgh->ny = nny;
  mgh->dx *= bin;
  mgh->dy *= bin;
          
  return(0);
}
