/****************************************************************************
** TIFF_IO.c
** =========

James F. Conway, LSBR/NIAMS/NIH July 1997

Function to read TIFF file:
        TIFF_ReadImageHeader
        TIFF_ReadImageData

Original source from Rob Ashmore (Purdue) 1997.

****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TIFF_io.h"
#include "jfctypes.h"

// Constants

#define kMAX_DIR_ENTRIES (100)  // max entries in TIFF file IFD.

// Define DEBUG to print extra informational statements
//#define DEBUG

// Structures

typedef struct __tifTag_t__ {
    int    tagid;
    int    dataType;
    int    dataCount;
    int    dataOffset;
  } tifTag_t;

typedef enum {
  kNOVERBOSE, kVERBOSE
  } VERBOSE;

typedef struct {
  int bitsPerSample;
  int compression;
  int pmi;
  int orientation;
  int samplesPerPixel;
  int planarConfig;
  int tileWidth;
  int tileLength;
  int tileOffsets;
  int tileByteCounts;
  int tileByteCountsType;
  int endianSwitch;
  int ncol;
  int nrow;
  int firstIFD;
  int newSubfileType;
  int stripOffsets;
  int rowsPerStrip;
  int stripByteCounts;
  int strips;
  int tiles;
  int sampleFormat;
} tifInfo_t;

// Local functions

static
int ReadImageHeader(FILE *fptr, tifInfo_t *tifInfo, VERBOSE verbose);
static
int ReadTiffTiles(FILE *fptr, UBYTE *buffer, int dx, int dy, tifInfo_t *tifInfo);
static
int ReadTiffStrips(FILE *fptr, UBYTE *buffer, int dx, int dy, tifInfo_t *tifInfo);
static
void CopyArray( UBYTE *dst, int ddx, int ddy
              , UBYTE *src, int sdx, int sdy
              , int dx0, int dy0
              );
static
void TiffTagType(FILE *fptr, tifTag_t tag, tifInfo_t *tifInfo, VERBOSE verbose);
static
int  Util_IAmBigEndian(void);
static
void DifferentEndian(UBYTE *inStr, int length, int endianSwitch);
static
char *TiffTagReadString(FILE *fptr, tifTag_t tag);
static
void ZeroTifInfo(tifInfo_t *tifInfo);

/****************************************************************************
** TIFF_ReadImageHeader
** ====================
*/
int TIFF_ReadImageHeader(FILE *fptr, int *dx, int *dy, int *bytesPerPixel)
{
  tifInfo_t tifInfo;

  ZeroTifInfo(&tifInfo);

// Read off the header

  if(fseek(fptr, 0, SEEK_SET))  {    // set to the beginning
    printf("TIFF_ReadImageHeader() error fseek-ing to file start\n");
    return(0);
  }

  if(ReadImageHeader(fptr, &tifInfo, kVERBOSE) == 0)  {
    printf("TIFF_ReadImageHeader() error reading header\n");
    return(0);
  }
  *dx = tifInfo.ncol;
  *dy = tifInfo.nrow;
  *bytesPerPixel = (int)(tifInfo.bitsPerSample / 8);

  return(1);
}
/****************************************************************************
** TIFF_ReadImageData
** ==================
*/
int TIFF_ReadImageData(FILE *fptr, UBYTE *buffer, int dx, int dy, 
		       int bytesPerPixel)
{
  tifInfo_t tifInfo;
  int error;

  ZeroTifInfo(&tifInfo);

// Read off the header

  if(fseek(fptr, 0, SEEK_SET))  {    // set to the beginning
    printf("TIFF_ReadImageData() error fseek-ing to file start\n");
    return(0);
  }

  if(ReadImageHeader(fptr, &tifInfo, kNOVERBOSE) == 0)  {
    printf("TIFF_ReadImageData() error reading header\n");
    return(0);
  }
  if(dx != tifInfo.ncol || dy != tifInfo.nrow)  {
    printf("TIFF_ReadImageData() error - dimensions don't match\n");
    printf("  Image header: (%d,%d)    Passed: (%d,%d)\n"
                        , tifInfo.ncol, tifInfo.nrow, dx, dy);
    return(0);
  }

  if(tifInfo.tiles && tifInfo.strips)  {
    printf("ERROR - TIFF image tags specify BOTH strips and tiles\n");
    error = 0;
  }
  else if(tifInfo.tiles && !tifInfo.strips)
    error = ReadTiffTiles(fptr, buffer, dx, dy, &tifInfo);
  else if(!tifInfo.tiles && tifInfo.strips)
    error = ReadTiffStrips(fptr, buffer, dx, dy, &tifInfo);
  else  {  // neither
    printf("ERROR - TIFF image tags specify NEITHER strips and tiles\n");
    error = 0;
  }
  return(error);
}
/****************************************************************************
** ReadTiffTiles
** =============
*/
static
int ReadTiffTiles( FILE *fptr
                 , UBYTE *buffer
                 , int dx, int dy
                 , tifInfo_t *tifInfo
                 )
{
  int m, imageX, imageY, tenth, nextPercent;
  int *tileTable;
  UBYTE *tileBuffer;
  int tilesAcross, tilesDown, tilesInImage;
  const int tileWidth =tifInfo->tileWidth;
  const int tileLength=tifInfo->tileLength;
  int depth,i,j;

  depth = (int)(tifInfo->bitsPerSample / 8);

// Gather tile information

  tilesAcross = (tifInfo->ncol + (tileWidth -1)) / tileWidth;
  tilesDown   = (tifInfo->nrow + (tileLength-1)) / tileLength;

  tilesInImage = tilesAcross * tilesDown;
  printf("Tiles: size=(%d,%d)  #across=%d  #down=%d  total=%d  \n"
        , tileWidth, tileLength
        , tilesAcross, tilesDown, tilesInImage);

  tileBuffer = (UBYTE *)malloc(sizeof(UBYTE) * tileWidth*tileLength*depth);
  if(tileBuffer==NULL)  {
    printf("TIFF_ReadImageData() - Can't allocate temp buffer \n");
    return(0);
  }
  tileTable = malloc(tilesInImage * sizeof(*tileTable));
  if(tileTable==NULL)  {
    printf("TIFF_ReadImageData() - Can't allocate tile table\n");
    free(tileBuffer);
    return(0);
  }

// Read tile offsets

  if(fseek(fptr, tifInfo->tileOffsets, SEEK_SET))  { // find tile table
    printf("TIFF_ReadTiffTiles() error fseek-ing to tileTable\n");
    free(tileBuffer);
    free(tileTable);
    return(0);
  }

  if(fread(tileTable, sizeof(int), tilesInImage, fptr) != tilesInImage)  {
    printf("TIFF_ReadTiffTiles() error reading tileTable\n");
    free(tileBuffer);
    free(tileTable);
  }

  for(m=0; m<tilesInImage; m++)         // Fix up endian
    DifferentEndian((UBYTE *) &tileTable[m], 4, tifInfo->endianSwitch);

// Read the image data

  printf("- Reading tiff data:");
  tenth = tilesInImage/10 + 1;  // make sure 10% or more, but not less
  nextPercent = 0;
  for(m=0; m<tilesInImage; m++)  {      // Look at all tiles
    if(m >= nextPercent)  {
      printf(" %d%%", (m*100)/tilesInImage);
      fflush(stdout);
      nextPercent += tenth;
    }

// JFC - Not sure what is going on here, but I have seen a tileTable entry
//       of zero, which puts the first tile at the start of the file. But
//       the first 8 bytes are the TIFF header. For now I will check for 
//       this, and add 8 if necessary. I am not sure what will happen to
//       the following tiles...probably I should check for overlaps, in case
//       all tile addresses need to be bumped by 8.

    if(tileTable[m] == 0)
      tileTable[m] = 8;

// Now read the tiles

    if(fseek(fptr, tileTable[m], SEEK_SET))  { // find tile m
      printf("TIFF_ReadTiffTiles() error fseek-ing to tile %d\n", m+1);
      free(tileBuffer);
      free(tileTable);
      return(0);
    }

    if(fread(tileBuffer, (sizeof(UBYTE)*tileWidth*depth)
                , tileLength, fptr) != tileLength)  {
      perror("TIFF_ReadTiffTiles - fread error on tile");
      free(tileBuffer);
      free(tileTable);
      return(0);
    }

// Now copy tile buffer into main array
    imageX= (m % tilesAcross) * tileWidth;
    imageY= (m / tilesAcross) * tileLength;
#ifdef ORIG_CONWAY_METHOD
    CopyArray( buffer, tifInfo->ncol, tifInfo->nrow
             , tileBuffer, tileWidth, tileLength
             , imageX, imageY
             );
#else
    if(depth==1)
      {/* Just use conway method */					
	CopyArray( buffer, tifInfo->ncol, tifInfo->nrow
		  , tileBuffer, tileWidth, tileLength
		  , imageX, imageY
		  );
      }/* Just use conway method */					
    else
      {/* Try it this way */
	for(i=0;i<tileWidth;i++)
	  for(j=0;j<tileLength;j++)
	    {
	      if((i+imageY < tifInfo->nrow)  &&
		 (j+imageX < tifInfo->ncol))
		{/* OK to do */
#ifdef sgi
		  buffer[((i+imageY)*(tifInfo->ncol)*depth)+
			 ((j+imageX)*depth)] =
			   (UBYTE)tileBuffer[((i*tileWidth*depth)+
					      ((j*depth)+1))];
		  buffer[((i+imageY)*(tifInfo->ncol)*depth)+
			 (((j+imageX)*depth)+1)] =
			   (UBYTE)tileBuffer[((i*tileWidth*depth)+
					      (j*depth))];
#else
		  buffer[((i+imageY)*(tifInfo->ncol)*depth)+
			 ((j+imageX)*depth)] =
			   (UBYTE)tileBuffer[((i*tileWidth*depth)+
					      (j*depth))];
		  buffer[((i+imageY)*(tifInfo->ncol)*depth)+
			 (((j+imageX)*depth)+1)] =
			   (UBYTE)tileBuffer[((i*tileWidth*depth)+
					      ((j*depth)+1))];
#endif
		}/* OK to do */
	    }
	  
      }/* Try it this way */

#endif

  }

  printf(" 100%%\n");
  free(tileBuffer);
  free(tileTable);
  return(1);
}
/****************************************************************************
** ReadTiffStrips
** ==============
*/
static
int ReadTiffStrips( FILE *fptr
                  , UBYTE *buffer
                  , int dx, int dy
                  , tifInfo_t *tifInfo
                  )
{
  int m, tenth, nextPercent;
  int *stripTable, *stripSizeTable;
  int stripsInImage, countRead;
  const int rowsPerStrip = tifInfo->rowsPerStrip;
  UBYTE *b;

// Gather strip information

  stripsInImage = (tifInfo->nrow + rowsPerStrip - 1) / rowsPerStrip;
  printf("Strips: %d rows/strip, %d strips/image\n", rowsPerStrip, stripsInImage);

  stripTable = malloc(stripsInImage * sizeof(*stripTable));
  if(stripTable==NULL)  {
    printf("TIFF_ReadImageData() - Can't allocate strip table\n");
    return(0);
  }
  stripSizeTable = malloc(stripsInImage * sizeof(*stripSizeTable));
  if(stripSizeTable==NULL)  {
    printf("TIFF_ReadImageData() - Can't allocate strip size table\n");
    free(stripTable);
    return(0);
  }

// Read strip offsets and sizes. If one strip, values are in pointers.

  if(stripsInImage == 1)  {
    *stripTable     = tifInfo->stripOffsets;
    *stripSizeTable = tifInfo->stripByteCounts;
  }
  else  {  // read tables
    if(fseek(fptr, tifInfo->stripOffsets, SEEK_SET))  { // find strip table
      printf("TIFF_ReadTiffStrips() error fseek-ing to strip table\n");
      free(stripTable);
      free(stripSizeTable);
      return(0);
    }

    if(fread(stripTable, sizeof(int), stripsInImage, fptr) != stripsInImage)  {
      printf("TIFF_ReadTiffStrips() error reading strip table\n");
      free(stripTable);
      free(stripSizeTable);
      return(0);
    }

    for(m=0; m<stripsInImage; m++)      // Fix up endian
      DifferentEndian((UBYTE *) &stripTable[m], 4, tifInfo->endianSwitch);

    if(fseek(fptr, tifInfo->stripByteCounts, SEEK_SET))  { // find strip size table
      printf("TIFF_ReadTiffStrips() error fseek-ing to strip size table\n");
      free(stripTable);
      free(stripSizeTable);
      return(0);
    }

    if(fread(stripSizeTable, sizeof(int), stripsInImage, fptr) != stripsInImage)  {
      printf("TIFF_ReadTiffStrips() error reading strip size table\n");
      free(stripTable);
      free(stripSizeTable);
      return(0);
    }

    for(m=0; m<stripsInImage; m++)      // Fix up endian
      DifferentEndian((UBYTE *) &stripSizeTable[m], 4, tifInfo->endianSwitch);
  }

// Read the image data

  printf("- Reading tiff data:");
  tenth = stripsInImage/10 + 1; // make sure 10% or more, but not less
  nextPercent = 0;
  b = buffer;
  for(m=0; m<stripsInImage; m++)  {     // Look at all strips
    if(m >= nextPercent)  {
      printf(" %d%%", (m*100)/stripsInImage);
      fflush(stdout);
      nextPercent += tenth;
    }

    if(fseek(fptr, stripTable[m], SEEK_SET))  { // find strip m
      printf("TIFF_ReadTiffStrips() error fseek-ing to strip %d\n", m+1);
      free(stripTable);
      free(stripSizeTable);
      return(0);
    }

    countRead = fread(b, sizeof(UBYTE), stripSizeTable[m], fptr);
    if(countRead != stripSizeTable[m])  {
      perror("ReadTiff - fread error");
      printf("ERROR - only %d of %d pixels read\n", countRead, stripSizeTable[m]);
      free(stripTable);
      free(stripSizeTable);
      return(0);
    }
    b += stripSizeTable[m];
  }
  printf(" 100%%\n");
  free(stripTable);
  free(stripSizeTable);
  return(1);
}
/****************************************************************************
** ReadImageHeader
** ===============
*/
static
int ReadImageHeader(FILE *fptr, tifInfo_t *tifInfo, VERBOSE verbose)
{
  unsigned short numDirEntries;
  tifTag_t taglist[kMAX_DIR_ENTRIES];
  int firstIFD, m;
  UWORD shortVal;
#ifdef DEBUG
  union {
    UWORD shortInt;
    UBYTE byte[2];
  } byteWord;
#endif
// Read off the header

  fread(&shortVal, sizeof(shortVal), 1, fptr);
  switch(shortVal)  {
  case 0x4949:  // little-endian ('II')
    if(Util_IAmBigEndian())
      tifInfo->endianSwitch = true;
    else
      tifInfo->endianSwitch = false;
    if(verbose)
      printf("File is little-endian, endian-swap = %s\n"
                        , (tifInfo->endianSwitch) ? "TRUE" : "FALSE");
    break;
  case 0x4d4d:  // big-endian ('MM')
    if(Util_IAmBigEndian())
      tifInfo->endianSwitch = false;
    else
      tifInfo->endianSwitch = true;
    if(verbose)
      printf("File is big-endian, endian-swap = %s\n"
                        , (tifInfo->endianSwitch) ? "TRUE" : "FALSE");
    break;
  default:
    printf("Error - file header starts with $'%4hX'", shortVal);
    printf("   but it must be either '$4949' (little-endian) or '$4D4D' (big-endian)\n");
    return(0);
  }
  fread(&shortVal, sizeof(shortVal), 1, fptr);
#ifdef DEBUG
  printf("Reading magic number (42) - raw value is %hd\n", shortVal);
  byteWord.shortInt = shortVal;
  printf("Bytes are %d and %d\n", (int) byteWord.byte[0], (int) byteWord.byte[1]);
#endif
  DifferentEndian((UBYTE *)&shortVal, 2, tifInfo->endianSwitch);
#ifdef DEBUG
  printf("Magic number (42) after checking byte-order is %hd\n", shortVal);
#endif
  if(verbose)
    printf("2nd word = %hd\n", shortVal);
  if(shortVal != 42)  {         // it should be 42
    printf("ERROR = 2nd word should be 42\n");
    if(shortVal == 0x2a00)      // 42 = 002a
      printf("Byte-order is being interpreted wrongly. Contact the author.\n");
    return(0);
  }
  fread(&firstIFD, sizeof(int), 1, fptr);
  DifferentEndian((UBYTE *)&firstIFD, 4, tifInfo->endianSwitch);

// Read TIFF ifd

  fseek(fptr, firstIFD, 0);
  fread(&numDirEntries, sizeof(short), 1, fptr);
  DifferentEndian((UBYTE *) &numDirEntries, 2, tifInfo->endianSwitch);
  if(numDirEntries > kMAX_DIR_ENTRIES)  {       // overflow
    printf("Too many tags [%hd]\n", numDirEntries);
    return(0);
  }
  for(m=0; m<numDirEntries; m++)  {     // Read each tag
    short thisShort;
    const int endianSwitch = tifInfo->endianSwitch;     // to save typing!
    fseek(fptr, (firstIFD+2+(m*12)), 0);
    fread(&thisShort, sizeof(short), 1, fptr);
    DifferentEndian((UBYTE *)&thisShort, 2, endianSwitch);
    taglist[m].tagid = thisShort;
    fread(&thisShort, sizeof(short), 1, fptr);
    DifferentEndian((UBYTE *)&thisShort, 2, endianSwitch);
    taglist[m].dataType = thisShort;
    fread(&(taglist[m].dataCount), sizeof(int), 1, fptr);
    DifferentEndian((UBYTE *)&(taglist[m].dataCount), 4, endianSwitch);
    fread(&(taglist[m].dataOffset), sizeof(int), 1, fptr);
    switch(taglist[m].dataType)  {
      case 1:  // unsigned byte
      case 6:  // signed byte
      case 7:  // undefined 8-bit
        break;    // do nothing
      case 3:  // unsigned short
      case 8:  // signed short
        DifferentEndian((UBYTE *)&taglist[m].dataOffset, 2, endianSwitch);
        break;
      case 2:  // offset to char (null-terminated string)
      case 4:  // unsigned long
      case 5:  // offset to RATIONAL (2 unsigned longs: numerator and denominator)
      case 9:  // signed long
      case 10: // offset to signed RATIONAL (2 signed longs: numerator and denominator)
      case 11: // float (4-byte IEEE)
      case 12: // offset to double (8-byte IEEE)
        DifferentEndian((UBYTE *)&(taglist[m].dataOffset), 4, endianSwitch);
        break;
      default:
        printf("TIFF-format error - unknown data type %d\n", taglist[m].dataType);
        break;
    }
#ifdef DEBUG
    {
      printf("id=%d  type=%d  count=%d  offset=%d ($%0x)\n"
        , taglist[m].tagid
        , taglist[m].dataType
        , taglist[m].dataCount
        , taglist[m].dataOffset
        , taglist[m].dataOffset
        );
    }
#endif
    TiffTagType(fptr, taglist[m], tifInfo, verbose);
  }

// Let's check to see we've got all the info we need

  if(tifInfo->tiles)  {
    if(tifInfo->tileWidth  == 0 ||
       tifInfo->tileLength == 0 ||
       tifInfo->ncol==0         ||
       tifInfo->nrow==0)  {
      printf("Not enough info in header to continue...\n");
      return(0);
    }
  }
  else if(tifInfo->strips)  {
    if(tifInfo->ncol==0         ||
       tifInfo->nrow==0)  {
      printf("Not enough info in header to continue...\n");
      return(0);
    }
  }
  else  {
    printf("Neither strips nor tiles defined\n");
    printf("Not enough info in header to continue...\n");
    return(0);
  }

  if(tifInfo->planarConfig != 1)  {
    printf("Not configured for this planar config [%d]\n", tifInfo->planarConfig);
    return(0);
  }

  if(tifInfo->sampleFormat > 1)  {
    printf("Not configured for this sample format [%d]\n", tifInfo->sampleFormat);
    return(0);
  }

  return(1);
}
/*****************************************************************************
** CopyArray
** =========
Copy all of src (size sdx,sdy) into dst (size ddx,ddy) starting at
dst(dx0,dy0). Make sure we copy only the areas that overlap.
*/
static
void CopyArray( UBYTE *dst, int ddx, int ddy
              , UBYTE *src, int sdx, int sdy
              , int dx0, int dy0
              )
{
  UBYTE *d, *s;
  const int dx1 = MIN(ddx, dx0+sdx);    // clip to dst boundaries
  const int dy1 = MIN(ddy, dy0+sdy);
  const int dinc = ddx - (dx1-dx0);
  const int sinc = sdx - (dx1-dx0);     // this may be zero
  int x, y;

  d = dst + dy0*ddx + dx0;              // start byte of dst(dx0,dy0)
  s = src;                              // start byte of src(0,0)
  for(y=dy0; y<dy1; y++, d+=dinc, s+=sinc)      // inc d&s to next line
    for(x=dx0; x<dx1; x++)
      *d++ = *s++;
}
/*****************************************************************************
** TiffTagType
** ===========
Note: dataOffset has not been endian-corrected. We need to check on the size
of any integer in it before doing this. 2-byte integers should only have the
1st 2 bytes swapped, but 4-byte integers should have all 4 bytes swapped.
*/
static
void TiffTagType(FILE *fptr, tifTag_t tag, tifInfo_t *tifInfo, VERBOSE verbose)
{
  int value=-1;

  if(verbose)
    printf("Tag %d: ", (int) tag.tagid);

  switch(tag.dataType)  {
    case 3:     // unsigned short
    {
      union {
        UWORD twoShorts[2];
        ULONG oneLong;
      } wordLong;
      wordLong.oneLong = tag.dataOffset;
      value = wordLong.twoShorts[0];
      break;
    }
    case 2:  // offset to char (null-terminated string)
    case 4:  // unsigned long
    case 5:  // offset to RATIONAL (2 unsigned longs: numerator and denominator)
    case 10: // offset to signed RATIONAL (2 signed longs: numerator and denominator)
    case 12: // offset to double (8-byte IEEE)
      value = tag.dataOffset;
      break;
    default:
      printf("TIFF TiffTagType does not handle data type %d\n", tag.dataType);
      return;
  }

  switch(tag.tagid)  {
    case 254:   // New Subfile Type
      tifInfo->newSubfileType = value;
      if(verbose)
        printf("NewSubfile type = %x\n", tifInfo->newSubfileType);
      break;
    case 256:   // width
      tifInfo->ncol = value;
      if(verbose)
        printf("Image width = %d\n", tifInfo->ncol);
      break;
    case 257:   // height
      tifInfo->nrow = value;
      if(verbose)
        printf("Image height = %d\n", tifInfo->nrow);
      break;
    case 258:   // bitspersample
      tifInfo->bitsPerSample = value;
      if(verbose)
        printf("Bits-per-sample = %d\n", tifInfo->bitsPerSample);
      break;
    case 259:   // compression
      tifInfo->compression = value;
      if(verbose)  {
        printf("Compression = ");
        switch(tifInfo->compression)  {
          case 1 :
            printf("uncompressed\n");
            break;
          case 2 :
            printf("CCITT 1D\n");
            break;
          case 3 :
            printf("CCITT grp3\n");
            break;
          case 4 :
            printf("CCITT grp4\n");
            break;
          case 5 :
            printf("lzw compression\n");
            break;
          case 6 :
            printf("jpeg\n");
            break;
          case 32771 :
            printf("uncompressed222\n");
            break;
          case 32773 :
            printf("packbits\n");
            break;
          default :
            printf("unknown\n");
            break;
        }
      }
      break;
    case 262:   // photometricinterpretation
      tifInfo->pmi = value;
      if(verbose)  {
        printf("Photometric Interpretation: ");
        switch(tifInfo->pmi) {
          case 0 :
            printf("White is Zero.\n");
            break;
          case 1 :
            printf("Black is zero.\n");
            break;
          case 2 :
            printf("rgb.\n");
            break;
          case 3 :
            printf("rgb pallette.\n");
            break;
          case 4 :
            printf("transparancey mask.\n");
            break;
          case 5 :
            printf("cmyk.\n");
            break;
          case 6 :
            printf("ycbcr.\n");
            break;
          case 8 :
            printf("cielab.\n");
            break;
          default :
            printf("unknown.\n");
            break;
        }
      }
      break;
    case 270:   // Comment string (null-terminated - count includes NULL)
      if(verbose)  {
        char *string = TiffTagReadString(fptr, tag);
        printf("Image Description String: ");
        if(string == NULL)
          printf("can't allocate memory to print it\n");      
        else  {
          printf("'%s'\n", string);
          free(string);
        }
      }
      break;
    case 273:   // StripOffsets
      tifInfo->stripOffsets = value;
      tifInfo->strips = true;
      if(verbose)
        printf("Strip Offsets = %d\n", tifInfo->stripOffsets);
      break;
    case 274:   // orientation
      tifInfo->orientation = value;
      if(verbose)  {
        printf("Orientation = %d (", tifInfo->orientation);
        switch(tifInfo->orientation)  {
        case 1:
          printf("portrait: top-down, left-to-right");
          break;
        case 2:
          printf("portrait: top-down, right-to-left");
          break;
        case 3:
          printf("portrait: bottom-up, right-to-left");
          break;
        case 4:
          printf("portrait: bottom-up, left-to-right");
          break;
        case 5:
          printf("landscape: left-to-right, top-down");
          break;
        case 6:
          printf("landscape: right-to-left, top-down");
          break;
        case 7:
          printf("landscape: right-to-left, bottom-up");
          break;
        case 8:
          printf("landscape: left-to-right, bottom-up");
          break;
        }
        printf(")\n");
      }
      break;
    case 277:   // SamplesPerPixel
      tifInfo->samplesPerPixel = value;
      if(verbose)
        printf("Samples per pixel = %d\n", tifInfo->samplesPerPixel);
      break;
    case 278:   // Rows/strip
      tifInfo->rowsPerStrip = value;
      tifInfo->strips = true;
      if(verbose)
        printf("Rows-per-strip = %d\n", tifInfo->rowsPerStrip);
      break;
    case 279:   // StripByteCounts
      tifInfo->stripByteCounts = value;
      tifInfo->strips = true;
      if(verbose)
        printf("Strip-byte-counts = %d\n", tifInfo->stripByteCounts);
      break;
    case 280: { // min sample value
        const short minValue = value;
        if(verbose)
          printf("MinSampleValue = %d\n", minValue);
      }
      break;
    case 281: { // max sample value
        const short maxValue = value;
        if(verbose)
          printf("MaxSampleValue = %d\n", maxValue);
      }
      break;
    case 282:   // x-resolution
      if(verbose)
        printf("X-resolution - skip\n");
      break;    // skip
    case 283:   // y-resolution
      if(verbose)
        printf("Y-resolution - skip\n");
      break;    // skip
    case 284:   // planar configuration
      tifInfo->planarConfig = value;
      if(verbose)
        printf("Planar configuration = %d\n", tifInfo->planarConfig);
      break;
    case 296:   // resolution unit
      if(verbose)
        printf("Resolution unit - skip\n");
      break;    // skip
    case 305:   // software
      if(verbose)  {
        char *string = TiffTagReadString(fptr, tag);
        printf("Software Description String: ");
        if(string == NULL)
          printf("can't allocate memory to print it\n");      
        else  {
          printf("'%s'\n", string);
          free(string);
        }
      }
      break;
    case 322:   // tileWidth
      tifInfo->tileWidth = value;
      tifInfo->tiles = true;
      if(verbose)
        printf("Tile width = %d\n", tifInfo->tileWidth);
      break;
    case 323:   // tileLength
      tifInfo->tileLength = value;
      tifInfo->tiles = true;
      if(verbose)
        printf("Tile length = %d\n", tifInfo->tileLength);
      break;
    case 324:   // tileOffsets
      tifInfo->tileOffsets = value;
      tifInfo->tiles = true;
      if(verbose)
        printf("Tile offsets = %d\n", tifInfo->tileOffsets);
      break;
    case 325:   // tileByteCounts
      tifInfo->tileByteCounts = value;
      tifInfo->tiles = true;
      if(verbose)
        printf("Tile byte-counts = %d\n", tifInfo->tileByteCounts);
      tifInfo->tileByteCountsType = tag.dataType;
      break;
    case 339:   // SampleFormat
      tifInfo->sampleFormat = value;
      if(verbose)  {
        printf("Sample Format = %d: ", tifInfo->sampleFormat);
        switch(tifInfo->sampleFormat)  {
          case 1:
            printf("unsigned integer data");
            break;
          case 2:
            printf("two's complement signed integer data");
            break;
          case 3:
            printf("IEEE floating point data");
            break;
          case 4:
            printf("undefined data format");
            break;
          default:
            printf("unknown sample format number");
            break;
        }
        printf("\n");
      }
      break;
    case 32997: // Terminating tag?
      if(verbose)
        printf("Terminating tag (32997)??? \n");
//    i = (UWORD)((tag.dataOffset & 0xFFFF0000) >> 16);
      break;
    default:
      if(verbose)
        printf("Don't know tag [%d]\n", tag.tagid);
      break;
  }
}
/*****************************************************************************
** TiffTagReadString
** =================
*/
static
char *TiffTagReadString(FILE *fptr, tifTag_t tag)
{
  char *string = NULL;

  if(tag.dataCount <= 0)
    return(string);

  if((string = malloc(tag.dataCount)) != NULL)  {
    if(tag.dataCount<=4)
      strncpy(string, (char *) &tag.dataOffset, tag.dataCount);
    else  {
      fseek(fptr, tag.dataOffset, SEEK_SET);
      fread(string, sizeof(*string), tag.dataCount, fptr);
    }
  }
  string[tag.dataCount] = '\0';   // ensure null-termination
  return(string);
}
/*****************************************************************************
** Util_IAmBigEndian
** =================
Return true if this is a big-endian machine (byte order is hi/lo), else false
(byte order is lo/hi).
*/
static
int Util_IAmBigEndian(void)
{
  static int firstTime = true;
  static int bigEndian;

  if(firstTime)  {
    union {
      unsigned char  bytes[2];
      unsigned short word;
    } byteWord;
    firstTime = false;
    byteWord.word = 1;
    bigEndian = (byteWord.bytes[1] == 1);
  }
  return(bigEndian);
}
/*****************************************************************************
** DifferentEndian
** ===============
*/
static
void DifferentEndian(UBYTE *inStr, int length, int endianSwitch)
{
  int i, lm1;
  UBYTE thisStr[8];

  if(endianSwitch)  {
    for(i=0; i<length; i++)
      thisStr[i] = inStr[i];
    lm1 = length - 1;
    for(i=0; i<length; i++)
      inStr[i] = thisStr[lm1-i];
  }
}
/*****************************************************************************
** ZeroTifInfo
** ===========
*/
static
void ZeroTifInfo(tifInfo_t *tifInfo)
{
  tifInfo->bitsPerSample        = 1;    // bit-map
  tifInfo->compression          = 1;    // none
  tifInfo->pmi                  = 1;    // strictly, no default
  tifInfo->orientation          = 1;    // portrait, top-down, left-to-right
  tifInfo->samplesPerPixel      = 1;    // 1 for bit,grey,indexed, 3 for rgb
  tifInfo->planarConfig         = 1;    // chunky
  tifInfo->tileWidth            = 0;
  tifInfo->tileLength           = 0;
  tifInfo->tileOffsets          = 0;
  tifInfo->tileByteCounts       = 0;
  tifInfo->tileByteCountsType   = 0;
  tifInfo->endianSwitch         = 0;
  tifInfo->ncol                 = 0;            // No default
  tifInfo->nrow                 = 0;            // No default
  tifInfo->firstIFD             = 0;
  tifInfo->newSubfileType       = 0;
  tifInfo->stripOffsets         = 0;            // No default
  tifInfo->rowsPerStrip         = 0x7FFFFFFF;   // 2^32 - 1 (as signed, 2^31-1)
  tifInfo->stripByteCounts      = 0;            // No default
  tifInfo->strips               = false;
  tifInfo->tiles                = false;
  tifInfo->sampleFormat         = 0;
}
/****************************************************************************/
