/******************************************************************************
** array.c
** =======

Array Utility Routines
James Conway, LMES/IBS, Grenoble, France (email: James.Conway@ibs.fr)

06-Mar-2000: Add CopyArrayByteToFloat, CopyArrayFloatToByte
             Rename old CopyArrayFloatToByte to CopyArrayFloatToUByte
             Add GetStatsByteArray, rename old GetStatsByteArray to GetStatsUByteArray
07-Mar-2000: Add GetStatsWordArray and GetStatsUWordArray
09-Mar-2000: Add GetStatsFloatArray3, change "char" to "signed char"

void   CopyArrayByteToFloat  (   signed char *src, float *dst, long size );
void   CopyArrayUByteToFloat ( unsigned char *src, float *dst, long size );
void   CopyArrayFloatToUByte ( float *src, unsigned char *dst, long size );
void   CopyArrayFloatToByte  ( float *src,   signed char *dst, long size );
void   CopyArrayShortToFloat ( short *src, float *dst, long size );
void   CopyArrayFloatToShort ( float *src, short *dst, long size );
void   CopyArrayFloatToFloat ( float *src, float *dst, long size );

void   GetStatsByteArray     (   signed char *array, long size,   signed char *min,   signed char *max, float *mean, float *stdv );
void   GetStatsUByteArray    ( unsigned char *array, long size, unsigned char *min, unsigned char *max, float *mean, float *stdv );
void   GetStatsWordArray     ( short *array, long size, short *min, short *max, float *mean, float *stdv );
void   GetStatsUWordArray    ( unsigned short *array, long size, unsigned short *min, unsigned short *max, float *mean, float *stdv );
void   GetStatsFloatArray    ( float *array, long size, float *min, float *max, float *mean, float *stdv );
void   GetStatsFloatArray2   ( float *array, long dx, long dy, float radius, float *min, float *max, float *mean, float *stdv );
void   GetStatsFloatArray3   ( float *array, long dx, long dy, float radius, float *min, float *max, float *mean, float *stdv );

******************************************************************************/

#include <math.h>
#include "array.h"

/******************************************************************************
** CopyArrayByteToFloat
** ====================
Copy array of signed byte to array of float. Arrays can be overlapping as
long as smaller (byte) data is at end of array space. 
*/
void CopyArrayByteToFloat(signed char *src, float *dst, long size)
{
  signed char *s;
  float *d;

  s = src + size;
  d = dst + size;
  while(s>src)
    *(--d) = (float) *(--s);
}
/******************************************************************************
** CopyArrayUByteToFloat
** =====================
Copy array of unsigned byte to array of float. Arrays can be overlapping as
long as smaller (byte) data is at end of array space. 
*/
void CopyArrayUByteToFloat(unsigned char *src, float *dst, long size)
{
  unsigned char *s;
  float *d;

  s = src + size;
  d = dst + size;
  while(s>src)
    *(--d) = (float) *(--s);
}
/******************************************************************************
** CopyArrayFloatToByte
** ====================
Copy array of float to array of signed byte. Arrays can be overlapping as
long as smaller (short) data is at beginning of array space.
*/
void CopyArrayFloatToByte(float *src, signed char *dst, long size)
{
  float *s, temp;
  signed char *d;

  s = src;
  d = dst;
  while((size--) > 0)  {
    temp = *(s++);
    if(temp<-128.0)
      *d = -128;
    else if(temp>127.0)
      *d = 127;
    else
      *d = (signed char) temp;
    d++;
  }
}
/******************************************************************************
** CopyArrayFloatToUByte
** =====================
Copy array of float to array of unsigned byte. Arrays can be overlapping as
long as smaller (short) data is at beginning of array space.
*/
void CopyArrayFloatToUByte(float *src, unsigned char *dst, long size)
{
  float *s, temp;
  unsigned char *d;

  s = src;
  d = dst;
  while((size--) > 0)  {
    temp = *(s++);
    if(temp<0.0)
      *d = 0;
    else if(temp>255.0)
      *d = 255;
    else
      *d = (unsigned char) temp;
    d++;
  }
}
/******************************************************************************
** CopyArrayShortToFloat
** =====================
Copy array of short to array of float. Arrays can be overlapping as long as
smaller (short) data is at beginning of array space. 
*/
void CopyArrayShortToFloat(short *src, float *dst, long size)
{
  short *s;
  float *d;

  s = src + size;
  d = dst + size;
  while(s>src)
    *(--d) = (float) *(--s);
}
/******************************************************************************
** CopyArrayFloatToShort
** =====================
Copy array of float to array of short. Arrays can be overlapping as long as
smaller (short) data is at beginning of array space. 
*/
void CopyArrayFloatToShort(float *src, short *dst, long size)
{
  float *s;
  short *d;

  s = src;
  d = dst;
  while((size--) > 0)
    *(d++) = (short) *(s++);
}
/******************************************************************************
** CopyArrayFloatToFloat
** =====================
Copy array of float to array of float. 
*/
void CopyArrayFloatToFloat(float *src, float *dst, long size)
{
  float *s;
  float *d;

  s = src;
  d = dst;
  while((size--) > 0)
    *d++ = *s++;
}
/******************************************************************************
** GetStatsByteArray
** =================
*/
void GetStatsByteArray( signed char *array, long size
                      , signed char  *min,  signed char  *max
                      , float *mean, float *stdv
                      )
{
  long i;
  signed long suml, sumh;
  unsigned long sum2l, sum2h;
  double sum, sum2;
  signed char lmin, lmax;

  suml  = 0;    // sum - lo 4 bytes
  sumh  = 0;    // sum - hi 4 bytes
  sum2l = 0;    // sum squared - lo 4 bytes
  sum2h = 0;    // sum squared - hi 4 bytes

  *min  = *max  = 0;
  *mean = *stdv = 0.0;    // defaults in case of size=0 or 1

  if(size<1)
    return;

  lmin = lmax = *array;

  for(i=0; i<size; i++)  {
    short value = *array++;
    if     (value < lmin)  lmin = value;
    else if(value > lmax)  lmax = value;
    suml  = suml  + value;
    sum2l = sum2l + value*value;
    if(suml>2000000000)  {        // avoid overflow of 4-byte integer
      sumh = sumh + 1;
      suml = suml - 2000000000;
    }
    else if(suml<-2000000000)  {  // avoid overflow of 4-byte integer
      sumh = sumh - 1;
      suml = suml + 2000000000;
    }
    if(sum2l>2000000000)  {       // avoid overflow of 4-byte integer
      sum2h = sum2h + 1;
      sum2l = sum2l - 2000000000;
    }
  }

  sum  = 2e9*((double) sumh)  + (double) suml;
  sum2 = 2e9*((double) sum2h) + (double) sum2l;

  *min = lmin;
  *max = lmax;

  *mean = sum / (double) size;
  if(size>1)
    *stdv = sqrt(  (sum2 - sum*sum/(double) size) / (double) (size-1)  );
}
/******************************************************************************
** GetStatsUByteArray
** ==================
*/
void GetStatsUByteArray( unsigned char *array, long size
                       , unsigned char *min, unsigned char *max
                       , float *mean, float *stdv
                       )
{
  long i;
  unsigned long suml, sumh, sum2l, sum2h;
  double sum, sum2;
  unsigned char lmin, lmax;

  suml  = 0;    // sum - lo 4 bytes
  sumh  = 0;    // sum - hi 4 bytes
  sum2l = 0;    // sum squared - lo 4 bytes
  sum2h = 0;    // sum squared - hi 4 bytes

  *min  = *max  = 0;
  *mean = *stdv = 0.0;    // defaults in case of size=0 or 1

  if(size<1)
    return;

  lmin = lmax = *array;

  for(i=0; i<size; i++)  {
    unsigned short value = *array++;
    if     (value < lmin)  lmin = value;
    else if(value > lmax)  lmax = value;
    suml  = suml  + value;
    sum2l = sum2l + value*value;
    if(suml>2000000000)  {        // avoid overflow of 4-byte integer
      sumh = sumh + 1;
      suml = suml - 2000000000;
    }
    if(sum2l>2000000000)  {       // avoid overflow of 4-byte integer
      sum2h = sum2h + 1;
      sum2l = sum2l - 2000000000;
    }
  }

  sum  = 2e9*((double) sumh)  + (double) suml;
  sum2 = 2e9*((double) sum2h) + (double) sum2l;

  *min = lmin;
  *max = lmax;

  *mean = sum / (double) size;
  if(size>1)
    *stdv = sqrt(  (sum2 - sum*sum/(double) size) / (double) (size-1)  );
}
/******************************************************************************
** GetStatsWordArray
** =================
*/
void GetStatsWordArray( short *array, long size
                      , short *min, short *max
                      , float *mean, float *stdv
                      )
{
  long i;
  signed long suml, sumh;
  unsigned long sum2l, sum2h;
  double sum, sum2;
  short lmin, lmax;

  suml  = 0;    // sum - lo 4 bytes
  sumh  = 0;    // sum - hi 4 bytes
  sum2l = 0;    // sum squared - lo 4 bytes
  sum2h = 0;    // sum squared - hi 4 bytes

  *min  = *max  = 0;
  *mean = *stdv = 0.0;    // defaults in case of size=0 or 1

  if(size<1)
    return;

  lmin = lmax = *array;

  for(i=0; i<size; i++)  {
    long value = *array++;
    if     (value < lmin)  lmin = value;
    else if(value > lmax)  lmax = value;
    suml  = suml  + value;
    sum2l = sum2l + value*value;
    if(suml>2000000000)  {  // avoid overflow of 4-byte integer
      sumh = sumh + 1;
      suml = suml - 2000000000;
    }
    else if(suml<-2000000000)  {  // avoid overflow of 4-byte integer
      sumh = sumh - 1;
      suml = suml + 2000000000;
    }
    if(sum2l>2000000000)  {  // avoid overflow of 4-byte integer
      sum2h = sum2h + 1;
      sum2l = sum2l - 2000000000;
    }
  }

  sum  = 2e9*((double) sumh)  + (double) suml;
  sum2 = 2e9*((double) sum2h) + (double) sum2l;

  *min = lmin;
  *max = lmax;

  *mean = sum / (double) size;
  if(size>1)
    *stdv = sqrt(  (sum2 - sum*sum/(double) size) / (double) (size-1)  );
}
/******************************************************************************
** GetStatsUWordArray
** ==================
*/
void GetStatsUWordArray( unsigned short *array, long size
                       , unsigned short *min, unsigned short *max
                       , float *mean, float *stdv
                       )
{
  long i;
  unsigned long suml, sumh, sum2l, sum2h;
  double sum, sum2;
  unsigned short lmin, lmax;

  suml  = 0;    // sum - lo 4 bytes
  sumh  = 0;    // sum - hi 4 bytes
  sum2l = 0;    // sum squared - lo 4 bytes
  sum2h = 0;    // sum squared - hi 4 bytes

  *min  = *max  = 0;
  *mean = *stdv = 0.0;    // defaults in case of size=0 or 1

  if(size<1)
    return;

  lmin = lmax = *array;

  for(i=0; i<size; i++)  {
    unsigned long value = *array++;
    if     (value < lmin)  lmin = value;
    else if(value > lmax)  lmax = value;
    suml  = suml  + value;
    sum2l = sum2l + value*value;
    if(suml>2000000000)  {  // avoid overflow of 4-byte integer
      sumh = sumh + 1;
      suml = suml - 2000000000;
    }
    if(sum2l>2000000000)  {  // avoid overflow of 4-byte integer
      sum2h = sum2h + 1;
      sum2l = sum2l - 2000000000;
    }
  }

  sum  = 2e9*((double) sumh)  + (double) suml;
  sum2 = 2e9*((double) sum2h) + (double) sum2l;

  *min = lmin;
  *max = lmax;

  *mean = sum / (double) size;
  if(size>1)
    *stdv = sqrt(  (sum2 - sum*sum/(double) size) / (double) (size-1)  );
}
/******************************************************************************
** GetStatsFloatArray
** ==================
*/
void GetStatsFloatArray( float *array, long size
                       , float *min, float *max
                       , float *mean, float *stdv
                       )
{
  double sum=0.0, sum2=0.0;
  long lsize = size;

  *min = *max = *mean = *stdv = 0.0;    // defaults in case of error
  if(size<1)
    return;
  *min = *max = *array;
  while(--lsize >= 0)  {
    if(*array < *min)  *min = *array;
    else if(*array > *max)  *max = *array;
    sum  += *array;
    sum2 += (*array) * (*array);
    array++;
  }
  *mean = sum / (double) size;
  if(size>1)
    *stdv = sqrt(  (sum2 - sum*sum/(double) size) / (double) (size-1)  );
}
/******************************************************************************
** GetStatsFloatArray2
** ===================
Same as above, but uses values within a radius centered in the array.
*/
void GetStatsFloatArray2( float *array, long dx, long dy, float radius
                        , float *min, float *max, float *mean, float *stdv)
{
  double sum=0.0, sum2=0.0;
  int x, y;
  float x2, y2;
  const float xcen = ((float)(dx-1))/2.0, ycen = ((float)(dy-1))/2.0;
  const float rad2 = radius * radius;
  const long size = dx * dy;

  *min = *max = *mean = *stdv = 0.0;    // defaults in case of error
  if(size<1)
    return;

  *min = *max = *array;
  for(y=0; y<dy; y++)  {
    y2 = ((float) y) - ycen;
    y2 *= y2;
    for(x=0; x<dx; x++, array++)  {
      x2 = ((float) x) - xcen;
      x2 *= x2;
      if(x2+y2 < rad2)  {
        if(*array < *min)  *min = *array;
        else if(*array > *max)  *max = *array;
        sum  += *array;
        sum2 += (*array) * (*array);
      }
    }
  }
  *mean = sum / (double) size;
  if(size>1)
    *stdv = sqrt(  (sum2 - sum*sum/(double) size) / (double) (size-1)  );
}
/******************************************************************************
** GetStatsFloatArray3
** ===================
Same as above, but uses values *outside* a radius centered in the array.
*/
void GetStatsFloatArray3( float *array, long dx, long dy, float radius
                        , float *min, float *max, float *mean, float *stdv)
{
  double sum=0.0, sum2=0.0;
  int x, y;
  float x2, y2;
  const float xcen = ((float)(dx-1))/2.0, ycen = ((float)(dy-1))/2.0;
  const float rad2 = radius * radius;
  const long size = dx * dy;

  *min = *max = *mean = *stdv = 0.0;    // defaults in case of error
  if(size<1)
    return;

  *min = *max = *array;
  for(y=0; y<dy; y++)  {
    y2 = ((float) y) - ycen;
    y2 *= y2;
    for(x=0; x<dx; x++, array++)  {
      x2 = ((float) x) - xcen;
      x2 *= x2;
      if(x2+y2 > rad2)  {
        if(*array < *min)  *min = *array;
        else if(*array > *max)  *max = *array;
        sum  += *array;
        sum2 += (*array) * (*array);
      }
    }
  }
  *mean = sum / (double) size;
  if(size>1)
    *stdv = sqrt(  (sum2 - sum*sum/(double) size) / (double) (size-1)  );
}
