/****************************************************************************
** pif.c
** =====

This version: 16-Jun-2001 (v1.4.9)
James Conway, LMES/IBS, Grenoble, France (email: James.Conway@ibs.fr)
Modified from source code provided by Rob Ashmore, Purdue University, Indiana.

List of things still to do:
These include:
1) Add error checking everywhere
2) Incorporate latest changes from Purdue
3) Add Purdue interfaces as a compatibility mode (in a separate file?)

Public functions:
    PIF_NewPifFileHeader  All fields inited to 'base' values
    PIF_NewPifImageHeader All fields inited to 'base' values
    PIF_OpenForWrite      Open file and write header
    PIF_WriteImage        Write image header and image
    PIF_WriteImageFloat   Write image header and image from float data
    PIF_KillHeader        Deallocate space for color table (if any)
    PIF_ModeString        Given the mode, return a descriptive string
    PIF_IAmBigEndian      Return 1 if the machine is big-endian, 0 if little.
    
To do:
    PIF_SetupImageHeader  Work out image mean, std dev and other fields

Important note: We work on the assumption that the data type 'int' is
32-bits wide. We test that this is so when opening files. If its not, we abort.

Changes:
1.4.9 16-Jun-2001 ArraySwapBytesInt2 was only working on the first element!
1.4.8 09-Mar-2000 Add PIF_NewPifFileHeader & PIF_NewPifImageHeader
1.4.7 03-Mar-2000 Always interpret magic number in fileid as little-endian 08 00 00 00.
1.4.6 08-Feb-2000 Fixed printing of '%'s in progress statements.
1.4.5 07-Jul-1999 Char16toDouble() was commented out - I forget why - but its
                  needed for reading the floatscalefactor in from a PIF file.
                  Extraordinary! It is now uncommented.
1.4.4 09-Jun-1999 Added PIF_DoNotify and PIF_DoError routines to avoid dependence
                  on printf (which does not work well with Mac toolbox model on
                  CodeWarrior). These routines should be provided in a separate
                  system-specific file, or surrounded by #ifdef's.
1.4.3 09-Jun-1999 Fixed bug in ArrayInt2toFloat4 introduced in 1.4.2 which did
                  not let byte-swapping happen.
1.4.2 08-Jun-1999 Fixed bug in ArrayInt2toFloat4 where an int4 varaible was used
                  by mistake in the swapping. This worked when swapping bytes on
                  a little-endian system (VMS) but not on a big-endian (Mac). I
                  also changed the method of swapping, and in ArrayInt4toFloat4
                  and SwapBytesInt4, so that the compiler could not optimize away
                  the int variable by explicitly tying it to the associated char
                  array in a UNION statement.

****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

#include "pif_private.h"    // for this file only
#include "pif.h"
#include "pif_errors.h"     // PIF_DoError() and PIF_DoNotify()

#define PIF_PROGRAM_VERSION "1.4.9"

// Macros

#define MIN(m,n) ( ((m) < (n)) ? (m) : (n) )
#define MAX(m,n) ( ((m) > (n)) ? (m) : (n) )

// The PIF_FILE_HEADER_RAW structure shows the layout of the 512-byte file header.

                        typedef struct {
  UCHAR fileId[8];              // BYTE*8   Magic bytes, always 08 00 00 00 08 00 00 00
  char  realScaleFactor[16];    // char[16] How to convert floatInt*4 into real*4
  UCHAR numImages[4];           // int*4    Number of images in this file
  UCHAR endianNess[4];          // int*4    What endian type of machine created this file
                                //          0 = VAX/DECstation (Little)
                                //          1 = SGI. Mac etc... (Big)
  char  genProgram[32];         // char[32] Program used to generate this file with version number
  UCHAR hType[4];               // int*4    1 - all projections have the same number of pixels & depth
                                //          0 - otherwise
  UCHAR nx[4];                  // int*4    Number of columns
  UCHAR ny[4];                  // int*4    Number of rows
  UCHAR nz[4];                  // int*4    Number of sections
  UCHAR mode[4];                // int*4    EM data type e.g.
                                //           0 = byte*1
                                //           1 = int*2
                                //           2 = floatInt*4
                                //           3 = complex int*2
                                //           4 = complex floatInt*4
                                //           5 = Structure Factors
                                //           6 = Boxed data
                                //           7 = floatInt*2
                                //           8 = complex floatint*2
                                //           9 = float*4
                                //          10 = complex float*4
                                //          97 = DEPTHCUED
} PIF_FILE_HEADER_RAW;

typedef struct {
  PIF_FILE_HEADER_RAW data;
  UCHAR futureUse[512-sizeof(PIF_FILE_HEADER_RAW)]; // Space up to 512 bytes reserved for future use
} PIF_FILE_HEADER_RAW_512;


// The PIF_IMAGE_HEADER_RAW structure shows the layout of the 512-byte image header

typedef struct {
  UCHAR nx[4];                  // int*4        Number of columns
  UCHAR ny[4];                  // int*4        Number of rows
  UCHAR nz[4];                  // int*4        Number of sections
  UCHAR mode[4];                // int*4        EM data type e.g.
                                //                       0 = byte*1
                                //                       1 = int*2
                                //                       2 = floatInt*4
                                //                       3 = complex int*2
                                //                       4 = complex floatInt*4
                                //                       5 = Structure Factors
                                //                       6 = Boxed data
                                //                       7 = floatInt*2
                                //                       8 = complex floatint*2
                                //                       9 = float*4
                                //                      10= complex float*4
                                //                      97= DEPTHCUED
  UCHAR bkgnd[4];               // int*4        Background value
  UCHAR packRadius[4];          // int*4        Radius for boxed images
  UCHAR nxstart[4];             // int*4        Number of first col in map
  UCHAR nystart[4];             // int*4        Number of first row in map
  UCHAR nzstart[4];             // int*4        Number of first section in map
  UCHAR mx[4];                  // int*4        Number of intervals along x
  UCHAR my[4];                  // int*4        Number of intervals along y
  UCHAR mz[4];                  // int*4        Number of intervals along z
  UCHAR xlength[4];             // floatInt*4   Cell Dimensions (Angstroms)
  UCHAR ylength[4];             // floatInt*4           "
  UCHAR zlength[4];             // floatInt*4           "
  UCHAR alpha[4];               // floatInt*4   Cell Angles (degrees)
  UCHAR beta[4];                // floatInt*4           "
  UCHAR gamma[4];               // floatInt*4           "
  UCHAR mapc[4];                // int*4        Which axis is col(1,2,3 = x,y,z)
  UCHAR mapr[4];                // int*4        Which axis is row(1,2,3 = x,y,z)
  UCHAR maps[4];                // int*4        Which axis is sections(1,2,3 = x,y,z)
  UCHAR min[4];                 // floatInt*4   Min density value
  UCHAR max[4];                 // floatInt*4   Max density value
  UCHAR mean[4];                // floatInt*4   Mean density value
  UCHAR stdDev[4];              // floatInt*4   StdDev of GrayLevels
  UCHAR ispg[4];                // int*4        Space group number
  UCHAR nsymbt[4];              // int*4        Number of bytes for symmetry ops
  UCHAR xorigin[4];             // floatInt*4   x origin
  UCHAR yorigin[4];             // floatInt*4   y origin
  char  title[80];              // char[80]     User defined description
  char  timeStamp[32];          // char[32]     Date/time data last modified
  char  uGraphDesignation[16];  // char[16]     Unique Micrograph Number
  char  scanNumber[8];          // char[8]      Scan Number of Micrograph
} PIF_IMAGE_HEADER_RAW;

typedef struct {
  PIF_IMAGE_HEADER_RAW data;
  UCHAR futureUse[512-sizeof(PIF_IMAGE_HEADER_RAW)]; // Space up to 512 bytes reserved for future use
} PIF_IMAGE_HEADER_RAW_512;


// Constants
// =========

static const int EM_PURDUE_PORTABLE = 8;
static const int PIF_FALSE = 0;
static const int PIF_TRUE  = 1;

//***************************************************************************
//          P U B L I C   F U N C T I O N S
/****************************************************************************
** PIF_NewPifFileHeader
** ====================
Return an inited PIF_FILE_HEADER.
*/
PIF_FILE_HEADER PIF_NewPifFileHeader(void)
{
  PIF_FILE_HEADER pifHeader;
  pifHeader.fileId[0] = pifHeader.fileId[4] = EM_PURDUE_PORTABLE; // int  fileId;   kind of data format,...
  pifHeader.fileId[1] = pifHeader.fileId[5] = 0;                  //     ...restricted to 8,0,0,0 and in...
  pifHeader.fileId[2] = pifHeader.fileId[6] = 0;                  //     ...two copies
  pifHeader.fileId[3] = pifHeader.fileId[7] = 0;
  pifHeader.floatScaleFactor = 1.0;     // double floatScaleFactor;      scale factor for floatInt*4 into real*4
  pifHeader.numImages        = 0;       // int  numImages;               number of images in this file
  pifHeader.endianNess       = PIF_IAmBigEndian();  // int  endianNess;  endian type: 0 = little (VMS), else = big (Unix)
  pifHeader.genProgram[0]    = '\0';    // char genProgram[33];          name of program used to generate this file
  pifHeader.hType            = 1;       // int  hType;       1 - all projections have the same size & depth
  pifHeader.nx               = 0;       // int  nx;          Number of columns
  pifHeader.ny               = 0;       // int  ny;          Number of rows
  pifHeader.nz               = 0;       // int  nz;          Number of sections
  pifHeader.mode             = -1;      // int  mode;        EM data type (-1 is invalid)
  pifHeader.colorMap         = NULL;    // short *colorMap;  Color map, only included with DepthCued (*256*3 = 1536 bytes)

  return(pifHeader);
}
/****************************************************************************
** PIF_NewPifImageHeader
** =====================
Return an inited PIF_IMAGE_HEADER.
*/
PIF_IMAGE_HEADER PIF_NewPifImageHeader(void)
{
  const PIF_IMAGE_HEADER imgHeader = {
                             0       // int   nx;         Number of columns
                           , 0       // int   ny;         Number of rows
                           , 0       // int   nz;         Number of sections
                           , -1      // int   mode;       EM data type (-1 is invalid)
                           , 0       // int   bkgnd;      Background value
                           , 0       // int   packRadius; Radius for boxed images
                           , 0       // int   nxstart;    Number of first col in map
                           , 0       // int   nystart;    Number of first row in map
                           , 0       // int   nzstart;    Number of first section in map
                           , 0       // int   mx;         Number of intervals along x
                           , 0       // int   my;         Number of intervals along y
                           , 0       // int   mz;         Number of intervals along z
                           , 0.0     // float xlength;    Cell Dimensions (Angstroms)
                           , 0.0     // float ylength;    Cell Dimensions (Angstroms)
                           , 0.0     // float zlength;    Cell Dimensions (Angstroms)
                           , 90.0    // float alpha;      Cell Angles (degrees)
                           , 90.0    // float beta;       Cell Angles (degrees)
                           , 90.0    // float gamma;      Cell Angles (degrees)
                           , 1       // int   mapc;       Which axis is col(1,2,3 = x,y,z)
                           , 2       // int   mapr;       Which axis is row(1,2,3 = x,y,z)
                           , 3       // int   maps;       Which axis is sections(1,2,3 = x,y,z)
                           , 0       // float min;        Min density value
                           , 0       // float max;        Max density value
                           , 0       // float mean;       Mean density value
                           , 0       // float stdDev;     StdDev of density value
                           , 0       // int   ispg;       Space group number
                           , 0       // int   nsymbt;     Number of bytes for symmetry ops
                           , 0.0     // float xorigin;    x origin
                           , 0.0     // float yorigin;    y origin
                           , "\0"    // char  title[81];  User defined description
                           , "\0"    // char  timeStamp[33];          Date/time data last modified
                           , "\0"    // char  uGraphDesignation[17];  Unique Micrograph Number
                           , "\0"    // char  scanNumber[9];          Scan Number of Micrograph
  };
  return(imgHeader);
}
/****************************************************************************
** PIF_OpenForWrite
** ================
Open PIF file and write header.
*/
FILE *PIF_OpenForWrite( char *fname                     // file name
                      , const PIF_FILE_HEADER *pifHeader// header info to write
                      )
{
  FILE *fptr;

  if(TestIntIs32Bits() != PIF_OK)
    return(NULL);

  fptr = fopen(fname, "wb");
  if(fptr == NULL)  {
    PIF_DoError("PIF_OpenForWrite - error opening file");
    return(NULL);
  }

  if(WriteFileHeader(fptr, pifHeader) != PIF_OK)  {     // write file header
    PIF_DoError("PIF_OpenForWrite - error writing header, closing file");
    fclose(fptr);
    return(NULL);
  }

  if(pifHeader->mode == kPIF_MODE_DEPTHCUED)            // write color map
    if(WriteDepthCueColorMap(fptr, pifHeader) != PIF_OK)  {
      PIF_DoError("PIF_OpenForWrite - error writing DepthCued colormap, closing file");
      fclose(fptr);
      return(NULL);
    }

  return(fptr);
}
/****************************************************************************
** PIF_WriteImageHeader
** ====================
Write image header to next position in PIF file.
*/
int PIF_WriteImageHeader( FILE *fptr                          // open PIF file pointer
                        , const PIF_FILE_HEADER  * pifHeader   // file header info
                        , const PIF_IMAGE_HEADER * imgHeader   // image header info
                        )
{
  if(WriteImageHeader(fptr, pifHeader, imgHeader) != PIF_OK)  {         // write image header
    PIF_DoError("PIF_WriteImageHeader - error writing header");
    return(PIF_ERROR);
  }
  return(PIF_OK);
}
/****************************************************************************
** PIF_WriteImageData
** ==================
Write image data, assuming header has already been written, in PIF file. The
output format is specified by the image header, and the image data is assumed
to be the same.
*/
int PIF_WriteImageData( FILE *fptr                          // open PIF file pointer
                      , const PIF_FILE_HEADER  *pifHeader   // file header info
                      , const PIF_IMAGE_HEADER *imgHeader   // image header info
                      , void *buffer                        // image data
                      )
{
  if(WriteImageData(fptr, pifHeader, imgHeader, buffer) != PIF_OK)  {   // write image data
    PIF_DoError("PIF_WriteImageData - error writing image data");
    return(PIF_ERROR);
  }

  return(PIF_OK);
}
/****************************************************************************
** PIF_WriteImage
** ==============
Write image and header to next position in PIF file. The output format is
specified by the image header, and the image data is assumed to be the same.
*/
int PIF_WriteImage( FILE *fptr                          // open PIF file pointer
                  , const PIF_FILE_HEADER  *pifHeader   // file header info
                  , const PIF_IMAGE_HEADER *imgHeader   // image header info
                  , void *buffer                        // image data
                  )
{
  if(WriteImageHeader(fptr, pifHeader, imgHeader) != PIF_OK)  {         // write image header
    PIF_DoError("PIF_WriteImage - error writing header");
    return(PIF_ERROR);
  }

  if(WriteImageData(fptr, pifHeader, imgHeader, buffer) != PIF_OK)  {   // write image data
    PIF_DoError("PIF_WriteImage - error writing image data");
    return(PIF_ERROR);
  }

  return(PIF_OK);
}
/****************************************************************************
** PIF_WriteImageFloat
** ===================
Write image and header to next position in PIF file. The output format is
specified by the image header, and the image data is assumed to be float.
*/
int PIF_WriteImageFloat( FILE *fptr                     // open PIF file pointer
                  , const PIF_FILE_HEADER  *pifHeader   // file header info
                  , const PIF_IMAGE_HEADER *imgHeader   // image header info
                  , float *buffer                       // image data
                  )
{
  if(WriteImageHeader(fptr, pifHeader, imgHeader) != PIF_OK)  {         // write image header
    PIF_DoError("PIF_WriteImageFloat - error writing header");
    return(PIF_ERROR);
  }

  if(WriteImageDataFloat(fptr, pifHeader, imgHeader, buffer) != PIF_OK)  {   // write image data
    PIF_DoError("PIF_WriteImageFloat - error writing image data");
    return(PIF_ERROR);
  }

  return(PIF_OK);
}
/****************************************************************************
** PIF_KillHeader
** ==============
Deallocate space that might have been allocated for PIF header info. Currently
the only candidate space is the DEPTHCUED colorMap.
*/
void PIF_KillHeader(PIF_FILE_HEADER *pifHeader)     // header info
{
  if(pifHeader->mode == kPIF_MODE_DEPTHCUED)
    if(pifHeader->colorMap != NULL)
      free(pifHeader->colorMap);
  pifHeader->colorMap = NULL;       // mark as not allocated
}
/****************************************************************************
** PIF_PixelSizeRaw
** ================
Determine the image depth (in bytes) as stored in the file.
*/
int PIF_PixelSizeRaw(int mode)
{
  return(FindImageRawDepth(mode));
}
/****************************************************************************
** PIF_PixelSizeFinal
** ==================
Determine the image depth (in bytes) after conversion of FLOATINTs.
*/
int PIF_PixelSizeFinal(int mode)
{
  return(FindImageFinalDepth(mode));
}
/****************************************************************************
** PIF_ModeString
** ==============
Given the mode, return the address of a descriptive string.
*/
const char *PIF_ModeString(int mode)
{
  switch(mode)  {
    case kPIF_MODE_INT1:
      return("Unsigned Byte");
    case kPIF_MODE_INT2:
      return("SignedInt2");
    case kPIF_MODE_FLOATINT4:
      return("FloatInt4");
    case kPIF_MODE_CMPLXINT2:
      return("ComplexInt2");
    case kPIF_MODE_CMPLXFLOATINT4:
      return("ComplexInt4");
    case kPIF_MODE_SF:
      return("Structure Factors");
    case kPIF_MODE_BOXED:
      return("Boxed-Unsigned Byte");
    case kPIF_MODE_FLOATINT2:
      return("FloatInt2");
    case kPIF_MODE_CMPLXFLOATINT2:
      return("Complex FloatInt2");
    case kPIF_MODE_FLOAT4:
      return("Float4");
    case kPIF_MODE_CMPLXFLOAT4:
      return("Complex FloatInt4");
    case kPIF_MODE_MAPFLOATINT2:
      return("Map-FloatInt2");
    case kPIF_MODE_MAPFLOATINT4:
      return("Map-FloatInt4");
    case kPIF_MODE_DEPTHCUED:
      return("DepthCued");
    default:
      return("Unknown Mode");
  }
}
/****************************************************************************
** PIF_IAmBigEndian
** ================
Return TRUE if this is a big-endian machine (byte order is hi/lo), else FALSE
(byte order is lo/hi).
*/
int PIF_IAmBigEndian(void)
{
  static int firstTime = 1;
  static int bigEndian;

  if(firstTime == 1)  {
    union {
      unsigned char  bytes[2];
      unsigned short word;
    } byteWord;
    firstTime = 0;
    byteWord.word = 1;
    bigEndian = (byteWord.bytes[1] == 1);
  }
  return(bigEndian);
}
/****************************************************************************
        P R I V A T E   F U N C T I O N S
        * 1) Miscellaneous
          2) I/O Routines - Read
          3) I/O Routines - Write
          4) I/O Routines - Misc
          5) Datatype conversion

1) Miscellaneous
      TestIntIs32Bits
      EndianModeMismatch
      FindImageRawDepth
      FindImageFinalDepth
      FindImageDimensions
      CalcImageRawSizeInBytes
      CalcImageFinalSizeInBytes
      CopyPad
      StripPad
      MakePrintable
*/
/****************************************************************************
** TestIntIs32Bits
** ===============
Make sure an int is 32-bits, else routines here will fail.
*/
static
int TestIntIs32Bits(void)
{
  if(sizeof(int) != 4)  {
    char msg[200];
    sprintf(msg, "PIF Error - 'int' data type is not 32-bits: %d", (int) (sizeof(int)*8));
    PIF_DoError(msg);
    PIF_DoError("The PIF i/o routines require 32-bit 'int's");
    return(PIF_ERROR);
  }
  return(PIF_OK);
}
/****************************************************************************
** EndianModeMismatch
** ==================
Check if there is an endian mode mismatch between the machine and the PIF file
(pifHeader->endianMode). 0 means little endian, none-zero means big-endian.
*/
static
int EndianModeMismatch(const PIF_FILE_HEADER *pifHeader)
{
  if((pifHeader->endianNess == 0) && (!PIF_IAmBigEndian())) // little and little
    return(PIF_FALSE);
  if((pifHeader->endianNess != 0) && (PIF_IAmBigEndian()))  // big and big
    return(PIF_FALSE);
  return(PIF_TRUE);
}
/****************************************************************************
** FindImageRawDepth
** =================
Work out the image depth (in bytes) as stored in the file.
*/
static
int FindImageRawDepth(int mode)
{
  int depth;

  switch(mode)  {
    case kPIF_MODE_INT1:
    case kPIF_MODE_BOXED:
      depth = 1;
      break;
    case kPIF_MODE_INT2:
    case kPIF_MODE_FLOATINT2:
    case kPIF_MODE_MAPFLOATINT2:
      depth = 2;
      break;
    case kPIF_MODE_FLOATINT4:           // 4-byte ints -> 4-byte reals
    case kPIF_MODE_MAPFLOATINT4:        // 4-byte ints -> 4-byte reals
    case kPIF_MODE_CMPLXINT2:           // two 2-byte ints
    case kPIF_MODE_CMPLXFLOATINT2:      // two 2-byte ints -> two 4-byte reals
    case kPIF_MODE_FLOAT4:              // 4-byte reals
//    case kPIF_MODE_SF:
      depth = 4;
      break;
    case kPIF_MODE_CMPLXFLOATINT4:      // two 4-byte ints -> two 4-byte reals
    case kPIF_MODE_CMPLXFLOAT4:         // two 4-byte reals
      depth = 8;
      break;
    case kPIF_MODE_DEPTHCUED:   // three bytes: red/green/blue indices into color tables
      depth = 3;
      break;
    default:
      depth = 0;                        // unknown
  }
  return(depth);
}
/****************************************************************************
** FindImageFinalDepth
** ===================
Work out the image depth (in bytes) after converting floatint's to float's.
*/
static
int FindImageFinalDepth(int mode)
{
  int depth;

  switch(mode)  {
    case kPIF_MODE_INT1:
    case kPIF_MODE_BOXED:
      depth = 1;
      break;
    case kPIF_MODE_INT2:
      depth = 2;
      break;
    case kPIF_MODE_FLOATINT2:       // 2-byte ints -> 4-byte reals
    case kPIF_MODE_FLOATINT4:       // 4-byte ints -> 4-byte reals
    case kPIF_MODE_MAPFLOATINT2:    // 2-byte ints -> 4-byte reals
    case kPIF_MODE_MAPFLOATINT4:    // 4-byte ints -> 4-byte reals
    case kPIF_MODE_CMPLXINT2:       // two 2-byte ints
    case kPIF_MODE_FLOAT4:          // 4-byte reals
//    case kPIF_MODE_SF:
      depth = 4;
      break;
    case kPIF_MODE_CMPLXFLOATINT2:  // two 2-byte ints -> two 4-byte reals
    case kPIF_MODE_CMPLXFLOATINT4:  // two 4-byte ints -> two 4-byte reals
    case kPIF_MODE_CMPLXFLOAT4:     // two 4-byte reals
      depth = 8;
      break;
    case kPIF_MODE_DEPTHCUED:
      depth = 3;
      break;
    default:
      depth = 0;   // unknown
  }
  return(depth);
}
/****************************************************************************
** FindImageDimensions
** ===================
*/
static
ULONG FindImageDimensions(const PIF_IMAGE_HEADER *imgHeader)
{
  const ULONG nx=imgHeader->nx;
  const ULONG ny=imgHeader->ny;
  const ULONG nz=imgHeader->nz;
//if (imgHeader->mode == 21) return (nx*ny);
//else 
  return(nx*ny*nz);
}
/****************************************************************************
** CalcImageRawSizeInBytes
** =======================
The size as stored in the file.
*/
static
ULONG CalcImageRawSizeInBytes(const PIF_IMAGE_HEADER *imgHeader)
{
  const ULONG size  = FindImageDimensions(imgHeader);
  const int   depth = FindImageRawDepth(imgHeader->mode);  // 0 if unknown mode
  const ULONG bytes = ((ULONG) depth) * size;
  return(bytes);
}
/****************************************************************************
** CalcImageFinalSizeInBytes
** =========================
The size after expanding any floatint's to float's.
*/
static
ULONG CalcImageFinalSizeInBytes(const PIF_IMAGE_HEADER *imgHeader)
{
  const ULONG size  = FindImageDimensions(imgHeader);
  const int   depth = FindImageFinalDepth(imgHeader->mode); // 0 if unknown mode
  const ULONG bytes = ((ULONG) depth) * size;
  return(bytes);
}
/****************************************************************************
** CopyPad
** =======
Copy characters, trancating if necessary, padding if necessary with spaces.
*/
static
void CopyPad(char *dest, const char *src, const int count)
{
  const int slen = strlen(src);
  if(slen<count)  {
    memcpy(dest, src, slen);
    memset(dest+slen, ' ', count-slen);
  }
  else
    memcpy(dest, src, count);
}
/****************************************************************************
** StripPad
** ========
Strip trailing spaces.
*/
static
void StripPad(char *str)
{
  char *s = str + strlen(str);
  while(--s >= str)
    if(!isspace(*s))  // Find first non-white-space char from end of string
      break;
  *(s+1) = '\0';    
}

/****************************************************************************
        P R I V A T E   F U N C T I O N S
          1) Miscellaneous
          2) I/O Routines - Read
        * 3) I/O Routines - Write
          4) I/O Routines - Misc
          5) Datatype conversion

3) I/O Routines - Write
      WriteFileHeader
      WriteDepthCueColorMap
      WriteImageHeader
      WriteImageData
      WriteImageDataFloat
      ConvertAndWriteImageData
      ConvertAndWriteImageDataFloat
*/
/****************************************************************************
** WriteFileHeader
** ===============
Write 512-byte header.
*/
static
int WriteFileHeader( FILE *fptr                         // file pointer
                   , const PIF_FILE_HEADER *pifHeader   // file header info
                   )
{
  PIF_FILE_HEADER_RAW_512 rawHeader512;
  PIF_FILE_HEADER_RAW * const rawHeader = &rawHeader512.data;

  memset(&rawHeader512, 0, sizeof(rawHeader512));     // initialize to 0's

// set fileID without respect to endian mode
  rawHeader->fileId[0] = rawHeader->fileId[4] = EM_PURDUE_PORTABLE;  // 1st & 5th bytes
  rawHeader->fileId[1] = rawHeader->fileId[5] =                      // rest
  rawHeader->fileId[2] = rawHeader->fileId[6] =                      // rest
  rawHeader->fileId[3] = rawHeader->fileId[7] = 0;                   // rest

  {
    char floatString[1024];
    double fValue;
    int size;
    if(pifHeader->floatScaleFactor == 0.0)
      fValue = 1.0;
    else
      fValue = pifHeader->floatScaleFactor;

    sprintf(floatString, "%16E"                , fValue);
    size = sizeof(rawHeader->realScaleFactor);
    strncpy(rawHeader->realScaleFactor, floatString, size);
  }

  Int4toChar4(pifHeader->numImages,  rawHeader->numImages);

  {
    const int myEndian = PIF_IAmBigEndian() ? 1 : 0;
    Int4toChar4(myEndian, rawHeader->endianNess);
  }

  {
    char programString[1024];
    sprintf(programString,"JFC-PIF %s of %s", PIF_PROGRAM_VERSION, __DATE__);
    CopyPad(rawHeader->genProgram, programString, sizeof(rawHeader->genProgram));
  }

  Int4toChar4(pifHeader->hType, rawHeader->hType);
  Int4toChar4(pifHeader->nx,    rawHeader->nx);
  Int4toChar4(pifHeader->ny,    rawHeader->ny);
  Int4toChar4(pifHeader->nz,    rawHeader->nz);
  Int4toChar4(pifHeader->mode,  rawHeader->mode);

// Write the 512-byte header out

  if(WriteBufferToFile(fptr, &rawHeader512, sizeof(rawHeader512)) != PIF_OK)  {
    PIF_DoError("PIF Error - WriteFileHeader(), writing file header.");
    return(PIF_ERROR);
  }

  return(PIF_OK);
}
/****************************************************************************
** WriteDepthCueColorMap
** =====================
Write the DEPTHCUED color map, which follows the PIF file header. It is only
in the file if the image is a DEPTHCUED type. If so, we must call this routine
after writing the file header with WriteFileHeader().

The color map has 1-byte entries, so endian mode is not an issue.
*/
static
int WriteDepthCueColorMap( FILE *fptr                           // file pointer
                         , const PIF_FILE_HEADER *pifHeader     // file info
                         )
{
  const long bytes = 3 * 256 * sizeof(*pifHeader->colorMap);

// Check we have a DEPTHCUED file

  if(pifHeader->mode != kPIF_MODE_DEPTHCUED)  {
    PIF_DoError("WriteDepthCueColorMap - error, not a DEPTHCUED file");
    return(PIF_ERROR);
  }

// Check for a non-Null pointer

  if(pifHeader->colorMap == NULL)  {
    PIF_DoError("WriteDepthCueColorMap - no colormap");
    return(PIF_ERROR);
  }

// Write out the color map.

  if(WriteBufferToFile(fptr, pifHeader->colorMap, bytes) != PIF_OK)  {
    PIF_DoError("WriteDepthCueColorMap - error writing colormap");
    return(PIF_ERROR);
  }
  
  return(PIF_OK);
}
/****************************************************************************
** WriteImageHeader
** ================
Write 512-byte header for a particular image.
*/
static
int WriteImageHeader( FILE *fptr                        // file pointer
                    , const PIF_FILE_HEADER  *pifHeader // file header info
                    , const PIF_IMAGE_HEADER *imgHeader // image header info
                    )
{
  PIF_IMAGE_HEADER_RAW_512 rawHeader512;
  PIF_IMAGE_HEADER_RAW * const rawHeader = &rawHeader512.data;

  memset(&rawHeader512, 0, sizeof(rawHeader512)); // initialize to 0's

  Int4toChar4(imgHeader->nx, rawHeader->nx);
  Int4toChar4(imgHeader->ny, rawHeader->ny);
  Int4toChar4(imgHeader->nz, rawHeader->nz);

  Int4toChar4(imgHeader->mode, rawHeader->mode);

  Int4toChar4(imgHeader->bkgnd,      rawHeader->bkgnd);
  Int4toChar4(imgHeader->packRadius, rawHeader->packRadius);

  Int4toChar4(imgHeader->nxstart, rawHeader->nxstart);
  Int4toChar4(imgHeader->nystart, rawHeader->nystart);
  Int4toChar4(imgHeader->nzstart, rawHeader->nzstart);

  Int4toChar4(imgHeader->mx, rawHeader->mx);
  Int4toChar4(imgHeader->my, rawHeader->my);
  Int4toChar4(imgHeader->mz, rawHeader->mz);

  FloatInt4toChar4(imgHeader->xlength, pifHeader, rawHeader->xlength);
  FloatInt4toChar4(imgHeader->ylength, pifHeader, rawHeader->ylength);
  FloatInt4toChar4(imgHeader->zlength, pifHeader, rawHeader->zlength);

  FloatInt4toChar4(imgHeader->alpha, pifHeader, rawHeader->alpha);
  FloatInt4toChar4(imgHeader->beta,  pifHeader, rawHeader->beta);
  FloatInt4toChar4(imgHeader->gamma, pifHeader, rawHeader->gamma);

  Int4toChar4(imgHeader->mapc, rawHeader->mapc);
  Int4toChar4(imgHeader->mapr, rawHeader->mapr);
  Int4toChar4(imgHeader->maps, rawHeader->maps);

  FloatInt4toChar4(imgHeader->min,    pifHeader, rawHeader->min);
  FloatInt4toChar4(imgHeader->max,    pifHeader, rawHeader->max);
  FloatInt4toChar4(imgHeader->mean,   pifHeader, rawHeader->mean);
  FloatInt4toChar4(imgHeader->stdDev, pifHeader, rawHeader->stdDev);

  Int4toChar4(imgHeader->ispg,   rawHeader->ispg);
  Int4toChar4(imgHeader->nsymbt, rawHeader->nsymbt);

  FloatInt4toChar4(imgHeader->xorigin, pifHeader, rawHeader->xorigin);
  FloatInt4toChar4(imgHeader->yorigin, pifHeader, rawHeader->yorigin);

  CopyPad(rawHeader->title, imgHeader->title, sizeof(rawHeader->title));
  CopyPad(rawHeader->uGraphDesignation, imgHeader->uGraphDesignation, sizeof(rawHeader->uGraphDesignation));
  CopyPad(rawHeader->scanNumber, imgHeader->scanNumber, sizeof(rawHeader->scanNumber));

  {
    char todayDate[1024];
    time_t now = time(NULL);
    struct tm *date = localtime(&now);
    strftime(todayDate, 100, "%d-%b-%Y %H:%M", date);
    CopyPad(rawHeader->timeStamp, todayDate, sizeof(rawHeader->timeStamp));
  }

// Write the 512-byte header out

  if(WriteBufferToFile(fptr, &rawHeader512, sizeof(rawHeader512)) != PIF_OK)  {
    PIF_DoError("PIF Error - WriteImageHeader(), writing image header.");
    return(PIF_ERROR);
  }

  return(PIF_OK);
}
/****************************************************************************
** WriteImageData
** ==============
Write out image data in chunks. Note that FloatInt4/2 & ComplexFloatInt4/2 need
to be converted from float4 to int4 or int2.
*/
static
int WriteImageData( FILE *fptr                          // file pointer
                  , const PIF_FILE_HEADER  *pifHeader   // file header info
                  , const PIF_IMAGE_HEADER *imgHeader   // image header info
                  , void *buffer                        // image data
                  )
{
  const ULONG count = FindImageDimensions(imgHeader);
  const ULONG bytes = CalcImageFinalSizeInBytes(imgHeader);
  const int   depth = FindImageFinalDepth(imgHeader->mode);

  if(bytes == 0)            // unknown mode
    return(PIF_ERROR);

  if(bytes < 1048576L)  {   // 1 MByte
    if(ConvertAndWriteImageData(fptr, pifHeader, imgHeader, count, buffer) != PIF_OK)  {
      PIF_DoError("WriteImageData - error writing image data");
      return(PIF_ERROR);
    }
  }
  else  {
    UCHAR *bptr = (UCHAR *) buffer; // pointer to byte-sized data
    ULONG total = 0;
    ULONG tenth = count / 10 + 1;   // for 10% (not less) progress reports
    
    PIF_DoNotify("Writing -");
    while(total < count)  {
      char msg[200];
      sprintf(msg, " %ld%%", total/(count/100));  // count/100 better than total*100 to prevent integer overflow
      PIF_DoNotify(msg);
      tenth = MIN(tenth, count-total);      // make sure we don't over-do it!
      if(ConvertAndWriteImageData(fptr, pifHeader, imgHeader, tenth, bptr) != PIF_OK)  {
        PIF_DoError("WriteImageData - error writing image data");
        return(PIF_ERROR);
      }
      bptr  += tenth * depth;
      total += tenth;
    }
    PIF_DoNotify(" 100%\n");
  }

  return(PIF_OK);
}
/****************************************************************************
** WriteImageDataFloat
** ===================
Write out image data in chunks. Data must be converted from float to the type
specified in the image header.
*/
static
int WriteImageDataFloat( FILE *fptr                          // file pointer
                       , const PIF_FILE_HEADER  *pifHeader   // file header info
                       , const PIF_IMAGE_HEADER *imgHeader   // image header info
                       , float *buffer                       // image data
                       )
{
  const ULONG count = FindImageDimensions(imgHeader);
  const ULONG bytes = CalcImageFinalSizeInBytes(imgHeader);

  if(bytes == 0)            // unknown mode
    return(PIF_ERROR);

  if(bytes < 1048576L)  {   // 1 MByte
    if(ConvertAndWriteImageDataFloat(fptr, pifHeader, imgHeader, count, buffer) != PIF_OK)  {
      PIF_DoError("WriteImageDataFloat - error writing image data");
      return(PIF_ERROR);
    }
  }
  else  {
    float *fdata = buffer;          // pointer to data
    ULONG total = 0;
    ULONG tenth = count / 10 + 1;   // for 10% (not less) progress reports
    
    PIF_DoNotify("Writing -");
    while(total < count)  {
      char msg[200];
      sprintf(msg, " %ld%%", total/(count/100));  // count/100 better than total*100 to prevent integer overflow
      PIF_DoNotify(msg);
      tenth = MIN(tenth, count-total);      // make sure we don't over-do it!
      if(ConvertAndWriteImageDataFloat(fptr, pifHeader, imgHeader, tenth, fdata) != PIF_OK)  {
        PIF_DoError("WriteImageDataFloat - error writing image data");
        return(PIF_ERROR);
      }
      fdata  += tenth;
      total += tenth;
    }
    PIF_DoNotify(" 100%\n");
  }

  return(PIF_OK);
}
/****************************************************************************
** ConvertAndWriteImageData
** ========================
FloatInt4/2 & ComplexFloatInt4/2 need to be converted from float4 to int4 or
int2. We want to leave the image data alone, so we will make copies.
*/
static
int ConvertAndWriteImageData( FILE *fptr
                            , const PIF_FILE_HEADER  *pifHeader
                            , const PIF_IMAGE_HEADER *imgHeader
                            , const ULONG count
                            , void *buffer
                            )
{
  int error = 0;
  const int   depth = FindImageRawDepth(imgHeader->mode);
  const ULONG bytes = count * depth;
  void *temp;
  const float scale = pifHeader->floatScaleFactor; // divide by this

  switch(imgHeader->mode)  {
    case kPIF_MODE_INT1:
    case kPIF_MODE_INT2:
    case kPIF_MODE_CMPLXINT2:
    case kPIF_MODE_SF:
    case kPIF_MODE_BOXED:
    case kPIF_MODE_FLOAT4:
    case kPIF_MODE_CMPLXFLOAT4:
    case kPIF_MODE_DEPTHCUED:
      error = WriteBufferToFile(fptr, buffer, bytes);
      break;
    case kPIF_MODE_FLOATINT4:
    case kPIF_MODE_MAPFLOATINT4:
    case kPIF_MODE_CMPLXFLOATINT4:
      if((temp = malloc(bytes)) == NULL)  {
        char msg[200];
        sprintf(msg, "ConvertAndWriteImageData - error allocating %ld bytes", bytes);
        PIF_DoError(msg);
        return(PIF_ERROR);
      }
      if(imgHeader->mode == kPIF_MODE_CMPLXFLOATINT4)
        ArrayFloat4toInt4(temp, count*2, scale, buffer);    // twice as many floats for COMPLEX
      else
        ArrayFloat4toInt4(temp, count, scale, buffer);
      error = WriteBufferToFile(fptr, temp, bytes);
      free(temp);
      break;
    case kPIF_MODE_FLOATINT2:
    case kPIF_MODE_MAPFLOATINT2:
    case kPIF_MODE_CMPLXFLOATINT2:
      if((temp = malloc(bytes)) == NULL)  {
        char msg[200];
        sprintf(msg, "ConvertAndWriteImageData - error allocating %ld bytes", bytes);
        PIF_DoError(msg);
        return(PIF_ERROR);
      }
      if(imgHeader->mode == kPIF_MODE_CMPLXFLOATINT2)
        ArrayFloat4toInt2(temp, count*2, scale, buffer);    // twice as many floats for COMPLEX
      else
        ArrayFloat4toInt2(temp, count, scale, buffer);
      error = WriteBufferToFile(fptr, temp, bytes);
      free(temp);
      break;
  }
  return(error);
}
/****************************************************************************
** ConvertAndWriteImageDataFloat
** =============================
Convert float data to the type specified in the image header, and write out.
We want to leave the image data alone, so we will make copies.
*/
static
int ConvertAndWriteImageDataFloat( FILE *fptr
                                 , const PIF_FILE_HEADER  *pifHeader
                                 , const PIF_IMAGE_HEADER *imgHeader
                                 , const ULONG count
                                 , float *buffer
                                 )
{
  int error = 0;
  const int   depth = FindImageRawDepth(imgHeader->mode);
  const ULONG bytes = count * depth;
  void *temp;
  const float scale = pifHeader->floatScaleFactor; // divide by this

  switch(imgHeader->mode)  {
    case kPIF_MODE_FLOAT4:
    case kPIF_MODE_CMPLXFLOAT4:
      error = WriteBufferToFile(fptr, buffer, bytes);
      break;
    case kPIF_MODE_INT1:
    case kPIF_MODE_BOXED:
      if((temp = malloc(bytes)) == NULL)  {
        char msg[200];
        sprintf(msg, "ConvertAndWriteImageDataFloat - error allocating %ld bytes", bytes);
        PIF_DoError(msg);
        return(PIF_ERROR);
      }
      ArrayFloat4toInt1(temp, count, 1.0, buffer);      // no scaling
      error = WriteBufferToFile(fptr, temp, bytes);
      free(temp);
      break;
    case kPIF_MODE_INT2:
    case kPIF_MODE_CMPLXINT2:
      if((temp = malloc(bytes)) == NULL)  {
      char msg[200];
        sprintf(msg, "ConvertAndWriteImageDataFloat - error allocating %ld bytes", bytes);
        PIF_DoError(msg);
        return(PIF_ERROR);
      }
      ArrayFloat4toInt2(temp, count, 1.0, buffer);      // no scaling
      error = WriteBufferToFile(fptr, temp, bytes);
      free(temp);
      break;
    case kPIF_MODE_SF:
      PIF_DoError("ConvertAndWriteImageDataFloat - sorry, MODE_SF not yet supported.");
      break;
    case kPIF_MODE_DEPTHCUED:
      PIF_DoError("ConvertAndWriteImageDataFloat - sorry, MODE_DEPTHCUED not yet supported.");
      break;
    case kPIF_MODE_FLOATINT4:
    case kPIF_MODE_MAPFLOATINT4:
    case kPIF_MODE_CMPLXFLOATINT4:
      if((temp = malloc(bytes)) == NULL)  {
        char msg[200];
        sprintf(msg, "ConvertAndWriteImageDataFloat - error allocating %ld bytes", bytes);
        PIF_DoError(msg);
        return(PIF_ERROR);
      }
      if(imgHeader->mode == kPIF_MODE_CMPLXFLOATINT4)
        ArrayFloat4toInt4(temp, count*2, scale, buffer);    // twice as many floats for COMPLEX
      else
        ArrayFloat4toInt4(temp, count, scale, buffer);
      error = WriteBufferToFile(fptr, buffer, bytes);
      free(temp);
      break;
    case kPIF_MODE_FLOATINT2:
    case kPIF_MODE_MAPFLOATINT2:
    case kPIF_MODE_CMPLXFLOATINT2:
      if((temp = malloc(bytes)) == NULL)  {
        char msg[200];
        sprintf(msg, "ConvertAndWriteImageDataFloat - error allocating %ld bytes", bytes);
        PIF_DoError(msg);
        return(PIF_ERROR);
      }
      if(imgHeader->mode == kPIF_MODE_CMPLXFLOATINT2)
        ArrayFloat4toInt2(temp, count*2, scale, buffer);    // twice as many floats for COMPLEX
      else
        ArrayFloat4toInt2(temp, count, scale, buffer);
      error = WriteBufferToFile(fptr, temp, bytes);
      free(temp);
      break;
  }
  return(error);
}
/****************************************************************************
        P R I V A T E   F U N C T I O N S
          1) Miscellaneous
          2) I/O Routines - Read
          3) I/O Routines - Write
        * 4) I/O Routines - Misc
          5) Datatype conversion

4) I/O Routines - Misc
      WriteBufferToFile
*/
/****************************************************************************
** WriteBufferToFile
** =================
*/
static
int WriteBufferToFile( FILE *fptr
                     , void *buffer
                     , ULONG bytes
                     )
{
  if(fwrite(buffer, bytes, 1, fptr) != 1)  {
    PIF_DoError("WriteBufferToFile");
    return(PIF_ERROR);
  }
  return(PIF_OK);
}
/****************************************************************************
        P R I V A T E   F U N C T I O N S
          1) Miscellaneous
          3) I/O Routines - Write
          4) I/O Routines - Misc
        * 5) Datatype conversion

5) Datatype conversion
      Char4toInt4
      Char4toFloatInt4
      Int4toChar4
      FloatInt4toChar4
      Char16toDouble
      ArraySwapBytesInt2
      SwapBytesInt4
      ArrayInt2toFloat4
      ArrayInt1toFloat4_Simple
      ArrayInt2toFloat4_Simple
      ArrayInt4toFloat4
      ArrayFloat4toInt4
      ArrayFloat4toInt2
      ArrayFloat4toInt1
*/
/****************************************************************************
** Char4toInt4
** ===========
Convert 4 (binary) bytes to an int, taking into account the endian mode of the
chars and the machine we are running on.
*/
static
int Char4toInt4( UCHAR const *cptr                      // 4-character array
               , const PIF_FILE_HEADER *pifHeader       // file header info
               )
{
  int value = *((int *) cptr);
  if(EndianModeMismatch(pifHeader))     // do we need to swap bytes?
    value = SwapBytesInt4(value);
  return(value);
}
/****************************************************************************
** Char4toFloatInt4
** ================
Convert 4 (binary) bytes to an int, taking into account the endian mode of the
chars. Then convert to a float by scaling as indicated.
*/
static
float Char4toFloatInt4( UCHAR const *cptr                       // 4-character array
                      , const PIF_FILE_HEADER *pifHeader        // file header info
                      )
{
  int   intValue;
  float floatValue;

  intValue = Char4toInt4(cptr, pifHeader);
  floatValue = ((double) intValue) * pifHeader->floatScaleFactor;

  return(floatValue);
}
/****************************************************************************
** Int4toChar4
** ===========
Convert an int to 4 (binary) bytes, using the endian mode of the machine running
this code.
*/
static
void Int4toChar4( int intVal            // integer
                , register UCHAR *char4 // place for output
                )
{
  register UCHAR *ichar4 = (UCHAR *) (&intVal);
  register int n;

  for(n=0; n<4; n++)
    *char4++ = *ichar4++;
}
/****************************************************************************
** FloatInt4toChar4
** ================
Convert a float to int4 (using the PIF file's global scale factor) and then to
4 (binary) bytes, using the natural endian mode of the machine running this code.
*/
static
void FloatInt4toChar4( float floatValue                 // float
                     , const PIF_FILE_HEADER *pifHeader // file header info
                     , UCHAR *char4                     // place for output
                     )
{
  int intValue = (int) (floatValue / pifHeader->floatScaleFactor);
  Int4toChar4(intValue, char4);
}
/****************************************************************************
** Char16toDouble
** ==============
Convert 16 character string to a double.
*/
static
double Char16toDouble(const char *cptr)
{
  char temp[17];
  double value;

  strncpy(temp, cptr, 16);
  temp[16] = '\0';

  if(sscanf(temp, "%lf", &value) != 1)  {
    fprintf(stderr, "PIF Error - Char16toDouble(), no conversion, 0.0 used\n");
    fprintf(stderr, "            '%s'\n", temp);
    value = 0.0;
  }

  value = atof(temp);
  return(value);
  
}
/****************************************************************************
** ArraySwapBytesInt2
** ==================
Swap the bytes in an array of 2-byte integers.
*/
static
void ArraySwapBytesInt2(void *array, long count)
{
  long n;
  UCHAR byte;
  struct {
    UCHAR b1;
    UCHAR b2;
  } *wptr = array;

  for(n=0; n<count; n++, wptr++)  {
    byte = wptr->b1;
    wptr->b1 = wptr->b2;
    wptr->b2 = byte;
  }
}
/****************************************************************************
** SwapBytesInt4
** =============
Swap the bytes in a 4-byte integer.
*/
static
int SwapBytesInt4(int reverse)
{
  union {
    int  iValue;
    char cValue[4];
  } swap;
  char *cptr = (char *) (&reverse);

  swap.cValue[3] = cptr[0];   // byte-swapping into iValue
  swap.cValue[2] = cptr[1];
  swap.cValue[1] = cptr[2];
  swap.cValue[0] = cptr[3];

  return(swap.iValue);
}
/****************************************************************************
** ArrayInt2toFloat4
** =================
Convert an array of Int2 to Float4 with byte-swapping if necessary. Assume the
array is already big enough for the two-fold expansion, and that the Int2's are
packed at the beginning.
*/
static
void ArrayInt2toFloat4( void *array
                      , const ULONG count
                      , const PIF_FILE_HEADER *pifHeader
                      )
{
  float *fptr = ((float *) array) + count, fValue;
  ULONG  n;
  const float scale = pifHeader->floatScaleFactor;

  if(EndianModeMismatch(pifHeader))  {  // do we need to swap bytes?
    union {
      short  iValue;
      char   cValue[2];
    } swap;
    char *cptr = ((char *) array) + count*2;
    for(n=0; n<count; n++)  {
      swap.cValue[0] = *--cptr;       // byte-swapping into iValue
      swap.cValue[1] = *--cptr;
      fValue = ((float) swap.iValue) * scale;
      *--fptr = fValue;
    }
  }
  else  {
    short   *iptr = ((short *) array) + count;
    for(n=0; n<count; n++)  {
      fValue = ((float) *--iptr) * scale;
      *--fptr = fValue;
    }
  }
}
/****************************************************************************
** ArrayInt1toFloat4_Simple
** ========================
Convert an array of Int1 to Float4. Assume the array is already big enough for
the four-fold expansion, and that the Int1's are packed at the beginning. Also
assume that no scaling is to be done.
*/
static
void ArrayInt1toFloat4_Simple( void *array
                             , register const ULONG count
                             )
{
  register float *fptr = ((float *) array) + count;
  register UCHAR *iptr = ((UCHAR *) array) + count;
  register ULONG  n;

  for(n=0; n<count; n++)
    *--fptr = (float) *--iptr;
}
/****************************************************************************
** ArrayInt2toFloat4_Simple
** ========================
Convert an array of Int2 to Float4. Assume the array is already big enough for
the two-fold expansion, and that the Int2's are packed at the beginning. Also
assume the int2's are properly byte-ordered and that no scaling is to be done.
*/
static
void ArrayInt2toFloat4_Simple( void *array
                             , register const ULONG count
                             )
{
  register float *fptr = ((float *) array) + count;
  register short *iptr = ((short *) array) + count;
  register ULONG  n;

  for(n=0; n<count; n++)
    *--fptr = (float) *--iptr;
}
/****************************************************************************
** ArrayInt4toFloat4
** =================
Convert an array of Int4 to Float4 with byte-swapping if necessary.
*/
static
void ArrayInt4toFloat4( void *array
                      , const ULONG count
                      , const PIF_FILE_HEADER *pifHeader
                      )
{
  float *fptr = array, fValue;
  ULONG  n;
  const float scale = pifHeader->floatScaleFactor;

  if(EndianModeMismatch(pifHeader))  {  // do we need to swap bytes?
    union {
      int  iValue;
      char cValue[4];
    } swap;
    char *cptr = array;
    for(n=0; n<count; n++)  {
      swap.cValue[3] = *cptr++;       // byte-swapping into iValue
      swap.cValue[2] = *cptr++;
      swap.cValue[1] = *cptr++;
      swap.cValue[0] = *cptr++;
      fValue = ((float) swap.iValue) * scale;
      *fptr++ = fValue;
    }
  }
  else  {
    int   *iptr = array;
    for(n=0; n<count; n++)  {
      fValue = ((float) *iptr++) * scale;
      *fptr++ = fValue;
    }
  }
}
/****************************************************************************
** ArrayFloat4toInt4
** =================
Clip to max and min values of (signed) int4.
*/
static
void ArrayFloat4toInt4( register int *dest
                      , ULONG count
                      , float scale
                      , register const float *src
                      )
{
  ULONG n;
  const float invScale = 1.0 / scale;
  const float kINT4MAX = (1UL<<31) - 1;
  const float kINT4MIN = -kINT4MAX - 1;
  float fValue;

  for(n=0; n<count; n++)  {
    fValue = (*src++ * invScale);
    if      (fValue>kINT4MAX)  fValue = kINT4MAX;
    else if (fValue<kINT4MIN)  fValue = kINT4MIN;
    *dest++ = (int) fValue;
  }
}
/****************************************************************************
** ArrayFloat4toInt2
** =================
Clip to max and min values of (signed) int2.
*/
static
void ArrayFloat4toInt2( register short *dest
                      , ULONG count
                      , float scale
                      , register const float *src
                      )
{
  ULONG n;
  const float invScale = 1.0 / scale;
  const float kINT2MAX = (1UL<<15) - 1;
  const float kINT2MIN = -kINT2MAX - 1;
  float fValue;

  for(n=0; n<count; n++)  {
    fValue = (*src++ * invScale);
    if      (fValue>kINT2MAX)  fValue = kINT2MAX;
    else if (fValue<kINT2MIN)  fValue = kINT2MIN;
    *dest++ = (short) fValue;
  }
}
/****************************************************************************
** ArrayFloat4toInt1
** =================
Clip to max and min values of (unsigned) int1.
*/
static
void ArrayFloat4toInt1( register UCHAR *dest
                      , ULONG count
                      , float scale
                      , register const float *src
                      )
{
  ULONG n;
  const float invScale = 1.0 / scale;
  const float kINT1MAX = 255.0;
  const float kINT1MIN = 0.0;
  float fValue;

  for(n=0; n<count; n++)  {
    fValue = (*src++ * invScale);
    if      (fValue>kINT1MAX)  fValue = kINT1MAX;
    else if (fValue<kINT1MIN)  fValue = kINT1MIN;
    *dest++ = (short) fValue;
  }
}
/****************************************************************************
        P R I V A T E   F U N C T I O N S   (End)
****************************************************************************/
