/****************************************************************************
** PIF.h
** =====

This version: 03-Mar-2000 (v1.4.8)
James Conway, LMES/IBS, Grenoble, France (email: James.Conway@ibs.fr)
Modified from source code provided by Rob Ashmore, Purdue University, Indiana.

****************************************************************************/

#ifndef __PIF_IO__      // Make sure we only include this file once
#define __PIF_IO__

#include <stdio.h>      // for FILE, used in public function prototypes

// Public Structure Definitions
// ============================

// The PIF_FILE_HEADER structure for general file info

typedef struct {
  unsigned char fileId[8];      // kind of data format, always 8,0,0,0,8,0,0,0
  double floatScaleFactor;      // scale factor for floatInt*4 into real*4
  int  numImages;               // number of images in this file
  int  endianNess;              // endian type: 0 = little (VMS), else = big (Unix)
  char genProgram[33];          // name of program used to generate this file
  int  hType;                   // 1 - all projections have the same size & depth
                                // 0 - otherwise
  int  nx;                      // number of columns
  int  ny;                      // number of rows
  int  nz;                      // number of sections
  int  mode;                    // EM data type e.g.
                                //    0 = byte*1
                                //    1 = int*2
                                //    2 = floatInt*4
                                //    3 = complex int*2
                                //    4 = complex floatInt*4
                                //    5 = Structure Factors
                                //    6 = Boxed data
                                //    7 = etc...
                                //   97 = DEPTHCUED
  short *colorMap;              // Color map, only included with DepthCued (*256*3 = 1536 bytes)
} PIF_FILE_HEADER;

// The PIF_IMAGE_HEADER structure for image-specific info

typedef struct {
  int   nx;                     // Number of columns
  int   ny;                     // Number of rows
  int   nz;                     // Number of sections
  int   mode;                   // EM data type e.g.
                                //    0 = byte*1
                                //    1 = int*2
                                //    2 = floatInt*4
                                //    3 = complex int*2
                                //    4 = complex floatInt*4
                                //    5 = Structure Factors
                                //    6 = Boxed data
                                //    7 = etc...
                                //   97 = DEPTHCUED
  int   bkgnd;                  // Background value
  int   packRadius;             // Radius for boxed images
  int   nxstart;                // Number of first col in map
  int   nystart;                // Number of first row in map
  int   nzstart;                // Number of first section in map
  int   mx;                     // Number of intervals along x
  int   my;                     // Number of intervals along y
  int   mz;                     // Number of intervals along z
  float xlength;                // Cell Dimensions (Angstroms)
  float ylength;                //              "
  float zlength;                //              "
  float alpha;                  // Cell Angles (degrees)
  float beta;                   //              "
  float gamma;                  //              "
  int   mapc;                   // Which axis is col(1,2,3 = x,y,z)
  int   mapr;                   // Which axis is row(1,2,3 = x,y,z)
  int   maps;                   // Which axis is sections(1,2,3 = x,y,z)
  float min;                    // Min density value
  float max;                    // Max density value
  float mean;                   // Mean density value
  float stdDev;                 // StdDev of density value
  int   ispg;                   // Space group number
  int   nsymbt;                 // Number of bytes for symmetry ops
  float xorigin;                // x origin
  float yorigin;                // y origin
  char  title[81];              // User defined description
  char  timeStamp[33];          // Date/time data last modified
  char  uGraphDesignation[17];  // Unique Micrograph Number
  char  scanNumber[9];          // Scan Number of Micrograph
} PIF_IMAGE_HEADER;

// Public Function Prototypes
// ==========================

PIF_FILE_HEADER   PIF_NewPifFileHeader  ( void );
PIF_IMAGE_HEADER  PIF_NewPifImageHeader ( void );

FILE *PIF_OpenForWrite( char *fname                      // name of file to open
                      , const PIF_FILE_HEADER *pifHeader // header info
                      );

int PIF_WriteImageHeader( FILE *fptr                          // open PIF file pointer
                        , PIF_FILE_HEADER  const * pifHeader   // file header info
                        , PIF_IMAGE_HEADER const * imgHeader   // image header info
                        );

int PIF_WriteImageData( FILE *fptr                          // open PIF file pointer
                      , const PIF_FILE_HEADER  *pifHeader   // file header info
                      , const PIF_IMAGE_HEADER *imgHeader   // image header info
                      , void *buffer                        // image data
                      );

int PIF_WriteImage( FILE *fptr                           // open PIF file pointer
                  , const PIF_FILE_HEADER *pifHeader     // file header info
                  , const PIF_IMAGE_HEADER *imgHeader    // image header info
                  , void *buffer                         // image data
                  );

int PIF_WriteImageFloat( FILE *fptr                      // open PIF file pointer
                  , const PIF_FILE_HEADER  *pifHeader    // file header info
                  , const PIF_IMAGE_HEADER *imgHeader    // image header info
                  , float *buffer                        // image data
                  );

const char *PIF_ModeString(int mode);

int PIF_IAmBigEndian(void);


// Public Constants
// ================

#define PIF_ERROR (1)
#define PIF_OK    (0)

// Values for the PIF header 'mode' fields (file and image)

#define kPIF_MODE_INT1              ( 0)    // byte*1 (unsigned)
#define kPIF_MODE_INT2              ( 1)    // int*2  (signed)
#define kPIF_MODE_FLOATINT4         ( 2)    // floatInt*4
#define kPIF_MODE_CMPLXINT2         ( 3)    // complex int*2
#define kPIF_MODE_CMPLXFLOATINT4    ( 4)    // complex floatInt*4
#define kPIF_MODE_SF                ( 5)    // Structure Factors
#define kPIF_MODE_BOXED             ( 6)    // Boxed data (unsigned byte)
#define kPIF_MODE_FLOATINT2         ( 7)    // floatInt*2
#define kPIF_MODE_CMPLXFLOATINT2    ( 8)    // complex floatInt*2
#define kPIF_MODE_FLOAT4            ( 9)    // float*4
#define kPIF_MODE_CMPLXFLOAT4       (10)    // complex float*4
#define kPIF_MODE_MAPFLOATINT2      (20)    // same as kPIF_MODE_FLOATINT2
#define kPIF_MODE_MAPFLOATINT4      (21)    // same as kPIF_MODE_FLOATINT4
#define kPIF_MODE_DEPTHCUED         (97)    // DEPTHCUED

// Finish up
// =========

#endif

/***************************************************************************/
