C*****************************************************************************
C
C    CTFFIND3 - determines defocus and astigmatism for images of
C    arbitrary size (MRC format). Astigmatic angle is measured form
C    x axis (same convetiones as in the MRC 2D image processing 
C    programs).
C
C    CARD 1: Input file name for image
C    CARD 2: Output file name to check result
C    CARD 3: CS[mm], HT[kV], AmpCnst, XMAG, DStep[um]
C    CARD 4: Box, ResMin[A], ResMax[A], dFMin[A], dFMax[A], FStep
C
C        The output image file to check the result of the fitting
C        shows the filtered average power spectrum of the input
C        image in one half, and the fitted CTF (squared) in the
C        other half. The two halfs should agree very well for a
C        sucessfull fit.
C
C        CS: Spherical aberration coefficient of the objective in mm
C        HT: Electron beam voltage in kV
C        AmpCnst: Amount of amplitude contrast (fraction). For ice
C                 images 0.07, for negative stain about 0.15.
C        XMAG: Magnification of original image
C        DStep: Pixel size on scanner in microns
C        Box: Tile size. The program devides the image into square
C             tiles and calculates the average power spectrum. Tiles
C             with a significatly higher or lower variance are 
C             excluded; these are parts of the image which are unlikely
C             to contain useful information (beam edge, film number 
C             etc). IMPORTANT: Box must have even pixel dimensions.
C        ResMin: Low resolution end of data to be fitted.
C        ResMaX: High resolution end of data to be fitted.
C        dFMin: Starting defocus value for grid search in Angstrom. 
C               Positive values represent an underfocus. The program
C               performs a systematic grid search of defocus values 
C               and astigmatism before fitting a CTF to machine 
C               precision.
C        dFMax: End defocus value for grid search in Angstrom.
C        FStep: Step width for grid search in Angstrom.
C
C*****************************************************************************
C    example command file (UNIX):
C
C    #!/bin/csh -x
C    #
C    #   ctffind3
C    #
C    time /public/image/bin/ctffind3.exe << eof
C    image.mrc
C    power.mrc
C    2.6,200.0,0.07,60000.0,28.0
C    128,100.0,15.0,30000.0,90000.0,5000.0
C    eof
C    #
C*****************************************************************************
C
      PROGRAM CTFFIND3
C
C    NIKO, 21 SEPTEMBER 2002
C
      IMPLICIT NONE
C
      INTEGER NXYZ(3),MODE,JXYZ(3),I,J,NX,NY,IX,IS,KXYZ(3)
      INTEGER K,CNT,ID,L,M,LL,MM,ITEST,IP,NBIN,IMP,IERR
      PARAMETER (NBIN=100)
      REAL DMAX,DMEAN,DRMS,RMS,WGH1,WGH2,RMSMIN
      REAL SCAL,MEAN,WL,PI,MIN,MAX,ANGAST
      REAL CS,KV,WGH,XMAG,DSTEP,RESMIN,RESMAX,DFMID1,DFMID2
      REAL THETATR,STEPR,RMIN2,RMAX2,HW,RMSMAX
      REAL RES2,CTF,CTFV,TMP,FLT,DFMIN,DFMAX,FSTEP,DRMS1,CMAX
      REAL,ALLOCATABLE :: AIN(:),ABOX(:),POWER(:),OUT(:),BUF1(:)
      REAL,ALLOCATABLE :: RMSA(:),BINS(:)
      PARAMETER (FLT=-0.1,PI=3.1415926535898)
      COMPLEX,ALLOCATABLE :: CBOXS(:)
      CHARACTER FILEIN*200,FILEOUT*200,TITLE*1600,CFORM,NCPUS*10
      LOGICAL EX
C
      WRITE(6,1000)
1000  FORMAT(/' CTF DETERMINATION, V3.4 (30-September-2010)',
     +       /' Distributed under the GNU',
     +        ' General Public License (GPL)')
C
      IMP=0
      CALL GETENV('OMP_NUM_THREADS',NCPUS)
      READ(NCPUS,*,ERR=111,END=111)IMP
111   CONTINUE
      IF (IMP.LE.0) THEN
        CALL GETENV('NCPUS',NCPUS)
        READ(NCPUS,*,ERR=112,END=112)IMP
112     CONTINUE
      ENDIF
      IF (IMP.LE.0) IMP=1
      IF (IMP.GT.1) THEN
        WRITE(*,7000) IMP
7000    FORMAT(/' Parallel processing: NCPUS =   ',I8)
      ENDIF
C
      WRITE(6,1002)
1002  FORMAT(/' Input image file name')
      READ(5,1010)FILEIN
1010  FORMAT(A)
      WRITE(6,1010)FILEIN
      WRITE(6,1060)
1060  FORMAT(/' Output diagnostic file name')
      READ(5,1010)FILEOUT
      WRITE(6,1010)FILEOUT
      WRITE(6,1020)
1020  FORMAT(/' CS[mm], HT[kV], AmpCnst, XMAG, DStep[um]')
      READ(5,*)CS,KV,WGH,XMAG,DSTEP
      WRITE(6,1031)CS,KV,WGH,XMAG,DSTEP
1031  FORMAT(F5.1,F9.1,F8.2,F10.1,F9.3)
      WRITE(6,1040)
1040  FORMAT(/' Positive defocus values for underfocus',
     +         /' Box, ResMin[A], ResMax[A], dFMin[A], dFMax[A], FStep')
      READ(5,*)JXYZ(1),RESMIN,RESMAX,DFMIN,DFMAX,FSTEP
      WRITE(6,1051)JXYZ(1),RESMIN,RESMAX,DFMIN,DFMAX,FSTEP
1051  FORMAT(I4,2F11.1,2F10.1,F7.1/)
      ITEST=JXYZ(1)/2
      IF (2*ITEST.NE.JXYZ(1)) THEN
          WRITE(6,1090)
1090    FORMAT(/' Box size must be even number')
          STOP
      ENDIF

      if(jxyz(1) > 2048) then
         write(*,*) 'Box size must be <= 2048'
         stop
      endif

      JXYZ(2)=JXYZ(1)
      JXYZ(3)=1
      IF (RESMIN.LT.RESMAX) THEN
          TMP=RESMAX
          RESMAX=RESMIN
          RESMIN=TMP
      ENDIF
      IF (DFMAX.LT.DFMIN) THEN
          TMP=DFMAX
          DFMAX=DFMIN
          DFMIN=TMP
      ENDIF
C
      CALL GUESSF(FILEIN,CFORM,EX)
      IF  (.NOT.EX) THEN
        WRITE(*,1001) FILEIN
1001    FORMAT(' File not found ',A80)
        STOP
      ENDIF
      CALL IOPEN(FILEIN,10,CFORM,MODE,NXYZ(1),NXYZ(2),
     +           NXYZ(3),'OLD',STEPR,TITLE)
C
      WRITE(*,1110)NXYZ(1),NXYZ(2)
1110  FORMAT(/,' READING IMAGE...'/' NX, NY= ',2I10)
C
      NX=NXYZ(1)/JXYZ(1)
      NY=NXYZ(2)/JXYZ(2)
C
      ALLOCATE(AIN(NXYZ(1)*JXYZ(2)),ABOX(JXYZ(1)*JXYZ(2)),
     +  OUT(JXYZ(1)*JXYZ(2)),
     +  BUF1(JXYZ(1)*JXYZ(2)),RMSA(NX*NY),BINS(NBIN),
     +  CBOXS(JXYZ(2)),STAT=IERR)
      IF (IERR.NE.0) THEN
        WRITE(*,*) ' ERROR: Memory allocation failed in MAIN'
        STOP ' Try reducing the tile size'
      ENDIF
C
      DRMS=0.0D0
      CNT=0
      DO 10 I=1,NY
        IP=(I-1)*JXYZ(2)
        DO 21 J=1,JXYZ(2)
          ID=1+NXYZ(1)*(J-1)
          CALL IREAD(10,AIN(ID),J+IP)
21      CONTINUE
        DO 10 J=1,NX
          IX=(J-1)*JXYZ(1)+1
          CALL BOXIMG(AIN,NXYZ,ABOX,JXYZ,IX,1,MEAN,RMS)
          DRMS=DRMS+RMS**2
            CNT=CNT+1
          IF (CNT.LE.50000) RMSA(CNT)=RMS
10    CONTINUE
      DRMS=SQRT(DRMS/CNT)
      CALL HISTO(CNT,NBIN,RMSA,BINS,MIN,MAX)
      RMSMIN = MIN
      RMSMAX = MAX
      CMAX=0
      DO 70 I=1,NBIN
        IF (BINS(I).GT.CMAX) THEN
          CMAX=BINS(I)
          J=I
        ENDIF
70    CONTINUE
      IF (J.GT.1) THEN
        DO 71 I=J-1,1,-1
          IF (BINS(I).LT.CMAX/10.0) THEN
            RMSMIN=(I-1)*(MAX-MIN)/(NBIN-1)+MIN
            GOTO 72
          ENDIF
71      CONTINUE
      ENDIF
72    CONTINUE
      IF (J.LT.NBIN) THEN
        DO 73 I=J+1,NBIN
          IF (BINS(I).LT.CMAX/10.0) THEN
            RMSMAX=(I-1)*(MAX-MIN)/(NBIN-1)+MIN
            GOTO 74
          ENDIF
73      CONTINUE
      ENDIF
74    CONTINUE
C
      KXYZ(1)=JXYZ(1)/2
      KXYZ(2)=JXYZ(2)
      KXYZ(3)=JXYZ(3)
      IF (2*(KXYZ(1)/2).NE.KXYZ(1)) KXYZ(1)=KXYZ(1)+1
      ALLOCATE(POWER(KXYZ(1)*KXYZ(2)),STAT=IERR)
      IF (IERR.NE.0) THEN
        WRITE(*,*) ' ERROR: Memory allocation failed for POWER'
        STOP ' Try reducing the tile size'
      ENDIF
      DO 30 K=1,KXYZ(1)*KXYZ(2)
          POWER(K)=0.0
30    CONTINUE
      SCAL=1.0/SQRT(REAL(JXYZ(1)*JXYZ(2)))
      CNT=0
      WRITE(*,1100)
1100  FORMAT(/,' TILING IMAGE...'/)
C
      DO 100 I=1,NY
        IP=(I-1)*JXYZ(2)
        DO 20 J=1,JXYZ(2)
            ID=1+NXYZ(1)*(J-1)
          CALL IREAD(10,AIN(ID),J+IP)
20    CONTINUE
          DO 100 J=1,NX
            IX=(J-1)*JXYZ(1)+1
            CALL BOXIMG(AIN,NXYZ,ABOX,JXYZ,IX,1,MEAN,RMS)
            IF ((RMS.LT.RMSMAX).AND.(RMS.GT.RMSMIN)) THEN
              CALL RLFT3(ABOX,CBOXS,JXYZ(1),JXYZ(2),1,1)
            DO 40 L=1,JXYZ(2)
                DO 41 K=1,JXYZ(1)/2
                ID=(K+JXYZ(1)/2*(L-1))*2
                IS=K+KXYZ(1)*(L-1)
                POWER(IS)=POWER(IS)
     +            +(ABOX(ID-1)**2+ABOX(ID)**2)*SCAL**2
41          CONTINUE
              IF (KXYZ(1).GT.JXYZ(1)/2)
     +          POWER(IS+1)=POWER(IS+1)+CABS(CBOXS(L)*SCAL)**2
40        CONTINUE
              CNT=CNT+1
            ENDIF
100   CONTINUE
      WRITE(*,*)' Total tiles and number used',NX*NY,CNT
C
      SCAL=1.0/CNT
      DO 60 K=1,KXYZ(1)*KXYZ(2)
          POWER(K)=SQRT(POWER(K)*SCAL)
60    CONTINUE
C
C       Filter power spectrum to remove slowly varying background
C       DMAX is maximum of filtered power spectrum (for later scaling)
C
      STEPR=DSTEP*(10.0**4.0)/XMAG
      CALL FILTER(JXYZ,KXYZ,POWER,BUF1,DMEAN,DRMS1,DMAX,STEPR)
C
      CS=CS*(10.0**7.0)                          ! Angstroms
      KV=KV*1000.0                               ! Volts
      WL=12.26/SQRT(KV+0.9785*KV**2/(10.0**6.0)) ! Angstroms
      WGH1=SQRT(1.0-WGH**2)
      WGH2=WGH
C
      THETATR=WL/(STEPR*JXYZ(1))
      RESMIN=STEPR/RESMIN
      RESMAX=STEPR/RESMAX
      IF (RESMIN.LT.STEPR/50.0) THEN
        RESMIN=STEPR/50.0
        WRITE(*,*)' Lower resolution limit reset to',
     +            STEPR/RESMIN,' A'
      ENDIF
      IF (RESMIN.GE.RESMAX)
     +  STOP ' RESMIN >= RESMAX; increase RESMAX'
C
      DFMID1=DFMIN
      DFMID2=DFMAX
      CALL SEARCH_CTF(CS,WL,WGH1,WGH2,THETATR,RESMIN,RESMAX,
     +              POWER,JXYZ,DFMID1,DFMID2,ANGAST,FSTEP)
      RMIN2=RESMIN**2
      RMAX2=RESMAX**2
      HW=-1.0/RMAX2
      hw=0.0

      CALL REFINE_CTF(DFMID1,DFMID2,ANGAST,POWER,
     +       CS,WL,WGH1,WGH2,THETATR,RMIN2,RMAX2,JXYZ,HW)
C
      DO 50 I=1,JXYZ(1)*JXYZ(2)
          OUT(I)=0.0
50    CONTINUE
      DO 200 L=1,JXYZ(1)/2
        LL=L-1
        DO 200 M=1,JXYZ(2)
          MM=M-1
          IF (MM.GT.JXYZ(2)/2) MM=MM-JXYZ(2)
          ID=L+JXYZ(1)/2*(M-1)
            I=L+JXYZ(1)/2
            J=M+JXYZ(2)/2
            IF (J.GT.JXYZ(2)) J=J-JXYZ(2)
            IS=I+JXYZ(1)*(J-1)
            OUT(IS)=POWER(ID)/DRMS1*SQRT(2.0*PI)
            IF (OUT(IS).GT.1.0) OUT(IS)=1.0
            IF (OUT(IS).lT.-1.0) OUT(IS)=-1.0
          RES2=(REAL(LL)/JXYZ(1))**2+(REAL(MM)/JXYZ(2))**2
          IF ((RES2.LE.RMAX2).AND.(RES2.GE.RMIN2)) THEN
            CTFV=CTF(CS,WL,WGH1,WGH2,DFMID1,DFMID2,
     +               ANGAST,THETATR,LL,MM)
             I=JXYZ(1)/2-L+1
              J=JXYZ(2)-J+2
              IF (J.LE.JXYZ(2)) THEN
                IS=I+JXYZ(1)*(J-1)
              OUT(IS)=CTFV**2
              ENDIF
          ENDIF
200   CONTINUE
C
      CALL ICLOSE(10)
      MODE=2
      IF (STEPR.NE.0.0) STEPR=1.0/STEPR
      CALL IOPEN(FILEOUT,10,CFORM,MODE,JXYZ(1),JXYZ(2),
     +           JXYZ(3),'NEW',STEPR,TITLE)
      DO 210 J=1,JXYZ(2)
        ID=1+JXYZ(1)*(J-1)
        CALL IWRITE(10,OUT(ID),J)
210   CONTINUE
      CALL ICLOSE(10)
C
      END
C
C**************************************************************************
      SUBROUTINE BOXIMG(AIN,NXYZ,ABOX,JXYZ,IX,IY,MEAN,RMS)
C**************************************************************************
C    Cuts out an area of size JXYZ from array AIN. IX, IY are
C    upper left corner of boxed area.
C**************************************************************************
C
      IMPLICIT NONE
C
      INTEGER I,J,NXYZ(3),JXYZ(3),IX,IY,ID,IDB,II,JJ
      REAL AIN(*),ABOX(*),MEAN,RMS,M1,M2,M3,M4
      REAL*8 DRMS,DMEAN,DM1,DM2,DM3,DM4,DTMP
C
      DMEAN=0.0
      DM1=0.0
      DM2=0.0
      DM3=0.0
      DM4=0.0
      DO 10 J=1,JXYZ(2)
        DO 10 I=1,JXYZ(1)
            II=I+IX-1
            JJ=J+IY-1
          ID=II+NXYZ(1)*(JJ-1)
          IDB=I+JXYZ(1)*(J-1)
          ABOX(IDB)=AIN(ID)
            DMEAN=DMEAN+ABOX(IDB)
            IF (I.EQ.1) DM1=DM1+ABOX(IDB)
            IF (I.EQ.JXYZ(1)) DM2=DM2+ABOX(IDB)
            IF (J.EQ.1) DM3=DM3+ABOX(IDB)
            IF (J.EQ.JXYZ(2)) DM4=DM4+ABOX(IDB)
10    CONTINUE
      DMEAN=DMEAN/JXYZ(1)/JXYZ(2)
      MEAN=DMEAN
      DM1=DM1/JXYZ(2)
      DM2=DM2/JXYZ(2)
      DM3=DM3/JXYZ(1)
      DM4=DM4/JXYZ(1)
      DRMS=0.0
      DO 20 I=1,JXYZ(1)*JXYZ(2)
          DRMS=DRMS+(ABOX(I)-DMEAN)**2
20    CONTINUE
      DRMS=SQRT(DRMS/JXYZ(1)/JXYZ(2))
      RMS=DRMS
      DO 30 J=1,JXYZ(2)
        DO 30 I=1,JXYZ(1)
          IDB=I+JXYZ(1)*(J-1)
          ABOX(IDB)=ABOX(IDB)-(DM1+(DM2-DM1)/(JXYZ(1)-1)*(I-1))
     +                 -(DM3+(DM4-DM3)/(JXYZ(2)-1)*(J-1))+DMEAN
30    CONTINUE
C
      RETURN
      END
C
C**************************************************************************
      SUBROUTINE SEARCH_CTF_S(CS,WL,WGH1,WGH2,THETATR,RMIN2,
     +          RMAX2,AIN,NXYZ,DF1,DF2,ANG,FSTEP,SUMS,HW,
     +            SUMMAX,DFMID1S,DFMID2S,ANGASTS,I,J,K,I1,I2)
C**************************************************************************
C
      IMPLICIT NONE
C
      INTEGER I,J,K,NXYZ(3),I1,I2,ID
      REAL CS,WL,WGH1,WGH2,THETATR,DF1(*),DF2(*),ANG(*)
      REAL RMIN2,RMAX2,SUMS(*),AIN(*),SUMMAX,FSTEP
      REAL DFMID1S,DFMID2S,ANGASTS,HW,PI,EVALCTF
      PARAMETER (PI=3.1415926535898)
C
      ID=I-I1+1+(I2-I1+1)*(J-I1)
      DF1(ID)=FSTEP*I
      DF2(ID)=FSTEP*J
      ANG(ID)=22.5*K
      ANG(ID)=ANG(ID)/180.0*PI
      SUMS(ID)=EVALCTF(CS,WL,WGH1,WGH2,DF1(ID),DF2(ID),
     +    ANG(ID),THETATR,HW,AIN,NXYZ,RMIN2,RMAX2)
C
      RETURN
      END      
C
C**************************************************************************
      REAL FUNCTION EVALCTF(CS,WL,WGH1,WGH2,DFMID1,DFMID2,ANGAST,
     +                THETATR,HW,AIN,NXYZ,RMIN2,RMAX2)
C**************************************************************************
C

C     Under certain conditions (typically smaller micrographs, larger
C     number of search steps, larger range of spatial frequencies used
C     in fit) this function can accounts for the majority of the
C     ctffind3 run time. Extensive optimizations were applied, including
C     the inlining of function ctf, reordering of loops, precalculation
C     of loop invariants, and splitting of inner loop to aid
C     vectorization of calls to intrinsic subroutines.

      IMPLICIT NONE
C
      INTEGER L,LL,M,MM,NXYZ(3),ID, LMIN, LMAX
      REAL CS,WL,WGH1,WGH2,DFMID1,DFMID2,ANGAST,THETATR,SUM1
      REAL SUM,AIN(*),RES2,RMIN2,RMAX2,CTF,CTFV,HW,SUM2
      real rad2, hangle2, c1, c2, df, angdif
      real expv, twopi_wli, ctfv2, dsum, ddif, rpart1, rpart2
      real half_thetatrsq, recip_nxyz1, recip_nxyz2, tmp
      real :: twopi=6.2831853071796

      real angspt(0:1023), ccos(0:1023)
      real chi(0:1023), sinchi(0:1023), coschi(0:1023)

      twopi_wli      = twopi/wl      
      dsum           = dfmid1 + dfmid2
      ddif           = dfmid1 - dfmid2
      half_thetatrsq = 0.5*thetatr*thetatr
      recip_nxyz1    = 1.0/nxyz(1)
      recip_nxyz2    = 1.0/nxyz(2)

      SUM  = 0.0
      SUM1 = 0.0
      SUM2 = 0.0

      DO M=1,NXYZ(2)
         MM=M-1
         IF (MM > NXYZ(2)/2) MM=MM-NXYZ(2)
         rpart2 = real(mm) * recip_nxyz2
         
C     Determine the bounds on the inner loop
         tmp  = (rmax2 - rpart2*rpart2)/(recip_nxyz1*recip_nxyz1)
         if(tmp < 0.0) then
            lmax = 0
         else
            tmp  = sqrt(tmp)
            lmax = floor(tmp)
            lmax = lmax + 1
            lmax = min(lmax, nxyz(1)/2)
         endif

         tmp  = (rmin2 - rpart2*rpart2)/(recip_nxyz1*recip_nxyz1)
         tmp = max(tmp, 0.0)
         tmp  = sqrt(tmp)
         lmin = ceiling(tmp)
         lmin = lmin + 1
         if(m == 1) then
            lmin = max(lmin, 2) ! To avoid L=M=1
         else
            lmin = max(lmin, 1)
         endif

C     Isolate loop over ATAN2 to aid vectorization
         DO LL=LMIN-1, LMAX-1
            ANGSPT(LL) = ATAN2(REAL(MM), REAL(LL))
         enddo

C     Isolate loop over COS to aid vectorization
         DO LL=LMIN-1, LMAX-1
            CCOS(LL) = cos(2.0*(angspt(ll)-angast))
         enddo

C     Calculate array of chi values
         DO LL=LMIN-1, LMAX-1
            RAD2    = LL*LL + MM*MM
            DF      = 0.5*(DSUM + CCOS(LL)*DDIF)
            HANGLE2 = rad2 * half_thetatrsq
            C1      = twopi_wli*HANGLE2
            C2      = -C1*CS*HANGLE2
            CHI(LL) = C1*DF + C2
         ENDDO

C     Isolate loop over SIN and COS to aid vectorization
         DO LL=LMIN-1, LMAX-1
            sinchi(ll) = sin(chi(ll))
            coschi(ll) = cos(chi(ll))
         enddo

C     Finish calcualtions and accumulate results
         DO L=LMIN, LMAX
            LL    = L-1
            CTFV  = -WGH1*SINCHI(LL)-WGH2*COSCHI(LL)
            ctfv2 = ctfv*ctfv
            ID    = L+NXYZ(1)/2*(M-1)
            
            if(hw == 0.0) then
               SUM  = SUM  + AIN(ID)*ctfv2
               SUM1 = SUM1 + ctfv2*ctfv2
               SUM2 = SUM2 + AIN(ID)*AIN(ID)
            else
               rpart1 = real(ll) * recip_nxyz1
               RES2   = rpart1*rpart1 + rpart2*rpart2
               expv   = exp(hw*res2)
               SUM    = SUM  + AIN(ID)*ctfv2*expv
               SUM1   = SUM1 + ctfv2*ctfv2
               SUM2   = SUM2 + AIN(ID)*AIN(ID)*expv*expv
            endif
         enddo

      enddo

      SUM=SUM/SQRT(SUM1*SUM2)
      EVALCTF=SUM

      RETURN
      END
C     
C**************************************************************************
      SUBROUTINE REFINE_CTF(DFMID1,DFMID2,ANGAST,POWER,
     +       CS,WL,WGH1,WGH2,THETATR,RMIN2,RMAX2,NXYZ,HW)
C**************************************************************************
C
      IMPLICIT NONE
C
      INTEGER NCYCLS,NXYZ(3)
      PARAMETER (NCYCLS=50)
      REAL DFMID1,DFMID2,ANGAST,XPAR(3),EPAR(3),RF,ESCALE,PI
      REAL POWER(*),CS,WL,WGH1,WGH2,THETATR,RMIN2,RMAX2,HW
      PARAMETER (PI=3.1415926535898)
      DATA EPAR/100.0,100.0,0.05/
      DATA ESCALE/100.0/
C
      WRITE(*,1000)
1000  FORMAT(/,' REFINING CTF PARAMETERS...'/,
     +       /,'      DFMID1      DFMID2      ANGAST          CC'/)
C
      XPAR(1)=DFMID1
      XPAR(2)=DFMID2
      XPAR(3)=ANGAST
      IF (XPAR(1).EQ.XPAR(2)) XPAR(1)=XPAR(1)+1.0
      CALL VA04A(XPAR,EPAR,3,RF,ESCALE,0,1,NCYCLS,POWER,
     +       CS,WL,WGH1,WGH2,THETATR,RMIN2,RMAX2,NXYZ,HW)
      DFMID1=XPAR(1)
      DFMID2=XPAR(2)
      ANGAST=XPAR(3)
      WRITE(*,1100)DFMID1,DFMID2,ANGAST/PI*180.0,-RF
1100  FORMAT(3F12.2,F12.5,'  Final Values')
C
      RETURN
      END
C
C******************************************************************************
C
      SUBROUTINE CALCFX(NX,XPAR,RF,AIN,
     +  CS,WL,WGH1,WGH2,THETATR,RMIN2,RMAX2,NXYZ,HW)
C
C     CALCULATES NEW VALUE FOR F TO INPUT TO SUBROUTINE VA04A
C
C******************************************************************************
C
      IMPLICIT NONE
C
      INTEGER NXYZ(3),NX,IXBMX
      PARAMETER (IXBMX=512)
      REAL CS,WL,WGH1,WGH2,THETATR,DFMID1,DFMID2,ANGAST
      REAL RMIN2,RMAX2,AIN(*),EVALCTF
      REAL DFMID1S,DFMID2S,ANGASTS,HW,PI,XPAR(3),RF
      PARAMETER (PI=3.1415926535898)
C
      RF=-EVALCTF(CS,WL,WGH1,WGH2,XPAR(1),XPAR(2),XPAR(3),
     +          THETATR,HW,AIN,NXYZ,RMIN2,RMAX2)
C
      RETURN
      END
C
C**************************************************************************
      SUBROUTINE VA04A(X,E,N,F,ESCALE,IPRINT,ICON,MAXIT,AIN,
     +          CS,WL,WGH1,WGH2,THETATR,RMIN2,RMAX2,NXYZ,HW)
C**************************************************************************
C  STANDARD FORTRAN 66 (A VERIFIED PFORT SUBROUTINE)
      DIMENSION W(40),X(1),E(1),AIN(*),NXYZ(3)
C    W[N*(N+3)]
      DDMAG=0.1*ESCALE
      SCER=0.05/ESCALE
      JJ=N*N+N
      JJJ=JJ+N
      K=N+1
      NFCC=1
      IND=1
      INN=1
      DO 1 I=1,N
      DO 2 J=1,N
      W(K)=0.
      IF(I-J)4,3,4
    3 W(K)=ABS(E(I))
      W(I)=ESCALE
    4 K=K+1
    2 CONTINUE
    1 CONTINUE
      ITERC=1
      ISGRAD=2
      CALL CALCFX(N,X,F,AIN,
     +  CS,WL,WGH1,WGH2,THETATR,RMIN2,RMAX2,NXYZ,HW)
      FKEEP=ABS(F)+ABS(F)
    5 ITONE=1
      FP=F
      SUM=0.
      IXP=JJ
      DO 6 I=1,N
      IXP=IXP+1
      W(IXP)=X(I)
    6 CONTINUE
      IDIRN=N+1
      ILINE=1
    7 DMAX=W(ILINE)
      DACC=DMAX*SCER
      DMAG=AMIN1(DDMAG,0.1*DMAX)
      DMAG=AMAX1(DMAG,20.*DACC)
      DDMAX=10.*DMAG
      GO TO (70,70,71),ITONE
   70 DL=0.
      D=DMAG
      FPREV=F
      IS=5
      FA=F
      DA=DL
    8 DD=D-DL
      DL=D
   58 K=IDIRN
      DO 9 I=1,N
      X(I)=X(I)+DD*W(K)
      K=K+1
    9 CONTINUE
      CALL CALCFX(N,X,F,AIN,
     +  CS,WL,WGH1,WGH2,THETATR,RMIN2,RMAX2,NXYZ,HW)
C
      NFCC=NFCC+1
      GO TO (10,11,12,13,14,96),IS
   14 IF(F-FA)15,16,24
   16 IF (ABS(D)-DMAX) 17,17,18
   17 D=D+D
      GO TO 8
   18 WRITE(6,19)
   19 FORMAT(5X,44HVA04A MAXIMUM CHANGE DOES NOT ALTER FUNCTION)
      GO TO 20
   15 FB=F
      DB=D
      GO TO 21
   24 FB=FA
      DB=DA
      FA=F
      DA=D
   21 GO TO (83,23),ISGRAD
   23 D=DB+DB-DA
      IS=1
      GO TO 8
   83 D=0.5*(DA+DB-(FA-FB)/(DA-DB))
      IS=4
      IF((DA-D)*(D-DB))25,8,8
   25 IS=1
      IF(ABS(D-DB)-DDMAX)8,8,26
   26 D=DB+SIGN(DDMAX,DB-DA)
      IS=1
      DDMAX=DDMAX+DDMAX
      DDMAG=DDMAG+DDMAG
      IF(DDMAX-DMAX)8,8,27
   27 DDMAX=DMAX
      GO TO 8
   13 IF(F-FA)28,23,23
   28 FC=FB
      DC=DB
   29 FB=F
      DB=D
      GO TO 30
   12 IF(F-FB)28,28,31
   31 FA=F
      DA=D
      GO TO 30
   11 IF(F-FB)32,10,10
   32 FA=FB
      DA=DB
      GO TO 29
   71 DL=1.
      DDMAX=5.
      FA=FP
      DA=-1.
      FB=FHOLD
      DB=0.
      D=1.
   10 FC=F
      DC=D
   30 A=(DB-DC)*(FA-FC)
      B=(DC-DA)*(FB-FC)
      IF((A+B)*(DA-DC))33,33,34
   33 FA=FB
      DA=DB
      FB=FC
      DB=DC
      GO TO 26
   34 D=0.5*(A*(DB+DC)+B*(DA+DC))/(A+B)
      DI=DB
      FI=FB
      IF(FB-FC)44,44,43
   43 DI=DC
      FI=FC
   44 GO TO (86,86,85),ITONE
   85 ITONE=2
      GO TO 45
   86 IF (ABS(D-DI)-DACC) 41,41,93
   93 IF (ABS(D-DI)-0.03*ABS(D)) 41,41,45
   45 IF ((DA-DC)*(DC-D)) 47,46,46
   46 FA=FB
      DA=DB
      FB=FC
      DB=DC
      GO TO 25
   47 IS=2
      IF ((DB-D)*(D-DC)) 48,8,8
   48 IS=3
      GO TO 8
   41 F=FI
      D=DI-DL
      DD=SQRT((DC-DB)*(DC-DA)*(DA-DB)/(A+B))
      DO 49 I=1,N
      X(I)=X(I)+D*W(IDIRN)
      W(IDIRN)=DD*W(IDIRN)
      IDIRN=IDIRN+1
   49 CONTINUE
      IF (DD.EQ.0.0) DD=1E-10
      W(ILINE)=W(ILINE)/DD
      ILINE=ILINE+1
      IF(IPRINT-1)51,50,51
   50 WRITE(6,52) ITERC,NFCC,F,(X(I),I=1,N)
   52 FORMAT (/1X,9HITERATION,I5,I15,16H FUNCTION VALUES,
     110X,3HF =,E21.14/(5E24.14))
      GO TO(51,53),IPRINT
   51 GO TO (55,38),ITONE
   55 IF (FPREV-F-SUM) 94,95,95
   95 SUM=FPREV-F
      JIL=ILINE
   94 IF (IDIRN-JJ) 7,7,84
   84 GO TO (92,72),IND
   92 FHOLD=F
      IS=6
      IXP=JJ
      DO 59 I=1,N
      IXP=IXP+1
      W(IXP)=X(I)-W(IXP)
   59 CONTINUE
      DD=1.
      GO TO 58
   96 GO TO (112,87),IND
  112 IF (FP-F) 37,37,91
   91 D=2.*(FP+F-2.*FHOLD)/(FP-F)**2
      IF (D*(FP-FHOLD-SUM)**2-SUM) 87,37,37
   87 J=JIL*N+1
      IF (J-JJ) 60,60,61
   60 DO 62 I=J,JJ
      K=I-N
      W(K)=W(I)
   62 CONTINUE
      DO 97 I=JIL,N
      W(I-1)=W(I)
   97 CONTINUE
   61 IDIRN=IDIRN-N
      ITONE=3
      K=IDIRN
      IXP=JJ
      AAA=0.
      DO 65 I=1,N
      IXP=IXP+1
      W(K)=W(IXP)
      IF (AAA-ABS(W(K)/E(I))) 66,67,67
   66 AAA=ABS(W(K)/E(I))
   67 K=K+1
   65 CONTINUE
      DDMAG=1.
      IF (AAA.EQ.0.0) AAA=1E-10
      W(N)=ESCALE/AAA
      ILINE=N
      GO TO 7
   37 IXP=JJ
      AAA=0.
      F=FHOLD
      DO 99 I=1,N
      IXP=IXP+1
      X(I)=X(I)-W(IXP)
      IF (AAA*ABS(E(I))-ABS(W(IXP))) 98,99,99
   98 AAA=ABS(W(IXP)/E(I))
   99 CONTINUE
      GO TO 72
   38 AAA=AAA*(1.+DI)
      GO TO (72,106),IND
   72 IF (IPRINT-2) 53,50,50
   53 GO TO (109,88),IND
  109 IF (AAA-0.1) 89,89,76
   89 GO TO (20,116),ICON
  116 IND=2
      GO TO (100,101),INN
  100 INN=2
      K=JJJ
      DO 102 I=1,N
      K=K+1
      W(K)=X(I)
      X(I)=X(I)+10.*E(I)
  102 CONTINUE
      FKEEP=F
      CALL CALCFX (N,X,F,AIN,
     +  CS,WL,WGH1,WGH2,THETATR,RMIN2,RMAX2,NXYZ,HW)
      NFCC=NFCC+1
      DDMAG=0.
      GO TO 108
   76 IF (F-FP) 35,78,78
   78 WRITE(6,80)
   80 FORMAT (5X,37HVA04A ACCURACY LIMITED BY ERRORS IN F)
      GO TO 20
   88 IND=1
   35 TMP=FP-F
      IF (TMP.GT.0.0) THEN
      DDMAG=0.4*SQRT(TMP)
      ELSE
      DDMAG=0.0
      ENDIF
      ISGRAD=1
  108 ITERC=ITERC+1
      IF (ITERC-MAXIT) 5,5,81
81    CONTINUE
C   81 WRITE(6,82) MAXIT
   82 FORMAT(I5,30H ITERATIONS COMPLETED BY VA04A)
      IF (F-FKEEP) 20,20,110
  110 F=FKEEP
      DO 111 I=1,N
      JJJ=JJJ+1
      X(I)=W(JJJ)
  111 CONTINUE
      GO TO 20
  101 JIL=1
      FP=FKEEP
      IF (F-FKEEP) 105,78,104
  104 JIL=2
      FP=F
      F=FKEEP
  105 IXP=JJ
      DO 113 I=1,N
      IXP=IXP+1
      K=IXP+N
      GO TO (114,115),JIL
  114 W(IXP)=W(K)
      GO TO 113
  115 W(IXP)=X(I)
      X(I)=W(K)
  113 CONTINUE
      JIL=2
      GO TO 92
  106 IF (AAA-0.1) 20,20,107
   20 RETURN
  107 INN=1
      GO TO 35
      END
