C*****************************************************************************
C
C    CTFTILT - determines defocus, astigmatism tilt axis and tilt angle
C    for images of arbitrary size (MRC format). Astigmatic angle is measured
C    from x axis (same conventions as in the MRC 2D image processing 
C    programs).
C
C    CARD 1: Input file name for image
C    CARD 2: Output file name to check result
C    CARD 3: CS[mm], HT[kV], AmpCnst, XMAG, DStep[um],PAve
C    CARD 4: Box, ResMin[A], ResMax[A], dFMin[A], dFMax[A], FStep[A], TiltA[deg], TiltR[deg]
C
C        The output image file to check the result of the fitting
C        shows the filtered average power spectrum of the input
C        image in one half, and the fitted CTF (squared) in the
C        other half. The two halves should agree very well for a
C        successful fit.
C
C        CS: Spherical aberration coefficient of the objective in mm
C        HT: Electron beam voltage in kV
C        AmpCnst: Amount of amplitude contrast (fraction). For ice
C                 images 0.07, for negative stain about 0.15.
C        XMAG: Magnification of original image
C        DStep: Pixel size on scanner in microns
C        PAve: Pixel averaging (PAve x PAve) for input image 
C        Box: Tile size. The program divides the image into square
C             tiles and calculates the average power spectrum. Tiles
C             with a significantly higher or lower variance are 
C             excluded; these are parts of the image which are unlikely
C             to contain useful information (beam edge, film number 
C             etc). IMPORTANT: Box must have a even pixel dimensions.
C        ResMin: Low resolution end of data to be fitted.
C        ResMaX: High resolution end of data to be fitted.
C        dFMin: Starting defocus value for grid search in Angstrom. 
C               Positive values represent an underfocus. The program
C               performs a systematic grid search of defocus values 
C               and astigmatism before fitting a CTF to machine 
C               precision.
C        dFMax: End defocus value for grid search in Angstrom.
C        FStep: Step width for grid search in Angstrom.
C        TiltA: Expected tilt angle in degrees
C        TiltR: Expected tilt angle uncertainty in degrees
C
C*****************************************************************************
C    example command file (UNIX):
C
C    #!/bin/csh -x
C    #
C    #   ctftilt
C    #
C    time /public/image/bin/ctftilt.exe << eof
C    image.mrc
C    power.mrc
C    2.6,200.0,0.07,60000.0,28.0,2
C    128,100.0,15.0,30000.0,90000.0,5000.0,0.0,5.0
C    eof
C    #
C*****************************************************************************
C
      PROGRAM CTFTILT
C
C    NIKO, 21 SEPTEMBER 2002
C
      IMPLICIT NONE
C
      INTEGER NXYZ(3),MODE,JXYZ(3),I,J,IS,KXYZ(3),NX,NY
      INTEGER ID,L,M,LL,MM,ITEST,IAVE,K,IERR
      INTEGER IX,IP,NBIN,NR,CNT,IY,IMP
      PARAMETER (NR=5,NBIN=100)
      REAL DMIN,DMAX,DMEAN,DRMS,WGH1,WGH2,SCAL,WL,MEAN,PI
      REAL CS,KV,WGH,XMAG,DSTEP,RESMIN,RESMAX,DFMID1,DFMID2
      REAL THETATR,STEPR,RMIN2,RMAX2,HW,TLTAXIS,ANGAST
      REAL RES2,CTF,CTFV,TMP,FLT,DFMIN,DFMAX,FSTEP,R,RMS
      REAL DRMS1,TANGLE,SIG2,TILTA,TILTR
      REAL MIN,MAX,RMSMIN,RMSMAX,CMAX
      REAL,ALLOCATABLE :: AIN(:),ABOX(:),POWER(:),OUT(:),BUF1(:)
      REAL,ALLOCATABLE :: BUF2(:),RMSA(:),BINS(:)
      PARAMETER (FLT=-0.1,PI=3.1415926535898)
      COMPLEX,ALLOCATABLE :: CBOXS(:)
      CHARACTER FILEIN*70,FILEOUT*70,TITLE*1600,CFORM,NCPUS*10
      LOGICAL EX
      COMMON/FUNC/CS,WL,WGH1,WGH2,THETATR,RMIN2,RMAX2,JXYZ,HW
      COMMON/FUNCB/NXYZ,STEPR,TILTA,TILTR,SIG2
C

C     RSS Initialization to prevent segmentation faults      
      RMSMIN = 0.0              

      WRITE(6,1000)
1000  FORMAT(/' CTF TILT DETERMINATION, V1.5 (19-August-2010)',
     +       /' Distributed under the GNU',
     +        ' General Public License (GPL)')
C                    
      IMP=0
      CALL GETENV('OMP_NUM_THREADS',NCPUS)
      READ(NCPUS,*,ERR=111,END=111)IMP
111   CONTINUE
      IF (IMP.LE.0) THEN
        CALL GETENV('NCPUS',NCPUS)
        READ(NCPUS,*,ERR=112,END=112)IMP
112     CONTINUE
      ENDIF
      IF (IMP.LE.0) IMP=1
      IF (IMP.GT.1) THEN
        WRITE(*,7000) IMP
7000    FORMAT(/' Parallel processing: NCPUS =   ',I8)
      ENDIF
C
C    Read in all input parameters and I/O files
C
      WRITE(6,1002)
1002  FORMAT(/' Input image file name')
      READ(5,1010)FILEIN
1010  FORMAT(A)
      WRITE(6,1010)FILEIN
      WRITE(6,1060)
1060  FORMAT(/' Output diagnostic file name')
      READ(5,1010)FILEOUT
      WRITE(6,1010)FILEOUT
      WRITE(6,1020)
1020  FORMAT(/' CS[mm], HT[kV], AmpCnst, XMAG, DStep[um], PAve')
      READ(5,*)CS,KV,WGH,XMAG,DSTEP,IAVE
      WRITE(6,1031)CS,KV,WGH,XMAG,DSTEP,IAVE
1031  FORMAT(F5.1,F9.1,F8.2,F10.1,F9.3,I5)
      WRITE(6,1040)
1040  FORMAT(/' Positive defocus values for underfocus',
     +       /' Box, ResMin[A], ResMax[A], dFMin[A], dFMax[A],',
     +        ' FStep[A], TiltA[deg], TiltR[deg]')
      READ(5,*)JXYZ(1),RESMIN,RESMAX,DFMIN,DFMAX,FSTEP,TILTA,
     +         TILTR
      WRITE(6,1051)JXYZ(1),RESMIN,RESMAX,DFMIN,DFMAX,FSTEP,
     +         TILTA,TILTR
C
C    Check input parameters
C
1051  FORMAT(I4,2F11.1,2F10.1,F10.1,2F12.1/)
      ITEST=JXYZ(1)/2
      IF (2*ITEST.NE.JXYZ(1)) THEN
          WRITE(6,1090)
1090    FORMAT(/' Box size must be even number')
          STOP
      ENDIF
C
C     JXYZ is size of box, X,Y dimensions equal
C
      JXYZ(2)=JXYZ(1)
      JXYZ(3)=1
C
C    Make sure RESMIN is larger than RESMAX
C
      IF (RESMIN.LT.RESMAX) THEN
          TMP=RESMAX
          RESMAX=RESMIN
          RESMIN=TMP
      ENDIF
C
C    Same for DFMIN, DFMAX
C
      IF (DFMAX.LT.DFMIN) THEN
          TMP=DFMAX
          DFMAX=DFMIN
          DFMIN=TMP
      ENDIF
C
C    Open input image
C
      CALL GUESSF(FILEIN,CFORM,EX)
      IF  (.NOT.EX) THEN
        WRITE(*,1001) FILEIN
1001    FORMAT(' File not found ',A80)
        STOP
      ENDIF
      CALL IOPEN(FILEIN,10,CFORM,MODE,NXYZ(1),NXYZ(2),
     +           NXYZ(3),'OLD',STEPR,TITLE)
C
C    STEPR is pixel size in Angstrom
C
      STEPR=DSTEP*(10.0**4.0)/XMAG
C
      WRITE(*,1110)NXYZ(1),NXYZ(2)
1110  FORMAT(/,' READING IMAGE...'/' NX, NY= ',2I10)
      IF (IAVE.GT.1) WRITE(*,1100)IAVE,IAVE
1100  FORMAT(/,' PIXEL AVERAGING = ',I1,' x ',I1,' ...')
C
      ALLOCATE(AIN(NXYZ(1)/IAVE*NXYZ(2)/IAVE+NXYZ(1)*IAVE),
     +         STAT=IERR)
      IF (IERR.NE.0) THEN
        WRITE(*,*) ' ERROR: Memory allocation failed in MAIN'
        STOP ' Try reducing the image size or increase PAVE'
      ENDIF
C
C    Read in image and compress at the same time
C
      DO 10 J=1,NXYZ(2)/IAVE
        IP=(J-1)*IAVE
        DO 11 K=1,IAVE
          ID=1+(NXYZ(1)/IAVE)*(J-1)+NXYZ(1)*(K-1)
          CALL IREAD(10,AIN(ID),K+IP)
          IF (K.GT.1) THEN
            DO 12 L=1,NXYZ(1)
              ID=L+(NXYZ(1)/IAVE)*(J-1)
              IS=L+(NXYZ(1)/IAVE)*(J-1)+NXYZ(1)*(K-1)
              AIN(ID)=AIN(ID)+AIN(IS)
12          CONTINUE
          ENDIF
11      CONTINUE
        DO 10 I=1,NXYZ(1)/IAVE
          R=0.0
          DO 13 K=1,IAVE
            ID=IAVE*(I-1)+K+(NXYZ(1)/IAVE)*(J-1)
            R=R+AIN(ID)
13        CONTINUE
          R=R/IAVE**2
          ID=I+(NXYZ(1)/IAVE)*(J-1)
          AIN(ID)=R
10    CONTINUE
C
      CALL ICLOSE(10)
C
      NXYZ(1)=NXYZ(1)/IAVE
      NXYZ(2)=NXYZ(2)/IAVE
      DSTEP=DSTEP*IAVE
      STEPR=STEPR*IAVE
C
      NX=NXYZ(1)/JXYZ(1)
      NY=NXYZ(2)/JXYZ(2)
C
      ALLOCATE(ABOX(JXYZ(1)*JXYZ(2)),
     +  OUT(JXYZ(1)*JXYZ(2)),
     +  BUF1(JXYZ(1)*JXYZ(2)),RMSA(NX*NY),BINS(NBIN),
     +  BUF2(JXYZ(1)*JXYZ(2)),CBOXS(JXYZ(2)),STAT=IERR)
      IF (IERR.NE.0) THEN
        WRITE(*,*) ' ERROR: Memory allocation failed in MAIN'
        STOP ' Try reducing the tile size'
      ENDIF
C
      DRMS=0.0D0
      CNT=0
      DO 101 I=1,NY
        DO 101 J=1,NX
          IX=(J-1)*JXYZ(1)+1
          IY=1+(I-1)*JXYZ(2)
          CALL BOXIMG(AIN,NXYZ,ABOX,JXYZ,IX,IY,MEAN,RMS)
          DRMS=DRMS+RMS**2
          CNT=CNT+1
          IF (CNT.LE.50000) RMSA(CNT)=RMS
101   CONTINUE
      DRMS=SQRT(DRMS/CNT)
      CALL HISTO(CNT,NBIN,RMSA,BINS,MIN,MAX)
      CMAX=0
      DO 70 I=1,NBIN
        IF (BINS(I).GT.CMAX) THEN
          CMAX=BINS(I)
          J=I
        ENDIF
70    CONTINUE
      IF (J.GT.1) THEN
        DO 71 I=J-1,1,-1
          IF (BINS(I).LT.CMAX/10.0) THEN
            RMSMIN=(I-1)*(MAX-MIN)/(NBIN-1)+MIN
            GOTO 72
          ENDIF
71      CONTINUE
      ENDIF
72    CONTINUE
      IF (J.LT.NBIN) THEN
        DO 73 I=J+1,NBIN
          IF (BINS(I).LT.CMAX/10.0) THEN
            RMSMAX=(I-1)*(MAX-MIN)/(NBIN-1)+MIN
            GOTO 74
          ENDIF
73      CONTINUE
      ENDIF
74    CONTINUE
C
C    Convert units
C
      CS=CS*(10.0**7.0)                          ! Angstroms
      KV=KV*1000.0                               ! Volts
      WL=12.26/SQRT(KV+0.9785*KV**2/(10.0**6.0)) ! Angstroms
C
C    Parameters for CTF calculation
C
      WGH1=SQRT(1.0-WGH**2)
      WGH2=WGH
      THETATR=WL/(STEPR*JXYZ(1))
C
C    Convert resolution into pixel resolution
C
      RESMIN=STEPR/RESMIN
      RESMAX=STEPR/RESMAX
      IF (RESMIN.LT.STEPR/50.0) THEN
        RESMIN=STEPR/50.0
        WRITE(*,*)
        WRITE(*,*)' Lower resolution limit reset to',
     +            STEPR/RESMIN,' A'
      ENDIF
      IF (RESMIN.GE.RESMAX)
     +  STOP ' RESMIN >= RESMAX; increase RESMAX'
C
C    KXYZ is dimension of power spectrum
C    This needs to be even for later filtering
C
      KXYZ(1)=JXYZ(1)/2
      KXYZ(2)=JXYZ(2)
      KXYZ(3)=JXYZ(3)
      IF (2*(KXYZ(1)/2).NE.KXYZ(1)) KXYZ(1)=KXYZ(1)+1
      ALLOCATE(POWER(KXYZ(1)*KXYZ(2)),STAT=IERR)
      IF (IERR.NE.0) THEN
        WRITE(*,*) ' ERROR: Memory allocation failed for POWER'
        STOP ' Try reducing the tile size'
      ENDIF
C
C
C    Find approx. direction of tilt axis
C    POWER is average power spectrum along tilt axis
C    TLTAXIS is angle between X-axis and tilt axis
C
      CALL FIND_TAXIS(AIN,NXYZ,RMSMIN,RMSMAX,
     +  JXYZ,POWER,KXYZ,TLTAXIS,NR)
C
C    If expected tilt angle is 0.0 deg, use entire image to do
C       initial CTF fit (as in CTFFIND3)
C
      IF (TILTA.LE.20.0) THEN
C
        WRITE(*,1102)
1102    FORMAT(/,' CALCULATING AVERAGE POWER SPECTRUM',/,
     +           '    OF ENTIRE IMAGE FOR INITIAL CTF FIT...'/)
        DO 30 K=1,KXYZ(1)*KXYZ(2)
          POWER(K)=0.0
30      CONTINUE
        SCAL=1.0/SQRT(REAL(JXYZ(1)*JXYZ(2)))
        CNT=0
C
        DO 300 I=1,NY
          DO 300 J=1,NX
            IX=(J-1)*JXYZ(1)+1
            IY=1+(I-1)*JXYZ(2)
            CALL BOXIMG(AIN,NXYZ,ABOX,JXYZ,IX,IY,MEAN,RMS)
            IF ((RMS.LT.RMSMAX).AND.(RMS.GT.RMSMIN)) THEN
              CALL RLFT3(ABOX,CBOXS,JXYZ(1),JXYZ(2),1,1)
              DO 340 L=1,JXYZ(2)
                DO 341 K=1,JXYZ(1)/2
                  ID=(K+JXYZ(1)/2*(L-1))*2
                  IS=K+KXYZ(1)*(L-1)
                  POWER(IS)=POWER(IS)
     +              +(ABOX(ID-1)**2+ABOX(ID)**2)*SCAL**2
341             CONTINUE
                IF (KXYZ(1).GT.JXYZ(1)/2)
     +            POWER(IS+1)=POWER(IS+1)+CABS(CBOXS(L)*SCAL)**2
340           CONTINUE
              CNT=CNT+1
            ENDIF
300     CONTINUE
C
        SCAL=1.0/CNT
        DO 360 K=1,KXYZ(1)*KXYZ(2)
          POWER(K)=SQRT(POWER(K)*SCAL)
360     CONTINUE
C
      ELSE
C
        WRITE(*,1101)
1101    FORMAT(/,' USING AVERAGE POWER SPECTRUM FROM',/,
     +           '    IMAGE CENTER FOR INITIAL CTF FIT...'/)
      ENDIF
C
C    Filter power spectrum to remove slowly varying background
C    DMAX is maximum of filtered power spectrum (for later scaling)
C
      CALL FILTER(JXYZ,KXYZ,POWER,BUF1,DMEAN,DRMS1,DMAX,STEPR)
C
C    Search for rough CTF parameters DFMID1, DFMID2, ANGAST
C
      DFMID1=DFMIN
      DFMID2=DFMAX
      CALL SEARCH_CTF(CS,WL,WGH1,WGH2,THETATR,RESMIN,RESMAX,
     +         POWER,JXYZ,DFMID1,DFMID2,ANGAST,FSTEP)
C
C    The following parameters are passed to the refinement
C    routine via common block FUNC
C    HW is exponent in low bass filter
C
      RMIN2=RESMIN**2
      RMAX2=RESMAX**2
C      HW=-1.0/RMAX2
C      HW=-1.0/0.4**2
      HW=0.0

      CALL REFINE_CTF(DFMID1,DFMID2,ANGAST,AIN,POWER)
      CALL EVALCTF(CS,WL,WGH1,WGH2,DFMID1,DFMID2,ANGAST,
     +        THETATR,HW,AIN,JXYZ,RMIN2,RMAX2,CMAX,SIG2)
C
C    Create diagnostic image showing power spectrum
C    and matching squared CTF
C
      DO 50 I=1,JXYZ(1)*JXYZ(2)
          OUT(I)=0.0
50    CONTINUE
      DO 200 L=1,JXYZ(1)/2
        LL=L-1
        DO 200 M=1,JXYZ(2)
          MM=M-1
          IF (MM.GT.JXYZ(2)/2) MM=MM-JXYZ(2)
          ID=L+JXYZ(1)/2*(M-1)
            I=L+JXYZ(1)/2
            J=M+JXYZ(2)/2
            IF (J.GT.JXYZ(2)) J=J-JXYZ(2)
            IS=I+JXYZ(1)*(J-1)
            OUT(IS)=POWER(ID)/DRMS1*SQRT(2.0*PI)
            IF (OUT(IS).GT.1.0) OUT(IS)=1.0
            IF (OUT(IS).lT.-1.0) OUT(IS)=-1.0
          RES2=(REAL(LL)/JXYZ(1))**2+(REAL(MM)/JXYZ(2))**2
          IF ((RES2.LE.RMAX2).AND.(RES2.GE.RMIN2)) THEN
            CTFV=CTF(CS,WL,WGH1,WGH2,DFMID1,DFMID2,
     +               ANGAST,THETATR,LL,MM)
             I=JXYZ(1)/2-L+1
              J=JXYZ(2)-J+2
              IF (J.LE.JXYZ(2)) THEN
                IS=I+JXYZ(1)*(J-1)
              OUT(IS)=CTFV**2
              ENDIF
          ENDIF
200   CONTINUE
C
C    Write out diagnostic image in mode 2 (floating point)
C
      MODE=2
      IF (STEPR.NE.0.0) DSTEP=1.0/STEPR
      CALL IOPEN(FILEOUT,10,CFORM,MODE,JXYZ(1),JXYZ(2),
     +           JXYZ(3),'NEW',DSTEP,TITLE)
      DO 210 J=1,JXYZ(2)
        ID=1+JXYZ(1)*(J-1)
        CALL IWRITE(10,OUT(ID),J)
210   CONTINUE
      CALL ICLOSE(10)
C
C    Replace array AIN with input image by local power spectra
C    by dividing up the image into tiles of size KXYZ and
C    calculating power spectra for each tile 
C
      CALL TILE(JXYZ,KXYZ,NXYZ,AIN,ABOX,ABOX,CBOXS,
     +             RMSMIN,RMSMAX,POWER,BUF1)
C
C    Search for the tilt angle that produces the highest
C    correlation coefficient between power spectra and
C    calculated squared CTF
C
      IF (TILTR.LT.2.5) TILTR=2.4999
      TILTA=ABS(TILTA/180.0*PI)
      TILTR=TILTR/180.0*PI
C
      CALL FIND_TANGLE(JXYZ,ABOX,CS,WL,
     +  WGH1,WGH2,THETATR,DFMID1,DFMID2,ANGAST,TLTAXIS,
     +  STEPR,RMIN2,RMAX2,HW,TANGLE,AIN)
C
C    Refine all parameters
C    Some parameters are passed through common block FUNCB
C
      CALL REFINE_TILT(DFMID1,DFMID2,ANGAST,TLTAXIS,TANGLE,AIN,POWER)
C
      GOTO 9999
999   STOP 'END-OF-FILE ERROR ON READ'
9999  CONTINUE
C
C    Close input file
C    Write out a little figure explaining the meaning of the
C    final refined parameters
C
      CALL FIGURE(DFMID1,DFMID2,ANGAST,TLTAXIS,TANGLE,NXYZ,IAVE,
     +            STEPR)
      END
C
C**************************************************************************
      SUBROUTINE FIGURE(DFMID1,DFMID2,ANGAST,TLTAXIS,TANGLE,
     +                  NXYZ,IAVE,PSIZE)
C**************************************************************************
C    Writes out a figure indicating the scanned image with defocus
C    values in the center and at the four corners. Also writes out
C    a formula for calculating the defocus at an arbitrary location
C**************************************************************************
      IMPLICIT NONE
C
      INTEGER NXYZ(3),IAVE,IW,I,CX,CY
      PARAMETER (IW=60)
      REAL DFMID1,DFMID2,ANGAST,TLTAXIS,TANGLE,PSIZE
      REAL DX,DY,N(2),DF,DFL11,DFL12,R,DFL21,DFL22
C
      CX=(NXYZ(1)*IAVE)/2
      CY=(NXYZ(2)*IAVE)/2
      N(1)=-SIN(TLTAXIS)
      N(2)=COS(TLTAXIS)
C
      DX=CX-1
      DY=CY-1
      R=(N(1)*DX+N(2)*DY)*PSIZE/IAVE
      DF=R*TAN(TANGLE)
      DFL11=DFMID1+DF
      DFL12=DFMID2+DF
      DX=CX-NXYZ(1)*IAVE
      DY=CY-1
      R=(N(1)*DX+N(2)*DY)*PSIZE/IAVE
      DF=R*TAN(TANGLE)
      DFL21=DFMID1+DF
      DFL22=DFMID2+DF
      WRITE(*,900)CX,CY,PSIZE/IAVE,N(1),N(2)
900   FORMAT(//,5X,'EQUATION FOR CALCULATING DEFOCUS ',
     +       'DFL1,DFL2 AT LOCATION NX,NY:',//,
     +       10X,'DFL1  = DFMID1 +DF',/,
     +       10X,'DFL2  = DFMID2 +DF',/,
     +       10X,'DF    = (N1*DX+N2*DY)*PSIZE*TAN(TANGLE)',/,
     +       10X,'DX    = CX-NX',/,
     +       10X,'DY    = CY-NY',/,
     +       10X,'CX    = CENTER_X = ',I12,/,
     +       10X,'CY    = CENTER_Y = ',I12,/,
     +       10X,'PSIZE = PIXEL SIZE [A] = ',F12.4,/,
     +       10X,'N1,N2 = TILT AXIS NORMAL:',/,
     +       13X,'N1 = -SIN(TLTAXIS) = ',F12.6,/,
     +       13X,'N2 =  COS(TLTAXIS) = ',F12.6,//)
      WRITE(*,1000)DFL11,DFL12,DFL21,DFL22
1000  FORMAT(F12.2,',',F12.2,4X,'<--(DFMID1,DFMID2)-->',
     +       3X,F12.2,',',F12.2)
      WRITE(*,1100)1,1,NXYZ(1)*IAVE,1
1100  FORMAT(I12,',',I12,4X,'<------(NX,NY)------>',
     +       3X,I12,',',I12)
      WRITE(*,1200)
1200  FORMAT(10X,'+-----------------------------',
     +            '-----------------------------+')
      DO 10 I=1,INT(REAL(NXYZ(2))/NXYZ(1)*IW/2)-2
        IF(I.EQ.INT(REAL(NXYZ(2))/NXYZ(1)*IW/4-1)) THEN
          WRITE(*,1300)DFMID1,DFMID2
1300      FORMAT(10X,'|',15X,F12.2,',',F12.2,18X,'|')
          WRITE(*,1400)CX,CY
1400      FORMAT(10X,'|',15X,I12,',',I12,18X,'|')
        ELSE
          WRITE(*,1500)
1500      FORMAT(10x,'|',58X,'|')
        ENDIF
10    CONTINUE
      WRITE(*,1200)
      DX=CX-1
      DY=CY-NXYZ(2)*IAVE
      R=(N(1)*DX+N(2)*DY)*PSIZE/IAVE
      DF=R*TAN(TANGLE)
      DFL11=DFMID1+DF
      DFL12=DFMID2+DF
      DX=CX-NXYZ(1)*IAVE
      DY=CY-NXYZ(2)*IAVE
      R=(N(1)*DX+N(2)*DY)*PSIZE/IAVE
      DF=R*TAN(TANGLE)
      DFL21=DFMID1+DF
      DFL22=DFMID2+DF
      WRITE(*,1100)1,NXYZ(2)*IAVE,NXYZ(1)*IAVE,NXYZ(2)*IAVE
      WRITE(*,1000)DFL11,DFL12,DFL21,DFL22
C
      RETURN
      END
C
C**************************************************************************
      SUBROUTINE REFINE_TILT(DFMID1,DFMID2,ANGAST,
     +       TLTAXIS,TANGLE,AIN,ABOX)
C**************************************************************************
C    Refines all five image parameters using Powell minimizer
C    VA04A
C**************************************************************************
      IMPLICIT NONE
C
      INTEGER NCYCLS
      PARAMETER (NCYCLS=50)
      REAL DFMID1,DFMID2,ANGAST,XPAR(5),EPAR(5),RF,ESCALE,PI
      REAL TLTAXIS,TANGLE,AIN(*),ABOX(*)
      PARAMETER (PI=3.1415926535898)
      DATA EPAR/0.05,0.05,100.0,100.0,0.5/
      DATA ESCALE/100.0/
      EXTERNAL CALCFXB
C
      WRITE(*,1000)
1000  FORMAT(/,' REFINING TILT PARAMETERS...'/,
     +       /,'      DFMID1      DFMID2      ANGAST',
     +         '     TLTAXIS      TANGLE          CC'/)
C
      XPAR(3)=DFMID1
      XPAR(4)=DFMID2
      XPAR(5)=ANGAST
      XPAR(1)=TLTAXIS
      XPAR(2)=TANGLE
      IF (XPAR(2).EQ.0.0) XPAR(2)=0.01
      IF (XPAR(3).EQ.XPAR(4)) XPAR(3)=XPAR(3)+1.0
      CALL VA04A(XPAR,EPAR,5,RF,ESCALE,0,1,NCYCLS,AIN,ABOX,CALCFXB)
      DFMID1=XPAR(3)
      DFMID2=XPAR(4)
      ANGAST=XPAR(5)
      TLTAXIS=XPAR(1)
      TANGLE=XPAR(2)
      WRITE(*,1100)DFMID1,DFMID2,ANGAST/PI*180.0,
     +             TLTAXIS/PI*180.0,TANGLE/PI*180.0,-RF
1100  FORMAT(/,5F12.2,F12.5,'  Final Values')
C
      RETURN
      END
C
C**************************************************************************
      SUBROUTINE FIND_TANGLE(JXYZ,ABOX,CS,WL,
     +  WGH1,WGH2,THETATR,DFMID1,DFMID2,ANGAST,TLTAXIS,
     +  PSIZE,RMIN2,RMAX2,HW,TANGLE,AIN)
C**************************************************************************
C    Search for tilt angle in 10 deg increments, starting from
C    -65 deg to +65 deg. The function EVAL_TILT calculates the
C    correlation coefficient between the calculated CTF for each
C    tile in AIN, and the power spectrum of the tile.
C**************************************************************************
      IMPLICIT NONE
C
      INTEGER JXYZ(*),NXYZ(3),K,ITILTA,ITILTR,IXMX
      REAL WGH1,WGH2,ANGAST,THETATR,EVAL_TILT,SIG2
      REAL AIN(*),ABOX(*),TLTAXIS,PI,TA,WL
      REAL DFMID1,DFMID2,PSIZE,CS,HW,TAR,TILTA,TILTR
      REAL SUM,SUMMAX,RMIN2,RMAX2,TANGLE,STEPR
      PARAMETER (PI=3.1415926535898)
      COMMON/FUNCB/NXYZ,STEPR,TILTA,TILTR,SIG2
C
      WRITE(*,1100)
1100  FORMAT(/,' SEARCHING FOR TILT ANGLE...'/)
C
      ITILTR=NINT(TILTR*180.0/PI/5.0)*5
      ITILTA=NINT(TILTA*180.0/PI)
      SUMMAX=-1.0E30
      DO 10 K=ITILTA-ITILTR,ITILTA+ITILTR,10
        TA=K/180.0*PI
        SUM=EVAL_TILT(JXYZ,NXYZ,AIN,CS,WL,
     +  WGH1,WGH2,THETATR,DFMID1,DFMID2,ANGAST,TLTAXIS,
     +  PSIZE,RMIN2,RMAX2,HW,TA)+TAR(SIG2,TA,TILTA,TILTR)
          IF (SUM.GT.SUMMAX) THEN
            SUMMAX=SUM
          TANGLE=K
            WRITE(*,1000)TANGLE,SUM
1000      FORMAT('  Tilt angle, CC = ',F5.1,F12.5)
          ENDIF
10    CONTINUE
      TANGLE=TANGLE/180.0*PI
C
      RETURN
      END
C
C**************************************************************************
      REAL FUNCTION EVAL_TILT(JXYZ,NXYZ,AIN,CS,WL,
     +  WGH1,WGH2,THETATR,DFMID1,DFMID2,ANGAST,TLTAXIS,
     +  PSIZE,RMIN2,RMAX2,HW,TA)
C**************************************************************************
C    Calculates the correlation coefficient between the calculated
C    CTF for each tile in AIN, and the power spectrum of the tile.
C**************************************************************************
      IMPLICIT NONE
C
      INTEGER NX,NY,JXYZ(*),NXYZ(*),I,J,IX,IY
      INTEGER CNT,CX,CY,KXYZ(3),IS,IERR,ID,K
      REAL WGH1,WGH2,ANGAST,THETATR
      REAL AIN(*),MEAN,RMS,HW,TLTAXIS
      REAL DFMID1,DFMID2,PSIZE,CS,PI,TA,WL
      REAL SUM,SUMMAX,RMIN2,RMAX2,N(2),SIG2
      REAL,ALLOCATABLE :: CCTF(:)
      PARAMETER (PI=3.1415926535898)
C
C    JXYZ is tile size
C    calculate number of tiles in X and Y
C
      NX=NXYZ(1)/JXYZ(1)
      NY=NXYZ(2)/JXYZ(2)
C
C    calculate center of image
C
      CX=NXYZ(1)/2
      CY=NXYZ(2)/2
C
C    KXYZ is size of power spectrum (same as KXYZ
C    but half the size in X)
C
      KXYZ(1)=JXYZ(1)/2
      KXYZ(2)=JXYZ(2)
      KXYZ(3)=JXYZ(3)
C
C    N is normal to tilt axis, indicates the direction
C    in which defocus varies most
C
      N(1)=-SIN(TLTAXIS)
      N(2)=COS(TLTAXIS)
C
C    check here if tilt angle TA happens to be +-90 deg.
C    if so, apply small offset to avoid degeneracy in
C    refinement of tilt axis
C
      IF (ABS(ABS(TA)-PI/2.0).LT.0.0001) THEN
        IF (TA.GT.0.0) TA=PI/2.0-0.0001
        IF (TA.LT.0.0) TA=-PI/2.0+0.0001
      ENDIF
C
      ALLOCATE(CCTF(NX*NY),STAT=IERR)
      IF (IERR.NE.0) THEN
        WRITE(*,*) ' ERROR: Memory allocation failed in EVAL_TILT'
        STOP ' Try reducing number of tiles by increasing tile size'
      ENDIF
C
C    loop over all tiles
C
      DO 100 I=1,NY
!$OMP PARALLEL DO
        DO 100 J=1,NX
          CALL EVAL_TILT_S(JXYZ,NXYZ,AIN,CS,WL,
     +      WGH1,WGH2,THETATR,DFMID1,DFMID2,ANGAST,PSIZE,
     +      RMIN2,RMAX2,HW,TA,CX,CY,KXYZ,N,NX,CCTF,I,J)
100     CONTINUE
C
      SUM=0.0
      CNT=0
C
      DO 101 I=1,NX
        DO 101 J=1,NY
          IX=(J-1)*JXYZ(1)+1
          IY=1+(I-1)*JXYZ(2)
          ID=IX-1+JXYZ(1)+NXYZ(1)*(IY-1+JXYZ(2)-1)
C
C     AIN(ID) is 1 if tile was good, otherwise 0 (see subroutine TILE)
C
          IF (AIN(ID).EQ.1.0) THEN
            IS=J+(I-1)*NX
              SUM=SUM+CCTF(IS)
            CNT=CNT+1
          ENDIF
101     CONTINUE
C
C    calculate average correlation coefficient
C
      SUM=SUM/CNT
C
      DEALLOCATE(CCTF)
      EVAL_TILT=SUM
C
      RETURN
      END
C
C**************************************************************************
      SUBROUTINE EVAL_TILT_S(JXYZ,NXYZ,AIN,CS,WL,
     +  WGH1,WGH2,THETATR,DFMID1,DFMID2,ANGAST,PSIZE,
     +  RMIN2,RMAX2,HW,TA,CX,CY,KXYZ,N,NX,CCTF,I,J)
C**************************************************************************
C    Calculates the correlation coefficient between the calculated
C    CTF for each tile in AIN, and the power spectrum of the tile.
C**************************************************************************
      IMPLICIT NONE
C
      INTEGER JXYZ(*),NXYZ(*),I,J,IX,IY,ID
      INTEGER CNT,CX,CY,KXYZ(3),NX,IS,IERR
      REAL WGH1,WGH2,ANGAST,THETATR,CCTF(*)
      REAL AIN(*),MEAN,RMS,HW,TA,WL
      REAL DFL1,DFL2,DFMID1,DFMID2,DF,DX,DY,R
      REAL RMIN2,RMAX2,N(2),SIG2,PSIZE,CS
      REAL,ALLOCATABLE :: ABOX(:)
C
C    calculate upper left corner of tile: IX, IY
C
      IX=(J-1)*JXYZ(1)+1
      IY=1+(I-1)*JXYZ(2)
C
C    calculate array element of AIN for lower right corner
C    of this tile: ID
C
      ID=IX-1+JXYZ(1)+NXYZ(1)*(IY-1+JXYZ(2)-1)
C
C     AIN(ID) is 1 if tile was good, otherwise 0 (see subroutine TILE)
C
      IF (AIN(ID).EQ.1.0) THEN
C
C    cut out power spectrum and put into ABOX
C
        ALLOCATE(ABOX(KXYZ(1)*KXYZ(2)),STAT=IERR)
        IF (IERR.NE.0) THEN
          WRITE(*,*) ' ERROR: Memory allocation failed in EVAL_TILT_S'
          STOP ' Try reducing the tile size'
        ENDIF
C
        CALL BOXIMGP(AIN,NXYZ,ABOX,KXYZ,IX,IY,MEAN,RMS)
C
C    calculate coordinates of center of tile relative to
C    center of image for which CTF values are being refined
C
        DX=CX-IX+JXYZ(1)/2
        DY=CY-IY+JXYZ(2)/2
C
C    calculate how far along tilt axis normal we are
C
        R=(N(1)*DX+N(2)*DY)*PSIZE
C
C    now calculate by how much the defocus changes over
C    the distance along the tilt axis normal
C
        DF=R*TAN(TA)
C
C    add this defocus change to DFMID1 and DFMID2
C
        DFL1=DFMID1+DF
        DFL2=DFMID2+DF
C
C    calculate the correlation coefficient between the power
C    spectrum of this tile and the calculated CTF.
C    HW is parameter in Gaussian filter for calculated
C    squared CTF, RMIN2,RMAX2 are resolution limits between
C    which to do the correlation analysis.
C
        IS=J+(I-1)*NX
        CALL EVALCTF(CS,WL,WGH1,WGH2,DFL1,DFL2,ANGAST,
     +    THETATR,HW,ABOX,JXYZ,RMIN2,RMAX2,CCTF(IS),SIG2)
C
        DEALLOCATE(ABOX)
      ENDIF
C
      RETURN
      END
C
C**************************************************************************
      SUBROUTINE TILE(JXYZ,KXYZ,NXYZ,AIN,ABOX,CBOX,CBOXS,
     +                RMSMIN,RMSMAX,POWER,BUF1)
C**************************************************************************
C    Divides the input image AIN into tiles of size KXYZ
C    and replaces each tile with its power spectrum.
C    Also calculates average power spectrum POWER.
C**************************************************************************
      IMPLICIT NONE
C
      INTEGER NX,NY,JXYZ(*),NXYZ(*),CNT,I,J,IX,IY,ID,IS,K,L
      INTEGER KXYZ(*),NW
      REAL SCAL,AIN(*),ABOX(*),MEAN,RMS,BUF1(*)
      REAL RMSMIN,RMSMAX,POWER(*),P
      COMPLEX CBOX(*),CBOXS(*)
C
      WRITE(*,1100)
1100  FORMAT(/,' TILING IMAGE...'/)
C
C    set power to zero
C
      DO 30 K=1,KXYZ(1)*KXYZ(2)
          POWER(K)=0.0
30    CONTINUE
C
C    NW is size of window for averaging to strip
C    power spectrum if smooth background
C
      NW=KXYZ(1)/10
C
C    SCAL is scale factor to keep numbers in reasonable range
C
      SCAL=1.0/SQRT(REAL(JXYZ(1)*JXYZ(2)))
C
C    NX, NY are numger of tiles in X, Y
C
      NX=NXYZ(1)/JXYZ(1)
      NY=NXYZ(2)/JXYZ(2)
      CNT=0
C
C    loop over all tiles
C
      DO 100 I=1,NY
          DO 100 J=1,NX
C
C    calculate upper left corner of tile: IX, IY
C
            IX=(J-1)*JXYZ(1)+1
          IY=1+(I-1)*JXYZ(2)
C
C    cut out tile and put into ABOX, tile size is JXYZ
C
            CALL BOXIMG(AIN,NXYZ,ABOX,JXYZ,IX,IY,MEAN,RMS)
C
C    exclude tile if STD is too small or too large
C
            IF ((RMS.LE.RMSMAX).AND.(RMS.GE.RMSMIN)) THEN
C
C    calculate power spectrum
C       store sqrt of power spectrum in BUF1 to reduce slope
C       of background (makes background subtraction in MSMOOTH
C       more accurate, background-subtracted power spectrum will
C       be squared after background subtraction)
C
              CALL RLFT3(ABOX,CBOXS,JXYZ(1),JXYZ(2),1,1)
            DO 40 L=1,JXYZ(2)
                DO 41 K=1,JXYZ(1)/2
                ID=K+JXYZ(1)/2*(L-1)
                IS=K+KXYZ(1)*(L-1)
                P=CABS(CBOX(ID)*SCAL)
                POWER(IS)=POWER(IS)+P**2
                BUF1(IS)=P
41          CONTINUE
C
C    if KXYZ(1) larger than JXYZ(1)/2 then need to store extra
C    line to get power spectrum with even dimensions
C
              IF (KXYZ(1).GT.JXYZ(1)/2) THEN
                P=CABS(CBOXS(L)*SCAL)
                POWER(IS+1)=POWER(IS+1)+P**2
                BUF1(IS+1)=P
              ENDIF
40        CONTINUE
C
C    strip smooth background from power spectrum and
C    put back into AIN to replace tile
C
            CALL MSMOOTH(BUF1,KXYZ,NW,ABOX)
            DO 62 L=1,KXYZ(2)
              DO 62 K=1,KXYZ(1)
                ID=IX-1+K+NXYZ(1)*(IY-1+L-1)
                IS=K+KXYZ(1)*(L-1)
                AIN(ID)=BUF1(IS)
62          CONTINUE
              CNT=CNT+1
C
C    set lower right corner of tile to 1 to indicate good tile
C    (this corner is not used to store data)
C
            ID=IX-1+JXYZ(1)+NXYZ(1)*(IY-1+JXYZ(2)-1)
            AIN(ID)=1.0
          ELSE
            ID=IX-1+JXYZ(1)+NXYZ(1)*(IY-1+JXYZ(2)-1)
C
C    set lower right corner of tile to 0 to indicate bad tile
C
            AIN(ID)=0.0
            ENDIF
100   CONTINUE
C
C    calculate sqrt of average power spectrum
C
      DO 60 K=1,KXYZ(1)*KXYZ(2)
          POWER(K)=SQRT(POWER(K))
60    CONTINUE
      SCAL=1.0/CNT
      CALL SCLIMG(POWER,KXYZ,SCAL)
C
      WRITE(*,*)' Total tiles and number used = ',NX*NY,CNT
C
      RETURN
      END
C
C**************************************************************************
      SUBROUTINE FIND_TAXIS(AIN,NXYZ,RMSMIN,RMSMAX,
     +                  JXYZ,POWER,KXYZ,TLTAXIS,NR)
C**************************************************************************
C    Finds tilt axis by calculating power spectra along lines
C    across input image AIN to minimize variance between spectra.
C    JXYZ gives tile size
C**************************************************************************
C
      IMPLICIT NONE
C
      INTEGER NXYZ(*),JXYZ(*),I,K,KXYZ(*),NR,IERR
      REAL AIN(*),RMSMIN,RMSMAX,PI
      REAL MINV,TLTAXIS,POWER(*)
      REAL,ALLOCATABLE :: VARP2(:)
      PARAMETER (PI=3.1415926535898)
C
      WRITE(*,1100)
1100  FORMAT(/,' SEARCHING FOR TILT AXIS...'/)
C
      MINV=1.0E30
C
C    search for tilt axis in 2 deg increments
C
      ALLOCATE(VARP2(179),STAT=IERR)
      IF (IERR.NE.0)
     +  STOP ' ERROR: Memory allocation failed in FIND_TAXIS'
C

C     RSS parallel loop causing seg faults !$OMP PARALLEL DO
      DO 100 I=1,179,2
        CALL FIND_TAXIS_S(AIN,NXYZ,RMSMIN,RMSMAX,
     +              JXYZ,POWER,KXYZ,NR,VARP2,I,0)
100   CONTINUE
C
C    find tilt axis with lowest variance
C
      DO 101 I=1,179,2
        IF (VARP2(I).LT.MINV) THEN
          MINV=VARP2(I)
          TLTAXIS=I
          K=I
          WRITE(*,*)' Angle between tilt axis and X-axis = ',
     +              TLTAXIS
        ENDIF
101   CONTINUE
      CALL FIND_TAXIS_S(AIN,NXYZ,RMSMIN,RMSMAX,
     +      JXYZ,POWER,KXYZ,NR,VARP2,K,1)
C
C    convert TLTAXIS into radiants
C
      TLTAXIS=TLTAXIS/180.0*PI
      DEALLOCATE(VARP2)
C
      RETURN
      END
C
C**************************************************************************
      SUBROUTINE FIND_TAXIS_S(AIN,NXYZ,RMSMIN,RMSMAX,
     +              JXYZ,POWER,KXYZ,NR,VARP2,I,STORE)
C**************************************************************************
C    Finds tilt axis by calculating power spectra along lines
C    across input image AIN to minimize variance between spectra.
C    JXYZ gives tile size
C**************************************************************************
C
      IMPLICIT NONE
C
      INTEGER NXYZ(*),JXYZ(*),NX,ID,I,J,K,IS,NY,IX,IY,JJ
      INTEGER CX,CY,CNT,KXYZ(*),L,CNT2,IR2,IRL2,NR,CNT3
      INTEGER STORE,IERR,CNT4
      REAL AIN(*),RMSMIN,RMSMAX,MEAN,RMS,PI,ALPHA
      REAL MINV,VARP,A2,POWER(*)
      REAL P,VARP2(*)
      PARAMETER (PI=3.1415926535898)
      REAL,ALLOCATABLE :: ABOX(:),BUF1(:),BUF2(:),OUT(:)
      COMPLEX,ALLOCATABLE :: CBOXS(:)
C
C    calculate number of tiles in X, Y
C
      NX=NXYZ(1)/JXYZ(1)
      NY=NXYZ(2)/JXYZ(2)
C
C    set NX to minimum of NX and NY to avoid lines
C    across image which are too long for some directions
C
      IF (NX.GT.NY) NX=NY
C
C    calculate center of image
C
      CX=NXYZ(1)/2
      CY=NXYZ(2)/2
C
C    restrict resolution of power spectra to half
C    the Nyquist frequency
C
      IRL2=(JXYZ(1)/2)**2
C
      ALPHA=I/180.0*PI
C
      ID=KXYZ(1)*KXYZ(2)
      ALLOCATE(ABOX(2*ID),BUF1(ID),BUF2(ID),OUT(ID),CBOXS(JXYZ(2)),
     +         STAT=IERR)
      IF (IERR.NE.0) THEN
        WRITE(*,*) ' ERROR: Memory allocation failed in FIND_TAXIS_S'
        STOP ' Try reducing the tile size'
      ENDIF
C
C    calculate power spectra along 2*NR+1 parallel lines (JJ=-NR...+NR).
C    lines are offset so that circles cut out by BOXIMAGE2 do not
C    overlap.
C
      CNT3=0
      CNT4=0
      DO 11 J=1,JXYZ(1)/2*JXYZ(2)
        OUT(J)=0.0
11    CONTINUE
C
      VARP2(I)=0.0
      DO 21 JJ=-NR,NR
C
C    set buffers to zero
C
      DO 10 J=1,JXYZ(1)/2*JXYZ(2)
        BUF1(J)=0.0
        BUF2(J)=0.0
10    CONTINUE
      CNT=0
C
      A2=ALPHA+J*60.0/180.0*PI
C
      DO 20 J=1,NX-ABS(JJ)
C
C    calculate upper left corner of tile
C
          IX=CX+COS(ALPHA)*(J-NX/2)*JXYZ(1)
     +       +ABS(JJ)*COS(A2)*JXYZ(1)-JXYZ(1)/2
          IY=CY+SIN(ALPHA)*(J-NX/2)*JXYZ(1)
     +       +ABS(JJ)*SIN(A2)*JXYZ(1)-JXYZ(2)/2
C
C    cut out tile and put into ABOX
C    the corners of the tile are set to the mean of the masked
C    circular area in center of tile
C
          CALL BOXIMG2(AIN,NXYZ,ABOX,JXYZ,IX,IY,MEAN,RMS)
C
C    exclude tile if STD is too small or too large
C
          IF ((RMS.LT.RMSMAX).AND.(RMS.GT.RMSMIN)) THEN
          CNT=CNT+1
          CNT3=CNT3+1
C
C    calculate power spectrum and accumulate sums in BUF1, BUF2
C
            CALL RLFT3(ABOX,CBOXS,JXYZ(1),JXYZ(2),1,1)
          DO 40 L=1,JXYZ(2)
              DO 41 K=1,JXYZ(1)/2
              ID=(K+JXYZ(1)/2*(L-1))*2
              IS=K+KXYZ(1)*(L-1)
              P=(ABOX(ID-1)**2+ABOX(ID)**2)/RMS**2
              BUF2(IS)=BUF2(IS)+P**2
              BUF1(IS)=BUF1(IS)+P
              OUT(IS)=OUT(IS)+P
41        CONTINUE
C
C    if KXYZ(1) larger than JXYZ(1)/2 then need to store extra
C    line to get power spectrum with even dimensions
C
            IF (KXYZ(1).GT.JXYZ(1)/2) THEN
              P=CABS(CBOXS(L)/RMS)**2
              BUF2(IS+1)=BUF2(IS+1)+P**2
              BUF1(IS+1)=BUF1(IS+1)+P
              OUT(IS+1)=OUT(IS+1)+P
            ENDIF
40      CONTINUE
          ENDIF
C
20    CONTINUE
C
      VARP=0.0
      CNT2=0
C
C    calculate total variance of power spectra
C
      IF (CNT.GT.1) THEN
        DO 30 L=1,KXYZ(2)
          DO 30 K=1,KXYZ(1)
            IS=K+KXYZ(1)*(L-1)
            BUF1(IS)=BUF1(IS)/CNT
            BUF2(IS)=BUF2(IS)/CNT
            BUF2(IS)=BUF2(IS)-BUF1(IS)**2
            IR2=(L-1)**2+(K-1)**2
C
C    exclude origin of power spectrum and any frequency larger
C    than IRL2
C
            IF ((L.GT.5).AND.(K.GT.5).AND.(IR2.LT.IRL2)) THEN
              VARP=VARP+BUF2(IS)
              CNT2=CNT2+1
            ENDIF
C
30      CONTINUE
        VARP=VARP/CNT2
        VARP2(I)=VARP2(I)+VARP
        CNT4=CNT4+1
      ENDIF
C
21    CONTINUE
      VARP2(I)=VARP2(I)/CNT4
C
      IF (STORE.NE.0) THEN
C
C    store power spectrum for new minimum
C       take sqrt to reduce slope of background (makes background
C       subtraction in MSMOOTH more accurate, background-subtracted
C       power spectrum will be squared after background subtraction)
C
        DO 31 K=1,KXYZ(2)*KXYZ(1)
          POWER(K)=SQRT(OUT(K)/CNT3)
31      CONTINUE
C
      ENDIF
C
      DEALLOCATE(ABOX,BUF1,BUF2,OUT,CBOXS)
C
      RETURN
      END
C
C**************************************************************************
      SUBROUTINE BOXIMG2(AIN,NXYZ,ABOX,JXYZ,IX,IY,MEAN,RMS)
C**************************************************************************
C    Cuts out an area of size JXYZ from array AIN. IX, IY are
C    upper left corner of boxed area.
C    Only central circular area is kept; pixels outside this area
C    are set to mean of pixels inside the area.
C**************************************************************************
      IMPLICIT NONE
C
      INTEGER I,J,NXYZ(3),JXYZ(3),IX,IY,ID,IDB,II,JJ
      INTEGER RAD2,D2,N
      REAL AIN(*),ABOX(*),MEAN,RMS,M1,M2,M3,M4
C
      MEAN=0.0
      M1=0.0
      M2=0.0
      M3=0.0
      M4=0.0
      N=0
      D2=JXYZ(1)**2
      DO 10 J=1,JXYZ(2)
        DO 10 I=1,JXYZ(1)
          II=I+IX-1
          JJ=J+IY-1
          IF (JJ.GT.NXYZ(2)) GOTO 99
          IF (II.GT.NXYZ(1)) GOTO 99
          IF (JJ.LT.1) GOTO 99
          IF (II.LT.1) GOTO 99
          ID=II+NXYZ(1)*(JJ-1)
          IDB=I+JXYZ(1)*(J-1)
          ABOX(IDB)=AIN(ID)
          RAD2=(J-JXYZ(2)/2-1)**2+(I-JXYZ(1)/2-1)**2
          IF (RAD2.LE.D2) THEN
            MEAN=MEAN+ABOX(IDB)
            N=N+1
          ENDIF
          IF (I.EQ.1) M1=M1+ABOX(IDB)
          IF (I.EQ.JXYZ(1)) M2=M2+ABOX(IDB)
          IF (J.EQ.1) M3=M3+ABOX(IDB)
          IF (J.EQ.JXYZ(2)) M4=M4+ABOX(IDB)
10    CONTINUE
      MEAN=MEAN/N
      M1=M1/JXYZ(2)
      M2=M2/JXYZ(2)
      M3=M3/JXYZ(1)
      M4=M4/JXYZ(1)
      DO 40 J=1,JXYZ(2)
        DO 40 I=1,JXYZ(1)
          IDB=I+JXYZ(1)*(J-1)
          ABOX(IDB)=ABOX(IDB)-(M1+(M2-M1)/(JXYZ(1)-1)*(I-1))
     +                   -(M3+(M4-M3)/(JXYZ(2)-1)*(J-1))+MEAN
40    CONTINUE
C
      DO 30 J=1,JXYZ(2)
        DO 30 I=1,JXYZ(1)
          RAD2=(J-JXYZ(2)/2-1)**2+(I-JXYZ(1)/2-1)**2
          IF (RAD2.GT.D2) THEN
            IDB=I+JXYZ(1)*(J-1) 
            ABOX(IDB)=0.0
          ENDIF
30    CONTINUE
C
      RMS=0.0
      DO 20 I=1,JXYZ(1)*JXYZ(2)
          RMS=RMS+ABOX(I)**2
20    CONTINUE
      RMS=SQRT(RMS/JXYZ(1)/JXYZ(2))
      RETURN
C
99    CONTINUE
      RMS=0.0
      MEAN=0.0
      RETURN
      END
C
C**************************************************************************
      SUBROUTINE BOXIMG(AIN,NXYZ,ABOX,JXYZ,IX,IY,MEAN,RMS)
C**************************************************************************
C    Cuts out an area of size JXYZ from array AIN. IX, IY are
C    upper left corner of boxed area.
C**************************************************************************
C
      IMPLICIT NONE
C
      INTEGER I,J,NXYZ(3),JXYZ(3),IX,IY,ID,IDB,II,JJ
      REAL AIN(*),ABOX(*),MEAN,RMS,M1,M2,M3,M4
C
      MEAN=0.0
      M1=0.0
      M2=0.0
      M3=0.0
      M4=0.0
      DO 10 J=1,JXYZ(2)
        DO 10 I=1,JXYZ(1)
            II=I+IX-1
            JJ=J+IY-1
          ID=II+NXYZ(1)*(JJ-1)
          IDB=I+JXYZ(1)*(J-1)
          ABOX(IDB)=AIN(ID)
            MEAN=MEAN+ABOX(IDB)
          IF (I.EQ.1) M1=M1+ABOX(IDB)
          IF (I.EQ.JXYZ(1)) M2=M2+ABOX(IDB)
          IF (J.EQ.1) M3=M3+ABOX(IDB)
          IF (J.EQ.JXYZ(2)) M4=M4+ABOX(IDB)
10    CONTINUE
      MEAN=MEAN/JXYZ(1)/JXYZ(2)
      M1=M1/JXYZ(2)
      M2=M2/JXYZ(2)
      M3=M3/JXYZ(1)
      M4=M4/JXYZ(1)
      RMS=0.0
      DO 20 I=1,JXYZ(1)*JXYZ(2)
          RMS=RMS+(ABOX(I)-MEAN)**2
20    CONTINUE
      RMS=SQRT(RMS/JXYZ(1)/JXYZ(2))
      DO 30 J=1,JXYZ(2)
        DO 30 I=1,JXYZ(1)
          IDB=I+JXYZ(1)*(J-1)
          ABOX(IDB)=ABOX(IDB)-(M1+(M2-M1)/(JXYZ(1)-1)*(I-1))
     +                   -(M3+(M4-M3)/(JXYZ(2)-1)*(J-1))+MEAN
30    CONTINUE
C
      RETURN
      END
C
C**************************************************************************
      SUBROUTINE BOXIMGP(AIN,NXYZ,ABOX,JXYZ,IX,IY,MEAN,RMS)
C**************************************************************************
C    Cuts out an area of size JXYZ from array AIN. IX, IY are
C    upper left corner of boxed area.
C**************************************************************************
C
      IMPLICIT NONE
C
      INTEGER I,J,NXYZ(3),JXYZ(3),IX,IY,ID,IDB,II,JJ
      REAL AIN(*),ABOX(*),MEAN,RMS
C
      MEAN=0.0
      DO 10 J=1,JXYZ(2)
        DO 10 I=1,JXYZ(1)
            II=I+IX-1
            JJ=J+IY-1
          ID=II+NXYZ(1)*(JJ-1)
          IDB=I+JXYZ(1)*(J-1)
          ABOX(IDB)=AIN(ID)
            MEAN=MEAN+ABOX(IDB)
10    CONTINUE
      MEAN=MEAN/JXYZ(1)/JXYZ(2)
      RMS=0.0
      DO 20 I=1,JXYZ(1)*JXYZ(2)
          RMS=RMS+(ABOX(I)-MEAN)**2
20    CONTINUE
      RMS=SQRT(RMS/JXYZ(1)/JXYZ(2))
C
      RETURN
      END
C
C**************************************************************************
      SUBROUTINE SEARCH_CTF_S(CS,WL,WGH1,WGH2,THETATR,RMIN2,
     +          RMAX2,AIN,NXYZ,DF1,DF2,ANG,FSTEP,SUMS,HW,
     +            SUMMAX,DFMID1S,DFMID2S,ANGASTS,I,J,K,I1,I2)
C**************************************************************************
      IMPLICIT NONE
C
      INTEGER I,J,K,NXYZ(3),I1,I2,ID
      REAL CS,WL,WGH1,WGH2,THETATR,DF1(*),DF2(*),ANG(*)
      REAL RMIN2,RMAX2,SUMS(*),AIN(*),SUMMAX,FSTEP
      REAL DFMID1S,DFMID2S,ANGASTS,HW,PI,SIG2
      PARAMETER (PI=3.1415926535898)
C
      ID=I-I1+1+(I2-I1+1)*(J-I1)
      DF1(ID)=FSTEP*I
      DF2(ID)=FSTEP*J
      ANG(ID)=22.5*K
      ANG(ID)=ANG(ID)/180.0*PI
      CALL EVALCTF(CS,WL,WGH1,WGH2,DF1(ID),DF2(ID),ANG(ID),
     +        THETATR,HW,AIN,NXYZ,RMIN2,RMAX2,SUMS(ID),SIG2)
C
      RETURN
      END      
C
C**************************************************************************
      SUBROUTINE EVALCTF(CS,WL,WGH1,WGH2,DFMID1,DFMID2,ANGAST,
     +        THETATR,HW,AIN,NXYZ,RMIN2,RMAX2,CCTF,SIG2)
C**************************************************************************
C    Calculates the correlation coefficient between an input power
C    spectrum AIN and the calculated squared CTF.
C    HW is a parameter in a Gaussian filter to attenuate the calculated
C    CTF towards higher resolution to obtain a better fit to the 
C    observed input spectrum. The comparison is done between the squared
C    resolution limits RMIN2 and RMAX2.
C**************************************************************************
      IMPLICIT NONE
C
      INTEGER L,LL,M,MM,NXYZ(*),ID
      REAL CS,WL,WGH1,WGH2,DFMID1,DFMID2,ANGAST,THETATR,SUM1
      REAL SUM,AIN(*),RES2,RMIN2,RMAX2,CTF,CTFV,HW,SUM2,CCTF
      REAL SIG2,A
C
      SUM=0.0
      SUM1=0.0
      SUM2=0.0
      DO 20 L=1,NXYZ(1)/2
          LL=L-1
          DO 20 M=1,NXYZ(2)
            MM=M-1
            IF (MM.GT.NXYZ(2)/2) MM=MM-NXYZ(2)
            RES2=(REAL(LL)/NXYZ(1))**2+(REAL(MM)/NXYZ(2))**2
            IF ((RES2.LE.RMAX2).AND.(RES2.GE.RMIN2)) THEN
              CTFV=CTF(CS,WL,WGH1,WGH2,DFMID1,DFMID2,
     +             ANGAST,THETATR,LL,MM)
              ID=L+NXYZ(1)/2*(M-1)
            CTFV=CTFV**2
              SUM=SUM+AIN(ID)*CTFV*EXP(HW*RES2)
            SUM1=SUM1+CTFV**2
            SUM2=SUM2+AIN(ID)**2*EXP(2.0*HW*RES2)
            ENDIF
20    CONTINUE
      A=SUM/SUM1
      SIG2=((SUM2/A+SUM1*A)/SUM-2.0)*2/NXYZ(1)/NXYZ(2)
      CCTF=SUM/SQRT(SUM1*SUM2)
C
      RETURN
      END
C
C**************************************************************************
      SUBROUTINE REFINE_CTF(DFMID1,DFMID2,ANGAST,AIN,ABOX)
C**************************************************************************
C    Refines defocus and astigmatism using Powell minimizer
C    VA04A
C**************************************************************************
      IMPLICIT NONE
C
      INTEGER NCYCLS
      PARAMETER (NCYCLS=50)
      REAL DFMID1,DFMID2,ANGAST,XPAR(3),EPAR(3),RF
      REAL ESCALE,PI,AIN(*),ABOX(*)
      PARAMETER (PI=3.1415926535898)
      DATA EPAR/100.0,100.0,0.5/
      DATA ESCALE/100.0/
      EXTERNAL CALCFX
C
      WRITE(*,1000)
1000  FORMAT(/,' REFINING CTF PARAMETERS...'/,
     +       /,'      DFMID1      DFMID2      ANGAST          CC'/)
C
      XPAR(1)=DFMID1
      XPAR(2)=DFMID2
      XPAR(3)=ANGAST
      IF (XPAR(1).EQ.XPAR(2)) XPAR(1)=XPAR(1)+1.0
      CALL VA04A(XPAR,EPAR,3,RF,ESCALE,0,1,NCYCLS,AIN,ABOX,CALCFX)
      DFMID1=XPAR(1)
      DFMID2=XPAR(2)
      ANGAST=XPAR(3)
      WRITE(*,1100)DFMID1,DFMID2,ANGAST/PI*180.0,-RF
1100  FORMAT(3F12.2,F12.5,'  Refined Values at Center')
C
      RETURN
      END
C
C******************************************************************************
C
      SUBROUTINE CALCFXB(NX,XPAR,RF,AIN,ABOX)
C
C     CALCULATES NEW VALUE FOR F TO INPUT TO SUBROUTINE VA04A
C
C**************************************************************************
C    Called by VA04A to refine defocus, astigmatism, tilt axis and
C    tilt angle
C**************************************************************************
      IMPLICIT NONE
C
      INTEGER JXYZ(3),NXYZ(3),NX,IXMX,IXBMX
      REAL WGH1,WGH2,THETATR,EVAL_TILT
      REAL AIN(*),ABOX(*),PI,WL
      REAL PSIZE,CS,XPAR(*),HW
      REAL RMIN2,RMAX2,RF,TAR,TILTA,TILTR,SIG2
      PARAMETER (PI=3.1415926535898)
      COMMON/FUNC/CS,WL,WGH1,WGH2,THETATR,RMIN2,RMAX2,JXYZ,HW
      COMMON/FUNCB/NXYZ,PSIZE,TILTA,TILTR,SIG2
C
      RF=-EVAL_TILT(JXYZ,NXYZ,AIN,CS,WL,WGH1,WGH2,
     +  THETATR,XPAR(3),XPAR(4),XPAR(5),XPAR(1),PSIZE,RMIN2,
     +  RMAX2,HW,XPAR(2))-TAR(SIG2,XPAR(2),TILTA,TILTR)
      WRITE(*,1000)XPAR(3),XPAR(4),XPAR(5)/PI*180.0
     +          ,XPAR(1)/pi*180.0,XPAR(2)/PI*180.0,-RF
1000  FORMAT(5F12.2,F12.5)
C
      RETURN
      END
C
C**************************************************************************
      REAL FUNCTION TAR(SIG2,TA,TILTA,TILTR)
C**************************************************************************
C       Calculates probability distribution for tilt angle
C**************************************************************************
      IMPLICIT NONE
C
      REAL TA,TILTA,TILTR,SIG2
C
      TAR=-SIG2*(ABS(TA)-TILTA)**2/2.0/TILTR**2
C
      RETURN
      END
C
C**************************************************************************
C
      SUBROUTINE CALCFX(NX,XPAR,RF,AIN,ABOX)
C
C     CALCULATES NEW VALUE FOR F TO INPUT TO SUBROUTINE VA04A
C
C**************************************************************************
C    Called by VA04A to refine defocus and astigmatism
C**************************************************************************
      IMPLICIT NONE
C
      INTEGER JXYZ(3),NX,IXBMX
      REAL CS,WL,WGH1,WGH2,THETATR,CCTF
      REAL RMIN2,RMAX2,AIN(*),ABOX(*)
      REAL HW,PI,XPAR(*),RF,SIG2
      PARAMETER (PI=3.1415926535898)
      COMMON/FUNC/CS,WL,WGH1,WGH2,THETATR,RMIN2,RMAX2,JXYZ,HW
C
      CALL EVALCTF(CS,WL,WGH1,WGH2,XPAR(1),XPAR(2),XPAR(3),
     +            THETATR,HW,ABOX,JXYZ,RMIN2,RMAX2,CCTF,SIG2)
      RF=-CCTF
C
      RETURN
      END
C
C**************************************************************************
      SUBROUTINE VA04A(X,E,N,F,ESCALE,IPRINT,ICON,MAXIT,
     +                 AIN,ABOX,FX)
C**************************************************************************
C  STANDARD FORTRAN 66 (A VERIFIED PFORT SUBROUTINE)
      DIMENSION W(40),X(1),E(1),AIN(*),ABOX(*)
      EXTERNAL FX
C    W[N*(N+3)]
      DDMAG=0.1*ESCALE
      SCER=0.05/ESCALE
      JJ=N*N+N
      JJJ=JJ+N
      K=N+1
      NFCC=1
      IND=1
      INN=1
      DO 1 I=1,N
      DO 2 J=1,N
      W(K)=0.
      IF(I-J)4,3,4
    3 W(K)=ABS(E(I))
      W(I)=ESCALE
    4 K=K+1
    2 CONTINUE
    1 CONTINUE
      ITERC=1
      ISGRAD=2
      CALL FX(N,X,F,AIN,ABOX)
      FKEEP=ABS(F)+ABS(F)
    5 ITONE=1
      FP=F
      SUM=0.
      IXP=JJ
      DO 6 I=1,N
      IXP=IXP+1
      W(IXP)=X(I)
    6 CONTINUE
      IDIRN=N+1
      ILINE=1
    7 DMAX=W(ILINE)
      DACC=DMAX*SCER
      DMAG=AMIN1(DDMAG,0.1*DMAX)
      DMAG=AMAX1(DMAG,20.*DACC)
      DDMAX=10.*DMAG
      GO TO (70,70,71),ITONE
   70 DL=0.
      D=DMAG
      FPREV=F
      IS=5
      FA=F
      DA=DL
    8 DD=D-DL
      DL=D
   58 K=IDIRN
      DO 9 I=1,N
      X(I)=X(I)+DD*W(K)
      K=K+1
    9 CONTINUE
      CALL FX(N,X,F,AIN,ABOX)
C
      NFCC=NFCC+1
      GO TO (10,11,12,13,14,96),IS
   14 IF(F-FA)15,16,24
   16 IF (ABS(D)-DMAX) 17,17,18
   17 D=D+D
      GO TO 8
   18 WRITE(6,19)
   19 FORMAT(5X,44HVA04A MAXIMUM CHANGE DOES NOT ALTER FUNCTION)
      GO TO 20
   15 FB=F
      DB=D
      GO TO 21
   24 FB=FA
      DB=DA
      FA=F
      DA=D
   21 GO TO (83,23),ISGRAD
   23 D=DB+DB-DA
      IS=1
      GO TO 8
   83 D=0.5*(DA+DB-(FA-FB)/(DA-DB))
      IS=4
      IF((DA-D)*(D-DB))25,8,8
   25 IS=1
      IF(ABS(D-DB)-DDMAX)8,8,26
   26 D=DB+SIGN(DDMAX,DB-DA)
      IS=1
      DDMAX=DDMAX+DDMAX
      DDMAG=DDMAG+DDMAG
      IF(DDMAX-DMAX)8,8,27
   27 DDMAX=DMAX
      GO TO 8
   13 IF(F-FA)28,23,23
   28 FC=FB
      DC=DB
   29 FB=F
      DB=D
      GO TO 30
   12 IF(F-FB)28,28,31
   31 FA=F
      DA=D
      GO TO 30
   11 IF(F-FB)32,10,10
   32 FA=FB
      DA=DB
      GO TO 29
   71 DL=1.
      DDMAX=5.
      FA=FP
      DA=-1.
      FB=FHOLD
      DB=0.
      D=1.
   10 FC=F
      DC=D
   30 A=(DB-DC)*(FA-FC)
      B=(DC-DA)*(FB-FC)
      IF((A+B)*(DA-DC))33,33,34
   33 FA=FB
      DA=DB
      FB=FC
      DB=DC
      GO TO 26
   34 D=0.5*(A*(DB+DC)+B*(DA+DC))/(A+B)
      DI=DB
      FI=FB
      IF(FB-FC)44,44,43
   43 DI=DC
      FI=FC
   44 GO TO (86,86,85),ITONE
   85 ITONE=2
      GO TO 45
   86 IF (ABS(D-DI)-DACC) 41,41,93
   93 IF (ABS(D-DI)-0.03*ABS(D)) 41,41,45
   45 IF ((DA-DC)*(DC-D)) 47,46,46
   46 FA=FB
      DA=DB
      FB=FC
      DB=DC
      GO TO 25
   47 IS=2
      IF ((DB-D)*(D-DC)) 48,8,8
   48 IS=3
      GO TO 8
   41 F=FI
      D=DI-DL
      DD=SQRT((DC-DB)*(DC-DA)*(DA-DB)/(A+B))
      DO 49 I=1,N
      X(I)=X(I)+D*W(IDIRN)
      W(IDIRN)=DD*W(IDIRN)
      IDIRN=IDIRN+1
   49 CONTINUE
      IF (DD.EQ.0.0) DD=1E-10
      W(ILINE)=W(ILINE)/DD
      ILINE=ILINE+1
      IF(IPRINT-1)51,50,51
   50 WRITE(6,52) ITERC,NFCC,F,(X(I),I=1,N)
   52 FORMAT (/1X,9HITERATION,I5,I15,16H FUNCTION VALUES,
     110X,3HF =,E21.14/(5E24.14))
      GO TO(51,53),IPRINT
   51 GO TO (55,38),ITONE
   55 IF (FPREV-F-SUM) 94,95,95
   95 SUM=FPREV-F
      JIL=ILINE
   94 IF (IDIRN-JJ) 7,7,84
   84 GO TO (92,72),IND
   92 FHOLD=F
      IS=6
      IXP=JJ
      DO 59 I=1,N
      IXP=IXP+1
      W(IXP)=X(I)-W(IXP)
   59 CONTINUE
      DD=1.
      GO TO 58
   96 GO TO (112,87),IND
  112 IF (FP-F) 37,37,91
   91 D=2.*(FP+F-2.*FHOLD)/(FP-F)**2
      IF (D*(FP-FHOLD-SUM)**2-SUM) 87,37,37
   87 J=JIL*N+1
      IF (J-JJ) 60,60,61
   60 DO 62 I=J,JJ
      K=I-N
      W(K)=W(I)
   62 CONTINUE
      DO 97 I=JIL,N
      W(I-1)=W(I)
   97 CONTINUE
   61 IDIRN=IDIRN-N
      ITONE=3
      K=IDIRN
      IXP=JJ
      AAA=0.
      DO 65 I=1,N
      IXP=IXP+1
      W(K)=W(IXP)
      IF (AAA-ABS(W(K)/E(I))) 66,67,67
   66 AAA=ABS(W(K)/E(I))
   67 K=K+1
   65 CONTINUE
      DDMAG=1.
      IF (AAA.EQ.0.0) AAA=1E-10
      W(N)=ESCALE/AAA
      ILINE=N
      GO TO 7
   37 IXP=JJ
      AAA=0.
      F=FHOLD
      DO 99 I=1,N
      IXP=IXP+1
      X(I)=X(I)-W(IXP)
      IF (AAA*ABS(E(I))-ABS(W(IXP))) 98,99,99
   98 AAA=ABS(W(IXP)/E(I))
   99 CONTINUE
      GO TO 72
   38 AAA=AAA*(1.+DI)
      GO TO (72,106),IND
   72 IF (IPRINT-2) 53,50,50
   53 GO TO (109,88),IND
  109 IF (AAA-0.1) 89,89,76
   89 GO TO (20,116),ICON
  116 IND=2
      GO TO (100,101),INN
  100 INN=2
      K=JJJ
      DO 102 I=1,N
      K=K+1
      W(K)=X(I)
      X(I)=X(I)+10.*E(I)
  102 CONTINUE
      FKEEP=F
      CALL FX(N,X,F,AIN,ABOX)
      NFCC=NFCC+1
      DDMAG=0.
      GO TO 108
   76 IF (F-FP) 35,78,78
   78 WRITE(6,80)
   80 FORMAT (5X,37HVA04A ACCURACY LIMITED BY ERRORS IN F)
      GO TO 20
   88 IND=1
   35 TMP=FP-F
      IF (TMP.GT.0.0) THEN
      DDMAG=0.4*SQRT(TMP)
      ELSE
      DDMAG=0.0
      ENDIF
      ISGRAD=1
  108 ITERC=ITERC+1
      IF (ITERC-MAXIT) 5,5,81
81    CONTINUE
C   81 WRITE(6,82) MAXIT
   82 FORMAT(I5,30H ITERATIONS COMPLETED BY VA04A)
      IF (F-FKEEP) 20,20,110
  110 F=FKEEP
      DO 111 I=1,N
      JJJ=JJJ+1
      X(I)=W(JJJ)
  111 CONTINUE
      GO TO 20
  101 JIL=1
      FP=FKEEP
      IF (F-FKEEP) 105,78,104
  104 JIL=2
      FP=F
      F=FKEEP
  105 IXP=JJ
      DO 113 I=1,N
      IXP=IXP+1
      K=IXP+N
      GO TO (114,115),JIL
  114 W(IXP)=W(K)
      GO TO 113
  115 W(IXP)=X(I)
      X(I)=W(K)
  113 CONTINUE
      JIL=2
      GO TO 92
  106 IF (AAA-0.1) 20,20,107
   20 RETURN
  107 INN=1
      GO TO 35
      END
