      subroutine ccf_fft (fft1, fft2, cc_map, idim1, idim2, res_min,
     $     res_max)
      
C     Calculate correlation coefficient map between image and
C     projection. The code looks a little more complicated than it
C     actually is owing to resolution tests and logic to minimize the
C     number of FTs that must be evaluated, but basically involves the
C     following steps
C     
C     (1) Multiply conjg(FFT1) by FFT2, element-by-element
C     (2) Calculate inverse transform of product
C     (3) Assign real part of inverse transform to cc_map
C

C     This routine has been completely overhauled and reorganized so
C     that logic is easier to follow. Complex arrays that had been
C     represented as real arrays with the leading dimension doubled are
C     now explicitly declared to be complex. Fortran 90 array syntax has
C     been used to condense code. For example,
C     
C     do i = 1,n        --->
C        x(i) = y(i,j)  --->   x(:) = y(:,j)
C     enddo             ---> 

      implicit none
      
      integer, intent(in)  :: idim1, idim2
      complex, intent(in)  :: fft1(idim1,idim2), fft2(idim1,idim2)
      real*4,  intent(in)  :: res_min, res_max
      real*4,  intent(out) :: cc_map(idim1,idim2)

      integer  i, j, k, iy, iymax, alloc_stat
      real     work, rsq, ysq, dimdim, div, res_minsq, res_maxsq
      logical  restest
      complex, allocatable :: col_array(:), row_array(:), tmp(:,:)

      allocate(tmp(idim1,idim2), row_array(idim2), col_array(idim1), 
     $     stat = alloc_stat)
      
      if(alloc_stat > 0) then
         write(*,*) 'memory allocation failed in ccf_fft'
         stop
      endif

C     Calculate parameters for resolution tests
      iymax = (idim2/2)+1
      if(res_min <= 0.0 .and. res_max <= 0.0) then
         restest = .false.
      else
         restest = .true.
         if(res_max < float(iymax)) iymax = nint(res_max)
         res_minsq = res_min*res_min
         res_maxsq = res_max*res_max
      endif

C     Do transform along first dimension for columns 1 through
C     (idim2/2)+1. Elements of the temporary workspace vector (and their
C     Friedel mates) corresponding to resolutions outside of range
C     res_min to res_max are set to zero

      iy     = 0
      dimdim = float(idim1)*float(idim2)
      div    = float(idim2/2) + 1.0

      do j = 1, (idim2/2)+1
         iy = iy+1
         if(iy <= iymax .and. real(iy) <= res_max) then
            col_array(:) = conjg(fft1(:,j))*fft2(:,j)/dimdim
            if(restest) then
               ysq = (iy-1)*(iy-1)
               do k = 1, (idim1/2)+1
                  rsq = ysq + (k-1)*(k-1)
                  if(rsq > res_maxsq .or. rsq < res_minsq) then
                     col_array(k) = (0.0, 0.0)
                     if(k .ne. 1) then
                        col_array(idim1+2-k) = (0.0, 0.0)
                     endif
                  endif
               enddo
            endif
            call fourt (col_array, idim1, 1, -1, 1, work)
            tmp(:,j) = col_array(:)
         else
            tmp(:,j) = (0.0, 0.0)
         endif
      enddo

C     Get the transforms of the remaining columns using Friedel relation
      k = idim2/2
      do j = (idim2/2)+2, idim2
         tmp(:,j) = conjg(tmp(:,k))
         k = k - 1
      enddo

C     Perform transforms along rows and assign real components to cc_map
      do i = 1, idim1
         row_array(:) = tmp(i,:)
         call fourt(row_array, idim2, 1, -1, 1, work)
         cc_map(i,:) = real(row_array(:))
      enddo
      
      deallocate(tmp, col_array, row_array)
      return
      end subroutine ccf_fft
