      logical function good(x)
C
C     Logical function good() returns the value TRUE if the orientation
C     specified by the direction cosine vector falls within the
C     icosahedral asymmetric unit (ASU). The outer surface of the ASU is
C     defined by the triangle formed from two adjacent points of
C     fivefold symmetry (90, +/- 31.7) and the nearest three-fold axis
C     (69.0, 0) The function does a rough check first on the direction
C     cosines of the view and then calculates the spherical measure of
C     the largest triangle formed by the test point and the bounding
C     symmetry axes. If this is larger than that of the asymmetric unit,
C     the view is outside the unit.
C
C     Author: Steve Fuller 10/8/88
C
C     Modified: Robert Sinkovits 6/28/06 - general cleanup and bringing
C     in line with Fortran 90 standard
C     Modified: Robert Sinkovits 12/8/06 - use save attribute so that
C     constants are calculated only first time function is called.

      implicit none

      real*4, intent(in) :: x(3)

      real*4 toler/1.0e-6/
      real*4 a3x, a5px, a5x, d3x, d5x, d5px, area_x
      real*4 sur,excess,a,b,c
      real*8 deg_to_rad, rad_to_deg, tau, arg
      real*8 theta_three, phi_five, dist_35, dist_55
      integer i

      real, save :: three_fold(3) ! direction cosines for three fold
      real, save :: five_fold(3)  ! direction cosines for five fold
      real, save :: ang_35,ang_55 ! angular distances btw 3fold and 5 folds
      real, save :: area_std      ! area of ASU in spherical measure
      logical, save :: first_time = .true.

C     Arithmetic function definitions
      sur(a,b,c) = (a+b+c)/2.0
      excess(a,b,c) =
     1     4.0 * atan(sqrt(abs(tan(sur(a,b,c)/2.)
     1     * tan((sur(a,b,c)-a)/2.0)
     1     * tan((sur(a,b,c)-b)/2.0)
     1     * tan((sur(a,b,c)-c)/2.0) ) )  )
      
      good = .false.

      if(first_time) then
         first_time = .false.

         deg_to_rad = acos(-1.0d0)/180.0d0
         rad_to_deg = 1.0d0/deg_to_rad

         tau = (1.+sqrt(5.0d0))/2.0
         arg =  (1.+tau*tau/3.d0-1./(3.d0*tau*tau))/
     $        (2.d0*tau/sqrt(3.0d0))

         theta_three = 90. - rad_to_deg * acos(arg)
         three_fold(1) = real(sin(theta_three*deg_to_rad))
         three_fold(2) = 0.0
         three_fold(3) = real(cos(theta_three*deg_to_rad))

         phi_five = rad_to_deg * atan(1.0/tau)
         five_fold(1) = real(cos(phi_five*deg_to_rad))
         five_fold(2) = real(sin(phi_five*deg_to_rad))
         five_fold(3) = 0.0      

         dist_35 = 0.0
         do i = 1,3
            dist_35=(three_fold(i)-five_fold(i))**2 + dist_35
         end do

         dist_35 = sqrt(dist_35)
         dist_55 = five_fold(2)*2.d0

         ang_55=real(2.0*asin(dist_55/2.))
         ang_35=real(2.0*asin(dist_35/2.))

         area_std  = 4*3.141592653589793/60. ! 1/60th sphere surface
      endif

      if(x(3) < -toler .or.
     $     x(3) > three_fold(3)+toler .or.
     $     abs(x(2)) > five_fold(2)+toler) then
         return
      end if

      d3x = 0.0
      d5x = 0.0
      d5px = 0.0

      do i = 1,3
         d3x=(x(i)-three_fold(i))**2 + d3x
         d5x=(x(i)-five_fold(i))**2 + d5x
         if(i == 2) then
            d5px=(x(i)+five_fold(i))**2 + d5px
         else
            d5px=(x(i)-five_fold(i))**2 + d5px
         end if
      end do
      
      d3x = sqrt(d3x)
      d5x = sqrt(d5x)
      d5px = sqrt(d5px)
      
      a5x=2.0*asin(d5x/2.)
      a5px=2.0*asin(d5px/2.)
      a3x=2.0*asin(d3x/2.)
      
      area_x = excess(a5x,a3x,ang_35) +
     1     excess(a5px,a3x,ang_35) +
     1     excess(a5x,a5px,ang_55) -
     1     min(excess(a5px,a3x,ang_35),excess(a5x,a5px,ang_55), 
     1     excess(a5x,a3x,ang_35))

      good = area_x < (area_std+toler)

      return
      end function good
