      subroutine interpl_3d(F3D, hkl_max, rmin, rmax,
     $     inv_rotmat, F2D, imgmagfactor, fft_ann)

C     Calculate central section (F2D) of the 3D FFT (F3D). Orientation
C     encoded in the inverse rotation matrix inv_rotmat; interpolations
C     evaluated using standard trilinear interpolation algorithm.
C
C     Note that the original version of this routine used the rotation
C     matrix rather than its inverse. The wavevector in the rotated
C     coordinate system was given by
C
C     h' = rotmat(1,1)*tcol + rotmat(1,2)*trow
C     k' = rotmat(2,1)*tcol + rotmat(2,2)*trow
C     l' = rotmat(3,1)*tcol + rotmat(3,2)*trow
C
C     The current version of this routine is based on the inverse
C     rotation matrix and uses the formula
C
C     h' = inv_rotmat(1,1)*tcol + inv_rotmat(2,1)*trow
C     k' = inv_rotmat(1,2)*tcol + inv_rotmat(2,2)*trow
C     l' = inv_rotmat(1,3)*tcol + inv_rotmat(2,3)*trow

      implicit none

      integer*4, intent(in) :: hkl_max, fft_ann
      real*4, intent(in)    :: inv_rotmat(3,3), imgmagfactor, rmin, rmax
      complex, intent(in)   :: F3D( -fft_ann:fft_ann, 
     $     -fft_ann:fft_ann, 0:fft_ann)
      complex, intent(out)  :: F2D( 0:hkl_max, -hkl_max:hkl_max)

      integer*4 row, col, row2, rmaxi, apply_friedel,
     $     ih_minus, ih_plus, ik_minus, ik_plus, il_minus, il_plus
      real*4 rmin2, rmax2, tmp, imgmagfactor2

      real*4 h_prime, l_prime, k_prime,
     +     h_prime_rowcont, l_prime_rowcont, k_prime_rowcont,
     +     del_h_minus, del_h_plus, del_k_minus, del_k_plus,
     +     del_l_minus, del_l_plus, trow, tcol
      complex   t1,t2,t3,t4,t5,t6,t7,t8

      rmaxi = min(hkl_max,int(rmax)+1)
      rmax2 = rmax*rmax
      rmin2 = rmin*rmin
      imgmagfactor2 = imgmagfactor * imgmagfactor

C     Image magnification factor is applied in the FFT domain. At end of
C     interpolation, projection is scaled appropriately.

      Do row = -rmaxi, rmaxi
         trow = row * imgmagfactor
         row2 = row * row
         
         h_prime_rowcont = inv_rotmat(2,1)*trow
         k_prime_rowcont = inv_rotmat(2,2)*trow
         l_prime_rowcont = inv_rotmat(2,3)*trow
         
         Do col = 0, rmaxi
            tcol = col * imgmagfactor

C     Restrict calculation of elements of the central section to those
C     wavevectors that are actually used in the comparison of the image
C     and the model.
            
            tmp = row2 + col*col

            if (tmp > rmax2 .or. tmp < rmin2) then
               F2D(col, row) = 0.
            else
               h_prime = inv_rotmat(1,1)*tcol + h_prime_rowcont
               k_prime = inv_rotmat(1,2)*tcol + k_prime_rowcont
               l_prime = inv_rotmat(1,3)*tcol + l_prime_rowcont
               
C     Friedel condition, F(-k) = F(k)*, was used so that we only need to
C     store half of the 3D FFT. If 3rd index of array is negative, take
C     complex conjugate of the FFT at the negative of the wavevector.

               if(h_prime < 0) then
                  h_prime = -h_prime
                  k_prime = -k_prime
                  l_prime = -l_prime
                  apply_friedel = 1
               else
                  apply_friedel = 0
               endif
               
               ih_minus    = floor(h_prime)
               ih_plus     = 1 + ih_minus
               del_h_minus = h_prime - ih_minus
               del_h_plus  = 1.0 - del_h_minus

               ik_minus    = floor(k_prime)
               ik_plus     = 1 + ik_minus
               del_k_minus = k_prime - ik_minus
               del_k_plus  = 1.0 - del_k_minus

               il_minus    = floor(l_prime)
               il_plus     = 1 + il_minus
               del_l_minus = l_prime - il_minus
               del_l_plus  = 1.0 - del_l_minus

               if(ih_plus > fft_ann)   then
                  ih_plus  = fft_ann
                  ih_minus = fft_ann - 1
               endif

               if(ik_plus > fft_ann)   then
                  ik_plus  = fft_ann
                  ik_minus = fft_ann - 1
               endif

               if(il_plus > fft_ann)   then
                  il_plus  = fft_ann
                  il_minus = fft_ann - 1
               endif

               if(ih_minus < -fft_ann) then
                  ih_minus = -fft_ann
                  ih_plus  = -fft_ann + 1
               endif

               if(ik_minus < -fft_ann) then
                  ik_minus = -fft_ann
                  ik_plus  = -fft_ann + 1
               endif

               if(il_minus < -fft_ann) then
                  il_minus = -fft_ann
                  il_plus  = -fft_ann + 1
               endif

               t1 = F3D(ik_minus,il_minus,ih_minus)
               t3 = F3D(ik_minus,il_plus,ih_minus)
               t5 = F3D(ik_plus,il_minus,ih_minus)
               t7 = F3D(ik_plus,il_plus,ih_minus)
               t2 = F3D(ik_minus,il_minus,ih_plus)
               t4 = F3D(ik_minus,il_plus,ih_plus)
               t6 = F3D(ik_plus,il_minus,ih_plus)
               t8 = F3D(ik_plus,il_plus,ih_plus)

               F2D(col,row) = 
     +              del_k_plus * 
     +              ((t1*del_h_plus + t2*del_h_minus)*del_l_plus +
     +              (t3*del_h_plus + t4*del_h_minus)*del_l_minus)
     +              +
     +              del_k_minus * 
     +              ((t5*del_h_plus + t6*del_h_minus)*del_l_plus +
     +              (t7*del_h_plus + t8*del_h_minus)*del_l_minus)
               
               F2D(col, row) = F2D(col, row) * imgmagfactor2

               if(apply_friedel == 1) then
                  F2D(col,row) = conjg(F2D(col,row))
               endif

            endif
         enddo
      enddo

      end subroutine interpl_3d
