      subroutine pfile_info(inunit)
      
C     Reads in particle parameter file header information
      
      implicit none
      integer, intent(in) :: inunit

      include '../include/sizes.inc'
      include '../include/allprogs.inc'
      include '../include/info.inc'

      include '../include/pifhead.inc'
      type (pifhead_t) imghead
      
      character(len=filename_len) :: fname, tempstr
      character(len=11)  :: subname
      character(len=1)   :: image_type, rw(2)
      integer :: kxyz(3), mxyz(3), jmode
      integer temp_units, total_views, ios, j, loc
      integer skipone, header_flag, imgcount
      real*4 :: dmin, dmax, dmean

      data subname /'pfile_info'/

      skipone = 1
      nparam         = 0    ! # of parameter files
      nimg           = 0    ! # of image files

C     Fortran 90 array syntax
      nimg_per_param = 0    ! # of image files per parameter file
      nptle_per_img  = 0    ! # of particles per image file
      first_index    = 0    ! Index of first record associated with image

      total_views = 0           ! count the number of images to search
      box_out     = -1          ! initialize maximum particle diameter
      pixsiz_out  =  100000.0   ! initialize pixsiz_out
      pixsiz_max  = -100000.0   ! initialize pixsiz_max

      do ! Loop over records in input file

C     Read parameter file name from standard input and check status  
C     ios < 0 : end of file reached, break out of loop
C     ios > 0 : error reading file, terminate program
         read (inunit, '(A)', iostat=ios) fname
         if(ios < 0) then
            exit ! break out of loop
         else if (ios > 0) then
            call error_stop(progname,subname,
     +           '* error reading parameter filename: exit program')
         endif
             
C     Get rid of leading spaces in parameter file name
         fname = trim(adjustl(fname))
            
C     Go to next iteration if blank line encountered
         if (fname(1:1) == ' ') cycle
         if (fname(1:1) == '#') cycle

C     Get rid of trailing comments
         loc = scan(fname,'#')
         if(loc > 0) then
            do j=loc,filename_len
               fname(j:j) = ' '
            enddo
         endif

C     Increment the number of parameter files
         nparam = nparam + 1
         param_fnames(nparam) = fname

C     Make sure that maximum number of parameter files not exceeded
         if(nparam > pfilesmx) then
            call error_stop(progname,subname,
     +           '* too many parameter files: exit program')
         endif
         
C     Open particle parameter file
         open(unit=3, file=fname, iostat=ios)
         if(ios .ne. 0) then
            call error_stop(progname,subname,
     $           '* error opening parameter file: exit program')
         endif

         header_flag = 0
         do ! Loop over records in particle parameter file

            read(3,'(A)', iostat=ios) tempstr
            if(ios < 0) then
               exit  ! break out of loop
            else if (ios > 0) then
               call error_stop(progname,subname,
     +              '* error processing parameter file: exit program')
            endif

            tempstr = adjustl(tempstr)
            if(tempstr(1:1) == ' ') cycle ! skip blank lines

C     Determine whether or not this is an image name
            if(index(tempstr,'.pif') > 0 .or. 
     $           index(tempstr,'.mrc') > 0) then

C     Distinguish between pif and mrc image formats based on extension
               if(index(tempstr,'.pif') > 0) then
                  image_type = 'p'
               else
                  image_type = 'm'
               endif

               header_flag = 1
               nimg = nimg + 1

C     Make sure that maximum number of parameter files not exceeded
               if(nimg > pfilesmx) then
                  call error_stop(progname,subname,
     +                 '* too many image files: exit program')
               endif

               nimg_per_param(nparam) = nimg_per_param(nparam) + 1

C     Get rid of leading whitspace
               img_fnames(nimg) = adjustl(tempstr)
               
C     The image file name may be specified with or without full path
C     information. For example /dir/subdir/image.pif or image.pif. To
C     allow greater flexibility, we first test to see if the image file
C     can be opened using the full path. If this fails, we then trim off
C     the path and try to open the file again.
               
C     Open the image file as specified in parameter file
               open(unit=25, file=img_fnames(nimg), status='old',
     $              form='unformatted', iostat=ios)
               close(25)
               
C     Test result of open. If failed, then try without path info
               if(ios == 0) then
                  write(*,*) 'img file OK: ', trim(img_fnames(nimg))
               else
                  loc = index(img_fnames(nimg), '/', .true.)
                  img_fnames(nimg) = 
     $                 img_fnames(nimg)(loc+1:filename_len)
                  open(unit=25, file=img_fnames(nimg), status='old',
     $                 form='unformatted', iostat=ios)
                  close(25)
                  if(ios == 0) then
                     write(*,*) 'image file path info removed: ',
     $                    trim(img_fnames(nimg))
                  else
                     write(*,*) 'Unable to open image file: ', 
     $                    img_fnames(nimg)
                     call error_stop(progname,subname,
     $                    '* Unable to open image file')
                  endif
               endif

            else if (header_flag == 1) then
               header_flag = 0
               
C     Get pixel size, units, etc. from second line of parameter file
               read(tempstr, *, iostat=ios) pixsizs(nimg), temp_units,
     $              volts(nimg), ampfacs(nimg), delf_majs(nimg),
     $              delf_mins(nimg), ang_majs(nimg), 
     $              cs_coeffs(nimg)
               if(ios .ne. 0) then
                  call error_stop(progname,subname,
     $                 '* err reading param file header: exit program')
               endif
               
C     Make sure that value of units is valid
               if(temp_units < 1 .or. temp_units > 2) then
                  call error_stop(progname,subname,
     +                 '* value for unit should be 1 or 2 *')
               endif

c     For purposes of calculations within this program convert all pixel
c     sizes to same units as "units"
               if(temp_units .ne. units) then
                  if(temp_units == 1) then
                     pixsizs(nimg) = pixsizs(nimg)/10.0 ! convert to nm
                  else
                     pixsizs(nimg) = pixsizs(nimg)*10.0 ! convert to ang
                  endif
               endif
               pixsiz_out           = min(pixsiz_out,pixsizs(nimg))
               pixsiz_max           = max(pixsiz_max,pixsizs(nimg))
               
c     Open image file and check size of each particle to determine
c     largest value of box_out. Count particles in parameter file and
c     find one with highest value. For MRC files, round up the x and y
c     dimensions of the image up to odd number as required by Baker lab
c     codes.
               
               if(image_type == 'p') then
                  imghead%lunum = pix_file_unit
                  rw(1) = 'r'
                  rw(2) = achar(0)
                  call pif_open(img_fnames(nimg), imghead%lunum, rw)
                  call pif_read_gh (imghead)
                  call pif_close(pix_file_unit)
                  npix_col   = imghead%numcols
                  npix_row   = imghead%numrows
                  imgcount   = imghead%nimg
                  pixhd_htyp = imghead%htype
               else
                  call imopen(pix_file_unit, img_fnames(nimg), 'ro')
                  call irdhdr(pix_file_unit, kxyz, mxyz, jmode, 
     $                 dmin, dmax, dmean)
                  call imclose(pix_file_unit)
                  npix_col   = kxyz(1) + mod((kxyz(1)+1),2)
                  npix_row   = kxyz(2) + mod((kxyz(2)+1),2)
                  imgcount   = kxyz(3)
                  pixhd_htyp = 1                  
               endif

               if (npix_col .ne. npix_row ) then
                  call error_stop(progname,subname,
     +                 ' * param npix_col != npix_row: exit program *')
               endif
               
               nbox(nimg) = npix_col
               box_out      = max(box_out, npix_col)
               
               if(imgcount == 0) then
                  call error_stop(progname,subname,
     $                 ' * img file contains no data: exit program *')
               endif
               
               nids(nimg) = imgcount
               total_views  = total_views + imgcount
                              
               if(nimg == 1) then
                  min_nids = imgcount
               else
                  min_nids = min(min_nids, imgcount)
               endif
            else
               if(first_index(nimg) <= 0) then
                  read(tempstr,*) first_index(nimg)
               endif
               if(  (oddeven .ne. 1 .and. oddeven .ne. 2) .or.
     $              (oddeven == 1 .and. skipone == 1) .or.
     $              (oddeven == 2 .and. skipone == 0) ) then
                  nptle_per_img(nimg) = nptle_per_img(nimg) + 1
               endif
               skipone = mod(skipone+1, 2) 
            endif
   
         enddo  ! End loop over records in parameter file
         close(unit=3)
      enddo  ! End loop over records in input file
      
      npix_row = box_out
      npix_col = box_out

      return
      end subroutine pfile_info
