      subroutine recenter_array_3d(arr, n, newcen, fillin)
C     
C     Three-dimensional version of routine that shifts elements of the
C     array ARR so that the element with the original indices
C     (newcen(1), newcen(2), newcen(3)) resides at the center (0,0,0) of
C     the array. Elements that are 'off the edge' of the array are given
C     the value fillin. An example in one dimension is shown below for
C     newcen(1) = 2 and n = 3
C     
C     shifted array:            -3  -2  -1   0   1   2   3
C                                ^   ^   ^   ^   ^   ^   ^
C                                |   |   |   |   |   |   |
C     original array:   -3  -2  -1   0   1   2   3   |   |
C                                                    |   |
C                                                    fillin
C
C     The purpose of this routine is to avoid unnecessary recalculations
C     of the elements of the array ARR when a new search for the optimal
C     origin or orientation is initiated starting from a new central
C     point.
C     
C     Author: R Sinkovits, 7/5/2006
      
      implicit none
      
      integer, intent(in) :: newcen(3), n
      real, intent(in)    :: fillin
      real, intent(inout) :: arr(-n:n, -n:n, -n:n)
      
      integer idx1, idx1_lb, idx1_ub
      integer idx2, idx2_lb, idx2_ub
      integer idx3, idx3_lb, idx3_ub
      real arr_new(-n:n, -n:n, -n:n)
      
C     Calculate bounds on indices for idx1, idx2, and idx3
      idx1_lb = max(-n - newcen(1), -n)
      idx1_ub = min( n - newcen(1),  n)
      
      idx2_lb   = max(-n - newcen(2), -n)
      idx2_ub   = min( n - newcen(2),  n)
      
      idx3_lb = max(-n - newcen(3), -n)
      idx3_ub = min( n - newcen(3),  n)
      
C     Initialize all elements of arr_new to fillin
      arr_new = fillin
      
C     Fill in elements of shifted array
      do idx1 = idx1_lb, idx1_ub
         do idx2 = idx2_lb, idx2_ub
            do idx3 = idx3_lb, idx3_ub
               arr_new(idx1, idx2, idx3) = 
     $              arr(idx1+newcen(1), idx2+newcen(2), idx3+newcen(3))
            enddo
         enddo
      enddo
      
C     Assign arr_new to arr
      arr = arr_new
      
      return
      end subroutine recenter_array_3d
      

      subroutine recenter_array_2d(arr, n, newcen, fillin)
C     
C     Two-dimensional version of routine that shifts elements of the
C     array ARR so that the element with the original indices
C     (newcen(1), newcen(2)) resides at the center (0,0) of the
C     array. Elements that are 'off the edge' of the array are given the
C     value fillin. An example in one dimension is shown below for
C     newcen(1) = 2 and n = 3
C
C     shifted array:            -3  -2  -1   0   1   2   3
C                                ^   ^   ^   ^   ^   ^   ^
C                                |   |   |   |   |   |   |
C     original array:   -3  -2  -1   0   1   2   3   |   |
C                                                    |   |
C                                                    fillin
C
C     The purpose of this routine is to avoid unnecessary recalculations
C     of the elements of the array ARR when a new search for the optimal
C     origin or orientation is initiated starting from a new central
C     point.
C     
C     Author: R Sinkovits, 7/5/2006
      
      implicit none
      
      integer, intent(in) :: newcen(2), n
      real, intent(in)    :: fillin
      real, intent(inout) :: arr(-n:n, -n:n)
      
      integer idx1, idx1_lb, idx1_ub
      integer idx2, idx2_lb, idx2_ub
      real arr_new(-n:n, -n:n)
      
C     Calculate bounds on indices for idx1 and idx2
      idx1_lb = max(-n - newcen(1), -n)
      idx1_ub = min( n - newcen(1),  n)
      
      idx2_lb   = max(-n - newcen(2), -n)
      idx2_ub   = min( n - newcen(2),  n)
      
C     Initialize all elements of arr_new to fillin
      arr_new = fillin
      
C     Fill in elements of shifted array
      do idx1 = idx1_lb, idx1_ub
         do idx2 = idx2_lb, idx2_ub
            arr_new(idx1, idx2) = 
     $           arr(idx1+newcen(1), idx2+newcen(2))
         enddo
      enddo
      
C     Assign arr_new to arr
      arr = arr_new
      
      return
      end subroutine recenter_array_2d
      

      subroutine recenter_array_1d(arr, n, newcen, fillin)
C
C     One-dimensional version of routine that shifts elements of the
C     array ARR so that the element with the original index (newcen)
C     resides at the center (0) of the array. Elements that are 'off the
C     edge' of the array are given the value fillin. An example in one
C     dimension is shown below for newcen(1) = 2 and n = 3
C
C     shifted array:            -3  -2  -1   0   1   2   3
C                                ^   ^   ^   ^   ^   ^   ^
C                                |   |   |   |   |   |   |
C     original array:   -3  -2  -1   0   1   2   3   |   |
C                                                    |   |
C                                                    fillin
C
C     The purpose of this routine is to avoid unnecessary recalculations
C     of the elements of the array ARR when a new search for the optimal
C     origin or orientation is initiated starting from a new central
C     point.
C     
C     Author: R Sinkovits, 7/5/2006

      implicit none
      
      integer, intent(in) :: newcen, n
      real, intent(in)    :: fillin
      real, intent(inout) :: arr(-n:n)
      
      integer idx1, idx1_lb, idx1_ub
      real arr_new(-n:n)
      
C     Calculate bounds on indices for idx1
      idx1_lb = max(-n - newcen, -n)
      idx1_ub = min( n - newcen,  n)
      
C     Initialize all elements of arr_new to fillin
      arr_new = fillin
      
C     Fill in elements of shifted array
      do idx1 = idx1_lb, idx1_ub
         arr_new(idx1) = arr(idx1)
      enddo
      
C     Assign arr_new to arr
      arr = arr_new
      
      return
      end subroutine recenter_array_1d
