      subroutine to_n22 (theta, phi, omega, z_sym)

C     Checks if orientation is within the specified n2 or n22 asymmetric
C     unit (ASU). If not, the equivalent position in the ASU is
C     calculated.
C     
C     geometric convention:
C     --------------------
C     In a right-handed x-y-z coordinate system, phi is a rotation about
C     the z-axis and theta is a rotation about the "new" y-axis (after
C     the phi rotation). Omega is a rotation about the normal
C     (determined by the theta, phi orientation).
C     
C     In this routine, the n-fold symmetry component is about the
C     z-axis. The value of n is input as z_sym. There is one 2-fold
C     symmetry axis about the x axis. This is assumed in the routine.
C     (for n22 symmetries there also is a 2-fold symmetry axis on the y
C     axis.) The minimum and maximum angles of phi (in the asymmetric
C     unit) are calculated from z_sym as phi_min and phi_max. phi_min
C     and phi_max differ only in sign, with the midpoint being phi =
C     0.0.  for all n2 and n22 asymmetric units, the minimum and maximum
C     theta angles are 0 and 90, respectively.
C     
C     For example, if you have 822 symmetry then the 8-fold symmetry
C     axis is defined along the z axis.  two two-fold axes are along
C     the x and y axes.  the 822 asymmetric unit is, in theta, 0 to 90
C     degrees, and in phi, -22.5 to +22.5 degrees.
C     
C     see belnap et al. (1997) j. struct. biol. 120:44-51 for a
C     definition of theta, phi, and omega angles.
C     
C     Purdue university dec 3000, 4 january 1994, D. Belnap
C     Added phi_dimension, 31 jan 1997, D. Belnap                            
C     General cleanup, 6/28/06, R Sinkovits

      implicit none

      integer*4, intent(in) :: z_sym
      real*4, intent(inout) :: theta, phi, omega

      real*4 phi_dimension, phi_min, phi_max

C     Calculate phi dimension of asymmetric unit
      phi_dimension = 360.0 / float(z_sym)

C     Calculate phi_min and phi_max
      phi_min  =  0.0  -  ( phi_dimension/2.0 )
      phi_max  =  0.0  +  ( phi_dimension/2.0 )

C     If theta at pole, phi = 0.0 and omega ranges between phi_min &
C     phi_max. Add or subtract phi_dimension if omega outside range
      if (theta == 0.0 .or. theta == 180.0) then
         phi = 0.0
         if (theta == 180.0) then
            theta = 0.0
            omega = -omega
         endif
         do while (omega < phi_min)
            omega = omega + phi_dimension
         enddo
         do while (omega > phi_max)
            omega = omega - phi_dimension
         enddo
         return
      endif

C     Determine if orientation is in ASU, exit if it is
      if (theta <= 90.0 .and. phi >= phi_min .and. phi <= phi_max) then
         return
      endif

C     Return stray orientations to ASU by adding or subtracting
C     phi_dimension
      do while (phi < phi_min)
        phi = phi + phi_dimension
      enddo
      do while (phi > phi_max)
        phi = phi - phi_dimension
      enddo

C     Return stray orientations to ASU by adjusting theta, phi, omega
      if (theta > 90.0) then
         theta = 180.0 - theta
         phi   = -phi
         omega = omega + 180.0
         if (omega >= 360.0) omega = omega - 360.0
      endif
      
      return
      end subroutine to_n22
