      subroutine to_nfold (theta, phi, omega, z_sym)
C
C     Tests whether of not a particular orientation falls within the
C     n-fold asymmetric unit (ASU), where the n-fold symmetry component
C     is about the z-axis. If the orientation does not fall within the
C     ASU, it's equivalent position in the ASU is calculated.  See the
C     description for subroutine to_n22 for geometric conventions.
C                                                                             
C     The value of n is input as z_sym and no other symmetry axes are
C     considered. The minimum and maximum angles of phi (in the ASU) are
C     calculated from z_sym as phi_min and phi_max.  phi_min and phi_max
C     differ only in sign, with the midpoint being phi = 0 for all
C     n-fold asymmetric units. The minimum and maximum theta angles are
C     0 and 180, respectively.
C                                                                             
C     For example, given 8-fold symmetry, the asymmetric unit is defined
C     by (0 <= theta <= 180) and |phi| <= 22.5 
C                                                                             
C     31 jan 1997, D. Belnap, modified from to_n22
C                                                                             
C     Modified 19-oct-2000 by TSB/MZ (check for omega > 360.0).
C     Modified 14-aug-2003 by MBS (theta -> 0 - 180 range).
C     Modified 28-jun-2006 by R Sinkovits - general cleanup

      implicit none

      integer*4, intent(in) :: z_sym
      real*4, intent(inout) :: theta, phi, omega

      real*4  phi_dimension, phi_min, phi_max

c     Bring theta into interval (0, 180); adjust phi accordingly
      if (theta < 0.0) then
         theta = -theta
         phi   = phi + 180.
      end if
      if (theta >= 360.0) theta = amod(theta,360.0)
      if(theta >= 180.0) then
         theta = 360.  - theta
         phi   = phi   + 180.
      end if

C     Calculate phi dimension of asymmetric unit
      phi_dimension = 360.0 / float(z_sym)

C     Calculate phi_min and phi_max
      phi_min  =  0.0  -  ( phi_dimension/2.0 )
      phi_max  =  0.0  +  ( phi_dimension/2.0 )

C     If theta at pole, phi = 0.0 and omega ranges between phi_min &
C     phi_max. Add or subtract phi_dimension if omega outside range
      if (theta == 0.0 .or. theta == 180.0) then
         phi = 0.0
         do while (omega < phi_min)
            omega = omega + phi_dimension
         enddo
         do while (omega > phi_max)
            omega = omega - phi_dimension
         enddo
         return
      endif
      
C     Bring omega into interval (0, 360)
      if(omega >= 360.0) omega = amod(omega,360.0)

C     Determine if orientation is in ASU, exit if it is
      if (phi >= phi_min .and. phi <= phi_max)  return

C     Return stray orientations to ASU by adding or subtracting
C     phi_dimension
      do while (phi < phi_min)
        phi = phi + phi_dimension
      enddo
      do while (phi > phi_max)
        phi = phi - phi_dimension
      enddo

      return
      end subroutine to_nfold
