      subroutine output_density(density_map, den_min_x, den_max_x,
     $     max_z_slab, map_min, map_max)

C     Writes out density map to a file. This routine is based on Hong
C     Lin's original routine (1998), but has been modified so that each
C     process writes its share of the map to the file. It also
C     circumvents a shortcoming of the original routine in that the
C     density_map on the root process is not overwritten.
C
C     Note that the logic for calculating the index of the map slice to
C     be written is somewhat confusing and should not be modified unless
C     you really know what you are doing. Also note that in order for
C     this routine to work correctly, a modified version of the PIF
C     library in which the data header is padded out to 512-bytes must
C     be used.

      implicit none

#ifdef MPI
      include 'mpif.h'
#endif

      include '../include/sizes.inc'
      INCLUDE '../include/allprogs.inc'
      INCLUDE '../include/info.inc'
      INCLUDE '../include/pifhead.inc'
      
      type (pifhead_t) maphead

      integer, intent(in) :: den_max_x, den_min_x, max_z_slab
      REAL, intent(in)    :: map_min, map_max

      real, intent(inout) ::   density_map( -den_min_x+den_max_x+1,
     +     -den_min_x+den_max_x+1, max_z_slab)
      
      integer sec, f_z, l_z, nx, ny, nz, mcw, i, j, indx
      real    scale1, map_scalef
      character(len=1) :: versionStr(11), rw(2)
      real    zerosec(den_min_x:den_max_x, den_min_x:den_max_x)

      integer*2 i2
      integer*4 i4

      character(len=14) :: subname
      data subname / 'output_density' /
      data versionStr(1) / 'l' /, versionStr(2) / 'i' /,
     +       versionStr(3) / 'b' /, versionStr(4) / 'P' /,
     +       versionStr(5) / 'I' /, versionStr(6) / 'F' /,
     +       versionStr(7) / ' ' /, versionStr(8) / '2' /,
     +       versionStr(9) / '.' /, versionStr(10) / '1' /,
     +       versionStr(11) / '6' /

      zerosec = 0.0 ! f90 array syntax

      nx = den_max_x * 2 +1
      ny = nx
      nz = nx

C     Make sure that all processes have necessary data and then 
C     set map header information

#ifdef MPI
      mcw = mpi_comm_world
      call mpi_bcast(pixsiz_out, 1, mpi_real, 0, mcw, mpi_err)
      call mpi_bcast(mapmagfactor, 1, mpi_real, 0, mcw, mpi_err)
      call mpi_bcast(pixhd_htyp, 1, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(out_file_name, filename_len, mpi_character,
     $     0, mcw, mpi_err)
#endif

      call pif_init_head(maphead)
      maphead%lunum = 8
         
      if (pixhd_htyp == 1) then
         maphead%numcols = ny
         maphead%numrows = nx
         maphead%numsecs = nz
      else
         call error_stop (progname, subname, 'Undefined pixhd_htype')
      endif
      
      maphead%nimg       =    1
      maphead%imgnum     =    1
      scale1             = 1.01*max(abs(map_min),abs(map_max))
      map_scalef         = real(huge(i4)) / scale1
      maphead%int4scalef = map_scalef
      if (outmode == 40) then
         maphead%mode     =   40
      elseif (outmode == 20) then
         maphead%mode     =   20
         map_scalef       = real(huge(i2)) / scale1
      else
         call error_stop (progname, subname, 'Undefined output mode')
      endif
      maphead%scalef     = map_scalef
      maphead%recipFloatScaleFactor = 1.0 / map_scalef
      maphead%endianNess = 0
      do i =1, 11
         maphead%versionStr(i) = versionStr(i)
      enddo
      maphead%img_bkg    = 0
      maphead%binfactor  = pixhd_htyp
      maphead%a_cell     = ny
      maphead%b_cell     = nx
      maphead%c_cell     = nz
      maphead%dataI4_min = int(map_min)
      maphead%dataI4_max = int(map_max)
      maphead%datar4_min = 0
      maphead%datar4_max = 0
C GCA: pixel size is more or less properly written to file
C      depending on the scale factor determined according
C      to the dynamic range of the voxel values
      maphead%pixelSize  = pixsiz_out / mapmagfactor
      if ( abs(int(maphead%pixelSize*maphead%int4scalef)
     1   /maphead%int4scalef - maphead%pixelSize)>10e-5) then
         maphead%pixelSize  = 0
      endif
         

C     ----- Write header from root process and leftmost -----
C     ----- sections that are not on root               -----

      if (iam == 0) then
         rw(1)  = 'w'
         rw(2) = achar(0)
         call pif_open(out_file_name, maphead%lunum, rw)
         call pif_write_gh(maphead)
         maphead%imgnum = 1
         call pif_write_dh(maphead)
      endif
      
      sec = 0
      f_z = max(frst_lst_z(1,0), den_min_x)
      l_z = min(frst_lst_z(2,0), den_max_x)

      if (f_z <= l_z .and. f_z > den_min_x) then
         do j = den_min_x, f_z - 1
            if(iam == 0) then
               call pif_write_map_sections(maphead, zerosec, outmode,
     $              nx, ny, nz, sec, 1)
            endif
            sec = sec +1
         enddo
      endif
      
      if(iam == 0) then
         call pif_close(maphead%lunum)
      endif

#ifdef MPI
      call mpi_barrier(mcw, mpi_err)
#endif

C     ----- Each process now appends its share of the map -----

      do i = 0, num_nodes-1
         f_z = max(frst_lst_z(1,i), den_min_x)
         l_z = min(frst_lst_z(2,i), den_max_x)
         
         if(iam == i) then
            rw(1)  = 'a'
            rw(2) = achar(0)
            call pif_open(out_file_name, maphead%lunum, rw)
         endif

         do j = f_z, l_z
            indx = j - frst_lst_z(1,iam) + 1
            if(iam == i) then
               call pif_write_map_sections(maphead, 
     $              density_map(1,1,indx), outmode,
     $              nx, ny, nz, sec, 1)
            endif
            sec = sec +1
         enddo
      
         if(iam == i) then
            call pif_close(maphead%lunum)
         endif
#ifdef MPI
         call mpi_barrier(mcw, mpi_err)
#endif
      enddo

C     ----- Write out the rightmost sections that are not on process n-1 -----
      if(iam == 0) then

         f_z = max(frst_lst_z(1,num_nodes-1), den_min_x)
         l_z = min(frst_lst_z(2,num_nodes-1), den_max_x)

         rw(1)  = 'a'
         rw(2) = achar(0)
         call pif_open(out_file_name, maphead%lunum, rw)
         if (l_z < den_max_x)  then
            do i = l_z + 1, den_max_x
               call pif_write_map_sections(maphead, zerosec, outmode,
     $              nx, ny, nz, sec, 1)
               sec = sec +1
            enddo
         endif
         
         call pif_close(maphead%lunum)
         if (sec .ne. nz) call error_stop(progname, subname, 
     +        'write map is not a cube')
      endif

      return
      end subroutine output_density
