/****************************************************************************
** Interp2D_standalone.c
** =====================

James Conway, 1 Jul 1996

Routine to align an image by rotation/scaling and translation. The
interpolation routine (rotscale3) by Michael Unser uses B-splines and is
called once to effect the transformation. We expect this to be called from
Fortran, hence arguments are passed in by address. We will not change them,
however, only the image array.

In this version (the standalone version), all the called code is included:
rotscale3, fiir, getrow.

Note: we want to align to a reference, so we pass in the particle's center and
a reference center. rotscale3() will scale around the array center, then apply
the translational offset we give it. The offset will return our transformed
cventer to the desired one. 

Remember that the TSB array origin (0,0) is at the bottom left - ie, y-axis
is inverted - but the image is also inverted wrt PIC format. The effect is to
change the sign of the rotation only. Note: rotscale3 expects +ve rotation for
CCW, and positive shifts are up and left.

Arguments
=========
float *array		2d array to be scaled, dx * dy
long  *dx, long *dy	dimensions of array
float *rx, float *ry	phase origin of reference - our phase origin will end
			  up here after the rotation/scaling
float *ix, float *iy	our image's phase origin (input origin)
float *rotation		rotation to apply to image - in degrees
float *scale		scale factor to apply to image
float *filler		filler value for off-image pixels

****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "../include/common.h"

// Macros

#define ROUND(a) ( ( (a)<0 ) ? (int) ((a) - 0.5) : (int) ( (a)+0.5 ) )

// Global variable declarations
// ============================

static const double kPI = 3.141592653589793238462643383278;

// Function prototypes
// ===================

int interp2d ( float *array		// 2d array to be scaled /*  */
	      , int  *dx, int *dy	// dimensions of array
	      , float *rx, float *ry	// (output) phase origin of reference
	      , float *ix, float *iy	// (input) phase origin of this image
	      , float *rotation		// rotation to apply to image
	      , float *scale		// scale factor
	      , float *filler		// filler for off-image pixels
              );
int interp2d_ ( float *array		// 2d array to be scaled
	      , int  *dx, int *dy	// dimensions of array
	      , float *rx, float *ry	// (output) phase origin of reference
	      , float *ix, float *iy	// (input) phase origin of this image
	      , float *rotation		// rotation to apply to image
	      , float *scale		// scale factor
	      , float *filler		// filler for off-image pixels
              );

// From rotscale3.h
static
int rotscale3 ( float *inImagePtr
              , int m, int n
              , float *outImagePtr
              , int mr, int nr
              , float tran_x, float tran_y
              , float deg
              , float scale
              , int   trans_first
              , float filler
              );
static double swf3( double x );
static int    newimrxx( float **ima             // Address of image pointer
              	      , short M, short N        // Image size
		      , short init              // Initialization value
		      );

// From FIIR.h

static
int Fiirlpr ( float *p1               // Input image
            , float *p2               // Output image
            , int nx, int ny, int nz  // Image size
            , float zi[]              // Roots
            , int nr                  // Number of roots
            , float gain              // DC-gain
            );
static
void Fiir1ds(float x[], float y[], short N, float c0, float zi[], short nz);

// From getrow.h

static void getxf(float *ip, int ky, int kz, int nx, int ny, float *x);
static void putxf(float *ip, int ky, int kz, int nx, int ny, float *x);
static void getyf(float *ip, int kx, int kz, int nx, int ny, float *x);
static void putyf(float *ip, int kx, int kz, int nx, int ny, float *x);
static void getzf(float *ip, int kx, int ky, int nx, int ny, int nz, float *x);
static void putzf(float *ip, int kx, int ky, int nx, int ny, int nz, float *x);

/*****************************************************************************
** Interp2D
** ========
TSB image is upside-down compared to PIC BP format, but y-axis is also upside
down. So x,y's are the same, but sense of rotation is reversed.
*/
int interp2d ( float *array		
	      , int  *dx, int *dy	
	      , float *rx, float *ry	
	      , float *ix, float *iy	
	      , float *rotation		
	      , float *scale		
	      , float *filler		
              )
{
  return(interp2d_ ( array		
	      , dx, dy	
	      , rx, ry	
	      , ix, iy	
	      , rotation		
	      , scale		
	      , filler		
              ));
}
int interp2d_ ( float *array		// 2d array to be scaled
	      , int  *dx, int *dy	// dimensions of array
	      , float *rx, float *ry	// (output)- phase origin of reference
	      , float *ix, float *iy	// (input) - phase origin of this image
	      , float *rotation		// rotation to apply to image
	      , float *scale		// scale factor
	      , float *filler		// filler for off-image pixels
              )
{
  double shiftX, shiftY, tx, ty;
  float iix, iiy;
  const double rads = -(*rotation) * kPI/180.0;	// convert to radians
  const double sinRot = sin(rads);
  const double cosRot = cos(rads);
  const int xCen = ((*dx)-1)/2;		// integer maths for center
  const int yCen = ((*dy)-1)/2;		// (This is how rotscale3 does it)
  const int rotFirstFlag   = 1;  	// => rotate first (in rotscale3)

/* correct for effect of rotation and scaling about array center (this is how
   rotscale3 does its rotation/scaling)
*/

  iix = (*ix) - xCen;		// position relative to rotation center
  iiy = (*iy) - yCen;
  tx = (*scale) * ( iix*cosRot + iiy*sinRot);	   // rotate and scale
  ty = (*scale) * (-iix*sinRot + iiy*cosRot);
  iix = tx + xCen;		// restore position relative to origin
  iiy = ty + yCen;

// determine shift relative to corresponding image in the first dataset

  shiftX = (*rx) - iix;
  shiftY = (*ry) - iiy;

// go do it

  return(rotscale3( array, *dx, *dy	// source
	   , array, *dx, *dy	// dest (same)
	   , shiftX, shiftY	// in pixels
	   , -(*rotation)	// in degrees (correct for TSB upside-down image
	   , *scale		// apply to image
	   , rotFirstFlag	// rotate/scale about center, then translate
	   , *filler		// fill off-image pixels
	   ));
}
/****************************************************************************
****************************************************************************
** rotscale3.c
** ===========

PURPOSE : Translates, rotates and scales an image using cubic spline
          interpolation. 
NOTE : this routine must have enough memory to allocate an intermediate float
       image used to store the B-spline coefficients.

M. Unser / BEIP  Nov-93
Adapted to Alpha OpenVMS 6.1 by James Conway, Aug-95

Additional files:
        rotscale3.h
        fiir.h ---- getrow.h

If good finish, return 1, else return 0
*/

int rotscale3 ( float *inImagePtr	// input  ( m by n )
              , int m, int n
              , float *outImagePtr	// output (mr by nr)
              , int mr, int nr
              , float tran_x, float tran_y
              , float deg
              , float scale
              , int trans_first
              , float filler
              )
{
  long   i, i1, i2, ii, i3, j, j1, j2, jj, j3, k, roundx, roundy;
  double xpix, x, y, fy, x1, y1, theta, sintheta, costheta;
  float  zi[3], *ixf=0, *optr;
  float x0, y0, x0r, y0r;
  const double pi=4.0*atan(1.0);
  const long   sx=2;
  const float  nz=1.0;
	
  x0  = (m-1)  / 2;
  y0  = (n-1)  / 2;
  x0r = (mr-1) / 2;
  y0r = (nr-1) / 2;

  theta = -pi * ((double) deg)/180.0;

  costheta = cos(theta) / scale;
  sintheta = sin(theta) / scale;

  if(trans_first==0)  {
    x1 = x0r - x0*costheta - y0*sintheta - tran_x;
    y1 = y0r + x0*sintheta - y0*costheta - tran_y;
  }
  else  {
    x1 = x0r - (x0 + tran_x)*costheta - (y0 + tran_y)*sintheta;
    y1 = y0r + (x0 + tran_x)*sintheta - (y0 + tran_y)*costheta;
  }

// Computes cubic B-spline coefficients

  if(newimrxx(&ixf, m, n, -1) == 0)
    return(0);
  {
    float *fptr=ixf;
    for(k=0; k<((long)m)*(long)n; k++)
      *fptr++ = *inImagePtr++;
  }

  zi[0] = -2.0 + sqrt(3.0);
  Fiirlpr(ixf, ixf, m, n, 1, zi, nz, 1.0);		

  for(j=0; j<nr; j++)  {
    optr = outImagePtr + mr*j;
    for(i=0; i<mr; i++)  {
      x = i*costheta + j*sintheta + x1;
      y = j*costheta - i*sintheta + y1;
      roundx = (int) x;
      roundy = (int) y;
      if(roundx>=0 && roundx<m && roundy>=0 && roundy<n)  {
        i1 = roundx - sx;
        i2 = roundx + sx;
        j1 = roundy - sx;
        j2 = roundy + sx;
	     
        xpix = 0.0;
        for(jj=j1; jj<=j2; jj++)  {
          j3 = jj;
          if(j3<0)
            j3 = -j3;
          if(j3>n-1)
            j3 = 2*(n-1) - j3;

          fy = swf3(y-jj);
	      
          k = j3 * m;
          for(ii=i1; ii<=i2; ii++)  {
            i3 = ii;
            if(i3<0) 
              i3 = -i3;
            if(i3>m-1)
              i3 = 2*(m-1) - i3;
            xpix += swf3(x-ii) * fy * ixf[k+i3];
          } // for ii
        }   // for jj

        *optr++ = xpix;
      }  // if(x>=0 && x<=mr-1 && y>=0 && y<=nr-1)
      else
        *optr++ = filler;
    } // for i
  }   // for j

  free(ixf);
  return(1);
}
/****************************************************************************
** swf3
** ====
*/
/* __inline static */
double swf3(double x)
{
  register double xx;

  xx = fabs(x);
  if(xx <= 1.0)
    return(2.0/3 - xx*xx*(1-xx/2.0));
  else if(xx < 2.0)  {
    xx = -xx + 2.0;
    return(xx*xx*xx/6.0);
  }
  else
    return(0.0);
}
/****************************************************************************
** newimrxx
** ========
PURPOSE : Makes sure there exists an MxN float array addressed by IX.
          If the required structure already exists it is reallocated otherwise
          a new pointer is created. The new array is optionally initilialized
          when init>=0.
Michael Unser / BEIB	JAN-90, MAY-92 (for ANSI C compatibility)
*/
static
int newimrxx( float **ima             // Address of image pointer
            , short M, short N        // Image size
            , short init              // Initialization value
            )
{
  float *p;
  long  k, kmax, size;
	
  if ((M==0)||(N==0))
    return 0;
  kmax = (long) M * N;
  size = kmax * sizeof(float);
  if(*ima==NULL)
    *ima = (float *) malloc(size);
  else {
    free(*ima);
    *ima = (float *) malloc(size);
  }
  if(*ima==NULL) {
    printf("newimrxx : Not enough memory");
    return 0;
  }
  if(init >= 0) {
    p = *ima ;
    for(k=0; k<kmax; k++)
      *p++ = init;
  }
  return 1;
}
/******************************************************************************
** Fiirlpr
** =======
Separable Symmetric infinite impulse response filter low pass filters.
Michael Unser / BEIB NOV-93 / MAY-94
*/
static
int Fiirlpr ( float *p1               // Input image
            , float *p2               // Output image
            , int nx, int ny, int nz  // Image size
            , float zi[]              // Roots
            , int nr                  // Number of roots
            , float gain              // DC-gain
            )
{
  int kx,ky,kz,i;
  long maxsize;
  float *x,*y;
  float c0,c1,c2,c3;

  maxsize=MAX(MAX(nx,ny),nz)*sizeof(float);
  x=(float*)malloc((size_t)maxsize);
  y=(float*)malloc((size_t)maxsize);
  if (y ==0) {
    if (x!=0) free(x);
    printf("fiirlpr: Out of memory");
    return (0);
  }
  c0=1.;
  for (i=0;i<nr;i++) {
    if (fabs(zi[i])>=1.) {
      printf("fiirlpr: Unstable filter");
      return (0);
    }
    c0=c0*(1.-zi[i])*(1.-zi[i]);
  }
  if (gain>0)
    c1=gain*c0;
  c2=c0;c3=c0;

// x-processing

  for (kz=0;kz<nz;kz++)
    for (ky=0;ky<ny;ky++) {
      getxf(p1,ky,kz,nx,ny,x);
      Fiir1ds(x,y,nx,c1,zi,nr);
      putxf(p2,ky,kz,nx,ny,y);
    }

// y-processing

 if (ny>1) {
   for (kz=0;kz<nz;kz++)
     for (kx=0;kx<nx;kx++) {
       getyf(p2,kx,kz,nx,ny,x);
       Fiir1ds(x,y,ny,c2,zi,nr);
       putyf(p2,kx,kz,nx,ny,y);
     }
  }

// z-processing

 if (nz>1) {
   for (ky=0;ky<ny;ky++)
     for (kx=0;kx<nx;kx++) {
       getzf(p2,kx,ky,nx,ny,nz,x);
       Fiir1ds(x,y,nx,c3,zi,nr);
       putzf(p2,kx,ky,nx,ny,nz,y);
     }
  }
  free(x);
  free(y);
  return (1);
}
/******************************************************************************
** Fiir1ds
** =======

PURPOSE : 1d symmetrical all pole recursive filter with clean reversible
mirror symmetric boundary conditions.

	H(z) = Product[ c0 / ( (1-zi[i]*z) (1-zi[i]/z) ), {i,1,nz}]

NOTE : x and y may occupy the same storage location.

Author : Michael Unser / BEIP        NOV-90/93 & APR-95 (full proof boundary conditions)

Example : to get the cubic B-spline coefficients {c[k]} for the
interpolation a sequence {x[k], k=0,...,nx-1} use the following call :

	alpha = -2. + sqrt(3.);
	c0 = (1-alpha)*(1-alpha);
	Fiir1ds(x, c, nx, c0, &alpha, 1);

Many other examples of the use of this subroutine may be found in :
[1]  M. Unser, A. Aldroubi and M. Eden, "B-spline signal processing. Part II :
   efficient design and applications", IEEE Trans. Signal Processing, vol. 41,
   pp. 834-848, February 1993.

******************************************************************************/
static
void Fiir1ds(float x[], float y[], short N, float c0, float zi[], short nz)
{
  float x0, xN;
  const float log10_epsi = -8.0;
  int k, kk, k0, k1, k00, iz;
  const int kn = 2*N - 2;

  for(iz=0; iz<nz; iz++)  {
    k0 = (int) (log10_epsi / log10(fabs(zi[iz])));
    k00 = k0 % kn;
    if(k00>=N)
      k00 = kn - k00;
    k1 = MIN(N-1, k0-1);
    if(iz==0)  {
      x0 = x[k00];
      xN = c0*x[N-1];
/* APR-95 : The case when k0 >= N */
      for(k=k0-1; k>=N; k--)  {
        kk = k % kn;
        if(kk>=N)
          kk = kn - kk;
        x0 = x[kk] + zi[iz]*x0;
      }
      for(k=k1; k>=0; k--)
        x0 = x[k] + zi[iz]*x0;
      y[0] = c0*x0;
      for(k=1; k<N; k++)
        y[k] = c0*x[k] + zi[iz]*y[k-1];
    }
    else  {
      x0 = y[k00];
      xN = y[N-1];
/* APR-95 : The case when k0 >= N */
      for(k=k0-1; k>=N; k--)  {
        kk = k % kn;
        if(kk>=N)
          kk = kn - kk;
        x0 = y[kk] + zi[iz]*x0;
      }
      for(k=k1; k>=0; k--)
        x0 = y[k] + zi[iz]*x0;
      y[0] = x0;
      for(k=1; k<N; k++)
        y[k] = y[k] + zi[iz]*y[k-1];
    }
    y[N-1] = (2.0*y[N-1] - xN) / (1.0-zi[iz]*zi[iz]);
    for(k=N-2; k>=0; k--)
      y[k] = y[k] + zi[iz]*y[k+1];
  }
}
/*************************************************************
 *	Title : getxf
 *	      : getyf
 *	      : getzf
 *	      : putxs
 *	      : putys
 *	      : putzs
 *
 *  PURPOSE : Set of routines for extracting or copying x, y, z
 *  vectors from a FLOAT 3D image of size (nx x ny x nz).
 *
 *************************************************************/

static
void getxf(float *ip, int ky, int kz, int nx, int ny, float *x)
{
  int i;
  ip += ((long) nx) * (ky + ((long) ny) * ((long) kz));
  for(i=0; i<nx; i++)
    *x++ = *ip++;
}

static
void putxf(float *ip, int ky, int kz, int nx, int ny, float *x)
{
  int i;
  ip += ((long) nx) * (ky + ((long) ny) * ((long) kz));
  for(i=0; i<nx; i++)
    *ip++ = *x++;
}

static
void getyf(float *ip, int kx, int kz, int nx, int ny, float *x)
{
  int i;
  ip += kx + ((long) nx) * ((long) ny) * ((long) kz);
  for(i=0; i<ny; i++, ip+=nx)
    *x++ = *ip;
}

static
void putyf(float *ip, int kx, int kz, int nx, int ny, float *x)
{
  int i;
  ip += kx + ((long) nx) * ((long) ny) * ((long) kz);
  for(i=0; i<ny; i++, ip+=nx)
    *ip = *x++;
}

static
void getzf(float *ip, int kx, int ky, int nx, int ny, int nz, float *x)
{
  int i;
  const long nxy = ((long) nx) * ((long) ny);
  ip += kx + ((long) nx) * ((long) ky);
  for(i=0; i<nz; i++, ip+=nxy)
    *x++ = *ip;
}

static
void putzf(float *ip, int kx, int ky, int nx, int ny, int nz, float *x)
{
  int i;
  const long nxy = ((long) nx) * ((long) ny);
  ip += kx + ((long) nx) * ((long) ky);
  for(i=0; i<nz; i++, ip+=nxy)
    *ip = *x++;
}
