C  ****************************************************************************
C  *                                                                          *
C  *      SFLIB.F SUBROUTINES                                                 *
C  *                                                                          *
C  *      SF_ABTOAP      - CONVERT SF DATA FROM APART/BPART TO AMP/PHASE      *
C  *      SF_APTOAB      - CONVERT SF DATA FROM AMP/PHASE TO APART/BPART      *
C  *      SF_AVGMATES       - AVERAGE TWO SYMMETRY MATES (CONSTRAIN TO BE REAL)  *
C  *      SF_BT_LOAD       - LOAD UP SF DATA FOR BACK-TRANSFORMING TO MAP DATA  *
C  *      SF_BT_MAPOUT   - OUTPUT BACK-TRANSFORMED SF DATA                    *
C  *      SF_BT_MINMAX       - COMPUTE MIN AND MAX OF BACK-TRANSFORMED DATA       *
C  *      SF_CHECKLIM       - CHECK IF LIMIT OF NLIM SPOTS HAS BEEN EXCEEDED     *
C  *      SF_FILL        - LOAD SF DATA INTO ARRAYS                           *
C  *      SF_FLIPPHASE   - INVERT PHASE (REVERSES CONTRAST)                   *
C  *      SF_HKL       - CONVERT 2D HKL INDICES TO 3D INDICES               *
C  *      SF_INIT       - INITIALIZE COUNTERS, ETC.                          *
C  *      SF_OPEN        - OPEN SF DATA FILE                                  *
C  *      SF_ORIGIN2       - FIND BEST 2-FOLD ORIGIN IN 2D LATTICE              *
C  *      SF_ORIGIN3       - FIND BEST 3- OR 6-FOLD ORIGIN IN 2D LATTICE        *
C  *      SF_ORIGIN4       - FIND BEST 4-FOLD ORIGIN IN 2D LATTICE              *
C  *      SF_PSHIFT       - APPLY PHASE SHIFT TO SF DATA                       *
C  *      SF_RESCALE     - MULTIPLY SF DATA BY A SCALE FACTOR                 *
C  *      SF_RFACTOR     - COMPUTE VARIOUS R-FACTORS BETWEEN SF DATA SETS     *
C  *      SF_SCALEFAC    - COMPUTE LINEAR SCALE FACTOR BETWEEN SF DATA SETS   *
C  *      SF_SETSHIFT       -                                                    *
C  *      SF_SORT        - SORT THE SF DATA ACCORDING TO HKL INDICES          *
C  *      SF_STATS       - COMPUTE VARIOUS SF DATA STATISTICS                 *
C  *      SF_SYM_P2      - IMPOSE p2 SYMMETRY ON SF DATA                      *
C  *      SF_SYM_PGG     - IMPOSE pgg SYMMETRY ON SF DATA                     *
C  *      SF_SYM_CMM     - IMPOSE cmm SYMMETRY ON SF DATA                     *
C  *      SF_SYM_P4      - IMPOSE p4 SYMMETRY ON SF DATA                      *
C  *      SF_SYM_P4M     - IMPOSE p4m SYMMETRY ON SF DATA                     *
C  *      SF_SYM_P4G     - IMPOSE p4g SYMMETRY ON SF DATA                     *
C  *      SF_SYM_P6      - IMPOSE p6 SYMMETRY ON SF DATA                      *
C  *      SF_SYM_P6M     - IMPOSE p6m SYMMETRY ON SF DATA                     *
C  *      SF_TFAC        - APPLY TEMPERATURE FACTOR TO SF DATA                *
C  *      SF_WRITE       - WRITE BINARY/ASCII SF DATA FILE                    *
C  *      SFREF_CMB      - Combine several data sets to obtain averaged data  *
C  *      SFREF_CMP      - Compare two data sets and find best common origin  *
C  *      SFREF_FLP      - Turn image by 180 deg or apply a vertical mirror   *
C  *      SFREF_WRT      - Save current structure factor data in output file  *
C  *                                                                          *
C  *    The following subroutines are from the EMSF3DBT program, which was    *
C  *    incorporated into EMSF (Dec 2001).                                    *
C  *                                                                          *
C  *      SF3DBT_FOURIER - Fourier summation routine                          *
C  *      SF3DBT_OUT     - Output the 3D MAP to disk                          *
C  *      SF3DBT_INPUT   - Reads input for driving SF3DBT routine             *
C  *      SF3DBT_INFO    - Read control parameters, cell constants, and       *
C  *                        symmetry cards                                    *
C  *      SF3DBT_TRANS   - Sections calculated parallel to a general plane    *
C  *                                                                          *
C  *      SF3DBT_CHEKUP                                                       *
C  *      SF3DBT_CROSS                                                        *
C  *      SF3DBT_GETARGS                                                      *
C  *      SF3DBT_GP1                                                          *
C  *      SF3DBT_MI3                                                          *
C  *      SF3DBT_NORM                                                         *
C  *      SF3DBT_ROT                                                          *
C  *      SF3DBT_SETU                                                         *
C  *      SF3DBT_STOAA                                                        *
C  *      SF3DBT_VM                                                           *
C  *      SF3DBT_VMV                                                          *
C  *                                                                          *
C  ****************************************************************************

C  ******************************************************************
C  *                                                                *
C  *            *****     SUBROUTINE SF_ABTOAP     *****            *
C  *                                                                *
C  *    CONVERT STRUCTURE FACTOR DATA FROM APART/BPART FORMAT TO    *
C  *     AMP/PHASE FORMAT.                                          *
C  *                                                                *
C  *    JULY 1986  PURDUE UNIVERSITY VAX 11/750, T. BAKER.          *
C  *    OCTOBER 1990  PURDUE UNIVERSITY VAX 8550, T. BAKER.         *
C  *    LAST MODIFIED 12-MAR-1992 BY T. BAKER.                      *
C  ******************************************************************
C
      SUBROUTINE SF_ABTOAP (SF_TYPE, NSF, AMP, PHI)
C
      IMPLICIT NONE
C
      INTEGER*4 I, NSF, SF_TYPE
      REAL*4    AMP(NSF), PHI(NSF), TMP1, TMP2, PI, RADIAN, DEGREE
C  ----------------------------------------------------------------------------
C
      CALL PIRADDEG (PI, RADIAN, DEGREE)
      SF_TYPE = 1                  ! SIGNIFIES SWITCH TO AMP/PHASE SF FORMAT
C
      DO I = 1,NSF
        TMP1   = AMP(I)
        TMP2   = PHI(I)
        AMP(I) = SQRT(TMP1**2+TMP2**2)
        IF(TMP1 == 0.0) THEN
           IF(TMP2.GT.0.0) PHI(I) =  90.0
           IF(TMP2.LT.0.0) PHI(I) = 270.0
         ELSE
           PHI(I) = ATAN2(TMP2,TMP1)/RADIAN
           PHI(I) = AMOD(PHI(I),360.0)
           IF(PHI(I).LT.0.0) PHI(I) = 360.0+PHI(I)
        ENDIF
      ENDDO
C
      RETURN
      END

C  ******************************************************************
C  *                                                                *
C  *            *****     SUBROUTINE SF_APTOAB     *****            *
C  *                                                                *
C  *    CONVERT STRUCTURE FACTOR DATA FROM AMP/PHASE FORMAT TO      *
C  *     APART/BPART FORMAT.                                        *
C  *                                                                *
C  *    JULY 1986  PURDUE UNIVERSITY VAX 11/750, T. BAKER.          *
C  *    OCTOBER 1990  PURDUE UNIVERSITY VAX 8550, T. BAKER.         *
C  *    LAST MODIFIED 12-MAR-1992 BY T. BAKER.                      *
C  ******************************************************************
C
      SUBROUTINE SF_APTOAB (SF_TYPE, NSF, AMP, PHI)
C
      IMPLICIT NONE
C
      INTEGER*4 I, NSF, SF_TYPE
      REAL*4    AMP(NSF), PHI(NSF), TMP1, TMP2
      real*8    deg_to_rad
C  ----------------------------------------------------------------------------
C
      deg_to_rad = acos(-1.0d0)/180.0d0
      SF_TYPE = 0                 ! SIGNIFIES SHIFT TO APART/BPART SF FORMAT
C
      DO I = 1,NSF
        TMP1   = AMP(I)
        TMP2   = PHI(I)
        AMP(I) = TMP1 + REAL(COS(deg_to_rad*TMP2))
        PHI(I) = TMP1 + REAL(SIN(deg_to_rad*TMP2))
        PHI(I) = AMOD(PHI(I),360.0)
        IF(PHI(I).LT.0.0) PHI(I) = 360.0+PHI(I)
      ENDDO
C
      RETURN
      END

C  ****************************************************************
C  *                                                              *
C  *             *****     SUBROUTINE SF_APAB     *****           *
C  *                                                              *
C  *    COMPUTES AMPLITUDE AND PHASE FROM INPUT APART, BPART      *
C  *     IF SFTYPE = 0 OR APART AND BPART FROM INPUT AMPLITUDE    *
C  *     AND PHASE IF SFTYPE = 1.                                 *
C  *                                                              *
C  *    OCTOBER 1990  PURDUE UNIVERSITY VAX 8550, T. BAKER.       *
C  *    LAST MODIFIED 10-OCT-1990 BY T. BAKER.                    *
C  ****************************************************************
C
      SUBROUTINE SF_APAB (SFTYPE,A,B,AMP,PHI,APART,BPART,RADIAN)
C                 
      IMPLICIT NONE
C
      INTEGER*4 SFTYPE
      REAL*4    A, B, AMP, PHI, APART, BPART, RADIAN
      real*8    deg_to_rad
C  ----------------------------------------------------------------------------
C
      deg_to_rad = acos(-1.0d0)/180.0d0

      IF(SFTYPE == 0) THEN                  ! A,B = APART,BPART
         APART = A
         BPART = B
         AMP   = SQRT(APART**2 + BPART**2)
         PHI   = 0.0
         IF(APART == 0.0) THEN
            IF(BPART.GT.0.0) PHI =  90.0
            IF(BPART.LT.0.0) PHI = 270.0
          ELSE
            PHI = ATAN2(BPART,APART)/RADIAN
            PHI = AMOD(PHI,360.0)
            IF(PHI.LT.0.0) PHI = 360.0 + PHI
         ENDIF
       ELSE                               ! A,B = AMP,PHI
         AMP   = A
         PHI   = B
         APART = REAL(AMP*COS(deg_to_rad*PHI))
         BPART = REAL(AMP*SIN(deg_to_rad*PHI))
      ENDIF
C
      RETURN
      END

C  ***********************************************************
C  *                                                         *
C  *       *****     SUBROUTINE SF_AVGMATES     *****        *
C  *                                                         *
C  *    AVERAGE TWO SYMMETRY MATES (CONSTRAIN TO BE REAL).   *
C  *                                                         *
C  *    JUNE 1986  PURDUE UNIVERSITY VAX 750, T. BAKER.      *
C  *    MARCH 1992  PURDUE UNIVERSITY VAX 8550, T. BAKER.    *
C  *    LAST MODIFIED 13-MAR-1992 BY T. BAKER.               *
C  ***********************************************************
C
      SUBROUTINE SF_AVGMATES (SF_TYPE,N,J,NSF,AMP,PHI,NPAIR,FLAG)
C
      IMPLICIT NONE
C
      INTEGER*4 NPAIR, J, N, NSF, FLAG(NSF), SF_TYPE
      REAL*4    AMP(NSF), PHI(NSF), P
      real*8    deg_to_rad
C  ----------------------------------------------------------------------------
C
      deg_to_rad = acos(-1.0d0)/180.0d0

      FLAG(N) = 1
      NPAIR   = NPAIR + 1
C
      IF(SF_TYPE == 0) THEN                          ! APART/BPART SF DATA
         AMP(J) = (AMP(J) + AMP(N)) /2.0
         AMP(N) = AMP(J)
         PHI(J) = 0.0
         PHI(N) = 0.0
      ELSE                      ! AMP/PHASE SF DATA
         AMP(J)  = REAL((AMP(J)*COS(deg_to_rad*PHI(J)) + 
     $         AMP(N)*COS(deg_to_rad*PHI(N))))/2.0
         P       = 0.0
         IF(AMP(J).LT.0.0) P = 180.0
         AMP(J) = ABS(AMP(J))
         AMP(N) = AMP(J)
         PHI(J) = P
         PHI(N) = P
      ENDIF
      RETURN
C
      END

C  ***********************************************************************
C  *                                                                     *
C  *              *****     SUBROUTINE SF_BT_LOAD     *****              *
C  *                                                                     *
C  *    LOAD STRUCTURE FACTOR DATA (AMPS AND PHASES) AFTER CONVERTING    *
C  *     TO A-PARTS AND B-PARTS AND GENERATE FRIEDEL MATES               *
C  *                                                                     *
C  *    LAST MODIFIED 10-MAR-1992 BY T. BAKER.                           *
C  ***********************************************************************
C
      SUBROUTINE SF_BT_LOAD (SF_TYPE,NSF,HKL,AMPS,PHIS,N1,N2,ARRAY,
     1 T_MATRIX)
C
      IMPLICIT NONE
C
      INTEGER*4 NSF, I, N1, N2, ITST, SF_TYPE, T_MATRIX(2,3)
      INTEGER*4 H, K, L, IH, IK, IHBAR, IKBAR, HKL(3,NSF)
      REAL*4    ARRAY(2,N1,N2), AMPP, PHII, AMPS(NSF), PHIS(NSF)
      REAL*4    APART, BPART
      real*8    deg_to_rad
C  ----------------------------------------------------------------------------
C
      deg_to_rad = acos(-1.0d0)/180.0d0
      ITST = (N1/2) - 1
C
C  *******************************************************
C  *    READ IN THE AMPLITUDES AND PHASES AND CONVERT    *
C  *     TO APARTS AND BPARTS OF STRUCTURE FACTORS       *
C  *******************************************************
      AMPP = 1.
      PHII = 0.
      DO I = 1,NSF
        IF(SF_TYPE == 0) THEN
           APART = AMPS(I)
           BPART = PHIS(I)
         ELSE
           AMPP = AMPS(I)
           PHII = PHIS(I)
        ENDIF
        H    = HKL(1,I)
        K    = HKL(2,I)
        L    = HKL(3,I)
        CALL HKL_TO_HK (H, K, L, IH, IK, T_MATRIX)
CCC        IF(IH == 0 .AND. IK == 0) AMPP = 0.0
        IF(ABS(IH).GT.ITST .OR. ABS(IK).GT.ITST) THEN
          WRITE(*,'(A,I2,A,I2,A)') 'Reflection [', IH, ',', IK, 
     1    '] rejected'
          GO TO 100
        ENDIF
C  *************************************************
C  *    SHIFT ORIGIN TO THE CENTER OF THE ARRAY    *
C  *************************************************
        IF(SF_TYPE == 0) THEN
           CALL BELL (3)
           PRINT *, '* CANNOT PROCESS A/B PART DATA'
           RETURN
         ELSE
           PHII = PHII - 180.0*(IH+IK)
           PHII = AMOD(PHII,360.0)
           IF(PHII.LT.0.0) PHII = PHII + 360.0
        ENDIF
        IH = IH + 1
        IK = IK + 1
        IF(IH.LE.0) IH = IH + N1
        IF(IK.LE.0) IK = IK + N1
        IF(SF_TYPE == 0) THEN
           ARRAY(1,IK,IH) = APART
           ARRAY(2,IK,IH) = BPART
         ELSE
           ARRAY(1,IK,IH) = REAL(AMPP*COS(deg_to_rad*PHII))
           ARRAY(2,IK,IH) = REAL(AMPP*SIN(deg_to_rad*PHII))
        ENDIF
C  ***************************************
C  *    NOW GENERATE THE FRIEDEL MATE    *
C  ***************************************
        IHBAR = -IH + 2
        IKBAR = -IK + 2
        IF(IHBAR.LE.0) IHBAR = IHBAR + N1
        IF(IKBAR.LE.0) IKBAR = IKBAR + N1
        ARRAY(1,IKBAR,IHBAR) =  ARRAY(1,IK,IH)
        ARRAY(2,IKBAR,IHBAR) = -ARRAY(2,IK,IH)
  100      ENDDO
C
      RETURN
      END

C  *************************************************************
C  *                                                           *
C  *        *****     SUBROUTINE SF_BT_MAPOUT     *****        *
C  *                                                           *
C  *    CREATES OUPUT FILE FOR PLOT AND CONTOUR PROGRAMS       *
C  *                                                           *
C  *    MODIFIED 10-MAR-1992 BY T. BAKER.                      *
C  *    LAST MODIFIED 12-DEC-1997 BY T. BAKER.                 *
C  *************************************************************
C
      SUBROUTINE SF_BT_MAPOUT (N1, N2, ARRAY, MAP_MIN, MAP_MAX, 
     1 T_MATRIX, A_CELL, B_CELL, C_CELL, MAP_ABANG, ALP, BET, GAM,
     2 AOVERB, ALP_STAR, BET_STAR, GAM_STAR, A_STAR, B_STAR, C_STAR,
     3 VOL, VOL_STAR, MAP_TITLE)
C
      IMPLICIT NONE
C
      INTEGER*4 N1, N2, NCOL, NROW, NSEC, I, J, ID
      INTEGER*4 T_MATRIX(2,3), MAP_TITLE(18)
      REAL*4    ARRAY(2,N1,N2), D1, D2, VOL, VOL_STAR
      REAL*4    MAP_MIN, MAP_MAX, MAP_ABANG, AOVERB
      REAL*4    ALP, BET, GAM, DELA, DELB, DELC, A_CELL, B_CELL, C_CELL
      REAL*4    A_STAR, B_STAR, C_STAR, ALP_STAR, BET_STAR, GAM_STAR
      CHARACTER FILENAME*64
C  ----------------------------------------------------------------------------
C
      NCOL = N1
      NROW = N2
      NSEC = 1
      D1 = 1.
      D2 = 1.
C
 100  WRITE(*,'(A)') '2D MAP OUTPUT FILENAME: '
      READ (*,'(A)',ERR=100) FILENAME
      OPEN (UNIT=2, FILE=FILENAME, STATUS='NEW', ACCESS='SEQUENTIAL',
     1 FORM='UNFORMATTED')
      WRITE(*,'(A)') 'HEADER: '
      READ (*,'(18A4)') MAP_TITLE
      ID = 3                        ! FILE IDENTIFIER = 3 FOR MAP FILES
      WRITE(2) ID
      WRITE(2) MAP_TITLE
C  *********************************************************
C  *    DETERMINE PARAMETERS FOR SCALING THE OUTPUT MAP    *
C  *********************************************************
      CALL REC_TO_REAL (A_CELL, B_CELL, C_CELL, ALP, BET, GAM, VOL,
     1 A_STAR, B_STAR, C_STAR, VOL_STAR, ALP_STAR, BET_STAR, GAM_STAR)
      CALL MAP_GET_ANGLE (T_MATRIX, A_CELL, B_CELL, C_CELL, MAP_ABANG, 
     1 ALP, BET, GAM)
C
      IF(A_CELL == 0.0) DELA = 0.0
      IF(T_MATRIX(1,1).NE.0) THEN
        D1   = 1.0/A_STAR
        DELA = A_CELL/NCOL
      ENDIF
      IF(T_MATRIX(2,1).NE.0) THEN
        D2   = 1.0/A_STAR
        DELA = A_CELL/NROW
      ENDIF
C
      IF(B_CELL == 0.0) DELB = 0.0
      IF(T_MATRIX(1,2).NE.0) THEN
        D1   = 1.0/B_STAR
        DELB = B_CELL/NCOL
      ENDIF
      IF(T_MATRIX(2,2).NE.0) THEN
        D2   = 1.0/B_STAR
        DELB = B_CELL/NROW
      ENDIF
C
      IF(C_CELL == 0.0) DELC = 0.0
      IF(T_MATRIX(1,3).NE.0) THEN
        D1   = 1.0/C_STAR
        DELC = C_CELL/NCOL
      ENDIF
      IF(T_MATRIX(2,3).NE.0) THEN
        D2   = 1.0/C_STAR
        DELC = C_CELL/NROW
      ENDIF
C
      AOVERB = D1/D2
      WRITE(2) NCOL, NROW, NSEC, MAP_MIN, MAP_MAX, AOVERB, MAP_ABANG
      WRITE(2) A_CELL, B_CELL, C_CELL, ALP, BET, GAM
      WRITE(2) DELA, DELB, DELC, T_MATRIX
C
      DO J = 1,NROW
        WRITE(2) (ARRAY(1,J,I),I=1,NCOL)
      ENDDO
C
      CLOSE (UNIT=2, STATUS='KEEP')            ! CLOSE AND SAVE MAP FILE
      RETURN
C
      END

C  ***********************************************************
C  *                                                         *
C  *      *****     SUBROUTINE SF_BT_MINMAX     *****        *
C  *                                                         *
C  *    THIS FINDS THE MINIMUM AND MAXIMUM VALUES IN ARRAY   *
C  *                                                         *
C  *    MODIFIED 10-MAR-1992 BY T. BAKER.                    *
C  *    LAST MODIFIED 12-DEC-1997 BY T. BAKER.               *
C  ***********************************************************
C
      SUBROUTINE SF_BT_MINMAX (N1, N2, ARRAY, MAP_MIN, MAP_MAX)
C
      IMPLICIT NONE

      INTEGER*4 N1, N2, I, ARRAY_SIZE
      REAL*4    ARRAY(2*N1*N2), MAP_MIN, MAP_MAX, X
C  ----------------------------------------------------------------------------
C
      MAP_MIN    = HUGE(MAP_MIN)
      MAP_MAX    = -HUGE(MAP_MAX)
      ARRAY_SIZE = 2*N1*N2
C
      DO I = 1,ARRAY_SIZE,2
        X       = ARRAY(I)
        MAP_MAX = AMAX1(MAP_MAX,X)
        MAP_MIN = AMIN1(MAP_MIN,X)
      ENDDO
C
      WRITE(*,'(A,2F10.1)') ' MAP_MIN,MAP_MAX = ', MAP_MIN, MAP_MAX
C
      RETURN
      END

C  **************************************************************
C  *                                                            *
C  *         *****     SUBROUTINE SF_CHECKLIM     *****         *
C  *                                                            *
C  *    Check if limit of NLIM spots has been exceeded          *
C  **************************************************************
C
      SUBROUTINE SF_CHECKLIM (NLIM, NEW)
C
      IMPLICIT NONE
C
      INTEGER*4 NLIM, NEW
C  ----------------------------------------------------------------------------
C
      IF(NEW.LE.NLIM) RETURN
      WRITE(*,'(A,I5,2A)') ' Limit of', NLIM, ' spots exceeded...',
     1 'EXIT program'
      CALL TSBEND
C
      END

C  *****************************************************************
C  *                                                               *
C  *                *****     SUBROUTINE SF_FILL *****             *
C  *                                                               *
C  *    READ SF DATA FROM UNIT LUN INTO ARRAYS.                    *
C  *                                                               *
C  *    September 1998  PURDUE UNIVERSITY, RW Ashmore              *
C  *    LAST MODIFIED 29-Nov-2001 (TSB/RWA Read correct # of Sfs)  *
C  *****************************************************************
C
      SUBROUTINE SF_FILL (EM_ERR, SF_TYPE, NSF, HKL, 
     1 AMP, PHI, SIG, DSTR, SF_DSTR_MIN, SF_DSTR_MAX, SF_SCALEF,
     2 SF_SUMFSQ, DSTR_MAX_user, nsf_in)
C
      IMPLICIT NONE

      INTEGER*4 EM_ERR, NSF, I, SF_TYPE, nsf_in, units
      INTEGER*4 HKL(3,NSF)
      REAL*4    AMP(NSF), PHI(NSF), SIG(NSF), DSTR(NSF), DSTR_MAX_user
      REAL*4    SF_DSTR_MIN, SF_DSTR_MAX, SF_SUMFSQ, SF_SCALEF
      real*4    voxsiz, reslim

      include '../include/pifhead.inc'
      TYPE(PIFHEAD_T)  MAPHEAD, IMGHEAD, PRJHEAD, PFTHEAD, SFHEAD
      COMMON /PIFHEAD/ MAPHEAD, IMGHEAD, PRJHEAD, PFTHEAD, SFHEAD

      include '../include/strucfac.inc'
      type(strucFac_T) stFacs(NSF)
C  ----------------------------------------------------------------------------
C
      PRINT *, '*********************************'
      PRINT *, '* Read in STRUCTURE FACTOR data *'
      PRINT *, '*********************************'
      PRINT '(I8,A)', nsf, ' total structure factors stored in file'

C  ************************************************
C  *    READ THE STRUCTURE FACTOR DATA FILE TO    *
C  *     DETERMINE ARRAY SIZE REQUIREMENTS        *
C  ************************************************
      units  = SFHEAD%PixSizUnits
      voxsiz = SFHEAD%PixelSize
      if(voxsiz.le.0.0) voxsiz = 1.0            ! Assume unitless pixels
      reslim = 2.0001*voxsiz
      if(units == 1) then
         WRITE(*,'(A,F8.3,A)') 'MAX resolution (<CR>=', reslim, 
     1        ' A) (F): '
       else if (units == 2) then
         WRITE(*,'(A,F8.3,A)') 'MAX resolution (<CR>=', reslim,
     1       ' nm) (F): '
       else
         WRITE(*,'(A,F8.3,A)') 'MAX resolution (<CR>=', reslim, 
     1       ' pixels) (F): '
      endif
      READ (*,*) DSTR_MAX_user
      if(dstr_max_user.le.0.0) dstr_max_user = reslim
      IF(dstr_max_user.LT.2.0001*VOXSIZ) THEN
        dstr_max_user = 2.0001*VOXSIZ
        WRITE(*,'(A,F8.3)') 'WARNING: res limit reduced to', 
     1      dstr_max_user
      ENDIF
      DSTR_MAX_USER = 1.0/DSTR_MAX_USER      ! Switch to reciprocal units
C
C  *******************************************
C  *    READ IN THE STRUCTURE FACTOR DATA    *
C  *******************************************
      SF_DSTR_MIN = HUGE(SF_DSTR_MIN)
      SF_DSTR_MAX = -HUGE(SF_DSTR_MAX)
C
      call pif_read_sf (sfHead,stFacs)
C
      nsf_in = 0            ! Counter for sf data
      DO 100 I = 1,NSF
        if(stFacs(i)%dstr.gt.DSTR_MAX_user) go to 100
        nsf_in        = nsf_in + 1
        hkl(1,nsf_in) = stFacs(i)%h
        hkl(2,nsf_in) = stFacs(i)%k
        hkl(3,nsf_in) = stFacs(i)%l
        amp(nsf_in)   = stFacs(i)%amp
        phi(nsf_in)   = stFacs(i)%phi
        sig(nsf_in)   = stFacs(i)%sig
        dstr(nsf_in)  = stFacs(i)%dstr
C
        SF_DSTR_MIN = MIN(SF_DSTR_MIN,DSTR(nsf_in))
        SF_DSTR_MAX = MAX(SF_DSTR_MAX,DSTR(nsf_in))
        IF(SF_TYPE == 0) THEN
           AMP(I)    = AMP(I)*SF_SCALEF                  ! APART, BPART
           PHI(I)    = PHI(I)*SF_SCALEF
           SF_SUMFSQ = SF_SUMFSQ + AMP(I)**2 + PHI(I)**2
         ELSE IF(SF_TYPE == 1) THEN
           AMP(I)    = AMP(I)*SF_SCALEF                  ! AMP, PHASE
             SF_SUMFSQ = SF_SUMFSQ + AMP(I)**2
        ENDIF
        IF(HKL(3,I).LT.0) THEN            ! MAKE SURE HKL DATA HAS L>=0
          HKL(1,I) = -HKL(1,I)
          HKL(2,I) = -HKL(2,I)
          HKL(3,I) = -HKL(3,I)
          IF(SF_TYPE == 0) THEN
             PHI(I) = -PHI(I)
           ELSE 
             PHI(I) = 360.0 - PHI(I)
          ENDIF
        ENDIF
        IF(MOD(I,50000) == 0) WRITE(*,'(I8,A)') I, ' read in'
  100      ENDDO
C
      WRITE (*,'(I8,A)') nsf_in, ' read in'
      WRITE (*,'(A,F10.5)') ' Max resolution =', SF_DSTR_MAX
      IF(SF_DSTR_MIN == 0.0) THEN
        WRITE (*,'(A,F10.5)') ' Min resolution =', SF_DSTR_MIN
       ELSE
        WRITE (*,'(A)') ' Min resolution =   INFINITY'
      ENDIF
C
      call pif_close (sfhead%luNum)
      EM_ERR = 0                              ! SUCCESSFUL COMPLETION
C
      RETURN
      END

C  ********************************************************************
C  *                                                                  *
C  *           *****     SUBROUTINE SF_FLIPPHASE     *****            *
C  *                                                                  *
C  *    REVERSE PHASES WITHIN RANGE RES_MIN TO RES_MAX TO REVERSE     *
C  *     IMAGE CONTRAST.                                              *
C  *     PHI(NEW) = PHI(OLD) - 180                                    *
C  *                                                                  *
C  *    SEPTEMBER 1990  PURDUE UNIVERSITY VAX 8550, T. BAKER.         *
C  *    LAST MODIFIED 10-MAR-1992 BY T. BAKER.                        *
C  ********************************************************************
C
      SUBROUTINE SF_FLIPPHASE (SF_TYPE, NSF, AMP, PHI, DSTR,
     1 DSTR_MIN, DSTR_MAX)

      IMPLICIT NONE

      INTEGER*4 NSF, I, SF_TYPE
      REAL*4    AMP(NSF), PHI(NSF), DSTR(NSF), RES, DSTR_MIN, DSTR_MAX

      IF(SF_TYPE == 0) THEN                              ! APART, BPART
         DO I = 1,NSF
           RES = DSTR(I)
           IF(RES.GE.DSTR_MIN) THEN
             IF(RES.LE.DSTR_MAX) THEN
               AMP(I) = -AMP(I)
               PHI(I) = -PHI(I)
             ENDIF
           ENDIF
         ENDDO
       ELSE                                          ! AMP, PHASE
         DO I = 1,NSF
           RES = DSTR(I)
           IF(RES.GE.DSTR_MIN) THEN
             IF(RES.LE.DSTR_MAX) THEN
               PHI(I) = AMOD((PHI(I)+180.0),360.0)
             ENDIF
           ENDIF
         ENDDO
      ENDIF
C
      RETURN
      END


C  *********************************************************
C  *                                                       *
C  *         *****     SUBROUTINE SF_HKL     *****         *
C  *                                                       *
C  *    CONVERT FROM 2D TO 3D OR 3D TO 2D HKL INDICIES.    *
C  *********************************************************
C
      SUBROUTINE SF_HKL (NSF, HKL, IOPT, T_MATRIX)
C
      IMPLICIT NONE
C
      INTEGER*4 NSF, I, IOPT, T_MATRIX(2,3)
      INTEGER*4 HKL(3,NSF), HH, KK, LL
C  ----------------------------------------------------------------------------
C
      IF(IOPT == 1) THEN                  ! CONVERT 3D TO 2D
         DO I = 1,NSF
           CALL HKL_TO_HK (HKL(1,I),HKL(2,I),HKL(3,I),HH,KK,T_MATRIX)
           HKL(1,I) = HH
           HKL(2,I) = KK
         ENDDO
       ELSE IF(IOPT == -1) THEN
         DO I = 1,NSF
           CALL HK_TO_HKL (HKL(1,I),HKL(2,I),HH,KK,LL,T_MATRIX)
           HKL(1,I) = HH
           HKL(2,I) = KK
           HKL(3,I) = LL
         ENDDO
      ENDIF
C
      RETURN
      END

C  **************************************************************
C  *                                                            *
C  *           *****     SUBROUTINE SF_INIT     *****           *
C  *                                                            *
C  *    INITIALIZE COUNTERS, ETC.                               *
C  **************************************************************
C
      SUBROUTINE SF_INIT (NSF, NEW, NREJ, NBAD, NPAIR, FLAG)
C
      IMPLICIT NONE
C
      INTEGER*4 NSF, NEW, NREJ, NBAD, NPAIR, FLAG(NSF), I
C  ----------------------------------------------------------------------------
C
      NEW   = NSF
      NREJ  = 0
      NBAD  = 0
      NPAIR = 0
      DO I = 1,NSF
        FLAG(I) = 0
      ENDDO
C
      RETURN
      END

C  *************************************************************
C  *                                                           *
C  *            *****     SUBROUTINE SF_OPEN *****             *
C  *                                                           *
C  *    OPENS PIF SF FILE                                      *
C  *                                                           *
C  *    September 1998  PURDUE UNIVERSITY, RW Ashmore.         *
C  *************************************************************
C
      SUBROUTINE SF_OPEN (LUN, NSF, SF_TYPE, SF_FILENAME, 
     1 A_STAR, B_STAR, C_STAR, ALP_STAR, BET_STAR, GAM_STAR, 
     2 SF_SCALEF, A_CELL, 
     3 B_CELL, C_CELL, ALP, BET, GAM, EM_NDIM,
     4 EM_ERR, t_matrix, pixelSize, pixSizUnits)

      IMPLICIT NONE

      include '../include/pifhead.inc'
      TYPE(PIFHEAD_T)  MAPHEAD, IMGHEAD, PRJHEAD, PFTHEAD, SFHEAD
      COMMON /PIFHEAD/ MAPHEAD, IMGHEAD, PRJHEAD, PFTHEAD, SFHEAD

      INTEGER*4 NSF, LUN, SF_TYPE
      INTEGER*4 EM_NDIM, EM_ERR, T_MATRIX(2,3), pixSizUnits
      REAL*4    A_STAR, B_STAR, C_STAR, ALP_STAR, BET_STAR, GAM_STAR
      REAL*4    SF_SCALEF, A_CELL, B_CELL, C_CELL, ALP, BET, GAM
      REAL*4    pixelSize
      CHARACTER SF_FILENAME*64
      character(len=1) RW(2)
C  ----------------------------------------------------------------------------
C
      IF(SF_FILENAME == ' ') THEN
        PRINT *, '************************************************'
        PRINT *, '*         ENTER NAME OF SF FILE TO OPEN        *'
        PRINT *, '* "RETURN" or <CR> = RETURN to calling routine *'
        PRINT *, '* "EXIT"           = terminate program         *'
        PRINT *, '************************************************'
        WRITE(*,'(A)') 'FILENAME: '
        READ (*,'(A)') SF_FILENAME
      ENDIF
      IF(SF_FILENAME(1:4) == 'EXIT') CALL TSBEND
      IF(SF_FILENAME(1:4) == 'RETURN') RETURN
C
      sfhead%luNum = lun
      rw(1) = 'r'
      rw(2) = achar(0)
      call pif_open (sf_filename, sfhead%luNum, rw)
      call pif_read_gh (sfHead)
      if(sfhead%mode .ne. 31 .and. sfhead%mode .ne. 32) then
         print *, 'Invalid PIF mode for SF data.', sfhead%mode
         print *, '(Should be 31 or 32.)'
         em_err = 1
         return
      endif
      sfhead%imgNum = 1
      call pif_read_dh (sfHead)
      nsf = sfhead%numRows
      if(sfhead%mode == 31) then
         sf_type = 1            ! STORING AMPS/PHASES
      else
         sf_type = 0            ! STORING APART/BPART
      endif
      if(sfhead%numCols == 2) then
         em_ndim = 2
      else
         em_ndim = 3
      endif
      A_CELL        = SFHEAD%A_CELL
      B_CELL        = SFHEAD%B_CELL
      C_CELL        = SFHEAD%C_CELL
      ALP           = SFHEAD%ALP
      BET           = SFHEAD%BET
      GAM           = SFHEAD%GAM
      A_STAR        = SFHEAD%A_STAR
      B_STAR        = SFHEAD%B_STAR
      C_STAR        = SFHEAD%C_STAR
      ALP_STAR      = SFHEAD%ALP_STAR
      BET_STAR      = SFHEAD%BET_STAR
      GAM_STAR      = SFHEAD%GAM_STAR
      T_MATRIX(1,1) = SFHEAD%T_MATRIX(1)
      T_MATRIX(1,2) = SFHEAD%T_MATRIX(2)
      T_MATRIX(1,3) = SFHEAD%T_MATRIX(3)
      T_MATRIX(2,1) = SFHEAD%T_MATRIX(4)
      T_MATRIX(2,2) = SFHEAD%T_MATRIX(5)
      T_MATRIX(2,3) = SFHEAD%T_MATRIX(6)
      pixelSize     = SFHEAD%PIXELSIZE
      pixSizUnits   = SFHEAD%PIXSIZUNITS

      IF(SF_SCALEF == 0.0) SF_SCALEF = 1.0

      print *, 'a_star,b_star,c_star       - ', a_star,b_star,c_star
      print *, 'alp_star,bet_star,gam_star - ',
     1         ALP_STAR,BET_STAR,GAM_STAR
      print *, 'a_cell,b_cell,c_cell       - ',a_cell,b_cell,c_cell
      print *, 'alp,bet,gam                - ',alp,bet,gam
C
      EM_ERR = 0
      RETURN
C
      END

C  **********************************************************
C  *                                                        *
C  *        *****     SUBROUTINE SF_ORIGIN2     *****       *
C  *                                                        *
C  *    FIND BEST TWO-FOLD ORIGIN IN 2-D CRYSTAL LATTICE    *
C  *                                                        *
C  *    LAST MODIFIED 12-MAR-1992, BY T. BAKER.             *
C  **********************************************************
C
      SUBROUTINE SF_ORIGIN2 (EM_NDIM, SF_TYPE, NSF, HKL, AMP, PHI, 
     1 XSHFT, YSHFT, AMPSQ, SF_SUMFSQ)
C
      IMPLICIT NONE
C
      INTEGER*4 NSF, IR(11), I, IX, IY, EM_NDIM, SF_TYPE
      INTEGER*4 HKL(3,NSF), T_MATRIX(2,3)
      REAL*4    AMP(NSF), PHI(NSF), AMPSQ(NSF), RES1, P, PHI2
      REAL*4    XSHFT, YSHFT, ZSHFT, DEL, XSH(11), YSH(11), R(11)
      REAL*4    PHDIFF, XMIN, YMIN, RMIN, SF_SUMFSQ
      LOGICAL   ANS
C  ----------------------------------------------------------------------------
C
      CALL SF_HKL (NSF, HKL, 1, T_MATRIX)         ! APPLY T_MATRIX TO HKL DATA
      DEL = 0.0
      DO I = 1,NSF
        AMPSQ(I) = AMP(I)**2.0
      ENDDO
C
  100      CALL SF_SETSHIFT (XSHFT, YSHFT, DEL, RMIN, XSH, YSH)
      WRITE(*,'(/9X,A,/)') 'RESIDUAL MAP (ASSUMING TWO-FOLD AXIS)'
C
      DO IY = 1,11
        DO IX = 1,11
          RES1 = 0.0
          DO I = 1,NSF
            P = PHI(I) - (360.0*(HKL(1,I)*XSH(IX)+HKL(2,I)*YSH(IY)))
            P = AMOD(P,360.0)
            IF(P.LT.0.0) P = P + 360.0
            PHI2   = INT(P/180.0+0.5)*180.0
            PHDIFF = (P-PHI2)**2*AMPSQ(I)
            RES1   = RES1 + PHDIFF
          ENDDO
C
          RES1   = SQRT(RES1/SF_SUMFSQ)
          R(IX)  = RES1
          IR(IX) = INT(RES1 + 0.5)
          IF(RES1.LT.RMIN) THEN
            RMIN = RES1
            XMIN = XSH(IX)
            YMIN = YSH(IY)
          ENDIF
        ENDDO
        WRITE(*,'(10X,11I3)') (IR(IX),IX=1,11)
      ENDDO
      WRITE(*,1000) XMIN, YMIN, RMIN
C
      XSHFT = XMIN
      YSHFT = YMIN
      WRITE(*,'(/,A)') 'CONTINUE REFINEMENT? <Y>/N: '
      CALL GET_YESNO (ANS, .TRUE.)
      IF(ANS) GO TO 100
C  **********************************************
C  *    APPLY CALCULATED PHASE SHIFT TO DATA    *
C  **********************************************
      CALL SF_PSHIFT (EM_NDIM, SF_TYPE, NSF, HKL, AMP, PHI, XSHFT, 
     1 YSHFT, ZSHFT)
      CALL SF_HKL (NSF, HKL, -1, T_MATRIX)      ! CONVERT BACK TO ORIGINAL HKL
      RETURN
C
 1000      FORMAT(/' Minimum residual at XMIN,YMIN =',2F8.3,' is',F6.2)
      END

C  ********************************************************************
C  *                                                                  *
C  *             *****     SUBROUTINE SF_ORIGIN3     *****            *
C  *                                                                  *
C  *    FIND BEST THREE OR SIX-FOLD ORIGIN IN 2D CRYSTAL LATTICE      *
C  *    ADAPTED FROM ORIG3 AND ORIG6 OF MRC PROCESSING SYSTEM         *
C  *                                                                  *
C  *    LAST MODIFIED 12-MAR-1992, BY T. BAKER.                       *
C  ********************************************************************
C
      SUBROUTINE SF_ORIGIN3 (EM_NDIM, SF_TYPE, IOPT, NSF, HKL, AMP,
     1 PHI, XSHFT, YSHFT, P2)
C
      IMPLICIT NONE
C
      INTEGER*4 NSF, NX, NY, J, N, N1, M, EM_NDIM, SF_TYPE
      INTEGER*4 T_MATRIX(2,3)
      REAL*4    AMP(NSF), PHI(NSF), P2(NSF), XSHFT, YSHFT, ZSHFT
      REAL*4    DEL, XSH(11), YSH(11), R(11), DJ, DN, ANG
      REAL*4    XMIN, YMIN, RMIN, DIFF, P, D
      INTEGER*4 KJ, KN, HJ, HN, IR(11), IOPT, KOUNT, HKL(3,NSF)
      LOGICAL   ANS
C  ----------------------------------------------------------------------------
C
      CALL SF_HKL (NSF, HKL, 1, T_MATRIX)         ! APPLY T_MATRIX TO HKL DATA
      DEL   = 0.0
      ANG   = 0.0
   10      DIFF  = 0.0
      KOUNT = 0
      CALL SF_SETSHIFT (XSHFT, YSHFT, DEL, RMIN, XSH, YSH)
      PRINT *
      IF(IOPT == 3) THEN
         PRINT *, '        RESIDUAL MAP (ASSUMING THREE-FOLD AXIS)'
       ELSE IF(IOPT == 6) THEN
         PRINT *, '         RESIDUAL MAP (ASSUMING SIX-FOLD AXIS)'
      ENDIF
      PRINT *
C
      DO NY = 1,11
        DO NX = 1,11
          IF(KOUNT.NE.0) THEN
            KOUNT = 0
            DIFF  = 0.0
          ENDIF
          DO N = 1,NSF
            P = PHI(N)-360.0*(XSH(NX)*HKL(1,N)+YSH(NY)*HKL(2,N))
            P = AMOD(P,360.0)
            IF(P.LT.0.) P = P+360.0
            P2(N) = P
          ENDDO
C
          DO J = 1,(NSF-1)
            HJ = HKL(1,J)
            KJ = HKL(2,J)
            N1  = J+1
            DO 600 N = N1,NSF
              HN = HKL(1,N)
              KN = HKL(2,N)
              DJ = HJ*HJ+HJ*KJ+KJ*KJ
              DN = HN*HN+HN*KN+KN*KN
              IF(DJ.NE.DN) GO TO 600
              IF(DJ.NE.0.0) ANG = (HJ*HN+KJ*KN+0.5*(HN*KJ+HJ*KN))/DJ
              IF(IOPT == 6) GO TO 300
              IF(ANG == -0.5) GO TO 200
              IF(ANG == 0.5) GO TO 100
              GO TO 600
  100              D = P2(J)+P2(N)-360.0
              GO TO 500
  200              D = P2(J)-P2(N)
              GO TO 500
  300              ANG = ABS(ANG)
              IF(ANG == 0.5) GO TO 400
              GO TO 600
  400              D = P2(J)-P2(N)
  500              D = ABS(D)
              IF(D.GT.180.0) D = 360.0-D
              DIFF  = D+DIFF
              KOUNT = KOUNT+1
  600            CONTINUE
          ENDDO
          IF(KOUNT.GT.0) THEN
             R(NX) = DIFF/KOUNT
           ELSE
             R(NX) = 0.0
          ENDIF
          IR(NX) = INT(R(NX)+0.5)
          IF(RMIN.GE.R(NX)) THEN
            RMIN = R(NX)
            XMIN = XSH(NX)
            YMIN = YSH(NY)
          ENDIF
        ENDDO
        WRITE(*,'(10X,11I3)') (IR(M),M=1,11)
      ENDDO
      WRITE(*,1000) KOUNT, XMIN, YMIN, RMIN
      XSHFT = XMIN
      YSHFT = YMIN
      WRITE(*,'(/,A)') 'CONTINUE REFINEMENT? <Y>/N: '
      CALL GET_YESNO (ANS, .TRUE.)
      IF(ANS) GO TO 10
C  **********************************************
C  *    APPLY CALCULATED PHASE SHIFT TO DATA    *
C  **********************************************
      CALL SF_PSHIFT (EM_NDIM, SF_TYPE, NSF, HKL, AMP, PHI, XSHFT, 
     1 YSHFT, ZSHFT)
      CALL SF_HKL (NSF, HKL, -1, T_MATRIX)      ! CONVERT BACK TO ORIGINAL HKL
      RETURN
C
 1000      FORMAT(/' Average phase error based on',I4,' comparisons',/
     1 ' Minimum residual at XMIN,YMIN =',2F8.3,' is',F6.2)
      END

C  ************************************************************
C  *                                                          *
C  *         *****     SUBROUTINE SF_ORIGIN4     *****        *
C  *                                                          *
C  *    FIND BEST FOUR-FOLD ORIGIN IN 2D CRYSTAL LATTICE      *
C  *    ADAPTED FROM ORIG4 OF MRC PROCESSING SYSTEM           *
C  *                                                          *
C  *    LAST MODIFIED 12-MAR-1992, BY T. BAKER.               *
C  ************************************************************
C
      SUBROUTINE SF_ORIGIN4 (EM_NDIM, SF_TYPE, IFLAG, NSF, HKL, AMP,
     1 PHI, XSHFT, YSHFT, P2)
C
      IMPLICIT NONE
C
      INTEGER*4 NSF, IFLAG(NSF), I, N, J, EM_NDIM, SF_TYPE
      INTEGER*4 T_MATRIX(2,3)
      INTEGER*4 HJ, HN, HJN, KJ, KN, KJN, IR(11), KOUNT, NX, NY, N1
      INTEGER*4 HKL(3,NSF)
      REAL*4    AMP(NSF), PHI(NSF), P2(NSF), P, D
      REAL*4    XSHFT, YSHFT, ZSHFT, DEL, XSH(11), YSH(11), R(11)
      REAL*4    XMIN, YMIN, RMIN, DIFF
      LOGICAL   ANS
C  ----------------------------------------------------------------------------
C
      CALL SF_HKL (NSF, HKL, 1, T_MATRIX)         ! APPLY T_MATRIX TO HKL DATA
      DEL   = 0.0
   10      DIFF  = 0.0
      KOUNT = 0
      CALL SF_SETSHIFT (XSHFT, YSHFT, DEL, RMIN, XSH, YSH)
      WRITE(*,'(/,8X,A,/)') 'RESIDUAL MAP (ASSUMING FOUR-FOLD AXIS)'
C  ********************************
C  *    STORE HSQ+KSQ IN IFLAG    *
C  ********************************
      DO I = 1,NSF
        IFLAG(I) = HKL(1,I)*HKL(1,I)+HKL(2,I)*HKL(2,I)
      ENDDO
C
      DO NY = 1,11
        DO NX = 1,11
          IF(KOUNT.NE.0) THEN
            KOUNT = 0
            DIFF  = 0.0
          ENDIF
          DO N = 1,NSF
            P = PHI(N)-360.0*(XSH(NX)*HKL(1,N)+YSH(NY)*HKL(2,N))
            P = AMOD(P,360.0)
            IF(P.LT.0.) P = P+360.0
            P2(N) = P
          ENDDO
C
          DO J = 1,(NSF-1)
            HJ = HKL(1,J)
            KJ = HKL(2,J)
            N1  = J+1
            DO N = N1,NSF
              IF(IFLAG(J) == IFLAG(N)) THEN
                HN  = HKL(1,N)
                KN  = HKL(2,N)
                HJN = HJ*HN
                KJN = KJ*KN
                IF(HJN == -KJN) THEN
                  D = P2(J)-P2(N)
                  D = ABS(D)
                  IF(D.GT.180.0) D = 360.0-D
                  DIFF  = D+DIFF
                  KOUNT = KOUNT+1
                ENDIF
              ENDIF
            ENDDO
          ENDDO
C
          R(NX)  = DIFF/KOUNT
          IR(NX) = INT(R(NX)+0.5)
          IF(RMIN.GE.R(NX)) THEN
            RMIN = R(NX)
            XMIN = XSH(NX)
            YMIN = YSH(NY)
          ENDIF
        ENDDO
        WRITE(*,'(10X,11I3)') (IR(I),I=1,11)
      ENDDO
C
        WRITE(*,1000) KOUNT, XMIN, YMIN, RMIN
      XSHFT = XMIN
      YSHFT = YMIN
      WRITE(*,'(/,A)') 'CONTINUE REFINEMENT? <Y>/N: '
      CALL GET_YESNO (ANS, .TRUE.)
      IF(ANS) GO TO 10
C  **********************************************
C  *    APPLY CALCULATED PHASE SHIFT TO DATA    *
C  **********************************************
      CALL SF_PSHIFT (EM_NDIM, SF_TYPE, NSF, HKL, AMP, PHI, XSHFT, 
     1 YSHFT, ZSHFT)
      CALL SF_HKL (NSF, HKL, -1, T_MATRIX)      ! CONVERT BACK TO ORIGINAL HKL
      RETURN
C
 1000      FORMAT(/' Average phase error based on',I4,' comparisons',/
     1 ' Minimum residual at XMIN,YMIN =',2F8.3,' is',F6.2)
        END

C  **************************************************************
C  *                                                            *
C  *          *****     SUBROUTINE SF_PSHIFT     *****          *
C  *                                                            *
C  *    APPLY PHASE SHIFT TO SF DATA.                           *
C  *                                                            *
C  *    LAST MODIFIED 12-MAR-1992 BY T. BAKER.                  *
C  **************************************************************
C
      SUBROUTINE SF_PSHIFT (EM_NDIM, SF_TYPE, NSF, HKL, AMP, PHI, 
     1 XSHFT, YSHFT, ZSHFT_PASS)

      IMPLICIT NONE

      INTEGER*4 EM_NDIM, NSF, I, SF_TYPE
      INTEGER*4 HKL(3,NSF)
      REAL*4    AMP(NSF), PHI(NSF), XSHFT, YSHFT, ZSHFT, A, P
      REAL*4    PI, RADIAN, DEGREE, ZSHFT_PASS
      real*8    deg_to_rad

      deg_to_rad = acos(-1.0d0)/180.0d0

      CALL PIRADDEG (PI, RADIAN, DEGREE)

      IF(EM_NDIM == 2) THEN
         ZSHFT = 0.0
      ELSE
         ZSHFT = ZSHFT_PASS
      ENDIF

      IF(SF_TYPE == 0) THEN
         DO I = 1,NSF
           A = SQRT(AMP(I)**2+PHI(I)**2)
           IF(AMP(I) == 0.0) THEN
              P =  90.0
              IF(PHI(I).LT.0.0) P = 270.0
            ELSE
              P = ATAN2(PHI(I),AMP(I))/RADIAN
              P = AMOD(P,360.0)
              IF(P.LT.0.0) P = 360.0+P
           ENDIF
           P = P-360.0*(HKL(1,I)*XSHFT+HKL(2,I)*YSHFT+HKL(3,I)*ZSHFT)
           P = AMOD(P,360.0)
           IF(P.LT.0.0) P = P+360.0
           AMP(I) = REAL(A*COS(deg_to_rad*P))
           PHI(I) = REAL(A*SIN(deg_to_rad*P))
         ENDDO
       ELSE
         DO I = 1,NSF
           P = PHI(I) - 360.0 *
     1        (HKL(1,I)*XSHFT + HKL(2,I)*YSHFT + HKL(3,I)*ZSHFT)
           P = AMOD(P,360.0)
           IF(P.LT.0.0) P = P+360.0
           PHI(I) = P
         ENDDO
      ENDIF
C
      RETURN
      END

C  ***************************************************************
C  *                                                             *
C  *           *****     SUBROUTINE SF_RESCALE     *****         *
C  *                                                             *
C  *    MULTIPLY STRUCTURE FACTOR DATA BY A SCALE FACTOR.        *
C  *                                                             *
C  *    FEBRUARY 1990  PURDUE UNIVERSITY VAX 8550, T. BAKER.     *
C  *    LAST MODIFIED 27-FEB-1990 BY T. BAKER.                   *
C  ***************************************************************
C
      SUBROUTINE SF_RESCALE (SF_TYPE, SF_SCALEF, SCALEF, NSF, AMP, 
     1 PHI, SIG)

      IMPLICIT NONE

      INTEGER*4 NSF, I, SF_TYPE
      REAL*4  AMP(NSF), PHI(NSF), SIG(NSF), SF_SCALEF, SCALEF

      SF_SCALEF = SCALEF

      IF(SF_TYPE == 0) THEN
         DO I = 1,NSF
           AMP(I) = AMP(I)*SF_SCALEF            ! APART, BPART
           PHI(I) = PHI(I)*SF_SCALEF
           SIG(I) = SIG(I)*SF_SCALEF
         ENDDO
       ELSE
         DO I = 1,NSF
           AMP(I) = AMP(I)*SF_SCALEF            ! AMP, PHASE
           SIG(I) = SIG(I)*SF_SCALEF
         ENDDO
      ENDIF

      SF_SCALEF = 1.0                        ! RESET SCALE FACTOR TO 1.0

      RETURN
      END

C  *************************************************************************
C  *                                                                       *
C  *                *****     SUBROUTINE SF_RFACTOR     *****              *
C  *                                                                       *
C  *    COMPUTE SEVERAL DIFFERENT COMPARISONS BETWEEN TWO SF DATA SETS.    *
C  *     1. PHASE DIFFERENCE                                               *
C  *     2. STANDARD RFACTOR                                               *
C  *     3. R-FACTOR FOR A-PARTS ONLY                                      *
C  *     4. R-FACTOR FOR A- & B-PARTS COMBINED                             *
C  *     5. CORRELATION COEFFICIENT.                                       *
C  *                                                                       *
C  *    FEBRUARY 1990  PURDUE UNIVERSITY VAX 8550, T. BAKER.               *
C  *    MODIFIED  7-JAN-1992 BY T. BAKER.                                  *
C  *    MODIFIED 15-APR-1999 BY RWA (Added PIF read).                      *
C  *    LAST MODIFIED 28-DEC-1999 By TSB (Clean up and minor bugs).        *
C  *************************************************************************
C
      SUBROUTINE SF_RFACTOR (NSF,
     1 SF_TYPE, SF_DSTR_MAX, HKL, AMP, PHI, A_STAR, 
     2 B_STAR, C_STAR, ALP_STAR, BET_STAR, GAM_STAR, 
     3 A_CELL, B_CELL, C_CELL, ALP, BET, GAM)

      IMPLICIT NONE

      include '../include/pifhead.inc'
      TYPE(PIFHEAD_T)  MAPHEAD, IMGHEAD, PRJHEAD, PFTHEAD, SFHEAD
      COMMON /PIFHEAD/ MAPHEAD, IMGHEAD, PRJHEAD, PFTHEAD, SFHEAD

      INTEGER*4 NBINS_LIM, I, I1, NSF
      INTEGER*4 SF_TYPE, SF_TYPE1, SF_TYPE2
      PARAMETER (NBINS_LIM = 100)
      INTEGER*4 NTOT(NBINS_LIM)
      INTEGER*4 HKL(3,NSF), NH, NK, NL, NBINS, BIN
      REAL*4    AMP(NSF), PHI(NSF), SF_DSTR_MAX
      REAL*4    F1SUM(NBINS_LIM), F2SUM(NBINS_LIM), DENOM1, DENOM2
      REAL*8    F1SQSUM(NBINS_LIM), F2SQSUM(NBINS_LIM)
      REAL*8    F1_F2SUM(NBINS_LIM), F1F2SUM(NBINS_LIM)
      REAL*4    SUMA(NBINS_LIM), SUMB(NBINS_LIM)
      REAL*4    SUMDA(NBINS_LIM), SUMDB(NBINS_LIM)
      REAL*4    AVG_PDIF(NBINS_LIM), RFAC(NBINS_LIM)
      REAL*4    CCOEF(NBINS_LIM), AVGF1(NBINS_LIM), AVGF2(NBINS_LIM)
      REAL*4    A, B, AMP1, AMP2, SIG2, TMP1, TMP2, TMP3
      REAL*4    APART1, BPART1, APART2, BPART2, SCALEF
      REAL*4    PHASE1, PHASE2, PDIF_TEST, RFACA, RFACAB
      REAL*4    DSTR_LO, DSTR_HI, DSTR2, DSTR_MAX, DSTR_STEP
      REAL*4    PI, RADIAN, DEGREE, PDIF, PERCENT, DIV
      REAL*4    A_STAR, B_STAR, C_STAR, ALP_STAR, BET_STAR, GAM_STAR
      REAL*4    A_CELL, B_CELL, C_CELL, ALP, BET, GAM
      LOGICAL   ANS, NOTIFY
      CHARACTER FILENAME*64

      include '../include/strucfac.inc'

      integer alloc_stat
      type(STRUCFAC_T), allocatable :: STFACS(:)
      character(len=1) RW(2)
      INTEGER*4 THISI

C  ----------------------------------------------------------------------------
C

      PRINT *, '**********************************************'
      PRINT *, '* Compute R-factors between two SF data sets *'
      PRINT *, '**********************************************'
      NOTIFY = .TRUE.
C
      IF(SCALEF == 0.0) SCALEF = 1.0
      DSTR_MAX  = -HUGE(DSTR_MAX)
      DSTR_LO   = 0.0
      DSTR_HI   = SF_DSTR_MAX
      DSTR_STEP = SF_DSTR_MAX
C
      OPEN (UNIT=3, FILE='EMSF_RFAC.TMP', STATUS='NEW',
     1        ACCESS='SEQUENTIAL', FORM='FORMATTED')
      CALL PIRADDEG (PI, RADIAN, DEGREE)
      SF_TYPE1 = SF_TYPE
 100  WRITE(*,'(A)') 'SECOND SF DATASET FILENAME: '
      READ (*,'(A)',ERR=100) FILENAME
      SFHEAD%LUNUM  = 12
      RW(1)  = 'r'
      RW(2) = achar(0)
      CALL PIF_OPEN (FILENAME, SFHEAD%LUNUM, RW) 
      CALL PIF_READ_GH (SFHEAD)
      SFHEAD%IMGNUM = 1
      CALL PIF_READ_DH (SFHEAD)
C
      A_STAR   = SFHEAD%A_STAR
      B_STAR   = SFHEAD%B_STAR
      C_STAR   = SFHEAD%C_STAR
      ALP_STAR = SFHEAD%ALP_STAR
      BET_STAR = SFHEAD%BET_STAR
      GAM_STAR = SFHEAD%GAM_STAR
      ALP      = SFHEAD%ALP
      BET      = SFHEAD%BET
      GAM      = SFHEAD%GAM
      A_CELL   = SFHEAD%A_CELL
      B_CELL   = SFHEAD%B_CELL
      C_CELL   = SFHEAD%C_CELL
      IF(SFHEAD%MODE == 31) THEN
         SF_TYPE = 1            ! STORING AMPS/PHASES
      ELSE
         SF_TYPE = 0            ! STORING APART/BPART
      ENDIF
      SF_TYPE2 = SF_TYPE

  200      PRINT *, '***********************************************'
      PRINT *, '* Enter resolution range of data for R-factor *'
      PRINT *, '*  calculation and resolution step size.      *'
      PRINT *, '***********************************************'
 300  WRITE(*,'(A,F8.5,A,F8.5,A,F8.5,A)')
     1     'DSTR_LO,DSTR_HI,DSTR_STEP <CR>=', DSTR_LO, ',',
     2     DSTR_HI, ',', DSTR_STEP, '): '
      READ (*,*,ERR=300) TMP1, TMP2, TMP3
      IF(TMP1.GT.0.0) DSTR_LO   = TMP1
      IF(TMP2.GT.0.0) DSTR_HI   = TMP2
      IF(TMP3.GT.0.0) DSTR_STEP = TMP3
      IF(DSTR_LO.GT.DSTR_HI) THEN
        PRINT *, '*********************************'
        PRINT *, '* DSTR_MIN cannot be > DSTR_MAX *'
        PRINT *, '*********************************'
        GO TO 300
      ENDIF
      NBINS = NINT ((DSTR_HI-DSTR_LO)/DSTR_STEP)
      IF(NBINS.LE.0) NBINS = 1
      IF(NBINS.GT.NBINS_LIM) THEN
        PRINT *, '*******************************'
        PRINT *, '* DSTR_STEP must be increased *'
        PRINT *, '*******************************'
        GO TO 300
      ENDIF
  400      WRITE(*,'(2A,F10.5,A)') 'LINEAR SCALE FACTOR FOR 2ND DATA',
     1 ' SET (<CR>=', SCALEF, '): '
      READ (*,*,ERR=400) TMP1
      IF(TMP1.GT.0.0) SCALEF = TMP1

 500  WRITE(*,'(A)') 'THRESHOLD FOR PHASE DIFFERENCES (<CR>=180.0): '
      READ (*,*,ERR=500) PDIF_TEST
      IF(PDIF_TEST.LE.0.0) PDIF_TEST = 180.0
C
      I1 = 1
      DO I = 1,NBINS                        ! INITIALIZE ARRAYS
        AVG_PDIF(I) = 0.0
        F1_F2SUM(I) = 0.0
        F1SUM(I)    = 0.0
        F2SUM(I)    = 0.0
        F1F2SUM(I)  = 0.0
        F1SQSUM(I)  = 0.0
        F2SQSUM(I)  = 0.0
        RFAC(I)     = 0.0
        NTOT(I)     =   0
        SUMA(I)     = 0.0
        SUMB(I)     = 0.0
        SUMDA(I)    = 0.0
        SUMDB(I)    = 0.0
      ENDDO
C
      IF(NOTIFY) WRITE(*,'(F6.1,A)') 0.0, '% done'
      DIV      = REAL(NSF)

      allocate(stfacs(SFHEAD%NUMROWS), stat=alloc_stat)
      if(alloc_stat > 0) then
         write(*,*) 'Unable to allocate STFACS'
         call tsbend
      endif

      CALL PIF_READ_SF (SFHEAD, STFACS)
      THISI = 0
  600      THISI = THISI + 1
      IF(THISI .GT. SFHEAD%NUMROWS) GO TO 700
      NH    = stFacs(thisI)%h
      NK    = stFacs(thisI)%k
      NL    = stFacs(thisI)%l
      A     = stFacs(thisI)%amp
      B     = stFacs(thisI)%phi
      SIG2  = stFacs(thisI)%sig
      DSTR2 = stFacs(thisI)%dstr
C      
      IF(DSTR2.GT.DSTR_HI .OR. DSTR2.LT.DSTR_LO) GO TO 600
      DSTR_MAX = MAX(DSTR_MAX,DSTR2)
C       *******************************************************
C       *    SEARCH FOR SAME REFLECTION (INCLUDING FRIEDEL)   *
C       *******************************************************
      DO I = I1,NSF
C         ************************************
C---------*    SEARCH FOR SAME REFLECTION    *
C         ************************************
        IF(NH == HKL(1,I).AND.NK == HKL(2,I).AND.NL == HKL(3,I)) THEN
          BIN = INT((DSTR2-DSTR_LO)/DSTR_STEP) + 1
          IF(BIN.GT.NBINS) BIN = NBINS
          CALL SF_APAB (SF_TYPE1, AMP(I), PHI(I), AMP1, PHASE1,
     1                 APART1, BPART1, RADIAN)
          CALL SF_APAB (SF_TYPE2, A, B, AMP2, PHASE2,
     1                 APART2, BPART2, RADIAN)
          PDIF   = ABS(PHASE1 - PHASE2)
          IF(PDIF.GT.180.0) PDIF = 360.0 - PDIF
          IF(PDIF.LE.PDIF_TEST) THEN
            AVG_PDIF(BIN) = AVG_PDIF(BIN) + PDIF
            F1_F2SUM(BIN) = F1_F2SUM(BIN) + ABS(AMP1 - AMP2*SCALEF)
            F1SUM(BIN)    = F1SUM(BIN) + AMP1
            F2SUM(BIN)    = F2SUM(BIN) + AMP2*SCALEF
            F1SQSUM(BIN)  = F1SQSUM(BIN) + AMP1*AMP1
            F2SQSUM(BIN)  = F2SQSUM(BIN) + (AMP2*SCALEF)**2
            F1F2SUM(BIN)  = F1F2SUM(BIN) + AMP1*AMP2*SCALEF
            NTOT(BIN)     = NTOT(BIN) + 1
          ENDIF
          APART2     = APART2*SCALEF
          BPART2     = BPART2*SCALEF
            SUMA(BIN)  = SUMA(BIN)  + ABS(APART1 + APART2)
            SUMB(BIN)  = SUMB(BIN)  + ABS(BPART1 + BPART2)
          SUMDA(BIN) = SUMDA(BIN) + ABS(APART1 - APART2)
          SUMDB(BIN) = SUMDB(BIN) + ABS(BPART1 - BPART2)
          I1         = I
          IF(NOTIFY) THEN
            IF(MOD(I,1000) == 0) THEN
              PERCENT = 100.0*REAL(I)/DIV
              WRITE(*,'(A,F5.1,A)') '+', PERCENT, '% done'
            ENDIF
          ENDIF
          GO TO 600
        ENDIF
C         ******************************************************
C---------*    SEARCH FOR FRIEDEL MATES BETWEEN TWO SF SETS    *
C         ******************************************************
        IF(-NH == HKL(1,I).AND.-NK == HKL(2,I).AND.-NL == HKL(3,I))
     1   THEN
          BIN = INT((DSTR2-DSTR_LO)/DSTR_STEP) + 1
          IF(BIN.GT.NBINS) BIN = NBINS
          CALL SF_APAB (SF_TYPE1, AMP(I), PHI(I), AMP1, PHASE1,
     1                 APART1, BPART1, RADIAN)
          CALL SF_APAB (SF_TYPE2, A, B, AMP2, PHASE2,
     1                 APART2, BPART2, RADIAN)
          PDIF   = ABS(PHASE1 - PHASE2)
          IF(PDIF.GT.180.0) PDIF = 360.0 - PDIF
          IF(PDIF.LE.PDIF_TEST) THEN
            AVG_PDIF(BIN) = AVG_PDIF(BIN) + PDIF
            F1_F2SUM(BIN) = F1_F2SUM(BIN) + ABS(AMP1 - AMP2*SCALEF)
            F1SUM(BIN)    = F1SUM(BIN) + AMP1
            F2SUM(BIN)    = F2SUM(BIN) + AMP2*SCALEF
            F1SQSUM(BIN)  = F1SQSUM(BIN) + AMP1*AMP1
            F2SQSUM(BIN)  = F2SQSUM(BIN) + (AMP2*SCALEF)**2
            F1F2SUM(BIN)  = F1F2SUM(BIN) + AMP1*AMP2*SCALEF
            NTOT(BIN)     = NTOT(BIN) + 1
          ENDIF
          APART2     = APART2*SCALEF
          BPART2     = BPART2*SCALEF
            SUMA(BIN)  = SUMA(BIN)  + ABS(APART1 + APART2)
            SUMB(BIN)  = SUMB(BIN)  + ABS(BPART1 - BPART2)
          SUMDA(BIN) = SUMDA(BIN) + ABS(APART1 - APART2)
          SUMDB(BIN) = SUMDB(BIN) + ABS(BPART1 + BPART2)
          I1         = I
          IF(NOTIFY) THEN
            IF(MOD(I,1000) == 0) THEN
              PERCENT = 100.0*REAL(I)/DIV
              WRITE(*,'(A,F5.1,A)') '+', PERCENT, '% done'
            ENDIF
          ENDIF
          GO TO 600
        ENDIF
      ENDDO
      GO TO 600
C
  700      WRITE(*,'(T43,A,T76,A,/,2A,/,2A)') 'AVG', 'CC',
     1         '  DSTAR    # SFs    AVGF1     AVGF2      P-dif',
     2         '     R-fac    R-fA   R-fAB  Coef',
     3         ' ------- -------  --------   -------    ------',
     4         '    ------  ------  ------  ----'
      WRITE(3,'(T43,A,T76,A,/,2A,/,2A)') 'AVG','CC',
     1         '  DSTAR    # SFs    AVGF1     AVGF2      P-dif',
     2         '     R-fac    R-fA   R-fAB  Coef',
     3         ' ------- -------  --------   -------    ------',
     4         '    ------  ------  ------  ----'
      DO I = 1,NBINS
        CCOEF(I) = 1.0
        AVGF1(I) = F1SUM(I)
        AVGF2(I) = F2SUM(I)
        IF(NTOT(I).GT.0) THEN
          RFAC(I)     = REAL(100.0*F1_F2SUM(I)/F1SUM(I))
          AVG_PDIF(I) = AVG_PDIF(I)/REAL(NTOT(I))
          AVGF1(I)    = AVGF1(I)/NTOT(I)
          AVGF2(I)    = AVGF2(I)/NTOT(I)
          DENOM1      = REAL(F1SQSUM(I) - (F1SUM(I)**2)/NTOT(I))
          DENOM2      = REAL(F2SQSUM(I) - (F2SUM(I)**2)/NTOT(I))
          IF(DENOM1.GT.0.0 .AND. DENOM2.GT.0.0) THEN
            CCOEF(I) = REAL((F1F2SUM(I) - (F1SUM(I)*F2SUM(I)/NTOT(I)))/
     1                      ( SQRT(DENOM1) * SQRT(DENOM2) ))
          ENDIF
        ENDIF
        RFACA  = 0.0
        RFACAB = 0.0
        IF(SUMA(I).GT.0.0) RFACA = (SUMDA(I)/SUMA(I))*100.0
        IF(SUMA(I).GT.0.0 .OR. SUMB(I).GT.0.0) THEN
            RFACAB = 100.0*(SQRT ((SUMDA(I)**2) + (SUMDB(I)**2)) /
     1                   SQRT ((SUMA(I)**2) + (SUMB(I)**2)) )
        ENDIF
        WRITE(*,1000) DSTR_STEP*I+DSTR_LO, NTOT(I), AVGF1(I), 
     1  AVGF2(I), AVG_PDIF(I), RFAC(I), RFACA, RFACAB, CCOEF(I)
        WRITE(3,1000) DSTR_STEP*I+DSTR_LO, NTOT(I), AVGF1(I), 
     1  AVGF2(I), AVG_PDIF(I), RFAC(I), RFACA, RFACAB, CCOEF(I)
      ENDDO
      WRITE(*,'(A)') 'RESET DSTR_LO,DSTR_HI,DSTR_STEP ? Y/<N>: '
      CALL GET_YESNO (ANS, .FALSE.)
      IF(ANS) THEN
        deallocate(STFACS)
        DSTR_MAX  = -HUGE(DSTR_MAX)
        GO TO 200
      ENDIF
C
      CLOSE (UNIT=3, STATUS='KEEP')
      CALL PIF_CLOSE (SFHEAD%LUNUM)
      CALL HITCR
      RETURN
C
 1000      FORMAT (F8.4,I8,2F10.0,2F10.2,2F8.2,F6.2)
      END

C  ********************************************************************
C  *                                                                  *
C  *              *****     SUBROUTINE SF_SCALEFAC    *****           *
C  *                                                                  *
C  *    DETERMINE LINEAR SCALE FACTOR BETWEEN TWO DATA SETS           *
C  *                                                                  *
C  *    FEBRUARY 1990  PURDUE UNIVERSITY VAX 8550, T. BAKER.          *
C  *    MODIFIED 17-DEC-1998 BY RWA (Added PIF Input).                *
C  *    LAST MODIFIED 30-DEC-1999 BY TSB (Keep track of SCALEF).      *
C  ********************************************************************
C
      SUBROUTINE SF_SCALEFAC (NSF,
     1 SF_TYPE, HKL, AMP, PHI, SF_DSTR_MAX, A_CELL,
     2 B_CELL, C_CELL, ALP, BET, GAM, A_STAR, B_STAR, C_STAR,
     3 ALP_STAR, BET_STAR, GAM_STAR, SCALEF)

      IMPLICIT NONE

      INTEGER*4 LUN, I1, NSF, I
      INTEGER*4 NBINS_LIM, SF_TYPE, SF_TYPE1, SF_TYPE2
      PARAMETER (NBINS_LIM = 100)
      INTEGER*4 NTOT(NBINS_LIM)
      INTEGER*4 HKL(3,NSF), NH, NK, NL, NBINS, BIN
      REAL*4    AMP(NSF), PHI(NSF), SF_DSTR_MAX
      REAL*4    F1SUM(NBINS_LIM), F2SUM(NBINS_LIM)
      REAL*4    AMP2, PHI2, SIG2, TMP1, TMP2, TMP3
      REAL*4    SCALEF
      REAL*4    PI, RADIAN, DEGREE, PERCENT, DIV
      REAL*4    DSTR_LO, DSTR_HI, DSTR2, DSTR_MAX, DSTR_STEP
      REAL*4    A_STAR, B_STAR, C_STAR, ALP_STAR, BET_STAR, GAM_STAR
      REAL*4    A_CELL, B_CELL, C_CELL, ALP, BET, GAM
      CHARACTER FILENAME*64

      integer alloc_stat
      include '../include/strucfac.inc'
      type(STRUCFAC_T), allocatable :: STFACS(:)

      include '../include/pifhead.inc'
      TYPE(PIFHEAD_T)  MAPHEAD, IMGHEAD, PRJHEAD, PFTHEAD, SFHEAD
      COMMON /PIFHEAD/ MAPHEAD, IMGHEAD, PRJHEAD, PFTHEAD, SFHEAD

      character(len=1) RW(2)
      INTEGER*4 THISI
      LOGICAL   NOTIFY

C  ----------------------------------------------------------------------------

      NOTIFY = .TRUE.
      PRINT *, '*********************************'
      PRINT *, '* Determine linear scale factor *'
      PRINT *, '*  between two SF data sets     *'
      PRINT *, '*********************************'
      DSTR_MAX  = -HUGE(DSTR_MAX)
      DSTR_LO   = 0.0
      DSTR_HI   = SF_DSTR_MAX
      DSTR_STEP = SF_DSTR_MAX
C
      CALL PIRADDEG (PI, RADIAN, DEGREE)
      CLOSE (UNIT=1, STATUS='KEEP')
      SF_TYPE1 = SF_TYPE
 200  WRITE(*,'(A)') 'SECOND SF DATASET FILENAME: '
      READ (*,*,ERR=200) FILENAME
      LUN          = 1
      sfhead%luNum = 12
      RW(1) = 'r'
      RW(2) = achar(0)
      call pif_open (filename, sfhead%luNum, rw) 

      call pif_read_gh(sfHead)
      sfhead%imgNum = 1
      call pif_read_dh(sfHead)

      a_star = sfhead%a_star
      b_star = sfhead%b_star
      c_star = sfhead%c_star
      alp_star = sfhead%alp_star
      bet_star = sfhead%bet_star
      gam_star = sfhead%gam_star
      alp = sfhead%alp
      bet = sfhead%bet
      gam = sfhead%gam
      a_cell = sfhead%a_cell
      b_cell = sfhead%b_cell
      c_cell = sfhead%c_cell
      if(sfhead%mode == 31) then
         sf_type = 1            ! STORING AMPS/PHASES
      else
         sf_type = 0            ! STORING APART/BPART
      endif

      SF_TYPE2 = SF_TYPE
      PRINT *, '*****************************'
      PRINT *, '* Enter range of data to be *'
      PRINT *, '*  scaled and bin size      *'
      PRINT *, '*****************************'
  300      WRITE(*,'(2A,F8.5,A,F8.5,A)') 'DSTR_MIN,DSTR_MAX,DSTR_STEP',
     1 ' (<CR>=0.0,', SF_DSTR_MAX, ',', SF_DSTR_MAX, '): '
      READ (*,*,ERR=300) TMP1, TMP2, TMP3
      IF(TMP1.GT.0.0) DSTR_LO   = TMP1
      IF(TMP2.GT.0.0) DSTR_HI   = TMP2
      IF(TMP3.GT.0.0) DSTR_STEP = TMP3
      IF(DSTR_LO.GT.DSTR_HI) THEN
        CALL BELL (2)
        PRINT *, '*********************************'
        PRINT *, '* DSTR_MIN cannot be > DSTR_MAX *'
        PRINT *, '*********************************'
        GO TO 300
      ENDIF
      NBINS = NINT ((DSTR_HI-DSTR_LO)/DSTR_STEP)
      IF(NBINS.LE.0) NBINS = 1
      IF(NBINS.GT.NBINS_LIM) THEN
        CALL BELL (2)
        PRINT *, '*******************************'
        PRINT *, '* DSTR_STEP must be increased *'
        PRINT *, '*******************************'
        GO TO 300
      ENDIF
C
        I1       = 1
      DO I = 1,NBINS                        ! INITIALIZE ARRAYS
        F1SUM(I) = 0.0
        F2SUM(I) = 0.0
        NTOT(I)  =   0
      ENDDO
C
      IF(NOTIFY) THEN
        PRINT *, '***********************************************'
        PRINT *, '* Compute scale factor between two data sets. *'
        PRINT *, '*                                             *'
        PRINT *, '* NOTE: If SF data are not sorted properly,   *'
        PRINT *, '*       results will probably be incorrect.   *'
        PRINT *, '***********************************************'
        WRITE(*,'(F6.1,A)') 0.0, '% done'
      ENDIF
Ca  400      IF(FILE_ID == 2) THEN            ! READ STRUCTURE FACTOR DATA FROM 2ND FILE
Ca         READ(1,'(3I5,3F10.4,F10.5)',END=500) NH, NK, NL, AMP2, PHI2,
Ca      1                                       SIG2, DSTR2
Ca       ELSE
Ca         READ(1,END=500) NH, NK, NL, AMP2, PHI2, SIG2, DSTR2
Ca      ENDIF

      allocate(stfacs(sfhead%numRows), stat=alloc_stat)
      if(alloc_stat > 0) then
         write(*,*) 'Unable to allocate STFACS'
         call tsbend
      endif

      call pif_read_sf (sfHead,stFacs)

      thisI = 0
  400      thisI = thisI + 1
      if(thisI .gt. sfhead%numRows) goto 500
      NH = stFacs(thisI)%h
      NK = stFacs(thisI)%k
      NL = stFacs(thisI)%l
      AMP2  = stFacs(thisI)%amp
      PHI2  = stFacs(thisI)%phi
      SIG2  = stFacs(thisI)%sig
      DSTR2 = stFacs(thisI)%dstr

      IF(DSTR2.GT.DSTR_HI .OR. DSTR2.LT.DSTR_LO) GO TO 400
      DSTR_MAX = MAX(DSTR_MAX,DSTR2)
C       *******************************************************
C       *    SEARCH FOR SAME REFLECTION (INCLUDING FRIEDEL)   *
C       *******************************************************
      DIV = REAL(NSF)
      DO I = I1,NSF
C         ************************************
C---------*    SEARCH FOR SAME REFLECTION    *
C         ************************************
        IF(NH == HKL(1,I).AND.NK == HKL(2,I).AND.NL == HKL(3,I)) THEN
          BIN = INT((DSTR2-DSTR_LO)/DSTR_STEP) + 1
          IF(BIN.GT.NBINS) BIN = NBINS
          IF(SF_TYPE1 == 0) THEN
             F1SUM(BIN) = F1SUM(BIN) + SQRT (AMP(I)**2+PHI(I)**2)
           ELSE
             F1SUM(BIN) = F1SUM(BIN) + AMP(I)
          ENDIF
          IF(SF_TYPE2 == 0) THEN
             F2SUM(BIN) = F2SUM(BIN) + SQRT (AMP2**2+PHI2**2)
           ELSE
             F2SUM(BIN) = F2SUM(BIN) + AMP2
          ENDIF
          NTOT(BIN) = NTOT(BIN) + 1
          I1        = I
          IF(NOTIFY) THEN
            IF(MOD(I,1000) == 0) THEN
              PERCENT = 100.0*REAL(I)/DIV
              WRITE(*,'(A,F5.1,A)') '+', PERCENT, '% done'
            ENDIF
          ENDIF
          GO TO 400
        ENDIF
C         ******************************************************
C---------*    SEARCH FOR FRIEDEL MATES BETWEEN TWO SF SETS    *
C         ******************************************************
        IF(-NH == HKL(1,I).AND.-NK == HKL(2,I).AND.-NL == HKL(3,I))
     1   THEN
          BIN = INT((DSTR2-DSTR_LO)/DSTR_STEP) + 1
          IF(BIN.GT.NBINS) BIN = NBINS
          IF(SF_TYPE1 == 0) THEN
             F1SUM(BIN) = F1SUM(BIN) + SQRT (AMP(I)**2+PHI(I)**2)
           ELSE
             F1SUM(BIN) = F1SUM(BIN) + AMP(I)
          ENDIF
          IF(SF_TYPE2 == 0) THEN
             F2SUM(BIN) = F2SUM(BIN) + SQRT (AMP2**2+PHI2**2)
           ELSE
             F2SUM(BIN) = F2SUM(BIN) + AMP2
          ENDIF
          NTOT(BIN) = NTOT(BIN) + 1
          I1        = I
          IF(NOTIFY) THEN
            IF(MOD(I,1000) == 0) THEN
              PERCENT = 100.0*REAL(I)/DIV
              WRITE(*,'(A,F5.1,A)') '+', PERCENT, '% done'
            ENDIF
          ENDIF
          GO TO 400
        ENDIF
      ENDDO
      GO TO 400
C
  500      SCALEF = 1.0
      WRITE(*,1000)
      DO I = 1,NBINS
        DSTR2 = I*DSTR_STEP + DSTR_LO
        IF(NTOT(I).GT.0) SCALEF = F1SUM(I)/F2SUM(I)
        WRITE(*,1010) DSTR2, F1SUM(I), F2SUM(I), SCALEF, NTOT(I)
      ENDDO
C
      CLOSE (UNIT=1, STATUS='KEEP')
      CALL HITCR
      RETURN
C
 1000      FORMAT('    DSTR      F1 SUM    F2 SUM   SCALE FACTOR 
     1  # COMPARISONS')
 1010      FORMAT(1X,F8.4,2E12.4,F10.5,I13)
      END

C  **********************************************************
C  *                                                        *
C  *       *****     SUBROUTINE SF_SETSHIFT     *****       *
C  *                                                        *
C  *    READ STARTING ORIGIN SHIFT AND INCREMENT            *
C  *     XSHFT,YSHFT = STARTING PHASE ORIGIN SHIFT          *
C  *     DEL = SIZE OF X AND Y INCREMENTS                   *
C  **********************************************************
C
      SUBROUTINE SF_SETSHIFT (XSHFT, YSHFT, DEL, RMIN, XSH, YSH)
C
      IMPLICIT NONE
C
      REAL*4    XSHFT, YSHFT, DEL, XSH(11), YSH(11), RMIN, TMP
      INTEGER*4 I
C  ----------------------------------------------------------------------------
C
      RMIN = 1.0E20
      IF(DEL == 0.0) THEN
         WRITE(*,'(A)') 'XSHFT,YSHFT,DSHFT <CR>=0.0,0.0,0.05: '
         READ (*,*) XSHFT, YSHFT, DEL
         IF(DEL.LE.0.0) DEL = 0.05
       ELSE
         TMP = DEL/2.0
         WRITE(*,'(A,F8.5,A)') 'DSHFT <CR> = ', TMP, ': '
         READ (*,*) DEL
         IF(DEL.LE.0.0) DEL = TMP
      ENDIF
C
      DO I = 1,11
        XSH(I) = XSHFT + (I-6)*DEL
        YSH(I) = YSHFT + (I-6)*DEL
      ENDDO
C
      RETURN
      END


      SUBROUTINE SF_STATS (NSF, AMP, DSTR, NBINS, SF_DSTR_MAX,
     $     SF_SCALEF)
C
      IMPLICIT NONE
C

      INTEGER*4 NSF, NBINS, I, J
      REAL*4    AMP(NSF), DSTR(NSF)
      REAL*4    DEL_RES, SF_DSTR_MAX, SF_SCALEF

      integer alloc_stat
      integer, allocatable :: count(:)
      real*4, allocatable :: avg(:)

C  ----------------------------------------------------------------------------
C
      allocate(count(nbins+1), stat=alloc_stat)
      allocate(avg(nbins+1), stat=alloc_stat)

      DEL_RES = SF_DSTR_MAX/REAL(NBINS-1)
C  *********************************************
C  *    INITIALIZE "COUNT" AND "AVG" ARRAYS    *
C  *********************************************
      DO I = 1,NBINS+1
        COUNT(I) =   0
        AVG(I)   = 0.0
      ENDDO
C
      DO I = 1,NSF
        J        = NINT(DSTR(I)/DEL_RES) + 1
        AVG(J)   = AVG(J) + AMP(I)*SF_SCALEF
        COUNT(J) = COUNT(J) + 1
      ENDDO
C  *******************************
C  *    NORMALIZE THE RESULTS    *
C  *******************************
      DO I = 1,NBINS+1
        IF(COUNT(I).GT.0) AVG(I) = AVG(I)/COUNT(I)
        WRITE(*,'(F10.4,F12.2)') (I-1)*DEL_RES, AVG(I)
      ENDDO
C
      deallocate(count)
      deallocate(avg)

      RETURN
      END

C  ***********************************************************
C  *                                                         *
C  *        *****     SUBROUTINE SF_SYM_P2     *****         *
C  *                                                         *
C  *    CONSTRAIN PHASES TO BE REAL.                         *
C  *                                                         *
C  *    JUNE 1986  PURDUE UNIVERSITY VAX 750, T. BAKER.      *
C  *    MARCH 1992  PURDUE UNIVERSITY VAX 8550, T. BAKER.    *
C  *    LAST MODIFIED 12-MAR-1992 BY T. BAKER.               *
C  ***********************************************************
C
      SUBROUTINE SF_SYM_P2 (SF_TYPE, NSF, AMP, PHI)
C
      IMPLICIT NONE
C
      INTEGER*4 NSF, I, NCOUNT, SF_TYPE
      REAL*4    AMP(NSF), PHI(NSF), P, SUMSQ, SUMSQ1, SUMSQ2, F
      REAL*4    PHIDIF, RMS, POW, PI, RADIAN, DEGREE
      real*8    deg_to_rad
C  ----------------------------------------------------------------------------
C
      deg_to_rad = acos(-1.0d0)/180.0d0
      
      PRINT *, '*********************'
      PRINT *, '* Apply p2 symmetry *'
      PRINT *, '*********************'
C
      CALL PIRADDEG (PI, RADIAN, DEGREE)
      SUMSQ  = 0.0
      SUMSQ1 = 0.0
      SUMSQ2 = 0.0
      NCOUNT =   0
C
      IF(SF_TYPE == 0) THEN                  ! APART/BPART STRUCTURE FACTORS
         DO I = 1,NSF
           IF(AMP(I).NE.0.0 .AND. PHI(I).NE.0.0) THEN
             NCOUNT = NCOUNT + 1
             SUMSQ1 = SUMSQ1 + AMP(I)**2 + PHI(I)**2
             SUMSQ2 = SUMSQ2 + AMP(I)**2
             PHIDIF = 0.0
             IF(AMP(I).NE.0.0) THEN
                PHIDIF = ABS( ATAN2(PHI(I),AMP(I))*DEGREE )
              ELSE IF(PHI(I).NE.0.0) THEN
                PHIDIF = 90.0
             ENDIF
             SUMSQ  = SUMSQ + PHIDIF*PHIDIF
             PHI(I) = 0.0
           ENDIF
         ENDDO
       ELSE                                ! AMP/PHASE STRUCTURE FACTORS
         DO I = 1,NSF
           IF(AMP(I).NE.0.0) THEN
              NCOUNT = NCOUNT + 1
              P      = PHI(I)
              SUMSQ1 = SUMSQ1 + AMP(I)*AMP(I)
              F      = REAL(COS(deg_to_rad*P)*AMP(I))
              P      = 0.0
              IF(F.LT.0.0) P = 180.0
              PHIDIF = ABS(P-PHI(I))
              IF(PHIDIF.GT.180.0) PHIDIF = 360.0 - PHIDIF
              SUMSQ  = SUMSQ + PHIDIF*PHIDIF
              SUMSQ2 = SUMSQ2 + F*F
              AMP(I) = ABS(F)
              PHI(I) = P
            ELSE
              AMP(I) = 0.0
              PHI(I) = 0.0
           ENDIF
         ENDDO
      ENDIF
C
      RMS = SQRT(SUMSQ/NCOUNT)
      POW = (1.0-(SUMSQ2/SUMSQ1))*100.0
      WRITE(*,1000) NCOUNT, RMS, POW
Ca      CALL HITCR
      RETURN
C
 1000 FORMAT(I5,' non-zero weight spots with RMS phase diff =',F8.1,
     1     ' degrees',/,' Total power reduced',F6.2,'%')
      END

C  ***********************************************************
C  *                                                         *
C  *        *****     SUBROUTINE SF_SYM_PGG     *****        *
C  *                                                         *
C  *    CONSTRAIN SF DATA TO PLANE GROUP PGG.                *
C  *                                                         *
C  *    JUNE 1986  PURDUE UNIVERSITY VAX 750, T. BAKER.      *
C  *    MARCH 1992  PURDUE UNIVERSITY VAX 8550, T. BAKER.    *
C  *    LAST MODIFIED 13-MAR-1992 BY T. BAKER.               *
C  *                                                         *
C  *    PLEASE NOTE: THIS SUBPROGRAM IS UNTESTED.            *
C  ***********************************************************
C
      SUBROUTINE SF_SYM_PGG (SF_TYPE, NLIM, NSF, HKL, AMP, PHI, SIG, 
     1 DSTR, FLAG, GAM_STAR)
C
      IMPLICIT NONE
C
      INTEGER*4 NEW, NBAD, NPAIR, NREJ, J, N, JJ, ISUM, N1
      INTEGER*4 NADD, NLIM, NSF, FLAG(NLIM), SF_TYPE
      INTEGER*4 HKL(3,NLIM)
      REAL*4    AMP(NLIM), PHI(NLIM), SIG(NLIM), DSTR(NLIM), GAM_STAR
C  ----------------------------------------------------------------------------
C
C  *********************************
C  *    FIRST APPLY P2 SYMMETRY    *
C  *********************************
      CALL SF_SYM_P2 (SF_TYPE, NSF, AMP, PHI)
      PRINT *, '**********************'
      PRINT *, '* Apply pgg symmetry *'
      PRINT *, '**********************'
      CALL SF_INIT (NSF, NEW, NREJ, NBAD, NPAIR, FLAG)
      GAM_STAR = 90.0                             ! ENFORCE ORTHOGONAL ANGLE
C
      DO 900 J = 1,(NSF-1)
        IF(FLAG(J).NE.1) GO TO 100
        IF(J.LT.(NSF-1)) GO TO 900
        IF(FLAG(J+1) == 1) GO TO 900
C         ************************************************
C         *    HERE IF LAST SPOT REMAINS WITH NO MATE    *
C         ************************************************
        JJ       = J + 1
        FLAG(JJ) = 1
        ISUM     = HKL(1,JJ) + HKL(2,JJ)
        ISUM     = ABS(MOD(ISUM,2))
        GO TO 650
  100        FLAG(J)  = 1
        ISUM     = HKL(1,J) + HKL(2,J)
        ISUM     = ABS(MOD(ISUM,2))
        IF(HKL(1,J).NE.0.AND.HKL(2,J).NE.0) GO TO 300
C  ******************************************
C  *    HERE IF SPOT LIES ON H OR K AXIS    *
C  ******************************************
        IF(ISUM == 0) GO TO 900
C  ************************************************************
C  *    HERE IF FOUND SPOT THAT DOES NOT OBEY PGG SYMMETRY    *
C  *    ALONG MERIDIAN OR EQUATOR                             *
C  ************************************************************
        NREJ   = NREJ + 1
        AMP(J) = 0.0
        GO TO 900
C
  300        N1 = J + 1
        DO 600 N = N1,NSF
          IF(FLAG(N) == 1) GO TO 600
          IF( (HKL(1,J) == HKL(1,N).AND.HKL(2,J) == -HKL(2,N)) .OR.
     1       (HKL(1,J) == -HKL(1,N).AND.HKL(2,J) == HKL(2,N)) ) THEN
C             **********************************************
C             *    HERE IF PAIR OF SYMMETRY MATES FOUND    *
C             **********************************************
            FLAG(N) = 1
            IF(SF_TYPE == 0) THEN                    ! APART/BPART SF DATA
               IF(ISUM == 0.AND.SIGN(1.,AMP(J)) == SIGN(1.,AMP(N)))
     1        GO TO 500
               IF(ISUM == 1.AND.SIGN(1.,AMP(J)).NE.SIGN(1.,AMP(N)))
     1        GO TO 500
             ELSE                                ! AMP/PHASE SF DATA
               IF(ISUM == 0.AND.PHI(J) == PHI(N)) GO TO 500
               IF(ISUM == 1.AND.PHI(J).NE.PHI(N)) GO TO 500
            ENDIF
C             ***************************************
C             *    HERE IF PAIR FAILS PHASE TEST    *
C             ***************************************
            NBAD = NBAD + 1
            IF(ISUM.NE.1) THEN                        ! H+K = EVEN
              CALL SF_AVGMATES (SF_TYPE,N,J,NSF,AMP,PHI,NPAIR,FLAG)
              GO TO 900
            ENDIF
            IF(SF_TYPE == 0) THEN                        ! H+K = ODD
               IF(ABS(AMP(J)).GE.ABS(AMP(N))) THEN
                  AMP(N) = -AMP(N)
                ELSE
                  AMP(J) = -AMP(J)
               ENDIF
               AMP(J) = (AMP(J) + AMP(N))/2.0
               AMP(N) = -AMP(J)
             ELSE                               ! AMP/PHASE SF DATA
               IF(AMP(J).GE.AMP(N)) THEN
                  PHI(N) = PHI(N) + 180.0
                ELSE
                  PHI(J) = PHI(J) + 180.0
               ENDIF
               AMP(J) = ABS(AMP(J) - AMP(N))/2.0
               AMP(N) = AMP(J)
            ENDIF
            GO TO 900
C             ********************************
C             *    AVERAGE SYMMETRY MATES    *
C             ********************************
  500            IF(SF_TYPE == 0) THEN                        ! APART/BPART
               IF(ISUM == 0) THEN
                  AMP(J) = (AMP(J) + AMP(N))/2.0
                  AMP(N) =  AMP(J)
                ELSE
                  AMP(J) = (AMP(J) - AMP(N))/2.0
                  AMP(N) = -AMP(J)
               ENDIF
             ELSE                                    ! AMP/PHASE
               AMP(J) = (AMP(J) + AMP(N))/2.0
               AMP(N) = AMP(J)
            ENDIF
            NPAIR  = NPAIR+1
            GO TO 900
          ENDIF
  600        CONTINUE
C         *********************************************************
C         *    HERE IF ONLY ONE SPOT OF SYMMETRY PAIR IS FOUND    *
C         *    GENERATE THE MISSING SPOT                          *
C         *********************************************************
        JJ      = J
  650        AMP(JJ) = AMP(JJ)/2.0
        NEW     = NEW + 1
        CALL SF_CHECKLIM (NLIM, NEW)
        HKL(1,NEW) = -HKL(1,JJ)
        HKL(2,NEW) =  HKL(2,JJ)
        AMP(NEW)   =  AMP(JJ)
        IF(SF_TYPE == 0.AND.ISUM == 1) AMP(NEW) = -AMP(JJ)
        PHI(NEW)  = PHI(JJ)
        IF(SF_TYPE.NE.0 .AND. ISUM == 1) THEN
          PHI(NEW) = PHI(JJ) + 180.0
          IF(PHI(NEW) == 360.0) PHI(NEW) = 0.0
        ENDIF
        SIG(NEW)  =  SIG(JJ)
        DSTR(NEW) = DSTR(JJ)
  900      CONTINUE
C
      NADD = NEW - NSF
      NSF  = NEW
      WRITE(*,1000) NREJ, NBAD, NPAIR, NADD
Ca      CALL HITCR
C
      RETURN
 1000 FORMAT(I4,' Spots along H and K axes discarded',/,I4,' out of',
     1     I4,' symmetry pairs FAIL phase test',/,I4,' missing spots ',
     2     'generated')
      END

C  ***********************************************************
C  *                                                         *
C  *        *****     SUBROUTINE SF_SYM_CMM     *****        *
C  *                                                         *
C  *    CONSTRAIN SF DATA TO PLANE GROUP CMM.                *
C  *                                                         *
C  *    JUNE 1986  PURDUE UNIVERSITY VAX 750, T. BAKER.      *
C  *    MARCH 1992  PURDUE UNIVERSITY VAX 8550, T. BAKER.    *
C  *    LAST MODIFIED 13-MAR-1992 BY T. BAKER.               *
C  *                                                         *
C  *    PLEASE NOTE: THIS SUBPROGRAM IS UNTESTED.            *
C  ***********************************************************
C
      SUBROUTINE SF_SYM_CMM (SF_TYPE, NLIM, NSF, HKL, AMP, PHI, SIG,
     1     DSTR, FLAG, NREJ, gamstr)
C
      IMPLICIT NONE
C
      INTEGER*4 NREJ, I, ISUM, NLIM, NSF, FLAG(NLIM), SF_TYPE
      INTEGER*4 HKL(3,NLIM)
      REAL*4    AMP(NLIM), PHI(NLIM), SIG(NLIM), DSTR(NLIM),gamstr
C  ----------------------------------------------------------------------------
C
      CALL SF_SYM_PGG(SF_TYPE,NLIM,NSF,HKL,AMP,PHI,SIG,DSTR,FLAG,gamstr)
C
      IF(SF_TYPE == 0) THEN                          ! APART/BPART SF DATA
         DO I = 1,NSF
           ISUM = HKL(1,I) + HKL(2,I)
           ISUM = ABS(MOD(ISUM,2))
           IF(ISUM.NE.0) THEN
             AMP(I) = 0.0
             PHI(I) = 0.0
             NREJ   = NREJ + 1
           ENDIF
         ENDDO
       ELSE                                        ! AMP/PHASE SF DATA
         DO I = 1,NSF
           ISUM = HKL(1,I) + HKL(2,I)
           ISUM = ABS(MOD(ISUM,2))
           IF(ISUM.NE.0) THEN
             AMP(I) = 0.0
             NREJ   = NREJ + 1
           ENDIF
         ENDDO
      ENDIF
C
      WRITE(*,'(I4,A)') NREJ, ' odd order spots rejected'
      RETURN
C
      END

C  ***********************************************************
C  *                                                         *
C  *        *****     SUBROUTINE SF_SYM_P4     *****         *
C  *                                                         *
C  *    CONSTRAIN SF DATA TO PLANE GROUP P4.                 *
C  *                                                         *
C  *    JUNE 1986  PURDUE UNIVERSITY VAX 750, T. BAKER.      *
C  *    MARCH 1992  PURDUE UNIVERSITY VAX 8550, T. BAKER.    *
C  *    LAST MODIFIED 13-MAR-1992 BY T. BAKER.               *
C  *                                                         *
C  *    PLEASE NOTE: THIS SUBPROGRAM IS UNTESTED.            *
C  ***********************************************************
C
      SUBROUTINE SF_SYM_P4 (SF_TYPE,NLIM,NSF,HKL,AMP,PHI,SIG,DSTR,FLAG,
     1 gam_star, aoverb)
C
      IMPLICIT NONE
C
      INTEGER*4 NEW, NBAD, NPAIR, NREJ, J, JJ, N, N1, JLIM, NADD
      INTEGER*4 NLIM, NSF, FLAG(NLIM), SF_TYPE
      INTEGER*4 HKL(3,NLIM)
      REAL*4    AMP(NLIM), PHI(NLIM), SIG(NLIM), DSTR(NLIM), TEST
      real*4    gam_star, aoverb
C  ----------------------------------------------------------------------------
C
      gam_star = 90.0
      aoverb = 1.0

      CALL SF_SYM_P2 (SF_TYPE, NSF, AMP, PHI)            ! APPLY P2 SYMMETRY FIRST
      PRINT *, '*********************'
      PRINT *, '* Apply p4 symmetry *'
      PRINT *, '*********************'
      CALL SF_INIT (NSF, NEW, NREJ, NBAD, NPAIR, FLAG)
C
      JLIM = NSF - 1
      DO 900 J = 1,JLIM
        IF(FLAG(J).NE.1) GO TO 100
        IF(J.LT.JLIM) GO TO 900
        IF(FLAG(J+1) == 1) GO TO 900
C         ************************************************
C---------*    HERE IF LAST SPOT REMAINS WITH NO MATE    *
C         ************************************************
        JJ       = J + 1
        FLAG(JJ) = 1
        GO TO 750
  100        FLAG(J) = 1
        N1      = J + 1
C         *********************************
C---------*    LOOK FOR SYMMETRY MATES    *
C         *********************************
        DO 700 N = N1,NSF
          IF(FLAG(N) == 1) GO TO 700
          IF(ABS(HKL(1,J)).NE.HKL(2,N) .OR. 
     1            HKL(2,J).NE.HKL(1,N)) THEN
            GO TO 700
          ENDIF
C           *************************************
C-----------*    HERE IF SYMMETRY MATE FOUND    *
C           *************************************
          IF(SF_TYPE == 0) THEN                    ! APART/BPART SF DATA
             TEST = AMP(J)*AMP(N)
             IF(TEST.LT.0.0) NBAD = NBAD + 1
           ELSE                                ! AMP/PHASE SF DATA
             IF(PHI(J).NE.PHI(N)) NBAD = NBAD + 1
          ENDIF
          CALL SF_AVGMATES (SF_TYPE,N,J,NSF,AMP,PHI,NPAIR,FLAG)
          GO TO 900
  700        CONTINUE
C  **********************************************
C  *    HERE IF ONLY ONE SYMMETRY MATE FOUND    *
C  **********************************************
        JJ      = J
  750        AMP(JJ) = AMP(JJ)/2.0
        NEW     = NEW + 1
        CALL SF_CHECKLIM (NLIM, NEW)
        HKL(1,NEW) = -HKL(2,JJ)
        HKL(2,NEW) =  HKL(1,JJ)
        AMP(NEW)   =  AMP(JJ)
        PHI(NEW)   =  PHI(JJ)
        SIG(NEW)   =  SIG(JJ)
        DSTR(NEW)  =  DSTR(JJ)
  900      CONTINUE
C
      NADD = NEW - NSF
      NSF  = NEW
      WRITE(*,1000) NBAD, NPAIR, NADD
      RETURN
C
 1000      FORMAT(I4,' of',I4,' symmetry pairs fail the phase test',/,
     1 I4,' symmetry related mates have been generated')
      END

C  ***********************************************************
C  *                                                         *
C  *        *****     SUBROUTINE SF_SYM_P4M     *****        *
C  *                                                         *
C  *    CONSTRAIN SF DATA TO PLANE GROUP P4M.                *
C  *                                                         *
C  *    JUNE 1986  PURDUE UNIVERSITY VAX 750, T. BAKER.      *
C  *    MARCH 1992  PURDUE UNIVERSITY VAX 8550, T. BAKER.    *
C  *    LAST MODIFIED 16-MAR-1992 BY T. BAKER.               *
C  *                                                         *
C  *    PLEASE NOTE: THIS SUBPROGRAM IS UNTESTED.            *
C  ***********************************************************
C
      SUBROUTINE SF_SYM_P4M (SF_TYPE,NLIM,NSF,HKL,AMP,PHI,SIG,DSTR,FLAG)
C
      IMPLICIT NONE
C
      INTEGER*4 NEW, NBAD, NPAIR, NREJ, J, JJ, N, N1, JLIM, NADD
      INTEGER*4 NLIM, NSF, FLAG(NLIM), SF_TYPE
      INTEGER*4 HKL(3,NLIM)
      REAL*4    AMP(NLIM), PHI(NLIM), SIG(NLIM), DSTR(NLIM)
C  ----------------------------------------------------------------------------
C
      CALL SF_SYM_P4 (SF_TYPE,NLIM,NSF,HKL,AMP,PHI,SIG,DSTR,FLAG,90.,1.)
      PRINT *, '**********************'
      PRINT *, '* Apply p4m symmetry *'
      PRINT *, '**********************'
      CALL SF_INIT (NSF, NEW, NREJ, NBAD, NPAIR, FLAG)
C
      JLIM = NSF - 1
      DO 900 J = 1,JLIM
        IF(FLAG(J).NE.1) GO TO 100
        IF(J.LT.JLIM) GO TO 900
        IF(FLAG(J+1) == 1) GO TO 900
C  ************************************************
C  *    HERE IF LAST SPOT REMAINS WITH NO MATE    *
C  ************************************************
        JJ       = J + 1
        FLAG(JJ) = 1
        GO TO 750
  100        FLAG(J) = 1
        IF(HKL(1,J) == 0.OR.HKL(2,J) == 0) GO TO 900
        N1 = J + 1
        DO 700 N = N1,NSF
          IF(FLAG(N) == 1) GO TO 700
          IF(HKL(1,J).NE.-HKL(1,N)) GO TO 700
C           *************************************
C-----------*    HERE IF FOUND SYMMETRY MATE    *
C           *************************************
          IF(PHI(J).NE.PHI(N)) NBAD = NBAD + 1
          CALL SF_AVGMATES (SF_TYPE,N,J,NSF,AMP,PHI,NPAIR,FLAG)
          GO TO 900
  700        CONTINUE
C         **********************************************
C---------*    HERE IF ONLY ONE SYMMETRY MATE FOUND    *
C         *    GENERATE THE MISSING MATE               *
C         **********************************************
        JJ      = J
  750        AMP(JJ) = AMP(JJ)/2.0
        NEW = NEW + 1
        CALL SF_CHECKLIM (NLIM, NEW)
        HKL(1,NEW) = -HKL(1,JJ)
        HKL(2,NEW) =  HKL(2,JJ)
        AMP(NEW)   =  AMP(JJ)
        PHI(NEW)   =  PHI(JJ)
        SIG(NEW)   =  SIG(JJ)
        DSTR(NEW)  =  DSTR(JJ)
  900      CONTINUE
C
      NADD = NEW - NSF
      NSF  = NEW
      WRITE(*,1000) NBAD, NPAIR, NADD
      RETURN
C
 1000      FORMAT(I4,' of',I4,' symmetry pairs fail the phase test',/,
     1 I4,' symmetry related mates have been generated')
      END

C  ***********************************************************
C  *                                                         *
C  *        *****     SUBROUTINE SF_SYM_P4G     *****        *
C  *                                                         *
C  *    CONSTRAIN SF DATA TO PLANE GROUP P4G.                *
C  *                                                         *
C  *    JUNE 1986  PURDUE UNIVERSITY VAX 750, T. BAKER.      *
C  *    MARCH 1992  PURDUE UNIVERSITY VAX 8550, T. BAKER.    *
C  *    LAST MODIFIED 16-MAR-1992 BY T. BAKER.               *
C  *                                                         *
C  *    PLEASE NOTE: THIS SUBPROGRAM IS UNTESTED.            *
C  ***********************************************************
C
      SUBROUTINE SF_SYM_P4G (SF_TYPE, NLIM, NSF, HKL, AMP, PHI,
     1 SIG, DSTR, FLAG)
C
      IMPLICIT NONE
C
      INTEGER*4 NLIM, NSF, FLAG(NLIM), SF_TYPE
      INTEGER*4 HKL(3,NLIM)
      REAL*4    AMP(NLIM), PHI(NLIM), SIG(NLIM), DSTR(NLIM), gamstr
C  ----------------------------------------------------------------------------
C
      CALL SF_SYM_P4 (SF_TYPE,NLIM,NSF,HKL,AMP,PHI,SIG,DSTR,FLAG,
     1    gamstr,1.)
      CALL SF_SYM_PGG (SF_TYPE,NLIM,NSF,HKL,AMP,PHI,SIG,DSTR,FLAG,
     1    gamstr)
C
      RETURN
      END

C  ***********************************************************
C  *                                                         *
C  *        *****     SUBROUTINE SF_SYM_P6     *****         *
C  *                                                         *
C  *    CONSTRAIN SF DATA TO PLANE GROUP P6.                 *
C  *                                                         *
C  *    JUNE 1986  PURDUE UNIVERSITY VAX 750, T. BAKER.      *
C  *    MARCH 1992  PURDUE UNIVERSITY VAX 8550, T. BAKER.    *
C  *    LAST MODIFIED 16-MAR-1992 BY T. BAKER.               *
C  *                                                         *
C  *    PLEASE NOTE: THIS SUBPROGRAM IS UNTESTED.            *
C  ***********************************************************
C
      SUBROUTINE SF_SYM_P6 (sf_type,NLIM,NSF,HKL,AMP,PHI,
     1 SIG, DSTR, FLAG, gam_star, aoverb)
C
      IMPLICIT NONE
C
      INTEGER*4 NLIM, NSF, FLAG(NLIM), DJ, DN, NADD, SF_TYPE
      INTEGER*4 NEW, NBAD, NPAIR, NREJ, J, N, N1, JLIM, NMISS, L
      INTEGER*4 H1, H2, HJ, HN, K1, K2, KN, KJ, KK, NSYM(3), ISYM
      INTEGER*4 HKL(3,NLIM)
      REAL*4    AMP(NLIM), PHI(NLIM), SIG(NLIM), DSTR(NLIM)
      REAL*4    AVG, ANG, P, gam_star, aoverb
      real*8    deg_to_rad
C  ----------------------------------------------------------------------------
      deg_to_rad = acos(-1.0d0)/180.0d0
C
C  *********************************
C  *    FIRST APPLY P2 SYMMETRY    *
C  *********************************
Ca      FORCE these values      
      if(abs(gam_star-60.0) .le. 5.0) then
         gam_star = 60.0  
      else
         if(abs(gam_star-120.0) .le. 5.0) then
            gam_star = 120.0 
         endif
      endif
      aoverb = 1.0

      CALL SF_SYM_P2 (SF_TYPE, NSF, AMP, PHI)
      PRINT *, '*********************'
      PRINT *, '* Apply p6 symmetry *'
      PRINT *, '*********************'
      CALL SF_INIT (NSF, NEW, NREJ, NBAD, NPAIR, FLAG)
C
      JLIM = NSF - 1
      DO 900 J = 1,JLIM
        IF(FLAG(J).NE.1) GO TO 100
   50        IF(J.LT.JLIM) GO TO 900
        IF(FLAG(J+1) == 1) GO TO 900
C         *************************************************
C---------*    HERE IF LAST SPOT REMAINS WITH NO MATES    *
C         *************************************************
        AVG        = REAL(AMP(J+1)*COS(deg_to_rad*PHI(J+1)))
        ISYM       = 1
        NSYM(ISYM) = J + 1
        FLAG(J+1)  = 1
        GO TO 250
  100        HJ         = HKL(1,J)
        KJ         = HKL(2,J)
        AVG        = REAL(AMP(J)*COS(deg_to_rad*PHI(J)))
        ISYM       = 1
        NSYM(ISYM) = J
        FLAG(J)    = 1
        N1         = J + 1
C         *********************************
C---------*    LOOK FOR SYMMETRY MATES    *
C         *********************************
        DO 200 N = N1,NSF
          HN = HKL(1,N)
          KN = HKL(2,N)
          DJ = HJ*HJ + HJ*KJ + KJ*KJ
          DN = HN*HN + HN*KN + KN*KN
          IF(DJ.NE.DN) GO TO 200
          IF(DJ == 0) GO TO 200
          ANG = (HJ*HN + KJ*KN + 0.5*(HN*KJ + HJ*KN))/DJ
          ANG = ABS(ANG)
          IF(ANG == 0.5) THEN
            ISYM       = ISYM + 1
            NSYM(ISYM) = N
            FLAG(N)    = 1
            AVG        = AVG + REAL(AMP(N)*COS(deg_to_rad*PHI(N)))
          ENDIF
 200   CONTINUE
C         ********************************
C---------*    AVERAGE SYMMETRY MATES    *
C         ********************************
  250        AVG = AVG/3.0
        P   = 0.0
        IF(AVG.LT.0.0) P = 180.0
        DO KK = 1,ISYM
          N      = NSYM(KK)
          PHI(N) = P
          AMP(N) = ABS(AVG)
        ENDDO
        IF(ISYM == 3) GO TO 50
C         ***************************************************
C---------*    HERE IF ONE OR TWO SYMMETRY MATES MISSING    *
C         *    GENERATE MISSING DATA                        *
C         ***************************************************
        NMISS = 3 - ISYM
        H1    = HKL(1,NSYM(1))
        K1    = HKL(2,NSYM(1))
  400        H2    = -K1
        K2    = H1 + K1
        IF(K2.GE.0) GO TO 600
  500        H1 = H2
        K1 = K2
        GO TO 400
  600        DO L = 1,ISYM
          IF(H2 ==  HKL(1,NSYM(L)).AND.K2 ==  HKL(2,NSYM(L))) THEN
            GO TO 500
          ENDIF
          IF(H2 == -HKL(1,NSYM(L)).AND.K2 == -HKL(2,NSYM(L))) THEN
            GO TO 500
          ENDIF
        ENDDO
C         *********************************************
C---------*    HERE IF FOUND MISSING MATE INDICIES    *
C         *********************************************
        NMISS = NMISS - 1
        ISYM  = ISYM + 1
        NEW   = NEW + 1
        CALL SF_CHECKLIM (NLIM, NEW)
        NSYM(ISYM) = NEW
        HKL(1,NEW) = H2
        HKL(2,NEW) = K2
        AMP(NEW)   = ABS(AVG)
        PHI(NEW)   = PHI(NSYM(1))
        SIG(NEW)   = SIG(NSYM(1))
        DSTR(NEW)  = DSTR(NSYM(1))
        IF(NMISS.GT.0) GO TO 500
        GO TO 50
  900      CONTINUE
C
      NADD = NEW - NSF
      WRITE(*,'(I4,A)') NADD, ' symmetry mates have been generated'
      NSF = NEW
C
      RETURN
      END

C  ***********************************************************
C  *                                                         *
C  *        *****     SUBROUTINE SF_SYM_P6M     *****        *
C  *                                                         *
C  *    CONSTRAIN SF DATA TO PLANE GROUP P6M.                *
C  *                                                         *
C  *    JUNE 1986  PURDUE UNIVERSITY VAX 750, T. BAKER.      *
C  *    MARCH 1992  PURDUE UNIVERSITY VAX 8550, T. BAKER.    *
C  *    LAST MODIFIED 16-MAR-1992 BY T. BAKER.               *
C  *                                                         *
C  *    PLEASE NOTE: THIS SUBPROGRAM IS UNTESTED.            *
C  ***********************************************************
C
      SUBROUTINE SF_SYM_P6M (SF_TYPE, NLIM, NSF, HKL, AMP, PHI, SIG,
     1 DSTR, FLAG, gam_star, aoverb)
C
      IMPLICIT NONE
C
      INTEGER*4 NEW, NBAD, NPAIR, NREJ, J, JJ, N, JLIM, NADD, IEND
      INTEGER*4 NLIM, NSF, FLAG(NLIM), HMATE, N1, SF_TYPE
      INTEGER*4 HKL(3,NLIM)
      REAL*4    AMP(NLIM), PHI(NLIM), SIG(NLIM), DSTR(NLIM)
      real*4    gam_star, aoverb
C  ----------------------------------------------------------------------------
C
      CALL SF_SYM_P6 (sf_type, NLIM, NSF, HKL, AMP, PHI, SIG, 
     1 DSTR, FLAG, gam_star, aoverb)
      PRINT *, '**********************'
      PRINT *, '* Apply p6m symmetry *'
      PRINT *, '**********************'
      CALL SF_INIT (NSF, NEW, NREJ, NBAD, NPAIR, FLAG)
      IEND = 0
      JLIM = NSF - 1
C
      DO 900 J = 1,JLIM
        IF(FLAG(J).NE.1) GO TO 100
        IF(J.LT.JLIM) GO TO 900
        JJ = J + 1
        IF(FLAG(JJ) == 1) GO TO 900
C         *************************************************
C---------*    HERE IF LAST SPOT REMAINS WITH NO MATES    *
C         *************************************************
        IF(HKL(2,JJ) == 0) GO TO 900
        FLAG(JJ) = 1
        HMATE    = -(HKL(1,JJ) + HKL(2,JJ))
        IF(HKL(1,JJ) == HMATE) GO TO 900
        GO TO 800
C
  100        IF(HKL(2,J) == 0) GO TO 900
        HMATE   = -(HKL(1,J) + HKL(2,J))
        IF(HKL(1,J) == HMATE) GO TO 900
        FLAG(J) = 1
        N1      = J + 1
        DO 700 N = N1,NSF
          IF(HKL(1,N).NE.HMATE.OR.HKL(2,N).NE.HKL(2,J)) GO TO 700
C           *************************************
C-----------*    HERE IF FOUND SYMMETRY MATE    *
C           *************************************
          IF(PHI(J).NE.PHI(N)) NBAD = NBAD + 1
          CALL SF_AVGMATES (SF_TYPE,N,J,NSF,AMP,PHI,NPAIR,FLAG)
          GO TO 900
  700        CONTINUE
C         ***************************************
C---------*    HERE IF SYMMETRY MATE MISSING    *
C         ***************************************
        JJ  = J
  800        NEW = NEW + 1
        CALL SF_CHECKLIM (NLIM, NEW)
        HKL(1,NEW) = HMATE
        HKL(2,NEW) = HKL(2,JJ)
        AMP(JJ)    = AMP(JJ)/2.0
        AMP(NEW)   = AMP(JJ)
        PHI(NEW)   = PHI(JJ)
        SIG(NEW)   = SIG(JJ)
        DSTR(NEW)  = DSTR(JJ)
  900      CONTINUE
C
      NADD = NEW - NSF
      WRITE(*,1000) NBAD, NPAIR, NADD
      NSF = NEW
C
      RETURN
 1000      FORMAT(I4,' of',I4,' symmetry pairs fail the phase test',
     1 /,I4,' symmetry mates have been generated')
      END

C  ****************************************************************************
C  *                                                                          *
C  *                  *****     SUBROUTINE SF_TFAC     *****                  *
C  *                                                                          *
C  *    APPLY TEMPERATURE FACTOR TO STRUCTURE FACTOR DATA.                    *
C  *                                                                          *
C  *    JANUARY 1985  PURDUE UNIVERSITY VAX 11/750, T. BAKER.                 *
C  *    JULY 1991  PURDUE UNIVERSITY VAX 8550, T. BAKER.                      *
C  *    LAST MODIFIED 10-MAR-1992 BY T. BAKER.                                *
C  ****************************************************************************
C
      SUBROUTINE SF_TFAC (SF_TYPE, NSF, HKL, AMP, PHI, DSTR, TEMPFAC)
C
      IMPLICIT NONE
C
      INTEGER*4 NSF, I, ITEST, SF_TYPE
      INTEGER*4 HKL(3,NSF)
      REAL*4    AMP(NSF), PHI(NSF), DSTR(NSF), TEMPFAC, DST, WEIGHT
      REAL*4    TMP
C  ----------------------------------------------------------------------------
C
      TMP   = 1.0/TEMPFAC**2
      ITEST = 0
C
      DO I = 1,NSF
        DST = DSTR(I)**2
        IF(DST.NE.0.0) THEN
           WEIGHT = EXP(-DST*TMP)
           IF(SF_TYPE == 0) THEN
              AMP(I) = AMP(I)*WEIGHT
              PHI(I) = PHI(I)*WEIGHT
            ELSE
              AMP(I) = AMP(I)*WEIGHT
            ENDIF
         ELSE
           ITEST = ITEST + 1
           IF(ITEST.LE.10) THEN
              WRITE(*,1000) HKL(1,I), HKL(2,I), HKL(3,I)
            ELSE
              CALL BELL (2)
              PRINT *, '*****************************************'
              PRINT *, 'Can''t process this SF data...EXIT ROUTINE'
              PRINT *, '*****************************************'
              CALL HITCR
              RETURN
           ENDIF
        ENDIF
      ENDDO
C
      RETURN
C
 1000      FORMAT(' No resolution data for HKL =',3I4)
      END

C  ***************************************************************
C  *                                                             *
C  *           *****     SUBROUTINE SF_WRITE_pif  *****          *
C  *                                                             *
C  *    ROUTINE TO SAVE 2D/3D STRUCTURE FACTORS IN A DISK FILE   *
C  *                                                             *
C  *    JUNE 1985 PURDUE UNIVERSITY VAX 11/750, T. BAKER.        *
C  *    APRIL 1989  PURDUE UNIVERSITY VAX 8550, T. BAKER.        *
C  *    MODIFIED 10-DEC-1998 by RWA (PIF format SF output).      *
C  *    MODIFIED 28-DEC-1999 by TSB (Clean up).                  *
C  *    LAST MODIFIED 29-Nov-2001 by TSB/RWA (Use nsf_in)        *
C  ***************************************************************
C
      SUBROUTINE SF_WRITE_pif (FILE_ID, EM_NDIM, SF_TYPE, SF_FILENAME, 
     1 SF_TITLE, SF_SCALEF, SF_DSTR_MIN, SF_DSTR_MAX, LUN, NSF, HKL, 
     2 AMP, PHI, SIG, DSTR, T_MATRIX, A_STAR, B_STAR, C_STAR, ALP_STAR, 
     3 BET_STAR, GAM_STAR, AX, AY, BX, BY, CX, CY, TILTAXIS, TILTANG, 
     4 A_CELL, B_CELL, C_CELL, ALP, BET, GAM, VOL, VOL_STAR, nsf_in)
C
      IMPLICIT NONE
C
      include '../include/strucfac.inc'
C
      INTEGER*4 NSF, I, LUN, N, NUMCHARI4, NCOUNT, EM_NDIM
      INTEGER*4 SF_TITLE(18), SF_TYPE, FILE_ID, T_MATRIX(2,3), JJ
      integer*4 nsf_in, nsf_pif_out
      INTEGER*4 HKL(3,NSF)
      REAL*4    AMP(NSF), PHI(NSF), DSTR(NSF), SIG(NSF), SF_SCALEF
      REAL*4    TMP1, TMP2, DSTR_LO, DSTR_HI
      REAL*4    A_STAR, B_STAR, C_STAR, ALP_STAR, BET_STAR, GAM_STAR
      REAL*4    AX, AY, BX, BY, CX, CY, TILTAXIS, TILTANG
      REAL*4    SF_DSTR_MIN, SF_DSTR_MAX, ABSMAX
      REAL*4    A_CELL, B_CELL, C_CELL, ALP, BET, GAM, VOL, VOL_STAR
      CHARACTER SF_FILENAME*64

      integer alloc_stat
      type(STRUCFAC_T), allocatable :: STFACS(:)

      include '../include/pifhead.inc'
      TYPE(PIFHEAD_T)  MAPHEAD, IMGHEAD, PRJHEAD, PFTHEAD, SFHEAD
      COMMON /PIFHEAD/ MAPHEAD, IMGHEAD, PRJHEAD, PFTHEAD, SFHEAD

      character(len=1) RW(2)
      LOGICAL   TF

C  ----------------------------------------------------------------------------
C
      PRINT *
      PRINT *, 'Enter STRUCTURE FACTOR output filename -'
      PRINT *
      PRINT *, 'Type "EXIT" to end program or "RETURN" to RETURN ',
     1        'to calling routine'
      PRINT *
 100  WRITE(*,'(A)') 'FILENAME: '
      READ (*,'(A)',ERR=100) SF_FILENAME
      IF(SF_FILENAME(1:4) == 'EXIT') CALL TSBEND
      IF(SF_FILENAME(1:4) == 'RETURN') RETURN
  200      WRITE(*,'(A)') 'Binary<CR> OR ASCII(1) output file: '
      READ (*,*,ERR=200) I
      IF(I.NE.1) THEN
C       Let's get dstr_min and dstr_max for later testing.
 240        WRITE(*,'(A,F8.5,A,F8.5,A)') 'DSTR_MIN,DSTR_MAX (<CR>=',
     1      SF_DSTR_MIN, ',', SF_DSTR_MAX, '): '
        READ (*,*,ERR=240) TMP1, TMP2
        DSTR_LO = SF_DSTR_MIN
        DSTR_HI = SF_DSTR_MAX
        IF(TMP1.GT.0.0) DSTR_LO = TMP1
        IF(TMP2.GT.0.0) DSTR_HI = TMP2
        IF(DSTR_LO.GT.DSTR_HI) THEN
           PRINT *, '*********************************'
           PRINT *, '* DSTR_MIN cannot be > DSTR_MAX *'
           PRINT *, '*********************************'
           GO TO 240
        ENDIF

C  *************************************************************
C  *    WRITE OUT HEADER RECORDS TO PIF FORMAT SF DATA FILE    *
C  *************************************************************
        CALL PIF_INIT_HEAD (SFHEAD)
        SFHEAD%LUNUM  = 2
        RW(1)  = 'w'
        RW(2) = achar(0)
        CALL PIF_OPEN (SF_FILENAME, SFHEAD%LUNUM, RW) 
        SFHEAD%MODE = 31
C       
        CALL REC_TO_REAL (A_CELL, B_CELL, C_CELL, ALP, BET, GAM, VOL,
     1   A_STAR,B_STAR,C_STAR,VOL_STAR,ALP_STAR,BET_STAR,GAM_STAR)
C
        print *, 'a,b,c cell dimensions - ',a_cell,b_cell,c_cell
        print *, 'a*,b*,c*              - ', a_star,b_star,c_star
C
        SFHEAD%A_CELL      = A_CELL
        SFHEAD%B_CELL      = B_CELL
        SFHEAD%C_CELL      = C_CELL
        SFHEAD%ALP         = ALP
        SFHEAD%BET         = BET
        SFHEAD%GAM         = GAM
        SFHEAD%A_STAR      = A_STAR
        SFHEAD%B_STAR      = B_STAR
        SFHEAD%C_STAR      = C_STAR
        SFHEAD%ALP_STAR    = ALP_STAR
        SFHEAD%BET_STAR    = BET_STAR
        SFHEAD%GAM_STAR    = GAM_STAR
        SFHEAD%NUMROWS     = nsf_in
        SFHEAD%NUMCOLS     = EM_NDIM
        
        allocate(stfacs(nsf), stat=alloc_stat)
        if(alloc_stat > 0) then
           write(*,*) 'Unable to allocate STFACS'
           call tsbend
        endif

        ABSMAX             = AMP(1)
C
        nsf_pif_out = 0
        DO JJ = 1,nsf_in

          IF(DSTR(jj).GE.DSTR_LO .AND. DSTR(jj).LE.DSTR_HI) THEN
            nsf_pif_out = nsf_pif_out + 1
            STFACS(nsf_pif_out)%H    = HKL(1,JJ)
            STFACS(nsf_pif_out)%K    = HKL(2,JJ)
            STFACS(nsf_pif_out)%L    = HKL(3,JJ)
            STFACS(nsf_pif_out)%AMP  = AMP(JJ)
            STFACS(nsf_pif_out)%PHI  = PHI(JJ)
            STFACS(nsf_pif_out)%SIG  = 0.0
            STFACS(nsf_pif_out)%DSTR = DSTR(JJ)
            IF(ABS(STFACS(nsf_pif_out)%AMP) .GT. ABSMAX) 
     1                 ABSMAX = ABS(STFACS(nsf_pif_out)%AMP)
            IF(ABS(STFACS(nsf_pif_out)%PHI) .GT. ABSMAX) 
     1                 ABSMAX = ABS(STFACS(nsf_pif_out)%PHI)
          endif

         ENDDO

        SFHEAD%NUMROWS     = nsf_pif_out

C
         SFHEAD%SCALEF = 2147483648.0 / (ABSMAX + 2.0)
         CALL PIF_WRITE_GH (SFHEAD)
         CALL PIF_WRITE_DH (SFHEAD)
C
       ELSE    ! ASCII output
         OPEN (UNIT=LUN, FILE=SF_FILENAME, STATUS='NEW',
     1      ACCESS='SEQUENTIAL', FORM='FORMATTED')
         FILE_ID = 2
      ENDIF
C
      IF(FILE_ID == 2) THEN
        DO I = 1,18
          IF(SF_TITLE(I) .NE. 32) THEN ! RSS Changed ' ' to ascii code
            TF = .TRUE.
            GO TO 300
          ENDIF
        ENDDO
        TF = .FALSE.
 300        IF(TF) THEN
           WRITE(*,'(A,/)') ' THE CURRENT HEADER RECORD READS:'
           WRITE(*,'(1X,18A4,/)') SF_TITLE
 400       WRITE(*,'(A)') 'USE SAME HEADER<CR> OR CHANGE(1): '
           READ (*,*,ERR=400) I
         ELSE
           I = 1
        ENDIF
        IF(I == 1) THEN
 500       WRITE(*,'(A)') 'NEW HEADER: '
          READ (*,'(18A4)',ERR=500) SF_TITLE
        ENDIF
        WRITE(*,'(/,1X,18A4)') SF_TITLE
 600    WRITE(*,'(A,F8.5,A,F8.5,A)') 'DSTR_MIN,DSTR_MAX (<CR>=',
     1       SF_DSTR_MIN, ',', SF_DSTR_MAX, '): '
        READ (*,*,ERR=600) TMP1, TMP2
        DSTR_LO = SF_DSTR_MIN
        DSTR_HI = SF_DSTR_MAX
        IF(TMP1.GT.0.0) DSTR_LO = TMP1
        IF(TMP2.GT.0.0) DSTR_HI = TMP2
        IF(DSTR_LO.GT.DSTR_HI) THEN
           PRINT *, '*********************************'
           PRINT *, '* DSTR_MIN cannot be > DSTR_MAX *'
           PRINT *, '*********************************'
           GO TO 600
        ENDIF
C       
        WRITE (LUN,*) FILE_ID
        WRITE (LUN,'(1X,18A4)') SF_TITLE
        WRITE (LUN,*) SF_TYPE, EM_NDIM
        WRITE (LUN,'(6I5)') T_MATRIX
        WRITE (LUN,'(6F10.4)')  A_STAR,   B_STAR,   C_STAR, 
     1                      ALP_STAR, BET_STAR, GAM_STAR
        WRITE (LUN,'(6F10.4)') AX, AY, BX, BY, CX, CY
        WRITE (LUN,'(3F10.4)') TILTAXIS, TILTANG, SF_SCALEF
      ENDIF
C
      NCOUNT = 0
      IF(FILE_ID == 2) THEN                            ! ASCII TYPE OUTPUT
         DO I = 1,nsf_in
           IF(DSTR(I).GE.DSTR_LO .AND. DSTR(I).LE.DSTR_HI) THEN
             WRITE(LUN,'(3I5,3F10.4,F10.5)') HKL(1,I), HKL(2,I), 
     1       HKL(3,I), AMP(I), PHI(I), SIG(I), DSTR(I)
             NCOUNT = NCOUNT + 1
           ENDIF
         ENDDO
         CLOSE (UNIT=LUN, STATUS='KEEP')
         N = NUMCHARI4(NCOUNT) + 1
Ca         WRITE(*,'(I<N>,2A)') NCOUNT, ' Reflections written to ASCII',
Ca      1                 ' output file'
         WRITE(*,'(I10,2A)') NCOUNT, ' Reflections written to ASCII',
     1                 ' output file'
       ELSE                                        ! PIF FORMAT OUTPUT
         CALL PIF_WRITE_SF (SFHEAD, STFACS)
         CALL PIF_CLOSE (SFHEAD%LUNUM)
         deallocate(STFACS)
      ENDIF
C
      RETURN
      END


C  *******************************************************************
C  *                                                                 *
C  *          *****     SUBROUTINE SF3DBT_FOURIER     *****          *
C  *                                                                 *
C  *    Fourier summation routine                                    *
C  *                                                                 *
C  *    THIS IS A MAJOR REVISION OF THE SUBROUTINE "SIG100"          *
C  *     OF THE PROGRAM FOUR3D.                                      *
C  *                                                                 *
C  *    MODIFIED  5-AUG-1998 BY TSB (Store MAP in 3D array and       *
C  *     eliminate need for EMSF3DBT.SCR scratch file).              *
C  *    LAST MODIFIED 23-DEC-1998 BY TSB (Scale MAP by 2.0)          *
C  *******************************************************************
C
      SUBROUTINE SF3DBT_FOURIER (MAP, TABLE_SIZE, NCD, NEF, TABLE1,
     1  TABLE2, NCOL, NROW, NSEC, AP, BP, HH, KK, LL, SINHX, COSHX, 
     2  SINKY, COSKY, MAP_MIN, MAP_MAX, NUM_UNIQUE, NEQUIV)

      IMPLICIT NONE

      INTEGER*4 TABLE_SIZE, NCOL, NROW, NSEC, NEF, NUM_UNIQUE, NEQUIV
      COMMON /CONST/ TRIG_TABLE, XMIN, YMIN, ZMIN, XMAX, YMAX, ZMAX,
     1 DX, DY, DZ, DCELL, ITRANS, HMAX, KMAX, KMIN, RCELL, SCALE,
     2 TS, U, UINV
      INTEGER*4 I, J, K, M, N, CE, CF, LOCEI, CC, CD
      INTEGER*4 HMAX, KMAX, KMIN, NH, NK, ITRANS, NCD, ISEC
      INTEGER*4 HH(NUM_UNIQUE*NEQUIV), KK(NUM_UNIQUE*NEQUIV)
      INTEGER*4 LL(NUM_UNIQUE*NEQUIV)
      INTEGER*4 SECTION_THIS_LOOP
      INTEGER*4 NSECS_THIS_LOOP
      REAL*4    TABLE1(TABLE_SIZE), TABLE2(NEF), TRIG_TABLE(1001)
      REAL*4    SINKY(NROW*(KMAX-KMIN+1)), COSKY(NROW*(KMAX-KMIN+1))
      REAL*4    SINHX(NCOL*(HMAX+1)), COSHX(NCOL*(HMAX+1))
      REAL*4    UINV(3,3), MAP_MIN, MAP_MAX
      REAL*4    XMIN, YMIN, ZMIN, XMAX, YMAX, ZMAX, DX, DY, DZ
      REAL*4    DCELL(9), RCELL(9), SCALE, TS(3,24), U(3,3)
      REAL*4    AP(NUM_UNIQUE*NEQUIV), BP(NUM_UNIQUE*NEQUIV)
      REAL*4    MAP(NCOL,NROW,NSEC)
      REAL*4    ARG, SINARG, COSARG, X, Y, Z, DEL_ARG
C  ----------------------------------------------------------------------------
C
      MAP_MIN         = HUGE(MAP_MIN)
      MAP_MAX         = -HUGE(MAP_MAX)
      ISEC            = 0
      NH              = HMAX + 1            ! # H INDICES
      NK              = KMAX - KMIN + 1      ! # OF K INDICES
      NSECS_THIS_LOOP = TABLE_SIZE/NCD      ! # SECTIONS HANDLED EACH LOOP
      Z               = ZMIN
      PRINT *, ' '
      WRITE(*,'(A)') '   Set up SINHX,COSHX tables'
C  ***********************************
C  *    SET UP SINHX,COSHX TABLES    *
C  ***********************************
      I       = 1
      DEL_ARG = 0.0
      X       = 0.0
      DO N = 1,NH
        ARG  = X
        DO M = 1,NCOL
          CALL SF3DBT_GETARGS (ARG, SINARG, COSARG, TRIG_TABLE)
          SINHX(I) = SINARG
          COSHX(I) = COSARG
          I        = I + 1
          ARG      = ARG + DEL_ARG
        ENDDO
        DEL_ARG = DEL_ARG + DX
        X       = X + XMIN
      ENDDO
      WRITE(*,'(A)') '   Set up SINHX,COSHX tables'
C  ***********************************
C  *    SET UP SINKY,COSKY TABLES    *
C  ***********************************
      I       = 1
      Y       = REAL(KMIN)*YMIN
      DEL_ARG = REAL(KMIN)*DY
      DO K = 1,NK
        ARG  = Y
        DO N = 1,NROW
          CALL SF3DBT_GETARGS (ARG, SINARG, COSARG, TRIG_TABLE)
          SINKY(I) = SINARG
          COSKY(I) = COSARG
          I        = I + 1
          ARG      = ARG + DEL_ARG
        ENDDO
        Y       = Y + YMIN
        DEL_ARG = DEL_ARG + DY
      ENDDO
C
 100  IF(NSECS_THIS_LOOP.GT.NSEC-ISEC) NSECS_THIS_LOOP = NSEC - ISEC
C  **************************
C  *    INITIALIZE TABLE    *
C  **************************
      WRITE(*,'(A)') '   Set up TABLE1            '
      DO I = 1,TABLE_SIZE
        TABLE1(I) = 0.0
      ENDDO
C  **********************
C  *    SET UP TABLE    *
C  **********************
      DO I = 1,NUM_UNIQUE
        CC      = HH(I)*NK + (KK(I) - KMIN) + 1
        CD      = CC + NCD/2
        ARG     = LL(I)*Z
        DEL_ARG = LL(I)*DZ
        DO J = 1,NSECS_THIS_LOOP
          CALL SF3DBT_GETARGS (ARG, SINARG, COSARG, TRIG_TABLE)
          TABLE1(CC) = TABLE1(CC) + AP(I)*COSARG + BP(I)*SINARG
          TABLE1(CD) = TABLE1(CD) - AP(I)*SINARG + BP(I)*COSARG
          CC         = CC + NCD
          CD         = CC + NCD/2
          ARG        = ARG + DEL_ARG
        ENDDO
      ENDDO
C
      CC = 1
      CD = CC + NCD/2
      DO SECTION_THIS_LOOP = 1,NSECS_THIS_LOOP
        ISEC = ISEC + 1
        if(section_this_loop.lt.nsecs_this_loop) then
           if(mod(section_this_loop,25) == 0) then
            WRITE(*,'(A,I4)') 'Fourier section', ISEC
           endif
         else
           WRITE(*,'(A,I4)') 'Fourier section', ISEC
        endif
        DO J = 1,NEF
          TABLE2(J) = 0.0
        ENDDO
        LOCEI = 1
        DO J = 1,NH
          I = 1
          DO K = 1,NK
            CE = LOCEI
            CF = CE + NEF/2
            DO N = 1,NROW
              TABLE2(CE) = TABLE2(CE) + TABLE1(CC)*COSKY(I)
     1                               + TABLE1(CD)*SINKY(I)
              TABLE2(CF) = TABLE2(CF) - TABLE1(CC)*SINKY(I)
     1                               + TABLE1(CD)*COSKY(I)
              CE = CE + HMAX + 1
              CF = CE + NEF/2
              I  = I + 1
            ENDDO
            CC    = CC + 1
            CD    = CC + NCD/2
          ENDDO
          LOCEI = LOCEI + 1
        ENDDO
C
        CE = 1
        CF = NEF/2 + 1
        DO J = 1,NROW
          DO K = 1,NCOL                              ! ZERO ROW OF MAP
            MAP(K,J,ISEC) = 0.0
          ENDDO
          I = 1
C    *******************************************************************
C    *  NOTE: SCALE MAP UP BY 2.0 TO ACCOMMODATE USING ONLY ONE HALF   *
C    *        OF THE TRANSFORM                                         *
C    *******************************************************************
          DO N = 1,NH
            DO M = 1,NCOL
              MAP(M,J,ISEC) = MAP(M,J,ISEC) + TABLE2(CE)*COSHX(I)*2.0
     1                                     + TABLE2(CF)*SINHX(I)*2.0
              I      = I + 1
            ENDDO
            CE = CE + 1
            CF = CE + NEF/2
          ENDDO
          DO M = 1,NCOL
            MAP_MIN = AMIN1(MAP_MIN,MAP(M,J,ISEC))
            MAP_MAX = AMAX1(MAP_MAX,MAP(M,J,ISEC))
          ENDDO
        ENDDO
        CC   = CC + NCD/2
        CD   = CC + NCD/2
        Z    = Z + DZ
      ENDDO                       ! DO SECTION_THIS_LOOP = 1,NSECS_THIS_LOOP
      IF(ISEC.LT.NSEC) GO TO 100
C
      CLOSE (UNIT=3, STATUS='DELETE')
      RETURN
      END

C  *******************************************************************
C  *                                                                 *
C  *           *****     SUBROUTINE SF3DBT_OUT     *****             *
C  *                                                                 *
C  *    WRITE THE 3D MAP TO A DISK FILE (UNIT 3).                    *
C  *                                                                 *
C  *    APRIL 1989  PURDUE UNIVERSITY VAX 8550, T. BAKER             *
C  *    LAST MODIFIED  5-AUG-1998 BY TSB (Store MAP in 3D array and  *
C  *     eliminate need for EMSF3DBT.SCR scratch file).              *
C  *******************************************************************
C
      SUBROUTINE SF3DBT_OUT (MAP, NORIEN, NCOL, NROW, NSEC, 
     1 A_CELL, B_CELL, C_CELL, ALP, BET, GAM, T_MATRIX, MAP_MIN, 
     2 MAP_MAX, MAP_FILENAME, MAP_TITLE, pixSizUnits)

      IMPLICIT NONE

      include '../include/pifhead.inc'
      TYPE(PIFHEAD_T) MAPHEAD

      INTEGER*4 NCOL, NROW, NSEC, I, NORIEN, T_MATRIX(6)
      INTEGER*4 pixSizUnits
      REAL*4    MAP(NCOL,NROW,NSEC), TMP, SCALE1, SCALE2
      REAL*4    A_CELL, B_CELL, C_CELL, ALP, BET, GAM
      REAL*4    MAP_MIN, MAP_MAX, MAP_SCALEF
      CHARACTER MAP_FILENAME*64
      character(len=1) RW(2), MAP_TITLE(80)
C  ----------------------------------------------------------------------------
C
      WRITE(*,'(2(A,E13.6),/)') ' MIN =', MAP_MIN, ' MAX =', MAP_MAX
C  ********************************
C  *    FINAL OUTPUT OF 3D MAP    *
C  ********************************
      CALL PIF_INIT_HEAD (MAPHEAD)
      DO I =1,80
        MAPHEAD%DESCRIPTIONSTR(I) = MAP_TITLE(I)
      ENDDO
      MAPHEAD%LUNUM = 3
      RW(1)  = 'w'
      RW(2) = achar(0)
      CALL PIF_OPEN (MAP_FILENAME, MAPHEAD%LUNUM, RW)
      MAPHEAD%MODE    =   20
      MAPHEAD%NUMCOLS = NCOL
      MAPHEAD%NUMROWS = NROW
      MAPHEAD%NUMSECS = NSEC
      IF(NORIEN == 1) THEN                        ! Z-sections
         MAPHEAD%MAP_ABANG = GAM
       ELSE IF(NORIEN == 10) THEN                  ! Y-sections
         MAPHEAD%MAP_ABANG = BET
         TMP               = B_CELL
         B_CELL            = C_CELL
         C_CELL            = TMP
         ELSE IF(NORIEN == 100) THEN                  ! X-sections
         MAPHEAD%MAP_ABANG = ALP
         TMP               = A_CELL
         A_CELL            = C_CELL
         C_CELL            = TMP
      ENDIF
      MAPHEAD%AOVERB    = 1.0
      MAPHEAD%A_CELL    = A_CELL                        !!! TIMES PIXSIZ
      MAPHEAD%B_CELL    = B_CELL
      MAPHEAD%C_CELL    = C_CELL
      MAPHEAD%DELA      = A_CELL/REAL(NCOL)
      MAPHEAD%DELB      = B_CELL/REAL(NROW)
      MAPHEAD%DELC      = C_CELL/REAL(NSEC)
      DO I = 1,6
        MAPHEAD%T_MATRIX(I)  = T_MATRIX(I)
      ENDDO
      MAPHEAD%PIXELSIZE   = MAPHEAD%DELA
      MAPHEAD%PIXSIZUNITS = pixSizUnits
      SCALE1          = 1.01*ABS(MAP_MIN/REAL(-32767))
C  **************************************************************************
C  *    We need to adjust SCALE1 to trap for instances when the maximum     *
C  *     value of alp,bet,gam etc... might exceed that of the density map   *
C  **************************************************************************
      scale1 = max(scale1, (360.0/real(32767))) ! Needed for PIF file format
      SCALE2          = 1.01*ABS(MAP_MAX/REAL(32767))
      MAP_SCALEF      = 1.0/MAX(SCALE1,SCALE2)
      MAPHEAD%SCALEF  = MAP_SCALEF
      call pif_set_i2Map(1)   ! Tell pif to store it on I2
      call determine_scaling(mapHead,map,ncol,nrow,nsec)
      
      CALL PIF_WRITE_GH (MAPHEAD)
      MAPHEAD%IMGNUM = 1
      CALL PIF_WRITE_DH (MAPHEAD)
C
      PRINT *, '**************************'
      PRINT *, '* Write out final 3D MAP *'
      PRINT *, '**************************'
      call pif_write_map_sections(maphead, map,
     $     maphead%mode, ncol, nrow, nsec, 0, nsec)
      CALL PIF_CLOSE (MAPHEAD%LUNUM)
C
      RETURN
      END

C  **************************************************************************
C  *                                                                        *
C  *               *****     SUBROUTINE SF3DBT_INPUT     *****              *
C  *                                                                        *
C  *    Reads input for driving SF3DBT routine.                             *
C  *                                                                        *
C  *    LAST MODIFIED  5-AUG-1998 BY TSB (Eliminate need for EMSF3DBT.TMP   *
C  *     scratch file for storing intermediate SF data).                    *
C  **************************************************************************
C
      SUBROUTINE SF3DBT_INPUT (NCOL, NROW, NSEC, NEQUIV, NORIEN, NCD,
     1 NEF, FS, HKL, AMP, PHI, nsf_in, AP, BP, HH, KK, LL, SF_TYPE,
     2 NREF_DUMP, NUM_UNIQUE, pixsizunits)
C
      IMPLICIT NONE
C
      INTEGER*4 NEQUIV, nsf_in
      COMMON /CONST/ TRIG_TABLE, XMIN, YMIN, ZMIN, XMAX, YMAX, ZMAX,
     1 DX, DY, DZ, DCELL, ITRANS, HMAX, KMAX, KMIN, RCELL, SCALE,
     2 TS, U, UINV
      REAL*4    XMIN, YMIN, ZMIN, XMAX, YMAX, ZMAX, DX, DY, DZ
      REAL*4    DCELL(9), RCELL(9), SCALE, TS(3,24), U(3,3)
      REAL*4    UINV(3,3), FS(3,3,24)
      REAL*4    TRIG_TABLE(1001), XSIN(16)
      REAL*4    AMP(nsf_in), PHI(nsf_in), HKLTMP(3)
      REAL*4    AP(nsf_in*NEQUIV), BP(nsf_in*NEQUIV)
      REAL*4    FJX, FKX, FLX
      REAL*4    APART, BPART, APARTS(24), BPARTS(24)
      real*8    deg_to_rad
      INTEGER*4 M1(24,3), M2(24,3), M3(24,3), SF_TYPE, NUM_UNIQUE
      INTEGER*4 NEXPAN, NTOTAL, NU, I, M, N
      INTEGER*4 BUFFER, IFIXR, NUCALC, MTX(24), MTY(24), MTZ(24)
      INTEGER*4 NCOL, NROW, NSEC, ITRANS, HMAX, KMAX, KMIN, pixsizunits
      INTEGER*4 NN, NREF_DUMP, NORIEN, NCD, NEF, itmp
      INTEGER*4 HH(nsf_in*NEQUIV), KK(nsf_in*NEQUIV), LL(nsf_in*NEQUIV)
      INTEGER*4 J(24), K(24), L(24), HKL(3,nsf_in)
      character(len=2) axis(3)
      character(len=1) xl, yl, zl

      DATA XSIN/0.,.5,.866025,1.,.866025,.5,0.,-.5,-.866025,-1.,
     1  -.866025,-.5,0.,.5,.866025,1./
      DATA AXIS/' Z',' Y',' X'/,   XL/'X'/,YL/'Y'/,ZL/'Z'/
C  ----------------------------------------------------------------------------
C
      deg_to_rad = acos(-1.0d0)/180.0d0

      NEXPAN = 0
      NTOTAL = 0
      BUFFER = 0
      HMAX   = 0
      KMAX   = 0
      KMIN   = 0
 111  IF(ITRANS.LE.1) THEN
         PRINT *, '****************************************************'
         PRINT *, '* INPUT PARAMETER DEFAULTS:                        *'
         PRINT *, '* Portion of cell desired XMIN,XMAX = 0.0,1.0      *'
         PRINT *, '*                         YMIN,YMAX = 0.0,1.0      *'
         PRINT *, '*                         ZMIN,ZMAX = 0.0,1.0      *'
         PRINT *, '* Pixel sizes              DX,DY,DZ = 1.0,1.0,1.0  *'
         PRINT *, '* If no values entered for DY,DZ, default to DX    *'
         PRINT *, '* Orientation                NORIEN =   1 (Z-secs) *'
         PRINT *, '*  (Default is Z-sections)   NORIEN =  10 (Y-secs) *'
         PRINT *, '*                            NORIEN = 100 (X-secs) *'
         PRINT *, '****************************************************'
         WRITE(*,'(A)') 'XMIN,XMAX,YMIN,YMAX,ZMIN,ZMAX: '
         READ (*,*) XMIN, XMAX, YMIN, YMAX, ZMIN, ZMAX
         IF(XMAX == 0.0) XMAX = 1.0
         IF(YMAX == 0.0) YMAX = 1.0
         IF(ZMAX == 0.0) ZMAX = 1.0
         WRITE(*,'(A)') 'DX,DY,DZ (<CR>=1.0,1.0,1.0): '
         READ (*,*) DX,   DY,   DZ
         IF(DX.LE.0.0) DX = 1.0
         IF(DY.LE.0.0) DY = DX
         IF(DZ.LE.0.0) DZ = DX
         WRITE(*,'(A)') 'NORIEN (<CR>=1): '
         READ (*,*) NORIEN
         IF(NORIEN.LE.0 .OR. NORIEN.GT.100) NORIEN = 1
         DX = DX/DCELL(1)
         DY = DY/DCELL(2)
         DZ = DZ/DCELL(3)
      ELSE
         NORIEN = 100
      ENDIF
      CALL SF3DBT_CHEKUP (XMIN, XMAX, NORIEN, 100, DX)
      CALL SF3DBT_CHEKUP (YMIN, YMAX, NORIEN,  10, DY)
      CALL SF3DBT_CHEKUP (ZMIN, ZMAX, NORIEN,   1, DZ)
      FJX = DX*DCELL(1)
      FKX = DY*DCELL(2)
      FLX = DZ*DCELL(3)
      I   = INT(ALOG10(REAL(NORIEN)+.1))+1
      WRITE(*,1010) AXIS(I)
      if(pixsizunits == 1) then
        WRITE(*,1020) XL, XMIN, XMAX, DX, FJX, YL, YMIN, YMAX, DY, FKX,
     1  ZL, ZMIN, ZMAX, DZ, FLX
       else if (pixsizunits == 2) then
        WRITE(*,1021) XL, XMIN, XMAX, DX, FJX, YL, YMIN, YMAX, DY, FKX,
     1  ZL, ZMIN, ZMAX, DZ, FLX
       else
        WRITE(*,1022) XL, XMIN, XMAX, DX, FJX, YL, YMIN, YMAX, DY, FKX,
     1  ZL, ZMIN, ZMAX, DZ, FLX
      endif
      IF(NORIEN.LT.10) THEN                         ! Z SECTIONS
C         CALL SF3DBT_ROT (XMIN,YMIN)                  ! SWITCH X & Y
C         CALL SF3DBT_ROT (XMAX,YMAX)
C         CALL SF3DBT_ROT (DX,DY)
       ELSE IF(NORIEN == 10) THEN                  ! Y SECTIONS
         CALL SF3DBT_ROT (YMIN,ZMIN)
         CALL SF3DBT_ROT (YMAX,ZMAX)
         CALL SF3DBT_ROT (DY,DZ)
       ELSE IF(NORIEN.GT.10) THEN                  ! X SECTIONS
         CALL SF3DBT_ROT (XMIN,ZMIN)
         CALL SF3DBT_ROT (XMAX,ZMAX)
         CALL SF3DBT_ROT (DX,DZ)
      ENDIF
      DO I = 1,NEQUIV
        DO M = 1,3
          M1(I,M) = INT(FS(1,M,I))
          M2(I,M) = INT(FS(2,M,I))
          M3(I,M) = INT(FS(3,M,I))
        ENDDO
        IF(M1(I,1)+M2(I,1) == 0) THEN
          IF(M1(I,1))         290,320,290
        ENDIF
        IF(M1(I,2)+M2(I,2)) 320,280,320
  280        IF(M1(I,2))         290,320,290
  290        IF(M1(I,1)+M2(I,2)) 310,300,310
  300        CALL SF3DBT_ROT_I4 (M1(I,2),M2(I,1))
        GO TO 320
  310        CALL SF3DBT_ROT_I4 (M1(I,1),M2(I,2))
  320        MTX(I) = NUCALC(IFIXR(12.*TS(1,I)))
        MTY(I) = NUCALC(IFIXR(12.*TS(2,I)))
        MTZ(I) = NUCALC(IFIXR(12.*TS(3,I)))
      ENDDO
      NCOL = INT((XMAX-XMIN)/DX+1.5)
      NROW = INT((YMAX-YMIN)/DY+1.5)
      NSEC = INT((ZMAX-ZMIN)/DZ+1.5)
      WRITE(*,'(3(A,I6))') ' MAP dimensions:', NCOL, ' by', NROW, 
     1 ' by', NSEC
      if(ncol*nrow*nsec.gt.400*400*400) then
         write(*,'(2A)') 'WARNING: Do you really want a map this',
     1         ' big? (Yes[1]/No[CR]): ' 
         read (*,*) itmp
         if(itmp.ne.1) go to 111
      endif
C
      IF(NREF_DUMP.GT.0) WRITE(6,1040)
      IF(ITRANS.GE.1) WRITE(6,1050)
C  *****************************************************************
C  *    CHANGES MADE TO READ HKL, AMP AND PHASE AND CONVERT TO     *
C  *      APARTS AND BPARTS   3-SEP-84  TSB                        *
C  *****************************************************************
      NN = 0
  400      NN = NN+1
      IF(NN.GT.nsf_in) GO TO 900
      IF(SF_TYPE == 1) THEN
         APART = REAL(AMP(NN)*COS(deg_to_rad*PHI(NN)))
         BPART = REAL(AMP(NN)*SIN(deg_to_rad*PHI(NN)))
       ELSE
         APART = AMP(NN)
         BPART = PHI(NN)
      ENDIF
      IF(APART == 0.0.AND.BPART == 0.0) GO TO 400
      HKLTMP(1) = HKL(1,NN)
      HKLTMP(2) = HKL(2,NN)
      HKLTMP(3) = HKL(3,NN)
      NTOTAL    = NTOTAL + 1
C
      DO N = 1,NEQUIV
        J(N) = M1(N,1)*HKL(1,NN)+M2(N,1)*HKL(2,NN)+M3(N,1)*HKL(3,NN)
        K(N) = M1(N,2)*HKL(1,NN)+M2(N,2)*HKL(2,NN)+M3(N,2)*HKL(3,NN)
        L(N) = M1(N,3)*HKL(1,NN)+M2(N,3)*HKL(2,NN)+M3(N,3)*HKL(3,NN)
        NU   = NUCALC(J(N)*MTX(N)+K(N)*MTY(N)+L(N)*MTZ(N)) + 1
        APARTS(N) = APART*XSIN(NU + 3) - BPART*XSIN(NU)
        BPARTS(N) = BPART*XSIN(NU + 3) + APART*XSIN(NU)
        IF(ITRANS.GE.1) THEN
          FJX  = J(N)
          FKX  = K(N)
          FLX  = L(N)
          J(N) = IFIXR( U(1,1)*FJX + U(1,2)*FKX + U(1,3)*FLX )
          K(N) = IFIXR( U(2,1)*FJX + U(2,2)*FKX + U(2,3)*FLX )
          L(N) = IFIXR( U(3,1)*FJX + U(3,2)*FKX + U(3,3)*FLX )
        ENDIF
        IF(NORIEN.LT.10) THEN                   ! Z SECTIONS
C           CALL SF3DBT_ROT_I2 (J(N),K(N))
         ELSE IF(NORIEN == 10) THEN                  ! Y SECTIONS
           CALL SF3DBT_ROT_I2 (K(N),L(N))
         ELSE IF(NORIEN.GT.10) THEN                  ! X SECTIONS
           CALL SF3DBT_ROT_I2 (J(N),L(N))
        ENDIF
        IF(J(N))680,650,710
  650        IF(K(N))690,660,710
  660        IF(L(N))700,670,710
  670        APARTS(N) = .5*APARTS(N)
        BPARTS(N) = .5*BPARTS(N)
        GO TO 710
  680        J(N) = -J(N)
  690        K(N) = -K(N)
  700        L(N) = -L(N)
        BPARTS(N) = -BPARTS(N)
  710        IF(APARTS(N).NE.0.0.OR.BPARTS(N).NE.0.0) THEN
          IF(J(N).GT.HMAX) HMAX = J(N)
          IF(K(N).GT.KMAX) KMAX = K(N)
          IF(K(N).LT.KMIN) KMIN = K(N)
        ENDIF
      ENDDO
      DO 860 I = 1,NEQUIV
        IF(APARTS(I) == 0.0.AND.BPARTS(I) == 0.0) GO TO 860
        M = I+1
        IF(M.LE.NEQUIV) THEN
          DO N = M,NEQUIV
           IF(APARTS(N) == 0.0) THEN
            IF(BPARTS(N).NE.0.0) THEN
             IF(J(I) == J(N).AND.K(I) == K(N).AND.L(I) == L(N)) THEN
              APARTS(N) = 0.0
              BPARTS(N) = 0.0
             ENDIF
            ENDIF
           ENDIF
          ENDDO
        ENDIF
        NEXPAN     = NEXPAN + 1
        BUFFER     = BUFFER + 1
        HH(BUFFER) = J(I)
        KK(BUFFER) = K(I)
        LL(BUFFER) = L(I)
        AP(BUFFER) = APARTS(I)
        BP(BUFFER) = BPARTS(I)
        IF(NREF_DUMP.GT.0) THEN
          IF(NORIEN.LT.10) THEN
C             WRITE(*,2020) K(I),J(I),L(I),APARTS(I),BPARTS(I)
             WRITE(*,2020) J(I),K(I),L(I),APARTS(I),BPARTS(I)
           ELSE IF(NORIEN == 10) THEN
             WRITE(*,2020) J(I),L(I),K(I),APARTS(I),BPARTS(I)
           ELSE IF(NORIEN.GT.10) THEN
             WRITE(*,2020) L(I),K(I),J(I),APARTS(I),BPARTS(I)
          ENDIF
        ENDIF
  860      CONTINUE
      IF(NREF_DUMP.GT.0) NREF_DUMP = NREF_DUMP - 1
      GO TO 400
C
 900  NCD = (HMAX+1)*(KMAX-KMIN+1)*2
      NEF = (HMAX+1)*NROW*2
      IF(NTOTAL.LE.0.OR.NEXPAN.LE.0) THEN
        PRINT *, '***********************************************'
        PRINT *, '* No data available for this summation...EXIT *'
        PRINT *, '* NTOTAL,NEXPAN =', NTOTAL, NEXPAN
        CALL TSBEND
      ENDIF
      WRITE(*,1060) NTOTAL, NEXPAN
      WRITE(*,1070) NCD, NEF
      NUM_UNIQUE = NEXPAN
      RETURN
C
 1010      FORMAT(' Compute sections along the',A2,' axis')
 1020      FORMAT(' ',A2,'Limits: ',F8.5,' to',F8.5,' in steps of',
     1  F8.5,' (',F6.3,' Ang)')
 1021      FORMAT(' ',A2,'Limits: ',F8.5,' to',F8.5,' in steps of',
     1  F8.5,' (',F6.3,' nms)')
 1022      FORMAT(' ',A2,'Limits: ',F8.5,' to',F8.5,' in steps of',
     1  F8.5,' (',F6.3,' pixels)')
 1040      FORMAT(' ',2X,'H',3X,'K',3X,'L',11X,'A',10X,'B'/)
 1050      FORMAT(20X,'(The indices are transformed to the new cell)')
 1060      FORMAT(' The',I8,' independent reflections have been',
     1  ' expanded to',I8,' terms')
 1070      FORMAT(' The CD table requires',I6,' memory locations'/,
     1  ' The EF table requires',I6,' memory locations'/)
 2020      FORMAT(3I4,2F10.2)
      END

C  *************************************************************************
C  *                                                                       *
C  *               *****     SUBROUTINE SF3DBT_INFO     *****              *
C  *                                                                       *
C  *   Read control parameters, cell constants, and symmetry cards         *
C  *   Calculate cosine table                                              *
C  *                                                                       *
C  *    LAST MODIFIED  5-AUG-1998 BY TSB (Rid of control parameter file).  *
C  *************************************************************************
C
      SUBROUTINE SF3DBT_INFO (NEQUIV, FS, AA, BB, A_CELL, B_CELL, 
     1 C_CELL, ALP, BET, GAM, VOL, VOL_STAR, NREF_DUMP)
C
      IMPLICIT NONE
C
      COMMON /CONST/ TRIG_TABLE, XMIN, YMIN, ZMIN, XMAX, YMAX, ZMAX,
     1 DX, DY, DZ, DCELL, ITRANS, HMAX, KMAX, KMIN, RCELL, SCALE,
     2 TS, U, UINV
      INTEGER*4 NEQUIV, I, M, N, ITRANS, HMAX, KMAX, KMIN
      INTEGER*4 NREF_DUMP
      REAL*4    XMIN, YMIN, ZMIN, XMAX, YMAX, ZMAX, DX, DY, DZ, V
      REAL*4    DCELL(9), RCELL(9), TS(3,24), U(3,3), UINV(3,3), SCALE
      REAL*4    TRIG_TABLE(1001), FS(3,3,24), AA(3,3), BB(3,3)
      REAL*4    A_CELL, B_CELL, C_CELL, ALP, BET, GAM, VOL, VOL_STAR
      REAL*8    DPI, DRADIAN, DDEGREE, ARG, DEL_ARG
C  ----------------------------------------------------------------------------
C
      CALL PIRADDEG_D (DPI, DRADIAN, DDEGREE)
      DEL_ARG = 2.0*DPI/1000.0
      ARG     = 0.D0
C
      DO I = 1,1001
        TRIG_TABLE(I) = COS(SNGL(ARG))
        ARG           = ARG + DEL_ARG
      ENDDO
C
      WRITE(*,'(A)') 'ITRANS, NEQUIV, SCALE <CR>=0,1,1.0'
      READ (*,*) ITRANS, NEQUIV, SCALE
      IF(ITRANS.LE.0 .OR. ITRANS.GT.2) ITRANS = 0
      IF(NEQUIV.LE.1 .OR. NEQUIV.GT.24) NEQUIV = 1
       IF(SCALE == 0.0) SCALE = 1.0
      DCELL(1) = A_CELL
      DCELL(2) = B_CELL
      DCELL(3) = C_CELL
      DCELL(4) = ALP
      DCELL(5) = BET
      DCELL(6) = GAM
      IF(DCELL(4).GE.1.0) THEN
        DO I = 4,6
          DCELL(I) = REAL(COS(DCELL(I)/DDEGREE))
        ENDDO
      ENDIF
      CALL SF3DBT_STOAA (DCELL, AA)
      CALL SF3DBT_MI3 (AA, BB, V)
      V = SQRT(V)
      RCELL(1) = SQRT(BB(1,1))
      RCELL(2) = SQRT(BB(2,2))      
      RCELL(3) = SQRT(BB(3,3))
      RCELL(4) = BB(2,3)/(RCELL(2)*RCELL(3))
      RCELL(5) = BB(1,3)/(RCELL(1)*RCELL(3))
      RCELL(6) = BB(1,2)/(RCELL(1)*RCELL(2))
      DO I = 4,6
        DCELL(I+3) = REAL(ATAN2(SQRT(1.-DCELL(I)*DCELL(I)),DCELL(I))
     1     *DDEGREE)
        RCELL(I+3) = REAL(ATAN2(SQRT(1.-RCELL(I)*RCELL(I)),RCELL(I))
     1     *DDEGREE)
      ENDDO
      IF(DCELL(1).LT.1.0) THEN
        V = 1./V
        DO I = 1,9
          CALL SF3DBT_ROT (DCELL(I), RCELL(I))
          CALL SF3DBT_ROT ( AA(I,1),  BB(I,1))
        ENDDO
      ENDIF
      WRITE(*,'(A,/,A)') '(TS(J),(FS(I,J),I=1,3),J=1,3)',
     1 'Defaults = 0.,1.,0.,0.,  0.,0.,1.,0.,  0.,0.,0.,1.: '
      READ (*,*) ((TS(M,N),(FS(I,M,N),I=1,3),M=1,3),N=1,NEQUIV)
      IF(NEQUIV == 1) THEN
        FS(1,1,1) = 1.0
        FS(2,2,1) = 1.0
        FS(3,3,1) = 1.0
      ENDIF
      WRITE(*,'(A)') 'NREF_DUMP <CR>=0: '
      READ (*,*) NREF_DUMP
C
      WRITE(*,'(A,I2,A,I2,A,F5.1)') ' ITRANS =', ITRANS, 
     1 '  NEQUIV =', NEQUIV, '  SCALE =', SCALE
      PRINT *, 'V,VOL,VOL_STAR', V,VOL,VOL_STAR
      WRITE(*,1010) DCELL
      WRITE(*,1020) RCELL
      WRITE(*,1030) NEQUIV
      WRITE(*,'(3(3X,F7.4,3F4.0))') ((TS(M,N),(FS(I,M,N),I=1,3),
     1                                M=1,3),N=1,NEQUIV)
      IF(ITRANS.NE.0) THEN
        PRINT *, 'Transform unit cell before Fourier summation'
      ENDIF
      RETURN
C
 1010 FORMAT(/,5X,'A',7X,'B',7X,'C',5X,'COS ALP',2X,'COS BET',2X,
     1     'COS GAM',3X,'ALP',4X,'BET',4X,'GAM',/' ',3F8.2,3F9.5,3F7.2)
 1020 FORMAT(/,5X,'A*',6X,'B*',6X,'C*',4X,'COS ALP*',1X,'COS BET*',1X,
     1     'COS GAM*',3X,'ALP*',3X,'BET*',3X,'GAM*',/  ' ',
     2     3F8.5,3F9.5,1X,3F7.2)
 1030 FORMAT(/' Number of equivalent reflections (HKL) is',I3,//,
     1     5X,'TRANSFORMED X',9X,'TRANSFORMED Y',9X,'TRANSFORMED Z'/)
      END

C  ****************************************************************************
C  *                                                                          *
C  *              *****     SUBROUTINE SF3DBT_TRANS     *****                 *
C  *                                                                          *
C  *    Sections calculated parallel to a general plane                       *
C  *     The U Matrix or the equation for the general plane is supplied       *
C  *                                                                          *
C  *    ROUTINE CALLED BY "EMSF3DBT" PROGRAM                                  *
C  ****************************************************************************
C
      SUBROUTINE SF3DBT_TRANS (DETU, AA, pixsizunits)
C
      IMPLICIT NONE
C
      COMMON /CONST/ TRIG_TABLE, XMIN, YMIN, ZMIN, XMAX, YMAX,
     1 ZMAX, DX, DY, DZ, DCELL, ITRANS, HMAX, KMAX, KMIN, RCELL,
     2 SCALE, TS, U, UINV
      REAL*4    XMIN, YMIN, ZMIN, XMAX, YMAX, ZMAX, DX, DY, DZ
      REAL*4    DCELL(9), RCELL(9), SCALE, TS(3,24), U(3,3)
      REAL*4    DETU, AA(3,3), TRIG_TABLE(1001)
      REAL*4    DPLANE, PLANE(3), DDCELL, CELL(3), GNEW(3,3)
      REAL*4    XYZMIN(3), XYZMAX(3), DXYZ(3), UINV(3,3)
      INTEGER*4 ITRANS, HMAX, KMAX, KMIN, I, J, K, KK, pixsizunits
      EQUIVALENCE  (XYZMIN,XMIN), (XYZMAX,XMAX), (DXYZ,DX)
C  ----------------------------------------------------------------------------
C
      IF(ITRANS == 1) THEN
         READ (*,*) ((U(I,J),J=1,3),I=1,3)
         WRITE(*,1000)
         DO I = 1,3
           DO J = 1,3
             GNEW(I,J) = 0.0
             DO K = 1,3
               DO KK = 1,3
                 GNEW(I,J) = GNEW(I,J) + U(I,K)*U(J,KK)*AA(K,KK)
               ENDDO
             ENDDO
           ENDDO
         ENDDO
         DCELL(1) = SQRT(GNEW(1,1))
         DCELL(2) = SQRT(GNEW(2,2))
         DCELL(3) = SQRT(GNEW(3,3))
         DCELL(6) = GNEW(1,2)/(DCELL(1)*DCELL(2))
         DCELL(5) = GNEW(1,3)/(DCELL(1)*DCELL(3))
         DCELL(4) = GNEW(2,3)/(DCELL(2)*DCELL(3))
       ELSE
          READ (*,*) PLANE(1), PLANE(2), PLANE(3), DPLANE, CELL(1),
     $         CELL(2), CELL(3)
         WRITE(*,1010) PLANE(1), PLANE(2), PLANE(3), DPLANE
         CALL SF3DBT_GP1 (PLANE, DPLANE, AA, U)
         READ (*,*) XMIN,XMAX,YMIN,YMAX,ZMIN,ZMAX,DX,DY,DZ
         DO I = 1,3
           if(pixsizunits == 1) then
              WRITE(*,1020) I, XYZMIN(I), XYZMAX(I), DXYZ(I)
            else if (pixsizunits == 2) then
              WRITE(*,1021) I, XYZMIN(I), XYZMAX(I), DXYZ(I)
            else
              WRITE(*,1022) I, XYZMIN(I), XYZMAX(I), DXYZ(I)
           endif
           DDCELL     = AMAX1(2.*ABS(XYZMIN(I)),2.*ABS(XYZMAX(I)),
     1                       10.,CELL(I))
           U(I,1)     = U(I,1)*DDCELL
           U(I,2)     = U(I,2)*DDCELL
           U(I,3)     = U(I,3)*DDCELL
           XYZMIN(I)  = XYZMIN(I)/DDCELL
           XYZMAX(I)  = XYZMAX(I)/DDCELL
           DXYZ(I)    = DXYZ(I)/DDCELL
           DCELL(I)   = DDCELL
           DCELL(I+3) = 0.0
         ENDDO
      ENDIF
      WRITE(*,1030) (DCELL(I),I=1,6)
      WRITE(*,1040) (I,(U(I,J),J=1,3),I=1,3)
      CALL SF3DBT_MI3 (U, UINV, DETU)
      WRITE(*,1050) DETU
      WRITE(*,1060) ((UINV(I,J),J=1,3),I=1,3)
      RETURN
C
 1000      FORMAT(' Transformation matrix read from control data file')
 1010      FORMAT(' Transformation matrix defined by the plane',/,11X,
     1 F10.5,' X +',F10.5,' Y +',F10.5,' Z =',F10.5)
 1020      FORMAT(' New axis',I2,' runs from',F10.5,' to',F10.5,
     1 ' in steps of',F10.5,' Angstroms')
 1021      FORMAT(' New axis',I2,' runs from',F10.5,' to',F10.5,
     1 ' in steps of',F10.5,' nms')
 1022      FORMAT(' New axis',I2,' runs from',F10.5,' to',F10.5,
     1 ' in steps of',F10.5,' pixels')
 1030      FORMAT(' New cell constants are',3F12.3,3F12.5)
 1040      FORMAT(11X,'NEWA',I1,' =',F12.6,' A1 +',F12.6,' A2 +',F12.6,
     1 ' A3')
 1050      FORMAT(' The determinant of the U matrix is',F15.6)
 1060      FORMAT(' The matrix inverse to U is as follows'
     $          ,/,(11X,3F15.6))
      END

C  ****************************************************************************
      SUBROUTINE SF3DBT_CHEKUP (XMIN, XMAX, NORIEN, N, DX)
C
      IMPLICIT NONE
C
      INTEGER*4 NORIEN, N, NX
      REAL*4    XMIN, XMAX, DX, DIFFX
C  ----------------------------------------------------------------------------
C
  100      DIFFX = XMAX-XMIN
      IF(DIFFX == 0.0) THEN                  ! COMPUTE ONLY 1 SECTION
        NORIEN =  N
        DX     = .000001
        RETURN
      ENDIF
      IF(DIFFX.LT.0.0) THEN                  ! SWITCH XMIN AND XMAX
        XMIN = XMAX
        XMAX = XMIN-DIFFX
        GO TO 100
      ENDIF
      IF(DIFFX.GT.1.0) THEN                  ! TRAP FOR XMAX-XMIN > 1.0
        XMAX = XMIN+1.0
      ENDIF
      IF(DX.LE.0.0) THEN
        XMAX = XMIN
        GO TO 100
      ENDIF
      IF(DIFFX == 1.0) THEN                  ! SET FOR INTEGRAL SAMPLING
        NX   = NINT(DIFFX/DX)
        DX   = DIFFX/NX
        XMAX = XMIN+DIFFX-DX
      ENDIF
C
      RETURN
      END

C  ****************************************************************************
      SUBROUTINE SF3DBT_CROSS (A, X, Y, Z)
C
      IMPLICIT NONE
C
      INTEGER*4 I
      REAL*4    A(3,3), X(3), Y(3), Z(3), ZZ(3)
C  ----------------------------------------------------------------------------
C
      ZZ(1) = X(2)*Y(3)-X(3)*Y(2)
      ZZ(2) = X(3)*Y(1)-X(1)*Y(3)
      ZZ(3) = X(1)*Y(2)-X(2)*Y(1)
      DO I = 1,3
        Z(I) = A(I,1)*ZZ(1)+A(I,2)*ZZ(2)+A(I,3)*ZZ(3)
      ENDDO
C
      RETURN
      END

C  ****************************************************************************
      SUBROUTINE SF3DBT_GETARGS (ARG, SINARG, COSARG, TRIG_TABLE)
C
      IMPLICIT NONE
C
      INTEGER*4 NARG1, NARG2
      REAL*4    TRIG_TABLE(1001), SINARG, COSARG, ARG
C  ----------------------------------------------------------------------------
C
      ARG = AMOD(ARG,1.0)
      IF(ARG.LT.0.0) ARG = ARG + 1.0
      NARG1 = INT(1000.*ARG + 1.5)
      NARG2 = NARG1 + 250
      IF(NARG2.GT.1000) NARG2 = NARG2 - 1000
      SINARG = -TRIG_TABLE(NARG2)
      COSARG =  TRIG_TABLE(NARG1)
C
      RETURN
      END

C  ***********************************************************************
C  *                                                                     *
C  *              *****     SUBROUTINE SF3DBT_GP1     *****              *
C  *                                                                     *
C  *    S/R TO FIND TRANSFORMATION MATRIX TO AXES APPROPIATE TO PLANE    *
C  *     CELL2=U*CELL1     CELL2(1) IS NORMAL TO PLANE                   *
C  ***********************************************************************
C
      SUBROUTINE SF3DBT_GP1 (PLANE, D, AA, U)
C
      INTEGER*4 K
      REAL*4    PLANE(3), U1(3), U2(3), U3(3), U(3,3)
      REAL*4    AA(3,3), BB(3,3), D, ALPHA
C  ----------------------------------------------------------------------------
C
      CALL SF3DBT_MI3 (AA, BB, DETA)
      IF(D) 10,20,20
   10      PLANE(1) = -PLANE(1)
      PLANE(2) = -PLANE(2)
      PLANE(3) = -PLANE(3)
   20      ALPHA    = 1./SQRT(SF3DBT_VMV(PLANE,BB,PLANE))
      PLANE(1) = ALPHA*PLANE(1)
      PLANE(2) = ALPHA*PLANE(2)
      PLANE(3) = ALPHA*PLANE(3)
      CALL SF3DBT_VM (PLANE, BB, U1)
      K = 1
      IF(ABS(PLANE(1)).GE.ABS(PLANE(2))) K = 2
      IF(ABS(PLANE(K)).GE.ABS(PLANE(3))) K = 3
      ALPHA = -SIGN(PLANE(K),U1(K))
      U2(1) = ALPHA*U1(1)
      U2(2) = ALPHA*U1(2)
      U2(3) = ALPHA*U1(3)
      U2(K) = U2(K)+1.0
      CALL SF3DBT_NORM (AA, U2)
      CALL SF3DBT_CROSS (BB, U1, U2, U3)
      CALL SF3DBT_NORM (AA, U3)
      DO K = 1,3
        U(1,K) = U1(K)
        U(2,K) = U2(K)
        U(3,K) = U3(K)
      ENDDO
C
      RETURN
      END

C  ****************************************************************************
      SUBROUTINE SF3DBT_MI3 (A, B, D)
C
      IMPLICIT NONE
C
      INTEGER*4 I, J
      REAL*4    A(3,3), B(3,3), D
C  ----------------------------------------------------------------------------
C
      B(1,1) =   A(2,2)*A(3,3) - A(2,3)*A(3,2)
      B(1,2) = -(A(1,2)*A(3,3) - A(1,3)*A(3,2))
      B(1,3) =   A(1,2)*A(2,3) - A(1,3)*A(2,2)
      B(2,1) = -(A(2,1)*A(3,3) - A(2,3)*A(3,1))
      B(2,2) =   A(1,1)*A(3,3) - A(1,3)*A(3,1)
      B(2,3) = -(A(1,1)*A(2,3) - A(1,3)*A(2,1))
      B(3,1) =   A(2,1)*A(3,2) - A(2,2)*A(3,1)
      B(3,2) = -(A(1,1)*A(3,2) - A(1,2)*A(3,1))
      B(3,3) =   A(1,1)*A(2,2) - A(1,2)*A(2,1)
      D      =   A(1,1)*B(1,1) + A(2,1)*B(1,2) + A(3,1)*B(1,3)
C
      IF(D == 0.0) RETURN
C
      DO J = 1,3
        DO I = 1,3
          B(I,J) = B(I,J)/D
        ENDDO
      ENDDO
C
      RETURN
      END

C  ****************************************************************************
      SUBROUTINE SF3DBT_NORM (A, X)
C
      IMPLICIT NONE
C
      REAL*4 ALPHA, SF3DBT_VMV, A(3,3),X(3)
C  ----------------------------------------------------------------------------
C
      ALPHA = SQRT(1.0/SF3DBT_VMV(X,A,X))
      X(1)  = ALPHA*X(1)
      X(2)  = ALPHA*X(2)
      X(3)  = ALPHA*X(3)
C
      RETURN
      END

C  ****************************************************************************
      SUBROUTINE SF3DBT_ROT (D, E)
C
      IMPLICIT NONE
C
      REAL*4 ALPHA, D, E
C  ----------------------------------------------------------------------------
C
      ALPHA = D
      D     = E
      E     = ALPHA
C
      RETURN
      END
C  ****************************************************************************
      SUBROUTINE SF3DBT_ROT_I2 (D, E)
C
      IMPLICIT NONE
C
      INTEGER*4 ALPHA, D, E
C  ----------------------------------------------------------------------------
C
      ALPHA = D
      D     = E
      E     = ALPHA
C
      RETURN
      END
C  ****************************************************************************
      SUBROUTINE SF3DBT_ROT_I4 (D, E)
C
      IMPLICIT NONE
C
      INTEGER*4 ALPHA, D, E
C  ----------------------------------------------------------------------------
C
      ALPHA = D
      D     = E
      E     = ALPHA
C
      RETURN
      END

C  ****************************************************************************
      SUBROUTINE SF3DBT_STOAA (Z, A)
C
      IMPLICIT NONE
C
      REAL*4 A(3,3), Z(6)
C  ----------------------------------------------------------------------------
C
      A(1,1) = Z(1)*Z(1)
      A(1,2) = Z(1)*Z(2)*Z(6)
      A(1,3) = Z(1)*Z(3)*Z(5)
      A(2,1) = A(1,2)
      A(2,2) = Z(2)*Z(2)
      A(2,3) = Z(2)*Z(3)*Z(4)
      A(3,1) = A(1,3)
      A(3,2) = A(2,3)
      A(3,3) = Z(3)*Z(3)
C
      RETURN
      END

C  ****************************************************************************
      SUBROUTINE SF3DBT_VM (X, A, Y)
C
      IMPLICIT NONE
C
      INTEGER*4 I
      REAL*4    A(3,3), X(3), Y(3), ZZ(3)
C  ----------------------------------------------------------------------------
C
      DO I = 1,3
        ZZ(I) = X(1)*A(1,I)+X(2)*A(2,I)+X(3)*A(3,I)
      ENDDO
      Y(1) = ZZ(1)
      Y(2) = ZZ(2)
      Y(3) = ZZ(3)
C
      RETURN
      END

C  ****************************************************************************
      REAL*4 FUNCTION SF3DBT_VMV (X, A, Y)
C
      IMPLICIT NONE
C
      REAL*4    X(3), A(3,3), Y(3)
      INTEGER*4 I
C  ----------------------------------------------------------------------------
C
      SF3DBT_VMV = 0.0
      DO I = 1,3
        SF3DBT_VMV = SF3DBT_VMV +
     1              X(I)*(A(I,1)*Y(1) + A(I,2)*Y(2) + A(I,3)*Y(3))
      ENDDO
C
      RETURN
      END

C  ****************************************************************************
      FUNCTION IFIXR (A_CELL)
C
      IMPLICIT NONE
C
      INTEGER*4 IFIXR
      REAL*4    A_CELL
C  ----------------------------------------------------------------------------
      IFIXR = IFIX(A_CELL+SIGN(.5,A_CELL))
C
      RETURN
      END

C  ****************************************************************************
      FUNCTION NUCALC (N)
C
      IMPLICIT NONE
C
      INTEGER*4 NUCALC, N
C  ----------------------------------------------------------------------------
      NUCALC = MOD(N,12)-(ISIGN(1,N)-1)*6
C
      RETURN
      END
