/*  File: libUtil.c 
 *
 * Generic utility routines.
 *
 * Rob Ashmore
 * Purdue University
 * October 2000
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "../include/common.h"
#include "../include/libUtil.h"


char *banner(char *application) 
{
  // function name: banner
  // Print the current version number and today's date to give the
  // executable a time/version stamp

  static char bannerOut[1024];
  pid_t pid;

  pid = getpid();
  sprintf(bannerOut,"%s-%s",application, EM_VERSION);
  printf("PID[%d][%s %s %s]\n",pid,bannerOut,__DATE__,__TIME__);
  return(bannerOut);
  
} // end function banner

char *banner_(char *application) {
  return(banner(application));
}

char *banner__(char *application) {
  return(banner(application));
}



double atof_fstr(char *fstr) {
  return(atof(fstr));
}

double atof_fstr_(char *fstr) {
  return(atof(fstr));
}

double atof_fstr__(char *fstr) {
  return(atof(fstr));
}



void flush_stdout() 
{
  // function name: flush_stdout
  // force buffered data to be written to standard out
  
  fflush(stdout);
  return;
  
} //end function flush_stdout



void flush_stdout_() 
{
  // function name: flush_stdout_
  // force buffered data to be written to standard out

  fflush(stdout);
  return;
  
} //end function flush_stdout_



void get_str(FILE *flptr, char brkchr, char *string)
{
  // function name: get_str
  // Get a string of ascii characters. Allocate any storage necessary
  // here locally. Trash any leading white space found.  Also trash
  // any trailing WS.

  char ans = ' ', *s1;
  
  while(!feof(flptr) && ans==' ')
    ans= fgetc(flptr);         // Eat the white space
  s1= string;
  while(!feof(flptr) && (ans!='\n' && ans!=brkchr)) {
    *(s1++)= ans;
    ans= fgetc(flptr);
  }
  while((*(s1-1)==' ' || *(s1-1)=='\t') && s1 > string)
    s1--;
  *s1= '\0';                   // Terminate the string
  
} // end function get_str



float ico_calcphimax(float theta)
{
  // function name: ico_calcphimax
  //
  // Calculate the maximum allowable value of the angle phi in the
  // icosahedral asymmetic unit for a given value of theta. Calculation
  // is based on spherical triognometry relations described in CRC
  // Standard Mathematical Tables
  //
  // Define a spherical triangle (on surface of a sphere) made up of
  // three sides (a,b,c) and opposite each side are three angles
  // (A,B,C). All side and angle measurements are in degrees. Each
  // side is an arc of a great circle. For the edge of the icosahedral
  // asymmetric unit between the 3-fold and 5-fold axes, the angle and
  // side lengths are as follows
  //
  // a = final latitude (theta) = 90.0                         
  // b = initial latitude (theta) = 69.1                         
  // c = arc between 3-fold and 5-fold axis                         
  //
  // A = initial bearing (at 3-fold axis)                         
  // B = final reverse bearing (at 5-fold axis)                   
  // C = change in longitude (phi) = 31.7                         
  //
  // D. Belnap, purdue university vax 4000/90 and dec 3000, december 1993
  // R. Sinkovits, 10/4/2010 Conversion to C

  float phimax, A, B, A_minus_B, A_plus_B, temp;
  float a           = 90.0;
  float b           = 69.09484255211070;  // theta for 3-fold axis
  float C           = 31.71747441146101;  // phi for 5-fold axis
  double deg_to_rad = acos(-1.0)/180.0;
  double rad_to_deg = 1.0/deg_to_rad;
  
  //  To calculate A, the initial bearing from 3-fold to 5-fold axis and
  //  B, the final reverse bearing, solve for A & B with napier's
  //  analogies
  //
  //          temp = tan[(A-B)/2], then temp = tan[(A+B)/2]                  
  //
  //  To calculate C, the distance (in degrees) from 3-fold to 5-fold
  //  axis, use gauss's formula
  //
  //         sin[C/2] = ( sin[(a+b)/2]sin[C/2] ) / cos[(A-B)/2]  
  //
  //  Only A and B are needed for the phimax calculation.

  temp = (1/tan(C*deg_to_rad/2.0)) * 
    (sin((a-b)*deg_to_rad/2.0) / sin((a+b)*deg_to_rad/2.0)) ;
  
  A_minus_B = 2*rad_to_deg*atan(temp);
  
  temp = (1/tan(C*deg_to_rad/2.0)) *
    (cos((a-b)*deg_to_rad/2.0) / cos((a+b)*deg_to_rad/2.0)) ;

  A_plus_B = 2*rad_to_deg*atan(temp) ;

  B = (A_plus_B - A_minus_B) / 2.0;
  A = A_minus_B + B;
  
  // Calculate phimax for theta (a)
  // Calculate B, use law of sines  temp = sin(B)
  // Calculate C (phimax), use Napier's analogies temp = tan(C/2)

  a    = theta;
  temp = sin(A*deg_to_rad)*sin(b*deg_to_rad) / sin(a*deg_to_rad);
  B    = rad_to_deg*asin(temp) ;
  temp = sin((a-b)*deg_to_rad/2.0) / 
    ( sin((a+b)*deg_to_rad/2.0) * tan((A-B)*deg_to_rad/2.0) ) ;
  phimax  = 2*rad_to_deg*atan(temp);

  return(phimax);

} // end function ico_calcphimax

float ico_calcphimax_(float theta) {
  return(ico_calcphimax(theta));
}
float ico_calcphimax__(float theta) {
  return(ico_calcphimax(theta));
}



float find_phimax(int *sym, float *theta)
{
  // function name: find_phimax
  // Get the maximin value of phi given symmetry and theta

  float phimax = 180.0;

  if(*sym == 532)
    phimax = ico_calcphimax(*theta);
  else if (*sym >= 2 && *sym <= 20)
    phimax = 360.0/(*sym*2);
  else if (*sym == 222) 
    phimax = 360.0/4.0;
  else if (*sym == 32) 
    phimax = 360.0/6.0;
  else if (*sym == 422) 
    phimax = 360.0/8.0;
  else if (*sym == 52) 
    phimax = 360.0/10.0;
  else if (*sym == 622) 
    phimax = 360.0/12.0;
  else if (*sym == 72) 
    phimax = 360.0/14.0;
  else if (*sym == 822) 
    phimax = 360.0/16.0;
  else if (*sym == 92) 
    phimax = 360.0/18.0;

  return(phimax);

} // end function find_phimax

float find_phimax_(int *sym, float *theta) {
  return(find_phimax(sym, theta));
}
float find_phimax__(int *sym, float *theta) {
  return(find_phimax(sym, theta));
}
