/************************************************************************
                             param_srt.c

        use heapsort to sort the EMPFT output data on different
        column in increasing order. param_srt() is the main routine,
        heapsort_rank() is a helper routine.
*************************************************************************
   Example usage:

        call from forturn program:
                call param_srt(key,nrow,%val(fids),%val(ids),%val(ts),
                1       %val(ps),%val(os),%val(xs),%val(ys),%val(mags),
                2       %val(pft_cc),%val(prj_cc),%val(cmp_cc))

        call from c program:
                param_srt(&key,&nrow,fids,ids,ts,ps,os,xs,ys,mags,
                          pft_cc,prj_cc,cmp_cc);

   param_srt():
	input:
	      int *key      : the address of an integer indicates the 
			      way to sort.
		     key = 1: sort by file IDs value (int fids[])
	             key = 2: sort by particle IDs value (int ids[])
		     key = 3: sort by theta value (float ts[]) 
	             key = 4: sort by phi value (float ps[])
		     key = 5: sort by omiga value (float os[])
		     key = 6: sort by center's x value (float xs[])
	             key = 7: sort by center's y value (float ys[])
		     key = 8: sort by magnification value (float mags[])
		     key = 9: sort by pft_cc value (float pft_cc[])
		     key =10: sort by prj_cc value (float prj_cc[])
		     key =11: sort by cmp_cc value (float cmp_cc[])
		     key =12: sort by file ID value first, then for 
		       	      same file ID, sort by particle ids again.
		     key =13: sort by deltaX value (float deltaX[])
		     key =14: sort by deltaY value (float deltaY[])
		     key =15: sort by abs(deltaX) value (float deltaX[])
		     key =16: sort by abs(deltaY) value (float deltaY[])
		     key =17: sort by distance value (float distance[])
		     key =18: sort by theta(float ts[]) THEN phi(float ps[])

	      int *nrow	   : address of an integer which is the 
			     number of rows in EMPFT output data.
	      int *fids	   : address of an integer array which keep
			     all file IDs.
	      int *ids	   : address of an integer array which keep 
			     all particle IDs.
	      float *ts    : address of a float array which keep
			     all theta values.
	      float *ps    : address of a float array which keep
			     all phi values.
	      float *os    : address of a float array which keep
			     all omiga values.
	      float *xs    : address of a float array which keep
			     all center's x values.
	      float *ys    : address of a float array which keep
			     all center's y values.
	      float *mags  : address of a float array which keep
			     all magnification values.
	      float *pft_cc: address of a float array which keep
			     all pft_cc values.
	      float *prj_cc: address of a float array which keep
			     all prj_cc values.
	      float *cmp_cc: address of a float array which keep
			     all cmp_cc values.
	output: 
	      	sorted data will be saved in original data arrays:
		fids[], ids[], ts[], ps[], os[], xs[], ys[], mags[]
		pft_cc[], prj_cc[], cmp_cc[].		

    heapsort_rank():
	input:
	      int n	  : number of rows in rank array need to be 
			    sorted, start form rank[0] to rank[n-1].
	      int *rank   : the integer array which indicates the  
			    rank order of reference data.
	      float *data : the reference data array, which need to
			    be re-ordered. 
	output:
	      sorted data rank will be saved in original array rank[].

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "../include/common.h"

/********************************************************************
 heapsort_rank(): param_srt() 's helper routine,
		  use Heapsort to sort the ranks of data rows.
*********************************************************************/
int heapsort_rank(int n, int *rank, float *data, int incSort){
	int 	i, j, heap_id, tail, temp_rank;

	/* sort ranks of data rows by related data value */
	if(n < 2) 
	  return 0;
	heap_id = (n >> 1);
	tail = n - 1;
	for(;;) {
	  if(heap_id > 0) {
	    temp_rank = rank[--heap_id];
	  }
	  else {
	    temp_rank = rank[tail];
	    rank[tail] = rank[0];
	    if( --tail == 0) {
	      rank[0] = temp_rank;
	      break;
	    }
	  }
	  i = heap_id;
	  j = heap_id + 1;
	  while (j <= tail) {
	    if(j < tail && 
	       ((incSort) ? (data[rank[j]] > data[rank[j+1]])
		: (data[rank[j]] < data[rank[j+1]]))
	       )
	      j++;
	    if((incSort) ? (data[temp_rank] > data[rank[j]])
		: (data[temp_rank] < data[rank[j]]))
	      {
		rank[i] = rank[j];
		i = j;
		j <<= 1;
	      } 
	    else
	      j = tail + 1;
	  }
	  rank[i] = temp_rank;
	}
	return 0;
}


int bubblesort_rank(int n, int *rank, float *data, int incSort,
		    int doAbs)
{/* BUBBLESORT_RANK */
  int i, sorted, tmpInt,nm1;

  if(doAbs)
    {/* DO it in absolute value */
      sorted = FALSE;
      nm1 = n - 1;
      while(!sorted)
	{/* Sort it */
	  sorted = TRUE;
	  for(i=0;i<nm1;i++)
	    {/* Look at all */
	      if((incSort) ? (fabs(data[rank[i]]) > fabs(data[rank[i+1]]))
		 : (fabs(data[rank[i]]) < fabs(data[rank[i+1]])))
		{/* Switch them */
		  tmpInt = rank[i];
		  rank[i] = rank[i+1];
		  rank[i+1] = tmpInt;
		  sorted = FALSE;
		}/* Switch them */
	    }/* Look at all */
	}/* Sort it */
    }/* DO it in absolute value */
  else
    {/* Normal */
      sorted = FALSE;
      nm1 = n - 1;
      while(!sorted)
	{/* Sort it */
	  sorted = TRUE;
	  for(i=0;i<nm1;i++)
	    {/* Look at all */
	      if((incSort) ? (data[rank[i]] > data[rank[i+1]])
		 : (data[rank[i]] < data[rank[i+1]]))
		{/* Switch them */
		  tmpInt = rank[i];
		  rank[i] = rank[i+1];
		  rank[i+1] = tmpInt;
		  sorted = FALSE;
		}/* Switch them */
	    }/* Look at all */
	}/* Sort it */
    }/* Normal */
  
  return(0);
}/* BUBBLESORT_RANK */



/********************************************************************
 param_srt_rwa_(): 	sort empft output data by centain value,
  		call heapsort_rank to sort the ranks of rows.
		(for unix version linker)
*********************************************************************/
int param_srt_rwa_(int *key_ptr, int *nrow_ptr, int *fids, int *ids, 
	       float *ts, float *ps, float *os, float *xs, 
	       float *ys, float *mags, float *pft_cc, 
	       float *prj_cc, float *cmp_cc, 
	       float *deltaX, float *deltaY, float *distance,
	       int *incSort) 
{
	int 	i, *rank, key, nrow, *subrank;
	float 	*buf, *fdata;
	int doAbs;

	nrow = *nrow_ptr;
	key = *key_ptr; 
	/* allocate memory for ranks array and initialize it */
	if((rank = (int *)malloc(nrow*sizeof(int))) == NULL) {
	  perror("param_srt:1no enough memory");
	  return -1;
	}
	if((subrank = (int *)malloc(nrow*sizeof(int))) == NULL) {
	  perror("param_srt:1no enough memory");
	  return -1;
	}
	for(i=0; i<nrow; i++) 
	  rank[i] = i;

	/* allocate memeory for float data buffer */
	if((buf = (float *)malloc(nrow*sizeof(float))) == NULL) {
	  perror("param_srt2:no enough memory");
	  free(rank);
	  free(subrank);
	  return -1;
	}

	/* I'm changing all of the heapsort_rank()
	   * calls below to bubblesort_rank() until we can debug the
	   * heapsort_rank() rtn.  heapsort_rank() gives
	   * up too quickly for large data sets
	   * e.g.  it sorts what sorts OK but it doesn't finish
	   * the job... ???
	   * Rob Ashmore   June 1998
	   */


	/* switch by the key to sort in different ways */
	doAbs = 0;
	switch (key) {
	  case 1:
		/* cast the fids[] into float array fdata[] */
		fdata = buf;	
		for(i=0;i<nrow;i++)
		  fdata[i] = (float)fids[i];	
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;
	  case 2:
		/* cast the ids[] into float array fdata[] */
		fdata = buf;
		for(i=0;i<nrow;i++)
		  fdata[i] = (float)ids[i];	
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;
	  case 3:
		fdata = ts;
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;
	  case 4:
		fdata = ps;
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;
	  case 5:
		fdata = os;
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;
	  case 6:
		fdata = xs;
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;
	  case 7:
		fdata = ys;
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;
	  case 8:
		fdata = mags;
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;
	  case 9:
		fdata = pft_cc;
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;
	  case 10:
		fdata = prj_cc;	
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;
	  case 11:
		fdata = cmp_cc;
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;
	  case 12:
		/* What we want is to sort on file id firsdt then in 
		 * particle id - Let's just kludge this for now by 
		 * multiplying the file id by 100000 and then adding the 
		 * particle id and then sorting on that.  This will be 
		 * quicker but has the drawback that they can't have more 
		 * than 100000 particles per file.
		 * Rob Ashmore June 1998
		 */
		/******* first level sorting on fids[] *******/
		/* cast the fids[] into float array fdata[] */
		fdata = buf;	
		for(i=0;i<nrow;i++)
		  fdata[i] = (float)((fids[i]*100000)+ids[i]);	

		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
	  case 13:
		fdata = deltaX;
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;

	  case 14:
		fdata = deltaY;
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;

	  case 15:
		fdata = deltaX;
		doAbs = 1;
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;

	  case 16:
		fdata = deltaY;
		doAbs = 1;
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;

	  case 17:
		fdata = distance;
		bubblesort_rank(nrow, rank, fdata, *incSort, doAbs);
		break;


		break;
	  default:
		printf("Undefined key value!\n");
		free(rank);
	  	free(subrank);
		free(buf);
		return -1;
	}
		
	
	/* fill in output array, in order of rank[] */
	for(i=0; i<nrow; i++) buf[i] = (float)fids[rank[i]];
	for(i=0; i<nrow; i++) fids[i] = (int)buf[i];
	for(i=0; i<nrow; i++) buf[i] = (float)ids[rank[i]]; 
	for(i=0; i<nrow; i++) ids[i] = (int)buf[i]; 
	for(i=0; i<nrow; i++) buf[i] = ts[rank[i]]; 
	for(i=0; i<nrow; i++) ts[i] = buf[i];
	for(i=0; i<nrow; i++) buf[i] = ps[rank[i]];
	for(i=0; i<nrow; i++) ps[i] = buf[i]; 
	for(i=0; i<nrow; i++) buf[i] = os[rank[i]]; 
	for(i=0; i<nrow; i++) os[i] = buf[i]; 
	for(i=0; i<nrow; i++) buf[i] = xs[rank[i]]; 
	for(i=0; i<nrow; i++) xs[i] = buf[i]; 
	for(i=0; i<nrow; i++) buf[i] = ys[rank[i]]; 
	for(i=0; i<nrow; i++) ys[i] = buf[i]; 
	for(i=0; i<nrow; i++) buf[i] = mags[rank[i]]; 
	for(i=0; i<nrow; i++) mags[i] = buf[i]; 
	for(i=0; i<nrow; i++) buf[i] = pft_cc[rank[i]]; 
	for(i=0; i<nrow; i++) pft_cc[i] = buf[i]; 
	for(i=0; i<nrow; i++) buf[i] = prj_cc[rank[i]]; 
	for(i=0; i<nrow; i++) prj_cc[i] = buf[i]; 
	for(i=0; i<nrow; i++) buf[i] = cmp_cc[rank[i]]; 
	for(i=0; i<nrow; i++) cmp_cc[i] = buf[i]; 
	for(i=0; i<nrow; i++) buf[i] = deltaX[rank[i]]; 
	for(i=0; i<nrow; i++) deltaX[i] = buf[i]; 
	for(i=0; i<nrow; i++) buf[i] = deltaY[rank[i]]; 
	for(i=0; i<nrow; i++) deltaY[i] = buf[i]; 
	for(i=0; i<nrow; i++) buf[i] = distance[rank[i]]; 
	for(i=0; i<nrow; i++) distance[i] = buf[i]; 
	    
	free(rank); 
 	free(subrank);
	free(buf);
	return 0;
}
	
/********************************************************************
 param_srt_rwa(): 	sort empft output data by centain value,
  		call heapsort_rank to sort the ranks of rows.
		(for vms version linker)
*********************************************************************/
int param_srt_rwa(int *key_ptr, int *nrow_ptr, int *fids, int *ids, 
	      float *ts, float *ps, float *os, float *xs, 
	      float *ys, float *mags, float *pft_cc, 
	      float *prj_cc, float *cmp_cc, 
	      float *deltaX, float *deltaY, float *distance,
	      int *incSort) 
{
	return param_srt_rwa_(key_ptr, nrow_ptr, fids, ids, ts, ps,
			  os, xs, ys, mags, pft_cc, prj_cc, cmp_cc,
			  deltaX, deltaY, distance, incSort);
}



