// ONED - Program to teach the principles of FFT in 1 dimension.
//    Tim Baker
//    Biology Department
//    Purdue University
//    January 1997

#include "../include/common.h"
#include "oned.h"
#include "../bitmaps/reset.xbm"
#include "../bitmaps/noreset.xbm"

// ----------------------------------------------------------------
// OS transfer point. The main routine does all the one-time setup
// and then calls XtAppMainLoop.
// ----------------------------------------------------------------

int main(int argc, char **argv)
{
  char pathenv[1024];
  FILE *fIn;
  String fallbacks[] =
    {"*fontList:-dec-terminal-bold-r-normal--14-140-75-75-c-80-iso8859-1=tag",
     NULL
    };
  
  // Show them our version number
  banner("oned");
  
  // Don't start listening to the windows until they are all up
  // and running
  inited= FALSE;
  
  // Initialize MRM
  MrmInitialize(); 
                
  // Initialize the application
  toplevel_widget= XtVaAppInitialize(&app_context, argv[0], NULL, 0, 
				     &argc, argv, fallbacks, NULL);

  // Remember the display we're using 
  display = XtDisplay(toplevel_widget);
  
  // Construct ':' delimited path(s) to be searched for uid file
  memset(pathenv, '\0', 1024);
  strcpy(pathenv, argv[0]);
  for (i=strlen(pathenv)-1; i>=0; i--) {
    if(pathenv[i] == '/') {
      pathenv[i] = ':';
      break;
    }
    else
      pathenv[i] = '\0';
  }
  strcat(pathenv, getenv("PATH"));

  db_filename_vec[0] = (char *)malloc(sizeof(char)*512);
  memset(db_filename_vec[0], '\0', 512);
  j = 0;

  // Check each element of path for uid file
  for(i=0; i<strlen(pathenv); i++) {
    if(pathenv[i] != ':') {
      db_filename_vec[0][j] = pathenv[i];
      j++;
    } 
    else {
      strcat(db_filename_vec[0], "/oned.uid");
      if( (fIn = fopen(db_filename_vec[0],"r")) != NULL ) {
	fclose(fIn);
	break;
      } 
      else {
	j = 0;
	memset(db_filename_vec[0], '\0', 512);
      }
    }
  }

  if(j == 0) {
    printf("Unable to find oned.uid in PATH\n");
    printf("Make sure that the uid file is in same directory as oned\n");
    exit(0);
  }

  printf("Full path to UID file: [%s]\n",db_filename_vec[0]);
  
  // Open UID file in hierarchy
  db_filename_num = (sizeof db_filename_vec / sizeof db_filename_vec [0]);
  if (MrmOpenHierarchy(db_filename_num, // Number of files
		       db_filename_vec, // Array of file names
		       NULL,            // Default OS extenstion
		       &s_MrmHierarchy) // Pointer to returned MRM ID
      !=MrmSUCCESS)
    s_error("can't open hierarchy");
  
  // Register the items (callbacks) MRM needs to bind for us.
  MrmRegisterNames(reglist, reglist_num);
  
  // Go get the main part of the application
  if (MrmFetchWidget(s_MrmHierarchy, "S_MAIN_WINDOW", toplevel_widget,
		     &main_window_widget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch main window");
  
  // Go get the infoScreen part of the application
  if (MrmFetchWidget(s_MrmHierarchy, "infoMessageBox", toplevel_widget,
		     &infoBox, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch info window");
  
  // Manage the main part and realize everything. The interface comes
  // up on the display now. Manage all of the screens now so we can
  // alter what we need to in XLibInitialize() - Unmanage the unneeded
  // ones thereafter.

  XtManageChild(main_window_widget);
  XtManageChild(infoBox);
  XFlush(display);

  XlibInitialize();
  XtUnmanageChild(infoBox);

  // Create our cursors
  XFlush(display);
  inited= TRUE;
   
  // Initialize our C code
  init_globals(argc, argv);

  // Sit around forever waiting to process X-events.  We never leave
  // XtAppMainLoop. From here on, we only execute our callback routines
  XtAppMainLoop(app_context);
  
}



// ***********************************************************************
// NON-Callback Routines - These rotuines are of the more generic C      *
// language variety and are not called directly from X.  The callback    *
// routines that are called by X do make heavy use of these, however.    *
// ***********************************************************************

void s_error(char *problem_string)
{
  // function name: s_error
  // All errors are fatal
  
  printf("%s\n", problem_string);
  exit(0);
  
} // end function s_error



void init_globals(int argc, char **argv)
{
  // function name: init_globals
  // Init out anything that needs it before we start executing. See
  // groundZero() for anything that needs initing later on in the
  // execution.

  int randNum;

  // Filename descriptor for FORTRAN
  fileName.addr = (char *)calloc(2048, sizeof(char));  
  fileName.length = 2048;
  fileName.type = 14;
  fileName.class = 1;

  curOned = 0;
  onedScrollValue = 0;

  for(i=0; i<12; i++) { // Do all scrolls
    modelState[i]        = FALSE;
    ampScrollValue[i]    = 0.0;
    phaseScrollValue[i]  = 0.0;
    wavnumScrollValue[i] = 1;
  }

  for(i=0; i<512; i++) {
    totModel[i].iy1 = 0;
    totModel[i].iy2 = 0;
  }
  
  formUp= MODEL;     // We start in MODEL mode
  filterSwitch = FALSE;
  floatSwitch = TRUE;
  numFilter= 8;
  numFilter2= 1;
  noiseSwitch = FALSE;
  tempSwitch = FALSE;
  allonSwitch = FALSE;
  numNoise= 1.0;
  numTemp= 256.0;         // Start it ridiculouly high
  numSigNoise= 10.0;

  numSamps= 64;
  numMotifs= 8.0;
  maxNumMotifs= 8.0;

  ndata= 256;
  
  // Seed the random number generator
  srandom(1);
  factor= 1.0 / DENOMINATOR;
      
  // Calculate the initial noise spectrum
  for(i=0;i<512;i++) { // Do all pixles
    randNum= random();
    // noiseSpectrum[i]= (float)(randNum*factor);
    noiseSpectrum[i]= (float)(((float)randNum)/
			      ((float)DENOMINATOR));
  }
  
  ranging= FALSE;
  beginAnn= 256;
  endAnn=512;

  for (i=0; i<512; i++) {
    filterRemember[i].iy1 = 0;
    filterRemember[i].iy2 = 0;
    noiseFactor[i] = 0.0;
  }

  annulusSwitch = FALSE;

} // end function init_globals



void createDefaultColormap()
{
  // function name: createDefaultColormap
  // Do whatever is necessary to get us the desired colormap.

  Bool contig;
  int tryColors, success;
  unsigned int nplanes;
  float scaleFactor;
  FILE *fColor;
  
  visual= DefaultVisual(display, DefaultScreen(display));
  visual_depth = (unsigned int)DefaultDepth(display, DefaultScreen(display));
  
  // Allocate the color cells
  pixels_return= (unsigned long *)malloc(sizeof(unsigned long)*256);

  // Allocate the color definition structs
  if (color_defs == NULL) 
    color_defs = (XColor *) malloc (1025 * sizeof (XColor) );
  
  color_map = XCreateColormap(display, DefaultRootWindow(display),
			      visual, AllocNone);
  
  if (color_map == 0) {
    fprintf (stderr, "Could not create color map for visual\n");
    exit (1);
  }
  
  // Now let's query the DISPLAY until we get the maximum number of
  // contiguous colorcell entries in the colormap that we can.
  
  success= FALSE;
  tryColors= 256;
  nplanes = 1;
  contig = True;
  nplanes = 0;   
  
  // Init the color defs to point to each entry
  for (i = 0; i <256 ; i++) { // Make the color map gray-scale
    color_defs[i].pixel= i;
    pixels_return[i] = i;
  }
  
  // Find out what's already there in the color table
  XQueryColors(display, color_map, color_defs, 256);
//GCA: not clear the reason ot write the file - commenting for now
/*  sprintf(outStr,"%s/cmap.robem",getenv("HOME"));
  fColor = fopen(outStr,"w");
  if(fColor) { // Write it
    fwrite(&tryColors,sizeof(int),1,fColor);
    for(i=0;i<tryColors;i++) { // Write it all out
      fwrite(&pixels_return[i],sizeof(unsigned long),1,fColor);
    }
  }
  else {
    printf("ERROR - Cannot write colormap file.\n");
  }
  fclose(fColor);
*/  
  // Now change the colormap for drawing in pixels_return[GREEN,RED,BLUE] 
  // First calculate some fudge factors.
  
  pixelColor= (65535.0/256.0);
  scaleFactor = 255.0 / ((float)tryColors - 7.0);
  
  // Make these entries in the colormap GRAYSCALE
  for(i=6;i<(tryColors);i++) { // Do the colors we have been assigned
    color_defs[pixels_return[i]].pixel = pixels_return[i];
    color_defs[pixels_return[i]].red   = (unsigned short)
      (((i-6.0) * scaleFactor) * pixelColor);
    color_defs[pixels_return[i]].green = (unsigned short)
      (((i-6.0) * scaleFactor) * pixelColor);
    color_defs[pixels_return[i]].blue  = (unsigned short)
      (((i-6.0) * scaleFactor) * pixelColor);
    color_defs[pixels_return[i]].flags = DoRed | DoGreen | DoBlue;
    color_defs[pixels_return[i]].pad = 0;
    success = XAllocColor (display, color_map, &color_defs[i]);
    pixels_return[i] = color_defs[i].pixel;
  }
  
  // Now define our robem colors
  color_defs[WHITE].pixel = WHITE;
  color_defs[WHITE].red   = 255 * pixelColor;
  color_defs[WHITE].green = 255 * pixelColor;
  color_defs[WHITE].blue  = 255 * pixelColor;
  color_defs[WHITE].flags = DoRed | DoGreen | DoBlue;
  XAllocColor (display, color_map, &color_defs[WHITE]);
  pixels_return[WHITE] = color_defs[WHITE].pixel;
  
  color_defs[YELLOW].pixel = YELLOW;
  color_defs[YELLOW].red   = 255 * pixelColor;
  color_defs[YELLOW].green = 255 * pixelColor;
  color_defs[YELLOW].blue  = 0 * pixelColor;
  color_defs[YELLOW].flags = DoRed | DoGreen | DoBlue;
  XAllocColor (display, color_map, &color_defs[YELLOW]);
  pixels_return[YELLOW] = color_defs[YELLOW].pixel;
  
  color_defs[GREEN].pixel = GREEN;
  color_defs[GREEN].red   = 0;
  color_defs[GREEN].green = 255 * pixelColor;
  color_defs[GREEN].blue  = 0;
  color_defs[GREEN].flags = DoRed | DoGreen | DoBlue;
  XAllocColor (display, color_map, &color_defs[GREEN]);
  pixels_return[GREEN] = color_defs[GREEN].pixel;

  color_defs[pixels_return[RED]].pixel = RED;
  color_defs[pixels_return[RED]].red   = 255 * pixelColor;
  color_defs[pixels_return[RED]].green = 0;
  color_defs[pixels_return[RED]].blue  = 0;
  color_defs[pixels_return[RED]].flags = DoRed | DoGreen | DoBlue;
  XAllocColor (display, color_map, &color_defs[pixels_return[RED]]);
  pixels_return[RED] = color_defs[RED].pixel;

  color_defs[BLUE].pixel = BLUE;
  color_defs[BLUE].red   = 0;
  color_defs[BLUE].green = 0;
  color_defs[BLUE].blue  = 255 * pixelColor;
  color_defs[BLUE].flags = DoRed | DoGreen | DoBlue;
  XAllocColor (display, color_map, &color_defs[BLUE]);
  pixels_return[BLUE] = color_defs[BLUE].pixel;

  color_defs[PLACEHOLDER].pixel = PLACEHOLDER;
  color_defs[PLACEHOLDER].red   = 0 * pixelColor;
  color_defs[PLACEHOLDER].green = 255 * pixelColor;
  color_defs[PLACEHOLDER].blue  = 0;
  color_defs[PLACEHOLDER].flags = DoRed | DoGreen | DoBlue;
  XAllocColor (display, color_map, &color_defs[PLACEHOLDER]);
  pixels_return[PLACEHOLDER] = color_defs[PLACEHOLDER].pixel;

  color_defs[BLACK].pixel = BLACK;
  color_defs[BLACK].red   = 0 * pixelColor;
  color_defs[BLACK].green = 0 * pixelColor;
  color_defs[BLACK].blue  = 0 * pixelColor;
  color_defs[BLACK].flags = DoRed | DoGreen | DoBlue;
  XAllocColor (display, color_map, &color_defs[BLACK]);
  pixels_return[BLACK] = color_defs[BLACK].pixel;
  
} // end function createDefaultColormap



void XlibInitialize()
{
  // function name: XlibInitialize
  // Initialize for Xlib calls. Setup the colormap we'll use and
  // graphics contexts. Create all of the Images and allocate memory
  // for all of our XmDrawingAreas.

  static const char dotted[8]= {4,4,4,4};
  XtCallbackRec callBck[5];    // Our callbacks
  
  // Init dimensions
  buffer_width  = image_width  = NUMCOLS;
  buffer_height = image_height = NUMROWS;
  oned_width  = ONED_WIDTH;
  oned_height = ONED_HEIGHT;
  image_colors = 3;

  // Get visual
  visual= DefaultVisual(display, DefaultScreen(display));
  visual_depth = (unsigned int)DefaultDepth(display, DefaultScreen(display));
  
  createDefaultColormap();
  
  // Create Image buffers for our drawingAreas
  // MAIN drawing area on main screen
  image_image1 = XCreateImage (display, visual, visual_depth, 
			       ZPixmap, 0, NULL, oned_width, 
			       oned_height, 32, 0);
  if (image_image1 == NULL) {
    fprintf (stderr, "XCreateImage (image) failed\n");
    exit(1);
  }
  
  // Get Window pointers from MOTIF
  image_window1= XtWindow(drawingWidget);
  
  // Now tell MOTIF what kind of events you'll listen to in this window.
  event_mask  = ButtonPressMask;
  event_mask |= ButtonReleaseMask   | KeyPressMask | KeyReleaseMask;
  event_mask |= PointerMotionMask   | LeaveWindowMask | ExposureMask;
  XSelectInput(display, image_window1, event_mask);
   
  // Make this colormap is the one that's used!
  XInstallColormap (display, color_map);  //Install the color map
  
  // Allocate for our image storage
  image_buffer1 = (BYTE *)malloc ((oned_width*4*oned_height)*
				  (sizeof(BYTE)));
  
  // Couldn't get allocate enough storage?
  if (image_buffer1 == NULL) {
    perror ("malloc image_buffer1");
    exit(1);
  }
  for(i=0; i<oned_height*4; i++)
    for(j=0; j<oned_width; j++)
      image_buffer1[(i*oned_width)+j]= (BYTE)PLACEHOLDER;
  
  // OK tell the image where to find it's data
  image_image1->data = (char *)image_buffer1;
  
  // Setup Graphics contexts
  fontstruct =  XLoadQueryFont(display, FONT);
  if (fontstruct  == NULL) {
    fprintf(stderr, "%s: display %s doesn't know font %s\n",
	    "oned", "DisplayString(display)", FONT);
    exit(1);
  }

  gc_mask = 0;
  gc_values.function = GXcopy;		gc_mask |= GCFunction;
  gc_values.plane_mask = AllPlanes;	gc_mask |= GCPlaneMask;
  gc_values.foreground = 1;	        gc_mask |= GCForeground;
  gc_values.background = black_pixel;	gc_mask |= GCBackground;
  gc_values.graphics_exposures = False;	gc_mask |= GCGraphicsExposures;
  gc_values.font = fontstruct->fid;     gc_mask |= GCFont; 
  image_gc = XCreateGC (display, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[RED];  gc_mask |= GCForeground;
  image_gcRED = XCreateGC (display, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[GREEN];  gc_mask |= GCForeground;
  image_gcGREEN = XCreateGC (display, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[BLUE];  gc_mask |= GCForeground;
  image_gcBLUE = XCreateGC (display, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[BLACK];  gc_mask |= GCForeground;
  image_gcBLACK = XCreateGC (display, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[YELLOW];  gc_mask |= GCForeground;
  image_gcYELLOW = XCreateGC (display, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[WHITE];  gc_mask |= GCForeground;
  image_gcWHITE = XCreateGC (display, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[GREEN];  gc_mask |= GCForeground;
  gc_values.line_style = LineOnOffDash; gc_mask |= GCLineStyle;
  image_gcGREENDASH = XCreateGC (display, image_window1, gc_mask, &gc_values);
  XSetDashes(display, image_gcGREENDASH, 0, dotted, 4);
  
  // Display the image in the window created above
  image_width    = DEF_DISPLAY_SIZE;   // This changes whenever a new image
  image_height   = DEF_DISPLAY_SIZE;   // is read - make sure we draw it all
  display_width  = image_width;
  display_height = image_height;
  
  XPutImage (display, image_window1, image_gc, image_image1, 0, 0,
	     0, 0, (image_width), image_height);
  
  XFlush (display);
  
  // Make reset Pixmaps
  resetPixmap = 
    XCreatePixmapFromBitmapData(display, image_window1,
				reset_bits, reset_width, reset_height,
				pixels_return[BLACK], pixels_return[WHITE],
				visual_depth);
  noresetPixmap = 
    XCreatePixmapFromBitmapData(display, image_window1,
				noreset_bits, noreset_width, noreset_height,
				pixels_return[BLACK], pixels_return[WHITE],
				visual_depth);

  for(i=0; i<11; i++) { // Do all
    XtSetArg(arglist[0], XmNx, 440);
    XtSetArg(arglist[1], XmNy, (8+(i*40)) );
    XtSetArg(arglist[2], XmNheight, 22);
    XtSetArg(arglist[3], XmNwidth, 22);
    XtSetArg(arglist[4], XmNlabelType, XmPIXMAP);
    XtSetArg(arglist[5], XmNlabelPixmap, resetPixmap);
    XtSetArg(arglist[6], XmNselectPixmap, noresetPixmap);
    XtSetArg(arglist[7], XmNrecomputeSize, FALSE);
    resetButton[i]= (Widget)XmCreatePushButton(onedscreen, "", 
					       arglist, 8);
    XtManageChild(resetButton[i]);
    // Add callbacks to our widgets
    callBck[0].callback= reset_activate;
    callBck[0].closure= 0;
    callBck[1].callback= 0;
    callBck[1].closure= 0;
    XtAddCallbacks(resetButton[i], XmNactivateCallback, callBck);  
    
  }
  
} // end function XlibInitialize



void filterFiles(char *filesStr)
{
  // function name: filterFiles
  // Set the file selection widget to filter for the fileStr files.

  XmString tcs;
  
  tcs = XmStringLtoRCreate(filesStr, XmSTRING_DEFAULT_CHARSET);
  XmFileSelectionDoSearch(fileSel_widget, tcs);
  //  XtFree(tcs);
  XFlush(display);
  
} // end function filterFiles


// *********************************************************************
// CALLBACK routines - Rotuines called directly from X whenever the    *
// appropriate event occurs.                                           *
// *********************************************************************


static int askThem(char *infoStr, int systemModal, int dialogType)
{
  // function name: askThem
  // This is the generic display a message and get them to respond
  // screen. We get a return value from globAsk which is set by
  // infoOK() or infoCancel() (see below).

  XmString simpleString;

  // Init and try to position the window
  globAsk = -99;
  XFlush(display);
  simpleString = XmStringCreateSimple(infoStr);
  XtSetArg(arglist[0], XmNmessageString, simpleString);
  XtSetArg(arglist[1], XmNdialogType, dialogType);
  XtSetArg(arglist[2], XmNx, 150);
  XtSetArg(arglist[3], XmNy, 150);
  XtSetValues(infoBox, arglist, 4);
  XFlush(display);
  XtFree((char *)simpleString);
  
  // Freeze the display until they give us an answer!!!  This is
  // important and we need an answer!!!!
  // Revisit this to make sure we absolutely need the answer.
  // e.g. They may be able to continue without the input ???

  if(systemModal) { // Make them respond to this
    XtSetArg(arglist[0], XmNdialogStyle, XmDIALOG_SYSTEM_MODAL);
    XtSetValues(infoBox, arglist, 1);
  }
  
  // Init it and pop it up sit here and wait until we get something back
  gFinished= FALSE;
  XtManageChild(infoBox);
  XtPopup(XtParent(infoBox), XtGrabNone);
  
  XmProcessTraversal(infoBox, XmTRAVERSE_CURRENT); 
  XtRealizeWidget(infoBox);  
  XFlush(display);
  
  // gFinished is set in the infoOK() or infoCancel() callback
  while(!gFinished)
    XtAppProcessEvent(app_context, XtIMAll);
  
  // Pop it down and return their response
  XtUnmanageChild(infoBox);
  XtPopdown(XtParent(infoBox));
  XtSetArg(arglist[0], XmNdialogStyle, XmDIALOG_MODELESS);
  XtSetValues(XtParent(infoBox), arglist, 1);
  XFlush(display);
  
  return(globAsk);
  
} // end function askThem



void infoOK(Widget widget, int *tag, 
	    XmAnyCallbackStruct *callback_data)
{
  // function name: infoOK
  // If the OK button is pressed from the infoscreen return here. Set
  // gFinished to cause the value of globAsk be return to the starting
  // point (see askThem() above).

  globAsk = TRUE;
  XtUnmanageChild(infoBox);
  gFinished= TRUE;

} // end function infoOK



void infoCancel(Widget  widget, int *tag, 
		XmAnyCallbackStruct *callback_data)
{
  // function name: infoCancel
  // If the CANCEL button is pressed from the infoscreen return here.
  // Set gFinished to cause the value of globAsk be return to the
  // starting point (see askThem() above).

  globAsk = FALSE;
  XtUnmanageChild(infoBox);
  gFinished= TRUE;

} // end function infoCancel



void new_activate(Widget widget, int *tag, 
		  XmAnyCallbackStruct *callback_data)
{
  // function name: new_activate
  // They selected the "New" submenu. Reinit the inerface so they
  // can start over.

  int i,j;
  XmString simpleString;
  
  for(j=0; j<11; j++) { // DO all models
    modelState[j]= FALSE;
    XtSetArg(arglist[0], XmNset, modelState[j]);
    XtSetValues(modelToggle[j], arglist, 1); 
    
    ampScrollValue[j]= 0.0;
    phaseScrollValue[j]= 0.0;
    
    XtSetArg(arglist[0], XmNlabelString, XmStringCreateLocalized("0"));
    XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
    XtSetValues(phaseLabel[j], arglist, 2); 
    XtSetValues(ampLabel[j], arglist, 2); 
    
    XtSetArg(arglist[0], XmNvalue, 0);
    XtSetValues(ampScroll[j], arglist, 1);
    XtSetArg(arglist[0], XmNvalue, 0);
    XtSetValues(phaseScroll[j], arglist, 1);
    
    XSync(display, FALSE);
    XFlush(display);
    
    for(i=0; i<515; i++) { // Do entire model
      model[(j*515)+i].ix1 = 0.0;
      model[(j*515)+i].ix2 = 0.0;
      model[(j*515)+i].ix3 = 0.0;
      model[(j*515)+i].iy1 = 0.0;
      model[(j*515)+i].iy2 = 0.0;
    }
    for(i=0; i<512; i++) {
      totModel[i].ix1 = 0.0;
      totModel[i].ix2 = 0.0;
      totModel[i].ix3 = 0.0;
      totModel[i].iy1 = 0.0; 
      totModel[i].iy2 = 0.0;
      sampCrystal[i].ix1 = 0.0;
    }
  }
  
  // Give feedback on main window title bar
  simpleString = XmStringCreateLocalized("ONED: [Unnamed]");
  XtSetArg(arglist[0], XmNdialogTitle, simpleString );
  XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
  XtSetValues(main_window_widget, arglist, 2); 
  XFlush(display);
  XmStringFree(simpleString); 
  XFlush(display);
  
} // end function new_activate



void read_activate(Widget widget, int *tag, 
		   XmAnyCallbackStruct *callback_data)
{
  // function name: read_activate
  // They selected the "Read" submenu. Put up file selection
  // interface so they can tell us what filename to use.

  readWrite= READING;
  XtManageChild(fileSel_widget);
  
} // end function read_activate



void save_activate(Widget widget, int *tag, 
		   XmAnyCallbackStruct *callback_data)
{
  // function name: save_activate
  // They selected the "Save" submenu. Put up file selection interface
  // so they can tell us what filename to use.
  
  readWrite= WRITING;
  XtManageChild(fileSel_widget);
  
} // end function save_activate



void exit_activate(Widget widget, int *tag, 
		   XmAnyCallbackStruct *callback_data)
{
  // function name: exit_activate
  // They said they wanted to exit the program
  
    exit(0);
  
} // end function exit_activate



void formResize(Widget widget, int *tag, 
		XmDrawingAreaCallbackStruct *callback_data)
{
  // function name: formResize
  // When the main screen formDialog is resized call here. Get the new
  // window size and setup our globals.

  int width, height;
  Window tWin;
  XWindowAttributes winAttrib;

  if(inited) {
    tWin = (Window)XtWindow(drawingWidget);
    
    // Get new window dimensions
    if(tWin > 0) { // Valid window
      XGetWindowAttributes(display,tWin,&winAttrib);
      width  = winAttrib.width;
      height = winAttrib.height;
    }
    
    // Remember the dimensions of this
    display_width = width;
    display_height= height; 
  }
  
} // end function formResize



void plotWindow_create(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{
  // function name: plotWindow_create
  // Remember the drawing area window pointer for use later. Setup the
  // file selection widget here.

  XmString tcs;
  XtCallbackRec callBck[5];    // Our callbacks
  
  // Remember the widget number for later
  drawingWidget= widget;
  
  // Setup the file selection widget
  fileSel_widget = XmCreateFileSelectionDialog(main_window_widget,
					       "Get file", NULL, 0);
  tcs = XmStringLtoRCreate("*.*", XmSTRING_DEFAULT_CHARSET);
  
  XtSetArg(arglist[0], XmNx, 10);
  XtSetArg(arglist[1], XmNy, 10);
  XtSetArg(arglist[2], XmNdirMask, tcs);  
  XtSetArg(arglist[3], XmNdefaultPosition, FALSE); 
  XtSetValues(fileSel_widget, arglist, 4);

  // Add callbacks to file selection widget
  callBck[0].callback = gotFilename;
  callBck[0].closure  = 0;
  callBck[1].callback = 0;
  callBck[1].closure  = 0;
  XtAddCallbacks(fileSel_widget, XmNokCallback, callBck);  
  callBck[0].callback = cancel_fileSel;
  callBck[0].closure  = 0;
  callBck[1].callback = 0;
  callBck[1].closure  = 0;
  XtAddCallbacks(fileSel_widget, XmNcancelCallback, callBck);  
  XmStringFree(tcs);

} // end function plotWindow_create



void calcModel(int modelNum)
{
  // function name: calcModel
  // Recalculate the model

  int i;
  float X, N;

  N = (float)modelNum;
  for(i=0; i<514; i++) {
    X = (ampScrollValue[modelNum]*0.01) * 
      cos(phaseScrollValue[modelNum]+(THISTWOPI*(N)*i)/256.0);
    model[((modelNum*515)+i)].ix1= X;
    model[((modelNum*515)+i)].iy1= (20 +(40 *modelNum)) - (int)(X*20);
    X = (ampScrollValue[modelNum]*0.01) * 
      cos(phaseScrollValue[modelNum]+(THISTWOPI*(N)*(i+1))/256.0);
    model[((modelNum*515)+i)].ix2= X;
    model[((modelNum*515)+i)].iy2= (20 +(40 *modelNum)) - (int)(X*20);
  }

} // end function calcModel



void DrawModel(int modelNum)
{
  // function name: DrawModel
  // Redraw the model

  int i;
  
  calcModel(modelNum);
  for(i=0; i<511; i++)
    XDrawLine(display, image_window1, image_gcRED,
	      i,   model[modelNum*515 + i].iy1,
	      i+1, model[modelNum*515 + i].iy2);
  
  XFlush (display);
  
} // end function DrawModel



void DrawWindow()
{
  // function name: DrawWindow
  // Redraw window e.g. on expose events or whenever it's prudent.

  int draw_width, draw_height, i, tot;
  float y1, y2, ymax, scaleFactor;
  char numStr[32];
  
  if(formUp == FFTSCREEN) {
    DrawFFTWindow();
  }
  else { // Draw model window
    // Draw all of image
    draw_width  = display_width;
    draw_height = display_height;
    
    // Hard code for now
    draw_width = draw_height = 512;
    
    XPutImage(display, image_window1, image_gc, image_image1, 0, 0,
	      0, 0,  draw_width,  draw_height);
    
    XFlush (display);
    
    // Draw axes
    XDrawLine(display, image_window1, image_gcGREEN,
	      0, 0, 512, 0);

    for(i=40; i<442; i+=40) { // Do all curves
      XDrawLine(display, image_window1, image_gcGREEN,
		0, i, 512, i);
      XDrawLine(display, image_window1, image_gcBLUE,
		0, (i-20), 512, (i-20));
      sprintf(numStr,"%d",(int)((i-1)/40));
      XDrawString(display, XtWindow(drawingWidget), image_gcWHITE,
		  2, (i-10), numStr, strlen(numStr));
    }
    
    // Draw TOT axes
    XDrawLine(display, image_window1, image_gcGREEN,
	      0, 511, 512, 511);
    XDrawLine(display, image_window1, image_gcBLUE,
	      0, 475, 512, 475);
    sprintf(numStr,"Tot");
    XDrawString(display, XtWindow(drawingWidget), image_gcWHITE,
		2, (511-10), numStr, strlen(numStr));

    for(i=20; i<472; i+=40) { // Do tick marks curves
      for(j=64; j<512; j+=64) { // Draw ticks on x axis
	if(i==460) { // DO tot one
	  if(j==256)
	    XDrawLine(display, image_window1, image_gcWHITE,
		      j,(i+15-10), j, (i+15+10));
	  else
	    XDrawLine(display, image_window1, image_gcBLUE,
		      j, (i+15-5), j, (i+15+5));
	}
	else { // normal 
	  if(j==256)
	    XDrawLine(display, image_window1, image_gcWHITE,
		      j, (i-10), j, (i+10));
	  else
	    XDrawLine(display, image_window1, image_gcBLUE,
		      j, (i-5), j, (i+5));
	}
      }
    }
    
    // Draw each model
    for(j=0;j<11;j++)
      DrawModel(j);
    
    // Now tot them up
    ymax = -99999.9;
    for(i=0; i<513; i++) {
      y1 = y2 = 0.0;
      tot = 0;
      for(j=0;j<11;j++) { // Do all models
	if(modelState[j]) { // Do this one
	  y1 += model[(j*515)+i].ix1;
	  y2 += model[(j*515)+i].ix2;
	  tot++;
	}
      }
      if(tot > 0) {
	totModel[i].ix1 = y1;
	totModel[i].ix2 = y2;
	ymax = MAX(ymax, fabs(y1));
	ymax = MAX(ymax, fabs(y2));
      }  
    }

    if(ymax!=0.0)
      scaleFactor= 1.0 / ymax;
    else 
      scaleFactor= 1.0;

    for(i=0;i<511;i++) { // Do all models
      if(tot > 0) {
	y1 = totModel[i].ix1 * scaleFactor;
	y2 = totModel[i].ix2 * scaleFactor;
	totModel[i].ix1 *= scaleFactor;
	totModel[i].ix2 *= scaleFactor;
	y1 = 475.0 - (y1*35.0);
	y2 = 475.0 - (y2*35.0);
	XDrawLine(display, image_window1, image_gcYELLOW,
		  i, (int)y1, i+1, (int)y2);
      }
	  
    }
  }
  
} // end function DrawWindow



void DrawFFTWindow()
{
  // function name: DrawFFTWindow
  // Redraw window e.g. on expose events or whenever it's prudent.

  int draw_width, draw_height, i;
  char numStr[32];
  
  // Draw all of image
  draw_width  = display_width;
  draw_height = display_height;

  // Hard code for now
  draw_width = draw_height = 512;
  XPutImage(display, image_window1, image_gc, image_image1, 0, 0,
	    0, 0,  draw_width,  draw_height);
  
  XFlush (display);
  
  // Calc the interps
  oned_terp_c();
  oned_fft_c();

  // Draw the FFT
  DrawFFTplot();

  sprintf(numStr,"Original");
  XDrawString(display, XtWindow(drawingWidget), image_gcWHITE,
	      2, 10, numStr, strlen(numStr));
  sprintf(numStr,"Reconstructed");
  XDrawString(display, XtWindow(drawingWidget), image_gcRED,
	      2, 138, numStr, strlen(numStr));
  sprintf(numStr,"1D FFT abs(Amp)");
  XDrawString(display, XtWindow(drawingWidget), image_gcGREEN,
	      2, 266, numStr, strlen(numStr));
  if(annulusSwitch)
    sprintf(numStr,"A");
  else 
    sprintf(numStr,"N");
  XDrawString(display, XtWindow(drawingWidget), image_gcBLUE,
	      2, 280, numStr, strlen(numStr));
  
  // Draw the backtransformed Crystal
  XDrawLine(display, image_window1, image_gcGREENDASH,
	    0, 192, 512, 192);

  for(i=0;i<510;i++) { // Do all of the window
    float thisIX1, thisIX2;
    thisIX1 = backCrystal[i].ix1-dfloatBack;
    thisIX2 = backCrystal[i+1].ix1-dfloatBack;
    thisIX1 = MIN(MAX(thisIX1,-1.1),1.1);
    thisIX2 = MIN(MAX(thisIX2,-1.1),1.1);
    XDrawLine(display, image_window1, image_gcRED,
	      i, (192 - (int)(thisIX1*63.0)),
	      i+1, (192 - (int)(thisIX2*63.0)));
  }
  
  // Draw the "perfect" crystal
  XDrawLine(display, image_window1, image_gcGREENDASH,
	    0, 64, 512, 64);

  for(i=0;i<511;i++) { // Do all of the window
    float thisIX1, thisIX2;
    thisIX1 = (perfCrystal[i].ix1 - dfloat);
    thisIX2 = (perfCrystal[i+1].ix1 - dfloat);
    XDrawLine(display, image_window1, image_gcWHITE,
	      i, (64 - (int)(thisIX1*63.0)),
	      i+1, (64 - (int)(thisIX2*63.0)));
  }

}// end function DrawFFTWindow



void DrawFFTplot()
{
  // function name: DrawFFTplot
  // Redraw fft plot
  
  int i;
  
  for(i=0; i<512; i++) { // Do all of the window
    float thisIX1, thisIX2;
    thisIX1 = fftCrystal[i].ix1;
    thisIX2 = fftCrystal[i+1].ix1;
    if(fftCrystal[i].iy1==0)
      XDrawLine(display, image_window1, image_gcGREEN,
		i, 510, i, (510 - (int)(thisIX1*250.0)));
    else if(fftCrystal[i].iy1==99)
      XDrawLine(display, image_window1, image_gcBLUE,
		i, 510, i, 505);
    if(fftCrystal[i].iy2==77)
      XDrawLine(display, image_window1, image_gcRED,
		i, 505, i, 500);
    
    if(tempSwitch && (i < 511) && (i>0)) { // draw temp factors
      XDrawLine(display, image_window1, image_gcYELLOW,
		i, 511-(int)(tempFactors[i]*256.0),
		i, 511-(int)(tempFactors[i+1]*256.0));
    }
  }
  
  XFlush(display);
  
} // end function DrawFFTplot



void imageClick(Widget widget, int *tag, 
		XmDrawingAreaCallbackStruct *draw_data)
{
  // function name: imageClick
  // Every time the mouse button is clicked call here. This routine
  // handles the selecting/deselecting of box points and ruler points
  // on the screen.
  //
  //    Button1 is the Left mouse button
  //    Button2 is the Middle mouse button
  //    Button3 is the Right mouse button

  char ans;
  int x, y, i;
  XEvent event;

  if(draw_data != 0) { // Only look at events meant for us
    event= *(draw_data->event);
    
    switch (event.type) {

    case ButtonRelease:
      // They are trying to find points
      x = (int)event.xbutton.x;      // Get x,y copords of point pos
      y = (int)event.xbutton.y;
      if(ranging) { // They are ending the range
	if(annulusSwitch) { // Just let the annulus pass through
	  for(i=256; i<512; i++) {
	    if(i>=beginAnn && i<=endAnn) {
	      fftCrystal[i].iy2=0;       
	      fftCrystal[(256-(i-256))].iy2=0;       
	    }
	    else {
	      fftCrystal[i].iy2=77;       
	      fftCrystal[(256-(i-256))].iy2=77;       
	    }
	  }
	}
	else { // Negative universe of the above
	  for(i=beginAnn; i<=endAnn; i++) {
	    fftCrystal[i].iy2=77;       
	    fftCrystal[(256-(i-256))].iy2=77;       
	  }
	}
      }
      ranging= FALSE;
      DrawWindow();
      break;
      
    case ButtonPress:
      // They are trying to find points
      x = event.xbutton.x;      // Get x,y copords of point pos
      y = event.xbutton.y;
      if(event.xbutton.button == Button1) { // They are starting the range
	if(x < 256)
	  beginAnn= 256+ (256-x);
	else 
	  beginAnn= x;
	for(i=0; i<512; i++)
	  fftCrystal[i].iy2= 0;
	ranging= TRUE;
      }
      if(event.xbutton.button == Button2) { // They are starting the range
	if(x < 256)
	  endAnn= 256+ (256-x);
	else 
	  endAnn= x;
	for(i=0; i<512; i++)
	  fftCrystal[i].iy2= 0;
	ranging= TRUE;
      }
      if(event.xbutton.button == Button3) {// Default annuli
	for(i=0;i<512;i++)
	  fftCrystal[i].iy2= 0;
	ranging= FALSE;
	beginAnn= 256;
	endAnn= 512;
      }
      break;
	  
    case KeyPress:
      ans= XLookupKeysym((XKeyEvent *)&event, 0);
      // printf ("Keyed[%c]\n",ans);
      if(ans=='a' || ans=='A')
	annulusSwitch = !annulusSwitch;
      if(annulusSwitch) { // Just let the annulus pass through
	for(i=256;i<512;i++) {
	  if(i>=beginAnn && i<=endAnn) {
	    fftCrystal[i].iy2=0;       
	    fftCrystal[(256-(i-256))].iy2=0;       
	  }
	  else {
	    fftCrystal[i].iy2=77;       
	    fftCrystal[(256-(i-256))].iy2=77;       
	  }
	}
      }
      else { // Negative universe of the above
	for(i=256; i<512; i++) { // Init all
	  fftCrystal[i].iy2=0;       
	  fftCrystal[(256-(i-256))].iy2=0;       
	}
	for(i=beginAnn; i<=endAnn; i++) {
	  fftCrystal[i].iy2=77;       
	  fftCrystal[(256-(i-256))].iy2=77;       
	}
      }
      DrawWindow();
      break;
	  
    default:
      break;
    }
  }
  
  XFlush(display);   // Make changes current
  
} // end function imageClick



void cancel_fileSel(Widget widget, char *tag, 
		    XmFileSelectionBoxCallbackStruct *callback_data)
{
  // function name: cancel_fileSel 
  // They want to abort the selection of a file name.
  
  XtUnmanageChild(fileSel_widget);
  
} // end function cancel_fileSel



void gotFilename(Widget w, char *client_data, 
		 XmFileSelectionBoxCallbackStruct *callback_data)
{
  // function name: gotFilename
  // Get the filename they specified in file selection widget. Branch
  // according to what operation they want performed on this filename.

  char oldFilename[256], outStr[128];
  char *s1, inStr[128];
  int status, newFormat;
  FILE *fIn;
  XmString simpleString;

  // Pop down the file selection widget
  XtUnmanageChild(fileSel_widget);
  XFlush(display);
  
  // Get the filename
  sprintf(oldFilename,"%s",fileName.addr);
  if(!XmStringGetLtoR(callback_data->value, 
		      XmSTRING_DEFAULT_CHARSET, 
		      &fileName.addr))
    printf("Error extracting filename.\n");
  
  sprintf(oldFilename,"%s",fileName.addr);
  s1= strrchr(oldFilename,';');
  if(s1)
    *s1= '\0';    // Get rid of version number
  
  if(readWrite==READING) { // Read the file
    
    // Give feedback on main window title bar
    sprintf(outStr,"ONED: [%s]",oldFilename);
    simpleString = XmStringCreateLocalized(outStr);
    XtSetArg(arglist[0], XmNdialogTitle, simpleString );
    XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
    XtSetValues(main_window_widget, arglist, 2); 
    XFlush(display);
    XmStringFree(simpleString); 
    XFlush(display);
    
    fIn= fopen(oldFilename,"r");
    if(fIn) {
      
      newFormat = FALSE;
      for(i=0;i<8;i++)
	fscanf(fIn,"%c",&inStr[i]);
      inStr[8]= '\0';
      if(EQU(inStr,"Portable")) {
	newFormat = TRUE;
	// Read other input
	fscanf(fIn,"%s",inStr);
	fscanf(fIn,"%s",inStr);
      }
      else
	rewind(fIn);
      
      for(j=0;j<11;j++) { // Do all models
	if(newFormat)
	  fscanf(fIn,"%d",&modelState[j]);
	else
	  fread(&modelState[j],sizeof(int),1,fIn);
	XtSetArg(arglist[0], XmNset, modelState[j]);
	XtSetValues(modelToggle[j], arglist, 1); 
	
	if(newFormat) {
	  fscanf(fIn,"%f",&ampScrollValue[j]);
	  fscanf(fIn,"%f",&phaseScrollValue[j]);
	}
	else {
	  fread(&ampScrollValue[j],sizeof(float),1,fIn);
	  fread(&phaseScrollValue[j],sizeof(float),1,fIn);
	}
	phaseScrollValue[j]*= ((float)360.0/THISTWOPI);
	sprintf(outStr,"%.0f",phaseScrollValue[j]);
	simpleString = XmStringCreateLocalized(outStr);
	XtSetArg(arglist[0], XmNlabelString, simpleString);
	XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
	XtSetValues(phaseLabel[j], arglist, 2); 
	XmStringFree(simpleString); 
	XFlush(display);
	sprintf(outStr,"%.0f",ampScrollValue[j]);
	simpleString = XmStringCreateLocalized(outStr);
	XtSetArg(arglist[0], XmNlabelString, simpleString);
	XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
	XtSetValues(ampLabel[j], arglist, 2); 
	XmStringFree(simpleString); 
	XFlush(display);
	
	XtSetArg(arglist[0], XmNvalue, (int)ampScrollValue[j]);
	XtSetValues(ampScroll[j], arglist, 1);
	XtSetArg(arglist[0], XmNvalue, (int)phaseScrollValue[j]);
	XtSetValues(phaseScroll[j], arglist, 1);
	
	phaseScrollValue[j]*= ((float)THISTWOPI/360.0);
	
	XSync(display,FALSE);
	XFlush(display);
	
	for(i=0;i<515;i++) { // Do entire model
	  if(newFormat) {
	    fscanf(fIn,"%f %f %f %d %d",
		   &model[(j*515)+i].ix1,
		   &model[(j*515)+i].ix2,
		   &model[(j*515)+i].ix3,
		   &model[(j*515)+i].iy1,
		   &model[(j*515)+i].iy2);
	  }
	  else {
	    fread(&model[(j*515)+i].ix1,sizeof(float),1,fIn);
	    fread(&model[(j*515)+i].ix2,sizeof(float),1,fIn);
	    fread(&model[(j*515)+i].ix3,sizeof(float),1,fIn);
	    fread(&model[(j*515)+i].iy1,sizeof(int),1,fIn);
	    fread(&model[(j*515)+i].iy2,sizeof(int),1,fIn);
	  }
	}
      }
    }
    fclose(fIn);
  }
  else { // Write the file
    status = TRUE;
    fIn= fopen(oldFilename,"r");
    if(fIn) { // Already exists
      sprintf(outStr,"Do you really want to overwrite [%s]?",oldFilename);
      status= askThem(outStr,FALSE,XmDIALOG_QUESTION);
      fclose(fIn);
    }
    if(status) { // OK  proceed
      
      // Give feedback on main window title bar
      sprintf(outStr,"ONED: [%s]",oldFilename);
      simpleString = XmStringCreateLocalized(outStr);
      XtSetArg(arglist[0], XmNdialogTitle, simpleString );
      XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
      XtSetValues(main_window_widget, arglist, 2); 
      XFlush(display);
      XmStringFree(simpleString); 
      XFlush(display);
      
      fIn= fopen(oldFilename,"w");
      if(fIn) {
	fprintf(fIn,"Portable\n");
	for(j=0;j<11;j++) { // DO all models
	  fprintf(fIn,"%d  %f  %f\n",
		  modelState[j], ampScrollValue[j], phaseScrollValue[j]);
	  
	  for(i=0; i<515; i++) {
	    fprintf(fIn,"%f  %f  %f  %d  %d\n",
		    model[(j*515)+i].ix1,
		    model[(j*515)+i].ix2,
		    model[(j*515)+i].ix3,
		    model[(j*515)+i].iy1,
		    model[(j*515)+i].iy2);
	  }
	}
      }
    }
    fclose(fIn);
  }
  
  DrawWindow();
  
} // end function gotFilename



void oned_activate(Widget widget, int *tag, 
		   XmAnyCallbackStruct *callback_data)
{
  // function name: oned_activate
  // They said they wanted to explore the 1D fft of this model.
  
  XtManageChild(onedscreen);
  
} // end function oned_activate



void oned_exit_activate(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{
  // function name: oned_exit_activate
  // Get rid of 1d interface from the display.
  
  XtUnmanageChild(onedscreen);
  
} // end function oned_exit_activate



void createScrolls(Widget widget, int *tag, 
		   XmAnyCallbackStruct *callback_data)
{
  // function name: createScrolls
  // Do what's necessary to create all of the scroll widgets in the
  // model building and 1d fft interfaces.  Remember other widget
  // numbers from UIL defns.

  int i;
  char outStr[64];
  XmString simpleString,ss2;
  XtCallbackRec callBck[5];    // Our callbacks
  XFontStruct *font;           // Holds font information
  XmFontList font_list;
  XmStringCharSet cs= "ISOLatin1";

  // Load all of these guys with proportional "9x15" font
  //  font = XLoadQueryFont(display, "9x15");
  font = XLoadQueryFont(display, "courier_bold14");
  font_list = XmStringCreateFontList(font, cs); 
  
  if(*tag==ONED)  { // Do model building widgets
    onedscreen = widget;
    
    simpleString = XmStringCreateLocalized("0");
    
    for(i=0; i<11; i++) {
      
      XtSetArg(arglist[0], XmNx, 105);
      XtSetArg(arglist[1], XmNy, (10+(i*40)) );
      XtSetArg(arglist[2], XmNorientation, XmHORIZONTAL);
      XtSetArg(arglist[3], XmNwidth, 150); 
      XtSetArg(arglist[4], XmNmaximum, 122); 
      XtSetArg(arglist[5], XmNminimum, -100);
      XtSetArg(arglist[6], XmNborderWidth, 2); 
      XtSetArg(arglist[7], XmNvalue, 0);
      
      ampScroll[i]= (Widget)XmCreateScrollBar(widget, "", arglist, 8);
      
      XtSetArg(arglist[0], XmNx, 280);
      XtSetArg(arglist[1], XmNy, (10+(i*40)) );
      XtSetArg(arglist[2], XmNorientation, XmHORIZONTAL);
      XtSetArg(arglist[3], XmNwidth, 150); 
      XtSetArg(arglist[4], XmNmaximum, 220); 
      XtSetArg(arglist[5], XmNminimum, -180); 
      XtSetArg(arglist[6], XmNborderWidth, 2); 
      XtSetArg(arglist[7], XmNvalue, 0);
      
      phaseScroll[i]= (Widget)XmCreateScrollBar(widget, "", arglist, 8);
      
      XtSetArg(arglist[0], XmNx, 105);
      XtSetArg(arglist[1], XmNy, (30+(i*40)) );
      XtSetArg(arglist[2], XmNlabelString, simpleString);
      XtSetArg(arglist[3], XmNwidth, 93);
      ampLabel[i]= (Widget)XmCreateLabel(widget, "", arglist, 4);
      
      XtSetArg(arglist[0], XmNx, 280);
      XtSetArg(arglist[1], XmNy, (30+(i*40)) );
      XtSetArg(arglist[2], XmNlabelString, simpleString);
      XtSetArg(arglist[3], XmNwidth, 93);
      phaseLabel[i]= (Widget)XmCreateLabel(widget, "", arglist, 4);
      
      sprintf(outStr,"%d",i);
      ss2 = XmStringCreateLocalized(outStr);
      XtSetArg(arglist[0], XmNx, 20);
      XtSetArg(arglist[1], XmNy, (10+(i*40)) );
      XtSetArg(arglist[2], XmNlabelString, ss2);
      XtSetArg(arglist[3], XmNwidth, 33);
      wavnumLabel[i]= (Widget)XmCreateLabel(widget, "", arglist, 4);
      XmStringFree(ss2);
      
      XtSetArg(arglist[0], XmNx, 60);
      XtSetArg(arglist[1], XmNy, (8+(i*40)) );
      XtSetArg(arglist[2], XmNset, 0);
      modelToggle[i]= (Widget)XmCreateToggleButton(widget, "", arglist, 3);
      
      // Now put them on the screen
      XtManageChild(ampScroll[i]);
      if(i>0)
	XtManageChild(phaseScroll[i]);
      
      XtManageChild(ampLabel[i]);
      if(i>0)
	XtManageChild(phaseLabel[i]);
      XtManageChild(wavnumLabel[i]);
      
      XtManageChild(modelToggle[i]);
      
      // Add callbacks to our widgets
      callBck[0].callback= ampScrollChanged;
      callBck[0].closure= 0;
      callBck[1].callback= 0;
      callBck[1].closure= 0;
      XtAddCallbacks(ampScroll[i], XmNvalueChangedCallback, callBck);  
      XtAddCallbacks(ampScroll[i], XmNdragCallback, callBck);  
      callBck[0].callback= phaseScrollChanged;
      XtAddCallbacks(phaseScroll[i], XmNvalueChangedCallback, callBck);  
      XtAddCallbacks(phaseScroll[i], XmNdragCallback, callBck);  
      
      callBck[0].callback= modelChanged;
      callBck[0].closure= 0;
      callBck[1].callback= 0;
      callBck[1].closure= 0;
      XtAddCallbacks(modelToggle[i], XmNvalueChangedCallback, callBck);  
   
    }

    XmStringFree(simpleString);
    
  }
  if(*tag==FFTSCREEN) {
    
    ss2 = XmStringCreateLocalized("64");
    XtSetArg(arglist[0], XmNx, 380);
    XtSetArg(arglist[1], XmNy, 380 );
    XtSetArg(arglist[2], XmNlabelString, ss2);
    XtSetArg(arglist[3], XmNwidth, 63);
    XtSetArg(arglist[4], XmNlabelFontList,  font_list);
    XtSetArg(arglist[5], XmNfontList,  font_list);
    numSampLabel= (Widget)XmCreateLabel(widget, "", arglist, 6);
    XmStringFree(ss2);
    
    ss2 = XmStringCreateLocalized("8.0");
    XtSetArg(arglist[0], XmNx, 380);
    XtSetArg(arglist[1], XmNy, 420 );
    XtSetArg(arglist[2], XmNlabelString, ss2);
    XtSetArg(arglist[3], XmNwidth, 63);
    XtSetArg(arglist[4], XmNlabelFontList,  font_list);
    XtSetArg(arglist[5], XmNfontList,  font_list);
    numMotifLabel= (Widget)XmCreateLabel(widget, "", arglist, 6);
    XmStringFree(ss2);
    
    XtManageChild(numSampLabel);
    XtManageChild(numMotifLabel);
    
    // Now create the filter label
    ss2 = XmStringCreateLocalized("8");
    XtSetArg(arglist[0], XmNx, 380);
    XtSetArg(arglist[1], XmNy, 275 );
    XtSetArg(arglist[2], XmNlabelString, ss2);
    XtSetArg(arglist[3], XmNwidth, 83);
    XtSetArg(arglist[4], XmNlabelFontList,  font_list);
    XtSetArg(arglist[5], XmNfontList,  font_list);
    filterLabel= (Widget)XmCreateLabel(widget, "", arglist, 6);
    XmStringFree(ss2);
    ss2 = XmStringCreateLocalized("8"); // We start nonsensitive so 8 here
    XtSetArg(arglist[0], XmNx, 380);
    XtSetArg(arglist[1], XmNy, 305 );
    XtSetArg(arglist[2], XmNlabelString, ss2);
    XtSetArg(arglist[3], XmNwidth, 83);
    XtSetArg(arglist[4], XmNlabelFontList,  font_list);
    XtSetArg(arglist[5], XmNfontList,  font_list);
    filter2Label= (Widget)XmCreateLabel(widget, "", arglist, 6);
    XmStringFree(ss2);
    ss2 = XmStringCreateLocalized("Spacing of Mask");
    XtSetArg(arglist[0], XmNx, 20);
    XtSetArg(arglist[1], XmNy, 275 );
    XtSetArg(arglist[2], XmNlabelString, ss2);
    f1Label= (Widget)XmCreateLabel(widget, "", arglist, 3);
    XmStringFree(ss2);
    ss2 = XmStringCreateLocalized("Size of Hole in Mask");
    XtSetArg(arglist[0], XmNx, 20);
    XtSetArg(arglist[1], XmNy, 305 );
    XtSetArg(arglist[2], XmNlabelString, ss2);
    f2Label= (Widget)XmCreateLabel(widget, "", arglist, 3);
    XmStringFree(ss2);
    
    XtManageChild(filterLabel);
    XtManageChild(filter2Label);
    XtManageChild(f1Label);
    XtManageChild(f2Label);
    
    // Now create the noise label
    // Now create the sigNoise label
    sprintf(outStr,"4.0");
    sprintf(saveSigStr,"%s",outStr);
    ss2 = XmStringCreateLocalized("Infinite");
    XtSetArg(arglist[0], XmNx, 380);
    XtSetArg(arglist[1], XmNy, 185 );
    XtSetArg(arglist[2], XmNlabelString, ss2);
    XtSetArg(arglist[3], XmNwidth, 83);
    XtSetArg(arglist[4], XmNlabelFontList,  font_list);
    XtSetArg(arglist[5], XmNfontList,  font_list);
    sigNoiseLabel= (Widget)XmCreateLabel(widget, "", arglist, 6);
    XmStringFree(ss2);
    
    ss2 = XmStringCreateLocalized("Signal/Noise");
    XtSetArg(arglist[0], XmNx, 20);
    XtSetArg(arglist[1], XmNy, 190 );
    XtSetArg(arglist[2], XmNlabelString, ss2);
    sigNoise2Label= (Widget)XmCreateLabel(widget, "", arglist, 3);
    XmStringFree(ss2);
    
    XtManageChild(sigNoiseLabel);
    XtManageChild(sigNoise2Label);
    
    // Now create the temp label
    sprintf(outStr,"0.25");
    sprintf(saveTempStr,"%s",outStr);
    ss2 = XmStringCreateLocalized("0.0");
    XtSetArg(arglist[0], XmNx, 380);
    XtSetArg(arglist[1], XmNy, 75 );
    XtSetArg(arglist[2], XmNlabelString, ss2);
    XtSetArg(arglist[3], XmNwidth, 83);
    XtSetArg(arglist[4], XmNlabelFontList,  font_list);
    XtSetArg(arglist[5], XmNfontList,  font_list);
    tempLabel= (Widget)XmCreateLabel(widget, "", arglist, 6);
    XmStringFree(ss2);
    XtManageChild(tempLabel);  
  }
  else if(*tag==NUMSAMP)
    numSampWidget = widget;
  else if(*tag==NUMMOTIF)
    numMotifWidget = widget;
  else if(*tag==TEMP)
    tempScroll = widget;
  else if(*tag==SIGNOISE)
    sigNoiseScroll = widget;
  else if(*tag==FILTER)
    filterScroll = widget;
  else if(*tag==FILTER2)
    filter2Scroll = widget;
  else if(*tag==ALLON)
    allonToggleW = widget;
  else if(*tag==NOISESPEC)
    noiseSpecToggleW = widget;
  
} // end function createScrolls



void ampScrollChanged(Widget widget, int *tag, 
		      XmScrollBarCallbackStruct *callback_data)
{
  // function name: ampScrollChanged
  // When the amp scroll[i] on the screen changes value call here.

  char outStr[64];
  int i;
  XmString simpleString;
  
  for(i=0; i<12; i++) {
    if(widget==ampScroll[i]) {
      ampScrollValue[i] = callback_data->value;
      sprintf(outStr,"%.0f",ampScrollValue[i]);
      simpleString = XmStringCreateLocalized(outStr);
      XtSetArg(arglist[0], XmNlabelString, simpleString);
      XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
      XtSetValues(ampLabel[i], arglist, 2); 
      XmStringFree(simpleString); 
      XFlush(display);
      DrawWindow();
    }
  }
  
} // end function ampScrollChanged



void phaseScrollChanged(Widget widget, int *tag, 
			XmScrollBarCallbackStruct *callback_data)
{
  // function name: phaseScrollChanged
  // When the phase scroll[i] on the screen changes value call here.

  int i;
  char outStr[64];
  XmString simpleString;

  for(i=0; i<12; i++) {
    if(widget==phaseScroll[i]) {
      phaseScrollValue[i] = callback_data->value;
      sprintf(outStr,"%.0f",phaseScrollValue[i]);
      phaseScrollValue[i]*= (THISTWOPI/360.0); // Convert to radians
      simpleString = XmStringCreateLocalized(outStr);
      XtSetArg(arglist[0], XmNlabelString, simpleString);
      XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
      XtSetValues(phaseLabel[i], arglist, 2); 
      XmStringFree(simpleString); 
      XFlush(display);
      DrawWindow();
    }
  }

} // end function phaseScrollChanged



void modelChanged(Widget widget, int *tag, 
		  XmToggleButtonCallbackStruct *callback_data)
{
  // function name: modelChanged
  // Whenever a toggle from the model interface is changed call here.
  // These toggle which wave number is used/not used.

  int i;

  for(i=0; i<12; i++) {
    if(widget==modelToggle[i]) {
      modelState[i] = !modelState[i];
      DrawWindow();
    }
  }
  
} // end function modelChanged



void toggle_activate(Widget widget, int *tag, 
		     XmToggleButtonCallbackStruct *callback_data)
{
  // function name: toggle_activate
  // Whenever the toggle state is changed call here.  

  XmString simpleString;
  char outStr[64];

  if(*tag==MAIN) { // Main toggle
    if(formUp == MODEL) {
      XtUnmanageChild(onedscreen);
      formUp = FFTSCREEN;
      sprintf(outStr,"Model Building");
    }
    else if(formUp == FFTSCREEN) {
      XtManageChild(onedscreen);
      formUp = MODEL;
      sprintf(outStr,"1D-FFT Screen");
    }
    simpleString = XmStringCreateLocalized(outStr);
    XtSetArg(arglist[0], XmNlabelString, simpleString);
    XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
    XtSetArg(arglist[2], XmNset, FALSE );
    XtSetValues(widget, arglist, 3); 
    XmStringFree(simpleString); 
    XFlush(display);
    
  }
  else if(*tag==FLOAT) {
    floatSwitch = !floatSwitch;
  }
  else if(*tag==TEMP) {
    tempSwitch = !tempSwitch;
    if(!tempSwitch) { // No noise
      XtSetArg(arglist[0], XmNsensitive, FALSE);
      XtSetValues(tempScroll, arglist, 1); 
      sprintf(outStr,"0.0");
    }
    else {
      XtSetArg(arglist[0], XmNsensitive, TRUE);
      XtSetValues(tempScroll, arglist, 1); 
      sprintf(outStr,"%s",saveTempStr);
    }
    simpleString = XmStringCreateLocalized(outStr);
    XtSetArg(arglist[0], XmNlabelString, simpleString);
    XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
    XtSetValues(tempLabel, arglist, 2); 
    XmStringFree(simpleString); 
    XSync(display,FALSE);
    XFlush(display);
  }
  else if(*tag==FILTER) { // DO something to filter the data
    filterSwitch = !filterSwitch;
    if(!filterSwitch) {
      // Set the label to indicate full ON
      sprintf(outStr,"%d",numFilter);
      simpleString = XmStringCreateLocalized(outStr);
      XtSetArg(arglist[0], XmNlabelString, simpleString);
      XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
      XtSetValues(filter2Label, arglist, 2); 
      XmStringFree(simpleString); 
      XSync(display,FALSE);
      XFlush(display);
      
      XtSetArg(arglist[0], XmNsensitive, FALSE);
      XtSetValues(filterScroll, arglist, 1); 
      XtSetValues(filter2Scroll, arglist, 1); 
      for(i=0; i<512; i++) {
	filterRemember[i].iy1 = fftCrystal[i].iy1;
	filterRemember[i].iy2 = fftCrystal[i].iy2;
	fftCrystal[i].iy1 = 0;
	fftCrystal[i].iy2 = 0;
      }
    }
    else {
      // Set the label back to what it was
      sprintf(outStr,"%d",numFilter2);
      simpleString = XmStringCreateLocalized(outStr);
      XtSetArg(arglist[0], XmNlabelString, simpleString);
      XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
      XtSetValues(filter2Label, arglist, 2); 
      XmStringFree(simpleString); 
      XSync(display,FALSE);
      XFlush(display);
      
      XtSetArg(arglist[0], XmNsensitive, TRUE);
      XtSetValues(filterScroll, arglist, 1); 
      XtSetValues(filter2Scroll, arglist, 1); 
      for(i=0; i<512; i++) {
	fftCrystal[i].iy1= (int)filterRemember[i].iy2;
	fftCrystal[i].iy2= (int)filterRemember[i].iy2;
      }
      doFilter(FILTER);
    }
  }
  else if(*tag==NOISE) {
    noiseSwitch = !noiseSwitch;
    if(!noiseSwitch) {
      XtSetArg(arglist[0], XmNsensitive, FALSE);
      XtSetValues(sigNoiseScroll, arglist, 1); 
      sprintf(outStr,"Infinite");
    }
    else {
      XtSetArg(arglist[0], XmNsensitive, TRUE);
      XtSetValues(sigNoiseScroll, arglist, 1); 
      sprintf(outStr,"%s",saveSigStr);
    }
    
    simpleString = XmStringCreateLocalized(outStr);
    XtSetArg(arglist[0], XmNlabelString, simpleString);
    XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
    XtSetValues(sigNoiseLabel, arglist, 2); 
    XmStringFree(simpleString); 
    XSync(display,FALSE);
    XFlush(display);
  }
  else if(*tag==NOISESPEC) {
    int randNum;
    
    for(i=0;i<512;i++) {
      randNum= random();
      // noiseSpectrum[i]= (float)(randNum*factor);
      noiseSpectrum[i]= (float)(((float)randNum)/
				((float)DENOMINATOR));
    }
    XtSetArg(arglist[0], XmNset, FALSE);
    XtSetValues(noiseSpecToggleW, arglist, 1); 
    XSync(display,FALSE);
    XFlush(display);
  }
  else if(*tag==ALLON) { // Switch them all ON or go back
    allonSwitch = !allonSwitch;
    if(allonSwitch) {
      for(i=0;i<11;i++) {
	rememberState[i]= modelState[i];
	modelState[i]= TRUE;
      }
      sprintf(outStr,"Go back");
    }
    else {
      for(i=0;i<11;i++)
	modelState[i]= rememberState[i];
      sprintf(outStr,"Try ALL curves");
    }
    simpleString = XmStringCreateLocalized(outStr);
    XtSetArg(arglist[0], XmNlabelString, simpleString);
    XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
    XtSetValues(allonToggleW, arglist, 2); 
    XmStringFree(simpleString); 
    XFlush(display);
    
    for(i=0;i<11;i++) { // Set toggles
      XtSetArg(arglist[0], XmNset, modelState[i]);
      XtSetValues(modelToggle[i], arglist, 1); 
      XSync(display,FALSE);
      XFlush(display);
    }
    
  }
  
  DrawWindow();
  
} // end function toggle_activate



void scrollChange(Widget widget, int *tag, 
		  XmScrollBarCallbackStruct *callback_data)
{
  // function name: scrollChange
  // When the fft scrolls on the screen changes value call here.
  
  char outStr[64];
  XmString simpleString;

  if(*tag==FILTER) { // Tweak the Filter label
    numFilter = callback_data->value;
    numFilter = MIN(numFilter, 512);
    numFilter = MAX(numFilter, 1);

    sprintf(outStr,"%d",numFilter);
    simpleString = XmStringCreateLocalized(outStr);
    XtSetArg(arglist[0], XmNlabelString, simpleString);
    XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
    XtSetValues(filterLabel, arglist, 2); 
    XmStringFree(simpleString); 
    XSync(display,FALSE);
    XFlush(display);
    
    // Now set the max for the filter2 scroll
    XtSetArg(arglist[0], XmNmaximum, (int)(numFilter*1.111));
    XtSetArg(arglist[1], XmNsliderSize, (1));
    XtSetArg(arglist[2], XmNvalue, (int)(1));
    XtSetValues(filter2Scroll, arglist, 3);
    
    XSync(display,FALSE);
    XFlush(display);
  }
  if(*tag==FILTER2) { // Tweak the Filter label
    numFilter2 = callback_data->value;
    numFilter2 = MAX(numFilter2, 1);
    numFilter2 = MIN(numFilter2, numFilter);
    
    sprintf(outStr,"%d",numFilter2);
    simpleString = XmStringCreateLocalized(outStr);
    XtSetArg(arglist[0], XmNlabelString, simpleString);
    XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
    XtSetValues(filter2Label, arglist, 2); 
    XmStringFree(simpleString); 
    XSync(display,FALSE);
    XFlush(display);
  }
  else if(*tag==TEMP) { // Tweak the Temp label
    numTemp = callback_data->value*0.01;
    numTemp = MAX(numTemp, 0.0);
    numTemp = MIN(numTemp, 7.0);

    sprintf(outStr,"%.2f",(pow(2.0,numTemp)/4.0));
    sprintf(saveTempStr,"%s",outStr);
    
    numTemp = 2.0/(pow(2.0,numTemp)/4.0);
    
    simpleString = XmStringCreateLocalized(outStr);
    XtSetArg(arglist[0], XmNlabelString, simpleString);
    XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
    XtSetValues(tempLabel, arglist, 2); 
    XmStringFree(simpleString); 
    XSync(display,FALSE);
    XFlush(display);
  }
  else if(*tag==SIGNOISE) { // Tweak the SigNoise label
    numSigNoise = callback_data->value*0.01;
    numSigNoise = MAX(numSigNoise, 0.0);
    numSigNoise = MIN(numSigNoise, 11.0);
    
    numSigNoise = (pow(2.0,numSigNoise)/10.0);
    sprintf(outStr,"%.2f",numSigNoise);
    sprintf(saveSigStr,"%s",outStr);
    simpleString = XmStringCreateLocalized(outStr);
    XtSetArg(arglist[0], XmNlabelString, simpleString);
    XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
    XtSetValues(sigNoiseLabel, arglist, 2); 
    XmStringFree(simpleString); 
    XSync(display,FALSE);
    XFlush(display);
    }
  else { // Do others
    if(*tag==NUMSAMP) { // Number of samples/motif
      numSamps = callback_data->value;
      numSamps = MAX(numSamps, 1);
      numSamps = MIN(numSamps, 256);
      
      if(numMotifs > (512.0/(float)numSamps)) { // Clip it to size
	numMotifs = (float)(512.0 / (float)numSamps);
	maxNumMotifs = (float)(512.0 / (float)numSamps);
      }
      else { // Keep it as is
	maxNumMotifs = (float)(512.0 / (float)numSamps);
      }
      XtSetArg(arglist[0], XmNmaximum, ((int)(maxNumMotifs*10.00)+10));
      XtSetArg(arglist[1], XmNsliderSize, (10));
      XtSetArg(arglist[2], XmNvalue, (int)(numMotifs*10));
      XtSetValues(numMotifWidget, arglist, 3);
      
      // Now set the filter Mask appropriately
      numFilter = (int)numMotifs;
      XtSetArg(arglist[0], XmNvalue, (int)(numFilter));
      XtSetValues(filterScroll, arglist, 1);
      sprintf(outStr,"%d",numFilter);
      simpleString = XmStringCreateLocalized(outStr);
      XtSetArg(arglist[0], XmNlabelString, simpleString);
      XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
      XtSetValues(filterLabel, arglist, 2); 
      XmStringFree(simpleString); 
      
      // Now instantiate the new filter
      doFilter(FILTER);
      
      XSync(display,FALSE);
      XFlush(display);
    }
    else if(*tag==NUMMOTIF) { // Number of motifs/crystal 
      numMotifs = callback_data->value*0.01;
      numMotifs = MAX(numMotifs, 0.1);
      numMotifs = MIN(numMotifs, maxNumMotifs);
      
      XSync(display,FALSE);
      XFlush(display);
    }
    
    sprintf(outStr,"%d",numSamps);
    simpleString = XmStringCreateLocalized(outStr);
    XtSetArg(arglist[0], XmNlabelString, simpleString);
    XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
    XtSetValues(numSampLabel, arglist, 2); 
    XmStringFree(simpleString); 
    XSync(display,FALSE);
    XFlush(display);
    
    sprintf(outStr,"%.2f",numMotifs);
    simpleString = XmStringCreateLocalized(outStr);
    XtSetArg(arglist[0], XmNlabelString, simpleString);
    XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
    XtSetValues(numMotifLabel, arglist, 2); 
    
    XmStringFree(simpleString); 
  }
  
  XSync(display,FALSE);
  XFlush(display);
  
  doFilter(*tag);
  DrawWindow(); 
  
} // end function scrollChange



void pageChange(Widget widget, int *tag, 
		XmScrollBarCallbackStruct *callback_data)
{
  // function name: pageChange
  // When the fft scrolls on the screen changes value BY A PAGE EVENT
  // call here.

  int thisVal;
  float floatVal;

  if(*tag==TEMP) { // Tweak the Temp label
    floatVal = (float)callback_data->value;
    // Round to integer
    thisVal= (int)((int)floatVal*0.01);
    //      thisVal= MAX(thisVal,0);
    floatVal= (float)((float)thisVal*100.0);
    callback_data->value= (float)floatVal;
    
    XtSetArg(arglist[0], XmNvalue, (int)(thisVal*100));
    XtSetValues(tempScroll, arglist, 1);
    }
  else if(*tag==SIGNOISE) { // Tweak the SigNoise label
    floatVal = (float)callback_data->value;
    // Round to integer
    thisVal= (int)(floatVal*0.01);
    //   thisVal= MAX(thisVal,1);
    floatVal= (float)((float)thisVal*100.0);
    callback_data->value= (float)floatVal;
    
    XtSetArg(arglist[0], XmNvalue, (int)(thisVal*100));
    XtSetValues(sigNoiseScroll, arglist, 1);
  }
  else if(*tag==NUMMOTIF) { // Number of motifs/crystal
    floatVal = (float)callback_data->value;
    // Round to integer
    thisVal= (int)(floatVal*0.1);
    thisVal= MAX(thisVal,1);
    floatVal= (float)((float)thisVal*10.0);
    callback_data->value= (float)floatVal;
    
    XtSetArg(arglist[0], XmNvalue, (int)(thisVal*10));
    XtSetValues(numMotifWidget, arglist, 1);
  }
  
  XSync(display,FALSE);
  XFlush(display);
  
  scrollChange(widget,tag,callback_data);
  
} // end function pageChange



void doFilter(int tag)
{
  // function name: doFilter
  // Do what's necessary to setup the filter mask.

  int thisIndx, halfNumTodo, i, j;

  if(filterSwitch) { // Do the filtering
    if(tag==FILTER || tag==FILTER2) { // Tweak the Filter
      // Init them all to OFF
      for(i=0; i<512; i++)
	fftCrystal[i].iy1= 99;
      
      halfNumTodo = (int)numFilter2 / (int)2;
      
      for(i=256;i<512;i+=numFilter) {
	// DO left half
	thisIndx = 256 - (i-256);
	thisIndx -= halfNumTodo;
	if(thisIndx < 0)
	  thisIndx= 0;
	for(j=(thisIndx);j<(thisIndx+numFilter2);j++) { // Do all in mask
	  fftCrystal[j].iy1= 0;
	}
	// Do right half
	thisIndx = i ;
	thisIndx -= halfNumTodo;
	for(j=(thisIndx);j<(thisIndx+numFilter2);j++) { // Do all in mask
	  fftCrystal[j].iy1= 0;
	}
      }

      // DO LAST left half
      thisIndx = 0;
      thisIndx -= halfNumTodo;
      for(j=(thisIndx);j<(thisIndx+numFilter2);j++) { // Do all in mask
	if(j>=0)
	  fftCrystal[j].iy1= 0;
      }
      
      // Do LAST right half
      thisIndx = 512 ;
      thisIndx -= halfNumTodo;
      for(j=(thisIndx);j<512;j++) { // Do all in mask
	fftCrystal[j].iy1= 0;
      }
    }
  }

} // end function doFilter



void oned_terp_c()
{
  // function name: oned_terp_c
  // Interpolate object (from ndata1 to xnp)
  // Ported from Fortran routine ONED_TERP

  int i, n, ix, indx, ndata1=256, thisPix;
  int tmmTnSamp, nmTnSamp, thisManyMotifs;
  float dtot=0.0, x, dx, scaleFactor;
  
  davg   = 0.0;
  dfloat = 0.0;

  dmin   =  9999999.9;
  dmax   = -9999999.9;
  xnp    = numSamps;
  ndata  = (int)(numSamps);
  xscl   = (float)(ndata1)/(float)xnp;

  // First determine what dfloat is gonna be
  if(floatSwitch)
    dfloat = (totModel[0].ix1+totModel[(ndata-1)].ix1)/2.0;
  else 
    dfloat = 0.0;

  // Now calc the sampled Crystal
  thisManyMotifs= 0;
  nmTnSamp= numMotifs * numSamps;
  for(i=0; i<=512; i+=numSamps) { // Do the entire line
    thisPix= 0;
    tmmTnSamp= (thisManyMotifs*numSamps);
    if(((i+numSamps)<=512) && (thisManyMotifs <= numMotifs)){
      // OK we can fit a whole 'nother motif in here
      for(n=0; n<numSamps; n++) {
	if(((tmmTnSamp)+n) < (nmTnSamp)) { // Put more ON
	  x         = ((float)n)*xscl;
	  ix        = (int)x;
	  indx      = 1 + (ix % ndata1);
	  dx        = x - (float)ix;
	  sampCrystal[i+thisPix].ix1 = totModel[indx].ix1+
	    dx*(totModel[indx+(int)(xscl)].ix1-
		totModel[indx].ix1);
	  dtot      = dtot + (float)sampCrystal[i+thisPix].ix1;
	  if(dmax < sampCrystal[i+thisPix].ix1)
	    dmax = sampCrystal[i+thisPix].ix1;
	  if(dmin > sampCrystal[i+thisPix].ix1)
	    dmin = sampCrystal[i+thisPix].ix1;
	}
	thisPix++;
      }
    }
    thisManyMotifs++;
  }
  
  // Calc davg
  davg = (dmax-dmin)/ndata;
  
  // Calc noise if needed
  for(i=0;i<513;i++)
    calcNoiseFactor(&noiseFactor[i], i);
  
  // What do we do if they only want less than one motif???
  if(numMotifs >= 1.0) { // Calc new dfloat
    if(floatSwitch)
      dfloat = (totModel[0].ix1+totModel[(255)].ix1)/2.0;
    else 
      dfloat = (totModel[0].ix1+totModel[(nmTnSamp)].ix1)/2.0;
  }
  
  // Remember the UN-noised perfect crystal - SCALE it
  dmax = MAX(fabs(dmax),fabs(dmin));
  if(dmax > 0.0)
    scaleFactor = 0.5/dmax;  // Damp it to compare with back FFT drawing
  else    
    scaleFactor= 1.0;
  for(i=0; i<513; i++)
    perfCrystal[i].ix1= sampCrystal[i].ix1 * scaleFactor;
  
  // Now add noise and scale it
  thisManyMotifs= 0;
  nmTnSamp= numMotifs * numSamps;
  dmin   =  9999999.9;
  dmax   = -9999999.9;
  for(i=0; i<=512; i+=numSamps) { // Do the entire line
    thisPix = 0;
    tmmTnSamp = thisManyMotifs*numSamps;
    if(((i+numSamps)<=512) && (thisManyMotifs <= numMotifs)) {
      //OK we can fit a whole 'nother motif in here
      for(n=0; n<numSamps; n++) {
	if(((tmmTnSamp)+n) < (nmTnSamp)) { // Put more ON
	  // Add noise
	  sampCrystal[i+thisPix].ix1 *= 0.5;
	  sampCrystal[i+thisPix].ix1 += (float)noiseFactor[i+thisPix];
	  if(dmax < sampCrystal[i+thisPix].ix1)
	    dmax = sampCrystal[i+thisPix].ix1;
	  if(dmin > sampCrystal[i+thisPix].ix1)
	    dmin = sampCrystal[i+thisPix].ix1;
	}
	thisPix++;
      }
    }
    thisManyMotifs++;
  }
  
  // Also make dfloat half as much
  dfloat *= 0.5;
  
  // Now reintroduce the dfloats without scaling
  thisManyMotifs= 0;
  nmTnSamp= numMotifs * numSamps;
  for(i=0; i<=512; i+=numSamps) { // Do the entire line
    thisPix= 0;
    tmmTnSamp= (thisManyMotifs*numSamps);
    if(((i+numSamps)<=512) && (thisManyMotifs <= numMotifs)) {
      // OK we can fit a whole 'nother motif in here
      for(n=0; n<numSamps; n++) {
	if(((tmmTnSamp)+n) < (nmTnSamp)) { // Put more ON
	  // Done above
	}
	else {
	  sampCrystal[i+thisPix].ix1 = dfloat; 
	  perfCrystal[i+thisPix].ix1 = dfloat; 
	  // dfloat==0.0 for !floatSwitch
	}
	thisPix++;
      }
    }
    else { // Just zero it out
      for(n=0; n<numSamps; n++) {
	sampCrystal[i+thisPix].ix1 = dfloat; 
	perfCrystal[i+thisPix].ix1 = dfloat; 
	thisPix++;
      }
    }
    thisManyMotifs++;
  }
  
  // DO the 512th one
  if(((int)thisManyMotifs-1) == (float)numMotifs) { // Do from model
    sampCrystal[512].ix1 = totModel[0].ix1*scaleFactor;
    perfCrystal[512].ix1 = totModel[0].ix1*scaleFactor;
  }
  else { // Init it
    sampCrystal[512].ix1 = dfloat;  // dfloat==0.0 for !floatSwitch
    perfCrystal[512].ix1 = dfloat;  // dfloat==0.0 for !floatSwitch
  }

  davg   = (float)(dmax-dmin)/(float)ndata;

} // end function oned_terp_c



void calcNoiseFactor(float *noiseFctr, int i)
{
  // function name: calcNoiseFactor
  // Calculate the noise factor needed.

  float isign;

  if(!noiseSwitch)
    *noiseFctr = 0.0;
  else { // Add noise
    isign = (float)(((int)(noiseSpectrum[i]*1000))%2);
    if(isign == 0.0)
      isign = -1.0;
    else 
      isign = 1.0;

    // Tim does this this way:
    // 
    // *noiseFctr = noiseSpectrum[i] * 
    //          ((float)davg / (float)numSigNoise)*isign;
    // 
    // which I'm supposing will do the scaling at the same time
    // as applying the noise.  What I'm seeing is that we're about
    // a factor of 4-16 off in this...  ???

    *noiseFctr = (float)((float)davg / (float)numSigNoise);
    *noiseFctr = *noiseFctr * (float)isign;
    *noiseFctr = *noiseFctr * (float)noiseSpectrum[i];
    // OLDWAY:
    // *noiseFctr = (float)((float)noiseSpectrum[i]
    //((float)davg / (float)numSigNoise)*(float)isign);

    *noiseFctr = *noiseFctr * 40.0;
    
  }
  
} // end function calcNoiseFactor



void oned_fft_c()
{
  // function name: oned_fft_c
  // 1D forward or backward transform - ported from Fortran source

  int firstArg, secondArg, thirdArg, fourthArg, fifthArg;
  int i, indx, indxT, tempOK;
  float data[512][2];
  float tmax, tmin, B, ttot, scaleFactor;

  // Do the forward transform
  for(i=0;i<512;i++) { // Stuff the FFT buffer
    data[i][0]= ((float)sampCrystal[i].ix1)-
      (float)dfloat; // dfloat ==0.0 for floatSwitch
    data[i][1]= 0.0;
  }
  
  firstArg=512; 
  secondArg=1; 
  thirdArg=1; 
  fourthArg=0; 
  fifthArg=0;

  fourt_ (data, &firstArg, &secondArg, &thirdArg, &fourthArg, &fifthArg);
  tmax = -99999.9;
  tmin = 999999.9;
  for(i=0; i<512; i++) {
    if(i<256)
      indx = 256+i;  // Do right half
    else 
      indx = i-256;  // Do left half
    
    data[indx][0] = data[indx][0]/256.0;
    data[indx][1] = data[indx][1]/256.0; 
    fftCrystal[i].ix2 = data[indx][0];
    fftCrystal[i].ix3 = data[indx][1];
    fftCrystal[i].ix1 = sqrt(data[indx][0]*data[indx][0] +
			     data[indx][1]*data[indx][1]);
    if(tmax < fftCrystal[i].ix1)
      tmax      = fftCrystal[i].ix1;
    if(tmin > fftCrystal[i].ix1)
      tmin      = fftCrystal[i].ix1;
  }

  // Now apply the temperature factor if desired
  if(tempSwitch)
    B = 4.0 / (float)(numTemp * numTemp);

  // Reverse transform
  tmax = -99999.9;
  tmin= 99999.9;
  ttot= 0.0;
  for(i=0; i<512; i++) {
    // Calc TEMP factor but remember to do it from the middle out
    if(tempSwitch) { // apply temp
      if(i<256)
	indxT= 256-i;    // Do right half
      else 
	indxT= i-256;    // Do left half
      if(indxT > 0)
	tempFactors[i] = exp(-B / (4.0 * (256.0 / indxT)*(256.0 / indxT)));
      else tempFactors[i] = 1.0;
    }
    else tempFactors[i] = 1.0;
    if(indxT!=0) {
      tmax = MAX(tempFactors[i],tmax);
      tmin = MIN(tempFactors[i],tmin);
      ttot += (tempFactors[i]*fftCrystal[i].ix1);
    }
  }
  
  tempOK = TRUE;
  
  // Now apply temp factors if appropriate
  if(tempOK) {
    tmax = -99999.9;
    tmin = 99999.9;
    for(i=0; i<512; i++) {
      if(i<256)
	indx= 256+i;    // Do right half
      else 
	indx= i-256;    // Do left half
      
      if(fftCrystal[i].iy1==0) { // Use it
	fftCrystal[i].ix1 *= tempFactors[i];
	fftCrystal[i].ix2 *= tempFactors[i];
	fftCrystal[i].ix3 *= tempFactors[i];
	data[indx][0]= fftCrystal[i].ix2;
	data[indx][1]= fftCrystal[i].ix3;
      }
      else { // Don't use it
	data[indx][0]= 0.0;
	data[indx][1]= 0.0;
      }
      
      if(tmax < fftCrystal[i].ix1)
	tmax = fftCrystal[i].ix1;
      if(tmin > fftCrystal[i].ix1)
	tmin = fftCrystal[i].ix1;
      
    }
    
    // Now scale FFT to 1.0
    if (tmax > tmin)
      scaleFactor = 1.0 / (tmax-tmin);
    else 
      scaleFactor = 1.0;
      
    for(i=0; i<512; i++) { // Scale them
      fftCrystal[i].ix1 = (fftCrystal[i].ix1 - tmin) * scaleFactor;
    }
  }
  
  // Now apply the filters
  for(i=0; i<512; i++) { // Scale them
    if(i<256)
      indx= 256+i;   // Do right half
    else 
      indx= i-256;   // Do left half
    
    if(fftCrystal[i].iy1==99 || fftCrystal[i].iy2==77) {
      // Don't use for back transform
      fftCrystal[i].ix1 = 0.0;
      data[indx][0]= 0.0;
      data[indx][1]= 0.0;
    }
  }

  firstArg= 512;
  secondArg= 1;
  thirdArg= -1;
  fourthArg= 1;
  fifthArg= 0;
  
  fourt_(data, &firstArg, &secondArg, &thirdArg, &fourthArg, &fifthArg);
  
  for(i=0; i<512; i++)
    backCrystal[i].ix1 = data[i][0]*0.5; // Now scale to 0.5

  dfloatBack = 0.0;
  
} // end function oned_fft_c



void reset_activate(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{
  // function name: reset_activate
  // They want to reset the scrolls for this curve.
  
  int i;
  
  for(i=0; i<11; i++) { // Search for the one
    if(widget==resetButton[i]) { // This is it
      phaseScrollValue[i] = 0.0;
      ampScrollValue[i] = 0.0;
      
      XtSetArg(arglist[0], XmNlabelString, XmStringCreateLocalized("0"));
      XtSetArg(arglist[1], XmNrecomputeSize, FALSE );
      XtSetValues(phaseLabel[i], arglist, 2); 
      XtSetValues(ampLabel[i], arglist, 2); 
      
      XtSetArg(arglist[0], XmNvalue, 0);
      XtSetValues(ampScroll[i], arglist, 1);
      XtSetArg(arglist[0], XmNvalue, 0);
      XtSetValues(phaseScroll[i], arglist, 1);
      XFlush(display);
      
      DrawWindow();
    }
  }
  
} // end function reset_activate
