#include <math.h>
#include <stdio.h>
#include <stdlib.h>

void ctf_nodes_misha(float wavelength, float focus,
		     float cs_coef, float amp_fac, int node,
		     float *d_star)
{
  // function name: ctf_nodes_misha
  // Misha Sherman's find CTF nodes function
  // Translated from Fortran to C by RSS

  float part1, part2, part3, focusH;
  
  focusH = focus*1000.0;
  part1 = node - asin(amp_fac)/M_PI ;
  part2 = cs_coef*1000000.0*wavelength;
  part3 = focusH*focusH - part2*part1;
  if(part3 < 0.0) {
    printf("Negative determinant for focus @node %f %d\n", focus, node);
    *d_star = 0.0;
  } else {
    *d_star = sqrt(2.0*(focusH - sqrt(part3))/(part2*wavelength));
  }
  
} // end function ctf_nodes_misha
  


void ctf_focus_misha (float wavelength, float *focus,
		      float cs_coef, float amp_fac, int node,
		      float d_star)
{
  // function name: ctf_focus_misha
  // Misha Sherman's find focus function
  // Translated from Fortran to C by RSS

  float part1, part2;
  
  part1 = (node - asin(amp_fac)/M_PI) / wavelength;
  part2 = 0.25 * cs_coef * 1000000.0 * wavelength * wavelength
    * (d_star*d_star*d_star*d_star);
  *focus = (part1 + part2) / (d_star * d_star);
  *focus = *focus * 0.001;

} // end function ctf_focus_misha



void convert_abpart(float apart, float bpart, 
		    float *amp, float *phi)
{
  // function name: convert_abpart
  // Convert complex number to amplitude and angle

  float pi, radian;
  pi = 4.0*atan2(1.0,1.0);
  radian = pi/180.0;

  *amp = sqrt(apart*apart + bpart*bpart);

  if(apart == 0.0) {
    if(bpart > 0.0)
      *phi = 90.0;
    else
      *phi = 270.0;
  } else {
    *phi = atan2(bpart,apart)/radian;
    *phi = fmod(*phi,360.0);
    if(*phi < 0.0)
      *phi += 360.0;
  }
} // end function convert_abpart



void fft_point (float *fft, int idim1, int idim2, int ix, int iy,
		float *radius, float *apart, float *bpart)
{
  // function name: fft_point
  // Return the real and imaginary components of the FFT along with
  // the distance of the point from center of transform
  
  int icount, offX, offY;

  *radius = sqrt((double)(ix*ix + iy*iy));
  
  if(ix < 0)
    offX = ix+idim1;
  else
    offX = ix;

  if(iy < 0) 
    offY = iy+idim2;
  else
    offY = iy;

  icount = offY*idim1*2 + (offX+1)*2 - 2;
  *apart  = fft[icount];
  *bpart  = fft[icount+1];

} // end function fft_point



void fft_setdim_local (int ncol, int nrow, int *idim1, int *idim2)
{
  // function name: fft_setdim_local
  // Set up Fourier transform dimensions to be a power of two
  // between 128 and 1024

  int itmp1, itmp2, id1, id2;

  // Determine minimum allowed fft size
  id1 = (int)((ncol-0.5)/128.0) + 1;
  id2 = (int)((nrow-0.5)/128.0) + 1;

  if (id1 == 3)
    id1 = 4;
  if (id2 == 3)
    id2 = 4;
  if(id1 >= 5 && id1 <= 7)
    id1 = 8;
  if(id2 >= 5 && id2 <= 7)
    id2 = 8;

  *idim1 = id1*128;
  *idim2 = id2*128;

  if(*idim1 == 1024 && *idim2 == 1024) {
    printf("******************************************\n");
    printf("* FFT size automatically set = 1024x1024 *\n");
    printf("******************************************\n");
    return;
  }

  itmp1 = *idim1;
  itmp2 = *idim2;
  if(itmp1 < *idim1 || itmp2 < *idim2) {
    printf("******************************************\n");
    printf("* FFT_SETDIM size too small*\n");
    printf("******************************************\n");
    exit(0);
  }

  if(itmp1%128 != 0 || itmp2%128 != 0) {
    printf("******************************************\n");
    printf("* FFT_SETDIM: Size not divisible by 128  *\n");
    printf("******************************************\n");
    exit(0);
  }

  if(itmp1 == 384 || itmp1 == 640 || itmp1 == 896 ||
     itmp2 == 384 || itmp2 == 640 || itmp2 == 896) {
    printf("***********************************************\n");
    printf("* FFT_SETDIM: Size not equal to allowed values*\n");
    printf("***********************************************\n");
    exit(0);
  }

  *idim1 = itmp1;
  *idim2 = itmp2;

} // end function fft_setdim_local
