/* emUnix.c
 * 
 * Holding area for code migrating from FORTRAN/VMS to UNIX/C.
 * 
 * Rob Ashmore
 * Purdue Univ.
 * rwa@bragg.bio.purdue.edu
 */

#include <math.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <Xm/FileSB.h>

#include "../include/common.h"
#include "lists.h"
#include "robem_def.h"
#include "robem.h"

#define COREMAPREF(x,y,z) coremap[((((int)z)*ncol*nrow)+(((int)y)*ncol)+((int)x))]

#define THRESHHOLD 1.0
#define SLIDE 0.4

static float threshold = THRESHHOLD;
static float slide = SLIDE;

int callXdy(double x, double y, double z, double minRad, double maxRad)
{
  return(0);
}



double disDotPlane(double x,  double y,  double z,
		   double x1, double y1, double z1,
		   double x2, double y2, double z2,
		   double x3, double y3, double z3)
{
  // function name: disDotPlane
  // Used for icosahedron drawing 
  
  double A, B, C, D, sum, d;
		
  A = (y2-y1)*(z3-z1)+(y3-y1)*(z1-z2);	
  B = (x3-x1)*(z2-z1)+(x2-x1)*(z1-z3);	
  C = (x2-x1)*(y3-y1)+(x3-x1)*(y1-y2);	
  D =  x1*(y1-y2)*(z3-z1)
    + z1*(x1-x2)*(y3-y1)
    + y1*(x1-x3)*(z2-z1)
    + x1*(y3-y1)*(z2-z1)
    + y1*(x2-x1)*(z3-z1)
    + z1*(x3-x1)*(y2-y1);
  sum = fabs(A*x+B*y+C*z+D);
  d = sum/(sqrt(A*A + B*B + C*C));	
  return d;

} // end function disDotPlane



int isOnPlane(double x,  double y,  double z,
	      double x1, double y1, double z1,
	      double x2, double y2, double z2,
	      double x3, double y3, double z3)
{
  // function name: isOnPlane
  // Used for icosahedron drawing 

  double dst = disDotPlane(x,y,z,x1,y1,z1,x2,y2,z2,x3,y3,z3);
  if (dst < threshold)
    return 1;
  else
    return 0;
  
} // end function isOnPlane



void crossPoint(double xa,  double ya,  double za,
		double xb,  double yb,  double zb,
		double x1,  double y1,  double z1,
		double x2,  double y2,  double z2,
		double x3,  double y3,  double z3, 
		double *Xx, double *Xy, double *Xz)
{
  // function name: crossPoint
  // Used for icosahedron drawing 
  // to obtain the cross point of a line and a plane
  // plane function: AX+BY+CZ+D = 0
  // line function: x/x2=y/y2=z/z2 = t; because x1=y1=z1 =0;

  double A, B, C, D, t;
  
  A = (y2-y1)*(z3-z1) - (y3-y1)*(z2-z1);
  B = (x3-x1)*(z2-z1) - (x2-x1)*(z3-z1);
  C = (x2-x1)*(y3-y1) - (x3-x1)*(y2-y1);
  D = -x1*(y2-y1)*(z3-z1) - z1*(x2-x1)*(y3-y1) - 
    y1*(x3-x1)*(z2-z1) + x1*(y3-y1)*(z2-z1) + 
    y1*(x2-x1)*(z3-z1) + z1*(x3-x1)*(y2-y1);
  
  t = (-D - A*xa- B*ya- C*za)/(A*(xb-xa) + B*(yb-ya) + C*(zb-za));
  *Xx = (xb-xa)*t + xa;
  *Xy = (yb-ya)*t + ya;
  *Xz = (zb-za)*t + za;
  
} // end function crossPoint	



int isOutPlane(double x,  double y,  double z,
	       double x1, double y1, double z1,
	       double x2, double y2, double z2, int PointHigherOut)
{
  // function name: isOutPlane
  // This function will check if this point is outside the plane will
  // check 3 planes which form the central cone in 1/8 space consider
  // the line parallel to z axis, will check the relationship between
  // the original point and the cross point the plane is decided by 3
  // points x1y1z1, x2y2z2, 000 if PointHigherOut is 1, this means if
  // the original point is higher than cross point then original point
  // is considered as outside the plane

  double Xx, Xy, Xz;
  crossPoint(x,y,0,x,y,z,0,0,0,x1,y1,z1,x2,y2,z2, &Xx, &Xy, &Xz);
  if(PointHigherOut && z > Xz)
    return 1;
  else if(!PointHigherOut && z < Xz)
    return 1;
  else
    return 0;
    
} // end function isOutPlane



int isOnBulge(double sphererad, double x, double y, double z,
	      double x1, double y1, double z1,
	      double x2, double y2, double z2,
	      double x3, double y3, double z3)
{
  // function name: isOnBulge
  // Used for icosahedron drawing 	
  double a, b, c;
  double dstcenter, dstsphere;
  double dstXcenter, dstIcosPoint, dstXsphere;
  double diff;
  
  crossPoint(x,y,z,0,0,0, x1,y1,z1,x2,y2,z2,x3,y3,z3, &a, &b, &c);
  dstcenter = sqrt(x*x + y*y + z*z);
  //distance between the cross point (of icosahedral and the line)
  //and the center
  dstXcenter = sqrt(a*a + b*b + c*c);	
  //distance between sphere to the point
  dstsphere = sphererad - dstcenter;	
  //distance between X point and the sphere;
  dstXsphere = sphererad - dstXcenter;	
  
  if(dstXcenter > dstcenter) 
    return 0;
  //if(dstXsphere < threshold)
  //	return 1;
  //the distance between the point and the cross point
  dstIcosPoint = dstcenter - dstXcenter; 	
  
  diff = dstXsphere * slide - dstIcosPoint;
  if( diff < threshold && diff > -threshold)
    return 1;
  return 0;

} // end function isOnBulge



int cutIcosFace(double x, double y, double z, double rad) 
{
  // function name: cutIcosFace
  // Cut out an icosahedral face
  // x,y,z is the cartesian position of the considered voxel
  // rad is the radius from the origin to the five fold vertex

  double x2, y2, z2, x4, y4, z4, x6, y6, z6;
  double rad85, rad52;
  double distance2;
  
  if (x < 0 || y < 0 || z < 0)	
    return 0;
  
  distance2 = x*x + y*y + z*z;
  if(distance2 > rad*rad)
    return 0;	
  
  //these are the cartessian values of six points enclosing 
  //the icosahedral shell
  rad85 = 0.85065*rad;
  rad52 = 0.52573*rad;
  
  x2 = 0.0;   y2 = rad85; z2 = rad52;  //5 fold point on yz plane
  x4 = rad85; y4 = rad52; z4 = 0.0;    //5-fold point which is on xy plane
  x6 = rad52; y6 = 0.0;   z6 = rad85;  //5-fold point on xz plane
  if(isOutPlane(x,y,z,x2,y2,z2,x6,y6,z6,0) && 
     isOutPlane(x,y,z,x2,y2,z2,x4,y4,z4,1) && 
     isOutPlane(x,y,z,x6,y6,z6,x4,y4,z4,1)) {
      return 1;
  }
  
  return 0;
  
} // end function  cutIcosFace



int cutAsymUnit(double x, double y, double z, double rad)	
{
  // function name: cutAsymUnit
  // cut out an assymetric unit

  double x4, y4, z4, x5, y5, z5, x6, y6, z6;
  double x46m, y46m, z46m;	

  // The middle point of the point 6 and 2-fold correspoinding point of
  // 4 around 5

  double rad85, rad52, distance2;

  if (x < 0 || y < 0 || z < 0)	
    return 0;

  distance2 = x*x + y*y + z*z;
  if(distance2 > rad*rad)
    return 0;  //the voxel is too close to center, 
  //speed up the process
  
  rad85 = 0.85065*rad;
  rad52 = 0.52573*rad;
  x4 = rad85; y4 = rad52; z4 = 0.0;    //5-fold point which is on xy plane
  x5 = rad85; y5 = 0.0;   z5 = 0.0;    //midege point on x axis
  x6 = rad52; y6 = 0.0;   z6 = rad85;  //5-fold point on xz plane
  x46m = (x6+x4)/2.0; 
  y46m = (y6-y4)/2.0; 
  z46m=(z6+z4)/2.0;
  
  if(isOutPlane(x,y,z,x4,y4,z4,x6,y6,z6,0) && 
     isOutPlane(x,y,z,x4,y4,z4,x46m,y46m,z46m,1)) {
      return 1;
  }
  return 0;
  
} // end function cutAsymUnit



int bulge(double x, double y, double z, double rad)
{  
  // function name: bulge
  // Used for icosahedron drawing 
  // x,y,z is the cartessian position of the considered voxel
  // rad is the radius from the origin to the five fold vertex

  double x1, y1, z1, x2, y2, z2, x3, y3, z3;
  double x4, y4, z4, x5, y5, z5, x6, y6, z6;
  double rad85, rad52, distance;

  distance = sqrt(x*x + y*y + z*z);
  
  // first we need to convert this voxel to the first 1/8 space	  
  x = fabs(x);
  y = fabs(y);
  z = fabs(z);

  // we need to get rid of the points whose position is out of range
  if( distance > (rad+threshold) || distance < (0.79465 * rad - threshold))
    return 0;	//the voxel is too close to center, 
  //speed up the process
  
  // these are the cartessian values of six points enclosing icosahedral shell
  rad85 = 0.85065*rad;
  rad52 = 0.52573*rad;
  x1 = 0.0;   y1 = 0.0;   z1 = rad85;   // mid-edge point on Z axis		
  x2 = 0.0;   y2 = rad85; z2 = rad52;   // 5 fold point on yz plane	
  x3 = 0.0;   y3 = rad85; z3 = 0.0;     // mid-edge point on y axis	
  x4 = rad85; y4 = rad52; z4 = 0.0;     // 5-fold point which is on xy plane
  x5 = rad85; y5 = 0.0;   z5 = 0.0;     // mid-edge point on x axis	
  x6 = rad52; y6 = 0.0;   z6 = rad85;   // 5-fold point on xz plane
  
  if(isOutPlane(x,y,z,x2,y2,z2,x6,y6,z6, 1)) {
    if(isOnBulge(rad,x,y,z,x1,y1,z1,x2,y2,z2,x6,y6,z6))
      return 1;
    else
      return 0;
  }
  else if(isOutPlane(x,y,z,x4,y4,z4,x6,y6,z6, 0)) {
    if(isOnBulge(rad,x,y,z,x4,y4,z4,x5,y5,z5,x6,y6,z6))
      return 1;
    else
      return 0;
  }
  else if (isOutPlane(x,y,z,x4,y4,z4,x2,y2,z2, 0)) {
    if(isOnBulge(rad,x,y,z,x2,y2,z2,x4,y4,z4,x3,y3,z3))
      return 1;
    else
      return 0;
  }	
  else if(isOnBulge(rad,x,y,z,x2,y2,z2,x6,y6,z6,x4,y4,z4))
    return 1;
  else
    return 0;	

} // end function bulge



int doIcosshell(double x, double y, double z, double rad)
{
  // function name: doIcosshell
  // Used for icosahedron drawing 

  double ka = -1.61804, kb = 1.61804, kc = -0.38211, kd = 0.61803;
  double x1, y1, z1, x2, y2, z2, x3, y3, z3;
  double x4, y4, z4, x5, y5, z5, x6, y6, z6;

  // x,y,z is the cartessian position of the considered voxel
  // rad is the radius from the origin to the five fold vertex
  // first we need to convert this voxel to the first 1/8 space  
  
  x = fabs(x);
  y = fabs(y);
  z = fabs(z);

  //we need to get rid of the points whose position is out of range
  if(x>(rad+threshold) || y>(rad+threshold) || z>(rad+threshold)) {
    return 0;
  }
  if(sqrt(x*x + y*y + z*z) < (0.79465 * rad - threshold)) {
      return 0;	//the voxel is too close to center, 
      //speed up the process
  }
  
  //these are the cartessian values of six points enclosing 
  //the icosahedral shell
  x1 = 0.0;
  y1 = 0.0;
  z1 = 0.85065*rad; //center to mid-edge
  
  x2 = 0.0;
  y2 = 0.85065*rad;
  z2 = 0.52573*rad; //half of the edge
  
  x3 = 0.0;
  y3 = 0.85065*rad;
  z3 = 0.0;
  
  x4 = 0.85065*rad; //center to mid-edge
  y4 = 0.52573*rad;
  z4 = 0.0;
  
  x5 = 0.85065*rad;
  y5 = 0.0;
  z5 = 0.0;
  
  x6 = 0.52573*rad;
  y6 = 0.0;
  z6 = 0.85065*rad;
    
  //we need to consider the projection of the voxel on xy plane
  //so we can decide which icosahedral plane is more close to the 
  //the voxel.
    
  if((x <= 0.525729*rad) && (y <= (ka * x + 0.85084*rad))  &&
     (isOnPlane(x,y,z,x1,y1,z1,x2,y2,z2,x6,y6,z6) == 1))
    return 1;
  else if((x >= 0.525729*rad && x <= 0.85065*rad) &&
	  (y <= (kb*x-0.85065*rad)) &&
	  (isOnPlane(x,y,z,x4,y4,z4,x5,y5,z5,x6,y6,z6) == 1))
    return 1;
  else if (( (x <= 0.85065*rad) &&
	     (y <= 0.85065*rad) &&
	     (z <= 0.85065*rad) &&
	     (y >= (ka * x + 0.85084*rad)) &&
	     (y >= (kb*x - 0.85065*rad)) &&
	     (y <= (kc * x + 0.85065*rad)) &&
	     (isOnPlane(x,y,z,x2,y2,z2,x4,y4,z4,x6,y6,z6) == 1)))
    return 1;
  else if( z <= (-kd*x + 0.52573*rad) &&
	   (isOnPlane(x,y,z,x2,y2,z2,x3,y3,z3,x4,y4,z4) == 1))
    return 1;

  return 0;
  
} // end function: doIcosshell



int icosshell(double x, double y, double z, double rad, int tag)
{
  // function name: icosshell
  // Display the appropriate icosahedron

  if(tag==CUTFACE)
    return cutIcosFace(x,y,z,rad);
  else if(tag==CUTASYM)
    return cutAsymUnit(x,y,z,rad);
  else if(tag==BULGE)
    return bulge(x,y,z,rad);
  else if(tag==ICOSNORMAL)
    return doIcosshell(x,y,z,rad);

  return 0;
} // end function icosshell



int checkMemory(void *status, int line, char *file)
{
  // function name: checkMemory
  // Check to make sure malloc did indeed give us the memory we wanted.

  int rtnVal=1;
  char outStr[512];
  
  if(status==NULL) {
    rtnVal= 0;
    printf("Error allocating memory [%d][%s]\n",line,file);
    sprintf(outStr,"Error allocating memory [%d][%s]\n\
You may want to restart the program\n",line,file);
    askThem(outStr, XmDIALOG_INFORMATION);
  }
  
  return(rtnVal);
  
} // end function checkMemory



double deg2rad(double degrees)
{
  // function name: deg2rad
  // Convert degrees to radians

  return((2.0 * M_PI) * (degrees / 360.0));
  
} // end function deg2rad



double rad2deg(double rads)
{
  // function name: rad2deg
  // convert radians to degrees

  return((rads/THISTWOPI) * 360.0);
  
} // end function rad2deg



float fsign(float arg1, float arg2)
{
  // function name: fsign
  // Mimic the FORTRAN SIGN intrinsic function. Transfers the sign of
  // the second arg to the first.

  int oneNeg=FALSE, twoNeg=FALSE;
  
  if(arg1<0.0)
    oneNeg= TRUE;
  if(arg2<0.0)
    twoNeg= TRUE;
  
  if(oneNeg == twoNeg) 
    return(arg1);           /* Already the same */
  else
    return((0.0 - arg1));   /* Flip the sign */
  
} // end function fsign



void emmap_rob_dphshd_C (float *map, float *final_map,
			 float *radial_map, float *shad_map,
			 float *origMap,
			 int   useForMinMax,
			 float surfaceDensity,
			 int ncol, int nrow, int firstSec, int nsec,
			 int col1, int col2, int row1, int row2,
			 char ikey, 
			 double thetaX, double thetaY,
			 double minRadius, double maxRadius,
			 double xCen, double yCen, double zCen,
			 float radialXcen, float radialYcen, float radialZcen,
			 int lowRes, int topRadialSec)
{
  // function name: emmap_rob_dphshd_C
  //
  //          *****     SUBROUTINE EMMAP_ROB_DPHSHD     *****            
  //                                                                     
  //  THIS IS ADAPTED FROM THE PROGRAM SURFACE WHICH PRODUCES
  //  DEPTH-CUE MAPS AND A SHADED SURFACE REPRESENTATION OF A
  //  THREE-DIMENSIONAL MAP.
  //                                                                    
  //  SUB-MENU TOUCH-KEY OPTIONS:                                          
  //                                                                       
  //      A   COMPUTE DEPTH-CUE MAP FROM 3D MAP                            
  //      B   COMPUTE SHADED MAP FROM DEPTH-CUE DATA                       
  //      D   DISPLAY DEPTH-CUE OR SHADED DATA ON GRAPHICS                 
  //      E   EXIT SUBROUTINE                                              
  //      W   WRITE DEPTH-CUE OR SHADED DATA TO "MAP" DISK FILE            
  //                                                                       
  //  UNITS: 3 = TEMPORARY DIRECT ACCESS SCRATCH FILE OF DEPTH-CUE IMAGES. 
  //                                                                       
  //  MAY 1988 PURDUE UNIVERSITY VAX 8550, T. BAKER.                       
  //  MODIFIED  9-MAY-1991 BY T. BAKER.                                    
  //  Modified 20-JAN-1994 by D. Belnap				     
  //
  //  SUBROUTINE EMMAP_ROB_DPHSHD (LUN,MAP,NCOL,NROW,NSEC,TMP,DPHSHD_MAP)
  //
  //   Ported to C May 1996  RWA
  //
  //   Main loop of the emmap cueing/shading routines.

  int num_views=1, sec1, sec2;
  float surf_dens;
  double theta;
  BYTE surf=FALSE, shift=FALSE;
  
  // INITIALIZE PARAMETER DEFAULTS
  sec1           = firstSec;
  sec2           = nsec;
  surf_dens      = surfaceDensity;
  
  theta = globTheta;     /* Passed from XWindows */
  
  surf_dens = fabs(surf_dens);
  
  /* we select 'd' for density from the dph_options() */
  /* this causes surf=TRUE to happen */
  surf = TRUE;
  
  /* DO depth cue ON_THE_FLY ... */
  if(radialMapShade)
    dph_nomem_tilt_fast(map, origMap, ncol, nrow, radial_map, 
			shift, surf, sec1, sec2, 
			col1, col2, row1, row2,
			num_views, 
			irev, surf_dens, useForMinMax, shad_map,
			thetaX, thetaY, minRadius, maxRadius,
			xCen, yCen, zCen,
			radialXcen, radialYcen, radialZcen,
			lowRes, topRadialSec);
  else 
    dph_nomem_tilt_fast(map, origMap, ncol, nrow, final_map, 
			shift, surf, sec1, sec2, 
			col1, col2, row1, row2,
			num_views, 
			irev, surf_dens, useForMinMax, shad_map,
			thetaX, thetaY, minRadius, maxRadius,
			xCen, yCen, zCen,
			radialXcen, radialYcen, radialZcen,
			lowRes, topRadialSec);
  
} // end function emmap_rob_dphshd_C



void emmap_rob_shd1 (cue_t *cuedMap,
		     double *tmp2d, double *tmp,
		     int ncol, int nrow,
		     int sec1, int sec2, 
		     float fract, 
		     int useForMinMax, 
		     float glbAngx, float stepVal,
		     float glbAngy, int whichStep,
		     float passMin, int angx, int angy, int thisFract)
{
  // function name: emmap_rob_shd1
  //
  //                *****     SUBROUTINE EMMAP_SHD     *****                
  //                                                                    
  // ADAPTED FROM MRREFL.FOR BY M. RADERMACHER OF ALBANY - STRAIGHT ON
  // ILLUMINATION
  //
  // UNIT: 3 = TEMPORARY DIRECT ACCESS SCRATCH FILE OF DEPTH-CUE MAP
  // (OPEN IF DPH_SCRATCH = .TRUE.)
  //                                                                    
  // IN EMMAP_SHD1, THE DEPTH_CUE MAP ALREADY EXISTS IN CORE MEMORY
  // (DPH_SCRATCH = .FALSE.) IN THE ARRAY "DPHSHD_MAP", OR IN A
  // SCRATCH FILE (DPH_SCRATCH = .TRUE.).
  //                                                                    
  // PROGRAMMERS AND USERS PLEASE NOTE: THE ALGORITHM FOR SHADING AT
  // ANGLES AWAY FROM THE VIEW DIRECTION IS NOT STRICTLY CORRECT IN
  // THE SENSE THAT NO ACCOUNT IS MADE FOR ELIMINATING LIGHT FROM
  // FALLING IN AREAS THAT WOULD OTHERWISE BE SHADED DUE TO
  // OBSTRUCTIONS TO THE PATH OF THE LIGHT SOURCE
  //                                                                    
  // LAST MODIFIED 22-APR-1991 BY T. BAKER.                              
  //
  // Ported to C May 1996 RWA
  //
  // Routine to shade depth/radial cued views.

  int       i, ii, j, jj, k;
  double    scaleIt, old_min;
  double    depth, depth1, depth2, depth3, shad, xx, yy, cs;
  double    thisMin=HUGE, thisMax=-HUGE;
  float     *thisMap,*radMap,*shadMap;
  int       shadStep;
  double    sinx, cosx, siny, cosy;
  
  map_min  = passMin;
  old_min  = (double)map_min;
  scaleIt  = (double)(map_max - map_min);   /*! to convert to 0-255 scale */
  
  // Determine the shading for the map
  
  for(k=(sec1-1); k<sec2; k++) {
    
    if(whichStep==angx)
      glbAngx += stepVal;
    else if(whichStep==angy)
      glbAngy -= stepVal;
    else if(whichStep==thisFract)
      fract += stepVal;
    
    sinx = sin(deg2rad(glbAngx));
    cosx = cos(deg2rad(glbAngx));
    siny = sin(deg2rad(glbAngy));
    cosy = cos(deg2rad(glbAngy));
    
    thisMap  = cuedMap[k].depth;
    radMap   = cuedMap[k].radial;
    shadMap  = cuedMap[k].shad;
    
    for(j=0; j<nrow; j++)
      for(i=0; i<ncol; i++)
	tmp2d[i+(j*ncol)] = (double)thisMap[i+(j*ncol)];
    
    for(j=0; j<nrow; j++) {
      if(j == (nrow-1))
	jj = 0;	     /*duplicate first value for wrap around*/
      else
	jj = j + 1;
      
      for(i=0; i<ncol; i++) {
	tmp[i] = 0.0;		/* reinitialize tmp array*/
	tmp[i+1] = 0.0;		/* reinitialize tmp array*/
	if(i == (ncol-1))
	  ii = 0;		    /* duplicate first value for wrap around*/
	else
	  ii = i + 1;
	
	depth1 = tmp2d[i+(j*ncol)]  - old_min;
	depth2 = tmp2d[ii+(j*ncol)] - old_min;
	depth3 = tmp2d[i+(jj*ncol)] - old_min;
	
	if(depth1 > 0.0 && depth2 > 0.0 && depth3 > 0.0) {
	  
	  /* OK here's where our "striping" problem happens when
	   * we try and rotate maps.  Suppose we are trying to 
	   * rotate a MAP 35 degrees in Y and we have some surfaces
	   * that have been "cut".  These cut surfaces are depth
	   * cued such that they have very small differences in
	   * their depth in the local area of depth1,depth2,depth3
	   * that we are looking at here.  Because small numbers
	   * cause HUGE differences in the calculation of shad
	   * below we get striping.  Here's some example points from
	   * a MAP I tried:
	   *
	   * j[145]i[0]
	   * depth1[72.000000]depth2[74.000000]depth3[72.000000]
	   * xx[-2.000000]yy[0.000000]cs[0.447214]shad[42.318746]
	   * (xx * sinx)[-0.347296] (yy * siny)[0.000000]  
	   * (cosx * cosy)[0.969846]
	   * j[145]i[1]
	   * depth1[74.000000]depth2[75.000000]depth3[74.000000]
	   * xx[-1.000000]yy[0.000000]cs[0.707107]shad[85.575559]
	   * (xx * sinx)[-0.173648] (yy * siny)[0.000000]  
	   * (cosx * cosy)[0.969846]
	   *
	   * Notice that the first point has a difference of -2 for
	   * xx and the second has a difference of -1.  These cause
	   * a big difference in the shad values of 42 and 85 
	   * respectively.
	   *
	   * RWA 8/6/97 - I'm wondering why the depths aren't
	   * showing up in fractional floating point - they are 
	   * always integer...  I'm going to check this out
	   * in dph_nomem_tilt_fast().  
	   *
	   * I'm now thinking it's hard to believe that getting any
	   * better depths out of dph_nomem_tilt_fast() will
	   * help this problem.  It's already giving us differences
	   * in depth that are very small.  I'm wondering if
	   * we need an entirely new algorithm for shading as
	   * in the four lines below.
	   */
	  
	  xx   = (depth1-depth2);
	  yy   = (depth1-depth3); 
	  cs   = 1.0/sqrt(xx*xx + yy*yy + 1.0);
	  shad = scaleIt * cs * (xx*sinx + yy*siny + cosx*cosy);
	  
	  if(shad < 0.0) 
	    shad = 0.0;
	  if(!radialMapShade)
	    shad  = shad*fract;  /* Fract has no meaning for radial */
	  if(radialMapShade) { /* We are shading radially here */
	    int steps,foundIt;
	    
	    depth = radMap[i + j*ncol];
	    
	    foundIt = FALSE;
	    shadStep = (int)(160.0/expStep);
	    for(steps=0;steps<expStep;steps++) {/* Calc amount of shade */
	      int cutOff;
	      cutOff = (int)(160 - (steps * shadStep));
	      if(!foundIt)
		if(shad > cutOff) {/* found it */
		  shad = 0.0 - steps;
		  foundIt = TRUE;
		}
	    }
	    if(!foundIt)
	      shad = (double)((double)0.0 - (double)(expStep-1));
	    
	    if(depth == 0.0)
	      shad = 0.0;
	    if((depth + shad)<0.0)
	      shad = 0.0;
	    
	  }/* We are shading radially here */
	  else 
	    depth = depth1 * (1.0 - fract);
	  
	  tmp[i] = shad + depth;
	  
	  if(thisMin > tmp[i])
	    thisMin = tmp[i];
	  if(thisMax < tmp[i])
	    thisMax = tmp[i];
	  
	}/* then */
	
      }/* For i */
      
      for(i = 0; i<ncol; i++)
	shadMap[i + j*ncol] = tmp[i];
      
    }/* For j */
    
    
  }/* For k */
  if(thisMax == thisMin)
    thisMin -= 0.01;
  
  if(useForMinMax) { // update globals
    map_min = thisMin;
    map_max = thisMax;
  }
  
} // end function emmap_rob_shd1



void dph_nomem_tilt_fast (float *coremap, float *origMap,
			  int ncol, int nrow, 
			  float *dphshd_map, 
			  BYTE shift, BYTE surf, 
			  int sec1, int sec2, 
			  int col1, int col2,
			  int row1, int row2,
			  int num_views,
			  int thisIrev, float surf_dens,
			  int useForMinMax,
			  float *shad_map,
			  double thetaX, double thetaY,
			  double minRadius, double maxRadius,
			  double xCen, double yCen, double zCen,
			  float radialXcen, float radialYcen, float radialZcen,
			  int lowRes, int topRadialSec)
{
  // function name: dph_nomem_tilt_fast
  //
  // Modified from dph_notilt_fast()
  // 
  // RW AShmore July 29, 1997 - I'm trying to incorporate the concept
  // that as the volume they want to look at gets smaller this routine
  // should get faster.  Suppose they cut the map down so that they
  // want to look at:
  //
  //        col   33   to   88   (col1 and col2)
  //        row   33   to   88   (row1 and row2)
  //        sec   0    to   55   (sec1 and sec2)
  //
  // So if we just alter our for(col=0;col<ncol;col++) loops below to
  // be for(col=col1,col<col2,col++) we will "clip off" the sections
  // below 33 until they've roatated into the col and row constraints.
  // We need to choose col1, col2, row1, row2, sec1 and sec2
  // intelligently such that when we rotate we see all we can.  I've
  // left it opened up for now so that we don't miss anything (e.g. 0
  // to ncols etc...) but with a little thought we might be able to
  // optimize this for speed.
  //
  // Suppose we are just rotating in Y - then we don't need to worry
  // to worry about the rows - they will always fall in the row1 row2
  // region.  Same for X - cols will be within region.  I've added
  // code below to stuff the startRow endRow startCol endCol
  // appropriately.
  //
  // This rtn has some problems with striping on the display of any
  // map that is cut and then rotated.  Not sure why it is happening.

  int      col, row;
  int      startCol,endCol,startRow,endRow;
  double    thisSecn,thisRow,thisCol;
  int      secn, topSec, bottomSec;
  int      colOffset,rowOffset,secOffset;
  double    origNumSecsM1,nrowM1,ncolM1,sec2M1,row2M1,col2M1;
  double    topSecMtmp1;
  double    tmp1;
  double   tmp_fract;
  double   dens1, dens2, dens_diff, absDensDiff;
  double   qSq, mapRadiusSq, minRadSq,maxRadSq;
  double   xSq,ySq,zSq,mapRadius;
  double   thisSecnF,thisRowF,thisColF;
  double   cthetaX,sthetaX,cthetaY,sthetaY;
  
  origNumSecsM1= (double)(origNumSecs - 1.0);
  nrowM1       = (double)(nrow - 1.0);
  ncolM1       = (double)(ncol - 1.0);
  sec2M1 = sec2 - 1.0;
  row2M1 = row2 - 1.0;
  col2M1 = col2 - 1.0;
  
  minRadSq= minRadius * minRadius;
  maxRadSq= maxRadius * maxRadius;
  
  cthetaX = cos(deg2rad(thetaX));
  sthetaX = sin(deg2rad(thetaX));
  cthetaY = cos(deg2rad(thetaY));
  sthetaY = sin(deg2rad(thetaY));
  
  /* How much do we have to offset for asymmetric maps */
  colOffset = (largestDim - ncol) / 2;
  rowOffset = (largestDim - nrow) / 2;
  secOffset = (largestDim - origNumSecs) / 2;
  xCen += colOffset;
  yCen += rowOffset;
  
  radialXcen += colOffset;
  radialYcen += rowOffset;
  
  /* Now try and optimize for speed */
  startCol= 0;
  endCol  = largestDim;
  startRow= 0;
  endRow  = largestDim;
  
  if(thetaX==0.0) { /* We don't have to worry about rows */
    startRow = row1;
    endRow   = row2;
    endRow   += rowOffset; 
  }
  if(thetaY==0.0) { /* We don't have to worry about cols */
    startCol = col1;
    endCol   = col2;
    endCol   += colOffset;
  }
  
  topSec = origNumSecs;
  if(!lowRes)
    topSec = largestDim;  /* Does this give us a better rotation??? */
  else 
    topSec = origNumSecs;
  
  /* Try this */
  topSec = largestDim;
  
  bottomSec = 0;
  startCol= 0;
  endCol  = largestDim;
  startRow= 0;
  endRow  = largestDim;
  
  if(!lowRes && (radialMapShade || saveRadialMapShade)) { /* Do for Highres */
    /* topSec *= 2.0; */     /* In case we have to rotate it */
    xCen = radialXcen;
    yCen = radialYcen;
    zCen = radialZcen; 
  }
  
  for(row = startRow;row<endRow;row++) { // process data row by row
    
    ySq = (row-yCen) * (row-yCen);
    
    for(col = startCol;col <endCol;col++) { // for col
      tmp1 = 0.0;
      
      xSq = (col-xCen) * (col-xCen);
      
      qSq = xSq+ySq;
      
      for(secn = topSec-1; secn >= bottomSec; secn--) {/* for secn */
	
	// Check to see if we're in radius
	zSq = (secn-zCen) * (secn-zCen);
	mapRadiusSq = (qSq+zSq);
	if(mapRadiusSq < minRadSq ||  mapRadiusSq > maxRadSq)
	  continue;
	
	// Init it
	dens1= surf_dens - 1.0;
	
	/* Get dens1 based on the rotation of thetaX and thetaY */
	/* First do rotation in Y */
	if(thetaY != 0.0) {/* rotate */
	  thisSecnF = (col-xCen)*(0.0-sthetaY) + (secn-zCen)*cthetaY;
	  thisRowF  = row;
	  thisColF  = (col-xCen)*cthetaY + (secn-zCen)*sthetaY;
	  thisSecnF  += zCen;
	  thisColF  += xCen;
	}
	else {/* No rotate */
	  thisSecnF = secn;
	  thisRowF  = row;
	  thisColF  = col;
	}
	/* Now do it in X */
	if(thetaX != 0.0) {/* rotate */
	  thisSecn = (thisRowF-yCen)*sthetaX + (thisSecnF-zCen)*cthetaX;
	  thisRow  = (thisRowF-yCen)*cthetaX + (thisSecnF-zCen)*(0.0-sthetaX);
	  thisCol  = thisColF;
	  thisRow  += yCen;
	  thisSecn += zCen;
	}
	else {/* Don't rotate */
	  thisSecn = thisSecnF;
	  thisRow  = thisRowF;
	  thisCol  = thisColF;
	}
	
	/* Center this thing */
	thisCol -= colOffset;
	thisRow -= rowOffset;
	
	/* October 21, 1998 */
	if(thisSecn >= sec1 && thisSecn < sec2M1 &&
	   thisRow  >= row1 && thisRow  < row2M1 &&
	   thisCol  >= col1 && thisCol  < col2M1) {/* OK to try this */
	  
	  double Ic,Jc,Kc,If,Jf,Kf;
	  double i,j,k;
	  
	  /* Add 0.0001 to all to force the ceiling to be
	   * the next pixel for thetas of modulus 90.0 */
	  Ic= ceil(thisCol);
	  Jc= ceil(thisRow);
	  Kc= ceil(thisSecn);
	  If = floor(thisCol);
	  Jf = floor(thisRow);
	  Kf = floor(thisSecn);
	  i = thisCol;
	  j = thisRow;
	  k = thisSecn;
	  if(Ic==If)
	    Ic = If + 1.0;
	  if(Jc==Jf)
	    Jc = Jf + 1.0;
	  if(Kc==Kf)
	    Kc = Kf + 1.0;
	  if(thetaY != 0.0    && thetaX != 0.0 ) {
	    dens1 = 
	      (Ic-i)*(Jc-j)*(Kc-k)*COREMAPREF((Ic-1),(Jc-1),(Kc-1)) +
	      (Ic-i)*(Jc-j)*(k-Kf)*COREMAPREF((Ic-1),(Jc-1),Kc) +
	      (i-If)*(Jc-j)*(Kc-k)*COREMAPREF(Ic,(Jc-1),(Kc-1)) +
	      (i-If)*(Jc-j)*(k-Kf)*COREMAPREF(Ic,(Jc-1),Kc) +
	      (Ic-i)*(j-Jf)*(Kc-k)*COREMAPREF((Ic-1),Jc,(Kc-1)) +
	      (Ic-i)*(j-Jf)*(k-Kf)*COREMAPREF((Ic-1),Jc,Kc) +
	      (i-If)*(j-Jf)*(Kc-k)*COREMAPREF(Ic,Jc,(Kc-1)) +
	      (i-If)*(j-Jf)*(k-Kf)*COREMAPREF(Ic,Jc,Kc);
	  }
	  else if(thetaY == 0.0 && thetaX != 0.0) {
	    dens1 = 
	      (Jc-j)*(Kc-k)*COREMAPREF(If,(Jc-1),(Kc-1)) +
	      (Jc-j)*(k-Kf)*COREMAPREF(If,(Jc-1),Kc) +
	      (j-Jf)*(Kc-k)*COREMAPREF(If,Jc,(Kc-1)) +
	      (j-Jf)*(k-Kf)*COREMAPREF(If,Jc,Kc);
	  }
	  else if(thetaY != 0.0 && thetaX == 0.0) {
	    dens1 =
	      (Ic-i)*(Kc-k)*COREMAPREF((Ic-1),Jf,(Kc-1)) +
	      (Ic-i)*(k-Kf)*COREMAPREF((Ic-1),Jf,Kc) +
	      (i-If)*(Kc-k)*COREMAPREF(Ic,Jf,(Kc-1)) +
	      (i-If)*(k-Kf)*COREMAPREF(Ic,Jf,Kc);
	  }
	  else {/* No rotation */
	    dens1 = coremap[(int)thisSecn*ncol*nrow+
			    (int)thisRow*ncol + (int)thisCol];
	  }
	  
	  dens1 *= thisIrev;
	  
	}/* OK to try this */
	else {/* Just go to next pixel */
	  dens1 = surf_dens - 1.0;
	}
	
	if(dens1 >= surf_dens) {/*then     ! reached threshold density level */
	  
	  if(TRUE) {/* Not top row/col/sec else secn */
	    double Ic,Jc,Kc,If,Jf,Kf;
	    double i,j,k;
	    
	    /* Init it */
	    dens2= surf_dens - 1.0;
	    
	    /* This mimics what's done for dens1 above */
	    /* Get dens2 here from rotation of thetaX
	     * and thetaY */
	    /* First do rotation in Y */
	    if(thetaY != 0.0) {/* rotate */
	      thisSecnF  = (col-xCen)*(0.0-sthetaY) + (secn+1.0-zCen)*cthetaY;
	      thisRowF   = row;
	      thisColF   = (col-xCen)*cthetaY + (secn+1.0-zCen)*sthetaY;
	      thisSecnF += zCen;
	      thisColF  += xCen;
	    }
	    else {/* No rotate */
	      thisSecnF = secn+1.0;
	      thisRowF  = row;
	      thisColF  = col;
	    }
	    /* Now do it in X */
	    if(thetaX != 0.0) {/* rotate */
	      thisSecn  = (thisRowF-yCen)*sthetaX + (thisSecnF-zCen)*cthetaX;
	      thisRow   = (thisRowF-yCen)*cthetaX + 
		(thisSecnF-zCen)*(0.0-sthetaX);
	      thisCol   = thisColF;
	      thisRow  += yCen;
	      thisSecn += zCen;
	    }
	    else {/* Don't rotate */
	      thisSecn = thisSecnF;
	      thisRow  = thisRowF;
	      thisCol  = thisColF;
	    }
	    
	    /* Center this thing */
	    thisCol -= colOffset;
	    thisRow -= rowOffset;
	    
	    /* We need to be able to see pixels outside the cut zones so
	     * do a wide search here */
	    
	    if(thisSecn >= 0.0 && thisSecn < origNumSecsM1 &&
	       thisRow  >= 0.0 && thisRow  < nrowM1 &&
	       thisCol  >= 0.0 && thisCol  < ncolM1 ) { // OK to proceed
	      /* Add 0.0001 to all to force the ceiling to be
	       * the next pixel for thetas of modulus 90.0 */
	      Ic= ceil(thisCol);
	      Jc= ceil(thisRow);
	      Kc= ceil(thisSecn);
	      
	      If = floor(thisCol);
	      Jf = floor(thisRow);
	      Kf = floor(thisSecn);
	      if(Ic==If)
		Ic = If + 1.0;
	      if(Jc==Jf)
		Jc = Jf + 1.0;
	      if(Kc==Kf)
		Kc = Kf + 1.0;
	      i = thisCol;
	      j = thisRow;
	      k = thisSecn;
	      
	      if(thetaY != 0.0 && thetaX != 0.0) {
		dens2 = 
		  (Ic-i)*(Jc-j)*(Kc-k) * COREMAPREF((Ic-1),(Jc-1),(Kc-1)) +
		  (Ic-i)*(Jc-j)*(k-Kf) * COREMAPREF((Ic-1),(Jc-1),Kc) +
		  (i-If)*(Jc-j)*(Kc-k) * COREMAPREF(Ic,(Jc-1),(Kc-1)) +
		  (i-If)*(Jc-j)*(k-Kf) * COREMAPREF(Ic,(Jc-1),Kc) +
		  (Ic-i)*(j-Jf)*(Kc-k) * COREMAPREF((Ic-1),Jc,(Kc-1)) +
		  (Ic-i)*(j-Jf)*(k-Kf) * COREMAPREF((Ic-1),Jc,Kc) +
		  (i-If)*(j-Jf)*(Kc-k) * COREMAPREF(Ic,Jc,(Kc-1)) +
		  (i-If)*(j-Jf)*(k-Kf) * COREMAPREF(Ic,Jc,Kc);
	      }
	      else if(thetaY == 0.0 && thetaX != 0.0) {
		dens2 =
		  (Jc-j)*(Kc-k) * COREMAPREF(If,(Jc-1.0),(Kc-1.0)) +
		  (Jc-j)*(k-Kf) * COREMAPREF(If,(Jc-1.0),Kc) +
		  (j-Jf)*(Kc-k) * COREMAPREF(If,Jc,(Kc-1.0)) +
		  (j-Jf)*(k-Kf)* COREMAPREF(If,Jc,Kc);
	      }
	      else if(thetaY != 0.0 && thetaX == 0.0) {
		dens2    =
		  (Ic-i)*(Kc-k) * COREMAPREF((Ic-1.0),Jf,(Kc-1.0)) +
		  (Ic-i)*(k-Kf) * COREMAPREF((Ic-1.0),Jf,Kc) +
		  (i-If)*(Kc-k) * COREMAPREF(Ic,Jf,(Kc-1.0)) +
		  (i-If)*(k-Kf) * COREMAPREF(Ic,Jf,Kc);
	      }
	      else {/* No rotation */
		dens2 = coremap[(int)thisSecn*ncol*nrow+
				(int)thisRow*ncol + (int)thisCol];
	      }
	      
	      dens2 *= thisIrev;
	      
	      /*******************************************
	       * This is it!!!!!  If you're going from the
	       * inside out then reverse the dens2 
	       * concept!!
	       ******************************************/
	      if(((secn <= zCen) && 
		  (mapRadiusSq <= minRadSq*1.1)) ||
		 (mapRadiusSq >= maxRadSq*0.95))
		dens2 *= -1.0;
	      
	      dens_diff = dens1 - dens2;
	      
	      tmp1      = topSec-secn;
	      
	      //trap for dens_diff close to 0.0
	      if(dens_diff < 0.0)
		absDensDiff = 0.0 - dens_diff;
	      else 
		absDensDiff = dens_diff;
	      
	      if(absDensDiff < 0.001) {
		/* Do nothing tmp1 = tmp1; */
	      }
	      else {/* else A*/
		double numer,denom;
		numer = surf_dens - dens2;
		denom = dens_diff;
		
		if(denom != 0.0)
		  tmp_fract = numer / denom;
		else 
		  tmp_fract = 0.0;
		
		if(dens1 <= dens2)
		  tmp_fract = 0.0;
		
		if(thetaY != 0.0 || thetaX != 0.0) {
		  if(thisCol  >= col2-2) //On a cutting edge, get fract diff
		    tmp_fract = thisCol - (col2-2);
		  if(thisRow  >= row2-2) //On a cutting edge, get fract diff
		    tmp_fract = thisRow - (row2-2);
		  if(thisSecn >= sec2-2) //On a cutting edge, get fract diff
		    tmp_fract = thisSecn - (sec2-2);
		}
		
		if(tmp_fract > 0.0) 
		  tmp1 = tmp1 + tmp_fract;
		
	      }/* else A*/
	      
	      if(radialMapShade) {/* Compute the radius here */
		if(!lowRes) {/* Do for Highres */
		  double qSq;
		  
		  xSq = (col-radialXcen)*(col-radialXcen);
		  ySq = (row-radialYcen)*(row-radialYcen);
		  qSq = xSq+ySq;
		  zSq = (secn-radialZcen)*(secn-radialZcen);
		  mapRadius = sqrt(qSq+zSq);
		  
		  /* Why do we have to halve it ??? */
		  /* mapRadius *= 0.5;   */
		  
		  tmp1= (topRadialSec - mapRadius) ;
		  /* If the center pixel is zero then
		   * it looks just like background -
		   * let's push it up above backgrd */
		  if(tmp1>=topRadialSec)
		    tmp1 = topRadialSec-1.0; 
		  
		  topSecMtmp1 = topRadialSec - tmp1;
		  topSecMtmp1 = mapRadius; 
		  
		}/* Do for Highres */
		else {/* Normal */
		  mapRadius = sqrt(mapRadiusSq);
		  tmp1= (topSec - mapRadius) ;
		  /* If the center pixel is zero then
		   * it looks just like background -
		   * let's push it up above backgrd */
		  if(tmp1==topSec)
		    tmp1 = topSec-1.0; 
		  
		}/* Normal */
	      }/* COmpute the radius here */
	      
	      if(lowRes || !radialMapShade) {/* OK to do these */
		if(tmp1 > topSec)
		  tmp1= topSec;
		
		topSecMtmp1 = topSec - tmp1;
	      }/* OK to do these */
	      else {/* We're high res and radialMapShade */
		/* Just use map radius */
	      }
	      
	      dphshd_map[(row*largestDim)+col] = topSecMtmp1;
	      shad_map[(row*largestDim)+col] = topSecMtmp1;
	      
	      if(useForMinMax) {/* If radial we may not want to include
				 * the depth-cued one here */
		
		if(topSecMtmp1 != 0.0) {/* DOn't include bckgrnd in min/max */
		  if(topSecMtmp1 < map_min)
		    map_min = topSecMtmp1; /*globmin */
		  if(topSecMtmp1 > map_max)
		    map_max = topSecMtmp1; /*globmax */
		}
		
	      }/* If radial we may not want to include
		* the depth-cued one here */
	      
	      /* Found the surface we're looking for -
	       * goto next column */
	      goto nextColumn;
	      
	    }/* OK to proceed */
	    
	  }/* Not top row/col/sec else secn */
	  
	}/*then     ! reached threshold density level */
	
      }/* for secn *//* ! step to next map section */
            
    nextColumn:;
    }/* for col ! step to next map column*/
    
    
  }/*! process data row by row*//*! step to next map row*/
  
} // end function dph_nomem_tilt_fast



void newTranslateThetaPhiOmega(double *thisCol, 
			       double *thisRow,
			       double *thisSecn,
			       double xCen, double yCen, double zCen,
			       int col, int row, int secn,
			       double t1, double t2, double t3,
			       double t4, double t5, double t6,
			       double t7, double t8, double t9)
{
  // function name: newTranslateThetaPhiOmega
  // Newer rtn for translating the roations.  Might be superior???

  double x,y,z;
  
  z = secn - zCen;
  y = row  - yCen;
  x = col  - xCen;
  
  *thisSecn = t7*x + t8*y + t9*z + zCen;
  *thisCol  = t1*x + t2*y + t3*z + xCen;
  *thisRow  = t4*x + t5*y + t6*z + yCen;
  
} // end function newTranslateThetaPhiOmega



void translateThetaPhiOmega(double *thisCol, double *thisRow, double *thisSecn,
			    double xCen, double yCen, double zCen,
			    int col,int row,int secn,
			    double theta, double phi, double omega,
			    double stheta, double ctheta,
			    double sphi, double cphi,
			    double somega, double comega)
{
  // function name: translateThetaPhiOmega
  // Orig way to do the rotation.  This rtn might be useful if we ever
  // get the x/y/z rotation implemented.

  float tmpX, tmpY, tmpZ, tx, ty, tz;
  float origX, origY, origZ;
  
  origX= tmpX = col - xCen;
  origY= tmpY = row - yCen;
  origZ= tmpZ = secn - zCen;
  
  if(theta != 0.0) { // theta
    tx=tmpX*ctheta  + tmpZ*stheta;
    tz=tmpX*-stheta + tmpZ*ctheta;
    tmpX = tx;
    tmpZ = tz;
  }
  
  if(phi != 0.0) { // phi
    ty=tmpY*cphi  + tmpZ*sphi;
    tz=tmpY*-sphi + tmpZ*cphi;
    tmpY = ty;
    tmpZ = tz;
  }
  
  if(omega != 0.0) { // Omega
    tx=tmpX*comega  + tmpY*somega;
    ty=tmpX*-somega +tmpY*comega;
    tmpX = tx;
    tmpY = ty;
  }
  
  tmpX += xCen;
  tmpY += yCen;
  tmpZ += zCen;
  
  *thisSecn = tmpZ;
  *thisRow  = tmpY;
  *thisCol  = tmpX;
  
} // end function translateThetaPhiOmega



double getVoxel(float *coremap,
		double thisCol, double thisRow, double thisSecn,
		double theta, double phi, double omega,
		int ncol, int nrow)
{
  // function name: getVoxel
  // Get the interpolated voxel at the desired location.  NOTE: We
  // need to add the omega interpolation here.

  double rtnVal, i, j, k, Ic, Jc, Kc, If, Jf, Kf;
  
  Ic = ceil(thisCol);
  Jc = ceil(thisRow);
  Kc = ceil(thisSecn);
  If = floor(thisCol);
  Jf = floor(thisRow);
  Kf = floor(thisSecn);

  if(Ic == If)
    Ic = If + 1.0;
  if(Jc == Jf)
    Jc = Jf + 1.0;
  if((Kc == Kf) && (Kf < (double)largestDim-1.0))
    Kc = Kf + 1.0;

  i = thisCol;
  j = thisRow;
  k = thisSecn;
  
  if(theta != 0.0 || phi != 0.0 || omega != 0.0)
    rtnVal = 
      (Ic-i)*(Jc-j)*(Kc-k)*COREMAPREF((Ic-1.0),(Jc-1.0),(Kc-1.0)) +
      (Ic-i)*(Jc-j)*(k-Kf)*COREMAPREF((Ic-1.0),(Jc-1.0),Kc)       +
      (i-If)*(Jc-j)*(Kc-k)*COREMAPREF(Ic,(Jc-1.0),(Kc-1.0))       +
      (i-If)*(Jc-j)*(k-Kf)*COREMAPREF(Ic,(Jc-1.0),Kc)             +
      (Ic-i)*(j-Jf)*(Kc-k)*COREMAPREF((Ic-1.0),Jc,(Kc-1.0))       +
      (Ic-i)*(j-Jf)*(k-Kf)*COREMAPREF((Ic-1.0),Jc,Kc)             +
      (i-If)*(j-Jf)*(Kc-k)*COREMAPREF(Ic,Jc,(Kc-1.0))             +
      (i-If)*(j-Jf)*(k-Kf)*COREMAPREF(Ic,Jc,Kc); 
  else // No rotation
    rtnVal    = coremap[(int)thisSecn*ncol*nrow +
			(int)thisRow*ncol + (int)thisCol];
    
  return(rtnVal);
  
} // end function getVoxel



void thetaPhi_dphshd_C(float *coremap, float *origMap,
		       int ncol, int nrow, 
		       float *dphshd_map, 
		       BYTE shift, BYTE surf, 
		       int sec1, int sec2, 
		       int col1, int col2,
		       int row1, int row2,
		       int num_views,
		       int thisIrev, float surf_dens,
		       int useForMinMax,
		       float *shad_map,
		       double phi, double theta, double omega,
		       double minRadius, double maxRadius,
		       double xCen, double yCen, double zCen,
		       float radialXcen, float radialYcen, float radialZcen,
		       int lowRes, int topRadialSec, int step)
{
  // function name: thetaPhi_dphshd_C
  //
  // This is the display of the "TRUE" phi rotation.  Use the
  // emmap_rob_dphsgd_C rtn for the "normal" way we do it.  Not sure
  // why they want 2 ways to do it but there seems to be controversy
  // as to which way is correct.  We are still getting some striping
  // on cut and rotated maps.  Modified from DPH_NOMEM_TILT_FAST

  int      col, row, r, s, printIts;
  int      startCol,endCol,startRow,endRow;
  double   thisSecn,thisRow,thisCol;
  double   diffit;
  int      secn, topSec, bottomSec;
  int      colOffset,rowOffset,secOffset;
  double   origNumSecsM1,nrowM1,ncolM1,sec2M1,row2M1,col2M1;
  double   topSecMtmp1;
  double   tmp1;
  double   tmp_fract,cutFract;
  double   dens1, dens2, dens_diff, absDensDiff;
  double   qSq, mapRadiusSq, minRadSq, maxRadSq;
  double   xSq,ySq,zSq,mapRadius;
  double   cphi,sphi,ctheta,stheta,comega,somega;
  double   t1,t2,t3,t4,t5,t6,t7,t8,t9;
  
  printIts = 0;

  origNumSecsM1 = origNumSecs - 1.0;
  nrowM1        = nrow - 1.0;
  ncolM1        = ncol - 1.0;
  sec2M1        = sec2 - 1.0;
  row2M1        = row2 - 1.0;
  col2M1        = col2 - 1.0;
  
  minRadSq = minRadius * minRadius;
  maxRadSq = maxRadius * maxRadius;
  
  /* We've got a bug whereby we've got to add 90 degress to
   * our omega values ... */

  // Avoiding exact cancellations in floating point operations
  theta += 0.0001234;
  phi   += 0.0001234;
  omega += 0.0001234;

  comega = cos(deg2rad(0.0-phi));
  somega = sin(deg2rad(0.0-phi));
  cphi   = cos(deg2rad(omega));
  sphi   = sin(deg2rad(omega));
  ctheta = cos(deg2rad(0.0-theta));
  stheta = sin(deg2rad(0.0-theta));
  
  // COMPUTE ELEMENTS OF ROTATION MATRIX
  t1 = ctheta*cphi*comega + sphi*somega;
  t2 = ctheta*sphi*comega - cphi*somega;
  t3 = -stheta*comega;
  t4 = ctheta*cphi*somega - sphi*comega;
  t5 = ctheta*sphi*somega + cphi*comega;
  t6 = -stheta*somega;
  t7 = stheta*cphi;
  t8 = stheta*sphi;
  t9 = ctheta;
  
  
  /* How much do we have to offset for asymmetric maps */
  colOffset = (largestDim - ncol) / 2;
  rowOffset = (largestDim - nrow) / 2;
  secOffset = (largestDim - origNumSecs) / 2;
  xCen += colOffset;
  yCen += rowOffset;
  
  radialXcen += colOffset;
  radialYcen += rowOffset;
  
  /* Now try and optimize for speed */
  startCol= 0;
  endCol  = largestDim;
  startRow= 0;
  endRow  = largestDim;
  
  if(omega==0.0) {
    if(phi==0.0) {/* We don't have to worry about rows */
      startRow = row1;
      endRow   = row2;
      endRow   += rowOffset; 
    }
    if(theta==0.0) {/* We don't have to worry about cols */
      startCol = col1;
      endCol   = col2;
      endCol   += colOffset;
    }
  }
  
  topSec = origNumSecs;
  if(!lowRes)
    topSec = largestDim;  /* Does this give us a better rotation??? */
  else 
    topSec = origNumSecs;
  
  /* Try this */
  topSec = largestDim;
  
  bottomSec = 0;
  /********************************
   * Why do we reset these here???
  startCol= 0;
  endCol  = largestDim;
  startRow= 0;
  endRow  = largestDim;
  *********************************/
  
  if(!lowRes && (radialMapShade || saveRadialMapShade)) {/* Do for Highres */
    /* topSec *= 2.0; */     /* In case we have to rotate it */
    xCen = radialXcen;
    yCen = radialYcen;
    zCen = radialZcen; 
  }
  
  for(row = startRow;row<endRow;row+=step) {/*! process data row by row*/
    ySq = (row-yCen) * (row-yCen);
    
    for(col = startCol;col <endCol;col+=step) {/* for col */
      tmp1 = 0.0;  /*	! initialize tmp1 */
      
      xSq = (col-xCen) * (col-xCen);
      
      qSq = (xSq+ySq);
      
      for(secn = (topSec-1);secn>=(bottomSec);secn--) {/* for secn */
	
	/* CHeck to see if we're in radius */
	zSq = (secn-zCen) * (secn-zCen);
	mapRadiusSq = qSq+zSq;
	cutFract = 0.0;
	if(mapRadiusSq < minRadSq || mapRadiusSq > maxRadSq) {/* OK to use this */
	  double sqrtVal;
	  // October 2005 RWA
	  // I added this concept of cutFract to help alleviate
	  // some of the pixel striping we were seeing when we
	  // cut the MAP radially.  Not sure this is really kosher
	  // but it looks like it helps ???
	  sqrtVal = sqrt(mapRadiusSq);
	  diffit = sqrtVal - maxRadius;
	  if(diffit < 1.0 && diffit > 0.0)
	    cutFract = -99.0;
	  diffit = minRadius - sqrtVal;
	  if(diffit < 1.0 && diffit > 0.0)
	    cutFract = 99.0;
	  if(sqrtVal < minRadius && sqrtVal <=1.0)
	    cutFract = 99.00;
	  if(cutFract==0.0)
	    continue;
	}
	
	/* Init it */
	dens1= surf_dens - 1.0;
	
	newTranslateThetaPhiOmega(&thisCol,&thisRow,&thisSecn,
				  xCen,yCen,zCen,
				  col,row,secn,
				  t1,t2,t3,t4,t5,t6,t7,t8,t9);
	
	thisCol -= colOffset;
	thisRow -= rowOffset;
	
	/* October 21, 1998 which worked OK but for first slice
	 * in a cutX,cutY,cutZ scenario 
	 if( thisSecn >= sec1 && thisSecn < sec2M1 &&
	 thisRow  >= row1 && thisRow  < row2M1 &&
	 thisCol  >= col1 && thisCol  < col2M1 )
	***********************************/
	if(thisSecn >= sec1 && thisSecn <= sec2M1 &&
	   thisRow  >= row1 && thisRow  <= row2M1 &&
	   thisCol  >= col1 && thisCol  <= col2M1 ) {
	  dens1 = getVoxel(coremap,thisCol,thisRow,thisSecn,
			   theta, phi, omega, ncol, nrow);
	  dens1 *= thisIrev;  
	}
	else
	  dens1 = surf_dens - 1.0;
	
	if(dens1 >= surf_dens) {/*then     ! reached threshold density level */
	  
	  /* Init it */
	  dens2= surf_dens - 1.0;
	  
	  newTranslateThetaPhiOmega(&thisCol,&thisRow,&thisSecn,
				    xCen,yCen,zCen,
				    col,row,(secn+1),
				    t1,t2,t3,t4,t5,t6,t7,t8,t9);
	  
	  /* Center this thing */
	  thisCol -= colOffset;
	  thisRow -= rowOffset;
	  
	  /* We need to be able to see pixels outside the cut 
	   * zones so do a wide search here */
	  
	  if(thisSecn >= 0.0 && thisSecn < origNumSecsM1 &&
	     thisRow  >= 0.0 && thisRow  < nrowM1 &&
	     thisCol  >= 0.0 && thisCol  < ncolM1 ) {/* OK to proceed */
	    
	    dens2 = getVoxel(coremap,thisCol,thisRow,thisSecn,
			     theta, phi, omega,
			     ncol, nrow);
	    dens2 *= thisIrev;
	    
	    /*******************************************
	     * This is it!!!!!  If you're going from the
	     * inside out then reverse the dens2 
	     * concept!!
	     ******************************************/
	    if(((secn < zCen) && 
		(mapRadiusSq <= minRadSq*1.1)) ||
	       (mapRadiusSq >= maxRadSq*0.95))
	      dens2 *= -1.0;
	    
	    dens_diff = dens1 - dens2;
	    
	    tmp1      = topSec-secn;
	    
	    /*
*****************************************
*    trap for dens_diff close to 0.0    *
*****************************************
*/
	    if(dens_diff < 0.0)
	      absDensDiff = 0.0 - dens_diff;
	    else 
	      absDensDiff = dens_diff;
	    
	    if(absDensDiff < 0.0000001) {/*then*/
	      /* Do nothing tmp1 = tmp1; */
	    }/*then*/
	    else {/* else A*/
	      double numer,denom;
	      numer = surf_dens-dens2;
	      denom = dens_diff;
	      // Why not this ?? denom = absDensDiff;
	      
	      if(denom != 0.0)
		tmp_fract = numer / denom;
	      else tmp_fract = 0.0;
	      
	      // Why do we eliminate the fraction if
	      // the densities are switched ??
	      if(dens1 <= dens2)
		tmp_fract = 0.0;
	      
	      // Get the fractional distance when on a cutting edge
	      if(theta != 0.0 || phi != 0.0 || omega != 0.0) {
		if(thisCol <= col1)
		  tmp_fract = col1 - thisCol;
		if(thisCol >= (col2-1))
		  tmp_fract = thisCol - (col2-1);
		if(thisRow < row1)
		  tmp_fract = row1 - thisRow;
		if(thisRow >= (row2-1))
		  tmp_fract = thisRow - (row2-1);
		if(thisSecn < sec1)
		  tmp_fract = thisSecn - sec1;
		if(thisSecn >= (sec2-1)) {
		  // Try this
		  tmp1      = topSec-thisSecn;
		  tmp_fract = thisSecn - (sec2-1);
		}
	      }
	      
	      if((cutFract == -99.0)) { // Do the MAX cut thing
		double newzSq, newTmp1;
		newzSq = maxRadSq - qSq;
		if(newzSq > 0.0)
		  newTmp1 = sqrt(newzSq)+zCen;
		else
		  newTmp1 = zCen - sqrt(fabs(newzSq));
		tmp1      = topSec - newTmp1;
	      }
	      else if((cutFract == 99.0)) {// Do the MIN cut thing
		double newzSq, newTmp1;
		newzSq = minRadSq - qSq;
		if(newzSq > 0.0)
		  newTmp1 = sqrt(newzSq)+zCen;
		else
		  newTmp1 = zCen - sqrt(fabs(newzSq));
		tmp1      = newTmp1;
	      }
	      else { // normal
		if(tmp_fract > 0.0) 
		  tmp1 = tmp1 + tmp_fract;
	      }
	    }/* else A*/
	    
	    if(radialMapShade) {/* COmpute the radius here */
	      if(!lowRes) {/* Do for Highres */
		double qSq;
		
		xSq = (col-radialXcen) * (col-radialXcen);
		ySq = (row-radialYcen) * (row-radialYcen);
		
		qSq = xSq + ySq;
		
		zSq = (secn-radialZcen) * (secn-radialZcen);
		mapRadius = sqrt(qSq+zSq);
		
		/* Why do we have to halve it ??? */
		/* mapRadius *= 0.5;   */
		
		tmp1= (topRadialSec - mapRadius) ;
		/* If the center pixel is zero then
		 * it looks just like background -
		 * let's push it up above backgrd */
		if(tmp1>=topRadialSec)
		  tmp1 = topRadialSec-1.0; 
		
		topSecMtmp1 = topRadialSec - tmp1;
		topSecMtmp1 = mapRadius; 
	      }
	      else {/* Normal */
		mapRadius = sqrt(mapRadiusSq);
		tmp1= (topSec - mapRadius) ;
		/* If the center pixel is zero then
		 * it looks just like background -
		 * let's push it up above backgrd */
		if(tmp1==topSec)
		  tmp1 = topSec-1.0; 
		
	      }
	    }/* COmpute the radius here */
	    
	    if(lowRes || !radialMapShade) {/* OK to do these */
	      if(tmp1 > topSec)
		tmp1= topSec;
	      
	      topSecMtmp1 = topSec - tmp1;
	    }
	    
	    dphshd_map[(row*largestDim)+col] = topSecMtmp1;
	    shad_map[(row*largestDim)+col] = topSecMtmp1;
	    
	    if(useForMinMax) {/* If radial we may not want to include
			       * the depth-cued one here */
	      
	      if(topSecMtmp1 != 0.0) {/* DOn't include bckgrnd in min/max */
		if(topSecMtmp1 < map_min)
		  map_min = topSecMtmp1; /*globmin */
		if(topSecMtmp1 > map_max)
		  map_max = topSecMtmp1; /*globmax */
	      }
	      
	    }/* If radial we may not want to include
	      * the depth-cued one here */
	    
	    for(r=0;r<step;r++) {/* propagate this for step */
	      for(s=0;s<step;s++) {/* Do all rows */
		dphshd_map[(row+s)*largestDim+col+r] = 
		  shad_map[(row+s)*largestDim+col+r] = 
		  topSecMtmp1;
	      }/* Do all rows */
	    }/* propagate this for step */
	    
	    /* Found the surface we're looking for -
	     * goto next column */
	    goto nextColumn;
	    
	  }/* OK to proceed */
	  
	  
	}/*then     ! reached threshold density level */
	
      }/* for secn *//* ! step to next map section */
      
      
    nextColumn:;
    }/* for col ! step to next map column*/
    
    
  }/*! process data row by row*//*! step to next map row*/  
  
} // end function thetaPhi_dphshd_C



void icosRadialProject(float *coremap, 
		       int ncol, int nrow, int ndepth, 
		       float *buf,
		       float x_org, float y_org, float z_org, 
		       float radius, float smallestDens,
		       float theta, float phi, float omega,
		       int colOffset, int rowOffset,
		       int thisIrev, int tag, float thresh, float slid,
		       int destroy, float minRadius)
{
  // function name: icosRadialProject
  // Do an icosahedral radial projection of the map.

  int rad_sec, dummyl, col, row, secn;
  int intCol, intRow;
  double rad, rad_sqr, min_rad_sqr, xx, yy, zz, sum;
  double sec_factor;
  double dens1, dens2, dens_diff1, dens_diff2;
  double thisSecn, thisRow, thisCol;
  double xCen, yCen, zCen;
  double thisRadius, xdyRow, xdyCol, xdySec;
  double cphi, sphi, ctheta, stheta, comega, somega;
  double t1, t2, t3, t4, t5, t6, t7, t8, t9;
  double calcRadius, thisZorg;;
  
  // Stuff globals
  threshold = thresh;
  slide = slid;
  
  // Make sure we're displaying density correctly
  thisIrev = 0.0 - thisIrev;
  
  //  We are calling this every time for every radius - stepping done
  //  Out side of this routine.
  
  // We've got a bug whereby we've got to add 90 degress to
  // our omega values

  // Avoiding exact cancellations in floating point operations
  theta += 0.0001234;
  phi   += 0.0001234;
  omega += 0.0001234;

  comega = cos(deg2rad(0.0-phi));
  somega = sin(deg2rad(0.0-phi));
  cphi   = cos(deg2rad(omega));
  sphi   = sin(deg2rad(omega));
  ctheta = cos(deg2rad(0.0-theta));
  stheta = sin(deg2rad(0.0-theta));
  
  // COMPUTE ELEMENTS OF ROTATION MATRIX
  t1 = ctheta*cphi*comega + sphi*somega;
  t2 = ctheta*sphi*comega - cphi*somega;
  t3 = -stheta*comega;
  t4 = ctheta*cphi*somega - sphi*comega;
  t5 = ctheta*sphi*somega + cphi*comega;
  t6 = -stheta*somega;
  t7 = stheta*cphi;
  t8 = stheta*sphi;
  t9 = ctheta;
  
  xCen = (ncol+1)/2   - 1.0;
  yCen = (nrow+1)/2   - 1.0;
  zCen = (ndepth+1)/2 - 1.0;
  
  map_min = 0.0;
  map_max = 0.0;
  
//  printf("radius[%f]  minRadius[%f]\n",radius,minRadius);
  
  rad = radius;
  rad_sqr = rad*rad;
  min_rad_sqr = minRadius*minRadius;
  
  if(destroy)
    thisZorg = ndepth;
  else 
    thisZorg = z_org;
  
  for (row=0; row<nrow; row++) {
    xx = (row - x_org) * (row - x_org);
    intRow = row + rowOffset;
    for(col=0; col<ncol; col++) {
      intCol = col + colOffset;
      buf[intCol+(intRow*ncol)] = 0.0;  // INITIALIZE BUF(COL,row)
      yy  = (col-y_org) * (col-y_org);
      sum = xx + yy;

      // I think we need to have sum < rad_sqr here to exclude 
      // looking at anything inside the radius NO - we need to look 
      // at all pixels!!!
      // if (sum > rad_sqr) 
      //    goto nextCol;

      // Now here  I'm thinking we need to iterate between z_org and 0 or
      // z_org and numSecs - e.g. Just look at half of the pixels.
      
      for(secn = 0; secn < (int)thisZorg; secn ++) {
	// dist_sum = sqrt(rad_sqr - sum);
	// dist     = z_org + dist_sum;
	// rad_sec  = (int)dist;

	zz = (secn - zCen)*(secn - zCen);
	
	//  Not sure how to setup sec_factor
	//     sec_factor = dist - ((float)rad_sec);
	//     if (rad_sec > ndepth) 
	//     rad_sec = ndepth;
	//     if (rad_sec < 1) 
	//     rad_sec = 1;

	sec_factor = 0.0;
	
	//  For some reason the row and col got messed in coremap???
	//  DENS1 = COREMAP(COL,ROW,RAD_SEC)
	//  DENS1 = COREMAP(rad_sec,col,row)
	dens1   = 0.0; 
	rad_sec = secn; 
	
	// Get dens1 based on the rotation of phi, omega, theta
	newTranslateThetaPhiOmega(&thisCol, &thisRow, &thisSecn,
				  xCen, yCen, zCen,
				  col, row, rad_sec,
				  t1, t2, t3, t4, t5, t6, t7, t8, t9);
	
	// Call XDY's rtn here with row,col,secn (or whatever we're 
	// iterating on for z)
	
	thisRadius = radius;
	xdyRow = thisCol - x_org;
	xdyCol = ndepth - thisSecn  - z_org;
	xdySec = nrow - thisRow - y_org;
	if(icosshell(xdyCol, xdyRow, xdySec, thisRadius, tag)) {
	  // YES this is on the icosahedral surface
	  dens1 = getVoxel(coremap, thisCol, thisRow, thisSecn,
			   theta, phi, omega, ncol, nrow);	  
	  dens1 *= thisIrev;
	  
	  if (rad_sec >= ndepth) {
	    buf[intCol+(intRow*ncol)] = dens1 - smallestDens;
	  }
	  else {
	    dens2 = 0.0; 
	    rad_sec = secn + 1;
	    
	    // Get dens2 based on the rotation of phi, omega, theta
	    newTranslateThetaPhiOmega(&thisCol, &thisRow, &thisSecn,
				      xCen, yCen, zCen,
				      col, row, rad_sec,
				      t1, t2, t3, t4, t5, t6, t7, t8, t9);
	    
	    dens2 = getVoxel(coremap, thisCol, thisRow, thisSecn,
			     theta, phi, omega, ncol, nrow);
	    dens2 *= thisIrev;
	    
	    dens_diff1 = dens1*(1.0-sec_factor) ;
	    dens_diff2 = dens2*sec_factor;
	    buf[intCol+(intRow*ncol)] = dens_diff1 + dens_diff2 - 
	      smallestDens;
	  }
	  
	  // GLOBAL MIN
	  map_min = MIN(map_min, buf[intCol+(intRow*ncol)]);  
	  map_max = MAX(map_max, buf[intCol+(intRow*ncol)]);  
	  goto nextCol;  // Go find another column
	}
	else if(destroy) { // Candidate for destroying
	  calcRadius = zz * sum;
	  if(calcRadius < min_rad_sqr || calcRadius > rad_sqr) // kill it
	    coremap[(secn*nrow*ncol)+(intRow*ncol)+intCol] = 0.0;
	}
	
      }
      
    nextCol:
      dummyl = 0;  // Do nothing
    }
    
  }
  
} // end function icosRadialProject



void rad_dens_fast_rob_C(float *coremap, 
			 int ncol, int nrow, int ndepth, 
			 float *buf,
			 float x_org, float y_org, float z_org, 
			 float radius, float smallestDens,
			 float theta, float phi, float omega,
			 int colOffset, int rowOffset, int largestDim,
			 int thisIrev)
{
  // function name: rad_dens_fast_rob_C
  // Do a spherical radial projection of the map.

  int rad_sec, col, row, secn, intCol, intRow;
  double rad, rad_sqr, xx, yy, sum;
  double sec_factor, dist;
  double dens1, dens2, dens_diff1, dens_diff2;
  double dist_sum;
  double thisSecn, thisRow, thisCol;
  double xCen, yCen, zCen;
  double cphi, sphi, ctheta, stheta, comega, somega;
  double t1, t2, t3, t4, t5, t6, t7, t8, t9;

  // Make sure we're displaying density correctly
  thisIrev = 0.0 - thisIrev;
  
  // We are calling this every time for every radius - stepping done
  // Out side of this routine.
  
  // We've got a bug whereby we've got to add 90 degress to
  // our omega values

  // Avoiding exact cancellations in floating point operations
  theta += 0.0001234;
  phi   += 0.0001234;
  omega += 0.0001234;

  comega = cos(deg2rad(0.0-phi));
  somega = sin(deg2rad(0.0-phi));
  cphi   = cos(deg2rad(omega));
  sphi   = sin(deg2rad(omega));
  ctheta = cos(deg2rad(0.0-theta));
  stheta = sin(deg2rad(0.0-theta));
  
  // COMPUTE ELEMENTS OF ROTATION MATRIX
  t1 = ctheta*cphi*comega + sphi*somega;
  t2 = ctheta*sphi*comega - cphi*somega;
  t3 = -stheta*comega;
  t4 = ctheta*cphi*somega - sphi*comega;
  t5 = ctheta*sphi*somega + cphi*comega;
  t6 = -stheta*somega;
  t7 = stheta*cphi;
  t8 = stheta*sphi;
  t9 = ctheta;
  
  xCen = (ncol+1)/2   - 1.0;
  yCen = (nrow+1)/2   - 1.0;
  zCen = (ndepth+1)/2 - 1.0;
  
  map_min = 0.0;
  map_max = 0.0;
  
  rad = radius;
  rad_sqr = rad*rad;

  for(row=0; row<nrow; row++) {
    intRow = row + rowOffset;
    xx = (row-x_org)*(row-x_org);
    for(col=0; col<ncol; col++) {
      intCol = col + colOffset;
      buf[intCol+(intRow*largestDim)] = 0.0;
      yy  = (col-y_org)*(col-y_org);
      sum = xx + yy;
      if (sum > rad_sqr) 
	continue;
      dist_sum = sqrt(rad_sqr - sum);
      dist     = z_org + dist_sum;
      rad_sec  = (int)dist;
      sec_factor = dist - rad_sec;
      if (rad_sec >= ndepth) 
	rad_sec = ndepth-1;
      if (rad_sec < 1) 
	rad_sec = 1;
      
      // For some reason the row and col got messed in coremap???
      // DENS1 = COREMAP(COL,ROW,RAD_SEC)
      // DENS1 = COREMAP(rad_sec,col,row)
      
      dens1 = 0.0; 
      secn = rad_sec;
      
      // Get dens1 based on the rotation of phi, omega, theta
      newTranslateThetaPhiOmega(&thisCol, &thisRow, &thisSecn,
				xCen, yCen, zCen,
				col, row, secn,
				t1, t2, t3, t4, t5, t6, t7, t8, t9);
      
      dens1 = getVoxel(coremap, thisCol, thisRow, thisSecn,
		       theta, phi, omega, ncol, nrow);
      dens1 *= thisIrev;
      
      if (rad_sec >= ndepth-1) {
	buf[intCol+(intRow*ncol)] = dens1 - smallestDens;
      }
      else { // DO other
	dens2 = 0.0; 
	
	secn = rad_sec + 1;
	
	// Get dens2 based on the rotation of phi, omega, theta
	newTranslateThetaPhiOmega(&thisCol, &thisRow, &thisSecn,
				  xCen, yCen, zCen,
				  col, row, secn,
				  t1, t2, t3, t4, t5, t6, t7, t8, t9);
	
	dens2 = getVoxel(coremap, thisCol, thisRow, thisSecn,
			 theta, phi, omega, ncol, nrow);
	dens2 *= thisIrev;
	
	dens_diff1 = dens1*(1.0-sec_factor) ;
	dens_diff2 = dens2*sec_factor;
	buf[intCol+(intRow*largestDim)] = (dens_diff1 + dens_diff2) - 
	  smallestDens;
      }
      
      // Global min and max
      map_min = MIN(map_min, buf[intCol+(intRow*largestDim)]);
      map_max = MAX(map_max, buf[intCol+(intRow*largestDim)]);
      
    }
    
  }
  
} // end function rad_dens_fast_rob_C



void dph_rescale_rob(int ncol, int nrow, int num_views,
		     cue_t *cuedMap, int tryColors, 
		     int *radialStepSize)
{
  // function name: dph_rescale_rob
  // Rescale all frames to consistent level (Ported from Fortran to C
  // by Rob Ashmore, 1996.

  int col, row, view;
  int thisRow, tvPtr, i;
  float *dphshd_map;
  float thisMin, thisMax;
  float scaleFactor;
  
  scaleFactor = ((float)tryColors-7) / (map_max-map_min);
  
  if(radialMapShade) { // Compute the stepSize
    
    *radialStepSize = map_max - map_min;
    *radialStepSize = map_max;
    // Minus 7 because we want to make sure we see everything
    // in the colormap
    scaleFactor = ((float)tryColors-7) / (map_max-map_min);
    scaleFactor /= (float)expStep;
    
    // Now let's compute what radius will be what colormap entry
    for (i=0; i<256; i++)
      radialValue[i]= 0.0;
    for(i=0; i<(tryColors-5); i++)
      radialValue[7+i]=  ((float)i/(float)expStep) / scaleFactor + map_min;
  }
  
  thisMin = HUGE;
  thisMax = -HUGE;
  for (view = 0; view<num_views; view++) {
      
    // Always scale the shaded one - if this is unshaded the .shad
    // buffer contains the cued data and it's the one being displayed
    
    if(radialMapShade && secondMapShade)
      dphshd_map = cuedMap[view].radial;
    else 
      dphshd_map = cuedMap[view].shad;
      
    for(row=0; row<nrow; row++) {
      thisRow = row*ncol;
      tvPtr = thisRow;
      for(col=0; col<ncol; col++) {
	if(dphshd_map[tvPtr+col] != 0.0) { // Then let's scale it
	  if(radialMapShade) {
	    // Do it in INT to make "holes" in colormap that
	    // we can exploit to put the shading into
	    dphshd_map[tvPtr+col] = 
	      (float)((int)((dphshd_map[tvPtr+col]-map_min)*scaleFactor)
		      * expStep);
	    
	    *radialStepSize = MIN(*radialStepSize,(int)dphshd_map[tvPtr+col]);
	    // We just need this to be the expStep don't we???
	    *radialStepSize = expStep;
	  }
	  else { // depth cued
	    dphshd_map[tvPtr+col] = 
	      (dphshd_map[tvPtr+col] - map_min) * scaleFactor;
	  }
	}
	
	if(dphshd_map[tvPtr+col]  > thisMax)
	  thisMax = dphshd_map[tvPtr+col];
	if(dphshd_map[tvPtr+col]  < thisMin)
	  thisMin = dphshd_map[tvPtr+col];
	
      }
    }
  }
  
  map_max = thisMax;
  map_min = thisMin;
  
} // end function dph_rescale_rob



void map_1d_rob(float *map, int ncol, int nrow, int nsec, 
		float xcen, float ycen, float zcen,
		int interp, int nterp, float max_r, 
		float *rint, int *irnum, float *sum_rint,
		float *sumsq_rint, float *stdev)
{
  // function name: map_1d_rob
  //
  // Compute spherically averaged density
  //
  // SEPTEMBER 1988  PURDUE UNIVERSITY VAX 8550, T. BAKER.
  // MODIFIED 29-MAR-1991 BY T. BAKER.
  // Converted to Xwindows, November 1993, by Eric Eltzroth
  // Modified 22 June 1995, by D. Belnap

  int    inum, ir, kTncolTnrow, jTncol;
  int    i, j, j1, j2, k;
  float  ysq, zsq;
  float  test, rsq_lim, fraction;
  float  sum_sq_over_n, numerator, denominator;
  float  thisFloat;
  
  inum    = nterp;
  rsq_lim = max_r*max_r;
  
  init_intarr(irnum, inum, 0);
  init_floatarr(rint, inum, 0.0);
  init_floatarr(sum_rint, inum, 0.0);
  init_floatarr(sumsq_rint, inum, 0.0);
  
  // Step through image and reduce to 1d circular average
  for (k=0; k<nsec; k++) {
    zsq   = (k-zcen)*(k-zcen);
    kTncolTnrow= k*ncol*nrow;
    for (j=0; j<nrow; j++) {
      ysq   = (j-ycen)*(j-ycen);
      jTncol = j * ncol;
      for (i=0; i<ncol; i++) {
	test = (i-xcen)*(i-xcen) + ysq + zsq;
	if(test <= rsq_lim) {
	  ir             = interp*sqrt(test) ;
	  irnum[ir]      = irnum[ir] + 1;
	  thisFloat      = map[kTncolTnrow+jTncol+i];
	  rint[ir]       = rint[ir] + thisFloat;
	  sumsq_rint[ir] = sumsq_rint[ir] + thisFloat*thisFloat;
	}
      }
    }
  }
  
  for(i=0;i<inum;i++) {
    if(irnum[i] > 1) {
      sum_rint[i] = rint[i];	      /*! for st. dev. calculation*/
      rint[i]     = rint[i]/irnum[i];
    }
  }
  
  // Now try to fill in the blanks
  for(i=1;i<(inum-2);i++) {
    /* I'm thinking we just want to go less than zero - e.g. the corners
     * are at zero so don't iunterpolate those ... ???
     if(irnum[i]<=0) 
    */
    if(irnum[i]<0) {
      j1 = i;
      while(irnum[j1]==0 && j1 > 0)	/* ! find last point */
	j1 = j1 - 1;
      j2 = i;
      while(irnum[j2]==0 && j2<inum)	/* ! find next point */
	j2 = j2 + 1;
      fraction = (float)(j2-i)/(float)(j2-j1);
      rint[i]  = rint[j1]*fraction + rint[j2]*(1.0-fraction);
    }
  }
  
  // Compute standard deviation at each radius where irnum > 1
  for(i=0;i<inum;i++) {
    if (irnum[i] > 1) {
      sum_sq_over_n = (sum_rint[i]*sum_rint[i])/irnum[i];
      numerator     = sumsq_rint[i] - sum_sq_over_n;
      denominator   = irnum[i] - 1;
      stdev[i]      = sqrt(numerator/denominator);
    }
    else
      stdev[i] = 0.0;
  }
  
} // end function map_1d_rob



void img_1d_rob(float *img,
		int ncol, int nrow,
		float xcen, float ycen,
		int interp, int nterp,
		float *rint, int *irnum,
		float img_bkg)
{
  // function name: img_1d_rob
  // Compute circularly-averaged density

  int i, j, inum, ir;
  float x, y, ysq;
  
  inum = nterp*2;
  
  init_floatarr(rint, inum+2, 0.0); // initialize arrays
  init_intarr(irnum, inum+2, 0);
  
  // step through image and reduce to 1d circular average
  for (j=0; j<nrow; j++) {
    y = j - ycen;
    ysq = y*y;
    for (i=0; i<ncol; i++) {
      x  = i - xcen;
      ir = interp*sqrt(x*x+ysq);
      irnum[ir] = irnum[ir] + 1;
      rint[ir]  = rint[ir] + img[(i*ncol)+j] - img_bkg;
    }
  }
  
  // normalize the 1d array
  for(i=0; i<inum; i++)
    if (irnum[i] > 1)  
      rint[i] = rint[i]/((float)irnum[i]);
  
  // interpolate density where irnum = 0, subtract background, output
  for(i=0; i<inum; i++) {
    if (irnum[i]==0 && i==0) 
      rint[0] = rint[1];
    if (irnum[i]==0 && (i!=0 && i<inum)) 
      rint[i] = (rint[i-1] + rint[i+1])/2;
  }
  
} // end function img_1d_rob



void ricardo_routine(int phaseGroupSymmetry, float **float_image_buffer,
                     int *cdNumCols, int *cdNumRows,
                     float origX, float origY)
{
  // function name: ricardo_routine
  // Program to impose plane group symmetry on an image motif to
  // create a whole unit cell.
  // April 30,1999, Ricardo Bernal

  float *temp_image_buffer, *newimage;
  int newCols, newRows;
  int i,j,k;                           /* loop control variables */
  float a11,a12,a13,a21,a22,a23;
  float b11,b12,b13,b21,b22,b23;
  float c11,c12,c13,c21,c22,c23;
  float d11,d12,d13,d21,d22,d23;
  float e11,e12,e13,e21,e22,e23;
  float f11,f12,f13,f21,f22,f23;   /* a-k are variables for       */
  float g11,g12,g13,g21,g22,g23;   /* matrix symmetry operations  */
  float h11,h12,h13,h21,h22,h23;   /* 13 & 23 are the translation */
  float i11,i12,i13,i21,i22,i23;
  float j11,j12,j13,j21,j22,j23;
  float k11,k12,k13,k21,k22,k23;
  
  int oldx,oldy,newx,newy;   /* coordinate variables           */
  int bgx;                   /* variables for background value */
  int bgy;
  float *thisBuffer;
  
  if(origX == 0.0)
    origX = 1.0;
  if(origY == 0.0)
    origY = 1.0;
  
  if(phaseGroupSymmetry != 1) {
    /** If its P1 then just fall out of routine **/
    newRows = *cdNumRows * 2;
    newCols = *cdNumCols * 2;
    if(phaseGroupSymmetry == 3) {
      newCols = *cdNumCols;
    }
    
    newimage=(float *)malloc(sizeof(float)*(newCols)*(newRows));
    
    /*********  Initialize the new array to make even background **********/
    
    bgx = *cdNumRows;
    bgy = *cdNumCols;
    thisBuffer = *float_image_buffer;
    
    for(i=*cdNumRows;i > (*cdNumRows-(*cdNumRows/20));i--) {
      for(j=*cdNumCols;j > (*cdNumCols-(*cdNumCols/20));j--) {
	if(thisBuffer[((i*(*cdNumCols))+j)] > 
	   thisBuffer[((bgx*(*cdNumCols))+bgy)]) {
	  bgx = i;
	  bgy = j;
	}
      }
    }
    for(i=0;i < newRows;i++) {
      for(j=0;j < newCols;j++) {
	newimage[(i*(newCols))+j]=thisBuffer[(bgx*(*cdNumCols))+bgy];
      }
    }
    
    /***************** Fill the new array with the old image **********/
    
    for(i=0; i < newRows; i++)
      for(j=0; j < newCols; j++)
	if(i < *cdNumRows && j < *cdNumCols)
	  newimage[(i*(newCols))+j]=thisBuffer[(i*(*cdNumCols))+j];
    
    /*************** Generate symmetry related motifs ***********/
    
    if(phaseGroupSymmetry==2) {/* p2 */
      a11 = -1*origX;   // the convention is positive x down (with rows)
      a12 =  0*origX;   // and positive y to the right (with columns)
      a21 =  0*origY;   //  a11  a12
      a22 = -1*origY;   //  a21  a22
      
      for (oldx=1; oldx <= (*cdNumRows); oldx++) {
	for (oldy=1; oldy <= (*cdNumCols); oldy++) {
	  newx = oldx*a11 + oldy*a21;
	  newy = oldx*a12 + oldy*a22;
	  if(newx < 0)
	    newx = (int)(newx + newRows*origX);
	  if(newy < 0)
	    newy = (int)(newy + newCols*origY);
	  
	  if(newimage[(((newx-1)*(newCols))+(newy-1))] > 
	     thisBuffer[((bgx*(*cdNumCols))+bgy)])
	    newimage[((newx-1)*(newCols))+(newy-1)] += 
	      (thisBuffer[(((oldx-1)*(*cdNumCols))+(oldy-1))]);
	  if(newimage[((newx-1)*(newCols))+(newy-1)] > 255)
	    newimage[((newx-1)*(newCols))+(newy-1)] = 255;
	  if(newimage[((newx-1)*(newCols))+(newy-1)] < 0)
	    newimage[((newx-1)*(newCols))+(newy-1)] = 0;
	  else if(newimage[(((newx-1)*(newCols))+(newy-1))] == 
		  thisBuffer[((bgx*(*cdNumCols))+bgy)])
	    newimage[(((newx-1)*(newCols))+(newy-1))] +=  
	      (thisBuffer[(((oldx-1)*(*cdNumCols))+(oldy-1))])-
	      thisBuffer[((bgx*(*cdNumCols))+bgy)];
	  if(newimage[((newx-1)*(newCols))+(newy-1)] > 255)
	    newimage[((newx-1)*(newCols))+(newy-1)] = 255;
	  if(newimage[((newx-1)*(newCols))+(newy-1)] < 0)
	    newimage[((newx-1)*(newCols))+(newy-1)] = 0;
	}
      }
    }
    else if(phaseGroupSymmetry==3) { /* pm */
      a11 = -1;      
      a12 =  0;      
      a21 =  0;
      a22 =  1;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*a11 + oldy*a21;
	  newy = oldx*a12 + oldy*a22;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
    }
    else if(phaseGroupSymmetry==4)  { /* pg */
      a11 = -1;      
      a12 =  0;      
      a13 =  0;
      a21 =  0;
      a22 =  1;
      a23 =  *cdNumCols;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*a11 + oldy*a21 + a13;
	  newy = oldx*a12 + oldy*a22 + a23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
    }
    else if(phaseGroupSymmetry==5) { /* cm */
      a11 = -1;      
      a12 =  0;      
      a13 =  0;
      a21 =  0;
      a22 =  1;
      a23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*a11 + oldy*a21 + a13;
	  newy = oldx*a12 + oldy*a22 + a23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      b11 =  1;
      b12 =  0;
      b13 =  *cdNumRows;
      b21 =  0;
      b22 =  1;
      b23 =  *cdNumCols;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*b11 + oldy*b21 + b13;
	  newy = oldx*b12 + oldy*b22 + b23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      c11 = -1;
      c12 =  0;
      c13 =  *cdNumRows;
      c21 =  0;
      c22 =  1;
      c23 =  *cdNumCols;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*c11 + oldy*c21 + c13;
	  newy = oldx*c12 + oldy*c22 + c23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
    }
    else if(phaseGroupSymmetry==6)  { /* p2mm */
      a11 = -1;      
      a12 =  0;      
      a13 =  0;
      a21 =  0;
      a22 = -1;
      a23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*a11 + oldy*a21 + a13;
	  newy = oldx*a12 + oldy*a22 + a23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      b11 = -1;
      b12 =  0;
      b13 =  0;
      b21 =  0;
      b22 =  1;
      b23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*b11 + oldy*b21 + b13;
	  newy = oldx*b12 + oldy*b22 + b23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      c11 =  1;
      c12 =  0;
      c13 =  0;
      c21 =  0;
      c22 = -1;
      c23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*c11 + oldy*c21 + c13;
	  newy = oldx*c12 + oldy*c22 + c23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
    }
    else if(phaseGroupSymmetry==7) { /* p2mg */
      a11 = -1;      
      a12 =  0;      
      a13 =  0;
      a21 =  0;
      a22 = -1;
      a23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*a11 + oldy*a21 + a13;
	  newy = oldx*a12 + oldy*a22 + a23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      b11 = -1;
      b12 =  0;
      b13 =  *cdNumRows;
      b21 =  0;
      b22 =  1;
      b23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*b11 + oldy*b21 + b13;
	  newy = oldx*b12 + oldy*b22 + b23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      c11 =  1;
      c12 =  0;
      c13 =  *cdNumRows;
      c21 =  0;
      c22 = -1;
      c23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*c11 + oldy*c21 + c13;
	  newy = oldx*c12 + oldy*c22 + c23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
    }
    else if(phaseGroupSymmetry==8) { /* p2gg */
      a11 = -1;      
      a12 =  0;      
      a13 =  0;
      a21 =  0;
      a22 = -1;
      a23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*a11 + oldy*a21 + a13;
	  newy = oldx*a12 + oldy*a22 + a23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      b11 = -1;
      b12 =  0;
      b13 =  *cdNumRows;
      b21 =  0;
      b22 =  1;
      b23 =  *cdNumCols;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*b11 + oldy*b21 + b13;
	  newy = oldx*b12 + oldy*b22 + b23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      c11 =  1;
      c12 =  0;
      c13 =  *cdNumRows;
      c21 =  0;
      c22 = -1;
      c23 =  *cdNumCols;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*c11 + oldy*c21 + c13;
	  newy = oldx*c12 + oldy*c22 + c23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
    }
    else if(phaseGroupSymmetry==9) { /* c2mm */
      a11 = -1;      
      a12 =  0;      
      a13 =  0;
      a21 =  0;
      a22 = -1;
      a23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*a11 + oldy*a21 + a13;
	  newy = oldx*a12 + oldy*a22 + a23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      b11 = -1;
      b12 =  0;
      b13 =  0;
      b21 =  0;
      b22 =  1;
      b23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*b11 + oldy*b21 + b13;
	  newy = oldx*b12 + oldy*b22 + b23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      c11 =  1;
      c12 =  0;
      c13 =  0;
      c21 =  0;
      c22 = -1;
      c23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*c11 + oldy*c21 + c13;
	  newy = oldx*c12 + oldy*c22 + c23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      d11 =  1;      
      d12 =  0;      
      d13 =  *cdNumRows;
      d21 =  0;
      d22 =  1;
      d23 =  *cdNumCols;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = oldx*d11 + oldy*d21 + d13;
	  newy = oldx*d12 + oldy*d22 + d23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      e11 = -1;
      e12 =  0;
      e13 =  *cdNumRows;
      e21 =  0;
      e22 = -1;
      e23 =  *cdNumCols;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*e11) + (oldy*e21)) + e13;
	  newy = ((oldx*e12) + (oldy*e22)) + e23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      f11 = -1;
      f12 =  0;
      f13 =  *cdNumRows;
      f21 =  0;
      f22 =  1;
      f23 =  *cdNumCols;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*f11) + (oldy*f21)) + f13;
	  newy = ((oldx*f12) + (oldy*f22)) + f23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      g11 =  1;
      g12 =  0;
      g13 =  *cdNumRows;
      g21 =  0;
      g22 = -1;
      g23 =  *cdNumCols;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*g11) + (oldy*g21)) + g13;
	  newy = ((oldx*g12) + (oldy*g22)) + g23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
    }
    else if(phaseGroupSymmetry==10) { /* p4 */
      a11 = -1;      
      a12 =  0;      
      a13 =  0;
      a21 =  0;
      a22 = -1;
      a23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*a11) + (oldy*a21)) + a13;
	  newy = ((oldx*a12) + (oldy*a22)) + a23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      b11 =  0;
      b12 =  1;
      b13 =  0;
      b21 = -1;
      b22 =  0;
      b23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*b11) + (oldy*b21)) + b13;
	  newy = ((oldx*b12) + (oldy*b22)) + b23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      c11 =  0;
      c12 = -1;
      c13 =  0;
      c21 =  1;
      c22 =  0;
      c23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*c11) + (oldy*c21)) + c13;
	  newy = ((oldx*c12) + (oldy*c22)) + c23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
    }
    else if(phaseGroupSymmetry==11) { /* p4mm */
      a11 = -1;     
      a12 =  0;      
      a13 =  0;
      a21 =  0;
      a22 = -1;
      a23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*a11) + (oldy*a21)) + a13;
	  newy = ((oldx*a12) + (oldy*a22)) + a23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      b11 =  0;
      b12 =  1;
      b13 =  0;
      b21 = -1;
      b22 =  0;
      b23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*b11) + (oldy*b21)) + b13;
	  newy = ((oldx*b12) + (oldy*b22)) + b23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      c11 =  0;
      c12 = -1;
      c13 =  0;
      c21 =  1;
      c22 =  0;
      c23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*c11) + (oldy*c21)) + c13;
	  newy = ((oldx*c12) + (oldy*c22)) + c23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      d11 = -1;      
      d12 =  0;      
      d13 =  0;
      d21 =  0;
      d22 =  1;
      d23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*d11) + (oldy*d21)) + d13;
	  newy = ((oldx*d12) + (oldy*d22)) + d23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      e11 =  1;
      e12 =  0;
      e13 =  0;
      e21 =  0;
      e22 = -1;
      e23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*e11) + (oldy*e21)) + e13;
	  newy = ((oldx*e12) + (oldy*e22)) + e23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      f11 =  0;
      f12 =  1;
      f13 =  0;
      f21 =  1;
      f22 =  0;
      f23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*f11) + (oldy*f21)) + f13;
	  newy = ((oldx*f12) + (oldy*f22)) + f23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      g11 =  0;
      g12 = -1;
      g13 =  0;
      g21 = -1;
      g22 =  0;
      g23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*g11) + (oldy*g21)) + g13;
	  newy = ((oldx*g12) + (oldy*g22)) + g23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
    }
    else if(phaseGroupSymmetry==12) { /* p4gm */
      a11 = -1;      
      a12 =  0;      
      a13 =  0;
      a21 =  0;
      a22 = -1;
      a23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*a11) + (oldy*a21)) + a13;
	  newy = ((oldx*a12) + (oldy*a22)) + a23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      b11 =  0;
      b12 =  1;
      b13 =  0;
      b21 = -1;
      b22 =  0;
      b23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*b11) + (oldy*b21)) + b13;
	  newy = ((oldx*b12) + (oldy*b22)) + b23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      c11 =  0;
      c12 = -1;
      c13 =  0;
      c21 =  1;
      c22 =  0;
      c23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*c11) + (oldy*c21)) + c13;
	  newy = ((oldx*c12) + (oldy*c22)) + c23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      d11 = -1;      
      d12 =  0;      
      d13 =  *cdNumRows;
      d21 =  0;
      d22 =  1;
      d23 =  *cdNumCols;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*d11) + (oldy*d21)) + d13;
	  newy = ((oldx*d12) + (oldy*d22)) + d23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      e11 =  1;
      e12 =  0;
      e13 =  *cdNumRows;
      e21 =  0;
      e22 = -1;
      e23 =  *cdNumCols;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*e11) + (oldy*e21)) + e13;
	  newy = ((oldx*e12) + (oldy*e22)) + e23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      f11 =  0;
      f12 =  1;
      f13 =  *cdNumRows;
      f21 =  1;
      f22 =  0;
      f23 =  *cdNumCols;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*f11) + (oldy*f21)) + f13;
	  newy = ((oldx*f12) + (oldy*f22)) + f23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      g11 =  0;
      g12 = -1;
      g13 =  *cdNumRows;
      g21 = -1;
      g22 =  0;
      g23 =  *cdNumCols;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*g11) + (oldy*g21)) + g13;
	  newy = ((oldx*g12) + (oldy*g22)) + g23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
    }
    else if(phaseGroupSymmetry==13) { /* p3 */
      a11 =  0;      
      a12 =  1;      
      a13 =  0;
      a21 = -1;
      a22 =  0;
      a23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*a11) + (oldy*a21)) + a13;
	  newy = (oldx - oldy) + a23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      b11 =  0;
      b12 = -1;
      b13 =  0;
      b21 = -1;
      b22 =  0;
      b23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = (-oldx + oldy) + b13;
	  newy = ((oldx*b12) + (oldy*b22)) + b23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
    }
    else if(phaseGroupSymmetry==14) { /* p3m1 */
      a11 =  0;      
      a12 =  1;      
      a13 =  0;
      a21 = -1;
      a22 =  0;
      a23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*a11) + (oldy*a21)) + a13;
	  newy = (oldx - oldy) + a23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      b11 =  0;
      b12 = -1;
      b13 =  0;
      b21 = -1;
      b22 =  0;
      b23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = (-oldx + oldy) + b13;
	  newy = ((oldx*b12) + (oldy*b22)) + b23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      c11 =  0;
      c12 = -1;
      c13 =  0;
      c21 = -1;
      c22 =  0;
      c23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*c11) + (oldy*c21)) + c13;
	  newy = ((oldx*c12) + (oldy*c22)) + c23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      d11 =  1;
      d12 =  0;
      d13 =  0;
      d21 =  0;
      d22 =  1;
      d23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = (-oldx + oldy) + d13;
	  newy = ((oldx*d12) + (oldy*d22)) + d23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      e11 =  1;
      e12 =  0;
      e13 =  0;
      e21 =  0;
      e22 =  1;
      e23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*e11) + (oldy*e21)) + e13;
	  newy = (oldx-oldy) + e23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
    }
    else if(phaseGroupSymmetry==15) { /* p31m */
      a11 =  0;      
      a12 =  1;      
      a13 =  0;
      a21 = -1;
      a22 =  0;
      a23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*a11) + (oldy*a21)) + a13;
	  newy = (oldx - oldy) + a23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      b11 =  0;
      b12 = -1;
      b13 =  0;
      b21 = -1;
      b22 =  0;
      b23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = (-oldx + oldy) + b13;
	  newy = ((oldx*b12) + (oldy*b22)) + b23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      c11 =  0;
      c12 =  1;
      c13 =  0;
      c21 =  1;
      c22 =  0;
      c23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*c11) + (oldy*c21)) + c13;
	  newy = ((oldx*c12) + (oldy*c22)) + c23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      d11 =  1;
      d12 =  0;
      d13 =  0;
      d21 =  0;
      d22 = -1;
      d23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = (oldx - oldy) + d13;
	  newy = ((oldx*d12) + (oldy*d22)) + d23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      e11 = -1;
      e12 =  0;
      e13 =  0;
      e21 =  0;
      e22 =  1;
      e23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*e11) + (oldy*e21)) + e13;
	  newy = (-oldx + oldy) + e23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
    }
    else if(phaseGroupSymmetry==16) { /* p6 */
      a11 =  0;      
      a12 =  1;      
      a13 =  0;
      a21 = -1;
      a22 =  0;
      a23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*a11) + (oldy*a21)) + a13;
	  newy = (oldx - oldy) + a23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      b11 =  0;
      b12 = -1;
      b13 =  0;
      b21 = -1;
      b22 =  0;
      b23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = (-oldx + oldy) + b13;
	  newy = ((oldx*b12) + (oldy*b22)) + b23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      c11 = -1;
      c12 =  0;
      c13 =  0;
      c21 =  0;
      c22 = -1;
      c23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*c11) + (oldy*c21)) + c13;
	  newy = ((oldx*c12) + (oldy*c22)) + c23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      d11 =  0;
      d12 =  1;
      d13 =  0;
      d21 =  1;
      d22 =  0;
      d23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*d11) + (oldy*d21)) + d13;
	  newy = (-oldx + oldy) + d23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      e11 =  0;
      e12 =  1;
      e13 =  0;
      e21 =  1;
      e22 =  0;
      e23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = (oldx - oldy) + e13;
	  newy = ((oldx*e12) + (oldy*e22)) + e23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
    }
    else if(phaseGroupSymmetry==17) { /* p6mm */
      a11 =  0;      
      a12 =  1;      
      a13 =  0;
      a21 = -1;
      a22 =  0;
      a23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*a11) + (oldy*a21)) + a13;
	  newy = (oldx - oldy) + a23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      b11 =  0;
      b12 = -1;
      b13 =  0;
      b21 = -1;
      b22 =  0;
      b23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = (-oldx + oldy) + b13;
	  newy = ((oldx*b12) + (oldy*b22)) + b23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      c11 = -1;
      c12 =  0;
      c13 =  0;
      c21 =  0;
      c22 = -1;
      c23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*c11) + (oldy*c21)) + c13;
	  newy = ((oldx*c12) + (oldy*c22)) + c23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      d11 =  0;
      d12 =  1;
      d13 =  0;
      d21 =  1;
      d22 =  0;
      d23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*d11) + (oldy*d21)) + d13;
	  newy = (-oldx + oldy) + d23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      e11 =  0;
      e12 =  1;
      e13 =  0;
      e21 =  1;
      e22 =  0;
      e23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = (oldx - oldy) + e13;
	  newy = ((oldx*e12) + (oldy*e22)) + e23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      f11 =  0;
      f12 = -1;
      f13 =  0;
      f21 = -1;
      f22 =  0;
      f23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*f11) + (oldy*f21)) + f13;
	  newy = ((oldx*f12) + (oldy*f22)) + f23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      g11 =  1;
      g12 =  0;
      g13 =  0;
      g21 =  0;
      g22 =  1;
      g23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = (-oldx + oldy) + g13;
	  newy = ((oldx*g12) + (oldy*g22)) + g23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      h11 =  1;
      h12 =  0;
      h13 =  0;
      h21 =  0;
      h22 =  1;
      h23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*h11) + (oldy*h21)) + h13;
	  newy = (oldx-oldy) + h23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      i11 =  0;
      i12 =  1;
      i13 =  0;
      i21 =  1;
      i22 =  0;
      i23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*i11) + (oldy*i21)) + i13;
	  newy = ((oldx*i12) + (oldy*i22)) + i23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      j11 =  1;
      j12 =  0;
      j13 =  0;
      j21 =  0;
      j22 = -1;
      j23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = (oldx - oldy) + j13;
	  newy = ((oldx*j12) + (oldy*j22)) + j23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
      k11 = -1;
      k12 =  0;
      k13 =  0;
      k21 =  0;
      k22 =  1;
      k23 =  0;
      
      for(oldx=1;oldx <= (*cdNumRows);oldx++) {
	for(oldy=1;oldy <= (*cdNumCols);oldy++) {
	  newx = ((oldx*k11) + (oldy*k21)) + k13;
	  newy = (-oldx + oldy) + k23;
	  if(newx < 0) newx = newx + newRows;
	  if(newy < 0) newy = newy + newCols;
	  i = (newx-1)*newCols + (newy-1);
	  j = bgx*(*cdNumCols) + bgy;
	  k = (oldx-1)*(*cdNumCols) + (oldy-1);
	  frr(&newimage[i], &thisBuffer[j], &thisBuffer[k]);
	}
      }
    }
    
    // ----- Transfer new values back -----
    free(*float_image_buffer);
    *float_image_buffer=(float *)malloc(sizeof(float)*newCols*newRows);
    
    temp_image_buffer=*float_image_buffer;
    for(i=0;i < newRows;i++)
      for(j=0;j < newRows;j++)
	temp_image_buffer[(i*newCols)+j]=newimage[(i*newCols)+j];
    
    free(newimage);
    *cdNumRows = newRows;
    *cdNumCols = newCols;
    *float_image_buffer = temp_image_buffer;
  }

}  // end function ricardo_routine



void frr(float *x, float *y, float *z) 
{
  // function name: frr
  // Utility routine used by ricardo_routine()

  if(*x > *y)
    *x += *z;
  else if(*x == *y)
    *x += *z - *y;

} // end function frr
