#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fitutil.h"
#define DEBUG1 0


// ---------- Code from Numerical Recipes starts here ----------


void covsrt(double **covar, int ma, int ia[], int mfit)
{
  // function name: covsrt
  //
  // Expand in storage the covariance matrix covar, so as to take into
  // account parameters that are being held fixed.  program: Numerical
  // Recipe in C 2nd edition, p675. Modified by Wei Zhang, change
  // float into double

  int i, j, k;
  double temp;
  
  for (i=mfit;i<ma;i++)
    for (j=0;j<=i;j++) 
      covar[i][j]=covar[j][i]=0.0;
  
  k=mfit;
  for (j=ma-1;j>=0;j--) {
    if (ia[j]) {
      k--;
      for (i=0;i<ma;i++) SWAP(covar[i][k], covar[i][j])
      for (i=0;i<ma;i++) SWAP(covar[k][i], covar[j][i])
    }
  }

} // end function covsrt



void gaussj(double **a, int n, double **b, int m)
{
  // function name: gausssj
  // 
  // Linear equation solution by Gauss-Jordan elimination,
  // a[1..n][1..n] is the input matrix. b[1..n][1..m] is input
  // containing the m right-hand side vectors. On output, a is
  // replaced by its matrix inverse, and b is replaced by the
  // corresponding set of solution vectors.  program: Numerical Recipe
  // in C 2nd edition, p39 modified by Wei Zhang, change float into
  // double

  int *indxc, *indxr, *ipiv;
  int i, icol, irow, j, k, l, ll;
  double big, dum, pivinv, temp;

  irow = icol = 0;
  indxc=(int *) malloc (n * sizeof(int));
  indxr=(int *) malloc (n * sizeof(int));
  ipiv =(int *) malloc (n * sizeof(int));
  for (j=0;j<n;j++) ipiv[j]=0;
  for (i=0;i<n;i++) {
    big=0.0;
    for (j=0;j<n;j++)
      if (ipiv[j] != 1) 
        for (k=0;k<n;k++) {
          if (ipiv[k]==0) {
            if (fabs(a[j][k]) >=big) {
              big=fabs(a[j][k]);
              irow=j;
              icol=k;
            }
          } else if (ipiv[k]>1)
	    {
	      fprintf(stderr, "yji_gaussj: Singular Matrix-1\n");
	      exit(1);
	    };
        }
    ++(ipiv[icol]);
    
    if (irow != icol) {
      for (l=0;l<n;l++) SWAP(a[irow][l], a[icol][l]);
      for (l=0;l<m;l++) SWAP(b[irow][l], b[icol][l]);
    }
    indxr[i]=irow;
    indxc[i]=icol;
    
    if (a[icol][icol] == 0.0) {
      fprintf(stderr, "yji_gaussj: Singular Matrix-2\n");
      exit(1);
    };
    
    pivinv=1.0/a[icol][icol];
    a[icol][icol]=1.0;
    for (l=0;l<n;l++) a[icol][l] *= pivinv;
    for (l=0;l<m;l++) b[icol][l] *= pivinv;
    for (ll=0;ll<n;ll++)
      if (ll != icol) {
        dum=a[ll][icol];
        a[ll][icol]=0.0;
        for (l=0;l<n;l++) a[ll][l] -= a[icol][l]*dum;
        for (l=0;l<m;l++) b[ll][l] -= b[icol][l]*dum;
      }
  }
  
  for (l=n-1;l>=0;l--) {
    if (indxr[l] != indxc[l])
      for (k=0;k<n;k++)
        SWAP(a[k][indxr[l]], a[k][indxc[l]]);
  }
  
  free(ipiv);
  free(indxr);
  free(indxc);

} // end function gaussj


// ---------- Code for 1D CTF fit starts here ----------

// ca1d_AVG   : array of avg value at certain fft radius
// ca1d_STD   : array of std value at certain fft radius
// ndata      : size of the choosen fft radius part
// poly_paras : initial parameters for Background , Decay,
//              defocus, also used for return values
// poly_order : size of poly_paras
// poly_orderb: size of parameters for Background
// poly_orderd: size of parameters for Decay
// wavelength : electron wavelength (units = x)
// hkl_max    : size of have fft size, ndata <= hkl_max
// A          : Percentage of Amplitude Contrast
// Cs         : Spherical aberration constant
// pixel_size : Pixel size (units = x)
// chisq      : return value
//
// Fortran call sequence
// 
//      call cmpt_fit(ca1d_AVG(annlow),ca1d_std(annlow), ndata,annhigh,
//	+              poly_paras, poly_order, poly_orderb, poly_orderd,
//	+              wavelength, hkl_max, A, Cs, pixel_size, chisq)


void cmpt_fit1d__(double ca1d_avg[], double ca1d_std[], int * ndata,
		int *annlow, int * annhigh,
		double poly_paras[], int * poly_order, 
		int * poly_orderb, int * poly_orderd,
		double * wavelength, int * hkl_max, double * A, double * Cs, 
		double * pixel_size, double * chisq)
{
  cmpt_fit1d(ca1d_avg, ca1d_std, ndata, annlow, annhigh,
	      poly_paras, poly_order, 
	      poly_orderb, 
	      poly_orderd, wavelength, hkl_max, A, Cs, pixel_size, chisq);
}

void cmpt_fit1d_(double ca1d_avg[], double ca1d_std[], int * ndata,
		int *annlow, int * annhigh,
		double poly_paras[], int * poly_order, 
		int * poly_orderb, int * poly_orderd,
		double * wavelength, int * hkl_max, double * A, double * Cs, 
		double * pixel_size, double * chisq)
{
  cmpt_fit1d(ca1d_avg, ca1d_std, ndata, annlow, annhigh,
	      poly_paras, poly_order, 
	      poly_orderb, 
	      poly_orderd, wavelength, hkl_max, A, Cs, pixel_size, chisq);
}



void cmpt_fit1d(double ca1d_avg[], double ca1d_std[], int * ndata, 
		 int *annlow, int *annhigh,
		 double poly_paras[], int * poly_order, int * poly_orderb, 
		 int * poly_orderd,
		 double * wavelength, int * hkl_max, double * A, double * Cs, 
		 double * pixel_size, double * chisq)
{
  // function name: cmpt_fit1d
  // Main function for CTF parameter fitting

  double *ca1d_id, **covar, **alpha;
  double big = 100000000.0, small = 0.000000001;
  double chisqold1, tmp5, chisqold2, defcusold;
  double alamda, condnum; 
  int    *ia, i, j, ma, masq, nmfit;
  
  ma = *ndata;
  ca1d_id = (double *) malloc (ma * sizeof(double ));
  for(i=0; i<ma; i++) ca1d_id[i] = i + (*annhigh) - ma + 1;
  
  ma  = *poly_order;
  ia  = (int *) malloc (ma * sizeof(int ));
  for(i=0; i<ma; i++) ia[i] = 1;
  
  covar = (double **) malloc (ma * sizeof(double ));
  alpha = (double **) malloc (ma * sizeof(double ));
  masq = ma * ma;
  covar[0] = (double *) malloc (masq * sizeof(double ));
  for (i=1; i<ma; i++) covar[i] = covar[i-1] + ma;
  alpha[0] = (double *) malloc (masq * sizeof(double ));
  for (i=1; i<ma; i++) alpha[i] = alpha[i-1] + ma;
  
  /**********************************************************             
   * In this iteration session, all parameters are adjusted *
   **********************************************************/
  alamda = -1.0;
  for (;;) {
    mrqmin_1d (ca1d_id, ca1d_avg, ca1d_std, *ndata, 
	       poly_paras, ia, *poly_order,
	       covar, alpha, chisq, fpoly_ctf_1d, &alamda,
	       *poly_orderb, *poly_orderd, *wavelength, *hkl_max,
	       *A, *Cs, *pixel_size, &nmfit, *annlow, *annhigh);
    if(alamda>big || alamda<small) break;
  }
  alamda = 0.0;
  mrqmin_1d (ca1d_id, ca1d_avg, ca1d_std, *ndata, poly_paras, ia, 
	     *poly_order,
	     covar, alpha, chisq, fpoly_ctf_1d, &alamda,
	     *poly_orderb, *poly_orderd, *wavelength, *hkl_max,
	     *A, *Cs, *pixel_size, &nmfit, *annlow, *annhigh);

  condnum = condnumcal_1d(covar, alpha, nmfit);
  printf("\nthe number of parameters for matrix covar is: %d\n", nmfit);
  printf("The condition number for matrix covar is: %e\n", condnum);
  
  
  for (i=0; i<*poly_orderb; i++)
    printf("\na[%d]= %e",i, poly_paras[i]);
  for (i=*poly_orderb; i<*poly_orderb+*poly_orderd; i++)
    printf("\nb[%d]= %e",i, poly_paras[i]);
  printf ("\ndefocus: %e \n",  poly_paras[*poly_order-1]);
  printf ("chisq: %e \n\n",  *chisq);
  
  /******************************************************************
   * In this iteration session, parameters for B, D and defocus value
   * are adjusted seperated *
   ******************************************************************/
  chisqold1 = *chisq;
  ma = *poly_order;

  for(;;) {

    // ***** Refine defocus value
    tmp5 = poly_paras[ma-1];
    chisqold2=*chisq;
    defcusold=poly_paras[ma-1];
    for (i=0; i<ma-1; i++) 
      ia[i]=0;
    ia[ma-1]=1;
    
    for (j=-20; j<=20;j++) {
      alamda = -1.0;
      poly_paras[ma-1] = tmp5+j*10.;
      for(;;) {
	mrqmin_1d (ca1d_id, ca1d_avg, ca1d_std, *ndata, poly_paras, 
		   ia, *poly_order,
		   covar, alpha, chisq, fpoly_ctf_1d, &alamda,
		   *poly_orderb, *poly_orderd, *wavelength, *hkl_max,
		   *A, *Cs, *pixel_size, &nmfit, *annlow, *annhigh);
	if(alamda>big || alamda<small) break;
      }
      alamda =0.0;
      mrqmin_1d (ca1d_id, ca1d_avg, ca1d_std, *ndata, poly_paras, 
		 ia, *poly_order,
		 covar, alpha, chisq, fpoly_ctf_1d, &alamda,
		 *poly_orderb, *poly_orderd, *wavelength, *hkl_max,
		 *A, *Cs, *pixel_size,&nmfit, *annlow, *annhigh);
      
      if(*chisq<=chisqold2) {
	chisqold2 = *chisq;
	defcusold = poly_paras[ma-1];
      }
    }
    poly_paras[ma-1]=defcusold;

    // ***** Refine back ground curve
    alamda =-1.0;
    for (i=0; i<*poly_orderb; i++)
      ia[i]=1;
    for (i=*poly_orderb; i<*poly_orderb + *poly_orderd; i++)
      ia[i]=0;
    ia[ma-1]=0;
    for(;;) {
      mrqmin_1d (ca1d_id, ca1d_avg, ca1d_std, *ndata, poly_paras, 
		 ia, *poly_order,
		 covar, alpha, chisq, fpoly_ctf_1d, &alamda,
		 *poly_orderb, *poly_orderd, *wavelength, *hkl_max,
		 *A, *Cs, *pixel_size, &nmfit, *annlow, *annhigh);
      if(alamda>big || alamda<small) break;
    }
    
    alamda =0.0;
    mrqmin_1d (ca1d_id, ca1d_avg, ca1d_std, *ndata, poly_paras, 
	       ia, *poly_order,
	       covar, alpha, chisq, fpoly_ctf_1d, &alamda,
	       *poly_orderb, *poly_orderd, *wavelength, *hkl_max,
	       *A, *Cs, *pixel_size, &nmfit, *annlow, *annhigh);
    
    // **** Refine signal strength decay curve
    alamda =-1.0;
    for (i=0; i<*poly_orderb; i++)
      ia[i]=0;
    for (i=*poly_orderb; i<*poly_orderb + *poly_orderd; i++)
      ia[i]=1;
    ia[ma-1]=0;
    for(;;) {
      mrqmin_1d (ca1d_id, ca1d_avg, ca1d_std, *ndata, poly_paras, 
		 ia, *poly_order,
		 covar, alpha, chisq, fpoly_ctf_1d, &alamda,
		 *poly_orderb, *poly_orderd, *wavelength, *hkl_max,
		 *A, *Cs, *pixel_size, &nmfit, *annlow, *annhigh);
      if(alamda>big || alamda<small) break;
    }
    
    alamda =0.0;
    mrqmin_1d (ca1d_id, ca1d_avg, ca1d_std, *ndata, poly_paras, 
	       ia, *poly_order,
	       covar, alpha, chisq, fpoly_ctf_1d, &alamda,
	       *poly_orderb, *poly_orderd, *wavelength, *hkl_max,
	       *A, *Cs, *pixel_size, &nmfit, *annlow, *annhigh);
      
    if((*chisq-chisqold1)<small) break;
    chisqold1 = *chisq;
  }
  
  for (i=0; i<*poly_orderb; i++)
    printf("\na[%d]= %e",i, poly_paras[i]);
  for (i=*poly_orderb; i<*poly_orderb+*poly_orderd; i++)
    printf("\nb[%d]= %e",i, poly_paras[i]);

  printf("\ndefocus: %e \n",  poly_paras[*poly_order-1]);
  printf("chisq: %e \n\n",  *chisq);
  printf("wavelength: %f\n",*wavelength);
  printf("Cs: %f\n",*Cs);
  printf("A: %f\n",*A);

  free(ca1d_id);
  free(covar[0]);
  free(covar);
  free(alpha[0]);
  free(alpha);
  free(ia);

} // end function cmpt_fit1d



double condnumcal_1d(double **matrix, double **matrix2, int n)
{  
  // function condnumcal_1d
  // Compute a condition number of a matrix. Equal to the
  // product of the norm of the matrix and the norm of its inverse

  double norm, norminv, condnum;
  int j, k;

  for (j=0; j<n; j++) 
    for (k=0; k<n; k++) 
      matrix2[j][k]=matrix[j][k];

  norm = normcal_1d(matrix, n);
  invmatrix_1d(matrix, matrix2, n);
  norminv = normcal_1d(matrix2, n);
  condnum = norm * norminv;
  return condnum;

} // end function condnumcal_1d



double normcal_1d (double **matrix, int n)
{
  // function name: normcal_1d
  // Compute the norm of a matrix

  double tmp, norm=-1.0;
  int j, k;

  for (j=0; j<n; j++) {
    tmp = 0.0;
    for (k=0; k<n; k++)
      tmp += abs(matrix[j][k]);
    if(tmp > norm) norm = tmp;
  }
  norm = sqrt(norm);	
  return norm;

} // end function normcal_1d



void fpoly_ctf_1d(double x, double *a, double *y, double *dyda, 
		  int na, int order_poly1, int order_poly2, 
		  double wavelength, int idim_half, double A, 
		  double Cs, double pixel_size,
		  int annlow, int annhigh) 
{
  // function name: fpoly_ctf_1d
  
  int i;
  double  arg, arg2, arg3, arg4, arg5, y1;
  double  fw, fw2, theta, defocus, ctf, ctf2;
  
  defocus = a[na-1];

  fw    = wavelength*x/(idim_half*2.0* pixel_size);
  fw2   = fw*fw;
  theta = M_PI*2.0*(Cs*fw2*fw2/4.0-defocus*fw2/2.0)/wavelength;
  ctf   = sqrt(1.0-A*A)*sin(theta)-A*cos(theta);
  ctf2  = ctf*ctf;

  arg  = 2.0*(x-annlow)/(annhigh-annlow) - 1.0;
  arg2 = arg*arg;
  arg3 = arg2*arg;
  arg4 = arg3*arg;
  arg5 = arg4*arg;

  dyda[0]  = 1;
  dyda[1]  = arg;
  dyda[2]  = (3.0*arg2 - 1.0)/2.0;
  dyda[3]  = (5.0*arg3 - 3.0*arg)/2.0;
  dyda[4]  = (35.0*arg4 - 30.0*arg2 + 3.0)/8.0;
  dyda[5]  = (63.0*arg5 - 70.0*arg3 + 15.0*arg)/8.0;

  dyda[6]  = ctf2*dyda[0];
  dyda[7]  = ctf2*dyda[1];
  dyda[8]  = ctf2*dyda[2];
  dyda[9]  = ctf2*dyda[3];
  dyda[10] = ctf2*dyda[4];
  dyda[11] = ctf2*dyda[5];

  *y = 0.0;
  for(i=0; i<12; i++)
    *y += a[i]*dyda[i];

  y1 = 0.0;
  for(i=6; i<12; i++)
    y1 += a[i]*dyda[i]/ctf2;
  
  dyda[12] = -2.0*ctf*y1*(sqrt(1.0-A*A)*cos(theta)+A*sin(theta))*
    M_PI*fw2/wavelength;
  
} // end function fpoly_ctf_1d



void nrerror_1d (char error_text[])
     /* Numerical Recipes standard error handler */
{
  fprintf(stderr, "Numerical Recipes run-time error...\n");
  fprintf(stderr, "%s\n",error_text);
  fprintf(stderr, "...now exiting to system...\n");
  exit(1);
}



/* Numerical recipes in C Page 48
   Modified by Wei Zhang, May 24, 2005*/

void invmatrix_1d(double **a, double **y, int n)
{	
  double d,*col;
  int i,j,*indx;
  indx=(int *) malloc (n * sizeof(int));
  col =(double *) malloc (n*sizeof(double));
  ludcmp_1d(a,n,indx,&d); /*Decompose the matrix just once.*/
  for(j=0;j<n;j++) { /*Find inverse by columns.*/
    for(i=0;i<n;i++) col[i]=0.0;
    col[j]=1.0;
    lubksb_1d(a,n,indx,col);
    for(i=0;i<n;i++) y[i][j]=col[i];
  }
  free(indx);
  free(col);
}

void lubksb_1d(double **a, int n, int *indx, double b[])
     /*Solves the set of n linear equations A·X = B. Here
a[1..n][1..n] is input, not as the matrix A but rather as its LU
decomposition, determined by the routine ludcmp. indx[1..n] is input
as the permutation vector returned by ludcmp. b[1..n] is input as the
right-hand side vector B, and returns with the solution vector X. a,
n, and indx are not modified by this routine and can be left in place
for successive calls with different right-hand sides b. This routine
takes into account the possibility that b will begin with many zero
elements, so it is efficient for use in matrix inversion.  Numerical
Recipes in C. P47 modified by Wei Zhang, May-24-2005, use double and
change index from (1,2,...n) to (0,1,...,n-1)*/
{
  int i,ii=-1,ip,j;
  double sum;
  for (i=0;i<n;i++) { /*When ii is set to a positive value, it will become the
			index of the first nonvanishing element of b. We now do the forward substitution, equation (2.3.6). The
			only new wrinkle is to unscramble the permutation as we go.*/
    ip=indx[i];
    sum=b[ip];
    b[ip]=b[i];
    if (ii>=0)
      for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
    else if (sum) ii=i; /*A nonzero element was encountered, so from now on we
			  will have to do the sums in the loop above*/
    b[i]=sum; 
  }
  for (i=n-1;i>=0;i--) { /*Now we do the backsubstitution, equation (2.3.7).*/
    sum=b[i];
    for (j=i+1;j<n;j++) sum -= a[i][j]*b[j];
    b[i]=sum/a[i][i]; /*Store a component of the solution vector X.*/
  } /*All done!*/
}

void ludcmp_1d(double **a, int n, int *indx, double *d)
     /*Given a matrix a[1..n][1..n], this routine replaces it by the
       LU decomposition of a rowwise permutation of itself. a and n
       are input. a is output, arranged as in equation (2.3.14) above;
       indx[1..n] is an output vector that records the row permutation
       effected by the partial pivoting; d is output as ±1 depending
       on whether the number of row interchanges was even or odd,
       respectively. This routine is used in combination with lubksb
       to solve linear equations or invert a matrix.  Numerical
       recipes in C, P 46-47 modified by Wei Zhang, change memory
       allocation and change float into double*/
{
  int i,imax,j,k;
  double big,dum,sum,temp;
  double *vv; /*vv stores the implicit scaling of each row.*/
  vv=(double *) malloc (n * sizeof(double));
  *d=1.0;    /*No row interchanges yet.*/
  imax = 0;
  for (i=0;i<n;i++) { /*Loop over rows to get the implicit scaling information.*/
    big= 0.0; 
    for (j=0;j<n;j++)
      if ((temp=fabs(a[i][j])) > big) big=temp;
    if (big == 0.0) nrerror_1d("Singular matrix in routine ludcmp");
    /*No nonzero largest element.*/
    vv[i]=1.0/big; /*Save the scaling.*/
  }
  for (j=0;j<n;j++) { /*This is the loop over columns of Crout's method.*/
    for (i=0;i<j;i++) { /*This is equation (2.3.12) except for i = j.*/
      sum=a[i][j];
      for (k=0;k<i;k++) sum -= a[i][k]*a[k][j];
      a[i][j]=sum;
    }
    big=0.0; /*Initialize for the search for largest pivot element.*/
    for (i=j;i<n;i++) { /*This is i = j of equation (2.3.12) and i = j+1. . .N
			  of equation (2.3.13).*/
      sum=a[i][j]; 
      for (k=0;k<j;k++)
	sum -= a[i][k]*a[k][j];
      a[i][j]=sum;
      if ( (dum=vv[i]*fabs(sum)) >= big) {
	/*Is the figure of merit for the pivot better than the best so far?*/
	big=dum;
	imax=i;
      }
    }
    if (j != imax) { /*Do we need to interchange rows?*/
      for (k=0;k<n;k++) { /*Yes, do so...*/
	dum=a[imax][k];
	a[imax][k]=a[j][k];
	a[j][k]=dum;
      }
      *d = -(*d); /*...and change the parity of d.*/
      vv[imax]=vv[j]; /*Also interchange the scale factor.*/
    }
    indx[j]=imax;
    if (a[j][j] == 0.0) a[j][j]=TINY;
    /*If the pivot element is zero the matrix is singular (at least to
      the precision of the algorithm). For some applications on
      singular matrices, it is desirable to substitute TINY for
      zero.*/
    if (j != n) { /*Now, finally, divide by the pivot element.*/
      dum=1.0/(a[j][j]);
      for (i=j+1;i<n;i++) a[i][j] *= dum;
    }
  } /*Go back for the next column in the reduction.*/
  free(vv);
}


void mrqcof_1d(double x[], double y[], double sig[], int ndata, double a[], 
	       int ia[],  int ma, double **alpha, double beta[], double *chisq,
	       void (*funcs)(double, double [], double *, double [], int, 
			     int, int, double, 
			     int,  double, double, double, int , int),
	       int num_gauss_bkg, int num_gauss_ctf, double wavelength, 
	       int idim_half, double A,
	       double Cs, double pixel_size, int annlow, int annhigh)
     
     /* Used by mrqmin to evaluate the linearized fitting matrix
	alpha, and vector beta and calculate chi square
	
	program: Numerical Recipe in C, p686 modified by Wei Zhang,
	pass more parameters and change float into double*/ 
{
  int i, j, k, l, m, mfit;
  double ymod, wt, sig2i, dy, *dyda;
  
  dyda=(double *) malloc (ma * sizeof(double));
  mfit = 0;
  for (j=0;j<ma;j++)
    if(ia[j]) mfit++;
  for(j=0;j<mfit;j++) {
    for (k=0;k<=j;k++) alpha[j][k]=0.0;
    beta[j]=0.0;
  }
  *chisq=0.0;
  for (i=0;i<ndata; i++) { 
    (*funcs)(x[i],a,&ymod,dyda,ma, num_gauss_bkg, 
	     num_gauss_ctf, wavelength, idim_half,
	     A, Cs, pixel_size, annlow, annhigh);    
    sig2i=1.0/(sig[i]*sig[i]);    
    dy=y[i]-ymod;
    for (j=-1,l=0;l<ma;l++) {
      if (ia[l]) {
	wt=dyda[l]*sig2i;
	for (j++,k=0,m=0;m<=l;m++)
	  if(ia[m]) alpha[j][k++]+=wt*dyda[m];
	beta[j] += dy*wt;
      }
    }
    *chisq +=dy*dy*sig2i;
  }
  for (j=1;j<mfit;j++)
    for (k=0;k<j; k++) alpha[k][j]=alpha[j][k];
  free(dyda);
}

void mrqmin_1d(double x[], double y[], double sig[], int ndata, 
	       double a[], int ia[],
	       int ma, double **covar, double **alpha, double *chisq,
	       void (*funcs)(double, double [], double *, double [], 
			     int, int, int, double,
			     int, double, double, double, int, int ), 
	       double *alamda, int num_gauss_bkg, int num_gauss_ctf, 
	       double wavelength,
	       int idim_half, double A, double Cs, double pixel_size, 
	       int *nmfit, int annlow, int annhigh)
     /*program: Numerical Recipe in C, p687 modified by Wei Zhang,
       pass more paramerters and change float into double*/
{
  int j, k, l;
  static int mfit;
  static double ochisq, *atry, *beta, *da, **oneda;
  
  if (*alamda < 0.0) {
    atry=(double *) malloc (ma * sizeof(double ));
    beta=(double *) malloc (ma * sizeof(double ));
    da  =(double *) malloc (ma * sizeof(double ));
    
    for (mfit=0,j=0;j<ma;j++)
      if (ia[j]) mfit++;
    *nmfit = mfit;
    oneda=(double **) malloc ( mfit * sizeof(double ));
    oneda[0] = (double *) malloc (mfit* sizeof(double ));
    for (j=1; j<mfit; j++) oneda[j] = oneda[j-1] + 1 ;
    *alamda=0.001;      
    mrqcof_1d(x,y,sig,ndata,a,ia,ma,alpha,beta,chisq,funcs,num_gauss_bkg,
	      num_gauss_ctf,
	      wavelength, idim_half, A, Cs, pixel_size,annlow, annhigh);
    ochisq=(*chisq);
    for(j=0;j<ma;j++) atry[j]=a[j];
  }
  
  for (j=0;j<mfit;j++) {
    for (k=0;k<mfit;k++) covar[j][k]=alpha[j][k];
    covar[j][j]=alpha[j][j]*(1.0+(*alamda));
    oneda[j][0]=beta[j];
  }
  
  /* call gauss solution*/
  gaussj(covar, mfit, oneda, 1);
  
  for (j=0;j<mfit;j++) da[j]=oneda[j][0];
  if (*alamda == 0.0) {
    covsrt(covar, ma, ia, mfit);
    free(oneda[0]);
    free(oneda);
    free(da);
    free(atry);
    return;
  }
  for (j=0, l=0; l<ma; l++)
    if(ia[l]) atry[l]=a[l]+da[j++];
  
  mrqcof_1d(x,y,sig,ndata,atry,ia,ma,covar,da,chisq,funcs,
	    num_gauss_bkg,num_gauss_ctf,
	    wavelength, idim_half, A, Cs, pixel_size, annlow, annhigh);
  if(*chisq<ochisq) {
    *alamda *= 0.1;
    ochisq=(*chisq);
    for (j=0;j<mfit; j++) {
      for (k=0;k<mfit;k++) alpha[j][k]=covar[j][k];
      beta[j]=da[j];
    }
    for(l=0;l<ma;l++) a[l]=atry[l];
  } else {
    *alamda *= 10.0;
    *chisq=ochisq;
  }
  
}


// ---------- Code for 2D CTF fit starts here ----------

/* 
   ca2d_AVG: an array of avg value at certain fft radius
   ca2d_STD: an array of std value at certain fft radius
   poly_paras : initial parameters for Background, Decay, defocus, angle also used for return values
   poly_order : is the size of poly_paras
   poly_orderb: is the size of parameters for Background
   poly_orderd: is the size of parameters for Decay
   wavelength :
   res_hi     :
   hkl_max    : is the size of have fft size
   A          : Percentage of Amplitude Contrast
   Cs         : Spherical aberration constant
   pixel_size : 
   chisq      : return value
*/
/* define used for fortran call method */

void cmpt_fit__(double ca2d_avg[], double ca2d_std[], 
	      int * hkl_max, double * res_hi,
	      double poly_paras[], int * poly_order, 
	      int * poly_orderb, int * poly_orderd,
	      double * wavelength, double * A, double * Cs, 
	      double * pixel_size, double * chisq, 
	      float *auto2DdrawBuf, float *auto2DdataBuf)
{
     cmpt_fit(ca2d_avg, ca2d_std, hkl_max, res_hi, 
               poly_paras, poly_order, poly_orderb, poly_orderd,
               wavelength, A, Cs, pixel_size, chisq, 
	       auto2DdrawBuf, auto2DdataBuf);
}

void cmpt_fit_(double ca2d_avg[], double ca2d_std[], 
	      int * hkl_max, double * res_hi,
	      double poly_paras[], int * poly_order, 
	      int * poly_orderb, int * poly_orderd,
	      double * wavelength, double * A, double * Cs, 
	      double * pixel_size, double * chisq, 
	      float *auto2DdrawBuf, float *auto2DdataBuf)
{
     cmpt_fit(ca2d_avg, ca2d_std, hkl_max, res_hi, 
               poly_paras, poly_order, poly_orderb, poly_orderd,
               wavelength, A, Cs, pixel_size, chisq, 
	       auto2DdrawBuf, auto2DdataBuf);
}

/* real code part for ctf fiting */
/* note ca2d_avg ca2d_std are stored in Fortran format
   ca2d_avg(0:hkl_max, -hkl_max + hkl_max + 1)
*/
void cmpt_fit(double ca2d_avg[], double ca2d_std[], int * hkl_max, 
	       double * res_hi,
	       double poly_paras[], int * poly_order, int * poly_orderb, 
	       int * poly_orderd,
	       double * wavelength, double * A, double * Cs, 
	       double * pixel_size, double * chisq, 
	       float *auto2DdrawBuf, float *auto2DdataBuf)
{
  
  int    *ca2d_id;
  double **covar, **alpha;
  double big=100000000.0, small=0.000000001;
  double fw, tmp1, tmp2;
  double alamda, defmajor, defminor;
  double angle, theta1, deg_rad, rad_deg, chisqtmp, angtmp;
  int    *ia, i, j, k, ma, masq, lenx, leny;
  double annlow, annhigh, ratio, annlow_max, annlow_min;
  double annhigh_max, annhigh_min, radius, def, ann;
  int kk;  

  deg_rad = M_PI / 180.0;
  rad_deg = 180.0 / M_PI;
  
  i = *poly_order;
  defmajor = poly_paras[i-3];
  defminor = poly_paras[i-2];
  angle   = poly_paras[i-1] * deg_rad; /* convert to radian */
  poly_paras[i-1]  = angle; /* save back angle in radian instead of degree*/
  
  /* allocate memory for store whether wanted data or not, ca2d_id[y][x] */
  lenx = * hkl_max + 1;
  leny = 2* (* hkl_max) + 1;
  ca2d_id = (int *) malloc ( lenx * leny * sizeof(int ));
  
  /* compute the annulus of first peak and annuals high*/
  fw = sqrt((defmajor - sqrt(defmajor*defmajor - 
			     (*Cs)*(*wavelength)*2.0))/(*Cs));
  annlow_max = fw*((*pixel_size)*2*(*hkl_max))/(*wavelength);
  if (annlow_max < 1.0) annlow_max = 1.0;
  fw = sqrt((defminor - sqrt(defminor*defminor - 
			     (*Cs)*(*wavelength)*2.0))/(*Cs));
  annlow_min = fw*((*pixel_size)*2*(*hkl_max))/(*wavelength);
  if (annlow_min < 1.0) annlow_min = 1.0;
  
  if (annlow_min > annlow_max) {
    annhigh_min = 2*(*hkl_max)*(*pixel_size)/(*res_hi);
    if( annhigh_min > (*hkl_max)) annhigh_min = (*hkl_max);
    annhigh_max = annhigh_min * annlow_max / annlow_min;
    annlow = annlow_max;
    annhigh = annhigh_min;
  } else {
    annhigh_max = 2*(*hkl_max)*(*pixel_size)/(*res_hi);
    if( annhigh_max > (*hkl_max)) annhigh_max = (*hkl_max);
    annhigh_min = annhigh_max * annlow_min / annlow_max;
    annlow = annlow_min;
    annhigh = annhigh_max;
  }
  ratio = annhigh_max / annlow_max;
  
  /* printf("wavelength = %f, annlow_min = %f, annlow_max= %f, 
     annhigh_min= %f, annhigh_max= %f\n",
     *wavelength, annlow_min, annlow_max, annhigh_min, annhigh_max);*/
  
  /* find which points is located in the elipse of annlow and annhigh */
  for (j=0; j<leny; j++) {
    k = j-lenx+1;   /* Y axis from -hkl_max to hkl_max */
    for (i=0; i<lenx; i++) {
      ca2d_id[j*lenx + i] = 0;
      radius = sqrt(i*i + k*k);
      if ( radius >= annlow && radius <= annhigh) {
	if (i>0) theta1 = atan2(k,i);
	else {
	  if (k >=0) theta1 = M_PI/2;
	  else theta1 = -M_PI/2;
	}
	theta1 -= angle;
	tmp1    = cos(theta1);
	tmp2   = sin(theta1);
	def    = defmajor * tmp1 * tmp1 + defminor * tmp2 * tmp2;
	fw     = sqrt((def - sqrt(def*def - (*Cs)*(*wavelength)*2.0))/(*Cs));
	ann    = fw*((*pixel_size)*2*(*hkl_max))/(*wavelength);
	if (radius>=ann && radius<=ann*ratio)  ca2d_id[j*lenx + i] = 1;
      }
    }
  }
  
  ma  = *poly_order;
  ia  = (int *) malloc (ma * sizeof(int ));
  covar = (double **) malloc (ma * sizeof(double ));
  alpha = (double **) malloc (ma * sizeof(double ));
  masq = ma * ma;
  covar[0] = (double *) malloc (masq * sizeof(double ));
  for (i=1; i<ma; i++) covar[i] = covar[i-1] + ma;
  alpha[0] = (double *) malloc (masq * sizeof(double ));
  for (i=1; i<ma; i++) alpha[i] = alpha[i-1] + ma;
  
  /***********************************************************************
   * In this iteration session, all parameters except angle are adjusted *
   ***********************************************************************/
  for(i=0; i<ma-1; i++) ia[i] = 1;
  ia[ma-1] = 0;
  for (kk=0;kk<1; kk++){
    alamda = -1.0;
    for (;;) {
      yji_mrqmin (ca2d_id, ca2d_avg, ca2d_std, poly_paras, ia, *poly_order,
		  covar, alpha, chisq, yji_fpoly_ctf, &alamda,
		  *poly_orderb, *poly_orderd, *wavelength, *hkl_max, *res_hi,
		  *A, *Cs, *pixel_size, auto2DdrawBuf, auto2DdataBuf);
      if(alamda>big || alamda<small) break;
    }
    alamda = 0.0;
    yji_mrqmin (ca2d_id, ca2d_avg, ca2d_std, poly_paras, ia, *poly_order,
                covar, alpha, chisq, yji_fpoly_ctf, &alamda,
                *poly_orderb, *poly_orderd, *wavelength, *hkl_max, *res_hi,
                *A, *Cs, *pixel_size, auto2DdrawBuf, auto2DdataBuf);
    
    /*  for (i=0; i<*poly_orderb; i++)
	printf("\na[%d]= %e",i, poly_paras[i]);
	for (i=*poly_orderb; i<*poly_orderb+*poly_orderd; i++)
	printf("\nb[%d]= %e",i, poly_paras[i]);*/
    printf ("\n\ndefocus major: %e \n",  poly_paras[*poly_order-3]);
    printf ("defocus minor: %e \n",  poly_paras[*poly_order-2]);
    printf ("angle: %e \n",  poly_paras[*poly_order-1]*rad_deg);
    printf ("chisq: %e \n\n",  *chisq);
  }
  /************************************************************             
   * In this iteration session, the best angle are determined *
   ************************************************************/
  for(i=0; i<ma-1; i++) ia[i] = 0;
  ia[ma-1]=1;
  /* Wei research all possibilities of angles */
  angle = poly_paras [*poly_order-1];
  angtmp = angle;
  chisqtmp = *chisq;
  for (i=0; i<90; i++) {
    poly_paras [*poly_order-1] = angle + deg_rad*(i-45.0);
    alamda = -1.0;
    for (;;) {
      yji_mrqmin (ca2d_id, ca2d_avg, ca2d_std, poly_paras, ia, *poly_order,
		  covar, alpha, chisq, yji_fpoly_ctf, &alamda,
		  *poly_orderb, *poly_orderd, *wavelength, *hkl_max, *res_hi,
		  *A, *Cs, *pixel_size, auto2DdrawBuf, auto2DdataBuf);
      if(alamda>big || alamda<small) break;
    }
    alamda = 0.0;
    yji_mrqmin (ca2d_id, ca2d_avg, ca2d_std, poly_paras, ia, *poly_order,
                covar, alpha, chisq, yji_fpoly_ctf, &alamda,
                *poly_orderb, *poly_orderd, *wavelength, *hkl_max, *res_hi,
                *A, *Cs, *pixel_size, auto2DdrawBuf, auto2DdataBuf);
    if (*chisq <chisqtmp) {
      chisqtmp = *chisq;
      angtmp = poly_paras[*poly_order-1];
    }
    i++;i++;i++;i++;/* calculate this cycle every 5 degrees to save time */
  }
  poly_paras[*poly_order-1] = angtmp;
  printf ("\nThe new angle is founded!\n");
  printf ("		angle: %e \n",  poly_paras[*poly_order-1]*rad_deg);
  /**********************************************************************
   now every thing except angle is recalculated
  ***********************************************************************/
  for(i=0; i<ma-1; i++) ia[i] = 1;
  ia[ma-1] = 0;
  alamda = -1.0;
  for (;;) {
    yji_mrqmin (ca2d_id, ca2d_avg, ca2d_std, poly_paras, 
		ia, *poly_order,
		covar, alpha, chisq, yji_fpoly_ctf, &alamda,
		*poly_orderb, *poly_orderd, *wavelength, 
		*hkl_max, *res_hi,
		*A, *Cs, *pixel_size, 
		auto2DdrawBuf, auto2DdataBuf);
    
    printf("Third pass [%f]\n",alamda);
    
    if(alamda>big || alamda<small) break;
  }
  alamda = 0.0;
  yji_mrqmin (ca2d_id, ca2d_avg, ca2d_std, poly_paras, ia, *poly_order,
	      covar, alpha, chisq, yji_fpoly_ctf, &alamda,
	      *poly_orderb, *poly_orderd, *wavelength, *hkl_max, *res_hi,
	      *A, *Cs, *pixel_size, auto2DdrawBuf, auto2DdataBuf);
  
  for (i=0; i<*poly_orderb; i++)
    printf("\na[%d]= %e",i, poly_paras[i]);
  for (i=*poly_orderb; i<*poly_orderb+*poly_orderd; i++)
    printf("\nb[%d]= %e",i, poly_paras[i]);
  printf ("\n\ndefocus major: %e \n",  poly_paras[*poly_order-3]);
  printf ("defocus minor: %e \n",  poly_paras[*poly_order-2]);
  printf ("angle: %e \n",  poly_paras[*poly_order-1]*rad_deg);
  printf ("chisq: %e \n\n",  *chisq);
  /*end here refine angle*/
  
  free(ca2d_id);
  free(covar[0]);
  free(covar);
  free(alpha[0]);
  free(alpha);
  free(ia);
}



void yji_fpoly_ctf(int idx, int idy, double a[], 
		   double *ymode, double dyda[], 
		   int na, 
		   int order_poly1,
		   int order_poly2, 
		   double  wavelength, 
		   int idim_half, 
		   double  A, 
		   double  Cs,
		   double  pixel_size )
{
  int i, j, k, orderb, orderd;
  double  arg1, arg2, dvalue, defmajor, defminor, angle, theta1;
  double  fw, fw2, theta, def, ctf, ctf2;
  double  cosph, sinph, costh, sinth, dctfdth;
  
  angle    = a[na-1];
  defminor = a[na-2];
  defmajor = a[na-3];
  
  if (idx!=0) theta1 = atan2(idy,idx);
  else {
    if (idy >=0) theta1 = M_PI/2;
    else theta1 = -M_PI/2;
  }

  theta1 -= angle;
  cosph   = cos(theta1);
  sinph   = sin(theta1);
  def     = defmajor*cosph*cosph + defminor*sinph*sinph;
  
  fw =wavelength*sqrt(idx*idx + idy*idy)/(idim_half*2.0* pixel_size );
  fw2= fw*fw;
  theta = M_PI*2.0*(Cs*fw2*fw2/4.0-def*fw2/2.0)/wavelength;
  costh = cos(theta);
  sinth = sin(theta);
  ctf = sqrt(1.0-A*A)*sinth-A*costh;
  ctf2 = ctf*ctf;
  dctfdth = sqrt(1.0-A*A)*costh+A*sinth;
  
  orderb = floor(sqrt(2*order_poly1)) - 1;
  orderd = floor(sqrt(2*order_poly2)) - 1;
  
  /* compute B + ctf*ctf*D
     we formulate the B and D in the order of x, 2nd order is y, 
     such as order 2:  a0 + a1*y + a2*y*y + a3*x + a4*x*y + a5*x*x  
     such as order 3:  a0 + a1*y + a2*y*y + a3*y*y*y + a4*x + a5*x*y + 
                            a6*x*y*y + a7*x*x +a8*x*x*y + a9*x*x*x 
  */
  *ymode=0.0;
  arg1 = 1;
  k    = 0;
  for (i = 0; i<= orderb; i++) {
    arg2  = 1;
    for (j = 0; j <= orderb - i; j++){
      dyda[k] = arg1 * arg2;
      *ymode += a[k] * dyda[k];
      arg2   *= idy;
      k++;
    }
    arg1  *= idx;
  }
  
  dvalue = 0.0;
  arg1 = 1;
  for (i = 0; i<= orderd; i++) {
    arg2  = 1;
    for (j = 0; j <= orderd - i; j++){
      dyda[k] = ctf2 * arg1 * arg2;
      *ymode += a[k] * dyda[k];
      dvalue += a[k] * arg1 * arg2;
      arg2   *= idy;
      k++;
    }
    arg1  *= idx;
  }
  
  /* compute using defocus major */
  dyda[na-3]= -2.0*M_PI*ctf*dvalue*fw2*dctfdth*cosph*cosph /wavelength;
  /* compute using defocus minor */
  dyda[na-2]= -2.0*M_PI*ctf*dvalue*fw2*dctfdth*sinph*sinph /wavelength ;
  /* compute using angle */
  dyda[na-1]= -4.0*M_PI*ctf*dvalue*fw2*(defmajor-defminor)*
    dctfdth*sinph*cosph /wavelength;
  if(DEBUG1) {   
    printf("\nX is %d, \n Y is %d\n", idx, idy);
    printf("   ymode = %f\n", *ymode);
    for(i=0;i<na; i++)  
      printf("   a[%d] = %f\n",i, a[i]);
    for(i=0;i<na; i++)
      printf("   dyda[%d] = %f\n",i, dyda[i]);
  } 
}



void yji_mrqcof(int id[], 
		double y[], 
		double sig[], 
		double a[], 
		int ia[],
		int ma, 
		double **alpha, 
		double beta[], 
		double *chisq,
		void (*funcs)(int, int, double [], double *, double [], 
			      int, int, int, double, int, 
			      double, double, double),
		int num_gauss_bkg, 
		int num_gauss_ctf, 
		double wavelength, 
		int idim_half, 
		double res_hi, 
		double A,
		double Cs, 
		double pixel_size,
		float *auto2DdrawBuf,
		float *auto2DdataBuf)
     
     /* Used by yji_mrqmin to evaluate the linearized fitting matrix
	alpha, and vector beta and calculate chi square
	
	program: Numerical Recipe in C, p686 modified by Wei Zhang,
	pass more parameters and change float into double*/ 
{
  int i, j, k, l, m, mfit, lenx, leny, len, idx, idy;
  double ymod, wt, sig2i, dy, *dyda;
  double radius;
  
  dyda=(double *) malloc (ma * sizeof(double));
  mfit = 0;
  for (j=0;j<ma;j++)
    if(ia[j]) mfit++;
  for(j=0;j<mfit;j++) {
    for (k=0;k<=j;k++) alpha[j][k]=0.0;
    beta[j]=0.0;
  }
  *chisq=0.0;
  lenx = idim_half + 1;
  leny = 2*idim_half + 1;
  len = lenx * leny;
  /* allocate memory for  recalculate sig */
  
  /*  caculate the max annuals at the fisrt elipse */
  /*  recalculate sig */
  
  for (i=0;i<len; i++) { 
    if (id[i]==1) {
      idx = i%lenx;
      idy = i/lenx - idim_half;
      (*funcs)(idx, idy, a, &ymod, dyda, ma, 
	       num_gauss_bkg, num_gauss_ctf, wavelength, idim_half,
               A, Cs, pixel_size);    
      if(idx < 256 && idy < 256 && idy > -256)
	{/* This is it */
	  auto2DdataBuf[(((idy+256)*256)+idx)] = (float)y[i];
	  auto2DdrawBuf[(((idy+256)*256)+idx)] = (float)ymod;
	}/* This is it */
      else
	printf("auto2DdrawBuf not found [%d][%d]\n",idx,idy);
      if(sig[i]==0) printf("sig is 0!\n");
      sig2i=1.0/(sig[i]*sig[i]);    
      dy=y[i]-ymod;
      for (j=-1,l=0;l<ma;l++) {
	if (ia[l]) {
	  wt=dyda[l]*sig2i;
	  for (j++,k=0,m=0;m<=l;m++)
	    if(ia[m]) alpha[j][k++]+=wt*dyda[m];
	  beta[j] += dy*wt;
	}
      }
      
      radius = sqrt(idx*idx + idy *idy);
      *chisq +=dy*dy*sig2i / (radius*radius);
    }
  }
  for (j=1;j<mfit;j++)
    for (k=0;k<j; k++) alpha[k][j]=alpha[j][k];
}

void yji_mrqmin(int x[], double y[], double sig[], double a[], int ia[],
		int ma, double **covar, double **alpha, double *chisq,
		void (*funcs)(int, int, double [], double *, 
			      double [], int, int, int, double,
			      int, double, double, double), 
		double *alamda, int num_gauss_bkg, int num_gauss_ctf, 
		double wavelength,
		int idim_half, double res_hi, double A, double Cs, 
		double pixel_size, 
		float *auto2DdrawBuf, float *auto2DdataBuf)
     /*program: Numerical Recipe in C, p687 modified by Wei Zhang,
       pass more paramerters and change float into double*/
{
  int j, k, l;
  static int mfit;
  static double ochisq, *atry, *beta, *da, **oneda;
  
  if (*alamda < 0.0) {
    atry=(double *) malloc (ma * sizeof(double ));
    beta=(double *) malloc (ma * sizeof(double ));
    da  =(double *) malloc (ma * sizeof(double ));
    for (mfit=0,j=0;j<ma;j++)
      if (ia[j]) mfit++;
    oneda=(double **) malloc ( mfit * sizeof(double ));
    oneda[0] = (double *) malloc (mfit* sizeof(double ));
    for (j=1; j<mfit; j++) oneda[j] = oneda[j-1] + 1 ;
    
    *alamda=0.001;      
    yji_mrqcof(x,y,sig,a,ia,ma,alpha,beta,chisq,
	       funcs,num_gauss_bkg,num_gauss_ctf,
	       wavelength, idim_half, res_hi, A, Cs, pixel_size,
	       auto2DdrawBuf, auto2DdataBuf);
    ochisq=(*chisq);
    for(j=0;j<ma;j++) atry[j]=a[j];
  }
  
  for (j=0;j<mfit;j++) {
    for (k=0;k<mfit;k++) covar[j][k]=alpha[j][k];
    covar[j][j]=alpha[j][j]*(1.0+(*alamda));
    oneda[j][0]=beta[j];
  }
  gaussj(covar, mfit, oneda, 1);
  
  for (j=0;j<mfit;j++) da[j]=oneda[j][0];
  if (*alamda == 0.0) {
    covsrt(covar, ma, ia, mfit);
    free(oneda[0]);
    free(oneda);
    free(da);
    free(atry);
    return;
  }
  
  for (j=0, l=0; l<ma; l++)
    if(ia[l]) atry[l]=a[l]+da[j++];
  yji_mrqcof(x,y,sig,atry,ia,ma,covar,da,chisq,
	     funcs,num_gauss_bkg,num_gauss_ctf,
	     wavelength, idim_half, res_hi, A, Cs, pixel_size,
	     auto2DdrawBuf, auto2DdataBuf);
  
  if(*chisq<ochisq) {
    *alamda *= 0.1;
    ochisq=(*chisq);
    for (j=0;j<mfit; j++) {
      for (k=0;k<mfit;k++) alpha[j][k]=covar[j][k];
      beta[j]=da[j];
    }
    for(l=0;l<ma;l++) a[l]=atry[l];
  } else {
    *alamda *= 10.0;
    *chisq=ochisq;
  }
}



void stuffrobembuf_ (float  *auto2DdrawBuf, 
		     double *a, 
		     int    *naCall,
		     double *waveCall, double *pixSizCall,
		     double *CsCall, double *ACall,
		     int *poly1Call, int *poly2Call)
{
  // function name: stuffrobembuf_
  // Do whatever it takes to make a display buffer for robem that
  // describes: s = b + ctf^2 * d
  
  int     i, j, k, orderb, orderd, idx, idy, idim_half;
  double  arg1, arg2, dvalue, defmajor, defminor, angle, theta1;
  double  fw, fw2, theta, def, ctf, ctf2;
  double  cosph, sinph, costh, sinth; //, dctfdth;
  double  ymode, *dyda;
  double  wavelength, pixel_size, Cs, A;
  int     na, order_poly1, order_poly2, x, y;
  float   minF,maxF,scaleF;
  
  na          = *naCall;
  wavelength  = *waveCall;
  pixel_size  = *pixSizCall;
  Cs          = *CsCall;
  A           = *ACall;
  order_poly1 = *poly1Call;
  order_poly2 = *poly2Call;    
  
  angle    = a[na-1];
  defminor = a[na-2];
  defmajor = a[na-3];
  
  dyda=(double *) malloc (na * sizeof(double));
  for(i=0;i<na;i++)
    dyda[i] = 0.0;
  
  for(y = 0 ; y < 512 ;y ++) {
    
    idy = 256 - y;
    if(idy < 1)
      idy = 0 - idy;
    
    for(x = 0 ; x < 256 ;x ++) {
      
      idx = x + 256;
      
      if (idx!=0) theta1 = atan2(idy,idx);
      else {
	if (idy >=0) theta1 = M_PI/2;
	else theta1 = -M_PI/2;
      }

      theta1 -= angle;
      cosph = cos(theta1);
      sinph = sin(theta1);

      def = defmajor*cosph*cosph + defminor*sinph*sinph;
      
      idim_half = 256; /* I think this is right for our 512 display */
      fw =wavelength*sqrt(idx*idx + idy*idy)/(idim_half*2.0* pixel_size );
      fw2= fw*fw;
      theta = M_PI*2.0*(Cs*fw2*fw2/4.0-def*fw2/2.0)/wavelength;
      costh = cos(theta);
      sinth = sin(theta);
      ctf = sqrt(1.0-A*A)*sinth-A*costh;
      ctf2 = ctf*ctf;
//      dctfdth = sqrt(1.0-A*A)*costh+A*sinth;
      
      orderb = floor(sqrt(2*order_poly1)) -1;
      orderd = floor(sqrt(2*order_poly2)) -1;
      
      /* compute B + ctf*ctf*D
	 we formulate the B and D in the order of x, 2nd order is y, 
	 such as order 2:  a0 + a1*y + a2*y*y + a3*x + a4*x*y + a5*x*x  
	 such as order 3:  a0 + a1*y + a2*y*y + a3*y*y*y + a4*x + a5*x*y + 
	                   a6*x*y*y + a7*x*x +a8*x*x*y + a9*x*x*x 
      */

      ymode=0.0;
      arg1 = 1;
      k    = 0;
      for (i = 0; i<= orderb; i++) {
	arg2  = 1;
	for (j = 0; j <= orderb - i; j++){
	  dyda[k] = arg1 * arg2;
	  ymode += a[k] * dyda[k];
	  arg2   *= idy;
	  k++;
	}
	arg1  *= idx;
      }
      
      dvalue = 0.0;
      arg1 = 1;
      for (i = 0; i<= orderd; i++) {
	arg2  = 1;
	for (j = 0; j <= orderd - i; j++) {
	  dyda[k] = ctf2 * arg1 * arg2;
	  ymode += a[k] * dyda[k];
	  dvalue += a[k] * arg1 * arg2;
	  arg2   *= idy;
	  k++;
	}
	arg1  *= idx;
      }
      
      auto2DdrawBuf[y*256 + x] = ymode;
      
    }
  }
  
  minF = maxF = auto2DdrawBuf[0];
  for(i=0;i<(256*512);i++) {
    minF = MIN(minF, auto2DdrawBuf[i]);
    maxF = MAX(maxF, auto2DdrawBuf[i]);
  }
  scaleF = 200.0 / (maxF - minF);
  
  for(y=0; y<512 ; y++)
    for(x=0; x<256 ; x++)
      auto2DdrawBuf[y*256 + x] = 10.0 + (auto2DdrawBuf[y*256 + x]-minF)*scaleF;
  
} // end function stuffrobembuf_
