#!/usr/bin/perl
use warnings;
use strict;

# per_ptle_ctf.pl
#
# Program takes a particle parameter file, boxed coordinate file, and
# CTFTILT output file as input and generates a new particle parameter
# file in which CTF parameters (major and minor defocus, astigmatism
# angle) are listed for each particle image.
#
# Author: Bob Sinkovits, SDSC
# Last modified: 12/11/07


# Print usage info if no arguments provided
if(@ARGV == 0 || $ARGV[0] =~ /^-help/i) {
print << "EOF";

NAME
  per_ptle_ctf - generates particle parameter file with defocus
  information provided on a per-particle image basis.

EXAMPLE
  per_ptle_ctf -dat dfile -bcrd bfile -ctf cfile -xcen x -ycen y -mode mode

DESCRIPTION
  -d[at]  dfile   particle parameter file
  -b[crd] bfile   RobEM bcrd file
  -c[tf]  cfile   CTFTILT output file
  -xcen   x       x-coordinate of box center (see notes below)
  -ycen   y       y-coordinate of box center (see notes below)
  -mode   mode    generate stigmated or astigmatic (default=astig)

   xcen and ycen only need to be specified if particle parameter files
   are empty and are ignored for non-empty files. By default, CTF
   parameters provided for astigmatic images. Using '-mode stig' sets
   defocus major and minor equal to average obtained from CTFTILT.

EOF
exit(0);
}

# Process command line args so that we can split on ' -'
my $command_line = join ' ', @ARGV;
$command_line =~ s/^\s+//;   # Get rid of leading whitespace
$command_line =~ s/\s+$//;   # Get rid of trailing whitespace
$command_line =~ s/=/ /g;    # Change '=' to ' '
$command_line =~ s/^-//;     # Get rid of leading '-'
@ARGV = split(/\s+-/, $command_line);

my %cmdline_arg;
foreach my $arg (@ARGV) {
    my($key, $value) = split(/\s+/, $arg);
    $key =~ tr/A-Z/a-z/;
    if ($key eq 'd') { 
	$key = 'dat';
    }
    if ($key eq 'b') {
	$key = 'bcrd';
    }
    if ($key eq 'c') {
	$key = 'ctf';
    }
    $cmdline_arg{$key} = $value;
}

my ($dat_file, $bcrd_file, $ctf_file, $xboxcen, $yboxcen, $mode);

if(defined $cmdline_arg{dat}) {
    $dat_file = $cmdline_arg{dat};
}

if(defined $cmdline_arg{bcrd}) {
    $bcrd_file = $cmdline_arg{bcrd};
}

if(defined $cmdline_arg{ctf}) {
    $ctf_file = $cmdline_arg{ctf};
}

if(defined $cmdline_arg{xcen}) {
    $xboxcen = $cmdline_arg{xcen};
}

if(defined $cmdline_arg{ycen}) {
    $yboxcen = $cmdline_arg{ycen};
}

if(defined $cmdline_arg{mode}) {
    $mode = $cmdline_arg{mode};
}

unless(defined $dat_file) {
    print "\n";
    print "Particle parameter file not specified\n";
    print "type 'per_ptle_ctf -help' for usage information\n";
    print "--- Exiting per_ptle_ctf\n\n";
    exit;
}
unless(defined $bcrd_file) {
    print "\n";
    print "bcrd file not specified\n";
    print "type 'per_ptle_ctf -help' for usage information\n";
    print "--- Exiting per_ptle_ctf\n\n";
    exit;
}    
unless(defined $ctf_file) {
    print "\n";
    print "CTFTILT file not specified\n";
    print "type 'per_ptle_ctf -help' for usage information\n";
    print "--- Exiting per_ptle_ctf\n\n";
    exit;
}

if(defined $mode) {
    unless ($mode =~ /^stig/i || $mode =~ /^astig/i) {
	print "\n";
	print "Allowed values are stig[mated] or astig[matic] - ";
	print "mode set to $mode\n";
	print "type 'per_ptle_ctf -help' for usage information\n";
	print "--- Exiting per_ptle_ctf\n\n";
	exit;
    }
} else {
    $mode = 'astig';
}

unless(-r $dat_file) {
    print "\n";
    print "$dat_file is not readable\n";
    print "--- Exiting per_ptle_ctf\n\n";
    exit;
}
unless(-r $bcrd_file) {
    print "\n";
    print "$bcrd_file is not readable\n";
    print "--- Exiting per_ptle_ctf\n\n";
    exit;
}
unless(-r $ctf_file) {
    print "\n";
    print "$ctf_file is not readable\n";
    print "--- Exiting per_ptle_ctf\n\n";
    exit;
}

# --- Parse CTFTILT file and store equation parameters
open(FH, $ctf_file);
my ($dfmid1, $dfmid2, $angast, $tangle, $cx, $cy, $n1, $n2, $psize);
while(<FH>) {
    s/^\s+//;
    s/\s+$//;

    if(/final\s+value/i) {
	($dfmid1, $dfmid2, $angast, undef, $tangle) = split(/\s+/, $_);
    }
    if(/center_x/i) {
	s/.*\s+//;
	$cx = $_;
    }
    if(/center_y/i) {
	s/.*\s+//;
	$cy = $_;
    }
    if(/pixel\s+size/i) {
	s/.*\s+//;
	$psize = $_;
    }
    if(/sin\(tltaxis\)/i) {
	s/.*\s+//;
	$n1 = $_;
    }
    if(/cos\(tltaxis\)/i) {
	s/.*\s+//;
	$n2 = $_;
    }
}
close(FH);

my $deg_to_rad = 0.0174533;
my $tan_tangle = sin($tangle*$deg_to_rad) / cos($tangle*$deg_to_rad);

# --- Parse BCRD file and store box coordinates ---
open(FH, $bcrd_file);
my (@x_bcrd, @y_bcrd);
while(<FH>) {
    s/^\s+//; # Trim leading whitespace
    s/\s+$//; # Trim trailing whitespace

    if($. <= 1) {
	next; # Skip first line
    }

    my ($x, $y) = split(/\s+/, $_);
    unless(defined $x && defined $y) {
	next;
    }

    push(@x_bcrd, $x);
    push(@y_bcrd, $y);
}
close(FH);

# --- Store records of dat file ---
open(FH, $dat_file);
my (@line, $header1, $header2);
while(<FH>) {
    s/\s+$//; # Trim trailing whitespace
    if($. == 1) {
	$header1 = $_;
	next;
    } elsif ($. == 2) {
	$header2 = $_;
	next;
    }
    if(/\w/) { # Make sure that we don't include empty lines
	push(@line, $_);
    }
}
close(FH);

# --- Confirm that the bcrd file is consistent with dat file ---
my $dat_count = @line;
my $bcrd_count = @x_bcrd;

if($dat_count != $bcrd_count && $dat_count > 0) {
    print "\n";
    print "NOTE - Number of records in bcrd and parameter files differ\n";
    print "       Be certain that this is what you had intended\n";
    print "# Boxed coordinate records   : $bcrd_count\n";
    print "# Particle parameter records : $dat_count\n";
    print "\n";
}

if($dat_count == 0 && (!defined $xboxcen || !defined $yboxcen)) {
    print "\n";
    print "For empty particle parameter files, ";
    print "must provide box center coordinates\n";
    print "type 'per_ptle_ctf -help' for usage information\n";
    print "--- Exiting per_ptle_ctf\n\n";
    exit;
}

# --- Write out new particle parameter file ---
my $dat_file_ext = $dat_file;
$dat_file_ext =~ s/\.(.*)$/_ppctf.$1/;

my $zero = 0.0;
my $one  = 1.0;

open(FH, ">$dat_file_ext");
print FH "$header1\n";
print FH "$header2\n";

# Calculate defocus for each bcrd record
my @dfl1 = ();
my @dfl2 = ();
for (my $i=0; $i<$bcrd_count; $i++) {
    my $df = ($n1*($cx-$x_bcrd[$i]) + $n2*($cy-$y_bcrd[$i]))*
	$psize*$tan_tangle;
    $dfl1[$i] = ($df + $dfmid1)/10000.0;
    $dfl2[$i] = ($df + $dfmid2)/10000.0;
    
    if($mode eq 'stig') {
	my $avg = ($dfl1[$i] + $dfl2[$i])/2.0;
	$dfl1[$i] = $avg;
	$dfl2[$i] = $avg;
    }
}
if($mode eq 'stig') {
    $angast = 0.0;
}

# Write out extended particle parameter files
if($dat_count > 0) {
    # Case 1 - dat file contains particle parameter records
    for (my $i=0; $i<$dat_count; $i++) {
	$line[$i] =~ s/^\s+//;
	$line[$i] =~ s/\s+$//;
	$line[$i] =~ s/,/ /g;
	my ($k, $t, $p, $o, $x, $y, $m, $cc1, $cc2, $cc3) = 
	    split(/\s+/,$line[$i]);
	
	unless (defined $cc1) {
	    $cc1 = 0.0;
	}
	unless (defined $cc2) {
	    $cc2 = 0.0;
	}
	unless (defined $cc3) {
	    $cc3 = 0.0;
	}
	
	printf FH "%6d, %8.3f, %7.3f, %7.3f, ", $k, $t, $p, $o;
	printf FH "%7.3f,%7.3f, ", $x, $y;
	printf FH "%5.3f,%7.3f,%7.3f,%7.3f, ", $m, $cc1, $cc2, $cc3;
	printf FH "%6.3f, %6.3f, %7.2f\n", $dfl1[$k-1], $dfl2[$k-1], $angast;
    }
} else {
    # Case 2 - empty dat file
    for (my $i=0; $i<$bcrd_count; $i++) {
	printf FH "%6d, %8.3f, %7.3f, %7.3f, ", $i+1, $zero, $zero, $zero;
	printf FH "%7.3f,%7.3f, ", $xboxcen, $yboxcen;
	printf FH "%5.3f,%7.3f,%7.3f,%7.3f, ", $one, $zero, $zero, $zero;
	printf FH "%7.3f, %7.3f, %7.2f\n", $dfl1[$i], $dfl2[$i], $angast;
    }
}

close(FH);

print "$dat_file_ext successfully generated\n";
