// *****************************************************************
// ROBEM - Rob's Electron Microscopy imaging software.
// 
// Program to take Purdue portable PIF image files, "box" them
// and write out Purdue portable PIF packed format images for use
// in reconstruction.  
//
// Also reads MAP files and displays with 3D or 2D rendering.
// 
// Rob Ashmore
// Biology Department
// Purdue University
// February 1996
//
// *****************************************************************

#define CNVPTX(x) ((fftZoomFactor==2) ? ((x-fftZoomX) * 2) : (x))
#define CNVPTY(y) ((fftZoomFactor==2) ? ((y-fftZoomY) * 2) : (y))
#define CNVPTXC(x) ((fftZoomFactor==2) ? ((x-fftZoomX) * 2 +1) : (x))
#define CNVPTYC(y) ((fftZoomFactor==2) ? ((y-fftZoomY) * 2 - fftZoomFactor) : (y-fftZoomFactor))

#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stddef.h>

#include <Xm/Xm.h>
#include <Xm/GadgetP.h>
#include <Xm/PrimitiveP.h>
#include <Xm/MessageB.h>
#include <Xm/ScrollBar.h>
#include <Xm/ToggleB.h>
#include <Xm/Label.h>
#include <Xm/DialogS.h>
#include <Xm/Text.h>
#include <Xm/List.h>
#include <X11/cursorfont.h>
#include <Xm/FileSB.h>

#include "../include/common.h"
#include "../include/libPIF.h"
#include "../include/libUtil.h"
#include "../include/appodize.h"
#include "init_array.h"
#include "stats.h"
#include "lists.h"
#include "robem_def.h"
#include "robem.h"
#include "robem_tt.h"
#include "emma_autobox.h"
#include "3dgraphics.h"
#include "gui_func.h"

#include "../bitmaps/zoom.bit"
#include "../bitmaps/zoommask.bit"
#include "../bitmaps/lantern.bit"
#include "../bitmaps/lanternmask.bit"
#include "../bitmaps/color.xbm"
#include "../bitmaps/uncolor.xbm"
#include "../bitmaps/section.xbm"
#include "../bitmaps/unsection.xbm"
#include "../bitmaps/locked.xbm"
#include "../bitmaps/unlocked.xbm"
#include "../bitmaps/play.xbm"
#include "../bitmaps/stop.xbm"
#include "../bitmaps/playing.xbm"
#include "../bitmaps/stopped.xbm"
#include "../bitmaps/doflashlight.xbm"
#include "../bitmaps/noflashlight.xbm"
#include "../bitmaps/dorock.xbm"
#include "../bitmaps/norock.xbm"
#include "../bitmaps/infini.xbm"
#include "../bitmaps/noinfini.xbm"
#include "../bitmaps/go.xbm"
#include "../bitmaps/nogo.xbm"
#include "../bitmaps/help.xbm"
#include "../bitmaps/nohelp.xbm"
#include "../bitmaps/exit.xbm"
#include "../bitmaps/noexit.xbm"
#include "../bitmaps/reset.xbm"
#include "../bitmaps/noreset.xbm"
#include "../bitmaps/slow.xbm"
#include "../bitmaps/noslow.xbm"
#include "../bitmaps/movie.xbm"
#include "../bitmaps/unmovie.xbm"




Boolean getToggleState(Widget w) {
  // function name: getToggleState
  // Obtain the state (on/off) of a toggleButton
  
  Boolean toggleState;
  Arg args[1];

  XtSetArg(args[0], XmNset, &toggleState);
  XtGetValues(w, args, 1);
  return(toggleState);

} // end function getToggleState



void doLabel(char *str, Widget wid, int flag)
{
  // function name: doLabel
  // Set the widget label with the input string.
  // flag has value of FALSE (0), TRUE (1), or DEFAULT (2)

  if(useGUI == FALSE) return;
  XtSetArg(arglst[0], XmNlabelString, XmStringCreateLocalized(str));

  if(flag == 0 || flag == 1) {
    XtSetArg(arglst[1], XmNrecomputeSize, flag);
    XtSetValues(wid, arglst, 2); 
  } else {
    XtSetValues(wid, arglst, 1); 
  }    
  XFlush(dsply);

} // end function doLabel



void doDialogTitle(char *str, Widget wid, int flag)
{
  // function name: doDialogTitle
  // Set the widget dialog title with the input string.
  // flag has value of TRUE (1) or FALSE (0)

  if(useGUI == FALSE) return;

  XtSetArg(arglst[0], XmNdialogTitle, XmStringCreateLocalized(str));
  XtSetArg(arglst[1], XmNrecomputeSize, flag );
  XtSetValues(wid, arglst, 2); 
  XFlush(dsply);

} // end function doDialogTitle



void doValue(char *str, Widget wid, int flag)
{
  // function name: doValue
  // Set the widget value with the input string.
  // flag has value of TRUE (1) or FALSE (0)

  if(useGUI == FALSE) return;

  XtSetArg(arglst[0], XmNvalue, str );
  XtSetArg(arglst[1], XmNrecomputeSize, flag );
  XtSetValues(wid, arglst, 2); 
  XFlush(dsply);

} // end function doValue



int main(int argc, char **argv)
{
  // function name: main
  // OS transfer point. The main routine does all the one-time setup
  // and then calls XtAppMainLoop.
  
  String fallbacks[] =
    { "*fontList:-dec-terminal-bold-r-normal--14-140-75-75-c-80-iso8859-1=tag",
      NULL
    };
  
  // Show them our version number
  sprintf(versionStr,"%s",banner("RobEM"));
  
  // Initialize our C code
  init_globals(argc, argv);
  curBox = 0;      // Assignments for create_editboxscreen widgets
  editBoxZoom = 1;
  paintSize = 10;
  inited= FALSE;
  showBoxingSwitch = FALSE;
  buffer_width  = defDispBufSize;
  buffer_height = defDispBufSize;
  image_width   = DEF_DISPLAY_SIZE;
  image_height  = DEF_DISPLAY_SIZE;
  allocInitBuffers();

  if(useGUI) {
    XtToolkitInitialize();
    app_context = XtCreateApplicationContext();
    
    XtAppSetFallbackResources(app_context, fallbacks);
    
    // Open the display
    dsply = XtOpenDisplay(app_context, NULL, "robem",
			  "robem", NULL, 0, &argc, argv);
    
    // Add our remapping of keys to widgets
    XtAppAddActions(app_context, rwa_action_table, 1);
    XtAppAddActions(app_context, box_action_table, 1);
    XtAppAddActions(app_context, editBoxText_action_table, 1);
    XtAppAddActions(app_context, map_abang_action_table, 1);
    XtAppAddActions(app_context, aoverb_action_table, 1);
    XtAppAddActions(app_context, fftImgNum_action_table, 1);
    XtAppAddActions(app_context, fftImgNumArrow_action_table, 1);
    XtAppAddActions(app_context, surfDens_action_table, 1);
    XtAppAddActions(app_context, zoom_action_table, 1);
    XtAppAddActions(app_context, zoomArrow_action_table, 1);
    XtAppAddActions(app_context, imgNum_action_table, 1);
    XtAppAddActions(app_context, radius_action_table, 1);
    XtAppAddActions(app_context, boxInit_action_table, 1);
    XtAppAddActions(app_context, fftStep_action_table, 1);
    XtAppAddActions(app_context, boxNum_action_table, 1);
    XtAppAddActions(app_context, paintSize_action_table, 1);
    XtAppAddActions(app_context, icosa_action_table, 1);
    XtAppAddActions(app_context, gbs_action_table, 1);
    XtAppAddActions(app_context, cmScroll_action_table, 1);
    XtAppAddActions(app_context, version_action_table, 1);
    XtAppAddActions(app_context, paramFile_action_table, 1);
    XtAppAddActions(app_context, paramMapFile_action_table, 1);
    XtAppAddActions(app_context, paramScanFile_action_table, 1);
    XtAppAddActions(app_context, projContour_action_table, 1);
    XtAppAddActions(app_context, projasu_action_table, 1);
    XtAppAddActions(app_context, ppin_action_table, 1);
    XtAppAddActions(app_context, stepPP_action_table, 1);
    XtAppAddActions(app_context, stepPPtext_action_table, 1);
    XtAppAddActions(app_context, ppintext_action_table, 1);
    XtAppAddActions(app_context, binPP_action_table, 1);
    XtAppAddActions(app_context, binPPtext_action_table, 1);
    XtAppAddActions(app_context, strtxpp_action_table, 1);
    XtAppAddActions(app_context, ppMinMax_action_table, 1);
    XtAppAddActions(app_context, ctfValue_action_table, 1);
    XtAppAddActions(app_context, ctfCurve_action_table, 1);
    XtAppAddActions(app_context, latText_action_table, 1);
    XtAppAddActions(app_context, ctfText_action_table, 1);
    XtAppAddActions(app_context, rotflt_action_table, 1);
    XtAppAddActions(app_context, cDisp_zArrow_act_table, 1);
    XtAppAddActions(app_context, subText_action_table, 1);
    XtAppAddActions(app_context, diffSec_action_table, 1);
    XtAppAddActions(app_context, smMMax_action_table, 1);
    XtAppAddActions(app_context, phase_action_table, 1);
    XtAppAddActions(app_context, lopText_action_table, 1);
    XtAppAddActions(app_context, hipText_action_table, 1);
    XtAppAddActions(app_context, coravg_action_table, 1);
    XtAppAddActions(app_context, lohiRadiusText_action_table, 1);
    
    // Explicitly call this class "robem"
    toplevel_widget = XtVaAppCreateShell(argv[0], "robem", 
					 topLevelShellWidgetClass, dsply, NULL);
    
    // Create our colormap
    visual = DefaultVisual(dsply, DefaultScreen(dsply));
    vdepth = (unsigned int)DefaultDepth(dsply, DefaultScreen(dsply));
    color_defs        =  NULL;
    mirror_color_defs =  NULL;
    gen_color_defs    =  NULL;
    if(vdepth != 8)
      createDefaultColormap();
    
    create_mainscreen();
    plotWindow_create(); // Creates additional main screen widgets
    create_binningscreen();
    create_rotatescreen();
    create_normalscreen();
    create_rescalescreen();
    create_statscreen();
    create_circavgscreen();
    create_appodizescreen();
    create_truncaddscreen();
    create_autodrscreen();
    create_dioscreen();
    create_fix2dscreen();
    create_pointscreen();
    create_manipppscreen();
    create_cavgppscreen();
    create_versionscreen();
    create_fixscreen();
    create_helpscreen();
    create_statusscreen();
    create_infoanswerscreen();
    create_infomessagebox();
    create_colorscreen();
    create_interpscreen();
    create_movementscreen();
    create_fftfeedbackscreen();
    create_domegascreen();
    create_dxyscreen();
    create_optionscreen();
    create_boxscreen();
    create_invertscreen();
    create_fixitscreen();
    create_subtractmapscreen();
    create_emcoravgscreen();
    create_refscreen();
    create_moviecontourscreen();
    create_mapshadescreen();
    create_batchjobscreen();
    create_modelscreen();
    create_fourierfiltscreen();
    create_crysdispscreen();
    create_imgconvscreen();
    create_partprocscreen();
    create_rotfltscreen();
    create_editboxscreen();
    create_fouraverscreen();
    create_auto2ddrscreen();
    create_fftscreen();
    fft_create(); // Creates additional FFT screen widgets
    
    XtUnmanageChild(XmMessageBoxGetChild(infoBoxW,XmDIALOG_HELP_BUTTON));
    
    // Doing this here avoids potential problems (but not quite sure why)
    TextSet(boxRadiusTextW, "0", FALSE); 
    TextSet(emptyShellTextW, "5", FALSE); 
    TextSet(emmaRadiusTextW, "45", FALSE); 
    TextSet(featherRadiusTextW, "101", FALSE); 
    
    // Now go looking for fonts
    if(robemFontList != 0) { // They've specified an override font
      XtSetArg(arglst[0], XmNfontList,  robemFontList);
      XtSetArg(arglst[1], XmNlabelFontList,  robemFontList);
      XtSetArg(arglst[2], XmNbuttonFontList,  robemFontList);
      XtSetValues(optionsScreenW, arglst, 3);
      XtSetValues(editBoxW, arglst, 3);
      XtSetValues(colorScreenW, arglst, 3);
      XtSetValues(mainWindowW, arglst, 3);
      XtSetValues(imgNumTextW, arglst, 3);
      XtSetValues(pointLabelW[1], arglst, 3);
      XtSetValues(pointLabelW[2], arglst, 3);
      XtSetValues(pointLabelW[3], arglst, 3);
      XtSetValues(pointLabelW[4], arglst, 3);
      XtSetValues(toplevel_widget, arglst, 3);
      XtSetValues(infoBoxW, arglst, 3);
      XtSetValues(imgNumLabelW, arglst, 3);
      XtSetValues(XtParent(mainWindowW), arglst, 3);
      XtSetValues(toplevel_widget, arglst, 3);
      XFlush(dsply);
    }
    
    // Manage GUIs so that attributes can be modified below
    XtManageChild(partprocScreenW);
    XtManageChild(editBoxScreenW);
    XtManageChild(mapShadeScreenW);
    XtManageChild(colorScreenW);
    XtManageChild(fftScreenW);
    XtManageChild(rotfltScreenW);
    XtManageChild(circAvgScreenW);
    XtManageChild(emcorAvgScreenW);
    XtManageChild(mainWindowW);
    XFlush(dsply);

    enableSaveUnder();
    XlibInitialize();  
    createCursors();
    createPixmaps();
    applyPixmaps();
    
    // Make following widgets not sensitive
    XtSetArg(arglst[0], XmNsensitive, FALSE);
    XtSetValues(sectionW, arglst, 1); 
    XtSetValues(stepToggleW, arglst, 1); 
    XtSetValues(stepToggle2W, arglst, 1); 
    XtSetValues(frameTextW, arglst, 1); 
    XtSetValues(stepTextW, arglst, 1); 
    XtSetValues(stepValueLabelW, arglst, 1); 
    XtSetValues(numFramesLabelW, arglst, 1); 
    XtSetValues(movieFormW, arglst, 1); 
    
    XtUnmanageChild(showBoxingStatsToggleW); 
    XtUnmanageChild(boxGrid2W); 
    XtUnmanageChild(partprocScreenW);
    XtUnmanageChild(editBoxScreenW);
    XtUnmanageChild(mapShadeScreenW);
    XtUnmanageChild(colorScreenW);
    XtUnmanageChild(fftScreenW); 
    XtUnmanageChild(rotfltScreenW); 
    XtUnmanageChild(circAvgScreenW);
    XtUnmanageChild(emcorAvgScreenW);    
    XtUnmanageChild(icosRadProjToggleW);
    XtUnmanageChild(radProjToggleW);
    XtUnmanageChild(radProjCSToggleW);
    XtUnmanageChild(parmTogsW);
    XtUnmanageChild(projasuW);

    XFlush(dsply);
    XSync(dsply,0);
    XFlush(dsply);
    
    inited= TRUE;
    ref_box_flag = FALSE;
    
    if(!liteSwitch) { // Don't let them choose display redraws
      XtUnmanageChild(drawLiteButtonW);
      XtUnmanageChild(drawFFTLiteButtonW);
      XtUnmanageChild(drawFFT2LiteButtonW);
    }

    // Read configurations from resource file (if exists)
    readResourceFile();
    
    // Put up version number
    sprintf(outStr,"Robem: [Version %s]",versionStr);
    doDialogTitle(outStr, mainWindowW, FALSE);
    
  } // End GUI initialization
  
  checkAutomatic(); // Check for command line options

  if(useGUI) {
    handle_cr();                // Override widget translations
    XtAppMainLoop(app_context); // Process X-events
  }

  return(0);
  
} // end function main



void checkAutomatic()
{
  // function name: checkAutomatic
  // Perform operations as specified by command line arguments

  char *s1;
  int i, outWidth, outHeight, *passTag;
  
  passTag = (int *)malloc(sizeof(int));

  if(forceAvgBkgSwitch) { // setup toggle
    XtVaSetValues(forceAvgBkgW, XmNset, TRUE, NULL); 
    forceAvgBkg = TRUE;
  }

  // Set the zoom stuff first
  if(zoomItSwitch) {
    sizeFactor = zoomItFactor;
    sprintf(outStr,"%.2f",sizeFactor);
    TextSet(zoomTextW, outStr, TRUE);
    
    zoomFactor = rwalog2(sizeFactor);
    zoomFactor = MAX(-4.0,MIN(zoomFactor,4.0));
    zoomItSwitch = FALSE;
  }

  // See if they want to preload something
  if(preloadSwitch) {
    readWrite= READING;
    gotFilename(0,0,0);
  }

  // See if they want normalize and linear gradient
  if(dumpNormStats) { // normalize stats
    *passTag = 1;
    normalize_activate(0, (XtPointer)passTag, 0);
    free(passTag);
    exit(0);
  }

  // See if they want spherical average
  if(dumpSpherAvg) { // spherical average
    FILE *fDump;
    float *thisRint, *thisStDev;
    thisRint  = circAvg[0].rint;
    thisStDev = circAvg[0].stDev;

    if ( (fDump = fopen(dumpSpherAvgFile,"w")) != NULL ) {
      for(i=0;i<(numRows/2)+1;i++)
	    fprintf(fDump,"%d %f %f\n", i, thisRint[i], thisStDev[i]);
      fclose(fDump);
    }
    exit(0);
  }
  
  // Do lingrad
  if(doLinAlso)	{ // linear gradient
    printf("Linear Gradient \n");
    linGrad_activate(NULL, NULL, NULL); 
  }

  // See if they want to fix values
  if(doFix) {
    printf("Remove blemishes\n");

    removeBlemish_activate(NULL, NULL, NULL);
    
    *passTag = FIX2D;
    fix2D_do_activate(0, (XtPointer)passTag, 0);
    free(passTag);
    
    if(!normLinSwitch) { // write it in normlin area
      s1 = strrchr(fileName.addr,'.');
      if(s1)
	*s1 = '\0';
      if(doOutFile)
	sprintf(outStr,"%s",outFname);
      else
	sprintf(outStr,"%s_fix.pif",fileName.addr);
      sprintf(fileName.addr,"%s",outStr);
      // Calc our dimensions
      if(packRadius <= 0)
	outHeight= outWidth = purdueHeader.numCols;
      else 
	outHeight= outWidth = (packRadius*2)+1;	
      if(purdueHeader.numCols != purdueHeader.numRows) {
	outHeight = purdueHeader.numRows;
	outWidth  = purdueHeader.numCols;}
      write_pif(outWidth, outHeight, fileName.addr, PACKEDPORTABLE);
      exit(0);
    }
  }

  // See if they want normalize and linear gradient
  if(normLinSwitch) { // normalize and linear gradient
    printf("Normalize\n");
    *passTag = 1;
    
    normalize_activate(0, (XtPointer)passTag, 0);
    free(passTag);
    normalize_activate_2(0, NULL, 0);
    
    if(!reBoxSwitch) { // write it in rebox area
      s1 = strrchr(fileName.addr,'.');
      if(s1)
	*s1 = '\0';
      if(doOutFile)
	sprintf(outStr,"%s",outFname);
      else
	sprintf(outStr,"%s_norm.pif",fileName.addr);
      sprintf(fileName.addr,"%s",outStr);
      // Calc our dimensions
      if(packRadius <= 0)
	outHeight= outWidth = purdueHeader.numCols;
      else 
	outHeight= outWidth = (packRadius*2)+1;	
      if(purdueHeader.numCols != purdueHeader.numRows) {
	outHeight = purdueHeader.numRows;
	outWidth  = purdueHeader.numCols;
      }
      write_pif(outWidth, outHeight, fileName.addr, PACKEDPORTABLE);
      exit(0);
    }
  }

  // See if they want to rebox an exisitng boxed PIF file
  if(reBoxSwitch) { // rebox it
    printf("Rebox\n");
    
    // Actually do the reboxing
    boxRadius = autoBoxBoxSize;
    featherRadius = autoBoxFeatherSize;
    
    s1 = strrchr(fileName.addr,'.');
    if(s1)
      *s1 = '\0';
    if(doOutFile)
      sprintf(outStr,"%s",outFname);
    else
      sprintf(outStr,"%s_rebox.pif",fileName.addr);
    sprintf(fileName.addr,"%s",outStr);
    // Calc our dimensions
    if(packRadius <= 0)
      outHeight= outWidth = purdueHeader.numCols;
    else 
      outHeight= outWidth = (packRadius*2)+1;	
    if(purdueHeader.numCols != purdueHeader.numRows) {
      outHeight = purdueHeader.numRows;
      outWidth  = purdueHeader.numCols;
    }
    write_pif(outWidth, outHeight, fileName.addr, PACKEDPORTABLE);
    exit(0);
  }

  // See if they want autoBoxing
  if(autoBoxSwitch) { // autobox
    char thisFileName[512];
    
    emmaRadius = autoBoxSize;
    emptyShellWidth = 5;
    boxRadius  = autoBoxBoxSize;
    featherRadius  = autoBoxFeatherSize;
    refCutoff = autoBoxThresh;

    // First autobox
    *passTag = AUTOBOX;
    autoBox_activate(0, (XtPointer)passTag, 0);

    if (autoBoxThresh >0.0) {
      *passTag = CMDREMOVEBOXES;
      autoMinMaxThreshold == AMMCORRCOEF;
      autobox_scroll_change(0, (XtPointer)passTag, 0);
    }

    free(passTag);

    if (autoBoxIter > 0) {
       // Then correlate
       for(i=0; i<numBoxes; i++)
          box[i].reference = 1; // Turn them all ON
       ref_correlate_activate();
    }

    // Let's write the box coords
    sprintf(thisFileName,"%s",fileName.addr);
    s1= strrchr(thisFileName,'.');
    if(s1)
      sprintf(s1,".bcrd");
    else 
      sprintf(thisFileName,"%s.bcrd",fileName.addr);
    writeBoxCoords(thisFileName, BOXCOORDS);
    
    if(useGUI) {
      raiseIt(statusScreenW);
      XSync(dsply,0);
    }
    
    // Let's write the file
    s1 = strrchr(fileName.addr,'.');
    if(s1)
      *s1 = '\0';
    if(doOutFile)
      sprintf(outStr,"%s",outFname);
    else
      sprintf(outStr,"%s_box.pif",fileName.addr);
    sprintf(fileName.addr,"%s",outStr);
    
    outHeight= outWidth = (boxRadius*2)+1;
    // Shouldn't we use the featherRadius here to determine
    // outWidth and outHeight for write_pif() ???
    outHeight= outWidth = (featherRadius*2)+1;
    write_pif(outWidth, outHeight, fileName.addr, CIRCLEPACKED);
    exit(0);
  }
  
  if(boxitSwitch) { // boxit
    readBoxCoords(boxitFname);
    
    // Override with what we've got from command line
    emmaRadius = boxitSize;
    emptyShellWidth = 5;
    boxRadius  = boxitBoxSize;
    featherRadius  = boxitFeatherSize;
    
    // Let's write the file
    s1 = strrchr(fileName.addr,'.');
    if(s1)
      *s1 = '\0';
    if(doOutFile)
      sprintf(outStr,"%s",outFname);
    else
      sprintf(outStr,"%s_box.pif",fileName.addr);
    sprintf(fileName.addr,"%s",outStr);

    outHeight= outWidth = (boxRadius*2)+1;
    // Shouldn't we use the featherRadius here to determine
    // outWidth and outHeight for write_pif() ???
    outHeight= outWidth = (featherRadius*2)+1;
    write_pif(outWidth, outHeight, fileName.addr, CIRCLEPACKED);
    exit(0);
  }

  // See if they want to add noise
  if(doNoise) { // noise
    *passTag = FIX2D;
    fix_do_activate(0, (XtPointer)passTag, 0);
    free(passTag);
    
    s1 = strrchr(fileName.addr,'.');
    if(s1)
      *s1 = '\0';
    if(doOutFile)
      sprintf(outStr,"%s",outFname);
    else {
      if(noiseType == GAUSSNOISE)
	sprintf(outStr,"%s_%.1fgausnoise.pif",fileName.addr,noiseStd);
      else
	sprintf(outStr,"%s_%.1fshotnoise.pif",fileName.addr,noiseStd);
    }
    sprintf(fileName.addr,"%s",outStr);
    // Calc our dimensions
    if(packRadius <= 0)
      outHeight= outWidth = purdueHeader.numCols;
    else 
      outHeight= outWidth = (packRadius*2)+1;	
    if(purdueHeader.numCols != purdueHeader.numRows) {
      outHeight = purdueHeader.numRows;
      outWidth  = purdueHeader.numCols;
    }
    if(pifMode==20)
      write_pif(outWidth, outHeight, fileName.addr, MAP);
    else
      write_pif(outWidth, outHeight, fileName.addr, PACKEDPORTABLE);
    exit(0);
  }

  // See if they want to apply CTF
  if(doApplyCtf) { // applyCTF
    idim1 = idim2 = 1024;   // MAX fft size
    CK(fftImage_gold= (float *)malloc(sizeof(float)*(idim1*(idim2*2))));
    writeFFTType = APPLYCTFR4 ;
    ctfFilter = applyCtfFilter;
    fft_activate(NULL, NULL, NULL);
    ctf_all_activate();
    exit(0);
  }

  // See if they want to trunc
  if(doTrunc) { // trunc
     
   if(fileType!=MAP) {
      printf("\n************************************************\n");
      printf("* Cannot TRUNC this file type [%d]\n",fileType);
      printf("************************************************\n");
      exit(0);
    }

    *passTag= 0;
    truncadd_do_activate(0, (XtPointer)passTag, 0);
    free(passTag);
    
    s1 = strrchr(fileName.addr,'.');
    if(s1)
      *s1 = '\0';
    if(doOutFile)
      sprintf(outStr,"%s",outFname);
    else
      sprintf(outStr,"%s_trunc.pif",fileName.addr);
    sprintf(fileName.addr,"%s",outStr);
    outHeight = purdueHeader.numRows;
    outWidth  = purdueHeader.numCols;
    write_pif(outWidth, outHeight, fileName.addr, MAP);
    exit(0);
  }

  // See if they want to interp
  if(doInterp) { // interp
    if(fileType!=EMPORTABLE && fileType!=BYTEPORTABLE && fileType!=PORTABLE) {
      printf("\n************************************************\n");
      printf("* Cannot interp this file type [%d]\n",fileType);
      printf("************************************************\n");
      exit(0);
    }
    
    interpolate_activate(NULL, NULL, NULL);
    
    s1 = strrchr(fileName.addr,'.');
    if(s1)
      *s1 = '\0';
    if(doOutFile)
      sprintf(outStr,"%s",outFname);
    else
      sprintf(outStr,"%s_interp.pif",fileName.addr);
    sprintf(fileName.addr,"%s",outStr);
    outHeight = purdueHeader.numRows;
    outWidth  = purdueHeader.numCols;
    write_pif(outWidth, outHeight, fileName.addr, EMPORTABLE);
    exit(0);
  }

} // end function checkAutomatic



void calcScaleFactor(float *floatScaleFactor, float *recipFloatScaleFactor, 
		     float *int4scaleF, float img_min, float img_max,
		     float img_avg, float img_std, int fileType)
{
  // function name: calcScaleFactor
  // Calculate the factor needed to convert floatIntegers to floats
  // when we read this file later.  Scan all of our values and pick a
  // value that maintains our dynamic range.

  double maxVal=360.0;

  // Get greatest magnitude
  maxVal= MAX(maxVal, fabs(img_min));
  maxVal= MAX(maxVal, fabs(img_max));
  maxVal= MAX(maxVal, fabs(img_avg));
  maxVal= MAX(maxVal, fabs(img_std));

  // Pad a little so we're guaranteed not to run over the top
  maxVal *= 1.02;
  
  // We're doing MAPs as shorts so make it smaller
  if(fileType==MAP)
    *floatScaleFactor = MAXSHORT / maxVal;
  else
    *floatScaleFactor = MAXINT / maxVal;

  *int4scaleF = MAXINT / maxVal;
  *recipFloatScaleFactor = 1.0 / (double)*floatScaleFactor; 

} // end function calcScaleFactor



void calcRectMedian(float *inbuffer, int *median, int width, int height)
{
  // function name: calcRectMedian 
  // Calculates median from the selected rectangle.

  int i, totSoFar=0, histVals[257];

  *median = 0;
  init_intarr(histVals, 257, 0); // initialize array
  
  // Calculate histogram
  for (i=0; i<height*width; i++)
    histVals[(int)inbuffer[i]]++;
  
  for (i=0; i<257; i++) { // Look for the median value
    totSoFar += histVals[i];
    if(totSoFar >= (width*height)/2) {
      *median = i;
      break;
    }
  }

} // end function calcRectMedian


void calcByteStats(BYTE *inbuffer, float *img_bkg, float *img_min,
	       float *img_max, float *img_avg, float *img_std,
	       int width, int height, int boxWidth)
{
  // function name: calcByteStats
  // Calculates statistics from the image

  int i, j, pixelValue, tot=0, totBkg=0;
  float totValues=0.0, totBkgValues=0.0;
  double var=0.0, sumsq=0.0;

  // Init all
  *img_bkg = 0;
  *img_min = INT_MAX;
  *img_max = INT_MIN;
  
  // Accumulate intermediate values

  for (i=0; i<height; i++) {
    for (j=0; j<width; j++) { // Look at all pixels
      pixelValue = (int)inbuffer[(i*width)+j];
      *img_min = MIN(*img_min,pixelValue);
      *img_max = MAX(*img_max,pixelValue);
      if(pixelValue > 0) { // Only use non-zeros
	sumsq += pixelValue*pixelValue;
	totValues +=  (float)pixelValue;
	tot++;
      }
      if(i==0 || j==0 || i==(height-1) || j==(width-1)) {
	// Accum for RECT background
	totBkgValues += pixelValue;
	totBkg++;
      }
    }
  }
  
  // Calc the background
  if(fileType==BYTEPORTABLE || fileType==EMPORTABLE || fileType==REBOX)
    *img_bkg= (float)
      calcBackground(inbuffer, width, height, boxWidth,
		     fileType, bkgrndMethod,
		     boxMethod, packRadius,
		     boxRadius, featherRadius,
		     recipSizeFactor, tryColors );
  else if(fileType==RECTPACKED && totBkg > 0) {
    *img_bkg = totBkgValues / totBkg;
  }
  else 
    *img_bkg = inbuffer[0]; // e.g. Don't recalc Packed boxes or MAPS!!
  
  // Calc stats
  *img_avg = totValues / (width * height);
  var = sumsq / (width*height) - (*img_avg * *img_avg);
  *img_std = sqrt(fabs(var));

} // end function calcByteStats



void particleCalcStats(int *part_min, int *part_max, 
		       float *part_avg, float *part_std)
{
  // function name: particleCalcStats
  // Calculates statistics from the particle X values.

  int pixelValue, tot=0;
  float totValues=0.0;
  double var=0.0, sumsq=0.0;
  P_INFO *thisInfo;

  // Init all
  *part_min = INT_MAX;
  *part_max = INT_MIN;
  *part_avg = 0;
  *part_std = 0;

  // Accumulate intermediate values
  thisInfo = info.info;
  while(thisInfo != NULL) { // Do all of them 
    pixelValue = (int)thisInfo->x;
    *part_min = MIN(*part_min,pixelValue);
    *part_max = MAX(*part_max,pixelValue);
    sumsq += pixelValue*pixelValue;
    totValues +=  (float)pixelValue;
    tot++;
    thisInfo = thisInfo->next;
  }
  
  // Calc stats
  *part_avg = totValues / tot;
  *part_std = 0.0;
  var = sumsq / tot - (*part_avg * *part_avg);
  *part_std = sqrt(fabs(var));

} // end function particleCalcStats



void calcSubScanStats(float *inbuffer, 
		      float *img_bkg, float *img_min,
		      float *img_max, float *img_avg, float *img_std,
		      int width, int height, int boxWidth, 
		      int offsetX, int offsetY)
{
  // function name: calcSubScanStats
  // Calculate statistics from the image but only from the subregion
  // determined by boxwidth and offset values

  int i, j;
  float pixelValue, totValues=0.0;
  double var=0.0, sumsq=0.0;

  // Init all except *img_bkg (may need value that's there)
  *img_min = INT_MAX;
  *img_max = INT_MIN;
  
  // Accumulate intermediate values
  for (i=0; i<height; i++){
    for (j=0; j<width; j++) {
      pixelValue = inbuffer[((i+offsetY)*boxWidth)+j+offsetX];
      *img_min = MIN(*img_min, pixelValue);
      *img_max = MAX(*img_max, pixelValue);
      sumsq += pixelValue*pixelValue;
      totValues += pixelValue;
    }
  }
  
  // Calc stats
  *img_avg = totValues / (width*height);
  var = sumsq / (width*height) - (*img_avg * *img_avg);
  *img_std = sqrt(fabs(var));
  
  *img_bkg = (int)*img_avg;
  
} // end function calcSubScanStats



void calcCircleStats(float *inbuffer, 
		     float *img_bkg, float *img_min,
		     float *img_max, float *img_avg, float *img_std,
		     int width, int height, int boxWidth)
{
  // function name: calcCircleStats
  // Calculates statistics from the image but only from the circle
  // inside the image.

  int i, j, tot=0;
  int xcen, ycen, iMycen, jMxcen, thisRadius;
  float pixelValue, distnce, totValues=0.0;
  double var=0.0, sumsq=0.0;

  // Init all except *img_bkg (may need value that's there)
  *img_min = INT_MAX;
  *img_max = INT_MIN;

  thisRadius = boxRadius;

  // Accumulate intermediate values
  xcen = width / 2;
  ycen = height / 2;

  for (i=0; i<height; i++) {
    iMycen = i-ycen;
    for (j=0; j<width; j++) { // Look at all pixels
      pixelValue = inbuffer[(i*boxWidth)+j];
      jMxcen = j-xcen;
      distnce = sqrt((double)(iMycen*iMycen + jMxcen*jMxcen)); 
      if(distnce<= thisRadius) { // we've found one in the circle
	*img_min = MIN(*img_min,pixelValue);
	*img_max = MAX(*img_max,pixelValue);
	sumsq += pixelValue*pixelValue;
	totValues +=  pixelValue;
	tot++;
      }
    }
  }
  
  // Calc the background
  if(fileType==BYTEPORTABLE || fileType==EMPORTABLE  || 
     fileType==PORTABLE     || fileType==REBOX) {
    // They are boxing from a scan
    if(backgroundMode==USECIRCLE)
      *img_bkg= (float)
	calcFloatBackground(inbuffer, width, height, boxWidth,
			    fileType, bkgrndMethod,
			    boxMethod, packRadius,
			    boxRadius, featherRadius,
			    recipSizeFactor, tryColors);
    else if(backgroundMode==USERECT) {
      printf("calcRectBackground not yet implemented.\n");
      // *img_bkg= (short)calcRectBackground(
      // inbuffer, width, 
      // height, boxWidth );
    }
    else if(backgroundMode==PICK)
      *img_bkg = inbuffer[0]; 
    else
      printf("Unknown background method [%d]\n",backgroundMode);
  }
  else if(packRadius <= 0) {
    // e.g. Don't recalc Packed boxes that have no radius!!!
      *img_bkg = purdueHeader.img_bkg;
  }
  else 
    *img_bkg = inbuffer[0]; // e.g. Don't recalc Packed boxes or MAPS!!

  // Calc stats - Only so many are used for tot
  *img_avg = totValues / tot;
  var = sumsq / tot - (*img_avg * *img_avg);
  // New MISHA way ??? var = (sumsq  - (*img_avg * *img_avg))  / (float)tot;
  // printf("var[%f]  avg[%f]\n",var,*img_avg);
  
  *img_std = sqrt(fabs(var));
  
} // end function calcCircleStats



void calcRealStats(float *inbuffer, 
		   float *img_bkg, float *img_min,
		   float *img_max, float *img_avg, float *img_std,
		   int width, int height, int boxWidth, int nimg)
{
  // function name: calcRealStats
  // Calculates statistics from the real image.

  int i, j, k, tot=0;
  int xcen, ycen, iMycen, jMxcen, thisRadius;
  float pixelValue, distnce, totValues=0.0;
  double var=0.0, sumsq=0.0 ;

  // Init all except *img_bkg (may need value that's there)
  *img_min = HUGE;
  *img_max = -HUGE;

  if(fileType==PACKEDPORTABLE || fileType==REBOX)
    thisRadius = (packRadius <= 0) ? height+width : packRadius;
  else 
    thisRadius = boxRadius;

  // Accumulate intermediate values
  xcen = ycen = thisRadius;

  for (k=0; k<nimg; k++) {
    for (i=0; i<height; i++) {
      iMycen = (i*recipSizeFactor)-ycen;
      for (j=0; j<width; j++) {
	pixelValue = inbuffer[(i*width)+j];
	jMxcen = (j*recipSizeFactor)-xcen;
	distnce = sqrt((double)(iMycen*iMycen + jMxcen*jMxcen)); 
	// Force this to always happen - e.g. does radius mean anything here?
	distnce = 0.0;
	if(distnce<= thisRadius) { // we've found one in the circle
	  *img_min = MIN(*img_min,pixelValue);
	  *img_max = MAX(*img_max,pixelValue);
	  // if(pixelValue > 0.0) { // Only use non-zeros
	  sumsq += (pixelValue * pixelValue);
	  totValues +=  pixelValue;
	  tot++;
	  // }
	}
      }
    }
  }
  printf("totValues[%f] to[%d]\n",totValues,tot);
  
  // Calc stats
  *img_avg = totValues / (width * height);
  var = sumsq / (width*height) - (*img_avg * *img_avg);
  *img_std = sqrt(fabs(var));
  
  // Accumulate intermediate values for background
  totValues = 0;
  tot = 2*height + 2*width;
  
  for(i=0;i<height;i++) {
    totValues += inbuffer[i*width];              // Left edge
    totValues += inbuffer[(i*width)+(width-1)];  // Right edge
  }
  
  for (i=0; i<width; i++) {
    totValues += inbuffer[i];                    // Bottom edge
    totValues += inbuffer[(height-1)*width + i]; // Top edge
  }
  
  // Calc the background - avg value around the perimeter of the square
  if(tot > 0)
    *img_bkg = totValues / tot;
  
} // end function calcRealStats



void raiseIt(Widget w)
{
  // function name: raiseIt
  // Do everything necessary to raise a window to the foreground on
  // the display.

  if(useGUI == FALSE) return;

  XtManageChild(w);
  XtPopup(XtParent(w), XtGrabNone);

  XmProcessTraversal(w, XmTRAVERSE_CURRENT); 
  XtRealizeWidget(w);  
  XRaiseWindow(dsply, XtWindow(XtParent(w)));
  XFlush(dsply);

} // end function raiseIt



void createDefaultColormap()
{
  // function name: createDefaultColormap
  // Do whatever is necessary to get us the desired colormap.
  
  int i;
  float scaleFactor;
//  FILE *fColor;

  visual= DefaultVisual(dsply, DefaultScreen(dsply));
  vdepth = (unsigned int)DefaultDepth(dsply, DefaultScreen(dsply));
  
  // Allocate the color cells
  pixels_return= (unsigned long *)malloc(sizeof(unsigned long)*256);

  // Allocate the color definition structs
  if (color_defs == NULL) 
    color_defs = (XColor *) malloc (1025 * sizeof (XColor));
  if (mirror_color_defs == NULL) 
    mirror_color_defs = (XColor *) malloc (1025 * sizeof (XColor));
  if (gen_color_defs == NULL) 
    gen_color_defs = (XColor *) malloc (1025 * sizeof (XColor));
  
  
  color_map = XCreateColormap(dsply, DefaultRootWindow(dsply),
			      visual, AllocNone);

//  printf("Using default colormap: ");
//  printf("visual = [%d], ",visual);
//  printf("DEC colormap = [%d]\n",color_map);

  if (color_map == (Colormap) NULL) {
    fprintf (stderr, "Could not create color map for visual\n");
    exit (1);
  }
  
  // Now let's query the DISPLAY until we get the maximum number of
  // contiguous colorcell entries in the colormap that we can .
  
  tryColors= 256;
  
  // Init the color defs to point to each entry
  for (i = 0; i <256 ; i++) // Make the color map gray-scale
    color_defs[i].pixel = gen_color_defs[i].pixel = pixels_return[i] = i;
  
  // Find out what's already there in the color table
  XQueryColors(dsply, color_map, color_defs, 256);
  XQueryColors(dsply, color_map, gen_color_defs, 256);
  
//GCA: not clear the reason ot write the file - commenting for now
/*  sprintf(outStr,"%s/cmap.robem",getenv("HOME"));      
  
  if ( (fColor = fopen(outStr,"w")) != NULL ) {
    fwrite(&tryColors,sizeof(int),1,fColor);
    for (i=0; i<tryColors; i++)
      fwrite(&pixels_return[i],sizeof(unsigned long),1,fColor);
  }
  else
    printf("ERROR - Cannot write colormap file.\n");

  fclose(fColor);
*/  
  // Now change the colormap for drawing in pixels_return[GREEN,RED,BLUE] 
  // First calculate some fudge factors.
  
  scaleFactor = 255.0 / (tryColors - 7.0);
  
  // Make these entries in the colormap GRAYSCALE
  for (i=6; i<tryColors; i++) {
    // Do the colors we have been assigned
    gen_color_defs[pixels_return[i]].pixel = pixels_return[i];
    gen_color_defs[pixels_return[i]].red   = (i-6)*scaleFactor*PIXELCOLOR;
    gen_color_defs[pixels_return[i]].green = (i-6)*scaleFactor*PIXELCOLOR;
    gen_color_defs[pixels_return[i]].blue  = (i-6)*scaleFactor*PIXELCOLOR;
    gen_color_defs[pixels_return[i]].flags = DoRed | DoGreen | DoBlue;
    gen_color_defs[pixels_return[i]].pad   = 0;

    // struct assignment
    color_defs[pixels_return[i]] = gen_color_defs[pixels_return[i]];

    XAllocColor (dsply, color_map, &color_defs[i]);
    pixels_return[i] = color_defs[i].pixel;
    gen_color_defs[i].pixel =  i;
  }
  
  // Now define our robem colors
  color_defs[WHITE].pixel = WHITE;
  color_defs[WHITE].red   = 255 * PIXELCOLOR;
  color_defs[WHITE].green = 255 * PIXELCOLOR;
  color_defs[WHITE].blue  = 255 * PIXELCOLOR;
  color_defs[WHITE].flags = DoRed | DoGreen | DoBlue;

  color_defs[YELLOW].pixel = YELLOW;
  color_defs[YELLOW].red   = 255 * PIXELCOLOR;
  color_defs[YELLOW].green = 255 * PIXELCOLOR;
  color_defs[YELLOW].blue  = 0;
  color_defs[YELLOW].flags = DoRed | DoGreen | DoBlue;
  
  color_defs[GREEN].pixel = GREEN;
  color_defs[GREEN].red   = 0;
  color_defs[GREEN].green = 255 * PIXELCOLOR;
  color_defs[GREEN].blue  = 0;
  color_defs[GREEN].flags = DoRed | DoGreen | DoBlue;
  
  color_defs[pixels_return[RED]].pixel = RED;
  color_defs[pixels_return[RED]].red   = 255 * PIXELCOLOR;
  color_defs[pixels_return[RED]].green = 0;
  color_defs[pixels_return[RED]].blue  = 0;
  color_defs[pixels_return[RED]].flags = DoRed | DoGreen | DoBlue;

  color_defs[BLUE].pixel = BLUE;
  color_defs[BLUE].red   = 0;
  color_defs[BLUE].green = 0;
  color_defs[BLUE].blue  = 255 * PIXELCOLOR;
  color_defs[BLUE].flags = DoRed | DoGreen | DoBlue;

  color_defs[PLACEHOLDER].pixel = PLACEHOLDER;
  color_defs[PLACEHOLDER].red   = 0;
  color_defs[PLACEHOLDER].green = 255 * PIXELCOLOR;
  color_defs[PLACEHOLDER].blue  = 0;
  color_defs[PLACEHOLDER].flags = DoRed | DoGreen | DoBlue;
 
  color_defs[BLACK].pixel = BLACK;
  color_defs[BLACK].red   = 0;
  color_defs[BLACK].green = 0;
  color_defs[BLACK].blue  = 0;
  color_defs[BLACK].flags = DoRed | DoGreen | DoBlue;

  XAllocColor (dsply, color_map, &color_defs[WHITE]);
  pixels_return[WHITE] = color_defs[WHITE].pixel;
  XAllocColor (dsply, color_map, &color_defs[YELLOW]);
  pixels_return[YELLOW] = color_defs[YELLOW].pixel;
  XAllocColor (dsply, color_map, &color_defs[GREEN]);
  pixels_return[GREEN] = color_defs[GREEN].pixel;
  XAllocColor (dsply, color_map, &color_defs[pixels_return[RED]]);
  pixels_return[RED] = color_defs[RED].pixel;
  XAllocColor (dsply, color_map, &color_defs[BLUE]);
  pixels_return[BLUE] = color_defs[BLUE].pixel;
  XAllocColor (dsply, color_map, &color_defs[PLACEHOLDER]);
  pixels_return[PLACEHOLDER] = color_defs[PLACEHOLDER].pixel;
  XAllocColor (dsply, color_map, &color_defs[BLACK]);
  pixels_return[BLACK] = color_defs[BLACK].pixel;

} // end function createDefaultColormap



void mainRobem_sub_UNIX(unsigned short **int_buffer, fString_t *fName,
			int *i, int *j)
{
  // function name: mainRobem_sub_UNIX
  // Complicate routine for reading images and updating GUI

  char descriptionStr[180] = {0};
  int k, l, m, z, one=1;
  int inWidth, inHeight, firstTime=TRUE;
  off_t imgOffset;
  float inBuf[MAXFILESIZE];
  FILE *fIn;
  forPifHead_t thisHead;
  fString_t inFile,rWrite;

  inFile.addr = (char *)malloc(sizeof(char)*264);
  rWrite.addr = (char *)malloc(sizeof(char)*64);
  
  if(useGUI) {
    XtSetArg(arglst[0], XmNsensitive, TRUE);
    XtSetValues(truncaddButtonW, arglst, 1); 
    XtSetValues(removeBlemishButtonW, arglst, 1); 
    XtSetValues(editBoxButtonW, arglst, 1); 
    XtSetValues(pcxButtonW, arglst, 1); 
    XtSetValues(correlateButtonW, arglst, 1);  
    XtSetValues(correlateButton2W, arglst, 1);  
    XtSetValues(autoBoxW, arglst, 1);  
    XtSetValues(linGradW, arglst, 1); 
    
    XtSetArg(arglst[0], XmNsensitive, FALSE);
    XtSetValues(paintMapButtonW, arglst, 1); 
    XtSetValues(mapMovieButtonW, arglst, 1); 
    XtSetValues(mapShadeButtonW, arglst, 1); 
    XtSetValues(invertButtonW, arglst, 1); 
    XtSetValues(rescaleButtonW, arglst, 1); 
    XtSetValues(modelW, arglst, 1); 
    XtSetValues(subtractMapButtonW, arglst, 1); 
    XtSetValues(calcSFButtonW, arglst, 1); 
    XtSetValues(hardRotateButtonW, arglst, 1); 
    XtSetValues(rotateButtonW, arglst, 1); 
    XtSetValues(fourFiltW, arglst, 1); 
    XtSetValues(addAverageW, arglst, 1); 
    XtSetValues(checkBkgW, arglst, 1); 
    XtSetValues(normalizeW, arglst, 1); 
    XtSetValues(normalizeBkgW, arglst, 1); 
    XFlush(dsply);
  }
  
  *i = *j = 0;
  
  if ( (fIn = fopen((*fName).addr,"r")) == NULL ) {
    sprintf(outStr,"Can't read [%s] ",(*fName).addr);
    askThem(outStr, XmDIALOG_INFORMATION);
    goto bugOut;
  }

  fclose(fIn);
  
  thisHead.luNum = 7;
  sprintf(inFile.addr,"%s",(*fName).addr);
  inFile.length = strlen((*fName).addr);
  sprintf(rWrite.addr,"r");
  
  pif_open(inFile.addr, &thisHead.luNum, rWrite.addr);
  pif_read_gh(&thisHead);
  
  thisHead.imgNum = 1;
  pif_read_dh(&thisHead);
  
  // Init featherMode
  if(useGUI) {
    XtVaSetValues(apOldW, XmNset, FALSE, NULL);
    XtVaSetValues(apLinearW, XmNset, FALSE, NULL);
    XtVaSetValues(apGaussianW, XmNset, FALSE, NULL);
    XtVaSetValues(apExponentialW, XmNset, FALSE, NULL);    
  }
  
  if(thisHead.borderType == 1) {
    featherMode = LINEAR;
    if(useGUI) XtVaSetValues(apLinearW, XmNset, TRUE, NULL);
  }
  else if(thisHead.borderType == 2) {
    featherMode = GAUSSIAN;
    if(useGUI) XtVaSetValues(apGaussianW, XmNset, TRUE, NULL);
  }
  else if(thisHead.borderType == 3) {
    featherMode = EXPONENTIAL;
    if(useGUI) XtVaSetValues(apExponentialW, XmNset, TRUE, NULL);
  }
  else {
    featherMode = NONE;
    if(useGUI) XtVaSetValues(apOldW, XmNset, TRUE, NULL);
  }
    
  fileType = readType;
  
  purdueHeader.numCols = thisHead.numCols;
  purdueHeader.numRows = thisHead.numRows;
  purdueHeader.nimg    = thisHead.nimg;
  purdueHeader.numSecs = thisHead.numSecs;
  
  pifMode = thisHead.mode;
  if(pifMode==20 || pifMode==21 || pifMode==22) { // MAP data
    purdueHeader.numCols = purdueHeader.numRows = *i = *j = 0;
    askThem("Use Read/MAP to read this file.", XmDIALOG_INFORMATION);
    goto bugOut;
  }
  
  if(pifMode==6 || pifMode==46 || pifMode==7) { // Boxed image files
    fileType          = PACKEDPORTABLE;
    boxRadiusDisp     = FALSE;
    emmaRadiusDisp    = FALSE;
    featherRadiusDisp = FALSE;
    numBoxes          = purdueHeader.nimg;
    dispMethod        = BOXES;
    if(useGUI) XtVaSetValues(vScrollW, XmNsensitive, FALSE, NULL); 
    if(useGUI) XtVaSetValues(hScrollW, XmNsensitive, FALSE, NULL); 
  }
  else if(pifMode==0 || pifMode==1 || pifMode ==2) {
    lastPackedNumImgs= 0;    
    if (pifMode==0) {
      purdueHeader.nimg = 1; // Single image
      fileType=BYTEPORTABLE;
    } 
    else
      fileType=PORTABLE;
    
    if(useGUI) XtVaSetValues(vScrollW, XmNsensitive, TRUE, NULL); 
    if(useGUI) XtVaSetValues(hScrollW, XmNsensitive, TRUE, NULL); 

    if(*int_buffer != NULL)
      free(*int_buffer);
    *int_buffer = (unsigned short *)
      malloc(defDispBufSize*defDispBufSize*sizeof(unsigned short) + 1);
  }

  maxPackPixel = -999999999.0;
  minPackPixel =  999999999.0;
  
  if (purdueHeader.nimg > MAXBOX) {
  	printf("Number of images (%d) higher than allowed (%d)!!!\n",purdueHeader.nimg, MAXBOX);
	exit(0);
  }
  for (m=0; m<purdueHeader.nimg; m++) { // Do each image
    
    sprintf(outStr,"Reading image[%d] of [%d]", (m+1),purdueHeader.nimg);
    if(checkStatus(outStr))
      goto bugOut;

    thisHead.imgNum = m+1;
    
    // [2] Read header for each image
    pif_read_dh(&thisHead);
    box[m].x = thisHead.origX;    
    box[m].y = thisHead.origY;   
    purdueHeader.img_bkg = thisHead.img_bkg;  

    if(fileType==PACKEDPORTABLE  && firstTime) { // Setup params
      firstTime     = FALSE;
      borderType    = thisHead.borderType;
      featherRadius = packRadius = thisHead.packRadius;
      boxRadius     = featherRadius;
      if(thisHead.borderWidth < thisHead.numCols)
	boxRadius = featherRadius - thisHead.borderWidth;
      if(useGUI) TextSetInt(featherRadiusTextW, 2*featherRadius + 1, FALSE); 
      if(useGUI) TextSetInt(boxRadiusTextW, featherRadius-boxRadius, FALSE); 
    }

    purdueHeader.img_max = thisHead.datai4_max;
    purdueHeader.img_min = thisHead.datai4_min;
    purdueHeader.img_avg = thisHead.datar4_avg;
    purdueHeader.img_std = thisHead.datar4_std;

    if(fileType==PACKEDPORTABLE) {
      freeNullFloat(&packImg[m]);
      packImg[m]= (float *)
	malloc(sizeof(float)*purdueHeader.numCols*purdueHeader.numRows);
    }

    headOff= 1024;

    if(fileType == PACKEDPORTABLE) {
      inWidth  = purdueHeader.numCols;
      inHeight = purdueHeader.numRows;
    } 
    else {
      inWidth  = MAX(0, MIN(defDispBufSize, purdueHeader.numCols));
      inHeight = MAX(0, MIN(defDispBufSize, purdueHeader.numRows));
    }
    // [3] Now read the image info
    if(fileType==PACKEDPORTABLE) {
      imgOffset  = thisHead.imgOffset;
      
      if(thisHead.mode==6) { // byte
	BYTE *byte_buffer = 
	  (BYTE *)malloc(sizeof(BYTE) * inWidth * inHeight);
	pif_read_byte_image(&thisHead.luNum, byte_buffer, 
			    &imgOffset, &thisHead.numCols, 
			    &thisHead.numRows, &thisHead.numSecs);
	for(z=0; z<(thisHead.numCols*thisHead.numRows); z++)
	  packImg[m][z] = (float)byte_buffer[z];
	free(byte_buffer);
      }
      else if(thisHead.mode==7) { // short
	short *short_buffer = (short *)
	  malloc(sizeof(short) * inWidth * inHeight);
	pif_read_short_image(&thisHead.luNum, short_buffer,
			     &imgOffset, &thisHead.numCols,
			     &thisHead.numRows, &thisHead.numSecs);
	for (z=0; z<(thisHead.numCols*thisHead.numRows); z++)
	  (packImg[m])[z] = (float)short_buffer[z];
	free(short_buffer);
      }
      else if(thisHead.mode==46)  // float
	pif_read_boxi4(&thisHead, packImg[m]);
      
      for (k=0; k<=(inHeight/2); k++) { // Switch it a row at a time
	int yOut, yIn;

	// Packed images on disk stored in Tim's orientation
	// (e.g. drawn from bottom up).If we ever convert to PCK
	// format already in correct convention
	
	yOut = (purdueHeader.numRows - 1 - k)*inWidth;
	yIn  = k*inWidth;
	
	for (l=0; l<purdueHeader.numCols; l++) { // Convert char to int
	  inBuf[0] = packImg[m][yOut+l];
	  inBuf[1] = packImg[m][yIn+l];
	  
	  packImg[m][yOut+l]= inBuf[1];
	  packImg[m][yIn+l] = inBuf[0];
	  
	  // Now find min/max
	  maxPackPixel = MAX(maxPackPixel, inBuf[0]);
	  minPackPixel = MIN(minPackPixel, inBuf[0]);
	  maxPackPixel = MAX(maxPackPixel, inBuf[1]);
	  minPackPixel = MIN(minPackPixel, inBuf[1]);
	}
      }
      
      deletePackedImage[m]= FALSE;
      
    }
    else if(fileType==PORTABLE || fileType==BYTEPORTABLE) { // PORTABLE
      imgOffset  = 1024;
      
      // Pass back dimensions of this image to calling routine
      file_width  = *i = purdueHeader.numCols;
      file_height = *j = purdueHeader.numRows;
      
      // Setup Scroll bars and display_wdith/height
      if(useGUI) formResize(NULL, NULL, NULL);   
      
      if(file_width > display_width || file_height > display_height)
	fullScreen= TRUE;
      else 
	fullScreen= FALSE;
      
      // This changes whenever a new image is read
      image_width = image_height = defDispBufSize;  
      
      inCore_buffer= (float *)malloc(sizeof(float)*file_width*file_height);
      
      if(inCore_buffer)
	imgInCore= TRUE;
      else  {
	imgInCore= FALSE;
	printf("Unable to allocate inCore_buffer!!!\n");
	exit(0);
      }
      
      sprintf(outStr,"Reading image[%d] of [%d] total.\n", 
	      m+1, purdueHeader.nimg);
      strcat(outStr, "Trying to read entire image into buffer...");
      if(checkStatus(outStr))
	goto bugOut;
       
      if(pifMode == 0) {
	BYTE *byte_buffer = (BYTE *)
	  malloc (sizeof(BYTE)*file_width*file_height);
	pif_read_byte_image(&thisHead.luNum, byte_buffer,
			    &imgOffset, &thisHead.numCols,
			    &thisHead.numRows, &thisHead.numSecs);
	for (z=0; z<file_width*file_height; z++)
	  inCore_buffer[z] = (float)(byte_buffer[z]);
	free(byte_buffer);
      }
      else if(pifMode==1) {
	short *short_buffer = (short *)
	  malloc(sizeof(short)*file_width*file_height);
	pif_read_short_image(&thisHead.luNum, short_buffer,
			     &imgOffset, &thisHead.numCols,
			     &thisHead.numRows, &thisHead.numSecs);
	for (z=0; z<file_width*file_height; z++)
	  inCore_buffer[z] = (float)(short_buffer[z]);
	free(short_buffer);
      }
      else if(pifMode==2)
	pif_read_mapi4(&thisHead, inCore_buffer, &one, 
		       &thisHead.numSecs, &thisHead.numCols, 
		       &thisHead.numRows, &thisHead.numSecs);
      
      if (purdueHeader.img_max == 0) {
	float min, max;
	farray_minmax(inCore_buffer, &min, &max, file_width*file_height);
	purdueHeader.img_max = max;
      } 
      
      sprintf(outStr, "Reading image[%d] of [%d]\nWriting to display now",
	      m+1, purdueHeader.nimg);
      if(checkStatus(outStr))
	goto bugOut;
      virtual_read_activate(-99,-99);
      
      if(useGUI) XtVaSetValues(interpolateW, XmNsensitive, TRUE, NULL); 
      if(useGUI) XtVaSetValues(binningW, XmNsensitive, TRUE, NULL); 
    }

  }
  
  if(minMaxColor && useGUI) { // Use option screen defaults
    printf("Using option defaults...\n");
    minPackPixel = wstrtod(minMapTextW);
    maxPackPixel = wstrtod(maxMapTextW);
  }
  
  // Now that we've read all of the packed images we can display them

  borderWidth = thisHead.borderWidth;

  if(map_abang == 0.0)
    map_abang = 90.0;
  if(aoverb == 0.0)
    aoverb = 1.0;
  
  if(checkStatus("Calculating stats ..."))
    goto bugOut;
  
  if(fileType==PACKEDPORTABLE) {
    
    // Let's let them interpolate them
    if(useGUI) XtVaSetValues(interpolateW, XmNsensitive, TRUE, NULL); 
    
    // Now clip this guy to our colormap
    if(minPackPixel < 0.0)
      scalePackFactor = (tryColors - 7.0)/(maxPackPixel-minPackPixel);
    else 
      scalePackFactor = (tryColors - 7.0)/maxPackPixel;
    
    printf("minPackPixel [%f]  max[%f]     scale[%f]\n",
	   minPackPixel, maxPackPixel, scalePackFactor);
    
    // Start at zero'th(1) image
    if(useGUI) stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
				packImg,purdueHeader.nimg, 1);
    
    *i = *j = defDispBufSize;
    
    doLabel("ImgNum", imgNumLabelW, FALSE);
    
    if(useGUI) {
      XtSetArg(arglst[0], XmNsensitive, TRUE);
      XtSetValues(addAverageW, arglst, 1); 
      XtSetValues(fourFiltW, arglst, 1); 
      XtSetValues(checkBkgW, arglst, 1); 
      XtSetValues(normalizeW, arglst, 1); 
      XtSetValues(normalizeBkgW, arglst, 1); 
      XtSetValues(linGradW, arglst, 1); 
      XFlush(dsply);      
      TextSet(imgNumTextW, "1", FALSE);
    }
    
    lastNumBox = 0;
    TextSetInt(boxInitW, lastNumBox+1, TRUE);

  // Old bug resulted in some PIF files having packRadius set to zero
  // or numCols. If encountered, set packRadius to numCols/2
    if(packRadius==0 || packRadius==purdueHeader.numCols) {
      featherRadius = packRadius = purdueHeader.numCols/2;
      if(thisHead.borderWidth < purdueHeader.numCols)
	boxRadius = featherRadius - thisHead.borderWidth;
      else 
	boxRadius = featherRadius;
      TextSetInt(featherRadiusTextW, 2*featherRadius + 1, FALSE); 
      TextSetInt(boxRadiusTextW, featherRadius-boxRadius, FALSE); 
    }
    
    calcCircleStats(packImg[0], &purdueHeader.img_bkg, 
		    &purdueHeader.img_min, &purdueHeader.img_max, 
		    &purdueHeader.img_avg, &purdueHeader.img_std,
		    purdueHeader.numCols, purdueHeader.numRows, 
		    purdueHeader.numCols);
    
    sprintf(outStr,"\nStatistics for file [%s] Image [%d]\n\n\
File size   [%d]x[%d]  nimg[%d]\n\n\
img_bkg     [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\n\n",
	    (*fName).addr, 1,
	    purdueHeader.numCols, purdueHeader.numRows, 
	    purdueHeader.nimg,
	    purdueHeader.img_bkg,
	    purdueHeader.img_min, purdueHeader.img_max, 
	    purdueHeader.img_avg, purdueHeader.img_std
	    );
    TextSet(statTextW, outStr, FALSE);
    lastPackedNumImgs = purdueHeader.nimg;
    if(useGUI) XtVaSetValues(useContrastW, XmNsensitive, FALSE, NULL); 
  }
  else {
    sprintf(outStr,"\nStatistics for file [%s] Image[%d]\n\n\
File size   [%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n",
	    (*fName).addr, 1,
	    purdueHeader.numCols, purdueHeader.numRows,
	    purdueHeader.img_bkg,
	    purdueHeader.img_min, purdueHeader.img_max,
	    purdueHeader.img_avg, purdueHeader.img_std);
    TextSet(statTextW, outStr, FALSE);
    lastPackedNumImgs = 0;
    if(useGUI) XtVaSetValues(useContrastW, XmNsensitive, TRUE, NULL); 
    if(useGUI) XtVaSetValues(histEqualW, XmNsensitive, TRUE, NULL); 
  }
  
  // Get baseline info
  pifHeader = thisHead;
  pifHeader.numCols    = purdueHeader.numCols;
  pifHeader.numRows    = purdueHeader.numRows;
  pifHeader.nimg       = purdueHeader.nimg;
  pifHeader.img_bkg    = purdueHeader.img_bkg;;
  pifHeader.datai4_min = purdueHeader.img_min;
  pifHeader.datai4_max = purdueHeader.img_max;
  pifHeader.datar4_avg = purdueHeader.img_avg;
  pifHeader.datar4_std = purdueHeader.img_std;
  sprintf(pifHeader.descriptionStr, "%s", descriptionStr);

 bugOut:;
  pif_close(&thisHead.luNum);

  if(fileType != MAP && fileType != PACKEDPORTABLE)
    if(useGUI) XtVaSetValues(truncaddButtonW, XmNsensitive, FALSE, NULL);

  free(inFile.addr);
  free(rWrite.addr);

  fflush(NULL);

} // end function mainRobem_sub_UNIX



int encput(unsigned char byt, unsigned char cnt, FILE *fid)
{
  // function name: encput
  // Used by writePCX()

  if (cnt) {
    if (cnt == 1 && (0xC0 != (0xC0 & byt))) {
      if (EOF == putc((int)byt, fid))
	return(0);     // disk write error (probably full)
      return(1);
    }
    else {
      if (EOF == putc((int)0xC0 | cnt, fid))
	return (0);      // disk write error
      if (EOF == putc((int )byt, fid))
	return (0);      // disk write error
      return (2);
    }
  }
  
  return (0);

} // end function encput



int encLine(unsigned char *inBuff, int inLen, FILE *fp)
{
  // function name: encLine
  //
  // Used by writePCX() rtn. This subroutine encodes one scanline and
  // writes it to a file. It returns number of bytes written into
  // outBuff, 0 if failed.
  
  unsigned char this, last;
  unsigned char runCount=1;     // max single runlength is 63
  int srcIndex, i, total=0;

  last = *(inBuff);

  // Find the pixel dimensions of the image by calculating 
  // [XSIZE = Xmax - Xmin + 1] and [YSIZE = Ymax - Ymin + 1].  
  // Then calculate how many bytes are in a "run"
  
  for (srcIndex = 1; srcIndex < inLen; srcIndex++) {
    this = *(++inBuff);
    if (this == last) {     // There is a "run" in the data, encode it
      runCount++;
      if (runCount == 63) {
	if (! (i = encput(last, runCount, fp)))
	  return (0);
	total += i;
	runCount = 0;
      }
    }
    else {                // No "run"  -  this != last
      if (runCount) {
	if (! (i = encput(last, runCount, fp)))
	  return(0);
	total += i;
      }
      last = this;
      runCount = 1;
    }
  }

  if (runCount) {        // finish up
    if (! (i = encput(last, runCount, fp)))
      return (0);
    return (total + i);
  }
  
  return (total);
  
} // end function encLine



void writePCX(char *filename, int startImg, int endImg, int movie)
{
  // function name: write PCX 
  // Write one or more PCX files using the data stored in packImg,
  // cuedMap.shad, editBox_display_buffer, or inCore_buffer.

  char name[80];
  int count1, count2; 
  int frame, i, lastImg, img, ncol, nrow;
  short stemp;
  float min, max, scale, minF, maxF, *thisBuf;
  BYTE temp, identifier=10, *PCX_Buffer;
  FILE *fp;
  
  // Set dimensions for PCX images
  if(movie) {
    if(!lowRes && !dispSections)
      ncol = nrow = largestDim;
    else  // low res images
      ncol = nrow = MAX(numCols, numRows);
  }
  else {
    ncol = purdueHeader.numCols;
    nrow = purdueHeader.numRows;
  }

  if(movie && !dispSections) {
    // Find global min/max for movies of radial projections
    // Start at beginning of first image and scan through number
    // of array elements equal to #images * size
    farray_minmax(&map_image_buffer[startImg*ncol*nrow], &minF, &maxF, 
		  (endImg-startImg+1)*ncol*nrow);
  }

  // If rockIt, step forward and reverse (e.g. 3,4,5,6,5,4,3)
  lastImg = (rockIt) ? 2*endImg - startImg - 1 : endImg;
  
  img = startImg - 1;
  for (frame=startImg; frame<lastImg; frame++) {
    if(movie)
      sprintf(name, "%s_%03d.pcx", filename, frame-startImg);
    else
      sprintf(name, "%s", filename);
      
    if ( (fp = fopen(name,"w")) == NULL ) {
      printf ("Error: Opening file %s\n",name);
      exit(0);
    }

    // Determine the image number corresponding to frame of movie
    if(frame >= endImg)
      img--;
    else 
      img++;

    // Write the header info
    fwrite(&identifier,sizeof(BYTE),1,fp);
    temp = 5; // Version 5 for PC Paintbrush for Win
    fwrite(&temp,sizeof(BYTE),1,fp);
    temp=1;  // Encoding = 1 for RLE encoded
    fwrite(&temp,sizeof(BYTE),1,fp);
    temp=8;  // Bits Per Pixel = 8 for 256 color
    fwrite(&temp,sizeof(BYTE),1,fp);
    stemp=0;  // Upper left corner is at 0,0
    fwrite(&stemp,sizeof(short),1,fp);
    fwrite(&stemp,sizeof(short),1,fp);
    
    stemp = ncol-1;
    fwrite(&stemp,sizeof(short),1,fp);
    stemp = nrow-1;
    fwrite(&stemp,sizeof(short),1,fp);
    stemp = ncol;
    fwrite(&stemp,sizeof(short),1,fp);
    stemp = nrow;
    fwrite(&stemp,sizeof(short),1,fp);
    
    temp=0;
    for (i=0; i<48; i++)
      fwrite(&temp,sizeof(BYTE),1,fp);
    fwrite(&temp,sizeof(BYTE),1,fp);
    temp=1; // Number of bit planes = 1
    fwrite(&temp,sizeof(BYTE),1,fp);
    stemp = (ncol%2==0) ? ncol : ncol-1;
    fwrite(&stemp,sizeof(short),1,fp);
    stemp=1; // PaletteType = 1 for color image
    fwrite(&stemp,sizeof(short),1,fp);
    temp=0; // Pad the rest with 0
    for (i=0; i<58; i++)
      fwrite(&temp,sizeof(BYTE),1,fp);
    
    // Allocate memory and load data into PCX_buffer
    CK(PCX_Buffer = (BYTE *)malloc(sizeof(BYTE)*ncol*nrow));
    if(movie) {
      if (!dispSections)  {  // If a 3D representation
	min=cuedMap[img].shad[0];
	max=cuedMap[img].shad[0];
	for (i=0; i<ncol*nrow; i++) {
	  min = MIN(min, cuedMap[img].shad[i]);
	  max = MAX(max, cuedMap[img].shad[i]);
	}
        
	if(!radialMapShade) { // Not radial
	  scale = (tryColors-8.0)/(max-min);
	  for (i=0; i<ncol*nrow; i++) 
	    PCX_Buffer[i] = (BYTE)((cuedMap[img].shad[i]-min)*scale)+8;
	}
	else { // Can't scale radially cued
	  scale = (tryColors-8.0)/(maxF-minF);
	  for (i=0; i<ncol*nrow; i++)
	    PCX_Buffer[i] = (BYTE)((cuedMap[img].shad[i]-minF)*scale)+8;
	}
      }
      else { // If a section representation
	stuffDisplayMapBuffer(img,1,0);
	for (count1=0,count2=0; count1<nrow*BOX_WIDTH;
	     count1+=BOX_WIDTH,count2+=ncol)
	  memcpy(PCX_Buffer+count2,editBox_display_buffer+count1,ncol);
      }
    } 
    else {
      thisBuf = (fileType==PACKEDPORTABLE) ? packImg[img-1] : inCore_buffer;
      for (i=0; i<ncol*nrow; i++)
	PCX_Buffer[i] = (BYTE)(thisBuf[i]);
    }

    for(i=0; i<nrow; i++) {
      if (ncol%2==0)
        encLine(&PCX_Buffer[i*ncol],ncol,fp);
      else
        encLine(&PCX_Buffer[i*ncol],ncol-1,fp);
    }
    
    // Write the palette
    temp=12;  // 12 precedes the palette
    fwrite(&temp,sizeof(BYTE),1,fp);
    
    for (i=0; i<256; i++) { // For all the palette
      temp = (BYTE)((unsigned short)color_defs[i].red/256);
      fwrite(&temp,sizeof(BYTE),1,fp);
      temp = (BYTE)((unsigned short)color_defs[i].green/256);
      fwrite(&temp,sizeof(BYTE),1,fp);
      temp = (BYTE)((unsigned short)color_defs[i].blue/256);
      fwrite(&temp,sizeof(BYTE),1,fp);
    }
    
    free(PCX_Buffer);
    fclose(fp);

    if(!movie) {
      sprintf(outStr,"PCX file [%s] sucessfully written",name);
      askThem(outStr, XmDIALOG_INFORMATION);
    }
  }
  
} // end function writePCX



void tsbworld_float(float *in_buffer, float *out_buffer,
		    int thisNcols, int thisNrows, int thisNsecs)
{
  // function name: tsbworld_float
  // Turn this float buffer upside down to conform with the way Tim
  // stores them on disk.  Tim's old Fortran code stored them in
  // reverse row order.

  int   tsbRow, yOut, yOut2, i, j, k;

  for (i=0; i<thisNsecs; i++) {
    for (j=0; j<thisNrows; j++) { // Do rows in TSB order
      tsbRow = thisNrows - j - 1;
      yOut = i*thisNcols*thisNrows + tsbRow*thisNcols;
      yOut2= i*thisNcols*thisNrows + j*thisNcols;
      for (k=0; k<thisNcols; k++) // Write it a short at a time
	out_buffer[yOut2+k]= in_buffer[yOut + k];
    }
  }
  
} // end function tsbworld_float



void tsbworld_float_box(float *in_buffer, float *out_buffer,
			int boxWidth, int ncol, int nrow)
{
  // function name: tsbworld_float_box
  // Turn this float buffer upside down to conform with the way Tim
  // stores them on disk. Tim's old Fortran code stored them in
  // reverse row order.

  int i, j;

  for (i=0; i<=(nrow/2); i++) {
    for (j=0; j<ncol; j++) {
      out_buffer[i*ncol + j] = in_buffer[(nrow-i-1)*boxWidth + j];
      out_buffer[(nrow-i-1)*ncol + j] = in_buffer[i*boxWidth + j];
    }
  }
  
} // end function tsbworld_float_box



void tsbworld_byte(BYTE *in_buffer, BYTE *out_buffer,
		   int box_width, int ncol, int nrow)
{
  // function name: tsbworld_byte
  // Turn this byte buffer upside down to conform with the way Tim
  // stores them on disk. Tim's old Fortran code stored them in
  // reverse row order.

  int i, j;

  for(i=0;i<=(nrow/2);i++) {
    for (j=0; j<ncol; j++) {
      out_buffer[i*ncol + j] = in_buffer[(nrow-i-1)*box_width + j];
      out_buffer[(nrow-i-1)*ncol + j] = in_buffer[i*box_width + j];
    }
  }
  
} // end function tsbworld_byte



void checkAdjustBoxes()
{
  // function name: checkAdjustBoxes
  // Move boxes that overlap the edge of the micrograph so that they
  // lie fully within the micrograph
  
  int beginX, endX, beginY, endY, origX, origY, i;

  for (i=0; i<numBoxes; i++) {
    beginY = box[i].y - featherRadius;
    beginX = box[i].x - featherRadius;
    endY   = box[i].y + featherRadius;
    endX   = box[i].x + featherRadius;
    origX = box[i].x;
    origY = box[i].y;

    if(beginX < 0)
      box[i].x -= beginX;
    if(beginY < 0)
      box[i].y -= beginY;
    if(endX >= file_width)
      box[i].x -= (endX + 1 - file_width);
    if(endY >= file_height)
      box[i].y -= (endY + 1 - file_height);
    
    if(box[i].x != origX || box[i].y != origY) {
      printf("Box %d overlaps edge of micrograph\n", i);
      printf("  Shifting from (%d, %d) to (%d, %d)\n",
	     origX, origY, box[i].x, box[i].y);
    }
  }
  
} // end function checkAdjustBoxes



void read_pif(int *ncol, int *nrow, int *nsec,
	      fString_t *inFileName)
{
  // function name: read_pif
  // Read image(s) in pif format. Future goal - get all input to robem
  // read here.

  int i=4, j, thisImg;
  int binCol, binRow, binSec, one=1;
  int strtCol, strtRow, strtSec;
  int convTsb;
  off_t imgOffset;
  fString_t rWrite;

  pifHeader.luNum = i;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 1;
  sprintf(rWrite.addr,"r");
  
  pif_open(inFileName->addr, &i, rWrite.addr);
  pif_read_gh(&pifHeader);
  nimg                  = pifHeader.nimg;
  recipFloatScaleFactor = pifHeader.recipFloatScaleFactor;
  numCols               = pifHeader.numCols;
  numRows               = pifHeader.numRows;
  numSecs               = pifHeader.numSecs;
  pifMode               = pifHeader.mode;
  fileMrcEven           = pifHeader.even;
  fileMrcX              = pifHeader.mrcX;
  fileMrcY              = pifHeader.mrcY;
  fileMrcZ              = pifHeader.mrcZ;
  
  if(pifMode!=20 && pifMode!=40 && pifMode!=21 && 
     pifMode!=22 && pifMode != 46) { // Not a map
    *nsec = -99;
    return;
  }

  pifHeader.imgNum = 1;
  pif_read_dh(&pifHeader);
  
  recipFloatScaleFactor = pifHeader.recipFloatScaleFactor;
  purdueHeader.img_bkg  = pifHeader.img_bkg;
  purdueHeader.img_min  = pifHeader.datai4_min;
  purdueHeader.img_max  = pifHeader.datai4_max;
  purdueHeader.map_min  = pifHeader.datar4_min;
  purdueHeader.map_max  = pifHeader.datar4_max;
  purdueHeader.img_avg  = pifHeader.datar4_avg;
  purdueHeader.img_std  = pifHeader.datar4_std;
  numCols    = pifHeader.numCols;
  numRows    = pifHeader.numRows;
  numSecs    = pifHeader.numSecs;
  imgOffset  = pifHeader.imgOffset; 
  pifMode    = pifHeader.mode; 
  sprintf(title,"%s",pifHeader.descriptionStr);
  aoverb     = pifHeader.aoverb; 
  map_abang  = pifHeader.map_abang;
  a_cell     = pifHeader.a_cell;
  b_cell     = pifHeader.b_cell;
  c_cell     = pifHeader.c_cell; 
  dela       = pifHeader.dela; 
  delb       = pifHeader.delb; 
  delc       = pifHeader.delc;
  binFactor  = pifHeader.binFactor;
  dthe       = pifHeader.dthe; 
  dphi_90    = pifHeader.dphi_90; 
  symmetry   = pifHeader.symmetry; 
  
  fileBinSecs = binSec  = numSecs;
  fileBinRows = binRow  = numRows;
  fileBinCols = binCol  = numCols;
  fileBinFactor = binFactor;
  
  if(pifMode==22) {// SPecial case for PFTS data - nrot*4 
    binRow  = numRows/binFactor;
    // We store it nrot/4 now ... 
    binCol  = numCols/binFactor;
  }
  else {// Normal 
    if(binFactor==2) {// 1/2 
      if(pifMode==20)
	binSec  = numSecs/2+1;  
      binRow  = numRows/2+1;
      binCol  = numCols/2+1;
      if(binSec%2 == 0)
	binSec--;
      if(binRow%2 == 0)
	binRow--;
      if(binCol%2 == 0)
	binCol--;
    }
  }

  if(float_image_buffer1 && float_image_buffer1 != lowRes_buffer1)
    free(float_image_buffer1);
  float_image_buffer1= NULL;

  freeNullFloat(&lowRes_buffer1);
  
  if(pifMode == 46)
    binSec = pifHeader.nimg;
  
  CK(float_image_buffer1 = (float *)malloc(sizeof(float)*
					   binCol*binRow*binSec));
  
  maxMapPixel = -HUGE;
  minMapPixel = HUGE;
  convTsb = TRUE;
  pif_tsbworld(convTsb);
  binFactor = 1;
  
  for (j=0; j<nimg; j++) {
    if(pifMode==20) {
      strtCol = strtRow = strtSec = 1;
      pif_read_map_short_image_(&i, float_image_buffer1,
				&imgOffset,
				&binCol, &binRow, &binSec,
				&recipFloatScaleFactor,
				&minMapPixel,&maxMapPixel, &binFactor,
				&strtCol, &strtRow, &strtSec,
				&binCol, &binRow, &binSec);
    }
    else if(pifMode==40) {
      pif_read_mapi4(&pifHeader, float_image_buffer1,
		     &one, &binSec,
		     &binCol, &binRow, &binSec);
      pifMode = 20;  // Kludge for now YJI 
      farray_minmax(float_image_buffer1, &minMapPixel, &maxMapPixel,
		    binCol*binRow*binSec);
    }
    else if (pifMode==21 || pifMode == 22) {
      pif_read_map_int_image_(&i, float_image_buffer1,
			      &imgOffset,
			      &binCol, &binRow, &binSec,
			      &recipFloatScaleFactor,
			      &minMapPixel,&maxMapPixel, &binFactor, 
			      &pifMode);
    }
    else if (pifMode == 46) {
      int localBinSec=1;  // We've only one section per img 
      thisImg = j + 1;    // For fortran 
      pifHeader.imgNum = thisImg;
      pif_read_dh(&pifHeader);
      imgOffset  = pifHeader.imgOffset = 512 + 512*(j+1) + (off_t) j*binCol*binRow*4;
      pif_read_map_int_image_(&i, (float_image_buffer1+(j*binCol*binRow)),
			      &imgOffset,
			      &binCol, &binRow, &localBinSec,
			      &recipFloatScaleFactor,
			      &minMapPixel,&maxMapPixel, &binFactor, 
			      &pifMode);
    }
    else
      askThem("Bad mode for MAP.", XmDIALOG_INFORMATION);
    
  }
  
  pif_close(&i);
  
  numCols = binCol;
  numRows = binRow;
  numSecs = binSec;
  origNumSecs= numSecs;
  
  if(pifMode == 20 && minMapPixel > -1.0 && maxMapPixel < 1.0) {
    // voxels have very small values 
    float  scaler=1.0;
    int done=FALSE, i;
    
    while(!done) {// find scaler 
      if(scaler*minMapPixel < -10.0 && scaler*maxMapPixel > 10.0)
	done = TRUE;
      else 
	scaler *= 10.0;
      if(scaler > 100000000.0)
	done = TRUE;
    }

    printf("Multiplying voxel densities by [%f].\n" ,scaler);
    
    for (i=0; i<(numCols*numRows*numSecs); i++)
      float_image_buffer1[i] *= scaler;
    
    minMapPixel *= scaler;
    maxMapPixel *= scaler; 
  }

  if(pifMode == 46) {
    // Let's try converting mode==46 to PACKEDPORTABLE 
    int i, k, maxI, minI;
    
    fileType = PACKEDPORTABLE;
    
    printf("MAX[%f]MIN[%f]   \n", maxMapPixel, minMapPixel);
    
    maxI = minI = (float_image_buffer1[0]);
    for (k=0; k<binSec; k++) {
      freeNullFloat(&packImg[k]);
      CK(packImg[k]= (float *)malloc(sizeof(float)*binCol*binRow));
      for (i=0; i<binRow*binCol; i++) {
	packImg[k][i] = float_image_buffer1[(k*binCol*binRow) + i] ;
	minI = MIN(minI, packImg[k][i]);
	maxI = MAX(maxI, packImg[k][i]);
      }
    }

    minPackPixel = minI;
    maxPackPixel = maxI;
    
    if(minMaxColor && useGUI) { // Use option screen defaults 
      printf("Using option defaults \n");
      minPackPixel = wstrtod(minMapTextW);
      maxPackPixel = wstrtod(maxMapTextW);
    }

    // Now clip this guy to our colormap  
    if(minPackPixel < 0.0)
      scalePackFactor = (tryColors - 7.0) / (maxPackPixel-minPackPixel);
    else 
      scalePackFactor = (tryColors - 7.0) / maxPackPixel;

    printf("minPackPixel [%f]  max[%f]     scale[%f]\n",
	   minPackPixel, maxPackPixel, scalePackFactor);
    
    purdueHeader.numCols = binCol;
    purdueHeader.numRows = binRow;
    purdueHeader.nimg    = binSec;
    packRadius = purdueHeader.numCols/2;
    stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		     packImg,purdueHeader.nimg, 1);
          
  }// Let's try converting mode==46 to PACKEDPORTABLE 
  
  // Global default 
  TextSet(imgNumTextW, "1", FALSE);
  
  if(pifMode==20) {
    imgNum = numCols/2; // Let's default to the middle section 
    purdueHeader.nimg = numSecs;  // Fool it to take our value 

    TextSetInt(imgNumTextW, imgNum+1, FALSE);
    if(useGUI) calcSF_work(calcSFs);
    
    // Now remember these forever 
    lowRes_buffer1 = float_image_buffer1;
    lowResCols = numCols;
    lowResRows = numRows;
    lowResSecs = numSecs;
    
    origWidthFromFile = numCols; 
  }

  purdueHeader.numCols = *ncol = numCols;
  purdueHeader.numRows = *nrow = numRows;
  purdueHeader.numSecs = *nsec = origNumSecs;
  
  xCen = numCols/2;
  yCen = numRows/2;
  zCen = origNumSecs/2;
  
  largestDim = MAX(numCols, numRows);
  largestDim = MAX(largestDim, origNumSecs);
  
  // We've got to make this larger for asymmetric maps 
  // and all maps that have information "outside" of the
  // normal sphere of our normal maps,
  // or else we just keep cliping the box into a sphere
  
  globMaxRadius = (int)(origNumSecs/2) - 1;
  globMinRadius = 0.0;
  
  free(rWrite.addr);
  
  purdueHeader.nimg = origNumSecs;

  if(useGUI == FALSE) return; // Done with GUI operations

  // Set scrolls on depthcue screen 
  // Give feedback 
  sprintf(outStr,"%.0f",globMinRadius);
  doValue(outStr, minRadiusValueW, FALSE);
  
  sprintf(outStr,"%.0f",globMaxRadius);
  doValue(outStr, maxRadiusValueW, FALSE);
  
  // Radius is origNumSecs/2 
  XtVaSetValues(maxRadiusScrollW, XmNsliderSize, 5, NULL);
  XtVaSetValues(maxRadiusScrollW, XmNmaximum, (int)((globMaxRadius)+5),NULL);
  XtVaSetValues(maxRadiusScrollW, XmNvalue, (int)(globMaxRadius),NULL);
  
  XtVaSetValues(minRadiusScrollW, XmNsliderSize, 5, NULL);
  XtVaSetValues(minRadiusScrollW, XmNmaximum, (int)((globMaxRadius)+5),NULL);
  XtVaSetValues(minRadiusScrollW, XmNvalue, (int)(globMinRadius), NULL);
  
  printf("MAX[%f]  MIN[%f]\n",maxMapPixel,minMapPixel);
  secMaxPixel = maxMapPixel;
  secMinPixel = minMapPixel;
  
  minSurf = MAX(minMapPixel,0.0);
  maxSurf = MAX(maxMapPixel,(0.0-minMapPixel));

  // We're doing the surfaceDensity in tenths now 
  XtVaSetValues(surfDensScrollW, XmNmaximum, maxSurf*10 + 50, NULL);
  XtVaSetValues(surfDensScrollW, XmNminimum, minSurf*10, NULL);

  if (pifMode == 46) {
    XtVaSetValues(surfDensScrollW, XmNsliderSize, 500, NULL);
    XtVaSetValues(surfDensScrollW, XmNpageIncrement, 500, NULL);
    XtVaSetValues(surfDensScrollW, XmNincrement, 100, NULL);
    XtVaSetValues(surfDensScrollW, XmNvalue, (int)(purdueHeader.img_bkg),NULL);
    }
  else if(maxSurf > 5000.0) {
    XtVaSetValues(surfDensScrollW, XmNsliderSize, (int)(maxSurf*0.01), NULL);
    XtVaSetValues(surfDensScrollW, XmNpageIncrement, (int)(maxSurf*0.01), NULL);
    XtVaSetValues(surfDensScrollW, XmNincrement, (int)(maxSurf*0.002), NULL);
    XtVaSetValues(surfDensScrollW, XmNvalue, 10, NULL);
  }
  else if(maxSurf > 500.0) {
    XtVaSetValues(surfDensScrollW, XmNsliderSize, 500, NULL);
    XtVaSetValues(surfDensScrollW, XmNpageIncrement, 500, NULL);
    XtVaSetValues(surfDensScrollW, XmNincrement, 100, NULL);
    XtVaSetValues(surfDensScrollW, XmNvalue, 10, NULL);
    }
  else {
    XtVaSetValues(surfDensScrollW, XmNsliderSize, 50, NULL);
    XtVaSetValues(surfDensScrollW, XmNpageIncrement, 50, NULL);
    XtVaSetValues(surfDensScrollW, XmNincrement, 10, NULL);
    XtVaSetValues(surfDensScrollW, XmNvalue, 10, NULL);  
  }
  surfaceDensity= 1.0;  // 1 tenth of 10 
  
  // Give feedback 
  doValue("-1", surfDensValueW, FALSE);
  doLabel("SecNum", imgNumLabelW, FALSE);
  
  // To make this work for now let's stuff the nimg field with the
  // number of sections.  Really we should change any and all references
  // to nimg to be origNumSecs. We really only work on one image at a time
  // so this may never have any affect.
    
  // Write stats 
  if (pifMode != 46)
    calcMapStats(float_image_buffer1,
		 &purdueHeader.img_bkg, 
		 &purdueHeader.map_min, &purdueHeader.map_max, 
		 &purdueHeader.img_avg, &purdueHeader.img_std,
		 &numCols, &numRows, &origNumSecs);
  
  // Write stats
  sprintf(outStr,"\nStatistics for file [%s]\n\
\n					      \
Title [%s]\n\
\n\
File size   [%d]x[%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
\n",
	  inFileName->addr,
	  title,
	  numCols,	  numRows, origNumSecs,
	  purdueHeader.img_bkg, 
	  minMapPixel, maxMapPixel,
	  purdueHeader.img_avg,	  purdueHeader.img_std
	  );
  TextSet(statTextW, outStr, TRUE);
  
} // end function read_pif



void write_pif(int ncol, int nrow, char *outFileName, int thisFileType)
{
  // function name: write_pif

  // Write image(s) in pif format. All PIF output from robem should
  // be written here.

  int writeIt=TRUE, noRectYet=TRUE, thisNimages=0, firstImg=0, zero=0, one=1;
  int mode, i, j, m, p, kk, ll, afterDeletingNumImages, numImages;
  int saveNimg, appendAndCurrentImgs, thisNcol, thisNrow;
  float *out_image_buffer, *fpointer;
  FILE *fOut;
  fString_t fName, rWrite;
  forPifHead_t thisHead;

  if(thisFileType==CIRCLEPACKED) // Check to see if we've got outliers
    checkAdjustBoxes();          // and move boxes if necessary

  appendAndCurrentImgs = purdueHeader.nimg;
  if(pifMode == 20)
    appendAndCurrentImgs = 1;
  else if(pifMode == 46)
    appendAndCurrentImgs = purdueHeader.nimg;
  else if(thisFileType==CIRCLEPACKED)
    appendAndCurrentImgs = numBoxes;
  
  thisHead.luNum = 4;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 4;
  fName.addr = malloc(sizeof(char)*264);
  fName.length = 264;
  sprintf(rWrite.addr,"r+");
  sprintf(fName.addr,"%s",outFileName);
  
  if(thisFileType==PACKEDPORTABLE) {
    // Find out how many we've got to write
    afterDeletingNumImages = purdueHeader.nimg;
    for (i=0; i<purdueHeader.nimg; i++)
      if(deletePackedImage[i])
	afterDeletingNumImages--;
  }
  
  if(thisFileType==PACKEDPORTABLE || thisFileType==CIRCLEPACKED ||
     thisFileType==RECTPACKED     || thisFileType==MAP          || 
     thisFileType==EMPORTABLE) {
    // PACKEDPORTABLE/emportable or MAP
    
    if(writeType==APPENDCIRCLEPACKED || writeType==APPENDRECT) {
      pif_open(fName.addr, &thisHead.luNum, rWrite.addr);
      
      pif_read_gh(&thisHead);
      thisHead.imgNum = 1;
      pif_read_dh(&thisHead);
      
      thisNcol    = thisHead.numCols;
      thisNrow    = thisHead.numRows;
      thisNimages = thisHead.nimg;
      appendAndCurrentImgs += thisNimages;
      mode        = thisHead.mode;
      firstImg    = thisHead.nimg;
      
      // Check for inconsistencies
      if(thisNcol != ncol || thisNrow != nrow) {
	sprintf(outStr, "Image sizes differ [%d][%d] in [%s] [%d][%d] other,",
		numCols, numRows, outFileName, thisNcol, thisNrow);
	askThem(outStr, XmDIALOG_INFORMATION);
	writeIt= FALSE;
      }
    
      // Now update the header to show new data at the end
      if(thisFileType==CIRCLEPACKED) {
	if(boxMethod==USECIRCLE)
	  purdueHeader.nimg = numBoxes;
	else if (boxMethod==USERECT || boxMethod==USESQUARE)
	  purdueHeader.nimg = 1;
      }
      else if(thisFileType == RECTPACKED || thisFileType != PACKEDPORTABLE)
	purdueHeader.nimg = 1;
      
      if(fileType==PACKEDPORTABLE)
	numImages = thisNimages + afterDeletingNumImages;
      else 
	numImages = thisNimages + purdueHeader.nimg;
      
      thisHead.nimg    = numImages;
      
            
      if(mode != 46) {
	printf("Can't append mode != 46 files!!\n");
	exit(0);
      }
      
      if(float_image_buffer1 == NULL)
	float_image_buffer1 = (float *)
	  malloc(sizeof(float)*appendAndCurrentImgs*
		 thisHead.numCols*thisHead.numRows);
      
      for (j=0; j<firstImg; j++) {
	int binFactor=1, pifMode=46, localBinSec=1;
	int binCol, binRow, thisImg;
	off_t imgOffset;
	recipFloatScaleFactor =  thisHead.recipFloatScaleFactor;
	thisImg = j + 1;  // For fortran
	thisHead.imgNum = thisImg;
	binCol          = thisHead.numCols;
	binRow          = thisHead.numRows;
	pif_read_dh(&thisHead);
	imgOffset  = thisHead.imgOffset = 512 + 512*(j+1) + (off_t) j*binCol*binRow*4;
	pif_read_map_int_image(&thisHead.luNum, 
			       float_image_buffer1 + j*binCol*binRow,
			       &imgOffset,
			       &binCol, &binRow, &localBinSec,
			       &recipFloatScaleFactor,
			       &minMapPixel,&maxMapPixel, &binFactor, 
			       &pifMode);
      }

      writeType = CIRCLEPACKED; // Just choose somethign else so
				// these data can be written as normal
      pif_close(&thisHead.luNum);
      
    }
    else if(!normLinSwitch && !autoBoxSwitch && !reBoxSwitch && 
	    !boxitSwitch && !doFix && !doInterp && !doNoise && !doTrunc) {

      // We always have the right filename for normLinSwitch
      // boxitSwitch doFix and autoBoxSwitch and doInterp Check to
      // make sure we're not overwriting it

      if ( (fOut = fopen(outFileName,"r")) != NULL ) {
	fclose(fOut);
	sprintf(outStr,
		"Do you really want to overwrite [%s]?",outFileName);
	writeIt= askThem(outStr, XmDIALOG_QUESTION);
      }
      fOut= NULL;
    }
    
    if(!writeIt)
      return;
      
    if(writeType==APPENDCIRCLEPACKED || writeType==APPENDRECT)
      sprintf(rWrite.addr,"r+");
    else {
      sprintf(rWrite.addr,"w");
      pif_init_head(&thisHead);
      thisHead.luNum = 4;
    }
    
    pif_open(fName.addr, &thisHead.luNum, rWrite.addr);
    
    
    // *******************************************
    //  WRIITING NEW GLOBAL PIF HEADER
    // *******************************************
    
    // [1] Write global file header
    
    // Pick appropriate statistics calculation based on data type
    
    if(thisFileType==CIRCLEPACKED) { // packed
      if(boxMethod==USECIRCLE) { // Circles
	stuffDisplayBoxBuffer(0, image_fileName.addr, 1);
	calcCircleStats(editBox_mirror_buffer,
			&purdueHeader.img_bkg, 
			&purdueHeader.img_min, &purdueHeader.img_max, 
			&purdueHeader.img_avg, &purdueHeader.img_std,
			ncol, nrow, BOX_WIDTH);  
      }
      else if(boxMethod==USERECT || boxMethod==USESQUARE) { // Rects
	CK(out_image_buffer = (float *) malloc(sizeof(float)*ncol*nrow));
	stuffRectangleBuffer(out_image_buffer, image_fileName.addr);
	
	calcStats(out_image_buffer, &purdueHeader.img_bkg, 
		  &purdueHeader.img_min, &purdueHeader.img_max, 
		  &purdueHeader.img_avg, &purdueHeader.img_std,
		  ncol, nrow, ncol, fileType, bkgrndMethod,
		  boxMethod, packRadius, boxRadius, featherRadius,
		  recipSizeFactor, tryColors);
	noRectYet = FALSE;
      }
    }
    else if(thisFileType==RECTPACKED) { // packed
      int saveFileType;
      CK(out_image_buffer = (float *)malloc(sizeof(float)*ncol*nrow));
      
      stuffRectangleBuffer(out_image_buffer, image_fileName.addr);
      
      saveFileType = fileType;
      fileType = RECTPACKED;
      calcStats(out_image_buffer, &purdueHeader.img_bkg, 
		&purdueHeader.img_min, &purdueHeader.img_max, 
		&purdueHeader.img_avg, &purdueHeader.img_std,
		ncol, nrow, ncol, fileType, bkgrndMethod,
		boxMethod, packRadius, boxRadius, featherRadius,
		recipSizeFactor, tryColors);
      fileType = saveFileType;
      noRectYet = FALSE;
    }
    else if(thisFileType==PACKEDPORTABLE) { // Packed Image file input
      int i, thisNsec=1;
      float img_bkg, img_min, img_max, img_avg, img_std;
      
      for (i=0; i<purdueHeader.nimg; i++) {
	sprintf(outStr,"Statistics calculating for image [%d].",i+1);
	checkStatus(outStr);
	calcMapStats(packImg[i], &img_bkg, 
		     &img_min, &img_max,  
		     &img_avg, &img_std,
		     &ncol, &nrow, &thisNsec);
	
	purdueHeader.img_bkg = MAX(purdueHeader.img_bkg, img_bkg);
	purdueHeader.img_min = MIN(purdueHeader.img_min, img_min);
	purdueHeader.img_max = MAX(purdueHeader.img_max, img_max);
	purdueHeader.img_avg = MAX(purdueHeader.img_avg, img_avg);
	purdueHeader.img_std = MAX(purdueHeader.img_std, img_std);
      }
      
      stuffDisplayPackBoxBuffer(0, 1, TRUE);
    }
    else if(thisFileType==MAP) {
      if(paintMapSwitch)
	calcMapStats(paintMap, &purdueHeader.img_bkg, 
		     &purdueHeader.map_min, &purdueHeader.map_max, 
		     &purdueHeader.img_avg, &purdueHeader.img_std,
		     &ncol, &nrow, &origNumSecs);
      else
	calcMapStats(float_image_buffer1, &purdueHeader.img_bkg, 
		     &purdueHeader.map_min, &purdueHeader.map_max, 
		     &purdueHeader.img_avg, &purdueHeader.img_std,
		     &ncol, &nrow, &origNumSecs);
    }
    else if(thisFileType==EMPORTABLE) {
      if(fileType==PORTABLE && pifMode==2) {
	origNumSecs = 1;
	calcMapStats(inCore_buffer, &purdueHeader.img_bkg, 
		     &purdueHeader.map_min, &purdueHeader.map_max, 
		     &purdueHeader.img_avg, &purdueHeader.img_std,
		     &ncol, &nrow, &origNumSecs);
      }
      else
	calcStats(inCore_buffer, &purdueHeader.img_bkg, 
		  &purdueHeader.img_min, &purdueHeader.img_max, 
		  &purdueHeader.img_avg, &purdueHeader.img_std,
		  ncol, nrow, ncol, fileType, bkgrndMethod,
		  boxMethod, packRadius, boxRadius, featherRadius,
		  recipSizeFactor, tryColors);
    }	      
    
    if(thisFileType==MAP)
      calcScaleFactor(&floatScaleFactor, &recipFloatScaleFactor, &int4scaleF,
		      purdueHeader.map_min-1.0, purdueHeader.map_max+1.0,
		      purdueHeader.img_avg, purdueHeader.img_std,
		      thisFileType);
    else if(fileType==PORTABLE && pifMode==2)
      calcScaleFactor(&floatScaleFactor, &recipFloatScaleFactor, &int4scaleF, 
		      purdueHeader.map_min-1.0, purdueHeader.map_max+1.0,
		      purdueHeader.img_avg, purdueHeader.img_std,
		      fileType);
    else
      calcScaleFactor(&floatScaleFactor, &recipFloatScaleFactor, &int4scaleF, 
		      purdueHeader.img_min, purdueHeader.img_max,
		      purdueHeader.img_avg, purdueHeader.img_std,
		      thisFileType);
    
    thisHead.recipFloatScaleFactor = recipFloatScaleFactor;
    thisHead.floatScaleFactor      = floatScaleFactor;
    thisHead.int4scaleF            = int4scaleF;
    thisHead.pixelSize             = pifHeader.pixelSize;	      
    thisHead.pixSizUnits           = pifHeader.pixSizUnits;	      
    
    // purdueHeader.nimg already knows how many we've got for
    // PACKEDPORTABLE except if some have been deleted
    
    if(thisFileType==CIRCLEPACKED) {
      if(boxMethod==USECIRCLE)
	purdueHeader.nimg = numBoxes;
      else if (boxMethod==USERECT || boxMethod==USESQUARE)
	purdueHeader.nimg = 1;
    }
    else if(thisFileType==RECTPACKED || thisFileType==MAP || 
	    thisFileType==EMPORTABLE || thisFileType != PACKEDPORTABLE)
      purdueHeader.nimg = 1;
    
    if(thisFileType == PACKEDPORTABLE)
      saveNimg = afterDeletingNumImages;
    else 
      saveNimg = purdueHeader.nimg;
    
    thisHead.nimg = saveNimg + firstImg;
    thisHead.htype= 1;
    
    if(fileType==MAP) {
      mrc_nx = fileBinCols;
      mrc_ny = fileBinRows;
      mrc_nz= fileBinSecs;
    }
    else {
      mrc_nx = ncol;
      mrc_ny = nrow;
      mrc_nz = 1;
    }
    
    thisHead.numCols = mrc_nx;
    thisHead.numRows = mrc_ny;
    thisHead.numSecs = mrc_nz;
    
    thisHead.even = fileMrcEven;
    thisHead.mrcX = fileMrcX;
    thisHead.mrcY = fileMrcY;
    thisHead.mrcZ = fileMrcZ;
    
    if(thisFileType==PACKEDPORTABLE || thisFileType==CIRCLEPACKED) {
      if(pifMode == 1 || pifMode==2   || pifMode==6 ||
	 pifMode == 7 || pifMode== 46 || fileType==BYTEPORTABLE)
	mode = 46;
      else
	mode= 7;
      if(forceByte)
	mode = 6;
    }
    else if(thisFileType==RECTPACKED)
      mode = 6;
    else if(thisFileType==EMPORTABLE) {
      if(fileType==PORTABLE && pifMode==2)
	mode = 2;
      else if(pifMode==1)
	mode = 1;
      else 
	mode = 0;
    }
    else if(thisFileType==MAP)
      mode = 20;
    
    thisHead.mode = mode;
    
    pif_write_gh(&thisHead);
    
    // ********************************************
    // 
    //  WRITE EACH Individually
    // 
    // ********************************************
    
    if(mode==46 && float_image_buffer1==NULL)
      CK(float_image_buffer1 = (float *)
	 malloc(sizeof(float)*appendAndCurrentImgs*
		thisHead.numCols*thisHead.numRows));
    
    saveNimg = -1;
    
//    printf("first [%d] nimg[%d]  tot[%d]\n",
//	   firstImg,purdueHeader.nimg,thisHead.nimg);
    
    for (m=firstImg; m<appendAndCurrentImgs; m++) {
      
      sprintf(outStr,"Processing image [%d] of [%d] total.",
	      m+1,thisHead.nimg);
      checkStatus(outStr);
      
      if((fileType==PACKEDPORTABLE && !deletePackedImage[m-firstImg]) 
	 || fileType!=PACKEDPORTABLE) { // OK to write
	saveNimg++;
	if(thisFileType==CIRCLEPACKED) { // packed
	  if(boxMethod==USECIRCLE) { // Circles
	    stuffDisplayBoxBuffer(m-firstImg, 
				  image_fileName.addr, 1);
	    calcCircleStats(editBox_mirror_buffer, &purdueHeader.img_bkg, 
			    &purdueHeader.img_min, &purdueHeader.img_max, 
			    &purdueHeader.img_avg, &purdueHeader.img_std,
			    ncol, nrow, BOX_WIDTH);
	    
	  }
	  else if((boxMethod==USERECT || boxMethod==USESQUARE) && noRectYet) { // Rects
	    out_image_buffer = (float *)malloc(sizeof(float)*ncol*nrow);
	    stuffRectangleBuffer(out_image_buffer, image_fileName.addr);
	    
	    calcStats(out_image_buffer, &purdueHeader.img_bkg, 
		      &purdueHeader.img_min, &purdueHeader.img_max, 
		      &purdueHeader.img_avg, &purdueHeader.img_std,
		      ncol, nrow, ncol, fileType, bkgrndMethod,
		      boxMethod, packRadius, boxRadius, featherRadius,
		      recipSizeFactor, tryColors);
	  }
	}
	else if(thisFileType==RECTPACKED && noRectYet) { // packed
	  out_image_buffer = (float *)malloc(sizeof(float)*ncol*nrow);
	  stuffRectangleBuffer(out_image_buffer, image_fileName.addr);
	  
	  calcStats(out_image_buffer, &purdueHeader.img_bkg, 
		    &purdueHeader.img_min, &purdueHeader.img_max, 
		    &purdueHeader.img_avg, &purdueHeader.img_std,
		    ncol, nrow, ncol, fileType, bkgrndMethod,
		    boxMethod, packRadius, boxRadius, featherRadius,
		    recipSizeFactor, tryColors);
	}
	else if(thisFileType==PACKEDPORTABLE) { // Packed file input
	  stuffDisplayPackBoxBuffer(m-firstImg, 1, TRUE);
	  calcCircleStats(editBox_mirror_buffer, &purdueHeader.img_bkg, 
			  &purdueHeader.img_min, &purdueHeader.img_max, 
			  &purdueHeader.img_avg, &purdueHeader.img_std,
			  ncol, nrow, BOX_WIDTH);
	}
	else if(thisFileType==MAP) { // MAP file input
	  if(paintMapSwitch)
	    calcMapStats(paintMap, &purdueHeader.img_bkg, 
			 &purdueHeader.map_min, &purdueHeader.map_max, 
			 &purdueHeader.img_avg, &purdueHeader.img_std,
			 &ncol, &nrow, &origNumSecs);
	  else
	    calcMapStats(float_image_buffer1, &purdueHeader.img_bkg, 
			 &purdueHeader.map_min, &purdueHeader.map_max, 
			 &purdueHeader.img_avg, &purdueHeader.img_std,
			 &ncol, &nrow, &origNumSecs);
	}
	else if(thisFileType==EMPORTABLE) { // IMG file input
	  calcStats(inCore_buffer, &purdueHeader.img_bkg, 
		    &purdueHeader.img_min, &purdueHeader.img_max, 
		    &purdueHeader.img_avg, &purdueHeader.img_std,
		    ncol, nrow, ncol, fileType, bkgrndMethod,
		    boxMethod, packRadius, boxRadius, featherRadius,
		    recipSizeFactor, tryColors);
	}
	
	// [2] Write header for each image
	
	if(fileType==MAP) { // Special
	  mrc_nx = fileBinCols;
	  mrc_ny = fileBinRows;
	}
	else { // Normal
	  mrc_nx = ncol;
	  mrc_ny = nrow;
	}
	
	if (fileType==MAP)
	  mrc_nz = fileBinSecs;
	else 
	  mrc_nz = 1;
	
	thisHead.numCols = mrc_nx;
	thisHead.numRows = mrc_ny;
	thisHead.numSecs = mrc_nz;
	thisHead.mode    = mode;
	
	thisHead.img_bkg    = 0;
	thisHead.packRadius = 0;
	
	if(thisFileType==CIRCLEPACKED) {
	  thisHead.img_bkg = purdueHeader.img_bkg;
	  thisHead.packRadius = featherRadius;
	}
	else if(thisFileType==RECTPACKED) {
	  thisHead.img_bkg = purdueHeader.img_bkg;
	  thisHead.packRadius = boxRadius;
	  thisHead.packRadius = -99; // Make it so we can write out new
	}
	else if(thisFileType==PACKEDPORTABLE) {
	  thisHead.img_bkg = purdueHeader.img_bkg;
	  thisHead.packRadius = packRadius;
	}
	
	thisHead.datai4_max = purdueHeader.img_max;
	thisHead.datar4_max = purdueHeader.img_max;
	thisHead.datai4_min = purdueHeader.img_min;
	thisHead.datar4_min = purdueHeader.img_min;
	thisHead.datar4_avg = purdueHeader.img_avg;
	thisHead.datar4_std = purdueHeader.img_std;
	
	thisHead.aoverb     = aoverb;
	thisHead.map_abang  = map_abang;
	thisHead.dela       = dela;
	thisHead.delb       = delb;
	thisHead.delc       = delc;
	thisHead.dthe       = dthe;
	thisHead.dphi_90    = dphi_90;
	thisHead.symmetry   = symmetry;
	thisHead.borderType = getBorderType(featherMode);
	
	// Carry through the original borderType
	if(fileType==PACKEDPORTABLE && thisHead.borderType == 0)
	  thisHead.borderType = borderType;
	thisHead.borderWidth = featherRadius-boxRadius;
	// We need to capture how far in the boxRadius is 
	// from the edge
	thisHead.borderWidth = thisHead.packRadius-boxRadius;
	
	thisHead.binFactor = (fileType==MAP) ? fileBinFactor : binFactor;
	
	// If we're appending we need to offset that here
	// with thisNimages - also we may not write ALL
	// of the images because they've "deleted" them
	// - use saveNimg for that.
	
	thisHead.imgNum = thisNimages+saveNimg+1;
	
	thisHead.transX = pifHeader.transX;
	thisHead.transY = pifHeader.transY;
	thisHead.transZ = pifHeader.transZ;
	thisHead.isTrunc = pifHeader.isTrunc;
	thisHead.origX = pifHeader.origX;
	thisHead.origY = pifHeader.origY;
	thisHead.origZ = pifHeader.origZ;
	
	pif_write_dh(&thisHead);
	thisHead.imgOffset += 512;  // Just wrote the header
	
	// Now call pif_write_?????_image() after we rework the
	// buffers so it's in Tim's row order.
	
	// [3] Now write the image info
	if(thisFileType==CIRCLEPACKED) { // packed
	  
	  // We keep these packed images on the disk in Tim's x,y
	  // orientation (e.g. drawn from bottom up so if we ever
	  // convert to .PCK format it's already in the right space.
	  
	  if(boxMethod==USECIRCLE) { // Circles
	    BYTE *tmpByte_buffer;
	    short *tmpShort_buffer;
	    float *tmpInt_buffer;
	    
	    CK(tmpByte_buffer=(BYTE *)malloc(sizeof(BYTE)*ncol*nrow));
	    CK(tmpInt_buffer=(float *)malloc(sizeof(float)*ncol*nrow));
	    CK(tmpShort_buffer=(short *)
	       malloc(sizeof(short)*ncol*nrow));
	    
	    tsbworld_float_box(editBox_mirror_buffer, tmpInt_buffer,
			       BOX_WIDTH, ncol, nrow);
	    
	    for (p=0; p<ncol*nrow; p++)
	      if(mode==6)
		tmpByte_buffer[p] = (BYTE)(tmpInt_buffer[p]);
	      else
		tmpShort_buffer[p] = (short)(tmpInt_buffer[p]);
	    
	    if(mode==6)
	      pif_write_byte_image(&thisHead.luNum, tmpByte_buffer,
				   &thisHead.imgOffset, &thisHead.numCols,
				   &thisHead.numRows, &thisHead.numSecs);
	    else { // Save for later
	      for (ll=0; ll<thisHead.numRows; ll++)
		for(kk=0; kk<thisHead.numCols; kk++)
		  float_image_buffer1[m*thisHead.numCols*thisHead.numRows +
				      ll*thisHead.numCols + kk]
		    = tmpInt_buffer[ll*thisHead.numCols + kk];  
	    }
	    
	    free(tmpByte_buffer);
	    free(tmpShort_buffer);
	    free(tmpInt_buffer);
	  }
	  else if (boxMethod==USERECT || boxMethod==USESQUARE) { // rects
	    BYTE *tmpByte_buffer;
	    float min, max;
	    CK(tmpByte_buffer=(BYTE *)malloc(sizeof(BYTE)*ncol*nrow));
	    // Do nothing
	    
	    farray_minmax(out_image_buffer, &min, &max, ncol*nrow);
	    
	    for (p=0; p<ncol*nrow; p++)
	      tmpByte_buffer[p] = (BYTE)(out_image_buffer[p]);
	    
	    pif_write_byte_image(&thisHead.luNum, tmpByte_buffer,
				 &thisHead.imgOffset, &thisHead.numCols,
				 &thisHead.numRows, &thisHead.numSecs);
	    
	    free(out_image_buffer);
	    free(tmpByte_buffer);
	  }
	}
	else if(thisFileType==RECTPACKED) { // packed
	  BYTE *tmpByte_buffer;
	  short *tmpShort_buffer;
	  float *tmpInt_buffer, min, max;
	  CK(tmpByte_buffer=(BYTE *)malloc(sizeof(BYTE)*ncol*nrow));
	  CK(tmpShort_buffer=(short *)
	     malloc(sizeof(short)*ncol*nrow));
	  CK(tmpInt_buffer=(float *)malloc(sizeof(float)*ncol*nrow));
	  tsbworld_float_box(out_image_buffer, tmpInt_buffer,
			     ncol, ncol, nrow);
	  
	  farray_minmax(tmpInt_buffer, &min, &max, ncol*nrow);
	  
	  for (p=0; p<ncol*nrow; p++)
	    if(mode==6)
	      tmpByte_buffer[p] = (BYTE)(tmpInt_buffer[p]);
	    else
	      tmpShort_buffer[p] = (short)(tmpInt_buffer[p]);
	  
	  if(mode==6)
	    pif_write_byte_image(&thisHead.luNum,     tmpByte_buffer,
				 &thisHead.imgOffset, &thisHead.numCols,
				 &thisHead.numRows,   &thisHead.numSecs);
	  else
	    pif_write_short_image(&thisHead.luNum,     tmpShort_buffer,
				  &thisHead.imgOffset, &thisHead.numCols,
				  &thisHead.numRows,   &thisHead.numSecs);
	  
	  free(tmpByte_buffer);
	  free(tmpShort_buffer);
	  free(tmpInt_buffer);
	  free(out_image_buffer);
	}
	else if(thisFileType==PACKEDPORTABLE) { // packed file input
	  BYTE *tmpByte_buffer;
	  short *tmpShort_buffer;
	  float *tmpFloat_buffer, *tmpInt_buffer, min, max, scaleF;
	  
	  CK(tmpShort_buffer=(short *)
	     malloc(sizeof(short)*ncol*nrow));
	  CK(tmpByte_buffer=(BYTE *)malloc(sizeof(BYTE)*ncol*nrow));
	  CK(tmpInt_buffer=(float *)malloc(sizeof(float)*ncol*nrow));
	  CK(tmpFloat_buffer=(float *)malloc(sizeof(float)*ncol*nrow));
	  
	  tsbworld_float_box(editBox_mirror_buffer, tmpInt_buffer,
			     BOX_WIDTH, ncol, nrow);
	  
	  farray_minmax(tmpInt_buffer, &min, &max, ncol*nrow);
	  
	  if(mode==6 || mode==46)
	    scaleF = 1.0;
	  else
	    scaleF = 32000.0 / max;
	  
	  for (p=0; p<ncol*nrow; p++)
	    if(mode==6)
	      tmpByte_buffer[p] = (BYTE)(tmpInt_buffer[p]*scaleF);
	    else if(mode==46)
	      tmpFloat_buffer[p] = tmpInt_buffer[p];
	    else
	      tmpShort_buffer[p] = (short)(tmpInt_buffer[p]*scaleF);
	  
	  if(mode==6)
	    pif_write_byte_image(&thisHead.luNum,     tmpByte_buffer,
				 &thisHead.imgOffset, &thisHead.numCols,
				 &thisHead.numRows,   &thisHead.numSecs);
	  else if(mode==46) {
	    for (ll=0; ll<thisHead.numRows; ll++)
	      for (kk=0; kk<thisHead.numCols; kk++) 
		float_image_buffer1[(m * 
				     thisHead.numCols * 
				     thisHead.numRows) +
				    (ll*thisHead.numCols)+kk] = 
		  tmpFloat_buffer[(ll*thisHead.numCols)+kk];
	    
	  }
	  else
	    pif_write_short_image(&thisHead.luNum,     tmpShort_buffer,
				  &thisHead.imgOffset, &thisHead.numCols,
				  &thisHead.numRows,   &thisHead.numSecs);
	  free(tmpByte_buffer);
	  free(tmpShort_buffer);
	  free(tmpInt_buffer);
	  free(tmpFloat_buffer);
	}
	else if(thisFileType==MAP) { // Do nothing for now
	  float *tmp_float_buffer;
	  int saveBinFactor, saveNC, saveNR, saveNS;
	  
	  CK(tmp_float_buffer=(float *)
	     malloc(sizeof(float)*numCols*numRows*origNumSecs));
	  
	  if(paintMapSwitch)
	    tsbworld_float(paintMap, tmp_float_buffer,
			   numCols, numRows, origNumSecs);
	  else
	    tsbworld_float(float_image_buffer1, tmp_float_buffer,
			   numCols, numRows, origNumSecs);
	  
	  // Suppose we've got a bin==2 map here we've read in.  The
	  // header contains the original rehydrated
	  // numcols/rows/secs.  SO just write it like it's bin==1
	  // with appropriate numcols/rows/secs
	  
	  saveBinFactor = thisHead.binFactor;
	  saveNC = thisHead.numCols;
	  saveNR = thisHead.numRows;
	  saveNS = thisHead.numSecs;
	  thisHead.numCols = numCols;
	  thisHead.numRows = numRows;
	  thisHead.numSecs = origNumSecs;
	  thisHead.binFactor = 1;
	  
	  pif_write_map_sections(&thisHead, tmp_float_buffer, 
				 &(thisHead.mode),
				 &numCols, &numRows, &origNumSecs, 
				 &zero, &origNumSecs);
	  
	  thisHead.binFactor = saveBinFactor;
	  thisHead.numCols = saveNC;
	  thisHead.numRows = saveNR;
	  thisHead.numSecs = saveNS;
	  free(tmp_float_buffer);
	  
	}
	else if(thisFileType==EMPORTABLE) { // Do nothing for now
	  float min,max,scaleF;
	  int m,ncol,nrow;
	  ncol = thisHead.numCols;
	  nrow = thisHead.numRows;
	  
	  farray_minmax(inCore_buffer, &min, &max, nrow*ncol);
	  
	  if(fileType==PORTABLE)
	    scaleF = (max > 32000.0) ? 32000.0/max : 1.0;
	  else
	    scaleF = 255.0 / max;
	  
	  thisHead.floatScaleFactor = scaleF;
	  thisHead.recipFloatScaleFactor = 1.0/scaleF;
	  pif_write_gh(&thisHead);
	  
	  if(fileType==PORTABLE && pifMode == 2) { // floatInt*4
	    pif_write_map_sections(&thisHead, inCore_buffer, &pifMode,
				   &thisHead.numCols, &thisHead.numRows, 
				   &thisHead.numSecs, &zero,
				   &thisHead.numSecs);
	  }
	  else if(fileType==PORTABLE) {
	    short *short_buffer = (short *) 
	      malloc(sizeof(short)*ncol*nrow);
	    for (m=0; m<nrow*ncol; m++)
	      short_buffer[m] = (short)(inCore_buffer[m]*scaleF);
	    pif_write_short_image(&thisHead.luNum,     short_buffer,
				  &thisHead.imgOffset, &thisHead.numCols,
				  &thisHead.numRows,   &thisHead.numSecs);
	    free(short_buffer);
	  }
	  else if(fileType==BYTEPORTABLE) {
	    BYTE *byte_buffer = (BYTE *)malloc(sizeof(BYTE)*ncol*nrow);
	    for (m=0; m<nrow*ncol; m++)
	      byte_buffer[m] = (BYTE)(inCore_buffer[m]*scaleF);
	    pif_write_byte_image(&thisHead.luNum,     byte_buffer,
				 &thisHead.imgOffset, &thisHead.numCols,
				 &thisHead.numRows,   &thisHead.numSecs);
	    free(byte_buffer);
	  }
	}
	
      }
      
    }
    
    if(mode==46) {
      // Rewrite the header for mode 46 to get floatScaelFactor correct
      int i, totWritten;
      
      if(origNumSecs == 0)
	origNumSecs = thisHead.nimg;
      
      // We calculated the scalefactors above  - we may have 
      // underestimated some outliers here ???

      calcMapStats(float_image_buffer1,   &purdueHeader.img_bkg, 
		   &purdueHeader.map_min, &purdueHeader.map_max, 
		   &purdueHeader.img_avg, &purdueHeader.img_std,
		   &ncol, &nrow, &thisHead.nimg);
      
      calcScaleFactor(&floatScaleFactor, &recipFloatScaleFactor, &int4scaleF,
		      purdueHeader.map_min-1.0, purdueHeader.map_max+1.0,
		      purdueHeader.img_avg, purdueHeader.img_std, fileType);

      thisHead.floatScaleFactor = floatScaleFactor;
      thisHead.recipFloatScaleFactor = recipFloatScaleFactor;
      thisHead.int4scaleF = int4scaleF;
      
      pif_write_gh(&thisHead);
      
      thisHead.datai4_max = purdueHeader.img_max;
      thisHead.datar4_max = purdueHeader.img_max;
      thisHead.datai4_min = purdueHeader.img_min;
      thisHead.datar4_min = purdueHeader.img_min;
      thisHead.datar4_avg = purdueHeader.img_avg;
      thisHead.datar4_std = purdueHeader.img_std;
      thisHead.imgNum = 1;
      
      totWritten = 0;

      for (i=0; i<appendAndCurrentImgs; i++) {
	// Just blow out the data - we've already stuffed the headers
	if(i < firstImg || !deletePackedImage[i-firstImg]) {
	  // Only do ones we want */
	  sprintf(outStr,"Writing image [%d] of [%d] total.",
		  totWritten+1,thisHead.nimg);
	  checkStatus(outStr);
	  // thisHead.imgNum = thisNimages + totWritten + 1;
	  thisHead.imgNum = totWritten + 1;
	  calcCircleStats((float_image_buffer1+
			   (i*(thisHead.numCols*thisHead.numRows))),
			  &purdueHeader.img_bkg, 
			  &purdueHeader.img_min,
			  &purdueHeader.img_max, 
			  &purdueHeader.img_avg, 
			  &purdueHeader.img_std,
			  ncol, nrow, ncol);

	  thisHead.img_bkg   = purdueHeader.img_bkg;
	  thisHead.datai4_max= purdueHeader.img_max;
	  thisHead.datar4_max= purdueHeader.img_max;
	  thisHead.datai4_min= purdueHeader.img_min;
	  thisHead.datar4_min= purdueHeader.img_min;
	  thisHead.datar4_avg= purdueHeader.img_avg;
	  thisHead.datar4_std= purdueHeader.img_std;
	  thisHead.origX     = box[i].x;
	  thisHead.origY     = box[i].y;
	  pif_write_dh(&thisHead);
	  fpointer = float_image_buffer1 + i*thisHead.numCols*thisHead.numRows;
	  pif_write_map_sections(&thisHead, fpointer,
				 &(thisHead.mode), &(thisHead.numCols), 
				 &(thisHead.numRows), &(thisHead.numSecs), 
				 &zero, &one);
	  totWritten ++ ;
	}
      }
      
    }
      
    pif_close(&thisHead.luNum);
    
    if(mode == 46)
      freeNullFloat(&float_image_buffer1);
    
  } // PACKEDPORTABLE/emportable or MAP
  
  // ****************************************
  //   Depth Cued images                    *
  // ****************************************

  else if(thisFileType==DEPTHCUED && shadedMaps) { // DEPTHCUED
    
    if ( (fOut = fopen(outFileName,"r")) != NULL ) {
      fclose(fOut);
      sprintf(outStr, "Do you really want to overwrite [%s]?", outFileName);
      writeIt= askThem(outStr, XmDIALOG_QUESTION);
    }
    fOut= NULL;
    
    if(!writeIt)
      return;
      
    sprintf(rWrite.addr,"w");     // Write it
    pif_init_head(&thisHead);
    thisHead.luNum = 5;
    
    pif_open(fName.addr, &thisHead.luNum, rWrite.addr);
    
    // [1] Write global file header
    stuffDisplayMapBuffer(0,1,0);
    calcStats(editBox_int_buffer, &purdueHeader.img_bkg, 
	      &purdueHeader.img_min, &purdueHeader.img_max, 
	      &purdueHeader.img_avg, &purdueHeader.img_std,
	      ncol, nrow, BOX_WIDTH, fileType, bkgrndMethod,
		  boxMethod, packRadius, boxRadius, featherRadius,
		  recipSizeFactor, tryColors);
    
    calcScaleFactor(&floatScaleFactor, &recipFloatScaleFactor, &int4scaleF,
		    purdueHeader.img_min, purdueHeader.img_max,
		    purdueHeader.img_avg, purdueHeader.img_std,
		    thisFileType);
    thisHead.recipFloatScaleFactor = recipFloatScaleFactor;
    thisHead.int4scaleF = int4scaleF;
    
    thisHead.nimg = purdueHeader.nimg = mapNumSecs;
    
    thisHead.htype = 1;
    thisHead.numCols = mrc_nx = ncol;
    thisHead.numRows = mrc_ny = nrow;
    thisHead.numSecs = mrc_nz = purdueHeader.nimg;
    thisHead.mode = mode = DEPTHCUED;
    
    pif_write_gh(&thisHead);
    
    for (m=0; m<purdueHeader.nimg; m++) {
      stuffDisplayMapBuffer(m,1,0);
      calcStats(editBox_int_buffer,    &purdueHeader.img_bkg, 
		&purdueHeader.img_min, &purdueHeader.img_max, 
		&purdueHeader.img_avg, &purdueHeader.img_std,
		ncol, nrow, BOX_WIDTH, fileType, bkgrndMethod,
		boxMethod, packRadius, boxRadius, featherRadius,
		recipSizeFactor, tryColors);
      
      // [2] Write header for each image
      thisHead.numCols = mrc_nx = ncol;
      thisHead.numRows = mrc_ny = nrow;
      thisHead.numSecs = mrc_nz = purdueHeader.nimg;
      
      thisHead.img_bkg = purdueHeader.img_bkg;
      thisHead.datai4_min = purdueHeader.img_min;
      thisHead.datai4_max = purdueHeader.img_max;
      thisHead.datar4_min = purdueHeader.img_min;
      thisHead.datar4_max = purdueHeader.img_max;
      thisHead.datar4_avg = purdueHeader.img_avg;
      thisHead.datar4_std = purdueHeader.img_std;
      thisHead.borderType = getBorderType(featherMode);
      thisHead.borderWidth = featherRadius-boxRadius;
      
      thisHead.imgNum = 1;
      pif_write_dh(&thisHead);
      
      // [3] Now write the image info
      pif_write_byte_image(&thisHead.luNum,     editBox_display_buffer,
			   &thisHead.imgOffset, &thisHead.numCols,
			   &thisHead.numRows,   &thisHead.numSecs);
      
    }
    pif_close(&thisHead.luNum);
  }
  
} // end function write_pif


// GCA: apparently this routine is not used anywhere
int calcShortBackground(unsigned short *inBuf, int width, int height, 
			int boxWidth )
{
  // function name: calcShortBackground
  // Calculate statistics about the image. Used by the
  // stuffDisplayBoxBuffer() routine(s) to mask the background of the
  // boxed images.  Note that this rtn throws out data > 2.0 STDev .
  
  int i, j, totPnts, xcen, ycen, iMycen, jMxcen, thisRadius;
  float absTmp, std, avg, var, distnce, cutOff, tmp;
  float back=0.0, sumSq=0.0;

  // Init the background value
  thisRadius = (fileType==PACKEDPORTABLE) ? packRadius : boxRadius;

  if(bkgrndMethod==USEAVGPIXEL) { // Average it
    if(fileType != PORTABLE &&
       (boxMethod==USECIRCLE && fileType!=EMPORTABLE) &&
       (width==height)) { // Calc for circle boxes
      xcen = ycen = thisRadius;
      totPnts = 0;

      for (i=0; i<height; i++) {
	for (j=0; j<width; j++) {
	  iMycen = i*recipSizeFactor - ycen;
	  jMxcen = j*recipSizeFactor - xcen;
	  distnce = sqrt((double)(iMycen*iMycen + jMxcen*jMxcen)); 
	  if(distnce>=thisRadius && (distnce-1.0)<= thisRadius) {
	    tmp  =  inBuf[(i*boxWidth)+j];
	    back += tmp;
	    totPnts++;
	    sumSq += (tmp*tmp);
	  }
	}
      }
      
      std = 0.0;
      avg = back/totPnts;
      var = sumSq/totPnts - avg*avg;
      if(var >= 0.0) 
	std = sqrt(var);
      
      back    = 0.0;
      totPnts = 0;
      sumSq   = 0.0;
      cutOff= avg + (2.0*std);

      // Throw out data points more than 2*STD from AVG      
      for (i=0; i<height; i++) {
	for (j=0; j<width; j++) {
	  tmp = inBuf[(i*boxWidth)+j];
	  iMycen = (i*recipSizeFactor)-ycen;
	  jMxcen = (j*recipSizeFactor)-xcen;
	  distnce = sqrt((double)(iMycen*iMycen + jMxcen*jMxcen)); 
	  if(distnce>=thisRadius && (distnce-1.0)<= thisRadius) {
	    if(tmp < avg)
	      absTmp = avg + (avg - (float)tmp);
	    else 
	      absTmp = (float)tmp;
	    if(absTmp <= cutOff) {
	      back   += tmp;
	      totPnts++;
	      sumSq  += (tmp*tmp);
	    }
	  }
	}
      }
      
      std = 0.0;
      avg = back/totPnts;
      var = sumSq/totPnts - avg*avg;
      if(var >= 0.0) 
	std = sqrt(var);
      
      back = (int)(avg);
    }
    else { // rectangles

      back = 0.0;
      for (i=0; i<width; i++) {
	back += inBuf[i];	                  // bottom row
	back += inBuf[((height-1)*boxWidth)+i];    // top row
      }
      for(i=1; i<(height-2); i++) {
	back += inBuf[i*boxWidth];                 // leftColumn
	back += inBuf[(i*boxWidth)+(height-1)];    // rightColumn
      }

      // Compute the average background pixel
      back    = back/((2*width)+(2*height)-2);
      
      // We're challenged here with calculating the background for 
      // a boxed particle saved as an "IMG" file.  Kludge for now
      // until we can get convert to pass the background from the 
      // header to us
      
    }
  }
  else if(bkgrndMethod==USERANDPIXEL) // Pass back random value
    back = (tryColors-7) * random() * (1.0/RAND_MAX);
  
  return((int)back);
  
} // end function calcShortBackground



void calcBoxSizeFactor(int thisZoomFactor)
{
  // function name: calcBoxSizeFactor
  // Calculate how much we have to correct for based on how zoomed we
  // are in the image. Used for display purposes.

  int k;

  sizeFactor= 1.0;

  if(thisZoomFactor >=1)
    for (k=1; k<thisZoomFactor; k++)
      sizeFactor *= 2.0;
  else
    for (k=thisZoomFactor; k<1; k++)
      sizeFactor *= 0.5;

  recipSizeFactor = 1.0 / sizeFactor ;

} // end function calcBoxSizeFactor



void calcSizeFactor(float thisZoomFactor)
{
  // function name: calcSizeFactor
  // Calculate how much we have to correct for based on how zoomed we
  // are in the image. Used for display purposes.
 
  if(fullScreen) { // Size it for full zoom
    float scaleX=1.0, scaleY=1.0; // Default values
    if(file_width > 0 && file_height > 0) { 
      scaleX = (float)display_width  / (float)file_width;
      scaleY = (float)display_height / (float)file_height;
    }
    sizeFactor = MIN(scaleX, scaleY);
    zoomFactor = rwalog2(sizeFactor);
  }
  else
    sizeFactor = pow(2.0,thisZoomFactor);
  
  recipSizeFactor = 1.0 / sizeFactor ;

} // end function calcSizeFactor



void evalSwitches(int argc, char **argv_mc)
{
  // function name: evalSwitches
  // Evaluate any switches they may have thrown from the command line.

  // argv_mc contains the original mixed case command line args
  // argv_lc is same as argv_mc except converted to lower case
  
  char **argv_lc;
  int i, j, len;
  
  // Make a lowercase copy of argv_mc - used in simplified tests
  argv_lc = (char **) malloc(sizeof(char *)*argc);
  for (i=0; i<argc; i++) {
    len = strlen(argv_mc[i]) + 1;
    argv_lc[i] = (char *) malloc(sizeof(char)*len);
    for (j=0; j<len; j++) {
      argv_lc[i][j] = tolower(argv_mc[i][j]);
    }
  }

  // Analyze switches
  scaleItSwitch     = FALSE;
  liteSwitch        = FALSE;
  shiftFFT          = FALSE;
  normLinSwitch     = FALSE;
  reBoxSwitch       = FALSE;
  autoBoxSwitch     = FALSE;
  boxitSwitch       = FALSE;
  doLinAlso         = FALSE;
  doFix             = FALSE;
  doNoise           = FALSE;
  doInterp          = FALSE;
  doTrunc           = FALSE;
  doOutFile         = FALSE;
  dumpNormStats     = FALSE;
  dumpSpherAvg      = FALSE;
  forceAvgBkgSwitch = FALSE;
  useGUI            = TRUE;
  
  for (i=0; i<argc; i++) {
    if(*(argv_lc[i]) != '-') // Looking for command line flags
      continue;
    
    if(strstr(argv_lc[i], "-scaleit") != 0)
      scaleItSwitch= TRUE;
    else if(strstr(argv_lc[i], "-nogui") != 0)
      useGUI = FALSE;
    else if(strstr(argv_lc[i], "-lite") != 0)
      liteSwitch = TRUE;
    else if(strstr(argv_lc[i], "-forceavgbkg") != 0)
      forceAvgBkgSwitch = TRUE;
    else if(strstr(argv_lc[i], "-rmblem") != 0) {
      doFix = TRUE;
      batchFixStd = 0.0;
      if((i+1)<argc)
	batchFixStd = strtod(argv_lc[i+1],0);
    }
    else if(strstr(argv_lc[i], "-noise") != 0) {
      doNoise = TRUE;
      noiseStd = 0.0;
      noiseType = GAUSSNOISE;
      if((i+1)<argc)
	noiseStd = strtod(argv_lc[i+1],0);
      if((i+2)<argc) {
	if(EQU(argv_lc[i+2],"shot"))
	  noiseType = NOISE;
	if(EQU(argv_lc[i+2],"gaussian"))
	  noiseType = GAUSSNOISE;
      }
    }
    else if(strstr(argv_lc[i], "-applyctf") != 0) {
      FILE *fIn;
      int units;
      doApplyCtf               = TRUE;
      applyCtfFilter           = 1;
      applyCtfWiener           = 0.0;
      applyCtfTFactor          = 0.0;
      applyCtfAngMajor         = 0.0;
      applyCtfPixelSize        = 0.0;
      applyCtfDefocus_in       = 0.0;
      applyCtfDefocus_minor_in = 0.0;
      applyCtfAmpContrast_in   = 0.0;
      applyCtfSphericalAbbr    = 0.0;
      applyCtfMode             = 1;
      applyCtfVolts            = 1.0;
      
      // paramFname 
      if((i+1)<argc) {// read paramFname 
	sprintf(outStr,"%s",argv_mc[i+1]);
	if ( (fIn = fopen(outStr,"r")) != NULL ) {
	  get_str(fIn, '\n', outStr);
	  if(!preloadSwitch) {// Just get the fname from the param file 
	    preloadSwitch = TRUE;
	    sprintf(preLoadStr,"%s",outStr);
	  }
	  get_str(fIn, ',', outStr);
	  applyCtfPixelSize = strtod(outStr,0);
	  get_str(fIn, ',', outStr);
	  units = (int)strtod(outStr,0);
	  if(units != 2)
	    applyCtfPixelSize  *= 0.1;  // always in nm here 
	  get_str(fIn, ',', outStr);
	  applyCtfVolts = strtod(outStr,0);
	  get_str(fIn, ',', outStr);
	  applyCtfAmpContrast_in = strtod(outStr,0);
	  get_str(fIn, ',', outStr);
	  applyCtfDefocus_in = strtod(outStr,0);
	  get_str(fIn, ',', outStr);
	  applyCtfDefocus_minor_in = strtod(outStr,0);
	  get_str(fIn, ',', outStr);
	  applyCtfAngMajor = strtod(outStr,0);
	  get_str(fIn, '\n', outStr);
	  applyCtfSphericalAbbr = strtod(outStr,0);
	  
	  fclose(fIn);
	} // read 
      } // read paramFname 
      
      if((i+2)<argc)
	applyCtfFilter = (int)strtod(argv_lc[i+2],0);
      if((i+3)<argc)
	applyCtfWiener = strtod(argv_lc[i+3],0);
      if((i+4)<argc)
	applyCtfTFactor = strtod(argv_lc[i+4],0);
      if((i+5)<argc)
	applyCtfMode = (int)strtod(argv_lc[i+5],0);
    }
    else if(strstr(argv_lc[i], "-interp") != 0) {
      doInterp = TRUE;
      doInterpOrigSize = doInterpOutputSize = 0.0;
      if((i+1)<argc)
	doInterpOrigSize = strtod(argv_lc[i+1],0);
      if((i+2)<argc)
	doInterpOutputSize = strtod(argv_lc[i+2],0);
    }
    else if(strstr(argv_lc[i], "-trunc") != 0) {
      doTrunc = TRUE;
      truncVal = 0.0;
      if((i+1)<argc)
	truncBegX = strtod(argv_lc[i+1],0);
      if((i+2)<argc)
	truncEndX = strtod(argv_lc[i+2],0);
      if((i+3)<argc)
	truncBegY = strtod(argv_lc[i+3],0);
      if((i+4)<argc)
	truncEndY = strtod(argv_lc[i+4],0);
      if((i+5)<argc)
	truncBegZ = strtod(argv_lc[i+5],0);
      if((i+6)<argc)
	truncEndZ = strtod(argv_lc[i+6],0);
      if((i+7)<argc)
	truncVal = strtod(argv_lc[i+7],0);
    }
    else if(strstr(argv_lc[i], "-lingrad") != 0) {
      doLinAlso = TRUE;
    }
    else if(strstr(argv_lc[i], "-outfile") != 0) {
      doOutFile = TRUE;
      if((i+1)<argc)
	sprintf(outFname,"%s",argv_mc[i+1]);
    }
    else if(strstr(argv_lc[i], "-box") != 0) {
      boxitSwitch = TRUE;
      boxitSize = 50;
      boxitBoxSize = 55;
      boxitFeatherSize = 55;
      if((i+1)<argc)
	sprintf(boxitFname,"%s",argv_mc[i+1]);
      else
	sprintf(boxitFname,"UNK");
      if((i+2)<argc) {
	if(EQU(argv_lc[i+2],"gaussian"))
	  featherMode = GAUSSIAN;
	else if(EQU(argv_lc[i+2],"exponential"))
	  featherMode = EXPONENTIAL;
	else if(EQU(argv_lc[i+2],"linear"))
	  featherMode = LINEAR;
	else if(EQU(argv_lc[i+2],"none"))
	  featherMode = OLD;
      }
      if((i+3)<argc)
	boxitSize = (int)strtod(argv_lc[i+3],0);
      if((i+4)<argc)
	boxitBoxSize = (int)strtod(argv_lc[i+4],0);
      if((i+5)<argc)
	boxitFeatherSize = (int)strtod(argv_lc[i+5],0);
      if(boxitBoxSize <= boxitSize)
	boxitBoxSize = boxitSize + 1;
      if(boxitFeatherSize < boxitBoxSize)
	boxitFeatherSize = boxitBoxSize ;
    }
    else if(strstr(argv_lc[i], "-rebox") != 0) {
      reBoxSwitch = TRUE;
      autoBoxSize = 50;
      autoBoxBoxSize = 55;
      autoBoxFeatherSize = 55;
      if((i+1)<argc) {
	if(EQU(argv_lc[i+1],"gaussian"))
	  featherMode = GAUSSIAN;
	else if(EQU(argv_lc[i+1],"exponential"))
	  featherMode = EXPONENTIAL;
	else if(EQU(argv_lc[i+1],"linear"))
	  featherMode = LINEAR;
	else if(EQU(argv_lc[i+1],"none"))
	  featherMode = OLD;
      }
      if((i+2)<argc)
	autoBoxBoxSize = (int)strtod(argv_lc[i+2],0);
      if((i+3)<argc)
	autoBoxFeatherSize = (int)strtod(argv_lc[i+3],0);
      if(autoBoxFeatherSize < autoBoxBoxSize)
	autoBoxFeatherSize = autoBoxBoxSize ;
    }
    else if(strstr(argv_lc[i], "-autobox") != 0) {
      autoBoxSwitch      = TRUE;
      autoBoxSize        = 50;
      autoBoxBoxSize     = 55;
      autoBoxFeatherSize = 55;
      autoBoxThin        = 1;
      autoBoxDilates     = 0;
      autoBoxIter        = 1;
      autoBoxThresh      = 0.;
      if((i+1)<argc) {
	if(EQU(argv_lc[i+1], "gaussian"))
	  featherMode = GAUSSIAN;
	else if(EQU(argv_lc[i+1], "exponential"))
	  featherMode = EXPONENTIAL;
	else if(EQU(argv_lc[i+1], "linear"))
	  featherMode = LINEAR;
	else if(EQU(argv_lc[i+1], "none"))
	  featherMode = OLD;
      }
      if((i+2)<argc)
	autoBoxSize = strtod(argv_lc[i+2],0);
      if((i+3)<argc)
	autoBoxBoxSize = strtod(argv_lc[i+3],0);
      if((i+4)<argc)
	autoBoxFeatherSize = strtod(argv_lc[i+4],0);
      if((i+5)<argc)
	autoBoxThin = strtod(argv_lc[i+5],0);
      if((i+6)<argc)
	autoBoxIter = strtod(argv_lc[i+6],0);
      if((i+7)<argc)
	autoBoxDilates = strtod(argv_lc[i+7],0);
      if((i+8)<argc)
	autoBoxThresh = (float) strtod(argv_lc[i+8],0);
      if(autoBoxBoxSize <= autoBoxSize)
	autoBoxBoxSize = autoBoxSize + 1;
      if(autoBoxFeatherSize < autoBoxBoxSize)
	autoBoxFeatherSize = autoBoxBoxSize ;
    }
    else if(strstr(argv_lc[i], "-dumpnorm") != 0) {
      dumpNormStats = TRUE;
      if((i+1)<argc)
	sprintf(dumpNormFile,"%s",argv_mc[i+1]);
    }
    else if(strstr(argv_lc[i], "-dumpspheravg") != 0) {
      dumpSpherAvg = TRUE;
      sprintf(dumpSpherAvgFile,"dump_spheravg.dat");
      if((i+1)<argc)
	sprintf(dumpSpherAvgFile,"%s",argv_mc[i+1]);
    }
    else if(strstr(argv_lc[i], "-norm") != 0) {
      normLinSwitch = TRUE;
      normLinMean = normLinStd = 0.0;
      if((i+1)<argc)
	normLinMean = strtod(argv_lc[i+1],0);
      if((i+2)<argc)
	normLinStd = strtod(argv_lc[i+2],0);
      printf("Setting norm mean/std [%f,%f]\n", normLinMean,normLinStd);
    }
    else if(strstr(argv_lc[i], "-pre") != 0) {
      preloadSwitch = TRUE;
      sprintf(preLoadStr,"%s",argv_mc[i+1]);
    }
    else if(strstr(argv_lc[i],"-zoom") != 0) {
      zoomItSwitch = TRUE;
      if((i+1) < argc) {
	zoomItFactor = strtod(argv_lc[i+1],0);
      }
      else {
	printf("ZOOM switch needs a zoomFactor argument like:\n\n");
	printf("-zoom 2\n\n");
	exit(0);
      }
    }
    else if(strstr(argv_lc[i], "-shade") != 0) {
      shadeItSwitch = TRUE;
      if((i+1) < argc) {
	shadeItDensity = strtod(argv_lc[i+1],0);
      }
      else {
	printf("SHADE switch needs a density argument like:\n\n");
	printf("-shade 16\n\n");
	exit(0);
      }
    }
    else if(strstr(argv_lc[i], "-shiftFFT") != 0) {
      shiftFFT = TRUE;
    }
    else if(strstr(argv_lc[i], "-roblite") != 0) {
      liteSwitch = TRUE;
      drawLite = FALSE;
    }
    else if(strstr(argv_lc[i], "-usage") != 0) {
      printf("robem \n\
     -APPLYCTF  paramFname filter wiener tFactor mode\n\
     -AUTOBOX  featherMode(GAUSSIAN,EXPONENTIAL,LINEAR,NONE)\n\
               emmaRadius boxRadius featherRadius numThin numIter numDilates NCCThreshold\n\
     -BOX  boxCoords featherMode(GAUSSIAN,EXPONENTIAL,LINEAR,NONE)\n\
               emmaRadius boxRadius featherRadius\n\
     -DUMPNORM dumpNormFile\n\
     -DUMPSPHERAVG dumpSpericalAverageFile\n\
     -FORCEAVGBKG\n\
     -INTERP origSize outputSize\n\
     -LINGRAD  (Used with the -NORM switch)\n\
     -NOISE noiseStd noiseType(SHOT,GAUSSIAN)\n\
     -NORM normMean normStd (if normMean==0.0 && normStd==0.0 then use what\n\
                             robEM calculates for those values)\n\
                             **Use -LINGRAD to do linear gradient also**\n\
     -OUTFILE fname\n\
     -PRELOAD fname\n\
     -REBOX  featherMode(GAUSSIAN,EXPONENTIAL,LINEAR,NONE)\n\
               boxRadius featherRadius\n\
     -RMBLEM stdDev  (if stDev==0.0 then use program default for stDev)\n\
     -SCALEIT\n\
     -SHADE density\n\
     -TRUNC begX endX begY endY begZ endZ fillValue\n\
     -USAGE\n\
     -ZOOM zoomFactor\n\\n");
      exit(0);
    }
  }
  
} // end function evalSwitches



void readResourceFile()
{
  // function name: readResourceFile
  // This needs to be developed further. Currently users can put
  // certain configuration defaults in a ~/.robemrc file that will
  // change the config of robem at startup. For example, some folks
  // like to have the immediateUpdate function on the 3D rendering
  // screen always be defaulted to ON whereas others don't like that
  // so much.

  char inStr[512], key[256], state[256];
  FILE *fIn;
  
  sprintf(outStr,"%s/.robemrc",getenv("HOME"));

  if ( (fIn = fopen(outStr,"r")) == NULL ) {
    printf("No resource file provided - using robem defaults\n");
    return;
  }
  
  while(!feof(fIn)) {
    get_str(fIn, '\n', inStr);
    
    sscanf(inStr,"%s",key);
    if(EQU(key,"IMMEDIATEUPDATE")) {
      sscanf((inStr+(strlen(key))), "%s",state);
      if(EQU(state,"AUTO")) {
	// Not yet - when lowRes YES 
	// immediateMapShade = TRUE; 
	// XtVaSetValues(immediateUpdateW, XmNset, TRUE, NULL); 
	immediateResource = FALSE;          
      }
      else if(EQU(state,"MANUAL")) {
	immediateResource = TRUE;          
	immediateMapShade = FALSE;
	XtVaSetValues(immediateUpdateW, XmNset, FALSE, NULL); 
      }
      else
	printf("I don't understand [%s][%s] from resource file\n", key, state);
    }
    else if(EQU(key,"TRUEPHI")) {
      sscanf((inStr+(strlen(key))), "%s",state);
      if(EQU(state,"ON")) {
	truePhi = TRUE;
	XtVaSetValues(truePhiW, XmNset, TRUE, NULL); 
      }
      else if(EQU(state,"OFF")) {
	truePhi = FALSE;
	XtVaSetValues(truePhiW, XmNset, FALSE, NULL); 
      }
      else
	printf("I don't understand [%s][%s] from resource file\n", key, state);
    }
    else if(EQU(key,"FORCEAVGBKG")) {
      sscanf((inStr+(strlen(key))), "%s",state); // RSSLOOKHERE
      if(EQU(state,"ON")) {
	forceAvgBkg = TRUE;
	XtVaSetValues(forceAvgBkgW, XmNset, TRUE, NULL); 
      }
      else if(EQU(state,"OFF")) {
	forceAvgBkg = FALSE;
	XtVaSetValues(forceAvgBkgW, XmNset, FALSE, NULL);
      }
      else
	printf("I don't understand [%s][%s] from resource file\n", key, state);
    }
    else if(EQU(key,"CALCSFS")) {
      sscanf((inStr+(strlen(key))), "%s",state);
      if(EQU(state,"ON")) {
	calcSFs = TRUE;
	XtVaSetValues(calcSfsW, XmNset, TRUE, NULL); 
      }
      else if(EQU(state,"OFF")) {
	calcSFs = FALSE;
	XtVaSetValues(calcSfsW, XmNset, FALSE, NULL); 
      }
      else
	printf("I don't understand [%s][%s] from resource file \n", key, state);
    }
    else if(EQU(key,"CALCBIN")) {
      sscanf((inStr+(strlen(key))), "%s",state);
      if(EQU(state,"ON")) {
	calcBin = FALSE;
	XtVaSetValues(calcBinW, XmNset, FALSE, NULL); 
      }
      else if(EQU(state,"OFF")) {
	calcBin = FALSE;
	XtVaSetValues(calcBinW, XmNset, FALSE, NULL); 
      }
      else
	printf("I don't understand [%s][%s] from resource file\n", key, state);
    }
    else if(EQU(key,"MAPSHADE")) {
      sscanf((inStr+(strlen(key))), "%s",state);
      if(EQU(state,"ON")) {
	secondMapShade = TRUE;
	XtVaSetValues(secondShadeW, XmNset, TRUE, NULL); 
      }
      else if(EQU(state,"OFF")) {
	secondMapShade = FALSE;
	XtVaSetValues(secondShadeW, XmNset, FALSE, NULL); 
      }
      else
	printf("I don't understand [%s][%s] from resource file\n", key, state);
      
      if(secondMapShade)
	XtSetArg(arglst[0], XmNsensitive, TRUE );
      else
	XtSetArg(arglst[0], XmNsensitive, FALSE );

      XtSetValues(fractScrollW, arglst, 1); 
      XtSetValues(angxScrollW, arglst, 1); 
      XtSetValues(angyScrollW, arglst, 1); 
      XtSetValues(fractLabelW, arglst, 1); 
      XtSetValues(angxLabelW, arglst, 1); 
      XtSetValues(angyLabelW, arglst, 1); 
      XtSetValues(fractValueW, arglst, 1); 
      XtSetValues(angxValueW, arglst, 1); 
      XtSetValues(angyValueW, arglst, 1); 
      XtSetValues(fractToggleW, arglst, 1); 
      XtSetValues(angxToggleW, arglst, 1); 
      XtSetValues(angyToggleW, arglst, 1); 
      
      // Turn radial Projections OFF
      radialProjections = FALSE;
      
      // Now reset toggles
      XtVaSetValues(radProjShadeW, XmNset, FALSE, NULL);
      XtUnmanageChild(icosRadProjToggleW);
      XtUnmanageChild(radProjToggleW);
      XtUnmanageChild(radProjCSToggleW);
      XtManageChild(radialCSToggleW);
      XtManageChild(radialTextW);
      XtManageChild(radialLabelW);
      XFlush(dsply);
      
      // Now activate the surface density widget 
      XtVaSetValues(surfDensScrollW, XmNsensitive, TRUE, NULL); 
      XtVaSetValues(surfDensLabelW, XmNsensitive, TRUE, NULL); 
      XtVaSetValues(surfDensValueW, XmNsensitive, TRUE, NULL);
      XtVaSetValues(surfDensToggleW, XmNsensitive, TRUE, NULL);
    }
    else if(*key=='!' || EQU(key,"")) {
      // Comment or blank line
    }
    else {
      printf("I don't understand [%s] from resource file.\n",key);
    }
  }
  
  fclose(fIn);

} // end function readResourceFile



void init_globals(int argc, char **argv)
{
  // function name: init_globals
  // Init out anything that needs it before we start executing. See
  // groundZero() for anything that needs initing later on in the
  // execution. e.g. init_globals() is done once at prg startup and
  // groundZero() is done whenever a new file is being read.

  int i, j, m;
  unsigned int seedVal;

  adraValues[0] = 1;
  adrbValues[0] = 1;
  for(i=1; i<6; i++) {
    adraValues[i] = 0;
    adrbValues[i] = 0;
  }
  global_defocus = 0.0;

  ctfMajorDefocusValue = 0.0;
  ctfMinorDefocusValue = 0.0;

  drawLite = TRUE;
  preloadSwitch = FALSE;
  shadeItSwitch = FALSE;
  zoomItSwitch = FALSE;
  featherMode = EXPONENTIAL;
  evalSwitches(argc,argv);
  
  if(liteSwitch)
    defDispBufSize = LITE_DEF_DISPLAY_BUFFER_SIZE;
  else
    defDispBufSize = DEF_DISPLAY_BUFFER_SIZE;
  
  // Colors
  redLevel =  greenLevel = blueLevel = 0;
  brightLevel = 255.0;

  // Generic counters etc...
  fullScreen= FALSE;
  borderType = 0;
  boxRadius  = 50;
  featherRadius  = 50;
  emmaRadius = 45;
  emptyShellWidth = 5;
  boxRadiusDisp = TRUE;
  emptyShellDisp = FALSE;
  emmaRadiusDisp = TRUE;
  featherRadiusDisp = TRUE;
  refineDisp = FALSE;
  clusterDisp = FALSE;
  referenceDisp = FALSE;
  imgNum= 1;
  curBox= numBoxes= numRects= numPoly= 0;
  findBox = -99;
  makeSquareBox = FALSE;
  boxZoom = USEPOINT;
  boxMethod = USECIRCLE;
  backgroundMode = USECIRCLE;
  bkgrndMethod= USEAVGPIXEL;
  dispMethod = CIRCLEDISP;
  histMethod= FALSE;
  histoEqualize = FALSE;
  writeType = EMPORTABLE;
  colorType = USEGENERIC;
  zoomFactor = 0.0;
  oldZoomFactor = -99.0;
  oldSizeFactor = -99.0;
  editBoxZoom= 1;
  stepCM = 256;
  virXoffset= virYoffset= xoffset= yoffset= 0;
  rectBox.drawable= FALSE;
  rectBox.width = rectBox.height= 0;
  buffer_width= buffer_height= 0;
  locked= FALSE;
  largeImage= FALSE;
  inited= FALSE;
  noCallbackScroll= FALSE;
  makingMovie= FALSE;
  memoryMiser = TRUE;
  imgInCore = FALSE;
  buildingCmap = FALSE;

  // Seed the random number generator
  seedVal = (unsigned int)time(NULL);
  srandom((unsigned int)seedVal);
      
  // Filename descriptor for FORTRAN
  image_fileName.addr = (char *)calloc(2048, sizeof(char));  
  image_fileName.length = 2048;
  fileName.addr = (char *)calloc(2048, sizeof(char));  
  fileName.length = 2048;
  image_fileName.type = 14;
  image_fileName.class = 1;
  fileName.type = 14;
  fileName.class = 1;

  for(i=0;i<MAXBOX;i++) {
    packImg[i]                = NULL;
    fix2DpackImg_buffer[i]    = NULL;
    fix2DpackImg_diff[i]      = NULL;
    box[i].editBox_buffer     = NULL;
    box[i].edit_mirror_buffer = NULL;
    box[i].inMemory           = FALSE;
  }
  lastPackedNumImgs= 0;
  lastDeleted= 0;
  
  // Map shading stuff
  secondMapShade    = FALSE;
  immediateMapShade = FALSE;
  immediateResource = FALSE;          
  radialMapShade    = FALSE;
  radialProjections = FALSE;
  stereoMapShade    = FALSE;
  lowRes = TRUE;
  globFract= 0.75;
  globAngx= 10.0;
  globAngy= -10.0;   // Why negative???
  globOmega =0.0;
  globTheta = 0.0;
  globPhi = 0.0;
  globMaxRadius = 0.0;
  globMinRadius = 0.0;
  globPixelSize = 1.0;
  rotateXYZ = FALSE;
  whichStep = NONE;
  shadedMaps = FALSE;
  for(m=0;m<MAXSECTIONS;m++)
    cuedMap[m].depth= cuedMap[m].radial= cuedMap[m].shad= NULL;
  tmpRotMap = rotMap = NULL;
  int_image_buffer1= NULL;
  float_image_buffer1= NULL;
  rwa_tmp2d_map= rwa_tmp_map= NULL;
  // Start these guys out small
  CK(rwa_tmp2d_map = (double *)malloc(2 * sizeof(double)));
  CK(rwa_tmp_map   = (double *)malloc(2 * sizeof(double)));
  CK(map_image_buffer   = (float *)malloc(2 * sizeof(float)));
  
  // Boxing stuff
  initLastNumBox = FALSE;
  lastNumBox = 0;
  showBoxingSwitch = FALSE;  // Not yet
  darkCircleRef = FALSE;
  cycleRef = FALSE;
  refReadFile = FALSE;
  
  autoZoom = TRUE;
  rockIt = FALSE;
  flashLightIt = FALSE;

  fullScreen_buffer.dispx = fullScreen_buffer.dispy = 0;

  irev= -1;
  rampSwitch = FALSE;
  rainbowSwitch = FALSE;

  // CircAvg stuff
  caInvertContrast= FALSE;  
  currentCircAvg= 0;
  selectedCircAvg= -99;
  circAvgSize= 1;
  caOffsetX = 0;

  maxRadiusSwitch = TRUE;
  movieSpeed= 0;
  flickSpeed = 0;
  totContours = 0;

  expStep = 8;

  // FFT stuff
  fftSize = 512;
  idim1= idim2= 1024;          // Allocate the buffer to 1024 - MAX size
  maxFFTvalue= 100000000.00;
  minFFTvalue= 0.00;
  tempFactor = 0;

  for(i=0;i<1024;i++)
    for(j=0;j<2048;j++)
      ctfValues[i][j] = tempFactors[i][j] = 1.0;

  fftMode= FALSE;
  // fftType= LINEARFFT;
  fftType= LOGFFT;
  // OLD way - let's do it all the time now 
  // backTransform = FALSE;
  backTransform = FALSE;
  pickLattice = 1;
  pickedLatPts = TRUE;
  latPt[0].x = 1;
  latPt[0].y = 0;
  latPt[1].x = 0;
  latPt[1].y = 1;
  latPt[0].measuredX = 290;
  latPt[0].measuredY = 256;
  latPt[1].measuredX = 256;
  latPt[1].measuredY = 222;
  calcLattice(TRUE);

  contour= TRUE;
  contourToggle= FALSE;
  lopHipSwitch = FALSE;
  filterToggle= FALSE;
  drawLatToggle= FALSE;
  drawHiLoToggle= TRUE;
  layerLines= FALSE;
  drawCTF = TRUE;
  ctfEnhance = TRUE;
  bumpContrast = FALSE;
  dispCenters = TRUE;
  measureFFT = TRUE;
  filtSizeValue = 5.0;
  icut= 0;                      // Start at Hard Cutoff
  for(i=0;i<5;i++)
    latScrollValue[i]= 0.0;
  pixSize = 1.0;                // FFT screen Pixel Size
  fftAngsPerPixel = pixSize*10.0;
  lowResoFilter = 320000;       // Ridiculously high
  hiResoFilter  = 0;            // Ridiculously low
  latLineOrder = 1;             // Start with something
  latValChanged = FALSE;
  averageFFT = FALSE;
  fftZoomFactor = 1;
  rasterStep = 1.00;
  ctfCurveNum = 0;
  maxCtfCurve = 1;
  applyCTF = FALSE;
  for(i=0;i<MAXCTFS;i++) {
    ctfPts[i].latX0 = 0;
    ctfPts[i].latY0 = 0;
    ctfPts[i].latX1 = 0;
    ctfPts[i].latY1 = 0;
    ctfPts[i].distA   = 0;
    ctfPts[i].distB   = 0;
    ctfPts[i].distAsq = 0;
    ctfPts[i].distBsq = 0;
    ctfPts[i].omega   = 0;
    ctfPts[i].pickedPts= FALSE;
  }
  for(i=0;i<3;i++) { 
    fftMeasures[i].x = 0; 
    fftMeasures[i].y = 0;
  }
  fftMeasures[0].x = 256; fftMeasures[0].y = 256; // Start in center
  fftImage_gold = NULL;
  fftImage_fourAver= NULL;
  fftImage_buffer2= NULL;
  fftMotionFeedback = FALSE;
  pixSize = 0.357;
  fftAngsPerPixel = pixSize*10.0;
  scanStep = 14.0;              // um
  magnification = 39200.0;      // X
  wavelength = 0.0037;          // nm-1
  spAberration = 2.0;           // mm

  // ROFLT stuff
  rotfltSwitch = FALSE;
  yFactor = 4.0;
  xFactor = 1.0;
  saveLinYFactor = 4;
  saveLogYFactor = 1;
  nrot = 5;
  lg_end = 100;
  nannuli = 30;
  logROTFLTplot = FALSE;
  fft_step = 1.0;
  phi0= omega= 0.0;
  jzero= 0;                     // 0 means to include it - believe it
  sump1= sump2= sump3= 0.0;
  backTransROTFLT= FALSE;
  stretchRotflt = TRUE;
  minRotfltScaleF= HUGE;
  
  // crysDisp stuff
  ncellx= 1;
  ncelly= 1;
  updateCrys = FALSE;
  crysDisping = FALSE;

  // fourAver stuff
  backSize = 2;
  fourAverRadius= 1.0;
  fourAverCut= 2.0;
  fourierAveragedCrysDisp = FALSE;
  refinePhaseOrigin= imposeSymmetry = 0;
  deltaPhaseShift = FALSE;
  
  // unblob
  unblobSwitch = FALSE;
  reverseSections= FALSE;
  negUniverseCut= FALSE;     // For cut
  cleanUnblob= FALSE;        // For unblob
  revRadProj = FALSE;
  icosRadProj = FALSE;
  radProjCS = FALSE;
  radialCS = FALSE;

  deltaCont = 0.0;
  numCont = 0;
  breakCont = 0.0;

  contrastStretch = FALSE;
  fftMinMaxLock = FALSE;
  disp1024fft = FALSE;
  fftDispType = INTENSITY;       
  fftMeasLineNum = 11;

  // Fix values stuff
  fixMode= FIXOVER;
  fixIcosFill = FALSE;

  calcFFT1DcircAvg= FALSE;
  disp1Dca = TRUE;
  ctf1dCalcIt = TRUE;
  ctf1dSaveFftAvg = NULL;
  ctf2dCalcIt = TRUE;
  ctf2dSaveFftasum = NULL;
  ctf2dSaveFftasumsq = NULL;

  // Autobox stuff
  nThins = 2;
  nDilates = 1;
//  prag = 1.5;
  refCutoffRatio = 1.15 ;       // Ratio cutoff to throw out particles

  // Param File stuff
  dispCenters = TRUE;
  haveParamFile = FALSE;
  symmetry = 532;
  icoDispMode = UNITCELL;

  // Alway set these to ON
  drawIcos    = FALSE;
  useOmega    = FALSE;
  findNearest = FALSE;
  icosaMotioning = TRUE;
  drawIcos = TRUE;
  useOmega = TRUE;
  findNearest = TRUE;
  icosaMotioning = FALSE;

  icosdist = 20;
  startbox = 0.000;
  endbox   = 1.000;
  yscale   = 480;
  linelength = 480.0;
  step = (endbox-startbox)/480.0;
  pixelstep = (step*linelength)/(endbox-startbox);
  ico_toolsScrollValue = 0;
  newXY = FALSE;
  idtoggle= thetatoggle= phitoggle= omegatoggle= xtoggle= ytoggle=
    magtoggle= cc1toggle= cc2toggle= cc3toggle= distancetoggle=
      deltaXtoggle = deltaYtoggle= absDeltaXtoggle= absDeltaYtoggle= FALSE;
  mapProj = mapProjContour = FALSE;
  incSort = TRUE;
  volumeDisp = FALSE;
  deltaStdX = deltaStdY= 0.0;

  info.info = NULL;

  pixelSizeFactor = 1.0;
  calcSFs   = FALSE;
  calcBin   = FALSE;
  truePhi   = TRUE;
  minMaxColor = forceAvgBkg =  forceByte = FALSE;
  dispIcosa = FALSE;
  pixelSizeSave = cutMin0Save = cutMax0Save = cutMin1Save =
      cutMax1Save = cutMin2Save = cutMax2Save = -99.0;
  icosaScaleF = 1.0;

  subtractingMap = FALSE;
  readEmcoravg = FALSE;
  readImgconv  = FALSE;

  doBinning = 0;
  doManipPP = 0;
  doCavgPP = 0;
  cavgPPcenter = 0;
  totCavgPPrint = 0;

  subtract_image_buffer1 = subtract_image_buffer2 = 0;
  // We've always got to have something allocated for this
  CK(subtract_map_image_buffer   = (float *)malloc(2*sizeof(float)));
  CK(subtract_map_image_buffer2  = (float *)malloc(2*sizeof(float)));
  subMapYmxb_buffer = diffSection_buffer = 0;

  diffSection = TRUE;
  subCanDrawWindow = TRUE;
  findMagHist_toggle= FALSE;
  findRadHist_toggle= FALSE;
  drawingFindRad = FALSE;
  appodizing = FALSE;
  writeFFTType = FALSE;
  rdensit= wrdensit= ccoefit= FALSE;
  wccoefit= TRUE;
  s_wccoefit= FALSE;
  RDoutStr[0] ='\0';
  WRDoutStr[0] ='\0';
  CCoutStr[0] ='\0';
  WCCoutStr[0] ='\0';
  S_WCCoutStr[0] ='\0';
  flickerIt = 12;
  threshScrollValue = -2000.0;
  magSecScrollValue = 1.0;
  diffSecScrollValue = 1.0;
  diffSec2ScrollValue= 0.0;
  pointColor = 0;
  totParamFiles= 0;
  for(i=0;i<MAXPACKEDFILES;i++)
    batchListStr[i]= 0;
  readPackedFile = TRUE;
  subMap2d3d= 2;
  saveSecondMap = FALSE;
  icosaQuarter = FALSE;
  showRadii = TRUE;
  threshDoit = TRUE;
  threshAbs = FALSE;
  submapFindRad  = FALSE;
  whichIcos = ICOSNORMAL;
  destroyIcos = FALSE;
  writeFourAverSF = FALSE;

  // Auto Defocus refine
  doingAutoDefocusRefine =  adrCircAvgDraw =  adrSDraw =  
    adrDDraw =   adrBDraw =  adrZoomDraw = FALSE;
  adrSminusBDraw = adrCtfsqDraw = TRUE;
  ca1d = ca1d_std = NULL;
  adr_s= adr_sminusb= adr_ctfsq= adr_d= adr_b= NULL;
  ca1d_cnt = NULL;
  doingAuto2DDefocusRefine = FALSE;

  paintMapSwitch= FALSE;
  paintMap= NULL;

  calcLowResMap= FALSE;
  globAccum = (int *)NULL;

  fix2DinCore_buffer= NULL;
  fix2DinCore_diff= NULL;
  autoMinMaxThreshold = AMMCORRCOEF;
  refCircAvg  = FALSE;
  refAutomate = FALSE;
  newOrigin = TRUE;
  rotfltCrosshair = FALSE;
  rotfltResiduals = FALSE;
  CK(sigma= (float *)malloc(sizeof(float)*22*22));
  coravgRadius = 10.0;
  lastPlotClickX = lastPlotClickY = 50;  // Just give it something
  coravgLowLimValue = 0.0;
  useAvgCorAvg = FALSE;
  showLowLim = FALSE;
  drawFix2Dgrid = FALSE;
  drawFix2Dstats = FALSE;
  flickerFix2D = FALSE;
  firstTimeLoHiRes = TRUE;
  blindWriting = FALSE;
  normalExcise = TRUE;
  ctfFilter = 1;

  numExp= 0;                                    // Debug exposures

  drawAuto2Dctf = FALSE;
  drawAuto2Dscurve = FALSE;
  drawAuto2Dctflines = FALSE;
  auto2DdrawBuf = (float *)malloc(sizeof(float)*256*512);
  auto2DdataBuf = (float *)malloc(sizeof(float)*256*512);
  emptyBox = FALSE;
  a2drSDraw=a2drSminusBDraw=a2drCtfsqDraw=a2drDDraw=a2drBDraw = FALSE;
  a2dr_s= a2dr_sminusb= a2dr_ctfsq= a2dr_d= a2dr_b= NULL;
  fftasum = poly_paras = NULL;

  modelType = SQUARE;
  model_undo_buffer = 0;
  mbImmediate = FALSE;
  isIcosRepl  = FALSE;

  firstAMMSTDEVtime = TRUE;
  showBoxingLevel = AMMCORRCOEF;
  threshLo= 0.0;
  threshHi= 99999999.9;

  // Other initializations moved from robem.h
  compress_factor = 1;
  snumb = 0.0;
  info.info = NULL;

  // Initialize pixel size information
  pifHeader.pixelSize = 0.0;
  pifHeader.pixSizUnits = 1;

} // end function init_globals



void groundZero()
{
  // function name: groundZero
  // Initialize these things when you do a major start over in the
  // code.  For example, when you read a new file

  int i;
  char infoStr[16];

  boxRadiusDisp = TRUE;
  emmaRadiusDisp = TRUE;
  featherRadiusDisp = TRUE;
  numBoxes = 0;
  dispMethod = CIRCLEDISP;

  // Initialize pixel size information
  pifHeader.pixelSize = 0.0;
  pifHeader.pixSizUnits = 1;

  infoStr[0]='\0';
  drawPointScreen(infoStr,1);
  drawPointScreen(infoStr,2);
  drawPointScreen(infoStr,3);
  drawPointScreen(infoStr,4);
  drawPointScreen(infoStr,5);

  if(projASU) { // But *do* get rid of any unwanted memory 
    for(i=0; i<purdueHeader.nimg; i++)
      freeNullFloat(&packImg[i]);
    projASU = FALSE;
    purdueHeader.nimg = 0;
    lastPackedNumImgs= 0;
  }
  
  if(haveParamFile) { // Get rid of any parameter file we've got
    haveParamFile = FALSE;
    inParamFilename[0]='\0';
    paramBoxesFilename[0]='\0';
    TextSet(paramFileTextW, inParamFilename, FALSE); 
    purdueHeader.nimg= 0;
    lastPackedNumImgs= 0;
    Free_Particle_Info(&info);
  }

  // Delete our last packed images
  for (i=0; i<lastPackedNumImgs; i++)
    freeNullFloat(&packImg[i]);
  
  lastPackedNumImgs= 0;
  for (i=0; i<MAXBOX; i++)
    packImg[i]= NULL;
  
  for (i=0; i<numBoxes; i++)
    if(box[i].inMemory==TRUE) {
      free(box[i].editBox_buffer);
      box[i].editBox_buffer= NULL;
      box[i].inMemory= FALSE;
    }
  
  numBoxes= numRects= numPoly= 0;
  xoffset= yoffset= 0;
  rectBox.drawable= FALSE; 
  rectBox.width = rectBox.height= 0;
  fullScreen= FALSE;
  histZoomFactor = 20;
  noCallbackScroll= FALSE;

  oldSizeFactor = -99.0;
  calcSizeFactor(zoomFactor); 
  makingMovie= FALSE;
  curBox= 0;
  packRadius= 0;
  averageFFT = FALSE;
  phaseOrigX = 0.0;
  phaseOrigY = 0.0;
  
  // Reset the FFT widgets
  XtVaSetValues(a128FFTW, XmNset, FALSE, NULL); 
  XtVaSetValues(a256FFTW, XmNset, FALSE, NULL); 
  XtVaSetValues(a1024FFTW, XmNset, FALSE, NULL); 
  XtVaSetValues(a512FFTW, XmNset, TRUE, NULL); 
  fftSize = 512;
  
  freeNullFloat(&inCore_buffer);
  imgInCore = FALSE;
  
  if(initLastNumBox) {
    lastNumBox = 0;
    TextSet(boxInitW, "1", TRUE);
  }

  // Free up MAP stuff
  for (i=0; i<MAXSECTIONS; i++) { // Do all section
    if(cuedMap[i].shad)
      free(cuedMap[i].shad);
    if(cuedMap[i].depth)
      free(cuedMap[i].depth);
    if(cuedMap[i].radial)
      free(cuedMap[i].radial);
    cuedMap[i].shad= cuedMap[i].depth= cuedMap[i].radial= NULL;
  }
  
  XtUnmanageChild(mapShadeScreenW);
  XtUnmanageChild(circAvgScreenW);
  XtUnmanageChild(boxScreenW);
  
  if(rotMap)
    free(rotMap);
  tmpRotMap = rotMap = NULL;
  shadedMaps = FALSE;
  
  if(int_image_buffer1)
    free(int_image_buffer1);
  int_image_buffer1= NULL;
  
  if(float_image_buffer1 && float_image_buffer1 != lowRes_buffer1)
    free(float_image_buffer1);
  if(lowRes_buffer1)
    free(lowRes_buffer1);
  
  float_image_buffer1= NULL;
  lowRes_buffer1 = NULL;
  
  if(subtract_image_buffer2)
    free(subtract_image_buffer2);
  subtract_image_buffer2 = 0;
  if(subtract_image_buffer1)
    free(subtract_image_buffer1);
  subtract_image_buffer1 = 0;
  if(diffSection_buffer)
    free(diffSection_buffer);
  diffSection_buffer = 0;
  if(subMapYmxb_buffer)
    free(subMapYmxb_buffer);
  subMapYmxb_buffer = 0;
  
  // Default menu choices
  XtSetArg(arglst[0], XmNsensitive, TRUE);
  XtSetValues(truncaddButtonW, arglst, 1); 
  XtSetValues(removeBlemishButtonW, arglst, 1); 
  XtSetValues(editBoxButtonW, arglst, 1); 
  XtSetValues(pcxButtonW, arglst, 1); 
  XtSetValues(correlateButtonW, arglst, 1);  
  XtSetValues(correlateButton2W, arglst, 1);  
  XtSetValues(autoBoxW, arglst, 1);  
  XtSetValues(rawButtonW, arglst, 1);   
  
  XtSetArg(arglst[0], XmNsensitive, FALSE);
  XtSetValues(paintMapButtonW, arglst, 1); 
  XtSetValues(mapMovieButtonW, arglst, 1); 
  XtSetValues(mapShadeButtonW, arglst, 1); 
  XtSetValues(invertButtonW, arglst, 1); 
  XtSetValues(rescaleButtonW, arglst, 1); 
  XtSetValues(calcSFButtonW, arglst, 1); 
  XtSetValues(modelW, arglst, 1); 
  XtSetValues(subtractMapButtonW, arglst, 1); 
  XtSetValues(hardRotateButtonW, arglst, 1); 
  XtSetValues(rotateButtonW, arglst, 1); 
  
  fullScreen_buffer.dispx = fullScreen_buffer.dispy = 0;
  
  // Sections or cued maps
  dispSections= TRUE;
  // Start without it
  XtVaSetValues(sectionW, XmNsensitive, FALSE, NULL);
  
  freeNullFloat(&fftImage_gold);
  freeNullFloat(&fftImage_fourAver);
  freeNullFloat(&fftImage_buffer2);

  fft_deactivate(NULL, NULL, NULL);
  rotflt_deactivate(NULL, NULL, NULL);
  crysdisp_deactivate(NULL, NULL, NULL);
  imgconv_deactivate(NULL, NULL, NULL);
  mapshade_deactivate(NULL, NULL, NULL);
  autodr_deactivate(NULL, NULL, NULL);
  emcoravg_deactivate(NULL, NULL, NULL);
  movie_deactivate(NULL, NULL, NULL);
  fouraver_deactivate(NULL, NULL, NULL);
  auto2ddr_deactivate(NULL, NULL, NULL);

  XtSetArg(arglst[0], XmNsensitive, FALSE);
  XtSetValues(interpolateW, arglst, 1); 
  XtSetValues(binningW, arglst, 1); 
  XtSetValues(addAverageW, arglst, 1); 
  XtSetValues(fourFiltW, arglst, 1); 
  XtSetValues(checkBkgW, arglst, 1); 
  XtSetValues(normalizeW, arglst, 1); 
  XtSetValues(normalizeBkgW, arglst, 1); 
  XtSetValues(linGradW, arglst, 1); 

  TextSet(frameTextW, "", FALSE); 

  XFlush(dsply);

  // Init the labels for autoboxing
  doLabel("     ", autoBoxLabelW, FALSE);
  doLabel("     ", refCorrLabelW, FALSE);
  doLabel("     ", autoCorrLabelW, FALSE);

  XtSetArg(arglst[0], XmNsensitive, FALSE);
  XtSetArg(arglst[1], XmNset, FALSE);
  XFlush(dsply);
  
  haveParamFile = FALSE;
  projASU       = FALSE;
  Free_Particle_Info(&info);
  // Make them read the map file second
  TextSet(paramMapFileTextW, "", FALSE); 
  TextSet(paramScanFileTextW, "", TRUE); 
  
  // Make them recalc any Hires buffers
  pixelSizeSave = cutMin0Save = cutMax0Save = cutMin1Save =
    cutMax1Save = cutMin2Save = cutMax2Save = -99.0;
  
  subCanDrawWindow = TRUE;
  drawingFindRad = FALSE;
  
  doingAuto2DDefocusRefine= doingAutoDefocusRefine= FALSE;
  
  paintMapSwitch= FALSE;
  freeNullFloat(&paintMap);
  
  ctf1dCalcIt = TRUE;
  ctf2dCalcIt = TRUE;
  freeNullFloat(&ctf1dSaveFftAvg);
  freeNullFloat(&ctf2dSaveFftasum);
  freeNullFloat(&ctf2dSaveFftasumsq);
  
  firstGridding = TRUE;  // Make a new grid first time empty boxing
  
  // reset these in case we did empties last time
  refAutomate = FALSE;
  darkCircleRef =  FALSE;
  
  XtVaSetValues(refDarkTW, XmNset, FALSE, NULL);
  XtVaSetValues(refAutomateW, XmNset, FALSE, NULL);
  
} // end function groundZero



void enableSaveUnder()
{
  // function name: enableSaveUnder
  // Enable save unders and backing stores for windows

  if( XDoesSaveUnders(XScreenOfDisplay(dsply, DefaultScreen(dsply))) &&
      XDoesBackingStore(XScreenOfDisplay(dsply, DefaultScreen(dsply))) ) {
    attrs.save_under = TRUE;
    attrs.backing_store = WhenMapped;
    printf("Enabling 'Backing Store' and 'SaveUnder' features\n");
    XChangeWindowAttributes(dsply, XtWindow(drawingW), 
			    (CWSaveUnder | CWBackingStore), &attrs);
    XChangeWindowAttributes(dsply, XtWindow(fftDrawingW), 
			    (CWSaveUnder | CWBackingStore), &attrs);
    XChangeWindowAttributes(dsply, XtWindow(editBoxW), 
			    (CWSaveUnder | CWBackingStore), &attrs);
    XChangeWindowAttributes(dsply, XtWindow(icosaW), 
			    (CWSaveUnder | CWBackingStore), &attrs);
    XFlush(dsply);
  }

} // end function enableSaveUnder



void allocInitBuffers() 
{
  // function name: allocInitBuffers
  // allocate and initialize buffers

  int i, j;
  int bxhw = BOX_HEIGHT * BOX_WIDTH;
  int bfhw = buffer_height * buffer_width;
  
  mirror_image_buffer1    = (BYTE *)malloc (sizeof(BYTE));
  emcoravg_buffer1        = (BYTE *)malloc (sizeof(BYTE)*4*256*256);
  circAvg_buffer1         = (BYTE *)malloc (sizeof(BYTE)*4*256*256);
  rotflt_buffer1          = (BYTE *)malloc (sizeof(BYTE)*4*512*512);
  fft_display_buffer1     = (BYTE *)malloc (sizeof(BYTE)*4*512*512);
  fft_buffer1             = (BYTE *)malloc (sizeof(BYTE)*4*512*512);
  fft_buffer2             = (BYTE *)malloc (sizeof(BYTE)*4*512*512);
  fft_gold_buffer1        = (BYTE *)malloc (sizeof(BYTE)*4*512*512);
  fft_contour_buffer      = (BYTE *)malloc (sizeof(BYTE)*4*128*128);
  fft_contour_buffer2     = (BYTE *)malloc (sizeof(BYTE)*4*128*128);
  color_buffer1           = (BYTE *)malloc (sizeof(BYTE)*4*32*256);
  test_color_buffer1      = (BYTE *)malloc (sizeof(BYTE)*4*32*32);
  hist_color_buffer1      = (BYTE *)malloc (sizeof(BYTE)*4*32*256);
  image_buffer1_disp      = (BYTE *)malloc (sizeof(BYTE)*4*bfhw);
  image_buffer1           = (BYTE *)malloc (sizeof(BYTE)*4*bfhw);
  editBox_display_buffer  = (BYTE *)malloc (sizeof(BYTE)*4*bxhw);
  editBox_display_buffer2 = (BYTE *)malloc (sizeof(BYTE)*4*bxhw);
  icosa_buffer1           = (BYTE *)malloc (sizeof(BYTE)*4*512*512);
  editBox_undo_buffer     = (float *)malloc(sizeof(float)*bxhw);
  editBox_mirror_buffer   = (float *)malloc(sizeof(float)*bxhw);
  editBox_int_buffer      = (float *)malloc(sizeof(float)*4*bxhw);
  
  init_floatarr(editBox_undo_buffer, BOX_WIDTH*BOX_HEIGHT, 0.0);
  init_floatarr(editBox_mirror_buffer, BOX_WIDTH*BOX_HEIGHT, 0.0);

  for(i=0; i<256; i++)
    for(j=0; j<32; j++)
      color_buffer1[i*32 + j] = 255 - i;
  
} // end function allocInitBuffers



void XlibInitialize()
{
  // function name: XlibInitialize
  // Initialize for Xlib calls.  Setup the colormap we'll use and
  // graphics contexts.  Create all of the Images and allocate memory
  // for all of our XmDrawingAreas.  Add event handlers for motion
  // events where needed and setup our cursors.

  int i, j;
  unsigned long event_mask;
  Window tWin;
  XWindowAttributes winAttrib;
  static const char dotted[8]= {4,4,4,4};

  // Create Image buffers for our drawingAreas
  // MAIN drawing area on main screen

  image_image1      = XCreateImage(dsply, visual, vdepth, ZPixmap, 0, NULL, 
				   buffer_width, buffer_height, 32, 0);
  
  hist_color_image1 = XCreateImage(dsply, visual, vdepth, ZPixmap, 0, NULL, 
				   32, 256, 32, 0);
  
  test_color_image1 = XCreateImage(dsply, visual, vdepth, ZPixmap, 0, NULL, 
				   32, 32, 32, 0);
  
  color_image1      = XCreateImage(dsply, visual, vdepth, ZPixmap, 0, NULL, 
				   32, 256, 32, 0);
  
  rotflt_image1     = XCreateImage(dsply, visual, vdepth, ZPixmap, 0, NULL, 
				   512, 512, 32, 0);
  
  fft_image1        = XCreateImage(dsply, visual, vdepth, ZPixmap, 0, NULL, 
				   512, 512, 32, 0);
  
  fftContour_image1 = XCreateImage(dsply, visual, vdepth, ZPixmap, 0, NULL, 
				   128, 128, 32, 0);
  
  emcoravg_image1   = XCreateImage(dsply, visual, vdepth, ZPixmap, 0, NULL, 
				   256, 256, 32, 0);
  
  circAvg_image1    = XCreateImage(dsply, visual, vdepth, ZPixmap, 0, NULL, 
				   256, 256, 32, 0);
  
  editBox_image1    = XCreateImage(dsply, visual, vdepth, ZPixmap, 0, NULL, 
				   BOX_WIDTH, BOX_HEIGHT, 32, 0);
  
  icosa_image1      = XCreateImage(dsply, visual, vdepth, ZPixmap, 0, NULL, 
				   512, 512, 32, 0);
  
  if (image_image1      == NULL || hist_color_image1 == NULL ||
      test_color_image1 == NULL || color_image1      == NULL ||
      rotflt_image1     == NULL || fft_image1        == NULL ||
      fftContour_image1 == NULL || emcoravg_image1   == NULL ||
      circAvg_image1    == NULL || editBox_image1    == NULL ||
      icosa_image1      == NULL) {
    fprintf (stderr, "XCreateImage (image) failed\n");
    exit(1);
  }
  
  // Get Window pointers from MOTIF
  image_window1= XtWindow(drawingW);
  rotflt_window1= XtWindow(rotfltDrawW); 
  fft_window1= XtWindow(fftDrawingW);
  fftContour_window1= XtWindow(contourW);
  emcoravg_window1= XtWindow(emcorAvgDrawW);
  circAvg_window1= XtWindow(circAvgW);
  color_window1= XtWindow(colormapW);
  test_color_window1= XtWindow(testcolormapW);
  hist_color_window1= XtWindow(histW);
  editBox_window1= XtWindow(editBoxW);
  icosa_window1= XtWindow(icosaW);
  
  // Tell MOTIF what kind of events you'll listen to in this window.
  event_mask  = ButtonPressMask;
  event_mask |= ButtonReleaseMask   | KeyPressMask | KeyReleaseMask;
  event_mask |= PointerMotionMask   | LeaveWindowMask | ExposureMask;
  XSelectInput(dsply, image_window1, event_mask);
  XSelectInput(dsply, emcoravg_window1, event_mask);
  XSelectInput(dsply, circAvg_window1, event_mask);
  XSelectInput(dsply, rotflt_window1, event_mask); 
  XSelectInput(dsply, fft_window1, event_mask);
  XSelectInput(dsply, color_window1, event_mask);
  XSelectInput(dsply, test_color_window1, event_mask);
  XSelectInput(dsply, hist_color_window1, event_mask);
  XSelectInput(dsply, editBox_window1, event_mask);
  XSelectInput(dsply, icosa_window1, event_mask);
  
  // OK tell the image where to find it's data
  image_image1->data      = (char *)image_buffer1;
  rotflt_image1->data     = (char *)rotflt_buffer1;
  fft_image1->data        = (char *)fft_display_buffer1;
  fftContour_image1->data = (char *)fft_contour_buffer2;
  emcoravg_image1->data   = (char *)emcoravg_buffer1;
  circAvg_image1->data    = (char *)circAvg_buffer1;
  color_image1->data      = (char *)color_buffer1;
  hist_color_image1->data = (char *)hist_color_buffer1;
  test_color_image1->data = (char *)test_color_buffer1;
  editBox_image1->data    = (char *)editBox_display_buffer2;
  icosa_image1->data      = (char *)icosa_buffer1;
  
  // Init up the main drawing Area to BLACK. 
  for (i=0; i<buffer_height; i++)
    for (j=0; j<buffer_width; j++)
      XPutPixel(image_image1, j, i, pixels_return[BLACK]); 
  
  // This is an optimization thing - just blow this out to the screen
  // numRows times whenever you need to make sure there's no unwanted
  // pixels showing.
  
  for(j=0; j<NUMCOLS; j++) {
    stufBuf[j]       = pixels_return[BLACK];
    stufIntBuf[j]    = pixels_return[BLACK];  
    stufBLKBuf[j]    = 0;
    stufBLKIntBuf[j] = (unsigned short)0;  
  }
  
  // Init up the TEST color drawing Area - we just show them this
  // placeholder value that's "mixed" by the scrolls beside it (RGB)

  for(i=0; i<32*32; i++)
    test_color_buffer1[i] = pixels_return[PLACEHOLDER];
  
  // Initialize various drawing areas to BLACK
  for(i=0; i<512*512; i++) {
    rotflt_buffer1[i] = pixels_return[BLACK];
    fft_buffer1[i] = BLACK;
    fft_gold_buffer1[i] = BLACK;
  }

  for(i=0; i<256*256; i++)
    emcoravg_buffer1[i] = circAvg_buffer1[i] = pixels_return[BLACK];

  for(i=0; i<128*128; i++)
    fft_contour_buffer[i] = BLACK;
  
  for(i=0; i<512; i++)
    icosa_buffer1[i] = pixels_return[BLACK];
  
  // Setup Graphics contexts
  if ((fontstruct = XLoadQueryFont(dsply, "9x15bold")) == NULL) {
    // Try generic fixed
    fprintf(stderr, "%s: display %s doesn't know font %s\n",
	    "robem", "DisplayString(dsply)", "9x15bold");
    if ((fontstruct = XLoadQueryFont(dsply, FONT)) == NULL) {
      fprintf(stderr, "%s: display %s doesn't know font %s\n",
	      "robem", "DisplayString(dsply)", FONT);
      exit(1);
    }
  }

  gc_mask = 0;
  gc_values.function = GXcopy;		gc_mask |= GCFunction;
  gc_values.plane_mask = AllPlanes;	gc_mask |= GCPlaneMask;
  gc_values.foreground = 1;	        gc_mask |= GCForeground;
  gc_values.background = black_pixel;	gc_mask |= GCBackground;
  gc_values.graphics_exposures = False;	gc_mask |= GCGraphicsExposures;
  
  image_gc      = XCreateGC(dsply, image_window1,      gc_mask, &gc_values);
  color_gc      = XCreateGC(dsply, color_window1,      gc_mask, &gc_values);
  rotflt_gc     = XCreateGC(dsply, rotflt_window1,     gc_mask, &gc_values);
  fft_gc        = XCreateGC(dsply, fft_window1,        gc_mask, &gc_values);
  fftContour_gc = XCreateGC(dsply, fftContour_window1, gc_mask, &gc_values);
  emcoravg_gc   = XCreateGC(dsply, emcoravg_window1,   gc_mask, &gc_values);
  circAvg_gc    = XCreateGC(dsply, circAvg_window1,    gc_mask, &gc_values);
  hist_color_gc = XCreateGC(dsply, hist_color_window1, gc_mask, &gc_values);
  test_color_gc = XCreateGC(dsply, test_color_window1, gc_mask, &gc_values);
  editBox_gc    = XCreateGC(dsply, editBox_window1,    gc_mask, &gc_values);
  icosa_gc      = XCreateGC(dsply, icosa_window1,      gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[WHITE];  gc_mask |= GCForeground;
  image_gcWHITE = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[RED];  gc_mask |= GCForeground;
  image_gcRED = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  image_rotfltgcRED = XCreateGC (dsply, rotflt_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[YELLOW];  gc_mask |= GCForeground;
  image_gcYELLOW = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[GREEN];  gc_mask |= GCForeground;
  image_gcGREEN = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[BLUE];  gc_mask |= GCForeground;
  image_gcBLUE = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[BLACK];  gc_mask |= GCForeground;
  image_gcBLACK = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[PLACEHOLDER];  gc_mask |= GCForeground;
  image_gcPLACEHOLDER = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  
  // Now setup for dashed lines
  gc_values.foreground = pixels_return[RED];	gc_mask |= GCForeground;
  gc_values.line_style = LineOnOffDash; gc_mask |= GCLineStyle;
  image_gcDOTred = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  XSetDashes(dsply, image_gcDOTred, 0, dotted, 4);
  gc_values.foreground = pixels_return[GREEN];	gc_mask |= GCForeground;
  gc_values.line_style = LineOnOffDash; gc_mask |= GCLineStyle;
  image_gcDOTgreen = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  XSetDashes(dsply, image_gcDOTgreen, 0, dotted, 4);
  gc_values.foreground = pixels_return[BLUE];	gc_mask |= GCForeground;
  gc_values.line_style = LineOnOffDash; gc_mask |= GCLineStyle;
  image_gcDOT = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  XSetDashes(dsply, image_gcDOT, 0, dotted, 4);
  
  for(i=0;i<256;i++) { // Make all of our GREY colors
    gc_mask = 0;
    gc_values.function = GXcopy;		gc_mask |= GCFunction;
    gc_values.plane_mask = AllPlanes;	gc_mask |= GCPlaneMask;
    if(i<tryColors)
      gc_values.foreground = pixels_return[i];     
    else
      gc_values.foreground = pixels_return[tryColors-1]; 
    gc_mask |= GCForeground;
    gc_values.background = black_pixel;	gc_mask |= GCBackground;
    gc_values.graphics_exposures = False;	gc_mask |= GCGraphicsExposures;
    gc_values.font = fontstruct->fid;     gc_mask |= GCFont; 
    
    image_gcGREY[i] = XCreateGC (dsply, fft_window1, gc_mask, &gc_values);
  }
  image_gcLAT = image_gcGREY[100];

  // Display the image in the window created above
  image_width = DEF_DISPLAY_SIZE;   // This changes whenever a new image 
  image_height = DEF_DISPLAY_SIZE;  // is read                           
  
  tWin = (Window)XtWindow(drawingW);
  // Get new window dimensions
  if(tWin > 0) {// Valid window
    XGetWindowAttributes(dsply,tWin,&winAttrib);
    display_width  = winAttrib.width;
    display_height = winAttrib.height;
  }
  
  // Let's also look for motion events in these windows
  XtAddEventHandler(fftDrawingW, PointerMotionMask, 
		    FALSE, (XtEventHandler)fftMotion, NULL);
  XtAddEventHandler(drawingW, PointerMotionMask, 
		    FALSE, (XtEventHandler)plotMotion, NULL);
  XtAddEventHandler(circAvgW, LeaveWindowMask, 
		    FALSE, (XtEventHandler)circAvgMotion, NULL);
  XtAddEventHandler(colormapW, LeaveWindowMask, 
		    FALSE, (XtEventHandler)colormapMotion, NULL);
  XtAddEventHandler(histW, LeaveWindowMask, 
		    FALSE, (XtEventHandler)colormapMotion, NULL);
  XtAddEventHandler(circAvgW, PointerMotionMask, 
		    FALSE, (XtEventHandler)circAvgMotion, NULL);
  XtAddEventHandler(colormapW, PointerMotionMask, 
		    FALSE, (XtEventHandler)colormapMotion, NULL);
  XtAddEventHandler(histW, PointerMotionMask, 
		    FALSE, (XtEventHandler)colormapMotion, NULL);
  XtAddEventHandler(editBoxW, PointerMotionMask, 
		    FALSE, (XtEventHandler)editBoxMotion, NULL);
  XtAddEventHandler(icosaW, PointerMotionMask, 
		    FALSE, (XtEventHandler)icosaMotion, NULL);
  XFlush (dsply);
  
  boxing_GC = image_gcGREEN;

} // end function XlibInitialize



void createCursors() 
{
  // function name: createCursors
  // Create cursors from bitmap data

  Pixmap shape, mask;

  shape = XCreateBitmapFromData(dsply, image_window1, lantern_bits, 
				lantern_width, lantern_height);
  mask = XCreateBitmapFromData(dsply, image_window1, lanternmask_bits, 
			       lanternmask_width, lanternmask_height);
  lanternCursor = XCreatePixmapCursor(dsply, shape, mask, 
				      &color_defs[BLACK], &color_defs[WHITE],
				      lantern_x_hot, lantern_y_hot);
  
  shape = XCreateBitmapFromData(dsply, image_window1, zoom_bits,
				zoom_width, zoom_height);
  mask = XCreateBitmapFromData(dsply, image_window1, zoommask_bits,
			       zoommask_width, zoommask_height);
  zoomCursor = XCreatePixmapCursor(dsply, shape, mask,
				   &color_defs[BLACK], &color_defs[RED],
				   zoom_x_hot, zoom_y_hot);
  
  watch_cursor    = XCreateFontCursor(dsply, XC_watch);
  cross_cursor    = XCreateFontCursor(dsply, XC_diamond_cross);
  bogosity_cursor = XCreateFontCursor(dsply, XC_hand1);
  XRecolorCursor(dsply, cross_cursor, &color_defs[YELLOW], &color_defs[BLACK]);

} // end function createCursors



void createPixmaps()
{
  // function name: createPixmaps
  // Create pixmaps from bitmap data

  lanternPixmap = 
    XCreatePixmapFromBitmapData(dsply, image_window1, 
				lantern_bits,
				lantern_width, lantern_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  zoomPixmap = 
    XCreatePixmapFromBitmapData(dsply, image_window1, 
				zoom_bits,
				zoom_width, zoom_height,
				pixels_return[BLACK], pixels_return[RED],
				vdepth);
  lockedPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(lockW)), 
				locked_bits,
				locked_width, locked_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  unlockedPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(lockW)), 
				unlocked_bits,
				unlocked_width, unlocked_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  playPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				play_bits, 
				play_width, play_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  playingPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				playing_bits, 
				playing_width, playing_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  infiniPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				infini_bits, 
				infini_width, infini_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  noinfiniPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				noinfini_bits, 
				noinfini_width, noinfini_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  flashLightPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				doflashlight_bits,
				doflashlight_width, doflashlight_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  noflashLightPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				noflashlight_bits,
				noflashlight_width, noflashlight_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  rockPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				dorock_bits,
				dorock_width, dorock_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  norockPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				norock_bits,
				norock_width, norock_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  stopPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				stop_bits,
				stop_width, stop_height,
				pixels_return[RED], pixels_return[WHITE],
				vdepth);
  stoppedPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				stopped_bits,
				stopped_width, stopped_height,
				pixels_return[RED], pixels_return[WHITE],
				vdepth);
  goPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				go_bits,
				go_width, go_height,
				pixels_return[GREEN], pixels_return[BLACK],
				vdepth);
  nogoPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				nogo_bits,
				nogo_width, nogo_height,
				pixels_return[GREEN], pixels_return[BLACK],
				vdepth);
  helpPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				help_bits,
				help_width, help_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  nohelpPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				nohelp_bits,
				nohelp_width, nohelp_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  exitPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				exit_bits,
				exit_width, exit_height,
				pixels_return[GREEN], pixels_return[BLACK],
				vdepth);
  noexitPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				noexit_bits,
				noexit_width, noexit_height,
				pixels_return[GREEN], pixels_return[BLACK],
				vdepth);
  moviePixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(surfDensToggleW)),
				movie_bits,
				movie_width, movie_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  unmoviePixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(surfDensToggleW)),
				unmovie_bits,
				unmovie_width, unmovie_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  slowPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				slow_bits,
				slow_width, slow_height,
				pixels_return[RED], pixels_return[BLACK],
				vdepth);
  noslowPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(playFormW)),
				noslow_bits,
				noslow_width, noslow_height,
				pixels_return[RED], pixels_return[BLACK],
				vdepth);
  resetPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(latResetButW)),
				reset_bits,
				reset_width, reset_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  noresetPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(latResetButW)),
				noreset_bits,
				noreset_width, noreset_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  sectionPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(sectionW)),
				section_bits,
				section_width, section_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);
  unsectionPixmap = 
    XCreatePixmapFromBitmapData(dsply, XtWindow(XtParent(sectionW)),
				unsection_bits,
				unsection_width, unsection_height,
				pixels_return[BLACK], pixels_return[WHITE],
				vdepth);

} // end function createPixmaps



void applyPixmaps()
{
  // function name: applyPixmaps
  // Apply pixmaps to widgets

  XtSetArg(arglst[0], XmNrecomputeSize, FALSE);

  XtSetArg(arglst[1], XmNlabelPixmap,   unlockedPixmap);
  XtSetArg(arglst[2], XmNselectPixmap,  lockedPixmap);
  XtSetValues(lockW, arglst, 3); 
  
  XtSetArg(arglst[1], XmNlabelPixmap,   sectionPixmap);
  XtSetArg(arglst[2], XmNselectPixmap,  unsectionPixmap);
  XtSetValues(sectionW, arglst, 3); 

  XtSetArg(arglst[1], XmNlabelPixmap,   playPixmap);
  XtSetArg(arglst[2], XmNselectPixmap,  playingPixmap);
  XtSetValues(playW, arglst, 3); 

  XtSetArg(arglst[1], XmNlabelPixmap,   stopPixmap);
  XtSetArg(arglst[2], XmNselectPixmap,  stoppedPixmap);
  XtSetValues(stopW, arglst, 3); 

  XtSetArg(arglst[1], XmNlabelPixmap,   stopPixmap);
  XtSetArg(arglst[2], XmNselectPixmap,  stoppedPixmap);
  XtSetValues(mapShadeStopW, arglst, 3); 

  XtSetArg(arglst[1], XmNlabelPixmap,   stopPixmap);
  XtSetArg(arglst[2], XmNselectPixmap,  stoppedPixmap);
  XtSetValues(boxStopW,     arglst, 3); 
  XtSetValues(versionStopW, arglst, 3); 
  XtSetValues(statusStopW,  arglst, 3); 
  XtSetValues(fftStopW,     arglst, 3); 
  
  XtSetArg(arglst[1], XmNlabelPixmap,   goPixmap);
  XtSetArg(arglst[2], XmNselectPixmap,  nogoPixmap);
  XtSetValues(subFlickUpW,   arglst, 3); 
  XtSetValues(movieArrowUpW, arglst, 3); 

  XtSetArg(arglst[1], XmNlabelPixmap,   slowPixmap);
  XtSetArg(arglst[2], XmNselectPixmap,  noslowPixmap);
  XtSetValues(subFlickDownW,   arglst, 3); 
  XtSetValues(movieArrowDownW, arglst, 3); 
  
  XtSetArg(arglst[1], XmNlabelPixmap,   exitPixmap);
  XtSetArg(arglst[2], XmNselectPixmap,  noexitPixmap);
  XtSetValues(dioExitW,         arglst, 3); 
  XtSetValues(fftFeedbackExitW, arglst, 3); 
  XtSetValues(modelExitW,       arglst, 3); 
  XtSetValues(binningExitW,     arglst, 3); 
  XtSetValues(interpExitW,      arglst, 3); 
  XtSetValues(autoDRExitW,      arglst, 3); 
  XtSetValues(auto2DDRExitW,    arglst, 3); 
  XtSetValues(normalExitW,      arglst, 3); 
  XtSetValues(fourAverExitW,    arglst, 3); 
  XtSetValues(truncaddExitW,    arglst, 3); 
  XtSetValues(rotateExitW,      arglst, 3); 
  XtSetValues(fixExitW,         arglst, 3); 
  XtSetValues(fixitExitW,       arglst, 3); 
  XtSetValues(fix2DExitW,       arglst, 3); 
  XtSetValues(pointExitW,       arglst, 3); 
  XtSetValues(boxExitW,         arglst, 3); 
  XtSetValues(optionsExitW,     arglst, 3); 
  XtSetValues(fftExitW,         arglst, 3); 
  XtSetValues(circAvgExitW,     arglst, 3); 
  XtSetValues(crysDispExitW,    arglst, 3); 
  XtSetValues(emcorAvgExitW,    arglst, 3); 
  XtSetValues(imgconvExitW,     arglst, 3); 
  XtSetValues(rotfltExitW,      arglst, 3); 
  XtSetValues(statExitW,        arglst, 3); 
  XtSetValues(movconExitW,      arglst, 3); 
  XtSetValues(colorExitW,       arglst, 3); 
  XtSetValues(mapShadeExitW,    arglst, 3); 
  XtSetValues(movieDoneW,       arglst, 3); 
  XtSetValues(versionExitW,     arglst, 3); 
  XtSetValues(partprocExitW,    arglst, 3); 
  XtSetValues(moveExitW,        arglst, 3); 
  XtSetValues(batchExitW,       arglst, 3); 
  XtSetValues(invertExitW,      arglst, 3); 
  XtSetValues(rescaleExitW,     arglst, 3); 
  XtSetValues(fourFiltExitW,    arglst, 3); 
  XtSetValues(helpExitW,        arglst, 3); 
  XtSetValues(subExitW,         arglst, 3); 
  XtSetValues(dxyExitW,         arglst, 3); 
  XtSetValues(domegaExitW,      arglst, 3); 
  XtSetValues(manipPPExitW,     arglst, 3); 
  XtSetValues(cavgPPExitW,      arglst, 3); 
  XtSetValues(appodizeExitW,    arglst, 3); 
  
  XtSetArg(arglst[1], XmNlabelPixmap,   resetPixmap);
  XtSetArg(arglst[2], XmNselectPixmap,  noresetPixmap);
  XtSetValues(refineResetButtonW, arglst, 3); 
  XtSetValues(apOrigResetW,       arglst, 3); 
  XtSetValues(adrabResetW,        arglst, 3); 
  XtSetValues(latResetButW,       arglst, 3); 
  XtSetValues(crysDispResetW,     arglst, 3); 
  XtSetValues(fourAverResetW,     arglst, 3); 
  XtSetValues(subSmResetW,        arglst, 3); 
  XtSetValues(subDiffRBW,         arglst, 3); 
  XtSetValues(subMagResetW,       arglst, 3); 
  XtSetValues(subRadRW,           arglst, 3); 
  XtSetValues(subMag1ResetW,      arglst, 3); 
  XtSetValues(subMag2ResetW,      arglst, 3); 
  XtSetValues(subMag3ResetW,      arglst, 3); 
  XtSetValues(unblobResetW,       arglst, 3); 

  XtSetArg(arglst[1], XmNlabelPixmap,   moviePixmap);
  XtSetArg(arglst[2], XmNselectPixmap,  unmoviePixmap);
  XtSetValues(fractToggleW,        arglst, 3); 
  XtSetValues(angxToggleW,         arglst, 3); 
  XtSetValues(angyToggleW,         arglst, 3); 
  XtSetValues(thetaToggleW,        arglst, 3); 
  XtSetValues(phiToggleW,          arglst, 3); 
  XtSetValues(omegaToggleW,        arglst, 3); 
  XtSetValues(maxRadiusToggleW,    arglst, 3); 
  XtSetValues(pixelSizeToggleW,    arglst, 3); 
  XtSetValues(unblobsizeToggleW,   arglst, 3); 
  XtSetValues(unblobradiusToggleW, arglst, 3); 
  XtSetValues(maxxToggleW,         arglst, 3); 
  XtSetValues(maxyToggleW,         arglst, 3); 
  XtSetValues(maxzToggleW,         arglst, 3); 
  XtSetValues(surfDensToggleW,     arglst, 3);  
  
  XtSetArg(arglst[1], XmNlabelPixmap,   infiniPixmap);
  XtSetArg(arglst[2], XmNselectPixmap,  noinfiniPixmap);
  XtSetValues(infiniW,        arglst, 3); 
  XtSetValues(versionScrollW, arglst, 3); 

  XtSetArg(arglst[1], XmNlabelPixmap,   noflashLightPixmap);
  XtSetArg(arglst[2], XmNselectPixmap,  flashLightPixmap);
  XtSetValues(flashLightW, arglst, 3); 

  XtSetArg(arglst[1], XmNlabelPixmap,   norockPixmap);
  XtSetArg(arglst[2], XmNselectPixmap,  rockPixmap);
  XtSetValues(rockW, arglst, 3); 

  XtSetArg(arglst[1], XmNlabelPixmap,   helpPixmap);
  XtSetArg(arglst[2], XmNselectPixmap,  nohelpPixmap);
  XtSetValues(imgconvHelpW,  arglst, 3); 
  XtSetValues(dioHelpW,      arglst, 3); 
  XtSetValues(rotfltHelpW,   arglst, 3); 
  XtSetValues(modelHelpW,    arglst, 3); 
  XtSetValues(binningHelpW,  arglst, 3); 
  XtSetValues(interpHelpW,   arglst, 3); 
  XtSetValues(autoDRHelpW,   arglst, 3); 
  XtSetValues(auto2DDRHelpW, arglst, 3); 
  XtSetValues(normalHelpW,   arglst, 3); 
  XtSetValues(fourAverHelpW, arglst, 3); 
  XtSetValues(boxHelpW,      arglst, 3); 
  XtSetValues(appodizeHelpW, arglst, 3); 
  XtSetValues(fourFiltHelpW, arglst, 3); 
  XtSetValues(fftHelpW,      arglst, 3); 
  XtSetValues(mapShadeHelpW, arglst, 3); 
  XtSetValues(helpToolW,     arglst, 3);  
  XtSetValues(fixHelpW,      arglst, 3); 
  XtSetValues(fix2DHelpW,    arglst, 3); 
  XtSetValues(rotateHelpW,   arglst, 3); 
  XtSetValues(truncaddHelpW, arglst, 3); 
  XtSetValues(subHelpW,      arglst, 3); 
  XtSetValues(dxyHelpW,      arglst, 3); 
  XtSetValues(domegaHelpW,   arglst, 3); 
  XtSetValues(manipPPHelpW,  arglst, 3); 
  XtSetValues(cavgPPHelpW,   arglst, 3); 

  XtSetArg(arglst[0], XmNrecomputeSize, TRUE);
  XtSetValues(ppHelpMenuEntryW, arglst, 3); 

} // end function applyPixmaps



int checkStatus(char *thisStr)
{
  // function name: checkStatus
  // Check to see if they want to stop this or not.  Very important if
  // they have asked for a long computation as in a 3D Render movie
  // and find out half way through it they either don't want to wait
  // that long or they setup the wrong params. This gives them a way
  // to abort by hitting the STOP pushbutton.

  if(useGUI == FALSE) return(FALSE);

  if(drawLite) {
    TextSet(statusTextW, thisStr, FALSE);
    
    // Frame the location of the stop button
    DrawX(XtWindow(statusStopW), image_gcRED, 16, 16, 4);
    XFlush(dsply);
    
    // See if they want to bugout or not
    if(XCheckTypedWindowEvent(dsply, XtWindow(statusStopW),
			      ButtonPress, &event)) {
      XFlush(dsply);
      return(TRUE);
    }
    else {
      return(FALSE);
    }
  }
  else
    return(FALSE);

} // end function checkStatus



void s_error(char *problem_string)
{
  // function name: s_error
  // All errors are fatal.

  printf("%s\n", problem_string);
  exit(0);

} // end function s_error



int calcHistLength(int histVal)
{
  // function name: calcHistLength
  // Use the histZoomFactor to make sure we don't draw in the
  // histogram window more than 32 pixels.

  return MIN(histogram[256-histVal]*histScaleFactor, 32);

} // end function calcHistLength



void stuffImage(unsigned short *thisBuffer, int i, int j, int scaleIt,
		int fullScreenFirst)
{
  // function name: stuffImage
  //
  // Take the integer buffer we've read from the .IMG disk file and
  // stuff it into our BYTE buffer for display.  Make the integer
  // buffer disk data "fit" into our colormap by scaling it.  e.g. You
  // may have data from 0 to 255 on the disk (per pixel) but we've
  // only got 237 colors available to us for our color table.
  //
  // We maintain a "mirror" buffer here for speed in rezooming the
  // image.  For example if we need to zoom the image it is easier to
  // read the data already converted to our colormap doman and in
  // character form than trying to reread the integer buffer and
  // reconverting it.  Keep that integer buffer around, though, to
  // stuff the box buffers etc...  We need to always write out the
  // original data NOT the colromap converted stuff when we write.

  int k, l, x, y, thisOffset, yMirror;
  int min, max, lastCol, lastRow;
  unsigned short *q;
  float scaleFactor;

  // Calculate zoominess
  calcSizeFactor(zoomFactor);

  if(thisBuffer==NULL)
    return;
  
  // Let's trap here for images less than image_width x image_height
  lastCol = display_width;
  lastRow = display_height;
  if(fullScreen) {
    if(numCols*sizeFactor < display_width)
      lastCol = purdueHeader.numCols*sizeFactor;
    if(numRows*sizeFactor < display_height)
      lastRow = purdueHeader.numRows*sizeFactor;
    if(disp1024fft && fftSize == 1024 && XtIsManaged(fftScreenW)) { // special
      lastCol = 1024*sizeFactor;
      lastRow = 1024*sizeFactor;
    }
  }
  
  // Reset offsets
  xoffset = yoffset = 0;
  
  // Just to make sure let's init this
  for(l=0; l<(buffer_height*3); l++)
    memcpy((&image_buffer1[l*buffer_width]),stufBuf,buffer_width);
  
  /// Now clip this guy to our colormap
  scaleFactor = (tryColors - 7.0) / 256.0;
  
  init_intarr(histogram, 257, 0); // initialize histogram
  q= (unsigned short *)  thisBuffer; 

  free(mirror_image_buffer1);
  CK(mirror_image_buffer1 = (BYTE *)malloc (i*j*sizeof(BYTE)));
  
  scaleItSwitch = scaleIt;
  if(scaleItSwitch) {
    min = INT_MAX;
    max = INT_MIN;
    for (y=0; y<lastRow; y++) {
      // Tim draws from the bottom up - We are opposite
      yMirror = y*i;
      for (x=0; x<lastCol; x++) { // Clip the entire image
	q = thisBuffer + yMirror + x;
	if(flickerFix2D && *q==30000) {
	  /* DO nothing */
	}
	else {
	  min = MIN((*q),min);
	  max = MAX((*q),max);
	}
      }  
    }

    if(fileType==PORTABLE)
      scaleFactor = scaleFactor * (65500.0 / max);
    else
      scaleFactor = scaleFactor * (255.0 / max);
    
  }
  else  { // No scaling
    if(fileType==BYTEPORTABLE) { // byte
      max=255; min=0;
      // What about if you want to maintian the values on the
      // disk - e.g. suppose you max BYTE value is 127 why
      // should you have to "dummy" it down???
      if(purdueHeader.img_max < tryColors) {
	max = (tryColors-7);
	min = 0;
      }
    }
    else { // short
      max = min = thisBuffer[0] ;
      for (k=0; k<image_height*image_width; k++) {
	min = MIN(thisBuffer[k],min);
	max = MAX(thisBuffer[k],max);
      }
    }
  }
  
  // Now clip this guy to our colormap - 8 bit colormap used to
  // reserve the first 7 spaces for defined colors like red,green,blue etc.
  scaleFactor = (tryColors  - 7.0) / (max - min);
  
  // Convert integer to char within our colormap and put it
  // in the display buffer(s)
  for (y=0; y<lastRow; y++) {
    // Tim draws from the bottom up - We are opposite
    yMirror = y*i;
    for(x=0; x<lastCol; x++) { // Clip the entire image
      
      q = thisBuffer + yMirror + x;
      
      // We've got to scale these for now because we may get one
      // that is larger than we've got colormap entries for.
      
      thisOffset =  ((float)*(q)- (float)min)*scaleFactor  + 6.0;

      if(thisOffset >=6 && thisOffset<tryColors) {
	XPutPixel(image_image1, x, y, pixels_return[thisOffset]);
	histogram[thisOffset] += 1;
      }

      if(flickerFix2D && *q == 30000) // Do flicker case
	XPutPixel(image_image1, x, y, pixels_return[RED]);
      
    }
  }
  
  if(histoEqualize && !fullScreenFirst) { // Histogram equalize
    init_intarr(histogram, 257, 0); // initialize histogram
    histEqualize(MIN(file_width, display_width),
		 MIN(file_height, display_height),
		 image_buffer1, 7, 245, int_image_buffer1,
		 min, max, buffer_width);
  }
  
} // end function stuffImage



void stuffMapImage(float *map_img_buf, float *thisBuffer, 
		   int ncols, int nrows, int nsecs,
		   int startImg, int passedOffset)
{
  // function name: stuffMapImage
  // Take the integer buffer we've read from the .MAP disk file and
  // stuff it into our BYTE buffer for display.  Make the integer
  // buffer disk data "fit" into our colormap by scaling it.  e.g. You
  // may have data from 0 to 255 on the disk (per pixel) but we've
  // only got 237 colors available to us for our color table.  For
  // this MAP data let's just start out with the first section.
  //
  // We maintain a "mirror" buffer here for speed in rezooming the
  // image.  For example if we need to zoom the image it is easier to
  // read the data already converted to our colormap doman and in
  // character form than trying to reread the integer buffer and
  // reconverting it.  Keep that integer buffer around, though, to
  // stuff the box buffers etc...  We need to always write out the
  // original data NOT the colromap converted stuff when we write.
  
  int l, x, y, thisOffset, yMirror, firstTime;
  int sectionsPerRow, rowsPerImage, thisNumSecs, thisX;
  float thisPixel;
  float *q;

  firstTime = TRUE;
  // Force it to always display the depth cued buffer if we've got it 
  // or want it (e.g. from the section toggle etc...)

  if(shadedMaps && !dispSections) { // Shaded maps
    thisBuffer= map_img_buf;
    ncols = largestDim;
    nrows = largestDim;
  }

  // Computer numbers from 0 not 1
  startImg--;
  
  // Calculate zoominess
  calcSizeFactor(zoomFactor);
  
  if(passedOffset == 0)
    for(l=0; l<buffer_height*3 ;l++)
      memcpy((&image_buffer1[l*buffer_width]),stufBuf,buffer_width);
  
  if(minMapPixel == maxMapPixel) {
    sprintf(outStr,"MAP file has bogus data min[%.2f]max[%.2f]",
	    minMapPixel, maxMapPixel);
    askThem(outStr, XmDIALOG_QUESTION);
    return;
  }
  
  // Now clip this guy to our colormap
  if(minMapPixel < 0.0)
    scaleMapFactor = (tryColors - 7.0) / (maxMapPixel-minMapPixel);
  else 
    scaleMapFactor = (tryColors - 7.0) / maxMapPixel;

  if(shadedMaps && !dispSections) { // scale it
    // We REALLY need to have this here for the radial shading
    // to work properly.  dph_rescale() sets it up properly.
    
    if(radialMapShade)
      scaleMapFactor= 1.0;  // Already done in dph_rescale_rob() 
  }
  
  // Now find out how we're going to put these on the display
  sectionsPerRow= MIN(buffer_width, display_width)/(int)(ncols*sizeFactor);
  sectionsPerRow= MAX(1,sectionsPerRow);

  rowsPerImage = MIN(buffer_height, display_width)/(int)(nrows*sizeFactor);
  rowsPerImage = MAX(1,rowsPerImage);

  thisNumSecs = (sectionsPerRow * rowsPerImage) + startImg;
  thisNumSecs = MAX((startImg+1),MIN(thisNumSecs,nsecs));

  // Stuff globals
  mapSectionsPerRow = sectionsPerRow;
  mapStartImage = startImg;
  
  init_intarr(histogram, 257, 0); // initialize histogram
  q = (float *) thisBuffer; 
  
  /// Convert integer to char within our colormap and put it
  // in the display buffer(s)
  lastImg= thisNumSecs;
  for (l=startImg; l<thisNumSecs; l++) {
    int rowOffset, colOffset;
    
    rowOffset = ((l-startImg) / sectionsPerRow) * (nrows*sizeFactor);
    colOffset = ((l-startImg) % sectionsPerRow) * (ncols*sizeFactor);
    
    colOffset += passedOffset;
    
    imgNums[l].x= colOffset;
    imgNums[l].y= rowOffset;
    
    for (y=0; y<(nrows*sizeFactor); y++) {
      // Tim draws from the bottom up - We are opposite
      yMirror = (int)(y*recipSizeFactor) * ncols;
      for(x=0; x<ncols*sizeFactor; x++) { // Clip the entire image
	
	q = thisBuffer + l*nrows*ncols + yMirror + (int)(x*recipSizeFactor);

	if(*q == -1234567.0)
	  // Special case for subtract map thresholding
	  thisOffset = GREEN;
	else if(*q == -1234568.0)
	  // Special case for subtract map thresholding
	  thisOffset = RED;
	else { // normal 
	  // We've got to scale these for now because we may get one
	  // that is larger than we've got colormap entries for.
	  
	  if(minMapPixel < 0.0)
	    thisPixel = *q - minMapPixel;
	  else 
	    thisPixel = *q;
	  
	  thisOffset = (int)(thisPixel * scaleMapFactor) + 6;
	}
	      
	if(thisOffset >=0 && thisOffset < 256) {
	  if(colOffset+x > display_width)
	    thisX = display_width - colOffset-1;
	  else
	    thisX = x;
	  if(useGUI) XPutPixel(image_image1, colOffset+thisX, rowOffset+y,
			       pixels_return[thisOffset]);
	  histogram[(int)thisOffset]++;
	}
	else {
	  if(firstTime)
	    printf ("colormap problems 1\nx[%d]y[%d] q[%d]\n",x,y,(int)*(q));
	  firstTime = FALSE;
	}
	
      }
    }
    
  }

  image_width  = ncols;
  image_height = nrows;
  
} // end function stuffMapImage



void pushBox(int x, int y, int findThisBox, int *thisBox)
{
  // function name: pushBox
  // We've got a new box.  Record the x,y coords as the coords of the
  // image on the disk (we may not have all of it displayed right
  // now). Also allocate the memory for the box image and get it from
  // what we've got in memory right now.  If we don't get the image
  // from memory now when they've selected the point we'll have to
  // read it of the disk later.
  //
  // Added 6-24-96: Check to see if the coords of this box are +- 10
  // pixels from another box - if so then just update the coords for
  // that box - they are "tuning" the coords for that box.

  int k, foundIt, thisRadius;
  
  // We also changed all of the references to sizeFactor below
  // to recipSizeFactor
  calcSizeFactor(zoomFactor);
  
  // Map this to original image coords
  box[numBoxes].x = (x*recipSizeFactor)+(xoffset*recipSizeFactor)+virXoffset;
  box[numBoxes].y = (y*recipSizeFactor)+(yoffset*recipSizeFactor)+virYoffset;
  box[numBoxes].inMemory= FALSE;  // They've got to "paint" it into memory
  
  // Add reference particle marker , Aug.20, 97, Lixin Zhu
  box[numBoxes].reference = ref_box_flag; 
  
  // Check to see if this one's been done before
  foundIt = FALSE;
  thisRadius = MAX(20, boxRadius);

  if(findThisBox) { // Find the box pointed to
    *thisBox = -99;
    for (k=0; k<numBoxes; k++) {
      if( box[k].x >= box[numBoxes].x-thisRadius &&
	  box[k].x <= box[numBoxes].x+thisRadius &&
	  box[k].y >= box[numBoxes].y-thisRadius &&
	  box[k].y <= box[numBoxes].y+thisRadius ) {
	if(!ref_box_flag) // New box
	  *thisBox = k;
      }
    }
  }
  else { // Try to add it
    if(*thisBox >= 0) { // Found it last time
      k = *thisBox;
      foundIt = TRUE;
      if(!ref_box_flag) { // New box 
	box[k].x = box[numBoxes].x;
	box[k].y = box[numBoxes].y;
      }
      box[k].reference = ref_box_flag;
    }
    else { // Look for it
      for(k=0;k<numBoxes;k++) {
	if( box[k].x >= box[numBoxes].x-thisRadius &&
	    box[k].x <= box[numBoxes].x+thisRadius &&
	    box[k].y >= box[numBoxes].y-thisRadius &&
	    box[k].y <= box[numBoxes].y+thisRadius ) {
	  foundIt = TRUE;
	  if(!ref_box_flag) { // New box 
	    box[k].x = box[numBoxes].x;
	    box[k].y = box[numBoxes].y;
	  }
	  box[k].reference = ref_box_flag;
	}
      }
    }
    
    // Check to see if we can add it to the end
    if(!foundIt) { // Add to end?
      if( box[numBoxes].x-boxRadius < 0 ||
	  box[numBoxes].y-boxRadius < 0 ||
	  box[numBoxes].x+boxRadius > (file_width-1) ||
	  box[numBoxes].y+boxRadius > (file_height-1) ) {
	// Can't add to list - it's out of the image
      }
      else { // OK add it
	numBoxes++;
      }
    }
  }
  
} // end function pushBox



void removeBox(int x, int y, int diskCoords)
{
  // function name: removeBox
  // Look for box that contains point (x,y) that had been selected,
  // delete box, and shift all subsequent boxes up one slot.

  int i, j, fs;

  calcSizeFactor(zoomFactor);
  fs = (boxRadius/2);

  if(!diskCoords) { // Map this to original image coords
    x = x*recipSizeFactor + xoffset*recipSizeFactor + virXoffset;
    y = y*recipSizeFactor + yoffset*recipSizeFactor + virYoffset;
  }

  for(i=0; i<numBoxes; i++) {
    if(x+fs > box[i].x && box[i].x > x-fs &&
       y+fs > box[i].y && box[i].y > y-fs ) {
      
      if(box[i].inMemory && box[i].edit_mirror_buffer) 
	free(box[i].edit_mirror_buffer); // free mem assoc w/ this box
      
      for (j=i; j<numBoxes; j++)  // Shift boxes up one slot
	box[j] = box[j+1]; // struct assignment

      numBoxes--;
      break;
    }
  }

} // end function removeBox



void pushRectBox(int button, int x, int y)
{
  // function name: pushRectBox
  // Whenever they are trying to make a rectangle (or square) box call
  // here. Check to see which button they've pressed and build the
  // box accordingly.

  int origX, origY;

  // Changed sizeFactor to recipSizeFactor below also
  calcSizeFactor(zoomFactor);

  // Map this to original image coords
  origX = (x*recipSizeFactor)+(xoffset*recipSizeFactor)+virXoffset;
  origY = (y*recipSizeFactor)+(yoffset*recipSizeFactor)+virYoffset;

  // Branch based on which button was pressed
  if(button == Button1 && !locked) { // Upper leftahnd corner
    if(rectBox.drawable==FALSE) { // Init these
      rectBox.x= origX;
      rectBox.y= origY;
      rectBox.lowx = origX+1;
      rectBox.lowy = origY+1;
    }
    else { // Check to make sure we've got it right
      if(origY > rectBox.lowy)
	rectBox.lowy= origY;
      else
	rectBox.y= origY;
      if(origX > rectBox.lowx)
	rectBox.lowx= origX;
      else
	rectBox.x= origX;
      
      // Now check for the SQUARE subset of rectangle
      if(boxMethod==USESQUARE) { // square it up
	int xDiff, yDiff;
	xDiff = rectBox.lowx - rectBox.x;
	yDiff = rectBox.lowy - rectBox.y;
	if( yDiff > xDiff )
	  rectBox.lowy = rectBox.y + xDiff;
	else if( xDiff > yDiff )
	  rectBox.lowx = rectBox.x + yDiff;
	if(makeSquareBox) { // They want it a certain size
	  rectBox.lowy = rectBox.y + makeSquareBox;
	  rectBox.lowx = rectBox.x + makeSquareBox;
	}
      }
    }

    // Now we have a valid starting point to pick a Lower
    // right from
    rectBox.drawable= TRUE;
    rectBox.width  = rectBox.lowx - rectBox.x;
    rectBox.height = rectBox.lowy - rectBox.y;      
  }
  else if(button == Button2 && !locked) { // Got the lower righthand or such
    if(rectBox.drawable) {
      // Make them start with leftmost button Force so that our
      // concept of lower right really is lower right
      if(origY > rectBox.y)
	rectBox.lowy= origY;
      else { // Reverse them
	rectBox.lowy= rectBox.y;
	rectBox.y= origY;
      }
      if(origX > rectBox.x)
	rectBox.lowx= origX;
      else { // Reverse them
	rectBox.lowx= rectBox.x;
	rectBox.x= origX;
      }
      
      // Now check for the SQUARE subset of rectangle
      if(boxMethod==USESQUARE) { // square it up
	int xDiff, yDiff;
	xDiff = rectBox.lowx - rectBox.x;
	yDiff = rectBox.lowy - rectBox.y;
	if( yDiff > xDiff )
	  rectBox.y = rectBox.lowy - xDiff;
	else if( xDiff > yDiff )
	  rectBox.x = rectBox.lowx - yDiff;
	if(makeSquareBox) { // They want it a certain size
	  rectBox.lowy = rectBox.y + makeSquareBox;
	  rectBox.lowx = rectBox.x + makeSquareBox;
	}
      }
    }
  }
  else if(button == Button3 && rectBox.drawable) {
    // Move it to new location - make the lower right hand corner the landmark
    rectBox.y = origY - (rectBox.lowy - rectBox.y);
    rectBox.x = origX - (rectBox.lowx - rectBox.x);
    rectBox.lowy = origY;
    rectBox.lowx = origX;
  }

} // end function pushRectBox



void pushPolyBox(int x, int y)
{
  // function name: pushPolyBox
  // Place the given coords on the polygon stack. Future goal - Try
  // and get this polyBox stuff debugged. It's my understanding that
  // noone has used it so far but it might come in handy later...???

  // Also changed sizeFactor to recipSizeFactor below
  calcSizeFactor(zoomFactor);
  
  // Map this to original image coords
  poly[numPoly].x = (x+xoffset)*recipSizeFactor + virXoffset;
  poly[numPoly].y = (y+yoffset)*recipSizeFactor + virYoffset;
  
  numPoly++;
  
} // end function pushPolyBox



void popPolyBox()
{
  // function name: popPolyBox
  // Pop the last one off the poly stack.

  numPoly--;
  numPoly = MAX(numPoly, 0);

} // end function popPolyBox



void filterFiles(char *filesStr, int tag)
{
  // function name: filterFiles
  // Set the file selection widget to filter for the fileStr files.

  char *directory, *pattern, *s1;
  static int lastFilter;
  XmString xmDirectory, xmPattern;

  // We need to filter every time to make sure we show them the latest
  // version of the file for reading OR writing.  e.g. If we wrote box
  // coords file rob.bcrd;3 before and now we're reading box coords we
  // want to filter the selection box to show rob.bcrd;3 to be read
  // instead of the rob.bcrd;2 which was the last filtering (e.g. ;2
  // has older data than ;3).

  XtSetArg(arglst[0], XmNdirectory, &xmDirectory);
  XtGetValues(fileSelW, arglst, 1);
  XmStringGetLtoR(xmDirectory, XmSTRING_DEFAULT_CHARSET, &directory);

  XtSetArg(arglst[0], XmNpattern, &xmPattern);
  XtGetValues(fileSelW, arglst, 1);
  XmStringGetLtoR(xmPattern, XmSTRING_DEFAULT_CHARSET, &pattern);

  strcpy(outStr, directory);
  strcat(outStr, pattern);

  if(tag != lastFilter) { // Only do if we're changing file types
    s1 = strrchr(outStr,'/');
    if(tag==ARCHIVE || tag==DEMO || tag==TESTARCH)
      sprintf(outStr,"%s",filesStr);
    else if(s1)
      sprintf((s1+1),"%s",filesStr);
    else 
      sprintf(outStr,"%s",filesStr);
  }
  
  XmFileSelectionDoSearch(fileSelW, XmStringCreateLocalized(outStr));
  XFlush(dsply);

  XmStringFree(xmDirectory); 
  XtFree(directory);
  XmStringFree(xmPattern); 
  XtFree(pattern);
  
  XFlush(dsply);

  lastFilter = tag;

} // end function filterFiles



void writeColormap(char *fName)
{
  // function name: writeColormap
  //
  // Write the current colormap out to an ASCII disk file.  NOTE: We
  // get a new set of colormap entries everytime we run this program
  // based on what other clients are using the display.  If we should
  // run again and get fewer colormap entries than we have now we may
  // have trouble with the colormap we've just written to disk.
  //
  // This mechanism was used in our 8 bit display environment where we
  // wanted multiple instances of robem running and displaying to the
  // same display.  We just allocated the colormap in the
  // masterProcess and then just shared the values of that colormap
  // between robem instances.

  int i, writeIt = TRUE;
  FILE *fOut;
  
  // Make sure we're not overwriting a good one
  if ( (fOut= fopen(fName,"r")) != NULL ) {
    char infoStr[512];
    sprintf(infoStr,"[%s] Already exists - OVERWRITE it?",fName);
    writeIt = askThem(infoStr, XmDIALOG_QUESTION);
    fclose(fOut);
    if(!writeIt)
      return;
  }
  
  if ( (fOut = fopen(fName,"w")) != NULL ) {
    askThem("Can't open file.", XmDIALOG_INFORMATION);
    return;
  }

  for(i=0; i<256; i++) {// Write the current colormap
    fprintf(fOut,"%lu %d %d %d %d\n",
	    color_defs[i].pixel, color_defs[i].red,
	    color_defs[i].green, color_defs[i].blue,
	    color_defs[i].flags);
  }
  fclose(fOut);
  
} // end function writeColormap



void readColormap(char *fName)
{
  // function name: readColormap
  // Read a range colormap from disk.
  // Set colors to appropriate rgb values based on radial range.
  // The disk file has format:
  // 
  //        radius red green blue
  //
  // Here's an example file:
  // 12  0     0      0
  // 45  0     0      65000
  // 75  0     35000  65000
  // 111 0     65000  0
  // 135 65000 0      0
  // 175 35000 65000  35000
  // 555 65000 0      35000
  //
  // In the above example up to radius 12 the colormap is set to BLACK.
  // Up to radius 45 the color is blue.  Up to 75 color greenish-blue.
  // Up to radius 111 then green etc...
  // 65535 is the MAX any color can be set to.
  
  int i, j, tot=0;
  int radius[512], red[512], green[512], blue[512], flags[512];
  char inStr[512];
  FILE *fIn;
  
  if(useGUI == FALSE) return;

  if ( (fIn= fopen(fName,"r")) != NULL ) {
    while(!feof(fIn)) { // Read the current colormap
      get_str(fIn,' ',inStr);
      radius[tot] = strtod(inStr,0);
      get_str(fIn,' ',inStr);
      red[tot] = strtod(inStr,0);
      get_str(fIn,' ',inStr);
      green[tot] = strtod(inStr,0);
      get_str(fIn,'\n',inStr);
      blue[tot] = strtod(inStr,0);
      flags[tot] = 7;
      if(radius[tot] > 0)
	tot++;
    }
    
    // Now update colormap
    for (i=0; i<256; i++) {
      for (j=0; j<tot; j++) {
	if(i < radius[j]) {
	  color_defs[i].pixel = i;
	  color_defs[i].red   = red[j];
	  color_defs[i].green = green[j];
	  color_defs[i].blue  = blue[j];
	  color_defs[i].flags = flags[j];
	  break;
	}
      }
    }

    // Set the current colormap on the display
    for (i=6; i<tryColors; i++) {
      XAllocColor (dsply, color_map, &color_defs[i]);
      pixels_return[i] = color_defs[i].pixel;
    }
    
    // Update display - Start at zero'th image
    if(fileType==PACKEDPORTABLE)
      stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		       packImg,purdueHeader.nimg, imgNum);  
    else if(fileType==MAP)
      stuffMapImage(map_image_buffer, float_image_buffer1, 
		    file_width,file_height, nsecs, imgNum, 0);
    
    DrawWindow(); 
  }
  else 
    askThem("Can't open file.", XmDIALOG_INFORMATION);
  
  fclose(fIn);
  
} // end function readColormap



void pushColormap()
{
  // function name: pushColormap
  // Save the current colormap in case we need to restore it later.

  int i;
  if(useGUI == FALSE) return;

  for(i=0; i<256; i++) // Remember the current colormap
    mirror_color_defs[i] = color_defs[i]; // struct assignment
  
} // end function pushColormap



void popColormap()
{
  // function name: popColormap
  // Restore saved colormap.

  int i;
  for(i=0; i<256; i++) // Remember the old current colormap
    color_defs[i] = mirror_color_defs[i]; // struct assignment
  
  for(i=6; i<tryColors; i++) { // Stuff the display with it
    XAllocColor (dsply, color_map, &color_defs[i]);
    pixels_return[i] = color_defs[i].pixel;
  }

  // Now redraw the histogram
  histogram_activate();
  
} // end function popColormap



void pointerLabelDraw(int x, int y)
{
  // function name: pointerLabelDraw
  // When they navigate the cursor over the image displayed this rtn
  // gives them feedback as to what the values for the underlying
  // pixel are.

  int thisX, thisY, rectX, rectY;
  int thisImg, sectY, sectX;
  float grayLevel;

  if(!inited) return;

  // Don't do this until we're up and running
  // Show them where they've started
  calcSizeFactor(zoomFactor);
  
  // Get these for this virtual page only
  thisX = (x+xoffset)*recipSizeFactor + virXoffset;
  thisY = (y+yoffset)*recipSizeFactor + virYoffset;
  
  grayLevel = findRealValue(x,y,&thisImg,&sectY,&sectX);

  if(fileType == PACKEDPORTABLE) {
    sprintf(outStr, "x,y,Img[%3d,%3d,%3d] [%f]",
	    sectX, sectY, thisImg, grayLevel);
  } 
  else {
    if(rectBox.drawable) { // OK we've got one 
      rectX = fabs(rectBox.x - rectBox.lowx);
      rectY = fabs(rectBox.y - rectBox.lowy);
      sprintf(outStr,"R[%dx%d]  %d,%d",
	      rectBox.width, rectBox.height, rectX, rectY);
    } 
    else {
      sprintf(outStr,"(x,y,z) = (%3d,%3d,%3d) value = %f",
	      sectX, sectY, thisImg, grayLevel);
    }
  }
  
  drawPointScreen(outStr,1);
  
  if(fftSize==1024 && disp1024fft && XtIsManaged(fftScreenW)) {
    int hereX, hereY;
    float thisRadius, apart, bpart, amp, phase;
    
    hereX = thisX - 512.0;
    hereY = 0.0 - (thisY - 512.0);
    
    if(fftImage_gold)
      fft_point(fftImage_gold, fftSize, fftSize,
		hereX, hereY, &thisRadius, &apart, &bpart);
    
    if(x != 512)
      sprintf(outStr,"x[%d]y[%d] radius[%.0f] Angs[%.2f]",
	      hereX, hereY, thisRadius,
	      ((1024.0/((float)(x-512.0)/zoomFactor)) * fftAngsPerPixel));
    else
      sprintf(outStr,"x[%d]y[%d] radius[%.0f] Angs[undef]",
	      hereX,hereY,thisRadius);
    
    drawPointScreen(outStr,2);
    
    sprintf(outStr,"a[%.2E] b[%.2E]", apart, bpart);
    drawPointScreen(outStr,3);
    
    convert_abpart(apart, bpart, &amp, &phase);
    sprintf(outStr,"amp[%.2E] phase[%.2E]",amp,phase);
    drawPointScreen(outStr,4);
  }
  
} // end function pointerLabelDraw



void handle_cr()
{
  // function name: handle_cr
  // Parse translation tables and assign translations to widgets. Used
  // primarily to handle carriage returns and clicks of buttons 2/3
  
  int i;
  XtTranslations parsed_t_table;

  parsed_t_table = XtParseTranslationTable(rwa_translation_table);
  XtOverrideTranslations(imgNumTextW, parsed_t_table);

  parsed_t_table = XtParseTranslationTable(box_translation_table);
  XtOverrideTranslations(boxInitW, parsed_t_table);

  parsed_t_table = XtParseTranslationTable(editBoxText_translation_table);
  XtOverrideTranslations(editBoxTextW, parsed_t_table);

  parsed_t_table = XtParseTranslationTable(aoverb_translation_table);
  XtOverrideTranslations(aoverbTextW,    parsed_t_table);
  XtOverrideTranslations(scaleXvalueW,   parsed_t_table);
  XtOverrideTranslations(adimTextW,      parsed_t_table);
  XtOverrideTranslations(bdimTextW,      parsed_t_table);
  XtOverrideTranslations(gammaXtalTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(map_abang_translation_table);
  XtOverrideTranslations(map_abangTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(fftImgNum_translation_table);
  XtOverrideTranslations(fftImgNumW,    parsed_t_table);  
  XtOverrideTranslations(rotfltImgNumW, parsed_t_table);

  parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
  XtOverrideTranslations(unblobRadiusScrollW, parsed_t_table);
  XtOverrideTranslations(unblobSizeScrollW,   parsed_t_table);
  XtOverrideTranslations(mapShadeScreenW,     parsed_t_table);
  XtOverrideTranslations(unblobFormW,         parsed_t_table);
  XtOverrideTranslations(cutMinScrollW[0],    parsed_t_table);       
  XtOverrideTranslations(cutMaxScrollW[0],    parsed_t_table);
  XtOverrideTranslations(cutMinScrollW[1],    parsed_t_table);      
  XtOverrideTranslations(cutMaxScrollW[1],    parsed_t_table);
  XtOverrideTranslations(cutMinScrollW[2],    parsed_t_table);       
  XtOverrideTranslations(cutMaxScrollW[2],    parsed_t_table);
  XtOverrideTranslations(maxRadiusScrollW,    parsed_t_table);
  XtOverrideTranslations(maxRadiusValueW,     parsed_t_table);   
  XtOverrideTranslations(minRadiusScrollW,    parsed_t_table);       
  XtOverrideTranslations(minRadiusValueW,     parsed_t_table);    
  XtOverrideTranslations(fractScrollW,        parsed_t_table);           
  XtOverrideTranslations(fractValueW,         parsed_t_table);       
  XtOverrideTranslations(angxScrollW,         parsed_t_table);            
  XtOverrideTranslations(angxValueW,          parsed_t_table);        
  XtOverrideTranslations(angyScrollW,         parsed_t_table);            
  XtOverrideTranslations(angyValueW,          parsed_t_table);        
  XtOverrideTranslations(phiScrollW,          parsed_t_table);             
  XtOverrideTranslations(phiValueW,           parsed_t_table);         
  XtOverrideTranslations(omegaScrollW,        parsed_t_table);           
  XtOverrideTranslations(omegaValueW,         parsed_t_table);       
  XtOverrideTranslations(thetaScrollW,        parsed_t_table);           
  XtOverrideTranslations(thetaValueW,         parsed_t_table);       
  XtOverrideTranslations(pixelSizeScrollW,    parsed_t_table);       
  XtOverrideTranslations(pixelSizeValueW,     parsed_t_table);   
  XtOverrideTranslations(surfDensScrollW,     parsed_t_table);        
  XtOverrideTranslations(surfDensValueW,      parsed_t_table);    
  XtOverrideTranslations(cutFormW,            parsed_t_table);

  parsed_t_table = XtParseTranslationTable(zoom_translation_table);
  XtOverrideTranslations(zoomTextW, parsed_t_table);

  parsed_t_table = XtParseTranslationTable(zoomArrow_translation_table);
  XtOverrideTranslations(zoomArrowUpW,   parsed_t_table);  
  XtOverrideTranslations(zoomArrowDownW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(cDisp_zArrow_trans_table);
  XtOverrideTranslations(cDispZoomArrowUpW,   parsed_t_table);  
  XtOverrideTranslations(cDispZoomArrowDownW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(boxNum_translation_table);
  XtOverrideTranslations(boxUpW,   parsed_t_table);  
  XtOverrideTranslations(boxDownW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(icosa_translation_table);
  XtOverrideTranslations(icosaUpW,   parsed_t_table);  
  XtOverrideTranslations(icosaDownW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(paintSize_translation_table);
  XtOverrideTranslations(paintUpW,   parsed_t_table);  
  XtOverrideTranslations(paintDownW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(version_translation_table);
  XtOverrideTranslations(versionTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(paramFile_translation_table);
  XtOverrideTranslations(paramFileTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(paramMapFile_translation_table);
  XtOverrideTranslations(paramMapFileTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(projasu_translation_table);
  XtOverrideTranslations(projasuTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(projContour_translation_table);
  XtOverrideTranslations(projContourTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(strtxpp_translation_table);
  XtOverrideTranslations(strtXPPW, parsed_t_table);  
  XtOverrideTranslations(endXPPW,  parsed_t_table);  
  XtOverrideTranslations(maxYPPW,  parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(ctfValue_translation_table);
  XtOverrideTranslations(ctfFminTW, parsed_t_table);  
  XtOverrideTranslations(ctfAmajTW, parsed_t_table); 
  XtOverrideTranslations(ctfFmajTW, parsed_t_table); 
  
  parsed_t_table = XtParseTranslationTable(latText_translation_table);
  for(i=0; i<4; i++)
    XtOverrideTranslations(latticeLabel2W[i], parsed_t_table);  
  
  parsed_t_table = XtParseTranslationTable(ctfText_translation_table);
  XtOverrideTranslations(ctfPixSizeW,  parsed_t_table);  
  XtOverrideTranslations(ctfMagW,      parsed_t_table);  
  XtOverrideTranslations(ctfScanStepW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(hipText_translation_table);
  XtOverrideTranslations(hipTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(lohiRadiusText_trans_table);
  XtOverrideTranslations(lopRadiusTextW, parsed_t_table);  
  XtOverrideTranslations(hipRadiusTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(lopText_translation_table);
  XtOverrideTranslations(lopTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(subText_translation_table);
  XtOverrideTranslations(subFile2TW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(diffSec_translation_table);
  XtOverrideTranslations(subDiff1TW,  parsed_t_table);   
  XtOverrideTranslations(subDiff2TW, parsed_t_table); 
  XtOverrideTranslations(subMagsecTW,   parsed_t_table);    
  XtOverrideTranslations(subMagStartTW,  parsed_t_table);       
  XtOverrideTranslations(subThreshTW,   parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(ppMinMax_translation_table);
  XtOverrideTranslations(ppMinTextW, parsed_t_table);  
  XtOverrideTranslations(ppMaxTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(binPP_translation_table);
  XtOverrideTranslations(binPPUpW,   parsed_t_table);  
  XtOverrideTranslations(binPPDownW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(stepPP_translation_table);
  XtOverrideTranslations(stepPPUpW,   parsed_t_table);  
  XtOverrideTranslations(stepPPDownW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(ppin_translation_table);
  XtOverrideTranslations(ppinUpW,   parsed_t_table);  
  XtOverrideTranslations(ppinDownW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(paramScanFile_translation_table);
  XtOverrideTranslations(paramScanFileTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(cmScroll_translation_table);
  XtOverrideTranslations(colormapScrollW,     parsed_t_table);  
  XtOverrideTranslations(colormapScrollDownW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(gbs_translation_table);
  XtOverrideTranslations(gbsUpArrowW,   parsed_t_table);  
  XtOverrideTranslations(gbsDownArrowW, parsed_t_table);

  parsed_t_table = XtParseTranslationTable(boxInit_translation_table);
  XtOverrideTranslations(boxinitArrowUpW,   parsed_t_table);  
  XtOverrideTranslations(boxinitArrowDownW, parsed_t_table); 
  
  parsed_t_table = XtParseTranslationTable(rotflt_translation_table);
  XtOverrideTranslations(xorigUpArrowW,   parsed_t_table);  
  XtOverrideTranslations(xorigDownArrowW, parsed_t_table); 
  XtOverrideTranslations(yorigUpArrowW,   parsed_t_table);  
  XtOverrideTranslations(yorigDownArrowW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(phase_translation_table);
  XtOverrideTranslations(phasXupW,   parsed_t_table);  
  XtOverrideTranslations(phasXdownW, parsed_t_table); 
  XtOverrideTranslations(phasYupW,   parsed_t_table);  
  XtOverrideTranslations(phasYdownW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(coravg_translation_table);
  XtOverrideTranslations(emcorAvgUAW,   parsed_t_table);  
  XtOverrideTranslations(emcorAvgDAW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(smMMax_translation_table);
  XtOverrideTranslations(subRadMinUpW,   parsed_t_table);  
  XtOverrideTranslations(subRadMinDownW, parsed_t_table); 
  XtOverrideTranslations(subRadMaxUpW,   parsed_t_table);  
  XtOverrideTranslations(subRadMaxDownW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(imgNum_translation_table);
  XtOverrideTranslations(imgnumArrowUpW,   parsed_t_table);  
  XtOverrideTranslations(imgnumArrowDownW, parsed_t_table); 
  
  parsed_t_table = XtParseTranslationTable(fftImgNumArrow_translation_table);
  XtOverrideTranslations(fftImgnumArrowUpW,      parsed_t_table);     
  XtOverrideTranslations(fftImgnumArrowDownW,    parsed_t_table); 
  XtOverrideTranslations(rotfltImgnumArrowUpW,   parsed_t_table);  
  XtOverrideTranslations(rotfltImgnumArrowDownW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(ctfCurve_translation_table);
  XtOverrideTranslations(ctfCurveTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(ppintext_translation_table);
  XtOverrideTranslations(ppinTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(binPPtext_translation_table);
  XtOverrideTranslations(binTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(stepPPtext_translation_table);
  XtOverrideTranslations(histStepTextW, parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(fftStep_translation_table);
  XtOverrideTranslations(fftStepTextW, parsed_t_table);  
  XtOverrideTranslations(xorigTextW,   parsed_t_table);  
  XtOverrideTranslations(yorigTextW,   parsed_t_table);
  
  parsed_t_table = XtParseTranslationTable(radius_translation_table);
  XtOverrideTranslations(boxRadiusArrowUpW,       parsed_t_table);      
  XtOverrideTranslations(boxRadiusArrowDownW,     parsed_t_table);  
  XtOverrideTranslations(emptyShellArrowUpW,      parsed_t_table);     
  XtOverrideTranslations(emptyShellArrowDownW,    parsed_t_table); 
  XtOverrideTranslations(emmaRadiusArrowUpW,      parsed_t_table);     
  XtOverrideTranslations(emmaRadiusArrowDownW,    parsed_t_table); 
  XtOverrideTranslations(featherRadiusArrowUpW,   parsed_t_table);  
  XtOverrideTranslations(featherRadiusArrowDownW, parsed_t_table);

} // end function handle_cr



void stripParens(char *inStr)
{
  // function name: stripParens
  // Strip of unwanted parens and commas

  char *s1;

  s1= strrchr(inStr,')');
  if(s1)
    *s1= ' ';
  s1= strrchr(inStr,'(');
  if(s1)
    *s1= ' ';
  s1= strrchr(inStr,',');
  if(s1)
    *s1= ' ';

  s1= strchr(inStr,')');
  if(s1)
    *s1= ' ';
  s1= strchr(inStr,'(');
  if(s1)
    *s1= ' ';
  s1= strchr(inStr,',');
  if(s1)
    *s1= ' ';

} // end function stripParens



void readBoxCoords(char *fName)
{
  // function name: readBoxCoords
  // Read box coords from a disk file.

  int i, emmaSwitch, newVersion;
  char firstChar, inStr[320];
  FILE *fIn;
  
  newVersion = emmaSwitch = FALSE;
  if ( (fIn= fopen(fName,"r")) != NULL ) {
    firstChar = (char)fgetc(fIn);
    if(firstChar == 'E')
      emmaSwitch = TRUE;
    if(firstChar == 'B')
      newVersion = TRUE;
    fclose(fIn);
  }

  if ( (fIn= fopen(fName,"r")) != NULL ) {
    if(!emmaSwitch) { // normal
      if(newVersion)
	get_str(fIn,'\n',inStr); 	  
      fscanf(fIn,"%d",&lastNumBox);
      fscanf(fIn,"%d",&numBoxes);
      fscanf(fIn,"%d",&boxRadius);
      if(newVersion)
	fscanf(fIn,"%d",&featherRadius);
      for (i=0; i<numBoxes; i++)  // Read the current BoxCoords
	fscanf(fIn,"%d %d", &box[i].x, &box[i].y);
    }
    else { // wing it
      // Scan off first "E" line
      fscanf(fIn,"%s",inStr);
      lastNumBox = 0;
      numBoxes= 0;
      boxRadius = 50;
      while(!feof(fIn)) {
	fscanf(fIn,"%s",inStr);
	stripParens(inStr);
	box[numBoxes].x = strtod(inStr,0);
	fscanf(fIn,"%s",inStr);
	stripParens(inStr);
	box[numBoxes].y = strtod(inStr,0);
	if(box[numBoxes].x > 0 || box[numBoxes].y > 0 )
	  numBoxes++;
      }
      numBoxes--;   // Last one is bogus
    }

    emmaRadius    = MIN(emmaRadius,(boxRadius-1));
    if(!newVersion)
      featherRadius = MIN(featherRadius,(boxRadius));
    printf("emma[%d]box[%d]fea[%d]\n", 
	   emmaRadius, boxRadius, featherRadius);
    
    TextSetInt(featherRadiusTextW, 2*featherRadius + 1, FALSE);
    TextSetInt(boxRadiusTextW, featherRadius-boxRadius, FALSE);
    TextSetInt(boxInitW, lastNumBox+1, TRUE);
    fclose(fIn);
    
    raiseIt(boxScreenW);
    DrawWindow();
    
  }
  else 
    askThem("Can't open file.", XmDIALOG_INFORMATION);
  
} // end function readBoxCoords



void readState(char *fName)
{
  // function name: readState
  // Read the options state for the program from an ASCII disk file.
  // Set the program "state" based on the values read. This was the
  // thinking early on that we would be able to store and restore that
  // state of all the program variables and display widgets or at
  // least the important ones. Not sure if it's needed or not..
  
  char version[64];
  FILE *fIn;
  
  if ( (fIn= fopen(fName,"r")) != NULL ) {
    fscanf(fIn,"%s",version);
    fscanf(fIn,"%d",&locked);
    fscanf(fIn,"%d",&boxRadius);
  }
  else 
    askThem("Can't open file.", XmDIALOG_INFORMATION);

  fclose(fIn);

  // Now change the "state" of the program - set the lock toggle
  XtVaSetValues(lockW, XmNset, locked, NULL);

  // Set the boxRadiusTextW to (un)editable
  XtVaSetValues(boxRadiusTextW, XmNeditable, !(locked), NULL);
  
  // New radius value
  TextSetInt(boxRadiusTextW, featherRadius-boxRadius, FALSE);
  
} // end function readState



void readSF(char *fName)
{
  // function name: readSF
  // Read the structure factors file and convert to a MAP on the fly.

  FILE *fIn;
  
  askThem("SF Not implemented", XmDIALOG_INFORMATION);

  if ( (fIn= fopen(fName,"r")) == NULL )
    askThem("Can't open file.", XmDIALOG_INFORMATION);
  fclose(fIn);

} // end function readSF



void writeState(char *fName)
{
  // function name: writeState
  // Write the current option state of the program to an ASCII file

  FILE *fOut;
  
  if ( (fOut= fopen(fName,"w")) != NULL ) {
    fprintf(fOut,"%s\n", EM_VERSION);
    fprintf(fOut,"%d\n",locked);
    fprintf(fOut,"%d\n",boxRadius);
  }
  else 
    askThem("Can't open file.", XmDIALOG_INFORMATION);

  fclose(fOut);

} // end function writeState



int askThem(char *infoStr, int dialogType)
{
  // function name: askThem
  // This is the generic display a message and get them to respond
  // screen.  We get a return value from globAsk which is set by
  // infoOK() or infoCancel() (see below).

  if(useGUI == FALSE) { // Behave like OK button had been pressed
    globAsk = TRUE;
    gFinished = TRUE;
    return(TRUE);
  }

  // Init and try to position the window
  globAsk = -99;
  XFlush(dsply);
  XtSetArg(arglst[0], XmNmessageString, XmStringCreateLocalized(infoStr));
  XtSetArg(arglst[1], XmNdialogType, dialogType);
  XtSetArg(arglst[2], XmNx, 150);
  XtSetArg(arglst[3], XmNy, 150);
  XtSetValues(infoBoxW, arglst, 4);
  XFlush(dsply);
  
  // Init it and pop it up sit here and wait until we get something back
  gFinished= FALSE;
  XtManageChild(infoBoxW);
  XtPopup(XtParent(infoBoxW), XtGrabNone);
  
  XmProcessTraversal(infoBoxW, XmTRAVERSE_CURRENT); 
  XtRealizeWidget(infoBoxW);  
  XFlush(dsply);
  
  // gFinished is set in the infoOK() or infoCancel() callback
  while(!gFinished)
    XtAppProcessEvent(app_context, XtIMAll);
  
  // Pop it down and return their response
  XtUnmanageChild(infoBoxW);
  XtPopdown(XtParent(infoBoxW));
  XtVaSetValues(XtParent(infoBoxW), XmNdialogStyle, XmDIALOG_MODELESS, NULL);
  XFlush(dsply);
  
  return(globAsk);
  
} // end function askThem



int askThemAnswer(char *infoStr, int maxAnswer)
{
  // function name: askThemAnswer
  // This is the generic display a message and get them to respond
  // screen. We get a return value from globAsk which is set by the
  // text widget on the screen.

  int i;

  // Init and try to position the window
  TextSet(infoPromptTextW, infoStr, FALSE); 
  globAskAnswer = -99;
  XFlush(dsply);
  
  // Init it and pop it up sit here and wait until we
  // get something back
  gFinished= FALSE;
  XtManageChild(infoAnswerScreenW);
  XtPopup(XtParent(infoAnswerScreenW), XtGrabNone);
  
  XmProcessTraversal(infoAnswerScreenW, XmTRAVERSE_CURRENT); 
  XtRealizeWidget(infoAnswerScreenW);  
  XFlush(dsply);
  
  // gFinished is set in the infoOK() or infoCancel() callback
  while(!gFinished)
    XtAppProcessEvent(app_context, XtIMAll);
  
  // Pop it down and return their response
  XtUnmanageChild(infoAnswerScreenW);
  XtPopdown(XtParent(infoAnswerScreenW));
  XtVaSetValues(XtParent(infoAnswerScreenW), XmNdialogStyle, 
		XmDIALOG_MODELESS, NULL);
  XFlush(dsply);
  
  globAskAnswer = MAX(MIN(globAskAnswer,maxAnswer-1),0);
  
  for(i=0;i<maxAnswer;i++) // Deleteing
    XmListDeletePos(infoListW,1);
  
  return(globAskAnswer);
  
} // end function askThemAnswer



void infoOK(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: infoOK
  // If the OK button is pressed from the infoscreen return here.  Set
  // gFinished to cause the value of globAsk be return to the starting
  // point (see askThem() above).

  globAsk = TRUE;
  XtUnmanageChild(infoBoxW);
  gFinished= TRUE;

} // end function infoOK



void infoCancel(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: infoCancel
  // If the CANCEL button is pressed from the infoscreen return here.
  // Set gFinished to cause the value of globAsk be return to the
  // starting point (see askThem() above).

  globAsk = FALSE;
  XtUnmanageChild(infoBoxW);
  gFinished= TRUE;

} // end function infoCancel



void write_activate(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: write_activate
  // Set up file browsing widget with appropriate default type

  char thisFileName[512], *s1, *suffix;
  int tag = *((int *) client_data);

  writeType = tag;
  readWrite = WRITING;

  suffix = (char *) malloc(sizeof(char)*10);
  suffix[0] = '\0';

  if(writeType == CIRCLEPACKED || writeType == IMG        ||
     writeType == APPENDRECT   || writeType == RECTPACKED ||
     writeType == APPENDCIRCLEPACKED || writeType == MAP) {
    if(writeType == CIRCLEPACKED)
      suffix = "_box.pif";
    filterFiles("*.pif", tag); 
  }
  else if(tag == BOXCOORDS) {
    suffix = ".bcrd";
    filterFiles("*.b*", tag); 
  }
  else if(tag == STATE) {
    suffix = ".sta";
    filterFiles("*.sta", tag); 
  }
  else if(tag == ARCHIVE)
    filterFiles("bragg4:[arch]*.*", tag); 
  else if(tag == TIFF) {
    suffix = ".raw";
    filterFiles("*.raw", tag); 
  }
  else if(tag == PCX) {
    suffix = ".pcx";
    filterFiles("*.pcx", tag); 
  }

  if(strlen(suffix)) {
    sprintf(thisFileName,"%s",fileName.addr);
    s1= strrchr(thisFileName,'.');
    if(s1)
      sprintf(s1,"%s",suffix);
    else
      sprintf(thisFileName,"%s%s",fileName.addr, suffix);

    XtSetArg(arglst[0], XmNtextString, XmStringCreateLocalized(thisFileName));
    XtSetValues(fileSelW, arglst, 1);
  }

  raiseIt(fileSelW);

} // end function write_activate



void read_activate(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: read_activate
  // They selected the "Read" submenu. Branch on the appropriate.

  int tag = *((int *) client_data);

  if(tag == ALL)
    // Try and setup the file browsing widget with the
    //appropriate default file type etc..
    filterFiles("*",ALL); 
  else if(tag == MRC)
    filterFiles("*.*",MRC); 
  else if(tag == PACKEDPORTABLE)
    filterFiles("*.pif",PACKEDPORTABLE); 
  else if(tag == SF)
    filterFiles("*.sf",SF); 
  else if(tag == BOXCOORDS)
    filterFiles("*.b*",BOXCOORDS); 
  else if(tag == STATE)
    filterFiles("*.sta",STATE); 
  else if(tag == ARCHIVE)
    filterFiles("/bio/baker11/archive/*",ARCHIVE); 
  else if(tag == TESTARCH)
    filterFiles("/bio/baker9d7/test/*",TESTARCH); 
  
  // Let gotFileName() find it out
  if(tag == BOXCOORDS || tag == SF || tag == STATE || tag == MRC)
    readType= tag;
  else 
    readType = 0;
  
  readWrite= READING;

// Make them read from partproc screen if they want particle processing
  XtUnmanageChild(partprocScreenW);   
  raiseIt(fileSelW);

} // end function read_activate



void exit_activate(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: exit_activate
  // Exit the program.

    exit(0);
  
} // end function exit_activate



void formResize(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: formResize
  // When the main screen formDialog is resized Get the new window
  // size and setup our globals.

  int width, height, pageIncW, pageIncH; 
  Window tWin;
  XWindowAttributes winAttrib;

  // This changes whenever a new image is read
  image_width  = defDispBufSize;   
  image_height = defDispBufSize;   

  if(!inited) return;
  
  tWin = (Window)XtWindow(drawingW);
  
  // Get new window dimensions
  if(tWin > 0) { // Valid window
    XGetWindowAttributes(dsply, tWin, &winAttrib);
    width  = winAttrib.width;
    height = winAttrib.height;
  }
  
  // Can't have the window larger than the image data we've got!!!
  width  = MIN(width, defDispBufSize);
  height = MIN(height, defDispBufSize);
  
  // First map coords to image coords
  calcSizeFactor(zoomFactor);
  
  // Remember the dimensions of this
  display_width = width;
  display_height= height;
  
  if(!largeImage) { // Setup for larger images
    scrollWidth  = NUMCOLS - DEF_DISPLAY_SIZE + 12;
    pageIncW     = display_height*sizeFactor;
    scrollHeight = NUMROWS - DEF_DISPLAY_SIZE + 12;
    pageIncH     = display_height*sizeFactor;
  }
  else { // Setup for smaller images
    scrollWidth  = file_width - width;
    pageIncW     = display_height;
    scrollHeight = file_height - height;
    pageIncH     = display_height;
  }
  
  xoffset = yoffset = 0;
  
  if(file_height>0) { // Now set the scrollbars
    XtVaSetValues(vScrollW, XmNmaximum, scrollHeight+(int)(pageIncH*0.9), NULL);
    XtVaSetValues(vScrollW, XmNminimum, 0, NULL);
    XtVaSetValues(vScrollW, XmNsliderSize, (int)(pageIncH*0.9), NULL);
    XtVaSetValues(vScrollW, XmNvalue, virYoffset, NULL);
    XtVaSetValues(vScrollW, XmNpageIncrement, (int)(pageIncH*0.9), NULL);
    
    XtVaSetValues(hScrollW, XmNmaximum, scrollWidth+(int)(pageIncW*0.9), NULL);
    XtVaSetValues(hScrollW, XmNminimum, 0, NULL);
    XtVaSetValues(hScrollW, XmNsliderSize, (int)(pageIncW*0.9), NULL);
    XtVaSetValues(hScrollW, XmNvalue, virXoffset, NULL);
    XtVaSetValues(hScrollW, XmNpageIncrement, (int)(pageIncW*0.9), NULL);  
  }   
  
  if(widget != 0) // Resized from interface
    virtual_read_activate(display_width/2, display_height/2);
  
} // end function formResize



double rwalog2(float thisSizeFactor)
{
  // function name: rwalog2
  // Compute the log base 2 for this number.

  return(log10(thisSizeFactor) / log10(2.0));

} // end function rwalog2



void zoomText_change(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{
  // function name: zoomText_change
  // They've changed ZOOM on the main screen.  

  if(inited) {
    zoomFactor = rwalog2(wstrtod(zoomTextW));;
    zoomFactor = MAX(-4.0,MIN(zoomFactor,4.0));
    
    // Now redraw with new zoominess
    fullScreen= FALSE;
    virtual_read_activate(display_width/2, display_height/2);
  }
  
} // end function zoomText_change



void toolMode_change(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: toolMode_change
  // They've changed modes on the main screen.  Remember the cursor
  // we're using and set the globals.

  int tag = *((int *) client_data);
  XmToggleButtonCallbackStruct *temp = (XmToggleButtonCallbackStruct *) call_data;

  boxZoom = tag;

  if(tag == USEZOOM) { // Zooming
    old_cursor = zoomCursor;
    updateCursor(drawingW, zoomCursor);
  }
  else if(tag == USEBOX) { // Boxing
    old_cursor = cross_cursor;
    updateCursor(drawingW, cross_cursor);
    if(temp->event != 0)
      raiseIt(boxScreenW);
  }
  else if(tag == USEPOINT) { // Pointing
    old_cursor = None;
    updateCursor(drawingW, None);
    if(temp->event != 0)
      raiseIt(pointScreenW);
  }
  
} // end function toolMode_change



void plotWindow_create()
{
  // function name: plotWindow_create
  // Remember the drawing area window pointer for use later.
  // Setup the file selection widget here.

  Arg arglst[32];
  XtCallbackRec callBck[5];    // Our callbacks
  
  // Setup the file selection widget
  fileSelW = (Widget)XmCreateFileSelectionDialog(mainWindowW,
						"Get file",NULL,0);
  XtSetArg(arglst[0], XmNx, 10);
  XtSetArg(arglst[1], XmNy, 10);
  XtSetArg(arglst[2], XmNdirMask, XmStringCreateLocalized("*.pif"));  
  XtSetArg(arglst[3], XmNdefaultPosition, FALSE); 
  XtSetArg(arglst[4], XmNfontList, robemFontList); 
  XtSetArg(arglst[5], XmNlabelFontList, robemFontList);
  XtSetArg(arglst[6], XmNbuttonFontList, robemFontList);
  XtSetValues(fileSelW, arglst, 4);

  // Add callbacks to file selection widget
  callBck[0].callback = (XtCallbackProc)gotFilename;
  callBck[0].closure  = 0;
  callBck[1].callback = 0;
  callBck[1].closure  = 0;
  XtAddCallbacks(fileSelW, XmNokCallback, callBck);  
  callBck[0].callback = (XtCallbackProc)cancel_fileSel;
  callBck[0].closure  = 0;
  callBck[1].callback = 0;
  callBck[1].closure  = 0;
  XtAddCallbacks(fileSelW, XmNcancelCallback, callBck);  
  XFlush(dsply);

} // end function plotWindow_create



void rwaXDrawCircle(Display *disp, Window drawWin, GC thisgc,
		    int xcen, int ycen, int rad)
{
  // function name: rwaXDrawCircle
  // Draw circles on screen  w/ option to draw thick circles

  int i, xpos, ypos, w;

  xpos = xcen - rad;
  ypos = ycen - rad;
  w = rad*2;

  if(getToggleState(boxThickW)) { // Draw thick circles
    for(i=-1; i<2; i++)
      XDrawArc(disp, drawWin, image_gcWHITE,
	       xpos-i, ypos-i, w+(i*2), w+(i*2), 0, 64*360);
  }
  else
    XDrawArc(disp, drawWin, thisgc, xpos, ypos, w, w, 0, 64*360);
  
} // end function rwaXDrawCircle



void DrawWindowExpose(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: DrawWindowExpose
  // Redraw window e.g. on expose events.

  DrawWindow();

} // end function DrawWindowExpose



void stuffPackedImage(int width, int height, float **pkImg, 
		      int nimg, int startImg)
{
  // function name: stuffPackedImage
  // For memory conservation reasons we've got to get these off the
  // disk. Currently we assume all of the data is in memory and
  // nothing is needed from the disk.

  int l, yOff, x, y;
  int yMirror, thisOffset, firstTime;
  int sectionsPerRow, rowsPerImage, thisNumSecs;
  float hTsf, wTsf;
  int  numRowsDone;
  float *thisBuffer, *thisBuffer2, thisPixel, *q, *w;
  BYTE *s;
  P_INFO *thisInfo, *foundID;

  firstTime = TRUE;
  if(fileType==PACKEDPORTABLE) { // Stuff the first one
    
    calcCircleStats(pkImg[startImg-1], &purdueHeader.img_bkg, 
		    &purdueHeader.img_min, &purdueHeader.img_max, 
		    &purdueHeader.img_avg, &purdueHeader.img_std,
		    purdueHeader.numCols, purdueHeader.numRows, 
		    purdueHeader.numCols);
    
    // Write stats
    sprintf(outStr,"\nStatistics for file [%s] Image [%d]\n\
\n\
File size   [%d]x[%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
\n\
\n",
	    fileName.addr, startImg,
	    purdueHeader.numCols, purdueHeader.numRows, 
	    purdueHeader.nimg, purdueHeader.img_bkg,
	    purdueHeader.img_min, purdueHeader.img_max, 
	    purdueHeader.img_avg, purdueHeader.img_std);
    TextSet(statTextW, outStr, FALSE);
  }
  
  // Computer numbers from 0 not 1
  startImg--;
  if(startImg > nimg)
    startImg = nimg-1;
  
  // Calculate zoominess
  calcSizeFactor(zoomFactor);

  // Just to make sure let's init this
  for(l=0; l<3*buffer_height; l++)
    memcpy((&image_buffer1[l*buffer_width]),stufBuf,buffer_width);
  
  // Now find out how we're going to put these on the display
  sectionsPerRow = MIN(buffer_width, display_width)/(int)(width*sizeFactor);
  sectionsPerRow= MAX(1,sectionsPerRow);

  rowsPerImage = MIN(buffer_height,display_height)/(int)(height*sizeFactor);
  rowsPerImage = MAX(1,rowsPerImage);

  thisNumSecs = (sectionsPerRow * rowsPerImage) + startImg;
  thisNumSecs = MAX(startImg, MIN(thisNumSecs,nimg));

  // Stuff globals
  packSectionsPerRow = sectionsPerRow;
  packStartImage = startImg;

  init_intarr(histogram, 257, 0); // initialize histogram
  s = (BYTE *)image_buffer1_disp;
  
  // Convert integer to char within our colormap and put it
  // in the display buffer(s)
  lastImg= thisNumSecs;
  if(haveParamFile && mapProj)
    lastImg = thisNumSecs = startImg +1;

  if(projASU) { // Setup for projected ASU
    int thisRow;
    thisRow = startImg / maxPhi;
    numRowsDone = thisRow-1;
    startImg = 0;  // We're just using the loop to cycle through them all
    // We have less to do if we start later  ???
    thisNumSecs -= (thisRow * maxPhi);
  }
  
  for (l=startImg; l<thisNumSecs; l++) { // Do all sections
    int rowOffset,colOffset,thisL;
    
    if(projASU) {
      // They have some projections in ASU form
      if(l%sectionsPerRow >= maxPhi) {
	// They've zoomed so small they have more than maxPhi perline
	thisL = maxPhi-1; // Just do a blank one
	thisNumSecs++;      // And make them do another one as
                            // this wasn't accounted for above
      }
      else if(l%sectionsPerRow == 0) { // Go to next line
	numRowsDone++;
	thisL = numRowsDone*maxPhi;
      }
      else { // Just do this line
	thisL = numRowsDone*maxPhi + l%sectionsPerRow;
      }
      
      if(thisL >= nimg) { // Don't even bother to do this one
	thisL = maxPhi-1;
	continue;
      }
      thisL = MIN(thisL,nimg-1);
    }
    else if(haveParamFile) { // They may have sorted the boxes
      foundID = NULL;
      thisInfo = info.info;
      while(thisInfo != NULL) {
	if(thisInfo->rank==(l+1)) { // Found it
	  foundID = thisInfo;
	  thisL = thisInfo->particle_number-1;
	  break;
	}
	else 
	  thisInfo = thisInfo->next;
      }
    }
    else { // normal
      thisL = l;
    }

    q = thisBuffer= (float *)pkImg[thisL]; 
    
    rowOffset = ((l-startImg) / sectionsPerRow) * height * sizeFactor;
    colOffset = ((l-startImg) % sectionsPerRow) * width  * sizeFactor;
    
    imgNums[l].x= colOffset;
    imgNums[l].y= rowOffset;
    if(thisInfo==NULL)
      imgNums[l].particle_number = l+1; /* Couldn't find it */
    else 
      imgNums[l].particle_number = thisL+1;
    
    hTsf = height*sizeFactor;
    wTsf = width *sizeFactor;
    for (y=0; y<hTsf; y++) {
      // Tim draws from the bottom up - We are opposite
      yOff= (rowOffset+y)*buffer_width + colOffset;
      yMirror = (int)(y*recipSizeFactor) * width;
      for (x=0; x<wTsf; x++) { // Clip the entire image
	
	s = image_buffer1_disp + yOff + x;
	q = thisBuffer + (yMirror + (int)(x*recipSizeFactor));
	
	// We've got to scale these for now because we may get one
	// that is larger than we've got colormap entries for.
	
	if(minPackPixel < 0.0)
	  thisPixel= *q + (0.0 - minPackPixel);
	else 
	  thisPixel= *q;
	      
	thisOffset = (int)(thisPixel*scalePackFactor) + 6;
	
	// This is done from the part proc screen so we can see
	// something next to the projected map image. This should
	// clip all but the darkest images at the top but it's better
	// than nothing

	if(bumpContrast) { // Let's just double it
	  thisOffset *= 2;
	  thisOffset = MAX(MIN(255,thisOffset),0);
	}
		  
	if(thisOffset >= 0 && thisOffset < 256) { // OK to do
	  *(s)= thisOffset;
	  if(useGUI) XPutPixel(image_image1, x+colOffset, y+rowOffset,
			       pixels_return[thisOffset]);
	  histogram[thisOffset]+= 1;
	}
	else {
	  if(firstTime)
	    printf ("colormap problems 2\nx[%d]y[%d] q[%d]\n",x,y,(int)*(q));
	  firstTime = FALSE;
	}

	if(flickerFix2D) { // DO for special case flickering
	  thisBuffer2= (float *)fix2DpackImg_diff[thisL]; 
	  w = thisBuffer2 + yMirror + (int)(x*recipSizeFactor);
	  if(*w) {
	    if(useGUI) XPutPixel(image_image1, x+colOffset, y+rowOffset,
		      pixels_return[RED]);
	  }
	}
	
      }
    }

    if(projASU) { // They have some projections in ASU form
    }
    else if(haveParamFile && dispCenters) { // Let's draw the centers
      int xZoom,yZoom;
      
      if(foundID != NULL) { // OK to draw
	   xZoom = foundID->x*sizeFactor - 4;
	   yZoom = (purdueHeader.numRows-foundID->y)*sizeFactor - 4;

	   // Draw one-half of the "X"
	   x = xZoom;
	   for(y=yZoom; y<=yZoom+8; y++) {
	     // Tim draws from the bottom up - We are opposite 
	     if(!foundID->toggle) {
	       if(useGUI) XPutPixel(image_image1, x+colOffset, y+rowOffset,
				 pixels_return[RED]);
	     } else {
	       if(useGUI) XPutPixel(image_image1, x+colOffset, y+rowOffset,
				 pixels_return[BLUE]);
          }
	     x++;
	   }

	   // Draw other-half of the "X"
	   y = yZoom;
	   for(x=xZoom+8; x>=xZoom; x--) {
	     // Tim draws from the bottom up - We are opposite
	     if(!foundID->toggle) {
	       if(useGUI) XPutPixel(image_image1, x+colOffset, y+rowOffset,
				 pixels_return[RED]);
	     } else {
	       if(useGUI) XPutPixel(image_image1, x+colOffset, y+rowOffset,
				 pixels_return[BLUE]);
	     }
           y++;
	   }
      }
      
    }

    if(projASU) { // They have some projections in ASU form
    }
    else if(haveParamFile && mapProj) { // Display a map projection
      int yOut1,yOut2;
      float thisOmega,thisTheta,thisPhi,thisSF;
      float maxF,minF,z1,z2,thisFloat;
      float *projn,*map2d,*thisQ;

      //  NOTE USE OF OMEGA = 90.0 TO GET CORRECT ORIENTATIONS
      
      if(foundID->projection==NULL) { // Then we've got to create it
	
	raiseIt(statusScreenW);
	if(checkStatus("Creating MAP projection ... ")) {
	  // Do nothing for now
	}
	
	// ORig way that looks 180 degrees off
	thisOmega = foundID->omega + 90.0;
	thisOmega = foundID->omega - 90.0;  // WORKS!!
	thisTheta = foundID->theta;
	thisPhi   = foundID->phi;
	z1 = 1.0;
	z2 = mapSecs;
	CK(projn = (float *)malloc(sizeof(float)*mapCols*(mapRows+2)));
	CK(map2d = (float *)malloc(sizeof(float)*mapCols*(mapRows+2)));
	
	map_prj__(map2d, float_image_buffer1,
		 &mapCols, &mapRows, &mapSecs,
		 &thisTheta, &thisPhi, &thisOmega, 
		 &z1, &z2);

	free(projn);
	// We've got to flip the projection in the Y direction.
	// Tim draws from the bottom up -we're opposite
	
	for (y=0; y<(mapRows/2); y++) { // Rows
	  yOut1 = (mapRows - y - 1)*mapCols;
	  yOut2 = y*mapCols;
	  for (x=0; x<mapCols; x++) { // Cols
	    thisFloat        = map2d[yOut2+x];
	    map2d[yOut2+x] = map2d[yOut1+x];
	    map2d[yOut1+x] = thisFloat;
	  }
	}

	foundID->projection = map2d;
	if(useGUI) XtUnmanageChild(statusScreenW);
	if(useGUI) XFlush(dsply);
      }

      map2d = foundID->projection;
      farray_minmax(map2d, &minF, &maxF, mapRows*mapCols);
      thisSF = (float)(tryColors-7.0)/(maxF-minF);
      
      // Now place this projection next to the original
      colOffset += (int)(width*sizeFactor);
      
      // Suppose we've got a binned MAP and a normal sized
      // particle file.  Let's account for that with sizeFactor
      
      sizeFactor *= (purdueHeader.numCols+1.0)/(mapCols+1.0);
      recipSizeFactor = 1.0 / sizeFactor;
      height = mapRows;
      width  = mapCols;
      for (y=0; y<(height*sizeFactor); y++) {
	// Tim draws from the bottom up - We are opposite
	yOff= ((rowOffset+y)*buffer_width + colOffset);
	yMirror = ((int)(y*recipSizeFactor))*width;
	for (x=0; x<(width*sizeFactor); x++) { // Clip the entire image
		  
	  thisQ = map2d + yMirror + (int)(x*recipSizeFactor);
	  
	  thisFloat= *thisQ;
	  
	  thisOffset = (int)((thisFloat-minF)*thisSF) + 6;
	  
	  if(thisOffset >= 0 && thisOffset<256) {
	    if(useGUI) XPutPixel(image_image1, x+colOffset, y+rowOffset,
				 pixels_return[thisOffset]);
	  }
	  else {
	    if(firstTime)
	      printf ("colormap probs 3\nx[%d]y[%d] q[%d]\n",x,y,(int)*(q));
	    firstTime = FALSE;
	  }
	  
	}
      }
      
      if(mapProjContour) {
	// Now draw the contour in the main window
	int saveEditBoxZoom,deltaX,deltaY;
	float thisTour;

	// drawContour uses editBoxZoom so let's stuff it
	// temporarily with the main window zoom factor
	// adjusted for any binning discrepancy between
	// particles and map

	saveEditBoxZoom = editBoxZoom;
	editBoxZoom = zoomFactor + 1;
	if(purdueHeader.numCols > mapCols)
	  editBoxZoom += 1;   // Map is binned
	calcSizeFactor(zoomFactor);
	deltaX = (foundID->x - purdueHeader.numCols/2.0) * sizeFactor;
	deltaY = (foundID->y - purdueHeader.numRows/2.0) * sizeFactor;
	// drawContour depends on these globals
	numCols = mapCols;
	numRows = mapRows;
	thisTour = wstrtod(projContourTextW);
	if(useGUI) drawContour(thisTour, TRUE, 0, drawingW, map2d,
			       deltaX, deltaY, image_buffer1);
	editBoxZoom = saveEditBoxZoom;
      }
    }
    
  }
  
} // end function stuffPackedImage



void stuffDisplayBoxBuffer(int thisBox, char *filename, 
			   int thisZoomFactor)
{
  // function name: stuffDisplayBoxBuffer
  // For memory conservation reasons we've got to get these off the
  // disk. We assume now that all data are in memory already so no
  // reading of the disk required.

  int i, j, l, k, outHeight, outWidth, logicalRecordLength, yOff;
  int yOut, yIn, xIn, beginY, beginX, min, max, endX, endY, thisBoxRadius;
  int thisPixel;
  float  tempBox_int_buffer[(MAXFILESIZE*2)];
  FILE *fIn;

  if(numBoxes <= 0)
    return;
    
  thisBoxRadius = featherRadius;
  
  init_intarr(histogram, 257, 0); // initialize histogram
  
  min = INT_MAX;
  max = INT_MIN;
  scaleBoxFactor = (tryColors - 7.0) / 256.0;
  
  calcBoxSizeFactor(thisZoomFactor);
  
  // Init display buffer
  for (k=0; k<BOX_HEIGHT; k++)
    for (l=0; l<BOX_WIDTH; l++)
      if(useGUI) XPutPixel(editBox_image1, l, k, pixels_return[BLACK]);
  
  // Get dimensions
  outHeight= outWidth = ((thisBoxRadius*2)+1)*sizeFactor;
  if(outHeight > BOX_HEIGHT || outWidth > BOX_WIDTH) { // Too large
    askThem("[STUFFDISPLAYBOXBUFFER] Error - box too large for buffer.",
	    XmDIALOG_INFORMATION);
    return;
  }
  
  outHeight = MAX(0,MIN(outHeight,BOX_HEIGHT));
  outWidth = MAX(0,MIN(outWidth,BOX_WIDTH));
  
  beginY = box[thisBox].y - thisBoxRadius;
  beginX = box[thisBox].x - thisBoxRadius;
  endY   = box[thisBox].y + thisBoxRadius;
  endX   = box[thisBox].x + thisBoxRadius;
  
  if(beginY < 0        || beginX < 0 ||
     endX >= file_width || endY >= file_height) { // Box is outside of file
    for (k=0; k<BOX_HEIGHT*BOX_WIDTH; k++) {
      editBox_mirror_buffer[k] = BLACK;
      editBox_undo_buffer[k]   = BLACK;
    }
    
    // Mark bad box with and X
    l = 0;
    for (k=0; k<outHeight; k++) {
      editBox_mirror_buffer[k*BOX_WIDTH + l] = 128.0;
      editBox_undo_buffer[k*BOX_WIDTH + l]   = 128.0;
      l++;
    }
    l = outWidth - 1;
    for (k=0; k<outHeight; k++) {
      editBox_mirror_buffer[k*BOX_WIDTH + l] = 128.0;
      editBox_undo_buffer[k*BOX_WIDTH + l]   = 128.0;
      l--;
    }
    
    sprintf(outStr,"Box number [%d] is outside of file ",thisBox+1);
    askThem(outStr, XmDIALOG_INFORMATION);
    return;
  }
  
  if(!box[thisBox].inMemory) { // Read it afresh from the disk
    if ( (fIn = fopen(filename,"r")) == NULL ) {
      askThem("[STUFFDISPLAYBOXBUFFER] Error reading file.",
	      XmDIALOG_INFORMATION);
      return;
    }
    
    // Read off the disk
    logicalRecordLength = BOX_WIDTH;
    beginY = box[thisBox].y - thisBoxRadius;
    beginX = box[thisBox].x - thisBoxRadius;
    for (k=0; k<outHeight; k++) { // Do the y direction
      
      // Use the +k because of EOR marker * 2.0 because it's 
      // shorts not chars
      if(fileType == PORTABLE) { // Don't use the EOR markers
	yOff = headOff + 2.0 * 
	  ((beginY+(int)(k*recipSizeFactor))*file_width +  beginX);
      }
      else if(fileType==BYTEPORTABLE) { // Don't use the EOR markers
	yOff = headOff + 
	  (beginY+(int)(k*recipSizeFactor))*file_width +  beginX;
      }
      else { // DO use the EOR markers
	yOff = headOff + 
	  2.0 * (
		  ((beginY  + (int)(k*recipSizeFactor)) * 
		   file_width) +
		  // Y direction
		  beginX +                            // X
		  (beginY + (int)(k*recipSizeFactor)) 
		  ); // EOR markers
      }
      // SEEK it
      sysError= fseek(fIn,yOff,0);
      if(sysError) { // Can't go on
	int k2,l2;
	askThem("[SDB seek] Error reading file.", XmDIALOG_INFORMATION);
	// Problems
	for (k2=0; k2<outHeight; k2++) { // Do the y direction
	  yOut = (k2 * BOX_WIDTH);
	  // Stuff all > radius in background character
	  for (l2=0; l2<outWidth; l2++) { // Stuff the outbuffer
	    editBox_int_buffer[yOut+l2]     = 0.0;
	    editBox_mirror_buffer[yOut+l2]  = 0.0;
	    editBox_undo_buffer[yOut+l2]    = 0.0;
	    editBox_display_buffer[yOut+l2] = 0.0;
	  }
	}
	fclose(fIn);
	return;
      }
      
      if(fileType==BYTEPORTABLE || fileType==PORTABLE) { // Read byte data
	int thisI, thisYoff;
	thisYoff = (beginY+(int)(k*recipSizeFactor))*file_width +  beginX;
	
	for(thisI=0; thisI<logicalRecordLength; thisI++) {
	  if(thisYoff+thisI < file_width*file_height)
	    tempBox_int_buffer[thisI] = inCore_buffer[thisYoff + thisI];
	  else
	    tempBox_int_buffer[thisI] = purdueHeader.img_bkg;
	}
	
	// Topmost point of box
	yOut = (k * BOX_WIDTH);
	// Stuff display and mirror buffers
	for (l=0; l<outWidth; l++) { // Stuff the outbuffer
	  // Leftmost point in box
	  xIn = (l * recipSizeFactor);
	  editBox_int_buffer[yOut+l]= 
	    editBox_undo_buffer[yOut+l]= 
	    editBox_mirror_buffer[yOut+l]= 
	    (float)tempBox_int_buffer[xIn];
	  thisPixel= (int)tempBox_int_buffer[xIn];
	  min = MIN(thisPixel,min);
	  max = MAX(thisPixel,max);
	}
      }
      else {
	askThem("[SDB fileType] I don't understand this type.",
		XmDIALOG_INFORMATION);
	fclose(fIn);
	return;
      }
      
    }
    
    fclose(fIn);
  }
  else { // We should already have it because it's not deleted yet
    int thisWidth;
    float *thisBoxBuf;
    
    thisBoxBuf = box[thisBox].edit_mirror_buffer;
    
    // Get dimensions
    thisWidth = (thisBoxRadius*2)+1;
    outHeight= outWidth = ((thisBoxRadius*2)+1)*sizeFactor;
    outHeight = MAX(0,MIN(outHeight,BOX_HEIGHT));
    outWidth = MAX(0,MIN(outWidth,BOX_WIDTH));
    
    for (i=0; i<outHeight; i++) {
      yIn = i * recipSizeFactor;
      yOut = (i * BOX_WIDTH);
      for (j=0; j<outWidth; j++) {
	xIn = j * recipSizeFactor;
	editBox_mirror_buffer[yOut+j]= 
	  editBox_undo_buffer[yOut+j]= 
	  editBox_int_buffer[yOut+j] = thisBoxBuf[(yIn*thisWidth)+xIn];
	thisPixel= (int)thisBoxBuf[(yIn*thisWidth)+xIn];
	min = MIN(thisPixel,min);
	max = MAX(thisPixel,max);
      }
    }
  }
  
  calcCircleStats(editBox_mirror_buffer, &purdueHeader.img_bkg, 
		  &purdueHeader.img_min, &purdueHeader.img_max, 
		  &purdueHeader.img_avg, &purdueHeader.img_std,
		  outWidth, outHeight, BOX_WIDTH);
  
  backChar = purdueHeader.img_bkg;
  appodize_it(editBox_int_buffer, outWidth, 
	      outHeight, BOX_WIDTH,
	      fileType, bkgrndMethod,
	      boxMethod, packRadius,
	      boxRadius, featherRadius,
	      recipSizeFactor, sizeFactor,
	      tryColors, featherMode,
	      &backChar, thisBoxRadius,
	      forceAvgBkg, purdueHeader.img_avg);
  
  // Stuff appropriate buffers
  for (k=0; k<outHeight; k++) { // Do the y direction
    // Where is the center?
    yOut  = (k * BOX_WIDTH);
    for(l=0; l<outWidth; l++) { // Stuff the outbuffer
      editBox_mirror_buffer[yOut+l]= 
	editBox_undo_buffer[yOut+l]= 
	editBox_int_buffer[yOut+l];
    }
  }
  
  // OK let's now fit this image to our colormap
  scaleBoxFactor = scaleBoxFactor * (255.0 / max);
//  printf("[%d]  max==[%f]  bkg[%f]\n", thisBox, (float)max, backChar);
  for (i=0; i<outHeight; i++) {
    yOut = (i * BOX_WIDTH);
    for (j=0; j<outWidth; j++) {
      editBox_display_buffer[yOut+j] = 
	(BYTE)(((float)editBox_int_buffer[yOut+j] * 
		scaleBoxFactor) + 6.0);
      if(useGUI) XPutPixel(editBox_image1, j, i, 
			   pixels_return[(int)(((float)
						editBox_int_buffer[yOut+j] * 
						scaleBoxFactor) + 6.0)]);
      
      histogram[(editBox_display_buffer[yOut+j])] += 1;
    }
  }
  
  // Now let's make sure our backDispChar is right
  backDispChar = (int)(backChar*scaleBoxFactor + 6.0);
  
  histogram_activate(); 
  
} // end function stuffDisplayBoxBuffer



void stuffRectangleBuffer(float *outBuf, char *filename)
{
  // function name: stuffRectangleBuffer
  // For memory conservation reasons we've got to get these off the
  // disk. Same - assume memory holds all data now - no disk read.

  int l, k, outHeight, outWidth, logicalRecordLength, yOff;
  int yOut, xIn, beginY, beginX;
  unsigned short tempBox_buffer[(MAXFILESIZE*2)] ;
  float recipSizeFactor;

  if(blindWriting) { // Just get from memory
    outWidth  = purdueHeader.numCols;
    outHeight = purdueHeader.numRows;
    for (k=0; k<outHeight*outWidth; k++) 
      outBuf[k] = packImg[imgNum-1][k]; // Leftmost point in box
    return;
  }
  
  // NO zoom
  recipSizeFactor= 1.0;
//  sizeFactor= 1.0;
  
  outHeight = rectBox.lowy - rectBox.y;
  outWidth  = rectBox.lowx - rectBox.x;
  rectBox.width  = rectBox.lowx - rectBox.x;
  rectBox.height = rectBox.lowy - rectBox.y;

  if(outWidth > MAXFILESIZE) { // Too large
    askThem("[STUFFRECTANGLEBUFFER] Box too large.",
	    XmDIALOG_INFORMATION);
    return;
  }

  // Read off the disk
  logicalRecordLength = outWidth;
  beginY = rectBox.y;
  beginX = rectBox.x;
  for (k=0; k<outHeight; k++) { // Do the y direction
      
    // Use the +k because of EOR marker * 2.0 because it's 
    // shorts not chars
    if(fileType == PORTABLE) { // Don't use the EOR markers
      yOff = (beginY+(int)(k*recipSizeFactor))*file_width +  beginX;
    }
    else if(fileType == BYTEPORTABLE) { // Don't use the EOR markers
      yOff = (beginY+(int)(k*recipSizeFactor))*file_width +  beginX;
    }
    else if(crysDisping) {
      // Just get the disply buffer - we're crysDisping a packed box
      yOff = 0;
    }
    else { // DO use the EOR markers
      yOff = 
	// ORIGway   (2.0 * (
	(1.0 * (
		((beginY  + (int)(k*recipSizeFactor)) * file_width) +
		// Y direction
		beginX +                            // X
		(beginY + (int)(k*recipSizeFactor)) // EOR markers
		));
    }
    
    if(crysDisping) {
      // Just get the disply buffer - we're crysDisping a packed box
    }
    else {
      // SEEK it
    }
    
    if(fileType==BYTEPORTABLE || fileType==PORTABLE) { // Read shorts from inCore_buffer
      yOut = k * outWidth; // Topmost point of box
      // Stuff display and mirrro buffers
      for (l=0; l<outWidth; l++) { // Stuff the outbuffer
	xIn = l * recipSizeFactor; // Leftmost point in box
	outBuf[yOut+l] = inCore_buffer[yOff+l];
      }
    }
    else if(crysDisping) {
      // Just get the disply buffer - we're crysDisping a packed box
      
      for(l=0;l<logicalRecordLength;l++)
	tempBox_buffer[l] = (float)image_buffer1[((k+rectBox.y)*
						  buffer_width)+
						 (rectBox.x) + l];
      yOut = k * outWidth; // Topmost point of box
      // Stuff display and mirrro buffers
      for (l=0; l<outWidth; l++) { // Stuff the outbuffer
	xIn = l; // Leftmost point in box
	outBuf[yOut+l] = (float)tempBox_buffer[xIn];
      }
    }
    else {
      askThem("[SRB fileType] I don't understand this type.",
	      XmDIALOG_INFORMATION);
      return;
    }
      
  }
  
} // end function stuffRectangleBuffer



void stuffDisplayMapBuffer(int thisSec, int thisZoomFactor, int boxOffset)
{
  // function name: stuffDisplayMapBuffer
  // Put the MAP section number passed into editBox_display_buffer
  // which will be put on the screen.

  int l, k, thisNumCols, thisNumRows;
  int y, x, lastRow, lastCol, thisOffset, saveBoxOffset;
  float *yMirror, *yMirror2, *yMirrorF, *q, *q2, *thisMap;
  float thisPixelF, thisPixel, offset;
  
  calcBoxSizeFactor(thisZoomFactor);
  
  if(liteSwitch) {
    saveBoxOffset = boxOffset;
    boxOffset = 1;
  }

  if(boxOffset==0) {
    for (k=0; k<BOX_HEIGHT; k++)
      for (l=0; l<BOX_WIDTH; l++)
	XPutPixel(editBox_image1, l, k, pixels_return[BLACK]);
  }
  
  if(liteSwitch)
    boxOffset = saveBoxOffset;

  q = float_image_buffer1; 

  // Convert float to char within our colormap and put it
  // in the display buffer(s)

  // Do only this section
  l = thisSec;
  offset = (minMapPixel < 0.0) ? -minMapPixel : 0.0;

  // Find the zoomed col/row values
  thisNumCols = numCols;
  thisNumRows = numRows;
  if(shadedMaps && !dispSections) { // May be an asymmetrical map
    thisNumCols = largestDim;
    thisNumRows = largestDim;
  }
  lastRow = thisNumRows * sizeFactor;
  lastCol = thisNumCols * sizeFactor;
  boxOffset *= sizeFactor;  // Got to get the offset into zoomed space
  lastRow= MIN(lastRow, BOX_HEIGHT);
  lastCol= MIN(lastCol, BOX_HEIGHT);
  
  init_intarr(histogram, 257, 0); // initialize histogram

  if(shadedMaps && !dispSections) { // We've got depth cued shaded data
    
    // Point to shaded buffer
    if(!radialProjections && !secondMapShade && !radialMapShade)
      thisMap = cuedMap[l].depth;
    else
      thisMap = cuedMap[l].shad;
    
    for (y=0; y<lastRow; y++) {
      // Tim draws from the bottom up - We are opposite
      yMirrorF = (float *)(thisMap + (int)(y*recipSizeFactor)*thisNumCols);
      for (x=0; x<lastCol; x++) { // Clip the entire image
	  
	q2 = (float *)(yMirrorF+ (int)(x*recipSizeFactor));
	
	// We've got to scale these for now because we may get one
	// that is larger than we've got colormap entries for.
	
	thisPixelF = (*q2==0.0) ? 0.0 : (*q2 + offset);
	
	thisOffset = (int)(thisPixelF * scaleMapFactor) + 6;
	
	if(thisOffset >= 0 && thisOffset < 256) {
	  histogram[(int)thisOffset]++;
	  editBox_display_buffer[(y*BOX_WIDTH)+x+boxOffset] = thisOffset;
	  XPutPixel(editBox_image1, x+boxOffset, y, pixels_return[thisOffset]);
	}
	
      }
    }
    
    // Now draw the new histogram
    histogram_activate();
    
  }
  else { // We've got section data
      
    for (y=0; y<lastRow; y++) {
      // Tim draws from the bottom up - We are opposite
      yMirror = (float *)(float_image_buffer1 + l*thisNumRows*thisNumCols +
			  (int)(y*recipSizeFactor)*thisNumCols);
      if(paintMapSwitch)
	yMirror2 = (float *)(paintMap + l*thisNumRows*thisNumCols +
			     (int)(y*recipSizeFactor)*thisNumCols);
      for (x=0; x<lastCol; x++) { // Clip the entire image
	
	q = yMirror + (int)(x*recipSizeFactor);
	
	// We've got to scale these for now because we may get one
	// that is larger than we've got colormap entries for.
	
	thisPixel= *(q) + offset;
	thisOffset = (int)(thisPixel * scaleMapFactor) + 6;
	
	if(thisOffset >= 0 && thisOffset < 256) {
	  histogram[thisOffset]++;
	  XPutPixel(editBox_image1, x+boxOffset, y, pixels_return[thisOffset]);
	  editBox_display_buffer[y*BOX_WIDTH + x + boxOffset] = thisOffset;	  
	}
	
	if(paintMapSwitch) { // special
	  q2 = yMirror2 + (int)(x*recipSizeFactor);
	  if((*q2) != 0.0) {
	    if(normalExcise) { // normal
	      XPutPixel(editBox_image1, x+boxOffset, y, pixels_return[RED]);
	      editBox_display_buffer[y*BOX_WIDTH + x + boxOffset] = RED;
	    }
	    else { // reverse (RSS - NOTE 'else' block same as 'if' block)
	      XPutPixel(editBox_image1, x+boxOffset, y, pixels_return[RED]);
	      editBox_display_buffer[y*BOX_WIDTH + x + boxOffset] = RED;
	    }
	  }
	}

      }
    }
  }
  
} // end function stuffDisplayMapBuffer



void stuffDisplayPackBoxBuffer(int thisSec, int thisZoomFactor, int doApod)
{
  // function name: stuffDisplayPackBoxBuffer
  // Put the packed box passed in on the screen.

  int l, k, thisOffset;
  int y, x, lastRow, lastCol, outWidth, outHeight;
  float *yMirror, *tmpPackImg, *q, thisPixel, offset;

  outWidth  = purdueHeader.numCols;
  outHeight = purdueHeader.numRows;
  tmpPackImg = (float *)malloc(sizeof(float)*outWidth*outHeight);

  for(k=0; k<outHeight*outWidth; k++)  // Do the y direction
    tmpPackImg[k] = packImg[thisSec][k];
  
  if(doApod) {
    calcCircleStats(tmpPackImg, &purdueHeader.img_bkg, 
		    &purdueHeader.img_min, &purdueHeader.img_max, 
		    &purdueHeader.img_avg, &purdueHeader.img_std,
		    outWidth, outHeight, outWidth);
    backChar = purdueHeader.img_bkg;

    appodize_it(tmpPackImg, outWidth, outHeight, outWidth,
		fileType, bkgrndMethod, boxMethod, packRadius,
		boxRadius, featherRadius, recipSizeFactor, sizeFactor,
		tryColors, featherMode, &backChar, featherRadius,
		forceAvgBkg, purdueHeader.img_avg);
  }
  
  calcBoxSizeFactor(thisZoomFactor);

  for (k=0; k<BOX_HEIGHT; k++)
    for (l=0; l<BOX_WIDTH; l++)
      if(useGUI) XPutPixel(editBox_image1, l, k, pixels_return[BLACK]);

  // Convert float to char within our colormap and put it in the
  // display buffer(s) Do only this section
  offset = (minPackPixel < 0.0) ? -minPackPixel : 0.0;

  // Find the zoomed col/row values
  lastRow = MIN(purdueHeader.numRows * sizeFactor, BOX_HEIGHT);
  lastCol = MIN(purdueHeader.numCols * sizeFactor, BOX_HEIGHT);
  
  init_intarr(histogram, 257, 0); // initialize histogram

  for (y=0; y<lastRow; y++) { // Tim draws from the bottom up - We are opposite
    yMirror = (float *)(tmpPackImg + ((int)(y*recipSizeFactor)*outWidth));
    for (x=0; x<lastCol; x++) { // Clip the entire image
	  
      q = yMirror + (int)(x*recipSizeFactor);
      
      // We've got to scale these for now because we may get one
      // that is larger than we've got colormap entries for.
      
      thisPixel= *q + offset;
      
      thisOffset = (int)(thisPixel * scalePackFactor) + 6;
      thisOffset = MAX(MIN(thisOffset, tryColors-1), 0);

      if(useGUI) XPutPixel(editBox_image1, x, y, pixels_return[thisOffset]);
      editBox_display_buffer[y*BOX_WIDTH + x] = (BYTE)thisOffset;
      editBox_mirror_buffer[y*BOX_WIDTH + x]  = *q;
      editBox_undo_buffer[y*BOX_WIDTH + x]    = *q;
      
      histogram[thisOffset]+= 1;
    }
  }
  
  backChar = editBox_undo_buffer[2*BOX_WIDTH + 2];
  backDispChar = (backChar * scalePackFactor) + 6.0;
  histogram_activate();
  
  free(tmpPackImg);
  
} // end function stuffDisplayPackBoxBuffer



void writeBoxCoords(char *fName, int append)
{
  // function name: writeBoxCoords
  // Write the current box coords (bcrd file) to an ASCII disk file.
  
  char infoStr[512];
  int i, writeIt=TRUE;
  int active=0, allzero=TRUE;
  FILE *fOut;
  
  if (append==PARAMSCAN) { // Just rewrite it
    sprintf(outStr, "cp %s boxcoords.save", fName);
    system(outStr); 
  }
  else { // Ask user before overwriting existing bcrd file
    if((fOut = fopen(fName,"r")) != NULL && !autoBoxSwitch && !reBoxSwitch) {
      sprintf(infoStr, "[%s] Already exists - OVERWRITE it?", fName);
      writeIt = askThem(infoStr, XmDIALOG_QUESTION);
      fclose(fOut);
    }
  }

  for (i=0; i<numBoxes; i++) {
    if(!deletePackedImage[i])
      active++;
    if(box[i].x != 0 || box[i].y != 0)
      allzero = FALSE;
  }

  if(allzero == TRUE) {
    sprintf(infoStr, "All box coords = (0,0) - Still write bcrd file?");
    writeIt = askThem(infoStr, XmDIALOG_QUESTION);
  }

  if(writeIt) {

    if( (fOut = fopen(fName,"w")) != NULL ) {
      fprintf(fOut,"BCRD V2.00\n");
      fprintf(fOut,"%d\n",lastNumBox);
      fprintf(fOut,"%d\n", active);
      fprintf(fOut,"%d\n",boxRadius);
      fprintf(fOut,"%d\n",featherRadius);
      for (i=0; i<numBoxes; i++)
	if(!deletePackedImage[i])
	  fprintf(fOut,"%d %d\n", box[i].x, box[i].y);
      fclose(fOut);
    }
    else 
      askThem("Can't open file.", XmDIALOG_INFORMATION);
  }
  
} // end function writeBoxCoords



void writeRAWcontour(char *fName, BYTE *byte_buffer, int ncol, int nrow)
{
  // function name: writeRAWcontour
  // Write the current image in RAW format. Used to get images to the
  // Mac e.g.

  BYTE outChar, *outBuffer, *tmpByte_buffer;
  int i, j, k;
  int writeIt, minByte, maxByte;
  FILE *fOut;
  fString_t forName,rWrite;
  forPifHead_t thisHead;
  
  writeIt= TRUE;
  if ( (fOut = fopen(fName,"r")) != NULL ) {
    sprintf(outStr, "Do you really want to overwrite [%s]?",fName);
    writeIt= askThem(outStr, XmDIALOG_QUESTION);
    fclose(fOut);
    if(!writeIt)
      return;
  }
  
  if ( (fOut= fopen(fName,"w")) == NULL ) {
    askThem("Can't open file.", XmDIALOG_INFORMATION);
    return;
  }
  
  CK(outBuffer= (BYTE *)malloc(sizeof(BYTE)*nrow*ncol));
  
  // Stuff outbuffer with each image
  for (i=0; i<nrow; i++)
    for (k=0; k<ncol; k++)
      outBuffer[i*ncol + k] = byte_buffer[i*BOX_WIDTH + k];
  
  // Now find the min and max
  minByte = 255;
  maxByte = 0;
  for (j=0; j<nrow*ncol; j++) {
    outChar = outBuffer[j];
    minByte = MIN(minByte,(int)outChar);
    maxByte = MAX(maxByte,(int)outChar);
  }
  
  // Now find scale factor
  minByte -= 5;      // We'll put our colors in 0-4
  
  // Now scale the data
  for (j=0; j<nrow*ncol; j++) {
    outChar = outBuffer[j];
    if(outChar == RED)
      outBuffer[j]= (BYTE)2;
    else if(outChar == GREEN)
      outBuffer[j]= (BYTE)3;
    else
      outBuffer[j]= outChar;
  }
  
  // Now find the min and max
  minByte = 255;
  maxByte = 0;
  for (j=0; j<nrow*ncol; j++) {
    outChar = outBuffer[j];
    minByte = MIN(minByte,(int)outChar);
    maxByte = MAX(maxByte,(int)outChar);
  }
	
  fwrite(outBuffer,sizeof(BYTE),(ncol*nrow),fOut);
  
  sprintf(outStr,"RAW file stats:\n\
\n				   \
     Filename = [%s]\n		   \
\n				   \
     Width = [%d]\n		   \
     Height= [%d]\n		   \
\n				   \
     8 bit byte pixels\n	   \
     NO header\n		   \
\n				   \
     Min[%d]\n			   \
     Max[%d]\n			   \
\n				   \
     RED pixel   = 2\n		   \
     GREEN pixel = 3\n",
	  fName,
	  (ncol),nrow,minByte,maxByte);
  
  askThem(outStr, XmDIALOG_INFORMATION);
  
  fclose(fOut);
  
  // Now write PIF file of the above
  pif_init_head(&thisHead);
  thisHead.luNum = 4;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 4;
  forName.addr = malloc(sizeof(char)*264);
  forName.length = 264;
  sprintf(rWrite.addr,"w");
  sprintf(forName.addr,"rawFile.pif");
  
  pif_open(forName.addr, &thisHead.luNum, rWrite.addr);
  
  // *******************************************
  //  WRIITING NEW GLOBAL PIF HEADER
  
  // [1] Write global file header
  
  // Pick appropriate statistics calculation based
  // on data type we have

  thisHead.datai4_min = 0;
  thisHead.datai4_max = 255;
  thisHead.numCols = ncol;
  thisHead.numRows = nrow;
  
  thisHead.mode = 6;
  
  pif_write_gh(&thisHead);
  
  thisHead.img_bkg     = 0;
  thisHead.packRadius  = ncol/2;
  thisHead.datai4_max  = 255;
  thisHead.datar4_max  = 255.0;
  thisHead.datai4_min  = 0;
  thisHead.datar4_min  = 0.0;
  thisHead.borderWidth = 0;
  thisHead.binFactor   = 1;
  thisHead.imgNum      = 1;
  
  pif_write_dh(&thisHead);
  thisHead.imgOffset += 512;  // Just wrote the header
  
  CK(tmpByte_buffer=(BYTE *)malloc(sizeof(BYTE)*
				   ncol*nrow));
  tsbworld_byte(outBuffer, tmpByte_buffer,
		ncol, ncol, nrow);
  pif_write_byte_image_(&thisHead.luNum,
			tmpByte_buffer,
			&thisHead.imgOffset,
			&thisHead.numCols,
			&thisHead.numRows,
			&thisHead.numSecs);
  free(tmpByte_buffer);
  free(outBuffer);
  free(forName.addr);
  free(rWrite.addr);
  pif_close(&thisHead.luNum);
  
} // end function writeRAWcontour



void writeRAW(char *fName)
{
  // function name: writeRAW
  // Write the current image in RAW format. Used to get images to the Mac e.g.

  BYTE outChar, *outBuffer;
  int i, j, k, writeIt;
  int minByte, maxByte, numColsTsizeFactor, numRowsTsizeFactor;
  int numChoices, numWanted, thisNsecs;
  float scaleFactor, minF, maxF, scaleF;
  FILE *fOut;

  XmListAddItem(infoListW, 
		XmStringCreateLocalized("Write current ImgNum"), 1);
  XmListAddItem(infoListW, 
		XmStringCreateLocalized("Write what's visible on screen"), 2);
  XmListAddItem(infoListW, 
		XmStringCreateLocalized("Write entire data set"), 3);
  XFlush(dsply);
  
  sprintf(outStr,"Choose how much to write below:");
  numChoices = 3;
  numWanted= askThemAnswer(outStr,numChoices);

  if ( (fOut = fopen(fName,"r")) != NULL ) {
    sprintf(outStr, "Do you really want to overwrite [%s]?",fName);
    writeIt= askThem(outStr, XmDIALOG_QUESTION);
    fclose(fOut);
    if(!writeIt)
      return;
  }
  
  if ( (fOut= fopen(fName,"w")) == NULL ) {
    askThem("Can't open file.", XmDIALOG_INFORMATION);
    return;
  }

  
  if(shadedMaps && !dispSections) { // Write cued map
    
    calcSizeFactor(zoomFactor);
    
    if(numWanted==1) { // Just do a screen dump
      
      numColsTsizeFactor = display_width;
      numRowsTsizeFactor = display_height;
      CK(outBuffer= (BYTE *)malloc(sizeof(BYTE)*
				   (numRowsTsizeFactor)*
				   (numColsTsizeFactor)));
      minF = maxF = image_buffer1_disp[0];
      
      for (i=0; i<numRowsTsizeFactor; i++) {
	for (k=0; k<numColsTsizeFactor; k++) {
	  maxF = MAX(maxF, image_buffer1_disp[(i*buffer_width)+k]);
	  minF = MIN(minF, image_buffer1_disp[(i*buffer_width)+k]);
	}
      }
      
      scaleF = 254.0 / (maxF-minF);
      for(i=0; i<numRowsTsizeFactor; i++)
	for(k=0; k<numColsTsizeFactor; k++)
	  outBuffer[(i*numColsTsizeFactor)+k]=
	    (BYTE)(((float)image_buffer1_disp
		    [(i*buffer_width)+k] -minF)*scaleF);
      
      // Now find the min and max
      minByte = INT_MAX;
      maxByte = INT_MIN;
      for(j=0; j<numRowsTsizeFactor*numColsTsizeFactor; j++) {
	outChar = outBuffer[j];
	minByte = MIN(minByte,(int)outChar);
	maxByte = MAX(maxByte,(int)outChar);
      }
      
      // Now find scale factor
      scaleFactor = 255.0 / (float)(maxByte-minByte);
      
      // Now scale the data
      for (j=0; j<numRowsTsizeFactor*numColsTsizeFactor; j++) {
	outChar = outBuffer[j];
	outBuffer[j] = (float)(outChar-minByte)*scaleFactor;
      }
      
      // Now find the min and max
      minByte = INT_MAX;
      maxByte = INT_MIN;
      for (j=0; j<numRowsTsizeFactor*numColsTsizeFactor; j++) {
	outChar = outBuffer[j];
	minByte = MIN(minByte,(int)outChar);
	maxByte = MAX(maxByte,(int)outChar);
      }
      
      fwrite(outBuffer, sizeof(BYTE),
	     (numRowsTsizeFactor*numColsTsizeFactor), fOut);
      
      sprintf(outStr,"RAW file stats:\n\
\n				       \
     Frames= [%d]\n		       \
     Width = [%d]\n		       \
     Height= [%d]\n		       \
\n				       \
     8 bit byte pixels\n	       \
     NO header\n		       \
\n				       \
     Min[%d]\n							\
     Max[%d]\n",1,(numColsTsizeFactor),numRowsTsizeFactor,
	      minByte,maxByte);
      askThem(outStr, XmDIALOG_INFORMATION);
      
    }
    else { // Do as many as wanted
      int firstJ, lastJ;
      
      if(numWanted == 0) {
	firstJ = imgNum - 1;
	lastJ  = firstJ + 1;
	thisNsecs = 1;
      }
      else {
	firstJ= 0;
	lastJ = nsecs;
	thisNsecs = nsecs;
      }
      
      numColsTsizeFactor = largestDim * sizeFactor;
      numRowsTsizeFactor = largestDim * sizeFactor;
      CK(outBuffer= (BYTE *)malloc(sizeof(BYTE)*
				   numRowsTsizeFactor*
				   numColsTsizeFactor*
				   thisNsecs));
      
      for (j=firstJ; j<lastJ; j++) { // Stuff outbuffer with each image
	// EditBox Zoom is off by 1.0
	stuffDisplayMapBuffer(j,(int)(zoomFactor+1.0),0);
	for(i=0; i<numRowsTsizeFactor; i++)
	  for(k=0; k<numColsTsizeFactor; k++)
	    outBuffer[(j-firstJ)*numColsTsizeFactor  +
		      i*numColsTsizeFactor*thisNsecs + 
		      k] = 
	      editBox_display_buffer[(i*BOX_WIDTH)+k];
      }
      
      // Now find the min and max
      minByte = INT_MAX;
      maxByte = INT_MIN;
      for (i=0; i<thisNsecs*numRowsTsizeFactor*numColsTsizeFactor; i++) {
	outChar = outBuffer[i];
	minByte = MIN(minByte,(float)outChar);
	maxByte = MAX(maxByte,(float)outChar);
      }
      
      // Now find scale factor
      scaleFactor = 254.0 / (float)(maxByte-minByte);
      
      // Now scale the data
      for (i=0; i<thisNsecs*numRowsTsizeFactor*numColsTsizeFactor; i++) {
	outChar = outBuffer[i];
	outBuffer[i] = (float)(outChar-minByte)*scaleFactor;
      }
      
      // Now find the min and max
      minByte = INT_MAX;
      maxByte = INT_MIN;
      for (i=0; i<thisNsecs*numRowsTsizeFactor*numColsTsizeFactor; i++) {
	outChar = outBuffer[i];
	minByte = MIN(minByte,(int)outChar);
	maxByte = MAX(maxByte,(int)outChar);
      }
      
      fwrite(outBuffer, sizeof(BYTE),
	     (numRowsTsizeFactor*numColsTsizeFactor*thisNsecs), fOut);
      
      sprintf(outStr,"RAW file stats:\n\
\n				       \
     Frames= [%d]\n		       \
     Width = [%d]\n		       \
     Height= [%d]\n		       \
\n				       \
     8 bit byte pixels\n	       \
     NO header\n		       \
\n				       \
     Min[%d]\n								\
     Max[%d]\n",thisNsecs,(thisNsecs*numColsTsizeFactor),numRowsTsizeFactor,
	      minByte,maxByte);
      askThem(outStr, XmDIALOG_INFORMATION);
      
      free(outBuffer);
    }
	
    fclose(fOut);
    
  }
  else if(projASU || fileType==PACKEDPORTABLE) { // PROJASU
    
    calcSizeFactor(zoomFactor);
    
    if(fileType==PACKEDPORTABLE) {
      mapCols = purdueHeader.numCols;
      mapRows = purdueHeader.numRows;
    }
    
    if(numWanted==0) { // Just write the current imgNum
      stuffDisplayPackBoxBuffer(0, (int)(zoomFactor+1.0),FALSE);
      numColsTsizeFactor = mapCols * sizeFactor;
      numRowsTsizeFactor = mapRows * sizeFactor;
      CK(outBuffer= (BYTE *)malloc(sizeof(BYTE)*
				   numRowsTsizeFactor*
				   numColsTsizeFactor));
      
      for (i=0; i<numRowsTsizeFactor; i++)
	for (k=0; k<numColsTsizeFactor; k++)
	  outBuffer[i*numColsTsizeFactor + k]=
	    editBox_display_buffer[i*BOX_WIDTH + k];
      
      // Now find the min and max
      minByte = INT_MAX;
      maxByte = INT_MIN;
      for (j=0; j<numRowsTsizeFactor*numColsTsizeFactor; j++) {
	outChar = outBuffer[j];
	minByte = MIN(minByte,(int)outChar);
	maxByte = MAX(maxByte,(int)outChar);
      }
      
      // Now find scale factor
      scaleFactor = 255.0 / (float)(maxByte-minByte);
      
      // Now scale the data
      for (j=0; j<numRowsTsizeFactor*numColsTsizeFactor; j++) {
	outChar = outBuffer[j];
	outBuffer[j] = (float)(outChar-minByte)*scaleFactor;
      }
      
      // Now find the min and max
      minByte = INT_MAX;
      maxByte = INT_MIN;
      for (j=0; j<numRowsTsizeFactor*numColsTsizeFactor; j++) {
	outChar = outBuffer[j];
	minByte = MIN(minByte,(int)outChar);
	maxByte = MAX(maxByte,(int)outChar);
      }
      
      fwrite(outBuffer,
	     sizeof(BYTE),(numRowsTsizeFactor*
			   numColsTsizeFactor),fOut);
      
      sprintf(outStr,"RAW file stats:\n\
\n				       \
     Frames= [%d]\n		       \
     Width = [%d]\n		       \
     Height= [%d]\n		       \
\n				       \
     8 bit byte pixels\n	       \
     NO header\n		       \
\n				       \
     Min[%d]\n							\
     Max[%d]\n",1,(numColsTsizeFactor),numRowsTsizeFactor,
	      minByte,maxByte);
      
      askThem(outStr, XmDIALOG_INFORMATION);
      
      
    }
    else if(numWanted==1) { // Just do a screen dump
      
      numColsTsizeFactor = display_width;
      numRowsTsizeFactor = display_height;
      CK(outBuffer= (BYTE *)malloc(sizeof(BYTE)*
				   numRowsTsizeFactor*
				   numColsTsizeFactor));
      
      minF = maxF = image_buffer1_disp[0];
      for (i=0; i<numRowsTsizeFactor; i++)
	for (k=0; k<numColsTsizeFactor; k++) {
	  maxF = MAX(maxF, image_buffer1_disp[i*buffer_width + k]);
	  minF = MIN(minF, image_buffer1_disp[i*buffer_width + k]);
	}
      
      scaleF = 254.0 / (maxF-minF);
      for (i=0; i<numRowsTsizeFactor; i++)
	for (k=0; k<numColsTsizeFactor; k++)
	  outBuffer[(i*numColsTsizeFactor)+k]=
	    (BYTE)(((float)image_buffer1_disp[i*buffer_width + k]-minF)
		   *scaleF);
      
      // Now find the min and max
      minByte = INT_MAX;
      maxByte = INT_MIN;
      for (j=0; j<numRowsTsizeFactor*numColsTsizeFactor; j++) {
	outChar = outBuffer[j];
	minByte = MIN(minByte,(int)outChar);
	maxByte = MAX(maxByte,(int)outChar);
      }
      
      // Now find scale factor
      scaleFactor = 255.0 / (float)(maxByte-minByte);
      
      // Now scale the data
      for (j=0; j<numRowsTsizeFactor*numColsTsizeFactor; j++) {
	outChar = outBuffer[j];
	outBuffer[j] = (float)(outChar-minByte)*scaleFactor;
      }
      
      // Now find the min and max
      minByte = INT_MAX;
      maxByte = INT_MIN;
      for (j=0; j<numRowsTsizeFactor*numColsTsizeFactor; j++) {
	outChar = outBuffer[j];
	minByte = MIN(minByte, (int)outChar);
	maxByte = MAX(maxByte, (int)outChar);
      }
      
      fwrite(outBuffer, sizeof(BYTE),
	     numRowsTsizeFactor*numColsTsizeFactor, fOut);
      
      sprintf(outStr,"RAW file stats:\n\
\n				       \
     Frames= [%d]\n		       \
     Width = [%d]\n		       \
     Height= [%d]\n		       \
\n				       \
     8 bit byte pixels\n	       \
     NO header\n		       \
\n				       \
     Min[%d]\n							\
     Max[%d]\n",1,(numColsTsizeFactor),numRowsTsizeFactor,
	      minByte,maxByte);
      askThem(outStr, XmDIALOG_INFORMATION);
      
    }
    else { // Dump entire set
      
      if(projASU) { // Calculated projections
	numColsTsizeFactor = mapCols * sizeFactor;
	numRowsTsizeFactor = mapRows * sizeFactor;
      }
      else { // Just boxed data
	numColsTsizeFactor = purdueHeader.numCols * sizeFactor;
	numRowsTsizeFactor = purdueHeader.numRows * sizeFactor;
      }
      
      CK(outBuffer= (BYTE *)malloc(sizeof(BYTE)*
				   numRowsTsizeFactor*
				   numColsTsizeFactor*
				   purdueHeader.nimg));
      
      for (i=0; i<purdueHeader.nimg; i++) { // Do all boxes
	int outRow,outCol;
	stuffDisplayPackBoxBuffer(i, (int)(zoomFactor+1.0),
				  FALSE);
	if(projASU) { // Calculated projections
	  outRow = (int)(i / maxPhi)*(numColsTsizeFactor*maxPhi)
	    *numRowsTsizeFactor;
	  outCol = (int)(i % maxPhi)*numColsTsizeFactor;
	}
	else { // Just boxed data
	  outRow = 0;
	  outCol = i*numColsTsizeFactor;
	}
	
	for(j=0; j<numRowsTsizeFactor; j++) {
	  for(k=0; k<numColsTsizeFactor; k++) {
	    if(projASU)
	      outBuffer[outRow + outCol +
			(j*numColsTsizeFactor*maxPhi)+k]=
		editBox_display_buffer[(j*BOX_WIDTH)+k];
	    else
	      outBuffer[outRow + outCol +
			(j*numColsTsizeFactor*
			 purdueHeader.nimg)+k]=
		editBox_display_buffer[(j*BOX_WIDTH)+k];
	  }
	}
      }
      
      // Now find the min and max
      minByte = INT_MAX;
      maxByte = INT_MIN;
      for(i=0; i<purdueHeader.nimg*numRowsTsizeFactor*numColsTsizeFactor; i++) {
	outChar = outBuffer[i];
	minByte = MIN(minByte, (int)outChar);
	maxByte = MAX(maxByte, (int)outChar);
      }
      
      // Now find scale factor
      scaleFactor = 255.0 / (float)(maxByte-minByte);
      
      // Now scale the data
      for(i=0; i<purdueHeader.nimg*numRowsTsizeFactor*numColsTsizeFactor; i++) {
	outChar = outBuffer[i];
	outBuffer[i] = (float)(outChar-minByte)*scaleFactor;
      }
      
      // Now find the min and max
      minByte = INT_MAX;
      maxByte = INT_MIN;
      for (i=0; i<purdueHeader.nimg*numRowsTsizeFactor*numColsTsizeFactor; i++) {
	outChar = outBuffer[i];
	minByte = MIN(minByte,(int)outChar);
	maxByte = MAX(maxByte,(int)outChar);
      }
      
      fwrite(outBuffer, sizeof(BYTE),
	     (purdueHeader.nimg*numRowsTsizeFactor*numColsTsizeFactor),
	     fOut);
      
      sprintf(outStr,"RAW file stats:\n\
\n				       \
     Frames= [%d]\n		       \
     Width = [%d]\n		       \
     Height= [%d]\n		       \
OR\n				       \
     1 Frame\n			       \
     Width = [%d]\n		       \
     Height= [%d]\n		       \
\n				       \
     8 bit byte pixels\n	       \
     NO header\n		       \
\n				       \
     Min[%d]\n								\
     Max[%d]\n",purdueHeader.nimg,(numColsTsizeFactor),numRowsTsizeFactor,
	      (projASU)?(maxPhi * numColsTsizeFactor)
	      :(purdueHeader.nimg * numColsTsizeFactor),
	      (projASU)?((purdueHeader.nimg/maxPhi) * 
			 numRowsTsizeFactor)
	      :(numRowsTsizeFactor),
	      minByte,maxByte);
      
      askThem(outStr, XmDIALOG_INFORMATION);
      
    }
    
    free(outBuffer);
    fclose(fOut);
    
  }
  else if(readType==MAP || readType==FLOATMAP || readType==PIFMAP) {
    // Try and write something
    
    if(numWanted==1) { // Just do a screen dump
      
      calcSizeFactor(zoomFactor);
      
      numColsTsizeFactor = display_width;
      numRowsTsizeFactor = display_height;
      CK(outBuffer= (BYTE *)malloc(sizeof(BYTE)*
				   numRowsTsizeFactor*
				   numColsTsizeFactor));
      
      minF = maxF = image_buffer1_disp[0];
      for (i=0; i<numRowsTsizeFactor; i++) {
	for (k=0; k<numColsTsizeFactor; k++) {
	  maxF = MAX(maxF, image_buffer1_disp[(i*buffer_width)+k]);
	  minF = MIN(minF, image_buffer1_disp[(i*buffer_width)+k]);
	}
      }
      
      scaleF = 254.0 / (maxF-minF);
      for (i=0; i<numRowsTsizeFactor; i++)
	for (k=0; k<numColsTsizeFactor; k++)
	  outBuffer[(i*numColsTsizeFactor)+k]=
	    (BYTE)(((float)image_buffer1_disp
		    [(i*buffer_width)+k]-minF)*scaleF);
      
      // Now find the min and max
      minByte = INT_MAX;
      maxByte = INT_MIN;
      for (j=0; j<numRowsTsizeFactor*numColsTsizeFactor; j++) {
	outChar = outBuffer[j];
	minByte = MIN(minByte,(int)outChar);
	maxByte = MAX(maxByte,(int)outChar);
      }
      
      // Now find scale factor
      scaleFactor = 255.0 / (float)(maxByte-minByte);
      
      // Now scale the data
      for (j=0; j<numRowsTsizeFactor*numColsTsizeFactor; j++) {
	outChar = outBuffer[j];
	outBuffer[j] = (float)(outChar-minByte)*scaleFactor;
      }
      
      // Now find the min and max
      minByte = INT_MAX;
      maxByte = INT_MIN;
      for (j=0; j<numRowsTsizeFactor*numColsTsizeFactor; j++) {
	outChar = outBuffer[j];
	minByte = MIN(minByte,(int)outChar); 
	maxByte = MAX(maxByte,(int)outChar);
      }
      
      fwrite(outBuffer, sizeof(BYTE),
	     (numRowsTsizeFactor*numColsTsizeFactor), fOut);
      
      sprintf(outStr,"RAW file stats:\n\
\n					   \
     Frames= [%d]\n			   \
     Width = [%d]\n			   \
     Height= [%d]\n			   \
\n					   \
     8 bit byte pixels\n		   \
     NO header\n			   \
\n								\
     Min[%d]\n							\
     Max[%d]\n",1,(numColsTsizeFactor),numRowsTsizeFactor,
	      minByte,maxByte);
      askThem(outStr, XmDIALOG_INFORMATION);
      fclose(fOut);
      
    }
    else { // DO this
      fclose(fOut);  // fOut must be closed before writeRAWcontour
      remove(fName); // and empty file should be removed
      //      stuffDisplayMapBuffer(curBox,editBoxZoom,0);
      stuffDisplayMapBuffer((imgNum-1),1,0);
      // calcBoxSizeFactor(editBoxZoom);
      writeRAWcontour(fName, editBox_display_buffer, numCols, numRows);
    }
    
  }
  else if((fileType==EMPORTABLE || fileType==BYTEPORTABLE ||
	   fileType==PORTABLE) && imgInCore) {
    
    fwrite(inCore_buffer, sizeof(float), file_width*file_height, fOut);
    
    sprintf(outStr,"RAW file stats:\n\nWidth = [%d]\nHeight= [%d]\n\n",
	    file_width, file_height);
    strcat(outStr, "32 bit float byte pixels\n");
    strcat(outStr, "NO header\n\n");
	   
    askThem(outStr, XmDIALOG_INFORMATION);  
    fclose(fOut);
  }
  
} // end function writeRAW



void virtual_read_activate(int newX, int newY)
{
  // function name: virtual_read_activate
  // For all images regardless of size read a new pages everytime they
  // move through the image with the scrollbars. Pay attention to how
  // we're zoomed as if we're zoomed smaller we can read more info off
  // the disk if we're zoomed larger we need to scroll more slowly
  // etc...  We implemented this way because we want to be able to
  // look at an infinitely large image. If we "page" it we can look at
  // it even if we don't have enough internal memory to hold it all.

  unsigned short *thisFSbuf;
  int i, y, l, m, n, yOut, yOff, logicalRecordLength, image_depth;
  int scaledVirYOffset, scaledVirXOffset, thisFileType;
  int thisInt, sizeFactorInt, yTsizeFactor;
  float *temp_int_image_buffer1;
  static float thisOldZoomFactor = 0.0;
  FILE *fIn;
  
  thisFileType = fileType;
  if(fftSize==1024 && disp1024fft && XtIsManaged(fftScreenW))
    thisFileType = PORTABLE; // Special display of 1024 fft
  
  // This may take awhile so give them the watch cursor
  updateCursor(drawingW, watch_cursor);
  
  if(rotfltSwitch)
    file_width = orig_file_width; // We are roflt'ing
  
  if(thisFileType==PACKEDPORTABLE) { // Just redraw for zoom
    if(fullScreen==99) {
      // Calc how much ZOOM we used to be in the file
      calcSizeFactor(thisOldZoomFactor);
      // Reset from full screen after we've calculated the coords
      // we were in before (in fullscreen mode)
      fullScreen= FALSE;
    }
    
    stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		     packImg, purdueHeader.nimg, imgNum);
    goto noImg;
  }
  else if(thisFileType==MAP || thisFileType==FLOATMAP) { 
    // Just redraw for zoom
    if(fullScreen==99) {
      // Calc how much ZOOM we used to be in the file
      calcSizeFactor(thisOldZoomFactor);
      // Reset from full screen after we've calculated the coords
      // we were in before (in fullscreen mode)
      fullScreen= FALSE;
    }
    
    stuffMapImage(map_image_buffer, float_image_buffer1, 
		  file_width, file_height, nsecs, imgNum, 0);
    goto noImg;
  }
  
  if(useGUI) {
    if((fullScreen==TRUE) && display_width==fullScreen_buffer.dispx &&
       display_height==fullScreen_buffer.dispy) {
      // Just use the one we calculated last time
      thisFSbuf = fullScreen_buffer.fullImage;
      for (i=0; i<image_height*image_width; i++)
	int_image_buffer1[i]=  (unsigned short)thisFSbuf[i];
      goto noImg;
    }
  }

  if(!imgInCore) { // Open disk file
    if ( (fIn = fopen(image_fileName.addr,"r")) == NULL ) {
      if(!EQU(fileName.addr,"")) {
	sprintf(outStr, "Can't read [%s] ",fileName.addr);
	askThem(outStr, XmDIALOG_INFORMATION);}
      goto noRead;
    }
  }
  
  // Calc zoom
  calcSizeFactor(zoomFactor);

  // Init out the image buffer
  for (i=0; i<buffer_height; i++)
    memcpy((&image_buffer1[i*buffer_width]),stufBLKBuf,buffer_width);
  
  for (i=0; i<image_height; i++)
    memcpy((&int_image_buffer1[i*image_width]),stufBLKIntBuf,
	   (image_width*2));

  // I think xoffset and yoffset will always be <=0 after this 
  // because we're clicking in the display_widht/height space.
  
  xoffset = (newX + xoffset)*recipSizeFactor - display_width/2.0;
  yoffset = (newY + yoffset)*recipSizeFactor - display_height/2.0;
  xoffset = MIN((MAXSIZE-display_width), MAX(xoffset,0));
  yoffset = MIN((MAXSIZE-display_height), MAX(yoffset,0));
  
  if(newY > 0 && newY > 0) {
    if(fullScreen==TRUE) { // We are Normal
      virXoffset = virYoffset = 0;
    }
    else { // We are zooming
      int fileSpaceX, fileSpaceY, dispXfileSpace, dispYfileSpace;

      // Calc how much ZOOM we used to be in the file
      calcSizeFactor(thisOldZoomFactor);
      // Reset from full screen after we've calculated the coords
      // we were in before (in fullscreen mode)
      if(fullScreen==99)
	fullScreen= FALSE;
      fileSpaceX = virXoffset + (int)(newX*recipSizeFactor);
      fileSpaceY = virYoffset + (int)(newY*recipSizeFactor);
      
      // Calc zoom for what we are now
      calcSizeFactor(zoomFactor);
      dispXfileSpace = (display_width / 2.0)  * recipSizeFactor;
      dispYfileSpace = (display_height / 2.0) * recipSizeFactor;
      
      // Now then virtual offsets are just what we are in the file
      // minus what we can display
      virXoffset = fileSpaceX - dispXfileSpace;
      virYoffset = fileSpaceY - dispYfileSpace; 
    }
  }
  
  scaledVirYOffset = virYoffset;
  scaledVirXOffset = virXoffset;
  
  // Larger images can have more readIn at lower zooms
  if(sizeFactor < 1.0) { // Try to get more off the disk
    logicalRecordLength = MIN(buffer_width, display_width) * recipSizeFactor;
    logicalRecordLength = MAX(defDispBufSize, 
			      MIN(file_width,logicalRecordLength));

    image_depth = MIN(buffer_height, display_height) * recipSizeFactor;
    image_depth = MIN(file_height, image_depth);
    
    scaledVirYOffset = MAX(0,MIN(scaledVirYOffset,
				 (file_height-image_depth)));
    scaledVirXOffset = MAX(0,MIN(scaledVirXOffset,
				 (file_width-logicalRecordLength)));
    virYoffset = scaledVirYOffset;
    virXoffset = scaledVirXOffset;
  }
  else { // MAX it out at 1 k
    if(buffer_width < display_width)
      logicalRecordLength = buffer_width * recipSizeFactor;
    else 
      logicalRecordLength = display_width * recipSizeFactor;
    logicalRecordLength = MAX(defDispBufSize,
			      MIN(file_width,logicalRecordLength));

    if(buffer_height < display_height) { // Buffer smaller
      image_depth = buffer_height * recipSizeFactor;
      scaledVirYOffset = MAX(0,MIN(scaledVirYOffset,
				   (MAX(file_height,buffer_height) -
				    ((int)(buffer_height*
					   recipSizeFactor)))));
      scaledVirXOffset = MAX(0,MIN(scaledVirXOffset,
				   (MAX(file_width,buffer_width) -
				    ((int)(buffer_width*
					   recipSizeFactor)))));
    }
    else { // normal
      image_depth = display_height * recipSizeFactor;
      scaledVirYOffset = MAX(0,MIN(scaledVirYOffset,
				   (MAX(file_height,display_height) -
				    ((int)(display_height*
					   recipSizeFactor)))));
      scaledVirXOffset = MAX(0,MIN(scaledVirXOffset,
				   (MAX(file_width,display_width) -
				    ((int)(display_width*
					   recipSizeFactor)))));
    }
    
    virYoffset = scaledVirYOffset;
    virXoffset = scaledVirXOffset;
  }
  
  image_depth = MAX(0,MIN((file_height-virYoffset), image_depth)); 
  logicalRecordLength = MAX(0,MIN((file_width-virXoffset),
				  logicalRecordLength));
  
  if(useGUI && (fullScreen || (newY>0 && newX>0))) { // Setup the scroll bars
    int maxVal ,thisPageSize;

    // Don't call us back on this just yet!!!
    noCallbackScroll= TRUE;
    
    // First map coords to image coords
    calcSizeFactor(zoomFactor);
    
    maxVal = file_height - (int)(display_height*recipSizeFactor);
    maxVal = MAX(((int)(display_height*recipSizeFactor)),
		 MIN(file_height,maxVal));

    // I *think* this is where the slider size for the main screen 
    // is defined. Make sure that if you increase sliderSize you also
    // increase XmNmaximum by slidereSize as in:
    //
    //     XmNmaximum = maxVal + sliderSize;
    //
    // Also make sure that if you change these you can still use the scrolls
    // to get to any part of the screen you want.
    // 
    // Also here is where you'd probably want to fix the paging 
    // problem Paul has. e.g. XmNpageIncrement now is calulated to be
    // display_height pages (dsply_height is the number of pixels 
    // you currently see on the scren) but maybe something like:
    // 
    //         XmNpageIncrement *= 0.90;  /+ 90.0%  +/
    // 
    // or such is what he wants???

    maxVal = MAX(maxVal, virYoffset+1);
    thisPageSize = display_height*recipSizeFactor*0.9;
    XtVaSetValues(vScrollW, XmNmaximum, maxVal+thisPageSize, NULL);
    XtVaSetValues(vScrollW, XmNminimum, 0, NULL);
    XtVaSetValues(vScrollW, XmNsliderSize, thisPageSize, NULL);
    XtVaSetValues(vScrollW, XmNvalue, virYoffset, NULL);
    XtVaSetValues(vScrollW, XmNpageIncrement, thisPageSize, NULL);
    
    if(buffer_width < display_width) { // Buffer is samller than display
      maxVal = file_width - (int)(buffer_width*recipSizeFactor);
      maxVal = MAX(((int)(buffer_width*recipSizeFactor)),
		   MIN(file_width,maxVal));
      maxVal = MAX(maxVal, virXoffset+1);
      thisPageSize = buffer_width*recipSizeFactor*0.9;
    }
    else { // Normal
      maxVal = file_width - (int)(display_width*recipSizeFactor);
      maxVal = MAX(((int)(display_width*recipSizeFactor)),
		   MIN(file_width,maxVal));
      maxVal = MAX(maxVal, virXoffset+1);
      thisPageSize = display_width*recipSizeFactor*0.9;
    }
    XtVaSetValues(hScrollW, XmNmaximum, maxVal+thisPageSize, NULL);
    XtVaSetValues(hScrollW, XmNminimum, 0, NULL);
    XtVaSetValues(hScrollW, XmNsliderSize, thisPageSize, NULL);
    XtVaSetValues(hScrollW, XmNvalue, virXoffset, NULL);
    XtVaSetValues(hScrollW, XmNpageIncrement, thisPageSize, NULL);
    XFlush(dsply);
  }

  sprintf(outStr,"%.2f",sizeFactor);
  TextSet(zoomTextW, outStr, TRUE);
  
  // Figure out how we're going to do fractional zooming when making
  // the image larger. Add 1.0 in case we're doing fractional zooms
  if(sizeFactor > (int)sizeFactor)
    sizeFactorInt = (int)sizeFactor + 1;
  
  // We should always calc sizefactorint??? even if we're not doing 
  // fractional zooms ???
  sizeFactorInt = (int)sizeFactor + 1;
  
  // Allocate our temp buffer and read
  CK(temp_int_image_buffer1= 
     (float *)malloc(sizeof(float)*file_width));

  for (y=0; y<image_depth; y++) { // Do all rows
    // Use the +k because of EOR marker * 2.0 because it's 
    // shorts not chars

    if(thisFileType == PORTABLE) { 
      // We're just using the inCore_buffer now 
      // so 1.0 not 2.0 for shorts like we'll need
      // Don't use the EOR markers

      yOff = headOff + (y+scaledVirYOffset)*file_width + scaledVirXOffset;
    }
    else if(thisFileType == BYTEPORTABLE) { // Don't use the EOR markers
      yOff = headOff + (y+scaledVirYOffset)*file_width + scaledVirXOffset;
    }
    else { // DO use the EOR markers
      yOff = headOff + 
	2.0 * ( (y+scaledVirYOffset) * file_width +   // Y direction
	        scaledVirXOffset +                    // X 
	        scaledVirYOffset + y );               // EOR markers
    }

    if(!imgInCore) { // SEEK it
      sysError= fseek(fIn,yOff,0);
      if(sysError) {
	askThem("[VRA] Error reading file.", XmDIALOG_INFORMATION);
	fclose(fIn);
	goto noRead;
      }
    }
    
    yTsizeFactor = y * sizeFactor;
    
    if(thisFileType==BYTEPORTABLE || thisFileType==PORTABLE) { // Read bytes
      if(!imgInCore) {
	// From disk
      }
      else { // Get from core memory
	int z;
	for (z=0; z<logicalRecordLength; z++)
	  temp_int_image_buffer1[z] = inCore_buffer[yOff-headOff+z];
      }
      
      if(sizeFactor > 1.0) {
	// We are zooming out = making larger than normal
	for (m=0; m<sizeFactorInt; m++) { // Do all zoomed lines
	  yOff = image_width * (yTsizeFactor + m) ;
	  for (l=0; l<logicalRecordLength; l++) { // Do all cols
	    thisInt= temp_int_image_buffer1[l];
	    yOut = yOff + (int)(l*sizeFactor);
	    // Let's do one more incase we're fractionally
	    // zooming smaller -> e.g. for zoomFactor = 1.123 e.g.
	    for(n=0; n<sizeFactorInt; n++) { // expand
	      *(int_image_buffer1+yOut+n) = (unsigned short)thisInt;
	    }
	    if(flickerFix2D) { // Special case
	      if(fix2DinCore_diff[((y+scaledVirYOffset) 
				   * file_width) +   // Y direction
				  scaledVirXOffset + // X
				  l]){
		for(n=0; n<sizeFactorInt ;n++) { // expand
		  *(int_image_buffer1+yOut+n)= 30000;
		}
	      }
	    }

	  }
	}
      }
      else { // We are zooming IN = making smaller than normal
	yOff = (image_width * (yTsizeFactor)) ;
	for (l=0; l<logicalRecordLength; l+=(int)recipSizeFactor) {
	  yOut = yOff + (int)(l*sizeFactor);
	  thisInt= 0;
	  for(n=0;n<(int)recipSizeFactor;n++)
	    thisInt += temp_int_image_buffer1[l+n];

	  *(int_image_buffer1+yOut)= (unsigned short)(thisInt/
						      (int)recipSizeFactor);
	  
	  if(flickerFix2D) { // Special case
	    if(fix2DinCore_diff[
				((y+scaledVirYOffset) 
				 * file_width) +   // Y direction
				scaledVirXOffset + // X
				l]) {
	      *(int_image_buffer1+yOut)= 30000;
	    }
	  }
		  
	}
      }
    }
    else { // Read shorts
      askThem("[VRA] I don't understand this type.", XmDIALOG_INFORMATION);
      fclose(fIn);
      goto noRead;
    }
  }

  free(temp_int_image_buffer1);
  
  if(fullScreen && display_width!=fullScreen_buffer.dispx &&
     display_height!=fullScreen_buffer.dispy) {
    // Remember this for next time
    
    // Stuff our display buffer
    stuffImage(int_image_buffer1, image_width, image_height, 
	       contrastStretch, TRUE);
    
    thisFSbuf = fullScreen_buffer.fullImage;
    if(thisFSbuf!=NULL)
      free(thisFSbuf);
    CK(thisFSbuf=(unsigned short *)malloc(sizeof(unsigned short)*
					  image_width*image_height));
    for (i=0; i<image_height*image_width; i++)
      thisFSbuf[i] = (unsigned short)int_image_buffer1[i];
    
    fullScreen_buffer.dispx     = display_width;
    fullScreen_buffer.dispy     = display_height; 
    fullScreen_buffer.fullImage = thisFSbuf;
  }
  
  thisOldZoomFactor = zoomFactor;
  
  if(!imgInCore)
    fclose(fIn);
  
 noImg:;
  // Stuff our display buffer
  if(thisFileType != PACKEDPORTABLE &&
     thisFileType != MAP && thisFileType!=FLOATMAP) { // OK to stuff it
    if(thisFileType==PORTABLE)
      stuffImage(int_image_buffer1, image_width, image_height, 
		 contrastStretch, FALSE);  //  WHY???
    else
      stuffImage(int_image_buffer1, image_width, image_height, 
		 contrastStretch, FALSE);
  }
  
  DrawWindow();
  
  // Redraw the histogram
  histogram_activate();
  
 noRead:;  
  updateCursor(drawingW, old_cursor);
  noCallbackScroll= FALSE;
  
} // end function virtual_read_activate



void drawFindRad()
{
  // function name: drawFindRad
  // Draw feedback info

  int   i, j, isAneg, binStep, thisY;
  float firstRad, firstCorr, *scaled_rdens;
  float minF, maxF, scaleF;

  if(ccoef==0) { // Haven't done this yet
    XFlush(dsply);
    XSync(dsply,0);
    return;
  }
  
  drawingFindRad = TRUE;
  
  XtUnmanageChild(statusScreenW);
  XFlush(dsply);
  
  // XClearWindow(dsply, XtWindow(drawingW));
  XFillRectangle(dsply, XtWindow(drawingW),
		 image_gcGREY[110], 0, 0,
		 display_width, display_height);
  XFlush(dsply);
  
  if(nbins > 0) {
    thickStepFR = 800/nbins;  // default
    thinStepFR  = 200/nbins;
    thinStepFR  = 3;          // Hard code for now
  }
  else {
    thickStepFR = 800/10;    // default
    thinStepFR  = 200/10;
    thinStepFR  = 3;         // Hard code for now
  }
  
  XDrawLine(dsply, XtWindow(drawingW), image_gcWHITE,
	    83, display_height-100,
	    83, display_height-700);
  XDrawLine(dsply, XtWindow(drawingW), image_gcWHITE,
	    83+(0*thickStepFR), display_height-100,
	    83+(nbins*thickStepFR), display_height-100);
  XDrawString(dsply, XtWindow(drawingW), image_gcWHITE,
	      83+(thickStepFR*((nbins/2)-1)), display_height-20, 
	      "Radius", 6);
  XDrawString(dsply, XtWindow(drawingW), image_gcWHITE,
	      10, display_height-720, 
	      "Correlation and Rel Dens", 24);

  // Don't use rdens to compute scale unlest it's the only one
  
  CK(scaled_rdens = (float *)malloc(sizeof(float)*nbins));

  if(rdensit) {
    minF = HUGE;
    maxF = -HUGE;
    for(i=0; i<nbins; i++) {
      if(rdens[i] < minF)
	minF = rdens[i];
      if(rdens[i] > maxF)
	maxF = rdens[i];
    }

    if(maxF == minF)
      scaleF = 1.0;
    else {
      if(minF > 0.0)
	scaleF = (1.0 / maxF);
      else {
	if(maxF > (0.0 - minF))
	  scaleF = (1.0 / maxF);
	else
	  scaleF = (1.0 / (0.0 - minF));
      }
    }
    
    for(i=0; i<nbins; i++)
      scaled_rdens[i] = rdens[i] * scaleF;
  }
  
  // Now let's scale everything
  minF = HUGE;
  maxF = -HUGE;
  for (i=0; i<nbins; i++) { // Search all data
    if(rdensit && scaled_rdens[i] > maxF)
      maxF = scaled_rdens[i];
    if(wrdensit && wrdens[i] > maxF)
      maxF = wrdens[i];
    if(ccoefit && ccoef[i] > maxF)
      maxF = ccoef[i];
    if(wccoefit && wccoef[i] > maxF)
      maxF = wccoef[i];
    if(s_wccoefit && s_wccoef[i] > maxF)
      maxF = s_wccoef[i];
    
    if(rdensit && scaled_rdens[i] < minF)
      minF = scaled_rdens[i];
    if(wrdensit && wrdens[i] < minF)
      minF = wrdens[i];
    if(ccoefit && ccoef[i] < minF)
      minF = ccoef[i];
    if(wccoefit && wccoef[i] < minF)
      minF = wccoef[i];
    if(s_wccoefit && s_wccoef[i] < minF)
      minF = s_wccoef[i];
  }

  minF = (minF < 0.0) ? -1.0 : 0.0;
  maxF = 1.0;
  scaleF = (maxF == minF) ? 1.0 : 600.0/(maxF-minF);

  // Write X axis labels
  firstRad = 0.0;
  binStep = nbins / 10;
  if(binStep <= 0)
    binStep = 1;

  for (i=0; i<nbins; i+=binStep) { // Do all
    firstRad += radstep;
    
    sprintf(outStr,"%d",i);
    XDrawString(dsply, XtWindow(drawingW), image_gcWHITE,
		83+(thickStepFR*i), display_height-65, 
		outStr, strlen(outStr));
    sprintf(outStr,"%.1f",firstRad);
    XFlush(dsply);  
  }
  
  // Draw Zero
  isAneg = FALSE;
  if(minF < 0.0) { // Show them where zero is
    isAneg = TRUE;
    XDrawLine(dsply, XtWindow(drawingW), image_gcWHITE,
	      83+(0*thickStepFR),
	      display_height-((int)((0.0-minF)*scaleF)+100),
	      83+(nbins*thickStepFR),
	      display_height-((int)((0.0-minF)*scaleF)+100));
  }

  // Now draw all curves
  for (i=0; i<nbins; i++) {
    
    // Draw tick mark
    for (j=0; j<(thinStepFR*4); j++)
      XDrawLine(dsply, XtWindow(drawingW), image_gcWHITE,
		83+(i*thickStepFR)+j, display_height-99,
		83+(i*thickStepFR)+j, display_height-97);
    
    for (j=0; j<thinStepFR; j++) { // Draw histogram
      if(isAneg) { // We do have negative values
	if(rdensit) {
	  thisY =  (int)((float)(scaled_rdens[i]-minF)*scaleF);
	  thisY = MIN(MAX(thisY,0),600);
	  XDrawLine(dsply, XtWindow(drawingW), image_gcWHITE,
		    83+(i*thickStepFR)+j,
		    display_height-((int)((0.0-minF)*scaleF)+100),
		    83+(i*thickStepFR)+j,
		    display_height-(thisY+100));
	}
	if(wrdensit) {
	  XDrawLine(dsply, XtWindow(drawingW), image_gcRED,
		    83+(i*thickStepFR)+j+3,
		    display_height-((int)((0.0-minF)*scaleF)+100),
		    83+(i*thickStepFR)+j+3,
		    display_height-
		    ((int)((wrdens[i]-minF)*scaleF)+100));
	}
	if(ccoefit) {
	  XDrawLine(dsply, XtWindow(drawingW), image_gcGREEN,
		    83+(i*thickStepFR)+j+6,
		    display_height-((int)((0.0-minF)*scaleF)+100),
		    83+(i*thickStepFR)+j+6,
		    display_height-
		    ((int)((ccoef[i]-minF)*scaleF)+100));
	}
	if(wccoefit) {
	  XDrawLine(dsply, XtWindow(drawingW), image_gcBLUE,
		    83+(i*thickStepFR)+j+9,
		    display_height-((int)((0.0-minF)*scaleF)+100),
		    83+(i*thickStepFR)+j+9,
		    display_height-
		    ((int)((wccoef[i]-minF)*scaleF)+100));
	}
	if(s_wccoefit) {
	  if(i < (nbins-1))
	    XDrawLine(dsply, XtWindow(drawingW), image_gcBLACK,
		      83+(i*thickStepFR)+j+4,
		      display_height-((int)((float)(s_wccoef[i]-minF)*
					    scaleF)+100),
		      83+((i+1)*thickStepFR)+j+4,
		      display_height-
		      ((int)((s_wccoef[i+1]-minF)*scaleF)+100));
	}
      }
      else { // Positive value
	if(rdensit) {
	  thisY =  (int)((float)(scaled_rdens[i]-minF)*scaleF);
	  thisY = MIN(MAX(thisY,0),600);
	  XDrawLine(dsply, XtWindow(drawingW), image_gcWHITE,
		    83+(i*thickStepFR)+j,
		    display_height-100,
		    83+(i*thickStepFR)+j,
		    display_height-(thisY+100)
		    );
	}
	if(wrdensit) {
	  XDrawLine(dsply, XtWindow(drawingW), image_gcRED,
		    83+(i*thickStepFR)+j+3,
		    display_height-100,
		    83+(i*thickStepFR)+j+3,
		    display_height-((int)((wrdens[i]-minF)*
					  scaleF)+100));
	}
	if(ccoefit) {
	  XDrawLine(dsply, XtWindow(drawingW), image_gcGREEN,
		    83+(i*thickStepFR)+j+6,
		    display_height-100,
		    83+(i*thickStepFR)+j+6,
		    display_height-((int)((ccoef[i]-minF)*
					  scaleF)+100));
	}
	if(wccoefit) {
	  XDrawLine(dsply, XtWindow(drawingW), image_gcBLUE,
		    83+(i*thickStepFR)+j+9,
		    display_height-100,
		    83+(i*thickStepFR)+j+9,
		    display_height-((int)((wccoef[i]-minF)*
					  scaleF)+100));
	}
	if(s_wccoefit) {
	  if(i < (nbins-1))
	    XDrawLine(dsply, XtWindow(drawingW), image_gcBLACK,
		      83+(i*thickStepFR)+j+4,
		      display_height-((int)((s_wccoef[i]-minF)*
					    scaleF)+
				      100),
		      83+((i+1)*thickStepFR)+j+4,
		      display_height-
		      ((int)((s_wccoef[i+1]-minF)*
			     scaleF)+100));
	}
      }
    
      XFlush(dsply);
    }
  }
  
  // Now do Y axis
  firstCorr = minF;
  while(firstCorr < maxF) {
    sprintf(outStr,"%.2f",firstCorr);
    XDrawString(dsply, XtWindow(drawingW), image_gcWHITE, 10,
		display_height - ((int)((firstCorr-minF)*scaleF)+100),
		outStr, strlen(outStr));
    firstCorr += (float)((maxF-minF)/10.0);
    XFlush(dsply);
  }
  
  // Do the last one
  sprintf(outStr,"%.2f",firstCorr);
  XDrawString(dsply, XtWindow(drawingW), image_gcWHITE, 10,
	      display_height - ((int)((firstCorr-minF)*scaleF)+100),
	      outStr, strlen(outStr));
  
  free(scaled_rdens);  
  XFlush(dsply);
  XSync(dsply,0);
  
} // end function drawFindRad



void drawMag()
{
  // function name: drawMag
  // Draw feedback info

  int   i, j, thickStep, thinStep, myTurn, isAneg, maxI=0, minI=0;
  float scaleF;
  
  if(coefs==0) { // Haven't done this yet
    XFlush(dsply);
    XSync(dsply,0);
    return;
  }

  drawingFindRad = FALSE;

  // XClearWindow(dsply, XtWindow(drawingW));
  // XDrawRectangle(dsply, XtWindow(drawingW),
  // image_gcBLACK,0,0,display_width,display_height); 
  
  XFillRectangle(dsply, XtWindow(drawingW),
		 image_gcGREY[110],0,0,
		 display_width, display_height);
  
  XFlush(dsply);
  
  if(magchk > 0) {
    thickStep = 800/magchk;  // default
    thinStep  = 400/magchk;
  }
  else {
    thickStep = 800/10;      // default
    thinStep  = 400/10;
  }
  
  for (i=0; i<magchk; i++) {
    if(coefs[i] > coefs[maxI])
      maxI = i;
    if(coefs[i] < coefs[minI])
      minI = i;
  }
  
  if (coefs[maxI] == coefs[minI])
    scaleF = 1.0;
  else
    scaleF = 600.0/(coefs[maxI]-coefs[minI]);
  
  XtUnmanageChild(statusScreenW);
  XFlush(dsply);
  
  XDrawLine(dsply, XtWindow(drawingW), image_gcYELLOW,
	    10, display_height-100,
	    10, display_height-700);
  XDrawLine(dsply, XtWindow(drawingW), image_gcYELLOW,
	    10+(0*thickStep), display_height-100,
	    10+(magchk*thickStep), display_height-100);
  XDrawString(dsply, XtWindow(drawingW), image_gcYELLOW,
	      10+(thickStep*((magchk/2)-1)), display_height-20, 
	      "Magnification", 13);
  XDrawString(dsply, XtWindow(drawingW), image_gcBLUE,
	      10, display_height-720, 
	      "Correlation", 11);
  isAneg = FALSE;

  if(coefs[minI] < 0.0) { // Show them where zero is
    isAneg = TRUE;
    XDrawLine(dsply, XtWindow(drawingW), image_gcBLUE,
	      10+(0*thickStep),
	      display_height-((int)((0.0-coefs[minI])*scaleF)+100),
	      10+(magchk*thickStep),
	      display_height-((int)((0.0-coefs[minI])*scaleF)+100));
  }
  
  myTurn = FALSE;
  for(i=0; i<magchk; i++) {
    
    for (j=0; j<thinStep; j++) { // Draw hsitogram
      if(isAneg) { // We do have negative values
	GC localThisGC;
	localThisGC = (coefs[i] < 0.0) ? image_gcGREEN : image_gcRED;
	XDrawLine(dsply, XtWindow(drawingW), localThisGC,
		  10+(i*thickStep)+j,
		  display_height-((int)((0.0-coefs[minI])*scaleF)+100),
		  10+(i*thickStep)+j,
		  display_height-
		  ((int)((coefs[i]-coefs[minI])*scaleF)+100));
      }
      else
	XDrawLine(dsply, XtWindow(drawingW), image_gcRED,
		  10+(i*thickStep)+j,
		  display_height-100,
		  10+(i*thickStep)+j,
		  display_height-((int)((coefs[i]-coefs[minI])*scaleF)+100));
    }
  }
  
  myTurn = FALSE;
  for(i=0;i<magchk;i++) {
    sprintf(outStr,"%.4f",mags[i]);
    // Offset the mag print-out so you can read it

    if(myTurn)
      XDrawString(dsply, XtWindow(drawingW), image_gcYELLOW,
		  10+(thickStep*i), display_height-50, 
		  outStr, strlen(outStr));
    else
      XDrawString(dsply, XtWindow(drawingW), image_gcYELLOW,
		  10+(thickStep*i), display_height-65, 
		  outStr, strlen(outStr));

    if(magchk > 10)
      myTurn = !myTurn; /* Only cycle for large numbers */
    sprintf(outStr,"%.4f",coefs[i]);
    if(coefs[i] < 0.0)
      XDrawString(dsply, XtWindow(drawingW), image_gcBLUE,
		  10+(thickStep*i), 
		  display_height-
		  ((int)((coefs[i]-coefs[minI])*scaleF)+90),
		  outStr, strlen(outStr));
    else
      XDrawString(dsply, XtWindow(drawingW), image_gcBLUE,
		  10+(thickStep*i), 
		  display_height-
		  ((int)((coefs[i]-coefs[minI])*scaleF)+110),
      outStr, strlen(outStr));

    XFlush(dsply);
  }
  
  XFlush(dsply);
  XSync(dsply,0);
  
} // end function drawMag



void showBoxingHistogram()
{
  // function name: showBoxingHistogram
  // Draw the histogram on screen showing histogram of different
  // boxing stats.

  int i, j, maxHits;
  float minVal, maxVal, scaleF, stepF, iter;
  bhist_t hist[512];

  for (i=0; i<256; i++)
    XDrawLine(dsply, XtWindow(drawingW), image_gcBLACK, i, 0, i, 256);
  
  // Now the axes
  XDrawLine(dsply, XtWindow(drawingW), image_gcYELLOW, 0, 220, 256, 220);
  XFlush(dsply);

  if(showBoxingLevel == AMMCORRCOEF) { // Do corr coef
    maxVal= minVal = box[0].correlation;
    for(i=0;i<numBoxes;i++) {
      minVal = MIN(box[i].correlation, minVal);
      maxVal = MAX(box[i].correlation, maxVal);
    }
  }
  else if (showBoxingLevel == AMMAVGDENS) { // Do avg dens
    maxVal= minVal = box[0].avgdensity;
    for (i=0; i<numBoxes; i++) {
      minVal = MIN(box[i].avgdensity, minVal);
      maxVal = MAX(box[i].avgdensity, maxVal);
    }
  }
  else if (showBoxingLevel == AMMSTDEV) { // Do stdev
    maxVal= minVal = box[0].stdev;
    for(i=0;i<numBoxes;i++) {
      minVal = MIN(box[i].stdev, minVal);
      maxVal = MAX(box[i].stdev, maxVal);
    }
  }
  stepF  = (maxVal-minVal) / 80.0;

  j = 0;
  for (iter=minVal; iter<maxVal; iter+=stepF) { // setup Hist
    hist[j].rangeVal = iter;
    hist[j].hits = 0;
    j++;
  }

  for (i=0; i<numBoxes; i++) {
    for (j=0; j<79; j++) { // Look at all hist cells
      if(showBoxingLevel== AMMCORRCOEF) { // Do corr coef
	if(box[i].correlation >= hist[j].rangeVal  &&
	   box[i].correlation < hist[j+1].rangeVal) {
	  hist[j].hits++;
	  goto foundIt;
	}
      }
      else if (showBoxingLevel==AMMAVGDENS) { // Do avg dens
	if(box[i].avgdensity >= hist[j].rangeVal  &&
	   box[i].avgdensity < hist[j+1].rangeVal) {
	  hist[j].hits++;
	  goto foundIt;
	}
      }
      else if (showBoxingLevel==AMMSTDEV) { // DO stdev
	if(box[i].stdev >= hist[j].rangeVal  &&
	   box[i].stdev < hist[j+1].rangeVal){
	  hist[j].hits++;
	  goto foundIt;
	}
      }
    }

    // If we got here than it's the last one
    hist[79].hits++;
  foundIt:;
  }

  // Now let's plot the histogram
  maxHits = hist[0].hits;
  for (j=0; j<80; j++)
    if(hist[j].hits > maxHits)
      maxHits = hist[j].hits;
  scaleF = 200.0 / (float)maxHits;
  for (j=0; j<80; j++) { // plot it
    GC thisGC;
    thisGC = image_gcBLUE;
    if(threshLo >= hist[j].rangeVal && threshLo <= hist[j+1].rangeVal)
      thisGC = image_gcRED;
    if(threshHi >= hist[j].rangeVal && threshHi <= hist[j+1].rangeVal)
      thisGC = image_gcGREEN;
    if(j==79 && threshHi >= hist[79].rangeVal)
      thisGC = image_gcGREEN;
    for(i=0;i<3;i++)
      XDrawLine(dsply, XtWindow(drawingW), thisGC,
		(j*3)+5+i,218,
		(j*3)+5+i,219 - (int)((float)hist[j].hits*scaleF)); 
  }

} // end function showBoxingHistogram



void DrawWindow()
{
  // function name: DrawWindow
  // Redraw window whenever it's prudent. Pay attention to different
  // program states and draw the appropriate when needed.

  char numStr[16];
  int i, j, l;
  int arcRadius,thisFileType;
  int drawX, drawY, drawDestX, drawDestY, width, height;
  
  if(useGUI == FALSE) return;
  if(!drawLite) return;

  thisFileType = fileType;
  if(fftSize==1024 && disp1024fft && XtIsManaged(fftScreenW))
    thisFileType = PORTABLE; // Special display of 1024 fft

  if(!subCanDrawWindow) { // DO this for subtracting maps
    if(submapFindRad) 
      drawFindRad();      // Draw the find radius stuff
    else
      drawMag();          // Draw the mag stuff
    return;
  }

  if(globAccum) { // Only do this for boxing
    // Stuff our display buffer  - we may have clusters displayed
    // from before
    stuffImage(int_image_buffer1, image_width, image_height, 
	       contrastStretch,TRUE);
  }
  
  // Show the clusters
  if(clusterDisp && globAccum) {
    float scaleFactor;
    int dispX, dispY, pntX, pntY, paint_factor;
    int autoboxHeight, autoboxWidth;
    
    scaleFactor = sizeFactor / (1.0 / (float)compress_factor);
    autoboxHeight = file_height/compress_factor;
    autoboxWidth  = file_width/compress_factor;
    for (j=1; j<autoboxHeight-1; j++) {
      dispY = (int)(j*scaleFactor) - (int)(virYoffset*sizeFactor);
      for (i=1; i<autoboxWidth-1; i++) {
	dispX = (int)(i*scaleFactor) - (int)(virXoffset*sizeFactor);
	if(dispY > 0 && dispX > 0 && dispY < image_height && 
	   dispX < image_width ) {
	  if(globAccum[j*autoboxWidth+i]) {// Paint it blue
	    paint_factor = scaleFactor * compress_factor;
	    paint_factor = MAX(paint_factor, 1);
	    for (pntY=0; pntY<paint_factor; pntY++)
	      for (pntX=0; pntX<paint_factor; pntX++) {
		XPutPixel(image_image1, (dispX+pntX), (dispY+pntY),
			  pixels_return[BLUE]);
	      }
	  }
	}
      }
    }
  }
  
  // Draw all of image
  if(XtIsManaged(mainWindowW))
    XPutImage(dsply, image_window1, image_gc, image_image1, 
	      xoffset, yoffset, 0, 0, display_width, display_height);
  
  // Get ZOOMiness factor
  calcSizeFactor(zoomFactor);
  
  if(haveParamFile) {
    // We're trying to do the particle processing thing 
  }
  else { // Do normal single file processing
    
    // Display any BOXes we may have
    if(boxMethod == USECIRCLE) { // Draw the circles
      // Account for this being a large image and the coords 
      // being in another "page" with virOffset
      
      for (i=0; i<numBoxes; i++) { // Draw the circles
	drawX = (box[i].x-virXoffset)*sizeFactor - xoffset;
	drawY = (box[i].y-virYoffset)*sizeFactor - yoffset;
	if(boxRadiusDisp) { // Draw box
	  arcRadius = sizeFactor * boxRadius;
	  rwaXDrawCircle(dsply, XtWindow(drawingW), image_gcRED,
			 drawX, drawY, arcRadius);
	}
	if(featherRadiusDisp) { // Draw feather
	  arcRadius = sizeFactor * featherRadius;
	  if(i==findBox || box[i].reference==1)
	    rwaXDrawCircle(dsply, XtWindow(drawingW), image_gcRED,
			   drawX, drawY, arcRadius);
	  else
	    rwaXDrawCircle(dsply, XtWindow(drawingW), image_gcBLUE,
			   drawX, drawY, arcRadius);

	  XDrawRectangle(dsply, XtWindow(drawingW), image_gcBLUE,
			 drawX-arcRadius, drawY-arcRadius,
			 2*arcRadius, 2*arcRadius);
	}
	if(emptyShellDisp) { // Draw empty width
	  arcRadius = sizeFactor * (emmaRadius-emptyShellWidth);
	  rwaXDrawCircle(dsply, XtWindow(drawingW), image_gcWHITE,
			 drawX, drawY, arcRadius);
	}
	if(emmaRadiusDisp) { // Draw emma
	  arcRadius = sizeFactor * emmaRadius;
	  if(i==findBox || box[i].reference==1)
	    rwaXDrawCircle(dsply, XtWindow(drawingW), image_gcRED,
			   drawX, drawY, arcRadius);
	  else
	    rwaXDrawCircle(dsply, XtWindow(drawingW), image_gcGREEN,
			drawX, drawY, arcRadius);
	}

	arcRadius = sizeFactor * boxRadius;
	sprintf(numStr,"%d",(lastNumBox+(i+1)));
	if(dispMethod == CIRCLEDISP) { // Draw circle numbers
	  if(i==findBox || box[i].reference==1)
	    XDrawString(dsply, XtWindow(drawingW), image_gcRED,
			drawX, drawY, numStr, strlen(numStr));
	  else
	    XDrawString(dsply, XtWindow(drawingW), boxing_GC,
			drawX, drawY, numStr, strlen(numStr));
	}
	
      }
    }
    else if((boxMethod==USERECT || boxMethod==USESQUARE) &&
	    rectBox.drawable) { // RECT or SQUARE
      // We use vir?offset so that we take into account that this 
      // guy might be on another virtual page 
      
      width = rectBox.lowx - rectBox.x;
      height = rectBox.lowy - rectBox.y;
      width = width*sizeFactor;
      height = height*sizeFactor;
      drawX = (rectBox.x-virXoffset)*sizeFactor - xoffset;
      drawY = (rectBox.y-virYoffset)*sizeFactor - yoffset;
      XDrawRectangle(dsply, XtWindow(drawingW), image_gcGREEN,
		     drawX, drawY, width, height);
      rectBox.width  = rectBox.lowx - rectBox.x;
      rectBox.height = rectBox.lowy - rectBox.y;
      // Draw rect dimensions
      sprintf(numStr,"[%d]x[%d]", rectBox.width, rectBox.height);
      XDrawString(dsply, XtWindow(drawingW), image_gcRED,
		  drawX+15, drawY+15, numStr, strlen(numStr));
    }
    else if(boxMethod==USEPOLY) { // POLY
      for(i=0;i<(numPoly-1);i++) { // Draw the lines
	// We use vir?offset so that we take into account that this 
	// guy might be on another virtual page 
	
	drawX = (int)((poly[i].x-virXoffset)*sizeFactor) - xoffset;
	drawY = (int)((poly[i].y-virYoffset)*sizeFactor) - yoffset;
	
	drawDestX = (int)((poly[i+1].x-virXoffset)*sizeFactor) - xoffset;
	drawDestY = (int)((poly[i+1].y-virYoffset)*sizeFactor) - yoffset;
	
	XDrawLine(dsply, XtWindow(drawingW), image_gcGREEN,
		  drawX, drawY,
		  drawDestX, drawDestY);
      }
    }
  }
  
  // Check to see if we should be drawing numbers for our packed images
  if(((thisFileType==PACKEDPORTABLE && dispMethod == BOXES) ||
      (thisFileType==MAP && dispMethod == CIRCLEDISP)) && !projASU) {
    for (i=(imgNum-1); i<lastImg; i++) { // Do all visible
      if(thisFileType==PACKEDPORTABLE)
	sprintf(numStr,"%d",imgNums[i].particle_number);
      else 
	sprintf(numStr,"%d",i+1);
      XDrawString(dsply, XtWindow(drawingW), image_gcPLACEHOLDER,
		  (imgNums[i].x+5), (imgNums[i].y+15), 
		  numStr, strlen(numStr));
    }
  }
  
  if(thisFileType==PACKEDPORTABLE) { // Packed file images may be deleted
    for (l=(imgNum-1); l<lastImg; l++) { // Look at all
      if(deletePackedImage[l]) { // Mark it so in the display
	int drawDestX,drawDestY,colOffset,rowOffset;
	colOffset= imgNums[l].x;
	rowOffset= imgNums[l].y;
	drawDestX= colOffset + (2*packRadius + 1) * sizeFactor;
	drawDestY= rowOffset + (2*packRadius + 1) * sizeFactor;
	XDrawLine(dsply, XtWindow(drawingW), image_gcRED,
		  colOffset, rowOffset,
		  drawDestX, drawDestY);
      }
    }
  }
  
  // Now check to see if we've got to draw the radius for correlation
  // averaging screen
  if(XtIsManaged(emcorAvgScreenW)) { // Draw radius
    arcRadius = sizeFactor * coravgRadius;
    rwaXDrawCircle(dsply, XtWindow(drawingW), image_gcRED,
		   lastPlotClickX, lastPlotClickY, arcRadius);
  }

  // Now check to see if we've got to draw the regions for fix2d
  if(XtIsManaged(fix2DScreenW)) { // OK to do
    int  i, j, stepX, stepY, thisX, thisY, afW, afH;
    int fixRegions=1;
    
    if(drawFix2Dgrid) { // draw it
      fixRegions = (int)wstrtod(fix2DRegionsTextW);
      fixRegions = MAX(MIN(fixRegions,MAXREGIONS),1);
      
      stepX = file_width  / fixRegions;
      stepY = file_height / fixRegions;
      
      afW = MIN(file_width*sizeFactor,  display_width);
      afH = MIN(file_height*sizeFactor, display_height);
      
      for(i=0;i<=file_width;i+=stepX) {
	for(j=0;j<=file_height;j+=stepY) {
	  thisX = (i - virXoffset) * sizeFactor;
	  thisY = (j - virYoffset) * sizeFactor;
	  if((thisX >= 0) && (thisY >= 0) &&
	     ((thisX <= display_width) || (thisY <= display_height))) {
	    if(thisY < display_height)
	      XDrawLine(dsply, XtWindow(drawingW), image_gcGREEN,
			0, thisY, afW, thisY);
	    if(thisX < display_width)
	      XDrawLine(dsply, XtWindow(drawingW), image_gcGREEN,
			thisX, 0, thisX, afH);
	  }
	  XFlush (dsply);
	  
	  if(drawFix2Dstats) {
	    int drawX, drawY;
	    drawX = i/stepX;
	    drawY = j/stepY;
	    if(drawX < fixRegions && drawY < fixRegions &&
	       thisX >= 0         && thisY >= 0) { // Only do valid
	      sprintf(numStr,"Mean[%.2f]", fix2Davg[drawX][drawY]);
	      XDrawString(dsply, XtWindow(drawingW), image_gcWHITE,
			  thisX+5, thisY+20, 
			  numStr, strlen(numStr));
	      sprintf(numStr,"Std[%.2f]", fix2Dstd[drawX][drawY]);
	      XDrawString(dsply, XtWindow(drawingW), image_gcWHITE,
			  thisX+5, thisY+40, 
			  numStr, strlen(numStr));
	      XFlush (dsply);
	    }
	  }
	}
      }
    }
  }
  
  if(showBoxingSwitch)
    showBoxingHistogram();
  
  XFlush (dsply);

} // end function DrawWindow



void plotClick(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: plotClick

  // Every time the mouse button is clicked call here. This routine
  // handles the selecting/deselecting of box points and ruler points
  // on the screen.  Do the appropriate based on the current state of
  // the program parameters selected from the interface.

  char ans;
  char infoStr[1024];
  int i, x, y, k, thisImg, sectY, sectX;
  static int thisBox;
  XEvent event;
  GC thisGC;

  XmDrawingAreaCallbackStruct *draw_data = 
    (XmDrawingAreaCallbackStruct *) call_data;

  if(draw_data == 0)
    return;
    
  event = *(draw_data->event);
  
  if(pointColor == 0)
    thisGC = image_gcRED;
  else if(pointColor == 1)
    thisGC = image_gcGREEN;
  else if(pointColor == 2)
    thisGC = image_gcBLUE;
  
  switch (event.type) {
    
  case ButtonRelease:
    x = event.xbutton.x;
    y = event.xbutton.y;
    
    if(boxZoom == USEBOX && isBoxing) { // Draw the appropriate box
      if(boxMethod==USECIRCLE) {
	if(event.xbutton.button == Button1)   // Push it
	  pushBox(x,y,FALSE,&thisBox);
	if(event.xbutton.button == Button2)   // Remove it
	  removeBox(x, y, FALSE);
      }
      else if(boxMethod==USERECT || boxMethod==USESQUARE)
	pushRectBox(event.xbutton.button, x, y);
      else if(boxMethod==USEPOLY) {
	if(event.xbutton.button == Button1)   // Push it
	  pushPolyBox(x,y);
	if(event.xbutton.button == Button2)   // Pop top of stack
	  popPolyBox();
	if(event.xbutton.button == Button3) { // Push last one on the list
	  // Write last leg of polygon
	  poly[numPoly].x = poly[0].x;
	  poly[numPoly].y = poly[0].y;
	  numPoly++;      
	}
      }
      
      isBoxing = FALSE;
      DrawWindow();
    }  
    break;
    
  case ButtonPress:
    // They are trying to find points
    x = event.xbutton.x;
    y = event.xbutton.y;
    
    lastPlotClickX = x;  // Rememebr these for things like emcoravg
    lastPlotClickY = y;  
    if(XtIsManaged(emcorAvgScreenW))
      DrawWindow(); // Show new location
    
    if(haveParamFile && newXY) { // Find a new x,y center for this
      int thisX,thisY;
      float newX,newY;
      P_INFO *thisInfo;
      
      newX = x * recipSizeFactor;
      newY = y * recipSizeFactor;
      newY = purdueHeader.numRows - newY;
      
      if(foundPart != NULL && foundPart->rank == imgNum) { // OK to go
	thisInfo = foundPart;
      }
      else { // Find it
	thisInfo = info.info;
	while(thisInfo != NULL) { // Find this one
	  if(thisInfo->rank==imgNum) // Found it
	    break;
	  thisInfo = thisInfo->next;
	}
      }
      
      if(event.xbutton.button == Button1) { // Change the position
	if(newX > purdueHeader.numCols || newY > purdueHeader.numRows) {
	  sprintf(outStr,
		  "[%d][%d] is outside the valid range [%d][%d]\n",
		  (int)newX,(int)newY,
		  purdueHeader.numCols, purdueHeader.numRows);
	  askThem(outStr, XmDIALOG_INFORMATION);
	}
	else { // OK to continue
	  thisInfo->x = newX;
	  thisInfo->y = newY;
	}
      }
      else if(event.xbutton.button == Button2) { // They want to toggle it
	thisInfo->x = purdueHeader.numCols/2.0 - 0.5;
	thisInfo->y = purdueHeader.numRows/2.0 - 0.5;
      }
      else if(event.xbutton.button == Button3) { // Restore original
	thisInfo->x = thisInfo->origX;
	thisInfo->y = thisInfo->origY;
      }
      
      // Recalc these in case they've changed
      thisInfo->deltaX = thisInfo->x - purdueHeader.numCols/2.0 - 0.5;
      thisInfo->deltaY = thisInfo->y - purdueHeader.numRows/2.0 - 0.5;

      // Display this on the parm screen
      convertXY(&thisX, &thisY, thisInfo->phi, thisInfo->theta);
      findNearestParticle(&thisX,&thisY,&foundPart,FALSE,
			  thisInfo->rank, TRUE,0);
    }
    else { // Normal
      if(boxZoom == USEPOINT) { // Draw the pointing info
	int imageX, imageY;
	float rulerLength;
	
	// Left button is just give me the coords and pixel value
	if(event.xbutton.button == Button1) { // Draw coords and pixel value

	  if (histMethod) { // If histogram is registered
	    int start_x=0, start_y=display_height-50;
	    
	    // Draw a line across the screen showing what 
	    // histogram they are seeing
	    XDrawLine(dsply,XtWindow(drawingW),
		      thisGC, 0, y, buffer_width, y);
	    
	    // Lets draw the histogram now across the bottom 
	    // of the screen
	    for (k=0; k<buffer_width; k++) {
	      XDrawLine(dsply, XtWindow(drawingW), thisGC,
			start_x, start_y, k,
			display_height-50-
			(2*(int)image_buffer1
			 [((y+yoffset)*buffer_width)+(k+xoffset)]));
	      start_x = k; 
	      start_y=display_height-50-
		(2*(int)image_buffer1[((y+yoffset)*buffer_width)+(k+xoffset)]);
	    }
	  }
	  else { // Not histogram
	    globX = x;
	    globY = y;
	    // First map coords to image coords
	    // Also changed sizeFactor to recipSizeFactor below
	    calcSizeFactor(zoomFactor);
	    
	    // Give coords in this virtual page only
	    oldGlobX = x*recipSizeFactor + xoffset*recipSizeFactor;
	    oldGlobY = y*recipSizeFactor + yoffset*recipSizeFactor;

	    // If they're measuring in their depth cued map
	    // they may be zoomed past image_width e.g.
	    findRealValue(x, y, &thisImg, &sectY, &sectX);
	    sprintf(infoStr,"Point 1: (x,y) = (%d,%d) value = %f", 
		    sectX, sectY,
		    findRealValue(x, y, &thisImg, &sectY, &sectX));
	    
	    DrawX(XtWindow(drawingW), thisGC, x, y, 1);
	    drawPointScreen(infoStr,2);
	    
	    infoStr[0]='\0'; // Blank out other strings
	    drawPointScreen(infoStr,3);
	    drawPointScreen(infoStr,4);
	  }
	}
	else if(event.xbutton.button == Button2) { // Draw ruler line
	  
	  if (histMethod) { // If histogram is registered
	    int start_x=0, start_y=display_height-50;
	    
	    // Draw a line down the screen to show where the 
	    // histogram was taken from
	    
	    XDrawLine(dsply, XtWindow(drawingW), image_gcGREEN,
		      x, 0, x, buffer_height);
	    
	    // Lets draw the histogram now across the bottom 
	    // of the screen
	    for (k=0; k<buffer_height; k++) {
	      XDrawLine(dsply, XtWindow(drawingW), image_gcGREEN,
			start_x,start_y,k,
			display_height-50-
			(2*(int)image_buffer1
			 [(k+yoffset)*buffer_width + x + xoffset]));
	      start_x=k; 
	      start_y=display_height-50-
		(2*(int)image_buffer1[(k+yoffset)*buffer_width + x + xoffset]);
	    }
	  }
	  else { // No histogram
	    // Also changed sizeFactor to recipSizeFactor below
	    DrawWindow();
	    calcSizeFactor(zoomFactor);
	    
	    // Get these for this virtual page only
	    imageX = (x+xoffset)*recipSizeFactor;
	    imageY = (y+yoffset)*recipSizeFactor;
	    rulerLength = sqrt((double)
			       ((imageX-oldGlobX)*(imageX-oldGlobX)) + 
			       ((imageY-oldGlobY)*(imageY-oldGlobY)));
	    
	    findRealValue(x, y, &thisImg, &sectY, &sectX);
	    sprintf(infoStr,"Point 2: (x,y) = (%d,%d) value = %f",
		    sectX, sectY,
		    findRealValue(x, y, &thisImg, &sectY, &sectX));
	    drawPointScreen(infoStr,3);
	    
	    XDrawLine(dsply, XtWindow(drawingW), thisGC, x, y, globX, globY);
	    
	    if(pifHeader.pixelSize > 0.0) {
	      sprintf(infoStr, "Length = %.3f pixels (%.3f Ang)", 
		      rulerLength, rulerLength * pifHeader.pixelSize);
	      drawPointScreen(infoStr, 4);
	      sprintf(infoStr, "(Pixel size = %.3f Ang)", pifHeader.pixelSize);
	      drawPointScreen(infoStr, 5);
	    }
	    else {
	      sprintf(infoStr, "Length = %.3f pixels", rulerLength);
	      drawPointScreen(infoStr, 4);
	      sprintf(infoStr, "(Pixel size not set in image file header)");
	      drawPointScreen(infoStr, 5);
	    }

	  }
	}
	else if(event.xbutton.button == Button3) { // reDraw window
	  DrawWindow();
	  infoStr[0]='\0';
	  drawPointScreen(infoStr,2);
	  drawPointScreen(infoStr,3);
	  drawPointScreen(infoStr,4);
	  drawPointScreen(infoStr,5);
	}
      }
      
      else if(boxZoom == USEBOX) { // Draw the appropriate box
	
	if(fileType==PACKEDPORTABLE && !crysDisping) { // (de)selecting boxes

	  int width, height, sectionsPerRow, rowsPerImage, thisSec;
	  
	  if(event.xbutton.button == Button1 || 
	     event.xbutton.button == Button2) {
	    
	    width  = purdueHeader.numCols;
	    height = purdueHeader.numRows;
	    
	    sectionsPerRow = MIN(buffer_width, display_width)/(width*sizeFactor);
	    sectionsPerRow = MAX(1,sectionsPerRow);
	    rowsPerImage = MIN(buffer_height, display_height)/(height*sizeFactor);
	    rowsPerImage = MAX(1,rowsPerImage);
	    
	    thisSec = (int)(x/(width*sizeFactor)) + 
	      sectionsPerRow * (int)(y/(height*sizeFactor));
	    // may be displaying images 42-78 so take that into account here
	    thisSec += (imgNum-1);
	    thisSec = MAX((imgNum-1), MIN(thisSec, purdueHeader.nimg));
	    
	    if(event.xbutton.button == Button2) { // (de)select range of boxes
	      for(i=(lastDeleted+1); i<=thisSec; i++)
		deletePackedImage[i] = !deletePackedImage[i];
	    } else {                              // (de)select single box
	      deletePackedImage[thisSec] = !deletePackedImage[thisSec];
	      lastDeleted= thisSec;  // Remember for button 2
	    }
	    stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
			     packImg,purdueHeader.nimg, imgNum);
	  }
	  else if(event.xbutton.button == Button3){ // (De)selecting all
	    for(i=0; i<purdueHeader.nimg; i++)
	      deletePackedImage[i] = !deletePackedImage[i];
	  }
	  
	  DrawWindow();
	}
	else { // Normal boxing
	  isBoxing = TRUE;
	  lastBoxY= lastBoxX= -999;
	  pushBox(x,y,TRUE, &thisBox);
	}
	
      }
      else if(boxZoom == USEZOOM) { // Zoom this puppy
	if(oldSizeFactor==-99.0) { // Init this guy
	  oldSizeFactor = sizeFactor;  
	  oldZoomFactor = zoomFactor;
	}
	if(event.xbutton.button == Button1 || event.xbutton.button == Button2) {
	  if(fullScreen) { // Calc coords
	    x *= (recipSizeFactor*oldSizeFactor);
	    y *= (recipSizeFactor*oldSizeFactor);
	  }

	  if(event.xbutton.button == Button1) // Zoom in
	    zoomFactor= (int)(zoomFactor+1.0);
	  else                                // Zoom out
	    zoomFactor= (int)(zoomFactor-1.0);
	  
	  zoomFactor = MIN(zoomFactor,  4.0);
	  zoomFactor = MAX(zoomFactor, -4.0);
	  fullScreen= FALSE;
	}
	else if(event.xbutton.button == Button3) { // Place Holder
	  if(fullScreen) { // Was fullscreen
	    zoomFactor = oldZoomFactor;
	    x *= (recipSizeFactor*oldSizeFactor);
	    y *= (recipSizeFactor*oldSizeFactor);
	    fullScreen= FALSE;
	  }
	  else { // Make it fullscreen
	    virXoffset = virYoffset = xoffset = yoffset = 0;
	    x = y = 0;
	    // zoomFactor = 0.0; May not have to set this
	    fullScreen= TRUE;
	  }
	}
	
	// Remember for next time
	oldZoomFactor = zoomFactor;
	oldSizeFactor = sizeFactor;
	
	calcSizeFactor(zoomFactor);   
	sprintf(outStr,"%.2f",  sizeFactor);
	TextSet(zoomTextW, outStr, FALSE); 
	
	virtual_read_activate(x,y);
      }
    }
    break;
    
  case KeyPress:
    ans= XLookupKeysym((XKeyEvent *)&event, 0);
    
    // Move box based on what key they've pressed from the keyboard
    if(boxZoom == USEBOX && rectBox.drawable && !locked &&
       (boxMethod==USERECT || boxMethod==USESQUARE)) {
      
      if(ans=='S') { // Right
	rectBox.lowx++;
	if(boxMethod==USESQUARE)
	  rectBox.lowy++;
      }
      else if(ans=='R') { // Up
	rectBox.lowy++;
	if(boxMethod==USESQUARE)
	  rectBox.lowx++;
      }
      else if(ans=='Q') { // Left
	rectBox.lowx--;
	if(boxMethod==USESQUARE)
	  rectBox.lowy--;
      }
      else if(ans=='T') { // Down
	rectBox.lowy--;
	if(boxMethod==USESQUARE)
	  rectBox.lowx--;
      }
      
      // Trap for end points
      if(rectBox.lowx <= rectBox.x)
	rectBox.lowx = rectBox.x + 1;
      if(rectBox.lowy <= rectBox.y)
	rectBox.lowy = rectBox.y + 1;
      DrawWindow();
      calcSizeFactor(zoomFactor);
      pointerLabelDraw(0,0);
    }
    break;
    
  default:
    break;
  }
  
  XFlush(dsply);  // Make changes current
  
} // end function plotClick



void plotMotion(Widget widget, caddr_t unused, 
		register XMotionEvent *pEvent)
{
  // function name: plotMotion
  // Whenever the pointer moves across our window call here. This is
  // where we redraw the circles if they're moving a box or print the
  // coords of the pointer as they move.  This rtn is called a LOT so
  // be very frugal in what you put in here because it will be
  // executed over and over again as the cursor travels over the
  // image(s).

  char infoStr[1024];
  int x, y;

  // Make all changes current
  XFlush(dsply);               

  if(pEvent->type != MotionNotify)
    return;
      
  // Get coordinates of pointer
  x = pEvent->x;     
  y = pEvent->y;
  
  if(!subCanDrawWindow && drawingFindRad) { // Draw for subtract map
    int thisBin;
    if(thickStepFR > 0) { // OK to do
      XFillRectangle(dsply, XtWindow(drawingW), image_gcGREY[110],
		     125, 0, 400, 80);
      XFlush(dsply);
      thisBin = (x - 83) / thickStepFR;
      thisBin = MIN(MAX(thisBin,0),nbins);
      sprintf(RDoutStr,  "Radial Density          [%f][%d]",
	      rdens[thisBin],thisBin);
      sprintf(WRDoutStr, "Weighted RadialDensity  [%f][%d]",
	      wrdens[thisBin],thisBin);
      sprintf(CCoutStr,  "Correlation Coefficient [%f][%d]",
	      ccoef[thisBin],thisBin);
      sprintf(WCCoutStr, "Weighted CCoeffficient  [%f][%d]",
	      wccoef[thisBin],thisBin);
      sprintf(S_WCCoutStr, "Smoothed Wghted CCoef   [%f][%d]",
	      s_wccoef[thisBin],thisBin);
      if(rdensit)
	XDrawString(dsply, XtWindow(drawingW), image_gcWHITE,
		    125, 10, RDoutStr, strlen(RDoutStr));
      if(wrdensit)
	XDrawString(dsply, XtWindow(drawingW), image_gcRED,
		    125, 25, WRDoutStr, strlen(WRDoutStr));
      if(ccoefit)
	XDrawString(dsply, XtWindow(drawingW), image_gcGREEN,
		    125, 40, CCoutStr, strlen(CCoutStr));
      if(wccoefit)
	XDrawString(dsply, XtWindow(drawingW), image_gcBLUE,
		    125, 55, WCCoutStr, strlen(WCCoutStr));
      if(s_wccoefit)
	XDrawString(dsply, XtWindow(drawingW), image_gcBLACK,
		    125, 70, S_WCCoutStr, strlen(S_WCCoutStr));
    }  
  }
  
  pointerLabelDraw(x, y);
  
  if(isBoxing && boxZoom == USEBOX &&  boxMethod == USECIRCLE) {
    // Draw the appropriate box
    int arcRadius, origX, origY, outOfBounds;
    
    if(pEvent->state & Button2Mask) // Remove it
      removeBox(x, y, FALSE);
    
    // Already done for us in pointerlabeldraw and done the 
    // way we want... 
    // calcSizeFactor(zoomFactor);
    
    // Map this to original image coords
    origX = (x+xoffset)*recipSizeFactor + virXoffset;
    origY = (y+yoffset)*recipSizeFactor + virYoffset;
    
    outOfBounds= FALSE;
    if( (origX-boxRadius) <= recipSizeFactor ||
	(origY-boxRadius) <= recipSizeFactor ||
	(origX+boxRadius) > (file_width-1)   ||
	(origY+boxRadius) > (file_height-1) ) 
      // Can't add to list - it's out of the image
      outOfBounds= TRUE;
    
    arcRadius = sizeFactor * featherRadius;
    if(lastBoxX>0 && lastBoxY>0) { // Redraw the image
      XPutImage(dsply, image_window1, image_gc, image_image1, 
		lastBoxX, lastBoxY,
		lastBoxX, lastBoxY,  
		(2*(arcRadius+3)),  
		(2*(arcRadius+3)));
    }
    
    if(featherRadiusDisp && (!outOfBounds)) { // Draw feather
      arcRadius = sizeFactor * featherRadius;
      rwaXDrawCircle(dsply, XtWindow(drawingW), image_gcPLACEHOLDER,
		     x, y, arcRadius);
      
      XDrawRectangle(dsply, XtWindow(drawingW), image_gcYELLOW,
		     (x-arcRadius), (y-arcRadius),
		     2*arcRadius, 2*arcRadius);
    }
    if(emptyShellDisp && !outOfBounds) {
      arcRadius = sizeFactor * (emmaRadius-emptyShellWidth);
      rwaXDrawCircle(dsply, XtWindow(drawingW), image_gcWHITE,
		     x, y, arcRadius);
    }
    if(emmaRadiusDisp && !outOfBounds) {
      arcRadius = sizeFactor * emmaRadius;
      rwaXDrawCircle(dsply, XtWindow(drawingW), image_gcYELLOW,
		     x, y, arcRadius);
    }
    if(boxRadiusDisp && !outOfBounds) {
      arcRadius = sizeFactor * boxRadius;
      rwaXDrawCircle(dsply, XtWindow(drawingW), image_gcPLACEHOLDER,
		     x, y, arcRadius);
    }
    
    // Do this in feather radius
    arcRadius = sizeFactor * featherRadius;
    lastBoxX = x - arcRadius - 1;
    lastBoxY = y - arcRadius - 1;
    if(buffer_width < display_width)
      lastBoxX = MIN(buffer_width, MAX(lastBoxX,0));
    else 
      lastBoxX = MIN(display_width, MAX(lastBoxX,0));
    if(buffer_height < display_height)
      lastBoxY = MIN(buffer_height, MAX(lastBoxY,0));
    else 
      lastBoxY = MIN(display_height ,MAX(lastBoxY,0));
  }
  else if(isBoxing && boxZoom == USEBOX && rectBox.drawable &&
	  (boxMethod==USERECT || boxMethod==USESQUARE)) {
    // Draw the appropriate RECT 
    int width, height, drawX, drawY;
    
    // Now check for the SQUARE subset of rectangle
    if(boxMethod==USESQUARE) { // square it up
      int xDiff, yDiff;
      if(pEvent->state & Button1Mask) { // x,y
	drawX = (rectBox.lowx-virXoffset)*sizeFactor - xoffset;
	drawY = (rectBox.lowy-virYoffset)*sizeFactor - yoffset;
	xDiff = drawX-x;
	yDiff = drawY-y;
	if( yDiff > xDiff )
	  y = drawY - xDiff;
	else if( xDiff > yDiff )
	  x = drawX - yDiff;
      }
      else if(pEvent->state & Button2Mask) { // x,y
	drawX = (rectBox.x-virXoffset)*sizeFactor - xoffset;
	drawY = (rectBox.y-virYoffset)*sizeFactor - yoffset;
	xDiff = x-drawX;
	yDiff = y-drawY;
	if( yDiff > xDiff )
	  y = drawY + xDiff;
	else if( xDiff > yDiff )
	  x = drawX + yDiff;
      }
      if(buffer_width < display_width)
	x= MIN(buffer_width, MAX(x,0));
      else 
	x= MIN(display_width, MAX(x,0));
      if(buffer_width < display_width)
	y= MIN(buffer_width, MAX(y,0));
      else 
	y= MIN(display_width, MAX(y,0));
    }
    
    // Redraw from last time
    width  = (rectBox.lowx - rectBox.x)*sizeFactor;
    height = (rectBox.lowy - rectBox.y)*sizeFactor;
    
    if(lastBoxX>0 && lastBoxY>0) { // Redraw the image
      XPutImage(dsply, image_window1, image_gc, image_image1, 
		lastBoxX-1, lastBoxY-1,
		lastBoxX-1, lastBoxY-1,  
		width+3, height+3);
      XFlush(dsply);
    }
    
    if(pEvent->state & Button3Mask)
      pushRectBox(Button3,x,y);
    else if(pEvent->state & Button2Mask)
      pushRectBox(Button2,x,y);
    else if(pEvent->state & Button1Mask)
      pushRectBox(Button1,x,y);
    
    // We use viroffset so that we take into account that this 
    // guy might be on another virtual page 
    
    width  = (rectBox.lowx - rectBox.x)*sizeFactor;
    height = (rectBox.lowy - rectBox.y)*sizeFactor;
    drawX  = (rectBox.x-virXoffset)*sizeFactor - xoffset;
    drawY  = (rectBox.y-virYoffset)*sizeFactor - yoffset;
    
    // Draw rectangle
    XDrawRectangle(dsply, XtWindow(drawingW), image_gcGREEN,
		   drawX, drawY, width, height);
    rectBox.width  = rectBox.lowx - rectBox.x;
    rectBox.height = rectBox.lowy - rectBox.y;
    // Draw rect dimensions
    sprintf(infoStr,"[%d]x[%d]", rectBox.width, rectBox.height);
    XDrawString(dsply, XtWindow(drawingW), image_gcRED,
		drawX+15, drawY+15, infoStr, strlen(infoStr));
    XFlush(dsply);
    
    if(pEvent->state & Button1Mask) { // button 1
      lastBoxX = x;
      lastBoxY = y;
    }
    else { // button 2 or 3
      lastBoxX = x - width;
      lastBoxY = y - height;
    }
    if(buffer_width < display_width)
      lastBoxX = MIN(display_width, MAX(lastBoxX,0)); // RSS this looks strange
    else 
      lastBoxX = MIN(display_width, MAX(lastBoxX,0));
    if(buffer_height < display_height)
      lastBoxY = MIN(buffer_height, MAX(lastBoxY,0));
    else 
      lastBoxY = MIN(display_height, MAX(lastBoxY,0));
  } 
  
  XFlush(dsply);   // Make changes current
  
} // end function plotMotion



void plotChange(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: plotChange
  // Called when scrollbars on main screen are moved. Also only for
  // scanned image boxing traversing. (RSS - what does that mean?)

  int thisWidth, thisHeight;
  XmScrollBarCallbackStruct *temp = (XmScrollBarCallbackStruct *) call_data;

  if(noCallbackScroll)
    return;
    
  // Don't do this if we're farbling with it. [xy]offset is just the
  // offset into the image buffer 2k x 2k we're seeing on the screen.
  // vir[XY]offset is the offset into the file image that we want to
  // read from the disk.
  
  fullScreen= FALSE;  // Scrolling not zooming
  
  if(fftMode) { 
    if(widget == hScrollW)
      xoffset = temp->value;
    if(widget == vScrollW)
      yoffset = temp->value;
  }
  else { // Read new one from disk
    if(widget == hScrollW) {
      virXoffset = temp->value;
      xoffset = 0;
    }
    if(widget == vScrollW) {
      virYoffset = temp->value;
      yoffset = 0;
    }
    thisWidth  = MIN(buffer_width, display_width) * 0.5;
    thisHeight = MIN(buffer_height, display_height) * 0.5;
    virtual_read_activate(thisWidth, thisHeight);
  }
  
  DrawWindow();
  
} // end function plotChange



void cancel_fileSel(Widget widget, int *tag, 
		    XmFileSelectionBoxCallbackStruct *callback_data)
{
  // function name: cancel_fileSel
  // They want to abort the selection of a file name.

  XtUnmanageChild(fileSelW);

} // end function cancel_fileSel



void whichFileType(char *fName, int *type)
{
  // function name: whichFileType
  // Determine which file type we've got so we can read it properly.
  // Only used for PIF files as of this writing.

  int i, mode;
  int inInt;
  FILE *fIn;
  fString_t inFile,rWrite;
  forPifHead_t thisHead;

  if(*type==BOXCOORDS || *type==SF || *type==STATE)
    return;
  
  if ( (fIn = fopen(fName,"r")) == NULL ) {
    sprintf(outStr, "Can't read [%s] ", fName);
    askThem(outStr, XmDIALOG_INFORMATION);
    return;
  }
  
  if(fIn) {
    fread(&inInt, sizeof(int),1,fIn);  //GCA: what is the purpose to read one integer?
    fclose(fIn);
    
    i= 4;
    thisHead.luNum = i;
    rWrite.addr = malloc(sizeof(char)*4);
    rWrite.length = 4;
    inFile.addr = malloc(sizeof(char)*264);
    inFile.length = 264;
    sprintf(rWrite.addr,"r");
    sprintf(inFile.addr,"%s",fName);
    
    pif_open(inFile.addr, &i, rWrite.addr);
    
    if(!pif_read_gh(&thisHead)) {
      askThem("I don't understand this file type.", XmDIALOG_INFORMATION);
      *type = -99;
      return;
    }
    
    mode = thisHead.mode;
    
    // RSS need to decide proper way to handle mode==1
    if(mode==20 || mode==21 || mode==22 || mode==40)
      *type = PIFMAP;
    else if (mode==0)
      *type = EMPORTABLE;
    else if (mode==1 || mode==2) // short or float Zeiss data
      *type = EMPORTABLEI2;  // short Zeiss data
    else if (mode==1 || mode==6 || mode==7)
      *type = EMPORTABLE;
    else if (mode==46)
      *type = EMPORTABLEI4;
    else
      askThem("I don't understand this mode", XmDIALOG_INFORMATION);
    
    free(rWrite.addr);
    free(inFile.addr);
  }
  
  pif_close(&thisHead.luNum);
  
} // end function whichFileType



void askThemForMode(int *mode)
{
  // function name: askThemForMode
  // Ask them how to interpret this MRC file. Future goal - try and
  // get this info from the MRC header.

  int numChoices, numWanted;
  
  XmListAddItem(infoListW, XmStringCreateLocalized("2D scanned image"), 1);
  XmListAddItem(infoListW, XmStringCreateLocalized("Set of 2D boxes"), 2);
  XmListAddItem(infoListW, XmStringCreateLocalized("3D Map"), 3);

  XFlush(dsply);
  
  sprintf(outStr,"How do you want me to interpret this file?: ");
  numChoices = 3;
  numWanted= askThemAnswer(outStr,numChoices);
  
  if(numWanted==0)
    *mode = 2;
  else if(numWanted==1)
    *mode = 46;
  else if(numWanted==2)
    *mode = 20;
  
} // end function askThemForMode



void gotFilename(Widget w, char *client_data, 
		  XmFileSelectionBoxCallbackStruct *callback_data)
{
  // function name: gotFilename

  // Get the filename they specified in file selection widget. Branch
  // according to what operation they want performed on this filename.
  // This rtn is called whenever a filename is selected from the
  // filename browsing widget.

  char oldFilename[1024], thisFileName[1024];
  char *s1, *s3;
  int i, j; 
  int interp, nterp;
  int *irnum;
  float xcen, ycen, zcen, max_r;
  float *sum_rint, *sumsq_rint, *mrcBuf;

  // Pop down the file selection widget
  if(useGUI) {
    XtUnmanageChild(fileSelW);
    XFlush(dsply);
    // This may take awhile so give them the watch cursor
    updateCursor(drawingW, watch_cursor);
  }
  if( readWrite == READPARAM || subtractingMap || readEmcoravg ||
      (refReadFile && readType==REFREADFILE) || readImgconv) {

    char *str;
    if(!XmStringGetLtoR(callback_data->value, XmSTRING_DEFAULT_CHARSET, &str))
      printf("Error extracting filename.\n");
    
    if(readWrite == READPARAM) {
      sprintf(gottenParamFname, "%s", str);
      paramFileRead_work();
    }
    else if(subtractingMap) {
      sprintf(subtractFilename, "%s", str);
      TextSet(subFile2TW, subtractFilename, FALSE);
      XtSetArg(arglst[0], XmNcursorPosition, strlen(str) );
      XtSetValues(subFile2TW, arglst, 1); 
      XFlush(dsply);
      subtractMap_read();
    }
    else if(readEmcoravg) {
      sprintf(emcoravgFilename, "%s", str);
      read_emcoravg(str);
    }
    else if(refReadFile && readType==REFREADFILE) {
      sprintf(refReadFileName, "%s", str);    
      sprintf(outStr, "%s", refReadFileName);
      doLabel(outStr, refReadLW, TRUE);
    }
    else if(readImgconv) {
      sprintf(imgconvFilename, "%s", str);
      read_imgconv(str);
    }

    XtFree(str);
    goto destructor;
  }
  
  // Get the filename
  sprintf(oldFilename,"%s",fileName.addr);

  if(preloadSwitch) { // Preload it
    sprintf(fileName.addr,"%s",preLoadStr); 
    preloadSwitch = FALSE;  // Done with it
  }
  else { // normal
    if(!XmStringGetLtoR(callback_data->value, 
			XmSTRING_DEFAULT_CHARSET, &fileName.addr))
      printf("Error extracting filename.\n");
  }

  if(useGUI) {
    raiseIt(statusScreenW);
    XSync(dsply,0);
  }
  
  if(readWrite==READING) { // Read from Input file
    sprintf(outStr,"Reading [%s]",fileName.addr);
    if(checkStatus(outStr))
      goto destructor;
    
    if(readType==MRC) { // read mrc file
      char fname[256], two_d[2];
      int m, n, o, thisNumSecs;
      
      sprintf(two_d,"Y");
      sprintf(fname,"%s",fileName.addr);
      
      read_mrc_header__(fname, &pifHeader, two_d);
      
      recipFloatScaleFactor = pifHeader.recipFloatScaleFactor;
      purdueHeader.img_bkg  = pifHeader.img_bkg;
      purdueHeader.img_min  = pifHeader.datai4_min;
      purdueHeader.img_max  = pifHeader.datai4_max;
      purdueHeader.map_min  = pifHeader.datar4_min;
      purdueHeader.map_max  = pifHeader.datar4_max;
      purdueHeader.img_avg  = pifHeader.datar4_avg;
      purdueHeader.img_std  = pifHeader.datar4_std;
      purdueHeader.numCols    = pifHeader.numCols;
      purdueHeader.numRows    = pifHeader.numRows;
      purdueHeader.numSecs    = pifHeader.numSecs;
      purdueHeader.nimg       = pifHeader.nimg;
      numCols    = pifHeader.numCols;
      numRows    = pifHeader.numRows;
      origNumSecs = numSecs    = pifHeader.numSecs;
      pifMode    = pifHeader.mode; 
      htype      = pifHeader.htype; 
      sprintf(title,"%s",pifHeader.descriptionStr);
      aoverb     = pifHeader.aoverb; 
      map_abang  = pifHeader.map_abang;
      a_cell     = pifHeader.a_cell;
      b_cell     = pifHeader.b_cell;
      c_cell     = pifHeader.c_cell; 
      dela       = pifHeader.dela; 
      delb       = pifHeader.delb; 
      delc       = pifHeader.delc;
      binFactor  = pifHeader.binFactor;
      dthe       = pifHeader.dthe; 
      dphi_90    = pifHeader.dphi_90; 
      symmetry   = pifHeader.symmetry; 
      
      fileBinSecs = numSecs;
      fileBinRows = numRows;
      fileBinCols = numCols;
      fileBinFactor = binFactor;
      
      askThemForMode(&pifHeader.mode);
      pifMode    = pifHeader.mode; 
      if(pifMode==20)
	numSecs = pifHeader.nimg;
      
      if(pifHeader.mode==46)
	thisNumSecs = pifHeader.nimg;
      else if(pifHeader.mode==2)
	thisNumSecs = 1;
      else { // mode == 20
	thisNumSecs = numSecs;
	pifHeader.nimg = numSecs;
	pifHeader.numSecs = numSecs;
	origNumSecs = numSecs;
	nsecs = numSecs;
      }
      // +2 in secs for even numbered sections in mrc
      CK(mrcBuf = (float *)malloc(sizeof(float)
				  *numCols*numRows*(thisNumSecs+2)));
      read_mrc_file__(fname, mrcBuf, two_d, 
		      &numCols,&numRows,&thisNumSecs);
      
      if(pifHeader.mode==46) { // DO for packed boxes
	float *thisF, *thisP;
	readType = MRCPACKED;
	fileType = PACKEDPORTABLE;
	minPackPixel = maxPackPixel = mrcBuf[123];
	for (m=0; m<pifHeader.nimg; m++) { // DO all
	  CK(thisP= packImg[m]= (float *)
	     malloc(sizeof(float)*numCols*numRows));
	  thisF = (float *)(mrcBuf + m*numCols*numRows);
	  for (n=0; n<numRows; n++) {
	    for (o=0; o<numCols; o++) {
	      thisP[(n*numCols)+o] = thisF[(numRows-n-1)*numCols + o];
	      minPackPixel = MIN(minPackPixel, thisP[n*numCols + o]);
	      maxPackPixel = MAX(maxPackPixel, thisP[n*numCols + o]);
	    }
	  }
	}
	// Now clip this guy to our colormap
	if(minPackPixel < 0.0)
	  scalePackFactor = (tryColors - 7.0) / (maxPackPixel-minPackPixel);
	else 
	  scalePackFactor = (tryColors - 7.0) / maxPackPixel;
	packRadius = purdueHeader.numCols/2;
	stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
			 packImg,purdueHeader.nimg, 1);
	
	free(mrcBuf);
      }
      else if(pifHeader.mode==20) { // DO for maps
	readType = MRCMAP;
	fileType = MAP;
	if(float_image_buffer1 && float_image_buffer1 != lowRes_buffer1)
	  free(float_image_buffer1);
	freeNullFloat(&lowRes_buffer1);
	float_image_buffer1 = mrcBuf;
	
	farray_minmax(float_image_buffer1, &minMapPixel, &maxMapPixel, 
		      numCols*numRows*numSecs);
	
	if(pifMode==20) { // We've got a map here 
	  imgNum = numCols/2; // Let's default to the middle section
	  purdueHeader.nimg = numSecs;  // Fool it to take our value
	  TextSetInt(imgNumTextW, imgNum+1, FALSE);
	  
	  calcSF_work(calcSFs);
	  
	  // Now remember these forever
	  lowRes_buffer1 = float_image_buffer1;
	  lowResCols = numCols;
	  lowResRows = numRows;
	  lowResSecs = numSecs;
	  
	  origWidthFromFile = numCols;  
	}

	xCen = numCols/2;
	yCen = numRows/2;
	zCen = origNumSecs/2;
	
	largestDim = numCols;
	largestDim = MAX(largestDim,numRows);
	largestDim = MAX(largestDim,origNumSecs);
	
	// We've got to make this larger for asymmetric maps 
	// and all maps that have information "outside" of the
	// normal sphere of our normal maps,
	// or else we just keep cliping the box into a sphere
	
	globMaxRadius = (int)(origNumSecs/2) - 1;
	globMinRadius = 0.0;
	
	// Set scrolls on depthcue screen  and give feedback
	sprintf(outStr,"%.0f",globMinRadius);
	doValue(outStr, minRadiusValueW, FALSE);
	
	sprintf(outStr,"%.0f",globMaxRadius);
	doValue(outStr, maxRadiusValueW, FALSE);
	
	// Radius is origNumSecs/2
	XtVaSetValues(maxRadiusScrollW, XmNsliderSize, 5, NULL);
	XtVaSetValues(maxRadiusScrollW, XmNmaximum, 
		      (int)((globMaxRadius)+5),NULL);
	XtVaSetValues(maxRadiusScrollW, XmNvalue, (int)(globMaxRadius),NULL);
	
	XtVaSetValues(minRadiusScrollW, XmNsliderSize, 5, NULL);
	XtVaSetValues(minRadiusScrollW, XmNmaximum, 
		      (int)((globMaxRadius)+5),NULL);
	XtVaSetValues(minRadiusScrollW, XmNvalue, (int)(globMinRadius),NULL);
	
	printf("MAX[%f]  MIN[%f]\n",maxMapPixel,minMapPixel);
	secMaxPixel = maxMapPixel;
	secMinPixel = minMapPixel;
	
	minSurf = MAX(minMapPixel,0.0);
	maxSurf = MAX(maxMapPixel,(0.0-minMapPixel));

	// We're doing the surfaceDensity in tenths now
	XtVaSetValues(surfDensScrollW, XmNmaximum, maxSurf*10 + 50, NULL);
	XtVaSetValues(surfDensScrollW, XmNminimum, minSurf*10, NULL);

	if (pifMode == 46) {
	  XtVaSetValues(surfDensScrollW, XmNsliderSize, 500, NULL);
	  XtVaSetValues(surfDensScrollW, XmNpageIncrement, 500, NULL);
	  XtVaSetValues(surfDensScrollW, XmNincrement, 100, NULL);
	  XtVaSetValues(surfDensScrollW, XmNvalue, 
			(int)(purdueHeader.img_bkg),NULL);
	}
	else if(maxSurf > 5000.0) {
	  XtVaSetValues(surfDensScrollW, XmNsliderSize, 
			(int)(maxSurf*0.01), NULL);
	  XtVaSetValues(surfDensScrollW, XmNpageIncrement, 
			(int)(maxSurf*0.01), NULL);
	  XtVaSetValues(surfDensScrollW, XmNincrement, 
			(int)(maxSurf*0.002), NULL);
	  XtVaSetValues(surfDensScrollW, XmNvalue, 10, NULL);
	}
	else if(maxSurf > 500.0) {
	  XtVaSetValues(surfDensScrollW, XmNsliderSize, 500, NULL);
	  XtVaSetValues(surfDensScrollW, XmNpageIncrement, 500, NULL);
	  XtVaSetValues(surfDensScrollW, XmNincrement, 100, NULL);
	  XtVaSetValues(surfDensScrollW, XmNvalue, 10, NULL);
	}
	else {
	  XtVaSetValues(surfDensScrollW, XmNsliderSize, 50, NULL);
	  XtVaSetValues(surfDensScrollW, XmNpageIncrement, 50, NULL);
	  XtVaSetValues(surfDensScrollW, XmNincrement, 10, NULL);
	  XtVaSetValues(surfDensScrollW, XmNvalue, 10, NULL);
	}
	surfaceDensity = 1.0;  // 1 tenth of 10
	      
	// Give feedback
	doValue("-1", surfDensValueW, FALSE);
	doLabel("SecNum", imgNumLabelW, FALSE);
	
	// To make this work for now let's stuff the nimg field with
	// the number of sections. Really we should change any and all
	// references to nimg to be origNumSecs. We really only work
	// on one image at a time so this may never have any affect.

	purdueHeader.nimg = origNumSecs;
	
	calcMapStats(float_image_buffer1,
		     &purdueHeader.img_bkg, 
		     &purdueHeader.map_min, &purdueHeader.map_max, 
		     &purdueHeader.img_avg, &purdueHeader.img_std,
		     &numCols, &numRows, &origNumSecs);
	
	// Write stats
	sprintf(outStr,"\nStatistics for file [%s]\n\
\n\
Title [%s]\n\
\n\
File size   [%d]x[%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
\n",
		fileName.addr, title,
		numCols, numRows, origNumSecs,
		purdueHeader.img_bkg, 
		minMapPixel, maxMapPixel,
		purdueHeader.img_avg, purdueHeader.img_std);
	TextSet(statTextW, outStr, TRUE);
      }
      else if(pifHeader.mode == 2) { // Do for scanned img
	file_width = numCols;
	file_height = numRows;
	inCore_buffer= (float *)mrcBuf;
	if(inCore_buffer)
	  imgInCore= TRUE;
	else {
	  imgInCore= FALSE;
	  printf("Trouble allocating inCore_buffer!!!\n");
	  exit(0);
	}
	if(int_image_buffer1 != NULL)
	  free(int_image_buffer1);
	int_image_buffer1 = (unsigned short *)
	  malloc(defDispBufSize*defDispBufSize*sizeof(unsigned short) + 1);
	readType = MRCSCAN;
	fileType = PORTABLE;
	virtual_read_activate(-99,-99);
      }
    } // end if readType==MRC
    else if(readType != PARAMFILE && readType != PARAMMAP && 
	    readType != BATCHJOB  && readType != PARAMSCAN)
      whichFileType(fileName.addr, &readType);

    if(readType==EMPORTABLE   || readType==EMPORTABLEI2 ||
       readType==EMPORTABLEI4 || readType==MRCPACKED) { // Read an image file
      // Remember this for the next time
      if(numBoxes > 0)
	lastNumBox = (lastNumBox + numBoxes);
      
      // Init for this new image
      if(readType != MRCPACKED)
	if(useGUI) groundZero();
      
      TextSetInt(boxInitW, lastNumBox+1, TRUE);
      
      // Remember what our filename was - especially if we had to 
      // change it to portable or etc...

      sprintf(image_fileName.addr,"%s",fileName.addr);
      
      // Let's see if we can't just use our own
      if(readType != MRCPACKED) { // We already read it
	mainRobem_sub_UNIX(&int_image_buffer1, &fileName, &i, &j);
	
	// Ooops no image get out of here
	if(i==0 || j==0)
	  goto destructor;
      }
      
      // Give feedback on main window title bar
      sprintf(outStr,"Robem: [%s] - %d images total.",fileName.addr,
	      purdueHeader.nimg);
      
      doDialogTitle(outStr, mainWindowW, FALSE);
      
      // Remember how big we are on disk
      file_width  = i;
      file_height = j;
      
      // Now clip this and only get 1k x 1k maximum at a time
      i = j = defDispBufSize;
      
      // Point to where the display data is
      if(useGUI) image_image1->data = (char *)image_buffer1;
      
    }
    else if(readType==MAP || readType==FLOATMAP || readType==PIFMAP
	    || readType==MRCMAP) { // Read a MAP file
      
      fileType= MAP;
      // Init for this new image
      if(readType != MRCMAP)
	if(useGUI) groundZero();
      
      doLabel("SecNum", imgNumLabelW, FALSE);
      TextSet(imgNumTextW, "1", FALSE);

      if(useGUI) {	
	XtSetArg(arglst[0], XmNsensitive, FALSE);
	XtSetValues(removeBlemishButtonW, arglst, 1); 
	XtSetValues(editBoxButtonW, arglst, 1); 
	XtSetValues(pcxButtonW, arglst, 1); 
	XtSetValues(correlateButtonW, arglst, 1);  
	XtSetValues(autoBoxW, arglst, 1);  
	XtSetValues(correlateButton2W, arglst, 1);  
	XtSetArg(arglst[0], XmNsensitive, TRUE);
	XtSetValues(paintMapButtonW, arglst, 1); 
	XtSetValues(mapMovieButtonW, arglst, 1); 
	XtSetValues(mapShadeButtonW, arglst, 1); 
	XtSetValues(truncaddButtonW, arglst, 1); 
	XtSetValues(rescaleButtonW, arglst, 1); 
	XtSetValues(modelW, arglst, 1); 
	XtSetValues(subtractMapButtonW, arglst, 1); 
	XtSetValues(calcSFButtonW, arglst, 1); 
	XtSetValues(hardRotateButtonW, arglst, 1); 
	XtSetValues(invertButtonW, arglst, 1); 
	XtSetValues(rotateButtonW, arglst, 1); 
	XtSetValues(rawButtonW, arglst, 1);   
      }
      
      // Read the MAP file
      if(readType==PIFMAP || readType==EMPORTABLEI4) { // PIF MAP file
	read_pif(&i, &j, &nsecs, &fileName);
	if(nsecs==-99) {
	  sprintf(outStr,"Error reading [%s] as MAP file.",
		  fileName.addr);
	  askThem(outStr, XmDIALOG_INFORMATION);
	  goto destructor;
	}
	fileType= MAP;
	if(readType==EMPORTABLEI4)
	  fileType = PACKEDPORTABLE;
      }
      else if(readType==MRCMAP)  {// MRC MAP
	fileType= MAP;
	nsecs = pifHeader.numSecs;
	i = numCols;
	j = numRows;
      }
      else { // Old TSB format MAP
	sprintf(outStr,"[%s] is not a PIF file.", fileName.addr);
	askThem(outStr, XmDIALOG_INFORMATION);
	goto destructor;
      }
      
      origNumSecs= numSecs= nsecs;  // stuff our global
      
      // Remember what our filename was - especially if we had to 
      // change it to portable or etc...
      
      sprintf(image_fileName.addr,"%s",fileName.addr);
      
      // Ooops no image get out of here
      if(i==0 || j==0)
	goto destructor;
      
      // Give feedback on main window title bar
      sprintf(outStr,"Robem: [%s] - [%d,%d,%d]",fileName.addr,
	      i, j, nsecs);

      doDialogTitle(outStr, mainWindowW, FALSE);
      
      // Let them toggle between sections or cued maps
      if(useGUI) XtVaSetValues(sectionW, XmNsensitive, TRUE, NULL); 
      if(useGUI) XFlush(dsply);
      
      // Remember how big we are on disk
      numCols = file_width  = i;
      numRows = file_height = j;
      
      // Point to where the display data is
      if(useGUI) image_image1->data = (char *)image_buffer1;
      
      if(fileType==FLOATMAP) { // DO crsytalDisplay
	int *passTag;
	passTag = (int *)malloc(sizeof(int));
	*passTag = 0;
	if(useGUI) crysDisp_screen_activate(0, (XtPointer)passTag, 0);
	free(passTag);
      }
      else { // regular MAP disp
	if(!calcSFs) {
	  // We stuff this in the calcSF_work() rtn
	  // Now stuff the display buffer from what was read off the
	  // disk - strt at zero'th(1) section
	  // printf("Buffers address: map_image[%p], float_image[%p]\n",map_image_buffer, float_image_buffer1); 
	  printf("Image size: [%d] [%d] [%d] / [%d]\n",
		 file_width, file_height, nsecs, imgNum);
	  
	  if(useGUI) stuffMapImage(map_image_buffer, float_image_buffer1, 
			file_width, file_height, nsecs, imgNum, 0);
	  
	  // Display the image in the window created above
	  if(useGUI) {
	    XPutImage (dsply, image_window1, image_gc, image_image1, 
		       0, 0, 0, 0, image_width, image_height);
	    XFlush (dsply);
	  }
	}     
      }

      if(maxMapPixel == minMapPixel)  // Can't do it
	goto destructor;
      
      // ----- Compute circular avearge -----

      if(pifMode==20 && currentCircAvg < MAXBOX-2) { // OK - we've got room
	s1= fileName.addr;
	s3= strrchr(s1,'/');
	
	circAvg[currentCircAvg].mapName = (char *)malloc(sizeof(char)*512);
	
	if(s3)
	  sprintf(circAvg[currentCircAvg].mapName,"%s",(s3+1));
	else 
	  sprintf(circAvg[currentCircAvg].mapName,"%s",fileName.addr);
	s1= circAvg[currentCircAvg].mapName;
	
	if(useGUI) {
	  XmListAddItem(circAvgListW, XmStringCreateLocalized(s1), 
			(currentCircAvg+1));  
	  XFlush(dsply);
	}
	
	xcen = numCols/2.0;
	ycen = numRows/2.0;
	zcen = numSecs/2.0;
	interp= 2;  // Original way 
	interp= 1;  // Check circAvgMotion() to make sure they agree

	// Let's go out to the corners
	nterp = numCols;
	max_r = numCols;
	circAvg[currentCircAvg].rint= (float*)malloc(sizeof(float)*nterp);
	circAvg[currentCircAvg].stDev= (float*)malloc(sizeof(float)*nterp);
	circAvg[currentCircAvg].nterp = nterp;
	
	irnum      = (int *)malloc(sizeof(int)*numCols*interp);
	sum_rint   = (float *)malloc(sizeof(float)*numCols*interp);
	sumsq_rint = (float *)malloc(sizeof(float)*numCols*interp);
	map_1d_rob(float_image_buffer1, numCols, numRows, numSecs,
		   xcen, ycen, zcen, interp, circAvg[currentCircAvg].nterp,
		   max_r, circAvg[currentCircAvg].rint, irnum,sum_rint,
		   sumsq_rint, circAvg[currentCircAvg].stDev);
	free(irnum);
	free(sum_rint);
	free(sumsq_rint);
	currentCircAvg++;
      }

    }
    else if(readType==BOXCOORDS) { // Read a box Coords
      if(fileType==PACKEDPORTABLE)
	askThem("Boxing of boxed (mode==6) PIF files not allowed.",
		XmDIALOG_INFORMATION);
      else
	readBoxCoords(fileName.addr);
    }
    else if(readType==STATE)
      readState(fileName.addr);
    else if(readType==SF)
      readSF(fileName.addr);
    else if(readType==PARAMFILE)
      read_param(fileName.addr, TRUE);
    else if(readType==PARAMMAP)
      read_param_map(fileName.addr);
    else if(readType==PARAMSCAN)
      read_param_scan(fileName.addr);
    else if(readType==BATCHJOB)
      read_batchjob(fileName.addr);
    
    // Let's setup what YELLOW means
    if(useGUI) {
      if(fileType==EMPORTABLE || fileType==PORTABLE || fileType==BYTEPORTABLE)
	color_defs[YELLOW].red = 255 * PIXELCOLOR;
      else 
	color_defs[YELLOW].red = 0;
      color_defs[YELLOW].pixel = YELLOW;
      color_defs[YELLOW].green = 255 * PIXELCOLOR;
      color_defs[YELLOW].blue  = 0;
      color_defs[YELLOW].flags = DoRed | DoGreen | DoBlue;
      
      XAllocColor (dsply, color_map, &color_defs[YELLOW]);
      pixels_return[YELLOW] = color_defs[YELLOW].pixel;
      
      DrawWindow(); // Update display
    }
    
  }
  else if(readWrite==WRITING) { // Write to output file
    
    sprintf(outStr,"Writing [%s]",fileName.addr);
    if(checkStatus(outStr))
      goto destructor;
    
    if(writeType==CIRCLEPACKED || writeType==APPENDCIRCLEPACKED) {
      
      if(fileType==PACKEDPORTABLE) { // We read a packed portable file!!!
	// Get all appodizing params before writing to write_pif()
	appodizing = PACKEDPORTABLE;
	// This is border width now
	TextSetInt(appodBoxRadW, featherRadius-boxRadius, FALSE);
	TextSetInt(appodFeaRadW, featherRadius, FALSE);
	raiseIt(appodizeScreenW);
      }
      else if(fileType==MAP) {
	sprintf(outStr, "I don't know how to box this type of PIF file [%d].",
		pifMode);
	askThem(outStr, XmDIALOG_INFORMATION);
      }
      else { // We boxed a scanned image
	// Write the Box coords by default - what if we're appending
	// here???  We need to write a bcrd file FOR THE ORIGINAL
	// image (stored in image_fileName.addr) also
	
	sprintf(thisFileName,"%s",image_fileName.addr);
	s1= strrchr(thisFileName,'.');
	if(s1)
	  sprintf(s1,".bcrd");
	else 
	  sprintf(thisFileName,"%s.bcrd",image_fileName.addr);
	writeBoxCoords(thisFileName, BOXCOORDS);
	
	// Get all appodizing params before writing to write_pif()
	appodizing = CIRCLEPACKED;
	// This is border width now
	TextSetInt(appodBoxRadW, featherRadius-boxRadius, FALSE);
	TextSetInt(appodFeaRadW, featherRadius, FALSE);
	raiseIt(appodizeScreenW);
      }
      
      updateCursor(drawingW, old_cursor); // Restore original cursor
    }
    else if(writeType==RECTPACKED || writeType==APPENDRECT ) {
      // Draw the rectangle
      if(fileType==PACKEDPORTABLE) { // We read a packed portable file
	
	// Forget appodizing
	blindWriting = TRUE;
	write_pif(purdueHeader.numCols, purdueHeader.numRows, 
		  fileName.addr, RECTPACKED);
	blindWriting = FALSE;
      }
      else if(rectBox.drawable)
	write_pif(rectBox.lowx - rectBox.x, rectBox.lowy - rectBox.y, 
		  fileName.addr, RECTPACKED); // Write out Rectangle
      else
	askThem("No rectangle to write.", XmDIALOG_INFORMATION);
    }
    else if(writeType==MAP) { // Write a MAP
      write_pif(numCols, numRows, fileName.addr, MAP);
    }
    else if(writeType==IMG) { // Write a scanned Img
      if(fileType==PACKEDPORTABLE) { // Fudge it so it works
	freeNullFloat(&inCore_buffer);
	inCore_buffer = (float *)malloc(numCols*numRows);
	
	for (i=0; i<numRows*numCols; i++)
	  inCore_buffer[i] = packImg[imgNum-1][i];
	
	write_pif(numCols, numRows, fileName.addr, EMPORTABLE);
	free(inCore_buffer);
	imgInCore = FALSE;
      }
      else // Normal
	write_pif(file_width, file_height, fileName.addr, EMPORTABLE);
    }
    else if(writeType == BOXCOORDS)
      writeBoxCoords(fileName.addr, BOXCOORDS);
    else if(writeType == STATE)
      writeState(fileName.addr);
    else if(writeType == PCX)
      writePCX(fileName.addr, imgNum, imgNum, 0); // last arg 0 --> non-movie
    else if(writeType == TIFF) {
      char fName[256];
      sprintf(fName, "%s", fileName.addr);
      if(!strrchr(fName,'.'))
	sprintf((fName + strlen(fileName.addr)), ".raw");      
      writeRAW(fName);  
    }
    else if(writeType==PARAMFILE) {
      sprintf(outParamFilename,"%s",fileName.addr);
      // ROBEM still thinks there's a packed boxes file there
      sprintf(fileName.addr,"%s",oldFilename);
      write_param(outParamFilename, 1);
    }
    else 
      askThem("[Error 583] Unknown write type", XmDIALOG_ERROR);
  }
  else if(readWrite==READCM) { // Read a colormap
    sprintf(outStr,"Reading [%s]",fileName.addr);
    if(checkStatus(outStr))
      goto destructor;
    // Save the old colormap
    pushColormap();
    readColormap(fileName.addr);
    histogram_activate();
  }
  else if(readWrite==WRITECM) { // Write a colormap
    sprintf(outStr,"Writing [%s]",fileName.addr);
    if(checkStatus(outStr))
      goto destructor;
    if(useGUI) writeColormap(fileName.addr);
  }
  
 destructor:;
  if(useGUI) {
    // Restore the old cursor whatever it was
    updateCursor(drawingW, old_cursor);
    
    XtUnmanageChild(statusScreenW);  
    
    if(shadeItSwitch) { // Automatically shade it
      mapShadeScreen_activate(NULL, NULL, NULL); // Now shade them
      
      // Give feedback on what the surface density *really* is
      // Make sure the user sees it with the right sign
      surfaceDensity = shadeItDensity;
      sprintf(outStr,"%.2f",irev*surfaceDensity);
      TextSet(surfDensValueW, outStr, FALSE); 
      
      mapShade_activate(NULL, NULL, NULL);   
      
      shadeItSwitch = FALSE;  // Don't do again next time
    }
  }
  
} // end function gotFilename



void imgNumText_changed(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: imgNumText_changed
  // When they change the imgNumText widget they need to press the
  // "imgNum" button next to it to get the change instantiated.

  int tag = *((int *) client_data);

  if(!locked) {
    if(tag==USEBOX) { // boxInit
      lastNumBox =  wstrtod(boxInitW);
      lastNumBox--;
      lastNumBox = MAX(lastNumBox, 0);
    }
    else if(tag==IMGNUMBUTTON) { // imgNum
      imgNum = wstrtod(imgNumTextW);
      imgNum = MAX(imgNum, 1);
      if(fileType==PACKEDPORTABLE || fileType==MAP)
	imgNum = MIN(imgNum,purdueHeader.nimg);
      imgNum = MAX(imgNum, 1);
      if(shadedMaps && !dispSections)
	imgNum = MIN(imgNum,mapNumSecs);
      
      XFlush(dsply);
    } 
  }
  else { // Maybe they're trying to find a certain box
    if(tag==USEBOX &&
       (fileType==EMPORTABLE || fileType==PORTABLE || 
	fileType==BYTEPORTABLE)) { // boxInit
      
      findBox =  wstrtod(boxInitW);
      findBox -= (lastNumBox+1);  // Humans number from 1 not 0
      findBox = MIN(MAX(findBox,0),(numBoxes-1));
    }
  }

} // end function imgNumText_changed



void doTheImgNum(Widget widget, int speedFactor)
{
  // function name: doTheImgNum
  // Called from imgNumArrowActivate and imgNumArrowActivateButton().
  // Do the work required to change the imgNum concept. speedFactor
  // determines how "fast" to change the value of imgNum

  if(widget==imgnumArrowUpW        || widget==fftImgnumArrowUpW   ||
     widget==rotfltImgnumArrowUpW)
    imgNum += speedFactor;
  else if(widget==imgnumArrowDownW || widget==fftImgnumArrowDownW ||
	  widget==rotfltImgnumArrowDownW)
    imgNum -= speedFactor;
  
  if(fileType==PACKEDPORTABLE || fileType==MAP)
    imgNum = MIN(imgNum,purdueHeader.nimg);
  imgNum = MAX(imgNum, 1);
  if(shadedMaps && !dispSections)
    imgNum = MIN(imgNum,mapNumSecs);
  
  TextSetInt(imgNumTextW, imgNum, FALSE);
  
  if(haveParamFile) { // Make the param screen consistent
    int thisX, thisY;
    P_INFO *thisInfo;
    
    thisInfo = info.info;
    while(thisInfo != NULL) { // Find this one
      if(thisInfo->rank==imgNum)
	break; // found it
      thisInfo = thisInfo->next;
    }
    
    // Display this on the parm screen
    convertXY(&thisX, &thisY, thisInfo->phi, thisInfo->theta);
    findNearestParticle(&thisX,&thisY,&foundPart,FALSE,imgNum,TRUE,0);
    
    TextSetInt(ppinTextW, thisInfo->rank, FALSE);
  }
  else { // Try these
    if(fileType==PACKEDPORTABLE)
      stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		       packImg,purdueHeader.nimg, imgNum);
    else if(fileType==MAP)
      stuffMapImage(map_image_buffer, float_image_buffer1, 
		    file_width,file_height, nsecs, imgNum, 0);
    
    DrawWindow(); 
  }
  
  if(XtIsManaged(subScreenW) && diffSection)
    diffSec_really_doit();

} // end function doTheImgNum



void imgNumArrowActivateButton(Widget widget, XEvent *event)
{
  // function name: imgNumArrowActivateButton
  // Instead of making them type the new imgNum just let them click
  // the arrow.  Change the imgNum accordingly.  This rtn is for the
  // second and third mouse buttons.
  
  if(!locked)
    doTheImgNum(widget, GetButtonVal(event->xbutton.button, 1, 2, 5));
  
} // end function imgNumArrowActivateButton



void boxInitActivate(Widget  widget, int *tag,
		     XmAnyCallbackStruct *callback_data)
{
  // function name: boxInitActivate
  // When the "BoxInit" button is pressed redisplay the window using
  // the new boxInit
  
  if(!locked)
    TextSetInt(boxInitW, lastNumBox+1, TRUE);

  DrawWindow();    
  
} // end function boxInitActivate



void boxInitArrowActivateButton(Widget widget, XEvent *event)
{
  // function name: boxInitArrowActivateButton
  // Instead of making them type the new first/find box just let them
  // click the arrow. Change the first find box variable accordingly.

  int speedFactor;

  if(widget==boxinitArrowDownW)
    speedFactor = -GetButtonVal(event->xbutton.button, 1, 2, 5);
  else
    speedFactor = GetButtonVal(event->xbutton.button, 1, 2, 5);

  if(!locked) {
    lastNumBox = MAX(lastNumBox + speedFactor, 0);
    
    TextSetInt(boxInitW, lastNumBox+1, TRUE);
    DrawWindow();   
  }
  else { // Maybe they're trying to find a certain box
    if((fileType==EMPORTABLE || fileType==PORTABLE || 
	fileType==BYTEPORTABLE)) { // boxInit
      findBox += speedFactor;
      findBox = MIN(MAX(findBox,0),(numBoxes-1));
      
      TextSetInt(boxInitW, findBox+lastNumBox+1, TRUE);
      DrawWindow(); 
    }  
  }
  
} // end function boxInitArrowActivateButton



void boxNumArrowActivateButton(Widget widget, XEvent *event)
{
  // function name: boxNumArrowActivateButton
  // Instead of making them type the new box/section number to edit on
  // the editBox screen, just let them click the arrow.

  int speedfactor;

  if(widget==boxDownW)
    speedfactor = -GetButtonVal(event->xbutton.button, 1, 2, 5);
  else
    speedfactor = GetButtonVal(event->xbutton.button, 1, 2, 5);
  
  curBox += speedfactor;
  
  if(makingMovie) { // Put the proper movie frame up there
    // Set the text widget to agree
    if(shadedMaps && stereoMapShade)
      curBox = MAX(0,MIN(curBox,(numSecs-2)));
    else 
      curBox = MAX(0,MIN(curBox,(numSecs-1)));
    TextSetInt(editBoxTextW, curBox+lastNumBox+1, TRUE);
    // Display the new box
    stuffDisplayMapBuffer(curBox,editBoxZoom,0);
    if(shadedMaps && stereoMapShade)
      stuffDisplayMapBuffer(curBox+1,editBoxZoom,numCols);
    DrawEditBoxWindow(NULL, NULL, NULL);
  }
  else { // Put box up
    // Set the text widget to agree
    if(fileType==PACKEDPORTABLE)
      curBox = MAX(0,MIN(curBox,(purdueHeader.nimg-1)));
    else 
      curBox = MAX(0,MIN(curBox,(numBoxes-1)));
    TextSetInt(editBoxTextW, curBox+lastNumBox+1, TRUE);
    // Display the new box
    if(fileType==PACKEDPORTABLE)
      stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
    else 
      stuffDisplayBoxBuffer(curBox,image_fileName.addr,editBoxZoom);
    DrawEditBoxWindow(NULL, NULL, NULL);
  }
  
} // end function boxNumArrowActivateButton



void paintSizeArrowActivateButton(Widget widget, XEvent *event)
{
  // function name: paintSizeArrowActivateButton
  // Instead of making them type the new paint brush size just let
  // them click the arrow.

  int speedFactor;
  speedFactor = GetButtonVal(event->xbutton.button, 1, 2, 5);

  if(widget == paintDownW)
    speedFactor *= -1;
  
  paintSize += speedFactor;
  paintSize = MAX(1,MIN(paintSize, 50));
  TextSetInt(editBoxPaintTextW, paintSize, FALSE); 
  
} // end function paintSizeArrowActivateButton



void cmScrollArrowActivateButton(Widget widget, XEvent *event)
{
  // function name: cmScrollArrowActivateButton
  // Scroll the colormap.

  int i, nextPixel, speedFactor;
  XColor tempColor[7];

  if(widget==colormapScrollDownW)
    speedFactor = -GetButtonVal(event->xbutton.button, 1, 2, 5);
  else
    speedFactor = GetButtonVal(event->xbutton.button, 1, 2, 5);

  if(speedFactor > 0) { // UP
    for (i=0; i<6; i++) {
      // Let's store all of these even though we mayn't need them all 
      nextPixel = (tryColors-1) - i;
      tempColor[i] = color_defs[nextPixel]; // struct assignment
      tempColor[i].pixel = nextPixel;
    }
      
    for (i=(tryColors-1); i>=(7+speedFactor); i--) { 
      // Do the colors we have been assigned
      nextPixel = i - speedFactor;
      color_defs[i] = color_defs[nextPixel]; // struct assignment
      color_defs[i].pixel = i;
      XAllocColor (dsply, color_map, &color_defs[i]);
      pixels_return[i] = color_defs[i].pixel;
    }

    for (i=7; i<(7+speedFactor); i++) { // Do the rest
      nextPixel = ((7+speedFactor-1)-i);
      color_defs[i] = tempColor[nextPixel]; // struct assignment
      color_defs[i].pixel = i;
      XAllocColor (dsply, color_map, &color_defs[i]);
      pixels_return[i] = color_defs[i].pixel;
    } 
  }
  else if(speedFactor < 0) { // DOWN
    for (i=0; i<6; i++) {
      // Let's store all of these even though we mayn't need them all
      nextPixel = (7+i);
      tempColor[i] = color_defs[nextPixel]; // struct assignment
      tempColor[i].pixel = nextPixel;
    }
    
    for (i=7; i<(tryColors+speedFactor); i++) {
      // Do the colors we have been assigned
      nextPixel = i - speedFactor;
      color_defs[i] = color_defs[nextPixel]; // struct assignment
      color_defs[i].pixel = i;
      XAllocColor (dsply, color_map, &color_defs[i]);
      pixels_return[i] = color_defs[i].pixel;
    }
    
    for(i=(tryColors + speedFactor); i<tryColors; i++) { // Do the rest
      nextPixel = i - (tryColors - 1 + speedFactor);
      color_defs[i] = tempColor[nextPixel]; // struct assignment
      color_defs[i].pixel = i;
      XAllocColor (dsply, color_map, &color_defs[i]);
      pixels_return[i] = color_defs[i].pixel;
    } 
  }
  
} // end function cmScrollArrowActivateButton



void imgNumActivate(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{
  // function name: imgNumActivate
  // When the "ImgNum" button is pressed redisplay the window using
  // the new imgNum. From main screen.

  if(!locked) { // Only do while not locked
    if(haveParamFile) { // Make the param screen consistent
      P_INFO *thisInfo;
      int thisX, thisY;
      
      thisInfo = info.info;
      while(thisInfo != NULL) { // Find this one
	if(thisInfo->particle_number==imgNum)
	  break;
	thisInfo = thisInfo->next;
      }
      
      // Display this on the parm screen
      convertXY(&thisX, &thisY, thisInfo->phi, thisInfo->theta);
      findNearestParticle(&thisX,&thisY,&foundPart,FALSE,imgNum,TRUE,0);  
    }
    else { // Try these
      if(fileType==PACKEDPORTABLE)
	stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
			 packImg,purdueHeader.nimg, imgNum);
      else if(fileType==MAP)
	stuffMapImage(map_image_buffer, float_image_buffer1, 
		      file_width, file_height, nsecs, imgNum, 0);
      
      DrawWindow();
    }
  }
  
} // end function imgNumActivate



void colormap_ramp_change(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: colormap_ramp_change
  // From colormap screen
  
  if(widget == colorRbowW)
    rainbowSwitch = !rainbowSwitch;
  else if(widget == colorRampW)
    rampSwitch = !rampSwitch;
  
} // end function colormap_ramp_change



void colormap_ramp_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{
  // function name: colormap_ramp_activate
  // They are modifying the colormap from the interface.

  int i, j;
  float percStep;

  buildingCmap = TRUE;
  if(colorType==USEBUILD && (firstBuildY > 6) && (lastBuildY > 6)) {
    // Only do if we're building
    
    if(!rampSwitch && !rainbowSwitch) {
      // Set the colormap to this color - NO RAMPING
      for (i=firstBuildY; i<=lastBuildY; i++) {
	// Do the colors we have been assigned
	color_defs[i].pixel = i;
	color_defs[i].red   = redLevel   * PIXELCOLOR;
	color_defs[i].green = greenLevel * PIXELCOLOR;
	color_defs[i].blue  = blueLevel  * PIXELCOLOR;
	color_defs[i].flags = DoRed | DoGreen | DoBlue;
	XAllocColor (dsply, color_map, &color_defs[i]);
	pixels_return[i] = color_defs[i].pixel;
      }
    }
    else if(rampSwitch && !rainbowSwitch) { // RAMPING
      
      if(lastBuildY > firstBuildY)
	percStep = (brightLevel/256.0) / (float)(lastBuildY - firstBuildY);
      else 
	percStep = 0.0;
      
      // Set the colormap to this color
      for (i=firstBuildY; i<=lastBuildY; i++) {
	// Do the colors we have been assigned
	int thisRedLevel, thisGreenLevel, thisBlueLevel;
	thisRedLevel = redLevel + 255.0*(i-firstBuildY)*percStep;
	thisRedLevel = MIN(255,thisRedLevel);
	thisGreenLevel = greenLevel + 255.0 * (i-firstBuildY)*percStep;
	thisGreenLevel = MIN(255,thisGreenLevel);
	thisBlueLevel = blueLevel + 255.0 * (i-firstBuildY)*percStep;
	thisBlueLevel = MIN(255,thisBlueLevel);
	color_defs[i].pixel = i;
	color_defs[i].red   = thisRedLevel   * PIXELCOLOR;
	color_defs[i].green = thisGreenLevel * PIXELCOLOR;
	color_defs[i].blue  = thisBlueLevel  * PIXELCOLOR;
	color_defs[i].flags = DoRed | DoGreen | DoBlue;
	XAllocColor (dsply, color_map, &color_defs[i]);
	pixels_return[i] = color_defs[i].pixel;
      }
	  
      // Now "darken" any of the pixels of a radially cued image
      if(secondMapShade && radialMapShade) { // darken appropriate
	int darkVal, q;

	// Use the fraction scroll to soften the shading
	darkVal = (65535/(expStep+2)) * globFract;

	// Shade every four
	for(i=6+1+radialStepSize; i< tryColors-expStep; i+= expStep) {
	  if(i>=firstBuildY && i <= lastBuildY) { // Only do marked
	    for (j=expStep-2; j>=0; j--) { // Ramp all
	      color_defs[i+j].pixel = i+j;

	      q = darkVal*(expStep-2-j+1);
	      color_defs[i+j].red = MAX(color_defs[i+expStep-1].red - q, 0);
	      color_defs[i+j].blue = MAX(color_defs[i+expStep-1].blue - q, 0);
	      color_defs[i+j].green = MAX(color_defs[i+expStep-1].green - q, 0);
	      color_defs[i+j].flags = DoRed | DoGreen | DoBlue;

	      XAllocColor(dsply,color_map, &color_defs[i+j]);
	      pixels_return[i+j] = color_defs[i+j].pixel;
	    }
	    
	    // Do the BLACK special case
	    color_defs[i].pixel = i;
	    color_defs[i].red   = 0;
	    color_defs[i].green = 0;
	    color_defs[i].blue = 0;
	    
	    XAllocColor (dsply, color_map, &color_defs[i]);
	    pixels_return[i] = color_defs[i].pixel;
	  }
	  
	}
	
      }
      
    }
    else if(!rampSwitch && rainbowSwitch) { // RAinbowING
      
      if(lastBuildY > firstBuildY)
	percStep = (brightLevel/256.0) / (float)(lastBuildY - firstBuildY);
      else 
	percStep = 0.0;
      
      // Set the colormap to this color
      for (i=firstBuildY; i<=lastBuildY; i++) {
	// Do the colors we have been assigned
	int thisRedLevel,thisGreenLevel,thisBlueLevel;
	thisRedLevel = redLevel + 382.0*(i-firstBuildY)*percStep;
	thisRedLevel = MAX(MIN(255,thisRedLevel),0);
	thisGreenLevel = greenLevel + 760.0*(i-firstBuildY)*percStep;
	if(thisGreenLevel > 511)
	  thisGreenLevel = (thisGreenLevel-511);
	else if(thisGreenLevel > 255)
	  thisGreenLevel = 255-(thisGreenLevel-255);
	thisGreenLevel = MAX(MIN(255,thisGreenLevel),0);
	thisBlueLevel = blueLevel + 
	  255.0 - 382.0*(i-firstBuildY)*percStep;
	if(thisBlueLevel < 0)
	  thisBlueLevel = 0 - (2*thisBlueLevel);
	thisBlueLevel = MIN(255,thisBlueLevel);
	
	color_defs[i].pixel = i;
	color_defs[i].red   = thisRedLevel   * PIXELCOLOR;
	color_defs[i].green = thisGreenLevel * PIXELCOLOR;
	color_defs[i].blue  = thisBlueLevel  * PIXELCOLOR;
	color_defs[i].flags = DoRed | DoGreen | DoBlue;
	XAllocColor (dsply, color_map, &color_defs[i]);
	pixels_return[i] = color_defs[i].pixel;
      }
    }
    else {
      askThem("Can't Ramp and Rainbow at the same time.",
	      XmDIALOG_INFORMATION);
      return;
    }
    
    // Redraw the histogram
    histogram_activate();
    
  }

} // end function colormap_ramp_activate



void paramFileGenerate_activate(Widget widget, XtPointer client_data, 
				XtPointer call_data)
{
  // function name: paramFileGenerate_activate
  // Put up the interface for doing paramFileGeneration. This gives a
  // default param file they can use for bootstrapping their
  // reconstruction.

  char *s1, outParamFile[1024], tmpParamFile[1024];
  int saveHPF, units;
  float pixelSize,volts,ampFac,delfMaj,delfMin,angMaj,csCoef;
  int tag = *((int *) client_data);
  
  sprintf(tmpParamFile,"%s",fileName.addr);
  s1 = strrchr(tmpParamFile,'.');
  if(!s1)
    s1 = tmpParamFile+ (strlen(fileName.addr)-1);
  *s1 = '\0';
  
  sprintf(outParamFile,"%s.dat_000",tmpParamFile);
  s1 = strchr(fileName.addr,'/');
  if(s1)
    sprintf(paramBoxesFilename,"%s",fileName.addr);
  else
    sprintf(paramBoxesFilename,"%s/%s",getenv("PWD"),fileName.addr);

  units = 1;  // Angstroms
  
  pixelSize = wstrtod(ctfPixSizeW);
  volts     = wstrtod(ctfKvW)*1000;
  ampFac    = wstrtod(ctfAmpFactorW);
  delfMaj   = wstrtod(ctfMajorDefocusW);
  delfMin   = wstrtod(ctfMinorDefocusW);
  angMaj    = wstrtod(ctfAngMajorW);
  csCoef    = wstrtod(ctfSpAberW);
  
  // Stuff appropriate widgets
  TextSetFloat(pixsizeW, pixelSize*10.0, FALSE);
  TextSetInt(unitsW, units, FALSE);
  TextSetFloat(voltsW, volts, FALSE);
  TextSetFloat(ampfacsW, ampFac, FALSE);
  TextSetFloat(delfMajW, delfMaj, FALSE);
  TextSetFloat(delfMinW, delfMin, FALSE);
  TextSetFloat(angMajW, angMaj, FALSE);
  TextSetFloat(csCoefW, csCoef, FALSE);
  
  saveHPF = haveParamFile;
  haveParamFile = TRUE;
  if(tag == FULL) 
    write_param(outParamFile, 1); // Write header + ptle records
  else
    write_param(outParamFile, 0); // Write header only
  haveParamFile = saveHPF;
  
} // end function paramFileGenerate_activate



void paramFileRead_work()
{
  // function name: paramFileRead_work
  // They've chosen a param file now read it from disk.

  char inStr[1024], tmpParamFile[1024];
  FILE *fIn;
  
  sprintf(tmpParamFile,"%s",gottenParamFname);

  if ( (fIn = fopen(tmpParamFile, "r")) != NULL ) {
    get_str(fIn, '\n', inStr); /* first line */
    
    get_str(fIn, ',', inStr); 
    TextSet(ctfPixSizeW, inStr, FALSE); 
    TextSet(pixsizeW, inStr, FALSE);
    
    get_str(fIn, ',', inStr); 
    TextSet(unitsW, inStr, FALSE);
    
    get_str(fIn, ',', inStr); 
    TextSet(ctfKvW, inStr, FALSE); 
    TextSet(voltsW, inStr, FALSE);

    get_str(fIn, ',', inStr); 
    TextSet(ctfAmpFactorW, inStr, FALSE); 
    TextSet(ampfacsW, inStr, FALSE);
    
    get_str(fIn, ',', inStr); 
    TextSet(ctfMajorDefocusW, inStr, FALSE); 
    TextSet(delfMajW, inStr, FALSE);
    
    get_str(fIn, ',', inStr); 
    TextSet(ctfMinorDefocusW, inStr, FALSE); 
    TextSet(delfMinW, inStr, FALSE);
    
    get_str(fIn, ',', inStr); 
    TextSet(ctfAngMajorW, inStr, FALSE); 
    TextSet(angMajW, inStr, FALSE);
    
    get_str(fIn, '\n', inStr); 
    TextSet(ctfSpAberW, inStr, FALSE); 
    TextSet(csCoefW, inStr, FALSE);
    
    fclose(fIn);
  }
  else {
    sprintf(outStr,"Cannot find file [%s].",tmpParamFile);
    askThem(outStr, XmDIALOG_INFORMATION);
  }
  
  XFlush(dsply);
  DrawFFTWindow(NULL, NULL, NULL);
  
} // end function paramFileRead_work



void auto2DDefocusRefine_activate(Widget widget, XtPointer client_data, 
				  XtPointer call_data)
{
  // function name: auto2DDefocusRefine_activate
  // Put up the interface for doing auto2DDefocusRefines.

  float thisHiRes, pixelSize;
  
  if(ctfMajorDefocusValue != 0.0 && ctfMinorDefocusValue != 0.0 ) {
    TextSetFloat(a2ddrfjW, ctfMajorDefocusValue, FALSE);
    TextSetFloat(a2ddrfnW, ctfMinorDefocusValue, FALSE);
    TextSetFloat(a2ddragW, angMajor, FALSE);
    
    pixelSize   = wstrtod(ctfPixSizeW);
    thisHiRes = pixelSize * 2.0 ;
    thisHiRes *= 10.0;   // Put it in Angstroms
    TextSetFloat(a2ddrtfW, thisHiRes, FALSE);
    
    freeNullDouble(&fftasum);
    freeNullDouble(&poly_paras);
    freeNullFloat(&a2dr_s);
    freeNullFloat(&a2dr_sminusb);
    freeNullFloat(&a2dr_ctfsq);
    freeNullFloat(&a2dr_d);
    freeNullFloat(&a2dr_b);

    fftasum      = (double *) malloc(sizeof(double) * 1024*2048);
    poly_paras   = (double *) malloc(sizeof(double) * 1024);
    a2dr_s       = (float *)  malloc(sizeof(float)  * 1024);
    a2dr_sminusb = (float *)  malloc(sizeof(float)  * 1024);
    a2dr_ctfsq   = (float *)  malloc(sizeof(float)  * 1024);
    a2dr_d       = (float *)  malloc(sizeof(float)  * 1024);
    a2dr_b       = (float *)  malloc(sizeof(float)  * 1024);
    
    raiseIt(auto2DDRScreenW); 
    doingAuto2DDefocusRefine= TRUE;
  }
  else {
    askThem("You must estimate the defocus first.", XmDIALOG_INFORMATION);
    doingAuto2DDefocusRefine= FALSE;
  }
  
} // end function auto2DDefocusRefine_activate



void auto2DDR_work()
{
  // function name: auto2DDR_work
  // Do what's nedded for auto 2d Defocus refinement.

  int ncol, nrow, nimg, i, j, k, y;
  float pixelSize, volts, defMajor, defMinor, angle;
  float ampContrast, Cs, res_hi, chisq;
  float minF, maxF, scaleF;

  res_hi = wstrtod(a2ddrtfW);
  
  if(res_hi <= 0.0) {
    askThem("You must enter a value for Highest Resolution.", 
	    XmDIALOG_INFORMATION);
    return;
  }
  
  ncol = purdueHeader.numCols;
  nrow = purdueHeader.numRows;
  nimg = purdueHeader.nimg;
  
  pixelSize   = wstrtod(ctfPixSizeW);
  pixelSize *= 10.0;     // Convert to Angs
  printf("pixelSize = [%f]\n",pixelSize);
  volts   = wstrtod(ctfKvW)  * 1000.0; // Convert from kv to volts
  Cs   = wstrtod(ctfSpAberW);
  // DONT DO THIS??? Cs *= 10000000.0;  /+ Put into Angstroms
  ampContrast = wstrtod(ctfAmpFactorW);
  
  angle = wstrtod(ctfAngMajorW);

  // We need to change the sign of angMajor here because our XWindows
  // based buffer is upside down.  When we write to disk we need to
  // use the sign just as given from the interface

  angle = 0.0 - angle ;
  printf("OLD angle [%f]\n",angle);

  defMajor = wstrtod(a2ddrfjW);
  defMinor = wstrtod(a2ddrfnW);
  angle    = wstrtod(a2ddragW);

  // We need to change the sign of angMajor here because our
  // XWindows based buffer is upside down.  When we write
  // to disk we need to use the sign just as given from
  // the interface 
  // DO WE REALLY NEED TO DO THIS ????
  // angle = 0.0 - angle ;

  printf("defmaj[%f] old[%f]\n",defMajor,ctfMajorDefocusValue);
  printf("defmin[%f] old[%f]\n",defMinor,ctfMinorDefocusValue);
  printf("NEW angle [%f]\n",angle);
  
  TextSet(a2ddrfjW, "Calculating ...", TRUE);
  TextSet(a2ddrfnW, "Calculating ...", TRUE);
  TextSet(a2ddragW, "Calculating ...", TRUE);
  TextSet(a2ddrcsqW, "Calculating ...", FALSE);

  // Yongchang floats all the data after it's read in
  CK(allPackImg = (float *)malloc(sizeof(float)*nimg*ncol*nrow));
  for(i=0;i<nimg;i++) {
    float *thisP,backG;
    int reverseIt;
    thisP = packImg[i];
    backG = *thisP;
    for(j=0;j<(nrow);j++) {
      // I'm suspicious that Yongchang also doesn't do the TSB "flip"
      // in the Y direction ... ???
      reverseIt = j;
      reverseIt = nrow - j -1;
      for(k=0;k<(ncol);k++) {
	allPackImg[(i*ncol*nrow)+(j*ncol)+k] = 
	  (thisP[(reverseIt*ncol)+k] - backG);
      }
    }
  }
  
  printf ("pixelSize= [%f]\n",pixelSize);
  printf ("volts= [%f]\n",volts);
  printf ("defMajor= [%f]\n",defMajor);
  printf ("defMinor= [%f]\n",defMinor);
  printf ("angle= [%f]\n",angle);
  printf ("ampContrast= [%f]\n",ampContrast);
  printf ("Cs= [%f]\n",Cs);
  printf ("res_hi= [%f]\n",res_hi);

  if(ctf2dSaveFftasum == NULL)
    CK(ctf2dSaveFftasum = (float *)malloc(sizeof(float)*fftSize*fftSize*2));
  if(ctf2dSaveFftasumsq == NULL)
    CK(ctf2dSaveFftasumsq = (float *)malloc(sizeof(float)*fftSize*fftSize*2));

  yji_ctf2dfit__(poly_paras, &ncol, &nrow, &nimg, &fftSize, &pixelSize,
		&volts, &res_hi, &defMajor, &defMinor, &angle,
		&ampContrast, &Cs,
		allPackImg, &chisq, auto2DdrawBuf, auto2DdataBuf,
		fftasum, &hkl_max, &order_b, &order_d,
		&ctf2dCalcIt,
		ctf2dSaveFftasum, ctf2dSaveFftasumsq);
  ctf2dCalcIt = FALSE;  // We just did
  
  freeNullFloat(&allPackImg);
  
  // Scale buffer for display 
  farray_minmax(auto2DdrawBuf, &minF, &maxF, 256*512);
  scaleF = 200.0 - (maxF-minF);

  for(y=0 ; y<512*256; y++)
    auto2DdrawBuf[y] = 10.0 + (auto2DdrawBuf[y] - minF)*scaleF;
  
  printf("New vals - defmajor[%f] defminor[%f] angle[%f] chisq[%f] \n",
	 defMajor, defMinor, angle, chisq);
  
  // Put it in the first 180 degrees
  if(angle > 180.0)
    angle = angle - 180.0;
  
  TextSetFloat(a2ddrfjW, defMajor, FALSE);
  TextSetFloat(a2ddrfnW, defMinor, FALSE);
  TextSetFloat(a2ddragW, angle, FALSE);
  TextSetFloat(a2ddrcsqW, chisq, TRUE);

  auto2Dangle    = angle;
  auto2DdefMajor = defMajor;
  auto2DdefMinor = defMinor;

} // end function auto2DDR_work



void auto2DDR_activate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: auto2DDR_activate
  // They want to do 2d defocus estimate.
  
  auto2DDR_work();
  
} // end function auto2DDR_activate



void auto2DDR_curves_activate(Widget widget, XtPointer client_data, 
			      XtPointer call_data)
{
  // function name: auto2DDR_curves_activate
  // What curve(s) do they want drawn as feedback.

  if(widget == a2ddrSW)
    a2drSDraw = !a2drSDraw ;
  else if(widget == a2ddrSMBW)
    a2drSminusBDraw = !a2drSminusBDraw ;
  else if(widget == a2ddrCTF2W)
    a2drCtfsqDraw = !a2drCtfsqDraw ;
  else if(widget == a2ddrDW)
    a2drDDraw = !a2drDDraw ;
  else if(widget == a2ddrBW)
    a2drBDraw = !a2drBDraw ;

  DrawFFTWindow(NULL, NULL, NULL);

} // end function auto2DDR_curves_activate



void auto2DDR_ctflines_activate(Widget widget, XtPointer client_data, 
				XtPointer call_data)
{
  // function name: auto2DDR_ctflines_activate
  // Toggle drawing the CTF lines.

  drawAuto2Dctflines = !drawAuto2Dctflines;
  drawCTF = (drawAuto2Dctflines) ? TRUE : FALSE;
  DrawFFTWindow(NULL, NULL, NULL);

} // end function auto2DDR_ctflines_activate



void auto2DDR_scurve_activate(Widget widget, XtPointer client_data, 
			      XtPointer call_data)
{
  // function name: auto2DDR_scurve_activate
  // Toggle drawing scurve

  drawAuto2Dscurve = !drawAuto2Dscurve;

  if(drawAuto2Dscurve) {
    pickLattice = -99;  
    drawCTF = FALSE; 
  }

  DrawFFTWindow(NULL, NULL, NULL);

} // end function auto2DDR_scurve_activate



void auto2DDR_drawmodified_activate(Widget widget, XtPointer client_data, 
				    XtPointer call_data)
{
  // function name: auto2DDR_drawmodified_activate
  // Toggle drawing modified display.

  drawAuto2Dctf = !drawAuto2Dctf;
  DrawFFTWindow(NULL, NULL, NULL);

} // end function auto2DDR_drawmodified_activate



void auto2DDR_update_activate(Widget widget, XtPointer client_data, 
			      XtPointer call_data)
{
  // function name: auto2DDR_update_activate
  // Update the display text widgets and draw FFT screen.
  
  TextSet(ctfFmajTW, XmTextGetString(a2ddrfjW), FALSE);
  TextSet(ctfFminTW, XmTextGetString(a2ddrfnW), FALSE);
  TextSet(ctfAmajTW, XmTextGetString(a2ddragW), FALSE);
  TextSet(ctfAngMajorW, XmTextGetString(a2ddragW), FALSE);
  TextSet(ctfTFactorW, XmTextGetString(a2ddrtfW), TRUE);

  DrawFFTWindow(NULL, NULL, NULL);
  
} // end function auto2DDR_update_activate



void autoDwritePIFhead_activate(Widget widget, XtPointer client_data, 
				XtPointer call_data)
{
  // function name: autoDwritePIFhead_activate
  // Put up the interface for doing autoDwritePIFheads.

  int i;
  float a[10];
  forPifHead_t thisHead;
  fString_t rWrite;

  // We have aproblem with PIF where the range of values for these CTF
  // params may exceed the range of values we have for the image
  // pixels. Therefore our floatScaleFactor won't work. Propose we
  // store these as char strings

  for (i=0; i<6; i++)
    a[i+1] = adraValues[i];
  
  printf("Found:\n\nfilename[%s]\n\n\
a0 [%12E]\na1 [%12E]\na2 [%12E]\na3 [%12E]\na4 [%12E]\na5 [%12E]\n\n",
	 fileName.addr, a[1], a[2], a[3], a[4], a[5], a[6]);
  
  rWrite.addr = (char *)malloc(sizeof(char)*16);
  rWrite.length = 1;
  i = 4;
  thisHead.luNum = i;
  
  // Read
  sprintf(rWrite.addr,"r");
  pif_open(fileName.addr, &i, rWrite.addr);
  pif_read_gh(&thisHead);
  pif_close(&thisHead.luNum);
  
  printf("Found from read header:\n\
a0 [%12E]\na1 [%12E]\na2 [%12E]\n\
a3 [%12E]\na4 [%12E]\na5 [%12E]\n\n",
	 thisHead.a0, thisHead.a1, thisHead.a2,
	 thisHead.a3, thisHead.a4, thisHead.a5 );
  
  thisHead.a0 = a[1];
  thisHead.a1 = a[2];
  thisHead.a2 = a[3];
  thisHead.a3 = a[4];
  thisHead.a4 = a[5];
  thisHead.a5 = a[6];
  
  // Write
  sprintf(rWrite.addr,"r+");
  rWrite.length = 2;
  pif_open(fileName.addr, &i, rWrite.addr);
  pif_write_gh_(&thisHead);
  pif_close(&thisHead.luNum);
  
  free(rWrite.addr);

} // end function autoDwritePIFhead_activate



void autoDefocusRefine_activate(Widget widget, XtPointer client_data, 
				XtPointer call_data)
{
  // function name: autoDefocusRefine_activate
  // Put up the interface for doing autoDefocusRefines.

  float thisHiRes, pixelSize;
  int tag = *((int *) client_data);

  if(tag == 2) { // Historic way
    printf("Historic\n");
    autoDefocus1Dmethod = HISTORIC; 
  }
  else
    autoDefocus1Dmethod = FALSE; 

  if(ctfMajorDefocusValue != 0.0 && ctfMinorDefocusValue != 0.0 ) {
    int ncol, nrow, nimg;
    
    ncol = purdueHeader.numCols;
    nrow = purdueHeader.numRows;
    nimg = purdueHeader.nimg;
    CK(allPackImg = (float *)malloc(sizeof(float)*nimg*ncol*nrow));

    pixelSize   = wstrtod(ctfPixSizeW);
    thisHiRes = (pixelSize * 2.0) * 10.0; // Nyquist - converted to Angstroms
    TextSetFloat(adrHiResW, thisHiRes, FALSE);
    raiseIt(autoDRScreenW); 
    doingAutoDefocusRefine= TRUE;
  }
  else {
    askThem("You must estimate the defocus first.", XmDIALOG_INFORMATION);
    doingAutoDefocusRefine= FALSE;
  }

} // end function autoDefocusRefine_activate



void auto1DDR_work(int orderBkg, int orderSig,
		   float *calcDefocus, float *calcChisq)
{
  // function name: auto1DDR_work
  // Do what's necessary to setup for 1D defocus estimation .

  int ncol, nrow, nimg, i, j, k;
  int bkg_poly_order, decay_poly_order, ma;
  float pixelSize, volts, defMajor, ampContrast, Cs, res_hi, chisq;
  float *a;

  if(autoDefocus1Dmethod==HISTORIC) { // Do old way
    printf("OLD method\n"); 
    return;
  }

  res_hi = wstrtod(a2ddrtfW);

  if(res_hi <= 0.0) {
    askThem("You must enter a value for Highest Resolution.",
	    XmDIALOG_INFORMATION);
    return;
  }
  
  ncol = purdueHeader.numCols;
  nrow = purdueHeader.numRows;
  nimg = purdueHeader.nimg;
  
  res_hi = wstrtod(adrHiResW);
  
  pixelSize   = wstrtod(ctfPixSizeW);
  pixelSize *= 10.0;     // Convert to Angs
  printf("pixelSize = [%f]\n",pixelSize);
  volts   = wstrtod(ctfKvW)  * 1000.0; // Convert from kv to volts
  Cs   = wstrtod(ctfSpAberW);
  // DONT DO THIS??? Cs *= 10000000.0;  /+ Put into Angstroms
  ampContrast = wstrtod(ctfAmpFactorW);

  bkg_poly_order   = orderBkg;
  decay_poly_order = orderSig;

  bkg_poly_order++; decay_poly_order++;	
  ma = bkg_poly_order+decay_poly_order+1;
  printf("bpo = %d  dpo = %d  ma = %d\n",
	 bkg_poly_order, decay_poly_order, ma);
  a = (float *)malloc(sizeof(float)*ma+5);
  for (i=1; i<=(bkg_poly_order+decay_poly_order); i++)  
    *(a+i) = 0.0;
  *(a) = 1.0;
  *(a+bkg_poly_order)=1.0;
  
  // Let's get these from interface
  for (i=0; i<6; i++) // a 
    a[i+1] = adraValues[i];  
  for (i=0; i<6; i++) // b
    a[i+8] = adrbValues[i];

  // Yongchang floats all the data after it's read in
  for (i=0; i<nimg; i++) {
    int reverseIt;
    float *thisP, backG;

    thisP = packImg[i];
    backG = *thisP;
    for (j=0; j<nrow; j++) {
      // I'm suspicious that Yongchang also doesn't do the TSB "flip"
      // in the Y direction ... ???
      reverseIt = j;
      reverseIt = nrow - j -1;
      for (k=0; k<ncol; k++) {
	allPackImg[(i*ncol*nrow)+(j*ncol)+k] = 
	  (thisP[(reverseIt*ncol)+k] - backG);
      }
    }
  }

  // Put in proper units
  Cs *= 10000000.0; 
  defMajor =   ctfMajorDefocusValue * 10000.0;  // Put it in Angstroms

  printf ("pixelSize= [%f]\n",pixelSize);
  printf ("volts= [%f]\n",volts);
  printf ("defMajor= [%f]\n",defMajor);
  printf ("ampContrast= [%f]\n",ampContrast);
  printf ("Cs= [%f]\n",Cs);
  printf ("res_hi= [%f]\n",res_hi);
  
  init_floatarr(adr_s, 256, 0.0);        // Initialize arrays
  init_floatarr(adr_sminusb, 256, 0.0);
  init_floatarr(adr_ctfsq, 256, 0.0);
  init_floatarr(adr_d, 256, 0.0);
  init_floatarr(adr_b, 256, 0.0);
  
  ca1d_save = (float *)malloc(sizeof(float)*1024);
  if(ctf1dSaveFftAvg == NULL)
    CK(ctf1dSaveFftAvg = (float *)malloc(sizeof(float)*fftSize*fftSize*2));

  yji_ctf1dfit__(&ncol, &nrow, &nimg, &fftSize, &pixelSize,
		&volts, &res_hi, &defMajor, 
		&ampContrast, &Cs, allPackImg, &chisq, 
		a, adr_s, adr_sminusb,
		adr_ctfsq, adr_d, adr_b,
		ca1d_save, &ctf1dCalcIt, ctf1dSaveFftAvg);

  ctf1dCalcIt = FALSE;  // We just did

  for (i=0; i<256; i++) 
    ca1d[i] = ca1d_save[i];
  free(ca1d_save);  

  printf("New 1D vals - defmajor[%f] chisq[%f] \n", defMajor, chisq);
  *calcDefocus = defMajor*10000.0; // We divide it later
  *calcChisq   = chisq;
  
  // Now put new array values on interface
  for(i=0; i<6; i++) { // a
    sprintf(outStr,"%12E",a[i]);
    XtVaSetValues(adraW[i], XmNlabelString, XmStringCreateLocalized(outStr), NULL);
    adraValues[i] = a[i];
  }
  for(i=0; i<6; i++) { // b
    sprintf(outStr,"%12E",a[i+6]);
    XtVaSetValues(adrbW[i], XmNlabelString, XmStringCreateLocalized(outStr), NULL);
    adrbValues[i] = a[i+6];
  }
  
} // end function auto1DDR_work



void autoDR_activate(Widget widget, XtPointer client_data, 
		     XtPointer call_data)
{
  // function name: autoDR_activate
  // Setup and do auto defocus refine.

  int i, j, iter;
  int orderBkg, orderSig, idim_min, thisIdim1, thisIdim2;
  int arg1, arg2, arg3, arg4, hiPixel, loPixel, linear;
  float map_min, map_max, map_avg, map_std;
  float calcDefocus, calcChisq;
  float *tmpBuffer;

  int tag = *((int *) client_data);

  for (iter=0; iter<tag; iter++) { // DO the calc
    
    // Now let's get the circAvg
    thisIdim1= thisIdim2 = 512;
    idim_min = MIN(thisIdim1,thisIdim2);

    freeNullFloat(&ca1d);
    freeNullFloat(&ca1d_std);
    freeNullFloat(&ca1d_cnt);

    ca1d     = (float *)malloc(sizeof(float)*(idim_min/2)+2);
    ca1d_std = (float *)malloc(sizeof(float)*(idim_min/2)+2);
    ca1d_cnt = (float *)malloc(sizeof(float)*(idim_min/2)+2);

    freeNullFloat(&adr_s);
    freeNullFloat(&adr_sminusb);
    freeNullFloat(&adr_ctfsq);
    freeNullFloat(&adr_d);
    freeNullFloat(&adr_b);

    adr_s       = (float *)malloc(sizeof(float) * 1024);
    adr_sminusb = (float *)malloc(sizeof(float) * 1024);
    adr_ctfsq   = (float *)malloc(sizeof(float) * 1024);
    adr_d       = (float *)malloc(sizeof(float) * 1024);
    adr_b       = (float *)malloc(sizeof(float) * 1024);
    
    arg1 = fftSize;
    arg2 = fftSize;
    arg3 = 512;
    arg4 = 512;
    loPixel = 8;
    hiPixel = tryColors-2;
    linear = (fftType==LINEARFFT) ? -1 : 0; // 0 --> log

    fft_dsp_c(fftImage_buffer2, 
	      fft_buffer2,
	      arg1, arg2, arg3, arg4,
	      globMaxFFTvalue, globMinFFTvalue,
	      hiPixel, loPixel, linear,
	      (fftDispType==AMPLITUDE)?TRUE:FALSE);
    
    tmpBuffer = (float *)malloc(sizeof(float)*thisIdim1*thisIdim2);
    for(i=0;i<thisIdim1;i++) // XXXX Index calculation may be wrong!!!
      for(j=0;j<thisIdim2;j++)
	tmpBuffer[(i*thisIdim1)+j]= (float)fft_buffer2[(i*thisIdim1)+j];

    map2d_stats__(tmpBuffer, &thisIdim1, &thisIdim2,
		 &map_min, &map_max,
		 &map_avg, &map_std,
		 ca1d, ca1d_std, ca1d_cnt,
		 &idim_min);
    
    ca1d_avg = map_avg;
    
    free(tmpBuffer);
        
    // Now let's calc the defocus
    // Let's just force it to be 4 for now
    // orderBkg = (int)strtod(XmTextGetString(orderBkgW), 0);
    // orderSig = (int)strtod(XmTextGetString(orderSigW), 0);
    
    // OLD way 
    // orderBkg = orderSig = 4;

    orderBkg = orderSig = 5;
    //RSSAAA - 1024^2 crash on auto1DDR_work
    auto1DDR_work(orderBkg, orderSig, &calcDefocus, &calcChisq);
    sprintf(outStr,"%f",calcDefocus/10000.0); // /10000.0 to convert to um
    XtVaSetValues(adrDefocusW, XmNlabelString, XmStringCreateLocalized(outStr), NULL);
    sprintf(outStr,"%f",calcChisq);
    XtVaSetValues(adrChisqW, XmNlabelString, XmStringCreateLocalized(outStr), NULL);
    
    global_defocus = calcDefocus/10000.0;

    XFlush(dsply);
  }
  
} // end function autoDR_activate



void autoDR_update_activate(Widget widget, XtPointer client_data, 
			    XtPointer call_data)
{
  // function name: autoDR_update_activate
  // Update text widgets with defocus estimate.

  TextSetFloat(ctfFmajTW, global_defocus, FALSE);
  TextSetFloat(ctfFminTW, global_defocus, TRUE);

  DrawFFTWindow(NULL, NULL, NULL);

} // end function autoDR_update_activate



void adrab_reset_activate(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: adrab_reset_activate
  // Reset param text widgets for auto defocus refine so a new
  // refeinement can be started.

  int i;

  XtVaSetValues(adraW[0], XmNlabelString, XmStringCreateLocalized("1.0"), NULL);
  XtVaSetValues(adrbW[0], XmNlabelString, XmStringCreateLocalized("1.0"), NULL);
  adraValues[0] = 1;
  adrbValues[0] = 1;

  for(i=1; i<6; i++) {
    XtVaSetValues(adraW[i], XmNlabelString, XmStringCreateLocalized("0.0"), NULL);
    XtVaSetValues(adrbW[i], XmNlabelString, XmStringCreateLocalized("0.0"), NULL);
    adraValues[i] = 0;
    adrbValues[i] = 0;
  }

} // end function adrab_reset_activate



void adr_toggle_change(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: adr_toggle_change
  // Toggle display of curves in auto defocus refine screen

  if(widget == adrCircW)
    adrCircAvgDraw= !adrCircAvgDraw;
  else if(widget == adrSW)
    adrSDraw= !adrSDraw;
  else if(widget == adrSMBW)
    adrSminusBDraw= !adrSminusBDraw;
  else if(widget == adrCtf2W)
    adrCtfsqDraw= !adrCtfsqDraw;
  else if(widget == adrDW)
    adrDDraw=  !adrDDraw;
  else if(widget == adrBW)
    adrBDraw=  !adrBDraw;
  else if(widget == adrZoomW)
    adrZoomDraw=  !adrZoomDraw;
  
  DrawFFTWindow(NULL, NULL, NULL);
  
} // end function adr_toggle_change



void fix2D_toggle_change(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: fix2d_toggle_change
  // Toggle fix2d display properties

  if(widget == fix2DGridW)
    drawFix2Dgrid = !drawFix2Dgrid;
  else if(widget == fix2DStatsW)
    drawFix2Dstats = !drawFix2Dstats;
  
  DrawWindow();

} // end function fix2D_toggle_change



void version_scroll_activate(Widget widget, XtPointer client_data, 
			     XtPointer call_data)
{
  // function name: version_scroll_activate
  // Scroll the version number widget. Useful on VMS but not sure if
  // applicable to UNIX - keep in case a need can be found.

  char *versionText, *s1;
  char thisFileName[512], rawFileName[512];
  int thisVersion;
  off_t imgOffset;
  XEvent event;
  forPifHead_t thisHead;
  FILE *fIn;
  fString_t rWrite,inFile;

  int tag = *((int *) client_data);

  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 1;
  sprintf(rWrite.addr,"r");
  inFile.addr = malloc(sizeof(char)*64);
  inFile.length = 1;
  
  if(inited) { // Don't do until we've got everything setup
    
    if(tag==SCROLL) { // Scroll them puppies
      
      sprintf(rawFileName,"%s",fileName.addr);
      if((s1= strrchr(rawFileName,';'))!=NULL)
	    *s1= '\0';
      thisVersion = 1;

      while(TRUE) { // Do until stopped
	fIn = 0;

	while(!fIn) { // Try this
	  sprintf(thisFileName,"%s;%d",rawFileName,thisVersion);
	  fIn = fopen(thisFileName, "r");
	  if(!fIn && thisVersion==1) { // Dont know what to do
	    sprintf(outStr,"Cannot find this file [%s]", thisFileName);
	    askThem(outStr, XmDIALOG_INFORMATION);
	    goto noMoreScroll;
	  }
	  if(!fIn)
	    thisVersion = 1;
	  else 
	    thisVersion++;
	}

	fclose(fIn);

	TextSetInt(versionTextW, thisVersion-1, TRUE);
	
	sprintf(outStr,"%d",(thisVersion-1));
	doLabel(outStr, versionTextW, FALSE);
	
	thisHead.luNum = 7;
	sprintf(inFile.addr,"%s",thisFileName);
	inFile.length = strlen(thisFileName);
	
	pif_open(inFile.addr, &thisHead.luNum, rWrite.addr);
	
	pif_read_gh(&thisHead);
	thisHead.imgNum = 1;
	pif_read_dh(&thisHead);
	
	purdueHeader.numCols= thisHead.numCols;
	purdueHeader.numRows= thisHead.numRows;
	purdueHeader.nimg=  thisHead.nimg;
	
	// If we can just read the whole thing into core memory then
	// virtual_read_activate() won't have to work as hard.

	freeNullFloat(&inCore_buffer);
	CK(inCore_buffer= (float *)malloc(sizeof(float)*
					  thisHead.numCols*thisHead.numRows));
	imgInCore = (inCore_buffer) ? TRUE : FALSE;
	
	if(imgInCore) { // OK to go
	  imgOffset = 1024;  // one image for now
	  if(thisHead.mode==1) { // short
	    short *short_buffer;
	    int z;
	    short_buffer = (short *)
	      malloc(sizeof(short)*
		     thisHead.numCols*thisHead.numRows);
	    pif_read_short_image(&thisHead.luNum,
				 short_buffer,
				 &imgOffset,
				 &thisHead.numCols,
				 &thisHead.numRows,
				 &thisHead.numSecs);
	    for(z=0;z<(thisHead.numCols*thisHead.numRows);z++)
	      inCore_buffer[z] = (float)short_buffer[z];

	    free(short_buffer);
	  }
	  else if(thisHead.mode==0) { // byte
	    BYTE *byte_buffer;
	    int z;
	    byte_buffer = (BYTE *)malloc(sizeof(BYTE)*
					 thisHead.numCols*thisHead.numRows);
	    pif_read_byte_image(&thisHead.luNum,
				byte_buffer,
				&imgOffset,
				&thisHead.numCols,
				&thisHead.numRows,
				&thisHead.numSecs);
	    for(z=0; z<thisHead.numCols*thisHead.numRows; z++)
	      inCore_buffer[z] = byte_buffer[z];
	    free(byte_buffer);
	  }
	}
	else {
	  sprintf(outStr,"Cannot allocate space for [%s]",
		  thisFileName);
	  askThem(outStr, XmDIALOG_INFORMATION);
	  goto noMoreScroll;
	}
	
	pif_close(&thisHead.luNum);
	
	virtual_read_activate(-99,-99);
	
	// Give feedback on main window title bar
	sprintf(outStr,"Robem: [%s] - %d images total.",
		inFile.addr, thisHead.nimg);

	doDialogTitle(outStr, mainWindowW, FALSE);
	
	// Frame the location of the stop button
	DrawX(XtWindow(versionStopW), image_gcRED, 16, 16, 4);
	XFlush(dsply);
	
	// See if they want to bugout or not
	if(XCheckTypedWindowEvent(dsply,
				  XtWindow(versionStopW),
				  ButtonPress,
				  &event)) { // STOP!!!
	  XFlush(dsply);
	  goto noMoreScroll;
	}
	
      }
    }
    else { // Just do this version number
      versionText = XmTextGetString(versionTextW); 
      
      if(!EQU(versionText,"")) { // Don't do in the interim state
	sprintf(rawFileName,"%s",fileName.addr);
	if((s1= strrchr(rawFileName,';'))!=NULL)
	    *s1= '\0';
	sprintf(thisFileName,"%s;%s",rawFileName,versionText);
	
	if ( (fIn = fopen(thisFileName, "r")) == NULL ) {
	  sprintf(outStr,"Cannot find this file [%s]",
		  thisFileName);
	  askThem(outStr, XmDIALOG_INFORMATION);
	  goto noMoreScroll;
	}
	fclose(fIn);
	
	thisHead.luNum = 7;
	sprintf(inFile.addr,"%s",thisFileName);
	inFile.length = strlen(thisFileName);
	
	pif_open(inFile.addr, &thisHead.luNum, rWrite.addr);
	
	pif_read_gh(&thisHead);
	thisHead.imgNum = 1;
	pif_read_dh(&thisHead);
	
	purdueHeader.numCols= thisHead.numCols;
	purdueHeader.numRows= thisHead.numRows;
	purdueHeader.nimg=  thisHead.nimg;
	
	// If we can just read the whole thing into core memory then
	// virtual_read_activate() won't have to work as hard.

	freeNullFloat(&inCore_buffer);
	CK(inCore_buffer= (float *)malloc(sizeof(float)*
					  thisHead.numCols*thisHead.numRows));
	imgInCore = (inCore_buffer) ? TRUE : FALSE;
	
	if(imgInCore) { // OK to go
	  imgOffset = 1024;  // one image for now
	  if(thisHead.mode==1) { // short
	    short *short_buffer;
	    int z;
	    short_buffer = (short *)
	      malloc(sizeof(short)*
		     thisHead.numCols*thisHead.numRows);
	    pif_read_short_image(&thisHead.luNum,
				 short_buffer,
				 &imgOffset,
				 &thisHead.numCols,
				 &thisHead.numRows,
				 &thisHead.numSecs);
	    for(z=0;z<(thisHead.numCols*thisHead.numRows);z++)
	      inCore_buffer[z] = (float)short_buffer[z];
	    free(short_buffer);
	  }
	  else if(thisHead.mode==0) { // byte
	    BYTE *byte_buffer;
	    int z;
	    byte_buffer = (BYTE *)malloc(sizeof(BYTE)*
					 (thisHead.numCols)*
					 (thisHead.numRows));
	    pif_read_byte_image(&thisHead.luNum,
				byte_buffer,
				&imgOffset,
				&thisHead.numCols,
				&thisHead.numRows,
				&thisHead.numSecs);
	    for(z=0;z<(thisHead.numCols*thisHead.numRows);z++)
	      inCore_buffer[z] = byte_buffer[z];
	    free(byte_buffer);
	  }
	}
	else { // error
	  sprintf(outStr,"Cannot allocate space for [%s]",
		  thisFileName);
	  askThem(outStr, XmDIALOG_INFORMATION);
	  goto noMoreScroll;
	}
	
	pif_close(&thisHead.luNum);
	
	virtual_read_activate(-99,-99);
	
	// Give feedback on main window title bar
	sprintf(outStr,"Robem: [%s] - %d images total.",
		inFile.addr, thisHead.nimg);
	
	doDialogTitle(outStr, mainWindowW, FALSE);
      }
      XtFree(versionText);
      
    }
    
  }

 noMoreScroll:;
  free(rWrite.addr);
  free(inFile.addr);
  
} // end function version_scroll_activate



void DrawColormapWindow(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: DrawColormapWindow
  // Redraw window e.g. on expose events or whenever it's prudent.
  
  if(inited && XtIsManaged(colorScreenW))
    XPutImage(dsply, color_window1, color_gc, color_image1, 0, 0,
	      0, 0, 32, 256);
  XFlush (dsply);
  
} // end function DrawColormapWindow



void DrawTestColormapWindow(Widget widget, XtPointer client_data, 
			    XtPointer call_data)
{
  // function name: DrawTestColormapWindow
  // Redraw window e.g. on expose events or whenever it's
  // prudent. Draws either the histogram area of the test color area
  // on the colormap screen.

  if(inited && XtIsManaged(colorScreenW)) {
    if(widget == histW)
      XPutImage(dsply, hist_color_window1, hist_color_gc, 
		hist_color_image1, 
		0, 0, 0, 0,  32,  256); 
    else if(widget == testcolormapW)
      XPutImage(dsply, test_color_window1, test_color_gc, 
		test_color_image1, 
		0, 0, 0, 0,  32,  32);
    
    XFlush (dsply);
  }
  
} // end function DrawTestColormapWindow



void colorScrollChange(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: colorScrollChange
  // Whenever they've changed the value of one of the scroll bars on
  // the Alter Colormap screen call here.  Set the level based on the
  // position of the scrollbar.  Set the PLACEHOLDER entry of our
  // colormap to the value of the RGB scrolls.  This will cause the
  // TEST drawing area of the Alter Colormap screen to change to this
  // new color as it is comprised of pixels with PLACEHOLDER value.

  int i;
  int tag = *((int *) client_data);
  XmScrollBarCallbackStruct *temp = (XmScrollBarCallbackStruct *) call_data;

  if(tag==RED)
      redLevel = temp->value;
  else if(tag==GREEN)
      greenLevel = temp->value;
  else if(tag==BLUE)
      blueLevel = temp->value;
  else if(tag==BRIGHT)
      brightLevel = temp->value;
  else if(tag==STEPCM) {
    int thisStep, thisSection, numSections;
    float colorPerSection;
    
    stepCM = (int)temp->value;
    sprintf(outStr,"%3d",stepCM);
    doLabel(outStr, stepCMValueW, FALSE);
    
    // Now step the colormap
    for (i=7; i<tryColors; i++) { // Do the colors we have been assigned
      // Divide the entire colormap in STEP sections 
      // stepCM is always >= 1

      numSections = (int)((tryColors-7.0)*(stepCM/256.0))+1;
      thisSection = (int)((i-7.0) * (numSections / (tryColors-7.0)));
      
      // Start the colorPerSection out as GRAYscale and override 
      // if desired
      if(numSections > 1)
	numSections -= 1;      // To make color brighter
      colorPerSection = 255.0 / numSections;
      
      color_defs[i].pixel = i;
      
      if(colorType==USEBUILD)
	colorPerSection = (float)redLevel / (float)numSections;
      thisStep = thisSection * colorPerSection;
      
      color_defs[i].red   = thisStep * PIXELCOLOR;
      
      if(colorType==USEBUILD)
	colorPerSection = (float)greenLevel / (float)numSections;
      thisStep = thisSection * colorPerSection;
      color_defs[i].green = thisStep * PIXELCOLOR;
      
      if(colorType==USEBUILD)
	colorPerSection = (float)blueLevel / (float)numSections;
      thisStep = thisSection * colorPerSection;
      color_defs[i].blue  = thisStep * PIXELCOLOR;
      
      color_defs[i].flags =  DoRed | DoGreen | DoBlue;
      
      XAllocColor (dsply, color_map, &color_defs[i]);
      pixels_return[i] = color_defs[i].pixel;
      XFlush(dsply);
    }
  }
  
  if(tag!=STEPCM) { // Give feedback
    sprintf(outStr,"%3d",temp->value);
    doLabel(outStr, colorValueW[tag], FALSE);
    
    // Stuff the PLACEHOLDER entry of our colormap with this new value
    color_defs[PLACEHOLDER].pixel =  PLACEHOLDER;
    color_defs[PLACEHOLDER].red   =  redLevel   * PIXELCOLOR;
    color_defs[PLACEHOLDER].green =  greenLevel * PIXELCOLOR;
    color_defs[PLACEHOLDER].blue  =  blueLevel  * PIXELCOLOR;
    color_defs[PLACEHOLDER].flags = DoRed | DoGreen | DoBlue;
    
    int thisI;
    XAllocColor(dsply,color_map, &color_defs[PLACEHOLDER]);
    pixels_return[PLACEHOLDER] = color_defs[PLACEHOLDER].pixel;
    XSetForeground(dsply, image_gcPLACEHOLDER, 
		   pixels_return[PLACEHOLDER]);
    // DO this ???
    for(thisI = 0; thisI<32; thisI++)
      XDrawLine(dsply, test_color_window1, image_gcPLACEHOLDER,
		0, thisI, 32, thisI);
    
    if(colorType==USEBUILD) // Update the colormap based on this new color
      colormap_ramp_activate(0,0,0);    
  }
  
} // end function colorScrollChange



void colorClick(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: colorClick
  // Every time the mouse button is clicked call here. This routine
  // handles the selecting/deselecting of points on the screen.

  char ans;
  int i, j;
  static int firstY, lastY;
  XEvent event;

  XmDrawingAreaCallbackStruct *draw_data = 
    (XmDrawingAreaCallbackStruct *) call_data;

  if(draw_data == 0)
    return;

  if(colorType==USEBUILD) { // Only do if we're building
    event= *(draw_data->event);
    
    switch (event.type) {
      
    case ButtonRelease:
      // Released Button1 means they are done picking points of the range
      if(event.xbutton.button == Button1) {
	// This y is the last one they want
	lastBuildY = event.xbutton.y;
	
	// Now convert the Y's into colormap space
	firstBuildY = tryColors - firstBuildY;
	// We display from bottom up
	lastBuildY = tryColors - lastBuildY;      
	firstBuildY = MIN((tryColors-1),MAX(firstBuildY,7));
	lastBuildY = MIN((tryColors-1),MAX(lastBuildY,7));
	if(firstBuildY > lastBuildY) { // switch them
	  int temp;
	  temp = lastBuildY;
	  lastBuildY= firstBuildY;
	  firstBuildY= temp;
	}
	
	colormap_ramp_activate(0,0,0);
	if(fileType==PACKEDPORTABLE)
	  stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
			   packImg,purdueHeader.nimg, imgNum);
	else if(fileType==MAP)
	  stuffMapImage(map_image_buffer, float_image_buffer1, 
			file_width, file_height, nsecs, imgNum, 0);
	else
	  virtual_read_activate(-99,-99);
	DrawWindow();
	histogram_activate();
	XFlush (dsply);
      }
      break;
      
    case ButtonPress:
      // They are picking the firsrt point in the range
      if(event.xbutton.button == Button1) {
//	firstX      = MIN(MAX(event.xbutton.x, 0),255);
	firstBuildY = MIN(MAX(event.xbutton.y, 0),255);
	
	// Save the old colormap
	pushColormap();
	// Restore histogram
	histogram_activate();
	
	// Show them where they've started
	i= MIN((tryColors-1),MAX((tryColors-firstBuildY),7));
	color_defs[i].pixel = i;
	color_defs[i].red   = redLevel   * PIXELCOLOR;
	color_defs[i].green = greenLevel * PIXELCOLOR;
	color_defs[i].blue  = blueLevel  * PIXELCOLOR;
	color_defs[i].flags = DoRed | DoGreen | DoBlue;
	XAllocColor (dsply, color_map, &color_defs[i]);
	pixels_return[i] = color_defs[i].pixel;
	
	// Make changes show up in histgram window
	// Now draw in image - the mirror image of normal
	i=(firstBuildY-1);
	for (j=0; j<32; j++)
	  XPutPixel(hist_color_image1, j, i, pixels_return[RED]);
	for (j=0; j<calcHistLength(i); j++)
	  XPutPixel(hist_color_image1, j, i, pixels_return[WHITE]);
	
	if(XtIsManaged(colorScreenW))
	  XPutImage(dsply, hist_color_window1, hist_color_gc, 
		    hist_color_image1, 0, 0,
		    0, 0,  32,  256);
	XFlush (dsply);
	
      }     
      else if(event.xbutton.button == Button2) { // Pop this puppy
	// They want to undo
	popColormap();
      }
      break;
      
    case KeyPress:
      ans= XLookupKeysym((XKeyEvent *)&event, 0);
      printf ("Keyed[%c]\n",ans);
      break;
      
    default:
      break;
    }
  }
  else if(colorType==USEFITTO) { // Try and make a better contrast
    event= *(draw_data->event);
    
    switch (event.type)  {
      
    case ButtonRelease:
      // They are done picking points they want
      // in the range for the new contrast colormap
      
      if(event.xbutton.button == Button1) {
	int min, max;
	float scaleFactor;
	
	// This y is the last in the range
	lastY = event.xbutton.y;
	
	// Map this to colormap space
	firstY = 255 - firstY;
	// We display from bottom up
	lastY = 255 - lastY;      
	firstY = MIN((255-1),MAX(firstY,7));
	lastY = MIN((255-1),MAX(lastY,7));
	if(firstY > lastY) { // switch them
	  int temp;
	  temp = lastY;
	  lastY= firstY;
	  firstY= temp;
	}
	
	// Now that we know the range let's expand our 
	// colormap accordingly
	
	min = MAX(firstY,7);
	max = MIN(lastY,255);
	
	if(min==max)
	  max++;
	scaleFactor = 255.0 / (max-min);  
	// scaleFactor = (max-min) /  256.0 ;
	
	// Make these entries in the colormap GRAYSCALE
	for (i=min; i<=max; i++) { // Do the colors we have been assigned
	  color_defs[i].pixel = i;
	  color_defs[i].red   = (i-min) * scaleFactor * PIXELCOLOR;
	  color_defs[i].green = (i-min) * scaleFactor * PIXELCOLOR;
	  color_defs[i].blue  = (i-min) * scaleFactor * PIXELCOLOR;
	  color_defs[i].flags = DoRed | DoGreen | DoBlue;
	  XAllocColor (dsply, color_map, &color_defs[i]);
	  pixels_return[i] = color_defs[i].pixel;
	}
	
	if(fileType==PACKEDPORTABLE)
	  stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
			   packImg,purdueHeader.nimg, imgNum);
	else if(fileType==MAP)
	  stuffMapImage(map_image_buffer, float_image_buffer1, 
			file_width, file_height, nsecs, imgNum, 0);
	else
	  virtual_read_activate(-99,-99);
	DrawWindow();
	histogram_activate();
	XFlush (dsply);
      }
      break;
      
    case ButtonPress:
      if(event.xbutton.button == Button1) {
	// This is the first point in the range for the new colormap
//	firstX = MIN(MAX(event.xbutton.x, 0),255);
	firstY = MIN(MAX(event.xbutton.y, 0),255);
	pushColormap();                   // Save the old colormap
	histogram_activate(); // Restore histogram
      }     
      else if(event.xbutton.button == Button2) { // Pop this puppy
	popColormap(); // They've changed their minds
      }
      break;
      
    case KeyPress:
      ans= XLookupKeysym((XKeyEvent *)&event, 0);
      printf ("Keyed[%c]\n",ans);
      break;
      
    default:
      break;
    }
  }
  
  XFlush(dsply);   // Make changes current
  
} // end function colorClick



void colormapMotion(Widget widget, caddr_t unused, 
		    register XMotionEvent *pEvent)
{
  // function name: colormapMotion

  // Whenever the pointer moves across our Alter Colormap drawing
  // area(s) call here. This rotuine basically takes care of drawing
  // that little black line you see when the cursor is dragged across
  // the colormap areas.

  static short lastRed, lastGreen, lastBlue;
  int y,lastDraw;
  int i, j, thisI;
  static int lastY= -999;
  static int lastI= -999;

  XFlush(dsply);  // Make all changes current

  switch (pEvent->type)  {
	  
    case LeaveNotify:
      // They've been here before redraw the old position
      // in the "normal" way so that we only see the current
      // position exhibited on the screen.
      if(lastY > 0) {
	// Let's redisplay the OLD position
	// Make changes show up in histgram window
	thisI = lastY;
	// Redraw the histogram window
	for (j=0; j<32; j++)
	  XPutPixel(hist_color_image1, j, thisI, pixels_return[WHITE]);
	lastDraw = calcHistLength(thisI);
	for(j=0; j<lastDraw; j++)
	  XPutPixel(hist_color_image1, j, thisI, pixels_return[RED]);
	
	// Redraw the colormap window
	for(j=0; j<32; j++)
	  XPutPixel(color_image1, j, thisI, pixels_return[256-thisI]);
	
	// Put them on the screen 
	XPutImage(dsply, hist_color_window1, hist_color_gc, 
		  hist_color_image1, 0, lastY,
		  0, lastY,  32,  1);
	XPutImage(dsply, color_window1, color_gc, 
		  color_image1, 0, lastY,
		  0, lastY,  32,  1);
      }
      if(lastI > 0 && !buildingCmap) { // Redo last ones
	color_defs[lastI].red   = lastRed;
	color_defs[lastI].green = lastGreen;
	color_defs[lastI].blue  = lastBlue;
	XAllocColor (dsply, color_map, &color_defs[lastI]);
	pixels_return[lastI] = color_defs[lastI].pixel;
      }
      lastY= -999;
      lastI= -999;
      break;
      
  case MotionNotify:
    // Get coords of pointer
//    x = pEvent->x;     
    y = pEvent->y;
    
    // Show them where they've started
    y = MAX(y,0);
    i= MIN(tryColors,MAX((tryColors-y),7));
    i= MIN(255,MAX((255-y),7));
    
    // Tell them what position in the colormap we are
    sprintf(outStr,"%3d",i);
    doLabel(outStr, colorValueW[ENTRY], FALSE);

    if(radialMapShade) { // Tell them what radius we are
      if(i<=pixels_return[7])
	sprintf(outStr,"%3d",0);
      else 
	sprintf(outStr,"%.1f",radialValue[i]);

      doLabel(outStr, radialValueW, FALSE);
    }

    // They've been here before redraw the old position
    // in the "normal" way so that we only see the current
    // position exhibited on the screen.
    if(lastY > 0) { // Let's redisplay the OLD position
      // Make changes show up in histgram window
      thisI=(lastY);
      // Redraw the histogram window
      for (j=0; j<32; j++)
	XPutPixel(hist_color_image1, j, thisI, pixels_return[WHITE]);
      lastDraw = calcHistLength(thisI);
      for (j=0; j<lastDraw; j++)
	XPutPixel(hist_color_image1, j, thisI, pixels_return[RED]);
      // Redraw the colromap window
      for (j=0; j<32; j++)
	XPutPixel(color_image1, j, thisI, pixels_return[256-thisI]);
      // Put them on the screen
      XPutImage(dsply, hist_color_window1, hist_color_gc, 
		hist_color_image1, 0, lastY,
		0, lastY,  32,  1);
      XPutImage(dsply, color_window1, color_gc, 
		color_image1, 0, lastY,
		0, lastY,  32,  1);
    }
    
    if(y > 0) { // Let's display the NEW position
      // Make changes show up in histgram window
      thisI=(y * 32);
      for (j=0; j<32; j++) { // Show them where it's at
	XPutPixel(hist_color_image1, j, y, pixels_return[BLACK]);
	XPutPixel(color_image1, j, y, pixels_return[BLACK]);
      }
      
      // Put changes on the screen
      XPutImage(dsply, hist_color_window1, hist_color_gc, 
		hist_color_image1, 0, y,
		0, y,  32,  1);
      XPutImage(dsply, color_window1, color_gc, 
		color_image1, 0, y,
		0, y,  32,  1);
      
      // Remeber what we've done for next time
      lastY= y;
      
      if(lastI > 0 && !buildingCmap) { // Redo last ones
	color_defs[lastI].red   = lastRed;
	color_defs[lastI].green = lastGreen;
	color_defs[lastI].blue  = lastBlue;
	XAllocColor (dsply, color_map, &color_defs[lastI]);
	pixels_return[lastI] = color_defs[lastI].pixel;
      }
      lastI = i;
      lastRed = color_defs[i].red;
      lastGreen = color_defs[i].green;
      lastBlue = color_defs[i].blue;
      color_defs[i].red    = 65000;
      color_defs[i].green  = 0;
      color_defs[i].blue   = 0;
      XAllocColor (dsply, color_map, &color_defs[i]);
      pixels_return[i] = color_defs[i].pixel;
      buildingCmap = FALSE;  // Done by now
    }
    break;
    
  default:
    break;
  }
  
  XFlush(dsply);  // Make changes current
  
} // end function colormapMotion



void contrast_enhance(int min, int max)
{
  // function name: contrast_enhance
  // Try and stretch the colormap so that we are looking at the
  // maximum contrast across the range of pixels of the image.

  int i;
  float scaleFactor;
  
  // Now that we know the range let's expand our colormap accordingly
  
  min = MAX(min, 7);
  max = MIN(max, 255);
  max = MAX(max, min+1);

  scaleFactor = 255.0 / (max-min);  
  //      scaleFactor = ((float)(max-min))  /  256.0 ;
  
  // Make these entries in the colormap GRAYSCALE
  for (i=min; i<=max; i++) { // Do the colors we have been assigned
    color_defs[i].pixel = i;
    color_defs[i].red   = (i-min) * scaleFactor * PIXELCOLOR;
    color_defs[i].green = (i-min) * scaleFactor * PIXELCOLOR;
    color_defs[i].blue  = (i-min) * scaleFactor * PIXELCOLOR;
    color_defs[i].flags = DoRed | DoGreen | DoBlue;
    XAllocColor (dsply, color_map, &color_defs[i]);
    pixels_return[i] = color_defs[i].pixel;
  }
  
} // end function contrast_enhance



void toggle_change(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: toggle_change
  // Whenever the toggle state is changed call here. Callback for the
  // colormap mode changes and also the lock on the main screen.

  int i, y, x, p, min, max;
  float factor;
  XmString simpleString;
  int tag = *((int *) client_data);

  colorType = tag;

  if(tag==USEGENERIC) { // Use the GENERIC colormap model
    for (i=7; i<tryColors; i++) { // Do the colors we have been assigned
      color_defs[i] = gen_color_defs[i]; // struct assignment
      XAllocColor (dsply, color_map, &color_defs[i]);
      pixels_return[i] = color_defs[i].pixel;
    }
  }
  else if(tag==USEREVERSE) { // Use the REVERSE Contrast colormap model
    for (i=7; i<tryColors; i++) { // Do the colors we have been assigned
      color_defs[i] = gen_color_defs[(tryColors-i)+7]; // struct assignment
      color_defs[i].pixel = i;
      XAllocColor (dsply, color_map, &color_defs[i]);
      pixels_return[i] = color_defs[i].pixel;
    }
  }
  else if(tag==USEFITTO) { // They want to fit colormap to image
    // First calculate min and max
    min= 65500;
    max= 0;
    for (y=0; y<image_height; y++) {
      for (x=0; x<image_width; x++) { // Visit each pixel
	p = image_buffer1[(y*buffer_width)+x];
	min = MIN(p, min);
	max = MAX(p, max);
      }
    }
    contrast_enhance(min, max);    
  }
  else if(tag==USERANDOM) { // They want a randomized colormap
    
    // Get our fraction of the random number
    factor= 1.0 / RAND_MAX;
    
    // Make these entries in the colormap GRAYSCALE
    for (i=7; i<tryColors; i++) { // Do the colors we have been assigned
      color_defs[i].pixel = i;
      
      color_defs[i].red   = (int)(random()*factor*255) * PIXELCOLOR;
      color_defs[i].green = (int)(random()*factor*255) * PIXELCOLOR;
      color_defs[i].blue  = (int)(random()*factor*255) * PIXELCOLOR;
      
      color_defs[i].flags = DoRed | DoGreen | DoBlue;

      XAllocColor (dsply, color_map, &color_defs[i]);
      pixels_return[i] = color_defs[i].pixel;
    } 
  }
  else if(tag==USEBUILD) // They want to build their own colormap
    pushColormap();
  else if(tag==READCM || tag==WRITECM) { // read or write colormap
    readWrite= tag;
    filterFiles("*.cmp", tag); 
    raiseIt(fileSelW);
  }
  else if(tag==STEPCM) { // They want to step the colormap
  }
  else if(tag==USELOCK) { // They want to lock their Box size
    locked = !locked;
    // Set the imgNumTextW to (un)editable
    XtVaSetValues(imgNumTextW, XmNeditable, !(locked), NULL);
    
    // Must reset label of the toggle to reflect whether we're
    // specifying the first box or in find box mode.  Also reset the
    // text widget for either the firstBox value or the find Box value
    
    if(locked)
      simpleString = XmStringCreateLocalized("Find Box:");
    else {
      simpleString = XmStringCreateLocalized("FirstBox:");
      findBox= -99; // Start again
    }
    
    XtVaSetValues(boxInitLabelW, XmNlabelString, simpleString, NULL);
    XFlush(dsply);
    XmStringFree(simpleString); 
    TextSetInt(boxInitW, lastNumBox+1, FALSE);
    
    DrawWindow();
  }
  else if(tag==NEWXY) { // find new x,y center
    newXY = !newXY;
    if(newXY) { // Set new xy cursor
      if(boxZoom==USEZOOM)
	old_newXY_cursor = zoomCursor;
      else if(boxZoom==USEBOX)
	old_newXY_cursor = cross_cursor;
      else if(boxZoom==USEPOINT)
	old_newXY_cursor = None;
      
      // Put up newXY cursor
      old_cursor = bogosity_cursor;
      updateCursor(drawingW, bogosity_cursor);
    }
    else { // Restore old cursor
      updateCursor(drawingW, old_newXY_cursor);
    }
  }
  else if(tag==PARAMPROJ) { // They want map projections
    mapProj= !mapProj;
    if(mapProj && EQU(paramMapFilename,"")) {
      mapProj= FALSE;
      XtVaSetValues(mapProjW, XmNset, FALSE, NULL);
      askThem("No MAP open can't do projection.", XmDIALOG_INFORMATION);
    }
    else { // Do it
      if(mapProj) {
/*     askThem(
        "MAP projections may be 90 degrees rotated in omega.\n\
        We're working on getting this resolved.",
        FALSE,XmDIALOG_INFORMATION);
*/	
	if(haveParamFile)
	  stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
			   packImg, purdueHeader.nimg, imgNum);
	else
	  askThem("Must read a param file first.", XmDIALOG_INFORMATION); 
      }
      DrawWindow(); 
    }
  }
  else if(tag==PROJCONTOUR) { // They want map projections contour
    mapProjContour= !mapProjContour;
    
    if(EQU(paramMapFilename,"")) {
      askThem("No MAP open can't contour the projection.", 
	      XmDIALOG_INFORMATION);
    }
    else { // Do it
      stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		       packImg,purdueHeader.nimg, imgNum);
      DrawWindow(); 
    }
  }
  else if(tag==SAVESECOND)
    saveSecondMap = !saveSecondMap;
  else if(tag==FOURAVERSF)
    writeFourAverSF = !writeFourAverSF;
  else if(tag==ICOSAQUARTER)
    icosaQuarter = !icosaQuarter;
  else if(tag==REFAUTOMATE)
    refAutomate = !refAutomate;
  else if(tag==REFCIRCAVG) {
    refCircAvg = !refCircAvg;
    if(refCircAvg) { // Force dark cirlce off
      darkCircleRef = FALSE;
      XtVaSetValues(refDarkTW, XmNset, FALSE, NULL);
      XtVaSetValues(refAllW, XmNsensitive, TRUE, NULL);
      XFlush(dsply);
    }
  }			       
  else if(tag==SHOWLOWLIM) // show lowlim?
    showLowLim = !showLowLim;
  else if(tag==REFREADFILE) { // ref boxing read file
    refReadFile = !refReadFile;
    if(refReadFile) { // get fname
      readType= REFREADFILE;
      readWrite= READING;
      raiseIt(fileSelW);
    }
    else { // no file
      doLabel("No file.", refReadLW, TRUE);
    }
  }
  else if(tag==USEAVGCORAVG) { // use the avg for coravg?
    int i;
    useAvgCorAvg = !useAvgCorAvg;
    if(useAvgCorAvg)
      for(i=0; i< emcoravgNcol*emcoravgNrow; i++)
	emcoravg_image_buffer1[i] = (BYTE)emcoravg_image_buffer_avg[i];
    else
      for(i=0; i< emcoravgNcol*emcoravgNrow; i++)
	emcoravg_image_buffer1[i] = (BYTE)emcoravg_image_buffer_disk[i];
    stuffEmcoravgImage();
  }
  else if(tag==SHOWRADII) {
    showRadii = !showRadii;
    if(diffSection)
      diffSec_really_doit();
  }
  else if(tag==THRESHDOIT) {
    threshDoit = !threshDoit;
    if(diffSection)
      diffSec_really_doit();
  }
  else if(tag==THRESHABS) {
    threshAbs = !threshAbs;
    if(diffSection)
      diffSec_really_doit();
  }
  else if(tag==SHIFTORIGX) {
    deltaPhaseShift = !deltaPhaseShift;
    if(deltaPhaseShift) { // Only allow delta
      XtSetArg(arglst[0], XmNsensitive, TRUE );
      XtSetValues(shiftOrigXW, arglst, 1); 
      XtSetValues(shiftOrigYW, arglst, 1); 
      XtSetValues(shiftOrigXlabelW, arglst, 1); 
      XtSetValues(shiftOrigYlabelW, arglst, 1); 
      XtSetArg(arglst[0], XmNsensitive, FALSE );
      XtSetValues(moveOrigXW, arglst, 1); 
      XtSetValues(moveOrigYW, arglst, 1); 
      XtSetValues(moveOrigXlabelW, arglst, 1); 
      XtSetValues(moveOrigYlabelW, arglst, 1); 
      XtSetValues(phasYdownW, arglst, 1); 
      XtSetValues(phasYupW, arglst, 1); 
      XtSetValues(phasXdownW, arglst, 1); 
      XtSetValues(phasXupW, arglst, 1); 
    }
    else { // Only allow change of real value
      XtSetArg(arglst[0], XmNsensitive, FALSE );
      XtSetValues(shiftOrigXW, arglst, 1); 
      XtSetValues(shiftOrigYW, arglst, 1); 
      XtSetValues(shiftOrigXlabelW, arglst, 1); 
      XtSetValues(shiftOrigYlabelW, arglst, 1); 
      XtSetArg(arglst[0], XmNsensitive, TRUE );
      XtSetValues(moveOrigXW, arglst, 1); 
      XtSetValues(moveOrigYW, arglst, 1); 
      XtSetValues(moveOrigXlabelW, arglst, 1); 
      XtSetValues(moveOrigYlabelW, arglst, 1); 
      XtSetValues(phasYdownW, arglst, 1); 
      XtSetValues(phasYupW, arglst, 1); 
      XtSetValues(phasXdownW, arglst, 1); 
      XtSetValues(phasXupW, arglst, 1); 
    }
  }
  
  if(tag==USEGENERIC || tag==USEREVERSE || tag==USEFITTO || 
     tag==USERANDOM  || tag==USEBUILD) {
    if(fileType==PACKEDPORTABLE)
      stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		       packImg,purdueHeader.nimg, imgNum);
    else if(fileType==MAP)
      stuffMapImage(map_image_buffer, float_image_buffer1, 
		    file_width, file_height, nsecs, imgNum, 0);
    else
      virtual_read_activate(-99,-99);
    DrawWindow();
    histogram_activate();
  }
  
  XFlush (dsply);
  
} // end function toggle_change



void reset_xyz_scrolls()
{
  // function name: reset_xyz_scrolls
  // Reset the x,y,z min and max scrolls on the 3D rendering screen.
  
  int i, thisMin, thisMax, thisDimension;
  
  for (i=0; i<3; i++) { // Do cut scrolls
    if(i==0)
      thisDimension = numCols;
    else if (i==1)
      thisDimension = numRows;
    else 
      thisDimension = origNumSecs;
    
    thisMin = 0;
    thisMax = thisDimension;
    if(lowRes) { // setup in case they've got other values
      if(i==2 && cutMin2Save>=0.0)
	thisMin = cutMin2Save;
      if(i==2 && cutMax2Save>=0.0)
	thisMax = cutMax2Save;
      if(i==1 && cutMin1Save>=0.0)
	thisMin = cutMin1Save;
      if(i==1 && cutMax1Save>=0.0)
	thisMax = cutMax1Save;
      if(i==0 && cutMin0Save>=0.0)
	thisMin = cutMin0Save;
      if(i==0 && cutMax0Save>=0.0)
	thisMax = cutMax0Save;
    }
    
    XtSetArg(arglst[0], XmNmaximum, thisDimension+4);
    XtSetArg(arglst[1], XmNminimum, 0);
    XtSetArg(arglst[2], XmNsliderSize, 5);
    XtSetArg(arglst[3], XmNvalue, thisMin);
    XtSetArg(arglst[4], XmNpageIncrement, 10);
    XtSetValues(cutMinScrollW[i], arglst, 5);
    XFlush(dsply);

    sprintf(outStr,"%d",thisMin+1);
    doLabel(outStr, cutMinValueW[i], DEFAULT);

    XtSetArg(arglst[0], XmNmaximum, thisDimension+6);
    XtSetArg(arglst[1], XmNminimum, 0);
    XtSetArg(arglst[2], XmNsliderSize, 5);
    XtSetArg(arglst[3], XmNvalue, (int)(thisMax));
    XtSetArg(arglst[4], XmNpageIncrement, 10);
    XtSetValues(cutMaxScrollW[i], arglst, 5);
    XFlush(dsply);

    sprintf(outStr,"%d",thisMax);
    doLabel(outStr, cutMaxValueW[i], DEFAULT);
    
    cutMinValues[i] = thisMin;
    cutMaxValues[i] = thisMax;
    perMinValues[i] = (float)thisMin / (float)thisDimension;
    perMaxValues[i] = (float)thisMax / (float)thisDimension;
    
    // Now do percent label
    sprintf(outStr,"%.1f%% to %.1f%%",      
	    perMinValues[i]*100.0, perMaxValues[i]*100.0);
    doLabel(outStr, perValueW[i], DEFAULT);
  }
  
} // end function reset_xyz_scrolls



void boxMode_change(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: boxMode_change
  // They want to change the boxing mode. Setup for
  // circle/rectangle/polygon or the way they want to have the
  // background done.

  int tag = *((int *) client_data);

  if(tag==USESQUARE || tag==USERECT || tag==USECIRCLE || tag==USEPOLY)
      boxMethod = tag;
  else if(tag==USEMAPSHADE) {
    secondMapShade = !secondMapShade;

    if(secondMapShade)
      XtSetArg(arglst[0], XmNsensitive, TRUE );
    else
      XtSetArg(arglst[0], XmNsensitive, FALSE );

    XtSetValues(fractScrollW, arglst, 1); 
    XtSetValues(angxScrollW, arglst, 1); 
    XtSetValues(angyScrollW, arglst, 1); 
    XtSetValues(fractLabelW, arglst, 1); 
    XtSetValues(angxLabelW, arglst, 1); 
    XtSetValues(angyLabelW, arglst, 1); 
    XtSetValues(fractValueW, arglst, 1); 
    XtSetValues(angxValueW, arglst, 1); 
    XtSetValues(angyValueW, arglst, 1); 
    XtSetValues(fractToggleW, arglst, 1); 
    XtSetValues(angxToggleW, arglst, 1); 
    XtSetValues(angyToggleW, arglst, 1); 
    
    // Turn radial Projections OFF
    radialProjections = FALSE;

    // Now reset toggles
    XtVaSetValues(radProjShadeW, XmNset, FALSE, NULL);
    XtUnmanageChild(icosRadProjToggleW);
    XtUnmanageChild(radProjToggleW);
    XtUnmanageChild(radProjCSToggleW);
    XtManageChild(radialCSToggleW);
    XtManageChild(radialTextW);
    XtManageChild(radialLabelW);
    XFlush(dsply);
    
    // Now activate the surface density widget
    XtSetArg(arglst[0], XmNsensitive, TRUE );
    XtSetValues(surfDensScrollW, arglst, 1); 
    XtSetValues(surfDensLabelW, arglst, 1); 
    XtSetValues(surfDensValueW, arglst, 1); 
    XtSetValues(surfDensToggleW, arglst, 1); 
    
  }
  else if(tag==USEIMMEDIATE)
    immediateMapShade = !immediateMapShade;
  else if(tag==MBUSEIMMEDIATE)
    mbImmediate = !mbImmediate;
  else if(tag==ICOSREPL) {
    isIcosRepl = !isIcosRepl;
    widgetList[0]  = modelYLW;
    widgetList[1]  = modelZLW;
    widgetList[2]  = mbCenteryScrollW;
    widgetList[3]  = mbCenteryTextW;
    widgetList[4]  = mbCenterzScrollW;
    widgetList[5]  = mbCenterzTextW;
    widgetList[6]  = mbRotatetScrollW;
    widgetList[7]  = mbRotatetTextW;
    widgetList[8]  = mbRotatepScrollW;
    widgetList[9]  = mbRotatepTextW;
    widgetList[10] = mbRotateoScrollW;
    widgetList[11] = mbRotateoTextW;
    if(isIcosRepl) { // replicate on the icosahedral 5-fold
      doLabel("Radius    ", modelXLW, FALSE);
      XtUnmanageChildren(widgetList, 12);
    }
    else { // normal
      doLabel("x         ", modelXLW, FALSE);
      XtManageChildren(widgetList, 12);
    }
    XFlush(dsply);
  }
  else if(tag==USESTEREO) // They want to use stereo mapping
    stereoMapShade = !stereoMapShade;
  else if(tag==LOHIRES) { // They want to use loHiResoltuion
    float thisPixSize;
    
    if(firstTimeLoHiRes) { // init
      calcSF_activate(NULL,NULL, NULL);
      firstTimeLoHiRes = FALSE;
    }
    
    lowRes = !lowRes;
    
    // Always reset the sampling in case they haven't
    TextSet(sampleRenderW, "1", FALSE); 
    if(globFract < 0.75) { // set shading fraction
      globFract = 0.75;
      sprintf(outStr,"%.2f",globFract);
      TextSet(fractValueW, outStr, FALSE); 
      XtSetArg(arglst[0], XmNvalue, (int)(globFract*1000.0));
      XtSetValues(fractScrollW, arglst, 1);
      XFlush(dsply);
    }
    
    if(lowRes) { // Give feedback
      
      if(!immediateResource) { // Turn immediate update ON
	XtVaSetValues(immediateUpdateW, XmNset, TRUE, NULL);
	immediateMapShade = TRUE;
	XFlush(dsply);
      }
      
      numCols = lowResCols;
      numRows = lowResRows;
      origNumSecs= numSecs = lowResSecs;
      
      fileBinSecs= numSecs;
      fileBinRows= numRows;
      fileBinCols= numCols;
      
      xCen = numCols/2;
      yCen = numRows/2;
      zCen = origNumSecs/2;
      
      largestDim = MAX(numCols, numRows);
      largestDim = MAX(largestDim, origNumSecs);
      
      // We've got to make this larger for asymmetric maps 
      // and all maps that have information "outside" of the
      // normal sphere of our normal maps,
      // or else we just keep cliping the box into a sphere
      
      globMaxRadius = (int)(origNumSecs/2) - 1;
      globMinRadius = 0.0;
      
      // Set scrolls on depthcue screen  and give feedback
      sprintf(outStr,"%.0f",globMinRadius);
      doValue(outStr, minRadiusValueW, FALSE);
      
      sprintf(outStr,"%.0f",globMaxRadius);
      doValue(outStr, maxRadiusValueW, FALSE);
      
      // Radius is origNumSecs/2
      XtVaSetValues(maxRadiusScrollW, XmNsliderSize, 5,NULL);
      XtVaSetValues(maxRadiusScrollW, XmNmaximum, 
		    (int)((globMaxRadius)+5),NULL);
      XtVaSetValues(maxRadiusScrollW, XmNvalue, (int)(globMaxRadius),NULL);
      
      XtVaSetValues(minRadiusScrollW, XmNsliderSize, 5,NULL);
      XtVaSetValues(minRadiusScrollW, XmNmaximum, 
		    (int)((globMaxRadius)+5),NULL);
      XtVaSetValues(minRadiusScrollW, XmNvalue, (int)(globMinRadius),NULL);
      
      if(whichStep==PIXELSIZE) { // We can't allow this for lowres
	XtSetArg(arglst[0], XmNset, FALSE );
	whichStep = NONE;
	TextSet(frameTextW, "", FALSE); 
	doLabel("NO Step:", stepValueLabelW, FALSE);
      }
      sprintf(outStr,"Low Resolution ON");
      XtVaSetValues(pixelSizeScrollW, XmNsensitive, FALSE, NULL);
      XtVaSetValues(pixelSizeValueW, XmNsensitive, FALSE, NULL);
      
      reset_xyz_scrolls();
      
    }
    else { // NOT lowres Give feedback
      
      // Make them recalc any new hires buffer
      pixelSizeSave = -99.0;
      
      // Save where they cut for use when generating the HiRes buffer
      cutMin0Save  = cutMinValues[0];
      cutMax0Save  = cutMaxValues[0];
      cutMin1Save  = cutMinValues[1];
      cutMax1Save  = cutMaxValues[1];
      cutMin2Save  = cutMinValues[2];
      cutMax2Save  = cutMaxValues[2];
      
      // Turn immediate update OFF
      XtVaSetValues(immediateUpdateW, XmNset, FALSE, NULL); 
      immediateMapShade = FALSE;
      XFlush(dsply);
      
      // Set the radius ridiculously high so they can
      // see all of it and then adjust as they need
      
      sprintf(outStr,"%.0f",(float)lowResSecs * 3.0);
      doValue(outStr, maxRadiusValueW, FALSE);
      
      globMaxRadius= lowResSecs * 3.0;
      
      XtVaSetValues(maxRadiusScrollW,XmNsliderSize, 5,NULL);
      XtVaSetValues(maxRadiusScrollW, XmNmaximum, (int)((lowResSecs*4)+5),NULL);
      XtVaSetValues(maxRadiusScrollW,XmNvalue, (int)(lowResSecs*3), NULL);
      
      sprintf(outStr,"High Resolution ON");
      XtSetArg(arglst[0], XmNsensitive, TRUE );
      XtSetValues(pixelSizeScrollW, arglst, 1); 
      XtSetValues(pixelSizeValueW, arglst, 1); 
    }
    
    // WHy are we doing this - it's for the movie???
    // XtSetValues(pixelSizeToggleW, arglst, 1); 
    
    // Now setup the pixelSize
    if(!lowRes) { // Setup for High Res
      thisPixSize= origWidthFromFile/lowResCols;
      XtSetArg(arglst[0], XmNvalue, (int)(thisPixSize*1000.0));
      XtSetValues(pixelSizeScrollW, arglst, 1);
      
      sprintf(outStr,"%.2f",thisPixSize);
      TextSet(pixelSizeValueW, outStr, TRUE); 
    }
    
  }
  else if(tag==USERADIAL) { // They want to use radial mapping
    radialMapShade = !radialMapShade;
    
    // Turn radial Projections OFF
    radialProjections = FALSE;

    // Now reset toggles
    XtVaSetValues(radProjShadeW, XmNset, FALSE, NULL);
    XtUnmanageChild(icosRadProjToggleW);
    XtUnmanageChild(radProjToggleW);
    XtUnmanageChild(radProjCSToggleW);
    XtManageChild(radialCSToggleW);
    XtManageChild(radialTextW);
    XtManageChild(radialLabelW);
    XFlush(dsply);
    
    // Now activate the surface density widget
    XtVaSetValues(surfDensScrollW, XmNsensitive, TRUE, NULL);
    XtVaSetValues(surfDensLabelW, XmNsensitive, TRUE, NULL);
    XtVaSetValues(surfDensValueW, XmNsensitive, TRUE, NULL);
    XtVaSetValues(surfDensToggleW, XmNsensitive, TRUE, NULL);
    
  }
  else if(tag==USERADPROJ) { // They want to use radial projections
    radialProjections = !radialProjections;
    if(radialProjections) { // Give feedback
      
      // Turn radial and shading off
      radialMapShade = FALSE;
      
      secondMapShade = FALSE;
      XtSetArg(arglst[0], XmNsensitive, FALSE );
      XtSetValues(fractScrollW, arglst, 1); 
      XtSetValues(angxScrollW, arglst, 1); 
      XtSetValues(angyScrollW, arglst, 1); 
      XtSetValues(fractLabelW, arglst, 1); 
      XtSetValues(angxLabelW, arglst, 1); 
      XtSetValues(angyLabelW, arglst, 1); 
      XtSetValues(fractValueW, arglst, 1); 
      XtSetValues(angxValueW, arglst, 1); 
      XtSetValues(angyValueW, arglst, 1); 
      XtSetValues(fractToggleW, arglst, 1); 
      XtSetValues(angxToggleW, arglst, 1); 
      XtSetValues(angyToggleW, arglst, 1); 
      
      // Now reset toggles
      XtVaSetValues(secondShadeW, XmNset, FALSE, NULL);
      XtVaSetValues(radialShadeW, XmNset, FALSE, NULL);
      
      // Now setup cut scrolls and values widgets
      reset_xyz_scrolls();
      
      // Now for radproj feedback
      XtManageChild(icosRadProjToggleW);
      XtManageChild(radProjToggleW);
      XtManageChild(radProjCSToggleW);
      XtUnmanageChild(radialCSToggleW);
      XtUnmanageChild(radialTextW);
      XtUnmanageChild(radialLabelW);
      XFlush(dsply);
      
      // Now DEactivate the surface density widget
      XtSetArg(arglst[0], XmNsensitive, FALSE );
      XtSetValues(surfDensScrollW, arglst, 1); 
      XtSetValues(surfDensLabelW, arglst, 1); 
      XtSetValues(surfDensValueW, arglst, 1); 
      XtSetValues(surfDensToggleW, arglst, 1); 
      XFlush(dsply);
      
    }
    else { // Give feedback
      XtUnmanageChild(icosRadProjToggleW);
      XtUnmanageChild(radProjToggleW);
      XtUnmanageChild(radProjCSToggleW);
      XtManageChild(radialCSToggleW);
      XtManageChild(radialTextW);
      XtManageChild(radialLabelW);
      XFlush(dsply);
      // Now activate the surface density widget
      XtSetArg(arglst[0], XmNsensitive, TRUE );
      XtSetValues(surfDensScrollW, arglst, 1); 
      XtSetValues(surfDensLabelW, arglst, 1); 
      XtSetValues(surfDensValueW, arglst, 1); 
      XtSetValues(surfDensToggleW, arglst, 1);  
    }
  }
  else if(tag==USEAVGPIXEL) { // They want to use average pixel bkgrnd
    bkgrndMethod = tag;
    
    // Redraw this puppy
    if(fileType==PACKEDPORTABLE)
      stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
    else 
      stuffDisplayBoxBuffer(curBox,image_fileName.addr,editBoxZoom);
    DrawEditBoxWindow(NULL, NULL, NULL); 
  }
  else if(tag==USERANDPIXEL) { // They want to use random pixel bkgrnd
    bkgrndMethod = tag;
    
    // Redraw this puppy
    if(fileType==PACKEDPORTABLE)
      stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
    else 
      stuffDisplayBoxBuffer(curBox,image_fileName.addr,editBoxZoom);
    DrawEditBoxWindow(NULL, NULL, NULL);
    
  }
  else if(tag==VOLUME) // They want to see what volume they are at
    volumeDisp = !volumeDisp;
  
} // end function boxMode_change



void histMode_change(Widget widget, XtPointer client_data, 
		     XtPointer call_data)
{
  // function name: histMode_change
  // They want to toggle whether they see a histogram or not 

  static int oldTag = 0;
  int tag = *((int *) client_data);

  if(tag==USEHIST) { // They want to see histogram
    histMethod = !histMethod;
  }
  else if(tag != oldTag) { // Don't do this twice
    if(tag==NOCONTRAST) { // No contrast stretch
      contrastStretch = FALSE;
      histoEqualize = FALSE;
    }
    else if(tag==USECONTRAST) { // Contrast stretch
      contrastStretch = TRUE;
      histoEqualize = FALSE;
    }
    else if(tag==USEHISTEQUAL) { // Use histogram equalization
      contrastStretch = FALSE;
      histoEqualize = TRUE;
    }

    if(tag==NOCONTRAST || tag==USECONTRAST || tag==USEHISTEQUAL) {
      if(fileType != PACKEDPORTABLE && fileType != MAP && fileType!=FLOATMAP)
	stuffImage(int_image_buffer1, image_width, image_height, 
		   contrastStretch, FALSE);
      DrawWindow();
      histogram_activate();
    }
  }
  oldTag = tag;
  
} // end function histMode_change



void dispMode_change(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: dispMode_change
  // They want to toggle whether they see the circle number displayed
  // in the middle of their circles or not.
  
  if(widget == boxDispnumW || widget == optionsCirW) {
    if(dispMethod==CIRCLEDISP) {
      dispMethod = FALSE;       // Toggle this guy
      XtSetArg(arglst[0], XmNset, FALSE);
    }
    else { 
      dispMethod= CIRCLEDISP;
      XtSetArg(arglst[0], XmNset, TRUE);
    }
    // We may want to set the toggle in the options screen
    XtSetValues(optionsCirW, arglst, 1);
    DrawWindow();
  }
  else if(widget == optionsBoxW)
    initLastNumBox = !initLastNumBox;
  else if(widget == optionsZoomW)
    autoZoom = !autoZoom;
  else if(widget == optionsFftW)
    fftMotionFeedback = !fftMotionFeedback;
  else if(widget == calcSfsW)
    calcSFs = !calcSFs;
  else if(widget == calcBinW)
    calcBin = !calcBin;
  else if(widget == truePhiW)
    truePhi = !truePhi;
  else if(widget == forceByteW)
    forceByte = !forceByte;
  else if(widget == forceAvgBkgW)
    forceAvgBkg = !forceAvgBkg;
  else if(widget == minmaxColorW)
    minMaxColor = !minMaxColor;
  else if(widget == optionsExW)
    normalExcise = !normalExcise;
  
} // end function dispMode_change



void subMapShadeScreen_activate(Widget widget, XtPointer client_data,
				XtPointer call_data)
{
  // function name: subMapShadeScreen_activate
  // They want to do 3D rendering of their subtracted map.
  
  int i;
  float *tmpSubtract_image_buffer2;
  
  // Compute third buffer
  CK(tmpSubtract_image_buffer2 = (float *)
     malloc(sizeof(float)*sub2binCol*sub2binRow*sub2binSec));
  
  for (i=0; i<sub2binSec*sub2binRow*sub2binCol; i++)
    subtract_image_buffer2[i] = float_image_buffer1[i]; 
  
  raiseIt(statusScreenW);
  
  subtractMap_doit(subtract_image_buffer2,tmpSubtract_image_buffer2);
  free(tmpSubtract_image_buffer2);
  
  XtUnmanageChild(statusScreenW);
  
  // Now shade them
  mapShadeScreen_activate(NULL, NULL, NULL);
  
} // end function subMapShadeScreen_activate



void mapShadeScreen_activate(Widget widget, XtPointer client_data, 
			     XtPointer call_data)
{
  // function name: mapShadeScreen_activate
  // They want to see their map sections displayed as a depth-cue
  // shade 3D surface.

  minUnblob = secMinPixel;
  maxUnblob =  -surfaceDensity;
  
  // Divide by 100 for percentage RSS - just setting to zero?
  unblobSize= ((numCols*numRows*origNumSecs) * 
		    (float)((float)0.000/100.0));

  // Now setup the scrolls accordingly
  XtSetArg(arglst[0], XmNmaximum, (numCols/2)+5);
  XtSetArg(arglst[1], XmNminimum, 0);
  XtSetArg(arglst[2], XmNsliderSize, 5);
  XtSetArg(arglst[3], XmNvalue, 0);
  XtSetArg(arglst[4], XmNpageIncrement, 10);
  XtSetValues(unblobRadiusScrollW, arglst, 5);
  XFlush(dsply);
  XtVaSetValues(unblobSizeScrollW, XmNvalue, 0, NULL);
  XFlush(dsply);

  // And the value labels
  doLabel("0", unblobRadiusValueW, DEFAULT);

  TextSet(unblobSizeValueW, "0.000", TRUE); 
  
  // Now setup cut scrolls and values widgets
  reset_xyz_scrolls();

  XtVaSetValues(paintMapButtonW, XmNsensitive, TRUE, NULL); 
  XtVaSetValues(mapMovieButtonW, XmNsensitive, TRUE, NULL); 
  XFlush(dsply);

  firstTimeLoHiRes = TRUE;

  if(!lowRes) { // Setup for High Res
    pixelSizeFactor= numCols/100.0;
    globPixelSize = pixelSizeFactor;
    
    XtSetArg(arglst[0], XmNvalue, (int)(globPixelSize*1000.0));
    XtSetValues(pixelSizeScrollW, arglst, 1);
    
    sprintf(outStr,"%.2f",globPixelSize);
    TextSet(pixelSizeValueW, outStr, TRUE); 
  }
  
  XtManageChild(mapShadeScreenW);
  raiseIt(mapShadeScreenW);

} // end function mapShadeScreen_activate



void updateNewVolume()
{
  // function name: updateNewVolume
  // Update the volume text widget based on what our new surface
  // density is.

  int foundIt;
  int i, j, maxI, thickStep, modStep, foundZero;
  float volume, thisSD, minRad, maxRad, scaleF;
  float *this_float_image_buffer1;
  GC thisGC;

  if(volumeDisp) {
    
    if(unblobSwitch)
      this_float_image_buffer1= unblobBuffer;
    else 
      this_float_image_buffer1= float_image_buffer1;

    // And now let's compute our volume arrays
    
    minRad = globMinRadius;
    maxRad = globMaxRadius;
    
    nbins = (int)(secMaxPixel+0.49999) - (int)(secMinPixel-0.49999);

    if(nbins <= 0) { // Impossible
      printf("Bad nbins - no volume possible.\n");
      return;
    }
    
    while(nbins > 200)
      nbins *= 0.5; 
    
    map_vol_rwa(this_float_image_buffer1, 
		  numCols, numRows, origNumSecs, 
		  secMinPixel, secMaxPixel,
		  volumeFraction, volumeBin, 
		  nbins, volumeCutoff,
		  minRad, maxRad);
    
    thisSD = irev * surfaceDensity;
    foundIt = -99;
    for (i=0; i<(nbins-1); i++) { // Look for this cutoff
      if(thisSD >= volumeCutoff[i] && thisSD < volumeCutoff[i+1]) {
	foundIt = i;
	break;
      }
    }
    
    if(foundIt >= 0) { // OK
      if(irev==-1)
	volume = 100.0 - volumeFraction[i];
      else 
	volume = volumeFraction[i];
      sprintf(outStr,"%.2f %% volume\n%.1E A**3",
	      volume,
	      (numCols*numRows*origNumSecs*pifHeader.pixelSize)*
	      (volume/100.0));
    }
    else  { // not OK
      volume = -99.9;
      outStr[0]='\0';
    }
    
    TextSet(volumeValueW, outStr, TRUE); 
    
    thickStep = 7;  /* default */
    thickStep = (display_width-50) / nbins;
    foundZero = 0;
    modStep   = 20;
    
    maxI=volumeBin[0];
    for (i=0; i<nbins; i++) {
      if(!((0.0 > volumeCutoff[i]) && (0.0 <= volumeCutoff[i+1])))
	maxI = MAX(maxI,volumeBin[i]);
      else
	foundZero = i;
    }
    scaleF = 400.0 / maxI;
    
    sprintf(outStr,"%.0f",volumeCutoff[0]);
    XDrawString(dsply, XtWindow(drawingW), image_gcYELLOW,
		10, display_height-50, outStr, strlen(outStr));
    sprintf(outStr,"%.0f",volumeCutoff[(nbins-1)]);
    XDrawString(dsply, XtWindow(drawingW), image_gcYELLOW,
		10+(thickStep*(nbins-1)), display_height-50, 
		outStr, strlen(outStr));
    
    for (j=0; j<thickStep; j++) { // DO large
      XDrawLine(dsply, XtWindow(drawingW), image_gcYELLOW,
		10+(0*thickStep)+j,display_height-100,
		10+(0*thickStep)+j,display_height-90);
      XDrawLine(dsply, XtWindow(drawingW), image_gcYELLOW,
		10+((nbins-1)*thickStep)+j,display_height-100,
		10+((nbins-1)*thickStep)+j,display_height-90);
    }
    
    for (i=0; i<nbins; i++) { // Do all
      if(irev==-1) { // This way
	if(i < foundIt)
	  thisGC = image_gcRED;
	else if(i==foundIt)
	  thisGC = image_gcBLUE;
	else 
	  thisGC = image_gcYELLOW;
      }
      else { // That way
	if(i > foundIt)
	  thisGC = image_gcRED;
	else if(i==foundIt)
	  thisGC = image_gcBLUE;
	else 
	  thisGC = image_gcYELLOW;
      }
	
      for (j=0; j<thickStep; j++) { // Draw histogram
	XDrawLine(dsply, XtWindow(drawingW), thisGC,
		  10+(i*thickStep)+j,
		  display_height-100,
		  10+(i*thickStep)+j,
		  display_height-((int)(volumeBin[i]*scaleF)+100));
	// DO ticks every modStep
	if((((int)fabs((float)i-(float)foundZero)%modStep)==0)   &&
	   (i > modStep) && (i < (nbins-modStep))&&
	   !((i <= foundZero+1) && (i >= foundZero-1))) { // Tick it
	  XDrawLine(dsply, XtWindow(drawingW), image_gcYELLOW,
		    10+(i*thickStep)+j,display_height-100,
		    10+(i*thickStep)+j,display_height-90);
	  sprintf(outStr,"%.0f",volumeCutoff[i]);
	  XDrawString(dsply, XtWindow(drawingW), image_gcYELLOW,
		      10+(thickStep*i), display_height-50, 
		      outStr, strlen(outStr));
	}
      }

      if(0.0 > volumeCutoff[i] && 0.0 <= volumeCutoff[i+1]) { // Draw the zero
	thisGC = image_gcYELLOW;
	sprintf(outStr,"%.0f",0.0);
	XDrawString(dsply, XtWindow(drawingW), thisGC,
		    10+(thickStep*i), display_height-50, 
		    outStr, strlen(outStr));
	for (j=0; j<thickStep; j++)
	  XDrawLine(dsply, XtWindow(drawingW),  thisGC,
		    10+(i*thickStep)+j, display_height-100,
		    10+(i*thickStep)+j, display_height-90);
      }
      
      XFlush(dsply);
    }
  }
  
} // end function updateNewVolume



void surfDensValueChange(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: surfDensValueChange
  // If they change the value in any of these XmTextWidgets this
  // remembers the value.

  int nFrames;
  float steps;

  // For theta/phi/omega value widgets
  // We've got to decide here if what we've got is XYZ or TPO
  // check the rotateXYZ logical and then call XDY's rtns as appropriate
  // XYZ-TPO
  // Add 360.0 because we start at -360.0 (which is zero for scroll

  if(widget == surfDensValueW) {
    surfaceDensity = wstrtod(surfDensValueW);
    
    // BRANT - make sure we don't have any bogus values here.
    // e.g. surfaceDensity <= 1.0 is undefined... ??? 
    // RWA July 29, 1998
    // OK - irev determines the sign on the surfaceDensity
    // sooo we'll always force it to be positive or 0.0.
    // Display in the text widget based on the irev so we can
    // show them what they are actually looking at when shaded.

    surfaceDensity = fabs(surfaceDensity);
    
    // BRANT - Now set the scroll bar to the value we just got.
    // This DOES NOT cause the scrollbar callback to be called.
    // We need to handle the immediateUpdate and I've done that
    // in surfDens_CR() below.  (Also XtAddActions() above)
    //
    // surfDens scroll done in tenths so times 10

    XtSetArg(arglst[0], XmNvalue, (int)(surfaceDensity*10.0));
    XtSetValues(surfDensScrollW, arglst, 1);
    XFlush(dsply);
    
    maxUnblob = 0.0 - surfaceDensity;
    
    updateNewVolume();
  }
  else if(widget == fractValueW) {
    globFract = wstrtod(fractValueW);
    
    // BRANT - Now set the scroll bar to the value we just got.
    // e.g. It has to be done for all widgets
    // This is times 1000
    XtSetArg(arglst[0], XmNvalue, (int)(globFract*1000.0));
    XtSetValues(fractScrollW, arglst, 1);
    XFlush(dsply);  
  }
  else if(widget == angxValueW) {
    globAngx = wstrtod(angxValueW);
    XtSetArg(arglst[0], XmNvalue, (int)(globAngx));
    XtSetValues(angxScrollW, arglst, 1);
    XFlush(dsply);
  }
  else if(widget == angyValueW) {
    globAngy = wstrtod(angyValueW);
    XtSetArg(arglst[0], XmNvalue, (int)(globAngy));
    XtSetValues(angyScrollW, arglst, 1);
    XFlush(dsply);
    globAngy = 0.0 - globAngy;
  }
  else if(widget == omegaValueW) {
    globOmega = wstrtod(omegaValueW);
    XtSetArg(arglst[0], XmNvalue, (int)(globOmega+360.0));
    XtSetValues(omegaScrollW, arglst, 1);
    XFlush(dsply);
  }
  else if(widget == phiValueW) {
    globPhi = wstrtod(phiValueW);
    XtSetArg(arglst[0], XmNvalue, (int)(globPhi+360.0));
    XtSetValues(phiScrollW, arglst, 1);
    XFlush(dsply);
  }
  else if(widget == thetaValueW) {
    globTheta = wstrtod(thetaValueW);
    XtSetArg(arglst[0], XmNvalue, (int)(globTheta+360.0));
    XtSetValues(thetaScrollW, arglst, 1);
    XFlush(dsply);  
  }
  else if(widget == minRadiusValueW) {
    globMinRadius = wstrtod(minRadiusValueW);
    XtSetArg(arglst[0], XmNvalue, (int)(globMinRadius));
    XtSetValues(minRadiusScrollW, arglst, 1);

    if(pifHeader.pixelSize > 0.0)
      sprintf(outStr, "%.1fA", globMinRadius*pifHeader.pixelSize);
    XtSetArg(arglst[0], XmNlabelString, XmStringCreateLocalized(outStr));
    XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
    XtSetArg(arglst[2], XmNalignment, XmALIGNMENT_BEGINNING);
    XtSetValues(minRadiusAngW, arglst, 3); 
    
    XFlush(dsply);
  }
  else if(widget == maxRadiusValueW) {
    globMaxRadius = wstrtod(maxRadiusValueW);
    XtSetArg(arglst[0], XmNvalue, (int)(globMaxRadius));
    XtSetValues(maxRadiusScrollW, arglst, 1);

    if(pifHeader.pixelSize > 0.0)
      sprintf(outStr, "%.1fA", globMaxRadius*pifHeader.pixelSize);
    XtSetArg(arglst[0], XmNlabelString, XmStringCreateLocalized(outStr));
    XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
    XtSetArg(arglst[2], XmNalignment, XmALIGNMENT_BEGINNING);
    XtSetValues(maxRadiusAngW, arglst, 3); 

    XFlush(dsply);
  }
  else if(widget == pixelSizeValueW) {
    globPixelSize = wstrtod(pixelSizeValueW);
    
    // BRANT - Now set the scroll bar to the value we just got.
    // e.g. It has to be done for all widgets...
    XtSetArg(arglst[0], XmNvalue, (int)(globPixelSize*1000.0));
    XtSetValues(pixelSizeScrollW, arglst, 1);
    XFlush(dsply);
  }
  else if(widget == stepToggleW) {
    
    if(whichStep==MAXRADIUS || whichStep==MINRADIUS)
      steps = 5.0;
    else if(whichStep==PIXELSIZE)
      steps = 1.0;
    else if(whichStep==ANGY || whichStep == ANGX)
      steps = 5.0;
    else if(whichStep==FRACT)
      steps = 0.05;
    else if(whichStep==SURFDENS) {
      // We're doing the surfaceDensity in tenths now
      if(maxSurf > 5000.0)
	steps = (maxSurf*0.002)*.01;
      else if(maxSurf > 500.0)
	steps = 10.0;
      else
	steps = 1.0;
    }
    else if(whichStep==UNBLOBSIZE)
      steps = 0.001;
    else if(whichStep==UNBLOBRADIUS ||
	    whichStep==MAXX  || whichStep==MAXY || whichStep==MAXZ ||
	    whichStep==THETA || whichStep==PHI  || whichStep==OMEGA)
      steps = 5.0;
    else 
      steps = 1.0;
    
    // Put the step value
    TextSetFloat(stepTextW, steps, FALSE); 
   
  }
  else if(widget == stepToggle2W) { // Setup for number of frames etc.
    int thisFrames;
    
    // Get the step value
    steps = wstrtod(stepTextW);
    
    // Get the frames value
    thisFrames  = wstrtod(frameTextW);
    
    if(steps ==0.0) { // Init it
      if(whichStep==MAXRADIUS || whichStep==MINRADIUS)
	steps = 5.0;
      else if(whichStep==PIXELSIZE)
	steps = 1.0;
      else if(whichStep==ANGY || whichStep == ANGX)
	steps = 5.0;
      else if(whichStep==FRACT)
	steps = 0.05;
      else if(whichStep==SURFDENS) {
	// We're doing the surfaceDensity in tenths now
	if(maxSurf > 5000.0)
	  steps = (maxSurf*0.002)*.01;
	else if(maxSurf > 500.0)
	  steps = 10.0;
	else
	  steps = 1.0;
      }
      else if(whichStep==UNBLOBSIZE)
	steps = 0.001;
      else if(whichStep==UNBLOBRADIUS)
	steps = 5.0;
      else if(whichStep==MAXX || whichStep==MAXY || whichStep==MAXZ)
	steps = 5.0;
      else if(whichStep==THETA || whichStep==PHI || whichStep==OMEGA)
	steps = 5.0;
      else 
	steps = 1.0;
      // Put the step value
      TextSetFloat(stepTextW, steps, FALSE); 
    }

    if(steps!=0.0) { // OK
      if(whichStep==MAXRADIUS || whichStep==MINRADIUS) {
	nFrames = (globMaxRadius - globMinRadius) / steps;
	nFrames++;
      }
      else if(whichStep==PIXELSIZE)
	nFrames = (thisFrames == 0) ? 1 : thisFrames;
      else if(whichStep==ANGY || whichStep == ANGX)
	nFrames = (thisFrames == 0) ? 1 : thisFrames;
      else if(whichStep==FRACT) {
	if(steps > 0)
	  nFrames = (1.00000001 - globFract) / steps;
	else 
	  nFrames = globFract / steps;
      }
      else if(whichStep==SURFDENS)
	nFrames = (thisFrames == 0) ? 1 : thisFrames;
      else if(whichStep==UNBLOBSIZE)
	nFrames = (thisFrames == 0) ? 1 : thisFrames;
      else if(whichStep==UNBLOBRADIUS)
	nFrames = (thisFrames == 0) ? 1 : thisFrames;
      else if(whichStep==MAXX || whichStep==MAXY || whichStep==MAXZ) {
	if(whichStep==MAXX)
	  nFrames = (cutMaxValues[0]-cutMinValues[0]) / steps;
	if(whichStep==MAXY)
	  nFrames = (cutMaxValues[1]-cutMinValues[1]) / steps;
	if(whichStep==MAXZ)
	  nFrames = (cutMaxValues[2]-cutMinValues[2]) / steps;
	if(steps > 0)
	  nFrames--;     // Always one less
	else 
	  nFrames++;  // Always one less
      }
      else if(whichStep==THETA || whichStep==PHI || whichStep==OMEGA)
	nFrames = (globPhi==0.0) ? 180.0/steps : 360.0/steps;
      else // default
	nFrames = 1;
    }
    else 
      nFrames = 1;
    
    nFrames = fabs((float)nFrames);
    
    TextSetInt(frameTextW, nFrames, TRUE); 
  }
  
} // end function surfDensValueChange



void surfDens_CR(Widget widget, int *tag, 
		 XmAnyCallbackStruct *callback_data)
{
  // function name: surfDens_CR
  // When a carriage return happens in any of the text widgets on the
  // depth/radial cueing screen call here.

  mapShade_activate(NULL, NULL, NULL);   // Don't do if just dragging

} // end function surfDens_CR



void surfDensScrollChange(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: surfDensScrollChange
  // Based on the change in the scroll bars below adjust the internal
  // program variable accordingly.
  
  XmScrollBarCallbackStruct *temp = (XmScrollBarCallbackStruct *) call_data;

  // For theta/phi/omega scroll widgets:
  // We've got to decide here if what we've got is XYZ or TPO
  // check the rotateXYZ logical and then call XDY's rtns as
  // appropriate XYZ-TPO
  
  if(widget == surfDensScrollW) {
    surfaceDensity= (float)temp->value;
    surfaceDensity *= 0.1;  /* Do it in tenths */
    maxUnblob = 0.0 - surfaceDensity;
    
    // RWA July 29, 1998 OK - irev determines the sign on the
    // surfaceDensity sooo we'll always force it to be positive or
    // 0.0.  Display in the text widget based on the irev so we can
    // show them what they are actually looking at when shaded.
    // Actually surfaceDenity should always be positive here as
    // we're getting it from the scroll bar.
    
    surfaceDensity = fabs(surfaceDensity);
    sprintf(outStr,"%.2f",irev*surfaceDensity);
    TextSet(surfDensValueW, outStr, FALSE); 
    updateNewVolume();
  }
  else if(widget == fractScrollW) {
    globFract = temp->value * 0.001;
    sprintf(outStr,"%.2f",globFract);
    doValue(outStr, fractValueW, FALSE);
  }
  else if(widget == angxScrollW) {
    globAngx = temp->value;
    sprintf(outStr,"%3d",temp->value);
    doValue(outStr, angxValueW, FALSE);
  }
  else if(widget == angyScrollW) {
    globAngy = -temp->value;
    sprintf(outStr,"%3d",temp->value);
    doValue(outStr, angyValueW, FALSE);
  }
  else if(widget == thetaScrollW) {
    globTheta = temp->value - 360.0;
    globTilt  = (globTheta != 0) ? TRUE : FALSE;    
    sprintf(outStr,"%.0f",globTheta);
    doValue(outStr, thetaValueW, FALSE);
  }
  else if(widget == omegaScrollW) {
    globOmega = temp->value - 360.0;
    globTilt  = (globOmega != 0) ? TRUE : FALSE;
    sprintf(outStr,"%.0f",globOmega);
    doValue(outStr, omegaValueW, FALSE);
  }
  else if(widget == phiScrollW) {
    globPhi  = temp->value - 360.0;
    globTilt = (globPhi != 0) ? TRUE : FALSE;
    sprintf(outStr,"%.0f",globPhi);
    doValue(outStr, phiValueW, FALSE);
  }
  else if(widget == maxRadiusScrollW) {
    globMaxRadius= temp->value;
    sprintf(outStr,"%.0f",globMaxRadius);
    doValue(outStr, maxRadiusValueW, FALSE);
  }
  else if(widget == minRadiusScrollW) {
    globMinRadius= temp->value;
    sprintf(outStr,"%.0f",globMinRadius);
    doValue(outStr, minRadiusValueW, FALSE);
  }
  else if(widget == pixelSizeScrollW) {
    globPixelSize= temp->value * 0.001;
    sprintf(outStr,"%.2f",globPixelSize);
    doValue(outStr, pixelSizeValueW, FALSE);
  }
  
  if(immediateMapShade)
    mapShade_activate(NULL, NULL, NULL);   // Don't do if just dragging
  
} // end function surfDensScrollChange



void mapShade_activate_work(float **float_img_buf1,
			    float **map_img_buf,
			    int calcMapPixel)
{
  // function name: mapShade_activate_work
  // They want to see their map sections displayed as a depth-cue
  // shade 3D surface.  this rtmn does all of the real work in setting
  // up for 3D rendering

  char ikey;
  int i, j, m;
  int col1, col2, row1, row2, sec1, sec2;
  int numFrames, secLeft, minLeft;
  int samRenStep;
  int topRadialSec;
  int thisExpStep, numFramesMm;
  float *thisMap, *thisRadialMap, *thisShadMap; 
  float stepVal;
  float saveSurfaceDensity, saveGlobFract;
  float saveGlobAngx, saveGlobAngy, saveUnblobSize; 
  float radialXcen, radialYcen, radialZcen;
  float thisMinMapPixel,thisMaxMapPixel;
  float saveCutMax0Value,saveCutMax1Value,saveCutMax2Value;
  double saveGlobTheta, saveGlobPhi, saveGlobOmega, saveGlobMaxRadius;
  double saveGlobMinRadius, saveGlobPixelSize;
  time_t tloc,oldtloc;

  // Give feedback on what the surface density *really* is...
  // Make sure the user sees it with the right sign
  sprintf(outStr,"%.2f",irev*surfaceDensity);
  TextSet(surfDensValueW, outStr, FALSE); 

  saveCutMax0Value = cutMaxValues[0];
  saveCutMax1Value = cutMaxValues[1];
  saveCutMax2Value = cutMaxValues[2];

  if(fileType==MAP) { // OK we've got map data
    
    // Get the sample rendering step
    samRenStep = (int)wstrtod(sampleRenderW);
    
    firstSec= 0;
    lastSec = origNumSecs;
    
    // Init the Unblob stuff
    minUnblob = secMinPixel;
    maxUnblob = 0.0 - surfaceDensity;
    
    // Ok they're gonna have a 3D image so we can write it as tiff 
    // if they want
    XtVaSetValues(rawButtonW, XmNsensitive, TRUE, NULL);
    XFlush(dsply);
    
    // Get the step value
    stepVal = wstrtod(stepTextW);
    if(whichStep==UNBLOBSIZE) { // This is done is percentages
      // Divide by 100 for percentage
      stepVal= (float)((numCols*numRows*origNumSecs) * (stepVal/100.0));
    }
    if(stepVal==0.0)
      stepVal = 1.0;
    
    thisExpStep = wstrtod(radialTextW);
    
    if(thisExpStep > 0)
      expStep = MIN(MAX(4,thisExpStep),12);
    TextSetInt(radialTextW, expStep, TRUE);
    
    if(whichStep==MAXRADIUS) { // Calculate how many frames you need
      numFrames = (globMaxRadius - globMinRadius)/ stepVal;
      // NO???  	    numFrames++;
      TextSetInt(frameTextW, numFrames, TRUE); 
    }
    if(whichStep==MAXX || whichStep==MAXY || whichStep==MAXZ) {
      // Calculate how many frames you need
      int maxFrames;
      
      // First see what they think they want
      numFrames = wstrtod(frameTextW);
      numFrames = MAX(numFrames,0);
      maxFrames = 1;
      
      if(whichStep==MAXX)
	maxFrames =  (cutMaxValues[0] - cutMinValues[0])/fabs(stepVal);
      else if(whichStep==MAXY)
	maxFrames =  (cutMaxValues[1] - cutMinValues[1])/fabs(stepVal);
      else if(whichStep==MAXZ)
	maxFrames =  (cutMaxValues[2] - cutMinValues[2])/fabs(stepVal);
      
      if(numFrames <= 0 || numFrames > maxFrames) { // Then calc it ourselves
	if(whichStep==MAXX)
	  numFrames = (cutMaxValues[0] - cutMinValues[0])/fabs(stepVal);
	if(whichStep==MAXY)
	  numFrames = (cutMaxValues[1] - cutMinValues[1])/fabs(stepVal);
	if(whichStep==MAXZ)
	  numFrames = (cutMaxValues[2] - cutMinValues[2])/fabs(stepVal);

	numFrames--;  // Always one less
	TextSetInt(frameTextW, numFrames, TRUE); 
      }

    }
    else { // Get how many movie frames they want
      numFrames = wstrtod(frameTextW);
      numFrames = MAX(numFrames,1);
    }
    
    // Stuff rotate text widgets appropriatly
    // We've got to decide here if what we've got is XYZ or TPO
    // check the rotateXYZ logical and then call XDY's rtns as appropriate
    // XYZ-TPO

    TextSetFloat(rotateThetaW, globTheta, FALSE); 
    TextSetFloat(rotatePhiW,   globPhi,   FALSE); 
    TextSetFloat(rotateOmegaW, globOmega, FALSE); 
    
    // ADjust numFrames if stereo
    if(stereoMapShade) { // Offset it for stereo
      numFrames *= 2;
      globTheta += 3;
    }
    
    // Init
    map_min = R4_MAX;
    map_max = R4_MIN;

    // First let's see if they want to do high resolution or not
    if(!lowRes) { // Calc the high res one
      
      if((globPixelSize + (stepVal*(float)numFrames)) <= 0.0) {
	askThem("I don't know how to do pixelSizes <= 0.0", 
		XmDIALOG_INFORMATION);
	goto allDone;
      }
      
      numFramesMm = numFrames;
      resolution_tryit_activate(0,&numFramesMm,0);
      // They may have changed the size of the buffer in
      // resolution_tryit() so account for that here
      saveCutMax0Value = cutMaxValues[0];
      saveCutMax1Value = cutMaxValues[1];
      saveCutMax2Value = cutMaxValues[2];
      
      if(whichStep==PIXELSIZE) { // Calc largestDim
	float tempValue;
	// pixelSizeFactor is numCols / 100.0 so let's
	// calc the largestDim as:
	//
	//   (100.0 * pixelSizeFactor) / pixelSize
	
	largestDim = (100.0*pixelSizeFactor*cutFactor) / globPixelSize;
	tempValue =  (100.0*pixelSizeFactor)/
	  (globPixelSize + stepVal*numFrames);
	// Now account for if it's cut down
	tempValue *= cutFactor;
	largestDim = MAX(largestDim,(int)tempValue);
      }
      else { // Just use this one
	largestDim = MAX(numCols, numRows);
	largestDim = MAX(largestDim, origNumSecs);
      }
      largestDim *= 1.1;  // Let's see if this helps
    }
    else { // Then let's just use the gold std
      if(*float_img_buf1 == float_image_buffer1) { // This is the kahuna
	if(*float_img_buf1 && *float_img_buf1 != lowRes_buffer1)
	  free(*float_img_buf1);
	*float_img_buf1 = lowRes_buffer1;
	numCols = lowResCols;
	numRows = lowResRows;
	origNumSecs= numSecs = lowResSecs;
      }
      else { // Maybe the subtract buffer
      }
      
      file_width  =  purdueHeader.numCols = numCols;
      file_height =  purdueHeader.numRows = numRows;
      nsecs = numSecs;
      
      fileBinSecs= numSecs;
      fileBinRows= numRows;
      fileBinCols= numCols;
      
      xCen = numCols/2;
      yCen = numRows/2;
      zCen = origNumSecs/2;
      
      largestDim = MAX(numCols, numRows);
      largestDim = MAX(largestDim, origNumSecs);
    }
    
    // Do unblob if necessary
    unblobSwitch = FALSE;
    if(whichStep == MAXX && stepVal >= 0.0)
      cutMaxValues[0] = cutMinValues[0] + 1;
    if(whichStep == MAXY && stepVal >= 0.0)
      cutMaxValues[1] = cutMinValues[1] + 1;
    if(whichStep == MAXZ && stepVal >= 0.0)
      cutMaxValues[2] = cutMinValues[2] + 1;

    if(whichStep == UNBLOBSIZE || whichStep == UNBLOBRADIUS ||
       unblobSize != 0.0 || reverseSections) { // OK unblob/cut it
      unblobSwitch = TRUE;
      unblobBuffer = NULL;
      unblobBuffer=(float *)malloc(sizeof(float)*
				   numCols*numRows*origNumSecs);
      if(!unblobBuffer) { // Can't allocate it
	if(askThem("Insufficient memory - do you want to unblob the\n\
Gold Standard memory buffer?  (This is a \n\
destructive operation - you'll have to re-read\n\
from disk to undo this.)", XmDIALOG_QUESTION))
	  unblobBuffer = (*float_img_buf1);
	else
	  unblobSwitch = FALSE;
      }
      if(unblobSwitch)
	unblob_tryit_activate(numFrames,0,0);
    }
    else
      unblobSwitch = FALSE;
    
    // Allocate main window buffer
    // We always keep something allocated for this
    free(*map_img_buf);
    CK((*map_img_buf)= (float *)malloc(sizeof(float)*
				       largestDim*largestDim
				       *numFrames));
    
    if((*map_img_buf)==NULL) { // Can't get memory
      askThem("Cannot allocate image buffer.", XmDIALOG_INFORMATION);
    }
    else { // Go ahead
      
      // Remember where we started
      saveSurfaceDensity = surfaceDensity; 
      saveGlobFract      = globFract; 
      saveGlobAngx       = globAngx; 
      saveGlobAngy       = globAngy; 
      saveGlobTheta      = globTheta; 
      saveGlobPhi        = globPhi; 
      saveGlobOmega      = globOmega; 
      saveGlobMaxRadius  = globMaxRadius; 
      saveGlobMinRadius  = globMinRadius; 
      saveGlobPixelSize  = globPixelSize; 
      saveUnblobSize     = unblobSize;
      
      ikey = 'M';
      
      // RWA July 29, 1998  We're trying to make this
      // consistent with the rotate_do_activate() routine.
      // globTheta  = 0.0 - globTheta;
      
      // Don't do this for X it's OK
      // globPhi = 0.0 - globPhi;
      
      if(whichStep==MAXRADIUS) { // Let's start off at the minRadius
	// WHY ???  globMaxRadius = globMinRadius+1;
	globMaxRadius = globMinRadius + stepVal;
      }
      
      shadedMaps = TRUE;
      XtVaSetValues(paintMapButtonW, XmNsensitive, TRUE, NULL); 
      XtVaSetValues(mapMovieButtonW, XmNsensitive, TRUE, NULL); 
      XFlush(dsply);
      
      // What time is it?
      time(&oldtloc);
      
      // Allocate these for the shading routine emmap_rob_shd1()
      free(rwa_tmp2d_map);
      free(rwa_tmp_map);
      CK(rwa_tmp2d_map = (double *)malloc((largestDim*largestDim + 22) *
					  sizeof(double)));
      CK(rwa_tmp_map   = (double *)malloc((largestDim + 22) * 
					  sizeof(double)));
      
      for (m=0; m<numFrames; m++) { // Alloc memory
	// We may have changed the dimensions of these by
	// going to a high resolution display (and back) etc. 
	// realloc these
	if(cuedMap[m].depth != NULL)
	  free(cuedMap[m].depth);
	if(cuedMap[m].shad != NULL)
	  free(cuedMap[m].shad);
	if(cuedMap[m].radial != NULL)
	  free(cuedMap[m].radial);
	
	CK(cuedMap[m].depth= (float *)malloc(sizeof(float)*
					     largestDim*largestDim));
	CK(cuedMap[m].shad= (float *)malloc(sizeof(float)*
					    largestDim*largestDim));
	CK(cuedMap[m].radial= (float *)malloc(sizeof(float)*
					      largestDim*largestDim));
      }
      
      for (m=0; m<numFrames; m++) { // Try different angles
	
	numFramesMm = numFrames - m;
	
	// Now estimate time
	time(&tloc);
	secLeft = numFrames * 1;    // a guess
	if(numCols > 240)
	  secLeft = numFrames * 12; // a guess
	else if(numCols > 190)
	  secLeft = numFrames * 8;  // a guess
	else if(numCols > 160)
	  secLeft = numFrames * 4;  // a guess
	
	if(m>0) { // Can't estimate unless we've already done one
	  secLeft = (((int)(tloc - oldtloc) + 1.0)/ m) * (numFrames - m);
	  // +1 For truncated seconds
	  
	  // Suppose we're taking less than one second to do a frame
	  // ???
	  // secLeft = MAX(secLeft,(numFrames-m));
	}
	
	// Frame the location of the stop button
	DrawX(XtWindow(mapShadeStopW), image_gcRED, 13, 13, 4);
	XFlush(dsply);
	
	if(cuedMap[m].depth && cuedMap[m].shad) { // We've enough memory
	  thisMap= cuedMap[m].depth;
	  thisShadMap= cuedMap[m].shad;
	  thisRadialMap = cuedMap[m].radial;
	  
	  if(m>0) { // OK we need to step something
	    if(whichStep==SURFDENS) { // surface density
	      surfaceDensity += stepVal;
	      if(unblobSwitch)
		unblob_tryit_activate((numFrames-m),0,0); 
	      // Need to recalc unblob if density changes
	    }
	    else if(whichStep==FRACT) {
	      globFract += stepVal;
	    }
	    else if(whichStep==PHI) {
	      if(!stereoMapShade || (m%2)==0) {
		if(memoryMiser)
		  globPhi += stepVal;
		else 
		  globPhi -= stepVal;
		if(globPhi >= 360.0)
		  globPhi -= 360.0;
		if(globPhi <= -360.0)
		  globPhi += 360.0;
	      }
	    }
	    else if(whichStep==OMEGA) {
	      if(!stereoMapShade || (m%2)==0) {
		if(memoryMiser)
		  globOmega += stepVal;
		else 
		  globOmega -= stepVal;
		if(globOmega >= 360.0)
		  globOmega -= 360.0;
		if(globOmega <= -360.0)
		  globOmega += 360.0;
	      }
	    }
	    else if(whichStep==THETA) {
	      if(!stereoMapShade || (m%2)==0) {
		if(memoryMiser)
		  globTheta += stepVal;
		else 
		  globTheta -= stepVal;
		if(globTheta >= 360.0)
		  globTheta -= 360.0;
		if(globTheta <= -360.0)
		  globTheta += 360.0;
	      }
	    }
	    else if(whichStep==MAXRADIUS)
	      globMaxRadius += stepVal;
	    else if(whichStep == UNBLOBSIZE) {
	      unblobSize += (int)stepVal;
	      unblob_tryit_activate((numFrames-m),0,0);
	      unblobSwitch = TRUE;
	    }
	    else if(whichStep == UNBLOBRADIUS) {
	      radiusUnblob += stepVal;
	      unblob_tryit_activate((numFrames-m),0,0);
	      unblobSwitch = TRUE;
	    }
	    else if(whichStep == MAXX) {
	      cutMaxValues[0] += stepVal;
	      resolution_tryit_activate(0, &numFramesMm, 0);
	    }
	    else if(whichStep == MAXY) {
	      cutMaxValues[1] += stepVal;
	      resolution_tryit_activate(0, &numFramesMm, 0);
	    }
	    else if(whichStep == MAXZ) {
	      cutMaxValues[2] += stepVal;
	      resolution_tryit_activate(0, &numFramesMm, 0);
	    }
	    else if(whichStep == PIXELSIZE) {
	      globPixelSize += stepVal;
	      if(lowRes) { // Do low res
		// Do nothing
	      }
	      else { // DO hires stuff
		resolution_tryit_activate(0,&numFramesMm,0);
		// Need to recalc unblob if we calc a new buffer
		if(unblobSwitch) { // We've got to unblob it
		  if(unblobBuffer != (*float_img_buf1))
		    free(unblobBuffer);
		  unblobBuffer= NULL;
		  unblobBuffer=(float *)malloc(sizeof(float)*
					       numCols*numRows*origNumSecs);
		  if(!unblobBuffer) { // Can't allocate it
		    askThem(
			    "Insufficient memory for unblob buffer\n\
 - continuing anyway without unblobbing\n",
			    XmDIALOG_INFORMATION);
		  }
		  else
		    unblob_tryit_activate((numFrames-m),
					  0,0); 
		}
	      }
	    }
	  }

	  // For stereo adjust the theta concept
	  if(stereoMapShade && ((m%2)==1)) { // theta
	    if(memoryMiser)
	      globTheta -= 6.0;
	    else 
	      globTheta += 6.0;
	    if(globTheta >= 360.0)
	      globTheta -= 360.0;
	    if(globTheta <= -360.0)
	      globTheta += 360.0;
	  }
	  
	  // Init this for unblob if we need it
	  maxUnblob = 0.0 - surfaceDensity;
	  
	  if(unblobSwitch)
	    this_float_image_buffer1= unblobBuffer;
	  else 
	    this_float_image_buffer1= (*float_img_buf1);
	  
	  // Show them how long they've got left
	  minLeft = (int)(float)secLeft / 60.0;
	  secLeft = secLeft - minLeft*60;
	  sprintf(outStr,
		  "[%d] Estimated time left (min:sec) -    %02d:%02d",
		  (numFrames-m),minLeft,secLeft);
	  
	  TextSet(timeLeftLabelW, outStr, TRUE);
	  
	  // No tilting in dph_tilt we do it all in 
	  // dph_nomem_tilt_fast()
	  globTilt = FALSE;
	  
	  init_floatarr(thisMap, largestDim*largestDim, 0.0);
	  init_floatarr(thisShadMap, largestDim*largestDim, 0.0);
	  init_floatarr(thisRadialMap, largestDim*largestDim, 0.0);
	  init_doublearr(rwa_tmp2d_map, largestDim*largestDim, 0.0);
	  init_doublearr(rwa_tmp_map, largestDim, 0.0);

	  // See if they want to bugout or not
	  if(XCheckTypedWindowEvent(dsply, XtWindow(mapShadeStopW),
				    ButtonPress, &event)) { // STOP!!!
	    XFlush(dsply);
	    goto noMoreCueing;
	  }

	  // Historical from FORTRAN - get rid of this later.
	  ikey = 'M';
	  
	  // Always compute the dpeth-cue even if you're doing
	  // radial cueing as you need depth to shade the radial.
	  saveRadialMapShade= radialMapShade;
	  radialMapShade = FALSE;
	  if(negUniverseCut) { // Stuff appropriately
	    col1 = 0;
	    col2 = numCols;
	    row1 = 0;
	    row2 = numRows;
	    sec1 = 0;
	    sec2 = origNumSecs;
	  }
	  else { // Optimize
	    col1 = cutMinValues[0];
	    col2 = cutMaxValues[0];
	    row1 = cutMinValues[1];
	    row2 = cutMaxValues[1];
	    sec1 = cutMinValues[2];
	    sec2 = cutMaxValues[2];
	  }
	  
	  // The concept of xcen/ycen/zcen is independent 
	  // for rotation and for radially shading - we may 
	  // need a varaible for both???
	  if(!lowRes) { // Compute center for high res images
	    float ifItWereEntireImageSize, perCentSize,
	      centerEntire, highResZero;
	    
	    perCentSize = (float)cutMaxValues[0]/(float)lowResCols -
	      (float)cutMinValues[0]/(float)lowResCols;
	    if(perCentSize <= 0.0)
	      perCentSize = 1.0;
	    ifItWereEntireImageSize = numCols / perCentSize;
	    centerEntire = ifItWereEntireImageSize / 2.0;
	    highResZero = ifItWereEntireImageSize * 
	      (float)cutMinValues[0]/(float)lowResCols;
	    radialXcen = (int)(centerEntire - highResZero);
	    
	    perCentSize = (float)cutMaxValues[1]/(float)lowResRows -
	      (float)cutMinValues[1]/(float)lowResRows;
	    if(perCentSize <= 0.0)
	      perCentSize = 1.0;
	    ifItWereEntireImageSize = numRows / perCentSize;
	    centerEntire = ifItWereEntireImageSize / 2.0;
	    highResZero = ifItWereEntireImageSize * 
	      (float)cutMinValues[1]/(float)lowResRows;
	    radialYcen = (int)(centerEntire - highResZero);
	    
	    perCentSize = (float)cutMaxValues[2]/(float)lowResSecs -
	      (float)cutMinValues[2]/(float)lowResSecs;
	    if(perCentSize <= 0.0)
	      perCentSize = 1.0;
	    ifItWereEntireImageSize = origNumSecs / perCentSize;
	    centerEntire = ifItWereEntireImageSize / 2.0;
	    highResZero = ifItWereEntireImageSize * 
	      (float)cutMinValues[2]/(float)lowResSecs;
	    radialZcen = (int)(centerEntire - highResZero);
	    
	    topRadialSec = (int)(highResZero +
				 ((ifItWereEntireImageSize *
				   perCentSize)*2.0));
	    // times 2.0 In case we rotate it
	    
	  }
	  else { // Just init them
	    radialXcen = radialYcen = radialZcen = 0;
	    topRadialSec = (int)origNumSecs;
	  }
	  
	  if(radialProjections) { // DO projections
	    float x_org, y_org, z_org;
	    float minF, maxF, scFactor, thisRadius, smallestDens;
	    float omega, phi, thetaY;
	    int colOffset, rowOffset;
	    
	    phi    = globPhi;
	    thetaY = globTheta;
	    omega  = globOmega;

	    x_org = (numCols + 1.0)/2.0     - 1.0;
	    y_org = (numRows + 1.0)/2.0     - 1.0;
	    z_org = (origNumSecs + 1.0)/2.0 - 1.0;
	    
	    thisRadius = globMaxRadius;
	    
	    if(thisRadius > largestDim * 0.5)
	      thisRadius = largestDim  * 0.5;
	    
	    smallestDens = secMinPixel-1.0;
	    
	    colOffset = (largestDim-numCols)/2;
	    rowOffset = (largestDim-numRows)/2;
	    
	    if(icosRadProj) { // Do icosahedral radial projections
	      float thresh,slid;

	      thresh = wstrtod(dioThreshTextW);
	      slid = wstrtod(dioSlideTextW);
	      icosRadialProject((float *)
				 this_float_image_buffer1,
				 numCols,numRows,origNumSecs,
				 (float *)thisShadMap,
				 x_org, y_org, z_org, 
				 thisRadius, smallestDens,
				 thetaY,phi,omega,
				 colOffset, rowOffset,
				 irev, whichIcos,
				 thresh,slid, destroyIcos,
				 globMinRadius);
	    }
	    else { // Do spherical
	      rad_dens_fast_rob_C ((float *)
				   this_float_image_buffer1,
				   numCols,numRows,origNumSecs,
				   (float *)thisShadMap,
				   x_org, y_org, z_org, 
				   thisRadius, smallestDens,
				   thetaY,phi,omega,
				   colOffset, rowOffset,
				   largestDim, irev);
	    }

	    minF = maxF = thisShadMap[0];
	    for (i=0; i<largestDim*largestDim; i++) {
	      if(thisShadMap[i] != 0.0) { // OK to use
		minF = MIN(minF,thisShadMap[i]);
		maxF = MAX(maxF,thisShadMap[i]);
	      }
	    }
	    
	    scFactor = (minF!=maxF) ? 255.0/(maxF-minF) : 1.0;
			
	    for (i=0; i<largestDim*largestDim; i++) {
	      if(thisShadMap[i] != 0.0) { // OK to use
		thisShadMap[i] = (thisShadMap[i] - minF) * scFactor;
		if(revRadProj) { // Reverse
		  // Now reverse the contrast - they do it
		  // backward for some reason
		  thisShadMap[i] = 255.0 - thisShadMap[i];
		}
	      }
	    }
	    
	  }
		    
	  else { // Do radial/depth-cued shading
	    
	    if(!truePhi) { // Use older routine
	      emmap_rob_dphshd_C(
				 (float *)this_float_image_buffer1,
				 (float *)thisMap,
				 (float *)thisRadialMap,
				 (float *)thisShadMap,
				 (float *)(*float_img_buf1),
				 (saveRadialMapShade) ? 
				 FALSE : TRUE,
				 surfaceDensity,
				 numCols, numRows,
				 sec1, sec2, col1, col2, row1, row2,
				 ikey,
				 globPhi, 
				 globTheta,
				 globMinRadius, globMaxRadius,
				 xCen, yCen, zCen,
				 radialXcen, radialYcen, radialZcen,
				 lowRes, topRadialSec);
	    }
	    else { // Try to do a TRUE phi rotation
	      thetaPhi_dphshd_C(
				(float *)this_float_image_buffer1,
				(float *)(*float_img_buf1),
				numCols, numRows,
				(float *)thisMap, // final
				FALSE, FALSE,
				sec1, sec2, col1, col2, row1, row2,
				1, irev, surfaceDensity, 
				(saveRadialMapShade) ? 
				FALSE : TRUE,
				(float *)thisShadMap,
				globPhi, globTheta, globOmega,
				globMinRadius, globMaxRadius,
				xCen, yCen, zCen,
				radialXcen, radialYcen, radialZcen,
				lowRes, topRadialSec, 
				samRenStep);
	    }
				
	  }
		    
	  // Now compute the radial if you need it
	  radialMapShade= saveRadialMapShade;
	  if(radialMapShade) { // radial
	    if(negUniverseCut) { // Stuff appropriately
	      col1 = 0;
	      col2 = numCols;
	      row1 = 0;
	      row2 = numRows;
	      sec1 = 0;
	      sec2 = origNumSecs;
	    }
	    else { // Optimize
	      col1 = cutMinValues[0];
	      col2 = cutMaxValues[0];
	      row1 = cutMinValues[1];
	      row2 = cutMaxValues[1];
	      sec1 = cutMinValues[2];
	      sec2 = cutMaxValues[2];
	    }

	    if(!truePhi) { // DO older rtn
	      emmap_rob_dphshd_C(
				 (float *)this_float_image_buffer1,
				 (float *)thisMap,
				 (float *)thisRadialMap,
				 (float *)thisShadMap,
				 (float *)(*float_img_buf1),
				 TRUE,
				 surfaceDensity,
				 numCols, numRows,
				 sec1,sec2, col1, col2, row1, row2,
				 ikey,
				 globPhi, globTheta,
				 globMinRadius, globMaxRadius,
				 xCen, yCen, zCen,
				 radialXcen, radialYcen, radialZcen,
				 lowRes, topRadialSec);
	    }
	    else { // Try a TRUE phi rotate
	      thetaPhi_dphshd_C(
				(float *)this_float_image_buffer1,
				(float *)(*float_img_buf1),
				numCols, numRows,
				(float *)thisRadialMap,
				FALSE, FALSE,
				sec1, sec2, col1, col2, row1, row2,
				1, irev, surfaceDensity, TRUE,
				(float *)thisShadMap,
				globPhi, globTheta, globOmega,
				globMinRadius, globMaxRadius,
				xCen, yCen, zCen,
				radialXcen, radialYcen, radialZcen,
				lowRes, topRadialSec,samRenStep);
	    }
	  }
		    
	  // Remember how many
	  numSecs= nsecs= m+1;
	  
	  // Now adjust back for next pass if stereo mode
	  if(stereoMapShade && ((m%2)==1)) { // theta
	    if(memoryMiser)
	      globTheta += 6.0;
	    else 
	      globTheta -= 6.0;
	    if(globTheta >= 360.0)
	      globTheta -= 360.0;
	    if(globTheta <= -360.0)
	      globTheta += 360.0;
	  }
		    
	}
	else { // no more
	  askThem("Trouble allocating memory.", XmDIALOG_INFORMATION);
	}
      }
      
      // For some reason emmap_dphshd_C() makes this too large sometimes
      if(lowRes) {
	map_max = MIN(map_max,(float)(origNumSecs/2)+1.0); 
      }
      
      // We've got to trap here in case we've set conditions such
      // that we don't have any data in the map to shade.
      if(map_max < R4_MIN*0.9 && map_min > R4_MAX*0.9) { // Have no data
	map_min= -20.0;
	map_max= 120.0;
      }
	    
      // Go in here if you want depth-cued shading
      if(secondMapShade && !radialMapShade) { // Do shading?
	
	emmap_rob_shd1((cue_t *)cuedMap,
		       rwa_tmp2d_map, rwa_tmp_map,
		       largestDim, largestDim,
		       1, numFrames,
		       saveGlobFract, TRUE,
		       saveGlobAngx,stepVal,
		       saveGlobAngy,whichStep,
		       (0.0), ANGX, ANGY, FRACT);
      }
      
      // Make sure we see the lowest value in the map as something other
      // than background
      if(radialMapShade)
	map_min -= fabs((map_max - map_min)*0.05);
      else {
	map_min -= fabs((map_max - map_min)*0.01);
	map_max += fabs((map_max - map_min)*0.01); 
      }
      
      if(map_max==map_min)
	map_max += 10.0;
      
      if(lowRes && (radialMapShade && !radialCS)) { // DOn't scale them
	// Reset these to the whole range
	// So we get the zeroth one displayed
	map_max = (float)(origNumSecs/2)+1.0;
	map_min = 0.0 - (map_max * 0.025); 
      }
      
      // Now scale them all the same
      dph_rescale_rob (largestDim, largestDim, numSecs, 
		       (cue_t *)cuedMap, tryColors, 
		       &radialStepSize);
      
      // Go in here if you want radial-cued shading
      if(secondMapShade && radialMapShade) { // Do shading for radial
	
	emmap_rob_shd1((cue_t *)cuedMap,
		       rwa_tmp2d_map, rwa_tmp_map,
		       largestDim, largestDim,
		       1, numFrames,
		       saveGlobFract, FALSE,
		       saveGlobAngx,stepVal,
		       saveGlobAngy,whichStep,
		       0.0, ANGX, ANGY, FRACT);
      }
	    
      thisMinMapPixel = HUGE;
      thisMaxMapPixel = -HUGE;
      
      // Stuff the main window display buffer -
      // Only stuff it on the last pass
      for (m=0; m<numFrames; m++) { // Do all frames
	float mapPixel;
	thisMap       = cuedMap[m].depth;
	thisShadMap   = cuedMap[m].shad;
	thisRadialMap = cuedMap[m].radial;
	
	for (i=0; i<largestDim; i++) {
	  for (j=0; j<largestDim; j++) {
	    if(!radialProjections && !secondMapShade && !radialMapShade)
	      mapPixel = thisMap[(i*largestDim)+j];
	    else
	      mapPixel = thisShadMap[(i*largestDim)+j];
	    
	    (*map_img_buf)[((m*(largestDim*largestDim))+
			    ((i*(largestDim))+j))]= 
	      (float)mapPixel;
	    
	    if(mapPixel != 0) { // Don't count background in this
	      if(mapPixel > thisMaxMapPixel)
		thisMaxMapPixel= mapPixel;
	      if(mapPixel < thisMinMapPixel)
		thisMinMapPixel= mapPixel;
	    }    
	  }
	}
      }
	    
      // Trap for empty display
      if(map_max==0.0 && map_min==0.0) {
	map_max = 20.0; 
	map_min = -10.0;
      }
      
      if(thisMaxMapPixel == thisMinMapPixel)
      // Just get it so that we've some kind of range
	thisMinMapPixel -= 0.001;  
      if(calcMapPixel) {
	// You can modify the concept of max and min pixel
	minMapPixel = MIN(thisMinMapPixel,minMapPixel);
	maxMapPixel = MAX(thisMaxMapPixel,maxMapPixel);
      }
      else { // Just stuff it
	minMapPixel = thisMinMapPixel;
	maxMapPixel = thisMaxMapPixel;
      }
      
      cuedMaxPixel = maxMapPixel;
      cuedMinPixel = minMapPixel;
      
      dispSections= FALSE;
      stuffMapImage((float *)(*map_img_buf), (float *)(*map_img_buf), 
		    file_width, file_height, nsecs, 1, 0);
      
      // Reset the concept of curBox in the movie/editBox screen so we
      // don't try and reference a part of the buffer we don't have.
      // Oh, and BTW we have no concept of lastNumBox with MAPs -
      // that's for scanned image boxes and should be zero here.

      curBox = MIN(curBox,(nsecs-1));
      TextSetInt(editBoxTextW, curBox+lastNumBox+1, TRUE);
      
      // Display the image in the window created above
      if(XtIsManaged(mainWindowW))
	XPutImage (dsply, image_window1, image_gc, image_image1, 0, 0,
		   0, 0, image_width, image_height);
      
      XFlush (dsply);
      
      DrawWindow();
      // Redraw the histogram
      histogram_activate();
      
      surfaceDensity = saveSurfaceDensity; 
      globFract = saveGlobFract; 
      globAngx = saveGlobAngx; 
      globAngy = saveGlobAngy; 
      globTheta = saveGlobTheta; 
      globPhi = saveGlobPhi; 
      globOmega = saveGlobOmega; 
      globMaxRadius = saveGlobMaxRadius; 
      globMinRadius = saveGlobMinRadius; 
      globPixelSize = saveGlobPixelSize; 
      unblobSize = saveUnblobSize;
      mapNumSecs = nsecs;
    }
	
    // Show them what they're seeing
    XtSetArg(arglst[0], XmNlabelPixmap, unsectionPixmap);
    XtSetArg(arglst[1], XmNselectPixmap, sectionPixmap);
    XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
    XtSetValues(sectionW, arglst, 3);
    TextSet(imgNumTextW, "1", FALSE);
    
    goto allDone;

  noMoreCueing:;
    surfaceDensity = saveSurfaceDensity; 
    globFract = saveGlobFract; 
    globAngx = saveGlobAngx; 
    globAngy = saveGlobAngy; 
    globTheta = saveGlobTheta; 
    globPhi = saveGlobPhi; 
    globOmega = saveGlobOmega; 
    globMaxRadius = saveGlobMaxRadius; 
    globMinRadius = saveGlobMinRadius; 
    globPixelSize = saveGlobPixelSize; 
    unblobSize = saveUnblobSize;
    mapNumSecs = nsecs;
    
  allDone:;
    TextSet(timeLeftLabelW, "Select parameters.", TRUE);
    
    // Make them remanage the map movie screen to get it setup properly
    // for this movie they've just calculated.
    XtUnmanageChild(editBoxScreenW);
    XFlush(dsply);
    
    // Don't trash this just yet -= they may want to zoom the
    // display in virtual_read_activate() -->> stuffMapImage()
    // so they can use this pointer before someone reallocates it...
    //
    // free((*map_img_buf)); 
    // (*map_img_buf)= NULL; 

    // Show them the volume stuff if they want it
    updateNewVolume();
    
    unblobSwitch = FALSE;
    if(unblobBuffer != (*float_img_buf1))
      free(unblobBuffer);
    unblobBuffer= NULL;
    
  }
  
  if(dispIcosa) { // Draw cisoahedron
    
    calcSizeFactor(zoomFactor);
    DrawIcosahedron(deg2rad(0.0-globTheta),
		    deg2rad(0.0-globPhi),
		    deg2rad(globOmega),
		    0, 0,
		    ((float)numCols*sizeFactor*2.0), 
		    image_window1); 
    
    DrawIcosahedron(deg2rad(0.0-globTheta),
		    deg2rad(0.0-globPhi),
		    deg2rad(globOmega),
		    500, 255,
		    numCols, XtWindow(mapShadeScreenW)); 
  }
  
  // Reset these in case we've trashed them etc.
  cutMaxValues[0] = saveCutMax0Value;
  cutMaxValues[1] = saveCutMax1Value;
  cutMaxValues[2] = saveCutMax2Value;

} // end function mapShade_activate_work



void mapMovie_show_activate(Widget widget, XtPointer client_data, 
			    XtPointer call_data)
{
  // function name: mapMovie_show_activate
  // They want to show their map sections displayed as a "movie".

  int i, j, k, incr;
  float saveContourLevel;
  XEvent event;
  int tag = *((int *) client_data);

  if(makingMovie) { // Don't do otherwise
    if(tag==MAPSTOP) { // Do nothing
      XtSetArg(arglst[0], XmNlabelPixmap, playPixmap);
      XtSetArg(arglst[1], XmNselectPixmap, playingPixmap);
      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
      XtSetValues(playW, arglst, 3); 
      XtSetArg(arglst[0], XmNlabelPixmap, stopPixmap);
      XtSetArg(arglst[1], XmNselectPixmap, stoppedPixmap);
      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
      XtSetValues(stopW, arglst, 3); 
    }
    else if(tag==FLASHLIGHT) { // flashLight the planet
      flashLightIt = ! flashLightIt;
      if(flashLightIt) { // Put up flashLight icon
	XtSetArg(arglst[0], XmNlabelPixmap, flashLightPixmap);
	XtSetArg(arglst[1], XmNselectPixmap, noflashLightPixmap);
      }
      else { // Put up noflashLight icon
	XtSetArg(arglst[0], XmNlabelPixmap, noflashLightPixmap);
	XtSetArg(arglst[1], XmNselectPixmap, flashLightPixmap);
      }
      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
      XtSetArg(arglst[3], XmNfillOnArm, FALSE);
      XtSetValues(flashLightW, arglst, 4); 
      if(flashLightIt)
	updateCursor(editBoxW, lanternCursor);
      else 
	updateCursor(editBoxW, None);
    }
    else if(tag==ROCKER) { // rock the planet
      rockIt = ! rockIt;
      if(rockIt) { // Put up rock icon
	XtSetArg(arglst[0], XmNlabelPixmap, rockPixmap);
	XtSetArg(arglst[1], XmNselectPixmap, norockPixmap);
      }
      else { // Put up norock icon
	XtSetArg(arglst[0], XmNlabelPixmap, norockPixmap);
	XtSetArg(arglst[1], XmNselectPixmap, rockPixmap);
      }
      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
      XtSetArg(arglst[3], XmNfillOnArm, FALSE);
      XtSetValues(rockW, arglst, 4); 
    }
    else if (tag==MAP || tag==INFINIMAP) { // Do the movie
      int goOn,switchThem,thisEndMovie;
      XtSetArg(arglst[0], XmNlabelPixmap, playingPixmap);
      XtSetArg(arglst[1], XmNselectPixmap, playPixmap);
      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
      XtSetArg(arglst[3], XmNfillOnArm, TRUE);
      XtSetValues(playW, arglst, 4); 
      XtSetArg(arglst[0], XmNlabelPixmap, noinfiniPixmap);
      XtSetArg(arglst[1], XmNselectPixmap, infiniPixmap);
      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
      XtSetArg(arglst[3], XmNfillOnArm, TRUE);
      XtSetValues(infiniW, arglst, 4); 
      goOn = TRUE;
      switchThem = FALSE;
      saveContourLevel= contourLevel;
      thisEndMovie = endMovie;
      if(shadedMaps && stereoMapShade)
	thisEndMovie = endMovie - 1;
      while(goOn) { // Do until done
	// Frame the location of the stop button
	DrawX(XtWindow(stopW), image_gcRED, 13, 13, 4);
	XFlush(dsply);

	incr = (shadedMaps && stereoMapShade) ? 2 : 1;

	for(i=startMovie; i<= thisEndMovie; i += incr) {

	  // if (switchThem), step backwards from thisEndMovie to startMovie
	  j = (switchThem) ? startMovie + thisEndMovie - i : i;
	  
	  TextSetInt(editBoxTextW, j+1, FALSE);
	  stuffDisplayMapBuffer(j,editBoxZoom,0);
	  if(shadedMaps && stereoMapShade)
	    stuffDisplayMapBuffer(j+1,editBoxZoom,numCols);

	  DrawEditBoxWindow(NULL, NULL, NULL);
	  if(getToggleState(movconTogW)) {
	    for (k=0; k<totContours; k++)
	      // Non-destructively draw the contour in the window buffer
	      drawContour(conTours[k], FALSE, j, editBoxW,
			  float_image_buffer1, 0, 0, editBox_display_buffer);
	    XFlush(dsply);
	  }

	  if(XCheckTypedWindowEvent(dsply, XtWindow(stopW),
				    ButtonPress, &event)) { // STOP!!!
	    XFlush(dsply);
	    goto noMore;
	  }
	  else if(XCheckTypedWindowEvent(dsply, XtWindow(movieArrowUpW),
					 ButtonPress, &event)) {
	    movieSpeed -= GetButtonVal(event.xbutton.button, 1, 2, 5);
	    movieSpeed = MAX(0,MIN(movieSpeed, 50));
	    DrawThermometer(); // Update thermometer to indicate movie speed
	  }
	  else if(XCheckTypedWindowEvent(dsply, XtWindow(movieArrowDownW),
					 ButtonPress, &event)) {
	    movieSpeed += GetButtonVal(event.xbutton.button, 1, 2, 5);
	    movieSpeed = MAX(0,MIN(movieSpeed, 50));
	    DrawThermometer(); // Update thermometer to indicate movie speed
	  }
	  else if(XCheckTypedWindowEvent(dsply, XtWindow(movieMinW),
					 ButtonPress, &event) ||
		  XCheckTypedWindowEvent(dsply, XtWindow(movieMaxW),
					 ButtonPress, &event) ) {
	    XPutBackEvent(dsply,&event);
	    XFlush(dsply);
	    if(switchThem) {      // RSS figure out why the two cases
	      XSync(dsply,FALSE); // RSS are handled differently
	      XFlush(dsply);
	    } 
	    else
	      goto noMore;
	  }
	  else if(XCheckTypedWindowEvent(dsply, XtWindow(movieDoneW),
					 ButtonPress, &event)) {
	    XPutBackEvent(dsply,&event);
	    XFlush(dsply);
	    XSync(dsply,FALSE);
	    goto noMore; 
	  }
	  XFlush(dsply);
	  em_sleep(movieSpeed);
	}

	if(rockIt) { // Rock it back the other way
	  switchThem = !switchThem;
	}
	goOn = (tag==INFINIMAP) ? TRUE : FALSE;
      }
      
    noMore:;
      contourLevel= saveContourLevel;
      XtSetArg(arglst[0], XmNlabelPixmap, playPixmap);
      XtSetArg(arglst[1], XmNselectPixmap, playingPixmap);
      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
      XtSetArg(arglst[3], XmNfillOnArm, FALSE);
      XtSetValues(playW, arglst, 4); 
      XtSetArg(arglst[0], XmNlabelPixmap, infiniPixmap);
      XtSetArg(arglst[1], XmNselectPixmap, noinfiniPixmap);
      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
      XtSetArg(arglst[3], XmNfillOnArm, FALSE);
      XtSetValues(infiniW, arglst, 4); 
    }
    XFlush(dsply);
  }
  
  // Find a better place to put this ...
  if(tag==USESECTION) { // They want to disp sections or cued map
    dispSections = !dispSections;
    if(dispSections) { // They want sections
      XtVaSetValues(paintMapButtonW, XmNsensitive, TRUE, NULL);
      XtVaSetValues(mapMovieButtonW, XmNsensitive, TRUE, NULL);
      XFlush(dsply);
      imgNum = numCols/2+1;
      TextSetInt(imgNumTextW, imgNum, FALSE);
      numSecs= nsecs= lastSec;
      maxMapPixel= secMaxPixel;
      minMapPixel= secMinPixel;
      XtSetArg(arglst[0], XmNlabelPixmap, sectionPixmap);
      XtSetArg(arglst[1], XmNselectPixmap, unsectionPixmap);
      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
      XtSetValues(sectionW, arglst, 3); 
      // Now stuff the display buffer from what was read off the
      // disk - strt at zero'th(1) section 
      stuffMapImage(map_image_buffer, float_image_buffer1, 
		    file_width, file_height, nsecs, imgNum, 0);  
    }
    else { // They want cued map
      if(shadedMaps) { // Only do if they've got any to disp
	TextSet(imgNumTextW, "1", FALSE);
	numSecs= nsecs= mapNumSecs;
	maxMapPixel= cuedMaxPixel;
	minMapPixel= cuedMinPixel;
	
	stuffMapImage(map_image_buffer, map_image_buffer, 
		      file_width, file_height, nsecs,1, 0);
	
	XtSetArg(arglst[0], XmNlabelPixmap, unsectionPixmap);
	XtSetArg(arglst[1], XmNselectPixmap, sectionPixmap);
	XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	XtSetValues(sectionW, arglst, 3);
	dispSections= FALSE;
      }
      else
	dispSections= TRUE;
    }
    
    // Display the image in the window created above
    XPutImage (dsply, image_window1, image_gc, image_image1, 0, 0,
	       0, 0, image_width, image_height);
    
    XFlush (dsply);
    
    DrawWindow();
    // Redraw the histogram
    histogram_activate();
  }
  
} // end function mapMovie_show_activate



void editBoxClick(Widget widget, XtPointer client_data, 
		  XtPointer call_data)
{
  // function name: editBoxClick
  // Every time the mouse button is clicked call here. This routine
  // handles the drawing of background into the box image or restoring
  // same in the Paint Boxes functionality.

  char ans;
  int x, y, i, j, yPaint, xPaint, xMax, yMax, thisRadius;
  int yPaintMap, xPaintMap;
  int thisOffset;
  float thisPixel, offset;
  XEvent event;
  XmDrawingAreaCallbackStruct *draw_data = 
    (XmDrawingAreaCallbackStruct *) call_data;

  if(draw_data == 0) {
    calcSizeFactor(zoomFactor);
    XFlush(dsply);   // Make changes current
    return;
  }
    
  thisRadius = (fileType==PACKEDPORTABLE) ? packRadius : boxRadius;
  event= *(draw_data->event);
  
  switch (event.type) {
    
  case ButtonRelease:	      
    x = event.xbutton.x;  // Get x,y copords of point pos
    y = event.xbutton.y;
    
    if(event.xbutton.button == Button1) {
      flashLightIt = FALSE;  // Put up noflashLight icon
      XtSetArg(arglst[0], XmNlabelPixmap, noflashLightPixmap);
      XtSetArg(arglst[1], XmNselectPixmap, flashLightPixmap);
      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
      XtSetArg(arglst[3], XmNfillOnArm, FALSE);
      XtSetValues(flashLightW, arglst, 4); 
      updateCursor(editBoxW, None);
    }
    break;
    
  case ButtonPress:
    x = event.xbutton.x;  // Get x,y coords of point pos
    y = event.xbutton.y;
    
    if(flashLightIt)
      flashLightIt=99;  // Turn it really ON
    
    calcBoxSizeFactor(editBoxZoom);
    
    // Button1 means draw in background value
    if(event.xbutton.button == Button1) {
      
      if(paintMapSwitch) { // We are painting a MAP
	// Get brush size
	yPaintMap = y * recipSizeFactor;
	xPaintMap = x * recipSizeFactor;
	yMax= xMax = numCols;
	
	xPaintMap = MAX(0,MIN(xPaintMap,xMax));
	yPaintMap = MAX(0,MIN(yPaintMap,yMax));
	
	// Repaint it
	if(normalExcise)
	  paintMap[curBox*numCols*numRows + yPaintMap*numCols + xPaintMap]= 
	    float_image_buffer1[((curBox*numCols*numRows)+
				 (yPaintMap*numCols)+xPaintMap)];
	else
	  paintMap[curBox*numCols*numRows + yPaintMap*numCols + xPaintMap]= 0.0;
	
	yPaint = y - y%(int)sizeFactor;
	xPaint = x - x%(int)sizeFactor;
	yMax = xMax = numCols*sizeFactor;
	
	xPaint = MAX(0,MIN(xPaint,xMax));
	yPaint = MAX(0,MIN(yPaint,yMax));
	
	for (i=yPaint; i<(yPaint+sizeFactor); i++)
	  for (j=xPaint; j<(sizeFactor+xPaint); j++)
	    editBox_display_buffer[(i*BOX_WIDTH)+j] = RED;
	
	stuffDisplayMapBuffer(curBox,editBoxZoom,0);
	DrawEditBoxWindow(NULL, NULL, NULL);
      }
      else { // Normal
	// How big is the brush?
	yPaint = y - (paintSize/2);
	xPaint = x - (paintSize/2);
	yMax= xMax = ((thisRadius*2)+1) * sizeFactor;
	
	xPaint = MAX(0,MIN(xPaint,(xMax-paintSize)));
	yPaint = MAX(0,MIN(yPaint,(yMax-paintSize)));
	
	// Paint it in brush size
	for (i=yPaint; i<(yPaint+paintSize); i++)
	  for (j=xPaint; j<(paintSize+xPaint); j++) 
	    if(bkgrndMethod==USEAVGPIXEL) { // avg
	      XPutPixel(editBox_image1, j, i, 
			pixels_return[(int)backDispChar]);
	      editBox_display_buffer[(i*BOX_WIDTH)+j] = backDispChar;
	      editBox_mirror_buffer[(i*BOX_WIDTH)+j]  = backChar;
	    }
	    else { // random
	      float thisVal;
	      thisVal = calcFloatBackground(editBox_int_buffer, 
					    xMax, yMax,
					    BOX_WIDTH,
					    fileType,
					    bkgrndMethod,
					    boxMethod,
					    packRadius,
					    boxRadius,
					    featherRadius,
					    recipSizeFactor,
					    tryColors);
	      XPutPixel(editBox_image1, j, i, pixels_return[6+(int)thisVal]);
	      
	      editBox_mirror_buffer[(i*BOX_WIDTH)+j]= 
		(float)
		pixels_return[(6+(int)
			       calcFloatBackground(
						   editBox_int_buffer, 
						   xMax, 
						   yMax,
						   BOX_WIDTH,
						   fileType,
						   bkgrndMethod,
						   boxMethod,
						   packRadius,
						   boxRadius,
						   featherRadius,
						   recipSizeFactor,
						   tryColors
						   ))];
	    }
      }
    }
    
    // Button2 is for replacing with image info kinda like
    // an undo feature
    else if(event.xbutton.button == Button2) {
      
      if(paintMapSwitch) { // We are painting a MAP
	// Get brush size
	yPaintMap = y * recipSizeFactor;
	xPaintMap = x * recipSizeFactor;
	yMax= xMax = numCols;
	
	xPaintMap = MAX(0,MIN(xPaintMap,xMax));
	yPaintMap = MAX(0,MIN(yPaintMap,yMax));
	
	// Repaint it
	if(normalExcise)
	  paintMap[((curBox*numCols*numRows)+
		    (yPaintMap*numCols)+xPaintMap)]= 0.0;
	else
	  paintMap[(curBox*numCols*numRows)+
		    (yPaintMap*numCols)+xPaintMap]= 
	    float_image_buffer1[(curBox*numCols*numRows)+
				 (yPaintMap*numCols)+xPaintMap];
	
	yPaint = y - y%(int)sizeFactor;
	xPaint = x - x%(int)sizeFactor;
	yMax = xMax = numCols*sizeFactor;
	
	xPaint = MAX(0,MIN(xPaint,xMax));
	yPaint = MAX(0,MIN(yPaint,yMax));
	
	offset = (minMapPixel < 0.0) ? -minMapPixel : 0.0;
	
	thisPixel=  float_image_buffer1[(curBox*numCols*numRows)+
					(yPaintMap*numCols)+
					xPaintMap] + offset;
	
	thisOffset = (int)(thisPixel * scaleMapFactor) + 6;
	
	for(i=yPaint; i<(yPaint+sizeFactor); i++) {
	  for(j=xPaint; j<(sizeFactor+xPaint); j++) {
	    XPutPixel(editBox_image1, j, i, pixels_return[thisOffset]);
	    editBox_display_buffer[i*BOX_WIDTH + j] = thisOffset;
	  }
	}
	stuffDisplayMapBuffer(curBox,editBoxZoom,0);
	DrawEditBoxWindow(NULL, NULL, NULL);
      }
      else { // Normal
	
	// Get brush size
	yPaint = y - paintSize/2;
	xPaint = x - paintSize/2;
	yMax = xMax = ((thisRadius*2)+1) * sizeFactor;
	
	xPaint = MAX(0,MIN(xPaint,(xMax-paintSize)));
	yPaint = MAX(0,MIN(yPaint,(yMax-paintSize)));
	
	// Repaint it
	for (i=yPaint; i<(yPaint+paintSize); i++) {
	  for (j=xPaint; j<(paintSize+xPaint); j++) {
	    editBox_mirror_buffer[(i*BOX_WIDTH)+j]=
	      editBox_undo_buffer[(i*BOX_WIDTH)+j];
	    if(fileType==PACKEDPORTABLE) {
	      XPutPixel(editBox_image1, j, i, 
			pixels_return
			[
			 (int)
			 ((editBox_undo_buffer[(i*BOX_WIDTH)+j]
			   *scalePackFactor)+6.0)]);
	      editBox_display_buffer[(i*BOX_WIDTH)+j]= 
		(BYTE)(int)
		((editBox_undo_buffer[(i*BOX_WIDTH)+j]
		  *scalePackFactor)+6.0);
	    }
	    else
	      editBox_display_buffer[(i*BOX_WIDTH)+j]= 
		pixels_return[(int)
			      ((editBox_undo_buffer[(i*BOX_WIDTH)+j]
				*scaleBoxFactor)+6.0)];
	  }
	}
      }
    }
    
    // Button3 is for filling in for paintMap.
    else if(event.xbutton.button == Button3) {
      
      if(paintMapSwitch) { // We are painting a MAP
	int i, j, xNew, yNew;
	int xLo, xHi, yLo, yHi;
	
	yLo = y - 2*sizeFactor;
	yHi = y + 2*sizeFactor;
	xLo = x - 2*sizeFactor;
	xHi = x + 2*sizeFactor;
	for (yNew=yLo; yNew<yHi; yNew++) {
	  for (xNew=xLo; xNew<xHi; xNew++) {
	    // Get brush size
	    yPaintMap = yNew * recipSizeFactor;
	    xPaintMap = xNew * recipSizeFactor;
	    yMax= xMax = numCols;
	    
	    xPaintMap = MAX(0,MIN(xPaintMap,xMax));
	    yPaintMap = MAX(0,MIN(yPaintMap,yMax));
	    
	    // Repaint it
	    if(normalExcise)
	      paintMap[(curBox*numCols*numRows)+
		       (yPaintMap*numCols)+xPaintMap]= 
		float_image_buffer1[(curBox*numCols*numRows)+
				    (yPaintMap*numCols)+
				    xPaintMap];
	    else
	      paintMap[(curBox*numCols*numRows)+
		       (yPaintMap*numCols)+xPaintMap]= 0.0;
	    
	    yPaint = yNew - yNew%(int)sizeFactor;
	    xPaint = xNew - xNew%(int)sizeFactor;
	    yMax = xMax = numCols*sizeFactor;
	    
	    xPaint = MAX(0,MIN(xPaint,xMax));
	    yPaint = MAX(0,MIN(yPaint,yMax));
	    for (i=yPaint; i<(yPaint+sizeFactor); i++) {
	      for (j=xPaint; j<(sizeFactor+xPaint); j++) {
		XPutPixel(editBox_image1, j, i, pixels_return[RED]);
		editBox_display_buffer[i*BOX_WIDTH + j] = RED;
	      }
	    }
	  }
	}
	
	stuffDisplayMapBuffer(curBox,editBoxZoom,0);
	DrawEditBoxWindow(NULL, NULL, NULL);
	
      }
    }
    
    DrawEditBoxWindow(NULL, NULL, NULL);
    break;
    
  case KeyPress:
    ans= XLookupKeysym((XKeyEvent *)&event, 0);
    printf ("Keyed[%c]\n",ans);
    break;
    
  default:
    break;
  }
  
  calcSizeFactor(zoomFactor);
  XFlush(dsply); // Make changes current
  
} // end function editBoxClick



void editBoxMotion(Widget widget, caddr_t unused, 
		   register XMotionEvent *pEvent)
{
  // function name: editBoxMotion

  // Whenever the pointer moves across our editBox area call
  // here. This is used to do the functions of editBoxclick above
  // except do them repeatedly while the pointer moves across the box.
  // Used in painting boxes.

  int i, j, x, y, xPaint, yPaint, xMax, yMax, thisRadius;
  int yPaintMap, xPaintMap, thisOffset;
  static int lastX=0;
  float thisPixel,offset;
  float xBoxCen, yBoxCen;
  float thisGlobAngx, thisGlobAngy;
  
  thisRadius = (fileType==PACKEDPORTABLE) ? packRadius : boxRadius;
  
  if (pEvent->type == MotionNotify) {
    
    x = pEvent->x;   // Get coords of pointer
    y = pEvent->y;
    
    calcBoxSizeFactor(editBoxZoom);
    
    // This is the Button1 mouse down state
    if(pEvent->state == 256) { // Button one = 256 find the name for this
      
      if(shadedMaps && !paintMapSwitch) { // We've got depth cued shaded data
	int xDiff, yDiff;
	if(flashLightIt >=99) { // OK to flashLight it
	  
	  xBoxCen = largestDim * sizeFactor * 0.5;
	  yBoxCen = largestDim * sizeFactor * 0.5;
	  
	  xDiff = x - (int)xBoxCen;
	  yDiff = y - (int)yBoxCen;
	  
	  if(xDiff != 0.0)
	    thisGlobAngx = atan((double)xDiff/xBoxCen);
	  else 
	    thisGlobAngx = 0.0;
	  
	  if(xBoxCen != 0.0)
	    thisGlobAngy = atan((double)yDiff/xBoxCen);
	  else 
	    thisGlobAngy = 0.0;
	  
	  sprintf(outStr,"%dx,%dy",
		  (int)rad2deg(thisGlobAngx),
		  (0-(int)rad2deg(thisGlobAngy)));
	  doLabel(outStr, lightSourceLabelW, DEFAULT);
	  
	  free(rwa_tmp2d_map);
	  free(rwa_tmp_map);
	  CK(rwa_tmp2d_map = (double *)malloc(((largestDim * 
						largestDim) + 
					       22) *
					      sizeof(double)));
	  CK(rwa_tmp_map   = (double *)malloc((largestDim + 22) * 
					      sizeof(double)));
	  emmap_rob_shd1((cue_t *)cuedMap,
			 rwa_tmp2d_map, rwa_tmp_map,
			 largestDim, largestDim,
			 (curBox+1), (curBox+1), 
			 globFract, FALSE,  // Don't use for min/max
			 rad2deg(thisGlobAngx),0.0,
			 rad2deg(thisGlobAngy),whichStep,
			 0.0, ANGX, ANGY, FRACT);
	  stuffDisplayMapBuffer(curBox,editBoxZoom,0);
	  DrawEditBoxWindow(NULL, NULL, NULL);
	}
	else { // Move movie frames based on motion
	  
	  if(x > lastX) // Move left
	    curBox--;
	  else // move right
	    curBox++;
	  
	  // Set the text widget to agree
	  if(shadedMaps && stereoMapShade)
	    curBox = MAX(0,MIN(curBox,(numSecs-2)));
	  else 
	    curBox = MAX(0,MIN(curBox,(numSecs-1)));
	  
	  // Human numbers from one not zero
	  TextSetInt(editBoxTextW, curBox+lastNumBox+1, TRUE);
	  // Display the new box
	  stuffDisplayMapBuffer(curBox,editBoxZoom,0);
	  if(shadedMaps && stereoMapShade)
	    stuffDisplayMapBuffer(curBox+1,editBoxZoom,numCols);
	  DrawEditBoxWindow(NULL, NULL, NULL);
	  lastX = x;
	}
	
      }
      else { // Regular boxing painting
	
	if(paintMapSwitch) { // We are painting a MAP
	  // Get brush size
	  yPaintMap = y * recipSizeFactor;
	  xPaintMap = x * recipSizeFactor;
	  yMax= xMax = numCols;
	  
	  xPaintMap = MAX(0,MIN(xPaintMap,xMax));
	  yPaintMap = MAX(0,MIN(yPaintMap,yMax));
	  
	  // Repaint it
	  if(normalExcise)
	    paintMap[((curBox*numCols*numRows)+
		      (yPaintMap*numCols)+xPaintMap)] = 
	      float_image_buffer1[((curBox*numCols*numRows)+
				   (yPaintMap*numCols)+xPaintMap)];
	  else
	    paintMap[((curBox*numCols*numRows)+
		      (yPaintMap*numCols)+xPaintMap)] = 0.0;
	  
	  
	  yPaint = y - y%(int)sizeFactor;
	  xPaint = x - x%(int)sizeFactor;
	  yMax = xMax = numCols*sizeFactor;
	  
	  xPaint = MAX(0, MIN(xPaint,xMax));
	  yPaint = MAX(0, MIN(yPaint,yMax));
	  for (i=yPaint; i<(yPaint+sizeFactor); i++) {
	    for (j=xPaint; j<(sizeFactor+xPaint); j++) {
	      XPutPixel(editBox_image1, j, i, pixels_return[RED]);
	      editBox_display_buffer[(i*BOX_WIDTH)+j] = (BYTE)RED;
	    }
	  }
	  
	  stuffDisplayMapBuffer(curBox,editBoxZoom,0);
	  DrawEditBoxWindow(NULL, NULL, NULL);
	  
	}
	else { // Normal
	  // get brush size
	  yPaint = y - paintSize/2;
	  xPaint = x - paintSize/2;
	  
	  yMax= xMax = ((thisRadius*2)+1) * sizeFactor;
	  
	  xPaint = MAX(0,MIN(xPaint,(xMax-paintSize)));
	  yPaint = MAX(0,MIN(yPaint,(yMax-paintSize)));
	  
	  // Paint appropriate
	  for (i=yPaint; i<(yPaint+paintSize); i++) {
	    for (j=xPaint; j<(paintSize+xPaint); j++) {
	      if(bkgrndMethod==USEAVGPIXEL) { // avg
		XPutPixel(editBox_image1, j, i, 
			  pixels_return[(int)backDispChar]);
		editBox_display_buffer[(i*BOX_WIDTH)+j]=
		  (BYTE)backDispChar;
		editBox_mirror_buffer[(i*BOX_WIDTH)+j]= backChar;
	      }
	      else { // random
		float thisVal;
		thisVal = calcFloatBackground(editBox_int_buffer, 
					      xMax, yMax,
					      BOX_WIDTH,
					      fileType,
					      bkgrndMethod,
					      boxMethod,
					      packRadius,
					      boxRadius,
					      featherRadius,
					      recipSizeFactor,
					      tryColors);
		XPutPixel(editBox_image1, j, i, 
			  pixels_return[6+(int)thisVal]);
		editBox_display_buffer[(y*BOX_WIDTH)+x] =
		  (BYTE)6+thisVal;
		editBox_mirror_buffer[(i*BOX_WIDTH)+j] = 6.0+thisVal;
	      }
	    }
	  }
	}
	
	XPutImage(dsply, editBox_window1, editBox_gc, editBox_image1, 
		  xPaint, yPaint,
		  xPaint, yPaint,  paintSize, paintSize);
	XFlush (dsply);
      } 
    }
    
    // This is the button2 mouse down state
    else if(pEvent->state == 512) {
      // Button two = 512 find the name for this!!!
      
      if(paintMapSwitch) { // We are painting a MAP
	// Get brush size
	yPaintMap = y * recipSizeFactor;
	xPaintMap = x * recipSizeFactor;
	yMax = xMax = numCols;
	
	xPaintMap = MAX(0,MIN(xPaintMap,xMax));
	yPaintMap = MAX(0,MIN(yPaintMap,yMax));
	
	// Repaint it
	if(normalExcise)
	  paintMap[(curBox*numCols*numRows)+
		   (yPaintMap*numCols)+xPaintMap]= 0.0;
	else
	  paintMap[(curBox*numCols*numRows)+
		   (yPaintMap*numCols)+xPaintMap]= 
	    float_image_buffer1[(curBox*numCols*numRows)+
				(yPaintMap*numCols)+xPaintMap];
	
	yPaint = y - y%(int)sizeFactor;
	xPaint = x - x%(int)sizeFactor;
	yMax = xMax = numCols*sizeFactor;
	
	xPaint = MAX(0, MIN(xPaint,xMax));
	yPaint = MAX(0, MIN(yPaint,yMax));
	
	offset = (minMapPixel < 0.0) ? -minMapPixel : 0.0;
	
	thisPixel=  float_image_buffer1[(curBox*numCols*numRows)+
					(yPaintMap*numCols)+
					xPaintMap] + offset;
	
	thisOffset = (int)(thisPixel * scaleMapFactor) + 6;
	
	for (i=yPaint; i<(yPaint+sizeFactor); i++) {
	  for (j=xPaint; j<(sizeFactor+xPaint); j++) {
	    XPutPixel(editBox_image1, j, i, 
		      pixels_return[thisOffset]);
	    editBox_display_buffer[(i*BOX_WIDTH)+j]=
	      (BYTE)thisOffset;
	  }
	}
	
	stuffDisplayMapBuffer(curBox,editBoxZoom,0);
	DrawEditBoxWindow(NULL, NULL, NULL);
      }
      else { // Normal
	// get brush size
	yPaint = y - paintSize/2;
	xPaint = x - paintSize/2;
	yMax = xMax = ((thisRadius*2)+1) * sizeFactor;
	
	xPaint = MAX(0, MIN(xPaint,(xMax-paintSize)));
	yPaint = MAX(0, MIN(yPaint,(yMax-paintSize)));
	
	// Paint appropriate
	for (i=yPaint; i<(yPaint+paintSize); i++) {
	  for (j=xPaint; j<(paintSize+xPaint); j++) {
	    int thisOffset;
	    editBox_mirror_buffer[(i*BOX_WIDTH)+j]=
	      editBox_undo_buffer[(i*BOX_WIDTH)+j];
	    if(fileType==PACKEDPORTABLE) {
	      thisOffset = (int)
		((editBox_undo_buffer[(i*BOX_WIDTH)+j]
		  *scalePackFactor)+6.0);
	      XPutPixel(editBox_image1, j, i, pixels_return[thisOffset]);
	      editBox_display_buffer[(i*BOX_WIDTH)+j]=
		(BYTE)thisOffset;
	    }
	    else {
	      thisOffset = (int)
		((editBox_undo_buffer[(i*BOX_WIDTH)+j]
		  *scaleBoxFactor)+6.0);
	      XPutPixel(editBox_image1, j, i, pixels_return[thisOffset]);
	      editBox_display_buffer[(i*BOX_WIDTH)+j]=
		(BYTE)thisOffset;
	    }
	  }
	}
      }
      
      XPutImage(dsply, editBox_window1, editBox_gc, editBox_image1, 
		xPaint, yPaint,
		xPaint, yPaint,  paintSize, paintSize);
      XFlush (dsply); 
    }
    
    // Button3 is for filling in for paintMap
    else if(pEvent->state == 1024) { // button 3
      int i, j, xNew, yNew;
      int xLo, xHi, yLo, yHi;
      xLo = yLo = 0;      //GCA: arbitrary values
      
      if(paintMapSwitch) { // We are painting a MAP
	
	yLo = y - 2*sizeFactor;
	yHi = y + 2*sizeFactor;
	xLo = x - 2*sizeFactor;
	xHi = x + 2*sizeFactor;
	for (yNew=yLo; yNew < yHi; yNew++) {
	  for (xNew=xLo; xNew < xHi; xNew++) {
	    // Get brush size
	    yPaintMap = yNew * recipSizeFactor;
	    xPaintMap = xNew * recipSizeFactor;
	    yMax= xMax = numCols;
	    
	    xPaintMap = MAX(0,MIN(xPaintMap,xMax));
	    yPaintMap = MAX(0,MIN(yPaintMap,yMax));
	    
	    // Repaint it
	    if(normalExcise)
	      paintMap[((curBox*numCols*numRows)+
			(yPaintMap*numCols)+xPaintMap)]= 
		float_image_buffer1[((curBox*numCols*
				      numRows)+
				     (yPaintMap*numCols)+
				     xPaintMap)];
	    else
	      paintMap[(curBox*numCols*numRows)+
		       (yPaintMap*numCols)+xPaintMap]= 0.0;
	    
	    
	    yPaint = yNew - yNew%(int)sizeFactor;
	    xPaint = xNew - xNew%(int)sizeFactor;
	    yMax = xMax = numCols*sizeFactor;
	    
	    xPaint = MAX(0,MIN(xPaint,xMax));
	    yPaint = MAX(0,MIN(yPaint,yMax));
	    
	    for(i=yPaint;i<(yPaint+(int)sizeFactor);i++) {
	      for(j=xPaint;j<((int)sizeFactor+xPaint);j++) {
		XPutPixel(editBox_image1, j, i, pixels_return[RED]);
		editBox_display_buffer[(i*BOX_WIDTH)+j] = RED;
	      }
	    }
	  }
	}
	
	stuffDisplayMapBuffer(curBox,editBoxZoom,0);
	DrawEditBoxWindow(NULL, NULL, NULL);
	
      }
      
      xLo -= sizeFactor;
      yLo -= sizeFactor;
      XPutImage(dsply, editBox_window1, editBox_gc, editBox_image1, 
		xLo, yLo, xLo, yLo,  
		6*(int)sizeFactor, 6*(int)sizeFactor);
      XFlush (dsply);  
    }
  }
    
  calcSizeFactor(zoomFactor);
  XFlush(dsply);   // Make changes current
  
} // end function editBoxMotion



void help_save_activate(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: help_save_activate
  // They have some new changes for the help document - save them.
  // Turn this function off later after the help has been fully
  // built/debugged.
  
  FILE *fOut;
  char *helpBuf;
  
  // Make sure we've got a backup
  //  system("cp help.doc help.doc_back");

  if ( (fOut = fopen("help.doc","w")) != NULL ) { // Write current buffer
    helpBuf= XmTextGetString(helpTextW);
    fwrite(helpBuf,sizeof(char),strlen(helpBuf),fOut);
    XtFree(helpBuf); 
  }
  
  fclose(fOut);
  system("mail sinkovit@sdsc.edu < help.doc");
  
} // end function help_save_activate



void context_help_screen_activate(Widget widget, XtPointer client_data, 
				  XtPointer call_data)
{
  // function name: context_help_screen_activate
  // We are trying a new way of providing help by giving them a
  // question mark "?" button on each screen that will eventually
  // contain some useful tips on how to use that screen.

  char srchStr[256], *helpBuf, *s1, *s2;
  int done;
  FILE *fIn;
  int tag = *((int *) client_data);

  // Make a backup
  // system("cp help.doc help.doc_orig");

  if(tag==UNITCELL)
    sprintf(srchStr,"UNITCELL");
  else if(tag==XYPLOT)
    sprintf(srchStr,"XYPLOT");
  else if(tag==HISTOGRAM)
    sprintf(srchStr,"HISTOGRAM");
  else if(tag==FOURIERFILT)
    sprintf(srchStr,"FOURIERFILT");
  else if(tag==USEMAPSHADE)
    sprintf(srchStr,"USEMAPSHADE");
  else if(tag==ROTFLTHELP)
    sprintf(srchStr,"ROTFLTHELP");
  else if(tag==HELPTOOLW)
    sprintf(srchStr,"HELPTOOLW");
  else if(tag==SUBTRACTMAP)
    sprintf(srchStr,"SUBTRACTMAP");
  else if(tag==TAHELP)
    sprintf(srchStr,"TRUNCADD");
  else if(tag==ROTATE)
    sprintf(srchStr,"ROTATEHELP");
  else if(tag==FIXADD)
    sprintf(srchStr,"FIXADD");
  else if(tag==FIX2D)
    sprintf(srchStr,"FIX2D");
  else if(tag==BOXHELP)
    sprintf(srchStr,"BOXHELP");
  else if(tag==APPODIZEHELP)
    sprintf(srchStr,"APPODIZEHELP");
  else if(tag==FOURAVERHELP)
    sprintf(srchStr,"FOURAVERHELP");
  else if(tag==NORMALHELP)
    sprintf(srchStr,"NORMALHELP");
  else if(tag==MANIPPP)
    sprintf(srchStr,"MANIPPPHELP");
  else if(tag==CAVGPP)
    sprintf(srchStr,"CAVGPPHELP");
  else if(tag==DELTAOMEGA)
    sprintf(srchStr,"DELTAOMEGAPPHELP");
  else if(tag==DELTAXY)
    sprintf(srchStr,"DELTAXYPARTPROCHELP");
  else if(tag==AUTODRHELP)
    sprintf(srchStr,"AUTODRHELP");
  else if(tag==AUTO2DDRHELP)
    sprintf(srchStr,"AUTO2DDRHELP");
  else if(tag==MODELHELP)
    sprintf(srchStr,"MODELHELP");
  else if(tag==BINNING)
    sprintf(srchStr,"BINNINGHELP");
  else if(tag==INTERP)
    sprintf(srchStr,"INTERPHELP");
  else if(tag==DIOHELP)
    sprintf(srchStr,"DIOHELP");
  else if(tag==IMGCONV)
    sprintf(srchStr,"IMGCONVHELP");
  else if(tag==FFTHELP) { // WHich kind of fft help
    if(measureFFT)
      sprintf(srchStr,"FFTMEASURE");
    else if(drawCTF)
      sprintf(srchStr,"CTF");
    else if(rotfltSwitch)
      sprintf(srchStr,"ROTFLT");
    else
      sprintf(srchStr,"LATREFINE");
  }
  else {
    sprintf(outStr,"[CONTEXT_HELP] I don't understand [%d]\n",tag);
    askThem(outStr, XmDIALOG_INFORMATION);
    return;
  }

  CK(s2= helpBuf= (char *)malloc((sizeof(char)*64000)));

  // Read the doc
  sprintf(outStr,"%s/context_help.mdoc",getenv("EMDIR"));
  if ( (fIn = fopen(outStr,"r")) != NULL ) {
    fread(helpBuf,sizeof(char),63000,fIn);
    fclose(fIn);
  }
  else {
    char tmpStr[512];
    sprintf(tmpStr,"Can't find context_help.mdoc file in EMDIR area. [%s]",
	    outStr);
    askThem(tmpStr, XmDIALOG_INFORMATION);
    return;
  }
  
  done = FALSE;
  s2= strstr(helpBuf,srchStr);
  
  if(s2) { // Found it
    s2 += strlen(srchStr);
    s1= helpBuf;
    sprintf(s1,"%s\n",srchStr);
    s1 += strlen(srchStr)+1;
    *s1++ = '\n';
    while(!done) { // Stuff and read
      *s1 = *s2;
      s1 ++;
      s2 ++;
      if((*s2 == 'D') && (*(s1-1)=='N') && (*(s1-2)=='E'))
	done = TRUE;
    }
    *s1++= 'D';
    *s1 = '\0';
    goto AllDone;
  }
  else
    sprintf(helpBuf,"No Help found.");
  
 AllDone:;
  
  // Stuff the text buffer
  TextSet(helpTextW, helpBuf, FALSE);
  free(helpBuf);
  
  raiseIt(helpScreenW);

} // end function context_help_screen_activate



void correlate_activate(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: correlate_activate
  // Auto-refine particle centers .

  int img_bkg, total_bkg;
  int i, j, k, cx, cy, dx, dy, yOff;
  int arraySize, Radius_old, outHeight, outWidth;
  int yOut, beginY, beginX;
  int outsideTotal, outsideNo, insideTotal, insideNo;
  float tempBox_byte_buffer[(MAXFILESIZE*2)];
  float *refineBox_buffer;
  float *refBox_byte_buffer;
  float outsideAvg, insideAvg, dividedAvg;
  FILE *fIn;
  int tag = *((int *) client_data);

  firstAMMSTDEVtime = TRUE;
  if(tag==CLEARBOXES) { // CLear all boxes

    for (i=0; i<numBoxes; i++) {
      if(box[i].inMemory==TRUE) { // Free it
	free(box[i].editBox_buffer);
	box[i].editBox_buffer= NULL;
	box[i].inMemory= FALSE;
      }
    }
    numBoxes= numRects= numPoly= 0;
    
    // Change cursor back to what it was
    updateCursor(drawingW, old_cursor);
    
    doLabel("     ", autoBoxLabelW, FALSE);
    doLabel("     ", refCorrLabelW, FALSE);
    doLabel("     ", autoCorrLabelW, FALSE);

    DrawWindow();
  }
  else if(tag==CORRELATE) { 
    //Auto_correlation method - correlate [1]
    
    // This may take awhile so give them the watch cursor
    updateCursor(drawingW, watch_cursor);
    
    printf("correlate_activate()[1] called.\n");
    TextSet(timeLeftBoxW,"Auto-correlation refinement.", TRUE);
    
    // Reset reference particles Marke
    for(i=0; i<numBoxes; i++)
      box[i].reference = FALSE;
    
    // Load in radius and enlarge it 25%
    if(fileType==PACKEDPORTABLE || fileType==PORTABLE ) {
      askThem("Sorry, Only box images can be refined.", XmDIALOG_INFORMATION);
      goto outCorr1;
      // return;
    }
    else {
      Radius_old = boxRadius;
      boxRadius = (int) (Radius_old * 1.25);
    }
    
    outHeight = outWidth = (boxRadius*2)+1;
    arraySize = outHeight * outWidth;
    
    // allocate space for the boxed particle image
    if((refineBox_buffer = (float *)
	calloc(arraySize, sizeof(float))) == (float *)NULL) {
      printf("\nError: Allocate particle image for refinement");
      TextSet(timeLeftBoxW,
		   "Error allocating buffers for refine centers.", TRUE);
      goto outCorr1;
      // return;
    }
    if((refBox_byte_buffer = (float *)
	calloc((BOX_WIDTH*2), sizeof(float))) == (float *)NULL) {
      printf("\nError: Allocate particle image for refinement[refBox]");
      TextSet(timeLeftBoxW,
		   "Error allocating buffers for refine centers.", TRUE);
      free(refineBox_buffer);
      goto outCorr1;
      // return;
    }
    
    // Open the disk file for load in boxed image
    if ( (fIn= fopen(image_fileName.addr,"r")) == NULL ) {
      printf(" Error reading image file for refine center.\n");
      free(refineBox_buffer);
      free(refBox_byte_buffer);
      boxRadius = Radius_old; 
      TextSet(timeLeftBoxW,
		   "Error opening image for refine centers.", TRUE);
      goto outCorr1;
      // return;
    }
    
    // Refine the center for each box
    for (i=0; i<numBoxes; i++) {
      // Give feedback
      sprintf(outStr,"Boxes left - [%d]",(numBoxes-i));
      TextSet(timeLeftBoxW, outStr, TRUE);
      // CHeck if they want to abort or not etc
      if(XCheckTypedWindowEvent(dsply, XtWindow(boxStopW),
				ButtonPress, &event)) { // STOP!!!
	TextSet(timeLeftBoxW, "Operation aborted by user.", TRUE);
	free(refineBox_buffer);
	free(refBox_byte_buffer);
	goto outCorr1;
      }

      // Load in particle image for refinement
      beginY = box[i].y - boxRadius;
      beginX = box[i].x - boxRadius;
      for(j=0; j<outHeight; j++) {
	if(!imgInCore) { // Get from disk
	  yOff = headOff + (beginY + j)*file_width + beginX;
	  sysError = fseek(fIn,yOff,0);
	  if(sysError) {
	    printf("Error: fseek in refine center.\n");
	    free(refineBox_buffer);
	    boxRadius = Radius_old; 
	    TextSet(timeLeftBoxW, 
			 "Error fseek() in refine centers.", TRUE);
	    free(refBox_byte_buffer);
	    goto outCorr1;
	    // return;
	  }
	  fread(refBox_byte_buffer, sizeof(BYTE),
		(outWidth),fIn);
	}
	else { // Get from core
	  yOff = (beginY + j)*file_width + beginX;
	  memcpy(refBox_byte_buffer,&inCore_buffer[(yOff)],
		 (outWidth*sizeof(float)));
	}
	      
	// Stuff the refineBox_buffer
	yOut = j*outWidth;
	for(k=0; k<outWidth; k++)
	  refineBox_buffer[yOut+k]= (short)refBox_byte_buffer[k];
      }
      
      // compute image background for each particle
      total_bkg = 0;
      
      for(j=0; j<outWidth; j++)
	total_bkg += (int)refineBox_buffer[j];       // top row
      for(j=arraySize-outWidth; j<arraySize; j++)
	total_bkg += (int)refineBox_buffer[j];       // bottom row
      for(j=outWidth; j<arraySize-outWidth; j+=outWidth)
	total_bkg += (int)refineBox_buffer[j];       // left column
      for(j=2*outWidth-1; j<arraySize-outWidth; j+=outWidth)
	total_bkg += (int)refineBox_buffer[j];       // right column
      
      img_bkg = (int) total_bkg/(2*outHeight+ 2*outWidth - 4);
      
      ccf_refine__(refineBox_buffer, &img_bkg, &outWidth,
		  &outHeight, &dx, &dy);
      
      // reset boxing center to new value
      box[i].x +=  dx;
      box[i].y +=  dy;
      
      // Remove bad particle
      // Load in refined particle image
      beginY = box[i].y - boxRadius;
      beginX = box[i].x - boxRadius;
      for(j=0; j<outHeight; j++) {
	if(!imgInCore) { // Get from disk
	  yOff = headOff + (beginY + j)*file_width + beginX;
	  sysError = fseek(fIn,yOff,0);
	  if(sysError) {
	    printf("Error: fseek in refine center.\n");
	    free(refineBox_buffer);
	    boxRadius = Radius_old;
	    TextSet(timeLeftBoxW,
			 "Error fseek() in refine centers.", TRUE);
	    free(refBox_byte_buffer);
	    goto outCorr1;
	    // return;
	  }
	  fread(tempBox_byte_buffer, sizeof(float),
		(outWidth),fIn);
	}
	else { // Get from core
	  yOff = (beginY + j)*file_width + beginX;
	  memcpy(tempBox_byte_buffer,&inCore_buffer[(yOff)],
		 (outWidth*sizeof(float)));
	}
	
	// Stuff the refineBox_buffer
	yOut = j*outWidth;
	for(k=0; k<outWidth; k++)
	  refineBox_buffer[yOut+k]= (float)tempBox_byte_buffer[k];
      }
	  
      // compute image background for each particle
      total_bkg = 0;
      
      for(j=0; j<outWidth; j++)
	total_bkg += (int)refineBox_buffer[j];       // top row
      for(j=arraySize-outWidth; j<arraySize; j++)
	total_bkg += (int)refineBox_buffer[j];       // bottom row
      for(j=outWidth; j<arraySize-outWidth; j+=outWidth)
	total_bkg += (int)refineBox_buffer[j];       // left column
      for(j=2*outWidth-1; j<arraySize-outWidth; j+=outWidth)
	total_bkg += (int)refineBox_buffer[j];       // right column
      
      img_bkg = (int) total_bkg/(2*outHeight+ 2*outWidth - 4);
      
      // Mark bad particles by comparing local background
      outsideTotal = 0;
      outsideNo = 0;
      insideTotal = 0;
      insideNo = 0;
      cx = cy = boxRadius; 
      
      for (j=0; j<outHeight; j++) {
	for (k=0; k<outWidth; k++) {
	  if((j-cy)*(j-cy)+(k-cx)*(k-cx) > Radius_old*Radius_old*1.1) {
	    outsideNo ++;
	    if((refineBox_buffer[j*outWidth+k]) < img_bkg)
	      outsideTotal ++;
	  }
	  if((j-cy)*(j-cy)+(k-cx)*(k-cx) < Radius_old*Radius_old) {
	    insideNo ++;
	    if((refineBox_buffer[j*outWidth+k]) < img_bkg)
	      insideTotal ++;
	  }
	}
      }
      
      outsideAvg = (float)outsideTotal/(float)outsideNo;
      insideAvg = (float)insideTotal/(float)insideNo;
      dividedAvg = insideAvg / outsideAvg;
      if( dividedAvg <=1.15 )
	box[i].reference = -1;
      
      box[i].correlation = dividedAvg;
      box[i].avgdensity = insideAvg/256.0;
      box[i].avgdensity = insideAvg;
      
      if(refineDisp) { // Give feedback as to which circle we're working on
	int arcRadius;
	calcSizeFactor(zoomFactor);
	// Which is best???
	arcRadius = sizeFactor * Radius_old;
	arcRadius = sizeFactor * boxRadius;
	rwaXDrawCircle(dsply, XtWindow(drawingW), image_gcBLUE,
		    box[i].x*sizeFactor - virXoffset*sizeFactor,
		    box[i].y*sizeFactor - virYoffset*sizeFactor,
		    arcRadius);
      }
    }

    // Remove bad particle
    i = 0; j = 0;
    do {
      if(box[i].reference == -1) {
	// Reset reference particles Marker
	box[i].reference = FALSE;
	i ++;
	numBoxes --;
      }
      else {
	box[j].x = box[i].x;
	box[j].y = box[i].y;
	// Reset reference particles Marker
	box[j].reference = FALSE;
	i ++; 
	j ++;
      }
    } while (j<numBoxes);
    
    // Recover the radius
    boxRadius = Radius_old;
    
    fclose(fIn);
    free(refBox_byte_buffer);
    free(refineBox_buffer);
    
  outCorr1:;
    
    sprintf(outStr,"Done.  [%d] boxes.",numBoxes);
    TextSet(timeLeftBoxW, outStr, TRUE);
    
    doLabel("Done.", autoCorrLabelW, FALSE);
    
    // Change cursor back to what it was
    updateCursor(drawingW, old_cursor);
    
    for(i=0; i<numBoxes; i++)
      box[i].reference = FALSE;
    DrawWindow();
    
  } 
  
  else if(tag==CORRELATE2 || tag==CORRELATE3) {
    // Reference based correlation - corr [2]
    printf("correlate_activate()[2] called.\n");
    TextSet(timeLeftBoxW,"Reference based refinement.", TRUE);

    if(tag==CORRELATE2) { // Make sure get some reference particles
      ref_box_flag = TRUE;
      for(i=0;i<numBoxes;i++)
	box[i].reference = FALSE;  // Reset reference particles Marker
      XtManageChild(refScreenW);
      sprintf(outStr,"\n\
Mouse modes:\n\
   Left - Select this particle for use in reference\n\
          OR Make a new particle at this lcoation and use it\n\
             for the reference.\n\
   Middle - Remove this particle altogether\n\
\n\
Use the Middle button to remove any particle that doesn't have an\n\
acceptable center and replace it using the Left button.\n\
\n\
When finished selecting press the 'DoIt' button.\n");
      TextSet(refTextW, outStr, FALSE);
      // When "doit" button is pushed, the routine 
      // ref_correlate_activate() will be called 
      
      // Manage proper children
      XtManageChild(refCircAvgW);
      XtManageChild(refAutomateW);
      XtManageChild(refReadfileW);
      XtManageChild(refReadLW);
      XtManageChild(refDarkTW);
      XtManageChild(refAllW);
      XtManageChild(refResetW);
      
      XtUnmanageChild(refRegridW);
      
      DrawWindow();
    }
    else if(tag==CORRELATE3) { // empty particle boxing
      ref_box_flag = TRUE;
      for(i=0;i<numBoxes;i++)
	box[i].reference = FALSE;  // Reset reference particles Marker
      XtManageChild(refScreenW);
      sprintf(outStr,"\n\
Empty Particle boxing:\n\
This routine tries to refine a grid of boxes using reference matching.\n\
The reference is built based on the Emma Radius and Empty Shell Width parameters.\n\
\n\
Choose how many iterations you'd like to try and press the DoIt button.\n\
Use the Stop Sign button on the 3D Rendering display to cancel this\n\
algorithm at any time.\n\
\n\
REGRID button - use this to start over with a new grid of boxes.\n\
\n");
      TextSet(refTextW, outStr, FALSE);
      // When "doit" button is pushed, the routine 
      // ref_correlate_activate() will be called
      
      refAutomate = TRUE;
      darkCircleRef =  99;
      
      XtVaSetValues(refDarkTW, XmNset, TRUE, NULL);
      XtVaSetValues(refAutomateW, XmNset, TRUE, NULL);
      
      // Manage proper children
      XtUnmanageChild(refCircAvgW);
      XtUnmanageChild(refAutomateW);
      XtUnmanageChild(refReadfileW);
      XtUnmanageChild(refReadLW);
      XtUnmanageChild(refDarkTW);
      XtUnmanageChild(refAllW);
      XtUnmanageChild(refResetW);
      
      XtManageChild(refRegridW);
      
      
      DrawWindow();
    }
    else { // Just go for it
      ref_correlate_activate();
    }
  }
  
} // end function correlate_activate



void ref_correlate_activate()
{
  // function name: ref_correlate_activate
  // Refine particle centers by reference.

  int i, j, k, dx, dy, fft_dim1, fft_dim2, iter, lastIter;
  int arraySize, Radius_old, outHeight, outWidth;
  int yOut, beginY, beginX, yOff;
  int numBoxes_ref, maxborder, cx, cy;
  int outsideTotal, outsideNo, insideTotal, insideNo;
  int totDx, totDy, numChanged, totalBugOut;
  float *refineBox_buffer;
  float *ref_refineBox_buffer;
  float *ref_refineBox_buffer_short;
  float *refBox_byte_buffer;
  float *ref_refineBox_fft;
  float *tmp_refineBox_buffer;
  float  maxP,minP;
  float varDx,varDy,sumsqDx,sumsqDy,sumDx,sumDy;
  float thisScaleFac,meanDx,meanDy,stdDx,stdDy;
  float img_bkg, total_bkg, ref_img_bkg;
  float outsideAvg, insideAvg, dividedAvg;
  FILE *fIn;

//  printf ("DBUG ra[%d]  dcr[%d]\n", refAutomate, darkCircleRef);

  // This may take awhile so give them the watch cursor
  updateCursor(drawingW, watch_cursor);
  
  // Load in radius and enlarge it 25%
  if(fileType==PACKEDPORTABLE) {
    askThem("Sorry, Only BYTE box scanned images can be refined.",
	    XmDIALOG_INFORMATION);
    return;
  }
  
  totalBugOut = FALSE;
  numChanged = INT_MAX;
  lastIter = 1;  // Default
  if(refAutomate) { // Try and do this a number of times
    lastIter = wstrtod(refIterTW);
    TextSetInt(refIterTW, lastIter, TRUE);
  }
  
  if(autoBoxSwitch)
    lastIter = autoBoxIter;
  
  // Init these so we can look at changes per cycle
  for(i=0; i<numBoxes; i++)
    box[i].nochange = 0;
  
  for (iter=0; iter<lastIter; iter++) { // Do iter times or until they opt out
    printf("Iter = [%d][%d][%d][%d][%d]\n",
	   iter,lastIter, emmaRadius, boxRadius,emptyShellWidth);
    
    Radius_old = boxRadius;
//GCA: why was the radius increased at each iteration?
//    boxRadius = Radius_old * 1.25;
    
    outHeight = outWidth = (boxRadius*2)+1;
    arraySize = outHeight * outWidth;

    // Remove particles outside file limits
    i = 0; j = 0;
    do {
      if(box[i].x+(boxRadius+1) > file_width || 
	 box[i].y+(boxRadius+1) > file_height ||
	 box[i].x-(boxRadius+1) < 0 ||
	 box[i].y-(boxRadius+1) < 0 ) {
	i ++;
	numBoxes --;
      }
      else {
	box[j].x = box[i].x;
	box[j].y = box[i].y;
	box[j].reference = box[i].reference;
	box[j].correlation = box[i].correlation;
	box[j].avgdensity = box[i].avgdensity;
	box[j].stdev = box[i].stdev;
	i ++; j ++;
      }
    } while (j<numBoxes);
    
    if(iter>0)
      sprintf(outStr,"Trying pass [%d]   Variations last time [%d]\n\
                MeanDx[%.1f]  MeanDy[%.1f]\n\
                StdDx [%.1f]  StdDy [%.1f]\n",
           iter+1, numChanged,
           meanDx, meanDy, stdDx, stdDy);
    else
      sprintf(outStr,"Trying pass [%d]",iter+1);
    printf("%s\n",outStr);
    if(checkStatus(outStr)) { // done
      iter= lastIter+11;
      goto bugOut1;
    }
    
    totDx = totDy = 0;
    sumDx = sumDy = 0.0;
    varDx = sumsqDx = 0.0;
    varDy = sumsqDy = 0.0;
    
    // allocate space for the boxed particle image 
    if((refineBox_buffer = (float *)
	calloc(arraySize,sizeof(float))) == (float *)NULL) {
      printf("\nError: Allocate particle image for refinement");
      return;
    }
    if((tmp_refineBox_buffer = (float *)
	calloc(arraySize,sizeof(float))) == (float *)NULL) {
      printf("\nError: Allocate particle image for refinement");
      free(refineBox_buffer);
      return;
    }
    if((ref_refineBox_buffer = (float *)
	calloc(arraySize,sizeof(float))) == (float *)NULL) {
      printf("\nError: Allocate particle image for refinement");
      free(refineBox_buffer);
      return;
    }
    if((ref_refineBox_buffer_short = (float *)
	calloc(arraySize,sizeof(float))) == (float *)NULL) {
      printf("\nError: Allocate particle image for refinement");
      free(refineBox_buffer);
      return;
    }
    if((refBox_byte_buffer = (float *)
	calloc(arraySize,sizeof(float))) == (float *)NULL) {
      printf("\nError: Allocate particle image for refinement[refBox]");
      free(refineBox_buffer);
      free(tmp_refineBox_buffer);
      free(ref_refineBox_buffer);
      free(ref_refineBox_buffer_short);
      return; 
    }
    
    // Open the disk file for load in boxed image
    if ( (fIn = fopen(image_fileName.addr,"r")) == NULL ) {
      printf(" Error reading image file for refine center.\n");
      boxRadius = Radius_old;
      free(refineBox_buffer);
      free(tmp_refineBox_buffer);
      free(ref_refineBox_buffer);
      free(ref_refineBox_buffer_short);
      return;
    }
    
    if(refReadFile) { // Just use a disk file as the reference
      fString_t rWrite,inFile;
      int nc, nr, halfHt, halfOut;
      float thisValue;
      float *tmpFltArray;
      forPifHead_t thisHead;
      
      rWrite.addr = malloc(sizeof(char)*4);
      rWrite.length = 1;
      sprintf(rWrite.addr,"r");
      inFile.addr = malloc(sizeof(char)*512);
      sprintf(inFile.addr,"%s",refReadFileName);
      inFile.length = strlen(refReadFileName);
      
      thisHead.luNum = 17;
      
      pif_open(inFile.addr, &thisHead.luNum, rWrite.addr);
      
      pif_read_gh(&thisHead);
      thisHead.imgNum = 1;
      pif_read_dh(&thisHead);
      
      if(thisHead.mode != 46) { // error
	sprintf(outStr,
		"Invalid PIF mode for reference image [%d].  Use mode==46",
		thisHead.mode);
	askThem(outStr, XmDIALOG_INFORMATION);
      }
      else { // OK to continue
	int yIn, xIn;
	nc = thisHead.numCols;
	nr = thisHead.numRows;
	tmpFltArray = (float *)malloc(sizeof(float)*nc*nr);
	pif_read_boxi4(&thisHead, tmpFltArray);
	
	printf("thisHead.bkg = [%f]\n", thisHead.img_bkg);
	// First pad with background
	for(i=0; i<(outHeight*outWidth); i++)
	  ref_refineBox_buffer_short[i] = thisHead.img_bkg;
	
	halfHt = nc / 2;
	halfOut = outHeight / 2;
	
	for(i=0; i<outHeight; i++) {
	  yOut = i*outWidth;
	  yIn  = halfHt - (halfOut - i);
	  for(j=0; j<outWidth; j++) { // Do all
	    xIn = halfHt - (halfOut -j);
	    
	    if( yIn<0 || yIn>(nr-1) || xIn<0 || xIn>(nc-1) ) {
	      // do nothing its already background
	    }
	    else
	      ref_refineBox_buffer_short[yOut+j] = tmpFltArray[(yIn*nc)+xIn];
	    
	    if(referenceDisp && useGUI) {
	      // Now stuff the display buffer so we can see it
	      thisValue = ref_refineBox_buffer_short[yOut+j];
	      XPutPixel(image_image1, j, i, pixels_return[(int)thisValue]);
	    }
	  }
	}
	
	if(referenceDisp && useGUI) { // Just put this stuff up
	  XPutImage(dsply, image_window1, image_gc, image_image1, 
		    0, 0, 0, 0,  outWidth, outHeight);
	  XFlush(dsply);
	}
	
	free(tmpFltArray);
	
      }
      
      free(rWrite.addr);
      free(inFile.addr);
      
    }
    else if(darkCircleRef) { // Just use a dark circle as the reference
      int thisDist, boxDist, ySq, halfHt, halfWid, emptyDist;
      float thisValue;
      boxDist   = emmaRadius * emmaRadius;
      emptyDist = (emmaRadius - emptyShellWidth)*(emmaRadius - emptyShellWidth);
      halfHt  = outHeight/2;
      halfWid = outWidth/2;
      for(i=0; i<outHeight; i++) {
	yOut = i*outWidth;
	ySq = (i-halfHt)*(i-halfHt);
	for(j=0; j<outWidth; j++) {
	  thisDist = ySq + (j-halfWid)*(j-halfWid);
	  if(thisDist < boxDist)
	    ref_refineBox_buffer[yOut+j] = 50.0;  // Dark
	  else
	    ref_refineBox_buffer[yOut+j] = 150.0;  // Background
	  // Check if we want an empty model
	  if(darkCircleRef==99 && thisDist < emptyDist)
	    ref_refineBox_buffer[yOut+j] = 175.0;   // Empty
	  // WORKS ??? 
	  // ref_refineBox_buffer[yOut+j] = 150.0;  // Empty
	  
	  ref_refineBox_buffer_short[yOut+j] = ref_refineBox_buffer[yOut+j];
	  
	  if(referenceDisp && useGUI) {
	    // Now stuff the display buffer so we can see it
	    thisValue = ref_refineBox_buffer[yOut+j];
	    XPutPixel(image_image1, j, i, pixels_return[(int)thisValue]);
	  }
	  
	}
      }
	  
      if(referenceDisp && useGUI) { // Just put this stuff up
	XPutImage(dsply, image_window1, image_gc, image_image1, 
		  0, 0, 0, 0,  outWidth, outHeight);
	XFlush(dsply);
      }
      
    }
    else { // Averaging reference particles
      numBoxes_ref = 0;
      for(i=0; i<numBoxes; i++) {
	if(box[i].reference==1) {
	  numBoxes_ref ++;
	  beginY = box[i].y - boxRadius;
	  beginX = box[i].x - boxRadius;
	  
	  for(j=0; j<outHeight; j++) {
	    if(!imgInCore) { // Get from disk
	      yOff = headOff + (beginY + j)*file_width + 
		beginX;
	      sysError = fseek(fIn,yOff,0);
	      if(sysError) {
		printf("Error: fseek in refine center.\n");
		boxRadius = Radius_old;
		free(refineBox_buffer);
		free(ref_refineBox_buffer);
		free(ref_refineBox_buffer_short);
		return;
	      }
	      fread(refBox_byte_buffer, sizeof(BYTE),
		    (outWidth),fIn);
	    }
	    else { // Get from core
	      yOff = (beginY + j)*file_width + beginX;
	      memcpy(refBox_byte_buffer,&inCore_buffer[(yOff)],
		     (outWidth*sizeof(float)));
	    }
	    
	    // Stuff the ref_refineBox_buffer
	    yOut = j*outWidth;
	    for(k=0; k<outWidth; k++) { // Do all cols
	      ref_refineBox_buffer[yOut+k] += (float)refBox_byte_buffer[k];
	    }
	  }
	}
      }
        
      if(numBoxes_ref == 0) {
	if(autoBoxSwitch) {
	  printf("\n\n---------------------------------------------------\n");
	  printf("No particles found during autoboxing: ");
	  printf("exiting robem\n");
	  printf("---------------------------------------------------\n\n");
	  exit(0);
	}
	sprintf(outStr,
		"You haven't chosen any reference particles yet.");
	askThem(outStr, XmDIALOG_INFORMATION);
	ref_refineBox_fft = malloc(1);
	iter = lastIter;
	goto bugOut1;
      }
      
      for(i=0; i<outHeight*outWidth; i++) {
	ref_refineBox_buffer[i] = 0.5 + ref_refineBox_buffer[i]/numBoxes_ref;
	ref_refineBox_buffer_short[i] = ref_refineBox_buffer[i];
      }
      
      if(refCircAvg && referenceDisp) { // Just put this stuff up
	farray_minmax(ref_refineBox_buffer, &minP, &maxP, outHeight*outWidth);
	thisScaleFac = (tryColors-11.0)/(maxP-minP);

	for(j=0; j<outHeight; j++) { // height
	  int thisValue;
	  yOut = j*outWidth;
	  for(k=0; k<outWidth; k++) { // Do all cols
	    thisValue=
	      (ref_refineBox_buffer[yOut+k]-minP)*thisScaleFac + 9;
	    if(useGUI) XPutPixel(image_image1, k+400, j, pixels_return[thisValue]);
	  }
	}

	if(useGUI) XPutImage(dsply, image_window1, image_gc, image_image1, 
			     400, 0, 400, 0,  outWidth, outHeight);
	if(useGUI) XFlush(dsply);
      }
      
      if(refCircAvg) { // Circular average the reference
	float *tmpBuffer,thisBkg;
	int ncol,nrow,interp,yOff,xOff,nterp;
	float xc,yc,dist;
	float *rint;
	int *irnum;
	
	ncol = outWidth;
	nrow = outHeight;
	xc = ncol/2.0;
	yc = nrow/2.0;
	interp= 2;  // Original way
	interp= 1;  // Check circAvgMotion() to make sure they agree
	nterp = interp * (MAX(ncol,nrow))/ 2 + 1;
	
	CK(rint= (float *)malloc(sizeof(float)*nterp*3));
	
	CK(irnum= (int *)malloc(sizeof(int)*nterp*3));
	CK(tmpBuffer= (float *)malloc(sizeof(float)*
				      outWidth*outHeight));
	
	// Stuff buffer
	for (i=0; i<outHeight*outWidth; i++)
	  tmpBuffer[i] = ref_refineBox_buffer[i];
	
	thisBkg = 0.0;
	img_1d_rob(tmpBuffer, ncol, nrow, xc, yc,
		   interp, nterp, rint, irnum, thisBkg);
	
	for (i=0; i<outHeight; i++) {
	  yOff = (i-yc)*(i-yc);
	  for (j=0; j<outWidth; j++) {
	    xOff = (j-xc)*(j-xc);
	    dist = sqrt((float)(xOff+yOff));
	    tmp_refineBox_buffer[(i*outWidth)+j] = rint[(int)dist];
	  }
	}
	
	free(irnum);
	free(rint);
	free(tmpBuffer);
	
	if(iter >= 0) { // debugging - try this
	  for(j=0; j<outHeight*outWidth; j++)
	    ref_refineBox_buffer[j] = tmp_refineBox_buffer[j];
	}
	      
      }
	  
      if(referenceDisp) { // Just put this stuff up
	farray_minmax(tmp_refineBox_buffer, &minP, &maxP, outHeight*outWidth);
	
	thisScaleFac = (float)(tryColors-11.0)/(float)(maxP-minP);
	for(j=0; j<outHeight; j++) {
	  int thisValue;
	  yOut = j*outWidth;
	  for(k=0; k<outWidth; k++) {
	    thisValue=
	      (tmp_refineBox_buffer[yOut+k]-minP)*thisScaleFac + 9;
	    if(useGUI) XPutPixel(image_image1, k, j, pixels_return[thisValue]);
	  }
	}
	
	if(useGUI) XPutImage(dsply, image_window1, image_gc, image_image1, 
			     0, 0, 0, 0,  outWidth, outHeight);
	if(useGUI) XFlush(dsply);
      }
      
    }

    // compute image background for averaged particle
    total_bkg = 0;
    for(j=0; j<outWidth; j++)
      total_bkg += ref_refineBox_buffer[j];       // top row
    for(j=arraySize-outWidth; j<arraySize; j++)
      total_bkg += ref_refineBox_buffer[j];       // bottom row
    for(j=outWidth; j<arraySize-outWidth; j+=outWidth)
      total_bkg += ref_refineBox_buffer[j];       // left column
    for(j=2*outWidth-1; j<arraySize-outWidth; j+=outWidth)
      total_bkg += ref_refineBox_buffer[j];       // right column
    
    ref_img_bkg = total_bkg/(float)(2.0*outHeight+ 2*outWidth - 4);
    
    // **** What is the best fft size?
    maxborder = MAX(outWidth, outHeight);
    fft_dim1 = 1;
    while(fft_dim1 < maxborder)
      fft_dim1 *= 2;
    fft_dim2 = fft_dim1; 
    
    // **** Apply fft on averaged reference partical once
    if((ref_refineBox_fft = (float *)
	calloc( (2*fft_dim1*fft_dim2),sizeof(float) )) == (float *)NULL) {
      printf("\nError: Allocate particle image for refinement");
      boxRadius = Radius_old;
      free(refineBox_buffer);
      free(tmp_refineBox_buffer);
      free(ref_refineBox_buffer);
      free(ref_refineBox_buffer_short);
      return;
    }
    
    init_floatarr(ref_refineBox_fft, 2*fft_dim1*fft_dim2, 0.0); // initialize
    fft_avg_ref__(ref_refineBox_buffer_short, ref_refineBox_fft, 
		 &ref_img_bkg, &outWidth, &outHeight, &fft_dim1, 
		 &fft_dim2);
    
    for (i=0; i<numBoxes; i++) {  
      // Give feedback to user
      if(lastIter > 1)
	sprintf(outStr,"Iteration[%d]  Boxes left - [%d]",
		iter+1,(numBoxes-i));
      else
	sprintf(outStr,"Boxes left - [%d]",(numBoxes-i));
      TextSet(timeLeftBoxW, outStr, TRUE);
      // CHeck if they want to abort or not etc...
      if(useGUI) {
	if(XCheckTypedWindowEvent(dsply, XtWindow(boxStopW),
				  ButtonPress, &event)) { // STOP!!!
	  TextSet(timeLeftBoxW, "Operation aborted by user.", TRUE);
	  totalBugOut = TRUE;
	  goto bugOut1;
	}
      }
      
      if(box[i].nochange < 2) { // Only do those who've changed a lot
	
	// Load in particle image for refinement
	beginY = box[i].y - boxRadius;
	beginX = box[i].x - boxRadius;
	
	for(j=0; j<outHeight; j++) {
	  if(!imgInCore) { // Get from disk
	    yOff = headOff + (beginY + j)*file_width + beginX;
	    sysError = fseek(fIn,yOff,0);
	    if(sysError) {
	      printf("Error: fseek in refine center.\n");
	      goto bugOut1;
	    }
	    fread(refBox_byte_buffer, sizeof(BYTE),
		  (outWidth),fIn);
	  }
	  else { // Get from core
	    yOff = (beginY + j)*file_width + beginX;
	    memcpy(refBox_byte_buffer,&inCore_buffer[(yOff)],
		   (outWidth*sizeof(float)));
	  }
	  
	  // Stuff the refineBox_buffer
	  yOut = j*outWidth;
	  for(k=0; k<outWidth; k++)
	    refineBox_buffer[yOut+k]= (float)refBox_byte_buffer[k];
	}
	
	// compute image background for each particle
	total_bkg = 0;
	
	for(j=0; j<outWidth; j++)
	  total_bkg += refineBox_buffer[j];       // top row
	for(j=arraySize-outWidth; j<arraySize; j++)
	  total_bkg += refineBox_buffer[j];       // bottom row
	for(j=outWidth; j<arraySize-outWidth; j+=outWidth)
	  total_bkg += refineBox_buffer[j];       // left column
	for(j=2*outWidth-1; j<arraySize-outWidth; j+=outWidth)
	  total_bkg += refineBox_buffer[j];       // right column
	
	img_bkg = total_bkg/(float)(2*outHeight+ 2*outWidth - 4);
	
	
	ref_ccf_refine__(ref_refineBox_fft, refineBox_buffer, &img_bkg,
			&outWidth, &outHeight, &fft_dim1, &fft_dim2, 
			&dx, &dy);
	
	if(dx==0 && dy==0) // No change
	  box[i].nochange++;
	else // Reset
	  box[i].nochange= 0;
	
	// Set boxing center to new value
	box[i].x +=  dx;  // x MAY be biased 1 to left??????
	box[i].y +=  dy;
	sumDx += (int)fabs((float)dx);
	sumDy += (int)fabs((float)dy);
	totDx ++;
	totDy ++;
	sumsqDx += dx*dx;
	sumsqDy += dy*dy;
	
	
	
	// **********  remove bad particle
	// Load in refined particle image
	beginY = box[i].y - boxRadius;
	beginX = box[i].x - boxRadius;
	if((beginX < 0 || beginY < 0) ||
       box[i].x+(boxRadius+1) > file_width || 
	   box[i].y+(boxRadius+1) > file_height
//	   (((box[i].y+outHeight)-file_height)>=0) ||
//	   (((box[i].x+outWidth)-file_width)>=0) 
	   ) { // Throw it out - it's off the edge
	   printf("# %d OUT LIMITS\n",i);
	  box[i].reference = -1;
	  goto byPassThrowOut;
	}

	for(j=0; j<outHeight; j++) {
	  if(!imgInCore) { // Get from disk
	    yOff = headOff + (beginY + j)*file_width + beginX;
	    sysError = fseek(fIn,yOff,0);
	    if(sysError) {
	      printf("Error: fseek in refine center.\n");
	      goto bugOut1;
	    }
	    fread(refBox_byte_buffer, sizeof(BYTE),
		  (outWidth),fIn);
	  }
	  else { // Get from core
	    yOff = (beginY + j)*file_width + beginX;
	    memcpy(refBox_byte_buffer,&inCore_buffer[(yOff)],
		   (outWidth*sizeof(float)));
	  }
	  
	  // Stuff the refineBox_buffer
	  yOut = j*outWidth;
	  for(k=0; k<outWidth; k++)
	    refineBox_buffer[yOut+k]= refBox_byte_buffer[k];
	}
	      
	// compute image background for each particle
	total_bkg = 0;
	
	for(j=0; j<outWidth; j++)
	  total_bkg += refineBox_buffer[j];       // top row
	for(j=arraySize-outWidth; j<arraySize; j++)
	  total_bkg += refineBox_buffer[j];       // bottom row 
	for(j=outWidth; j<arraySize-outWidth; j+=outWidth)
	  total_bkg += refineBox_buffer[j];       // left column
	for(j=2*outWidth-1; j<arraySize-outWidth; j+=outWidth)
	  total_bkg += refineBox_buffer[j];       // right column
	
	img_bkg = total_bkg/(float)(2*outHeight+ 2*outWidth - 4);
	
	// Mark bad particles by comparing local background
	outsideTotal = 0;
	outsideNo = 0;
	insideTotal = 0;
	insideNo = 0;
	cx = cy = boxRadius;
	
	for (j=0; j<outHeight; j++) {
	  for (k=0; k<outWidth; k++) {
	    if((j-cy)*(j-cy)+(k-cx)*(k-cx) > Radius_old*Radius_old*1.05) {
	      outsideNo ++;
	      if((refineBox_buffer[j*outWidth+k]) < img_bkg)
		outsideTotal ++;
	    }
	    if((j-cy)*(j-cy)+(k-cx)*(k-cx) < Radius_old*Radius_old) {
	      insideNo ++;
	      if((refineBox_buffer[j*outWidth+k]) < img_bkg)
		insideTotal ++;
	    }
	  }
	}
	
	outsideAvg = (float)outsideTotal/(float)outsideNo;
	insideAvg = (float)insideTotal/(float)insideNo;
	if(outsideAvg != 0.0)
	  dividedAvg = insideAvg / outsideAvg;
	else 
	  dividedAvg = refCutoffRatio+1.00;  // Make sure it's included
	
      byPassThrowOut:;
	
	box[i].correlation = dividedAvg;
//	box[i].avgdensity = insideAvg/256.0;
	box[i].avgdensity = insideAvg;
	
	if(refineDisp) { // Give feedback as to which circle we're working on
	  int arcRadius;
	  calcSizeFactor(zoomFactor);
	  arcRadius = sizeFactor * Radius_old;
	  if(useGUI) rwaXDrawCircle(dsply, XtWindow(drawingW), image_gcBLUE,
				    box[i].x*sizeFactor-virXoffset*sizeFactor, 
				    box[i].y*sizeFactor-virYoffset*sizeFactor,
				    arcRadius);
	}
	      
      }
      else { // Show them we're done with this one
	if(refineDisp) { // Give feedback as to which circle we're working on
	  int arcRadius;
	  calcSizeFactor(zoomFactor);
	  arcRadius = sizeFactor * Radius_old;
	  if(useGUI) rwaXDrawCircle(dsply, XtWindow(drawingW), image_gcWHITE,
				    box[i].x*sizeFactor-virXoffset*sizeFactor,
				    box[i].y*sizeFactor-virYoffset*sizeFactor,
				    arcRadius);
	}
      }
	  
    }
      
    // Let's try and throw out duplicate boxes
    for(j=0; j<numBoxes; j++) { // Look through all boxes
      int dist;
      for(i=0;i<numBoxes;i++) { // second
	if((i != j) && (box[j].reference!=-1)) {// Dont compare this to itself
	  dist = (box[i].x - box[j].x)*(box[i].x - box[j].x) +
	    (box[i].y - box[j].y)*(box[i].y - box[j].y);
//GCA: why a fixed distance?
//	  if(dist < 80) { // Throw it out - it's duplicate
	  if(dist < boxRadius*boxRadius) { // Throw it out - it's duplicate
	    box[i].reference = -1;
	    // Now average the coords of the 2 particles
	    box[j].y = (int)(((float)box[j].y +(float)box[i].y)*0.5);
	    box[j].x = (int)(((float)box[j].x +(float)box[i].x)*0.5);
	  }
	}
      }
    }

    // Remove bad particle
    i = 0; j = 0;
    do {
      if(box[i].reference == -1) {
	i ++;
	numBoxes --;
      }
      else {
	box[j].x = box[i].x;
	box[j].y = box[i].y;
	box[j].reference = box[i].reference;
	box[j].correlation = box[i].correlation;
	box[j].avgdensity = box[i].avgdensity;
	box[j].stdev = box[i].stdev;
	i ++; j ++;
      }
    } while (j<numBoxes);
    
    // Let's see if we've changed much from last time...
    // Now let's do stats
    meanDy = sumDy / totDy;
    stdDy  = 0.0;
    varDy  = sumsqDy/totDy - meanDy*meanDy;
    stdDy = sqrt(fabs(varDy));
    meanDx = sumDx / totDx;
    stdDx  = 0.0;
    varDx  = sumsqDx/totDx - meanDx*meanDx;
    stdDx = sqrt(fabs(varDx));
    numChanged = sumDx + sumDy;
    if(numChanged < (int)(numBoxes*0.02)) { // Let's stop
      sprintf(outStr,"Stopping - not much changed. [Only %d boxes]",
	      numChanged);
      printf("%s\n",outStr);
      checkStatus(outStr);
      iter= lastIter+11;
    }
    
  bugOut1:
    // Recover the radius
    boxRadius = Radius_old;
    
    free(ref_refineBox_fft);
    
    fclose(fIn);
    free(refBox_byte_buffer);
    free(refineBox_buffer);
    free(tmp_refineBox_buffer);
    free(ref_refineBox_buffer);
    free(ref_refineBox_buffer_short);
    if(totalBugOut)
      goto totalBugOut2;
    
    DrawWindow();    
  }
 totalBugOut2:;
  
  printf("Completed correlation refinement\n");
  printf("Done.  [%d] boxes.\n",numBoxes);
  sprintf(outStr,"Done.  [%d] boxes.",numBoxes);
  TextSet(timeLeftBoxW,outStr, TRUE);
  
  doLabel("Done.", refCorrLabelW, FALSE);
  
  // Setup for next time
  for(i=0; i<numBoxes; i++)
    box[i].reference = FALSE;
  
  DrawWindow();
  
  // Change cursor back to what it was
  updateCursor(drawingW, old_cursor);
  
} // end function ref_correlate_activate



void ref_select_activate(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: ref_select_activate
  // Mark all boxes as candidates for reference correlation standard.
  
  int i, passTag;

  if (widget == refAllW) { // Mark all reference particles Marker
    for(i=0; i<numBoxes; i++)
      box[i].reference = TRUE;
  }
  else if (widget == refResetW) { // Reset reference particles Marker
    for(i=0; i<numBoxes; i++)
      box[i].reference = FALSE;
  }
  else if (widget == refRegridW) { // Reset grid
    for(i=0; i<numBoxes; i++)
      box[i].reference = FALSE;
    
    // Clear all boxes
    for(i=0;i<numBoxes;i++)
      if(box[i].inMemory==TRUE) { // Free it
	free(box[i].editBox_buffer);
	box[i].editBox_buffer= NULL;
	box[i].inMemory= FALSE;
      }
    numBoxes= numRects= numPoly= 0;
    
    passTag = -99;
    boxgrid_activate(widget, &passTag, call_data);
  }
  
  XFlush(dsply);
  
  DrawWindow();
  
} // end function ref_select_activate



void displayPifHead(forPifHead_t thisHead)
{
  // function name: displayPifHead
  // Show them exactly what we think are the contents of the header(s)
  // of the PIF file.
  
  char pifHeadStr[4028], pixSizUnitsStr[32];
  
  if(thisHead.pixSizUnits==1)
    sprintf(pixSizUnitsStr,"Angs");
  else if(thisHead.pixSizUnits==2)
    sprintf(pixSizUnitsStr,"nm");
  else
    sprintf(pixSizUnitsStr,"Unk");
  
  sprintf(pifHeadStr,"\nPIF Header for file [%s]\n\
\n\
    nimg                    =   [%d]\n\
    imgNum                  =   [%d]\n\
    floatScaleFactor        =   [%f]\n\
    recipFloatScaleFactor   =   [%f]\n\
    int4scaleF              =   [%f]\n\
    htype                   =   [%d]\n\
    numCols                 =   [%d]\n\
    numRows                 =   [%d]\n\
    numSecs                 =   [%d]\n\
    mode                    =   [%d]\n\
    endianNess              =   [%d]\n\
    versionStr              =   [%s]\n\
    img_bkg                 =   [%f]\n\
    datai4_min              =   [%d]\n\
    datai4_max              =   [%d]\n\
    datar4_min              =   [%f]\n\
    datar4_max              =   [%f]\n\
    datar4_avg              =   [%f]\n\
    datar4_std              =   [%f]\n\
    descriptionStr          =   [%s]\n\
    timeStr                 =   [%s]\n\
    micrographStr           =   [%s]\n\
    scanNumberStr           =   [%s]\n\
    aoverb                  =   [%f]\n\
    map_abang               =   [%f]\n\
    a_cell                  =   [%f]\n\
    b_cell                  =   [%f]\n\
    c_cell                  =   [%f]\n\
    alp                     =   [%f]\n\
    bet                     =   [%f]\n\
    gam                     =   [%f]\n\
    dela                    =   [%f]\n\
    delb                    =   [%f]\n\
    delc                    =   [%f]\n\
    pixelSize               =   [%f]\n\
    pixSizUnits             =   [%d][%s]\n\
    t_matrix                =   [%d][%d][%d]\n\
                                [%d][%d][%d]\n\
    packRadius              =   [%d]\n\
    binFactor               =   [%d]\n\
    dthe                    =   [%f]\n\
    dphi_90                 =   [%f]\n\
    symmetry                =   [%d]\n\
    res_out                 =   [%f]\n\
    ctfMode                 =   [%d]\n\
    tempFactor              =   [%f]\n\
    wiener                  =   [%f]\n\
    borderType              =   [%d]\n\
    borderWidth             =   [%d]\n\
    rad_lo                  =   [%d]\n\
    rad_hi                  =   [%d]\n\
    pifVersion              =   [%f]\n\
    even                    =   [%d]\n\
    mrcX                    =   [%d]\n\
    mrcY                    =   [%d]\n\
    mrcZ                    =   [%d]\n\
    transX                  =   [%d]\n\
    transY                  =   [%d]\n\
    transZ                  =   [%d]\n\
    isTrunc                 =   [%d]\n\
    origX                   =   [%d]\n\
    origY                   =   [%d]\n\
    origZ                   =   [%d]\n\
    a0                      =   [%12E]\n\
    a1                      =   [%12E]\n\
    a2                      =   [%12E]\n\
    a3                      =   [%12E]\n\
    a4                      =   [%12E]\n\
    a5                      =   [%12E]\n\n\n",
	  fileName.addr,
	  thisHead.nimg,
	  thisHead.imgNum,
	  thisHead.floatScaleFactor,
	  thisHead.recipFloatScaleFactor,
	  thisHead.int4scaleF,
	  thisHead.htype,
	  thisHead.numCols, thisHead.numRows, thisHead.numSecs,
	  thisHead.mode,
	  thisHead.endianNess,
	  thisHead.versionStr,
	  thisHead.img_bkg,
	  thisHead.datai4_min, thisHead.datai4_max,
	  thisHead.datar4_min, thisHead.datar4_max,
	  thisHead.datar4_avg, thisHead.datar4_std,
	  thisHead.descriptionStr,
	  thisHead.timeStr,
	  thisHead.micrographStr,
	  thisHead.scanNumberStr,
	  thisHead.aoverb,
	  thisHead.map_abang,
	  thisHead.a_cell, thisHead.b_cell, thisHead.c_cell,
	  thisHead.alp,    thisHead.bet,    thisHead.gam,
	  thisHead.dela,   thisHead.delb,   thisHead.delc,
	  thisHead.pixelSize,
	  thisHead.pixSizUnits,
	  pixSizUnitsStr,
	  thisHead.t_matrix[0], thisHead.t_matrix[1],
	  thisHead.t_matrix[2], thisHead.t_matrix[3],
	  thisHead.t_matrix[4], thisHead.t_matrix[5],
	  thisHead.packRadius,
	  thisHead.binFactor,
	  thisHead.dthe,
	  thisHead.dphi_90,
	  thisHead.symmetry,
	  thisHead.res_out,
	  thisHead.ctfMode,
	  thisHead.tempFactor,
	  thisHead.wiener,
	  thisHead.borderType,
	  thisHead.borderWidth,
	  thisHead.rad_lo, thisHead.rad_hi,
	  thisHead.pifVersion,
	  thisHead.even,
	  thisHead.mrcX,   thisHead.mrcY,   thisHead.mrcZ,
	  thisHead.transX, thisHead.transY, thisHead.transZ,
	  thisHead.isTrunc,
	  thisHead.origX,  thisHead.origY,  thisHead.origZ,
	  thisHead.a0,     thisHead.a1,     thisHead.a2, 
	  thisHead.a3,     thisHead.a4,     thisHead.a5);
  
  TextSet(statTextW, pifHeadStr, FALSE);
  TextSet(statTitleW, thisHead.descriptionStr, TRUE);

} // end function displayPifHead

  

void pifhead_activate(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: pifhead_activate
  // List the PIF header.

  displayPifHead(pifHeader);

} // end function pifhead_activate



void save_title_activate(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: save_title_activate
  // Update PIF file with this new title.

  char *thisStr;
  off_t imgOffset;
  int thisImgNum;
  FILE *fOut;

  thisStr = XmTextGetString(statTitleW); 
  if(strlen(thisStr) > 79)
    thisStr[79]= '\0';
  sprintf(pifHeader.descriptionStr,"%s",thisStr);
  XtFree(thisStr);
  XFlush(dsply);

  fOut = fopen(fileName.addr,"r+");
	  
  thisImgNum = (fileType==PACKEDPORTABLE) ? imgNum : 1;
  
  // Offset = global header + preceding images + offset to description string
  imgOffset = 512 + (thisImgNum-1) *
    (off_t) (512 + purdueHeader.numRows * purdueHeader.numCols * purdueHeader.numSecs)
    + 116;

  // Now update the header to show new data at the end
  sysError= fseek(fOut,imgOffset,0);
  if(sysError) // Can't go on
    askThem("[SAVE_TITLE_ACTIVATE] Error seeking file header.",
	    XmDIALOG_INFORMATION);
  else // Update title
    fwrite(pifHeader.descriptionStr,sizeof(char),80,fOut);

  fclose(fOut);

  displayPifHead(pifHeader);

} // end function save_title_activate



void stat_calc_activate(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: stat_calc_activate
  // Calc the stats.

  if(fileType==MAP || fileType==FLOATMAP) {
    
    calcMapStats(float_image_buffer1,
		 &purdueHeader.img_bkg, 
		 &purdueHeader.map_min,
		 &purdueHeader.map_max, 
		 &purdueHeader.img_avg, 
		 &purdueHeader.img_std,
		 &numCols, &numRows, &origNumSecs);
    
    sprintf(outStr,"\nStatistics for file [%s]\n\
\n\
Title [%s]\n\
\n\
File size   [%d]x[%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
\n",
	    fileName.addr,
	    title,
	    numCols,	  numRows, origNumSecs,
	    purdueHeader.img_bkg, 
	    minMapPixel, maxMapPixel,
	    purdueHeader.img_avg, purdueHeader.img_std
	    );
    TextSet(statTextW, outStr, TRUE);
  }
  else if(fileType==EMPORTABLE || fileType==BYTEPORTABLE
	  || fileType==PORTABLE) { // Scanned iamge
    int useRectangle, thisNcol, thisNrow, median;
    
    useRectangle = rectBox.drawable;
    if(useRectangle) {
      int saveFileType;
      float *char_buffer;
      
      thisNrow = rectBox.lowy - rectBox.y;
      thisNcol = rectBox.lowx - rectBox.x;
      
      CK(char_buffer= (float *)malloc(sizeof(float)*thisNcol*thisNrow));
      
      stuffRectangleBuffer(char_buffer, 
			   image_fileName.addr);
      
      saveFileType = fileType;
      fileType = RECTPACKED;
      calcStats(char_buffer, &purdueHeader.img_bkg, 
		&purdueHeader.img_min, &purdueHeader.img_max, 
		&purdueHeader.img_avg, &purdueHeader.img_std,
		thisNcol, thisNrow, thisNcol, fileType, bkgrndMethod,
		boxMethod, packRadius, boxRadius, featherRadius,
		recipSizeFactor, tryColors);
      fileType = saveFileType;
      
      calcRectMedian(char_buffer,&median,
		     thisNcol, thisNrow);
      
      free(char_buffer);
      
      sprintf(outStr,"\nStatistics for file [%s] RECTANGLE\n\
\n\
Rect size   [%d]x[%d]\n\
Rect offset x[%d]y[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
Median      [%d]\n\
AVG,STD     [%f],[%f]\n\
",
	      fileName.addr,
	      thisNcol, thisNrow,
	      rectBox.x,rectBox.y,
	      purdueHeader.img_bkg,
	      purdueHeader.img_min,purdueHeader.img_max,
	      median,
	      purdueHeader.img_avg,purdueHeader.img_std
	      );
    }
    else { // No can do
      calcSubScanStats(inCore_buffer,           &purdueHeader.img_bkg, 
		       &purdueHeader.img_min,   &purdueHeader.img_max, 
		       &purdueHeader.img_avg,   &purdueHeader.img_std,
		       file_width, file_height, file_width, 0, 0);
      
       sprintf(outStr,"\nStatistics for file [%s] Image[%d]\n\
\n\
File size   [%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
",
	      fileName.addr,1,
	      purdueHeader.numCols,	  purdueHeader.numRows,
	      purdueHeader.img_bkg,
	      purdueHeader.img_min,purdueHeader.img_max,
	      purdueHeader.img_avg,purdueHeader.img_std
	      );
    }
    
    TextSet(statTextW, outStr, FALSE);
  }
  else if(fileType==PACKEDPORTABLE) { // packed boxes
    
    calcCircleStats(packImg[(imgNum-1)],   &purdueHeader.img_bkg, 
		    &purdueHeader.img_min, &purdueHeader.img_max, 
		    &purdueHeader.img_avg, &purdueHeader.img_std,
		    purdueHeader.numCols,  purdueHeader.numRows, 
		    purdueHeader.numCols);
    
    sprintf(outStr,"\nStatistics for file [%s] Image [%d]\n\
\n\
File size   [%d]x[%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
\n\
\n",
	    fileName.addr, imgNum,
	    purdueHeader.numCols,	  purdueHeader.numRows, 
	    purdueHeader.nimg,
	    purdueHeader.img_bkg,
	    purdueHeader.img_min, purdueHeader.img_max, 
	    purdueHeader.img_avg, purdueHeader.img_std
	    );
    TextSet(statTextW, outStr, FALSE);
  }
  
} // end function stat_calc_activate



void movie_contour_state(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: movie_contour_state
  // They want to do something with their section and contour plot.

  char *s1;
  int i;
  
  if(widget == movconClearW) {
    stuffDisplayMapBuffer(curBox,editBoxZoom,0);
    DrawEditBoxWindow(NULL, NULL, NULL);
    XFlush(dsply);
  }
  if(widget == movconResetW) {
    stuffDisplayMapBuffer(curBox,editBoxZoom,0);
    DrawEditBoxWindow(NULL, NULL, NULL);
    XFlush(dsply);
    totContours = 0;
    sprintf(contourStr," ");
    TextSet(movconTextW, contourStr, TRUE); 
  }
  else if(widget == movconUseW) {
    if(numCont > 0) {
      for(i=0; i<numCont; i++) {
	conTours[totContours]= contourLevel + i*deltaCont;
	if(totContours < 50)
	  totContours++;
	else 
	  askThem("MAX contour levels exceeded", XmDIALOG_INFORMATION);
      }
    }
    else { // Do individual
      conTours[totContours]= contourLevel;
      if(totContours < 50)
	totContours++;
	else 
	  askThem("MAX contour levels exceeded", XmDIALOG_INFORMATION);
    }
    
    s1= contourStr;
    for(i=0;i<totContours;i++) {
      sprintf(s1,"%d  ",conTours[i]);
      s1= (char *)(contourStr+strlen(contourStr));
    }

    TextSet(movconTextW, contourStr, TRUE);
  }
  else if(widget == movconWriteW) {
    int saveEditBoxZoom;
    float thisContour;
    
    // Get a fresh buffer
    saveEditBoxZoom= editBoxZoom;
    // editBoxZoom = 1;  ONLY if you want to force it to 1
    stuffDisplayMapBuffer(curBox,editBoxZoom,0);
    
    // thisStr = XmTextGetString(movconTextW);
    for(i=0; i<totContours; i++) { // Stuff all
      // sscanf(thisStr,"%d",numStr);
      // thisContour = strtod(numStr,0);
      thisContour = conTours[i];
      printf("%f\n",thisContour);
      
      // Now destructively draw the contour in the window buffer
      drawContour(thisContour,TRUE,curBox,editBoxW,
		  float_image_buffer1,0,0,
		  editBox_display_buffer);
    }
    // XtFree(thisStr);
    
    calcBoxSizeFactor(editBoxZoom);
    writeRAWcontour("contourSection.raw",editBox_display_buffer,
		    (numCols*sizeFactor),(numRows*sizeFactor));
    
    editBoxZoom= saveEditBoxZoom;
  }
  
} // end function movie_contour_state



void drawContour(float contourLvl, int destructive, int thisCurBox,
		 Widget thisWidget, float *thisFloatBuf,
		 int deltaX, int deltaY, BYTE *this_display_buffer)
{
  // function name: drawContour
  // Draw the contour over the image already displayed.

  int i, numPts, numColsTsizeFactor, nCols, nRows;
  int box_width, icx=0, icy=0;
  short *short_display_image;
  short ix[40000], iy[40000], ix1[40000], iy1[40000];
  float *xp, *yp, *thisBuffer;
  float sclx, scly, tour;

  if(this_display_buffer == editBox_display_buffer)
    box_width = BOX_WIDTH;
  else 
    box_width = buffer_width;
  
  cosang = 0.0;     // 90.0 degrees
  sinang = 1.0;     // 90.0 degrees
  
  // ??? deg
  cosang = 1.0;
  sinang = 1.0;
  
  // These are for scalex and scale Y - get from zoomfactor
  calcBoxSizeFactor(editBoxZoom);
  sclx = sizeFactor;
  scly = sizeFactor;
  
  tour = contourLvl;
  numPts = 1;
  
  for(i=0;i<40000;i++) // Init 
    ix[i] = iy[i] = ix1[i] = iy1[i]= 0;
  
  if(shadedMaps && !dispSections) { // SHow the shaded map
    int colOffset, rowOffset;
    
    CK(xp = (float *)malloc(sizeof(float)*(largestDim+1)));
    CK(yp = (float *)malloc(sizeof(float)*(largestDim+1)));
    CK(short_display_image = (short *)
       malloc(sizeof(short)*(largestDim+1)*(largestDim+1)));
    
    // Need numCols and numRows here because we want to use the section
    // contour info
    thisBuffer= (float *)(thisFloatBuf+(cutMaxValues[2]*
					numCols*numRows));
    nCols = numCols;
    nRows = numRows;
    
    contur_a_rob__(thisBuffer,
		  short_display_image,
		  xp, yp,
		  &nCols, &nRows,
		  &sclx, &scly, &tour, ix, iy, ix1, iy1, &numPts,
		  &cosang, &sinang, &icx, &icy);
    
    numColsTsizeFactor = numCols * sizeFactor;
    colOffset = (largestDim-numCols)/2;
    rowOffset = (largestDim-numRows)/2;
    colOffset *= sizeFactor;
    rowOffset *= sizeFactor;
    
    for(i=0; i<numPts-1; i++) { // Draw all points
      
      if(ix[i] >= (numColsTsizeFactor*3) || 
	 ix1[i] >= (numColsTsizeFactor*3)) { // Into next cell
	ix[i]  -= (numColsTsizeFactor*3);
	ix1[i] -= (numColsTsizeFactor*3);
      }
      else if(ix[i] >= (numColsTsizeFactor*2) || 
	      ix1[i] >= (numColsTsizeFactor*2)) { // Into next cell
	ix[i]  -= (numColsTsizeFactor*2);
	ix1[i] -= (numColsTsizeFactor*2);
      }
      else if(ix[i] >= (numColsTsizeFactor) || 
	      ix1[i] >= (numColsTsizeFactor)) { // Into next cell
	ix[i]  -= numColsTsizeFactor;
	ix1[i] -= numColsTsizeFactor;
      }
      
      ix[i]  += colOffset;
      ix1[i] += colOffset;
      iy[i]  += rowOffset;
      iy1[i] += rowOffset;
      
      if(!destructive) { // Just draw with X
	GC thisGC;
	if(contourLvl > breakCont)
	  thisGC = image_gcRED;
	else if(contourLvl == breakCont)
	  thisGC = image_gcBLUE;
	else
	  thisGC = image_gcGREEN;

	XDrawLine(dsply, XtWindow(thisWidget), thisGC,
		  ix[i], iy[i], ix1[i], iy1[i]);
      }
      
      if(destructive) { // Now draw this in the buffer itself
	float slope, intercept;
	int j, y, startX, endX;
	
	// Corrupt it
	// Need to draw a line with value==pixels_return[RED] or
	// [GREEN] in the buffer below based on ix,iy,ix1,iy1
	if(ix1[i]==ix[i])
	  slope = 9999.9;
	else 
	  slope = (float)(iy1[i] - iy[i]) / (float)(ix1[i] - ix[i]);
	intercept = (float)iy[i] - (float)(slope * (float)ix[i]);
	startX = MIN(ix[i],ix1[i]);
	endX   = MAX(ix[i],ix1[i]);
	
	startX = MAX(startX,0);
	endX   = MIN(endX,numColsTsizeFactor);
	
	for(j=startX;j<=endX;j++) { // Do all X
	  y= (int)((slope * (float)j) + intercept);
	  y = MAX(y,0);
	  y = MIN(y,numColsTsizeFactor);
	  
	  if(contourLvl > 0.0)
	    this_display_buffer[(y*box_width)+j] = pixels_return[RED];
	  else if(contourLvl == 0.0)
	    this_display_buffer[(y*box_width)+j] = pixels_return[BLUE];
	  else
	    this_display_buffer[(y*box_width)+j] = pixels_return[GREEN];
	}
	      
      } 
    }
      
    if(destructive) { // Draw it
      DrawEditBoxWindow(NULL, NULL, NULL);
      XFlush(dsply);
    }
    
    free(xp);
    free(yp);
    free(short_display_image);
    
  }
  else { // Do sections
    
    nCols = numCols;
    nRows = numRows;
    
    numColsTsizeFactor = numCols * sizeFactor;
    
    CK(xp = (float *)malloc(sizeof(float)*(numCols+1)));
    CK(yp = (float *)malloc(sizeof(float)*(numRows+1)));
    CK(short_display_image = (short *)malloc(sizeof(short)*
					     (numCols+1)*
					     (numRows+1)));
    thisBuffer= (float *)(thisFloatBuf+(thisCurBox*
					numCols*numRows));
    
    contur_a_rob__(thisBuffer,
		  short_display_image,
		  xp, yp,
		  &nCols, &nRows,
		  &sclx, &scly, &tour, ix, iy, ix1, iy1, &numPts,
		  &cosang, &sinang, &icx, &icy);
    
    for(i=0; i<numPts-1; i++) { // Draw all points
      
      if(ix[i] >= (numColsTsizeFactor*3) || 
	 ix1[i] >= (numColsTsizeFactor*3)) { // Into next cell
	ix[i]  -= (numColsTsizeFactor*3);
	ix1[i] -= (numColsTsizeFactor*3);
      }
      else if(ix[i] >= (numColsTsizeFactor*2) || 
	      ix1[i] >= (numColsTsizeFactor*2)) { // Into next cell
	ix[i]  -= (numColsTsizeFactor*2);
	ix1[i] -= (numColsTsizeFactor*2);
      }
      else if(ix[i] >= (numColsTsizeFactor) || 
	      ix1[i] >= (numColsTsizeFactor)) { // Into next cell
	ix[i]  -= numColsTsizeFactor;
	ix1[i] -= numColsTsizeFactor;
      }
      
      ix[i]  += deltaX;
      ix1[i] += deltaX;
      iy[i]  += deltaY;
      iy1[i] += deltaY;
      
      if(!destructive) { // Just draw with X
	GC thisGC;
	
	if(contourLvl > breakCont)
	  thisGC = image_gcRED;
	else if(contourLvl == breakCont)
	  thisGC = image_gcBLUE;
	else
	  thisGC = image_gcGREEN;
	
	XDrawLine(dsply, XtWindow(thisWidget), thisGC,
		  ix[i], iy[i], ix1[i], iy1[i]);
      }
      
      if(destructive) { // Now draw this in the buffer itself
	float slope, intercept;
	int j, y, startX, endX;
	      
	// Corrupt it
	// Need to draw a line with value==pixels_return[RED] or
	// [GREEN] in the buffer below based on ix,iy,ix1,iy1
	if(ix1[i]==ix[i])
	  slope = 9999.9;
	else 
	  slope = (float)(iy1[i] - iy[i]) / (float)(ix1[i] - ix[i]);
	intercept = (float)iy[i] - (float)(slope * (float)ix[i]);
	startX = MIN(ix[i],ix1[i]);
	endX   = MAX(ix[i],ix1[i]);
	
	startX = MAX(startX,0);
	endX   = MIN(endX,numColsTsizeFactor);
	
	for (j=startX; j<=endX; j++) { // Do all X
	  y = (int)((slope * (float)j) + intercept);
	  y = MAX(y,0);
	  y = MIN(y,numColsTsizeFactor);
	  
	  if(contourLvl > 0.0)
	    this_display_buffer[(y*box_width)+j] = pixels_return[RED];
	  else if(contourLvl == 0.0)
	    this_display_buffer[(y*box_width)+j] = pixels_return[BLUE];
	  else
	    this_display_buffer[(y*box_width)+j] = pixels_return[GREEN];
	}
	
      }  
    }
    
    if(destructive) { // Draw it
      DrawEditBoxWindow(NULL, NULL, NULL);
      XFlush(dsply);
    }
    
    free(xp);
    free(yp);
    free(short_display_image);
  }
  
} // end function drawContour



void fft_dsp_c(float *fft, BYTE *img, 
	       int fftSize1, int fftSize2,
	       int ncol, int nrow, 
	       float thresh_hi, float thresh_lo,
	       int hiPixel, int loPixel, 
	       int linear, int amps)
{
  // function name: fft_dsp_c

  // Derived from the TSB fortran FFT_DSP(). We're trying to do
  // 1024x1024 FFT's now and display them in 512x512. Also be aware
  // of the problesm involving 256x256 and 128x128 fft's...
  //
  // ************************************************************************
  // *                                                                      *
  // *                *****     SUBROUTINE FFT_DSP_rob *****                *
  // *                                                                      *
  // *   THE INTENSITIES FROM EMIMGFFT ARE DISPLAYED USING A 64 GREY-LEVEL  *
  // *   SCALE.  THE TOP TWO BITS ARE USED BY THE MASK PROGRAM FOR          *
  // *     SUBSEQUENT PSEUDO-OPTICAL FILTERING (USING EMFFT).               *
  // *                                                                      *
  // *   ONLY VALUES BETWEEN THRESH_LO AND THRESH_HI ARE MAPPED INTO THE    *
  // *    COLOR TABLE, WITH OUTLIERS SET TO THE OUTSIDE LIMITS OF THE CT.   *
  // *                                                                      *
  // *   LAST MODIFIED  3-MAR-1992 BY T. BAKER.                             *
  // ************************************************************************

  int i, j, fftSize2D2, fftSize2T2, jT2;
  int lastRow, lastCol, ii, jj, thisOutRow;
  float xmult, ymult, testlo, testhi;
  float slopeb, value;
  
//  printf("fftSizes [%d,%d]\n",fftSize1,fftSize2);
//  printf("ncol,nrow[%d,%d]\n",ncol,nrow);

  testhi = hiPixel;
  testlo = loPixel;
  
  xmult = (float)ncol / (float)fftSize1;
  ymult = (float)nrow / (float)fftSize2;

  lastCol = MAX((int)xmult,1);
  lastRow = MAX((int)ymult,1);

  
  fftSize2D2 = fftSize2 / 2;
  fftSize2T2 = fftSize2 * 2;

  slopeb = (testhi-testlo)/(thresh_hi-thresh_lo);
  
  // FILL UP IMG ARRAY AFTER SCALING

  if(xmult >= 1.0 && ymult >= 1.0) {
    // The FFT is smaller or equal to the display size
    
    for(i=0;i<(fftSize1);i++) { // do i = 1,((fftSize2/2)+1)
      
      if(i==0)
	thisOutRow = fftSize2D2 - 1;
      else if(i <= fftSize2D2)
	thisOutRow = fftSize2D2 - 1 + i;
      else
	thisOutRow = i - fftSize2D2 - 1;
      thisOutRow = MAX(thisOutRow,0);
      thisOutRow = MIN(thisOutRow,(fftSize1-1));
      
      for(j=0;j<(fftSize2D2);j++) { // do j = 1,jtest ! right side of fft
	
	jT2 = 2 * j;
	fft[jT2+1+(i*fftSize2T2)] *= 
	  tempFactors[(int)(i*ymult)][2*(int)(j*xmult)];
	fft[jT2+(i*fftSize2T2)] *= 
	  tempFactors[(int)(i*ymult)][2*(int)(j*xmult)];
	value = fft[jT2+1+(i*fftSize2T2)] * fft[jT2+1+(i*fftSize2T2)] + 
	  fft[jT2+(i*fftSize2T2)] * fft[jT2+(i*fftSize2T2)];
	if(amps != 0) value = sqrt(value);
	if(linear==0 && 
	   value != 0.0)value = log10(value);
	
	value = (value - thresh_lo)*slopeb + testlo;
	
	if(value < testlo) 
	  value = testlo;
	if(value > testhi) 
	  value = testhi;
	
	for(ii=0; ii<lastRow; ii++)    // Expand rows
	  for(jj=0; jj<lastCol; jj++)  // Expand cols
	    img[((int)(((thisOutRow*ymult)+ii)*ncol)+
		 (int)((j+fftSize2D2)*xmult)+jj)] = (BYTE)(value);
	
      }
	  
      
      for(j=(fftSize2/2);j<(fftSize2);j++) {
	// do j = (jtest+1),fftSize1 ! left side of fft
	jT2 = 2 * j;
	fft[jT2+1+(i*fftSize2T2)] *= 
	  tempFactors[(int)(i*ymult)][2*(int)(j*xmult)];
	fft[jT2+(i*fftSize2T2)] *= 
	  tempFactors[(int)(i*ymult)][2*(int)(j*xmult)];
	value = fft[jT2+1+(i*fftSize2T2)] * fft[jT2+1+(i*fftSize2T2)] + 
	  fft[jT2+(i*fftSize2T2)] * fft[jT2+(i*fftSize2T2)];
	
	if(amps != 0) value = sqrt(value);
	if(linear==0 &&
	   value != 0.0)value = log10(value);
	
	value = ((value - thresh_lo)*slopeb) + testlo;
	
	if(value < testlo) value = testlo;
	if(value > testhi) value = testhi;
	
	for(ii=0;ii<lastRow;ii++)    // Expand rows
	  for(jj=0;jj<lastCol;jj++)  // Expand cols
	    img[((int)(((thisOutRow*ymult)+ii)*ncol)+
		 (int)((j-fftSize2D2)*xmult)+jj)] = (BYTE)(value);
	
      }
    }
  }
  else if(xmult < 1.0 && ymult < 1.0) {
    // The FFT is larger than the display size
    int step, fftSize2D2mult;
    float totValue, numPnts;
    
    step = 1.0 / xmult;
    fftSize2D2mult = fftSize2D2 * xmult;
    for(i=0; i<fftSize1; i+=step) { // do all rows
	
      if(i==0)
	thisOutRow = fftSize2D2mult - 1;
      else if(i <= fftSize2D2)
	thisOutRow = fftSize2D2mult - 1 + (i*xmult);
      else
	thisOutRow = (i*xmult) - fftSize2D2mult - 1;
      thisOutRow = MAX(thisOutRow,0);
      thisOutRow = MIN(thisOutRow,((int)(fftSize1*xmult)-1)); 
      
      for (j=0; j<fftSize2D2; j+=step) { // do right side of fft
	
	numPnts  = 0.0;
	totValue = 0.0;
	for (ii=0; ii<step; ii++) { // Expand it out	  
	  for (jj=0; jj<step; jj++) { // Expand cols
	    jT2 = 2 * (j+jj);
	    fft[jT2 + 1 + (i+ii)*fftSize2T2] *= 
	      tempFactors[(int)(i*ymult)][2*(int)(j*xmult)];
	    fft[jT2 + (i+ii)*fftSize2T2] *= 
	      tempFactors[(int)(i*ymult)][2*(int)(j*xmult)];
	    value = fft[jT2+1+(i+ii)*fftSize2T2] * fft[jT2+1+(i+ii)*fftSize2T2]
	      + 
	      fft[jT2+(i+ii)*fftSize2T2] * fft[jT2+(i+ii)*fftSize2T2];
	    
	    if(amps != 0) value = sqrt(value);
	    if(linear==0 && 
	       value != 0.0)value = log10(value);
	    
	    value = ((value - thresh_lo)*slopeb) + testlo;
	    
	    if(value < testlo) 
	      value = testlo;
	    if(value > testhi) 
	      value = testhi;
	    
	    totValue += value;
	    numPnts  += 1.0;
	  }
	}
	
	img[thisOutRow*ncol + (int)((j+fftSize2D2)*xmult)] = (BYTE)(totValue/numPnts);
	
      }
	  
      
      for (j=fftSize2D2; j<fftSize2; j+=step) { // do left side of fft
	      
	numPnts  = 0.0;
	totValue = 0.0;
	for (ii=0; ii<step; ii++) { // Expand it out
	  for (jj=0; jj<step; jj++) { // Expand cols
	    jT2 = 2 * (j+jj);
	    fft[jT2+1 + (i+ii)*fftSize2T2] *= 
	      tempFactors[(int)(i*ymult)][2*(int)(j*xmult)];
	    fft[jT2 + (i+ii)*fftSize2T2] *= 
	      tempFactors[(int)(i*ymult)][2*(int)(j*xmult)];
	    value = fft[jT2+1+(i+ii)*fftSize2T2] * fft[jT2+1+(i+ii)*fftSize2T2]
	      + 
	      fft[jT2+(i+ii)*fftSize2T2] * fft[jT2+(i+ii)*fftSize2T2];
	    
	    if(amps != 0) value = sqrt(value);
	    if(linear==0 && 
	       value != 0.0)value = log10(value);
	    
	    value = (value - thresh_lo)*slopeb + testlo;
	    value = MAX(value, testlo);
	    value = MIN(value, testhi);
	    
	    totValue += value;
	    numPnts  += 1.0;
	  }
	}
	
	img[thisOutRow*ncol + (int)((j-fftSize2D2)*xmult)] = (BYTE)(totValue/numPnts);
	
      }  
    }
    
  }
  else
    askThem("[FFT_DSP_C] Don't know how to do this...", XmDIALOG_QUESTION);
  
} // end function fft_dsp_c



void calcFFT(int doFFT, int thisBox, int tryBackground)
{
  // function name: calcFFT
  // Stuff the buffer and calc the FFT. Display based on scroll values.

  int i, j, thisPixel;
  int arg1, arg2, arg3, arg4, jindx;
  int iTfftSize2;
  int strtAvg, endAvg, thisImg, thisNimg, useRectangle;
  int ksign, iform, linear;
  int thisMaxInt, thisMinInt, sliderSize, hiPixel, loPixel;
  float thisMin, thisMax, maxF, minF, offset, recipThisNimg;
  float theoreticalMax, theoreticalMin;
  float scaleFactor, thisFloat;
  float recipSize;
  float thisBackground;
  float thisVal, thisValP1;
  float *temp_buffer, *char_buffer, *thisBuffer;

  numCols = purdueHeader.numCols;
  numRows = purdueHeader.numRows;

  useRectangle = rectBox.drawable;

  if(useRectangle) {
    numRows = rectBox.lowy - rectBox.y;
    numCols = rectBox.lowx - rectBox.x;
  }
  
  if(numCols > 1024 || numRows > 1024) {
    askThem("Image sizes greater than 1024x1024 not yet implemented.",
	    XmDIALOG_QUESTION);
    return;
  }

  CK(temp_buffer= (float *)malloc(
			 sizeof(float)*numCols*numRows));

  if(doFFT) { // First time - calc the fft
    
    // We've got a fresh FFT here so no shifting of the origin
    // has occurred yet.
    apOrigXtot = apOrigYtot = 0.0;
    
    // Find out what fftSize they want and get the reciprical for
    // use later
    recipSize = (float)fftSize / (float)idim1;
    
    init_floatarr(fftImage_gold, idim1*idim2*2, 0.0);   // initialize arrays
    init_floatarr(fftImage_buffer2, idim1*idim2*2, 0.0);

    CK(char_buffer= (float *)malloc(sizeof(float)*numCols*numRows));
    
    if(averageFFT) { // Setup for doing all of the images
      strtAvg = 0;
      endAvg  = purdueHeader.nimg;
      thisNimg = 0;
      // Find out how many we've got
      for(i=strtAvg;i<endAvg;i++)
	if(!deletePackedImage[i])
	  thisNimg++;
      if(thisNimg<=0)
	return;
    }
    else { // Just do this one
      strtAvg = imgNum-1;
      endAvg  = imgNum;
      thisNimg = 1;
    }

    recipThisNimg = 1.0 / (float)thisNimg;
    
    for (thisImg=strtAvg; thisImg<endAvg; thisImg++) { // Do all images wanted
	
      if(!averageFFT || (averageFFT && !deletePackedImage[thisImg])) {
	  
	if(fileType==PACKEDPORTABLE || fileType==MAP)
	  sprintf(outStr,"%d",thisImg+1);
	else
	  sprintf(outStr,"%d",1);
	TextSet(fftImgNumW, outStr, FALSE);
	TextSet(rotfltImgNumW, outStr, TRUE);

	thisBuffer= (float *)packImg[(thisImg)]; 
	
	if(useRectangle)
	  stuffRectangleBuffer(char_buffer, image_fileName.addr);
	
	for(i=0;i<numRows;i++) {
	  for(j=0;j<numCols;j++) {
	    if(useRectangle)
	      temp_buffer[(i*numCols)+j]= char_buffer[(i*numCols)+j];
	    else if(fileType==PACKEDPORTABLE)
	      temp_buffer[(i*numCols)+j]= thisBuffer[(i*numCols)+j];
	    else if(fileType==MAP || fileType==FLOATMAP)
	      temp_buffer[(i*numCols)+j]= (float)
		float_image_buffer1[(thisImg*numRows*numCols)+
				     (i*numCols)+j];
	    else { // Just get it from here
	      temp_buffer[(i*numCols)+j]= int_image_buffer1[(i*image_width)+j];
	      char_buffer[(i*numCols)+j]= temp_buffer[(i*numCols)+j];
	    }
	  }
	}
	      
	if(tryBackground) { // Try this one
	  thisBackground = fftBkgValue;
	  printf("Trying bkg[%f]\n",thisBackground);
	}
	else { // Get global
	  if(fileType==PACKEDPORTABLE && !useRectangle) {
	    // Don't do this it messes up gap.pif
	    // printf("ibkg[%f] pI[%f]\n",
	    // purdueHeader.img_bkg,(*packImg[thisImg]));
	    // purdueHeader.img_bkg = (float)(*packImg[thisImg]);

	    if(packRadius > 0)
	      purdueHeader.img_bkg = (int)(*packImg[thisImg]);
	  }
	  else if((fileType==EMPORTABLE || fileType==BYTEPORTABLE ||
		   fileType==PORTABLE ) && !useRectangle) { // Do nothing
	    //  just use what we read from the file
	    purdueHeader.img_bkg = pifHeader.img_bkg;
	    // purdueHeader.img_bkg= (short)purdueHeader.img_avg;
	  }
	  else if(fileType==MAP || fileType==FLOATMAP ) {
	    // purdueHeader.img_bkg= 0; Don't do this
	    // EMPORTABLEI4 images have
	    // a valid background
	  }
	  else { // Unknown source or userectangle
	    if(useRectangle) {
	      int saveFileType;
	      saveFileType = fileType;
	      fileType = RECTPACKED;
	      calcStats(char_buffer, &purdueHeader.img_bkg, 
			&purdueHeader.img_min, &purdueHeader.img_max, 
			&purdueHeader.img_avg, &purdueHeader.img_std,
			numCols, numRows, numCols, fileType, bkgrndMethod,
		    boxMethod, packRadius, boxRadius, featherRadius,
		    recipSizeFactor, tryColors);
	      fileType = saveFileType;
	    }
	    else { // Else just use what we read from the file
	      purdueHeader.img_bkg = pifHeader.img_bkg;
	    }
	  }
	  thisBackground = purdueHeader.img_bkg;
	  
//	  printf("Using bkg[%f]\n",thisBackground);
	  
	  // Setup the FFT Background widget
	  fftBkgValue = purdueHeader.img_bkg;
	  sliderSize = 5;
	  XtSetArg(arglst[0], XmNmaximum, (int)(purdueHeader.img_max+
						(float)sliderSize));
	  XtSetArg(arglst[1], XmNminimum, (int)
		   (purdueHeader.img_min));
	  XtSetArg(arglst[2], XmNsliderSize, sliderSize);
	  XtSetArg(arglst[3], XmNvalue, (int)fftBkgValue);
	  XtSetArg(arglst[4], XmNpageIncrement, sliderSize);
	  XFlush(dsply);
	  
	  sprintf(outStr,"%d",fftBkgValue);
	}
	
	ksign = 1;  // Do a "forward" FFT
	iform = 1;  // Data is imaginary

	if(fileType == MAP) { // Do for MAP
	  thisBackground = purdueHeader.img_bkg;
	  r4_img_fft_fill__((float_image_buffer1+
				    (thisImg*numCols*numRows)), 
				   &numCols, &numRows,
				   fftImage_buffer2, 
				   &fftSize, &fftSize, 
				   &thisBackground); 
	}
	else { // Do for images
	  r4_img_fft_fill__(temp_buffer, &numCols, &numRows, 
				   fftImage_buffer2, 
				   &fftSize, &fftSize, 
				   &thisBackground);
	}

	// Check to see if they want to stop
	// Frame the location of the stop button
	DrawX(XtWindow(fftStopW), image_gcRED, 16, 16, 4);
	XFlush(dsply);
	
	// See if they want to bugout or not
	if(XCheckTypedWindowEvent(dsply,
				  XtWindow(fftStopW),
				  ButtonPress,
				  &event)) { // STOP!!!
	  XFlush(dsply);
	  free(char_buffer);
	  free(temp_buffer);
	  return;
	}
	else { // OK to continue
	}

	// Now do the FFT
	fft_2d_rob__(fftImage_buffer2, &fftSize, &fftSize,
		     &ksign, &iform);
	
	// This is curious - We need to have the FFT buffer 
	// built by stuffing this way for the Fourier Averaging
	// rtn but we need it the other way for the rtn that
	// calculates ax,ay,bx,by (latcen2)

	if(!averageFFT) { // Don't do for averaged FFT
	  // Let's try putting this into TSB space and 
	  // see if it helps
	  for (i=0; i<numRows/2; i++) {
	    float tmpB;
	    for (j=0; j<numCols; j++) {
	      tmpB = temp_buffer[(i*numCols)+j];
	      temp_buffer[(i*numCols)+j]= 
		temp_buffer[((numRows-i-1)*numCols)+j] ;
	      temp_buffer[((numRows-i-1)*numCols)+j] = tmpB ; 
	    }
	  }
	  r4_img_fft_fill__(temp_buffer, &numCols, &numRows, 
				   fftImage_fourAver, 
				   &fftSize, &fftSize, 
				   &thisBackground);
	  
	  fft_2d_rob__(fftImage_fourAver, &fftSize, &fftSize,
		       &ksign, &iform);
	}
	      
	// This is only valid when averaging now
	// ctfEnhance = FALSE;
	
	// Now accumulate this one into the GOLD
	for (i=0; i<fftSize; i++) {

	  iTfftSize2 = i * (fftSize*2);  
	  for (j=0; j<(fftSize*2); j+=2) { // Accumulate all
	    
	    if(averageFFT) { // average
	      // This is only valid when averaging now ... 
	      // what follows is an "incoherent" average of
	      // the data - makes the CTF rings look pretty.
	      // ctfEnhance = TRUE;
	      
	      if(ctfEnhance) { // DO this kludge - why does it enhance???
		thisVal   = fftImage_buffer2[iTfftSize2+j];
		thisValP1   = fftImage_buffer2[iTfftSize2+j+1];
		
		fftImage_gold[(iTfftSize2+j)] += 
		  (((thisVal*thisVal)+(thisValP1*thisValP1)) * 
		   recipThisNimg);
		fftImage_gold[iTfftSize2+j+1] = 0.0; 
	      }
	      else { // "normal"
		thisVal = fftImage_buffer2[iTfftSize2+j];
		fftImage_gold[iTfftSize2+j] += thisVal * recipThisNimg;
		thisValP1   = fftImage_buffer2[iTfftSize2+j+1];
		fftImage_gold[iTfftSize2+j+1] += thisValP1 * recipThisNimg;
	      }
	    }
	    else { // Individual
	      fftImage_gold[iTfftSize2+j]   = fftImage_buffer2[iTfftSize2+j];
	      fftImage_gold[iTfftSize2+j+1] = fftImage_buffer2[iTfftSize2+j+1];
	    }
	    
	    thisVal = 
	      fftImage_gold[iTfftSize2+j]   * fftImage_gold[iTfftSize2+j] +
	      fftImage_gold[iTfftSize2+j+1] * fftImage_gold[iTfftSize2+j+1];
	    
	    if(thisVal < 0.0)
	      printf("Found one [%d][%d]-[%f]\n",i,j,thisVal);
	    
	  }
	}
	
      }
      
    }
    
    free(char_buffer);
    
  }
  
  thisMin = R4_MAX;
  thisMax = R4_MIN;
  recipSize = 512.0 / (float)idim1;

  for (i=0; i<fftSize; i++) {
    for (j=0; j<fftSize*2; j+=2) { // Do all image
      float thisValue;
      thisPixel = i*fftSize*2 + j;
      thisValue = 
	fftImage_gold[thisPixel]   * fftImage_gold[thisPixel] +
	fftImage_gold[thisPixel+1] * fftImage_gold[thisPixel+1];
      
      // NOTE - Shouldn't we just apply the tempFactor here ???
      thisValue *= tempFactors[(int)(i*recipSize)][(int)(j*recipSize)];
      if(tempFactors[(int)(i*recipSize)][(int)(j*recipSize)] < 0.0) {
	printf("Found one [%d][%d]-[%f][%f]\n",
	       i,j,thisValue,
	       tempFactors[(int)(i*recipSize)][(int)(j*recipSize)]);
	printf("temp[i,j] = [%d,%d]\n",
	       (int)(i*recipSize), (int)(j*recipSize));
      }
      if(fftDispType==AMPLITUDE)
	thisValue = sqrt(thisValue);
      if(fftType == LOGFFT && thisValue > 0.0)
	thisValue = log10(thisValue);
      thisMax = MAX(thisValue,thisMax);
      thisMin = MIN(thisValue,thisMin);
    }
  }
  
  if(averageFFT)  // Do special
    TextSet(fftImgNumW, "Aver", FALSE);
  
  // Now setup the scroll bars to hold the new max/min
  // We can't exceed what a 32 bit signed integer will hold!! 
  // - Really 2.17 Gig but let's be safe... 
  
  theoreticalMax = thisMax;  // Capture these so we can stuff the label
  theoreticalMin = thisMin;
  if (thisMin == 0.0)
    thisMin = 1.0;
  if(thisMax <= thisMin)
    thisMax = thisMin + 1.0;
  if(doFFT && !fftMinMaxLock) { // OK to change fft min max value
    globMinFFTvalue = minFFTvalue = thisMin;
    globMaxFFTvalue = maxFFTvalue = thisMax;
  }
  
  // OK so suppose they've changed to Amplitudes from Intensities
  // we need to adjust the min/max values in case the range is less
  // now
  if(!fftMinMaxLock) { // keep it in range
    minFFTvalue = MAX(minFFTvalue,thisMin);
    maxFFTvalue = MIN(maxFFTvalue,thisMax);
  }
  
  thisMin *= (fftType==LINEARFFT) ? 0.001 : 100.0;
  thisMax *= (fftType==LINEARFFT) ? 0.001 : 100.0;
  
  thisMaxInt = (int)MIN(thisMax,1800000000.0);   // Why?????
  imageMaxFFTvalue = thisMax;
  thisMinInt = (int)thisMin;
  
  if(thisMaxInt <= thisMinInt)
    thisMaxInt = thisMinInt + 1;
  
  sliderSize = (thisMaxInt - thisMinInt) / 100;
  sliderSize = MAX(sliderSize,1);
  
  XtSetArg(arglst[0], XmNmaximum, (int)(thisMaxInt+(float)sliderSize));
  XtSetArg(arglst[1], XmNminimum, thisMinInt);
  XtSetArg(arglst[2], XmNsliderSize, sliderSize);
  
  if(!fftMinMaxLock) { // OK to change fft min max value
    if(fftType==LINEARFFT) // 1/1000th
      XtSetArg(arglst[3], XmNvalue, (int)(minFFTvalue/1000.0));
    else 
      XtSetArg(arglst[3], XmNvalue, (int)(minFFTvalue*100.0));
    XtSetArg(arglst[4], XmNpageIncrement, sliderSize);
    if(sliderSize > 11.0)
      XtSetArg(arglst[5], XmNincrement, (int)(sliderSize/10.0));
    else 
      XtSetArg(arglst[5], XmNincrement, 1.0);
    XtSetValues(minFFTscrollW, arglst, 6);
  }
  else { // Not OK
    XtSetArg(arglst[3], XmNpageIncrement, sliderSize);
    if(sliderSize > 11.0)
      XtSetArg(arglst[4], XmNincrement, (int)(sliderSize/10.0));
    else 
      XtSetArg(arglst[4], XmNincrement, 1);
    XtSetValues(minFFTscrollW, arglst, 5);
  }
  
  XtSetArg(arglst[0], XmNmaximum, (int)(thisMaxInt+(float)sliderSize));
  XtSetArg(arglst[1], XmNminimum, thisMinInt);
  XtSetArg(arglst[2], XmNsliderSize, sliderSize);
  if(!fftMinMaxLock) { // OK to change fft min max value
    if(fftType==LINEARFFT) // 1/1000th
      XtSetArg(arglst[3], XmNvalue, (int)(maxFFTvalue/1000.0));
    else 
      XtSetArg(arglst[3], XmNvalue, (int)(maxFFTvalue*100.0));
    XtSetArg(arglst[4], XmNpageIncrement, sliderSize);
    if(sliderSize > 11.0)
      XtSetArg(arglst[5], XmNincrement, (int)(sliderSize/10.0));
    else 
      XtSetArg(arglst[5], XmNincrement, 1);
    XtSetValues(maxFFTscrollW, arglst, 6);
  }
  else { // Not OK
    XtSetArg(arglst[3], XmNpageIncrement, sliderSize);
    if(sliderSize > 11.0)
      XtSetArg(arglst[4], XmNincrement, (int)(sliderSize/10.0));
    else 
      XtSetArg(arglst[4], XmNincrement, 1);
    XtSetValues(maxFFTscrollW, arglst, 5);
  }
  
  XSync(dsply,FALSE);
  XFlush(dsply);
  
  // Stuff the current values of these
  sprintf(outStr,"%.2f",minFFTvalue);   
  doLabel(outStr, minFFTW, DEFAULT);

  sprintf(outStr,"%.2f",maxFFTvalue);     
  doLabel(outStr, maxFFTW, DEFAULT);
  
  // Now stuff the theoretical MAX/MIN labels
  sprintf(outStr,"Range Min:  %.2f",theoreticalMin);  
  doLabel(outStr, theominFFTW, DEFAULT); 

  sprintf(outStr,"Max:  %.2f",theoreticalMax);    
  doLabel(outStr, theomaxFFTW, DEFAULT);
  
  XSync(dsply,FALSE);
  XFlush(dsply);
  
  
  // STuff the mirror buffer - we don't have to sew it together here
  // let's just do it for i=0 to < idim1 and idim2 etc...
  for (i=0; i<idim1*idim2*2; i++)
    fftImage_buffer2[i] = fftImage_gold[i] ;
  
  if(lopHipSwitch) { // Just do a lop hip thing
    int thisTrue = 1;
    float res_max, res_min, sizeF, cutmaxLo, cutmaxHi;
    
    cutmaxLo  = wstrtod(lopTextW);
    cutmaxHi  = wstrtod(hipTextW);
    sizeF = (float)fftSize/512.0;
    res_max =  sqrt((double)lowResoFilter);
    res_max *= sizeF;
    res_min =  sqrt((double)hiResoFilter);
    res_min *= sizeF;
    
    fft_lop__(fftImage_buffer2, &fftSize, &fftSize, 
	      &res_max, &cutmaxLo, &thisTrue);
    fft_hip__(fftImage_buffer2, &fftSize, &fftSize, 
	      &res_min, &cutmaxHi, &thisTrue);
    
  }
  else { // Normal
    if(filterToggle && pickedLatPts) { // filter it
      
      if(!layerLines) { // Do a 2d lattice
	short *thisMask;
	
	CK(thisMask = (short *)malloc(sizeof(short)*fftSize*fftSize*4));
	
	arg1 = arg2 = fftSize;
	
	fft_fltmsk_rob__(fftImage_buffer2, thisMask, &arg1, &arg2,
			&axL, &ayL, &bxL, &byL,
			&filtSizeValue, &icut,  // icut ==1 for soft cutoff
			&lowResoFilter, &hiResoFilter);
	
	free(thisMask);
      }
      else if(pickedLatPts) { // DO layer lines
	float space, fall, slope, intercept;
	
	if(latLineOrder > 0) { // OK to do
	  fall  = (float)filtSizeValue;
	  // printf("fall = %f\n",fall);

	  slope = ((float)
		   ((256-latPt[0].measuredY)-
		    (256-latPt[1].measuredY)))/
	    ((float)(latPt[0].measuredX-latPt[1].measuredX));

	  intercept = (256 - latPt[0].measuredY) - slope*(latPt[0].measuredX - 256);

	  if(pickLattice != -99 && !latValChanged) { // calc it
	    latScrollValue[1] = 
	      dstar = dstarSave = (fabs((float)(intercept)*
					cos(atan2(slope,1.0))) / 
				   (float)latLineOrder);    
	    latScrollValue[3]= omega= omegaSave = 
	      rad2deg(atan2((0.0-slope),1.0));
	    latReset_activate(NULL, NULL, NULL); 
	  }
	  else { // Get from scroll
	    dstar = latScrollValue[1]; 
	    omega = latScrollValue[3]; 
	  }
	  space = dstar;
	  // fltllmsk needs these in degrees not radians
	  slope = omega;
	  slope = 0.0 - omega;  // I think this is correct ???
	  // Try this
	  slope = 0.0 - deg2rad(omega);
	  
	  fft_fltllmsk_rob__(fftImage_buffer2, 
			     &fftSize, &fftSize, 
			     &slope, 
			     &lowResoFilter, &hiResoFilter,
			     &space, &fall, &icut);
	  
	}
	else askThem("Must have a positive Lattice Line Order.",
		     XmDIALOG_INFORMATION);
	
      }
      
    }
  }
  
  // Stuff (filtered) buffer back for display
  arg1 = fftSize;
  arg2 = fftSize;
  arg3 = 1024;
  arg4 = 1024;

  // Try this with fftImage_buffer2 after we clean up how it's stuff
  // in the sew-together stuff above
  loPixel = 8;
  hiPixel = tryColors-2;
  
  linear = (fftType==LINEARFFT) ? -1 : 0; // 0 --> log
  
  if(disp1024fft && fftSize == 1024) { // display it
    int i;
    BYTE *byte_inCore_buffer;
    byte_inCore_buffer = (BYTE *)malloc(sizeof(BYTE)*1024*1024);
    if(int_image_buffer1 != NULL)
      free(int_image_buffer1);
    int_image_buffer1 = (unsigned short *)
      malloc(defDispBufSize*defDispBufSize*sizeof(unsigned short) + 1);
    freeNullFloat(&inCore_buffer);
    CK(inCore_buffer = (float *)malloc(sizeof(float)*1024*1024));
    imgInCore = TRUE;
    file_width = 1024;
    file_height = 1024;
    fullScreen= TRUE;
    
    fft_dsp_c(fftImage_buffer2, 
	      byte_inCore_buffer,
	      arg1, arg2, arg3, arg4,
	      maxFFTvalue, minFFTvalue,
	      hiPixel, loPixel,
	      linear,
	      (fftDispType==AMPLITUDE)?TRUE:FALSE);
    
    for(i=0;i<(1024*1024);i++)
      inCore_buffer[i] = byte_inCore_buffer[i];

    free(byte_inCore_buffer);
    virtual_read_activate(-99,-99);
    
  }

  // Stuff (filtered) buffer back for display
  arg1 = fftSize;
  arg2 = fftSize;
  arg3 = 512;
  arg4 = 512;
  
  // Try this with fftImage_buffer2 after we clean up how it's stuff
  // in the sew-together stuff above
  loPixel = 8;
  hiPixel = tryColors-2;
  
  linear = (fftType==LINEARFFT) ? -1 : 0; // 0 --> log

//  printf("maxFFT[%f]  min[%f]\n",maxFFTvalue,minFFTvalue);
  fft_dsp_c(fftImage_buffer2, 
	    fft_buffer1,
	    arg1, arg2, arg3, arg4,
	    maxFFTvalue, minFFTvalue,
	    hiPixel, loPixel,
	    linear,
	    (fftDispType==AMPLITUDE)?TRUE:FALSE);
  
  for(i=0; i<512*512; i++)
    fft_gold_buffer1[i] = fft_buffer1[i];
  
  if(backTransform) { // Show them what the back transform is
    float *fftImage_copy;
    // Now put it back in "normal" space
    ksign = 0-1;     // Do a "reverse" FFT
    iform = 1;       // Data is imaginary
    
    // Back transform into a copy of the filtered FFT just in case
    // they want to do Fourier Averaging or etc
    CK(fftImage_copy= (float *)malloc(sizeof(float)*idim1*idim2*2));
    
    for(i=0; i<idim1*idim2*2; i++)
      fftImage_copy[i] = fftImage_buffer2[i] ;
    
    fft_2d_rob__(fftImage_copy, &fftSize, &fftSize, &ksign, &iform);
    
    // Let's look at what scaling needs done
    scaleFactor = 0.0;
    findMaxMin(fftImage_copy, &maxF, &minF, 
	       &offset, fftSize, fftSize);
    
    // Defloat it
    maxF += purdueHeader.img_bkg;
    minF += purdueHeader.img_bkg;
    
    if(applyCTF) { // Do for packed boxes
      if(pifMode==6) { // byte
	if(maxF==minF)
	  scaleFactor= 1.0;
	else 
	  scaleFactor = 255.0 / (maxF-minF);
      }
      else { // short
	if(maxF==minF)
	  scaleFactor= 1.0;
	else 
	  scaleFactor = 65500.0 / (maxF-minF);
      }
    }
    else { // Normal
      if(pifMode==6) { // byte
	if(maxF==minF)
	  scaleFactor= 1.0;
	else 
	  scaleFactor = (tryColors - 10.0) / (maxF-minF);
      }
      else { // short
	if(maxF==minF)
	  scaleFactor= 1.0;
	else 
	  scaleFactor = (tryColors - 10.0) / (maxF-minF);
      }
    }
    
    offset = minF;
    
    // Now put the back-transformed image into the display buffer
    if(applyCTF)
      thisBuffer = packImg[thisBox];
    
    calcSizeFactor(zoomFactor);
    for (j=0; j<numRows; j++) { // Rows
      int yOut, xOut, l, m, thisChar;
      yOut = j*sizeFactor;
      for (i=0; i<numCols; i++) { // Do entire image
	// Offset this one beside the original
	xOut = i*sizeFactor + numCols*sizeFactor;
	
	// If we need to adjust for TSB orientation 
	// jindx   = (((numRows-j-1)*(int)(fftSize*2))+(i*2));

	jindx   = j*fftSize*2 + i*2;
	
	// Definitely need to add img_bkg back in!!! (To defloat)
	thisFloat= fftImage_copy[jindx]+purdueHeader.img_bkg;
	thisChar = (int)(
			 (float)(((float)((thisFloat)-
					  offset) 
				  * scaleFactor)));
	if(applyCTF) { // Stuff this into the packed box
	  thisBuffer[j*numCols + i]=  thisFloat;
	}
	else { // Normal
	  for (l=0; l<=(int)sizeFactor; l++) {
	    for (m=0; m<=(int)sizeFactor; m++) {
	      XPutPixel(image_image1,
			(xOut+m), (yOut+l),
			pixels_return[9+thisChar]);
	      image_buffer1_disp[(((yOut+l)*buffer_width)+
				  (xOut+m))]= 
		(unsigned char)(thisChar); 
	    }
	  }
	}
      }
    }
    XFlush(dsply);
    free(fftImage_copy);
    
    if(applyCTF && stuffBackground) {
      // Stuff outside of circle with background
      int oldFileType, thisRad, k, l, outHeight, outWidth;
      int xCent, yCent, dist, xSq, ySq, yOut;
      float backChar;
      
      oldFileType = fileType;
      fileType = REBOX;
      thisBuffer = packImg[thisBox];
      calcCircleStats(thisBuffer,
		      &purdueHeader.img_bkg, 
		      &purdueHeader.img_min, &purdueHeader.img_max, 
		      &purdueHeader.img_avg, &purdueHeader.img_std,
		      numCols, numRows, numCols);
      fileType = oldFileType;
      
      // Calc the background
      backChar= purdueHeader.img_bkg;
      
      // Draw the background  for all > radius in box
      // Where is the center?
      yCent = numRows / 2;
      xCent = numCols / 2;
      thisRad = xCent * xCent;
      outHeight = numRows;
      outWidth  = numCols;

      for (k=0; k<outHeight; k++) { // Do the y direction
	yOut = (k * numCols);
	ySq = (k - yCent) * (k - yCent);
	// Stuff all > radius in background character
	for (l=0; l<outWidth; l++) { // Stuff the outbuffer
	  xSq = (l - xCent) * (l - xCent);
	  dist = ySq+xSq;
	  if(dist > thisRad)
	    thisBuffer[yOut+l]=  backChar;
	}
      }
	  
    }
    
    if(fileType==MAP)  // Show more
      XPutImage (dsply, image_window1, image_gc, image_image1, 0, 0,
		 0, 0, (image_width*3), (image_height*2));
    else
      XPutImage (dsply, image_window1, image_gc, image_image1, 0, 0,
		 0, 0, (image_width), image_height);
    
    XFlush (dsply);  
  }
  
  free(temp_buffer);
  //   DrawFFTWindow(NULL, NULL, NULL);
  
} // end function calcFFT



void fft_writeFB_activate(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: fft_writeFB_activate
  // Write various FFT buffers to disk.

  char outFileName[256], rawOutFileName[256];
  int yOut, i, j, m;
  off_t imgOffset;
  int luNum, zero=0;
  float img_bkg, img_min, img_max, img_avg, img_std;
  float minF, maxF, scaleFactor;
  float *tempF_buffer;
  float floatScaleFactor,recipFloatScaleFactor;
  BYTE *temp_buffer;
  FILE *fOut;
  forPifHead_t thisHead;
  fString_t fName,readWrite;
  int tag = *((int *) client_data);

  fName.addr = (char *)malloc(sizeof(char)*64);
  readWrite.addr = (char *)malloc(sizeof(char)*64);
  
  pif_init_head_(&thisHead);
  
  if(tag==WRITEFFT) { // init
    sprintf(outFileName,"fftImage.pif");
    sprintf(rawOutFileName,"fftImage.raw");
    thisHead.numCols =  512;
    thisHead.numRows =  512;
    CK(temp_buffer = (BYTE *)malloc(sizeof(BYTE)*
				    thisHead.numRows*
				    thisHead.numCols));
    for (j=0; j<512; j++) { // Rows
      yOut = 511 - j;
      for (i=0; i<512; i++) { // Do entire image
	// Remember to put it into TSB space
	temp_buffer[(j*512)+i]= (BYTE)fft_buffer1[(yOut*512)+i];
      }
    }
    
  }
  else { // init back
    sprintf(outFileName,"backImage.pif");
    sprintf(rawOutFileName,"backImage.raw");
    thisHead.numCols =  numCols;
    thisHead.numRows =  numRows;
    
    CK(temp_buffer = (BYTE *)malloc(sizeof(BYTE)*numRows*numCols));
    CK(tempF_buffer = (float *)malloc(sizeof(float)*numRows*numCols));
    
    calcSizeFactor(zoomFactor);
    for (j=0; j<numRows; j++) { // Rows
      int yOut, xOut;
      // Remember to put it into TSB space
      yOut = (numRows-1-j)*sizeFactor;
      for (i=0; i<numCols; i++) { // Do entire image
	// Offset this one beside the original
	xOut = i*sizeFactor + numCols*sizeFactor;
	temp_buffer[(j*numCols)+i]= (BYTE)
	  image_buffer1_disp[(yOut*buffer_width)+xOut];
	tempF_buffer[(j*numCols)+i]= (float)
	  image_buffer1_disp[(yOut*buffer_width)+xOut];
      }
    }
    
  }

  calcByteStats(temp_buffer,
		&img_bkg, &img_min,
		&img_max, &img_avg, &img_std,
		thisHead.numCols,thisHead.numRows,
		thisHead.numCols);
  
  minF = img_min;
  maxF = img_max;
  scaleFactor = 255.0 / (maxF-minF);
  
  // Now scale it
  for (j=0; j<thisHead.numRows*thisHead.numCols; j++) { // Rows
    temp_buffer[j]= (BYTE)(((float)temp_buffer[j]-minF) * scaleFactor);
  }
  
  fOut = fopen(rawOutFileName,"w");

  sprintf(fName.addr,"%s",outFileName);
  sprintf(readWrite.addr,"w");

  luNum = 7;
  thisHead.luNum = luNum;
  pif_open(fName.addr, &luNum, readWrite.addr);

  calcMapStats(tempF_buffer,
	       &img_bkg, &img_min, &img_max, 
	       &img_avg, &img_std,
	       &thisHead.numCols, &thisHead.numRows, &thisHead.numSecs);
  calcScaleFactor(&floatScaleFactor,		     
		  &recipFloatScaleFactor, 
		  &int4scaleF, 
		  img_min,img_max,
		  img_avg,img_std,
		  PACKEDPORTABLE);
  
  thisHead.floatScaleFactor = floatScaleFactor;
  thisHead.recipFloatScaleFactor = recipFloatScaleFactor;
  thisHead.int4scaleF = int4scaleF;
  thisHead.nimg = 1;
  thisHead.htype= 1;
  thisHead.numSecs =  1;  // No maps here
  
  thisHead.mode = (tag==WRITEBACK) ? 46 : 6;
  
  pif_write_gh_(&thisHead);
  
  m=0;  // Only 1 image
  
  thisHead.datai4_min = img_min;
  thisHead.datar4_min = (int)img_min;
  thisHead.datai4_max = img_max;
  thisHead.datar4_max = img_max;
  thisHead.datar4_avg = img_avg;
  thisHead.datar4_std = img_std;
  thisHead.img_bkg = img_bkg;
  thisHead.packRadius = packRadius;
  thisHead.borderType = getBorderType(featherMode);
  thisHead.borderWidth = featherRadius-boxRadius;
  
  thisHead.imgNum = m+1;
  pif_write_dh_(&thisHead);
  
  // [3] Now write the image info
  imgOffset = 
    512+                   // Gloal header
    ((m+1)*512)+           // Image header
    ((m)*                  // The image itself
     (off_t) thisHead.numCols*
     thisHead.numRows)*
    sizeof(BYTE);
  
  if(tag==WRITEFFT) { // Write the FFT
    pif_write_byte_image_(&thisHead.luNum,
			  temp_buffer,
			  &imgOffset,
			  &thisHead.numCols,
			  &thisHead.numRows,
			  &thisHead.numSecs);
  }
  else if(tag==WRITEBACK) { // Write the Back transform
    pif_write_map_sections(&thisHead, tempF_buffer, &thisHead.mode,
			   &thisHead.numCols, &thisHead.numRows, 
			   &thisHead.numSecs, &zero,
			   &thisHead.numSecs);
  }
  
  // Flip it back from TSB space for RAW file
  for (j=0; j<(thisHead.numRows/2); j++) { // Rows
    int yOut;
    BYTE thisChar;
    yOut = thisHead.numRows - j - 1;
    for (i=0; i<(thisHead.numCols); i++) { // Do entire image
      thisChar= (BYTE)temp_buffer[(j*thisHead.numCols)+i];
      temp_buffer[(j*thisHead.numCols)+i]= (BYTE)
	temp_buffer[(yOut*thisHead.numCols)+i];
      temp_buffer[(yOut*thisHead.numCols)+i]= (BYTE)thisChar;
    }
  }
  
  fwrite(temp_buffer,sizeof(BYTE),(thisHead.numCols*
				   thisHead.numRows*
				   thisHead.numSecs),fOut);
      
  fclose(fOut);
  pif_close(&thisHead.luNum);
  
  free(temp_buffer);
  
  sprintf(outStr,"Successfully wrote [%s] and [%s]\nDimensions [%d][%d]\n",
	  outFileName, rawOutFileName,
	  thisHead.numCols, thisHead.numRows);

  askThem(outStr, XmDIALOG_INFORMATION);
  
} // end function fft_writeFB_activate



void ctf_all_activate()
{
  // function name: ctf_all_activate
  // Apply a ctf multiplication or division.

  int i;
  int units, fft_full;
  int mode,filter;
  int j;
  int jindx;
  int strtAvg,endAvg,thisImg,thisNimg;
  int ksign,iform;
  int icount;
  int luNum;
  off_t imgOffset;
  int thisBoxRadius;
  int m;
  int zero=0, one=1;

  float volts,pixelSize,ampContrast_in,defocus_in;
  float defocus_minor_in,sphericalAbbr;
  float angMajor,tFactor,wiener,pi;
  float radian,degree;
  float maxF,minF,offset;
  float scaleFactor, thisFloat;
  float thisBackground;
  float *temp_buffer;
  float *char_buffer, *thisBuffer;
  float *real_buffer;
  float *fftImage_copy;
  float img_bkg,img_min,img_max;
  float img_minF,img_maxF;
  float img_avg,img_std;
  float cc, dE, beta, filt_fac[10];
  float min,max,scaleF;

  BYTE *byte_buffer;
  forPifHead_t thisHead;
  fString_t fName,readWrite;
  
  if(!doApplyCtf) { // ok to check
    sprintf(outStr,"I'm using CTF filter==[%d]  Do you want me to continue?",
	    ctfFilter);
    if(!askThem(outStr, XmDIALOG_INFORMATION))
      return;
  }

  numCols = purdueHeader.numCols;
  numRows = purdueHeader.numRows;

  CK(temp_buffer= (float *)malloc(
			 sizeof(float)*numCols*numRows));
  CK(real_buffer= (float *)malloc(
			 sizeof(float)*numCols*numRows*purdueHeader.nimg));

  CK(char_buffer= (float *)malloc(sizeof(float)*numCols*numRows));
  CK(fftImage_copy= (float *)malloc(sizeof(float)*idim1*idim2*2));
  
  // Open up the output file
  fName.addr = (char *)malloc(sizeof(char)*264);
  readWrite.addr = (char *)malloc(sizeof(char)*64);
  pif_init_head(&thisHead);
  sprintf(fName.addr,"ctf_corrected_imgs.pif");
  sprintf(readWrite.addr,"w");
  
  luNum = 7;
  thisHead.luNum = luNum;
  
  pif_open(fName.addr, &luNum, readWrite.addr);
  
  strtAvg = 0;
  endAvg  = purdueHeader.nimg;
  thisNimg = 0;
  // Find out how many we've got
  for(i=strtAvg;i<endAvg;i++)
    if(!deletePackedImage[i])
      thisNimg++;
  
  icount = 0;  // Count the images
  for(thisImg=strtAvg;thisImg < endAvg;thisImg++) { // Do all images wanted
    
    if(!deletePackedImage[thisImg]) { // OK to do
      
      // Tell them what we're doing
      TextSetInt(fftImgNumW, thisImg+1, TRUE);
      
      if(fileType==PACKEDPORTABLE || fileType==MAP)
	TextSetInt(fftImgNumW, thisImg+1, TRUE);
      else
	TextSetInt(fftImgNumW, 1, TRUE);
      
      init_floatarr(fftImage_gold, idim1*idim2*2, 0.0); // Initialize array
      
      if(fileType == MAP) { // DO for MAP
	thisBackground = purdueHeader.img_bkg;
	
	// NOTE:  Use tsb orientation when writing to disk
	map_fft_fill_tsborien__((float_image_buffer1+
				(thisImg*numCols*numRows)), 
			       &numCols, &numRows,
			       fftImage_gold, 
			       &fftSize, &fftSize, 
			       &thisBackground);
      }
      else { // Do for packed images
	thisBuffer= (float *)packImg[(thisImg)]; 
	
	for (i=0; i<numRows; i++)
	  for (j=0; j<numCols; j++) { // Do all
	      // Why oh why do we have to do this here ??? - when to
	      // shift to TSB orientation and when not to ????? We
	      // need to do this because we will write this in TSB
	      // orientation on disk.
	    
	    // temp_buffer[(((numRows-i-1)*numCols)+j)]= (unsigned short)
	    // thisBuffer[((i*numCols)+j)];
	    // temp_buffer[((i*numCols)+j)]= (unsigned short)
	    // thisBuffer[((i*numCols)+j)];
	    
	    temp_buffer[((numRows-i-1)*numCols)+j]= (float)
	      thisBuffer[(i*numCols)+j];
	  }
	
	// Get global
	if(packRadius > 0)
	  purdueHeader.img_bkg = (float)(*packImg[thisImg]);
	thisBackground = purdueHeader.img_bkg;
	
	r4_img_fft_fill__(temp_buffer, &numCols, &numRows, 
				 fftImage_gold, 
				 &fftSize, &fftSize, &thisBackground);
      } // Do for packed images
      
      ksign = 1;   // Do a "forward" FFT
      iform = 1;   // Data is imaginary
      
      fft_2d_rob__(fftImage_gold, &fftSize, &fftSize,
		   &ksign, &iform);
      
      // Apply CTF to current FFT and back transform
      
      volts   = wstrtod(ctfKvW) * 1000.0; // Convert from kv to volts
      
      pixelSize   = wstrtod(ctfPixSizeW);
      
      // These should be in microns ???
      ctfMajorDefocusValue   = wstrtod(ctfMajorDefocusW);
      ctfMinorDefocusValue   = wstrtod(ctfMinorDefocusW);
      defocus_in        = ctfMajorDefocusValue;
      defocus_minor_in  = ctfMinorDefocusValue;
      
      sphericalAbbr   = wstrtod(ctfSpAberW);
      wavelength = wavel_(&volts);
      ampContrast_in = wstrtod(ctfAmpFactorW);
      
      if(doApplyCtf)
	pixelSize  = applyCtfPixelSize;
      
      if(pixelSize != 0.0) { // OK to do
	
	units = 2;  // e.g. we're using nanometers for pixelSize
	fft_full = 1;
	pi = M_PI;
	radian = 0.0;
	piraddeg_(&pi, &radian, &degree);
	
	mode = (int)wstrtod(ctfModeW);
	wiener = wstrtod(ctfWienerW);
	if(wiener==0.0)
	  wiener= 0.001;
	TextSetFloat(ctfWienerW, wiener, TRUE);
	angMajor = wstrtod(ctfAngMajorW);

	// We DO NOT need to change the sign of angMajor here. We are
	// writing to disk and we need to use the sign just as given
	// from the interface.  e.g. The buffer was stuffed in TSB
	// orientation above so no flipping needed.  
	// angMajor = 0.0 - angMajor;

	tFactor = wstrtod(ctfTFactorW);
	
	if(doApplyCtf) { // batch mode
	  wiener     = applyCtfWiener;
	  tFactor    = applyCtfTFactor;
	  mode       = applyCtfMode;      // (int)
	  angMajor   = applyCtfAngMajor;  // in nm
	  pixelSize  = applyCtfPixelSize;
	  defocus_in = applyCtfDefocus_in;
	  defocus_minor_in = applyCtfDefocus_minor_in;
	  ampContrast_in   = applyCtfAmpContrast_in;
	  sphericalAbbr    = applyCtfSphericalAbbr;
	  volts            = applyCtfVolts;  // int in volts NOT kV
	}
	
	// Init these for now - get from interface eventually
	cc = 0.0;
	dE= 0.0;
	beta = 0.0;
	
	filt_fac[0] = wiener;
	filt_fac[1] = 0.1;
	filter = ctfFilter;
	if(writeFFTType==APPLYCTF || writeFFTType==APPLYCTFR4)
	  ctf_multiply__(&mode,
			fftImage_gold,
			&fftSize, &fftSize, 
			&units, 
			&pixelSize,
			&pi, &radian,
			&wavelength,
			&defocus_in, &defocus_minor_in,
			&angMajor,
			&ampContrast_in,
			&sphericalAbbr, 
			&cc,&beta,&dE,&volts,
			&tFactor, &fft_full);
	else if(writeFFTType==APPLYINVERSECTF || 
		writeFFTType==APPLYINVERSECTFR4)
	  ctf_divide__(&mode,
		      &filter,
		      fftImage_gold,
		      &fftSize, &fftSize, 
		      &units, 
		      &pixelSize,
		      &pi, &radian,
		      &wavelength,
		      &defocus_in, &defocus_minor_in,
		      &angMajor,
		      &ampContrast_in,
		      &sphericalAbbr, 
		      filt_fac,
		      &tFactor,
		      &fft_full,
		      &cc,&beta,&dE,&volts);
	
	applyCTF = FALSE;
	
	// Show them what the back transform is
	// Now put it back in "normal" space
	ksign = 0-1;   // Do a "reverse" FFT
	iform = 1;     // Data is imaginary
	      
	// Back transform into a copy of the filtered FFT just in case
	// they want to do Fourier Averaging or etc...
	
	for (i=0; i<idim1*idim2*2; i++)
	  fftImage_copy[i] = fftImage_gold[i] ;
	
	if(writeFFTType==APPLYCTF || writeFFTType==APPLYINVERSECTF) {
	  // DO it in BYTE
	  
	  fft_2d_rob__(fftImage_copy, &fftSize, &fftSize, 
		      &ksign, &iform);
	  
	  // Let's look at what scaling needs done
	  scaleFactor = 0.0;
	  findMaxMin(fftImage_copy, &maxF, &minF, 
		     &offset, fftSize, fftSize);
	  
	  // Defloat it
	  maxF += purdueHeader.img_bkg;
	  minF += purdueHeader.img_bkg;
	  
	  scaleFactor = (maxF==minF) ? 1.0 : 255.0/(maxF-minF); 
	  offset = minF;
	  
	  // Now put the back-transformed image into the 
	  // display buffer
	  thisBuffer = char_buffer;
	  
	  calcSizeFactor(zoomFactor);

	  for (j=0; j<numRows; j++) { // Rows
	    for (i=0; i<numCols; i++) { // Do entire image
	      // Offset this one beside the original
	      
	      jindx   = j*fftSize*2 + i*2;
	      
	      // Definitely need to add img_bkg back in!!! 
	      // (To defloat)
	      thisFloat= fftImage_copy[jindx]+purdueHeader.img_bkg;
	      thisBuffer[(j*numCols)+i]=  
		(unsigned char)((thisFloat-minF)*scaleFactor);   
	    }
	  }
		  
	  calcCircleStats(char_buffer,
			  &img_bkg, 
			  &img_min, &img_max,
			  &img_avg, &img_std,
			  numCols, numRows, numCols);
	  
	  calcScaleFactor(&floatScaleFactor,		     
			  &recipFloatScaleFactor, 
			  &int4scaleF, 
			  img_min, img_max,
			  img_avg, img_std,
			  PACKEDPORTABLE);
	  
	  thisBuffer = char_buffer;
	  
	  thisBoxRadius = featherRadius;
	  backChar = img_bkg;
	  appodize_it(thisBuffer, 
		      numCols, 
		      numRows, numCols,
		      fileType,
		      bkgrndMethod,
		      boxMethod,
		      packRadius,
		      boxRadius,
		      featherRadius,
		      recipSizeFactor,
		      sizeFactor,
		      tryColors,
		      featherMode,
		      &backChar,
		      thisBoxRadius,
		      forceAvgBkg,
		      img_avg);
	  
	  thisHead.numCols = numCols;
	  thisHead.numRows = numRows;
	  thisHead.numSecs = 1;
	  // Is this OBsolete???  Don't we ALWAYS write float data 
	  // from now on??? 
	  thisHead.mode = 6;
	  thisHead.nimg = thisNimg;
	  thisHead.floatScaleFactor = floatScaleFactor;
	  thisHead.recipFloatScaleFactor = recipFloatScaleFactor;
	  thisHead.int4scaleF = int4scaleF;
	  thisHead.htype= 1;
	  
	  pif_write_gh_(&thisHead);
	  
	  m=icount;
	  icount++;
	  
	  thisHead.datai4_min = img_min;
	  thisHead.datar4_min = (int)img_min;
	  thisHead.datai4_max = img_max;
	  thisHead.datar4_max = img_max;
	  thisHead.datar4_avg = img_avg;
	  thisHead.datar4_std = img_std;
	  thisHead.img_bkg = img_bkg;
	  thisHead.packRadius = packRadius;
	  thisHead.borderType = getBorderType(featherMode);
	  thisHead.borderWidth = featherRadius-boxRadius;
	  
	  thisHead.imgNum = m+1;
	  pif_write_dh_(&thisHead);
	  
	  // [3] Now write the image info
	  imgOffset = 
	    512+                   // Gloal header
	    ((m+1)*512)+           // Image header
	    ((m)*                  // The image itself
	     (off_t) thisHead.numCols*
	     thisHead.numRows)*
	    sizeof(BYTE);
		  
	  byte_buffer = (BYTE *)malloc(sizeof(BYTE)*
				       thisHead.numCols*thisHead.numRows);
	  
        min = max = char_buffer[0];
	  for (m=0; m<thisHead.numRows*thisHead.numCols; m++) {
	    min = MIN(min,char_buffer[m]);
	    max = MAX(max,char_buffer[m]);
	  }
		  
	  scaleF = 255.0 / max;
	  
	  for (m=0; m<thisHead.numRows*thisHead.numCols; m++) {
	    byte_buffer[m] = (BYTE)(char_buffer[m] * scaleF);
	  }
	  
	  pif_write_byte_image_(&thisHead.luNum,
				byte_buffer,
				&imgOffset,
				&thisHead.numCols,
				&thisHead.numRows,
				&thisHead.numSecs);
	  
	  free(byte_buffer);
	}
	else { // DO it R4
	  thisBackground = purdueHeader.img_bkg;	  
	  fft_2dbt_r4__(fftImage_copy, &fftSize, &fftSize, 
		       (real_buffer+(thisImg*numCols*numRows)),
		       &numCols, &numRows,
		       &fft_full,
		       &thisBackground);  
	}
	      
      }
      
    }
  }
  
  if(writeFFTType==APPLYCTFR4 || writeFFTType==APPLYINVERSECTFR4) {
    // Write R4 images
    
    if(doApplyCtf)
      nimg = thisNimg;
    img_bkg = img_minF = img_maxF = img_avg = img_std = 0.0;
    calcRealStats(real_buffer,
		  &img_bkg, 
		  &img_minF, &img_maxF,
		  &img_avg, &img_std,
		  numCols, numRows, numCols, nimg);
    printf("img_max/min[%f,%f]\n",img_maxF,img_minF);
    printf("nc,nr,nimg=[%d,%d,%d]\n",numCols,numRows,nimg);
    
    thisHead.numCols = numCols;
    thisHead.numRows = numRows;
    thisHead.numSecs = 1;
    thisHead.mode = 46;
    thisHead.nimg = thisNimg;
    thisHead.floatScaleFactor = MAXINT / 
      (2.0 + MAX((img_maxF+2.0), fabs(img_minF-2.0)));

    thisHead.recipFloatScaleFactor = 1.0/(thisHead.floatScaleFactor);
    thisHead.int4scaleF = thisHead.floatScaleFactor;
    thisHead.htype= 1;
    thisHead.datai4_min = img_minF;
    thisHead.datar4_min = img_minF;
    thisHead.datai4_max = img_maxF;
    thisHead.datar4_max = img_maxF;
    thisHead.datar4_avg = img_avg;
    thisHead.datar4_std = img_std;
    
    pif_write_gh_(&thisHead);
    
    icount = 0;
    for (thisImg=strtAvg; thisImg<endAvg; thisImg++) { // Do all images wanted
      if(!deletePackedImage[thisImg]) { // OK to do
	m=icount;
	icount++;
	
	// Now get stats on just this one
	calcRealStats((real_buffer+(thisImg*numCols*numRows)),
		      &img_bkg, 
		      &img_minF, &img_maxF,
		      &img_avg, &img_std,
		      numCols, numRows, numCols, 1);
	
	thisHead.datai4_min = img_minF;
	thisHead.datar4_min = img_minF;
	thisHead.datai4_max = img_maxF;
	thisHead.datar4_max = img_maxF;
	thisHead.datar4_avg = img_avg;
	thisHead.datar4_std = img_std;
	thisHead.floatScaleFactor = MAXINT / 
	  (2.0 + MAX((img_maxF+2.0), fabs(img_minF-2.0)));
	
	thisHead.recipFloatScaleFactor = 1.0/(thisHead.floatScaleFactor);
	thisHead.int4scaleF = thisHead.floatScaleFactor;
	thisHead.img_bkg = img_bkg;
	thisHead.packRadius = packRadius;
	thisHead.borderType = getBorderType(featherMode);
	thisHead.borderWidth = featherRadius-boxRadius;
	
	thisHead.imgNum = m+1;
	
	pif_write_dh_(&thisHead);
	
	// WHY ????
	thisHead.imgOffset= 512 +
	  (512 * (thisImg+1)) +
	  ((off_t)(thisImg) * numCols * numRows *4);

	pif_write_map_sections(&thisHead, real_buffer+thisImg*numCols*numRows,
			       &(thisHead.mode), &(thisHead.numCols),
			       &(thisHead.numRows), &(thisHead.numSecs),
			       &zero, &one);
	
      }
    }
  }
  
  pif_close(&thisHead.luNum);
  
  if(!doApplyCtf)
    askThem("Images written to ctf_corrected_imgs.pif file.",
	    XmDIALOG_INFORMATION);
  
  free(fName.addr);
  free(readWrite.addr);
  free(char_buffer);
  free(temp_buffer);
  free(real_buffer);
  free(fftImage_copy);
  
} // end function ctf_all_activate



void fftClick(Widget widget, XtPointer client_data, XtPointer call_data)
{ 
  // function name: fftClick
  // Every time the mouse button is clicked call here. This routine
  // handles the selecting of lattice points for the FFT.
  
  char ans, outStr[16];
  int x, y, recalcAxBxAyBy;
  XEvent event;

  XmDrawingAreaCallbackStruct *draw_data = 
    (XmDrawingAreaCallbackStruct *) call_data;

  if(draw_data == 0) 
    return;
  
  // Turn this off
  calcFFT1DcircAvg= FALSE;
  
  event= *(draw_data->event);
  
  switch (event.type) {
    
  case ButtonRelease:
    x = event.xbutton.x;   // Get x,y coords of point pos
    y = event.xbutton.y;
    break;
    
  case ButtonPress:
    x = event.xbutton.x;   // Get x,y coords of point pos
    y = event.xbutton.y;
    
    recalcAxBxAyBy= TRUE;
    
    // Put coords back into "normal" space
    if(doingAutoDefocusRefine && adrZoomDraw) { // special
      x = x/2 + 256;
    }
    else if(widget!=contourW &&  fftZoomFactor==2) { // Adjust for zoom
      x= fftZoomX + x/2;
      y= fftZoomY + y/2;
    }
    
    if(drawAuto2Dscurve) { // remember where they clicked
      scurveClick.x = x;
      scurveClick.y = y;
      DrawFFTWindow(NULL, NULL, NULL);
    }
    
    if(measureFFT) { // They are measuring the FFT here
      if(event.xbutton.button == Button1) { // Pick first point
	fftMeasures[0].x = x;
	fftMeasures[0].y = y;
	fftMeasureDraw(event.xbutton.button);
      }
      if(event.xbutton.button == Button2) { // Pick second point
	fftMeasures[1].x = x;
	fftMeasures[1].y = y;
	fftMeasureDraw(event.xbutton.button);
      }
      if(event.xbutton.button == Button3) { // measure
	fftMeasures[0].x = 256;
	fftMeasures[0].y = 256;
	fftMeasures[1].x = x;
	fftMeasures[1].y = y;
	fftMeasureDraw(event.xbutton.button);
      }
      XFlush(dsply);
    }
    else if(pickLattice != -99) { // OK to keep on trying
      latValChanged = FALSE;
      
      if(widget==contourW) { // They've selected from CONTOUR drawing area
	if(contour) { // Don't do if not intended
	  if(event.xbutton.button == Button1) { // Pick appropriate Lattice point
	    if(pickLattice==1) { // Get first points
	      latPt[0].measuredX += x/8 - 8;
	      latPt[0].measuredY += y/8 - 8 + 1;
	      stuffContourBuffer();
	    }
	    else if(pickLattice==2) { // Get second points
	      latPt[1].measuredX += x/8 - 8;
	      latPt[1].measuredY += y/8 - 8 + 1;
	      stuffContourBuffer();
	    }
	    else if(pickedLatPts) {
	      // Just find out which one we're dealing with
	      // findLatPt(x,y);
	      latPt[currentLatPt].measuredX += x/8 - 8;
	      latPt[currentLatPt].measuredY += y/8 - 8 + 1;
	      stuffContourBuffer();		   
	    }
	  }
	}
      }
      else { // Main fft screen
	if(event.xbutton.button == Button1) {
	  // Pick appropriate Lattice point
	}
	else if(event.xbutton.button == Button2) { // yongning
	  printf("x,y[%d,%d]\n",x,y);
	  recalcAxBxAyBy= FALSE;
	}
	else if(event.xbutton.button == Button3) { // Give feedback
	  short ix, iy, jh, jk;
	  float xtmp, ytmp, fmax, degree;
	  int ly_lim, ncol, nrow;
	  float axold, bxold, ayold, byold;
	  float *apart, *bpart;
	  
	  // Now give some info on what's there
	  ix = x - 255;
	  iy = 512 - y - 255;
	  ly_lim = 512;
	  axold = axL;
	  ayold = ayL;
	  bxold = bxL;
	  byold = byL;
	  fmax = sqrt(pow(10.0,(double)maxFFTvalue)) * 0.1; // Why times 0.1 ???
	  if(fileType==PACKEDPORTABLE) { // packed box
	    ncol = purdueHeader.numCols;
	    nrow = purdueHeader.numRows;
	  }
	  else { // other
	    ncol = numCols;
	    nrow = numRows;
	  }
	  CK(apart = (float *)malloc(sizeof(float)*fftSize));
	  CK(bpart = (float *)malloc(sizeof(float)*fftSize));
	  
	  fft_latref2_index__(&ix, &iy, 
			      &fftSize, &fftSize, 
			      &jh, &jk,
			      &xtmp, &ytmp, 
			      &ly_lim, 
			      &axold, &ayold, &bxold, &byold);
	  
	  fft_latref2_spot__(fftImage_buffer2, 
			     &fftSize, &fftSize, 
			     apart, bpart,
			     &ix, &iy, 
			     &fmax, &degree,
			     &ncol, &nrow);
	  
	  free(apart);
	  free(bpart);
	}
	
	if(event.xbutton.button != Button3) {
	  // if it wasn't the third button they clicked above try
	  // and find a lattice point for it ...
	  if(pickLattice==1) { // Get first points
	    latPt[0].measuredX = x;
	    latPt[0].measuredY = y;
	    currentLatPt=0;
	    stuffContourBuffer();		   
	    pickLattice= 2;
	  }
	  else if(pickLattice==2) { // Get second points
	    latPt[1].measuredX = x;
	    latPt[1].measuredY = y;
	    currentLatPt=1;
	    stuffContourBuffer();		   
	    pickLattice= 0;
	    pickedLatPts= TRUE;
	  }
	  else if(pickedLatPts) { // find out which one we're dealing with
	    findLatPt(x,y);
	    latPt[currentLatPt].measuredX = x;
	    latPt[currentLatPt].measuredY = y;
	    stuffContourBuffer();		   
	  }
	}
      }
      
      // Calc lattice - they may have moved where they want it
      if(!layerLines && pickedLatPts && !drawCTF) { // OK to calc
	calcLattice(recalcAxBxAyBy);
      }
      
      if(drawCTF && disp1Dca ) {
	// Just default it to where they clicked
	latPt[0].measuredX = x;
	latPt[0].measuredY = y;
	latPt[1].measuredX = x;
	latPt[1].measuredY = y;
	currentLatPt=1;
	stuffContourBuffer();		   
	pickLattice= 0;
	pickedLatPts= TRUE;
      }
      
      // **** DO the CTF stuff here ****
      if(pickedLatPts && drawCTF) { // DO CTF
	int x1, y1, x2, y2;
	float ang1, distAsq, distBsq, distA, distB;
	
	ctfPts[ctfCurveNum].latX0 = latPt[0].measuredX;
	ctfPts[ctfCurveNum].latY0 = latPt[0].measuredY;
	ctfPts[ctfCurveNum].latX1 = latPt[1].measuredX;
	ctfPts[ctfCurveNum].latY1 = latPt[1].measuredY;
	ctfPts[ctfCurveNum].pickedPts = TRUE;
	
	if(latPt[0].measuredX==256)
	  latPt[0].measuredX=257;
	
	if(latPt[0].measuredX != 256)
	  ang1 = atan( fabs((double)(latPt[0].measuredY-256))/
		       fabs((double)(latPt[0].measuredX-256)) );
	else
	  ang1 = deg2rad(90.0);
	
	// If we're greater than 90 degrees add that in
	if(latPt[0].measuredX < 256)
	  ang1 = deg2rad(90.0) + (deg2rad(90.0) - ang1);
	// If we're greater than 180 degrees add that in
	if(latPt[0].measuredY > 256)
	  ang1 = deg2rad(180.0) + (deg2rad(180.0) - ang1);
	
	// If 1D CTF (Assuming stigmated image), set ang to zero
	if(disp1Dca)
	  ang1 = 0.0;

	x1= latPt[0].measuredX;
	x2= latPt[1].measuredX;
	y1= latPt[0].measuredY;
	y2= latPt[1].measuredY;
	
	distAsq = (x1-256)*(x1-256) + (y1-256)*(y1-256);
	distBsq = (x2-256)*(x2-256) + (y2-256)*(y2-256);
	distA = sqrt(distAsq);
	distB = sqrt(distBsq);
	
	ctfPts[ctfCurveNum].distAsq = distAsq;
	ctfPts[ctfCurveNum].distA   = distA;
	ctfPts[ctfCurveNum].distBsq = distBsq;
	ctfPts[ctfCurveNum].distB   = distB;
	ctfPts[ctfCurveNum].omega   = ang1;
	
	ctfFocusMinvalue = distA / distB;
	ctfAngMajorValue = ang1;
	
	calcDefocus(&ctffocusMajValue, &ctfFocusMinvalue);
	printf("fftClick [%f,%f]\n", ctffocusMajValue, ctfFocusMinvalue);
	
	// Now setup the scroll bars
	XtSetArg(arglst[0], XmNincrement, 10);
	XtSetArg(arglst[1], XmNpageIncrement, 100);
	XtSetArg(arglst[2], XmNvalue, (int)(rad2deg(ctfAngMajorValue)*100.0));
	XtSetArg(arglst[3], XmNmaximum, 18888);
	XtSetArg(arglst[4], XmNminimum, -18000);
	XtSetValues(ctfAmajSW, arglst, 5);
	
	XtSetArg(arglst[0], XmNvalue, (int)(ctfFocusMinvalue*100.0));
	XtSetArg(arglst[1], XmNincrement, 1);
	XtSetArg(arglst[2], XmNpageIncrement, 10);
	XtSetArg(arglst[3], XmNmaximum, 11111);
	XtSetArg(arglst[4], XmNminimum, -10000);
	XtSetValues(ctfFminSW, arglst, 5);
	
	XtSetArg(arglst[0], XmNvalue, (int)(ctffocusMajValue*100.0));
	XtSetArg(arglst[1], XmNincrement, 1);
	XtSetArg(arglst[2], XmNpageIncrement, 10);
	XtSetArg(arglst[3], XmNmaximum, 11111);
	XtSetArg(arglst[4], XmNminimum, -10000);
	XtSetValues(ctfFmajSW, arglst, 5);
	
	TextSetFloat(ctfFminTW,       ctfFocusMinvalue, FALSE);
	TextSetFloat(ctfMinorDefocusW,ctfFocusMinvalue, FALSE);
	sprintf(outStr,"%.2f",rad2deg(ctfAngMajorValue));  
	TextSet(ctfAmajTW, outStr, FALSE);
	TextSet(ctfAngMajorW, outStr, FALSE); 
	TextSetFloat(ctfFmajTW,        ctffocusMajValue, FALSE);
	TextSetFloat(ctfMajorDefocusW, ctffocusMajValue, FALSE);
	
      }
      
      // Draw new Curve
      DrawFFTWindow(NULL, NULL, NULL);
      
    }
    else if(filterToggle || lopHipSwitch) { // Get resolution filter point
      
      // No reso filter with layer lines yet
      if(event.xbutton.button == Button1) { // Select new resolution
	resoX= x;
	resoY= y;
	lowResoFilter = (x-256)*(x-256) + (y-256)*(y-256);
      }
      if(event.xbutton.button == Button2) { // Select new resolution
	resoX= x;
	resoY= y;
	hiResoFilter = (x-256)*(x-256) + (y-256)*(y-256);
      }
      else if(event.xbutton.button == Button3) { // Reset resolution
	resoX= 0;
	resoY= 0;
	lowResoFilter = 320000;  // Ridiculously high
	hiResoFilter  = 0;       // Ridiculously lo
      }
      
      sprintf(outStr,"%d",(int)sqrt((float)lowResoFilter));
      TextSet(lopRadiusTextW, outStr, FALSE);
      sprintf(outStr,"%d",(int)sqrt((float)hiResoFilter));
      TextSet(hipRadiusTextW, outStr, FALSE);
      
      calcFFT(FALSE, 0, 0);
      DrawFFTWindow(NULL, NULL, NULL);
      
    }
    break;
    
  case KeyPress:
    ans= XLookupKeysym((XKeyEvent *)&event, 0);
    printf ("Keyed[%c]\n",ans);
    break;
    
  default:
    break;
  }
  
  XFlush(dsply);   // Make changes current
  
} // end function fftClick



void autoBox_activate(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: autoBox_activate
  // Do the autoboxing type that has been selected

  int abh, abw, aba; // autoBox height, width and area
  int dataRange, thisRadius;
  int i, j, k, yOut, yOff;
  unsigned short level;
  float *ab_buf, *prob_p, *sum_p;
  float  *ab_buf_raw, *temp_byte_buffer;
  float img_bkg, minF, maxF, img_avg, img_std, scaleF;
  FILE *fIn;
  int tag = *((int *) client_data);
  char *error_msg;
   
  firstAMMSTDEVtime = TRUE;

  // ************** Autobox [1] (EMMA method)
  if(tag==AUTOBOX || tag==CORRELATE) {
    
    DrawWindow();
    
    printf("autoBox_activate() called [1].\n");
    TextSet(timeLeftBoxW, "Autoboxing", TRUE);
    
    // ***** Load in current radius
    if(fileType==PACKEDPORTABLE) {
      askThem("Sorry, Only scanned images can be autoBoxed.",
	      XmDIALOG_INFORMATION);
      return;
    } else {
      thisRadius = emmaRadius;
    }

    if(autoBoxSwitch) {
      nThins = autoBoxThin;
      nDilates = autoBoxDilates;
    }

    // reset data
    freeNullInt(&globAccum);

    if((globAccum=(int *)malloc(file_height*file_width*sizeof(int)))==(int *)NULL) {
      printf("\ncannot allocate globAccum");
      TextSet(timeLeftBoxW, "Err alloc buffers for globAccum", TRUE);
	  return;
    }
    
    error_msg = emma_autobox(inCore_buffer, globAccum, box, &compress_factor,
                          &numBoxes, thisRadius, emptyBox, nThins, nDilates,
                          file_height, file_width, fileType,
                          bkgrndMethod, boxMethod, packRadius, boxRadius,
                          featherRadius, recipSizeFactor, tryColors);
    if (error_msg) {
       TextSet(timeLeftBoxW, error_msg, TRUE);
       return;
    }

    if(tag==CORRELATE) {
      // DO the reference correlation automatically 
      for(i=0; i<numBoxes; i++)
	     box[i].reference = TRUE;
      
      ref_correlate_activate();
      
      // Remember for next time
      for(i=0; i<numBoxes; i++)
	     boxRefLast[i] = box[i].reference ;
      
      doLabel("Done.", autoCorrLabelW, FALSE);
    }
    
    DrawWindow();
    
    sprintf(outStr,"Done.  [%d] boxes.",numBoxes);
    TextSet(timeLeftBoxW, outStr, TRUE);
    
    doLabel("Done.", autoBoxLabelW, FALSE);
  }

   // ******************* Autobox [2] (New method) **************
  
  else if(tag==AUTOBOX2) {
    printf("\nautoBox_activate called [2].\n");
    
    // Load in current radius
    thisRadius = emmaRadius;
    
    abh = file_height;
    abw = file_width;
    aba = abh*abw;
    
    // Open the disk file for load in raw image
    if ( (fIn= fopen(image_fileName.addr,"r")) == NULL ) {
      printf(" Error open image file for refine center.\n");
      TextSet(timeLeftBoxW, "Error opening image", TRUE);
      return;
    }
    
    // ***** allocate space for image and temp byte buffer
    if((ab_buf_raw=  (float *)malloc(aba*sizeof(float)))==(float *)NULL) {
      printf("\nError: Allocate particle image for autoboxing");
      TextSet(timeLeftBoxW, "Err alloc particle buffers", TRUE);
      return;
    }
    
    if((ab_buf = (float *)calloc(aba, sizeof(float))) == (float *)NULL) {
      printf("\nError: Allocate particle image for autoboxing");
      free(ab_buf_raw);
      TextSet(timeLeftBoxW, "Err alloc ptle buffers", TRUE);
      return;
    }
    
    if((temp_byte_buffer=(float *)malloc(abw*sizeof(float)))==(float *)NULL) {
      printf("\nError: Allocate particle image for autoboxing");
      free(ab_buf_raw);
      free(ab_buf);
      TextSet(timeLeftBoxW, "Err alloc particle buffers", TRUE);
      return;
    }
    
    // ****** Image compression regarding to radius
    
    // Let's scale this
    calcStats(inCore_buffer, &img_bkg, &minF,
	      &maxF, &img_avg, &img_std,
	      file_width, file_height, file_width,
	      fileType, bkgrndMethod, boxMethod,
	      packRadius, boxRadius, featherRadius,
		  recipSizeFactor, tryColors);
    scaleF = 255.0/(maxF-minF);
    
    // Load in raw image
    for (j=0; j<abh; j++) {
      if(!imgInCore) { // Get from disk 
	yOff = headOff + j*abw;
	sysError = fseek(fIn,yOff,0);
	if(sysError) {
	  printf("Error: fseek in auto box.\n");
	  free(ab_buf_raw);
	  free(ab_buf);
	  free(temp_byte_buffer);
	  TextSet(timeLeftBoxW, "fseek() err in autobox", TRUE);
	  return;
	}
	fread(temp_byte_buffer, sizeof(BYTE), (abw), fIn);
      } else { // Get from core
	yOff = j*abw;
	memcpy(temp_byte_buffer,&inCore_buffer[(yOff)],
	       (abw*sizeof(float)));
      }
      
      // Stuff the ab_buf_raw
      yOut = j*abw;
      for(k=0; k<abw; k++)
        ab_buf_raw[yOut+k]= ((temp_byte_buffer[k]-minF)*scaleF)/2.0;
    }
    
    fclose(fIn);
    free(temp_byte_buffer);
    
    // *********** Histogram equalization **********
    dataRange = 128;
    if((prob_p=(float*)malloc(dataRange*sizeof(float)))==(float*)NULL ||
       (sum_p=(float*)malloc(dataRange*sizeof(float)))==(float*)NULL) {
      printf("\nError: Allocate buffers for Histogram equalization\n");
      TextSet(timeLeftBoxW, "Error alloc prob_p buffers", TRUE);
      return;
    }
    
    init_floatarr(prob_p, dataRange, 0.0);
    init_floatarr(sum_p, dataRange, 0.0);
    
    for (k=0; k<abh*abw; k++)
      prob_p[(int)ab_buf_raw[k]]++;
    
    for (i=0; i<dataRange; i++)
      prob_p[i] = prob_p[i]/(float)(aba);
    
    for (i=0; i<dataRange; i++)
      for (j=0; j<=i; j++)
        sum_p[i] += prob_p[j];
    
    for (k=0; k<abh*abw; k++) {
      level = (unsigned short)(0.5+(dataRange - 1)*sum_p[(int)ab_buf_raw[k]]);
      if(level >= dataRange)
	level = dataRange - 1;
      ab_buf_raw[k] = level;
    }
    
    free(ab_buf_raw);
    free(prob_p);
    free(sum_p);
    free(ab_buf);
    
    DrawWindow();
    
    printf("\nFinish automatic boxing\n");
  }
  
} // end function autoBox_activate



void histArrowActivate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: histArrowActivate
  // Expand histogram on Alter Colormap Screen
 
  if (widget == colorHistUpW)
    histZoomFactor++;
  else if (widget == colorHistDnW)
    histZoomFactor--;

  // Clip it accordingly
  histZoomFactor = MIN(histZoomFactor, 39);
  histZoomFactor = MAX(histZoomFactor, 1);
  
  histogram_activate();
} // end function histArrowActivate



void histogram_activate()
{
  // function name: histogram_activate
  // Plot histogram on colormap screen

  int i, j, minP, maxP;
  float thisZoomFactor = 1.0;
  
  if(useGUI == FALSE) return;

  minP = maxP = histogram[0];
  for(i=0; i<256; i++) {
    minP = MIN(minP, histogram[i]);
    maxP = MAX(maxP, histogram[i]);
  }    

  // Zoom this guy appropriately
  if (histZoomFactor >= 20) { // ZoomIn
    for (i=20; i<histZoomFactor; i++)
      thisZoomFactor *= 2.0;
  } else { // ZoomOut
    for (i=histZoomFactor; i<20; i++)
      thisZoomFactor /= 2.0;
  }

  // Make sure we don't over step our bounds
  if(maxP > 0 && thisZoomFactor > 0)
    histScaleFactor = 32.0 / (maxP / thisZoomFactor);
  else 
    histScaleFactor = 0;
  
  // Draw histogram
  for(i=0; i<256; i++) {
    for(j=0; j<32; j++) {
      if(j < calcHistLength(i))
	XPutPixel(hist_color_image1, j, i, pixels_return[RED]);
      else
	XPutPixel(hist_color_image1, j, i, pixels_return[WHITE]);
    }
  }

  // Redraw the colormap window
  for (i=255; i>=0; i--)
    for (j=0; j<32; j++)
      XPutPixel(color_image1, j, i, pixels_return[(256-i)]);
  
  if(XtIsManaged(colorScreenW)) {
    XPutImage(dsply, hist_color_window1, hist_color_gc, hist_color_image1, 
	      0, 0, 0, 0, 32, 256);
    XPutImage(dsply, color_window1, color_gc, color_image1, 
	      0, 0, 0, 0, 32, 256);
    XFlush (dsply);
  }
  
  // Draw the colormap window for good measure
  DrawColormapWindow(NULL, NULL, NULL);

} // end function histogram_activate



void mapMovie_activate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: mapMovie_activate
  // Display map sections as a movie

  int tag = *((int *) client_data);

  // Put up the proper widgets in the Form
  makingMovie= TRUE;
  XtManageChild(playFormW);
  XtManageChild(movieSaveW);
  XtManageChild(movieMinW);
  XtManageChild(movieMaxW);
  XtManageChild(maxLabelW);
  XtManageChild(minLabelW);
  XtUnmanageChild(undoW);
  XtUnmanageChild(savePaintW);
  XtUnmanageChild(paintFormW);
  XtSetArg(arglst[0], XmNsensitive, TRUE);
  XtSetValues(playW, arglst, 1); 
  XtSetValues(stopW, arglst, 1); 
  XtSetValues(infiniW, arglst, 1); 
  XtSetValues(rockW, arglst, 1); 
  XtSetValues(flashLightW, arglst, 1); 

  doDialogTitle("Movie Display", editBoxScreenW, FALSE);
  
  // Setup the movie min/max scroll bars
  startMovie = 0;
  endMovie = (dispSections) ? numSecs-1 : mapNumSecs-1;

  sprintf(outStr,"%d",startMovie+1);
  doLabel(outStr, minLabelW, DEFAULT);
  
  sprintf(outStr,"%d",endMovie+1);
  doLabel(outStr, maxLabelW, DEFAULT);
  
  XtSetArg(arglst[0], XmNmaximum, numSecs);
  XtSetArg(arglst[1], XmNminimum, 0);
  XtSetArg(arglst[2], XmNsliderSize, 1);
  XtSetArg(arglst[3], XmNpageIncrement, 5);

  XtSetArg(arglst[4] ,XmNvalue, 0);
  XtSetValues(movieMinW, arglst, 5);
  XtSetArg(arglst[4] ,XmNvalue, numSecs-1);
  XtSetValues(movieMaxW, arglst, 5);
  
  raiseIt(editBoxScreenW);
  stuffDisplayMapBuffer(0,editBoxZoom,0);
  if(shadedMaps && stereoMapShade)
    stuffDisplayMapBuffer(1,editBoxZoom,numCols);
  DrawEditBoxWindow(NULL, NULL, NULL);
  XFlush(dsply);
  
  if(tag==PAINTMAP) {
    int i;
    paintMapSwitch = TRUE;
    XtUnmanageChild(movieSaveW); // Doesn't make sense
    if(!paintMap) {
      paintMap = (float *)malloc(sizeof(float)*numCols*numRows*origNumSecs);
      for(i=0; i<origNumSecs*numRows*numCols; i++)
	paintMap[i] = (normalExcise) ? 0.0 : float_image_buffer1[i];
    }
    
    askThem("Mouse modes for Excise Cuts:\n\nLeft=Small Cut  \nMiddle=Small UNDO  \nRight=LARGE Cut  \n\nUse File/Write/MAP to write cutout to disk.",
	    XmDIALOG_INFORMATION); 
  }

} // end function mapMovie_activate



void mapShade_activate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: mapShade_activate
  // Display map sections as depth-cue shaded 3D surface
  
  if(paintMapSwitch) { // Just do the paintMap
    mapShade_activate_work(&paintMap, &map_image_buffer, FALSE);
  } 
  else { // Normal mode of operation
    mapShade_activate_work(&float_image_buffer1, 
			   &map_image_buffer, FALSE);
    
    if(subtract_image_buffer2) // Surface shade the second one also
      mapShade_activate_work(&subtract_image_buffer2,
			     &subtract_map_image_buffer2, TRUE);
    
    if(subtract_image_buffer1) // Surface shade the second one also
      mapShade_activate_work(&subtract_image_buffer1,
			     &subtract_map_image_buffer, TRUE);
    
    if(subtract_image_buffer2 || subtract_image_buffer1)
      diffSec_really_doit();
  }
} // end function mapShade_activate



void rotate_xyz_activate(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: rotate_xyz_activate
  
  // Implement an xyz rotation for the 3D rendering screen based on
  // the values chosen in the theta/phi/omega widgets.  We were
  // planning on just doing an XYZ to TPO conversion and then letting
  // the 3D rendering rtns just go on their merry way using the TPO
  // like normal. We have yet to devise the XYZ to TPO conversion
  // scheme yet.

  rotateXYZ = !rotateXYZ;
  
  // Check rotateXYZ and call XDY's routines as appropriate e.g. We
  // might want to convert globPhi back and forth and update the text
  // widgets as appropriate here
  
  if(rotateXYZ) {
    doLabel("X", thetaButtonW, FALSE);
    doLabel("Y", phiButtonW,   FALSE);
    doLabel("Z", omegaButtonW, FALSE);
  }
  else {
    doLabel("Theta", thetaButtonW, FALSE);
    doLabel("Phi",   phiButtonW,   FALSE);
    doLabel("Omega", omegaButtonW, FALSE);
  }

} // end function rotate_xyz_activate



void foldAxis_activate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: foldAxis_activate
  // Change viewing axis

  globPhi = 360.0;
  globOmega = 360.0;

  // Set globTheta based on choice of viewing axis
  if(widget == fold5W)       // five fold
    globTheta= 360.0+32.0;
  else if(widget == fold3W)  // three fold
    globTheta = 360.0+69.0;
  else if(widget == fold2W)  // two fold
    globTheta = 360.0+0.0;
  
  // We've got to decide here if what we've got is XYZ or TPO
  // check the rotateXYZ logical and then call XDY's rtns as appropriate
  // XYZ-TPO

  // Now set the scrollbar to the new value
  XtVaSetValues(thetaScrollW, XmNvalue, (int)globTheta, NULL);
  XtVaSetValues(phiScrollW,   XmNvalue, (int)globPhi,   NULL); 
  XtVaSetValues(omegaScrollW, XmNvalue, (int)globOmega, NULL); 

  globTheta -= 360.0;
  globTilt = (globTheta != 0.0) ? TRUE : FALSE;
  
  globPhi -= 360.0;
  globTilt = (globPhi != 0.0) ? TRUE : FALSE;
  
  globOmega -= 360.0;
  globTilt = (globOmega != 0.0) ? TRUE : FALSE;
  
  // Give feedback
  sprintf(outStr,"%.0f",globTheta);
  doValue(outStr, thetaValueW, FALSE);
  
  sprintf(outStr,"%.0f",globOmega);
  doValue(outStr, omegaValueW, FALSE);
  
  sprintf(outStr,"%.0f",globPhi);
  doValue(outStr, phiValueW, FALSE);
  
  if(immediateMapShade)
    mapShade_activate(NULL, NULL, NULL);

} // end function foldAxis_activate



void DrawThermometer()
{
  // function name: DrawThermometer
  // Draw the "thermometer" to indicate movie speed on editBox window

  XDrawLine(dsply, XtWindow(playFormW), image_gcRED,   1, 1, 50,  1);
  XDrawLine(dsply, XtWindow(playFormW), image_gcRED,   1, 2, 50,  2);
  XDrawLine(dsply, XtWindow(playFormW), image_gcGREEN, 1, 1, 50-movieSpeed, 1);
  XDrawLine(dsply, XtWindow(playFormW), image_gcGREEN, 1, 2, 50-movieSpeed, 2);
  XFlush(dsply);

} // end function DrawThermometer



void DrawFlickThermometer()
{
  // function name: DrawFlickThermometer
  // Draw "thermometer" to indicate the flicker speed on Subtract Map window

  int tmp;
  tmp = 450 - flickSpeed;

  XDrawLine(dsply, XtWindow(subScreenW), image_gcRED,   401,95,450,95);
  XDrawLine(dsply, XtWindow(subScreenW), image_gcRED,   401,96,450,96);
  XDrawLine(dsply, XtWindow(subScreenW), image_gcGREEN, 401,95,tmp,95);
  XDrawLine(dsply, XtWindow(subScreenW), image_gcGREEN, 401,96,tmp,96);
  XFlush(dsply);

} // end function DrawFlickThermometer



void flickArrowActivate(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: flickArrowActivate
  // Change speed of flicker

  int numClicks;

  XmPushButtonCallbackStruct *temp = (XmPushButtonCallbackStruct *) call_data;
  numClicks = temp->click_count;

  if(widget == subFlickUpW)
    flickSpeed -= numClicks;
  else if(widget == subFlickDownW)
    flickSpeed += numClicks;
  
  flickSpeed = MAX(0,MIN(flickSpeed, 50));
  DrawFlickThermometer();
  
} // end function flickArrowActivate



void mapMovieArrowActivate(Widget widget, XtPointer client_data, 
			   XtPointer call_data)
{
  // function name: mapMovieArrowActivate
  // Change speed at which frames of movie are displayed
  
  int numClicks;
  XmPushButtonCallbackStruct *temp = (XmPushButtonCallbackStruct *) call_data;
  numClicks = temp->click_count;
  
  if(widget == movieArrowUpW)
    movieSpeed -= numClicks;
  else if(widget == movieArrowDownW)
    movieSpeed += numClicks;
  
  movieSpeed = MAX(0,MIN(movieSpeed, 50));
  DrawThermometer();
  
} // end function mapMovieArrowActivate



void editBox_activate(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: editBox_activate
  //  They want to edit their boxes. Stuff the first one up there for
  //  them and show them the screen. The editBox screen is also used
  //  for painting and showing 3D rendered movies.

  TextSetInt(editBoxTextW, curBox+lastNumBox+1, TRUE);
  
  makingMovie= FALSE;
  doDialogTitle("Box Editor", editBoxScreenW, FALSE);
  raiseIt(editBoxScreenW);
  
  XtManageChild(undoW);
  XtManageChild(savePaintW);
  XtManageChild(paintFormW);
  XtUnmanageChild(playFormW);
  XtUnmanageChild(movieSaveW); 
  XtUnmanageChild(movieMinW);
  XtUnmanageChild(movieMaxW);
  XtUnmanageChild(maxLabelW);
  XtUnmanageChild(minLabelW);
  
  if(fileType==PACKEDPORTABLE)
    stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
  else 
    stuffDisplayBoxBuffer(curBox, image_fileName.addr, editBoxZoom);
  
  DrawEditBoxWindow(NULL, NULL, NULL);

} // end function editBox_activate



void editBox_save_activate(Widget widget, XtPointer client_data, 
			   XtPointer call_data)
{
  // function name: editBox_save_activate
  // Save all edits on this nbox. Save in memory buffer on the screen

  int i, j, thisZoomFactor, thisRadius, yOut, yIn, outHeight, outWidth;
  float *thisBuf;

  if(widget == savePaintW) {
    if(fileType != MAP) { // Save the display buffer info
      
      // Calc the zoom
      box[curBox].boxZoom= thisZoomFactor = editBoxZoom;
      calcBoxSizeFactor(thisZoomFactor);
      
      box[curBox].inMemory = TRUE;  // Make them paint it again
      
      if(fileType == PACKEDPORTABLE) { // packed file input
	thisBuf = packImg[curBox];
	thisRadius = packRadius;
	outHeight = purdueHeader.numRows;
	outWidth  = purdueHeader.numCols;
      } else { // boxing from image file
	thisBuf = box[curBox].edit_mirror_buffer;
	thisRadius = boxRadius;
	outHeight = (thisRadius*2)+1;
	outWidth  = (thisRadius*2)+1;
      }
      
      // Reallocate the memory in case we had a larger one last time
      freeNullFloat(&thisBuf);

      CK(thisBuf = (float *)malloc(sizeof(float)*(outWidth*outHeight)));
      
      for (i=0; i<outHeight; i++) {
	yOut = i*outWidth;
	yIn  = (i*sizeFactor) * BOX_WIDTH;
	for (j=0; j<outWidth; j++)
	  thisBuf[yOut+j] = editBox_mirror_buffer[yIn+(int)(j*sizeFactor)];
      }
      
      if(fileType == PACKEDPORTABLE)
	packImg[curBox]= thisBuf;
      else 
	box[curBox].edit_mirror_buffer= thisBuf;
      
    } // Save the display buffer info    
    DrawEditBoxWindow(NULL, NULL, NULL);

  } else if(widget == movieSaveW) {
    
    system("rm robem*.pcx"); 
    
    // last arg == 1 --> movie mode
    writePCX("robem", startMovie, endMovie, 1);

    // Following code not working ...
    // system("rm \\~/robem\\*.pcx"); 
    // sprintf(outStr,"mv robem\\*.pcx ~/");
    // printf("[%s]\n",outStr);
    // system(outStr);
    // system("convert robem*.pcx movie.gif ");
    // system("rm robem*.pcx");  

    system("convert robem*.pcx movie.gif ");
    askThem("Movie saved in PCX format in robem*.pcx files.", 
	    XmDIALOG_INFORMATION); 
  }
  
} // end function editBox_save_activate



void editBox_undo_activate(Widget widget, XtPointer client_data, 
			   XtPointer call_data)
{
  // function name: editBox_undo_activate
  // Undo edits on this box. Restuff the box from disk and show them
  // the screen.

  int i, j;

  if(fileType != PACKEDPORTABLE) { // Make them paint it again
    box[curBox].inMemory = FALSE;  
    free(box[curBox].editBox_buffer);
  } else { // redraw from the mirror_buffer
    for (i=0; i<BOX_HEIGHT; i++) {
      for (j=0; j<BOX_WIDTH; j++) {
	
	XPutPixel(editBox_image1, x, y, 
		  pixels_return[(int)
				((editBox_undo_buffer[((i*BOX_WIDTH)+j)]
				  *scalePackFactor)+6.0)]);
	editBox_display_buffer[((y*BOX_WIDTH)+x)]=
	  (BYTE)(int)((editBox_undo_buffer[((i*BOX_WIDTH)+j)]
		       *scalePackFactor)+6.0);
      }
    }
  }

  if(fileType==PACKEDPORTABLE)
    stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
  else 
    stuffDisplayBoxBuffer(curBox,image_fileName.addr,editBoxZoom);
  
  DrawEditBoxWindow(NULL, NULL, NULL);
  
} // end function editBox_undo_activate



void DrawEditBoxWindow(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: DrawEditBoxWindow
  // Redraw window e.g. on expose events or whenever it's prudent for
  // the editBox drawing area.

  int draw_width = BOX_WIDTH, draw_height = BOX_HEIGHT;

  if((numBoxes > 0 || makingMovie) ||
     (fileType==PACKEDPORTABLE && purdueHeader.nimg > 0)) {
    
    if(liteSwitch){
      draw_width  = numCols;
      draw_height = numRows;
    }
    
    if(XtIsManaged(editBoxScreenW))
      XPutImage(dsply, editBox_window1, editBox_gc, editBox_image1, 0, 0,
		0, 0, draw_width, draw_height);
    XFlush (dsply);
    
    DrawThermometer();
  }
  
} // end function DrawEditBoxWindow



void editBoxText_changed(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{
  // function name: editBoxText_changed
  // Whenever they change the value of box number text widget call
  // here and make the interface display the appropriate box.

  char *editBoxText;

  if(!inited) return;

  editBoxText = XmTextGetString(editBoxTextW); 
  
  if(!EQU(editBoxText,"")) { // Don't do in the interim state
    if(makingMovie) { // Display the movie frame
      curBox = (int)strtod(editBoxText, 0);
      if(shadedMaps && stereoMapShade)
	curBox = MAX(1,MIN(curBox,(numSecs-1)));
      else 
	curBox = MAX(1,MIN(curBox,(numSecs)));
      curBox--;
      
      // Draw new box
      stuffDisplayMapBuffer(curBox,editBoxZoom,0);
      if(shadedMaps && stereoMapShade)
	stuffDisplayMapBuffer(curBox+1,editBoxZoom,numCols);
      
      DrawEditBoxWindow(NULL, NULL, NULL);
    }
    else { // Display box
      curBox = (int)strtod(editBoxText, 0);
      if(fileType==PACKEDPORTABLE)
	curBox = MAX(1,MIN(curBox,(purdueHeader.nimg)));
      else
	curBox = MAX((1), MIN(curBox,(numBoxes)));
      // Computer numbers from zero not one
      curBox--;
      
      // Draw new box
      if(fileType==PACKEDPORTABLE)
	stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
      else 
	stuffDisplayBoxBuffer(curBox,image_fileName.addr, editBoxZoom);
      
      DrawEditBoxWindow(NULL, NULL, NULL);
    }
  }
  XtFree(editBoxText);
  
  TextSetInt(editBoxTextW, curBox+lastNumBox+1, TRUE);
  
} // end function editBoxText_changed



void editBoxZoomText_changed(Widget widget, XtPointer client_data, 
			     XtPointer call_data)
{
  // function name: editBoxZoomText_changed
  // Whenever they change the value of the editBox zoom text widget
  // call here. Redraw the box based on the zoom.

  if(!inited) return;

  XFlush(dsply);
  editBoxZoom = (int)wstrtod(editBoxZoomTextW);
  
  if(fileType==PACKEDPORTABLE)
    stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
  else if(makingMovie) {
    stuffDisplayMapBuffer(curBox, editBoxZoom, 0);
    if(shadedMaps && stereoMapShade)
      stuffDisplayMapBuffer(curBox+1, editBoxZoom, numCols);
  }
  else
    stuffDisplayBoxBuffer(curBox, image_fileName.addr, editBoxZoom);
  
  DrawEditBoxWindow(NULL, NULL, NULL);
  
} // end function editBoxZoomText_changed



void editBoxZoomArrowActivate(Widget widget, XtPointer client_data, 
			      XtPointer call_data)
{ 
  // function name: editBoxZoomArrowActivate
  // Change current concept of zoom
  
  if(widget == editBoxZoomUpW)
    editBoxZoom++;
  else if(widget == editBoxZoomDownW)
    editBoxZoom--;
  
  TextSetInt(editBoxZoomTextW, editBoxZoom, FALSE);
  
  if(fileType == PACKEDPORTABLE)
    stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
  else if(makingMovie) {
    stuffDisplayMapBuffer(curBox, editBoxZoom, 0);
    if(shadedMaps && stereoMapShade)
      stuffDisplayMapBuffer(curBox+1, editBoxZoom, numCols);
  }
  else
    stuffDisplayBoxBuffer(curBox, image_fileName.addr, editBoxZoom);
  
  DrawEditBoxWindow(NULL, NULL, NULL);
  XFlush(dsply);
  
} // end function editBoxZoomArrowActivate



void editBoxPaintText_changed(Widget widget, XtPointer client_data, 
			      XtPointer call_data)
{
  // function name: editBoxPaintText_changed
  // Change brush size for painting

  paintSize = (int)wstrtod(editBoxPaintTextW);
  paintSize = MAX(1,MIN(paintSize, 50));

} // end function editBoxPaintText_changed



void em_sleep(int count)
{
  // function name: em_sleep
  // Do some calculations that use up time
  int i,j,k;
  float a,b,c,d;

  for(i=0; i<count; i++) {
    for (j=0; j<10000; j++) {
      a = j / (1 + 2.2);
      for (k=0; k<1000; k++) {
	d = k / (1 + 5.5);
	b = a + d;
	c = b / 7.7;
      }
    }
  }

} // end function em_sleep



void boxgrid2_activate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: boxgrid2_activate
  // Try putting a default grid of boxes out there and then do a
  // reference refine on them. This is an experiment for autoboxing
  // empty particles.
  
  int i, j, m, n;
  int save_file_width, save_file_height, saveRefAuto, saveDCR;
  int save_boxRadius, save_emmaRadius, save_emptyShellWidth;
  int shrinkFactor, numAccum, compressed;
  float *save_inCore_buffer, accumVal;
  
  //  Save what we've got now 
  save_inCore_buffer   = inCore_buffer;
  save_file_width      = file_width;
  save_file_height     = file_height;
  save_boxRadius       = boxRadius;
  save_emmaRadius      = emmaRadius; 
  save_emptyShellWidth = emptyShellWidth; 
  
  // Compute how much we need to shrink
  compressed   = FALSE;
  shrinkFactor = 1;
  while (!compressed) {
    if(boxRadius/shrinkFactor < 20)
      compressed = TRUE;
    else
      shrinkFactor *= 2;
  }
  
  printf("Shrinking by [%d]\n",shrinkFactor);
  
  // Now shrink it all 
  if(shrinkFactor > 1) {
    file_height      /= shrinkFactor;
    file_width       /= shrinkFactor;
    boxRadius        /= shrinkFactor;
    emmaRadius       /= shrinkFactor;
    emptyShellWidth  /= shrinkFactor;
    
    CK(inCore_buffer = (float *)malloc(sizeof(float)*
				       file_height*file_width));
    numAccum = shrinkFactor * shrinkFactor;
    
    for (i=0; i<file_height; i++) {
      for (j=0; j<file_width; j++) {
	accumVal = 0.0;
	for (m=0; m<shrinkFactor; m++) {
	  for (n=0; n<shrinkFactor; n++) {
	    accumVal += 
	      save_inCore_buffer[(((i*shrinkFactor)+m)*
				  save_file_width)+
				 ((j*shrinkFactor)+n)];
	  }
	}
	inCore_buffer[(i*file_width)+j] = accumVal / (float)numAccum;
      }
    }
  }
  
  fullScreen = FALSE;
  virtual_read_activate(-99,-99);
  set_boxgrid_default();  
  DrawWindow();
  
  firstAMMSTDEVtime = TRUE;
  saveRefAuto = refAutomate;
  refAutomate = TRUE;
  saveDCR = darkCircleRef;
  darkCircleRef =  99;
  ref_correlate_activate();
  refAutomate = saveRefAuto;
  darkCircleRef = saveDCR;
  
  //  Restore what we had
  if(shrinkFactor > 1)
    free(inCore_buffer);
  
  inCore_buffer   = save_inCore_buffer;
  file_width      = save_file_width;
  file_height     = save_file_height;
  boxRadius       = save_boxRadius;
  emmaRadius      = save_emmaRadius; 
  emptyShellWidth = save_emptyShellWidth;
  
  //  Rejigger all of the box coords
  for (i=0; i<numBoxes; i++) {
    box[i ].x *= shrinkFactor;
    box[i ].y *= shrinkFactor;
  }
  
  fullScreen = FALSE;
  virtual_read_activate(-99,-99);
  
} // end function boxgrid2_activate



void set_boxgrid_default() 
{
  // function name: set_boxgrid_default
  // Set the default grid of boxes
  
  int i, j, aveX, aveY;
  
  // Using 1.5 below may result in too many ptles according to old comments
  aveX = emmaRadius * 1.5;
  aveY = emmaRadius * 1.5;
  
  for (i=emmaRadius; i<file_height; i+=aveY) {
    for (j=emmaRadius; j<file_width; j += aveX) {
      if(numBoxes < MAXBOX) {
	box[numBoxes].x = j;
	box[numBoxes].y = i;
	box[numBoxes].reference = FALSE;
	box[numBoxes].correlation = 1.0;
	box[numBoxes].avgdensity = 0.0;
	box[numBoxes].stdev = 0.0;
	numBoxes ++;
      }
    }
  }
  
} // end function set_boxgrid_default



void boxgrid_activate(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: boxgrid_activate
  // Put a default grid of boxes and do a reference on them

  int *passTag;
  int tag = *((int *) client_data);
  passTag = (int *)malloc(sizeof(int));
  
  if(tag==-99 || firstGridding) {
    set_boxgrid_default();
    DrawWindow();
  }
  
  firstGridding = FALSE;
  
  if(tag != -99) { // do first time
    firstAMMSTDEVtime = TRUE;
    *passTag= CORRELATE3;
    correlate_activate(widget, (XtPointer)passTag, call_data);
    free(passTag);
  }

} // end function boxgrid_activate



void stepMode_button_change(Widget widget, XtPointer client_data, 
			    XtPointer call_data)
{
  // function name: stepMode_button_change
  // Remember parameters stepped in the movie module

  int tag = *((int *) client_data);

  if((tag!=SURFDENS)  && (tag!=FRACT)      && (tag!=ANGX) &&
     (tag!=ANGY)      && (tag!=THETA)      && (tag!=PHI) &&
     (tag!=OMEGA)     && (tag!=UNBLOBSIZE) && (tag!=UNBLOBRADIUS) &&
     (tag!=MAXX)      && (tag!=MAXY)       && (tag!=MAXZ) &&
     (tag!=MAXRADIUS) && (tag!=PIXELSIZE) ) { 
    return; // Valid tag not found
  }

  sprintf(outStr,"Theta Step:");
  XtSetArg(arglst[0], XmNsensitive, TRUE);
  XtSetValues(movieFormW, arglst, 1); 
  XtSetValues(stepToggleW, arglst, 1); 
  XtSetValues(stepToggle2W, arglst, 1); 
  XtSetValues(frameTextW, arglst, 1); 
  XtSetValues(stepTextW, arglst, 1); 
  XtSetValues(stepValueLabelW, arglst, 1); 
  XtSetValues(numFramesLabelW, arglst, 1); 
  XFlush(dsply);

  if(whichStep==tag) { // They are toggling the same one
    if(whichStep != NONE) { // They turned it off
      whichStep = NONE;
      sprintf(outStr,"NO Step:");
      XtSetArg(arglst[0], XmNsensitive, FALSE);
      XtSetValues(stepToggleW, arglst, 1); 
      XtSetValues(stepToggle2W, arglst, 1); 
      XtSetValues(frameTextW, arglst, 1); 
      XtSetValues(stepTextW, arglst, 1); 
      XtSetValues(stepValueLabelW, arglst, 1); 
      XtSetValues(numFramesLabelW, arglst, 1); 
      XtSetValues(movieFormW, arglst, 1); 
    }
  }
  else {
    whichStep = tag;
    if(tag==SURFDENS)
      sprintf(outStr,"SurfDens Step:");
    else if(tag==FRACT)
      sprintf(outStr,"Fract Step:");
    else if(tag==ANGX)
      sprintf(outStr,"Angx Step:");
    else if(tag==ANGY)
      sprintf(outStr,"Angy Step:");
    else if(tag==PHI)
      sprintf(outStr,"Phi Step:");
    else if(tag==OMEGA)
      sprintf(outStr,"Omega Step:");
    else if(tag==THETA)
      sprintf(outStr,"Theta Step:");
    else if(tag==UNBLOBSIZE)
      sprintf(outStr,"UnblobSize Step:");
    else if(tag==UNBLOBRADIUS)
      sprintf(outStr,"UnblobRadius Step:");
    else if(tag==MAXX)
      sprintf(outStr,"Cut X Step:");
    else if(tag==MAXY)
      sprintf(outStr,"Cut Y Step:");
    else if(tag==MAXZ)
      sprintf(outStr,"Cut Z Step:");
    else if(tag==MAXRADIUS)
      sprintf(outStr,"MaxRad Step:");
    else if(tag==PIXELSIZE)
      sprintf(outStr,"PixelSize Step:");
  }
  
  XtSetArg(arglst[0], XmNlabelPixmap, moviePixmap);
  XtSetArg(arglst[1], XmNselectPixmap, unmoviePixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(surfDensToggleW, arglst, 3); 
  XtSetValues(fractToggleW, arglst, 3);
  XtSetValues(angxToggleW, arglst, 3);
  XtSetValues(angyToggleW, arglst, 3);
  XtSetValues(thetaToggleW, arglst, 3);
  XtSetValues(phiToggleW, arglst, 3);
  XtSetValues(omegaToggleW, arglst, 3);
  XtSetValues(maxRadiusToggleW, arglst, 3);
  XtSetValues(pixelSizeToggleW, arglst, 3);
  XtSetValues(unblobsizeToggleW, arglst, 3);
  XtSetValues(unblobradiusToggleW, arglst, 3);
  XtSetValues(maxxToggleW, arglst, 3);
  XtSetValues(maxyToggleW, arglst, 3);
  XtSetValues(maxzToggleW, arglst, 3);
  
  // Now turn this one on
  if(whichStep != NONE) { // Set this one ON
    XtSetArg(arglst[0], XmNlabelPixmap, unmoviePixmap);
    XtSetArg(arglst[1], XmNselectPixmap, moviePixmap);
    XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
    XtSetValues(widget, arglst, 3); 
  }
  
  // Now make them decide on a new numFrames and step value
  TextSet(frameTextW, "", FALSE); 
  
  doLabel(outStr, stepValueLabelW, FALSE);
  
  if(immediateMapShade)
    mapShade_activate(NULL, NULL, NULL);

} // end function stepMode_button_change



void stepMode_change(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: stepMode_change
  // Remember parameters stepped in movie

  int tag = *((int *) client_data);

  if(tag!=IREV)
    return; // Valid tag not found
  
  sprintf(outStr,"Theta Step:");
  
  if(whichStep==tag) { // They are toggling the same one
    if(whichStep != NONE) { // They turned it off
      whichStep = NONE;
      sprintf(outStr,"NO Step:");
    }
  }
  
  // These don't have to do with the whichStep ones
  irev = -irev; // Only allowed values +/- 1
  
  // Give feedback on what the surface density *really* is...
  // Make sure the user sees it with the right sign
  sprintf(outStr,"%.2f",irev*surfaceDensity);
  TextSet(surfDensValueW, outStr, FALSE);
  
  if(immediateMapShade)
    mapShade_activate(NULL, NULL, NULL);
  
} // end function stepMode_change



void drawFFTcontour()
{
  // function name: drawFFTcontour
  // Draw contour lines in FFT contour window

  int i, numPts, numColsTsizeFactor, nCols, nRows;
  int icx=0, icy=0;
  short *short_display_image;
  float *xp, *yp;
  float sclx, scly, tour;
  short ix[40000], iy[40000], ix1[40000], iy1[40000];
  float minF, maxF;
  float *tempFFTcontour_buffer;

  if(contourToggle) {
    cosang = 0.0;          // 90.0 degrees
    sinang = 1.0;          // 90.0 degrees
    
    // ??? deg
    cosang = 1.0;
    sinang = 1.0;
    
    // These are for scalex and scale Y - get from zoomfactor
    sclx = 1.0;
    scly = 1.0;
    
    nCols = 128;
    nRows = 128;
    
    numColsTsizeFactor = 128;
    
    CK(xp = (float *)malloc(sizeof(float)*(128+1)));
    CK(yp = (float *)malloc(sizeof(float)*(128+1)));
    CK(short_display_image = (short *)malloc(sizeof(short)*(128+1)*
					     (128+1)));
    CK(tempFFTcontour_buffer = (float *)malloc(sizeof(float)*(128+1)*
					       (128+1)));
    
    minF =  HUGE;
    maxF = -HUGE;  

    for (i=0; i<128*128; i++) {
      tempFFTcontour_buffer[i] = (float)fft_contour_buffer[i];
      minF= MIN(tempFFTcontour_buffer[i], minF);
      maxF= MAX(tempFFTcontour_buffer[i], maxF);
    }

    for(tour = (maxF-1.0);tour > (minF);tour -= ((maxF-minF)/4.0)) {
      
      numPts = 1;
      
      for(i=0;i<40000;i++)
	ix[i]=iy[i]=ix1[i]=iy1[i]= 0;
	  
      contur_a_rob__(tempFFTcontour_buffer, short_display_image,
		    xp, yp, &nCols, &nRows, &sclx, &scly, &tour, 
		    ix, iy, ix1, iy1, &numPts,
		    &cosang, &sinang, &icx, &icy);
      
      for (i=0; i<(numPts-1); i++) { // Draw all points
	
	if(ix[i]  >= (numColsTsizeFactor*3) || 
	   ix1[i] >= (numColsTsizeFactor*3)) {
	  ix[i]  -= (numColsTsizeFactor*3);
	  ix1[i] -= (numColsTsizeFactor*3);
	}
	else if(ix[i]  >= (numColsTsizeFactor*2) || 
		ix1[i] >= (numColsTsizeFactor*2)) {
	  ix[i]  -= (numColsTsizeFactor*2);
	  ix1[i] -= (numColsTsizeFactor*2);
	}
	else if(ix[i]  >= (numColsTsizeFactor) || 
		ix1[i] >= (numColsTsizeFactor)) {
	  ix[i]  -= numColsTsizeFactor;
	  ix1[i] -= numColsTsizeFactor;
	}
	
	ix[i]  *= 1;    // Put it back into 128 space
	ix1[i] *= 1;
	iy[i]  *= 1;
	iy1[i] *= 1;
	
	// Just draw with X
	if(tour < (maxF*0.5))
	  XDrawLine(dsply, fftContour_window1, image_gcRED,
		    (int)ix[i], (int)iy[i], (int)ix1[i], (int)iy1[i]);
	else if((tour < (maxF-1.0)) && (tour >= (maxF*0.5)))
	  XDrawLine(dsply, fftContour_window1, image_gcBLUE,
		    (int)ix[i], (int)iy[i], (int)ix1[i], (int)iy1[i]);
	else
	  XDrawLine(dsply, fftContour_window1, image_gcGREEN,
		    (int)ix[i], (int)iy[i], (int)ix1[i], (int)iy1[i]);
	
      }	  
    }
    
    free(xp);
    free(yp);
    free(short_display_image);
  }

} // end function drawFFTcontour



void RwaActivateWidget(Widget w)
{
  // function name: RwaActivateWidget
  // Activate the widget passed in. This routine makes the interface
  // behave just as if you had pressed a push button to activate it.

  if(XmIsGadget(w))
    (*((XmGadgetClass)XtClass(w))->gadget_class.arm_and_activate)(w,0,0,0);
  else if(XmIsPrimitive(w))
    (*((XmPrimitiveWidgetClass)
       XtClass(w))->primitive_class.arm_and_activate)(w,0,0,0);
  
  return;
  
} // end function RwaActivateWidget



void movieMinMax_change(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: movieMinMax_change
  // Limit movie frames from min to max

  int tag = *((int *) client_data);
  XmScrollBarCallbackStruct *temp = (XmScrollBarCallbackStruct *) call_data;

  if(tag == MOVIEMIN) { // surf changed
    startMovie = MIN(temp->value, endMovie);
    XtVaSetValues(movieMinW, XmNvalue, startMovie, NULL);
  }
  else if(tag == MOVIEMAX) { // surf changed
    endMovie = MAX(temp->value, startMovie);
    XtVaSetValues(movieMaxW, XmNvalue, endMovie, NULL);
  }
  else if(tag == BREAKPOINT) { // delta changed
    breakCont = contourLevel;
    sprintf(outStr,"%.0f",breakCont);
    doLabel(outStr, movconBrkLW, DEFAULT);
  }
  else if(tag == DELTA) { // delta changed
    deltaCont = temp->value;
    sprintf(outStr,"%.0f",deltaCont);
    doLabel(outStr, movconDelLW, DEFAULT);
  }
  else if(tag == NUMCONT) { // numcont changed
    numCont = temp->value;
    sprintf(outStr,"%d",numCont);
    doLabel(outStr, movconNumLW, DEFAULT);
  }
  else if(tag == CONTOUR) { // contour changed
    stuffDisplayMapBuffer(curBox,editBoxZoom,0);
    DrawEditBoxWindow(NULL, NULL, NULL);
    XFlush(dsply);
    contourLevel = temp->value;
    sprintf(outStr,"  %.0f  ",contourLevel);
    doLabel(outStr, movconConLW, DEFAULT);
    
    // Non-destructively draw it
    XFlush(dsply);
    drawContour(contourLevel, FALSE, curBox, editBoxW,
		float_image_buffer1,0,0,
		editBox_display_buffer);
  }    
  else if(tag == TRYIT) {
    int i;
    if(crysDisping) { // Special for crysDisp
      for (i=0; i<totContours; i++) // Non-destructively draw it
	drawContour(conTours[i], TRUE, 0, drawingW,
		    crys_float_image_buffer, 0, 0, image_buffer1);
    }
    else { // Normal
      for (i=0; i<numCont; i++) //  Non-destructively draw it
	drawContour((contourLevel+((float)i*deltaCont)), 
		    FALSE, curBox, editBoxW,float_image_buffer1,
		    0,0,editBox_display_buffer);
    }
  }
  
  if(tag == MOVIEMAX || tag == MOVIEMIN) { // OK to do this
//    int tag;
    sprintf(outStr,"%d",startMovie+1);  // +1 for humans
    doLabel(outStr, minLabelW, DEFAULT);
    
    sprintf(outStr,"%d",endMovie+1);    // +1 for humans
    doLabel(outStr, maxLabelW, DEFAULT);

//    tag= INFINIMAP;
    
    // How do we get it to start the infiniLoop again???
    // mapMovie_show_activate(0, &tag, 0);
    // RwaActivateWidget(infiniW); This messes up the mwm 
  }
  
} // end function movieMinMax_change



void  circAvg_pifOut_activate(Widget widget, XtPointer client_data, 
			     XtPointer call_data)
{
  // function name: circAvg_pifOut_activate
  // PifOut the circAvg

  int i, j, xcen, ycen, idist, luNum;
  off_t imgOffset;
  float *thisRint, *outBufF, maxF, minF;
  float scaleFactor, img_bkg, img_min, img_max, img_avg, img_std;
  fString_t fName, readWrite;
  char outFileName[256];
  BYTE *outBuf;
  forPifHead_t thisHead;
  
  fName.addr     = (char *)malloc(sizeof(char)*64);
  readWrite.addr = (char *)malloc(sizeof(char)*64);
  
  pif_init_head_(&thisHead);
  
  thisHead.numCols =  numCols;
  thisHead.numRows =  numSecs;
  CK(outBuf = (BYTE *)malloc(sizeof(BYTE)*
			     thisHead.numRows*thisHead.numCols));
  CK(outBufF = (float *)malloc(sizeof(float)*
			       thisHead.numRows*thisHead.numCols));
  
  xcen = numCols/2;
  ycen = numRows/2;
  thisRint = circAvg[selectedCircAvg].rint;

  for (j=0; j<numRows; j++) {
    for (i=0; i<numCols; i++) {
      idist = sqrt((float)( (j-ycen)*(j-ycen) + (i-xcen)*(i-xcen) ));
      outBufF[(j*numCols)+i]= thisRint[idist];
    }
  }
  
  calcStats(outBufF, &img_bkg,  &img_min, &img_max, &img_avg, &img_std,
	    thisHead.numCols, thisHead.numRows, thisHead.numCols, fileType,
	    bkgrndMethod, boxMethod, packRadius, boxRadius, featherRadius,
		recipSizeFactor, tryColors);
  
  minF = img_min;
  maxF = img_max;
  scaleFactor = 255.0 / (maxF-minF);

  // Now scale it
  for (j=0; j<thisHead.numRows*thisHead.numCols; j++)
    outBuf[j]= (BYTE)(((float)outBufF[j]-minF) * scaleFactor);
  
  calcByteStats(outBuf, 
		&img_bkg, &img_min, &img_max, &img_avg, &img_std,
		thisHead.numCols, thisHead.numRows, thisHead.numCols);  

  sprintf(outFileName,"spherAver.pif");
  sprintf(fName.addr,"%s",outFileName);
  sprintf(readWrite.addr,"w");

  luNum = 7;
  thisHead.luNum = luNum;

  pif_open(fName.addr, &luNum, readWrite.addr);

  thisHead.nimg    = 1;
  thisHead.htype   = 1;
  thisHead.numSecs = 1;  // No maps here

  thisHead.mode = 6;

  pif_write_gh_(&thisHead);
      
  thisHead.datai4_min = img_min;
  thisHead.datar4_min = (int)img_min;
  thisHead.datai4_max = img_max;
  thisHead.datar4_max = img_max;
  thisHead.datar4_avg = img_avg;
  thisHead.datar4_std = img_std;
  thisHead.img_bkg    = img_bkg;
  thisHead.imgNum     = 1;

  pif_write_dh_(&thisHead);
      
  // [3] Now write the image info
  imgOffset = 1024;  
  
  pif_write_byte_image_(&thisHead.luNum, outBuf, &imgOffset,
			&thisHead.numCols,
			&thisHead.numRows,
			&thisHead.numSecs);
  
  free(fName.addr);
  free(readWrite.addr);
  free(outBuf);
  free(outBufF);
  
  pif_close(&thisHead.luNum);
  
  askThem("spherAver.pif file writtten.", XmDIALOG_INFORMATION);
  
} // end function circAvg_pifOut_activate



void circAvg_print_activate(Widget widget, XtPointer client_data, 
			    XtPointer call_data)
{
  // function name: circAvg_print_activate
  // Print the circAvg

  int i, j, thisNterp;
  float *thisRint, *thisStDev;
  FILE *fOut; 

  if ( (fOut = fopen("spherical_average.dat","w")) != NULL ) {
    for (j=0; j<currentCircAvg; j++) {
      thisRint   = circAvg[j].rint;
      thisStDev  = circAvg[j].stDev;
      thisNterp  = circAvg[j].nterp;
      fprintf(fOut,"Map [%s] - number of points:[%d]\n",
	      circAvg[j].mapName, circAvg[j].nterp);
      for (i=0; i<thisNterp; i++) {
	fprintf(fOut,"%f  %f\n",thisRint[i],thisStDev[i]);
      }
    }
    fclose(fOut);
    askThem("spherical_average.dat file writtten.", XmDIALOG_INFORMATION);
  } else {
    askThem("Can't open spherical_average.dat file.", XmDIALOG_INFORMATION);
  }

} // end function circAvg_print_activate



void circAvg_screen_activate(Widget widget, XtPointer client_data, 
			     XtPointer call_data)
{
  // function name: circAvg_screen_activate
  // Put the circAvgscreen on the display

  int i, j, thisNterp;
  float *thisRint;

  // Draw the circular average already calculated when the MAP was read in

  minCA= R4_MAX;
  maxCA= R4_MIN;
  
  for (j=0; j<currentCircAvg; j++) {
    thisRint  = circAvg[j].rint;
    thisNterp = circAvg[j].nterp;
    for (i=0; i<thisNterp; i++) {
      if(thisRint[i] != 0.0) {
	minCA = MIN(minCA, thisRint[i]);
	maxCA = MAX(maxCA, thisRint[i]);
      }
    }
  }
  
  scaleCircFactor = 245.0 / (maxCA-minCA); 
  raiseIt(circAvgScreenW);

} // end function circAvg_screen_activate



void DrawCircAvgWindow(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: DrawCircAvgWindow
  // Redraw window on expose events or whenever prudent

  int i, j, scaledY, scaledYp1, thisX, thisXp1, thisNterp;
  float *thisRint;

  if(widget == circAvgClearW) {
    for (j=0; j<3; j++) {
      doLabel(" ", circAvgLabelW[j], FALSE);
      XtSetArg(arglst[0], XmNlabelString, XmStringCreateLocalized("---"));
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetArg(arglst[2], XmNforeground, drawCircs[j].pixelColor );
      XtSetValues(circAvgLineW[j], arglst, 3); 
      XFlush(dsply);
    }
    numCircAvgDraw= 0; 
  }
  
  if(XtIsManaged(circAvgScreenW))
    XPutImage(dsply, circAvg_window1, circAvg_gc, circAvg_image1, 0, 0,
	      0, 0, 256, 256);

  // Now draw axes
  XDrawLine(dsply, XtWindow(circAvgW), image_gcWHITE, 5, 0, 5, 256);
  XDrawLine(dsply, XtWindow(circAvgW), image_gcWHITE, 0, 251, 256, 251);

  // Draw zero line
  scaledY = 251.0 - ((0.0-minCA)*scaleCircFactor);
  XDrawLine(dsply, XtWindow(circAvgW), image_gcWHITE,
	    6, scaledY, 1024, scaledY);
  
  XFlush (dsply);
 
  for (j=0; j<numCircAvgDraw; j++) {
    thisRint  = circAvg[drawCircs[j].index].rint;
    thisNterp = circAvg[drawCircs[j].index].nterp;  
    for (i=caOffsetX; i<(thisNterp-1); i++) {
      scaledY   = (thisRint[i]-minCA)*scaleCircFactor;
      scaledYp1 = (thisRint[i+1]-minCA)*scaleCircFactor;
      scaledY   = 245.0 - scaledY + 6;
      scaledYp1 = 245.0 - scaledYp1 + 6;
      thisX   = (i-caOffsetX)*circAvgSize + 6;
      thisXp1 = (i-caOffsetX)*circAvgSize + 7;
      XDrawLine(dsply, XtWindow(circAvgW), drawCircs[j].colorGC,
		thisX, scaledY, thisXp1, scaledYp1);
    }
  }
  
  XFlush (dsply);
  
} // end function DrawCircAvgWindow



void circAvg_list_proc(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: circAvg_list_proc
  // Call here everytime they make a new choice form the circAvg
  // master list. Plot the appropriate circAvg already calculated
  // when the map was read in.

  Arg arglst[5];
  int foundOne, *passTag;
  char *thisStr;

  XmListCallbackStruct *list = (XmListCallbackStruct *) call_data;

  passTag = (int *)malloc(sizeof(int));

  foundOne = (int) (list->item_position);  // Numbers from 1 not 0
  foundOne--;  // Computer numbers from 0 not 1
  selectedCircAvg= foundOne;
  
  // We'll always fill up the last place regardless
  if(numCircAvgDraw==3)
    numCircAvgDraw= 2;

  if(numCircAvgDraw==0)
    drawCircs[numCircAvgDraw].colorGC = image_gcRED;
  else if(numCircAvgDraw==1)
    drawCircs[numCircAvgDraw].colorGC = image_gcGREEN;
  else if(numCircAvgDraw==2)
    drawCircs[numCircAvgDraw].colorGC = image_gcBLUE;


  // Remember these
  if(!XmStringGetLtoR(list->item, XmSTRING_DEFAULT_CHARSET, &thisStr))
    printf("Error extracting filename.\n");

  sprintf(drawCircs[numCircAvgDraw].fName,"%s",thisStr);
  XtFree(thisStr);
  drawCircs[numCircAvgDraw].index   = foundOne;
  drawCircs[numCircAvgDraw].pixelColor = pixels_return[(RED+numCircAvgDraw)];
  
  sprintf(outStr,"[%s] ",drawCircs[numCircAvgDraw].fName);
  doLabel(outStr, circAvgLabelW[numCircAvgDraw], FALSE);
  
  XtSetArg(arglst[0], XmNlabelString, XmStringCreateLocalized("---"));
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetArg(arglst[2], XmNforeground, drawCircs[numCircAvgDraw].pixelColor);
  XtSetValues(circAvgLineW[numCircAvgDraw], arglst, 3); 
  XFlush(dsply);
  
  numCircAvgDraw++;
  if(numCircAvgDraw > 2)
    numCircAvgDraw= 3;
  
  *passTag= 0;
  DrawCircAvgWindow(0, (XtPointer)passTag, 0);
  free(passTag);
  
} // end function circAvg_list_proc



void circAvgMotion(Widget widget, caddr_t unused, 
		   register XMotionEvent *pEvent)
{
  // function name: circAvgMotion
  // Whenever the pointer moves across our CircAvg drawing area(s)
  // call here. This rotuine basically takes care of drawing that
  // little white line you see when the cursor is dragged across the
  // circAvg areas.

  int x, y,foundIt,thisEntry;
  static int lastX= -999;
  static int lastI= -999;
  static short lastRed, lastGreen, lastBlue;
  int i, thisI;
  float *thisRint;

  // Make all changes current
  XFlush(dsply);               

  switch (pEvent->type) {
    
  case LeaveNotify:
    if(lastX > -99) {
      XDrawLine(dsply, XtWindow(circAvgW), image_gcBLACK,
		lastX, 252, lastX, 256);
      XDrawLine(dsply, XtWindow(circAvgW), image_gcBLACK,
		lastX, 1, lastX, 6);
    }
    if(lastI > -99) {
      color_defs[lastI].red    = lastRed;
      color_defs[lastI].green  = lastGreen;
      color_defs[lastI].blue   = lastBlue;
      XAllocColor (dsply, color_map, &color_defs[lastI]);
      pixels_return[lastI] = color_defs[lastI].pixel;
    }
    break;
    
  case MotionNotify:  
    // Get coords of pointer
    x = pEvent->x;     
    y = pEvent->y;
    
    if(lastX > -99) {
      XDrawLine(dsply, XtWindow(circAvgW), image_gcBLACK,
		lastX, 252, lastX, 256);
      XDrawLine(dsply, XtWindow(circAvgW), image_gcBLACK,
		lastX, 1, lastX, 6);
    }
    XDrawLine(dsply, XtWindow(circAvgW), image_gcWHITE,
	      x, 252, x, 256);
    XDrawLine(dsply, XtWindow(circAvgW), image_gcWHITE,
	      x, 1, x, 6);
    
    for (i=0; i<numCircAvgDraw; i++) {
      int thisX,thisNterp;
      float *thisRint;
      thisX= ((x-6)*(1.0/circAvgSize))+caOffsetX;
      thisRint  = circAvg[drawCircs[i].index].rint;
      thisNterp = circAvg[drawCircs[i].index].nterp;
      
      if(thisX < thisNterp)
	sprintf(outStr,"[%s] %d,%.2f",drawCircs[i].fName,thisX,
		thisRint[thisX]);
      else
	sprintf(outStr,"[%s] %d,UNDEF",drawCircs[i].fName,thisX);
      
      doLabel(outStr, circAvgLabelW[i], FALSE);
      
      XtSetArg(arglst[0], XmNlabelString, XmStringCreateLocalized("---"));
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetArg(arglst[2], XmNforeground, drawCircs[i].pixelColor );
      XtSetValues(circAvgLineW[i], arglst, 3); 
      XFlush(dsply);
    }
      
    if(radialMapShade && !dispSections && selectedCircAvg>=0) {
      // Change the colormap at this radius
      if(lastI > -99) {
	color_defs[lastI].red    = lastRed;
	color_defs[lastI].green  = lastGreen;
	color_defs[lastI].blue   = lastBlue;
	XAllocColor (dsply, color_map, &color_defs[lastI]);
	pixels_return[lastI] = color_defs[lastI].pixel;
      }
      
      // (/ circAvgSize) because we use (* circAvgSize) for display 
      // in circAvg_list_proc()
      y = (x - 6) / circAvgSize;
      y = MAX(y,0);
      i= MIN(tryColors-1,MAX(y,6));
      // i += pixels_return[0];
      thisI= i;
      
      thisRint = circAvg[selectedCircAvg].rint;
      foundIt = FALSE;	
      for (thisEntry=0; thisEntry<256; thisEntry++) {
	if(!foundIt) { // see if this is it
	  if((radialValue[thisEntry]<= (i)) &&  
	     (radialValue[thisEntry+1]>= (i)) ) {
	    thisI= thisEntry+pixels_return[0];
	    foundIt= TRUE;
	  }
	}
      }
      
      // Make sure we're in range
      thisI= MIN(pixels_return[tryColors-1], MAX(thisI,pixels_return[6]));
      
      // Tell them what position in the colormap we are
      sprintf(outStr,"%3d",thisI);
      doLabel(outStr, colorValueW[ENTRY], FALSE);
      
      if(thisI==pixels_return[0]+6)
	sprintf(outStr,"%3d",0);
      else 
	sprintf(outStr,"%.1f",radialValue[thisI]);
      
      doLabel(outStr, radialValueW, FALSE);
      
      if(foundIt) { // Then we can change a color
	lastRed  = color_defs[thisI].red;
	lastGreen= color_defs[thisI].green;
	lastBlue = color_defs[thisI].blue;
	color_defs[thisI].red    = 65000;
	color_defs[thisI].green  = 0;
	color_defs[thisI].blue   = 0;
	XAllocColor (dsply, color_map, &color_defs[thisI]);
	pixels_return[thisI] = color_defs[thisI].pixel;
	lastI = thisI;
      } 
    }
    
    lastX = x;  
    break;
    
  default: 
    break;
  }
  
  XFlush(dsply); // Make changes current

} // end function circAvgMotion



void circAvgArrowActivate(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: circAvgArrowActivate
  // change the current circavg displayed by using the arrows.

  int *passTag;
  passTag = (int *)malloc(sizeof(int));

  if(widget == circAvgUpW)
    circAvgSize++;
  else if(widget == circAvgDownW)
    circAvgSize--;
  
  circAvgSize = MAX(1,MIN(circAvgSize, 4));
  TextSetInt(circAvgTextW, circAvgSize, FALSE);
  
  *passTag= 0;
  DrawCircAvgWindow(0, (XtPointer)passTag, 0);
  free(passTag);

} // end function circAvgArrowActivate



void circAvg_change(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: circAvg_change
  // change the offset in the X direction for the circAvg screen.
  
  int i, j, *passTag;
  float *thisRint;
  int thisNterp;
  XmScrollBarCallbackStruct *temp = (XmScrollBarCallbackStruct *) call_data;

  passTag = (int *)malloc(sizeof(int));

  if(widget == circAvgToggleW) { // toggle invert contrast
    caInvertContrast = !caInvertContrast;
    
    for (j=0; j<numCircAvgDraw; j++) {
      thisRint  = circAvg[drawCircs[j].index].rint;
      thisNterp = circAvg[drawCircs[j].index].nterp;
      
      for(i=caOffsetX; i<thisNterp; i++)
	thisRint[i] *= -1.0;

    } 
  } else { // Do scrollbar
    caOffsetX = (int)temp->value;
  }
  
  // Recalc the Y scaling factors based on this new offset. We may be
  // able to show more of the curves out here if we've got a huge
  // noise spike at the center of the map e.g.

  minCA= R4_MAX;
  maxCA= R4_MIN;
  for (j=0; j<currentCircAvg; j++) {
    thisRint  = circAvg[j].rint;
    thisNterp = circAvg[j].nterp;
    for (i=caOffsetX; i<(thisNterp); i++) {
      if(thisRint[i] != 0.0) { // OK to use
	minCA = MIN(minCA,thisRint[i]);
	maxCA = MAX(maxCA,thisRint[i]);
      }
    }
  }
  
  if(maxCA==minCA)
    minCA = maxCA - 0.01;
  
  scaleCircFactor = 245.0 / (maxCA-minCA); 
  
  *passTag= 0;
  DrawCircAvgWindow(0, (XtPointer)passTag, 0);
  free(passTag);

} // end function circAvg_change



void circAvgText_changed(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: circAvgText_changed
  // Change the circAvg size

  circAvgSize = (int)wstrtod(circAvgTextW);
  circAvgSize = MAX(1, MIN(circAvgSize, 4));

} // end function circAvgText_changed



void findMaxMin(float *fftImg_buffer, float *maxF, float *minF, 
		float *offset, int fftWidth, int fftHeight)
{
  // function name: findMaxMin
  // Generic routine for finding the min/max in the given fft buffer.

  int i, j, jindx;

  *maxF = -HUGE;
  *minF =  HUGE;
  
  for (j=0; j<fftHeight; j++) {
    for (i=0; i<fftWidth*2; i+= 2) {
      jindx   = j*fftWidth*2 + i;
      *maxF = MAX(*maxF, fftImg_buffer[jindx]);
      *minF = MIN(*minF, fftImg_buffer[jindx]);
    }
  }
  
  *offset = *minF;

} // end function findMaxMin



void calcNewLat(int numScroll)
{
  // function name: calcNewLat
  // Calculate new lattice ax,ay,bx,by based on the values they've
  // chosen from the scrollbars for astar,bstar,gamstr, and omega(OLD
  // gammaLat). See latref_calc() in latref.for for code from which
  // this was taken.
  
  if(!layerLines) { // 2D lattice
    if(numScroll==2) { // gammaStr == latScroll
      gamstr = latScrollValue[2];
      map_abang = 180.0 - gamstr;
      // Don't forget to add omega also ...
      anga = angaSave + omega;
      angb = latScrollValue[2] + anga;
      
      byL  = bstar * sin(deg2rad(angb));
      bxL  = bstar * cos(deg2rad(angb));
      ayL  = astar * sin(deg2rad(anga));
      axL  = astar * cos(deg2rad(anga));
      // omega= 180.0 - gamstr;
    }
    else if(numScroll==0) { // astar
      astar = latScrollValue[0];
      if(bstar != 0.0 && astar != 0.0)
	aoverb = 1.0 / (astar / bstar);
      
      ayL = astar * sin(deg2rad(anga));
      axL = astar * cos(deg2rad(anga));
      // adim = idim1*raster/(astar*sind(gamstr));
    }
    else if(numScroll==1) { // bstar
      bstar = latScrollValue[1];
      if(bstar != 0.0)
	aoverb = 1.0 / (astar / bstar);
      
      byL = bstar * sin(deg2rad(angb));
      bxL = bstar * cos(deg2rad(angb));
      // bdim = idim1*raster/(bstar*sind(gamstr));
    }
    else if(numScroll==3) { // omega
      omega = latScrollValue[3];
      
      // Shouldn't we do this from the anga and angb ??
      // what if the gamstr has changed ???
      // Don't forget gamstr 
      anga = angaSave + omega;
      angb = latScrollValue[2] + anga;
      
      ayL  = astar * sin(deg2rad(anga));
      axL  = astar * cos(deg2rad(anga));
      byL  = bstar * sin(deg2rad(angb));
      bxL  = bstar * cos(deg2rad(angb));
    }
  }
  else { // layerlines
    if(numScroll==3)
      omega = latScrollValue[3];
    else if(numScroll==1)
      dstar = latScrollValue[1];
  }

} // end function calcNewLat



void read_fftbuffer_activate(Widget widget, XtPointer client_data, 
			     XtPointer call_data)
{
  // function name: read_fftbuffer_activate
  // Read an fft buffer from disk. Used for debuggin other programs
  // mainly.

  int saveBackTransform;
  FILE *fIn;
  
  if ( (fIn = fopen("fft_buffer.dat", "r")) != NULL ) {
    fread(fftImage_gold,sizeof(float),((fftSize*2)*fftSize),fIn);
    
    applyCTF = FALSE;
    saveBackTransform = backTransform;
    backTransform = TRUE;  // Let's back transform it
    
    calcFFT(FALSE, 0, FALSE);
    
    XSync(dsply,FALSE);
    XFlush(dsply);
    DrawFFTWindow(NULL, NULL, NULL);
    raiseIt(mainWindowW); 
    raiseIt(fftScreenW);
    
    // Now stuff it back the way it used to be
    backTransform = saveBackTransform;

    fclose(fIn);
  } else {
    sprintf(outStr,
	    "Can't find fft_buffer.dat in directory you started robem from.");
    askThem(outStr, XmDIALOG_INFORMATION);
  }
  
} // end function read_fftbuffer_activate



void fft_activate(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: fft_activate
  // Put the fftscreen on the display.
  
  int i;
    
  if(purdueHeader.numCols <= 0 || purdueHeader.numRows <= 0)
    return;

  averageFFT = FALSE; // Don't want average anymore
  
  raiseIt(fftScreenW);
  
  XtVaSetValues(a128FFTW,  XmNsensitive, TRUE, NULL); 
  XtVaSetValues(a256FFTW,  XmNsensitive, TRUE, NULL);
  XtVaSetValues(a512FFTW,  XmNsensitive, TRUE, NULL);
  XtVaSetValues(a1024FFTW, XmNsensitive, TRUE, NULL);
  
  // Start as small as we can get
  fftSize = 128;
  if(rectBox.drawable) { // Look at their rectangle
    askThem("I'm using the rectangle box you defined\n\
earlier - NOT the entire image.",
	    XmDIALOG_INFORMATION);
    if(rectBox.width>128 || rectBox.height>128) { 
      // Can't do 128x128
      XtVaSetValues(a128FFTW, XmNsensitive, FALSE, NULL);
      fftSize = 256;
    }
    if(rectBox.width>256 || rectBox.height>256) { 
      // Can't do 256 x256
      XtVaSetValues(a256FFTW, arglst, XmNsensitive, FALSE, NULL);
      fftSize = 512;
    }
  }
  else { // Use whole image
    if(purdueHeader.numCols>128 || purdueHeader.numRows>128) {
      // Can't do 128 x 128
      XtVaSetValues(a128FFTW, XmNsensitive, FALSE, NULL); 
      fftSize = 256;
    }
    if(purdueHeader.numCols>256 || purdueHeader.numRows>256) {
      // Can't do 256 x 256
      XtVaSetValues(a256FFTW, XmNsensitive, FALSE, NULL);
      fftSize = 512;
    }
  }
  
  // Let's change the lopHip scrolls accordingly
  XtSetArg(arglst[0], XmNmaximum, (int)(fftSize*1.222));
  XtSetArg(arglst[1], XmNminimum, 0);
  XtSetArg(arglst[2], XmNpageIncrement, (int)(fftSize/32));
  XtSetValues(lopScrollW, arglst, 3);
  XtSetValues(hipScrollW, arglst, 3);
  
  // This may have been reset by rotflt e.g.
  idim1= idim2= 1024;   // MAX fft size
  CK(fftImage_gold= (float *)malloc(sizeof(float)*idim1*idim2*2));
  CK(fftImage_fourAver= (float *)malloc(sizeof(float)*idim1*idim2*2));
  CK(fftImage_buffer2= (float *)malloc(sizeof(float)*idim1*idim2*2));
  
  fftMode= TRUE;
  
  if(measureFFT) { // Unmanage these from the FFT screen  
    lopHipSwitch = FALSE;
    drawCTF = FALSE;
    measureFFT = TRUE;
    pickLattice = -99;  // Make them draw it all over again
    backTransform = FALSE;
    currentLatPt=0;
    // Turn interface off
    XtSetArg(arglst[0], XmNset, FALSE);
    XtSetValues(backFFTToggleW, arglst, 1);  
    
    XtUnmanageChild(lophipFormW);
    XtUnmanageChild(ctfContFormW);
    XtUnmanageChild(ctfCurveFormW);
    XtUnmanageChild(ctfScrollsFormW);
    XtUnmanageChild(ctfParamsW);
    
    XtUnmanageChild(latPointsFormW);
    XtUnmanageChild(latTogglesFormW);
    XtUnmanageChild(latResetButW);
    XtUnmanageChild(latResetLabelW);
    
    XtManageChild(fftMeasureFormW);
    XtManageChild(fftFeedbackFormW);
    
    for (i=0; i<4; i++) {
      XtUnmanageChild(latticeScrollW[i]);
      XtUnmanageChild(latticeLabel1W[i]);
      XtUnmanageChild(latticeLabel2W[i]);
    }
    
    // Give feedback on FFT window title bar
    doDialogTitle("FFT - Measure FFT", fftScreenW, FALSE); 
    
  }// Unmanage these from the FFT screen 
  
  // Force the colormap to normal grayscale 
  for (i=7; i<tryColors; i++) { // Do the colors we have been assigned
    color_defs[i] = gen_color_defs[i]; // struct assignment
    
    XAllocColor (dsply, color_map, &color_defs[i]);
    pixels_return[i] = color_defs[i].pixel;
  }// Do the colors we have been assigned 

  if(fftSize == 128 | fftSize == 256 || fftSize == 512 || fftSize == 1024) {
    XtSetArg(arglst[0], XmNmaximum, (int)(fftSize*1.222));
    XtSetArg(arglst[1], XmNminimum, 0);
    XtSetArg(arglst[2], XmNpageIncrement, (int)(fftSize/32));
    XtSetValues(lopScrollW, arglst, 3);
    XtSetValues(hipScrollW, arglst, 3);
  }

  XtVaSetValues(a128FFTW,   XmNset, FALSE, NULL); 
  XtVaSetValues(a256FFTW,   XmNset, FALSE, NULL); 
  XtVaSetValues(a512FFTW,   XmNset, FALSE, NULL); 
  XtVaSetValues(a1024FFTW,  XmNset, FALSE, NULL); 

  if(fftSize==128)
    XtVaSetValues(a128FFTW,  XmNset, TRUE, NULL); 
  else if(fftSize==256)
    XtVaSetValues(a256FFTW,  XmNset, TRUE, NULL); 
  else if(fftSize==512)
    XtVaSetValues(a512FFTW,  XmNset, TRUE, NULL); 
  else if(fftSize==1024)
    XtVaSetValues(a1024FFTW, XmNset, TRUE, NULL); 

  calcFFT(TRUE, 0, 0);
  DrawFFTWindow(NULL, NULL, NULL);
  
} // end function fft_activate



void ctf_reset_activate(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: ctf_reset_activate
  // Initialize CTF parameters based on microscope choice
  
  int spCoExp;
  float afractThisValue, pfractThisValue ,
    avoltageThisValue, 
    spaberrationThisValue, chaberrationThisValue, 
    spCoherenceThisValue, spCoFactor ,
    tmCoherenceThisValue, majDefocusThisValue, 
    minDefocusThisValue, majAngThisValue, 
    wienerThisValue, defocusThisValue;
  char inStr[256];
  int tag = *((int *) client_data);

  // Values common to all microscopes
  majDefocusThisValue = 1.0;
  minDefocusThisValue = 1.0;
  majAngThisValue = 0.0;
  wienerThisValue = 0.0;
  defocusThisValue = 1.000;

  if(tag == ALL) { // normal
    afractThisValue = 0.0;
    avoltageThisValue = 100000.0;
    avoltageThisValue = 200000.0;  // Default to 200 kV for FEG
    spaberrationThisValue = 2.0;
    chaberrationThisValue = 2.0;
    spCoherenceThisValue = 5.0;
    spCoExp = -3;
    spCoFactor = 0.001;
    tmCoherenceThisValue = 0.0;
  }
  else if(tag == FEG) { // CM200 feg
    afractThisValue = 0.05;
    avoltageThisValue = 200000.0;
    spaberrationThisValue = 2.0;
    chaberrationThisValue = 2.0;
    spCoherenceThisValue = 5.0;
    spCoExp = -2;
    spCoFactor = 0.01;
    tmCoherenceThisValue = 0.5;
  }
  else if(tag == FEG300) { // feg 300
    afractThisValue = 0.04;
    avoltageThisValue = 300000.0;
    spaberrationThisValue = 2.0;
    chaberrationThisValue = 2.0;
    spCoherenceThisValue = 5.0;
    spCoExp = -3;
    spCoFactor = 0.001;
    tmCoherenceThisValue = 0.5;
  }
  else if(tag == EM300) { // nermag
    afractThisValue = 0.07;
    avoltageThisValue = 100000.0;
    spaberrationThisValue = 1.6;
    chaberrationThisValue = 1.3;
    spCoherenceThisValue = 5.0;
    spCoExp = -3;
    spCoFactor = 0.001;
    tmCoherenceThisValue = 1.6;
  }
  else if(tag == EM420W) { // em420
    afractThisValue = 0.07;
    avoltageThisValue = 120000.0;
    spaberrationThisValue = 2.0;
    chaberrationThisValue = 2.0;
    spCoherenceThisValue = 1.0;
    spCoFactor = 1.0;
    spCoExp = 0;
    tmCoherenceThisValue = 1.6;
  }
  else if(tag == EM420L) { // em420 LaB6
    afractThisValue = 0.07;
    avoltageThisValue = 120000.0;
    spaberrationThisValue = 2.0;
    chaberrationThisValue = 2.0;
    spCoherenceThisValue = 0.5;
    spCoExp = 0;
    spCoFactor = 1.0;
    tmCoherenceThisValue = 0.8;
  }
  else if(tag == JEOL_EM_2010F_HT_FEG) { // JEOL_EM-2010F-HT_FEG
    afractThisValue = 0.05;
    avoltageThisValue = 200000.0;
    spaberrationThisValue = 1.4;
    chaberrationThisValue = 1.8;
    spCoherenceThisValue = 5.0;
    spCoExp = -3;
    spCoFactor = 0.001;
    tmCoherenceThisValue = 0.5;
  }
  else if(tag == JEOL_EM_2010F_UHR_FEG) { // JEOL_EM-2010F-UHR_FEG
    afractThisValue = 0.05;
    avoltageThisValue = 200000.0;
    spaberrationThisValue = 0.5;
    chaberrationThisValue = 1.1;
    spCoherenceThisValue = 5.0;
    spCoExp = -3;
    spCoFactor = 0.001;
    tmCoherenceThisValue = 0.5;
  }
  else if(tag == JEOL1010) { // JEOL1010
    afractThisValue = 0.05;
    avoltageThisValue = 100000.0;
    spaberrationThisValue = 1.9;
    chaberrationThisValue = 1.9;
    spCoherenceThisValue = 5.0;
    spCoExp = -3;
    spCoFactor = 0.001;
    tmCoherenceThisValue = 0.5;
  }
  else if(tag == Hitachi_H9000) { // Hitachi_H9000
    afractThisValue = 0.05;
    avoltageThisValue = 120000.0;
    spaberrationThisValue = 0.9;
    chaberrationThisValue = 2.0;
    spCoherenceThisValue = 0.5;
    spCoExp = -3;
    spCoFactor = 0.001;
    tmCoherenceThisValue = 0.5;
  }
  else if(tag == Hitachi_HF2000) { // Hitachi_HF2000
    afractThisValue = 0.05;
    avoltageThisValue = 200000.0;
    spaberrationThisValue = 1.2;
    chaberrationThisValue = 1.5;
    spCoherenceThisValue = 0.2;
    spCoExp = -3;
    spCoFactor = 0.001;
    tmCoherenceThisValue = 0.3;
  }
  else if(tag == FEI_Tecnai_G2_Polara) { // FEI_Tecnai_G2_Polara
    afractThisValue = 0.04;
    avoltageThisValue = 300000.0;
    spaberrationThisValue = 2.3;
    chaberrationThisValue = 2.0;
    spCoherenceThisValue = 0.1;
    spCoExp = -3;
    spCoFactor = 0.001;
    tmCoherenceThisValue = 0.7;
  }
  else if(tag == FEI_Tecnai_G2_Sphera) { // FEI_Tecnai_G2_Sphera
    afractThisValue = 0.05;
    avoltageThisValue = 200000.0;
    spaberrationThisValue = 2.26;
    chaberrationThisValue = 2.0;
    spCoherenceThisValue = 1.0;
    spCoExp = -3;
    spCoFactor = 0.001;
    tmCoherenceThisValue = 1.0;
  }
  
  pfractThisValue = 1.0 - afractThisValue;
  
  sprintf(inStr,"%.2f",afractThisValue);
  TextSet(ctfAmpFactorW, inStr, FALSE); 
  sprintf(inStr,"%.0f",avoltageThisValue*0.001);  // Units = Kv
  TextSet(ctfKvW, inStr, FALSE); 
  sprintf(inStr,"%.1f",spaberrationThisValue);
  TextSet(ctfSpAberW, inStr, FALSE); 
  sprintf(inStr,"%.1f",wienerThisValue);
  TextSet(ctfWienerW, inStr, FALSE); 

} // end function ctf_reset_activate



void fft_mode_activate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: fft_mode_activate
  // Select FFT mode

  int i, saveBackTransform;
  int ctfMode, view, units, fft_full, filter, mode, thisTag;
  float volts, pixelSize, ampContrast_in, defocus_in;
  float defocus_minor_in,input_phi,sphericalAbbr;
  float angMajor, tFactor, wiener, d_star, pi;
  float filt_fac[10], cc, beta, dE, radian, degree;
  
  int tag = *((int *) client_data);

  thisTag = tag;
  if(tag == APPLYINVERSE1CTF) {
    ctfFilter = 1;
    thisTag = APPLYINVERSECTF;
  }
  else if(tag == APPLYINVERSE2CTF){
    ctfFilter = 2;
    thisTag = APPLYINVERSECTF;
  }
  else if(tag == APPLYINVERSE3CTF) {
    ctfFilter = 3;
    thisTag = APPLYINVERSECTF;
  }
  else if(tag == APPLYINVERSE4CTF) {
    ctfFilter = 4;
    thisTag = APPLYINVERSECTF;
  }
  else if(tag == APPLYINVERSE5CTF) {
    ctfFilter = 5;
    thisTag = APPLYINVERSECTF;
  }

  if(thisTag==APPLYCTF || thisTag==APPLYINVERSECTF || 
     thisTag==APPLYINVERSECTF2 || thisTag==APPLYINVERSECTF3 || 
     thisTag==BACKTRANS) {
    // Apply CTF to current FFT and back transform

    ctfMode = 1;
    view    = 1;
    
    volts   = wstrtod(ctfKvW)  * 1000.0; // Convert from kv to volts
    
    pixelSize   = wstrtod(ctfPixSizeW);
    
    // These should be in microns ???
    ctfMajorDefocusValue   = wstrtod(ctfMajorDefocusW);
    ctfMinorDefocusValue   = wstrtod(ctfMinorDefocusW);
    defocus_in        = ctfMajorDefocusValue;
    defocus_minor_in  = ctfMinorDefocusValue;
    
    input_phi   =  0.0;
    
    sphericalAbbr   = wstrtod(ctfSpAberW);
    wavelength = wavel_(&volts);
    ampContrast_in = wstrtod(ctfAmpFactorW);
    
    if(pixelSize != 0.0) {

      units = 2;  // using nanometers for pixelSize
      fft_full = 1;
      d_star = 0.025;
      pi = M_PI;
      radian = 0.0;
      piraddeg_(&pi, &radian, &degree);
      
      mode = (int)wstrtod(ctfModeW);
      wiener = wstrtod(ctfWienerW);
      if(wiener==0.0)
	wiener= 0.001;
      TextSetFloat(ctfWienerW, wiener, TRUE);
      angMajor = wstrtod(ctfAngMajorW);

      // We need to change the sign of angMajor here because our
      // XWindows based buffer is upside down.  When we write to disk
      // we need to use the sign just as given from the interface

      angMajor = 0.0 - angMajor ;
      tFactor = wstrtod(ctfTFactorW);

      // Init these for now - get from interface eventually
      cc   = 0.0;
      dE   = 0.0;
      beta = 0.0;
      
      filt_fac[0] = wiener;
      filt_fac[1] = 0.1;

      if(thisTag == APPLYCTF) {
	ctf_multiply__(&mode, fftImage_gold, &fftSize, &fftSize, 
		      &units, &pixelSize, &pi, &radian,
		      &wavelength, &defocus_in, &defocus_minor_in,
		      &angMajor, &ampContrast_in, &sphericalAbbr, 
		      &cc, &beta, &dE, &volts, &tFactor, &fft_full);
      }
      else if(thisTag == APPLYINVERSECTF) {
	filter = ctfFilter;
	ctf_divide__(&mode, &filter, fftImage_gold, &fftSize, &fftSize, 
		    &units, &pixelSize, &pi, &radian, &wavelength,
		    &defocus_in, &defocus_minor_in, &angMajor,
		    &ampContrast_in, &sphericalAbbr, filt_fac,
		    &tFactor, &fft_full, &cc,&beta,&dE,&volts);
      }
      
      applyCTF = FALSE;
      saveBackTransform = backTransform;
      backTransform = TRUE;  // Let's back transform it
      
      calcFFT(FALSE, 0, FALSE);
      
      XSync(dsply,FALSE);
      XFlush(dsply);
      DrawFFTWindow(NULL, NULL, NULL);
      XFlush(dsply);
      
      // Now stuff it back they way it used to be
      backTransform = saveBackTransform;
      
    } else {
      askThem("You've got to enter a valid PixelSize.", XmDIALOG_INFORMATION);
    }
    
  }
  else if(thisTag==LOPHIP) { // Manage all of the LOP/HIP stuff
    
    lopHipSwitch = TRUE;
    drawCTF = FALSE;
    pickLattice = -99;  // Make them draw it all over again
    // We will force them to pick lattice points here
    pickedLatPts= FALSE;
    backTransform = FALSE;
    XtVaSetValues(drawHiLoToggleW, XmNsensitive, TRUE, NULL);
    
    currentLatPt=0;
    // Turn interface off
    XtSetArg(arglst[0], XmNset, FALSE);
    XtSetValues(backFFTToggleW, arglst, 1);  
    
    XtManageChild(lophipFormW);
    XtUnmanageChild(ctfContFormW);
    XtUnmanageChild(ctfCurveFormW);
    XtUnmanageChild(ctfScrollsFormW);
    XtUnmanageChild(ctfParamsW);
    
    XtUnmanageChild(latPointsFormW);
    XtUnmanageChild(latTogglesFormW);
    XtUnmanageChild(latResetButW);
    XtUnmanageChild(latResetLabelW);
    
    XtUnmanageChild(fftMeasureFormW);
    XtUnmanageChild(fftFeedbackFormW);
    
    for (i=0; i<4; i++) {
      XtUnmanageChild(latticeScrollW[i]);
      XtUnmanageChild(latticeLabel1W[i]);
      XtUnmanageChild(latticeLabel2W[i]);
    }
    
    measureFFT = FALSE;
    
    // Give feedback on FFT window title bar
    doDialogTitle("FFT - LOP/HIP FIltering", fftScreenW, FALSE); 
  }
  else if(thisTag == CTF) { // Manage all of the CTF stuff
    lopHipSwitch = FALSE;
    drawCTF = TRUE;

    if(getToggleState(pickCTFtoggleW))
      pickLattice = 1;
    else
      pickLattice = -99;
    
    // We will force them to pick latticw points here
    pickedLatPts= FALSE;
    backTransform = FALSE;
    currentLatPt=0;
    // Turn interface off
    XtVaSetValues(backFFTToggleW, XmNset, FALSE, NULL);
    
    XtManageChild(ctfContFormW);
    XtManageChild(ctfCurveFormW);
    XtManageChild(ctfScrollsFormW);
    XtManageChild(ctfParamsW);
    
    XtUnmanageChild(lophipFormW);
    XtUnmanageChild(latPointsFormW);
    XtUnmanageChild(latTogglesFormW);
    XtUnmanageChild(latResetButW);
    XtUnmanageChild(latResetLabelW);
    
    XtUnmanageChild(fftMeasureFormW);
    XtUnmanageChild(fftFeedbackFormW);
    
    for (i=0; i<4; i++) {
      XtUnmanageChild(latticeScrollW[i]);
      XtUnmanageChild(latticeLabel1W[i]);
      XtUnmanageChild(latticeLabel2W[i]);
    }
    
    measureFFT = FALSE;
    
    // Give feedback on FFT window title bar
    doDialogTitle("FFT - CTF Estimation", fftScreenW, FALSE); 
  }
  else if(thisTag == LATREFINE) { // Manage lattice refinement stuff

    // Force it into 512 size mode
    if(fftSize != 512) {
      fftSize = 512;
      RwaActivateWidget(a512FFTW); 
      calcFFT(TRUE, 0, 0);
      DrawFFTWindow(NULL, NULL, NULL);
      askThem("Lattice Refinement and Filtering only in 512x512 FFTs for now.",
	      XmDIALOG_INFORMATION);
    }
    
    lopHipSwitch = FALSE;
    drawCTF = FALSE;
    pickLattice = 1;  // Make them draw it all over again
    backTransform = FALSE;
    currentLatPt=0;
    // Turn interface off
    XtVaSetValues(backFFTToggleW, XmNset, FALSE, NULL);
    
    XtUnmanageChild(lophipFormW);
    XtUnmanageChild(ctfContFormW);
    XtUnmanageChild(ctfCurveFormW);
    XtUnmanageChild(ctfScrollsFormW);
    XtUnmanageChild(ctfParamsW);
    
    XtManageChild(latPointsFormW);
    XtManageChild(latTogglesFormW);
    XtManageChild(latResetButW);
    XtManageChild(latResetLabelW);
       
    XtUnmanageChild(fftMeasureFormW);
    XtUnmanageChild(fftFeedbackFormW);
    
    for(i=0; i<4; i++) {
      XtManageChild(latticeScrollW[i]);
      XtManageChild(latticeLabel1W[i]);
      XtManageChild(latticeLabel2W[i]);
    }
    
    measureFFT = FALSE;
    
    // Give feedback on FFT window title bar
    doDialogTitle("FFT - Lattice Refinement", fftScreenW, FALSE); 
  }
  else if(thisTag == FFTMEASURE) {// UnManage all

    lopHipSwitch = FALSE;
    drawCTF = FALSE;
    measureFFT = TRUE;
    pickLattice = -99;  // Make them draw it all over again
    backTransform = FALSE;
    currentLatPt=0;
    // Turn interface off
    XtVaSetValues(backFFTToggleW, XmNset, FALSE, NULL);
    
    XtUnmanageChild(lophipFormW);
    XtUnmanageChild(ctfContFormW);
    XtUnmanageChild(ctfCurveFormW);
    XtUnmanageChild(ctfScrollsFormW);
    XtUnmanageChild(ctfParamsW);
    
    XtUnmanageChild(latPointsFormW);
    XtUnmanageChild(latTogglesFormW);
    XtUnmanageChild(latResetButW);
    XtUnmanageChild(latResetLabelW);
    
    XtManageChild(fftMeasureFormW);
    XtManageChild(fftFeedbackFormW);
    
    doLabel("[] Pixels.", fftMeasPixW, DEFAULT);

    doLabel("[] Angstroms.", fftMeasAngW, DEFAULT);
    
    for (i=0; i<4; i++) {
      XtUnmanageChild(latticeScrollW[i]);
      XtUnmanageChild(latticeLabel1W[i]);
      XtUnmanageChild(latticeLabel2W[i]);
    }
    
    // Give feedback on FFT window title bar
    doDialogTitle("FFT - Measure FFT", fftScreenW, FALSE); 
  }
  else {
    fft_deactivate(NULL, NULL, NULL);
  }
	
  // Make them do a new circavg
  calcFFT1DcircAvg= FALSE;
 
} // end function fft_mode_activate



void ctf_appodize_activate(Widget widget, XtPointer client_data, 
			   XtPointer call_data)
{
  // function name: ctf_appodize_activate
  // Get appodizing info for writing this from FFT.

  int tag = *((int *) client_data);

  appodizing = WRITEFFT;
  writeFFTType = tag;
  // This is border width now
  TextSetInt(appodBoxRadW, featherRadius-boxRadius, FALSE);
  TextSetInt(appodFeaRadW, featherRadius, FALSE);
  raiseIt(appodizeScreenW);

} // end function ctf_appodize_activate



void fft_1DCA_activate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: fft_1DCA_activate
  // Do a 1d circular average

  int i, j, scaledY, scaledYp1, thisX, thisXp1;
  int interp, nterp, ncol, nrow;
  int *irnum;
  float xc, yc, minCA, maxCA, scaleCircFactor;
  float tmpBuffer[512*512], thisBkg, *rint;
  FILE  *fOut;

  ncol = 512;
  nrow = 512;
  xc = ncol/2.0;
  yc = nrow/2.0;

  interp= 2;  // Original way
  interp= 1;  // Check circAvgMotion() to make sure they agree
  nterp = interp*(int)((int)((MAX(ncol,nrow)))/(int)2) + 1;
  
  CK(rint= (float*)malloc(sizeof(float)*nterp*3));
  CK(irnum= (int*)malloc(sizeof(int)*nterp*3));

  // Stuff buffer
  for (i=0; i<512*512; i++)
    tmpBuffer[i]= (float)fft_gold_buffer1[i];
  
  thisBkg = 0.0;
  img_1d_rob(tmpBuffer, ncol, nrow, xc, yc,
	     interp, nterp, rint, irnum, thisBkg);
  
  free(irnum);
  
  // Recalc the Y scaling factors based on this new offset. We may be
  // able to show more of the curves out here if we've got a huge
  // noise spike at the center of the map e.g

  minCA= R4_MAX;
  maxCA= R4_MIN;

  // Open up out file
  fOut = fopen("fft1dca.txt","w");
  
  rint[0]= 0.0;  // Hard code for now
  for (i=0; i<256; i++) {
    if(rint[i]!=0.0) {
      minCA = MIN(minCA,rint[i]);
      maxCA = MAX(maxCA,rint[i]);
    }
    if(fOut && i > 0) { // Print it
      float numAngs;
      numAngs = ((512.0 / (float)i)*fftAngsPerPixel);
      if(fftAngsPerPixel != 0.0)
	fprintf(fOut,"%f,%f\n",numAngs,rint[i]);
      else fprintf(fOut,"%d,%f\n",i,rint[i]);
    }
  }
  
  if(fOut) {
    fclose(fOut);
    // askThem("1D CircAvg written to sys$login:fft1dca.txt file.",
    // XmDIALOG_INFORMATION);
  }
  else {
    askThem("Cannot open sys$login:fft1dca.txt file.", XmDIALOG_INFORMATION);
  }
  
  if(disp1Dca) {// Display the 1D circavg on the fft window 
    int xSq,ySq;
    float dist;
    
    scaleCircFactor = (float)(tryColors-10.0) / (maxCA-minCA); 
    
    // Init
    for (i=0; i<512*512; i++)
      fft_buffer1[i]= 8;
    
    for (i=0; i<512; i++) {
      ySq = (i-256)*(i-256);
      for(j=0;j<512;j++) { // Circ avg the buffer
	xSq = (j-256)*(j-256);
	dist = sqrt((double)(xSq+ySq));
	if(dist <= 256.0)
	  fft_buffer1[(i*512)+j] = (rint[(int)dist]-minCA)*scaleCircFactor;
      }
    }
    
    if(XtIsManaged(fftScreenW))
      XPutImage(dsply, fft_window1, fft_gc, fft_image1, 0, 0,
		0, 0,  512, 512);
    XFlush (dsply);
    
  }
  else {// Just put back the fft buffer
    DrawFFTWindow(NULL, NULL, NULL);
  }
  
  scaleCircFactor = 505.0 / (maxCA-minCA); 
  
  for (i=256; i<512; i++) {
    scaledY = (rint[(int)((float)(i-256)/(float)fftZoomFactor)]-
	       minCA)*scaleCircFactor;
    scaledYp1 = (rint[(int)((float)((i-256)+1)/(float)fftZoomFactor)]-
		 minCA)*scaleCircFactor;
    scaledY   = 510.0 - scaledY;
    scaledYp1 = 510.0 - scaledYp1;
    thisX= i;
    thisXp1= i+1;
    if(i==256)
      scaledY = 256;
    XDrawLine(dsply, XtWindow(fftDrawingW), image_gcGREEN,
	      thisX, scaledY, thisXp1, scaledYp1);
  }
  
  free(rint);
  
  calcFFT1DcircAvg= TRUE;
  
} // end function fft_1DCA_activate



void calcNodes()
{
  // function name: calcNodes
  // Calculate CTF nodes

  int i;
  float distA, distB, pixelSize, ampFactor;

  ctfMajorDefocusValue = ctffocusMajValue ;
  ctfMinorDefocusValue = ctfFocusMinvalue ;
  
  pixelSize = wstrtod(ctfPixSizeW);
  ampFactor = wstrtod(ctfAmpFactorW);

  // Get all distances for all nodes
  for (i=0; i<MAXCTFS; i++) {
    ctf_nodes_misha(wavelength, ctfMajorDefocusValue, spAberration,
		    ampFactor, i+1, &distA);
    ctf_nodes_misha(wavelength, ctfMinorDefocusValue, spAberration,
		    ampFactor, i+1, &distB);
    
    distA *= 512.0*pixelSize;
    distB *= 512.0*pixelSize;
    ctfPts[i].distA   = distA;
    ctfPts[i].distAsq = distA * distA;
    ctfPts[i].distB   = distB;
    ctfPts[i].distBsq = distB * distB;
    ctfPts[i].omega   = ctfAngMajorValue;
  }
  
} // end function calcNodes



void calcTFpixel(float focusMaj, float focusMin)
{
  // function name: calcTFpixel
  // Calculate the transform pixel length.
  
  float a, c, alpha, spatialFreq, distance;

  a = (spAberration * 1000000.0) / 2.0;
  c = wavelength / M_PI;

  alpha = sqrt((focusMaj + sqrt(focusMaj*focusMaj - 4.0*a*c)) / (2.0*a));
  spatialFreq = alpha / wavelength;
  distance = (512.0 * (scanStep * 1000.0) * spatialFreq) / magnification;
  ctfPts[ctfCurveNum].distA = distance;
  
  alpha = sqrt((focusMin + sqrt(focusMin*focusMin - 4.0*a*c)) / (2.0*a));
  spatialFreq = alpha / wavelength;
  distance = (512.0 * (scanStep * 1000.0) * spatialFreq) / magnification;
  ctfPts[ctfCurveNum].distB = distance;
  
} // end function calcTFpixel



void calcDefocus(float *majorValue, float *minorValue)
{
  // function name: calcDefocus
  // Calculate the major and minor defocus values

  int nodes;
  float distA, distB, focus, spatialFreq, pixelSize, ampFactor, volts;

  distA   = ctfPts[ctfCurveNum].distA;
  distB   = ctfPts[ctfCurveNum].distB;

  pixelSize = wstrtod(ctfPixSizeW);
  ampFactor = wstrtod(ctfAmpFactorW);
  volts     = wstrtod(ctfKvW) * 1000.0;

  wavelength = wavel_(&volts);

  nodes = ctfCurveNum + 1;
  
  // Only do if points have been picked
  if(pickLattice != -99 && distA != 0 && distB != 0) { 
    spatialFreq =  (distA / 512.0) / pixelSize;
    ctf_focus_misha(wavelength, &focus, spAberration, 
		     ampFactor, nodes, spatialFreq);
    ctfMajorDefocusValue = focus;
    
    spatialFreq =  (distB / 512.0) / pixelSize;
    ctf_focus_misha(wavelength, &focus, spAberration, 
		     ampFactor, nodes, spatialFreq);
    ctfMinorDefocusValue = focus;
  }
  
  // Now set up the text widgets
  *majorValue = ctfMajorDefocusValue;
  *minorValue = ctfMinorDefocusValue;
  TextSetFloat(ctfMajorDefocusW, ctfMajorDefocusValue, FALSE);
  TextSetFloat(ctfMinorDefocusW, ctfMinorDefocusValue, FALSE);
  
  calcNodes();
  
} // end function calcDefocus



void redo_ctfFocMajMin()
{
  // function name: redo_ctfFocMajMin
  
  calcDefocus(&ctffocusMajValue, &ctfFocusMinvalue);
  
  TextSetFloat(ctfFminTW,        ctfFocusMinvalue, FALSE);
  TextSetFloat(ctfMinorDefocusW, ctfFocusMinvalue, FALSE);
  sprintf(outStr,"%.2f",rad2deg(ctfAngMajorValue));  
  TextSet(ctfAmajTW, outStr, FALSE);
  TextSet(ctfAngMajorW, outStr, FALSE); 
  TextSetFloat(ctfFmajTW,        ctffocusMajValue, FALSE);
  TextSetFloat(ctfMajorDefocusW, ctffocusMajValue, FALSE);
  calcNodes();
  calcFFT(FALSE, 0, FALSE);
  DrawFFTWindow(NULL, NULL, NULL);
  XSync(dsply,FALSE);
  
} // end function redo_ctfFocMajMin



void calcCTF()
{
  // function name: calcCTF
  // Calculate CTF curve

  int i, j, thisJ, thisRowSq, thisColSq;
  float res_step, defocus, alpha, chi_alpha, thisCs, tmp;
  float pfract, afract, dist, wave, thisDefocus;

  res_step = 1.0 / 32.0;
  defocus  = ctffocusMajValue;
  pfract   = 1.0;
  afract   = 0.0;

  // From ctfDisp 
  // wavelength = 1230.0/sqrt(voltage + (1.0e-6)*(voltage*voltage));
  // wave = wavelength/1000.0;		// convert to nm units

  // Convert ALL to nm units
  wave = wavelength/1000.0;		 
  thisDefocus = defocus * 1000.0;
  thisCs = spAberration * 1000000.0;
  
  for (i=0; i<512; i++) {
    thisRowSq = (i-256)*(i-256);
    for (j=0; j<1024; j+=2) {
      //Calc CTF factor but remember to do it from distance from the middle
      thisJ = j;
      thisColSq = (thisJ/2 - 256)*(thisJ/2 - 256);
      dist = sqrt((float)(thisRowSq+thisColSq));
      
      alpha     = dist*wave*res_step;
      chi_alpha = (THISTWOPI/wave) *
	((thisDefocus*alpha*alpha/2.0) - 
	 (thisCs*(alpha*alpha*alpha*alpha)/4.0));
      tmp       = pfract*sin(chi_alpha) + afract*cos(chi_alpha);
      
      // Orig 
      ctfValues[i][j] = tmp;
      ctfValues[i][j+1] = dist;
      
      // Why this???
      // ctfValues[i][j] = 1.0 - fabs(tmp);
      // ctfValues[i][j+1] = 1.0 - fabs(tmp);    
    }
  }
  
  printf("leaving calcCTF\n");

} // end function calcCTF



void drawSCURVEonFFT()
{
  // function name: drawSCURVEonFFT
  // Draw CTF curve on FFT screen

  int i;
  float res_step, defocus, thisCs, tmp;
  float pfract, afract, dist, wave, thisDefocus;
  float lastVal;
  float slope, minPixel, maxPixel, scaleF;
    
  // Draw background
  for (i=0; i<256; i++)
    XDrawLine(dsply, XtWindow(fftDrawingW), image_gcGREY[100],
	      CNVPTX(256+i),256, CNVPTX(256+i),512);      
  
  if(scurveClick.x != 256)
    dist = sqrt( (256.0 - scurveClick.y) * (256.0 - scurveClick.y) +
		 (256.0 - scurveClick.x) * (256.0 - scurveClick.x) );
  else
    dist = abs(((float)scurveClick.x - 256.0));
  
  // Do we really want this ???  dist *=  (float)((float)fftSize / 512.0);

  if(scurveClick.x != 256)
    slope = (256.0 - scurveClick.y) / (scurveClick.x - 256.0);
  else
    slope = 0.0;

  if(slope==0.0)
    slope = 0.00000001;
  
  res_step = 1.0 / 32.0;
  defocus  = ctffocusMajValue;
  pfract   = 1.0;
  afract   = 0.0;
  
  // From ctfDisp 
  // wavelength = 1230.0/sqrt(voltage + (1.0e-6)*(voltage*voltage));
  // wave = wavelength/1000.0;		/+! convert to nm units +/

  // Convert ALL to nm units
  wave = wavelength/1000.0;		 
  thisDefocus = defocus * 1000.0;
  thisCs   = spAberration * 1000000.0;
  
  // Get stats for data
  farray_minmax(auto2DdataBuf, &minPixel, &maxPixel, 512*256);
  scaleF = 255.0 / (maxPixel-minPixel);
  
  // First draw all data on screen
  for (i=0; i<(int)dist; i++) {
    int x, y;
    float xStep, spatialI;
    
    spatialI = i * (fftSize / 512.0);
    xStep = sqrt( 1.0/(slope*slope + 1.0) );
    x = spatialI*xStep;
    y = spatialI*xStep*slope;
    // printf("i[%d] xStep[%f] x[%d] y[%d]\n",i,xStep,x,y);
    
    // Now draw scurve
    if(y > 255 || y<-254)
      tmp = 511;
    else
      tmp = 511 - (int)((auto2DdataBuf[(256-y)*256 + x] - minPixel)*scaleF);
    
    XDrawLine(dsply, XtWindow(fftDrawingW), image_gcBLUE,
	      CNVPTX(256+i), 511, CNVPTX(256+i), (int)tmp);
  }
  
  // Get stats for curve
  farray_minmax(auto2DdrawBuf, &minPixel, &maxPixel, 512*256);
  scaleF = 255.0 / (maxPixel-minPixel);
  
  lastVal = 128.0 + 256.0;
  for (i=0; i<(int)dist; i++) {
    int x,y;
    float xStep, spatialI;
    
    spatialI = i * ((float)fftSize / 512.0);
    xStep = sqrt( 1.0/(slope*slope + 1.0) );
    x = spatialI*xStep;
    y = spatialI*xStep*slope;
    
    // Now draw scurve
    if(y > 255 || y < -254)
      tmp = 511;
    else
      tmp = 511 - (int)((auto2DdrawBuf[(256-y)*256 + x] - minPixel)*scaleF);
    
    XDrawLine(dsply, XtWindow(fftDrawingW), image_gcGREEN,
	      CNVPTX(256+i), (int)lastVal, CNVPTX(256+i+1), (int)tmp);
    
    lastVal = tmp;
  }  

} // end function drawSCURVEonFFT



void ctf_ioparms_activate(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: ctf_ioparms_activate
  // write CTF params to a disk file for sharing with the ctfDisp program

  FILE *fIO;
  char inStr[256], fileStr[512];

  if(widget == ctfCurveOutBW) { // OUTPUT
    sprintf(fileStr,"%s/ctf_ioparms.dat",getenv("HOME"));
    if ( (fIO = fopen(fileStr,"w")) != NULL ) {
      fprintf(fIO,"%s\n", XmTextGetString(ctfAmpFactorW)); 
      fprintf(fIO,"%s\n", XmTextGetString(ctfSpAberW)); 
      fprintf(fIO,"%s\n", XmTextGetString(ctfMajorDefocusW)); 
      fprintf(fIO,"%s\n", XmTextGetString(ctfMinorDefocusW)); 
      fprintf(fIO,"%s\n", XmTextGetString(ctfKvW)); 
      fprintf(fIO,"%s\n", XmTextGetString(ctfWienerW)); 
      fprintf(fIO,"%s\n", XmTextGetString(ctfAngMajorW)); 
      fclose(fIO);
    }
    else {
      askThem("Can't open ctf_ioparms.dat file.", XmDIALOG_INFORMATION);
    }
  }
  else if (widget == ctfCurveInBW) { // INPUT
    sprintf(fileStr,"%s/ctf_ioparms.dat",getenv("HOME"));
    if ( (fIO = fopen(fileStr,"r")) != NULL ) {
      get_str(fIO, '\n', inStr);
      TextSet(ctfAmpFactorW, inStr, FALSE); 
      get_str(fIO, '\n', inStr);
      TextSet(ctfSpAberW, inStr, FALSE); 
      get_str(fIO, '\n', inStr);
      TextSet(ctfMajorDefocusW, inStr, FALSE); 
      get_str(fIO, '\n', inStr);
      TextSet(ctfMinorDefocusW, inStr, FALSE); 
      get_str(fIO, '\n', inStr);
      TextSet(ctfKvW, inStr, FALSE); 
      get_str(fIO, '\n', inStr);
      TextSet(ctfWienerW, inStr, FALSE); 
      get_str(fIO, '\n', inStr);
      TextSet(ctfAngMajorW, inStr, FALSE); 
      fclose(fIO);
    }
    else {
      askThem("Can't open ctf_ioparms.dat file.", XmDIALOG_INFORMATION);
    }
  }

} // end function ctf_ioparms_activate



void ctfCurve_cr_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{
  // function name: ctfCurve_cr_activate
  // Called when Return key hit while in ctfCurve text widget
  
  ctfCurveNum = (int)wstrtod(ctfCurveTextW);
  ctfCurveNum--;     // Human numbers from 1

  if(ctfCurveNum >= MAXCTFS)
    ctfCurveNum = MAXCTFS-1;

  ctfCurveNum = MAX(ctfCurveNum, 0);

  maxCtfCurve = MAX((ctfCurveNum+1),maxCtfCurve);

  // Restuff the old points
  if(!ctfPts[ctfCurveNum].pickedPts) { // Guess some new ones
    ctfPts[ctfCurveNum].latX0 = latPt[0].measuredX;
    ctfPts[ctfCurveNum].latY0 = latPt[0].measuredY;
    ctfPts[ctfCurveNum].latX1 = latPt[1].measuredX;
    ctfPts[ctfCurveNum].latY1 = latPt[1].measuredY;
    // ctfPts[ctfCurveNum].distA = ctfPts[0].distA;
    // ctfPts[ctfCurveNum].distB = ctfPts[0].distB;
  } else { // restuff
    latPt[0].measuredX = ctfPts[ctfCurveNum].latX0;
    latPt[0].measuredY = ctfPts[ctfCurveNum].latY0;
    latPt[1].measuredX = ctfPts[ctfCurveNum].latX1;
    latPt[1].measuredY = ctfPts[ctfCurveNum].latY1;
  }

  ctfPts[ctfCurveNum].pickedPts = TRUE;
  TextSetInt(ctfCurveTextW, ctfCurveNum+1, FALSE);
  DrawFFTWindow(NULL, NULL, NULL);
  
} // end function ctfCurve_cr_activate



void fftImgNum_cr_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{
  // function name: fftImgNum_cr_activate
  // Called when Return key hit while in fftImgNum text widget.
  // Work on the image they've entered

  averageFFT= FALSE;

  // Could be fftImgNumW or rotfltImgNumW
  imgNum = wstrtod(widget);

  if(imgNum > purdueHeader.nimg)
    imgNum = purdueHeader.nimg;
  imgNum = MAX(imgNum, 1);

  TextSetInt(imgNumTextW,   imgNum, FALSE);
  TextSetInt(fftImgNumW,    imgNum, FALSE);
  TextSetInt(rotfltImgNumW, imgNum, FALSE);
  if(fileType==PACKEDPORTABLE)
    stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		     packImg,purdueHeader.nimg, imgNum);
  DrawWindow(); 

  calcFFT(TRUE, 0, 0);
  DrawFFTWindow(NULL, NULL, NULL);
  
  // Make them do a new circavg
  calcFFT1DcircAvg= FALSE;

  // DO if rotflt'ing
  newOrigin = TRUE;
  if(XtIsManaged(rotfltScreenW)) {
    iopt[0]= 1;      // iopt[0] == 1 --> Refine symmetry axis
    iopt[1]= 0;
    iopt[2]= 0;      // iopt[2] == 1 --> plot power spectrum
    iopt[3]= 0;  
    rotflt_tryit_activate(0,0,0);
  }
  
} // end function fftImgNum_cr_activate



void fft_change(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: fft_change
  // Limit the display of the FFT from min to max.  Also other FFT
  // parameters are changed here.

  int i, j;
  int tag = *((int *) client_data);
  XmScrollBarCallbackStruct *temp = (XmScrollBarCallbackStruct *) call_data;

  if(!inited) return;

  if(tag == FFTMIN || tag == FFTMAX) {// MIN/MAX scrolls
    if(tag == FFTMIN) {
      if(fftType == LINEARFFT) {
	minFFTvalue = temp->value*1000.0; // 1/1000th
	minFFTvalue = MIN(minFFTvalue,(maxFFTvalue-1000.0));
      }
      else {
	minFFTvalue = temp->value/100.0; // 1/100th
	minFFTvalue = MIN(minFFTvalue,(maxFFTvalue-0.01));
      }
    }
    else if(tag == FFTMAX) {
      if(fftType == LINEARFFT) {
	maxFFTvalue = temp->value*1000.0; // 1/1000th 
	maxFFTvalue = MAX(maxFFTvalue,(minFFTvalue+1000.0));
      }
      else {
	maxFFTvalue = temp->value/100.0; // 1/100th 
	maxFFTvalue = MAX(maxFFTvalue,(minFFTvalue+0.01));
      }
    }
    
    if(minFFTvalue >= maxFFTvalue)
      maxFFTvalue = minFFTvalue + 1.0;
    
    sprintf(outStr,"%.2f",minFFTvalue);  
    doLabel(outStr, minFFTW, DEFAULT);
    
    sprintf(outStr,"%.2f",maxFFTvalue);    
    doLabel(outStr, maxFFTW, DEFAULT);
  }
  else if(tag == FFTTFAC) { // New Temp factor
    tempFactor = (int)temp->value;
    
    sprintf(outStr,"%d",tempFactor);  
    doLabel(outStr, tfacFFTW, DEFAULT);
    
    // Calc all temp factors
    if(tempFactor == 0) { // No temp factor
      for(i=0; i<1024; i++)
	for(j=0; j<2048; j++)
	  tempFactors[i][j] = 1.0;
    }
    else { // Calc it
      float B,dist,numTemp,floatVal;
      int thisRowSq,thisColSq,thisJ;
      
      floatVal = (float)temp->value;
      
      numTemp = floatVal*0.1;
      if(numTemp > 10.0)
	numTemp= 10.0;
      
      numTemp = 4.0/(pow(2.0,numTemp));
      B = 4.0 / (numTemp * numTemp);
      for (i=0; i<512; i++) {
	thisRowSq = (i < 256) ? i*i : (512-i)*(512-i);
	for (j=0; j<1024; j+=2) { // Remember to use distance from middle
	  thisJ = j;
	  thisJ = (j - 512); // Try this 
	  if(j > 512)
	    thisJ -= 512;
	  else 
	    thisJ += 512;
	  thisColSq = (thisJ/2)*(thisJ/2);
	  dist = (float)(thisRowSq+thisColSq);
	  if(dist > 0.0)
	    tempFactors[i][j] = exp(-B / (4.0 * (65536.0 / dist)*
					  (65536.0 / dist)));
	  else 
	    tempFactors[i][j] = 1.0;
	  if(tempFactors[i][j] == 0.0)
	    tempFactors[i][j]= 0.00000001;
	  tempFactors[i][j] = 1.0 / tempFactors[i][j];
	  tempFactors[i][j+1] = tempFactors[i][j];
	}
      }
    } 
  }
  else if(tag == CTF_ANGMAJOR) { // CTF omega scroll
    ctfAngMajorValue = temp->value;
    ctfAngMajorValue *= 0.01;
    ctfAngMajorValue = deg2rad(ctfAngMajorValue);  
    changeCTFtext = TRUE;
    sprintf(outStr,"%.2f",rad2deg(ctfAngMajorValue));  // +1 for humans
    TextSet(ctfAmajTW, outStr, FALSE);
    TextSet(ctfAngMajorW, outStr, FALSE); 
  }
  else if(tag == CTF_FOCUSMAJ) { // CTF mean Defocus scroll
    ctffocusMajValue = temp->value;
    ctffocusMajValue *= 0.01;
    // OBSOLETE???  calcTFpixel(ctffocusMajValue, ctfFocusMinvalue);
    calcNodes();
    // calcMeanTFpixel(ctffocusMajValue);
    changeCTFtext = TRUE;
    TextSetFloat(ctfFmajTW,        ctffocusMajValue, FALSE);
    TextSetFloat(ctfMajorDefocusW, ctffocusMajValue, FALSE);
  }
  else if(tag == CTF_FOCUSMIN) { // CTF a/b scroll
    ctfFocusMinvalue = temp->value;
    ctfFocusMinvalue *= 0.01;
    // OBSOLETE???  calcTFpixel(ctffocusMajValue, ctfFocusMinvalue);
    calcNodes();
    changeCTFtext = TRUE;
    TextSetFloat(ctfFminTW,        ctfFocusMinvalue, FALSE);
    TextSetFloat(ctfMinorDefocusW, ctfFocusMinvalue, FALSE);
  }
  else if(tag == FILTSIZE) { // FILTER Size scroll
    filtSizeValue = temp->value;
    filtSizeValue *= 0.5;
    sprintf(outStr,"%.0f",filtSizeValue*2.0);  // +1 for humans
    doLabel(outStr, filtLabelW, DEFAULT);
  }
  else if (tag == LATCOLOR) { // lat color
    int thisVal;
    thisVal = (int)temp->value;
    image_gcLAT = image_gcGREY[thisVal];
  }
  else if (tag == HIP) { // hip
    int thisVal;
    thisVal = (int)temp->value;
    TextSetInt(hipTextW, thisVal, FALSE);
  }
  else if (tag == LOP) { // lop
    int thisVal;
    thisVal = (int)temp->value;
    TextSetInt(lopTextW, thisVal, FALSE);
  }
  
  if(temp->event != 0) {
    calcFFT(FALSE, 0, FALSE);
    
    XSync(dsply,FALSE);
    XFlush(dsply);
    DrawFFTWindow(NULL, NULL, NULL);
    XSync(dsply,FALSE);
    XFlush(dsply);
  }
 
} // end function fft_change



void fft_change_toggle(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: fft_change_toggle
  // Change the FFT size

  int i, j;
  XmString simpleString;
  int tag = *((int *) client_data);
  XmPushButtonCallbackStruct *temp = (XmPushButtonCallbackStruct *) call_data;
  
  if(!inited || temp->event == 0) return;
  
  if(tag == A1024FFT) { 
    rotflt_deactivate(NULL, NULL, NULL);
    fftSize = 1024;
  }
  else if(tag == A512FFT) { 
    rotflt_deactivate(NULL, NULL, NULL);
    fftSize = 512;
  }
  else if(tag == A256FFT) {
    rotflt_deactivate(NULL, NULL, NULL);
    fftSize = 256;
  }
  else if(tag == A128FFT) {
    rotflt_deactivate(NULL, NULL, NULL);
    fftSize = 128;
  }
  else if(tag == AMPLITUDE) { 
    if(fftDispType == AMPLITUDE) {
      fftDispType = INTENSITY;
      simpleString = XmStringCreateLocalized("Intensity");
    } else {
      fftDispType = AMPLITUDE;
      simpleString = XmStringCreateLocalized("Amplitude");
    }
    
    // XtSetArg(arglst[0], XmNset, FALSE);
    XtVaSetValues(amplitudeToggleW, XmNlabelString, simpleString, NULL);
    XFlush(dsply);
    XmStringFree(simpleString);
  }
  else if (tag == FFTZOOM) {
    if(getToggleState(fftZoomToggleW)) { // Turn the others off
      fftZoomFactor = 2;
      fftZoomX = 128;  // Upper left hand corner
      fftZoomY = 128;
      fftZoomX = MIN(MAX(fftZoomX,0),(512-128));
      fftZoomY = MIN(MAX(fftZoomY,0),(512-128));
    } else {
      fftZoomFactor = 1;
      fftZoomX= fftZoomY= 0;
    }
  }
  else if(tag == DISP1024FFT) { 
    disp1024fft = !disp1024fft; 
    printf("disp1024[%d]\n",disp1024fft);
  }
  else if(tag == FILTER) {
    filterToggle = !filterToggle;
    if(filterToggle) {
      // XtVaSetValues(filtSizeW, XmNsensitive, TRUE, NULL);
      XtVaSetValues(drawHiLoToggleW, XmNsensitive, TRUE, NULL);
      XtVaSetValues(pick2LW, XmNsensitive, FALSE, NULL);
      // Turn off pick lattice mode
      pickLattice = -99;
    } else {
      // XtVaSetValues(filtSizeW, XmNsensitive, FALSE, NULL);
      XtVaSetValues(drawHiLoToggleW, XmNsensitive, FALSE, NULL);
      XtVaSetValues(pick2LW, XmNsensitive, TRUE, NULL);
      // Turn on pick lattice mode
      pickLattice = 1;
      currentLatPt=0; 
      contour= TRUE;
    }
  }
  else if(tag == PICKCTF) {
    if(tag == PICKCTF) { // Turn it WAY off for pick ctf points
      if(pickLattice != -99) {
	pickLattice = -99;  
	drawCTF = FALSE; 
      } else {
	pickLattice = 1;
	drawCTF = TRUE;
      }
    } else {
      pickLattice = 1;
    }
    
    contour= TRUE;
    //	  pickedLatPts= FALSE; They may have some from before
    currentLatPt=0; 
    stuffContourBuffer();
    // Turn off Filtering 
    filterToggle = FALSE;
    // XtVaSetValues(filtSizeW, XmNsensitive, FALSE, NULL);
    XtVaSetValues(drawHiLoToggleW, XmNsensitive, FALSE, NULL);
  }
  else if(tag == LAYERLINES || tag == TWODLINES) {
    layerLines = (tag == LAYERLINES) ? TRUE : FALSE;

    filterToggle = FALSE;
    // XtVaSetValues(filtSizeW, XmNsensitive, FALSE, NULL);
    XtVaSetValues(drawHiLoToggleW, XmNsensitive, FALSE, NULL);

    // Turn it on
    pickLattice = 1;
    contour= TRUE;
    //	  pickedLatPts= FALSE; They may have some from before
    currentLatPt=0; 
    stuffContourBuffer();
    
    XtVaSetValues(filterFFTtoggleW, XmNset, FALSE, NULL); 
    
    if(layerLines) {
      // Reinit these
      TextSet(lat1XtextW, "", FALSE);
      TextSet(lat1YtextW, "", FALSE);
      TextSet(lat2XtextW, "", FALSE);
      TextSet(lat2YtextW, "", FALSE);
      
      // No resolution filter with layer lines yet
      resoX= 0;
      resoY= 0;
      lowResoFilter = 320000;  // Ridiculously high
      hiResoFilter  = 0;       // Ridiculously low
      
      sprintf(outStr,"Draw Layer Lines"); 
      XtSetArg(arglst[0], XmNlabelString, XmStringCreateLocalized(outStr));
      XtSetArg(arglst[1], XmNset, TRUE);
      XtSetValues(drawLatToggleW, arglst, 2); 
      // We have to force the scrolls to be inited 
      drawLatToggle = TRUE;
      XFlush(dsply);
      sprintf(outStr,"Filter Mask Slit Diameter"); 
      simpleString = XmStringCreateLocalized(outStr);
      XtVaSetValues(filtDescW, XmNlabelString, simpleString, NULL);
      XFlush(dsply);
      XmStringFree(simpleString); 
      XtUnmanageChild(lat1YtextW);
      XtUnmanageChild(lat2XtextW);
      XtUnmanageChild(lat2YtextW);
      XtUnmanageChild(lat2LabelW);
      XtUnmanageChild(millerLabelW);
      doLabel("LayrLine#", lat1LabelW, DEFAULT);

      if(latLineOrder <= 1)
	latLineOrder = 1;

      TextSetInt(lat1XtextW, latLineOrder, TRUE);
      XtVaSetValues(latticeScrollW[0], XmNsensitive, FALSE, NULL); 
      XtVaSetValues(latticeScrollW[2], XmNsensitive, FALSE, NULL); 
      TextSet(latticeLabel2W[0], "n/a", FALSE); 
      TextSet(latticeLabel2W[2], "n/a", FALSE); 
      
      // Why do we need to do this ???
      // latPt[2].x = latPt[0].x;
      // latPt[2].y = latPt[0].y;
      // latPt[2].measuredX = latPt[0].measuredX;
      // latPt[2].measuredY = latPt[0].measuredY;
      // latPt[3].x = latPt[1].x;
      // latPt[3].y = latPt[1].y;
      // latPt[3].measuredX = latPt[1].measuredX;
      // latPt[3].measuredY = latPt[1].measuredY;
      
      latReset_activate(NULL, NULL, NULL);
    } else {
      // Reinit these
      TextSet(lat1XtextW, "", FALSE);
      TextSet(lat1YtextW, "", FALSE);
      TextSet(lat2XtextW, "", FALSE);
      TextSet(lat2YtextW, "", FALSE);
      
      doLabel("Draw 2D Lattice", drawLatToggleW, DEFAULT);
      doLabel("Filter Mask Hole Diameter:", filtDescW, DEFAULT);

      XtManageChild(lat1YtextW);
      XtManageChild(lat2YtextW);
      XtManageChild(lat2XtextW);
      XtManageChild(lat2LabelW);
      XtManageChild(millerLabelW);

      doLabel("h1, k1", lat1LabelW, DEFAULT);

      XtVaSetValues(latticeScrollW[0], XmNsensitive, TRUE, NULL);
      XtVaSetValues(latticeScrollW[2], XmNsensitive, TRUE, NULL);
      XFlush(dsply);
      // Why do we ned to do this ???
      // latPt[0].x = latPt[2].x;
      // latPt[0].y = latPt[2].y;
      // latPt[0].measuredX = latPt[2].measuredX;
      // latPt[0].measuredY = latPt[2].measuredY;
      // latPt[1].x = latPt[3].x;
      // latPt[1].y = latPt[3].y;
      // latPt[1].measuredX = latPt[3].measuredX;
      // latPt[1].measuredY = latPt[3].measuredY;

      latReset_activate(NULL, NULL, NULL);
    }
  }
  else if(tag == FFTCONTOUR)
    contourToggle = !contourToggle;
  else if (tag == DISP1DCA)
    disp1Dca = !disp1Dca;
  else if (tag == CTFENHANCE)
    ctfEnhance = !ctfEnhance;
  else if(tag == FFTLOCK)
    fftMinMaxLock = !fftMinMaxLock; 
  else if(tag == LINEARFFT)
    fftType = LINEARFFT; 
  else if(tag == LOGFFT)
    fftType = LOGFFT; 
  else if(tag == NONLINFFT)
    fftType = NONLINFFT;
  else if(tag == BACKFFT)
    backTransform = !backTransform; 
  else if(tag == DRAWLAT)
    drawLatToggle = !drawLatToggle;
  else if(tag == DRAWHILO)
    drawHiLoToggle = !drawHiLoToggle;
  else if (tag == SOFT)
    icut = 1;
  else if (tag == HARD)
    icut = 0;

  if(tag == LINEARFFT || tag == LOGFFT) { // LOG or LINEAR fft
    float thisMin, thisMax;
    int thisMaxInt, thisMinInt, fftHeight,fftWidth, thisPixel, sliderSize;
    
    fftHeight = fftWidth = 512;
    thisMin = R4_MAX;
    thisMax = R4_MIN;
    for (i=0; i<fftHeight; i++) {
      for (j=0; j<fftWidth*2; j+=2) {
	float thisValue;
	thisPixel = i*fftWidth*2 + j;
	thisValue = (fftImage_gold[thisPixel]   * fftImage_gold[thisPixel] +
		     fftImage_gold[thisPixel+1] * fftImage_gold[thisPixel+1] );
	thisValue *= tempFactors[(int)(i*0.5)][(int)(j*0.5)];
	if(fftType == LOGFFT && thisValue > 0.0)
	  thisValue = log10(thisValue);
	
	thisMax = MAX(thisValue,thisMax);
	thisMin = MIN(thisValue,thisMin);
      }
    }
    
    if(averageFFT)  // Do special
      TextSet(fftImgNumW, "Aver", FALSE);
    
    // Now setup the scroll bars to hold the new max/min
    // We can't exceed what a 32 bit signed integer will hold!! 
    // - Really 2.17 Gig but let's be safe... 
    // Suppose thisMin was 65 for linear but 1.8 for log ???
    
    minFFTvalue = thisMin;
    maxFFTvalue = thisMax;
    
    if(fftType == LINEARFFT)
      thisMax *= 0.001;  // Let's get rid of a lot of the top stuff
    
    thisMax *= (fftType == LINEARFFT) ? 0.001 : 100.0;
    thisMin *= (fftType == LINEARFFT) ? 0.001 : 100.0;
    
    thisMaxInt = (int)MIN(thisMax, 1800000000.0);
    imageMaxFFTvalue = (float)thisMax;
    thisMinInt = (int)thisMin;
    
    if(thisMaxInt <= thisMinInt)
      thisMaxInt = thisMinInt + 1;
    
    sliderSize = (thisMaxInt - thisMinInt) / 100;
    sliderSize = MAX(sliderSize,1);
    
    XtSetArg(arglst[0], XmNmaximum, (int)(thisMaxInt+(float)sliderSize));
    XtSetArg(arglst[1], XmNminimum, thisMinInt);
    XtSetArg(arglst[2], XmNsliderSize, sliderSize);

    if(!fftMinMaxLock) { // OK to change fft min max value
      if(fftType == LINEARFFT)// 1/1000th 
	XtSetArg(arglst[3], XmNvalue, (int)(minFFTvalue/1000.0));
      else 
	XtSetArg(arglst[3], XmNvalue, (int)(minFFTvalue*100.0));
      XtSetArg(arglst[4], XmNpageIncrement, (int)(sliderSize));
      if(sliderSize > 11.0)
	XtSetArg(arglst[5], XmNincrement, (int)(sliderSize/10.0));
      else 
	XtSetArg(arglst[5], XmNincrement, 1.0);
      XtSetValues(minFFTscrollW, arglst, 6);
    } else {
      XtSetArg(arglst[3], XmNpageIncrement, sliderSize);
      if(sliderSize > 11.0)
	XtSetArg(arglst[4], XmNincrement, (int)(sliderSize/10.0));
      else 
	XtSetArg(arglst[4], XmNincrement, 1.0);
      XtSetValues(minFFTscrollW, arglst, 5);
    }

    XtSetArg(arglst[0], XmNmaximum, (int)(thisMaxInt+(float)sliderSize));
    XtSetArg(arglst[1], XmNminimum, thisMinInt);
    XtSetArg(arglst[2], XmNsliderSize, sliderSize);

    if(!fftMinMaxLock) { // OK to change fft min max value
      if(fftType == LINEARFFT) // 1/1000th
	XtSetArg(arglst[3], XmNvalue, (int)(maxFFTvalue/1000.0));
      else 
	XtSetArg(arglst[3], XmNvalue, (int)(maxFFTvalue*100.0));
      XtSetArg(arglst[4], XmNpageIncrement, (int)(sliderSize));
      if(sliderSize > 11.0)
	XtSetArg(arglst[5], XmNincrement, (int)(sliderSize/10.0));
      else 
	XtSetArg(arglst[5], XmNincrement, 1.0);
      XtSetValues(maxFFTscrollW, arglst, 6);
    } else {
      XtSetArg(arglst[3], XmNpageIncrement, sliderSize);
      if(sliderSize > 11.0)
	XtSetArg(arglst[4], XmNincrement, (int)(sliderSize/10.0));
      else 
	XtSetArg(arglst[4], XmNincrement, 1.0);
      XtSetValues(maxFFTscrollW, arglst, 5);
    }
    
    XSync(dsply,FALSE);
    XFlush(dsply);
    
    sprintf(outStr,"%.2f",minFFTvalue);  
    doLabel(outStr, minFFTW, DEFAULT);
    
    sprintf(outStr,"%.2f",maxFFTvalue);    
    simpleString = XmStringCreateLocalized(outStr);
    XtVaSetValues(maxFFTW, XmNlabelString, simpleString, NULL); 
    XSync(dsply,FALSE);
    XFlush(dsply);
    XmStringFree(simpleString); 
  }
  
  // Change the lopHip scrolls accordingly
  if(tag == A1024FFT || tag == A512FFT ||
     tag == A256FFT  || tag == A128FFT ) { 
    XtSetArg(arglst[0], XmNmaximum, (int)(fftSize*1.222));
    XtSetArg(arglst[1], XmNminimum, 0);
    XtSetArg(arglst[2], XmNpageIncrement, (int)(fftSize/32));
    XtSetValues(lopScrollW, arglst, 3);
    XtSetValues(hipScrollW, arglst, 3);
  } 
  
  // Recalc FFT if tag is not set to FFTLOCK
  if(tag != FFTLOCK) {
    if(tag == A1024FFT || tag == A512FFT  || tag == A256FFT  ||
       tag == A128FFT  || tag == RECALCFFT) {
      averageFFT = FALSE;
      calcFFT(TRUE, 0, 0); 
    } else {
      calcFFT(FALSE, 0, 0); 
    }
    XFlush(dsply);
    DrawFFTWindow(NULL, NULL, NULL);
    XFlush(dsply);
  }
  
  // Make them do a new circavg
  calcFFT1DcircAvg= FALSE;
  
} // end function fft_change_toggle



void latScrollChanged(Widget widget, int *tag, 
		      XmScrollBarCallbackStruct *callback_data)
{
  // function name: latScrollChanged
  // Call here when lat scroll value on screen changes

  int i, thisOne;
  float floatVal;

  thisOne= -99;
  for (i=0; i<12; i++)
    if(widget==latticeScrollW[i])
      thisOne= i;
  
  if(thisOne >=0) {
    latValChanged = TRUE;
    
    floatVal = (float)callback_data->value;
    if(thisOne == 3) // omega
      latScrollValue[thisOne] = floatVal / 10.0;
    else 
      latScrollValue[thisOne] = floatVal / 100.0;
    sprintf(outStr,"%.2f",latScrollValue[thisOne]);
    
    TextSet(latticeLabel2W[thisOne], outStr, FALSE); 
    
    // Reasons this is called are drag and value_changed - 
    // see /usr/include/Xm/Xm.h for defn's
    calcNewLat(thisOne);
    
    calcFFT(FALSE, 0, 0);
    DrawFFTWindow(NULL, NULL, NULL); 
  }
  
} // end function latScrollChanged



void latTextChanged(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{
  // function name: latTextChanged
  // Call here when lat text value on screen changes
  
  int i, thisOne;
  float floatVal;
  
  thisOne= -99;
  for(i=0;i<12;i++)
    if(widget==latticeLabel2W[i])
      thisOne= i;
  
  if(thisOne >=0) { 
    latValChanged = TRUE;
    
    floatVal = wstrtod(latticeLabel2W[thisOne]);
    
    if(thisOne==3) // Omega // RSS This looks strange
      latScrollValue[thisOne] = floatVal;
    else 
      latScrollValue[thisOne] = floatVal;
    
    // Don't change you'll get an infinite loop !!
    // XtSetArg(arglst[0], XmNvalue, (int)(floatVal*100.0));
    // XtSetValues(latticeScrollW[thisOne], arglst, 1);

    // Reasons this is called is value_changed - 
    // see /usr/include/Xm/Xm.h for defn's

    calcNewLat(thisOne);

    calcFFT(FALSE, 0, 0);
    DrawFFTWindow(NULL, NULL, NULL);
  }

} // end function latTextChanged



void hipTextChanged(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{
  // function name: hipTextChanged
  // When the HIP text widgets on the screen change value call here.
  // Only called when the CR is pressed in the text widget.

  scanStep = (int)wstrtod(widget);
  XtVaSetValues(hipScrollW, XmNvalue, (int)scanStep, NULL);  
  
  calcFFT(FALSE, 0, 0);
  DrawFFTWindow(NULL, NULL, NULL);
      
} // end function hipTextChanged



void lopTextChanged(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{
  // function name: lopTextChanged
  // When the LOP text widgets on the screen change value call here.
  // Only called when the CR is pressed in the text widget.

  scanStep = (int)wstrtod(widget);
  XtVaSetValues(lopScrollW, XmNvalue, (int)scanStep, NULL);  
  
  calcFFT(FALSE, 0, 0);
  DrawFFTWindow(NULL, NULL, NULL);
  
} // end function lopTextChanged



void lohiRadiusChanged(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{
  // function name: lohiRadiusChanged
  // When the LOP or HIP radius text widgets on the screen change
  // value call here. Only called when the CR is pressed in the text
  // widget.

  scanStep = (int)wstrtod(lopRadiusTextW);
  lowResoFilter = scanStep * scanStep;
  
  scanStep = (int)wstrtod(hipRadiusTextW);
  hiResoFilter = scanStep * scanStep;
  
  sprintf(outStr,"%d",(int)sqrt((float)lowResoFilter));
  TextSet(lopRadiusTextW, outStr, FALSE);
  sprintf(outStr,"%d",(int)sqrt((float)hiResoFilter));
  TextSet(hipRadiusTextW, outStr, FALSE);

  calcFFT(FALSE, 0, 0);
  DrawFFTWindow(NULL, NULL, NULL);
      
} // end function lohiRadiusChanged



void ctfTextChanged(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{
  // function name: ctfTextChanged 
  // When the CTF text widgets on the screen change value call here.
  // Only called when the CR is pressed in the text widget.

  if(widget == ctfScanStepW) { // Get scan step
    scanStep = wstrtod(widget);
    if(scanStep > 0.0) {
      pixSize = (scanStep*1000.0)/magnification;
      fftAngsPerPixel = pixSize*10.0;
      sprintf(outStr,"%.4f",pixSize);
      TextSet(ctfPixSizeW, outStr, FALSE);
      sprintf(outStr,"%.3f",fftAngsPerPixel);
      TextSet(fftMeasurePSTextW, outStr, TRUE);
      redo_ctfFocMajMin();   
    }
  }
  else if(widget == ctfPixSizeW) {// Get pixsize
    pixSize = wstrtod(widget);
    fftAngsPerPixel = pixSize*10.0;
    if(pixSize > 0.0) {
      sprintf(outStr,"%.3f",fftAngsPerPixel);
      TextSet(fftMeasurePSTextW, outStr, FALSE);
      magnification = (scanStep*1000.0)/pixSize;
      sprintf(outStr,"%.0f",magnification);
      TextSet(ctfMagW, outStr, TRUE);	
      redo_ctfFocMajMin();
    }
  }
  else if(widget == ctfMagW) { // Get mag
    magnification = wstrtod(widget);
    if(magnification > 0.0) {
      pixSize = (scanStep*1000.0)/magnification;
      fftAngsPerPixel = pixSize*10.0;
      sprintf(outStr,"%.4f",pixSize);
      TextSet(ctfPixSizeW, outStr, FALSE);	
      sprintf(outStr,"%.3f",fftAngsPerPixel);
      TextSet(fftMeasurePSTextW, outStr, TRUE);
      redo_ctfFocMajMin();
    }
  }

} // end function ctfTextChanged



void drawCross(int drawX,int drawY,int drawXfin,int drawYfin,
	       GC colorGC)
{
  // function name: drawCross
  // Draw Cross to indicate x,y point selected.
  
  if(pickLattice != -99) {
    XDrawLine(dsply, XtWindow(fftDrawingW), colorGC,
	      drawX,drawY, drawXfin,drawYfin);
  }
  
} // end function drawCross



void fft_flicker_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{
  // function name: fft_flicker_activate
  // Flicker the drawing of the CTF with the raw FFT.

  int saveDrawCtf;

  saveDrawCtf = drawCTF;
  if(*tag==1)
    drawCTF = FALSE;
  
  DrawFFTWindow(NULL, NULL, NULL);
  drawCTF = saveDrawCtf;

} // end function fft_flicker_activate



void DrawFFTWindow(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: DrawFFTWindow
  // Redraw window on expose events or whenever it's prudent.
  
  int drawX, drawY, drawXfin, drawYfin;
  int nx, ny, ik, ix, iy, ih, ix1, ix2, iy1, iy2;
  int i, j, thisY, thisPixel;
  int thisX, is1D, is2D;
  float xcen, ycen;
  float x1, x2, y1, y2, thisNode;
  float scaleCircFactor, maxCA, minCA;

#define DRAWBLANK(x) ((doingAutoDefocusRefine && adrZoomDraw) ? ((x-256) * 2) : (x))
#define DRAWBLANKY(y) ((doingAutoDefocusRefine && adrZoomDraw) ? (256-((256-y) * 2)) : (y))

  if(!drawLite)
    return;

  // Stuff the display buffer based on how zoomed we are
  if(fftZoomFactor == 1) { // Normal
    for (i=0; i<512; i++) {
      for (j=0; j<512; j++) { // Zoom it normal
	thisPixel = i*512 + j;
	XPutPixel(fft_image1, j, i,
		  pixels_return[fft_gold_buffer1[thisPixel]]);
	fft_buffer1[thisPixel] = fft_gold_buffer1[thisPixel];
      }
    }
  } else { // Zoomed
    for (i=0; i<512; i++) {
      for (j=0; j<512; j++) { // Zoom it up
	thisX = fftZoomX + j/2;
	thisY = fftZoomY + i/2;
	thisPixel = i*512 + j;
	XPutPixel(fft_image1, j, i,
		  pixels_return[fft_gold_buffer1[(thisY*512)+thisX]]);
	fft_buffer1[thisPixel] = fft_gold_buffer1[(thisY*512)+thisX];
      }
    }
  }
  
  if(doingAutoDefocusRefine && adrZoomDraw) { // Black it out
    for (i=0; i<512; i++) {
      for (j=0; j<512; j++) { // Zoom it up
	thisPixel = i*512 + j;
	fft_buffer1[thisPixel]= fft_gold_buffer1[((128+(i/2))*512)+
						    (256-(j/2))];
      }
    }
  }
  
  is1D = XtIsManaged(autoDRScreenW); 
  is2D = XtIsManaged(auto2DDRScreenW); 

  if(is2D && drawAuto2Dctf) { // Draw the auto2Dctf feedback
    for (i=0; i<512; i++)
      for (j=0; j<256; j++)
	XPutPixel(fft_image1, 256+j, i, 
		  pixels_return[(int)auto2DdrawBuf[(i*256)+j]]);
    XFlush(dsply);
  }
  
  if(XtIsManaged(fftScreenW))
    XPutImage(dsply, fft_window1, fft_gc, fft_image1, 0, 0,
	      0, 0, 512, 512);
  XFlush (dsply);

  if(is1D || drawCTF) { // Draw the CTF elipse
    int beginEl, endEl, *dX1P, *dY1P, *drawX1, *drawY1;
    float distAsq, distBsq, distA, distB;
    float ang1, angThis;
    
    // We calculate drawX1 and drawY1 below 0 and above
    // 1024 so make thee arrays to have entries there or
    // we get memory violations/trashing

    dX1P = (int *)malloc(sizeof(int)*2048);
    dY1P = (int *)malloc(sizeof(int)*2048);
    drawX1 = dX1P + 512;
    drawY1 = dY1P + 512;
    
    for (j= 0; j<MAXCTFS; j++) {
      if(pickedLatPts) { // Only do if there's two 
	
	thisNode = j + 1.0;
	
	distB = ctfPts[j].distB;
	distBsq = distB * distB;
	distA = ctfPts[j].distA;
	distAsq = distA * distA;
	ang1 = ctfAngMajorValue;
	
	beginEl= 512-((int)distA); // We offset this by 256 pixels
	endEl= 512+((int)distA);
	// Don't threshold it 
	// beginEl = MAX(0,beginEl);
	// endEl = MIN(endEl,1023);
	
	for (i=beginEl; i<=endEl; i++) { // Do all Y
	  int thisX, thisY, thisI;
	  double tempF, thisR;
	  
	  thisI = i-256;  // We're offsetting this by 256 pixels 
	  
	  tempF = distBsq * (1.0 - ((thisI-256.0)*(thisI-256.0))/ distAsq);
	  tempF= sqrt(fabs(tempF)); 
	  
	  if(thisI == beginEl-256 || thisI == endEl-256)
	    tempF= 0.0;
	  
	  if(thisI==256)
	    angThis = atan(tempF/0.0000001);
	  else
	    angThis = atan(tempF/(256.0-thisI));
	  
	  thisX = 256 - thisI;
	  thisY = tempF;
	  thisR = sqrt((double)(thisX*thisX + thisY*thisY));
	  
	  // Now rotate in ang1
	  drawX1[i] = 256 + thisR*cos(angThis-ang1);
	  drawY1[i] = 256 + thisR*sin(angThis-ang1);
	  
	  drawX1[i] = DRAWBLANK(drawX1[i]);
	  drawY1[i] = DRAWBLANKY(drawY1[i]);
	}
	
	// Now draw them
	for (i=beginEl ;i<endEl ;i++) { 
	  // i==256 draws a line down the middle
	  if(i != 512) { // Then draw them
	    if(j == ctfCurveNum) { // Special
	      XDrawLine(dsply, XtWindow(fftDrawingW), image_gcRED,
			CNVPTXC(drawX1[i]), CNVPTYC(drawY1[i]),
			CNVPTXC(drawX1[i+1]), CNVPTYC(drawY1[i+1]));
	      if(!doingAutoDefocusRefine || // RSS
		 (doingAutoDefocusRefine && !adrZoomDraw))
		XDrawLine(dsply, XtWindow(fftDrawingW), image_gcRED,
			  CNVPTXC(256 + (256-drawX1[i])),
			  CNVPTYC(256 - (drawY1[i]-256)),
			  CNVPTXC(256 + (256-drawX1[i+1])),
			  CNVPTYC(256 - (drawY1[i+1]-256)));  
	    } else { // Normal
	      XDrawLine(dsply, XtWindow(fftDrawingW), image_gcLAT,
			CNVPTXC(drawX1[i]), CNVPTYC(drawY1[i]),
			CNVPTXC(drawX1[i+1]), CNVPTYC(drawY1[i+1]));
	      
	      if(!doingAutoDefocusRefine ||
		 (doingAutoDefocusRefine && !adrZoomDraw))
		XDrawLine(dsply, XtWindow(fftDrawingW), image_gcLAT,
			  CNVPTXC(256 + (256-drawX1[i])),
			  CNVPTYC(256 - (drawY1[i]-256)),
			  CNVPTXC(256 + (256-drawX1[i+1])),
			  CNVPTYC(256 - (drawY1[i+1]-256)));  
	    }
	    XFlush(dsply);  
	  }
	}
	      
	XFlush (dsply);
      }  
    }
    free(dX1P);
    free(dY1P);
      
    if(pickedLatPts) { // Only do if there's two
      // Reset these just in case
      latPt[0].measuredX= ctfPts[ctfCurveNum].latX0;
      latPt[0].measuredY= ctfPts[ctfCurveNum].latY0;
      latPt[1].measuredX= ctfPts[ctfCurveNum].latX1;
      latPt[1].measuredY= ctfPts[ctfCurveNum].latY1;
    }
    
    // Now see if there is any autoDR (Automatic Defocus Refinement
    // stuff to draw
    if(is1D && doingAutoDefocusRefine) {
	  
      if(ca1d) { // Don't do if nothing there
	maxCA = minCA = ca1d[3];
	minCA = 0.0;
	for (i=3; i<256; i++) {
	  minCA = MIN(ca1d[i], minCA);
	  maxCA = MAX(ca1d[i], maxCA);
	}
	if(maxCA != minCA) {
	  scaleCircFactor = 505.0 / (maxCA-minCA); 
	}
      }

      if(adrCircAvgDraw && ca1d)
	drawCtfCurve(ca1d, 510.0, minCA, scaleCircFactor, 1, image_gcGREEN);

      if(adrBDraw && adr_b)
	drawCtfCurve(adr_b, 510.0, minCA, scaleCircFactor, 1, image_gcBLUE);
      
      if(adrSDraw && adr_s)
	drawCtfCurve(adr_s, 510.0, minCA, scaleCircFactor, 1, image_gcRED);
      
      if(adrSminusBDraw && adr_sminusb) {
	farray_minmax(adr_sminusb, &minCA, &maxCA, 256);
	if(maxCA != minCA) {
	  scaleCircFactor = 250.0 / (maxCA-minCA); 
	  drawCtfCurve(adr_sminusb, 255.0, minCA, scaleCircFactor, 
		       1, image_gcBLUE);	  
	}
      }
      
      if(adrDDraw && adr_d) { // Use S-B for scaling
	farray_minmax(adr_sminusb, &minCA, &maxCA, 256);
	if(maxCA != minCA) {
	  scaleCircFactor = 250.0 / (maxCA-minCA); 
	  drawCtfCurve(adr_d, 255.0, minCA, scaleCircFactor, 
		       1, image_gcWHITE);	  
	}
      }

      if(adrCtfsqDraw && adr_ctfsq) {
	farray_minmax(adr_ctfsq, &minCA, &maxCA, 256);
	if(maxCA != minCA) {
	  scaleCircFactor = 250.0 / (maxCA-minCA); 
	  drawCtfCurve(adr_ctfsq, 255.0, minCA, scaleCircFactor, 
		       1, image_gcGREEN);	  
	}
      }
	  
    } // end if (is1D && doingAutoDefocusRefine)

    if(is2D &&  doingAuto2DDefocusRefine) { // Draw the axes 2D refine
      float tmpAng, curAngle;
      ix1 = 256;
      iy1 = 256;
      
      curAngle = wstrtod(ctfAngMajorW);
      if(curAngle > 180.0)
	curAngle -= 180.0;
      
      // First do major
      if(curAngle < 90.0) {
	ix2 = 512;
	iy2 = 256.0*(1.0 - tan(deg2rad(curAngle)));
      }
      else if(curAngle == 90.0) {
	ix2 = ix1;
	iy2 = 0;
      }
      else {
	tmpAng = 90.0 - (curAngle-90.0);
	ix2 = 0;
	iy2 = 256.0*(1.0 - tan(deg2rad(tmpAng)));
      }

      XDrawLine(dsply, XtWindow(fftDrawingW), image_gcLAT,
		ix1, iy1, ix2, iy2);
      
      // Now minor
      curAngle -= 90.0;
      ix2 = 512;
      iy2 = 256.0*(1.0 - tan(deg2rad(curAngle)));

      XDrawLine(dsply, XtWindow(fftDrawingW), image_gcLAT,
		ix1, iy1, ix2, iy2);

    } // end if (is2D && doingAutoDefocusRefine)
    
  } // endif if (is1D || drawCTF
  else { // Lattice refinement
    if(!layerLines) { // 2D draw
      if(drawLatToggle && pickedLatPts && !measureFFT) {
	
	xcen= ycen = 256.0;
	
	if(axL != 0.0)
	  nx = abs((int)(512.0/(axL*fftZoomFactor))) + 11;
	if(byL != 0.0)
	  ny = abs((int)(512.0/(byL*fftZoomFactor))) + 11;
	
	if(ny < 50000 && nx < 50000) {
	  // Don't do ridiculous ones
	  //  **************************
	  //  *    astar grid lines    *
	  //  **************************

	  ik = -(ny/2) - 2;
	  for (iy=0; iy<ny; iy++) {
	    ik = ik + 1;
	    ih = -(nx/2) - 1;
	    x1 = xcen + ih*axL*fftZoomFactor + ik*bxL*fftZoomFactor;
	    
	    y1 = ycen - ih*ayL*fftZoomFactor - ik*byL*fftZoomFactor;
	    y1--;
	    
	    for (ix=0; ix<nx; ix++) {
	      x2 = x1 + axL*fftZoomFactor;
	      y2 = y1 - ayL*fftZoomFactor;
	      ix1 = x1+0.5;
	      iy1 = y1+0.5;
	      ix2 = x2+0.5;
	      iy2 = y2+0.5;
	      XDrawLine(dsply, XtWindow(fftDrawingW), image_gcLAT,
			ix1, iy1, ix2, iy2);
	      
	      x1 = x2;
	      y1 = y2;
	      ih = ih + 1;
	      
	    }
	  }
	      
	  // c  **************************
	  // c  *    bstar grid lines    *
	  // c  **************************

	  ih = -(nx/2) - 2;
	  for (ix=0; ix<nx; ix++) {
	    ih = ih + 1;
	    ik = -(ny/2) - 1;
	    x1 = xcen + ih*axL*fftZoomFactor + ik*bxL*fftZoomFactor;
	    
	    y1 = ycen - ih*ayL*fftZoomFactor - ik*byL*fftZoomFactor;
	    y1--;
	    
	    for (iy=0; iy<ny; iy++) {
	      x2 = x1+bxL*fftZoomFactor;
	      y2 = y1-byL*fftZoomFactor;
	      ix1 = x1+0.5;
	      iy1 = y1+0.5;
	      ix2 = x2+0.5;
	      iy2 = y2+0.5;
	      XDrawLine(dsply, XtWindow(fftDrawingW), image_gcLAT,
			ix1,iy1,ix2,iy2);

	      x1 = x2;
	      y1 = y2;
	      ik = ik+1;
	    }
	  }
	} // Don't do ridiculous ones
	
      } // Draw 2D lattice lines
    } // 2D draw
    else if(drawLatToggle && pickedLatPts && !measureFFT) {
      // Just draw layerlines
      float space, fall, slope, intercept, thisI;
      float cosang, tanang;
      int ix1, ix2, iy1, iy2;
      if(latLineOrder > 0) {
	fall  = filtSizeValue;
	slope = (float)(latPt[1].measuredY - latPt[0].measuredY) / 
	  (float)(latPt[0].measuredX - latPt[1].measuredX);

	intercept = (256 - latPt[0].measuredY) - 
	  slope*(latPt[0].measuredX - 256);

	if(pickLattice != -99 && !latValChanged) {
	  latScrollValue[1] = 
	    dstar = dstarSave= (fabs(intercept*
				     cos(atan2(slope,1.0))) / 
				(float)latLineOrder);    
	  latScrollValue[3]= omega= omegaSave = 
	    rad2deg(atan2((slope),1.0));

	  // Do we really need to do this here??
	  // May cause infinite looping  SIgnal no redraw with
	  // -99 tag

	  latReset_activate(NULL, NULL, NULL);  
		   
	} // Calc it
	else { // Get from scroll
	  dstar = latScrollValue[1]; 
	  omega = latScrollValue[3]; 
	  
	  slope     = tan(deg2rad(omega)); 
	  intercept = cos(deg2rad(omega)); 
	}

	space = dstar;  // ORIG way 
	cosang = cos(deg2rad(omega));
	tanang = tan(deg2rad(omega));
	
	if(cos(omega) != 0.0)
	  intercept = dstar / cos(deg2rad(omega));
	
	thisI = 256.0;
	while(thisI < 2750.0) {
		  
	  // Draw in positive direction
	  ix1= 0;
	  iy1= thisI - slope*(ix1-256.0);
	  ix2= 511;
	  iy2= thisI - slope*(ix2-256.0);

	  // Why do we have to do this???
	  iy1--;
	  iy2--;
	  
	  XDrawLine(dsply, XtWindow(fftDrawingW), image_gcLAT,
		    ix1, iy1, ix2, iy2);
	  
	  if(thisI != 256.0) { // Draw in negative direction.
	    ix1= 0;
	    iy1= 256.0 - (thisI-256.0) - slope*(ix1-256.0);
	    ix2= 511;
	    iy2= 256.0 - (thisI-256.0) - slope*(ix2-256.0);
	    // Why do we have to do this???
	    iy1--;
	    iy2--;
	    XDrawLine(dsply, XtWindow(fftDrawingW), image_gcLAT,
		      ix1, iy1, ix2, iy2);
	  }
	  
	  // ORIGWAY thisI += fabs(space);
	  thisI += fabs(intercept)*fftZoomFactor;
	  
	}
	
      }
      else 
	askThem("Must have a positive Lattice Line Order.", 
		XmDIALOG_INFORMATION);
      
    } // Just draw layerlines
    
  }// Lattice refinement
  
  
  if(!doingAutoDefocusRefine) {
    // Only draw these if not auto defocus refining
    if( (pickLattice > 0 || pickedLatPts) &&
	!filterToggle && !measureFFT) { // Draw what they've selected
      drawX= CNVPTX(latPt[0].measuredX) - 5;
      drawY= CNVPTY(latPt[0].measuredY)-1;
      drawXfin= CNVPTX(latPt[0].measuredX) + 5;
      drawYfin= CNVPTY(latPt[0].measuredY)-1;
      if(pickLattice != -99)
	drawCross(drawX,drawY,drawXfin,drawYfin, image_gcGREEN);
      drawX= CNVPTX(latPt[0].measuredX);
      drawY= CNVPTY(latPt[0].measuredY) - 5;
      drawXfin= CNVPTX(latPt[0].measuredX);
      drawYfin= CNVPTY(latPt[0].measuredY) + 5; 
      if(pickLattice != -99)
	drawCross(drawX,drawY,drawXfin,drawYfin, image_gcGREEN);
      else 
	drawCross(CNVPTX(latPt[0].measuredX),CNVPTY(latPt[0].measuredY),
		  0,0,image_gcGREEN);
      
      drawX= CNVPTX(latPt[1].measuredX) - 5;
      drawY= CNVPTY(latPt[1].measuredY)-1;
      drawXfin= CNVPTX(latPt[1].measuredX) + 5;
      drawYfin= CNVPTY(latPt[1].measuredY)-1;
      if(pickLattice != -99)
	drawCross(drawX,drawY,drawXfin,drawYfin, image_gcRED);
      drawX= CNVPTX(latPt[1].measuredX);
      drawY= CNVPTY(latPt[1].measuredY) - 5;
      drawXfin= CNVPTX(latPt[1].measuredX);
      drawYfin= CNVPTY(latPt[1].measuredY) + 5;
      if(pickLattice != -99)
	drawCross(drawX,drawY,drawXfin,drawYfin, image_gcRED);
      else 
	drawCross(CNVPTX(latPt[1].measuredX),CNVPTY(latPt[1].measuredY),
		  0,0,image_gcRED);
    }
  }
  
  if( (filterToggle && drawHiLoToggle) ||
      (lopHipSwitch && drawHiLoToggle) ) {
    // Draw the low/hi pass filters
    int arcRadius;
    if(hiResoFilter > 0) {
      drawX = 256;
      drawY = 256;
      drawY= 255;
      if(fftSize == 128) {
	drawX = 257;
	drawY = 254;
      }
      else if(fftSize == 256) {
	drawX = 256;
	drawY = 255;
      }
      arcRadius = sqrt((double)hiResoFilter) * fftZoomFactor;
      XDrawArc(dsply, XtWindow(fftDrawingW), image_gcRED,
	       (drawX-arcRadius), (drawY-arcRadius),
	       (arcRadius*2), (arcRadius*2), 0, (360*64));
    }

    if(lowResoFilter < 320000) {
      drawX = 256;
      drawY = 256;
      drawY= 255;
      if(fftSize == 128) {
	drawX = 257;
	drawY = 254;
      }
      else if(fftSize==256) {
	drawX = 256;
	drawY = 255;
      }
      
      arcRadius = sqrt((double)lowResoFilter) * fftZoomFactor;
      XDrawArc(dsply, XtWindow(fftDrawingW), image_gcGREEN,
	       (drawX-arcRadius), (drawY-arcRadius),
	       (arcRadius*2), (arcRadius*2), 0, (360*64));
    }

    if(lopHipSwitch && drawHiLoToggle) {
      // Draw dotted circles where the drop off is
      float hiCutmax, loCutmax;
      
      loCutmax  = wstrtod(lopTextW);
      hiCutmax  = wstrtod(hipTextW);

      if(hiResoFilter > 0) {
	drawX = 256;
	drawY = 256;
	drawY= 255;
	if(fftSize == 128) {
	  drawX = 257;
	  drawY = 254;
	}
	else if(fftSize==256) {
	  drawX = 256;
	  drawY = 255;
	}
	
	arcRadius = (sqrt((double)hiResoFilter)-hiCutmax) * fftZoomFactor;
	XDrawArc(dsply, XtWindow(fftDrawingW), image_gcDOTred,
		 (drawX-arcRadius), (drawY-arcRadius),
		 (arcRadius*2),(arcRadius*2), 0, (360*64));
      }

      if(lowResoFilter < 320000) {
	drawX = 256;
	drawY = 256;
	drawY= 255;
	if(fftSize == 128) {
	  drawX = 257;
	  drawY = 254;
	}
	else if(fftSize==256) {
	  drawX = 256;
	  drawY = 255;
	}
	
	arcRadius = (sqrt((double)lowResoFilter)+loCutmax) * fftZoomFactor;
	XDrawArc(dsply, XtWindow(fftDrawingW), image_gcDOTgreen,
		 (drawX-arcRadius), 
		 (drawY-arcRadius),
		 (arcRadius*2),(arcRadius*2),
		 0, (360*64));
      }	  
    }

  } // Draw the low/hi pass filters
  
  if(contour) { // Put up contour image
    if(XtIsManaged(fftScreenW))
      XPutImage(dsply, fftContour_window1, fftContour_gc, 
		fftContour_image1, 0, 0, 0, 0, 128, 128);
    drawFFTcontour();
    XFlush (dsply);
  }
  
  if(is2D &&  doingAuto2DDefocusRefine  && 
     (drawAuto2Dscurve || a2drSDraw || a2drSminusBDraw ||
       a2drCtfsqDraw || a2drBDraw || a2drDDraw)) {
    // Draw the auto2Dscurve feedback
    float psi, phi, pi, radian, degree, defocus;
    float thisAngle, ratio, xDiff, yDiff, saveDef;
    
    // Now let's draw the appropriate s curve
    piraddeg_(&pi,&radian,&degree);
    
    phi = auto2Dangle * radian;
    
    // Now calc the angle on the screen
    psi = ratio = 0.0;
    if(scurveClick.x < 512 && scurveClick.y < 256 && scurveClick.y > 0) {
      yDiff = 256.0 - scurveClick.y;
      xDiff = fabs(scurveClick.x - 256.0);
      ratio = (xDiff != 0.0) ? yDiff/xDiff : 999999999.9;
    }

    thisAngle = atan((float)ratio);
    if(scurveClick.x < 256.0)
      thisAngle = (90.0*radian) + (90.0*radian - atan(ratio));
    // If degrees needed 
    // psi = thisAngle * degree;
      psi = thisAngle ;
      
      if(ratio > 0.0) {
	focus_astig__(&defocus, &auto2DdefMajor, &auto2DdefMinor, &psi, &phi);
	
	sprintf(outStr,"[%f] defocus",defocus);
	XDrawString(dsply, XtWindow(fftDrawingW), image_gcRED,
		    15, 15, outStr, strlen(outStr));
	
	XDrawLine(dsply, XtWindow(fftDrawingW), image_gcRED,
		  257,255, scurveClick.x,scurveClick.y);
	XFlush(dsply);
	
	if(drawAuto2Dscurve) {
	  saveDef = ctffocusMajValue;
	  ctffocusMajValue = defocus;
	  drawSCURVEonFFT();
	  ctffocusMajValue = saveDef;
	}
      }

  } // Draw the auto2Dscurve feedback

  if(is2D && doingAuto2DDefocusRefine  && 
     (a2drSDraw || a2drSminusBDraw || a2drCtfsqDraw ||
      a2drBDraw || a2drDDraw)) {

      float defMajor, defMinor, angle, slope, angdraw;
      float   volts, ampContrast, Cs, pixelSize;
      float minS, maxS, scaleS;
      int xfac;
      
      defMajor = wstrtod(a2ddrfjW);
      defMinor = wstrtod(a2ddrfnW);
      angle = wstrtod(a2ddragW);
      volts   = wstrtod(ctfKvW) * 1000.0; // Converting to kV
      Cs   = wstrtod(ctfSpAberW);
      // DONT DO THIS??? Cs *= 10000000.0;  // Put into Angstroms
      ampContrast = wstrtod(ctfAmpFactorW);
      pixelSize   = wstrtod(ctfPixSizeW);
      pixelSize *= 10.0;     // Convert to Angs
      
      if(scurveClick.x != 256.0)
	slope = (256.0 - scurveClick.y) / (scurveClick.x - 256.0);
      else
	slope = 0.0;

      if(slope==0.0)
	slope = 0.00000001;
      angdraw = atan(slope);
      angdraw = rad2deg(angdraw);
      
      robangle_curves__(&angdraw, &angle, poly_paras, &defMajor, &defMinor,
		       &hkl_max, &order_b, &order_d, fftasum, 
		       &volts, &ampContrast, &Cs, &pixelSize,
		       a2dr_s, a2dr_sminusb, a2dr_ctfsq, a2dr_d, a2dr_b);
      
      xfac = 512.0 / fftSize;
      farray_minmax(&a2dr_s[2], &minS, &maxS, 256-2);
      if(maxS != minS)
	scaleS = 505.0 / (maxS-minS);
      
      if(a2drSDraw && a2dr_s) {
	maxCA = maxS;
	minCA = minS;
	if(maxCA != minCA) {
	  scaleCircFactor = scaleS;
	  drawCtfCurve(a2dr_s, 510.0, minCA, scaleCircFactor, 
		       xfac, image_gcRED);	  
	}
      }

      if(a2drSminusBDraw && a2dr_sminusb) {
	farray_minmax(&a2dr_sminusb[2], &minCA, &maxCA, 256-2);
	if(maxCA != minCA) {
	  scaleCircFactor = 250.0 / (maxCA-minCA); 
	  drawCtfCurve(a2dr_sminusb, 255.0, minCA, scaleCircFactor, 
		       xfac, image_gcBLUE);	  
	}
      }

      if(a2drCtfsqDraw && a2dr_ctfsq) {
	farray_minmax(&a2dr_ctfsq[2], &minCA, &maxCA, 256-2);
	if(maxCA != minCA) {
	  scaleCircFactor = 250.0 / (maxCA-minCA); 
	  drawCtfCurve(a2dr_ctfsq, 255.0, minCA, scaleCircFactor, 
		       xfac, image_gcGREEN);	  
	}
      }

      if(a2drDDraw && a2dr_d) {
	farray_minmax(&a2dr_d[2], &minCA, &maxCA, 256-2);
	if(maxCA != minCA) {
	  scaleCircFactor = 250.0 / (maxCA-minCA); 
	  drawCtfCurve(a2dr_d, 255.0, minCA, scaleCircFactor, 
		       xfac, image_gcWHITE);
	}
      }

      if(a2drBDraw && a2dr_b) {
	maxCA = maxS;
	minCA = minS;
	if(maxCA != minCA) {
	  scaleCircFactor = 250.0 / (maxCA-minCA); 
	  drawCtfCurve(a2dr_b, 510.0, minCA, scaleCircFactor, 
		       xfac, image_gcBLUE);
	}  
      }
      
  } // Get curves 
  
} // end function DrawFFTWindow



void drawCtfCurve(float *farr, float shift, float minCA,
		  float scale, int xfac, GC color)
{
  // function name: drawCtfCurve
  // Draws a curve on the FFT window

  int i, thisX, thisXp1, scaledY, scaledYp1;

  for (i=0; i<255; i++) {
    scaledY   = shift - (farr[i]   - minCA)*scale;
    scaledYp1 = shift - (farr[i+1] - minCA)*scale;
    thisX   = i*xfac + 256;
    thisXp1 = (i+1)*xfac + 256;
    if(i == 0)
      scaledY = 256;
    thisX   = DRAWBLANK(thisX);
    thisXp1 = DRAWBLANK(thisXp1);
    XDrawLine(dsply, XtWindow(fftDrawingW), color,
	      CNVPTX(thisX), scaledY, CNVPTX(thisXp1), scaledYp1);
  }

} // end function drawCtfCurve



void latText_changed(Widget widget, XtPointer client_data, 
		     XtPointer call_data)
{
  // function name: latText_changed
  // Call here whenever lattice text widgets changed

  int tag = *((int *) client_data);

  if(tag == RASTERSTEP)
    rasterStep = wstrtod(widget);
  else if(tag == FARADIUS)
    fourAverRadius = wstrtod(widget);
  else if(tag == FACUT)
    fourAverCut = wstrtod(widget);
  else if(tag == KV) {
    float volts,calcWavelength;
    kv = wstrtod(widget);
    volts = kv * 1000.0;
    if(volts > 0.0) {
      calcWavelength = 1.23 / sqrt(volts + volts*volts*1.0e-6);
      sprintf(outStr,"%8.6f",calcWavelength);
      XtVaSetValues(ctfWaveW, XmNlabelString, 
		    XmStringCreateLocalized(outStr), NULL);
      XFlush(dsply);
    }
  }
  else if(tag == WIENER)
    wiener = wstrtod(widget);
  else if(tag == ANGMAJOR)
      angMajor = wstrtod(widget);
  else if(tag == TFACTOR)
      ctfTFactor = wstrtod(widget);
  else if(tag == MODE) 
      ctfDivideMode = wstrtod(widget);
  else if(tag == SPABER)
      spAberration = wstrtod(widget);
  else if(tag == CTFCURVENUM) {
    ctfCurveNum = wstrtod(widget);
    ctfCurveNum--;   // Human numbers from 1 not 0
    if(ctfCurveNum >= MAXCTFS)
      ctfCurveNum = MAXCTFS-1;
    ctfCurveNum = MAX(ctfCurveNum, 0);
  }

  if(tag != FFTMEASURE  && tag != CTFCURVENUM) {
    // Don't do for measurement
    if(layerLines) { // Get lattice line order
      latLineOrder = wstrtod(widget);
    }
    else { // 2D lattice
      if(tag == LAT1X) { 
	latPt[0].x = wstrtod(widget);
	latLineOrder = wstrtod(widget); // Remeber for when we need it
      }
      else if(tag == LAT1Y)
	latPt[0].y = wstrtod(widget);
      else if(tag == LAT2X)
	latPt[1].x = wstrtod(widget);
      else if(tag == LAT2Y)
	latPt[1].y = wstrtod(widget);
    }
  }
  
} // end function latText_changed



void ctfText_changed(Widget widget, XtPointer client_data, 
		     XtPointer call_data)
{
  // function name: ctfText_changed
  // Call here whenever any of the ctf value widgets change


  // In all cases - don't set the scrollbar it gives very funky results

  if(widget == ctfFminTW) { // Get new ab ratio
    ctfFocusMinvalue = wstrtod(widget);
    calcNodes();
    // calcTFpixel(ctffocusMajValue, ctfFocusMinvalue); Also funky
    XtSetArg(arglst[0], XmNvalue, (int)(ctfFocusMinvalue*100.0));
    if(!changeCTFtext)
      XtSetValues(ctfFminSW, arglst, 1);  
    TextSetFloat(ctfMinorDefocusW, ctfFocusMinvalue, FALSE);
  }
  else if(widget == ctfAmajTW) { // Get omega
    ctfAngMajorValue = wstrtod(widget);
    XtSetArg(arglst[0], XmNvalue, (int)(ctfAngMajorValue*100.0));
    if(!changeCTFtext)
      XtSetValues(ctfAmajSW, arglst, 1); 
    ctfAngMajorValue = deg2rad(ctfAngMajorValue);
    sprintf(outStr,"%.2f",rad2deg(ctfAngMajorValue));
    TextSet(ctfAngMajorW, outStr, FALSE); 
  }
  else if(widget == ctfFmajTW) { // Get defocus
    ctffocusMajValue = wstrtod(widget);
    calcNodes();
    // calcTFpixel(ctffocusMajValue, ctfFocusMinvalue); Also funky
    // calcMeanTFpixel(ctffocusMajValue);
    XtSetArg(arglst[0], XmNvalue, (int)(ctffocusMajValue*100.0));
    if(!changeCTFtext)
      XtSetValues(ctfFmajSW, arglst, 1); 
    TextSetFloat(ctfMajorDefocusW, ctffocusMajValue, FALSE);
  }
  
  changeCTFtext = FALSE; // Toggle to false since we just did change text
  XFlush(dsply);

} // end function ctfText_changed



void ctfValue_change(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{
  // function name: ctfValue_change
  // Call here when the ctf is changed to recalculate the CTF and then
  // redisplay.
  
  calcFFT(FALSE, 0, FALSE);
  XSync(dsply,FALSE);
  XFlush(dsply);
  DrawFFTWindow(NULL, NULL, NULL);
  XSync(dsply,FALSE);
  XFlush(dsply);

} // end function ctfValue_change



void stuffContourBuffer()
{
  // function name: stuffContourBuffer
  // Stuff the current concepts of x,y for lattice points in zoomed up
  // contour buffer

  int i, j, thisX, thisY;
  
  if(currentLatPt < 2) { // first
    // contour = TRUE;
    for (i=0; i<128; i++) {
      for (j=0; j<128; j++) {
	thisX = latPt[currentLatPt].measuredX - 8 + (j/8);
	thisY = latPt[currentLatPt].measuredY - 8 + (i/8);
	
	fft_contour_buffer[(i*128)+j]= 
	  fft_gold_buffer1[((latPt[currentLatPt].measuredY-8+(i/8))*512)+
			   (latPt[currentLatPt].measuredX-8+(j/8))];
	XPutPixel(fftContour_image1, j, i,
		  pixels_return[fft_contour_buffer[(i*128)+j]]);
      }
    }
    
    // Now draw the contour plot in the window
    // USe contour_a_rob() fortran routine
    // drawFFTcontour(); Done in DrawFFTWindow(NULL, NULL, NULL)
    
    if(currentLatPt==0)
      XtSetArg(arglst[0], XmNborderColor, pixels_return[GREEN]);
    else 
      XtSetArg(arglst[0], XmNborderColor, pixels_return[RED]);
    
    XtSetValues(contourW, arglst, 1); 
    XFlush(dsply);
    
    if(layerLines)
      sprintf(outStr,"[%1d] xy[%d,%d]", currentLatPt+1,
	      latPt[currentLatPt].measuredX, latPt[currentLatPt].measuredY);
    else if(drawCTF)
      sprintf(outStr,"[%1d][%d,%d] xy[%d,%d]", currentLatPt+1,
	      latPt[currentLatPt].x, latPt[currentLatPt].y,
	      latPt[currentLatPt].measuredX-256,
	      256-latPt[currentLatPt].measuredY);    
    else
      sprintf(outStr,"[%1d][%d,%d] xy[%d,%d]", currentLatPt+1,
	      latPt[currentLatPt].x, latPt[currentLatPt].y,
	      latPt[currentLatPt].measuredX,
	      latPt[currentLatPt].measuredY);    

    doLabel(outStr, fftFeedbackLabel2W, DEFAULT); 
    doLabel("", fftFeedbackLabel1W, DEFAULT);
    doLabel("", fftFeedbackLabel3W, DEFAULT);
  }
  else { // none 
    // contour = FALSE;
    for (i=0; i<128; i++) {
      for (j=0; j<128; j++) {
	XPutPixel(fftContour_image1, j, i, pixels_return[BLACK]);
	fft_contour_buffer[(i*128)+j] = BLACK;
      }
    }
  }
  
} // end function stuffContourBuffer



void findLatPt(int x, int y)
{
  // function name: findLatPt
  // Which lattice point are they wanting?

  int dist1, dist2;

  dist1 = (latPt[0].measuredX - x) * (latPt[0].measuredX - x) +
          (latPt[0].measuredY - y) * (latPt[0].measuredY - y);
  dist2 = (latPt[1].measuredX - x) * (latPt[1].measuredX - x) +
          (latPt[1].measuredY - y) * (latPt[1].measuredY - y);
  
  currentLatPt = (dist2 > dist1) ? 0 : 1;

} // end function findLatPt



void fftMotion(Widget widget, caddr_t unused, 
	       register XMotionEvent *pEvent)
{
  // function name: fftMotion
  // Whenever the pointer moves across our FFT window call here. This
  // is where we draw the zoomed up display in the contour window.

  int x, y, i, j, thisX, thisY;
  static int lastX=-99, lastY=-99;
  float thisRadius, sizeF, apart, bpart, amp, phase;

  // Make all changes current
  XFlush(dsply);               

  if(pEvent->type != MotionNotify)
    return;
  
  // Get coords of pointer
  x = pEvent->x;     
  y = pEvent->y;
  
  if(pickLattice == -99) {
    // Don't do if they're trying to pick points
    
    sizeF = fftSize / 512.0;
    // ORIG way   thisX = (int)(((float)x-256.0));
    thisX = (x-256)/fftZoomFactor;
    if(thisX < 0 && fftSize < 512)
      thisX -= (int)(512.0/fftSize)-1;
    // ORIG way   thisY = (int)((255.0-(float)y));
    thisY = (float)(255-y)/fftZoomFactor;
    if(thisY < 0 && fftSize < 512)
      thisY -= (int)(512/fftSize)-2;
    thisX = MIN(MAX(thisX,-256),255);
    thisY = MIN(MAX(thisY,-256),255);
    thisX = thisX * sizeF;
    thisY = thisY * sizeF;
    
    if(fftImage_buffer2)
      fft_point(fftImage_buffer2, fftSize, fftSize,
		thisX, thisY, &thisRadius, &apart, &bpart);
    
    if(x != 256)
      sprintf(outStr,"x[%d]y[%d] radius[%.0f] Angs[%.2f]",
	      thisX, thisY, thisRadius,
	      (512.0/((x-256.0)/fftZoomFactor) * fftAngsPerPixel));
    else
      sprintf(outStr,"x[%d]y[%d] radius[%.0f] Angs[undef]",
	      thisX,thisY,thisRadius);
    
    doLabel(outStr, fftFeedbackLabel1W, DEFAULT);
    
    sprintf(outStr,"a[%.2E] b[%.2E]", apart, bpart);
    doLabel(outStr, fftFeedbackLabel2W, DEFAULT);
    
    convert_abpart(apart,bpart,&amp,&phase);
    sprintf(outStr,"amp[%.2E] phase[%.2E]", amp, phase);
    doLabel(outStr, fftFeedbackLabel3W, DEFAULT);
    
    if(fftZoomFactor==2) { // Adjust for zoom
      x= fftZoomX + (x / 2);
      y= fftZoomY + (y / 2);
    }
    
    for (i=0; i<128; i++) {
      int thisX, thisY;
      for (j=0; j<128; j++) {
	thisX = x - 8 + (j/8);
	thisY = y - 8 + (i/8);
	if(thisX < 512 && thisY < 512 && thisX >= 0 && thisY >= 0) {
	  fft_contour_buffer[i*128 + j] = 
	    fft_gold_buffer1[thisY*512 + thisX];
	  XPutPixel(fftContour_image1, j, i,
		    pixels_return[fft_contour_buffer[i*128 + j]]);
	} else {
	  fft_contour_buffer[i*128 + j]= BLACK;
	  XPutPixel(fftContour_image1, j, i,
		    pixels_return[fft_contour_buffer[BLACK]]);
	}
      }
    }
    
    if(XtIsManaged(fftScreenW)) { // Draw the contour window
      XPutImage(dsply, fftContour_window1, fftContour_gc, 
		fftContour_image1, 0, 0, 0, 0, 128, 128);
      XDrawLine(dsply, fftContour_window1, image_gcBLACK,
		64, 60, 64, 68);
      XDrawLine(dsply, fftContour_window1, image_gcBLACK,
		60, 64, 68, 64);
    }
    
    XFlush (dsply);
  }
  
  if(doingAutoDefocusRefine || calcFFT1DcircAvg) { // Give feedback
    int distX;
    
    if(doingAutoDefocusRefine && adrZoomDraw)
      sprintf(outStr,"Radius = [%d] Pixels.",(x/2));
    else
      sprintf(outStr,"Radius = [%d] Pixels.",(x-256));
    
    doLabel(outStr, fftMeasPixW, DEFAULT);
    
    // First calc dist on X axis only
    if(doingAutoDefocusRefine && adrZoomDraw) { // special
      if(x > 1)
	sprintf(outStr,"[%.2f] Angs",
		((512.0/((float)(x/2)))*fftAngsPerPixel));
      else
	sprintf(outStr,"[undef] Angs");
    } else { // normal
      if(x != 256)
	sprintf(outStr,"[%.2f] Angs",
		((512.0/((float)(x-256.0)))*fftAngsPerPixel));
      else
	sprintf(outStr,"[undef] Angs");
    }
    
    XtSetArg(arglst[0], XmNlabelString, XmStringCreateLocalized(outStr));
    XtSetArg(arglst[1], XmNforeground, pixels_return[RED] );
    XtSetValues(fftMeasAngW, arglst, 2); 
    XtSetValues(ctfMeasureLabelW, arglst, 2); 
    XFlush(dsply);
    
    // Now calc dist based on pointer
    distX = sqrt((x-256.0)*(x-256.0) + (y-256.0)*(y-256.0));
    
    if(doingAutoDefocusRefine && adrZoomDraw) { // special
      if(distX > 1)
	sprintf(outStr,"[%.2f] Angs", (512.0/(distX/2))*fftAngsPerPixel );
      else
	sprintf(outStr,"[undef] Angs");
    } else { // normal
      if(distX != 256)
	sprintf(outStr,"[%.2f] Angs", (512.0/distX)*fftAngsPerPixel);
      else
	sprintf(outStr,"[undef] Angs");
    }
    
    XtSetArg(arglst[0], XmNlabelString, XmStringCreateLocalized(outStr));
    XtSetArg(arglst[1], XmNforeground, pixels_return[BLUE] );
    XtSetValues(ctfMeasureLabel2W, arglst, 2); 
    XFlush(dsply);
    
    // Don't do it 	  x = CNVPTX(x);
    
    if(lastX >=0 && lastY >= 0 && fftMotionFeedback) { // redraw last one
      XDrawLine(dsply, XtWindow(fftDrawingW), 
		image_gcBLACK, lastX, 512, lastX, 505);
      XDrawLine(dsply, XtWindow(fftDrawingW), 
		image_gcBLACK, lastX, 255, lastX, 256);
    }
    
    // Now put thisX in drawable coords
    x = MAX(256, pEvent->x);     
    sizeF = 1.0;
    thisX = 256 + (int)((x-256.0) * sizeF);
    if(fftMotionFeedback) {
      XDrawLine(dsply, XtWindow(fftDrawingW), image_gcRED,
		thisX, 512, thisX, 505);
      XDrawLine(dsply, XtWindow(fftDrawingW), image_gcRED,
		thisX, 255, thisX, 256);
    }
    lastX = thisX;
    lastY = y;
    
    // Now draw where the pointer has been
    x = pEvent->x;     
    y = pEvent->y;     
    if(fftMotionFeedback)
      XDrawLine(dsply, XtWindow(fftDrawingW), image_gcBLUE, x, y, x, y-1);
  }
  
  XFlush(dsply); // Make changes current

} // end function fftMotion



void fftMeasureDraw(int button)
{
  // function name: fftMeasureDraw
  // Draw the things needed for the FFT Measurement

  int x, y, ix1, ix2, iy1, iy2;
  float lngth;

  DrawFFTWindow(NULL, NULL, NULL); // Refresh the FFT in window

  // Now draw all points
  x = fftMeasures[0].x;
  y = fftMeasures[0].y;
  XDrawPoint(dsply, XtWindow(fftDrawingW), image_gcRED, x, y);
  drawCross(CNVPTX(x), CNVPTY(y), CNVPTX(x+2), CNVPTY(y+2), image_gcRED);

  x = fftMeasures[1].x;
  y = fftMeasures[1].y;
  XDrawPoint(dsply, XtWindow(fftDrawingW), image_gcBLUE, x, y);
  drawCross(CNVPTX(x),CNVPTY(y), CNVPTX(x+2), CNVPTY(y+2), image_gcBLUE);
  
  // Zoom it for drawing 
  ix1= CNVPTX(fftMeasures[0].x);
  ix2= CNVPTX(fftMeasures[1].x);
  iy1= CNVPTY(fftMeasures[0].y);
  iy2= CNVPTY(fftMeasures[1].y);
  XDrawLine(dsply, XtWindow(fftDrawingW), image_gcGREEN, ix1, iy1, ix2, iy2);
  
  // DON'T Zoom it for measuring
  ix1= fftMeasures[0].x;
  ix2= fftMeasures[1].x;
  iy1= fftMeasures[0].y;
  iy2= fftMeasures[1].y;
  lngth = sqrt((double)((ix2-ix1)*(ix2-ix1)) + ((iy2-iy1)*(iy2-iy1)));
  
  sprintf(outStr,"[%.2f] Pixels.   xy[%d,%d] to xy[%d,%d]",
	  lngth*(fftSize/512.0), ix1, iy1, ix2, iy2);
  doLabel(outStr, fftMeasPixW, DEFAULT);
  
  if(lngth==0.0)
    sprintf(outStr,"[n/a] Angstroms.");
  else
    sprintf(outStr,"[%f] Angstroms.",((512.0/lngth)*fftAngsPerPixel));
  
  doLabel(outStr, fftMeasAngW, DEFAULT);
} // end function fftMeasureDraw



void calcLattice(int recalcAxBxAyBy)
{
  // function name: calcLattice
  // Calc the lattice based on prg settings determined by other callbacks.

  float xcen=256.0, ycen=256.0;
  float x1, x2, y1, y2;
  int h1, k1, h2, k2;
  
  // Compute distance from origin
  h1 = latPt[0].x;
  k1 = latPt[0].y;
  h2 = latPt[1].x;
  k2 = latPt[1].y;
  
  x1 = latPt[0].measuredX;
  y1 = latPt[0].measuredY;
  x2 = latPt[1].measuredX;
  y2 = latPt[1].measuredY;
  
  // Put it in TSB space
  y1 = 512.0 - latPt[0].measuredY;
  y2 = 512.0 - latPt[1].measuredY;
  
  if(recalcAxBxAyBy) { // Not recalculated above
    latcen2_(&h1, &k1, &x1, &y1, &h2, &k2, &x2, &y2,
	     &axL, &ayL, &bxL, &byL, &xcen, &ycen);
    
    // printf("Values gotten from latcen2:\nax[%f]ay[%f]bx[%f]by[%f]\n",
    // axL,ayL,bxL,byL);
    
    // Let's always store anga angb gamstr and omega in degrees
    anga= rad2deg(atan2((double)ayL,(double)axL));
    angb= rad2deg(atan2((double)byL,(double)bxL));
    
    // Let's try this ... ??? 
    // anga = 0.0 - anga;
    // angb = 0.0 - angb;
    
    ang    = angb-anga;
    astar  = sqrt(axL*axL+ayL*ayL);
    bstar  = sqrt(bxL*bxL+byL*byL);
    gamstr = angb-anga; 
  }
  
  // Stuff this for things like the crysDisp screen
  map_abang = 180.0 - gamstr;
  if(bstar != 0.0 && astar != 0.0)
    aoverb = 1.0 / (astar / bstar);
  
  omega  = 180.0 - gamstr;
  
  angaSave    = anga;
  angbSave    = angb;
  angSave     = ang;
  
  axLSave     = axL;
  ayLSave     = ayL;
  bxLSave     = bxL;	
  byLSave     = byL;
  
  astarSave   = astar;
  bstarSave   = bstar;
  gamstrSave  = gamstr;
  omegaSave   = omega;
  
  latScrollValue[0] = astar;
  latScrollValue[1] = bstar;
  latScrollValue[2] = gamstr;
  latScrollValue[3] = omega;
  
  latReset_activate(NULL, NULL, NULL);
  
} // end function calcLattice



void ctfArrowActivate(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: ctfArrowActivate
  // Change the current CTF by using the arrows.

  if(widget == ctfCurveUpW)
    ctfCurveNum++;
  else if(widget == ctfCurveDownW)
    ctfCurveNum--;
  
  ctfCurveNum = MIN(ctfCurveNum, MAXCTFS-1);
  ctfCurveNum = MAX(ctfCurveNum, 0);

  maxCtfCurve = MAX((ctfCurveNum+1), maxCtfCurve);

  if(ctfPts[ctfCurveNum].pickedPts) { // Restuff
    latPt[0].measuredX = ctfPts[ctfCurveNum].latX0;
    latPt[0].measuredY = ctfPts[ctfCurveNum].latY0;
    latPt[1].measuredX = ctfPts[ctfCurveNum].latX1;
    latPt[1].measuredY = ctfPts[ctfCurveNum].latY1;
  }
  else { // Guess new points
    ctfPts[ctfCurveNum].latX0 = latPt[0].measuredX;
    ctfPts[ctfCurveNum].latY0 = latPt[0].measuredY;
    ctfPts[ctfCurveNum].latX1 = latPt[1].measuredX;
    ctfPts[ctfCurveNum].latY1 = latPt[1].measuredY;
  }
  
  ctfPts[ctfCurveNum].pickedPts = TRUE;
  TextSetInt(ctfCurveTextW, ctfCurveNum+1, FALSE);
  DrawFFTWindow(NULL, NULL, NULL);

} // end function ctfArrowActivate



void icosaArrowActivateButton(Widget widget, XEvent *event)
{
  // function name: icosaArrowActivateButton
  // Arrows change the speed at which the icosahedron is expanded or
  // contracted.

  float speedFactor;

  if(widget == icosaDownW)
    speedFactor = (float) GetButtonVal(event->xbutton.button, 99, 95, 90) / 100.;
  else
    speedFactor = (float) GetButtonVal(event->xbutton.button, 101, 105, 110) / 100.;
    
  icosaScaleF *= speedFactor;
  
  if(dispIcosa) {
    DrawWindow();
    calcSizeFactor(zoomFactor);
    
    DrawIcosahedron(deg2rad(0.0-globTheta), deg2rad(0.0-globPhi), 
		    deg2rad(globOmega),
		    0, 0, numCols*sizeFactor*2.0, 
		    image_window1); 
    
    DrawIcosahedron(deg2rad(0.0-globTheta), deg2rad(0.0-globPhi),
		    deg2rad(globOmega),
		    500, 255, numCols,
		    XtWindow(mapShadeScreenW));  
  }

  
} // end function icosaArrowActivateButton



void samRenArrowActivate(Widget widget, XtPointer client_data,
			 XtPointer call_data)
{
  // function name: samRenArrowActivate
  // Change the sampling for 3D render.

  int thisVal;
  thisVal = wstrtod(sampleRenderW);
  
  if(widget == sampleUW)
    thisVal++; 
  else if(widget == sampleDW)
    thisVal--; 
  else 
    printf("Unknown SAMREN_ARROW.\n");
  
  thisVal = MAX(MIN(thisVal,10),1);

  TextSetInt(sampleRenderW, thisVal, FALSE);
  
  if(thisVal > 1 && globFract >= 0.75) { // set shading fraction
    globFract = 0.25;
    sprintf(outStr,"%.2f", globFract);
    TextSet(fractValueW, outStr, FALSE); 
    XtSetArg(arglst[0], XmNvalue, (int)(globFract*1000.0));
    XtSetValues(fractScrollW, arglst, 1);
    XFlush(dsply);
  }
  
} // end function samRenArrowActivate



void fftImgNumArrowButton(Widget widget, XEvent *event)
{
  // function name: fftImgNumArrowButton
  // Change the current image to FFT by using the arrows.

  int speedFactor;

  if(!locked) { // Can't do if we're locked out
    speedFactor = GetButtonVal(event->xbutton.button, 1, 2, 5);
    
    if(widget==rotfltImgnumArrowUpW || widget==fftImgnumArrowUpW)
      doTheImgNum(imgnumArrowUpW, speedFactor);
    else 
      doTheImgNum(imgnumArrowDownW, speedFactor);
  }
  
  averageFFT= FALSE;       // Don't want average anymore

  calcFFT(TRUE, 0, 0);
  DrawFFTWindow(NULL, NULL, NULL);

  calcFFT1DcircAvg= FALSE; // Make them do a new circavg

  // Do if rotflt'ing
  newOrigin = TRUE;
  if(XtIsManaged(rotfltScreenW)) {
    iopt[0]= 1;      // iopt[0] == 1 --> Refine symmetry axis
    iopt[1]= 0;
    iopt[2]= 0;      // iopt[2] ==1  --> Plot power spectrum
    iopt[3]= 0;
    rotflt_tryit_activate(0,0,0);
  }

} // end function fftImgNumArrowButton



void latReset_activate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: latReset_activate
  // Reset axL, ayL, bxL and byLto the original values they were
  // calculated to by when the cursor selected the points.

  int thisMin, thisMax, i, sliderSize;

  axL = axLSave;
  ayL = ayLSave;
  bxL = bxLSave;
  byL = byLSave;

  anga = angaSave;
  angb = angbSave;

  astar  = astarSave;
  bstar  = bstarSave;
  gamstr = gamstrSave;
  omega  = omegaSave;
  dstar  = dstarSave;

  // Stuff this for things like the crysDisp screen
  map_abang = 180.0 - gamstr;
  if(bstar != 0.0 && astar != 0.0)
    aoverb = 1.0 / (astar / bstar);
  
  latScrollValue[0]= astarSave;
  latScrollValue[1] = (layerLines) ? dstarSave : bstarSave;
  latScrollValue[2]= gamstrSave;
  latScrollValue[3]= omegaSave;
  if(!layerLines)
    latScrollValue[3]= 0.0;    // Omega starts at 0 now

  if(!inited) return;

  // Now setup the scrolls and labels
  for (i=0; i<4; i++) {
    if(!layerLines || (layerLines && (i==3||i==1))) {
      sprintf(outStr,"%.2f",(latScrollValue[i]));
      TextSet(latticeLabel2W[i], outStr, FALSE); 
      
      // Now setup the scroll bars to hold the new max/min
      if(latScrollValue[i] < 0.0) {
	thisMax = (latScrollValue[i]-(latScrollValue[i]*0.5))*100.0;
	thisMin = (latScrollValue[i]+(latScrollValue[i]*0.5))*100.0;
      } else {
	thisMax = (latScrollValue[i]+(latScrollValue[i]*0.5))*100.0;
	thisMin = (latScrollValue[i]-(latScrollValue[i]*0.5))*100.0;
      }
    }
    
    if(i==3 && !layerLines) { // omega
      thisMax= 6000;  // 60 degrees times 100
      thisMin= -6000;
      latScrollValue[i]= 0.0;
      omega = 0.0;
      sprintf(outStr,"%.2f",(latScrollValue[i]));
	TextSet(latticeLabel2W[i], outStr, FALSE); 
    }
    
    if(!layerLines || (layerLines && (i==3 || i==1))) {
      if(thisMax == thisMin) {// Make them separate
	thisMax += 500;
	thisMin -= 500;
      }
      if(layerLines && (i==3 || i==1)) {
	if(i==3) {
	  thisMax += 1800;   // +/- 180 degrees
	  thisMin -= 1800; 
	} else {
	  thisMax += 500; 
	  thisMin -= 500; 
	}
      }
      sliderSize = (thisMax - thisMin) / 100;
      sliderSize = MAX(sliderSize,1);
      
      XtSetArg(arglst[0], XmNmaximum, (int)(thisMax+(float)sliderSize));
      XtSetArg(arglst[1], XmNminimum, thisMin);
      XtSetArg(arglst[2], XmNsliderSize, sliderSize);
      
      if(i==3)  // omega
	XtSetArg(arglst[3], XmNvalue, (int)(latScrollValue[i]*10.0));
      else 
	XtSetArg(arglst[3], XmNvalue, (int)(latScrollValue[i]*100.0));
      
      XtSetValues(latticeScrollW[i], arglst, 4);
    }
    
    XSync(dsply,FALSE);
    XFlush(dsply);
  }
  
} // end function latReset_activate



void fft_average_activate(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: fft_average_activate
  // Look at the average FFT of all images .

  averageFFT = TRUE;
  calcFFT(TRUE, 0, 0);
  DrawFFTWindow(NULL, NULL, NULL);
  
} // end function fft_average_activate



void fourAver_work()
{
  // function name: fourAver_work
  // Generate the Structure factors

  int i, lastSF, ncol, nrow;
  float axPass, ayPass, bxPass, byPass;
  float x1, x2, y1, y2, thisMin, thisMax;

  if(axL == 0.0 && bxL == 0.0 && ayL == 0.0 && byL == 0.0) {
    askThem("Must create a lattice first.", XmDIALOG_INFORMATION);
  } 
  else {
    pif_init_head(&faHeader);
    
    // Let's try these
    axPass = axL;
    ayPass = ayL;
    bxPass = bxL;
    byPass = byL; 
    
    x1 = axPass;
    y1 = ayPass;
    x2 = bxPass;
    y2 = byPass;
    
    printf("ax[%f]ay[%f]bx[%f]by[%f]\n", x1, y1, x2, y2);
    
    if(fileType == PACKEDPORTABLE) {// packed box
      ncol = purdueHeader.numCols;
      nrow = purdueHeader.numRows;
    } else { // other
      ncol = numCols;
      nrow = numRows;
    }

    emfft_spt_rob1__(&fftSize, &fftSize,
		     &axPass, &ayPass, &bxPass, &byPass, &nsf,
		     &x1, &x2, &y1, &y2, &ncol, &nrow,
		     &rasterStep, &fourAverRadius, &fourAverCut);
    
    if(sFacs != NULL)
      free(sFacs);
    sFacs = (strucFact_t *)malloc(sizeof(strucFact_t)*nsf*2);
    if(fa_sFacs != NULL)
      free(fa_sFacs);
    fa_sFacs = (strucFact_t *)malloc(sizeof(strucFact_t)*nsf*2);
    
    emfft_spt_rob2__(fftImage_fourAver, &fftSize, &fftSize,
		     &axPass, &ayPass, &bxPass, &byPass, &nsf, fa_sFacs,
		     &x1, &x2, &y1, &y2, &ncol, &nrow,
		     &faHeader.a_star, &faHeader.b_star, &faHeader.c_star, 
		     &faHeader.alp_star,
		     &faHeader.bet_star,
		     &faHeader.gam_star,
		     &rasterStep, &fourAverRadius, &fourAverCut,
		     &shiftFFT);
    
      rec_to_real__(&faHeader.a_cell, &faHeader.b_cell, &faHeader.c_cell, 
		   &faHeader.alp, &faHeader.bet, &faHeader.gam, 
		   &vol,
		   &faHeader.a_star, &faHeader.b_star, &faHeader.c_star, 
		   &vol_star, 
		   &faHeader.alp_star, &faHeader.bet_star, &faHeader.gam_star);
      
      thisMin = thisMax = fa_sFacs[0].a_amp;
      origNsf = nsf;  // We may expand this in the impose rtns so remember 

      for (i=0; i<nsf; i++) { // Check all SFs
	// Copy gold std into temp area
	sFacs[i].h     = fa_sFacs[i].h;
	sFacs[i].k     = fa_sFacs[i].k;
	sFacs[i].l     = fa_sFacs[i].l;
	sFacs[i].a_amp = fa_sFacs[i].a_amp;
	sFacs[i].b_phi = fa_sFacs[i].b_phi;
	sFacs[i].dstr  = fa_sFacs[i].dstr;
	
	thisMax = MAX(fa_sFacs[i].a_amp,thisMax);
	thisMax = MAX(fa_sFacs[i].b_phi,thisMax);
	thisMax = MAX(fa_sFacs[i].sig,thisMax);
	thisMax = MAX(fa_sFacs[i].dstr,thisMax);
	
	thisMin = MIN(fa_sFacs[i].a_amp,thisMin);
	thisMin = MIN(fa_sFacs[i].b_phi,thisMin);
	thisMin = MIN(fa_sFacs[i].sig,thisMin);
	thisMin = MIN(fa_sFacs[i].dstr,thisMin);
	
	if(fa_sFacs[i].a_amp == 0.0 && fa_sFacs[i].b_phi == 0.0)
	  lastSF = i;
      }      
  }
  
} // end function fourAver_work



void fourAver_backit(int firstTime)
{
  // function name: fourAver_backit
  // Perform back transform based on the SFs.

  int i, j, backSizeA, backSizeB;
  int ncol, nrow, sf_type, em_ndim;
  float scaleFac, thisMap_abang;

  sf_type = 1;
  em_ndim = 2;  // 2d data here
  
  // Convert Structure factors back into a 2D MAP

  if((float_image_buffer1) && (float_image_buffer1 != lowRes_buffer1))
    free(float_image_buffer1);
  
  // These are already calculated 
  // map_abang = 90.0;
  // aoverb = 1.0;
  // thisMap_abang = 180.0 - map_abang; 

  thisMap_abang = map_abang;
  
  adim = bdim = 256; 
  backSizeA = adim;
  backSizeB = bdim;

  float_image_buffer1 = (float *)malloc(sizeof(float)*2*backSizeA*backSizeB);

  emsf_backt_rob__(&sf_type, &em_ndim, &nsf, sFacs, &map_min, &map_max,
		  float_image_buffer1, &backSizeA, &backSizeB);

  cdNumCols = ncol = backSizeA;
  cdNumRows = nrow = backSizeB;
  cdNumSecs = 1;
  
  printf("\n\nNew dimensions ->  %d,%d,%d\n",ncol,nrow,1);
  
  // Scale the buffer to fit a BYTE output
  if(map_max == map_min)
    scaleFac = 1.0;
  else 
    scaleFac = 254.0 / (map_max - map_min);

  // Scale them all
  // j*2 because the fourt() rtn stuffs it in the first position

  for (i=0; i<cdNumRows; i++)
    for (j=0; j<cdNumCols; j++)
      float_image_buffer1[i*cdNumCols*2 + j*2] = 
	(float_image_buffer1[i*cdNumCols*2 + j*2] - map_min)*scaleFac;
  
  fourierAveragedCrysDisp = TRUE;
  crysDisp_screen_activate(0, (XtPointer)&firstTime, 0); 
  
} // end function fourAver_backit



void cDisp_newAdimBdim_changed(Widget widget, XtPointer client_data, 
			       XtPointer call_data)
{
  // function name: cDisp_newAdimBdim_changed
  // Redraw since adim and/or bdim concept changed

  fourAver_backit(FALSE);

} // end function cDisp_newAdimBdim_changed



void fourAver_refine_impose(int tag)
{
  // function name: fourAver_refine_impose
  // Impose symmetry

  char  outFilename[256];
  //  int lastSF;
  int sf_type, em_ndim, ikey, nlim;
  float thisMin, thisMax;

  sf_type = 1;
  em_ndim = 2;  // 2d data here
  
  // Check to see if we want to do anything to these structure factors

  if(tag==1) { // Refine phase origins
    if(refinePhaseOrigin) { // Refine the SF's for 2,3,4,or 6 symmetry
      float sf_sumfsq;
      ikey = refinePhaseOrigin;
      emsf_origins_rob__(&em_ndim, &sf_type, &nsf, &sf_sumfsq, sFacs, &ikey); 
    }   
  }
  else if(tag==2) { // Impose symmetries
    float thisAoverb;
    if(imposeSymmetry) {
      nlim = nsf * 2 ;

      if(imposeSymmetry == P2)
	ikey = 1;
      else if(imposeSymmetry == PGG)
	ikey = 8;
      else if(imposeSymmetry == CMM)
	ikey = 9;
      else if(imposeSymmetry == P4)
	ikey = 10;
      else if(imposeSymmetry == P4M)
	ikey = 11;
      else if(imposeSymmetry == P4G)
	ikey = 12;
      else if(imposeSymmetry == P6)
	ikey = 16;
      else if(imposeSymmetry == P6M)
	ikey = 17;
      else {
	sprintf(outStr, "Invalid Symmetry [%d]", imposeSymmetry);
	askThem(outStr, XmDIALOG_INFORMATION);
	goto noSymm;
      }

      thisAoverb = aoverbTemp;
      
      emsf_sym_rob__(&em_ndim, &sf_type, &nlim, &nsf, 
		    sFacs, &gamstr, &ikey, &thisAoverb);
      
      if(thisAoverb != aoverbTemp) { // Set it
	aoverbTemp = thisAoverb;
	TextSetFloat(aoverbTextW, aoverbTemp, FALSE);
      }
      
      if(map_abangTemp != (180.0 - gamstr)) { // Set it
	map_abangTemp = 180.0 - gamstr;
	TextSetFloat(map_abangTextW, map_abangTemp, FALSE);
	// Now setup map and ab scrolls and tryit
	crysDisp_cr_activate(0,0,0);
      }

    noSymm:;
      
    } 
  }
  
  if(writeFourAverSF) { // Write the sfs
    // Do we really need to clip the last zero ones ???
    // nsf = lastSF;
    
    // Let's write out a temporary PIF file to see what's going on
    sprintf(outFilename,"four_aver_sf.psf");
    
    pif_open(outFilename,&faHeader.luNum,"w");
    
    faHeader.nimg  = 1;
    thisMax = MAX(thisMax,fabs(thisMin));
    thisMax = MAX(thisMax,360.0);
    printf("thisMax[%f]\n",thisMax);
    faHeader.floatScaleFactor = (double)MAXINT / (double)thisMax;
    faHeader.recipFloatScaleFactor   = (1.0/faHeader.floatScaleFactor);
    faHeader.int4scaleF              = faHeader.floatScaleFactor;
    faHeader.numCols   = 2;  // We're 2D here
    faHeader.numRows   = nsf;
    faHeader.mode      = 31;
    
    printf("map_abang here - [%f]\n",map_abang);
    faHeader.map_abang = map_abang;
    faHeader.aoverb = aoverb;

    faHeader.borderType = getBorderType(featherMode);

    faHeader.borderWidth = featherRadius-boxRadius;
            
    pif_write_gh(&faHeader);
    faHeader.imgNum    = 1;
    pif_write_dh(&faHeader);
    pif_write_sf(&faHeader, sFacs);
    pif_close(&faHeader.luNum);
    
    sprintf(outStr,"Wrote structure factors to file [four_aver_sf.psf]");
    askThem(outStr, XmDIALOG_INFORMATION);
  }
  
  // Now convert Structure factors back into a 2D MAP
  fourAver_backit(FALSE);
  
} // end function fourAver_refine_impose



void fourAver_manage(Widget widget, XtPointer client_data, 
		     XtPointer call_data)
{
  // function name: fourAver_manage
  // Get the fourAver screen up
  
  int tag2;
  
  if(!averageFFT) { // Don't do for averaged FFT
    sprintf(outStr,"No symmetry imposed yet");
    sprintf(outStr,"%s",outStr);
    doLabel(outStr, imposeLabelW, FALSE);
    
    doLabel("No refinement yet", refineLabelW, FALSE);
    
    imposeSymmetry = refinePhaseOrigin = 0;
    
    // Generate gold standard structure factors
    fourAver_work();
    
    // Just give us a default view of it
    adim = bdim = 256;
    fourAver_backit(TRUE);
    
    raiseIt(mainWindowW);
    raiseIt(fourAverScreenW);
    
    // Set the zoom 
    TextSet(scaleXvalueW, "1.0", FALSE);  
    
    tag2 = TRUE;
    crysDisp_screen_activate(0, (XtPointer)&tag2, 0); 
  }
  
} // end function fourAver_manage



void fourAver_resetSfs(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: fourAver_resetSfs
  // Reset to last gold standard of SFs.
  
  int i;
  
  phaseOrigX = 0.0;
  phaseOrigY = 0.0;
  TextSet(moveOrigXW, "0.0", FALSE);
  TextSet(moveOrigYW, "0.0", FALSE);
  TextSet(shiftOrigXW, "", FALSE);
  TextSet(shiftOrigYW, "", FALSE);
  
  nsf = origNsf;  // We may have expanded this in the impose routines
  for(i=0;i<nsf;i++) {
    // Copy gold std into temp area
    sFacs[i].h     = fa_sFacs[i].h;
    sFacs[i].k     = fa_sFacs[i].k;
    sFacs[i].l     = fa_sFacs[i].l;
    sFacs[i].a_amp = fa_sFacs[i].a_amp;
    sFacs[i].b_phi = fa_sFacs[i].b_phi;
    sFacs[i].dstr  = fa_sFacs[i].dstr;
  }

  doLabel("No symmetry imposed yet", imposeLabelW, FALSE);
  doLabel("No refinement yet", refineLabelW, FALSE);
  
  // Reset map_abang and aoverb
  crysDisp_reset_activate(NULL, NULL, NULL);
  
  fourAver_backit(FALSE);
  
} // end function fourAver_resetSfs



void movePhaseOrigin_activate(Widget widget, XtPointer client_data, 
			      XtPointer call_data)
{
  // function name: movePhaseOrigin_activate
  // Move the concept of phase origin.
  
  int i;
  float moveX, moveY;
  
  if(deltaPhaseShift) { // Do relative
    moveY = wstrtod(shiftOrigYW);
    moveX = wstrtod(shiftOrigXW);
    
    phaseOrigX += moveX;
    phaseOrigY += moveY;
    TextSetFloat(moveOrigXW, phaseOrigX, FALSE);
    TextSetFloat(moveOrigYW, phaseOrigY, FALSE);
  } else {// Do absolute
    moveY = wstrtod(moveOrigYW);
    moveX = wstrtod(moveOrigXW);
    
    phaseOrigX = moveX;
    phaseOrigY = moveY;
    
    nsf = origNsf;  // We may have expanded this in the impose routines
    for(i=0; i<nsf; i++) {
      // Copy gold std into temp area
      sFacs[i].h     = fa_sFacs[i].h;
      sFacs[i].k     = fa_sFacs[i].k;
      sFacs[i].l     = fa_sFacs[i].l;
      sFacs[i].a_amp = fa_sFacs[i].a_amp;
      sFacs[i].b_phi = fa_sFacs[i].b_phi;
      sFacs[i].dstr  = fa_sFacs[i].dstr;
    }
  }

  // Shift the structure factors
  rob_sf_pshift__(sFacs, &nsf, &moveX, &moveY);

  // Init for next time
  TextSet(shiftOrigXW, "", FALSE); 
  TextSet(shiftOrigYW, "", FALSE); 
 
  fourAver_backit(FALSE);

} // end function movePhaseOrigin_activate



void impose_changed(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: impose_changed
  // New symmetry has been selected
  
  int tag = *((int *) client_data);
  imposeSymmetry = tag;

  if(tag == P2)
    sprintf(outStr, "P2 Imposed");
  else if(tag == PGG)
    sprintf(outStr, "PGG Imposed");
  else if(tag == CMM)
    sprintf(outStr, "CMM Imposed");
  else if(tag == P4)
    sprintf(outStr, "P4 Imposed");
  else if(tag == P4M)
    sprintf(outStr, "P4M Imposed");
  else if(tag == P4G)
    sprintf(outStr, "P4G Imposed");
  else if(tag == P6)
    sprintf(outStr, "P6 Imposed");
  else if(tag == P6M)
    sprintf(outStr, "P6M Imposed");
  else 
    sprintf(outStr,"No symmetry imposed yet");
  
  doLabel(outStr, imposeLabelW, FALSE);
  fourAver_refine_impose(2);

} // end function impose_changed



void refine_changed(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: refine_changed 
  // Changed the concept of folding.

  int tag = *((int *) client_data);
  refinePhaseOrigin = tag;

  if(tag == 1)
    sprintf(outStr, "2 Fold Refined");
  else if(tag == 2)
    sprintf(outStr, "3 Fold Refined");
  else if(tag == 3)
    sprintf(outStr, "4 Fold Refined");
  else if(tag == 4)
    sprintf(outStr, "6 Fold Refined");
  else 
    sprintf(outStr, "No refinement yet ");

  doLabel(outStr, refineLabelW, FALSE);  
  fourAver_refine_impose(1);

} // end function refine_changed



void rotflt_screen_activate(Widget widget, XtPointer client_data, 
			    XtPointer call_data)
{
  // function name: rotflt_screen_activate
  // Put the rotflt screen on the display.

  int thisMaxInt, thisMinInt,  sliderSize, i, idim1Calc;

  orig_file_width = file_width;
  rotfltSwitch    = TRUE;
  firstRotFlt     = TRUE;
  boxMethod       = USERECT;
  newOrigin       = TRUE;
  winScale        = 1.0;

  idim1Calc = 1;
  for (i=0; i<10; i++) {
    idim1Calc *= 2;
    if(purdueHeader.numCols > idim1Calc)
      idim1Rotflt = idim1Calc;
  }
  
  idim1Rotflt *= 2;
  idim1Rotflt = fftSize;
  
  if(fft_step <= 0.0)
    fft_step = 1.0;
  
  thisMaxInt = (int)(((idim1Rotflt / 2) - 2)/fft_step);
  thisMinInt = (int)(1);
  sliderSize = 5;
  nannuli = (int)(idim1Rotflt / 4);
  XtSetArg(arglst[0], XmNmaximum, (int)(thisMaxInt+(float)sliderSize));
  XtSetArg(arglst[1], XmNminimum, thisMinInt);
  XtSetArg(arglst[2], XmNsliderSize, sliderSize);
  XtSetArg(arglst[3], XmNvalue, nannuli);

  if(thisMaxInt > 200)
    XtSetArg(arglst[4], XmNpageIncrement, (int)(thisMaxInt/100));
  else if(thisMaxInt > 20)
    XtSetArg(arglst[4], XmNpageIncrement, (int)(thisMaxInt/10));
  else 
    XtSetArg(arglst[4], XmNpageIncrement, (int)(thisMaxInt));

  XtSetValues(nannuliScrollW, arglst, 5);

  // Calc max symmetry Limit
  // Don't set lg_end variable here - it's done on the interface

  thisMaxInt = (3.1415 * purdueHeader.numCols) * 0.5;
  XtSetArg(arglst[0], XmNmaximum, (int)(thisMaxInt+(float)sliderSize));
  XtSetArg(arglst[1], XmNsliderSize, sliderSize);

  if(thisMaxInt > 200)
    XtSetArg(arglst[2], XmNpageIncrement, (int)(thisMaxInt/10));
  else if(thisMaxInt > 20)
    XtSetArg(arglst[2], XmNpageIncrement, (int)(thisMaxInt/10));
  else 
    XtSetArg(arglst[2], XmNpageIncrement, (int)(thisMaxInt/10));

  XtSetValues(lgendScrollW, arglst, 3);
  
  sprintf(outStr,"%.2f",fft_step);
  TextSet(fftStepTextW, outStr, FALSE);
  
  omega = 0.0;
  XtVaSetValues(phi0ScrollW, XmNvalue, (int)(omega), NULL);
  
  iopt[0]= 1;      // iopt[0] == 1 --> Refine symmetry axis
  iopt[1]= 0;
  iopt[2]= 0;      // iopt[2] == 1 --> Plot power spectrum
  iopt[3]= 0;
  
  raiseIt(rotfltScreenW);
  rotflt_tryit_activate(0,0,0);
  
} // end function rotflt_screen_activate



void rotflt_write_activate(Widget widget, XtPointer client_data, 
			   XtPointer call_data)
{
  // function name: rotflt_write_activate
  // Write the rotflt image to disk

  int i, j, ns, maxI, minI, radiusM1M1, scaleF;
  forPifHead_t thisHead;
  fString_t rWrite, forName;
  BYTE *outBuf, *tmpByte_buffer;
  FILE *fOut;

  CK(outBuf = (BYTE *)malloc(sizeof(BYTE)*numCols*numRows));
  CK(tmpByte_buffer = (BYTE *)malloc(sizeof(BYTE)*numCols*numRows));
  pif_init_head(&thisHead);
  thisHead.luNum = 4;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 4;
  forName.addr = malloc(sizeof(char)*64);
  forName.length = 64;
  sprintf(rWrite.addr,"w");
  sprintf(forName.addr,"rotflt_fimg.pif");
  
  pif_open(forName.addr, &thisHead.luNum, rWrite.addr);
  
  // Pick appropriate statistics calculation based on data type we have
  thisHead.numCols = numCols;
  thisHead.numRows = numRows;
  thisHead.numSecs = 1;
  
  thisHead.mode = 6;
  
  ns = 1;
  radiusM1M1=  (2 * radiusM1) - 1;
  calcMapStats(rho, &thisHead.img_bkg, 
	       &thisHead.datar4_min, &thisHead.datar4_max, 
	       &thisHead.datar4_avg, &thisHead.datar4_std,
	       &radiusM1M1, &radiusM1M1, &ns);
  thisHead.datar4_max  += 1.0;
  thisHead.datar4_min  -= 1.0;
  scaleF = 225.0 / (thisHead.datar4_max - thisHead.datar4_min);
  maxI = minI = 128;

  for (i=0; i<numRows; i++)
    for (j=0; j<numCols; j++) {
      outBuf[i*numCols + j] = (BYTE)((rho[i*radiusM1M1 + j] - 
					thisHead.datar4_min)*scaleF);
      minI = MIN(minI,outBuf[i*numCols + j]);
      maxI = MAX(maxI,outBuf[i*numCols + j]);
    }
  
  calcByteStats(outBuf, &thisHead.img_bkg, 
		&thisHead.datar4_min, &thisHead.datar4_max, 
		&thisHead.datar4_avg, &thisHead.datar4_std,
		numCols, numRows, numCols);
  thisHead.datai4_max = thisHead.datar4_max;
  thisHead.datai4_min = thisHead.datar4_min;
  
  pif_write_gh(&thisHead);
  
  thisHead.imgNum = 1;
  
  pif_write_dh(&thisHead);
  thisHead.imgOffset += 512;  // Just wrote the header
  
  tsbworld_byte(outBuf, tmpByte_buffer, numCols, numCols, numRows);
  pif_write_byte_image_(&thisHead.luNum, tmpByte_buffer,
			&thisHead.imgOffset, &thisHead.numCols,
			&thisHead.numRows, &thisHead.numSecs);
  
  free(forName.addr);
  free(rWrite.addr);
  pif_close(&thisHead.luNum);

  fOut = fopen("rotflt_fimg.raw","w");
  fwrite(outBuf, sizeof(BYTE),(numRows*numCols),fOut);
  fclose(fOut);

  sprintf(outStr,"RAW file stats:\n\
\n\
     Filename= [rotflt_fimg.raw]\n\
\n\
     Frames= [%d]\n\
     Width = [%d]\n\
     Height= [%d]\n\
\n\
     8 bit byte pixels\n\
     NO header\n\
\n\
     Min[%d]\n\
     Max[%d]\n",1,(numCols),numRows,
	  minI,maxI);

  askThem(outStr, XmDIALOG_INFORMATION);
  
  askThem("File rotflt_fimg.pif written to disk.", XmDIALOG_INFORMATION);
  
  free(outBuf);
  free(tmpByte_buffer);
  
} // end function rotflt_write_activate



void DrawROTFLTWindow(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: DrawROTFLTWindow
  // Redraw window on expose events or whenever it's prudent.

  int drawX, drawY, i;
  int lastVal, lastX, lastXaccum;
  float thisMin, thisMax, scaleFactor;
  char numStr[128];

  if(XtIsManaged(rotfltScreenW))
    XPutImage(dsply, rotflt_window1, rotflt_gc, rotflt_image1, 0, 0,
	      0, 0, 512, 512);
  XFlush (dsply);

  // Now draw axes
  XDrawLine(dsply, XtWindow(rotfltDrawW), image_gcWHITE,
	    25, 0, 25, 512);
  XDrawLine(dsply, XtWindow(rotfltDrawW), image_gcWHITE,
	    0, 485, 512, 485);

  if(logROTFLTplot) { // Draw markers for log plot
    XDrawString(dsply, XtWindow(rotfltDrawW), image_gcRED,
		1, 10, "100", 3);
    XFlush (dsply);
    for (i=0; i<yFactor; i++) { // Horizontal grid
      int offset;
      offset = 475 / yFactor;
      XDrawLine(dsply, XtWindow(rotfltDrawW), image_gcGREY[150],
		25, ((i+1)*offset), 512, ((i+1)*offset));
      drawY = (i+1)*offset;
      sprintf(numStr,"%.2f",(float)(100.0 / (pow(10.0,(i+1)))));
      XDrawString(dsply, XtWindow(rotfltDrawW), image_gcRED,
		  1, drawY, numStr, strlen(numStr));
      XFlush (dsply);
    }
  } else { // Draw markers for linear
    for (i=0; i<4; i++) { // Horizontal grid
      int offset;
      float maxVal, tickVal;
      maxVal  =  100.0 / yFactor;
      tickVal = maxVal / 5.0;
      sprintf(numStr,"%.1f",maxVal);
      XDrawString(dsply, XtWindow(rotfltDrawW), image_gcRED,
		  1, 10, numStr, strlen(numStr));
      XFlush (dsply);
      offset = 475 / 5;
      XDrawLine(dsply, XtWindow(rotfltDrawW), image_gcGREY[150],
		25,((i+1)*offset),512,((i+1)*offset));
      drawY = (i+1)*offset;
      sprintf(numStr,"%.1f",(float)((maxVal-tickVal)-((float)i*tickVal)));
      XDrawString(dsply, XtWindow(rotfltDrawW), image_gcRED,
		  1, drawY, numStr, strlen(numStr));
      XFlush (dsply);
    }
  }

  // Vertical grid
  for (i=0; i<(radiusM1*2*5); i+=5) {
    int yp1;

    yp1 = (i%25 == 0) ? 490 : 485;
    if(i%50 == 0)
      XDrawLine(dsply, XtWindow(rotfltDrawW), image_gcGREY[150],
		i+26, 0, i+26, yp1);
    else if((i%25)==0)
      XDrawLine(dsply, XtWindow(rotfltDrawW), image_gcGREY[100],
		i+26, 0, i+26, yp1);
  }
  
  // Draw tick mark numbers
  for(i=0;i<(radiusM1*2*5);i+=5) {
    if(i%50 == 0) { // Do every tenth
      drawX = 20+i;
      sprintf(numStr,"%2d",(int)(xFactor*(float)((float)i/5.0)));
      XDrawString(dsply, XtWindow(rotfltDrawW), image_gcRED,
		  drawX, 500, numStr, strlen(numStr));
      XFlush (dsply);
    }
  }

  XFlush (dsply);
  
  // Draw power spectrum
  thisMin = HUGE;
  thisMax = -HUGE;

  for (i=0; i<(radiusM1*2)+1; i++) {
    if(logROTFLTplot) { // log
      if(powSpec[i] < 0.001)
	powSpec[i]= 0.001;
      thisMin = MIN(log(powSpec[i]),thisMin);
      thisMax = MAX(log(powSpec[i]),thisMax);
    } else { // linear
      thisMin = MIN(powSpec[i],thisMin);
      thisMax = MAX(powSpec[i],thisMax);
    }
  }

  if(logROTFLTplot && (logCutoff > 0.0))
    thisMin = log(logCutoff);
  
  scaleFactor = (thisMax > thisMin) ? 400.0/(thisMax-thisMin) : 1.0;
  
  if(sump1 > 0.0) { // Don't do unless we've got a power spectrum
    sprintf(numStr,"%.2f%% of azimuthal power is %3d-fold.",
	    sump3, nrot);
    XDrawString(dsply, XtWindow(rotfltDrawW), image_gcGREEN,
		100, 10, numStr, strlen(numStr));

    sprintf(numStr,"This plus G-zero is %6.2f%% of total power.", 
	    sump2);
    XDrawString(dsply, XtWindow(rotfltDrawW), image_gcGREEN,
		100, 25, numStr, strlen(numStr));

    sprintf(numStr,"Threshold of noise = %.2f%%", ((1.0/nrot)*100.0));
    XDrawString(dsply, XtWindow(rotfltDrawW), image_gcGREEN,
		100, 40, numStr, strlen(numStr));
  }

  lastVal = 99;
  lastX = -5;
  lastXaccum = 10;

  for(i=0; i<(radiusM1*2*5); i+=5) {
    int y, yp1, thisPlotVal, nextVal;
    float x, xp1;
    thisPlotVal = (int)(xFactor * (float)((float)i/5.0));
    nextVal = (int)xFactor;
    if(xFactor < 1.0) { // Do special
      if((int)(xFactor * (float)((float)(i+5)/5.0)) > lastVal)
	nextVal = 1;
    }
    lastVal = thisPlotVal + nextVal;
    if(nextVal > 0) { // For xfactors < 1 this will prevent multiple draws
      if(logROTFLTplot) { // log
	x   = MAX(log(powSpec[thisPlotVal]),thisMin);
	xp1 = MAX(log(powSpec[thisPlotVal+((int)nextVal)]),thisMin);
	y   = (int)475-(((x - thisMin)*scaleFactor));
	yp1 = (int)475-(((xp1 - thisMin)*scaleFactor));
      } else { // linear
	x   = powSpec[thisPlotVal];
	xp1 = powSpec[thisPlotVal+((int)nextVal)];
	y   = (int)475-(yFactor*((x - thisMin)*scaleFactor));
	yp1 = (int)475-(yFactor*((xp1 - thisMin)*scaleFactor));
      }
      XDrawLine(dsply, XtWindow(rotfltDrawW), image_rotfltgcRED,
		((lastX)+26+5), (int)y, ((lastX)+26+lastXaccum), (int)yp1);
      XDrawLine(dsply, XtWindow(rotfltDrawW), image_gcBLUE,
		0, 0, 0, 1);
      lastX = i;
      lastXaccum = 5;
    }
    lastXaccum +=5;
  }

} // end function DrawROTFLTWindow



void rotfltClick(Widget widget, XtPointer client_data, 
		 XtPointer call_data)
{
  // function name: rotfltClick
  // Every time the mouse button is clicked call here.  

  char ans;
  int x, y;
  XEvent event;
  XmDrawingAreaCallbackStruct *draw_data = 
    (XmDrawingAreaCallbackStruct *) call_data;

  if(draw_data == 0)
    return;
  
  event= *(draw_data->event);
  
  switch (event.type) {
    
  case ButtonRelease:
    x = event.xbutton.x;      // Get x,y copords of point pos
    y = event.xbutton.y;
    break;
    
  case ButtonPress:
    x = event.xbutton.x;      // Get x,y copords of point pos
    y = event.xbutton.y;
    printf("[%d,%d] clicked\n", x, y);
    break;
    
  case KeyPress:
    ans= XLookupKeysym((XKeyEvent *)&event, 0);
    printf ("Keyed[%c]\n", ans);
    break;
    
  default:
    break;
  }
  
  XFlush(dsply);    // Make changes current
  
} // end function rotfltClick



void fft_step_changed(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{
  // function name: fft_step_changed
  // Call here when fft_step text widget gets a carriage return event

  int thisMaxInt, thisMinInt, sliderSize;

  if(widget == fftStepTextW) {
    
    fft_step = wstrtod(widget);
    if(fft_step < 0.0) {
      askThem("Cannot have an FFT step less than or equal to zero",
	      XmDIALOG_INFORMATION);
      fft_step = fabs(fft_step);
      sprintf(outStr,"%.2f",fft_step);
      TextSet(fftStepTextW, outStr, TRUE);
    }
    else if (fft_step == 0.0) {
      askThem("Cannot have an FFT step equal to zero\nReset to 1.0",
	      XmDIALOG_INFORMATION);
      fft_step = 1.0;
      sprintf(outStr, "%.2f", fft_step);
      TextSet(fftStepTextW, outStr, TRUE);
    }
    
    // We should be able to bump up the nannuli 
    // if we've got a fft_step less than 1.0 (or
    // also we should have to reduce nannuli if
    // we've got an fft_step > 1.0.  The fortran
    // code doesn't handle this very well yet so let's
    // just not use this until we can get emrotflt_main()
    // to do the right thing.
    // nannuli = MIN(nannuli,thisMaxInt);
    
    thisMaxInt = (int)(idim1Rotflt / 2) - 1;
    thisMaxInt = (int)((float)thisMaxInt / fft_step)-1;
    thisMinInt = 1;
    sliderSize = 5;
    nannuli = MIN(nannuli,thisMaxInt);
    XtSetArg(arglst[0], XmNmaximum, (int)(thisMaxInt+(float)sliderSize));
    XtSetArg(arglst[1], XmNminimum, thisMinInt);
    XtSetArg(arglst[2], XmNsliderSize, sliderSize);
    XtSetArg(arglst[3], XmNvalue, nannuli);

    if(thisMaxInt > 200)
      XtSetArg(arglst[4], XmNpageIncrement, (int)(thisMaxInt/100));
    else if(thisMaxInt > 20)
      XtSetArg(arglst[4], XmNpageIncrement, (int)(thisMaxInt/10));
    else 
      XtSetArg(arglst[4], XmNpageIncrement, thisMaxInt);

    XtSetValues(nannuliScrollW, arglst, 5);
    
    sprintf(outStr,"%d",nannuli);
    doLabel(outStr, nannuliLabelW, FALSE);
  }
  else if(widget==xorigTextW) {
    // Do nothing
  }
  else if(widget==yorigTextW) {
    // Do nothing
  }
  
  // if(backTransROTFLT)
  rotflt_tryit_activate(0,0,0);
  
} // end function fft_step_changed



void rotfltArrowActivate(Widget widget, XEvent *event)
{
  // function name: rotfltArrowActivate
  // Controls behavior of arrow on rotational filtering screen

  float speedFactor;
  speedFactor = (float) GetButtonVal(event->xbutton.button, 100, 10, 1) / 100.;

  if(widget == xorigUpArrowW || widget == xorigDownArrowW) {
    if(widget == xorigUpArrowW)
      sprintf(outStr,"%.2f", wstrtod(xorigTextW) + speedFactor);
    else
      sprintf(outStr,"%.2f", wstrtod(xorigTextW) - speedFactor);
    TextSet(xorigTextW, outStr, TRUE);
  }
  else if(widget == yorigUpArrowW || widget == yorigDownArrowW) {
    if(widget == yorigUpArrowW)
      sprintf(outStr,"%.2f", wstrtod(yorigTextW) + speedFactor);
    else
      sprintf(outStr,"%.2f", wstrtod(yorigTextW) - speedFactor);
    TextSet(yorigTextW,outStr, TRUE);
  }
  else 
    printf("Unknown widget value.\n");

  rotflt_tryit_activate(0,0,0);

} // end function rotfltArrowActivate



void rotflt_forceTryit_activate(Widget widget, XtPointer client_data, 
				XtPointer call_data)
{
  // function name: rotflt_forceTryit_activate
  // Try to Rotationally Filter this FFT.

  int saveVal;

  saveVal = backTransROTFLT;
  backTransROTFLT = TRUE;  // Force it
  rotflt_tryit_activate(0,0,0);
  backTransROTFLT = saveVal;

} // end function rotflt_forceTryit_activate



void rotflt_tryit_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{
  // function name: rotflt_tryit_activate
  // Try to Rotationally Filter this FFT.
  // RSS the loops in this routine are a mess!

  int i, irad, max_bo, radiusM1M1, lg_start, numPixels;
  int y, m, l, n, thisInt;
  float thisMin, thisMax, scaleFactor;
  float del_br, del_lr, twodd, fft_origx, fft_origy;
  float pow_scale, xorigSave,yorigSave;
  float elipX=1.0, elipY=1.0, ascal=1.0, bscal=1.0;
  float *thisBuffer;

  // Get these from interface now 
  // nrot = 5;                    
  // nannuli = 30;        
  // fft_step = 1.0;
  // phi0= 0.0;
  //  jzero= 0;

  // Give them feed_back on what it actually is
  sprintf(outStr,"%.2f",fft_step);
  TextSet(fftStepTextW, outStr, TRUE);
  
  phi0 = omega;   // What was gotten from the interface?

  radiusM1 = purdueHeader.numCols/2;

  // Get del_lr from scroll eventually
  del_lr = 1.0;    // Don't Try and calc in rotflt_info()
  del_br = 1.0;    // Already calculated in rotflt_info()

  // Only init once!!!!  Use last xorig and origY each successive time!!!
  // oldway:  if(iopt[0]==1)
  
  if(newOrigin) { // Recalc the origin
    xorig = (purdueHeader.numCols/2);   // FFT origx and origy
    yorig = (purdueHeader.numCols/2);      
    sprintf(outStr,"%.2f",xorig);
    TextSet(xorigTextW, outStr, FALSE);
    sprintf(outStr,"%.2f",yorig);
    TextSet(yorigTextW, outStr, FALSE);
    newOrigin = FALSE;
    newStart = TRUE;
  } else { // Get from last time
    newStart = FALSE;
    xorig = wstrtod(xorigTextW);
    xorig -= 1.0;  // Xorig is one off
    yorig = wstrtod(yorigTextW);
    // Let's put it in tsb space
    yorig = purdueHeader.numRows - yorig;
  }

  pow_scale = 100.0;

  lg_start = 0;         // Calculated in rotflt_info()
  // lg_end   = 0;      // Symmetry Limit

  // Doesn't matter what it's set to - recalculated in rotflt_info_rob()
  max_bo = 0;                   
  
  if(fileType == PACKEDPORTABLE) {
    thisBuffer= (float *)packImg[(imgNum-1)]; 
    numCols = purdueHeader.numCols;
    numRows = purdueHeader.numRows;
  } else { // Scanned
    thisBuffer= (float *)inCore_buffer;
    numCols = file_width;
    numRows = file_height;
  }

  // *******************************************************
  // RHO is defined as: 
  //
  //     REAL*4  RHO(2*NLR-1,2*NLR-1)
  //
  // in robemF.f so make sure we've got enough allocated here to
  // allow for any discrepancies 
  // *********************************************************

  freeNullFloat(&rho);
  CK(rho= (float *)malloc(sizeof(float)*(numCols)*(numRows)*2));
  
  calcSizeFactor(zoomFactor); // Calc zoom

  init_floatarr(powSpec, 1000, 0.0); // initialize array

  // These may get corrupted
  xorigSave = xorig;
  yorigSave = yorig;

  emrotflt_main__(powSpec, &elipX, &elipY, &irad,
		 &idim1Rotflt, &phi0, &ascal, &bscal,
		 &twodd, &jzero, &radiusM1, &max_bo,
		 &fft_step, &nannuli, &del_br, rho, &del_lr,
		 iopt, &lg_start, &lg_end, &xorig, &yorig,
		 &pow_scale, &nrot, &numCols, &numRows,
		 &sump1, &sump2, &sump3,
		 &fft_origx, &fft_origy, &winScale, sigma,
		 fftImage_buffer2, 
		 &newStart, &sigMax, &sigMaxTot);
  
  sprintf(outStr,"%.3f ",winScale);
  doLabel(outStr, refineResetLabelW, FALSE);
  
  // Give feedback on nannuli
  XtVaSetValues(nannuliScrollW, XmNvalue, nannuli, NULL);
  sprintf(outStr,"%d",(nannuli));
  doLabel(outStr, nannuliLabelW, FALSE);
  
  if(iopt[0] == 1) { // origin refined
    // Put it back in XWindows space (in TSB space above)
    yorig = purdueHeader.numRows - yorig;
    // Why do this here ???
    xorig = xorig + (1.0+fft_origx);
    // It's in TSB space
    yorig = yorig - fft_origy;
    sprintf(outStr,"%.2f",xorig);
    TextSet(xorigTextW, outStr, FALSE);
    sprintf(outStr,"%.2f",yorig);
    TextSet(yorigTextW, outStr, FALSE);
  } else { // Just use our originals
    xorig = xorigSave;
    yorig = yorigSave;
    // Put it back in tsb space
    yorig = purdueHeader.numRows - yorig;
  }
  
  firstRotFlt = FALSE;
  
  radiusM1M1=  (2 * radiusM1) - 1;
  file_width = orig_file_width + radiusM1M1;

  farray_minmax(rho, &thisMin, &thisMax, radiusM1M1*radiusM1M1);

  if(thisMax > thisMin)
    scaleFactor = (tryColors-14.0)/(thisMax-thisMin);
  else 
    scaleFactor = 1.0;

  if(scaleFactor < minRotfltScaleF)
    minRotfltScaleF = scaleFactor;
  if(!stretchRotflt)  
    scaleFactor = minRotfltScaleF;
  
  // Calc zoom
  calcSizeFactor(zoomFactor);
  
  if(backTransROTFLT)  { // Put rotflt image beside original

    // Redraw original in case it's got a crosshair on it or etc...
    virtual_read_activate(-99,-99);
    
    // 1 Put rotflt image beside original here

    for (y=0; y<radiusM1M1; y++) { // Do lines	  
      if(sizeFactor > 1.0) { // Zooming out
	for (m=0; m<(int)sizeFactor; m++) { // Do all zoomed lines
	  for (l=0; l<radiusM1M1; l++) { // Do all cols
	    for (n=0; n<(int)sizeFactor; n++) {
	      XPutPixel(image_image1,
			(radiusM1M1*sizeFactor)+
			(l*sizeFactor)+n, 
			(int)(y*sizeFactor)+m,
			pixels_return[(int)
				      ((rho[((y*radiusM1M1)+l)]-
					thisMin)*scaleFactor)+6]);
	    }
	  }
	}
      }
      else { // We are zooming IN
	for (l=0; l<radiusM1M1; l+=(int)recipSizeFactor) { // Do all cols
	  thisInt = 0;
	  for (n=0; n<(int)recipSizeFactor; n++) {
	    thisInt += (int)((rho[((y*radiusM1M1)+l+n)]-
			      thisMin)*scaleFactor);
	  }
	  XPutPixel(image_image1,
		    (l*sizeFactor)+(radiusM1M1*sizeFactor), 
		    (int)(y*sizeFactor),
		    pixels_return[(int)(thisInt/
					(int)recipSizeFactor)+6]);
	}
      }
    }
  }
  
  if(rotfltResiduals) { // Put up contour plot of residuals
    farray_minmax(sigma, &thisMin, &thisMax, 17*17);
    if((thisMax-thisMin)==0.0)
      scaleFactor = 1.0;
    else 
      scaleFactor = (tryColors-9) / (thisMax -thisMin);
    numPixels = (int)((radiusM1M1 *sizeFactor) / 17);
    
    for (y=0; y<(17*numPixels); y+=numPixels) { // Do lines
      for (m=0; m<(17*numPixels); m+=numPixels) {// Do all zoomed lines
	for (l=0; l<numPixels; l++) { // y
	  for (n=0; n<numPixels; n++) {
	    XPutPixel(image_image1,
		      (radiusM1M1*sizeFactor) + 
		      (radiusM1M1*sizeFactor)+m+n, 
		      (int)(y+l),
		      pixels_return[(int)((int)(
						((sigma[(((y/numPixels)*17)+
							 (m/numPixels))] - 
						  thisMin) * scaleFactor))
					  +6)]);
	  }
	}
      }
    }
  }
 
  if(rotfltCrosshair) { // Draw crosshair in image
    float y,x;
    
    y = (float)numRows - (float)yorig;  
    x = (float)xorig;
    
    if(sizeFactor > 1.0) { // We are zooming out
      for (m=0; m<(int)sizeFactor; m++) { // Do all zoomed lines
	for (l=0; l<radiusM1M1; l++) // Do all cols
	  for (n=0; n<(int)sizeFactor; n++)
	    XPutPixel(image_image1, (int)(l*sizeFactor)+n, 
		      (int)(y*sizeFactor)+m, pixels_return[RED]);
	for (l=0; l<radiusM1M1; l++) // Do all rows
	  for (n=0; n<(int)sizeFactor; n++) 
	    XPutPixel(image_image1, (int)(x*sizeFactor)+n, 
		      (int)(l*sizeFactor)+m, pixels_return[RED]);
      }
    }
    else { // We are zooming IN
      for (l=0; l<radiusM1M1; l+=(int)recipSizeFactor) // Do all cols
	XPutPixel(image_image1, l*sizeFactor, 
		  (int)(y*sizeFactor), pixels_return[RED]);
      for (l=0; l<radiusM1M1; l+=(int)recipSizeFactor) //Do all rows
	XPutPixel(image_image1, xorig*sizeFactor, 
		  (int)(l*sizeFactor), pixels_return[RED]);
    }
    
  }

  // We may not want to refine the origin next time ...
  iopt[0] = 0;  // Make them force it again
  
  XPutImage (dsply, image_window1, image_gc, image_image1, 0, 0,
	     0, 0, (file_width*sizeFactor), (file_height*sizeFactor));
  XFlush (dsply);
  DrawWindow(); 
  
  if(widget != (Widget)-99)
    DrawROTFLTWindow(NULL, NULL, NULL);
  
} // end function rotflt_tryit_activate



void rotflt_change(Widget widget, XtPointer client_data, 
		   XtPointer call_data)
{
  // function name: rotflt_change
  // Whenever rotflt scroll widgets changed call here.

  char outStr[64];
  XmScrollBarCallbackStruct *temp = (XmScrollBarCallbackStruct *) call_data;

  if(!inited) return;
  
  if(widget == nannuliScrollW) {
    nannuli = MAX(temp->value, 1);
    sprintf(outStr,"%d",(nannuli));
    doLabel(outStr, nannuliLabelW, FALSE);
    if((nannuli*fft_step) >= (idim1Rotflt/2)) {// Reset fft_step 
      printf("Resetting fft_step\n");
      fft_step = (float)(idim1Rotflt/2)/(float)(nannuli+1); // +1 to make sure 
      sprintf(outStr,"%.2f",fft_step);
      TextSet(fftStepTextW, outStr, TRUE);
    }
    // if(backTransROTFLT) 
    rotflt_tryit_activate(0,0,0);
  }
  else if(widget == lgendScrollW) {
    lg_end = (int)temp->value;
    if(lg_end < (nrot)) {
      lg_end = nrot;
      XtVaSetValues(lgendScrollW, XmNvalue, lg_end, NULL);
      XFlush(dsply);
    }
    sprintf(outStr,"%d",lg_end);
    doLabel(outStr, lgendLabelW, FALSE);
    //if(backTransROTFLT) 
    rotflt_tryit_activate(0,0,0);
  }
  else if(widget == nrotScrollW) {
    nrot = (int)temp->value;
    sprintf(outStr,"%d",nrot);
    doLabel(outStr, nrotLabelW, FALSE);
    if(lg_end < nrot) {
      lg_end = nrot;
      XtVaSetValues(lgendScrollW, XmNvalue, lg_end, NULL);
      XFlush(dsply);
      sprintf(outStr,"%d",(lg_end));
      doLabel(outStr, lgendLabelW, FALSE);
    }
    //if(backTransROTFLT) 
    rotflt_tryit_activate(0,0,0);
  }
  else if (widget == xscaleScrollW) {
    xFactor = (float)temp->value;
    if(xFactor < 1.0) {// Do it in 0.0 - 1.0 range 
      float tmpF;
      tmpF = 0.9 - ((0.0 - xFactor)/10.0);
      tmpF = MAX(tmpF,0.1);
      xFactor = tmpF;
    }
    DrawROTFLTWindow(NULL, NULL, NULL);
  }
  else if (widget == yscaleScrollW) {
    yFactor = (float)temp->value;
    if(logROTFLTplot) {// Now setup yscale scroll and logCutoff 
      logCutoff = (10.0/ pow(10.0,(yFactor-1)));
      saveLogYFactor = yFactor;
    }
    else {// Linear 
      if(yFactor < 1.0) {// Do it in 0.0 - 1.0 range 
	float tmpF;
	tmpF = 0.9 - ((0.0 - yFactor)/10.0);
	tmpF = MAX(tmpF,0.1);
	yFactor = tmpF;
      }
      saveLinYFactor = yFactor;
    }
    DrawROTFLTWindow(NULL, NULL, NULL);
  }
  else if (widget == jZerotoggleW) {
    jzero = (jzero == 0) ? 1 : 0;
    rotflt_tryit_activate(0,0,0);
  }
  else if (widget == phi0ScrollW) {
    omega = (float)temp->value;
    sprintf(outStr,"%.0f",(omega));
    doLabel(outStr, phi0LabelW, FALSE);
    // if(backTransROTFLT) 
    rotflt_tryit_activate((Widget)-99,0,0);
    // DrawROTFLTWindow(NULL, NULL, NULL); 
  }
  else if (widget == logROTFLTtoggleW) {
    logROTFLTplot = TRUE;
    logCutoff = 10.0;
    yFactor= saveLogYFactor;
    logCutoff = (10.0/ pow(10.0,(yFactor-1)));
    XtSetArg(arglst[0], XmNmaximum, (5+1));
    XtSetArg(arglst[1], XmNminimum, 1);
    XtSetArg(arglst[2], XmNsliderSize, 1);
    XtSetArg(arglst[3], XmNvalue, (int)(yFactor));
    XtSetArg(arglst[4], XmNpageIncrement, 1);
    XtSetValues(yscaleScrollW, arglst, 5);
    XFlush(dsply);
    XtVaSetValues(logROTFLTtoggleW, XmNset, TRUE, NULL);
    XtVaSetValues(linROTFLTtoggleW, XmNset, FALSE, NULL);
    XFlush(dsply);
    DrawROTFLTWindow(NULL, NULL, NULL); 
  }
  else if (widget == linROTFLTtoggleW) {
    logROTFLTplot = FALSE;
    yFactor= saveLinYFactor;
    XtSetArg(arglst[0], XmNmaximum, (100+5));
    XtSetArg(arglst[1], XmNminimum, 1);
    XtSetArg(arglst[2], XmNsliderSize, 5);
    XtSetArg(arglst[3], XmNvalue, (int)(yFactor));
    XtSetArg(arglst[4], XmNpageIncrement, 5);
    XtSetValues(yscaleScrollW, arglst, 5);
    XFlush(dsply);
    XtVaSetValues(logROTFLTtoggleW, XmNset, FALSE, NULL);
    XtVaSetValues(linROTFLTtoggleW, XmNset, TRUE, NULL);
    XFlush(dsply);
    DrawROTFLTWindow(NULL, NULL, NULL); 
  }
  else if (widget == rotfltStretchW) {
    stretchRotflt = !stretchRotflt;
    rotflt_tryit_activate(0,0,0);
  }
  else if (widget == backTtoggleW) {
    backTransROTFLT= !backTransROTFLT;
    if(backTransROTFLT) // Turn button to insenitive 
      XtVaSetValues(tryitButtonW, XmNsensitive, FALSE, NULL); 
    else // Turn button sensitive 
      XtVaSetValues(tryitButtonW, XmNsensitive, TRUE, NULL); 
    // if(backTransROTFLT) 
    rotflt_tryit_activate(0,0,0);
  }
  else if (widget == refineButtonW) {
    XtVaSetValues(refineButtonW, XmNset, FALSE, NULL); 
    iopt[0]= 1;
    rotflt_tryit_activate(0,0,0);
  }
  else if (widget == rotfltXhairW)
    rotfltCrosshair = !rotfltCrosshair;
  else if (widget == rotfltContW)
    rotfltResiduals = !rotfltResiduals;
  else if (widget == refineResetButtonW) {
    winScale = 1.0;
    sprintf(outStr,"%.3f ",winScale);
    doLabel(outStr, refineResetLabelW, FALSE);
  }
  
} // end function rotflt_change



void cDisp_zoom_activate(Widget widget, XEvent *event)
{
  // function name: cDisp_zoom_activate
  // change the current zoom factor.

  float thisF, speedFactor;

  if(!inited) return;

  speedFactor = (float) GetButtonVal(event->xbutton.button, -1, 5, 1);
  thisF = wstrtod(scaleXvalueW);

  if(speedFactor <= 0.0) { // Button 1 
    if(widget == cDispZoomArrowUpW)
      thisF *= 2.0;
    else if(widget == cDispZoomArrowDownW)
      thisF *= 0.5;
    
    // Now give us a clean power of two 
    thisF = pow(2.0,(double)((int)rwalog2(thisF)));
  }
  else { // Buttons 2 or 3
    if(widget == cDispZoomArrowUpW)
      thisF += (0.01*speedFactor);
    else if(widget == cDispZoomArrowDownW)
      thisF -= (0.01*speedFactor);
  }

  if(thisF != 0.0) {
    x_size = (int)((float)thisF * (float)cdNumCols); 
    sprintf(outStr,"%.2f",thisF);
    TextSet(scaleXvalueW, outStr, TRUE);
  }
  
  // Now redraw with new zoominess
  crysDisp_cr_activate(0,0,0);

} // end function cDisp_zoom_activate



void cDispSymm_changed(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: cDispSymm_changed
  //
  // Callback that happens whenever the user makes a choice from the
  // "Plane Group Symmetry" menu on the Crystal Display screen. All
  // this does is call your routine with the appropriate args and then
  // tries to redraw the crystal.
  //
  // phaseGroupSymmetry == 1    ->   p1
  // phaseGroupSymmetry == 2    ->   p2
  // phaseGroupSymmetry == 3    ->   pm
  // phaseGroupSymmetry == 4    ->   pg

  int phaseGroupSymmetry = *((int *) client_data);
  float origX, origY;

  origY = wstrtod(cDispSymmOrigYW);
  origX = wstrtod(cDispSymmOrigXW);
    
  ricardo_routine(phaseGroupSymmetry, &crys_float_image_buffer,
		  &cdNumCols, &cdNumRows, origX, origY);
  
  crysDisp_tryit_activate(0,0,0);  // Now tryit
  
} // end function cDispSymm_changed



void crysDisp_screen_activate(Widget widget, XtPointer client_data, 
			      XtPointer call_data)
{
  // function name: crysDisp_screen_activate
  // Put the crysDispscreen on the display.

  int i, j;
  static int oldXsize=-99;
  int tag = *((int *) client_data);

  crysDisping = TRUE;

  if(fileType==PACKEDPORTABLE && !fourierAveragedCrysDisp) {
    // Do a little farbling here
    numCols   = purdueHeader.numCols;
    numRows   = purdueHeader.numRows;
    cdNumCols = purdueHeader.numCols;
    cdNumRows = purdueHeader.numRows;
    cdNumSecs = 1;
    // We've got a bug whereby these are bogus
    // aoverb = 1.0;
    // map_abang = 90.0;
  }
  
  if(!fourierAveragedCrysDisp) {
    cdNumCols = purdueHeader.numCols;
    cdNumRows = purdueHeader.numRows;
    cdNumSecs = 1;
  }

  if(cdNumCols > 0) {
    CK(crys_float_image_buffer= (float *)malloc(sizeof(float)*
						cdNumCols*cdNumRows));
    for (i=0; i<cdNumRows; i++) {
      for (j=0; j<cdNumCols; j++) { // Redo in TSB format
	if(fileType==PACKEDPORTABLE && !fourierAveragedCrysDisp) {
	  crys_float_image_buffer[(i*cdNumCols)+j]= 
	    packImg[0][(((numRows-1)-i)*numCols)+j]; 
	  // We've got to trap for zero values for B&W images
	  if(crys_float_image_buffer[(i*cdNumCols)+j]<=2.0)
	    crys_float_image_buffer[(i*cdNumCols)+j]= 2.0;
	}
	else if(fourierAveragedCrysDisp) {
	  // j*2 because the fourt() rtn stuffs it in the first 
	  // position
	  // Also - why do I have to rotate this 90 degrees ???
	  // OLDWAY works
	  // crys_float_image_buffer[((i*cdNumCols)+j)]= 
	  // float_image_buffer1[((((cdNumRows-1)-i)*(cdNumCols*2))+
	  // (j*2))];
	  // 
	  // Works 3/31/99 
	  // crys_float_image_buffer[(((cdNumRows-1-j)*
	  // cdNumCols)+i)]= 
	  // float_image_buffer1[((i*(cdNumCols*2))+
	  // (j*2))];  
	  // Works 3/31/99
	  // Works 4/1/99
	  // crys_float_image_buffer[((j*cdNumCols)+i)]= 
	  // float_image_buffer1[((i*(cdNumCols*2))+
	  // (j*2))];
	  //
	  // Original orientation gotten from FOURT rtn 
	  // crys_float_image_buffer[((i*cdNumCols)+j)]= 
	  // float_image_buffer1[((i*(cdNumCols*2))+
	  // (j*2))];
	  //
	  // Rotate 90 degrees - The fourt() rtn stuffs this
	  // buffer one column at a time - so we've got to
	  // compensate for that here.

	  crys_float_image_buffer[j*cdNumCols + i]= 
	    float_image_buffer1[i*cdNumCols*2 + j*2];
	}
	else {
	  crys_float_image_buffer[i*cdNumCols + j]= 
	    float_image_buffer1[(numRows-1-i)*numCols + j];
	}
      }
    }
    
    
    // Don't display those annoying numnbers
    dispMethod = FALSE;
    
    if(cdNumCols != oldXsize) {
      x_size= cdNumCols;
      npixy= cdNumRows;
      
      // Now setup the scroll accordingly
      XtSetArg(arglst[0], XmNmaximum, (int)((x_size*10)+5));
      XtSetArg(arglst[1], XmNminimum, 2);
      XtSetArg(arglst[2], XmNsliderSize, 5);
      XtSetArg(arglst[3], XmNvalue, x_size);
      XtSetArg(arglst[4], XmNpageIncrement, (int)(x_size/5));
      XtSetArg(arglst[5], XmNincrement, (int)MAX((x_size/50),1));
      XtSetArg(arglst[4], XmNpageIncrement, (int)(x_size/10));
      XtSetValues(adimScrollW, arglst, 5);
      XFlush(dsply);
      adim = (int)x_size;
      bdim = cdNumRows;
      XtSetArg(arglst[4], XmNpageIncrement, (int)(bdim/10));
      XtSetValues(bdimScrollW, arglst, 5);
      gammaXtal = 90.0;
      XtSetArg(arglst[0], XmNmaximum, 365);
      XtSetArg(arglst[1], XmNminimum, 0);
      XtSetArg(arglst[2], XmNsliderSize, 10);
      XtSetArg(arglst[3], XmNvalue, (int)(gammaXtal));
      XtSetArg(arglst[4], XmNpageIncrement, 10);
      XtSetValues(gammaXtalScrollW, arglst, 5);
      XFlush(dsply);
      
      // And the value label
      sprintf(outStr,"%.2f", (float)x_size/(float)cdNumCols); 
      TextSet(scaleXvalueW, outStr, FALSE);
      TextSetInt(adimTextW, adim, FALSE);
      TextSetInt(bdimTextW, bdim, FALSE);
      sprintf(outStr,"%.2f",(float)(gammaXtal)); 
      TextSet(gammaXtalTextW, outStr, TRUE);
    }
    oldXsize = cdNumCols;
    
    if(tag == 1) {
      if(map_abang == 0.0)
	map_abang = 90.0;
      map_abangTemp = map_abang;
      TextSetFloat(map_abangTextW, map_abangTemp, FALSE);
      aoverbTemp = aoverb;
      TextSetFloat(aoverbTextW, aoverbTemp, FALSE);
    }
    
    // Put screen on display
    raiseIt(crysDispScreenW);
    
    // Setup map and ab scrolls and tryit
    crysDisp_cr_activate(0,0,0); 
  }
  else
    crysDisping= FALSE;
  
} // end function crysDisp_screen_activate



void crysDisp_reset_activate(Widget widget, XtPointer client_data, 
			     XtPointer call_data)
{
  // function name: crysDisp_reset_activate
  // Called after RESET button is clicked. Reset delta scrolls to
  // zero, reinit map_abangTemp, and aoverbTemp and call tryit().

  // Reinit gloabls
  if(map_abang == 0.0)
    map_abang = 90.0;
  map_abangTemp = map_abang;
  sprintf(outStr, "%.2f", map_abangTemp);
  TextSet(map_abangTextW, outStr, FALSE);
  aoverbTemp = aoverb;
  sprintf(outStr, "%.2f", aoverbTemp);
  TextSet(aoverbTextW, outStr, FALSE);
      
  // Now setup the scrolls accordingly
  deltaMapABang= 0.0;
  XtSetArg(arglst[0], XmNvalue, (int)map_abangTemp);
  XtSetValues(mapScrollW, arglst, 1);
  XFlush(dsply);
  // Add the value label
  sprintf(outStr,"%.2f",(float)(map_abangTemp)); 
  doLabel(outStr, mapValueW, DEFAULT);
  
  // Now setup the scrolls accordingly
  deltaAoverB= 0.0;
  XtSetArg(arglst[0], XmNvalue, (int)aoverbTemp*200);
  XtSetValues(abScrollW, arglst, 1);
  XFlush(dsply);
  // And the value label
  sprintf(outStr,"%.2f",(float)(aoverbTemp)); 
  doLabel(outStr, abValueW, DEFAULT);

  crysDisp_tryit_activate(0,0,0);
  
} // end function crysDisp_reset_activate



void crysDisp_cr_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{
  // function name: crysDisp_cr_activate
  // Called when Return key hit while in the map_abang or aoverb Text
  // widgets. Reset delta scrolls to zero and call tryit().

  // Now setup the scrolls accordingly
  deltaMapABang= 0.0;
  XtSetArg(arglst[0], XmNvalue, (int)map_abangTemp);
  XtSetValues(mapScrollW, arglst, 1);
  XFlush(dsply);
  // Add the value label
  sprintf(outStr,"%.2f",(float)(map_abangTemp)); 
  doLabel(outStr, mapValueW, DEFAULT);

  // Now setup the scrolls accordingly
  deltaAoverB= 0.0;
  XtSetArg(arglst[0], XmNvalue, (int)aoverbTemp*200);
  XtSetValues(abScrollW, arglst, 1);
  XFlush(dsply);
  // Add the value label
  sprintf(outStr,"%.2f",(float)(aoverbTemp)); 
  doLabel(outStr, abValueW, DEFAULT);
  
  crysDisp_tryit_activate(0,0,0);
  
} // end function crysDisp_cr_activate



void crysDisp_tryit_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{
  // function name: crysDisp_tryit_activate
  // Try to Angularly display this float map image.

  unsigned short *disp_buffer;
  int y_size, thisPixel, icenx, iceny;
  int i, j;
  int ny, section, plotX, plotY; 
  float xstrt, ystrt, new_min, angle, aoverbPass, sinang, cosang;
  float scaleFactor, thisMin, thisMax, scale, passGamma;

  angle = map_abangTemp;
  aoverbPass = aoverbTemp;
  
  if(angle == 0.0)
    angle = 90.0;
  
  if(aoverbPass == 0.0)
    aoverbPass = 1.0;

  sinang = sin(deg2rad(angle));
  cosang = cos(deg2rad(angle));

  // Now convert angle to radians
  angle = deg2rad(angle);

  aDscale = scale  = (float)(x_size)/(float)(cdNumCols);
  ny      = scale*(float)(cdNumRows)/aoverbPass;
  y_size  = sinang*ny;

  icenx = MIN(buffer_width, display_width) / 2;
  iceny = MIN(buffer_height, display_height) / 2;

  xstrt = icenx - ((float)(x_size*ncellx) + 
		   ncelly*cosang*ny)/2.0;
  ystrt = (float)((float)iceny - 
		  ((float)((float)y_size*(float)ncelly)/2.0));

  // Center it better
  xstrt -= x_size;
  ystrt -= y_size;
  
  npixx = x_size;
  npixy = y_size;  
  
  section = 1;
  new_min = minMapPixel;
  
  plotX = MIN(buffer_width, display_width);
  plotY = MIN(buffer_height, display_height);
  
  CK(disp_buffer= (unsigned short *)
     malloc(sizeof(unsigned short)*(plotX*plotY)));
  
  for (i=0; i<plotY*plotX; i++)
    disp_buffer[i]= 0;
  
  // Put gamma into radians
  passGamma = deg2rad(gammaXtal);
  
  mapdsp_gplot_rob__(crys_float_image_buffer, disp_buffer, &aDscale, 
		    &cdNumCols, &cdNumRows, &cdNumSecs, 
		    &npixx, &npixy, &xstrt, &ystrt, &ny, 
		    &ncellx, &ncelly, &section, &new_min,
		    &angle, &aoverbPass, &plotX,&plotY,
		    &adim, &bdim, &passGamma);
  
  thisMin =  HUGE;
  thisMax = -HUGE;
  
  // Put it right beside the orig
  for(i=0; i<plotY*plotX; i++) {
    thisPixel= (float)disp_buffer[i];
    thisMin = MIN(thisPixel, thisMin);
    thisMax = MAX(thisPixel, thisMax);
  }
  
  if(thisMax > thisMin)
    scaleFactor = (tryColors-7.0)/(thisMax-thisMin);
  else
    scaleFactor = 1.0;
  
  // Init it to black
  for (i=0; i<buffer_height; i++)
    for (j=0; j<buffer_width; j++)
      XPutPixel(image_image1, j, i, pixels_return[BLACK]);
  
  // Put it right beside the orig
  for (i=0; i<plotY; i++)
    for(j=0; j<plotX; j++)
      XPutPixel(image_image1, j, ((plotY-1)-i), 
		pixels_return[(int)((disp_buffer[((i*plotX)+j)]-
				     thisMin)*scaleFactor)+6]);
  
  XFlush (dsply);
  DrawWindow();
  
  free(disp_buffer);
  
} // end function crysDisp_tryit_activate



void crystalDisp_changed(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: crystalDisp_changed
  // Call here whenever crysDisp text widgets changed

  int thisI;
  float thisF;

  if(widget == crysDispImmW)
    updateCrys = !updateCrys;
  else if (widget == scaleXvalueW) {
    thisF = wstrtod(widget);
    if(thisF != 0.0) {
      x_size = (int)((float)thisF * (float)cdNumCols);
      // XtSetArg(arglst[0], XmNvalue, (int)x_size);
      // XFlush(dsply);
      if(aoverbTemp != 0.0) {
	thisI = (int)((float)((float)cdNumCols) * thisF);
	if(thisI != 0) {
	  adim = thisI;
	  XtVaSetValues(adimScrollW, XmNvalue, adim, NULL);
	  TextSetInt(adimTextW, adim, TRUE);
	}
	thisI = ((float)((float)cdNumRows / aoverbTemp) * thisF);
	if(thisI != 0) {
	  bdim = thisI;
	  XtVaSetValues(bdimScrollW, XmNvalue, bdim, NULL);
	  TextSetInt(bdimTextW, bdim, TRUE);
	}
      }
      if(aoverbTemp != 0.0) // Wait until calculated
	crysDisp_tryit_activate(0,0,0);
    }
  }
  else if (widget == aoverbTextW) {
    aoverbTemp = wstrtod(widget);
    if(aoverbTemp != 0.0) {
      thisI = (int)((float)cdNumRows / aoverbTemp)*
	(float)((float)x_size/(float)cdNumCols);
      if(thisI != 0) {
	bdim = thisI;
	XtVaSetValues(bdimScrollW, XmNvalue, bdim, NULL);
	TextSetInt(bdimTextW, bdim, TRUE);
      }
    }
  }
  else if (widget == map_abangTextW)
    map_abangTemp = wstrtod(widget);
  else if (widget == adimTextW) {
    thisI = (int)wstrtod(widget);
    if(thisI != 0) {
      adim = thisI;
      XtVaSetValues(adimScrollW, XmNvalue, adim, NULL);
      XFlush(dsply);
      if(crys_float_image_buffer != 0 &&
	 aoverbTemp != 0.0) // Wait until calculated
	crysDisp_tryit_activate(0,0,0);
    }
  }
  else if (widget == bdimTextW) {
    thisI = (int)wstrtod(widget);
    if(thisI != 0) {
      bdim = thisI;
      XtVaSetValues(bdimScrollW, XmNvalue, bdim, NULL);
      XFlush(dsply);
      if(crys_float_image_buffer != 0 &&
	 aoverbTemp != 0.0) // Wait until calculated
	crysDisp_tryit_activate(0,0,0);
    }
  }
  else if (widget == gammaXtalTextW) {
    thisF = wstrtod(widget);
    if(thisF != 0.0) {
      gammaXtal = thisF;
      XtVaSetValues(gammaXtalScrollW, XmNvalue, (int)gammaXtal, NULL);
      XFlush(dsply);
      if(aoverbTemp != 0.0) // Wait until calculated
	crysDisp_tryit_activate(0,0,0);
    }
  }
  
} // end function crystalDisp_changed



void crysDisp_scroll_change(Widget widget, XtPointer client_data, 
			    XtPointer call_data)
{
  // function name: crysDisp_scroll_change
  // Called whenever crysDisp_scroll scroll widgets changed

  XmScrollBarCallbackStruct *temp = (XmScrollBarCallbackStruct *) call_data;
  char outStr[64];
  
  if(!inited) return;

  if (widget == ncellxScrollW) {
    ncellx = temp->value;
    sprintf(outStr,"%d",ncellx); 
    doLabel(outStr, ncellxvalueW, DEFAULT);
  }
  else if (widget == ncellyScrollW) {
    ncelly = temp->value;
    sprintf(outStr,"%d",ncelly); 
    doLabel(outStr, ncellyvalueW, DEFAULT);
  }
  else if (widget == mapScrollW) {
    map_abangTemp = temp->value;
    sprintf(outStr,"%.2f", map_abangTemp);
    doLabel(outStr, mapValueW, DEFAULT);
    TextSet(map_abangTextW, outStr, TRUE);
  }
  else if (widget == abScrollW) {
    aoverbTemp = temp->value;
    aoverbTemp *= 0.005;    // Do it in 5/1000's
    sprintf(outStr,"%.2f", aoverbTemp);
    doLabel(outStr, abValueW, DEFAULT);
    TextSet(aoverbTextW, outStr, TRUE);
  }
  else if (widget == adimScrollW) {
    adim = temp->value;
    TextSetInt(adimTextW, adim, TRUE);
  }
  else if (widget == bdimScrollW) {
    bdim = temp->value;
    TextSetInt(bdimTextW, bdim, TRUE);
  }
  else if (widget == gammaXtalScrollW) {
    gammaXtal = temp->value;
    sprintf(outStr,"%.2f",(gammaXtal));
    TextSet(gammaXtalTextW, outStr, TRUE);
  }

  if(updateCrys || (temp->reason != XmCR_DRAG))
    crysDisp_tryit_activate(0,0,0);
  
} // end function crysDisp_scroll_change



void linGrad_activate(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: linGrad_activate
  // Linearly gradient the image(s).

  int   gradRadius, i, j, k, thisPixel, saveMaxPackPixel;
  int   xcen, ycen, xSq, ySq, jMycen, kMxcen, thisRad, distnce;
  float img_avg, img_std, img_min, img_max, img_bkg;
  float *tmpBuf, *thisPackBuf;
  float scaleFactor, expFac, minP, maxP;
  
  if(fileType == PACKEDPORTABLE) {
    
    xcen = ycen = packRadius;
    thisRad = packRadius * packRadius;
    
    numCols = purdueHeader.numCols;
    numRows = purdueHeader.numRows;
    
    CK(tmpBuf= (float *)malloc(sizeof(float)*numCols*numRows));
    
    saveMaxPackPixel = abs(maxPackPixel)>abs(minPackPixel)?abs(maxPackPixel):abs(minPackPixel);
    maxPackPixel = INT_MIN;
    minPackPixel = INT_MAX;
    for (i=0; i<purdueHeader.nimg; i++) { // Accumlate all images
      
      if(!deletePackedImage[i]) {
	calcCircleStats(packImg[i], &img_bkg, &img_min, &img_max,
			&img_avg, &img_std, numCols, numRows, numCols);

	// Stuff packed byte image into short image buffer
	thisPackBuf = packImg[i];
	for(j=0; j<numRows*numCols; j++) {
	  tmpBuf[j] = thisPackBuf[j];
	}
	
	// img_lingrad depends on the background being right
	if(packRadius > 0.0)
	  gradRadius = numCols / 2 ;
	else 
	  gradRadius = 0; // prevent img_lingrad_rob() from doing circles

	// img_bkg_short = (short)img_bkg;
	
	// In img_lingrad_rob():
	// March 1998 RWA - I've had to multiply the IMG() pixel values by
	// 100.0 to increase the dynamic range so that the TRUS*() rtns 
	// work properly.  Trouble is we're passing in BYTE info in this
	// short buffer and only using 0-255 when TRUS expects 32000 short
	// values.  Changed it to pass in the expFac (expanding Factor).
	// Seems with packed boxes we can't use this large of an expanding
	// factor (100.0).  Let's try 10.0 * numCols/500.0 so a 5000x5000
	// scanned image woudl have our 100.0 factor and a 50x50 packed box
	// would have an expFac==1.0.

	expFac = 10.0 * (numCols /100.0);
	expFac = MAX(expFac, 1.0);
	
	// Let's just force this to be 1.0 for now - I think
	// expFac is more important for large scanned images???
	//expFac = 1.0;

	img_lingrad_rob__(tmpBuf, &numCols, &numRows, 
			 &gradRadius, &img_bkg, &expFac);
	
	// Find min/max
	farray_minmax(tmpBuf, &minP, &maxP, numRows*numCols);
	minP = (int)minP; // RSS casting being done for consistency
	maxP = (int)maxP; // with results from original version of code

	// Why tryColors-50.0 here ???  Why not 255.0 ???
	if(maxP == minP) { // no contrast
	  scaleFactor = 1.0;
	  minP = 1.0;
	} else { // normal
	  scaleFactor = (tryColors-50.0) / (maxP-minP);
	  scaleFactor = saveMaxPackPixel / (float)(maxP-minP);
	}
	
	// Stuff short image buffer back into Pack array
	thisPackBuf = packImg[i];
	for (j=0; j<numRows*numCols; j++)
	  thisPackBuf[j]= (tmpBuf[j]-minP)*scaleFactor + 1.0;
	
	if(packRadius > 0.0) { // Now recalculate the backGround
	  calcCircleStats(packImg[i], &img_bkg, &img_min, &img_max,
			  &img_avg, &img_std, numCols, numRows, numCols);

	  // And apply new background
	  for (j=0; j<numRows; j++) {
	    jMycen = j - ycen;
	    ySq    = jMycen*jMycen;
	    for (k=0; k<numCols; k++) {
	      kMxcen = k - xcen;
	      xSq    = kMxcen*kMxcen;
	      distnce = xSq+ySq;
	      thisPixel = j*numCols + k;
	      if(distnce > thisRad)
		thisPackBuf[thisPixel]= img_bkg;
	      
	      // Now find min/max
	      maxPackPixel= (MAX(maxPackPixel, thisPackBuf[thisPixel]));
	      minPackPixel= (MIN(minPackPixel, thisPackBuf[thisPixel]));
	    }
	  }
	  
	}
	else { // Just get what min max we've got

	  for (j=0; j<numRows*numCols; j++) {
	    maxPackPixel= (MAX(maxPackPixel, thisPackBuf[j]));
	    minPackPixel= (MIN(minPackPixel, thisPackBuf[j]));
	  }

	}
	
      }
    }
    
    // Now clip this guy to our colormap 
    if(minPackPixel < 0.0)
      scalePackFactor = (tryColors - 7.0) / (maxPackPixel-minPackPixel);
    else 
      scalePackFactor = (tryColors - 7.0) / maxPackPixel;
    
    free(tmpBuf);
    
    // Start at radius'th image
    stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		     packImg, purdueHeader.nimg, imgNum);  
    
    DrawWindow(); 
    
    if(normLinSwitch || doLinAlso)
      printf("Linear Gradient finished OK.\n");
    else
      askThem("Linear Gradient finished OK.", XmDIALOG_INFORMATION);
  }
  else if((fileType == EMPORTABLE || fileType == BYTEPORTABLE ||
	   fileType == PORTABLE) && imgInCore) {
    
    numCols = file_width;
    numRows = file_height;
    
    CK(tmpBuf= (float *)malloc(sizeof(float)*numCols*numRows));
    
    calcStats(inCore_buffer, &img_bkg, &img_min, &img_max,
	      &img_avg, &img_std, numCols, numRows, numCols,
	      fileType, bkgrndMethod, boxMethod, packRadius,
	      boxRadius, featherRadius, recipSizeFactor, tryColors);
    
    // Stuff packed byte image into short image buffer
    for (j=0; j<numRows*numCols; j++)
      tmpBuf[j]= inCore_buffer[j];
    
    // img_lingrad depends on the background being right
    gradRadius = MAX(numRows, numCols) * 1.5;

    // img_bkg_short = (short)img_bkg;
    
    
    // In img_lingrad_rob():
    // March 1998 RWA - I've had to multiply the IMG() pixel values by
    // 100.0 to increase the dynamic range so that the TRUS*() rtns 
    // work properly. Trouble is we're passing in BYTE info in this
    // short buffer and only using 0-255 when TRUS expects 32000 short
    // values. Changed it to pass in the expFac (expanding Factor).
    // Seems with packed boxes we can't use this large of an expanding
    // factor (100.0). Let's try 10.0 * numCols/500.0 so a 5000x5000
    // scanned image woudl have our 100.0 factor and a 50x50 packed box
    // would have an expFac==1.0.

    expFac = 10.0 * ((float)numCols /500.0);
    expFac = 100.0;
    img_lingrad_rob__(tmpBuf, &numCols, &numRows, 
		     &gradRadius, &img_bkg, &expFac);
    
    // Stuff short image buffer back into Pack array
    farray_minmax(tmpBuf, &minP, &maxP, numRows*numCols);
    minP = (int)minP; // RSS casting being done for consistency
    maxP = (int)maxP; // with results from original version of code
    scaleFactor = (tryColors-9.0) / (maxP-minP);
    
    for (j=0; j<numRows*numCols; j++)
      inCore_buffer[j] = (tmpBuf[j]-minP)*scaleFactor;
    
    free(tmpBuf);
    
    // Make them regenerate a full-screen image
    // fullScreen = TRUE;
    fullScreen_buffer.dispx = fullScreen_buffer.dispy = 0;
    
    virtual_read_activate(0,0);
    
    askThem("Linear Gradient finished OK.", XmDIALOG_INFORMATION);
  }
  else {
    askThem("Cannot perform Linear gradient for this data.",
	    XmDIALOG_INFORMATION);
  }
  
} // end function linGrad_activate



void normalize_activate(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: normalize_activate

  // Normalize images across the entire data set. *tag==2 means they
  // want normalization to happen for backgrounds instead of the usual
  // way using means.

  int i, tot;
  int xcen, ycen, thisRad;
  float *img_bkg, *img_min,*img_max, *img_avg, *img_std;
  float totMean, totStd;
  float globMin, globMax, globMean, globStd;
  int tag = *((int *) client_data);

  kindOfNormalize = tag;

  if(fileType == PACKEDPORTABLE) {
    xcen = ycen = packRadius;
    thisRad = packRadius * packRadius;
    
    if(tag==2) { // They want background normalization
      if(!askThem("Normalization of Backgrounds should only be used\n\
for cosmetic purposes.  For programs like PFT\n\
you should use Normalize Means&Vars.\n\
Continue?",
		  XmDIALOG_QUESTION))
	goto noCalc; 
    }

    CK(img_bkg= (float *)malloc(sizeof(float)*purdueHeader.nimg));
    CK(img_min= (float *)malloc(sizeof(float)*purdueHeader.nimg));
    CK(img_max= (float *)malloc(sizeof(float)*purdueHeader.nimg));
    CK(img_avg= (float *)malloc(sizeof(float)*purdueHeader.nimg));
    CK(img_std= (float *)malloc(sizeof(float)*purdueHeader.nimg));
    
    numCols = purdueHeader.numCols;
    numRows = purdueHeader.numRows;
    
    tot = 0;
    totMean = totStd= 0.0;
    globMin = HUGE;
    globMax = -HUGE;
    
    for (i=0; i<purdueHeader.nimg; i++) { // Accumlate all images
      if(!deletePackedImage[i]) {
	calcCircleStats(packImg[i], &img_bkg[i], &img_min[i],
			&img_max[i], &img_avg[i], &img_std[i],
			numCols, numRows, numCols);

	// printf("Norm min[i]=%d  max[i]=%d  avg=%f  std=%f\n",
	// img_min[i], img_max[i], img_avg[i], img_std[i]);
	
	totMean += (tag==2) ? img_bkg[i] : img_avg[i];
	
	totStd  += img_std[i];	
	globMin = MIN(globMin,img_min[i]);
	globMax = MAX(globMax,img_max[i]);
	tot++;
      }
    }

    if(tot > 0) { // OK to calc
      globMean = totMean / tot;
      globStd  = totStd  / tot;
    }
    else { // Error
      if(normLinSwitch)
	printf("Nothing to calc.\n");
      else
	askThem("Nothing to calc.", XmDIALOG_INFORMATION);
      goto noCalc;
    }
    
    // printf("\n\nglobMin=%d  max=%d\n",globMin,globMax); 
    // printf("GlobMean=%f  globStd=%f\n",globMean,globStd);

    printf("Done globSTD changed to [%f]\n",globStd);    
    printf("     globMean           [%f]\n",globMean);    
    
    free(img_bkg);
    free(img_min);
    free(img_max);
    free(img_avg);
    free(img_std);
    
    if(dumpNormStats) { // dump to file
      FILE *fDump;
      if ( (fDump = fopen(dumpNormFile,"a")) != NULL) {
	fprintf(fDump,"[%s] mean[%f] stDev[%f]\n",
		fileName.addr, globMean, globStd);
	fclose(fDump);
      }
    }
    
    // Suppose you get the gloabl mean and StDev from the interface
    // and then just apply that here... ???

    TextSetFloat(normalMeanW, globMean, FALSE);
    TextSetFloat(normalStdW, globStd, FALSE);
    if(!normLinSwitch)
      raiseIt(normalScreenW);
    
  }
  else {
    if(normLinSwitch)
      printf("Normalize Means&Vars only for Packed boxes.\n");
    else
      askThem("Normalize Means&Vars only for Packed boxes.",
	      XmDIALOG_INFORMATION);
  }
  
 noCalc:;
  
} // end function normalize_activate



void normalize_activate_2(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: normalize_activate_2
  // Normalize images across the entire data set. *tag==2 means they
  // want normalization to happen for Backgrounds instead of the usual
  // way using means.

  int   i, j, k, new_min, new_max;
  int   xcen, ycen, xSq, ySq, jMycen, kMxcen, thisRad, distnce, thisPixel;
  float *img_bkg, *img_min, *img_max, *img_avg, *img_std;
  float globMean, globStd, stdAdjustFactor;
  float *thisPackBuf;

  if(useGUI) globMean = wstrtod(normalMeanW);
  if(useGUI) globStd = wstrtod(normalStdW);
  
  if(normLinSwitch) {
    if(normLinMean==0.0 && normLinStd==0.0) {
      printf("Using calculated values Mean[%f] Std[%f]\n",
	     globMean,globStd);
    }
    else { // Use command line vals
      globMean = normLinMean;
      globStd  = normLinStd;
      printf("Using command line values Mean[%f] Std[%f]\n",
	     globMean,globStd);
    }
  }
  
  xcen = ycen = packRadius;
  thisRad = packRadius * packRadius;
  
  CK(img_bkg= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_min= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_max= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_avg= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_std= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  
  numCols = purdueHeader.numCols;
  numRows = purdueHeader.numRows;
  
  for (i=0; i<purdueHeader.nimg; i++)
    if(!deletePackedImage[i])
      calcCircleStats(packImg[i], &img_bkg[i], &img_min[i], &img_max[i],
		      &img_avg[i], &img_std[i], numCols, numRows, numCols);
  
  // Suppose you get the gloabl mean and StDev from the interface
  // and then just apply that here... ???
  
  printf("Using globSTD  == [%f]\n",globStd);    
  printf("      globMean == [%f]\n",globMean);    
  
  // Now scale each image to the new min/max and add the constant
  // to bring it in line with the global Mean.
  
  minPackPixel = maxPackPixel = globMean;
  
  for (i=0; i<purdueHeader.nimg; i++) {
    if(!deletePackedImage[i]) {
      stdAdjustFactor = (img_std[i] != 0.0) ? globStd/img_std[i] : 1.0;
	  
      if(kindOfNormalize == 2) { // BKG
	new_min = (img_min[i]-img_bkg[i])*stdAdjustFactor + 
	  img_avg[i] + 0.49999;
	new_max = (img_max[i]-img_bkg[i])*stdAdjustFactor + 
	  img_avg[i] + 0.49999;
      }
      else { // AVG
	new_min = (img_min[i]-img_avg[i])*stdAdjustFactor + 
	  img_avg[i] + 0.49999;
	new_max = (img_max[i]-img_avg[i])*stdAdjustFactor + 
	  img_avg[i] + 0.49999;
      }
      
      for (j=0; j<numRows*numCols; j++) {
	// Get this to expand/contract around the new global Mean 
	if(kindOfNormalize == 2)
	  packImg[i][j] = globMean + (packImg[i][j]-img_bkg[i])*stdAdjustFactor;
	else
	  packImg[i][j] = globMean + (packImg[i][j]-img_avg[i])*stdAdjustFactor;
	
	maxPackPixel= MAX(maxPackPixel, packImg[i][j]);
	minPackPixel= MIN(minPackPixel, packImg[i][j]);
      }
      
      // Now recalc the backGround
      calcCircleStats(packImg[i], &img_bkg[i], &img_min[i], &img_max[i],
		      &img_avg[i], &img_std[i], numCols, numRows, numCols);

      // And apply new background
      thisPackBuf = packImg[i];
      for (j=0; j<numRows; j++) {
	jMycen = j - ycen;
	ySq    = jMycen*jMycen;
	for (k=0; k<numCols; k++) {
	  kMxcen = k - xcen;
	  xSq    = kMxcen*kMxcen;
	  distnce = xSq+ySq;
	  if(distnce > thisRad) { // Stuff it
	    thisPixel = j*numCols + k;
	    thisPackBuf[thisPixel] = (forceAvgBkg) ? img_avg[i] : img_bkg[i];
	  }  
	}
      }
      
    } 
  }

  minPackPixel -= 1.5;
  maxPackPixel += 1.5;
  if(minPackPixel < 0.0)
    scalePackFactor = (tryColors - 7.0) / (maxPackPixel-minPackPixel);
  else 
    scalePackFactor = (tryColors - 7.0) / maxPackPixel;

  free(img_bkg);
  free(img_min);
  free(img_max);
  free(img_avg);
  free(img_std);
  
  // Start at radius'th image
  stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		   packImg,purdueHeader.nimg, imgNum);
  
  DrawWindow(); 
  
  if(kindOfNormalize == 2)
    askThem("Normalize Bkgs&Vars finished OK.", XmDIALOG_INFORMATION);
  else {
    if(normLinSwitch)
      printf("Normalize Means&Vars finished OK.\n");
    else
      askThem("Normalize Means&Vars finished OK.", XmDIALOG_INFORMATION);
  }

  if(useGUI) XtUnmanageChild(normalScreenW);
  
} // end function normalize_activate_2



void checkBkg_activate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: checkBkg_activate
  // Check boxes for bad backgrounds.

  int i, j, k, tot;
  int xcen, ycen, xSq, ySq, jMycen, kMxcen, thisRad, distnce, thisPixel;
  float *img_bkg, *img_min,*img_max, *img_avg,*img_std;
  float totMean, *thisPackBuf;
	      
  // This may take awhile so give them the watch cursor
  updateCursor(drawingW, watch_cursor);
  
  CK(img_bkg= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_min= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_max= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_avg= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_std= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  
  numCols = purdueHeader.numCols;
  numRows = purdueHeader.numRows;
  
  tot = 0; 
  totMean= 0.0;

  for (i=0; i<purdueHeader.nimg; i++) {
      calcCircleStats(packImg[i], &img_bkg[i], &img_min[i], &img_max[i],
		      &img_avg[i], &img_std[i], numCols, numRows, numCols);
      if(img_bkg[i] != (float)*(packImg[i])) {
	tot++;
	totMean += (float)((float)*(packImg[i])-img_bkg[i]);
      }
  }
  
  if(tot > 0) {
    
    sprintf(outStr,"I've found [%d] boxes with bogus background values.\n\
The mean difference from the theoretical background is [%f].\n\n\
Do you want me to change them?\n",tot,(float)(totMean/(float)tot));
    
    if(askThem(outStr, XmDIALOG_QUESTION)) {
      for (i=0; i<purdueHeader.nimg; i++) {    
	xcen = ycen = packRadius;
	thisRad = packRadius * packRadius;
	
	// Now recalc the backGround
	calcCircleStats(packImg[i], &img_bkg[i], &img_min[i], &img_max[i],
			&img_avg[i], &img_std[i], numCols, numRows, numCols);
	
	if(img_bkg[i] != (float)*(packImg[i])) {
	  thisPackBuf = packImg[i];
	  for (j=0; j<numRows; j++) {
	    jMycen = j - ycen;
	    ySq    = jMycen*jMycen;
	    for (k=0; k<numCols; k++) {
	      kMxcen = k - xcen;
	      xSq    = kMxcen*kMxcen;
	      distnce = xSq+ySq;
	      if(distnce > thisRad) {
		thisPixel = j*numCols + k;
		thisPackBuf[thisPixel]= img_bkg[i];
	      }
	    }
	  }
	}
      }
    }
  } else {
    askThem("All backgrounds checked out OK.", XmDIALOG_INFORMATION);
  }

  free(img_bkg);
  free(img_min);
  free(img_max);
  free(img_avg);
  free(img_std);
  
  // Start at radius'th image
  stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		   packImg,purdueHeader.nimg, imgNum);  
  
  DrawWindow(); 
  
  // Change cursor back to what it was
  updateCursor(drawingW, old_cursor);
  
} // end function checkBkg_activate



void mb_scroll_change(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: mb_scroll_change
  // Update text widgets in model builder screen when scroll bars activated

  XmScrollBarCallbackStruct *temp = (XmScrollBarCallbackStruct *) call_data;

  sprintf(outStr, "%d", (int)temp->value);

  if(widget      == mbRotatetScrollW)
    TextSet(mbRotatetTextW, outStr, FALSE); 
  else if(widget == mbRotatepScrollW)
    TextSet(mbRotatepTextW, outStr, FALSE); 
  else if(widget == mbRotateoScrollW)
    TextSet(mbRotateoTextW, outStr, FALSE); 
  else if(widget == mbDimensionxScrollW)
    TextSet(mbDimensionxTextW, outStr, FALSE); 
  else if(widget == mbDimensionyScrollW)
    TextSet(mbDimensionyTextW, outStr, FALSE); 
  else if(widget == mbDimensionzScrollW)
    TextSet(mbDimensionzTextW, outStr, FALSE); 
  else if(widget == mbCenterxScrollW)
    TextSet(mbCenterxTextW, outStr, FALSE); 
  else if(widget == mbCenteryScrollW)
    TextSet(mbCenteryTextW, outStr, FALSE); 
  else if(widget == mbCenterzScrollW)
    TextSet(mbCenterzTextW, outStr, FALSE); 
  
  if(TRUE && mbImmediate)
    model_activate(widget, client_data, call_data);
  
} // end function mb_scroll_change



void modelObject_change(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: modelObject_change
  // Change modelObject
  
  int tag = *((int *) client_data);
  modelType = tag;

  widgetList[0] = mbDimensionyScrollW;
  widgetList[1] = mbDimensionyTextW;
  widgetList[2] = mbDimensionzScrollW;
  widgetList[3] = mbDimensionzTextW;
  widgetList[4] = mbRotatetScrollW;
  widgetList[5] = mbRotatetTextW;
  widgetList[6] = mbRotatepScrollW;
  widgetList[7] = mbRotatepTextW;
  widgetList[8] = mbRotateoScrollW;
  widgetList[9] = mbRotateoTextW;

  if(tag == SQUARE   || tag == CIRCLE || tag == ICOS ||
     tag == CYLINDER || tag == TRIANGLE)
    XtManageChildren(widgetList, 10);
  else {
    printf ("MODELOBJECT_CHANGE - Unknown object\n");
    modelType = SQUARE;  // Default it here
    doLabel("Bug in object code.", modelLabelW, DEFAULT);
    return;
  }

  if(tag == SQUARE)
    doLabel("Object to build:  Cube", modelLabelW, DEFAULT);
  else if(tag == TRIANGLE)
    doLabel("Object to build:  Triangle", modelLabelW, DEFAULT);
  else if(tag == CIRCLE) {
    doLabel("Object to build:  Sphere", modelLabelW, DEFAULT);
    XtUnmanageChildren(widgetList, 10);
  }
  else if(tag == ICOS) {
    doLabel("Object to build:  Icosahedron", modelLabelW, DEFAULT);
    XtUnmanageChild(mbDimensionyScrollW);
    XtUnmanageChild(mbDimensionyTextW);
  }
  else if(tag == CYLINDER) {
    doLabel("Object to build:  Cylinder", modelLabelW, DEFAULT);
    XtUnmanageChild(mbDimensionyScrollW);
    XtUnmanageChild(mbDimensionyTextW);
  }
    
} // end function modelObject_change



void model_screen_activate(Widget widget, XtPointer client_data, 
			   XtPointer call_data)
{
  // function name: model_screen_activate
  // Get the Model building screen up on the display.

  int i, xCen, yCen, zCen;

  freeNullFloat(&model_undo_buffer);

  CK(model_undo_buffer = (float *)malloc(sizeof(float)*
					 numCols*numRows*numSecs));

  for (i=0; i<(numSecs*numRows*numCols); i++)
    model_undo_buffer[i] = float_image_buffer1[i];
  
  raiseIt(statusScreenW);
  XSync(dsply,0);
  checkStatus("Configuring widgets ...");
  
  xCen = (numCols / 2) + 1 ;      // +1 for human user
  yCen = (numRows / 2) + 1 ;      // +1 for human user
  zCen = (origNumSecs / 2) + 1 ;  // +1 for human user
  XtVaSetValues(mbDimensionxScrollW,XmNmaximum, numCols+10, NULL);
  XtVaSetValues(mbDimensionxScrollW,XmNsliderSize, 10,NULL);
  
  XtVaSetValues(mbDimensionyScrollW,XmNmaximum, numRows+10, NULL);
  XtVaSetValues(mbDimensionyScrollW,XmNsliderSize, 10,NULL);
  
  XtVaSetValues(mbDimensionzScrollW,XmNmaximum, origNumSecs+10, NULL);
  XtVaSetValues(mbDimensionzScrollW,XmNsliderSize, 10,NULL);
  
  XtVaSetValues(mbCenterxScrollW,XmNmaximum, numCols+10, NULL);
  XtVaSetValues(mbCenterxScrollW,XmNsliderSize, 10, NULL);
  XtVaSetValues(mbCenterxScrollW,XmNvalue, xCen, NULL);
  
  XtVaSetValues(mbCenteryScrollW,XmNmaximum, numRows+10, NULL);
  XtVaSetValues(mbCenteryScrollW,XmNsliderSize, 10, NULL);
  XtVaSetValues(mbCenteryScrollW,XmNvalue, yCen, NULL);
  
  XtVaSetValues(mbCenterzScrollW,XmNmaximum, origNumSecs+10, NULL);
  XtVaSetValues(mbCenterzScrollW,XmNsliderSize, 10, NULL);
  XtVaSetValues(mbCenterzScrollW,XmNvalue, zCen, NULL);
  
  TextSetInt(mbCenterxTextW, xCen, FALSE);
  TextSetInt(mbCenteryTextW, yCen, FALSE);
  TextSetInt(mbCenterzTextW, zCen, FALSE);
  
  raiseIt(mapShadeScreenW);
  raiseIt(statusScreenW);
  XSync(dsply,0);
  
  mapShadeScreen_activate(NULL, NULL, NULL);
  XFlush(dsply);
  raiseIt(modelScreenW);

  mapShade_activate(NULL, NULL, NULL);
  XtUnmanageChild(statusScreenW);

} // end function model_screen_activate



void model_activate(Widget widget, XtPointer client_data, 
		    XtPointer call_data)
{
  // function name: model_activate
  // This is the "Tryit" button.

  int i, j, k;
  int xCen, yCen, zCen, dimX, dimY, dimZ;
  int theta, phi, omega;
  int begX, begY, begZ, endX, endY, endZ;
  int zSq, ySq, xSq, dimXsq;
  int iter, numIter;
  double thisCol, thisRow, thisSecn;
  double t1, t2, t3, t4, t5, t6, t7, t8, t9;
  double ctheta, cphi, comega, stheta, sphi, somega;
  float modelFillValuePlus, thetas[25], phis[25];
  
  modelFillValue =  wstrtod(modelFillTW);
  modelFillValuePlus = modelFillValue*1.01;
  
  xCen =  wstrtod(mbCenterxTextW);
  yCen =  wstrtod(mbCenteryTextW);
  zCen =  wstrtod(mbCenterzTextW);
  dimX =  wstrtod(mbDimensionxTextW);
  dimY =  wstrtod(mbDimensionyTextW);
  dimZ =  wstrtod(mbDimensionzTextW);
  theta = wstrtod(mbRotatetTextW);
  phi   = wstrtod(mbRotatepTextW);
  omega = wstrtod(mbRotateoTextW);

  dimXsq = dimX*dimX;
  
  xCen--;  // Now in computer space
  yCen--;
  zCen--;
  begX = xCen - dimX/2;
  begY = yCen - dimY/2;
  begZ = zCen - dimZ/2;
  endX = begX + dimX;
  endY = begY + dimY;
  endZ = begZ + dimZ;

  // We've got a bug whereby we've got a lot of problems with
  // the original map_prj_all() rtn that does the rotation.
  // WHY does phi not have to have the sign reversed here like
  // it does in emUnix.c in the thetaPhi_dphshd_C() rtn???
  
  comega = cos(deg2rad(phi));
  somega = sin(deg2rad(phi));
  cphi   = cos(deg2rad(omega));
  sphi   = sin(deg2rad(omega));
  ctheta = cos(deg2rad(0.0-theta));
  stheta = sin(deg2rad(0.0-theta));
  
  // COMPUTE ELEMENTS OF ROTATION MATRIX 
  
  t1 = ctheta*cphi*comega + sphi*somega;
  t2 = ctheta*sphi*comega - cphi*somega;
  t3 = -stheta*comega;
  t4 = ctheta*cphi*somega - sphi*comega;
  t5 = ctheta*sphi*somega + cphi*comega;
  t6 = -stheta*somega;
  t7 = stheta*cphi;
  t8 = stheta*sphi;
  t9 = ctheta;
  
  raiseIt(statusScreenW);
  
  // Always refresh from last saved
  for(i=0; i<origNumSecs*numRows*numCols; i++)
    float_image_buffer1[i] = model_undo_buffer[i] ;
  
  numIter = 1;
  if(isIcosRepl) { // Let's find the vertices we need
    float thisRad;
    int thisXcen, thisYcen, thisZcen;

    printf("Doing icos repl \n");
    thisRad = (float)xCen;
    thisXcen = numCols / 2;
    thisYcen = numRows / 2;
    thisZcen = origNumSecs / 2;
    
    // Stuff the icosahedron values
    init_vertices();
    
    // OK let's change the coords now
    for (i=0; i<icosabox.num_points; i++) {
      icosabox.vertex[i].lx *= thisRad;
      icosabox.vertex[i].ly *= thisRad;
      icosabox.vertex[i].lz *= thisRad;
      // Now put it in our box
      icosabox.vertex[i].lx += thisXcen;
      icosabox.vertex[i].ly += thisYcen;
      icosabox.vertex[i].lz += thisZcen;
      
      thetas[i] = (float)(icosabox.vertex[i].lx - thisXcen) /
	(float)(icosabox.vertex[i].lz - thisZcen) ;
      thetas[i] = atan(thetas[i]);
      thetas[i] = rad2deg(thetas[i]);
      phis[i] = (float)(icosabox.vertex[i].ly - thisYcen) /
	(float)(icosabox.vertex[i].lx - thisXcen) ;
      phis[i] = atan(phis[i]);
      phis[i] = rad2deg(phis[i]); 
    }
      
    numIter = icosabox.num_points;  
  }
  
  // OK - create a new object for every replication we may need
  for (iter=0; iter<numIter; iter++) {
    if(isIcosRepl) { // Let's find the centers
      sprintf(outStr,"Placing Icos position[%d].",iter+1);
      checkStatus(outStr);
      xCen = icosabox.vertex[iter].lx;
      yCen = icosabox.vertex[iter].ly;
      zCen = icosabox.vertex[iter].lz;
      begX = xCen - dimX/2;
      begY = yCen - dimY/2;
      begZ = zCen - dimZ/2;
      endX = begX + dimX;
      endY = begY + dimY;
      endZ = begZ + dimZ;
      
      // Now do the rotation
      omega = 0.0;

      // Why do we have to do this for this case ??? 
      if(thetas[iter] > -0.0001 && thetas[iter] < 0.0001) {
	if(iter ==  5 || iter == 7)
	  thetas[iter] = 0.0 - (90.0 - 31.717468);
	else
	  thetas[iter] = 90.0 - 31.717468;
      }

      comega = cos(deg2rad(phis[iter]));
      somega = sin(deg2rad(phis[iter]));
      cphi   = cos(deg2rad(omega));
      sphi   = sin(deg2rad(omega));
      ctheta = cos(deg2rad(0.0-thetas[iter]));
      stheta = sin(deg2rad(0.0-thetas[iter]));
      
      // COMPUTE ELEMENTS OF ROTATION MATRIX
      t1 = ctheta*cphi*comega + sphi*somega;
      t2 = ctheta*sphi*comega - cphi*somega;
      t3 = -stheta*comega;
      t4 = ctheta*cphi*somega - sphi*comega;
      t5 = ctheta*sphi*somega + cphi*comega;
      t6 = -stheta*somega;
      t7 = stheta*cphi;
      t8 = stheta*sphi;
      t9 = ctheta; 
    }
    
    if(modelType==SQUARE) { // do cube      
      for (i=begZ; i<endZ; i++) {
	for (j=begY; j<endY; j++) {
	  for (k=begX; k<endX; k++) {
	    newTranslateThetaPhiOmega(&thisCol,&thisRow,&thisSecn,
				      xCen,yCen,zCen,k,j,i,
				      t1,t2,t3,t4,t5,t6,t7,t8,t9);
	    float_image_buffer1[(((int)thisSecn*numCols*numRows)+
				 ((int)thisRow*numCols)+ 
				 (int)thisCol)] = modelFillValue;
	  }
	}
      }
      
      // We have to corrupt a corner pixel to make sure our
      // shading algorithm can display it when not rotated
      //      newTranslateThetaPhiOmega(&thisCol,&thisRow,&thisSecn,
      //			xCen,yCen,zCen,
      //			begX,begY,(endZ-1),
      //			t1,t2,t3,t4,t5,t6,t7,t8,t9);
      //      float_image_buffer1[(((int)thisSecn*numCols*numRows)+
      //		   ((int)thisRow*numCols)+ 
      //		   (int)thisCol)] = 0.0;
    }
    else if(modelType==CIRCLE) { // do sphere  
      for (i=0; i<origNumSecs; i++) {
	zSq = (i-zCen)*(i-zCen);
	for (j=0; j<numRows; j++) {
	  ySq = (j-yCen)*(j-yCen);
	  for (k=0; k<numCols; k++) {
	    xSq = (k-xCen)*(k-xCen) + ySq + zSq;
	    if(xSq < dimXsq)
	      float_image_buffer1[(i*numCols*numRows)+
				  (j*numCols)+ k] = modelFillValue;
	  }
	}
      }
    }
    else if(modelType==TRIANGLE) { // do triangle
      int *triMask;
      float slopeLeft, interLeft, slopeRight, interRight;
      float thisY;
      
      triMask = (int *)malloc(numCols*numRows*sizeof(int));
      
      for(j=0; j<numRows*numCols; j++)
	triMask[j] = TRUE;
      
      if(xCen != begX && xCen != endX) {
	slopeLeft = (float)(endY - begY) / (float)(xCen - begX);
	interLeft = (float)begY - (float)(slopeLeft * (float)begX);
	slopeRight = (float)(endY - begY) / (float)(xCen - endX);
	interRight = (float)begY - (float)(slopeRight*(float)endX);
	
	for (j=begY; j<endY; j++) {
	  for (k=begX; k<endX; k++) {
	    if(k <= xCen) { // left
	      thisY = k*slopeLeft+interLeft;
	      if(j >= (int)thisY)
		triMask[j*numCols + k] = FALSE;
	    }
	    else if(k > xCen) { // right
	      thisY = k*slopeRight+interRight;
	      if(j >= (int)thisY)
		triMask[j*numCols + k] = FALSE;
	    }
	    else
	      triMask[j*numCols + k] = FALSE;
	  }
	}
	      
	// Now that we have the mask just rotate it and fill buffer
	for(i=begZ;i<endZ;i++) {
	  for(j=begY;j<endY;j++) {
	    for(k=begX;k<endX;k++) {
	      if(triMask[j*numCols + k]) {
		newTranslateThetaPhiOmega(&thisCol,&thisRow,&thisSecn,
					  xCen,yCen,zCen,k,j,i,
					  t1,t2,t3,t4,t5,t6,t7,t8,t9);
		float_image_buffer1[(((int)thisSecn*numCols*numRows)+
				     ((int)thisRow*numCols)+ 
				     (int)thisCol)] = modelFillValue;
	      }
	    }
	  }
	}

	// We have to corrupt a corner pixel to make sure our
	// shading algorithm can display it when not rotated
	// newTranslateThetaPhiOmega(&thisCol,&thisRow,&thisSecn,
	//		    xCen,yCen,zCen,
	//		    begX+1,begY+1,(endZ-1),
	//		    t1,t2,t3,t4,t5,t6,t7,t8,t9);
	//float_image_buffer1[(((int)thisSecn*numCols*numRows)+
	//	       ((int)thisRow*numCols)+ 
	//	       (int)thisCol)] = 0.0;
      }
	  
      free(triMask);
    }
    else if(modelType==CYLINDER) { // do cylinder
      
      for (i=begZ; i<endZ; i++) {
	for (j=0; j<numRows; j++) {
	  ySq = (j-yCen)*(j-yCen);
	  for (k=0; k<numCols; k++) {
	    xSq = (k-xCen)*(k-xCen) + ySq;
	    if(xSq < dimXsq) {
	      newTranslateThetaPhiOmega(&thisCol,&thisRow,&thisSecn,
					xCen,yCen,zCen,k,j,i,
					t1,t2,t3,t4,t5,t6,t7,t8,t9);
	      float_image_buffer1[(((int)thisSecn*numCols*numRows)+
				   ((int)thisRow*numCols)+ 
				   (int)thisCol)] = modelFillValue;
	    }
	  }
	}
      }

      // We have to corrupt a corner pixel to make sure our
      // shading algorithm can display it when not rotated
      // newTranslateThetaPhiOmega(&thisCol,&thisRow,&thisSecn,
      //			xCen,yCen,zCen,
      //			begX+1,yCen,(endZ-1),
      //		    	t1,t2,t3,t4,t5,t6,t7,t8,t9);
      // float_image_buffer1[(((int)thisSecn*numCols*numRows)+
      //	   ((int)thisRow*numCols)+ 
      //		   (int)thisCol)] = 0.0;  
    }
    else if(modelType==ICOS) { // do icos
      double thisX, thisY, thisZ, thisRad;
      int iter;
      
      checkStatus("Calculating Icos shell.");
      
      for (iter=0; iter<dimZ; iter++) {
	dimX -= 1;
	if(dimX >= 0) {
	  for (i=0; i<origNumSecs; i++) {
	    for (j=0; j<numRows; j++) {
	      ySq = (j-yCen)*(j-yCen);
	      for (k=0; k<numCols; k++) {
		xSq = (k-xCen)*(k-xCen) + ySq;
		thisX = k - xCen;
		thisY = j - yCen;
		thisZ = i - zCen;
		thisRad = dimX;
		if(doIcosshell(thisX,thisY,thisZ,thisRad)) {
		  newTranslateThetaPhiOmega(&thisCol,&thisRow,&thisSecn,
					    xCen,yCen,zCen,k,j,i,
					    t1,t2,t3,t4,t5,
					    t6,t7,t8,t9);
		  float_image_buffer1[(((int)thisSecn*numCols*
					numRows)+
				       ((int)thisRow*numCols)+ 
				       (int)thisCol)] = modelFillValue;
		}
	      }
	    }
	  }
	}
      }
    }

  } // DO all iter
  
  // We've got "holes" in the objects that are rotated to
  // the icosahedral vertices.  Do a nearest neighbor search
  // to fill in the holes.

  if(isIcosRepl) { // Fill in the blanks
    int numGood, thisPixel, curPixel, nrnc;
    
    checkStatus("Cleaning up.");
    
    nrnc = numRows*numCols;
    
    for (i=1; i<origNumSecs-1; i++) {
      for (j=1; j<numRows-1; j++) {
	for (k=1; k<numCols-1; k++) {
	  numGood = 0;
	  thisPixel = i*nrnc+j*numCols + k;
	  if(float_image_buffer1[thisPixel] != modelFillValue) {
	    // Now accum the nearest neighbors
	    
	    // (i-1)
	    curPixel = (i-1)*nrnc + (j-1)*numCols + (k-1);
	    if(float_image_buffer1[curPixel] == modelFillValue)
	      numGood++;
	    curPixel = (i-1)*nrnc + (j-1)*numCols + k;
	    if(float_image_buffer1[curPixel] == modelFillValue)
	      numGood++;
	    curPixel = (i-1)*nrnc + (j-1)*numCols + (k+1);
	    if(float_image_buffer1[curPixel] == modelFillValue)
	      numGood++;
	    curPixel = (i-1)*nrnc + j*numCols + (k-1);
	    if(float_image_buffer1[curPixel] == modelFillValue)
	      numGood++;
	    curPixel = (i-1)*nrnc + j*numCols + k;
	    if(float_image_buffer1[curPixel] == modelFillValue)
	      numGood++;
	    curPixel = (i-1)*nrnc + j*numCols + (k+1);
	    if(float_image_buffer1[curPixel] == modelFillValue)
	      numGood++;
	    curPixel = (i-1)*nrnc + (j+1)*numCols + (k-1);
	    if(float_image_buffer1[curPixel] == modelFillValue)
	      numGood++;
	    curPixel = (i-1)*nrnc + (j+1)*numCols + k;
	    if(float_image_buffer1[curPixel] == modelFillValue)
	      numGood++;

	    if(numGood > 1) {
	      // We can't pass unless we've gotten one by now
	      curPixel = (i-1)*nrnc + (j+1)*numCols + (k+1);
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      
	      // (i)
	      curPixel = i*nrnc + (j-1)*numCols + (k-1);
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = i*nrnc + (j-1)*numCols + k;
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = i*nrnc + (j-1)*numCols + (k+1);
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = i*nrnc + j*numCols + (k-1);
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = i*nrnc + j*numCols + k;
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = i*nrnc + j*numCols + (k+1);
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = i*nrnc + (j+1)*numCols + (k-1);
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = i*nrnc + (j+1)*numCols + (k);
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = i*nrnc + (j+1)*numCols + (k+1);
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      
	      // (i+1)
	      curPixel = (i+1)*nrnc + (j-1)*numCols + (k-1);
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = (i+1)*nrnc + (j-1)*numCols + k;
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = (i+1)*nrnc + (j-1)*numCols + (k+1);
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = (i+1)*nrnc + j*numCols + (k-1);
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = (i+1)*nrnc + j*numCols + k;
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = (i+1)*nrnc + j*numCols + (k+1);
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = (i+1)*nrnc + (j+1)*numCols + (k-1);
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = (i+1)*nrnc + (j+1)*numCols + k;
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      curPixel = (i+1)*nrnc + (j+1)*numCols + (k+1);
	      if(float_image_buffer1[curPixel] == modelFillValue)
		numGood++;
	      
	      if(numGood > 18)
		float_image_buffer1[thisPixel] = modelFillValue;
	    }
	  }
	}
      }
    }
    
      
  } // Fill in the blanks
  
  XtUnmanageChild(statusScreenW);
  XFlush(dsply);
  mapShade_activate(NULL, NULL, NULL);
  
} // end function model_activate



void model_save_activate(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: model_save_activate
  int i;

  for (i=0; i<(origNumSecs*numRows*numCols); i++)
    model_undo_buffer[i] = float_image_buffer1[i];
  
} // end function model_save_activate



void model_undo_activate(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: model_undo_activate
  // Undo last model operation.

  int i;

  for (i=0; i<(origNumSecs*numRows*numCols); i++)
    float_image_buffer1[i] = model_undo_buffer[i] ;

  mapShade_activate(NULL, NULL, NULL);

} // end function model_undo_activate



void model_clear_activate(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: model_clear_activate
  // Clear all of the MAP volume.

  init_floatarr(float_image_buffer1, origNumSecs*numRows*numCols, 0.0);
  mapShade_activate(NULL, NULL, NULL);

} // end function model_clear_activate



void binning_activate(Widget widget, XtPointer client_data,
		      XtPointer call_data)
{
  // function name: binning_activate
  // Do the binning operation.
  
  float totPixels;
  int i, j, m, n, newHeight, newWidth, numGrid;

  if(doBinning >= 2 && doBinning <= 4) {
    printf("Do %d x %d Binning\n", doBinning, doBinning);
    newHeight = file_height / doBinning;
    newWidth  = file_width / doBinning;
  }
  else {
    askThem("You must choose the Binning Size first.",
	    XmDIALOG_INFORMATION);
    return;
  }
  
  XtUnmanageChild(binningScreenW); 
  XFlush(dsply);
  
  numGrid = doBinning * doBinning;
  
  for (i=0; i<file_height; i+=doBinning) {
    for (j=0; j<file_width; j+=doBinning) {
      if(i+doBinning <file_height && j+doBinning <file_width) {
	totPixels = 0.0;
	for (m=0; m<doBinning; m++) {
	  for (n=0; n<doBinning; n++) {
	    totPixels += inCore_buffer[(i+m)*file_width + j + n] ;
	  }
	}
	inCore_buffer[(((i/doBinning)*newWidth)+(j/doBinning))] = 
	  (float)totPixels / numGrid;
      }
    }
  }
  
  // Now adjust concept of image size
  purdueHeader.numCols = file_width  = newWidth;
  purdueHeader.numRows = file_height = newHeight;
  
  // Make it generate a new fullscreen buffer
  fullScreen_buffer.dispx = fullScreen_buffer.dispy = 0;
  virtual_read_activate(0,0);
  
  DrawWindow(); 
  
} // end function binning_activate



void interpolate_activate(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: interpolate_activate
  // Do the actual interpolation

  int i, j, offset, new_width, new_height, success;
  int iter, thisHeight, thisWidth;
  float maxF, minF, scaleF, range, origSize, outputSize;
  float rx, ry, ix, iy, rotation, scale, filler;
  float *tmpBuf, *thisBuf;

  if(useGUI) XtUnmanageChild(interpScreenW); 
  if(useGUI) XFlush(dsply);

  // Get pixel sizes from interface
  if(useGUI) origSize =  wstrtod(interpOrigTextW);
  if(useGUI) outputSize =  wstrtod(interpOutTextW);

  if(doInterp) { // Do in batch mode
    origSize   = doInterpOrigSize;
    outputSize = doInterpOutputSize;
  }

  if(outputSize == 0.0 || origSize == 0.0 || outputSize < origSize) {
    if(outputSize == 0.0 || origSize == 0.0)
      askThem("Error - Pixel size of 0.0 not allowed.", XmDIALOG_INFORMATION);
    else
      askThem("Error - Output pixSize is greater than original pixSize.",
	      XmDIALOG_INFORMATION);
  }
  else { // OK to go on
    if(fileType == PACKEDPORTABLE) { // DO for packed files
      
      thisHeight = purdueHeader.numRows;
      thisWidth  = purdueHeader.numCols;
      // Now adjust concept of image size
      scale = origSize / outputSize;
      new_width  = thisWidth  * scale;
      new_height = thisHeight * scale;
      
      for(iter=0; iter<purdueHeader.nimg; iter++) { // DO all boxes
	printf("Doing image[%d]\n",iter);
	printf("tw,th[%d,%d]  nw,nh[%d,%d]  scale[%f]\n",
	       thisWidth, thisHeight, new_width, new_height, scale);
	tmpBuf = (float *)malloc(sizeof(float)*thisHeight*thisWidth);
	thisBuf = packImg[iter];
	
	farray_minmax(thisBuf, &minF, &maxF, thisHeight*thisWidth);
	for (i=0; i<thisHeight*thisWidth; i++)
	  tmpBuf[i] = thisBuf[i] ;
	
	range  = maxF - minF;
	offset = (int)minF;
	
	rx = ry = ix = iy = 0.0;
	rotation = 0.0;
	scale = origSize / outputSize;
	filler = tmpBuf[0];
	
	// OK Let's Get rid of some memory to help inter2d out
	free(thisBuf);
	
	success= interp2d(tmpBuf, &thisWidth,  &thisHeight,
			  &rx, &ry, &ix, &iy,
			  &rotation, &scale, &filler);
	printf("Success=[%d]\n",success);
	
	// Now stuff interpolated buffer back into BYTE
	farray_minmax(tmpBuf, &minF, &maxF, thisHeight*thisWidth);
	scaleF = range / (maxF - minF);
	
	if(!success) { // Can't do it
	  askThem("Error - Can't get enough memory.", XmDIALOG_INFORMATION);
	  // Pretend we didn't do anything
	  scale = 1.0;
	}
	
	// OK Let's Get back the memory we freed to help 
	// inter2d out
	thisBuf = packImg[iter] = (float *)malloc(sizeof(float)*
						  new_width*new_height);
	
	for (i=0; i<new_height; i++)
	  for (j=0; j<new_width; j++)
	    thisBuf[i*new_width + j] = offset + 
	      (tmpBuf[i*thisWidth + j]-minF)*scaleF;

	free(tmpBuf);      
      }
      
      // Now adjust concept of image size
      purdueHeader.numCols = thisWidth  = new_width;
      purdueHeader.numRows = thisHeight = new_height;
      
      DrawWindow(); 
      
      if(!doInterp)
	askThem("Finished.  \n", XmDIALOG_INFORMATION);
      
    } // DO for packed files
    else { // normal
      
      tmpBuf = (float *)malloc(sizeof(float)*file_height*file_width);
      
      if(tmpBuf) { // OK to go on
	
	minF = maxF = (float)inCore_buffer[0];
	for (i=0; i<file_height*file_width; i++) {
	  tmpBuf[i] = inCore_buffer[i] ;
	  minF = MIN(minF,tmpBuf[i]);
	  maxF = MAX(maxF,tmpBuf[i]);
	}
	range  = maxF - minF;
	offset = (int)minF;
	      
	rx = ry = ix = iy = 0.0;
	rotation = 0.0;
	scale = origSize / outputSize;
	filler = tmpBuf[0];
	
	// OK Let's Get rid of some memory to help inter2d out
	free(inCore_buffer);
	
	success= interp2d(tmpBuf, &file_width,  &file_height,
			  &rx, &ry, &ix, &iy,
			  &rotation, &scale, &filler);
	printf("Success=[%d]\n",success);
	
	// Now stuff interpolated buffer back into BYTE
	farray_minmax(tmpBuf, &minF, &maxF, file_height*file_width);
	scaleF = range / (maxF - minF);
	
	if(!success) { // Can't do it
	  askThem("Error - Can't get enough memory.", XmDIALOG_INFORMATION);
	  // Pretend we didn't do anything
	  scale = 1.0;
	}
	
	// Now adjust concept of image size
	new_width  = file_width  * scale;
	new_height = file_height * scale;
	
	// OK Let's Get back the memory we freed to help 
	// inter2d out
	inCore_buffer = (float *)malloc(sizeof(float)*
					new_width*new_height);
	
	for (i=0; i<new_height; i++) {
	  for (j=0; j<new_width; j++) {
	    inCore_buffer[i*new_width + j] = 
	      offset + (tmpBuf[i*file_width + j]-minF)*scaleF;
	  }
	}
	
	// Now adjust concept of image size
	purdueHeader.numCols = file_width  = new_width;
	purdueHeader.numRows = file_height = new_height;
	
	free(tmpBuf);
	
	if(success) { // OK to ask them about box coords
	}
	
	// Make it generate a new fullscreen buffer
	fullScreen_buffer.dispx = fullScreen_buffer.dispy = 0;
	virtual_read_activate(0,0);
	
	DrawWindow(); 
	
	if(!doInterp)
	  askThem("Finished.  \n\
    Now you MUST write out this new file before you can box.",
		  XmDIALOG_INFORMATION);
	
      }
      else {
	askThem("Insufficient memory for interpolation",
		XmDIALOG_INFORMATION);
      }
    }
      
  }
  
} // end function interpolate_activate



void addAverage_activate(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: addAverage_activate
  // Add an averaged image to dataset.

  int i, j, ncol, nrow, nimg;

  nrow = purdueHeader.numRows;
  ncol = purdueHeader.numCols;
  nimg = purdueHeader.nimg;

  if(fileType != PACKEDPORTABLE || nimg <= 0) return;
  
  // Allocate memory for averaged image
  freeNullFloat(&packImg[nimg]);
  packImg[nimg]= (float *)malloc(sizeof(float) * ncol * nrow);
  
  init_floatarr(packImg[nimg], ncol*nrow, 0.0);
  
  for (i=0; i<nimg; i++)
    for (j=0; j<nrow*ncol; j++)
      packImg[nimg][j] += packImg[i][j];
  
  for (j=0; j<nrow*ncol; j++)
    packImg[nimg][j] /= nimg;
  
  purdueHeader.nimg++; // Increment for new average image
  stuffPackedImage(ncol, nrow, packImg, nimg, imgNum);
  DrawWindow(); 
  
  sprintf(outStr,"Averaged Image is now image number [%d]",
	  purdueHeader.nimg);
  askThem(outStr, XmDIALOG_INFORMATION);
  
} // end function addAverage_activate



void fourierFilt_activate(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: fourierFilt_activate
  // Setup the fourFilter screen.

  int idim1;
  
  TextSet(resMinTextW, "0.0", FALSE);
  TextSet(cutMinTextW, "0.0", FALSE);
  TextSet(cutMaxTextW, "0.0", FALSE);
  
  if(purdueHeader.numCols > 512)
    idim1 = 1024;
  else if(purdueHeader.numCols > 256)
    idim1 = 512;
  else if(purdueHeader.numCols > 128)
    idim1 = 256;
  else if(purdueHeader.numCols > 64)
    idim1 = 128;
  else 
    idim1 = 64;
  
  // And the value labels
  sprintf(outStr,"FFT size = %d x %d",idim1,idim1);
  doLabel(outStr, fourFiltLabelW, DEFAULT);
  
  TextSetFloat(resMaxTextW, 1.41*idim1, FALSE);
  
  raiseIt(fourFiltScreenW);
  
} // end function fourierFilt_activate



void fourierFilt_work(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: fourierFilt_work
  // Do the fourier filtering.

  int i, j;
  int ncol, nrow, idim1, idim2, imgBkgP;
  float *img, *thisBuf, *fft;
  float resmin, resmax, cutmin, cutmax;

  // This may take awhile so give them the watch cursor
  updateCursor(drawingW, watch_cursor);
  
  ncol = purdueHeader.numCols;
  nrow = purdueHeader.numRows;
  
  if(purdueHeader.numCols > 512)
    idim1 = 1024;
  else if(purdueHeader.numCols > 256)
    idim1 = 512;
  else if(purdueHeader.numCols > 128)
    idim1 = 256;
  else if(purdueHeader.numCols > 64)
    idim1 = 128;
  else 
    idim1 = 64;
  
  idim2 = idim1;

  resmin  = wstrtod(resMinTextW);
  resmax  = wstrtod(resMaxTextW);
  cutmin  = wstrtod(cutMinTextW);
  cutmax  = wstrtod(cutMaxTextW);
  
  CK(img= (float *)malloc(sizeof(float)*ncol*nrow));
  CK(fft= (float *)malloc(sizeof(float)*idim1*idim2*2));
  
  for (i=0; i<purdueHeader.nimg; i++) {
    thisBuf = packImg[i];
    imgBkgP = *(thisBuf);
      
    for(j=0; j<nrow*ncol; j++)
      img[j] = thisBuf[j];
    
    img_filt__(img, &ncol, &nrow, fft, &idim1, &idim2,
	       &imgBkgP, &cutmin, &resmin, &cutmax, &resmax);
    
    for(j=0; j<nrow*ncol; j++)
      thisBuf[j] = img[j]; 
  }
  
  free(img);
  free(fft);
  
  // Always force mode==46 output after we do this!!
  fileType = 46;
  
  // Start at radius'th image
  stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		   packImg,purdueHeader.nimg, imgNum);  
  
  DrawWindow(); 
  
  XtUnmanageChild(fourFiltScreenW);
  
  // Change cursor back to what it was
  updateCursor(drawingW, old_cursor);
  
  askThem("Fourier Filtering finished OK.", XmDIALOG_INFORMATION);
  
} // end function fourierFilt_work



void unblob_text_change(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: unblob_text_change
  // Call here whenever unblob text widgets changed

  if(inited)
    unblobSize = numCols*numRows*origNumSecs*wstrtod(unblobSizeValueW)*0.01;

} // end function unblob_text_change



void unblob_scroll_change(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: unblob_scroll_change
  // Call here whenever unblob scroll widgets changed

  int thisVal;
  float thisFloat;
  int tag = *((int *) client_data);
  XmScrollBarCallbackStruct *temp = (XmScrollBarCallbackStruct *) call_data;
  
  if (tag == UNBLOBRADIUS) {
    thisVal = temp->value;
    sprintf(outStr,"%d",thisVal); 
    radiusUnblob= thisVal;
    doLabel(outStr, unblobRadiusValueW, DEFAULT);
  }
  else if (tag == UNBLOBSIZE) {
    thisFloat = temp->value;
    thisFloat *= 0.001;
    sprintf(outStr,"%.3f%%",thisFloat); 
    // Divide by 100 for percentage 
    unblobSize= numCols*numRows*origNumSecs*thisFloat/100.0;
    TextSet(unblobSizeValueW, outStr, TRUE);
  }
  else if (tag == MINX || tag == MINY || tag == MINZ) {
    int indx_diff;

    if(tag == MINX)
      indx_diff = 0;
    else if(tag == MINY)
      indx_diff = 1;
    else if(tag == MINZ)
      indx_diff = 2;

    thisVal = (int)temp->value;
    sprintf(outStr,"%d",thisVal+1); 

    if(tag == MINY)
      cutMaxValues[indx_diff]= numRows-thisVal;
    else
      cutMinValues[indx_diff]= thisVal;

    doLabel(outStr, cutMinValueW[indx_diff], DEFAULT);

    // Now do percent label 
    if(tag == MINX)
      perMinValues[indx_diff]= (float)thisVal/(float)numCols;
    else if(tag == MINY)
      perMinValues[indx_diff]= (float)thisVal/(float)numRows;
    else if(tag == MINZ)
      perMinValues[indx_diff]= (float)thisVal/(float)origNumSecs;

    sprintf(outStr,"%.1f%% to %.1f%%",perMinValues[indx_diff]*100.0,
	    perMaxValues[indx_diff]*100.0); 
    doLabel(outStr, perValueW[indx_diff], DEFAULT);
  }
  else if (tag == MAXX || tag == MAXY || tag == MAXZ) {
    int indx_diff;

    if(tag == MAXX)
      indx_diff = 0;
    else if(tag == MAXY)
      indx_diff = 1;
    else if(tag == MAXZ)
      indx_diff = 2;

    thisVal = (int)temp->value;

    if(thisVal > 2)
      sprintf(outStr,"%d",thisVal-1); 
    else 
      sprintf(outStr,"%d",1); 

    if(tag == MAXY)
      cutMinValues[indx_diff]= numRows-thisVal;
    else
      cutMaxValues[indx_diff]= thisVal;

    doLabel(outStr, cutMaxValueW[indx_diff], DEFAULT);

    // Now do percent label 
    if(tag == MAXX)
      perMaxValues[indx_diff]= (float)thisVal/(float)numCols;
    else if(tag == MAXY)
      perMaxValues[indx_diff]= (float)thisVal/(float)numRows;
    else if(tag == MAXZ)
      perMaxValues[indx_diff]= (float)thisVal/(float)origNumSecs;
    
    sprintf(outStr,"%.1f%% to %.1f%%",perMinValues[indx_diff]*100.0,
	    perMaxValues[indx_diff]*100.0); 
    doLabel(outStr, perValueW[indx_diff], DEFAULT);
  }
  
  if((temp->reason != XmCR_DRAG)) {
    // Do something
  }
  
  if(immediateMapShade)
    mapShade_activate(NULL, NULL, NULL);  // Don't do if just dragging
  
} // end function unblob_scroll_change



void unblob_tryit_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{
  // function name: unblob_tryit_activate
  // unblob the map
  
  int bCnt, uniqueBlobs, status, killFlag;
  int i, j, k, thisPixel, thatPixel;
  int cutSecs, cutRows, cutCols, clipIt, thisUnblobSize;
  float bgMin, bgMax, minIntensity, maxIntensity;
  float killValue, contMin, contMax;
  float *unblobitBuf, thisFloat;
  blobStructure_t blobList[BMAX];

  // Give feedback
  sprintf(outStr,"Calculating Unblob.");
  TextSet(timeLeftLabelW, outStr, TRUE);

  if(lowRes) { // Don't do this for High Resolution images
    // Clear our temp buffer
    if(unblobBuffer != float_image_buffer1) {
      
      // Don't zero out our gold standard!!!
      if(unblobBuffer != float_image_buffer1)
	init_floatarr(unblobBuffer, origNumSecs*numRows*numCols, 0.0);
      
      if(negUniverseCut) {// Cut out the negative space
	// First stuff the buffer with everything
	
	for (i=0; i<origNumSecs; i++) {
	  for (j=0; j<numRows; j++) {
	    for (k=0; k<numCols; k++) {
	      if(reverseSections) { // Reverse the sections
		unblobBuffer[(origNumSecs-i)*numRows*numCols+
			     (j*numCols)+k] =
		  float_image_buffer1[(i*numRows*numCols)+(j*numCols)+k];
	      } else { // Normal 
		unblobBuffer[(i*numRows*numCols)+(j*numCols)+k] =
		  float_image_buffer1[(i*numRows*numCols)+(j*numCols)+k];
	      }
	    }
	  }
	}
	    
	// Now cut out the negative space
	// Now cut it if appropriate else just stuff it
	
	for (i=cutMinValues[2]; i<cutMaxValues[2]; i++) {
	  for (j=cutMinValues[1]; j<cutMaxValues[1]; j++) {
	    for (k=cutMinValues[0]; k<cutMaxValues[0]; k++) {
	      if(reverseSections) { // Reverse the sections
		unblobBuffer[(origNumSecs-i)*numRows*numCols+
			     (j*numCols)+k] = 0.0;
	      }
	      else { // Normal
		unblobBuffer[(i*numRows*numCols)+(j*numCols)+k] = 0.0;
	      }
	    }
	  }
	}
	
      }// Cut out the negative space
      else {// Normal
	// Now cut it if appropriate else just stuff it
	
	for (i=cutMinValues[2]; i<cutMaxValues[2]; i++) {
	  for (j=cutMinValues[1]; j<cutMaxValues[1]; j++) {
	    for (k=cutMinValues[0]; k<cutMaxValues[0]; k++) {
	      if(reverseSections) { // Reverse the sections
		unblobBuffer[(origNumSecs-i)*numRows*numCols+
			     (j*numCols)+k] =
		  float_image_buffer1[(i*numRows*numCols)+(j*numCols)+k];
	      }
	      else { // Normal
		unblobBuffer[(i*numRows*numCols)+(j*numCols)+k] =
		  float_image_buffer1[(i*numRows*numCols)+(j*numCols)+k];
	      }
	    }
	  }
	}
	
      }// Normal
    }// Don't zero out our gold standard!!!
    else {// unblobBuffer is the gold standard float_image_buffer1
      if(negUniverseCut) {// Cut out the negative space
	// First stuff the buffer with everything
	// Already done - it *IS* the gold std!!
	
	// Now cut out the negative space
	// Now cut it if appropriate else just stuff it
	
	for (i=cutMinValues[2]; i<cutMaxValues[2]; i++) {
	  for (j=cutMinValues[1]; j<cutMaxValues[1]; j++) {
	    for (k=cutMinValues[0]; k<cutMaxValues[0]; k++) {
	      if(reverseSections) { // Reverse the sections
		unblobBuffer[(origNumSecs-i)*numRows*numCols +
			     (j*numCols) + k] = 0.0;
	      }
	      else { // Normal
		unblobBuffer[(i*numRows*numCols)+(j*numCols)+k] = 0.0;
	      }
	    }
	  }
	}
	
      } // Cut out the negative space
      else {// Normal
	// Now zero out the cut stuff if appropriate
	
	for (i=0; i<origNumSecs; i++) {
	  for (j=0; j<numRows; j++) {
	    for (k=0; k<numCols; k++) {
	      if((i < cutMinValues[2] || i >= cutMaxValues[2]) ||
		 (j < cutMinValues[1] || j >= cutMaxValues[1]) ||
		 (k < cutMinValues[0] || k >= cutMaxValues[0])) {
		if(reverseSections) { // Reverse the sections
		  unblobBuffer[(origNumSecs-i)*numRows*numCols+
			       (j*numCols)+k] = 0.0;
		}
		else { // Normal
		  unblobBuffer[(i*numRows*numCols)+(j*numCols)+k] = 0.0;
		}
	      }
	    }
	  }
	}
	
      }// Normal
    }// unblobBuffer is the gold standard float_image_buffer1
  }// Don't do this for High Resolution images
  else {// We've got to stuff the buffer for high res images

    if(unblobBuffer != float_image_buffer1)
      for(i=0; i<origNumSecs*numRows*numCols; i++)
	unblobBuffer[i]= float_image_buffer1[i];
    
  }// We've got to stuff the buffer for high res images
  
  
  if(cleanUnblob) { // Just cut out anything of lesser density
    for (i=0; i<origNumSecs*numRows*numCols; i++) {
      // Use ">" here because of sign difference in TSB density values.
      if(unblobBuffer[i] > maxUnblob)
	unblobBuffer[i] = 0.0;
    }
  }
  
  if(radiusUnblob > 0 || unblobSize > 0) { // They want to unblob
    
    killValue  = wstrtod(unblobFillValueW);
    printf("UB fill [%f]\n",killValue);
    
    contMin = (float)minUnblob*(irev * -1);
    contMax = (float)maxUnblob*(irev * -1);
    if(contMin > contMax) { // swap contMin and contMax
      float tmpF;
      tmpF = contMin;
      contMin = contMax;
      contMax = tmpF;
    }
    
    thisFloat  = wstrtod(unblobSizeValueW);
    unblobSize = numCols*numRows*origNumSecs*thisFloat/100.0;
    
    clipIt = (cutMinValues[0] != 0.0        ||
	      cutMinValues[1] != 0.0        ||
	      cutMinValues[2] != 0.0        ||
	      cutMaxValues[0] != numCols    ||
	      cutMaxValues[1] != numRows    ||
	      cutMaxValues[2] != origNumSecs);
    
    if(negUniverseCut)
      clipIt= FALSE;
    if(!lowRes)
      clipIt = FALSE;
    
    if(clipIt) { // Pass smaller cube to unblob rtns
      cutSecs = cutMaxValues[2] - cutMinValues[2];
      cutRows = cutMaxValues[1] - cutMinValues[1];
      cutCols = cutMaxValues[0] - cutMinValues[0];
      unblobitBuf = NULL;
      CK(unblobitBuf = (float *)malloc(sizeof(float)*
				       cutSecs*cutRows*cutCols));
      if(!unblobitBuf) {
	askThem("[UNBLOB_TRYIT_ACTIVATE]  Error allocating unblobitBuf.",
		XmDIALOG_INFORMATION);
	return;
      }
      
      for (i=0; i<cutSecs; i++) {
	for (j=0; j<cutRows; j++) {
	  for (k=0; k<cutCols; k++) {
	    thisPixel = (i*cutRows*cutCols)+(j*cutCols)+k;
	    thatPixel = (i+cutMinValues[2])*numRows*numCols +
	      (j+cutMinValues[1])*numCols + k+cutMinValues[0]; 
	    unblobitBuf[thisPixel]= unblobBuffer[thatPixel];
	  }
	}
      }
  
    } else { // Stuff with what we've got
      cutSecs = origNumSecs;
      cutRows = numRows;
      cutCols = numCols;
      unblobitBuf = unblobBuffer;
    }
    
    bCnt = BMAX;
    for (i=0; i<BMAX; i++) {
      blobList[i].id     = 0;
      blobList[i].size   = 0;
      blobList[i].minInt = 0.0;
      blobList[i].maxInt = 0.0;
    }
    
    killFlag = FALSE;
    
    thisUnblobSize = unblobSize;
    
    count_kill_blobs_rob__(&bCnt, &uniqueBlobs, &contMin, &contMax,
			  &killFlag, &thisUnblobSize, &killValue,
			  &bgMin, &bgMax, &status, unblobitBuf,
			  &cutCols, &cutRows, &cutSecs, blobList,
			  &radiusUnblob, &cleanUnblob);
    
    getblobminmaxint_rob__(blobList, &bCnt, &thisUnblobSize, &killValue, 
			  &minIntensity, &maxIntensity);
    
    minIntensity = MIN(bgMin,minIntensity);
    maxIntensity = MAX(bgMax,maxIntensity);
    
    killFlag = TRUE;
    count_kill_blobs_rob__(&bCnt, &uniqueBlobs, &contMin, &contMax,
			  &killFlag, &thisUnblobSize, &killValue,
			  &bgMin, &bgMax, &status, unblobitBuf,
			  &cutCols, &cutRows, &cutSecs, blobList,
			  &radiusUnblob, &cleanUnblob);
    
    if(clipIt) {
      for (i=0; i<cutSecs; i++) {
	for (j=0; j<cutRows; j++) {
	  for (k=0; k<cutCols; k++) {
	    thisPixel = (i*cutRows*cutCols)+(j*cutCols)+k;
	    thatPixel = (i+cutMinValues[2])*numRows*numCols +
	      (j+cutMinValues[1])*numCols + k+cutMinValues[0];
	    unblobBuffer[thatPixel]= unblobitBuf[thisPixel];
	  }
	}
      }
      freeNullFloat(&unblobitBuf);
    }
    
  }
  
} // end function unblob_tryit_activate



void unblob_reset_activate(Widget widget, XtPointer client_data, 
			   XtPointer call_data)
{
  // function name: unblob_reset_activate
  // Go back to the last gold STD map to undo any unblobbing.
  
  XtVaSetValues(unblobSizeScrollW, XmNvalue, 0, NULL);
  XFlush(dsply);
  TextSet(unblobSizeValueW, "0.000 %%", TRUE); 
  unblobSize = 0;
  
  if(immediateMapShade)
    mapShade_activate(NULL, NULL, NULL);
  
} // end function unblob_reset_activate



void unblob_save_activate(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: unblob_save_activate
  // Save this unblobbed MAP as the new gold STD MAP.
  
  int i;
  
  if(askThem("This operation destructively overwrites\n\
the Map in core with the UNBLOB Map.  You must\n\
reread the disk Map file to undo this.", XmDIALOG_QUESTION)) {
    
    // Calculate the unblob buffer
    CK(unblobBuffer=(float *)malloc(sizeof(float)*
				    numCols*numRows*origNumSecs));
    
    unblobSwitch = TRUE;
    if(!unblobBuffer) {
      if(askThem("Insufficient memory - do you want to unblob the\n\
Gold Standard memory buffer?  (This is a \n\
destructive operation - you'll have to re-read\n\
from disk to undo this.)", XmDIALOG_QUESTION))
	unblobBuffer = float_image_buffer1;
      else
	unblobSwitch = FALSE;
    }
    
    if(unblobSwitch) { // unblob it
      unblob_tryit_activate(0,0,0);
      
      if(unblobBuffer != float_image_buffer1) { // Save our temp buffer
	for (i=0; i<origNumSecs*numRows*numCols; i++) {
	  float_image_buffer1[i] = unblobBuffer[i];
	}
	free(unblobBuffer);
      }
      
    }
    unblobBuffer = NULL;
    unblobSwitch = FALSE;
    
    TextSet(timeLeftLabelW, "Select parameters.", TRUE);
  }
  
} // end function unblob_save_activate



void reverse_sections(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: reverse_sections
  // Set switches to do 3D Rendering properly.

  if(widget == unblobToggleW)
    cleanUnblob = !cleanUnblob;
  else if(widget == mapIcosTogW) {
    dispIcosa = !dispIcosa;
    if(!dispIcosa) { // Init the display
      int i, sx, sy, ex, ey, xoffset, yoffset;
      xoffset = 500;
      yoffset = 290;
      sx = xoffset - 35;
      ex = xoffset + 35;
      for (i=0; i<70; i++) {
	sy = (yoffset - 35) + i;
	ey = sy;
	XDrawLine(dsply, XtWindow(mapShadeScreenW), image_gcGREY[100],
		  sx, sy, ex, ey);
      }
    }
  }
  else if(widget == radProjToggleW)
    revRadProj = !revRadProj;
  else if(widget == icosRadProjToggleW) {
    icosRadProj = !icosRadProj;
    raiseIt(dioScreenW);
  }
  else if(widget == radialCSToggleW)
    radialCS = !radialCS;
  else if(widget == radProjCSToggleW)
    radProjCS = !radProjCS;
  
} // end function reverse_sections



void fixit_changed(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: fixit_changed
  // Change properties of FFT scrollbars

  if(widget == minFixTextW)
    minFixValue = wstrtod(widget);
  else if (widget == maxFixTextW)
    maxFixValue = wstrtod(widget);
  else if (widget == pageFixTextW)
    pageFixValue = wstrtod(widget);
  else if (widget == incFixTextW)
    incFixValue = wstrtod(widget);
  
} // end function fixit_changed



void fixit_activate(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: fixit_activate
  // Setup the fixit screen on display.

  if(widget == fftMinFixW)
    fixitW = minFFTscrollW;
  else if(widget == fftMaxFixW)
    fixitW = maxFFTscrollW;

  XtSetArg(arglst[0], XmNminimum, &minFixValue);
  XtSetArg(arglst[1], XmNmaximum, &maxFixValue);
  XtSetArg(arglst[2], XmNincrement, &incFixValue);
  XtSetArg(arglst[3], XmNpageIncrement, &pageFixValue);
  XtGetValues(fixitW, arglst, 4);
  XFlush(dsply);
  
  TextSetInt(minFixTextW,  minFixValue,  FALSE);
  TextSetInt(maxFixTextW,  maxFixValue,  FALSE);
  TextSetInt(pageFixTextW, pageFixValue, FALSE);
  TextSetInt(incFixTextW,  incFixValue,  TRUE);
  
  raiseIt(fixitScreenW);
  
} // end function fixit_activate



void movie_contour_activate(Widget widget, XtPointer client_data, 
			    XtPointer call_data)
{
  // function name: movie_contour_activate
  // Get movie Contour screen up on display.

  // Setup the scrolls accordingly. Always use the section min and max
  // because we cue off of the sections to contour shaded views

  XtSetArg(arglst[0], XmNmaximum, (int)(secMaxPixel+5.0));
  XtSetArg(arglst[1], XmNminimum, (int)(secMinPixel));
  XtSetArg(arglst[2], XmNsliderSize, 5);
  XtSetArg(arglst[3], XmNvalue, 0);
  XtSetArg(arglst[4], XmNpageIncrement, 5);
  XtSetArg(arglst[5], XmNincrement, 1);
  XtSetValues(movconConScrollW, arglst, 6);
  XFlush(dsply);
  
  // Add the value labels
  sprintf(outStr,"%d", 0);
  doLabel(outStr, movconConLW, DEFAULT);
  
  raiseIt(movconScreenW);
  
} // end function movie_contour_activate



void histEqualize(int dim1, int dim2, unsigned char *image, 
		  int firstPixel, int ncolors, unsigned char *outBuffer,
		  int min, int max, int buffer_width)
{
  // function name: histEqualize
  // Gotten from EMMA in hist.c module. Sept 1997

   int i, j, k, *hist;
   unsigned short level, *short_image;
   float *prob, *sum, scaleFactor;

   CK(hist = (int *)malloc(ncolors*sizeof(int)));
   CK(prob = (float *)malloc(ncolors*sizeof(float)));
   CK(sum = (float *)malloc(ncolors*sizeof(float)));

   init_intarr(hist, ncolors, 0);     // initialize arrays
   init_floatarr(prob, ncolors, 0.0); 
   init_floatarr(sum, ncolors, 0.0);
   
   // dcm  This code is replaced by the following
   // for(j=0;j<dim1;j++)
   // for(k=0;k<dim2;k++)
   // hist[image[k*dim1+j]]++;
   
   short_image = (unsigned short *)image;
   max = 65500;
   scaleFactor= (float)(ncolors-1)/(float)(max-min);
   for (k=0; k<dim2; k++)
     for (j=0; j<dim1; j++)
       hist[(int)((short_image[k*buffer_width+j]-min)*scaleFactor)]++;
   
   for (i=0; i<ncolors; i++)
     prob[i] = (float)hist[i]/(float)(dim1*dim2);
   
   for (i=0; i<ncolors; i++)
     for (j=0; j<=i; j++)
       sum[i] += prob[j];
   
   for (i=0; i<ncolors; i++)
     level = (unsigned short)(0.5+(ncolors-1)*sum[i]);

   // dcm  j and k loops have been reversed
   for (k=0; k<dim2; k++) {
     int kTdim1, kTdim1Pj, thisVal;
     kTdim1 = k*buffer_width;
     for (j=0; j<dim1; j++) {
       kTdim1Pj = kTdim1 + j;
       level = (unsigned short)
	 (0.5+(ncolors-1)*sum[(int)
			      ((short_image[kTdim1Pj]-min)*scaleFactor)]);
       if(level >= ncolors)
	 level = ncolors-1;
       
       // transfImage[currImage][kTdim1+j] = (float)level;
       thisVal = (int)((float)firstPixel + (float)level);
       XPutPixel(image_image1, j, k, pixels_return[thisVal]);
       histogram[thisVal]++;
     }
   }
   
   free(hist);
   free(prob); 
   free(sum);
   
} // end function histEqualize



void updateDimensions()
{
  // function name: updateDimensions
  // Change dims on truncate or addto
  
  int diffVal;
  
  diffVal = (int)wstrtod(taEndXW) - (int)wstrtod(taBegXW) + 1;
  if(diffVal%2 == 0)
    sprintf(outStr,"%d Columns EVEN",diffVal);
  else 
    sprintf(outStr,"%d Columns",diffVal);
  doLabel(outStr, taDimXW, FALSE);
  
  diffVal = (int)wstrtod(taEndYW) - (int)wstrtod(taBegYW) + 1;
  if(diffVal%2 == 0)
    sprintf(outStr,"%d Rows EVEN",diffVal);
  else 
    sprintf(outStr,"%d Rows",diffVal);
  doLabel(outStr, taDimYW, FALSE);
  
  diffVal = (int)wstrtod(taEndZW) - (int)wstrtod(taBegZW) + 1;
  if(diffVal % 2 == 0)
    sprintf(outStr,"%d Sections EVEN",diffVal);
  else 
    sprintf(outStr,"%d Sections",diffVal);
  if(fileType!=PACKEDPORTABLE)
    doLabel(outStr, taDimZW, FALSE);

  XFlush(dsply);

} // end function updateDimensions



void gbsArrowActivateButton(Widget widget, XEvent *event)
{
  // function name: gbsArrowActivateButton
  // Global box size arrows

  int begVal, endVal, speedFactor;

  if(locked)
    return;

  speedFactor = GetButtonVal(event->xbutton.button, 1, 2, 5);

  if(widget==gbsDownArrowW)
    speedFactor *= -1;
  
  begVal  = (int)wstrtod(taBegXW) - speedFactor;
  TextSetInt(taBegXW, begVal, FALSE); 
  endVal  = (int)wstrtod(taEndXW) + speedFactor;
  TextSetInt(taEndXW, endVal, FALSE); 
  
  begVal  = (int)wstrtod(taBegYW) - speedFactor;
  TextSetInt(taBegYW, begVal, FALSE); 
  endVal  = (int)wstrtod(taEndYW) + speedFactor;
  TextSetInt(taEndYW, endVal, FALSE); 
  
  if(fileType!=PACKEDPORTABLE) {
    begVal  = (int)wstrtod(taBegZW) - speedFactor;
    TextSetInt(taBegZW, begVal, FALSE); 
    endVal  = (int)wstrtod(taEndZW) + speedFactor;
    TextSetInt(taEndZW, endVal, FALSE); 
  }
  
  updateDimensions();
    
} // end function gbsArrowActivateButton



void truncadd_half_activate(Widget widget, XtPointer client_data, 
				  XtPointer call_data)
{
  // function name: truncadd_half_activate
  // Truncate or add to half of the dataset.

  if(widget == taFrontW  && fileType != PACKEDPORTABLE) {
    TextSet(taBegZW, "1", FALSE); 
    TextSetInt(taEndZW, (origNumSecs/2)+1, FALSE); 
  }
  else if(widget == taBackW  && fileType != PACKEDPORTABLE) {
    TextSetInt(taBegZW, (origNumSecs/2)+1, FALSE); 
    TextSetInt(taEndZW, largestDim, FALSE); 
  }
  else if(widget == taLeftW) {
    TextSet(taBegXW, "1", FALSE); 
    TextSetInt(taEndXW, (numCols/2)+1, FALSE); 
  }
  else if(widget == taRightW) {
    TextSetInt(taBegXW, (numCols/2)+1, FALSE); 
    TextSetInt(taEndXW, largestDim, FALSE); 
  }
  else if(widget == taTopW) {
    TextSet(taBegYW, "1", FALSE); 
    TextSetInt(taEndYW, (numRows/2)+1, FALSE); 
  }
  else if(widget == taBottomW) {
    TextSetInt(taBegYW, (numRows/2)+1, FALSE); 
    TextSetInt(taEndYW, largestDim, FALSE); 
  }
  else if(widget == taWholeW) {
    TextSet(taBegXW, "1", FALSE); 
    TextSet(taBegYW, "1", FALSE); 
    if (fileType != PACKEDPORTABLE)
      TextSet(taBegZW, "1", FALSE); 
    TextSetInt(taEndXW, largestDim, FALSE); 
    TextSetInt(taEndYW, largestDim, FALSE); 
    if (fileType != PACKEDPORTABLE)
      TextSetInt(taEndZW, largestDim, FALSE); 
  }
  
  updateDimensions();
  XFlush(dsply);
  
} // end function truncadd_half_activate



void truncadd_activate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: truncadd_activate
  // Get the truncate/addto screen up.
  
  TextSet(taBegXW, "1", FALSE); 
  TextSet(taBegYW, "1", FALSE); 
  TextSet(taBegZW, "1", FALSE); 
  if(fileType == PACKEDPORTABLE)
    sprintf(outStr,"%d",purdueHeader.numCols);
  else
    sprintf(outStr,"%d",largestDim);
  TextSet(taEndXW, outStr, FALSE); 
  if(fileType == PACKEDPORTABLE)
    sprintf(outStr,"%d",purdueHeader.numRows);
  TextSet(taEndYW, outStr, FALSE); 
  if(fileType == PACKEDPORTABLE)
    sprintf(outStr,"%d",purdueHeader.nimg);
  TextSet(taEndZW, outStr, FALSE);
  
  if(fileType == PACKEDPORTABLE) // Turn off z plane
    XtSetArg(arglst[0], XmNsensitive, FALSE);
  else // Turn on z plane
    XtSetArg(arglst[0], XmNsensitive, TRUE);
  
  XtSetValues(taEndZW, arglst, 1);  
  XtSetValues(taBegZW, arglst, 1);  
  
  sprintf(outStr,"%.2f",(float)pifHeader.img_bkg);
  if(EQU(outStr,"nan"))
    sprintf(outStr,"0.0");
  TextSet(taFillW, outStr, FALSE); 
  updateDimensions();
  
  raiseIt(truncaddScreenW);
  mapshade_deactivate(NULL, NULL, NULL);
  
} // end function truncadd_activate



void truncadd_do_activate(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: truncadd_do_activate
  // Actually do the truncation or the adding to.
  
  int i, j, k;
  int begX, begY, begZ, endX, endY, endZ, thisX, thisY, thisZ;
  int centerX, centerY, centerZ, origX, origY, origZ;
  int residX, residY, residZ, newNumCols, newNumRows, newNumSecs;
  float thisPixel, fillValue, *temp_float_buffer;

  if(doTrunc) { // Use values specified on cmd line with -trunc
    begX = truncBegX;
    endX = truncEndX;
    begY = truncBegY;
    endY = truncEndY;
    begZ = truncBegZ;
    endZ = truncEndZ;
    fillValue = truncVal;
  } else {     // Use values set in GUI
    begX = wstrtod(taBegXW);
    begY = wstrtod(taBegYW);
    begZ = wstrtod(taBegZW);
    endX = wstrtod(taEndXW);
    endY = wstrtod(taEndYW);
    endZ = wstrtod(taEndZW);
    fillValue = wstrtod(taFillW);
    printf("Fill value = [%f]\n",fillValue);
  }

// GCA: it does not seem relevant to set centerX/Y/Z properly
// since they are used only for residX/Y/Z, which are not used for anything 
  centerX = (begX+endX)/2; // GCA: arbitrary value
  centerY = (begY+endY)/2; // GCA: arbitrary value
  centerZ = (begZ+endZ)/2; // GCA: arbitrary value
  // Humans number from 1 not zero
  begX--;
  begY--;
  begZ--;
  centerX--; // RSS - Note that center[XYZ] are never set
  centerY--;
  centerZ--;

  newNumCols = endX - begX;
  newNumRows = endY - begY;
  newNumSecs = endZ - begZ;
  numCols = purdueHeader.numCols;
  featherRadius += (newNumCols - numCols)/2;
  boxRadius     += (newNumCols - numCols)/2;
  
  // Y is a special case here - Tim numbers row starting from the bottom.

  if(fileType == PACKEDPORTABLE)
    newNumSecs = 1;  // We don't have secs here
  
  if(newNumCols <= 0 || newNumRows <=0 || newNumSecs <=0) {
    sprintf(outStr,"Can't do this:  [cols=%d,rows=%d,secs=%d]",
	    newNumCols, newNumRows, newNumSecs);
    if(doTrunc)
      printf("%s\n",outStr);
    else
      askThem(outStr, XmDIALOG_INFORMATION);
    return;
  }
  
  if(fileType == PACKEDPORTABLE) {
    int m, nc, nr;
    float *thisPackImg, *tmpInt;
    
    // Now calc how far we are from original
    residY = centerY - (purdueHeader.numRows/2);
    residX = centerX - (purdueHeader.numCols/2);
    
    nc = purdueHeader.numCols;
    nr = purdueHeader.numRows;
    for (m=0; m<purdueHeader.nimg; m++) {
      CK(tmpInt = (float *)malloc(sizeof(float)*newNumCols*newNumRows));
      
      for (j=0; j<newNumRows*newNumCols; j++)
	tmpInt[j] = fillValue;
      
      thisPackImg = packImg[m];
      fillValue = thisPackImg[0];
      printf("Fill value is bkg value = [%f]\n",fillValue);
      
      for (j=begY; j<endY; j++) {
	 // TSB orientation :
	// thisY = (newNumRows-(j-begY)-1)*newNumCols;
	// origY = (nr - j) - 1 ;
	// We've changed it to be XWindows coords as the code
	// in the 3D Rendering screen will be VERY challenging
	// to make that consistent with this convention.
	// RWa March 2004 
	// * RWA April 2004 Here's the code to make it XWindows:
	// thisY = (begY-1)*newNumCols;
	// origY = j;
	// I've changedf back to TSB orientation for all
	// in robem

	thisY = (newNumRows-(j-begY)-1)*newNumCols;
	origY = nr - j - 1 ;
	
	for (k=begX; k<endX; k++) {
	  thisX = k-begX;
	  origX = k;
	  
	  if(origY<nr && origX<nc && origY>=0 && origX>=0)
	    tmpInt[thisY+thisX] = thisPackImg[(origY*nc)+origX];
	  else
	    tmpInt[thisY+thisX] = fillValue;

	  thisPixel = tmpInt[thisY+thisX];
	}
      }
      
      packImg[m] = tmpInt;
      free(thisPackImg);
    }
    
    // Now clip this guy to our colormap
    if(minPackPixel < 0.0)
      scalePackFactor = (tryColors - 7.0)/(maxPackPixel-minPackPixel);
    else 
      scalePackFactor = (tryColors - 7.0)/maxPackPixel;

    printf("minPackPixel [%f]  max[%f]     scale[%f]\n",
	   minPackPixel, maxPackPixel, scalePackFactor);
    
    purdueHeader.numCols = newNumCols;
    purdueHeader.numRows = newNumRows;
    packRadius = purdueHeader.numCols/2;
    stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		     packImg,purdueHeader.nimg, imgNum);
    DrawWindow(); 
    
    if(useGUI) XtUnmanageChild(truncaddScreenW);
    
    sprintf(outStr, "Boxes successfully truncated/added to");

    if(doTrunc)
      printf("%s\n",outStr);
    else
      askThem(outStr, XmDIALOG_INFORMATION);
    
    // Give feedback on main window title bar
    sprintf(outStr,"Robem: [%s] - [%d,%d,%d]",fileName.addr,
	    purdueHeader.numCols,purdueHeader.numRows,
	    purdueHeader.nimg);
    
    doDialogTitle(outStr, mainWindowW, FALSE);
  }
  else { // MAP
    
    // Now calc how far we are from original
    residZ = centerZ - origNumSecs/2;
    residY = centerY - numRows/2;
    residX = centerX - numCols/2;
    
    temp_float_buffer = float_image_buffer1;
    CK(float_image_buffer1 = (float *)malloc(sizeof(float)*
					     newNumCols*
					     newNumRows*
					     newNumSecs));
    
    maxMapPixel = -HUGE;
    minMapPixel = HUGE;

    for (i=begZ; i<endZ; i++) {
      thisZ = (i-begZ)*newNumCols*newNumRows;
      origZ = i;
      for (j=begY; j<endY; j++) {
	      
	// Original way :  
	// thisY = (j-begY)*newNumCols;
	// origY = j; 
	
	// TSB orientation :
	// thisY = (newNumRows-(j-begY)-1)*newNumCols;
	// origY = (numRows - j) - 1 ;
	// We've changed it to be XWindows coords as the code
	// in the 3D Rendering screen will be VERY challenging
	// to make that consistent with this convention.
	// RWa March 2004 
	// RWA April 2004 Here's the code to make it XWindows:
	// thisY = (j-begY)*newNumCols;
	// origY = j;
	//  I've changedf back to TSB orientation for all in robem

	thisY = (newNumRows-(j-begY)-1)*newNumCols;
	origY = (numRows - j) - 1 ;
	
	for (k=begX; k<endX; k++) {
	  thisX = k-begX;
	  origX = k;
	  
	  if(origZ < origNumSecs && origY < numRows &&
	     origX < numCols     && origZ >= 0      && 
	     origY >= 0          && origX >= 0)
	    thisPixel= float_image_buffer1[thisZ+thisY+thisX] =
	      temp_float_buffer[origZ*numCols*numRows +
				origY*numCols + origX];
	  else
	    thisPixel= 
	      float_image_buffer1[thisZ+thisY+thisX] = fillValue;
	  
	  if(thisPixel > maxMapPixel)
	    maxMapPixel= thisPixel;
	  if(thisPixel < minMapPixel)
	    minMapPixel= thisPixel;
	  
	}
      }
    }
    
    // Now get rid of the old
    free(temp_float_buffer);
    
    // Keep track for traslation stuff
    pifHeader.transX += begX;
    pifHeader.transY += begY;
    pifHeader.transZ += begZ;
    if(!pifHeader.isTrunc) { // First truncation
	pifHeader.origX= numCols;
	pifHeader.origY= numRows;
	pifHeader.origZ= numSecs;
    }
    pifHeader.isTrunc = TRUE;

    numCols = newNumCols;
    numRows = newNumRows;
    numSecs = newNumSecs;
    origNumSecs= numSecs;
    
    // Now remember these forever
    lowRes_buffer1 = float_image_buffer1;
    lowResCols = numCols;
    lowResRows = numRows;
    lowResSecs = numSecs;
    
    // Remember how big we are on disk
    file_width  = numCols;
    file_height = numRows;
    
    purdueHeader.numCols = numCols;
    purdueHeader.numRows = numRows;
    purdueHeader.numSecs = origNumSecs;
    
    // For writing new file
    fileBinSecs  = numSecs;
    fileBinRows  = numRows;
    fileBinCols  = numCols;
    
    xCen = numCols/2;
    yCen = numRows/2;
    zCen = origNumSecs/2;
    
    largestDim = MAX(numCols, numRows);
    largestDim = MAX(largestDim, origNumSecs);
    
    nsecs= nimg= numSecs;
    
    // We've got to make this larger for asymmetric maps 
    // and all maps that have information "outside" of the
    // normal sphere of our normal maps,
    // or else we just keep cliping the box into a sphere
    
    globMaxRadius = (int)(origNumSecs/2) - 1;

    globMinRadius = 0.0;
    
    // Set scrolls on depthcue screen  and give feedback
    sprintf(outStr,"%.0f",globMinRadius);
    doValue(outStr, minRadiusValueW, FALSE);
    
    sprintf(outStr,"%.0f",globMaxRadius);
    doValue(outStr, maxRadiusValueW, FALSE);
    
    // Radius is origNumSecs/2
    if(useGUI) {
      XtVaSetValues(maxRadiusScrollW, XmNsliderSize, 5, NULL);
      XtVaSetValues(maxRadiusScrollW, XmNmaximum, 
		    (int)((globMaxRadius)+5),NULL);
      XtVaSetValues(maxRadiusScrollW, XmNvalue, (int)(globMaxRadius),NULL);
      
      XtVaSetValues(minRadiusScrollW, XmNsliderSize, 5,NULL);
      XtVaSetValues(minRadiusScrollW, XmNmaximum, 
		    (int)((globMaxRadius)+5),NULL);
      XtVaSetValues(minRadiusScrollW, XmNvalue, (int)(globMinRadius),NULL);
    }
    
    printf("MAX[%f]  MIN[%f]\n",maxMapPixel,minMapPixel);
    secMaxPixel = maxMapPixel;
    secMinPixel = minMapPixel;
    
    minSurf = MAX(minMapPixel,0.0);
    maxSurf = MAX(maxMapPixel,(0.0-minMapPixel));

    // We're doing the surfaceDensity in tenths now
    if(useGUI) {
      XtVaSetValues(surfDensScrollW, XmNmaximum, maxSurf*10 + 50, NULL);
      XtVaSetValues(surfDensScrollW, XmNminimum, minSurf*10, NULL);
      XtVaSetValues(surfDensScrollW, XmNvalue, 10, NULL);
    
      if(maxSurf > 5000.0) {
	XtVaSetValues(surfDensScrollW, XmNsliderSize, 
		      (int)(maxSurf*0.01), NULL);
	XtVaSetValues(surfDensScrollW, XmNpageIncrement, 
		      (int)(maxSurf*0.01), NULL);
	XtVaSetValues(surfDensScrollW, XmNincrement, 
		      (int)(maxSurf*0.002), NULL);
      }
      else if(maxSurf > 500.0) {
	XtVaSetValues(surfDensScrollW, XmNsliderSize, 500, NULL);
	XtVaSetValues(surfDensScrollW, XmNpageIncrement, 500,NULL);
	XtVaSetValues(surfDensScrollW, XmNincrement, 100,NULL);
      }
      else {
	XtVaSetValues(surfDensScrollW, XmNsliderSize, 50, NULL);
	XtVaSetValues(surfDensScrollW, XmNpageIncrement, 50, NULL);
	XtVaSetValues(surfDensScrollW, XmNincrement, 10, NULL);
      }
    }

    surfaceDensity= 1.0;  // 1 tenth of 10
    
    // Give feedback
    sprintf(outStr,"%d",-1);
    doValue(outStr, surfDensValueW, FALSE);
    
    doLabel("SecNum", imgNumLabelW, FALSE);
    
    TextSet(imgNumTextW, "1", FALSE);
    
    // To make this work for now let's stuff the nimg field with the
    // number of sections.  Really we should change any and all references
    // to nimg to be origNumSecs. We really only work on one image at a 
    // time so this may never have any affect.
    
    nsecs= purdueHeader.nimg = origNumSecs;
    
    // Write stats
    calcMapStats(float_image_buffer1, &purdueHeader.img_bkg, 
		 &purdueHeader.map_min, &purdueHeader.map_max, 
		 &purdueHeader.img_avg, &purdueHeader.img_std,
		 &numCols, &numRows, &origNumSecs);

    sprintf(outStr,"\nStatistics for file [%s]\n\
\n\
Title [%s]\n\
\n\
File size   [%d]x[%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
\n",
	    fileName.addr,
	    title,
	    numCols, numRows, origNumSecs,
	    purdueHeader.img_bkg, minMapPixel, maxMapPixel,
	    purdueHeader.img_avg, purdueHeader.img_std);

    TextSet(statTextW, outStr, TRUE);
    if(useGUI) XtUnmanageChild(truncaddScreenW);
    
    sprintf(outStr, "Map successfully truncated/added to");
    
    if(doTrunc)
      printf("%s\n",outStr);
    else
      askThem(outStr, XmDIALOG_INFORMATION);
    
    // Their shaded views are no longer valid - show new sections
    shadedMaps= FALSE;
    dispSections = TRUE;
    if(useGUI) {
      XtSetArg(arglst[0], XmNlabelPixmap, sectionPixmap);
      XtSetArg(arglst[1], XmNselectPixmap, unsectionPixmap);
      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
      XtSetValues(sectionW, arglst, 3);
    }
    if(dispSections) {
      // Start at zero'th image
      stuffMapImage(map_image_buffer, float_image_buffer1, 
		    file_width, file_height, nsecs, imgNum, 0);  
      DrawWindow(); 
    }
    
    // Give feedback on main window title bar
    sprintf(outStr,"Robem: [%s] - [%d,%d,%d]",fileName.addr,
	    numCols,numRows,numSecs);
    
    doDialogTitle(outStr, mainWindowW, FALSE);
  }
  
} // end function truncadd_do_activate



void rotate_do_activate(Widget widget, XtPointer client_data, 
			XtPointer call_data) 
{
  // function name: rotate_do_activate
  // Do the rotate and then display it.

  int  two=2;
  float rotateTheta  = -wstrtod(rotateThetaW); // Note minus sign
  float rotatePhi    = wstrtod(rotatePhiW);
  float rotateOmega  = wstrtod(rotateOmegaW);
  
  phys_map_rotate(float_image_buffer1, float_image_buffer1,
		  rotateTheta, rotatePhi, rotateOmega);
  
  if(dispSections) { // Start at zero'th image
    stuffMapImage(map_image_buffer, float_image_buffer1, 
		  file_width, file_height, nsecs, imgNum, 0);  
    DrawWindow(); 
  }
  else
    foldAxis_activate(widget, &two, call_data); // Force 2-fold display
  
} // end function rotate_do_activate



void phys_map_rotate(float *gold_buffer, float *temp_buffer,
		     float theta, float phi, float omega)
{
  // function name: phys_map_rotate
  // Actually rotate the MAP in memory.
  // RSS Make sure that the choice for axis is correct

  char axis;
  int i;
  
  if(gold_buffer != temp_buffer)
    for (i=0; i<origNumSecs*numRows*numCols; i++)
      temp_buffer[i] = gold_buffer[i];
  
  axis = 'Z'; // Do Phi
  if(phi != 0.0)
    map_rot__(temp_buffer, &numCols, &numRows, &origNumSecs, &phi, &axis);
  
  axis = 'Y'; // Do Theta
  if(theta != 0.0)
    map_rot__(temp_buffer, &numCols, &numRows, &origNumSecs, &theta, &axis);

  axis = 'Z'; // Do Omega
  if(omega != 0.0)
    map_rot__(temp_buffer, &numCols, &numRows, &origNumSecs, &omega, &axis);

} // end function phys_map_rotate



void new_phys_map_rotate(float *gold_buffer, float *temp_buffer,
			 float theta, float phi, float omega)
{
  // function name: new_phys_map_rotate
  // Better way to rotate???

  int  i, j, k, yOff, zOff, xOff, ncol, nrow;
  double t1, t2, t3, t4, t5, t6, t7, t8, t9;
  double cphi, sphi, ctheta, stheta, comega, somega;
  double thisCol, thisRow, thisSecn;
  double xCen, yCen, zCen;

  // We've got a bug whereby we've got a lot of problems with
  // the original map_prj_all() rtn that does the rotation.
  // WHY does phi not have to have the sign reversed here like
  // it does in emUnix.c in the thetaPhi_dphshd_C() rtn???
  
  comega = cos(deg2rad(phi));
  somega = sin(deg2rad(phi));
  cphi   = cos(deg2rad(omega));
  sphi   = sin(deg2rad(omega));
  ctheta = cos(deg2rad(0.0-theta));
  stheta = sin(deg2rad(0.0-theta));
  
  //  COMPUTE ELEMENTS OF ROTATION MATRIX
  t1 = ctheta*cphi*comega + sphi*somega;
  t2 = ctheta*sphi*comega - cphi*somega;
  t3 = -stheta*comega;
  t4 = ctheta*cphi*somega - sphi*comega;
  t5 = ctheta*sphi*somega + cphi*comega;
  t6 = -stheta*somega;
  t7 = stheta*cphi;
  t8 = stheta*sphi;
  t9 = ctheta;
  
  ncol = numCols;
  nrow = numRows;
  xCen = numCols/2;
  yCen = numRows/2;
  zCen = origNumSecs/2;

  for (i=0; i<origNumSecs; i++) {
    zOff = (i*numCols*numRows);
    for (j=0; j<numRows; j++) {
      yOff = zOff+(j*numCols);
      for (k=0; k<numCols; k++) {
	xOff = yOff + k;
	newTranslateThetaPhiOmega(&thisCol, &thisRow, &thisSecn,
				  xCen, yCen, zCen, k, j, i,
				  t1, t2, t3, t4, t5, t6, t7, t8, t9);

	if(thisCol >= 1.0 && thisCol <= (numCols-2) &&
	   thisRow >= 1.0 && thisRow <= (numRows-2) &&
	   thisSecn >= 1.0 && thisSecn <= (origNumSecs-2))
	  temp_buffer[xOff] = getVoxel(gold_buffer,
				       thisCol, thisRow, thisSecn,
				       theta, phi, omega, ncol, nrow);
	else 
	  temp_buffer[xOff] = 0.0;
	
      }
    }
  }

} // end function new_phys_map_rotate



void fix2D_flicker_activate(Widget widget, XtPointer client_data, 
			    XtPointer call_data)
{
  // function name: fix2D_flicker_activate
  // Flicker the new fixed 2D image with what was on disk.

  flickerFix2D = *((int *) client_data);
  
  if(fileType==PACKEDPORTABLE)
    // Start at radius'th image
    stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		     packImg, purdueHeader.nimg, imgNum);  
  else {
    // Force a recalc of full screen
    fullScreen_buffer.dispx = 0;
    fullScreen_buffer.dispy = 0;
    virtual_read_activate(-99,-99);
  }
  
  DrawWindow();
  
} // end function fix2D_flicker_activate



void fix2D_do_activate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: fix2D_do_activate
  // Do the fixing of this image

  int  i, j, k, fixRegions=1;
  float cutHigh, cutLow;
  float fixStd=5.0;
  float thisBkg, thisMin, thisMax, thisAvg, thisStd;

  int tag = *((int *) client_data);

  raiseIt(statusScreenW);
  checkStatus("Setting up to fix values.");
  
  if(useGUI) fixRegions = (int)wstrtod(fix2DRegionsTextW);
  if(useGUI) fixStd     = wstrtod(fix2DStdTextW);

  // We're in batch mode and they've asked for a specific stDev
  if(doFix && batchFixStd != 0.0)
    fixStd = batchFixStd;
  
  fixRegions = MAX(MIN(fixRegions,MAXREGIONS),1);
  fixStd     = MAX(MIN(fixStd,100.0),0.1);
  printf("Using stDev==[%f]\n",fixStd);
  
  if(fileType==PACKEDPORTABLE)
    fixRegions = 1;
  
  TextSetInt(fix2DRegionsTextW, fixRegions, FALSE);
  sprintf(outStr,"%.2f",fixStd);
  TextSet(fix2DStdTextW, outStr, FALSE);
  
  // Let's reload all data in case it's corrupted from last time
  // GCA: when the data could be corrupted? Otherwise memcpy not needed
  if(fileType==PACKEDPORTABLE) { // packed
    for (i=0; i<purdueHeader.nimg; i++) {
      memcpy(packImg[i], fix2DpackImg_buffer[i],
	     (purdueHeader.numCols*purdueHeader.numRows*sizeof(float)));
      init_floatarr(fix2DpackImg_diff[i], 
		    purdueHeader.numRows*purdueHeader.numCols, 0.0);
    }
  }
  else {// scanned
    memcpy(inCore_buffer, fix2DinCore_buffer,
	   (file_width*file_height*sizeof(float)));
    init_floatarr(fix2DinCore_diff, file_height*file_width, 0.0);
  }
  
  if(tag == FIXRESTORE) // Just get out we've restored it above
    goto restoreDone;

  if(tag==FIXPREVIEW || tag==FIX2D) { // Do it non-destructively
    
    if(fileType==PACKEDPORTABLE) { // packed
      for (k=0; k<purdueHeader.nimg; k++) {
	sprintf(outStr,"Processing Image [%d].",k);
	checkStatus(outStr);

	if (purdueHeader.numCols==purdueHeader.numRows)
	  calcCircleStats(packImg[k], &thisBkg, &thisMin, &thisMax,
			&thisAvg, &thisStd, purdueHeader.numCols,
			purdueHeader.numRows, purdueHeader.numCols);
	else
	  calcSubScanStats(packImg[k], &thisBkg, &thisMin,
			 &thisMax, &thisAvg, &thisStd,
			 purdueHeader.numCols, 
			 purdueHeader.numRows, 
			 purdueHeader.numCols,
			 0, 0);
	
	cutHigh = thisAvg + fixStd*thisStd;
	cutLow  = thisAvg - fixStd*thisStd;

	for (i=0; i<purdueHeader.numRows*purdueHeader.numCols; i++) {
	  if(packImg[k][i] > cutHigh || packImg[k][i] < cutLow) {
	    // Cut it off
	    packImg[k][i]  = thisAvg;
	    fix2DpackImg_diff[k][i] = (BYTE)TRUE;
	  }
	}
	
      }      
    } else { // scanned
      int thisX, thisY, xStep, yStep, thisOne;
      int drawX, drawY;
      
      xStep = file_width  / (float)fixRegions;
      yStep = file_height / (float)fixRegions;
      // Make sure we get the last niblet in last region
      xStep++; yStep++;  
      
      for (thisY=0,thisOne=1; thisY<file_height; thisY+= yStep,thisOne++) {
	sprintf(outStr,"Processing region [%d].",thisOne);
	checkStatus(outStr);
	
	for (thisX=0; thisX<file_width; thisX+=xStep) {
	    
	  calcSubScanStats(inCore_buffer, &thisBkg, &thisMin,
			   &thisMax, &thisAvg,  &thisStd,
			   ((thisX+xStep) > file_width)?
			   (file_width-thisX) :xStep, 
			   ((thisY+yStep) > file_height)?
			   (file_height-thisY) :yStep, 
			   file_width, thisX, thisY);
	  
	  drawX = thisX/xStep;
	  drawY = thisY/yStep;
	  
	  fix2Davg[drawX][drawY] = thisAvg;
	  fix2Dstd[drawX][drawY] = thisStd;
	  
	  cutHigh = thisAvg + fixStd*thisStd;
	  cutLow  = thisAvg - fixStd*thisStd;
	  
	  for (i=thisY; i<thisY+yStep; i++) {
	    for (j=thisX; j<thisX+xStep; j++) {
	      if((inCore_buffer[(i*file_width)+j] > cutHigh) ||
		 (inCore_buffer[(i*file_width)+j] < cutLow)) {
		// Cut it off
		if(i<file_height && j<file_width) {
		  inCore_buffer[(i*file_width)+j]    = thisAvg;
		  fix2DinCore_diff[(i*file_width)+j] = (BYTE)TRUE;
		}
	      }
	    }
	  }

	}
      }
    }

  } // DO it non-destructively
  
 restoreDone:;
  
  checkStatus("Writing to screen.");
  
  if(fileType==PACKEDPORTABLE)
    // Start at radius'th image
    stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		     packImg, purdueHeader.nimg, imgNum);  
  else {
    // Force a recalc of full screen
    fullScreen_buffer.dispx = 0;
    fullScreen_buffer.dispy = 0;
    virtual_read_activate(-99,-99);
  }
  
  DrawWindow(); 
  
  if(tag == FIX2D) { // Destroy the data
    if(fileType != MAP) { // 2d
      if(fileType == PACKEDPORTABLE) {
	for (i=0; i<purdueHeader.nimg; i++) {
	  freeNullFloat(&fix2DpackImg_diff[i]);
	  freeNullFloat(&fix2DpackImg_buffer[i]);
	}
      }
      else {
	freeNullFloat(&fix2DinCore_diff);
	freeNullFloat(&fix2DinCore_buffer);
      }
    }
    if(useGUI) XtUnmanageChild(fix2DScreenW);
  }
  
  if(useGUI) XtUnmanageChild(statusScreenW);
  
} // end function fix2D_do_activate



void removeBlemish_activate(Widget widget, XtPointer client_data, 
			    XtPointer call_data)
{
  // function name: removeBlemish_activate
  // Remove statistically outlying data from this scanned image -
  // maybe there's a speck of dust on the micrograph when it's
  // scanned?

  int i;
  
  // Get rid of 3D rendering if it's there
  if(useGUI && XtIsManaged(mapShadeScreenW))
    mapshade_deactivate(NULL, NULL, NULL);
  
  // Fill the corners
  sprintf(outStr,"%d",(int)((float)largestDim*1.5));
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr,"%d",(int)(purdueHeader.numRows*1.5));
  TextSet(fixMaxW, outStr, TRUE);
  
  // 2d
  for (i=0; i<purdueHeader.nimg; i++) {
    freeNullFloat(&fix2DpackImg_diff[i]);
    freeNullFloat(&fix2DpackImg_buffer[i]);
  }

  freeNullFloat(&fix2DinCore_diff);
  freeNullFloat(&fix2DinCore_buffer);

  if(fileType == PACKEDPORTABLE) {// packed
    for(i=0; i<purdueHeader.nimg; i++) {
      fix2DpackImg_diff[i] = (float *)
	malloc(sizeof(float)*purdueHeader.numCols*purdueHeader.numRows);
      fix2DpackImg_buffer[i] = (float *)
	malloc(sizeof(float)*purdueHeader.numCols*purdueHeader.numRows);
      memcpy(fix2DpackImg_buffer[i],packImg[i],
	     (purdueHeader.numCols*purdueHeader.numRows*sizeof(float)));
    }
  }
  else { // scanned
    CK(fix2DinCore_diff = (float *)
       malloc(sizeof(float)*file_width*file_height));
    fix2DinCore_buffer = (float *)
      malloc(sizeof(float)*file_width*file_height);
    memcpy(fix2DinCore_buffer,inCore_buffer,
	   (file_width*file_height*sizeof(float)));
  }
  
  raiseIt(fix2DScreenW);
  
} // end function removeBlemish_activate



void fix_activate(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: fix_activate
  // Get fix screen up.
  
  // Get rid of 3D rendering if it's there
  if(XtIsManaged(mapShadeScreenW))
    mapshade_deactivate(NULL, NULL, NULL);
  
  TextSet(fixBegXW, "1", FALSE); 
  TextSet(fixBegYW, "1", FALSE); 
  TextSet(fixBegZW, "1", FALSE); 
  
  sprintf(outStr,"%d",largestDim);
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr,"%d",purdueHeader.numCols);
  TextSet(fixEndXW, outStr, FALSE); 
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr,"%d",purdueHeader.numRows);
  TextSet(fixEndYW, outStr, FALSE); 
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr,"%d",purdueHeader.nimg);  // No sections - boxes
  TextSet(fixEndZW, outStr, FALSE); 
  
  // Fill the corners
  sprintf(outStr, "%d", (int)(largestDim*1.5));
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr, "%d", (int)(purdueHeader.numRows*1.5));
  TextSet(fixMaxW, outStr, FALSE);
  TextSet(fixMinW, "1", TRUE);
  
  raiseIt(fixScreenW);
  
} // end function fix_activate



void fix_change(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: fix_change
  // Call here whenever text widgets changed
  
  fixMode = *((int *) client_data);
  
  if(fixMode == NOISE || fixMode == GAUSSNOISE)  // Init fixValue field
    TextSet(fixTextW, "1.0 STD", TRUE); 
  
  if(fixMode == GAUSSNOISE) {
    XtManageChild(fixGaussStdLabelW);
    XtManageChild(fixGaussStdTextW);
    TextSetFloat(fixGaussStdTextW, purdueHeader.img_std, FALSE);
  }
  else {
    XtUnmanageChild(fixGaussStdLabelW);
    XtUnmanageChild(fixGaussStdTextW);
  }
  
} // end function fix_change



void fix_do_activate(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: fix_do_activate
  // Do the fix

  int i, j, k, xcen, ycen, zcen, lastZ;
  int begX, begY, begZ, endX, endY, endZ, tmpY, thisNumSecs;
  int icount, ncol, nrow;
  float minRad, maxRad, thisRad, thisPixel, saveMinRad, saveMaxRad;
  float xSq, ySq, zSq;
  float noise, fixValue;
  float *thisPackImg, *tmpS, *gauss;
  float maxF, minF, scaleF, stDev;
  double factor;
  
  fixValue = wstrtod(fixTextW);
  
  saveMinRad= minRad = wstrtod(fixMinW);
  minRad *= minRad;  // Do this squared

  saveMaxRad = maxRad = wstrtod(fixMaxW);
  maxRad = maxRad*maxRad;  // Do this squared

  begX = wstrtod(fixBegXW);
  begY = wstrtod(fixBegYW);
  begZ = wstrtod(fixBegZW);
  endX = wstrtod(fixEndXW);
  endY = wstrtod(fixEndYW);
  endZ = wstrtod(fixEndZW);

  // Humans number from 1 not zero
  begX--;
  begY--;
  begZ--;

  // Always do fixValue STD
  noise  = purdueHeader.img_std * fixValue;   
  if(fixMode==GAUSSNOISE) // Check to see what's entered
    noise = wstrtod(fixGaussStdTextW) * fixValue;
  
  // We're in batch mode and they've asked for a specific stDev
  if(doNoise) {
    fixValue = noiseStd;
    noise    = purdueHeader.img_std * fixValue ;
    fixMode  = noiseType;
    begX = begY = begZ = 0;
    endX = purdueHeader.numCols;
    endY = purdueHeader.numRows;
    endZ = purdueHeader.nimg;
  }

  printf("begEnd [%d,%d,%d,%d,%d,%d]\n", begX, begY, begZ , endX, endY ,endZ);
  
  // Get our fraction of the random number
  factor= 1.0 / RAND_MAX;
  thisNumSecs = 1;   // GCA: arbitrary choice

  if(fileType==PACKEDPORTABLE) { // Do packed boxes
    // Y is a special case - Tim numbers row starting from the bottom.
    
    tmpY = endY;
    endY = purdueHeader.numRows - begY;
    begY = purdueHeader.numRows - tmpY;
    
    begX = MAX(begX, 0);
    begY = MAX(begY, 0);
    begZ = MAX(begZ, 0);
    endX = MIN(endX, purdueHeader.numCols);
    endY = MIN(endY, purdueHeader.numRows);
    endZ = MIN(endZ, purdueHeader.nimg);
  }
  else if(fileType==MAP) { // MAPs
    // Y is a special case - Tim numbers row starting from the bottom.
    
    tmpY = endY;
    endY = numRows - begY;
    begY = numRows - tmpY;
    
    begX = MAX(begX, 0);
    begY = MAX(begY, 0);
    begZ = MAX(begZ, 0);
    endX = MIN(endX, largestDim);
    endY = MIN(endY, largestDim);
    endZ = MIN(endZ, largestDim);
    // We may have shaded map display where numSecs is less than 
    // largestDim so account for that here
    thisNumSecs = numSecs;
    if(shadedMaps)
      thisNumSecs = largestDim;
    
    endZ = MIN(endZ,thisNumSecs);  // We may have a PRJ file here that has
				   // reduced number of sections
  }
  else if(fileType==BYTEPORTABLE || fileType==PORTABLE) {
    // Add noise to scan
    begZ = 0;  // Only one scan
    endZ = 1;
    begX = 0;
    endX = file_width;
    begY = 0;
    endY = file_height;
    thisNumSecs = 1;
  }

  if(fixMode==GAUSSNOISE && 
     (fileType==PACKEDPORTABLE || fileType==BYTEPORTABLE || fileType==MAP)) {
    // Add gaussian noise to packed boxes/scans
    
    if(fileType==BYTEPORTABLE) { // Setup for scans
      calcSubScanStats(inCore_buffer,
		       &purdueHeader.img_bkg, &purdueHeader.img_min,
		       &purdueHeader.img_max, &purdueHeader.img_avg, 
		       &purdueHeader.img_std,
		       file_width, file_height, file_width, 0, 0);
      noise  = purdueHeader.img_std * fixValue;   
      stDev = purdueHeader.img_std;
      ncol  = file_width;
      nrow  = file_height;
      tmpS = (float *)malloc(sizeof(float) * file_width*file_height);
    }
    else if(fileType==MAP) {
      stDev = purdueHeader.img_std;
      ncol  = numCols;
      nrow  = numRows;
      tmpS = (float *)malloc(sizeof(float)*origNumSecs*numCols*numRows);
    }
    else { // Setup for packed boxes 
      stDev = purdueHeader.img_std;
      ncol  = purdueHeader.numCols;
      nrow  = purdueHeader.numRows;
      tmpS = (float *)malloc(sizeof(float) *
			     purdueHeader.nimg *
			     purdueHeader.numCols *
			     purdueHeader.numRows);
    }
    
    
    icount = 1000;
    gauss = (float *)malloc(sizeof(float) * icount);
    
    for (i=begZ; i<endZ; i++) {

      for (j=0; j<nrow; j++)
	for (k=0; k<ncol; k++)
	  tmpS[i*ncol*nrow + j*ncol +k] = 0.0;
      
      if(fileType==PACKEDPORTABLE) { // packed image
	thisPackImg = packImg[i];
	
	for (j=begY;  j<endY;  j++)
	  for (k=begX;  k<endX;  k++)
	    tmpS[i*ncol*nrow + j*ncol +k] = thisPackImg[j*ncol + k];
	
      }
      else if(fileType==MAP) { // map image
	
	for (j=begY; j<endY; j++)
	  for (k=begX; k<endX; k++)
	    tmpS[i*ncol*nrow + j*ncol + k] = 
	      float_image_buffer1[i*ncol*nrow + j*ncol + k];
	
      }
      else { // scanned image
	
	for (j=begY; j<endY; j++)
	  for (k=begX; k<endX; k++)
	    tmpS[i*ncol*nrow + j*ncol + k] = inCore_buffer[j*ncol + k];
	
      }// scanned image
      
      add_gauss_noise(tmpS+(i*ncol*nrow), ncol*nrow, 0.0, noise, 3.0);

    }// imgs
    
    minF = maxF = tmpS[begZ*nrow*ncol + begY*ncol + begX];
    
    for (i=begZ; i<endZ; i++) {
      for (j=begY; j<endY; j++) {
	for (k=begX; k<endX; k++) {
	  minF = MIN(tmpS[i*nrow*ncol + j*ncol +k], minF);
	  maxF = MAX(tmpS[i*nrow*ncol + j*ncol +k], maxF);
	}
      }
    }
    
    scaleF = (tryColors - 7.0) / (maxF - minF);
    
    // Now put it back
    for (i=begZ; i<endZ; i++) {
      if(fileType==MAP) {
	maxMapPixel = maxF;
	minMapPixel = minF;
	
	for (j=begY; j<endY; j++)
	  for (k=begX; k<endX; k++)
	    float_image_buffer1[i*ncol*nrow + j*ncol + k] =
	      tmpS[i*ncol*nrow + j*ncol + k];
	
      }
      else {
	thisPackImg = (fileType==PACKEDPORTABLE) ? packImg[i] : inCore_buffer;
	for (j=begY; j<endY; j++)
	  for (k=begX; k<endX; k++)
	    thisPackImg[j*ncol + k] = tmpS[i*ncol*nrow + j*ncol + k];
	
      }
    }
    
    // Now get min/max
    minF = maxF = tmpS[0];
    if(fileType!=MAP) {
      for (i=0; i<purdueHeader.nimg; i++) {
	thisPackImg = (fileType==PACKEDPORTABLE) ? packImg[i] : inCore_buffer;
	for (j=0; j<nrow*ncol; j++) {
	  minF = MIN(thisPackImg[j],minF);
	  maxF = MAX(thisPackImg[j],maxF);
	}
      }
    }
    
    printf("min,max = [%f,%f]\n",minF,maxF);
    
    free(tmpS);
    free(gauss);
    
    if(fileType==PACKEDPORTABLE) { // Setup packed stuff
      maxPackPixel = maxF;
      minPackPixel = minF;
      
      // Now clip this guy to our colormap 
      if(minPackPixel < 0.0)
	scalePackFactor = (tryColors - 7.0) / (maxPackPixel-minPackPixel);
      else 
	scalePackFactor = (tryColors - 7.0) / maxPackPixel;

      stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		       packImg,purdueHeader.nimg, imgNum);
    }// Setup packed stuff
    else if(fileType==MAP) { // Map
      // Make sure we have some kind of range
      if(maxMapPixel==minMapPixel)
	maxMapPixel = minMapPixel + 1.0;
      printf("MAX[%f]  MIN[%f]\n",maxMapPixel,minMapPixel);
      secMaxPixel = maxMapPixel;
      secMinPixel = minMapPixel;
      
      calcMapStats(float_image_buffer1,
		   &purdueHeader.img_bkg, &purdueHeader.map_min,
		   &purdueHeader.map_max, &purdueHeader.img_avg, 
		   &purdueHeader.img_std,
		   &numCols, &numRows, &origNumSecs);
      stuffMapImage(map_image_buffer, float_image_buffer1, 
		    file_width, file_height, nsecs, imgNum, 0);
      DrawWindow();
    }
    else { //Scan
      // Make it generate a new fullscreen buffer
      fullScreen_buffer.dispx = fullScreen_buffer.dispy = 0;
      virtual_read_activate(0,0);
    }
    
    DrawWindow(); 
    
    goto allDone;
    
  }
  else if(fixMode==GAUSSNOISE || fileType==BYTEPORTABLE) {
    askThem("Not yet.", XmDIALOG_INFORMATION);
    goto byPass;
  }
  
  if(fileType==PACKEDPORTABLE || fileType==PORTABLE) { // Do for packed boxes
    float newValue;
    float tmp1, maxValue;
    float *thisTmpS;
    xcen =   purdueHeader.numCols/2;
    ycen =   purdueHeader.numRows/2;
    
    thisTmpS = (float *)malloc(sizeof(float) *
			       purdueHeader.nimg *
			       purdueHeader.numCols *
			       purdueHeader.numRows);
    
    for (i=begZ; i<endZ; i++) {
      thisPackImg = (fileType==PORTABLE) ? inCore_buffer : packImg[i];
      
      ncol  = purdueHeader.numCols;
      nrow  = purdueHeader.numRows;
      maxValue = thisPackImg[0];
      
      for (j=0; j<nrow; j++) {
	for (k=0; k<ncol; k++) {
	  thisTmpS[(i*nrow*ncol)+(j*ncol)+k] = thisPackImg[(j*ncol)+k];
	  maxValue = MAX(maxValue,thisPackImg[(j*ncol)+k]);
	}
      }
      
      for (j=begY; j<endY; j++) {
	ySq = (j-ycen)*(j-ycen);
	
	for (k=begX; k<endX; k++) {
	  xSq = (k-xcen)*(k-xcen);
	  thisRad = xSq+ySq;
	  if((thisRad >= minRad && thisRad <=maxRad)) {
	    newValue = thisPackImg[(j*purdueHeader.numCols)+k];
	    if(fixMode==FIXADD)
	      newValue +=  fixValue;
	    else if(fixMode==FIXMULT)
	      newValue =  (newValue * fixValue);
	    else if(fixMode==FIXOVER)
	      newValue =  fixValue;
	    else if(fixMode==INVERT) {
	      if(fileType==PORTABLE)
		newValue = maxValue - newValue;
	      else  // For maps???
		newValue = 0.0 - newValue;
	    }
	    else if(fixMode==NOISE) {
	      tmp1 = (((float)random()*factor)-0.5)*2.0*noise;
	      newValue +=  tmp1;
	    }
	    
	    thisTmpS[i*nrow*ncol + j*purdueHeader.numCols + k]= newValue;
	  }
	}
      }
      
    }
    
    if(fixMode==NOISE) {
      minF = maxF = thisTmpS[(begZ*nrow*ncol)+(begY*ncol)+begX]; 
      for (i=begZ; i<endZ; i++) {
	for (j=begY; j<endY; j++) {
	  for (k=begX; k<endX; k++) {
	    minF = MIN(thisTmpS[(i*nrow*ncol)+(j*ncol)+k],minF);
	    maxF = MAX(thisTmpS[(i*nrow*ncol)+(j*ncol)+k],maxF);
	  }
	}
      }
      scaleF = (tryColors - 7.0) / (maxF - minF);
    }
    
    // Now put it back
    
    for (i=begZ; i<endZ; i++) {
      thisPackImg = (fileType==PORTABLE) ? inCore_buffer : packImg[i];

      for (j=begY; j<endY; j++) {
	for (k=begX; k<endX; k++) {
	  if(fixMode==NOISE)
	    thisPackImg[j*ncol + k] = thisTmpS[i*ncol*nrow + j*ncol + k];
	  else
	    thisPackImg[j*ncol + k] = thisTmpS[i*ncol*nrow + j*ncol + k];
	}
      }
    }
    
    // Now get min/max
    // Why only for noise???       if(fixMode==NOISE)

    minF = maxF = thisTmpS[0];
    
    for (i=0; i<purdueHeader.nimg; i++) {
      if(fileType==PORTABLE)
	thisPackImg = inCore_buffer;
      else
	thisPackImg = packImg[i];
      for (j=0; j<nrow*ncol; j++) {
	minF = MIN(thisPackImg[j],minF);
	maxF = MAX(thisPackImg[j],maxF);
      }
    }
    
    maxPackPixel = maxF;
    minPackPixel = minF;
    
    // Now clip this guy to our colormap
    if(minPackPixel < 0.0)
      scalePackFactor = (tryColors - 7.0) / (maxPackPixel-minPackPixel);
    else 
      scalePackFactor = (tryColors - 7.0) / maxPackPixel;
    
    free(thisTmpS);
    if(fileType==PORTABLE) { // Do nothing
      fullScreen= FALSE;
      virtual_read_activate(-99,-99);
    }
    else {
      stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		       packImg, purdueHeader.nimg, imgNum);
    }
    DrawWindow(); 
    
  }
  else if(fileType==MAP) { // Do for MAP
    
    xcen = numCols/2;
    ycen = numRows/2;
    zcen = numSecs/2;
    
    for (i=begZ; i<endZ; i++) {
      zSq = (float)((i-zcen)*(i-zcen));  
      for (j=begY; j<endY; j++) {
	ySq = (float)((j-ycen)*(j-ycen));
	for (k=begX; k<endX; k++) {
	  xSq= (float)((k-xcen)*(k-xcen));
	  thisRad = zSq + (xSq+ySq);
	  if((fixIcosFill && callXdy((double)i,(double)j,(double)k,
				     (double)saveMinRad,
				     (double)saveMaxRad)) ||
	     (thisRad >= minRad && thisRad <=maxRad)) {
	    
	    int gindx;
	    gindx = i*numRows*numCols + j*numCols + k;
	    
	    if(fixMode == FIXADD)
	      float_image_buffer1[gindx] += fixValue;
	    else if(fixMode == FIXMULT)
	      float_image_buffer1[gindx] *= fixValue;
	    else if(fixMode == FIXOVER)
	      float_image_buffer1[gindx]  = fixValue;
	    else if(fixMode == INVERT)
	      float_image_buffer1[gindx] *= -1.0;
	    else if(fixMode == NOISE)
	      float_image_buffer1[gindx] += 
		((random()*factor)-0.5)*2.0*noise;
	    
	  } 
	}
      }
    }
    
    // Find the new min/max - we may be only
    // doing a part of the MAP here so we need to include all of
    // the ones that aren't procesed here in the min/max.
    maxMapPixel = -HUGE;
    minMapPixel = HUGE;
    
    if(thisNumSecs < numRows) // Not symmetrical e.g. may be a PRJ file
      lastZ = thisNumSecs;
    else
      lastZ = largestDim;
    
    for (i=0; i<lastZ; i++) {
      for (j=0; j<largestDim; j++) {
	for (k=0; k<largestDim; k++) {
	  thisPixel= float_image_buffer1[i*numRows*numCols + j*numCols + k];
	  if(thisPixel > maxMapPixel)
	    maxMapPixel = thisPixel;
	  if(thisPixel < minMapPixel)
	    minMapPixel = thisPixel;
	}
      }
    }
    
    
    minSurf = MAX(minMapPixel,0.0);
    maxSurf = MAX(maxMapPixel,(0.0-minMapPixel));

    // We're doing the surfaceDensity in tenths now
    XtVaSetValues(surfDensScrollW, XmNmaximum, maxSurf*10 + 50, NULL);
    XtVaSetValues(surfDensScrollW, XmNminimum, minSurf*10, NULL);
    XtVaSetValues(surfDensScrollW,XmNvalue, 10, NULL);
    
    if(maxSurf > 5000.0) {
      XtVaSetValues(surfDensScrollW, XmNsliderSize, 
		    (int)(maxSurf*0.01),NULL);
      XtVaSetValues(surfDensScrollW, XmNpageIncrement, 
		    (int)(maxSurf*0.01),NULL);
      XtVaSetValues(surfDensScrollW, XmNincrement, 
		    (int)(maxSurf*0.002),NULL);
    }
    else if(maxSurf > 500.0) {
      XtVaSetValues(surfDensScrollW, XmNsliderSize, 500, NULL);
      XtVaSetValues(surfDensScrollW, XmNpageIncrement, 500, NULL);
      XtVaSetValues(surfDensScrollW, XmNincrement, 100, NULL);
    }
    else {
      XtVaSetValues(surfDensScrollW, XmNsliderSize, 50, NULL);
      XtVaSetValues(surfDensScrollW, XmNpageIncrement, 50, NULL);
      XtVaSetValues(surfDensScrollW, XmNincrement, 10, NULL);
    }
    
    // Their shaded views are no longer valid - show the new sections
    shadedMaps= FALSE;
    dispSections = TRUE;
    XtSetArg(arglst[0], XmNlabelPixmap, sectionPixmap);
    XtSetArg(arglst[1], XmNselectPixmap, unsectionPixmap);
    XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
    XtSetValues(sectionW, arglst, 3); 
    nsecs = origNumSecs;
    if(dispSections) {
      // Update the sections - if shaded views then they have to recalc
      // Now stuff the display buffer from our new FIXed map
      stuffMapImage(map_image_buffer, float_image_buffer1, 
		    file_width, file_height, nsecs, imgNum, 0);
      DrawWindow();
    }
    
  }
  else {
    askThem("FIX not yet implemented for this file type.",
	    XmDIALOG_INFORMATION);
    goto byPass;
  }
  
 allDone:;
  
  if(fixMode==FIXADD)
    sprintf(outStr, "Successfully %s %f", "Added", fixValue);
  else if(fixMode==FIXMULT)
    sprintf(outStr, "Successfully %s %f", "Multiplied by", fixValue);
  else if(fixMode==FIXOVER)
    sprintf(outStr, "Successfully %s %f", "Overwrote with", fixValue);
  else if(fixMode==INVERT)
    sprintf(outStr, "Successfully INVERTED voxels");
  else if(fixMode==NOISE || fixMode==GAUSSNOISE) {
    sprintf(outStr,
	    "Successfully added %.1f STD of noise",fixValue);
    printf("%s\n",outStr);
  }
  
  //   XtUnmanageChild(fixScreenW);
  if(!doNoise)
    askThem(outStr, XmDIALOG_INFORMATION);
  
 byPass:;
  
} // end function fix_do_activate



void point_color_activate(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: point_color_activate
  // Change the color of both the point screen label and lines on display

  pointColor = (++pointColor)%3;

  if(pointColor == 0) 
    XtVaSetValues(widget, XmNforeground, pixels_return[RED], NULL);
  else if(pointColor == 1) 
    XtVaSetValues(widget, XmNforeground, pixels_return[GREEN], NULL);
  else if(pointColor == 2) 
    XtVaSetValues(widget, XmNforeground, pixels_return[BLUE], NULL);

} // end function point_color_activate



void drawPointScreen(char *thisStr, int pos)
{
  // function name: drawPointScreen
  // Draw this string on the pointscreen.

  XtSetArg(arglst[0], XmNlabelString, XmStringCreateLocalized(thisStr));
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetArg(arglst[2], XmNalignment, XmALIGNMENT_BEGINNING);
  XtSetValues(pointLabelW[pos], arglst, 3); 

  XFlush(dsply);

} // end function drawPointScreen



float findRealValue(int posX, int posY, int *thisImg, int *sectY, int *sectX)
{
  // function name: findRealValue
  // Return the value for this pixel
  
  int nrows, ncols, sectRow, sectCol, section, offX, offY;
  
  calcSizeFactor(zoomFactor);
  nrows= purdueHeader.numRows;
  ncols= purdueHeader.numCols;
  
  if(fileType==EMPORTABLE || fileType==BYTEPORTABLE || fileType==PORTABLE) {
    *sectX = virXoffset + posX * recipSizeFactor;
    *sectY = virYoffset + posY * recipSizeFactor;
    *thisImg = 0;
    if(posY < image_height && posX < image_width)
      return(int_image_buffer1[posY*image_width + posX]);
  }
  else if(fileType==PACKEDPORTABLE || fileType==MAP) {
    sectRow = posY / (int)(nrows*sizeFactor);
    sectCol = posX / (int)(ncols*sizeFactor);
    offY = (posY % (int)(nrows*sizeFactor)) * recipSizeFactor;
    offX = (posX % (int)(ncols*sizeFactor)) * recipSizeFactor;
    *thisImg = 0;
    *sectY   = 0;
    *sectX   = 0;
    if(fileType==PACKEDPORTABLE) {
      section = packStartImage + sectRow*packSectionsPerRow + sectCol;
      if(section < purdueHeader.nimg) {
	*thisImg = section + 1;
	*sectY   = purdueHeader.numRows - offY;
	*sectX   = offX + 1;
	return(packImg[section][offY*ncols + offX]);
      }
    }
    else if(fileType==MAP) {
      section = mapStartImage + sectRow*mapSectionsPerRow + sectCol;
      if(section < nsecs) {
	*thisImg = section + 1;
	*sectY   = purdueHeader.numRows - offY; 
	*sectX   = offX + 1;
	if(shadedMaps && !dispSections)
	  return(map_image_buffer[section*nrows*ncols + offY*ncols + offX]);
	else
	  return(float_image_buffer1[section*nrows*ncols + offY*ncols + offX]);
      }
    }
  }

  return(0.0);
  
} // end function findRealValue



void button_zoom_activate(Widget widget, XEvent *event)
{
  // function name: button_zoom_activate
  // Change the current displayed box by using the arrows.

  float thisSizeFactor, speedFactor;

  if(!inited) return;

  speedFactor = (float) GetButtonVal(event->xbutton.button, 1, 2, 5);
  calcSizeFactor(zoomFactor);

  if(widget == zoomArrowUpW)
    thisSizeFactor = sizeFactor + 0.01*speedFactor;
  else if(widget == zoomArrowDownW)
    thisSizeFactor = sizeFactor - 0.01*speedFactor;
  else 
    thisSizeFactor = sizeFactor;
  
  zoomFactor = rwalog2(thisSizeFactor);  
  zoomFactor = MAX(-4.0,MIN(zoomFactor,4.0));
  
  sprintf(outStr,"%.2f",thisSizeFactor);
  TextSet(zoomTextW, outStr, FALSE);
  
  // Now redraw with new zoominess
  fullScreen= FALSE;
  virtual_read_activate(display_width/2, display_height/2);
  
} // end function button_zoom_activate



void radiusText_changed(Widget widget, XtPointer client_data, 
			XtPointer call_data)
  
{
  // function name: radiusText_changed
  // When they change the radiusText widget they need to press the
  // "radius" button next to it to get the change instantiated.  We
  // change the radius here but they need to call the
  // radius_activate() callback (called when the button is pressed) to
  // get it displayed.

  if(inited && !locked) {
    if(widget == boxRadiusTextW) {
      boxRadius = featherRadius - (int)wstrtod(boxRadiusTextW);
      boxRadius = MAX(boxRadius, 1);
      XFlush(dsply);
    }
    else if(widget == emptyShellTextW) {
      emptyShellWidth = (int)wstrtod(emptyShellTextW);
      emptyShellWidth = MAX(emptyShellWidth, 1);
      if(emptyShellWidth >= emmaRadius)
	emptyShellWidth = emmaRadius - 1;
      XFlush(dsply); 
    }
    else if(widget == emmaRadiusTextW) {
      emmaRadius = (int)wstrtod(emmaRadiusTextW);
      emmaRadius = MAX(emmaRadius, 1);
      if(emmaRadius >= featherRadius) {
	boxRadius += (emmaRadius+1)-featherRadius;
	featherRadius = emmaRadius+1;
      }
      XFlush(dsply);
    }
    else if(widget == featherRadiusTextW) {
      int oldSize;
      oldSize = featherRadius - boxRadius;	    
      featherRadius = (int)wstrtod(featherRadiusTextW);
      featherRadius = (featherRadius -1)/2;
      featherRadius = MAX(featherRadius, 1);
      boxRadius = featherRadius - oldSize;
      XFlush(dsply); 
    }
  }
  
  DrawWindow(); 

} // end function radiusText_changed



void radiusArrowActivateButton(Widget widget, XEvent *event)
{
  // function name: radiusArrowActivateButton
  // Instead of making them type the new radius just let them clikc
  // the arrow. Change the radius accordingly. This routine is for the
  // second and third mouse buttons.

  int oldSize, speedFactor;

  if(locked)
    return;

  speedFactor = GetButtonVal(event->xbutton.button, 1, 2, 5);
  oldSize = featherRadius - boxRadius;
  
  if(widget == boxRadiusArrowUpW)
    boxRadius = MAX(boxRadius - speedFactor, 1);
  else if(widget == boxRadiusArrowDownW) {
    boxRadius = MIN(boxRadius + speedFactor, featherRadius);
    boxRadius = MAX(boxRadius, 1);
  }
  else if(widget == emptyShellArrowUpW) {
    emptyShellWidth = MAX(emptyShellWidth + speedFactor, 1);
    emptyShellWidth = MIN(emptyShellWidth, emmaRadius-1);
  }
  else if(widget == emptyShellArrowDownW)
    emptyShellWidth = MAX(emptyShellWidth - speedFactor, 1);
  else if(widget == emmaRadiusArrowUpW) {
    emmaRadius = MAX(emmaRadius+speedFactor, 1);
    if(emmaRadius >= featherRadius) { // Inc it
      featherRadius = emmaRadius + 1;
      boxRadius = featherRadius - oldSize;
    }
  }
  else if(widget == emmaRadiusArrowDownW)
    emmaRadius = MAX(emmaRadius - speedFactor, 1);
  else if(widget == featherRadiusArrowUpW) {
    featherRadius = MAX(featherRadius + speedFactor, 1);
    boxRadius = featherRadius - oldSize;
  }
  else if(widget == featherRadiusArrowDownW) {
    featherRadius = MAX(featherRadius - speedFactor, 1);
    featherRadius = MAX(featherRadius, emmaRadius);
    boxRadius = featherRadius - oldSize;
  }
  
  TextSetInt(boxRadiusTextW, featherRadius-boxRadius, FALSE);
  TextSetInt(emptyShellTextW, emptyShellWidth, FALSE);
  TextSetInt(emmaRadiusTextW, emmaRadius, FALSE);
  TextSetInt(featherRadiusTextW, 2*featherRadius + 1, FALSE);
  DrawWindow(); 
  
} // end function radiusArrowActivateButton



void appodize_activate(Widget widget, XtPointer client_data, 
				  XtPointer call_data)
{
  // function name: appodize_activate
  // Get appodize screen up.

  int outHeight, outWidth ;  

  XtUnmanageChild(appodizeScreenW);
  raiseIt(statusScreenW);
  XSync(dsply,0);
  
  featherRadius = (int)wstrtod(appodFeaRadW);

  // This is now border Width
  boxRadius = featherRadius - (int)wstrtod(appodBoxRadW);
  
  // Make sure we don't get thresholded out with the emma radius
  if(emmaRadius >= (boxRadius-1))
    emmaRadius = boxRadius-2;  
  
  // Now setup boxing screen for new values
  TextSetInt(featherRadiusTextW, (2*featherRadius)+1,     FALSE); 
  TextSetInt(boxRadiusTextW,     featherRadius-boxRadius, FALSE); 
  
  if(appodizing == CIRCLEPACKED) {
    outHeight = outWidth = boxRadius*2 + 1;
    // Shouldn't we use the featherRadius here to determine
    // outWidth and outHeight for write_pif() ???
    outHeight= outWidth = featherRadius*2 + 1;
    write_pif(outWidth, outHeight, fileName.addr, CIRCLEPACKED);
  }
  else if(appodizing == PACKEDPORTABLE) {
    if(packRadius <= 0)
      outHeight = outWidth = purdueHeader.numCols;
    else 
      outHeight= outWidth = packRadius*2 + 1;	
    if(purdueHeader.numCols != purdueHeader.numRows) {
      outHeight = purdueHeader.numRows;
      outWidth  = purdueHeader.numCols;
    }
    write_pif(outWidth, outHeight, fileName.addr, PACKEDPORTABLE);
  }
  else if(appodizing == WRITEFFT) {
    ctf_all_activate();
  }

  XtUnmanageChild(statusScreenW);

} // end function appodize_activate



void boxcolor_change(Widget widget, XtPointer client_data, 
		     XtPointer call_data)
{
  // function name: boxcolor_change
  // Get the boxing color

  if(boxing_GC == image_gcGREEN) {
    boxing_GC = image_gcRED;
    XtSetArg(arglst[0], XmNforeground, pixels_return[RED]);
  }
  else if(boxing_GC == image_gcRED) {
    boxing_GC = image_gcBLUE;
    XtSetArg(arglst[0], XmNforeground, pixels_return[BLUE]);
  }
  else if(boxing_GC == image_gcBLUE) {
    boxing_GC = image_gcYELLOW;
    XtSetArg(arglst[0], XmNforeground, pixels_return[YELLOW]);
  }
  else if(boxing_GC == image_gcYELLOW) {
    boxing_GC = image_gcWHITE;
    XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE]);
  }
  else if(boxing_GC == image_gcWHITE) {
    boxing_GC = image_gcBLACK;
    XtSetArg(arglst[0], XmNforeground, pixels_return[BLACK]);
  }
  else if(boxing_GC == image_gcBLACK) {
    boxing_GC = image_gcPLACEHOLDER;
    XtSetArg(arglst[0], XmNforeground, pixels_return[PLACEHOLDER]);
  }
  else if(boxing_GC == image_gcPLACEHOLDER) {
    boxing_GC = image_gcGREEN;
    XtSetArg(arglst[0], XmNforeground, pixels_return[GREEN]);
  }
  else {
    boxing_GC = image_gcGREEN;
    XtSetArg(arglst[0], XmNforeground, pixels_return[GREEN]);
  }

  XtSetValues(widget, arglst, 1);   
  DrawWindow();
  
} // end function boxcolor_change



void box_change(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: box_change
  // Redraw window if changes made to particle boxing parameters

  int tag = *((int *) client_data);

  if(tag == CYCLEREF)
    cycleRef = !cycleRef;
  else if(tag == DARKREF) {
    darkCircleRef = !darkCircleRef;
    if(darkCircleRef) {
      refCircAvg = FALSE;
      XtVaSetValues(refCircAvgW, XmNset, FALSE, NULL);
      XtVaSetValues(refAllW, XmNsensitive, FALSE, NULL);
      XFlush(dsply);
    }
    else {
      XtVaSetValues(refAllW, XmNsensitive, TRUE, NULL);
      XFlush(dsply);
    }
  }
  else if(tag == SHOWBOXINGSTATS)
    showBoxingSwitch = !showBoxingSwitch;
  else if (tag == BOXRADIUSBUTTON)
    boxRadiusDisp = !boxRadiusDisp;
  else if (tag == EMPTYSHELLBUTTON)
    emptyShellDisp = !emptyShellDisp;
  else if (tag == EMMARADIUSBUTTON)
    emmaRadiusDisp = !emmaRadiusDisp;
  else if (tag == FEATHERRADIUSBUTTON)
    featherRadiusDisp = !featherRadiusDisp;
  else if (tag == DRAWREFINE)
    refineDisp = !refineDisp;
  else if (tag == DRAWCLUSTERS)
    clusterDisp = !clusterDisp;
  else if (tag == DRAWREFERENCE)
    referenceDisp = !referenceDisp;
  else if (tag == DISPCENTERS)
    dispCenters = !dispCenters;
  else if (tag == AMMCORRCOEF || tag == AMMAVGDENS || tag == AMMSTDEV) {
    autoMinMaxThreshold = tag;
    if (tag == AMMCORRCOEF)
      sprintf(outStr,"Correlation Coef Cutoffs");
    else if (tag == AMMAVGDENS)
      sprintf(outStr,"Average Density Cutoffs");
    else if (tag == AMMSTDEV)
      sprintf(outStr,"Stdev Density Cutoffs");

    doLabel(outStr, autoMinMaxLabelW, FALSE);
    showBoxingLevel = tag; 
  }
  
  DrawWindow();
  
} // end function box_change



void autobox_scroll_change(Widget widget, XtPointer client_data, 
			   XtPointer call_data)
{
  // function name: autobox_scroll_change
  // Call here whenever autobox scroll widgets changed

  int i, removeThem, yOff;
  static float minScrollVal= 0.0, maxScrollVal=1.0;
  float minVal, maxVal;
  int tag = *((int *) client_data);

  XmScrollBarCallbackStruct *temp = (XmScrollBarCallbackStruct *) call_data;

  if(tag == THINNING) { // Change number of thinning layers
    nThins = temp->value;
    sprintf(outStr, "%d", nThins); 
    doLabel(outStr, thinValueW, DEFAULT);
  }
  else if(tag == DILATE) { // Change number of dilation layers
    nDilates = temp->value;
    sprintf(outStr,"%d",nDilates); 
    doLabel(outStr, dilateValueW, DEFAULT);
  }
  else { // Do min/max/remove thing
    removeThem = FALSE;

    if(tag == AUTOBOX)
      minScrollVal = temp->value/100.0;
    else if (tag == CORRELATE)
      maxScrollVal = temp->value/100.0;
    else if (tag == REMOVEBOXES)
      removeThem = TRUE;
    else if (tag == CMDREMOVEBOXES){
      removeThem = TRUE;
      minScrollVal = refCutoff;
      maxScrollVal = 1.0;
    }

    if(autoMinMaxThreshold == AMMCORRCOEF) {
          
      // First find MIN and MAX
      maxVal= minVal = box[0].correlation;
      for (i=0; i<numBoxes; i++) {
	box[i].reference = FALSE;
	minVal = MIN(box[i].correlation,minVal);
	maxVal = MAX(box[i].correlation,maxVal);
      }
      
      threshHi = minVal + (maxVal-minVal)*maxScrollVal;
      threshLo = minVal + (maxVal-minVal)*minScrollVal;
      
      // Now mark those outside range
      i = 0;
      do { // Check all
	if(box[i].correlation<threshLo ||
	   box[i].correlation>threshHi ) { // Candidate for removal
	  if(removeThem) { // Seek and destroy
	    removeBox(box[i].x, box[i].y, TRUE);
	    // DON'T Go to next one - we've got to check this new 
	    // one that's taken i's place...
	  }
	  else { // Just mark it
	    box[i].reference = TRUE;
	    i++;  // Go to next one
	  }
	}
	else // Go to next one
	  i++; 
      }
      while(i<numBoxes);
      
    }
    else if(autoMinMaxThreshold == AMMAVGDENS) { // Avg density
      int beginY, beginX, outHeight, outWidth, j, k;
      float tmpBuffer[20555], totVal;
      
      // The current value of avgdensity is incorrest I'm thinking
      // so let's recalc it here

      outHeight = outWidth = (boxRadius*2)+1;
      for (i=0; i<numBoxes; i++) {
	beginY = box[i].y - boxRadius;
	beginX = box[i].x - boxRadius;
	totVal = 0.0;
	for (j=0; j<outHeight; j++) {
	  yOff = (beginY+j)*file_width + beginX;
	  memcpy(tmpBuffer, &inCore_buffer[yOff], outWidth*sizeof(float));
	  for (k=0; k<outWidth; k++)
	    totVal += tmpBuffer[k];
	}
	box[i].avgdensity = totVal ;
      }
      
      // First find MIN and MAX
      maxVal= minVal = box[0].avgdensity;
      for (i=0; i<numBoxes; i++) {
	box[i].reference = FALSE;
	minVal = MIN(box[i].avgdensity, minVal);
	maxVal = MAX(box[i].avgdensity, maxVal);
      }
      
      threshHi = minVal + (maxVal-minVal)*maxScrollVal;
      threshLo = minVal + (maxVal-minVal)*minScrollVal;
      
      // Now mark those outside range
      i = 0;
      do { // Check all
	if(box[i].avgdensity<threshLo ||
	   box[i].avgdensity>threshHi) { // CAndidate for removal
	  if(removeThem) { // Seek and destroy
	    removeBox(box[i].x, box[i].y, TRUE);
	    // DON'T Go to next one - we've got to check this new 
	    // one that's taken i's place...
	  }
	  else { // Just mark it
	    box[i].reference = TRUE;
	    i++;  // Go to next one
	  }
	}
	else // Go to next one
	  i++; 
      }
      while(i<numBoxes);
	  
    }
    else if(autoMinMaxThreshold == AMMSTDEV) { // stdev
      int beginY, beginX, outHeight, outWidth;
      float totVal;
      
      outHeight = outWidth = (boxRadius*2)+1;
      if(firstAMMSTDEVtime) { // only do once
	for (i=0; i<numBoxes; i++) {
	  beginY = box[i].y - boxRadius;
	  beginX = box[i].x - boxRadius;
	  totVal = 0.0;
	  stuffDisplayBoxBuffer(i, image_fileName.addr, 1);
	  box[i].stdev = purdueHeader.img_std ;
	  printf("box[%d] stdev[%f]\n",i,box[i].stdev);
	}
      }
      firstAMMSTDEVtime = FALSE;

      // First find MIN and MAX
      maxVal= minVal = box[0].stdev;
      for (i=0; i<numBoxes; i++) {
	box[i].reference = FALSE;
	minVal = MIN(box[i].stdev, minVal);
	maxVal = MAX(box[i].stdev, maxVal);
      }
	  
      threshHi = minVal + (maxVal-minVal)*maxScrollVal;
      threshLo = minVal + (maxVal-minVal)*minScrollVal;
      
      // Now mark those outside range
      i = 0;
      do { // Check all
	if(box[i].stdev<threshLo ||
	   box[i].stdev>threshHi ) { // Candidate for removal
	  if(removeThem) { // Seek and destroy
	    removeBox(box[i].x, box[i].y, TRUE);
	    // DON'T Go to next one - we've got to check this new 
	    // one that's taken i's place...
	  }
	  else { //
	    box[i].reference = TRUE;
	    i++;  // Go to next one
	  }
	}
	else // Go to next one
	  i++; 
      }
      while(i<numBoxes);
      
    }
      
    printf("Limits adjusted\n");
    printf("Done.  [%d] boxes.\n",numBoxes);
    if(useGUI) {
      if(temp->reason != XmCR_DRAG)
        DrawWindow();
    
      sprintf(outStr,"Done.  [%d] boxes.",numBoxes);
      TextSet(timeLeftBoxW, outStr, TRUE);
    
      if (tag == REMOVEBOXES) { // Reset this for next time
        minScrollVal = 0.0;
        maxScrollVal = 1.0;
      
      // Now setup the scrolls accordingly
        XtVaSetValues(minBoxScrollW, XmNvalue, 0, NULL);
        XtVaSetValues(maxBoxScrollW, XmNvalue, 100, NULL);
        XFlush(dsply);
      }
    }
  }
  
} // end function autobox_scroll_change



void drawFFTLiteActivate(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: drawFFTLiteActivate
  // They want to draw the window over the net. drawLite causes a
  // smaller drawing area and also forces the user to use the "D"
  // button when he wants a fresh display.

  drawLite = TRUE;
  int tag = *((int *) client_data);

  DrawFFTWindow(NULL, NULL, NULL);
  if(tag == 2)
    DrawWindow();

  drawLite = FALSE;

} // end function drawFFTLiteActivate



void drawLiteActivate(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: drawLiteActivate
  // Draw the window over the net.

  drawLite = TRUE;
  DrawWindow();
  drawLite = FALSE;

} // end function drawLiteActivate



void convertDeltaXY(int *convX, int *convY, float fX, float fY) 
{
  // function name: convertDeltaXY
  // convert X,Y coord to display coords.

  float factor;

  factor =1;   //GCA: arbitrary value
  allMax = MAX(deltaMax,(0.0-deltaMin));
  if(allMax != 0.0)
    factor = 220.0 / allMax;
  *convX = 256.0 + fX*factor;
  *convY = 256.0 - fY*factor;
  
} // end function convertDeltaXY



void convertXY(int *convX, int *convY, float fX, float fY) 
{
  // function name: convertXY
  // Convert X,Y coord to display coords.

  float xFactor, yFactor, tempTheta, sinTheta;

  if(symmetry == 5) { // Draw a diamond shape
    tempTheta = (fY > 90.0) ? 90.0 - (fY-90.0) : fY;
    sinTheta = sin(deg2rad(tempTheta));
    xFactor = (7.0*sinTheta) * (32.0/phiMax);
  }
  else
    xFactor = 7.0 * (32.0/phiMax);

  yFactor = 21.0 * (90.0 - 69.0)/(theta2 - theta1);

  *convX = fX*xFactor + 256;
  *convY = (fY-theta1)*yFactor + 20;
  
} // end function convertXY



void batchjob_allasu_activate(Widget widget, XtPointer client_data, 
			      XtPointer call_data)
{
  // function name: batchjob_allasu_activate
  // Batch process particles.
  
  int i;
  for (i=0; i<totParamFiles; i++)
    read_param(batchListStr[i], FALSE);

  DrawICOSWindow(NULL, NULL, NULL);
  raiseIt(partprocScreenW);

} // end function batchjob_allasu_activate



void batchjob_submit_activate(Widget widget, XtPointer client_data, 
			      XtPointer call_data)
{
  // function name: batchjob_submit_activate
  // Submit a batch job for all particles at which you've been looking.

  char sysCall[128], mailNode[128];
  FILE *fOut;
  
  if ( (fOut = fopen(paramBatchJobFilename,"w")) != NULL ) {
    fprintf(fOut, "%s\n", XmTextGetString(batchTextW));
    fclose(fOut);
    
    sprintf(mailNode,"%s@baker1",getenv("USER"));
    
    sprintf(sysCall,"%s < /dev/null | mail %s &",
	    paramBatchJobFilename, mailNode);
    
    system(sysCall);
    askThem("Batch Job submitted to OS for execution.", XmDIALOG_INFORMATION);
  }

} // end function batchjob_submit_activate



void batch_focus(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: batch_focus
  // Give helpful feedback about the meaning of each widget.

  if(widget==batchModeW)
    sprintf(outStr,"Enter MODE.");
  else if(widget==batchBinW)
    sprintf(outStr,"Enter BIN.");
  else if(widget==batchSymW)
    sprintf(outStr,"Enter SYM.");
  else if(widget==batchDtheW)
    sprintf(outStr,"Enter DTHE.");
  else if(widget==batchDphi90W)
    sprintf(outStr,"Enter DPHI90.");
  else if(widget==batchIlistW)
    sprintf(outStr,"Enter ILIST.");
  else if(widget==batchRadLoW)
    sprintf(outStr,"Enter RADLO.");
  else if(widget==batchRadHiW)
    sprintf(outStr,"Enter RADHI.");
  else if(widget==batchRadStepW)
    sprintf(outStr,"Enter RADSTEP.");
  else if(widget==batchAnnLoW)
    sprintf(outStr,"Enter ANNLO.");
  else if(widget==batchAnnHiW)
    sprintf(outStr,"Enter ANNHI.");
  else if(widget==batchResLoW)
    sprintf(outStr,"Enter RESLO.");
  else if(widget==batchResHiW)
    sprintf(outStr,"Enter RESHI.");
  else if(widget==batchJCutW)
    sprintf(outStr,"Enter JCUT.");
  else if(widget==batchSigW)
    sprintf(outStr,"Enter SIG.");
  else if(widget==batchMagCenW)
    sprintf(outStr,"Enter MAGCEN.");
  else if(widget==batchMagStepW)
    sprintf(outStr,"Enter MAGSTEP.");
  else if(widget==batchMagNumW)
    sprintf(outStr,"Enter MAGNUM.");
  else if(widget==batchMagNormW)
    sprintf(outStr,"Enter MAGNORM.");
  else if(widget==batch3DMapW)
    sprintf(outStr,"Enter 3D MAP Filename.");
  else if(widget==batchPrjsFileW)
    sprintf(outStr,"Enter PRJS Filename.");
  else if(widget==batchPftsFileW)
    sprintf(outStr,"Enter PFTS Filename.");

  doLabel(outStr, batchHelpLabelW, DEFAULT);

} // end function batch_focus



void batchparam_list_proc(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: batchparam_list_proc
  // Read the specific file from the batch job and display.

  char *thisStr;
  XmListCallbackStruct *list = (XmListCallbackStruct *) call_data;

  if(!XmStringGetLtoR(list->item, XmSTRING_DEFAULT_CHARSET, &thisStr))
    printf("Error extracting List value.\n");
  
  read_param(thisStr,TRUE);
  
  // Update display
  DrawWindow();
  DrawICOSWindow(NULL, NULL, NULL);

  XtFree(thisStr);  
  raiseIt(partprocScreenW);

} // end function batchparam_list_proc



void partproc_delta_work(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: partproc_delta_work
  // Add Delta x,y to all particles. Do the work here.

  P_INFO *thisInfo;
  
  thisInfo = info.info;
  while(thisInfo != NULL) {
    thisInfo->x = thisInfo->x + wstrtod(dxyXTextW);
    thisInfo->y = thisInfo->y + wstrtod(dxyYTextW);
    thisInfo->origX = thisInfo->x;
    thisInfo->origY = thisInfo->y;
    thisInfo->deltaX = thisInfo->x - (purdueHeader.numCols/2.0 - 0.5);
    thisInfo->deltaY = thisInfo->y - (purdueHeader.numRows/2.0 - 0.5);

    deltaMax = MAX(thisInfo->deltaX, deltaMax);
    deltaMax = MAX(thisInfo->deltaY, deltaMax);
    deltaMin = MIN(thisInfo->deltaX, deltaMin);
    deltaMin = MIN(thisInfo->deltaY, deltaMin);
    deltaXmin= MIN(thisInfo->deltaX, deltaXmin);
    deltaYmin= MIN(thisInfo->deltaY, deltaYmin);
    deltaXmax= MAX(thisInfo->deltaX, deltaXmax);
    deltaYmax= MAX(thisInfo->deltaY, deltaYmax);
    thisInfo->distance = sqrt(thisInfo->deltaX*thisInfo->deltaX +
			      thisInfo->deltaY*thisInfo->deltaY);
    distanceMin = MIN(distanceMin, thisInfo->distance);
    distanceMax = MAX(distanceMax, thisInfo->distance);
    
    thisInfo = thisInfo->next;
  }
  
  XtUnmanageChild(dxyScreenW);
  stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		   packImg, purdueHeader.nimg, imgNum);
  DrawWindow();
  DrawICOSWindow(NULL, NULL, NULL);
  XtUnmanageChild(dxyScreenW);
  
} // end function partproc_delta_work



void partproc_deltaomega_work(Widget widget, XtPointer client_data, 
			      XtPointer call_data)
{
  // function name: partproc_deltaomega_work
  // Add Delta omega to all particles. Do the work here.

  float deltaOmega, omega;
  P_INFO *thisInfo;

  deltaOmega = wstrtod(domegaTextW);

  thisInfo = info.info;
  while(thisInfo != NULL) { 
    omega = thisInfo->omega + deltaOmega;
    if(omega > 360.0)
      omega -= 360.0;
    if(omega < 0.0)
      omega += 360.0;
    
    thisInfo->omega = omega; 
    thisInfo = thisInfo->next;
  }

  XtUnmanageChild(dxyScreenW);
  stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		   packImg, purdueHeader.nimg, imgNum);
  DrawWindow();
  DrawICOSWindow(NULL, NULL, NULL);
  XtUnmanageChild(domegaScreenW);

} // end function partproc_deltaomega_work



void manipPP_change(Widget widget, XtPointer client_data,  XtPointer call_data)
{
  // function name: manipPP_change
  // Change partproc concepts

  doManipPP = *((int *) client_data);

  if(doManipPP == RANDCENTER)  // Radomize the centers
    TextSet(manipPPValueW, "1.0 STD", FALSE);
  else if(doManipPP == UNKPP)  // future use
    printf("Not yet.\n");

} // end function manipPP_change



void partproc_manip_work(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: partproc_manip_work
  // Do the work here

  int pMin, pMax;
  float numStd, factor, noise, mean, stDev, tmp1;
  P_INFO *thisInfo;
  
  if(doManipPP==RANDCENTER) {// Randomize the centers
    numStd = wstrtod(manipPPValueW);
    
    particleCalcStats(&pMin, &pMax, &mean, &stDev);
    
    noise  = stDev * numStd;   
    // Get our fraction of the random number
    factor= 1.0 / RAND_MAX;
    
    thisInfo = info.info;
    while(thisInfo != NULL) {
      tmp1 = (random()*factor - 0.5)*2.0*noise;
      thisInfo->x += tmp1;
      
      tmp1 = (random()*factor - 0.5)*2.0*noise;
      thisInfo->y += tmp1;
      
      thisInfo->x = MIN(MAX(thisInfo->x,0.0), purdueHeader.numCols);
      thisInfo->y = MIN(MAX(thisInfo->y,0.0), purdueHeader.numRows);
      
      thisInfo = thisInfo->next;
    }   
  }

  stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		   packImg, purdueHeader.nimg, imgNum);
  DrawWindow();
  DrawICOSWindow(NULL, NULL, NULL);
  XtUnmanageChild(manipPPScreenW);
  
} // end function partproc_manip_work



void partproc_deltaoldNew_work(Widget widget, XtPointer client_data, 
			      XtPointer call_data)
{
  // function name: partproc_deltaoldNew_work
  // Add Delta omega to all particles to convert old parameter files
  // into new ones (July 1999). Do the work here.

  float deltaOmega = 90.0, omega;
  P_INFO *thisInfo;

  thisInfo = info.info;
  while(thisInfo != NULL) {
      omega = thisInfo->omega + deltaOmega;
      if(omega > 360.0)
	omega -= 360.0;
      if(omega < 0.0)
	omega += 360.0;

      thisInfo->omega = omega; 
      thisInfo = thisInfo->next;
  }

  XtUnmanageChild(dxyScreenW);
  stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		   packImg, purdueHeader.nimg, imgNum);
  DrawWindow();
  DrawICOSWindow(NULL, NULL, NULL);
  XtUnmanageChild(domegaScreenW);
  
} // end function partproc_deltaoldNew_work



void partproc_rebox_activate(Widget widget, XtPointer client_data, 
			     XtPointer call_data)
{
  // function name: partproc_rebox_activate
  // Rebox the specifed particle(s)

  int tag = *((int *) client_data);
  int i, k, l, m, yOut, startI, endI, box_width;
  int yIn, outHeight, outWidth, ok2go;
  int firstCol, firstRow, thisBox, thisX, thisY;
  static int beenWarned = FALSE;
  float deltaX, deltaY, *thisBoxBuf;
  FILE *fOut, *fIn;
  
  ok2go = TRUE;
  if(EQU(paramScanFilename,"")) { // No scan file read
    askThem("No scan file read.", XmDIALOG_INFORMATION);
    ok2go = FALSE;
  }
  
  if(ok2go) {
    if(ok2go && 
       askThem("This is a destructive operation.  The particle\n\
disk file will be permanently changed as well as\n\
the box coordinates file.  Do you want to continue?",
	       XmDIALOG_QUESTION)) { // destruct?
      ok2go = TRUE;
    }
    else 
      ok2go = FALSE;
  }
  
  beenWarned = TRUE;
  
  if(ok2go && tag==ALL) { // destruct?
    if(askThem("You've requested that ALL particles be reboxed.\n\
Is this what you *really* want?",
	       XmDIALOG_QUESTION))
      ok2go = TRUE;
    else 
      ok2go = FALSE;
  }
  
  if(ok2go) {

    if(tag==0) { // Do a single one
      startI = imgNum;
      endI   = startI + 1;
    }
    else if(tag==ALL) { // Do them all
      startI = 1;
      endI   = purdueHeader.nimg+1;
    }
    
    for (m=startI; m<endI; m++) {  
      thisX = 0;
      thisY = 0;
      
      findNearestParticle(&thisX,&thisY,&foundPart,FALSE,i,FALSE,0);
      
      deltaX = foundPart->x - (purdueHeader.numCols/2+1);
      deltaY = foundPart->y - (purdueHeader.numRows/2+1);
      
      thisBox = foundPart->particle_number-1;
      boxRadius = purdueHeader.numCols / 2;
      box[thisBox].x += deltaX;
      box[thisBox].y += deltaY;
      
      // Get these so that they are NINT() rounded
      box[thisBox].x += 0.0499999;
      box[thisBox].y += 0.0499999;
      
      firstCol = box[thisBox].x - boxRadius;
      firstRow = box[thisBox].y - boxRadius;
      
      if(box[thisBox].edit_mirror_buffer)
	free(box[thisBox].edit_mirror_buffer);
      
      box[thisBox].inMemory = FALSE;
      thisBoxBuf = packImg[thisBox];
      
      // Get dimensions
      outHeight= outWidth = boxRadius*2 + 1;
      box_width = purdueHeader.numCols;
      
      fIn  = fopen(paramScanFilename, "r");
      for (i=0; i<outHeight; i++) {
	yIn = 1024 + (firstRow+i)*scanCols + firstCol; 
	sysError = fseek(fIn,yIn,0);
	if(sysError) {
	  askThem("Error: fseek in reading new particle.", 
		  XmDIALOG_INFORMATION);
	  goto noWriteThis;
	}
	yOut = (i * box_width);
	// We need to read the floatInt*4 format here for mode==46!!
	fread(&thisBoxBuf[yOut], sizeof(BYTE), box_width, fIn);
      }
      
      // Calc the background
      backChar= (float)
	calcFloatBackground(thisBoxBuf, outWidth, outHeight, 
			    box_width, fileType, bkgrndMethod,
			    boxMethod, packRadius, boxRadius,
			    featherRadius, recipSizeFactor, tryColors);
      
      // Draw the background  for all > radius in box
      for (k=0; k<outHeight; k++) { // Do the y direction
	int yOut, yCent, xCent;
	float dist;
	
	// Where is the center?
	yOut = k * box_width;
	yCent = boxRadius;
	xCent = boxRadius;
	
	// Stuff all > radius in background character
	for (l=0; l<outWidth; l++) { // Stuff the outbuffer
	  dist = sqrt((k-yCent)*(k-yCent) + (l-xCent)*(l-xCent));
	  
	  if(dist > boxRadius+1.0)
	    thisBoxBuf[yOut+l]=  backChar; 
	}
      }
      
      fOut = fopen(paramBoxesFilename,"r+");
      // Now write this new box to the disk file
      yOut = 512 + 512*m + (m-1)*outWidth*outHeight;
      sysError = fseek(fOut,yOut,0);
      if(sysError) {
	askThem("Error: fseek in writing new particle.", XmDIALOG_INFORMATION);
	goto noWriteThis;
      }
      fwrite(thisBoxBuf, sizeof(BYTE), (outWidth*outHeight), fOut);
      
      // Now write new Box Coords file
      writeBoxCoords(paramBoxCoordsFilename, PARAMSCAN);
      
      foundPart->x = boxRadius + 1.0;
      foundPart->y = boxRadius + 1.0;
      
    noWriteThis:;
      fclose(fOut);
      fclose(fIn);  
    }
    
    findNearestParticle(&thisX,&thisY,&foundPart,FALSE,imgNum,TRUE,0);
  }
  
} // end function partproc_rebox_activate



void cavgPP_change(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: cavgPP_change
  // Change circavg concepts in partproc screen.
  
  int tag = *((int *) client_data);

  if(tag == CAVGDO) { // Circ Avg
    doCavgPP = tag;
    XtUnmanageChild(cavgPPNflW);
    XtUnmanageChild(cavgPPNftW);
  }
  else if(tag == NSYMDO) { // nsym
    doCavgPP = tag;
    XtManageChild(cavgPPNflW);
    XtManageChild(cavgPPNftW);
  }
  else if(tag == PARAMXY || tag == BOXXY) { // param/box xy
    cavgPPcenter = tag;
  }

} // end function cavgPP_change



void cavgPP_do_activate(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: cavgPP_do_activate
  // Do circular average
  
  int nFold, i, j, numAlloc, Ncol, Nrow, inum;
  float *thisRint, xcen, ycen;
  float *thisImg, *pkImg;
  FILE *fOut;
  
  nFold = wstrtod(cavgPPNftW);
  XFlush(dsply);
  
  if(doCavgPP == 0) { // warning
    askThem("Must choose whether you want Circ Averaging or nSymm applied.",
	    XmDIALOG_INFORMATION);
    return;
  }
  
  if(cavgPPcenter == 0) {
    askThem("You must choose a method for centers (param file or box).",
	    XmDIALOG_INFORMATION);
    return;
  }
  
  if ( (fOut = fopen("circular_avg.asc","w")) == NULL ) {
    askThem("Cannot open file circular_avg.asc.",
	    XmDIALOG_INFORMATION);
    return;
  }
  
  if(nFold > 99 || nFold < 2) {
    askThem("N Fold symmetry outside limits - resetting.",
	    XmDIALOG_INFORMATION);
    nFold = MIN(MAX(nFold,2),99);
    TextSetInt(cavgPPNftW, nFold, FALSE);
  }
  
  // OK now call routine to compute
  Ncol = purdueHeader.numCols;
  Nrow = purdueHeader.numRows;
  
  if(totCavgPPrint > 0)
    for (i=0; i<totCavgPPrint; i++)
      free(cavgPPrint[i]);
  
  totCavgPPrint = purdueHeader.nimg;
  
  // malloc() by *6 because we want to go to the edges of the box and
  // it has an interg==3 in img_sym()
  numAlloc = purdueHeader.numCols * 6;
  for(i=0; i<totCavgPPrint; i++)
    cavgPPrint[i] = (float *)malloc(sizeof(float)*numAlloc);
  
  for (i=0; i<totCavgPPrint; i++) {
    thisRint = cavgPPrint[i];
    init_floatarr(thisRint, numAlloc, 0.0);
  }
  
  thisImg = malloc(sizeof(float)*purdueHeader.numCols*purdueHeader.numRows);
  
  // 2.0 for interp in img_sym() and 0.5 for radius not diameter 
  // 0.90 for 90% - we're plotting too much ... ???
  
  inum = (int)
    ((2.0*0.5*0.90)*sqrt((double)
			 (purdueHeader.numCols * purdueHeader.numCols +
			  purdueHeader.numRows * purdueHeader.numRows)));
  
  fprintf(fOut,"Total particles                 = %d\n",
	  purdueHeader.nimg);
  fprintf(fOut,"Number data points per particle = %d\n",
	  inum);
  fprintf(fOut,"*******************************************\n");
  
  for(i=0; i<purdueHeader.nimg; i++) {
    int nrot, img_bkg;
    
    if(doCavgPP == CAVGDO)
      nrot = 101;
    else 
      nrot = nFold;
    
    pkImg = packImg[i];
    thisRint = cavgPPrint[i];
    
    for (j=0; j<purdueHeader.numRows*purdueHeader.numCols; j++)
      thisImg[j] = pkImg[j];
    
    if(cavgPPcenter == PARAMXY) { // Find param center
      P_INFO *thisInfo;
      
      thisInfo = info.info;
      while(thisInfo != NULL) {
	if(thisInfo->particle_number == i+1) { // Found it
	  xcen = (float)(thisInfo->x);
	  ycen = (float)(thisInfo->y);
	  
	  printf("Found id[%d,%d]  x[%f] y[%f]\n",
		 thisInfo->particle_number, i,
		 thisInfo->x, thisInfo->y);
	}
	thisInfo = thisInfo->next;
      }
    }
    else { // Use box center
      xcen = purdueHeader.numCols/2.0;
      ycen = purdueHeader.numRows/2.0;
    }
    
    r4_img_sym__(thisImg, &Ncol, &Nrow, &xcen, &ycen, &nrot, 
		&img_bkg, thisRint, &numAlloc);
    
    fprintf(fOut,"Particle #%d\n",i+1);
    // 2.0 because that's the interp used in img_sym
    for (j=0; j<inum; j++)
      fprintf(fOut,"%f  %f\n",(j / 2.0),thisRint[j]);
    fprintf(fOut,"*******************************************\n");
    
  }
  
  free(thisImg);
  
  fclose(fOut);
  askThem("Circular averages written to file 'circular_avg.asc'.",
	  XmDIALOG_INFORMATION);
  
} // end function cavgPP_do_activate



void partproc_sort_activate(Widget widget, XtPointer client_data, 
			    XtPointer call_data)
{
  // function name: partproc_sort_activate
  // Sort particles

  icosaMotioning = FALSE;
  XtSetArg(arglst[0], XmNset, FALSE );

  XtSetValues(idtW, arglst, 1); 
  XtSetValues(thetatW, arglst, 1); 
  XtSetValues(phitW, arglst, 1); 
  XtSetValues(omegatW, arglst, 1); 
  XtSetValues(xtW, arglst, 1); 
  XtSetValues(ytW, arglst, 1); 
  XtSetValues(deltaXTW, arglst, 1); 
  XtSetValues(deltaYTW, arglst, 1); 
  XtSetValues(absDeltaXTW, arglst, 1); 
  XtSetValues(absDeltaYTW, arglst, 1); 
  XtSetValues(distanceTW, arglst, 1); 
  XtSetValues(magtW, arglst, 1); 
  XtSetValues(pftccW, arglst, 1); 
  XtSetValues(prjccW, arglst, 1); 
  XtSetValues(cmpccW, arglst, 1); 

  XFlush(dsply);

  idtoggle = thetatoggle = phitoggle = omegatoggle = xtoggle = ytoggle=
    magtoggle = cc1toggle = cc2toggle = cc3toggle = distancetoggle =
    deltaXtoggle = deltaYtoggle = absDeltaXtoggle = absDeltaYtoggle = FALSE;

  XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE] );
  XtSetValues(magtW, arglst, 1); 
  XtSetValues(pftccW, arglst, 1); 
  XtSetValues(prjccW, arglst, 1); 
  XtSetValues(cmpccW, arglst, 1); 

  doLabel("Choose key for SORT: ", parmTogsLabelW, DEFAULT);

  XtManageChild(parmTogsW);
  projASU = FALSE;
  XtUnmanageChild(projasuW);
  XtUnmanageChild(parmheadW);
  TextSet(partPointTextW, "Choose SORT key.", FALSE);
  icoDispMode = ICOSORT;
  icoSortLevel = 0;
  DrawICOSWindow(NULL, NULL, NULL);

} // end function partproc_sort_activate



void paramFile_browse_activate(Widget widget, XtPointer client_data, 
			       XtPointer call_data)
{
  // function name: paramFile_browse_activate
  // Read file from disk

  int tag = *((int *) client_data);
  filterFiles("*",ALL); 
  
  if(tag == PARAMFILE || tag == PARAMMAP ||
     tag == PARAMSCAN || tag == BATCHJOB)
    readType = tag;

  readWrite= READING;
  raiseIt(fileSelW);
  
} // end function paramFile_browse_activate



void projContour_cr_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{
  // function name: projContour_cr_activate
  // Draw the contour on the display.

  if(EQU(paramMapFilename,"")) {
    askThem("No MAP open can't contour the projection.",
	    XmDIALOG_INFORMATION);
  }
  else {
    stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		     packImg, purdueHeader.nimg, imgNum);
    DrawWindow(); 
  }
  
} // end function projContour_cr_activate



void paramScanFile_read_activate(Widget widget, int *tag, 
				 XmAnyCallbackStruct *callback_data)
{
  // function name: paramScanFile_read_activate
  // Show status screen and read the param file.

  raiseIt(statusScreenW);
  XSync(dsply,0);

  // Now read the SCAN file
  read_param_scan(XmTextGetString(paramScanFileTextW)); 

  XFlush(dsply);

} // end function paramScanFile_read_activate



void paramMapFile_read_activate(Widget widget, int *tag, 
				XmAnyCallbackStruct *callback_data)
{
  // function name: paramMapFile_read_activate
  // Show status and read MAP listed in the param file.

  raiseIt(statusScreenW);
  XSync(dsply,0);

  // Now read the MAP file
  read_param_map(XmTextGetString(paramMapFileTextW)); 
  
  XFlush(dsply);

} // end function paramMapFile_read_activate



void paramFile_read_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{
  // function name: paramFile_read_activate
  // Callback to actually read the param file.

  read_param(XmTextGetString(paramFileTextW), TRUE);
  XFlush(dsply);

} // end function paramFile_read_activate



void write_param(char *fName, int flag)
{
  // function name: write_param
  //
  // Write particle parameter file
  // flag == True,  write header plus ptle records
  // flag == False, write header only
  
  P_INFO *thisInfo;
  FILE *fOut;
  
  if(haveParamFile) {
    
    if ( (fOut = fopen(fName,"r")) != NULL ) {
      fclose(fOut);
      sprintf(outStr,"[%s] Already exists - Overwrite anyway?",fName);
      if(!askThem(outStr, XmDIALOG_QUESTION))
	return;
    }
    
    if ( (fOut = fopen(fName,"w")) != NULL ) {
      
      // Write packed boxes name
      fprintf(fOut,"%s\n",paramBoxesFilename);
      
      // Write second line
      fprintf(fOut,"%s, ", XmTextGetString(pixsizeW));
      fprintf(fOut,"%s, ", XmTextGetString(unitsW));
      fprintf(fOut,"%s, ", XmTextGetString(voltsW));
      fprintf(fOut,"%s, ", XmTextGetString(ampfacsW));
      fprintf(fOut,"%s, ", XmTextGetString(delfMajW));
      fprintf(fOut,"%s, ", XmTextGetString(delfMinW));
      fprintf(fOut,"%s, ", XmTextGetString(angMajW));
      fprintf(fOut,"%s\n", XmTextGetString(csCoefW));
      
      // Now write any untoggled particles
      if(flag) {
	thisInfo = info.info;
	while(thisInfo != NULL) {
	  if(!thisInfo->toggle) {
	    fprintf
	      (fOut,
	       "%7d,%8.3f,%8.3f,%8.3f,%8.3f,%7.3f,%6.3f,%6.3f,%6.3f,%6.3f\n",
	       thisInfo->particle_number, thisInfo->theta,
	       thisInfo->phi, thisInfo->omega,
	       thisInfo->x, thisInfo->y, thisInfo->mag,
	       thisInfo->cc1, thisInfo->cc2, thisInfo->cc3);
	  }
	  else {
	    fprintf
	      (fOut,
	       "%7d,%8.3f,%8.3f,%8.3f,%8.3f,%7.3f,%6.3f,%6.3f,%6.3f,%6.3f\n",
	       (0 - thisInfo->particle_number), thisInfo->theta,
	       thisInfo->phi, thisInfo->omega,
	       thisInfo->x, thisInfo->y, thisInfo->mag,
	       thisInfo->cc1, thisInfo->cc2, thisInfo->cc3);
	  }
	  thisInfo = thisInfo->next;
	}
	
	
	if(thisInfo==NULL) { 
	  // They called us from FFT screen - just default it 
	  int ll;
	  float factor, randNum, thisTheta, thisPhi, thisOmega;
	  
	  // Get our fraction of the random number
	  factor= 1.0 / RAND_MAX;
	  randNum= (random() * factor);
	  
	  for (ll=0; ll<purdueHeader.nimg; ll++) {
	    randNum= random() * factor;
	    thisTheta = 69.0 + 21.0*randNum;
	    randNum= random() * factor;
	    thisPhi = ((thisTheta - 69.0)*31.7/21.0) * randNum;
	    
	    // XDY wants some negative values for phi to
	    // fill the ASU ...  randNum is between 0.0 and 1.0
	    randNum= random() * factor;
	    if(randNum > 0.5)
	      thisPhi *= -1.0;
	    
	    randNum= random() * factor;
	    thisOmega = 360.0 * randNum;
	    fprintf
	      (fOut,
	       "%7d,%8.3f,%8.3f,%8.3f,%8.3f,%7.3f,%6.3f,%6.3f,%6.3f,%6.3f\n",
	       ll+1, thisTheta,
	       thisPhi, thisOmega,
	       (float)(purdueHeader.numCols/2), (float)(purdueHeader.numRows/2),
	       1.000,
	       -99.9, -99.9, -99.9);
	  }
	}
      }

      fclose(fOut);
      
      sprintf(outStr,"[%s] file written.",fName);
      askThem(outStr, XmDIALOG_INFORMATION);
    }
    else {
      sprintf(outStr,"Can't open [%s] for writing.",fName);
      askThem(outStr, XmDIALOG_INFORMATION);
    }
  }
  else {
    sprintf(outStr,"No param file open.");
    askThem(outStr, XmDIALOG_INFORMATION);
  }
  
} // end function write_param



void read_batchjob(char *fName)
{
  // function name: read_batchjob
  // Read the batch file - VMS specific

  int i, beginParms;
  char inStr[5120];
  char parm1[32], parm2[32], parm3[32], parm4[32];
  char parm5[32], parm6[32], parm7[32];

  FILE *fIn;
  
  TextSet(batchFileTextW, fName, FALSE); 
  sprintf(paramBatchJobFilename,"%s",fName);
  
  if ( (fIn= fopen(fName,"r")) != NULL ) {
    totParamFiles= 0;
    
    for (i=0; i<MAXPACKEDFILES; i++) {
      if(batchListStr[i])
	free(batchListStr[i]);
      batchListStr[i]= 0;
    }
    
    fread(inStr,sizeof(char),5100,fIn);
    fclose(fIn);      
    
    TextSet(batchTextW, inStr, FALSE); 
    
    // Now read it one field at a time
    fIn= fopen(fName,"r");
    
    XmListDeleteAllItems(batchParamFileW);
    beginParms = 0;
    while(!feof(fIn)) {
      get_str(fIn, '\n', inStr);
      if(inStr[0]=='e' && inStr[1]=='c' &&
	 inStr[2]=='h' && inStr[3]=='o') {

	if(beginParms == 0) { // First line
	  sscanf(inStr,"%s%s%s%s%s%s%s",
		 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
	  TextSet(batchModeW,   parm2, FALSE);                    
	  TextSet(batchBinW ,   parm3, FALSE);
	  TextSet(batchSymW,    parm4, FALSE);
	  TextSet(batchDtheW,   parm5, FALSE);
	  TextSet(batchDphi90W, parm6, FALSE);
	  TextSet(batchIlistW,  parm7, FALSE);
	}
	else if(beginParms == 1) { // Second line
	  sscanf(inStr,"%s%s%s%s%s%s%s",
		 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
	  TextSet(batchRadLoW,   parm2, FALSE);
	  TextSet(batchRadHiW,   parm3, FALSE);
	  TextSet(batchRadStepW, parm4, FALSE);
	  TextSet(batchAnnLoW,   parm5, FALSE);
	  TextSet(batchAnnHiW,   parm6, FALSE);
	}
	else if(beginParms == 2) { // thrid line 3
	  sscanf(inStr,"%s%s%s%s%s%s%s",
		 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
	  TextSet(batchResLoW, parm2, FALSE);
	  TextSet(batchResHiW, parm3, FALSE);
	  TextSet(batchJCutW,  parm4, FALSE);
	  TextSet(batchSigW,   parm5, FALSE);
	}
	else if(beginParms == 3) { // fourth line 4
	  sscanf(inStr,"%s%s%s%s%s%s%s",
		 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
	  TextSet(batchMagCenW,  parm2, FALSE);
	  TextSet(batchMagStepW, parm3, FALSE);
	  TextSet(batchMagNumW,  parm4, FALSE);
	  TextSet(batchMagNormW, parm5, FALSE);
	}
	else if(beginParms == 4) { // fifth line 5
	  sscanf(inStr,"%s%s%s%s%s%s%s",
		 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
	  TextSet(batch3DMapW, parm2, FALSE);
	  read_param_map(parm2);
	}
	else if(beginParms == 5) { // sixth line
	  sscanf(inStr,"%s%s%s%s%s%s%s",
		 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
	  TextSet(batchPrjsFileW, parm2, FALSE);
	}
	else if(beginParms == 6) { // seventh line 7
	  sscanf(inStr,"%s%s%s%s%s%s%s",
		 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
	  TextSet(batchPftsFileW, parm2, FALSE);
	}
	else if(beginParms >= 7) { // lines 8+
	  sscanf(inStr,"%s%s%s%s%s%s%s",
		 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
	  if(!EQU(parm2,"")) {
	    XmListAddItem(batchParamFileW, XmStringCreateLocalized(parm2), 
			  (beginParms-7)+1);  
	    XFlush(dsply);
	  }
	}
	
	beginParms++;
	
      }
    }

    fclose(fIn);      
    
  }
  else {
    TextSet(batchFileTextW, "", FALSE); 
    paramBatchJobFilename[0]='\0';
    sprintf(outStr,"Can't open [%s]\n",fName);
    askThem(outStr, XmDIALOG_INFORMATION);
    goto destructor;
  }
  
 destructor:;
  XtUnmanageChild(statusScreenW);
  XFlush(dsply);
  
} // end function read_batchjob



void read_param_scan(char *fName)
{
  // function name: read_param_scan
  // Read the scan file

  int i;
  char *s1;
  forPifHead_t thisHead;
  fString_t rWrite, inFile;
  FILE *fIn;
  
  if(!haveParamFile) {
    askThem("You must read a parameter file first.", XmDIALOG_INFORMATION);
    XtUnmanageChild(statusScreenW);
    return;
  }
  
  TextSet(paramScanFileTextW, fName, FALSE); 
  sprintf(paramScanFilename,"%s",fName);
  
  if ( (fIn = fopen(fName,"r")) != NULL ) {
    fclose(fIn);      
    i = 7;
    
    rWrite.addr = malloc(sizeof(char)*4);
    rWrite.length = 1;
    inFile.addr = malloc(sizeof(char)*64);
    inFile.length = 64;
    
    sprintf(inFile.addr,"%s",fName);
    sprintf(rWrite.addr,"r");
    
    pif_open(inFile.addr, &i, rWrite.addr);
    
    free(rWrite.addr);
    free(inFile.addr);
    
    thisHead.luNum = i;
    thisHead.mode  = -99;
    pif_read_gh(&thisHead);
    
    if(thisHead.mode != 0) { // BAD mode
      TextSet(paramScanFileTextW, "", FALSE); 
      paramScanFilename[0]='\0';
      sprintf(outStr,"Invalid PIF mode for scanned data [%s]\n",fName);
      askThem(outStr, XmDIALOG_INFORMATION);
      pif_close(&i);
      XtUnmanageChild(statusScreenW);
      return;
    }
    
    scanCols = thisHead.numCols;
    scanRows = thisHead.numRows;
    pif_close(&i);
  }
  else {
    TextSet(paramScanFileTextW, "", FALSE); 
    paramScanFilename[0]='\0';
    sprintf(outStr,"Can't open [%s]\n",fName);
    askThem(outStr, XmDIALOG_INFORMATION);
    XtUnmanageChild(statusScreenW);
    return;
  }
  
  s1 = strrchr(fName,'.');
  if(s1)
    *s1 = '\0';
  
  sprintf(paramBoxCoordsFilename,"%s.bcrd",fName);
  sprintf(outStr,"Reading [%s]",paramBoxCoordsFilename);
  if(checkStatus(outStr)) {
    XtUnmanageChild(statusScreenW);
    return;
  }

  readBoxCoords(paramBoxCoordsFilename);  
  XtUnmanageChild(statusScreenW);
  
} // end function read_param_scan



void read_param_map(char *fName)
{
  // function name: read_param_map
  // Read the map file

  char outStr[512];
  int i, strtCol, strtRow, strtSec;
  int thisNcol, thisNrow, thisNsec;
  off_t imgOffset;
  int convTsb, binCol, binRow, binSec;

  FILE *fIn;
  forPifHead_t thisHead;
  fString_t rWrite, inFile;

  printf("Reading [%s]\n",fName);
  TextSet(paramMapFileTextW, fName, FALSE); 
  sprintf(paramMapFilename,"%s",fName);

  if ( (fIn= fopen(fName,"r")) == NULL ) {
    sprintf(outStr,"Can't open [%s]\n",fName);
    askThem(outStr, XmDIALOG_INFORMATION);
    return;
  }

  fclose(fIn);
  
  sprintf(outStr,"Reading [%s]",fName);
  if(checkStatus(outStr)) {
    XtUnmanageChild(statusScreenW);
    return;
  }
  
  if(projASU) { // But *do* get rid of any unwanted memory
    for (i=0; i<purdueHeader.nimg; i++)
      freeNullFloat(&packImg[i]);
    projASU = FALSE;
    purdueHeader.nimg = 0;
    lastPackedNumImgs= 0;
  }
  
  thisHead.luNum= 4;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 1;
  inFile.addr = malloc(sizeof(char)*64);
  inFile.length = 64;
  sprintf(inFile.addr,"%s",fName);
  sprintf(rWrite.addr,"r");
  
  pif_open(inFile.addr, &thisHead.luNum, rWrite.addr);
  
  pif_read_gh(&thisHead);
  pifMode  = thisHead.mode;
  thisNcol = thisHead.numCols;
  thisNrow = thisHead.numRows;
  thisNsec = thisHead.numSecs;
  
  if(pifMode!=20) { // Not a map!!
    sprintf(outStr,"Invalid MAP mode [%s]\n",fName);
    askThem(outStr, XmDIALOG_INFORMATION);
    XtUnmanageChild(statusScreenW);
    return;
  }
  
  thisHead.imgNum = 1;
  pif_read_dh(&thisHead);
  imgOffset = thisHead.imgOffset;
  recipFloatScaleFactor = thisHead.recipFloatScaleFactor;
  
  fileBinSecs = binSec  = thisNsec;
  fileBinRows = binRow  = thisNrow;
  fileBinCols = binCol  = thisNcol;
  fileBinFactor = binFactor= thisHead.binFactor;
  
  // Normal
  if(binFactor ==2 ) { // 1/2
    binSec  = thisNsec/2+1;
    binRow  = thisNrow/2+1;
    binCol  = thisNcol/2+1;
    if(binSec%2 == 0)
      binSec--;
    if(binRow%2 == 0)
      binRow--;
    if(binCol%2 == 0)
      binCol--;
  }
  
  freeNullFloat(&float_image_buffer1);
  CK(float_image_buffer1 = (float *)malloc(sizeof(float)*
					   binCol*binRow*binSec));
  
  maxMapPixel = -HUGE;
  minMapPixel = HUGE;
  convTsb = TRUE;
  pif_tsbworld(convTsb);
  
  binFactor = 1;
  strtCol = strtRow = strtSec = 1;
  pif_read_map_short_image_(&thisHead.luNum, 
			    float_image_buffer1,
			    &imgOffset,
			    &binCol, &binRow, &binSec,
			    &recipFloatScaleFactor,
			    &minMapPixel,&maxMapPixel, 
			    &binFactor,
			    &strtCol, &strtRow, &strtSec,
			    &binCol, &binRow, &binSec);
  
  pif_close(&thisHead.luNum);
  
  thisNcol = binCol;
  thisNrow = binRow;
  thisNsec = binSec;
  origNumSecs= thisNsec;  
  
  free(rWrite.addr);
  free(inFile.addr);
  
  xCen = thisNcol/2;
  yCen = thisNrow/2;
  zCen = origNumSecs/2;
  
  largestDim = MAX(thisNcol, thisNrow);
  largestDim = MAX(largestDim, origNumSecs);
  
  mapCols = binCol;
  mapRows = binRow;
  mapSecs = binSec;      

  XtUnmanageChild(statusScreenW);
  
} // end function read_param_map



void read_param(char *fName, int readPacked)
{
  // function name: read_param
  // Read the parameter file

  char inStr[512], tFname[512], *s1;
  int i, j, tiltFormat;
  FILE *fIn;
  
  printf("Reading [%s]\n",fName);
  TextSet(paramFileTextW, fName, FALSE); 
  sprintf(inParamFilename,"%s",fName);

  if ( (fIn= fopen(fName,"r")) != NULL ) {
    // Get packed boxes filename
    // What are we passed?
    
    readPackedFile = readPacked;
    fscanf(fIn,"%s",inStr);
    tiltFormat = FALSE;
    if(EQU(inStr,"TILT_FORMAT")) {
      readPackedFile = FALSE;
      tiltFormat = TRUE;
    }
    if(!EQU(inStr,"") && readPackedFile) { // OK to read file
	
      sprintf(tFname,"%s",inStr);
      sprintf(fileName.addr,"%s",tFname);
      sprintf(paramBoxesFilename,"%s",tFname);
      raiseIt(statusScreenW);
      XSync(dsply,0);
      
      sprintf(outStr,"Reading [%s]",fileName.addr);
      if(checkStatus(outStr))
	goto destructor;
      
      // Remember this for the next time
      if(numBoxes > 0)
	lastNumBox = (lastNumBox + numBoxes);
      
      // Don't do this - we need to be able to keep
      // MAPs and such around etc...
      // Init for this new image
      // groundZero();
      
      if(projASU) { // But *do* get rid of any unwanted memory
	for (i=0; i<purdueHeader.nimg; i++)
	  freeNullFloat(&packImg[i]);
	projASU = FALSE;
	purdueHeader.nimg = 0;
      }
      
      // Suppose they've got a previous packed file read
      for(i=0;i<purdueHeader.nimg;i++)
	freeNullFloat(&packImg[i]);
      purdueHeader.nimg = 0;

      TextSetInt(boxInitW, lastNumBox+1, TRUE);
      
      // Remember what our filename was - especially if we had to
      // change it to portable or etc...

      sprintf(image_fileName.addr,"%s",fileName.addr);

      // Let's see if we can't just use our own
      mainRobem_sub_UNIX(&int_image_buffer1, 
			 &fileName, &i, &j);
      
      // Ooops no image get out of here
      if(i==0 || j==0)
	goto destructor;
      
      // Give feedback on main window title bar
      sprintf(outStr,"Robem: [%s][%d][%d] - %d images total.",
	      fileName.addr,
	      purdueHeader.numCols, purdueHeader.numRows,
	      purdueHeader.nimg);
      
      doDialogTitle(outStr, mainWindowW, FALSE);
      
      // Remember how big we are on disk
      file_width = i;
      file_height = j;
      lastPackedNumImgs = purdueHeader.nimg;
      
      // Now clip this and only get 1k x 1k maximum at a time
      i = defDispBufSize;
      j = defDispBufSize;
      
      // Point to where the display data is
      image_image1->data = (char *)image_buffer1;
      
    }

    if(checkStatus("Reading parameters"))
      goto destructor;
    
    // Get second line
    fscanf(fIn,"%s",inStr);
    if(s1=strrchr(inStr,','))
      *s1= '\0';
    TextSet(pixsizeW, inStr, FALSE);
    pixSizeP= strtod(inStr,0);
    fscanf(fIn,"%s",inStr);
    if(s1=strrchr(inStr,','))
      *s1= '\0';
    TextSet(unitsW, inStr, FALSE);
    unitsP= (int)strtod(inStr,0);
    fscanf(fIn,"%s",inStr);
    if(s1=strrchr(inStr,','))
      *s1= '\0';
    TextSet(voltsW, inStr, FALSE);
    volts= strtod(inStr,0);
    fscanf(fIn,"%s",inStr);
    if(s1=strrchr(inStr,','))
      *s1= '\0';
    TextSet(ampfacsW, inStr, FALSE);
    ampFacs= strtod(inStr,0);
    fscanf(fIn,"%s",inStr);
    if(s1=strrchr(inStr,','))
      *s1= '\0';
    TextSet(delfMajW, inStr, FALSE);
    delf_maj= strtod(inStr,0);
    fscanf(fIn,"%s",inStr);
    if(s1=strrchr(inStr,','))
      *s1= '\0';
    TextSet(delfMinW, inStr, FALSE);
    delf_min= strtod(inStr,0);
    fscanf(fIn,"%s",inStr);
    if(s1=strrchr(inStr,','))
      *s1= '\0';
    TextSet(angMajW, inStr, FALSE);
    ang_maj= strtod(inStr,0);
    fscanf(fIn,"%s",inStr);
    if(s1=strrchr(inStr,','))
      *s1= '\0';
    TextSet(csCoefW, inStr, FALSE);
    cs_coeff= strtod(inStr,0);
    
    rewind(fIn);
    if(readPackedFile || tiltFormat) {
      Free_Particle_Info(&info);
      deltaXmin = deltaYmin = deltaMin = 99999999.9;
      deltaXmax = deltaYmax = deltaMax = -99999999.9;
    }
    
    Get_Particle_Info(&info,fIn,purdueHeader.numCols,purdueHeader.numRows);
    find_nearest_neighbors(&info);

    // Why do we have to add missing particles ???
    // add_missing_particles(&info, purdueHeader.nimg,
    // purdueHeader.numCols,purdueHeader.numRows);

    // We've got one now
    haveParamFile = TRUE;
    
    // Start at zero'th(1) image
    if(readPackedFile)
      stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		       packImg, purdueHeader.nimg, 1);

    DrawICOSWindow(NULL, NULL, NULL);
    
  }
  else {
    sprintf(outStr,"Can't open [%s]\n",fName);
    askThem(outStr, XmDIALOG_INFORMATION);
  }

 destructor:;
  XtUnmanageChild(statusScreenW);
  fclose(fIn);
  
} // end function read_param



void unitCell_background()
{
  // function name: unitCell_background
  // Draw the basic background for the unit cell

  char str[256];
  int x, y, xp1, yp1;
  float thisTheta=90.0, thisPhi=0.0, negPhiMax;
  
  phiMax    = find_phimax(&symmetry, &thisTheta);
  negPhiMax = 0.0 - phiMax;
  get_t1t2p1__(&symmetry, &theta1, &theta2, &thisPhi);

  if(symmetry==5) { // Draw a diamond w/ dotted lines to show plotting "space"
    convertXY(&x, &y, 0.0, 0.0);
    convertXY(&xp1, &yp1, negPhiMax, 90.0);
    XDrawLine(dsply, icosa_window1, image_gcDOT, x, y, xp1, yp1);
    convertXY(&x, &y, negPhiMax, 90.0);
    convertXY(&xp1, &yp1, 0.0, 180.0);
    XDrawLine(dsply, icosa_window1, image_gcDOT, x, y, xp1, yp1);
    convertXY(&x, &y, 0.0, 180.0);
    convertXY(&xp1, &yp1, phiMax, 90.0);
    XDrawLine(dsply, icosa_window1, image_gcDOT, x, y, xp1, yp1);
    convertXY(&x, &y, phiMax, 90.0);
    convertXY(&xp1, &yp1, 0.0, 0.0);
    XDrawLine(dsply, icosa_window1, image_gcDOT, x, y, xp1, yp1);
  }
  else { // Draw a triangle w/ dotted lines to show plotting "space"
    convertXY(&x, &y, negPhiMax, theta2);
    convertXY(&xp1, &yp1, 0.0, theta1);
    XDrawLine(dsply, icosa_window1, image_gcDOT, x, y, xp1, yp1);
    convertXY(&x, &y, negPhiMax, theta2);
    convertXY(&xp1, &yp1, phiMax, theta2);
    XDrawLine(dsply, icosa_window1, image_gcDOT, x, y, xp1, yp1);
    convertXY(&x, &y, 0.0, theta1);
    convertXY(&xp1, &yp1, phiMax, theta2);
    XDrawLine(dsply, icosa_window1, image_gcDOT, x, y, xp1, yp1);
  }
  
  // ----- Draw tick marks -----

  y = 490; // y already in dsply space
  x = negPhiMax;
  sprintf(str,"%d",x);
  convertXY(&x, &yp1, negPhiMax, 490.0);
  XDrawLine(dsply,icosa_window1,image_gcGREEN, x, y, x, y+5);
  XDrawString(dsply, icosa_window1, image_gcWHITE, x,  y+15, str, strlen(str));
  
  x = 0;
  sprintf(str, "%d", x);
  convertXY(&x, &yp1, 0.0, 490.0);
  XDrawLine(dsply,icosa_window1,image_gcGREEN, x, y, x, y+5);
  XDrawString(dsply, icosa_window1, image_gcWHITE, x, y+15, str, strlen(str));
  
  x = phiMax;
  sprintf(str, "%d", x);
  convertXY(&x, &yp1, phiMax, 490.0);
  XDrawLine(dsply,icosa_window1,image_gcGREEN, x, y, x, y+5);
  XDrawString(dsply, icosa_window1, image_gcWHITE, x, y+15, str, strlen(str));
  
  x = YAXIS; // x already in dsply space
  y = (int)theta1;
  sprintf(str, "%d", y);
  convertXY(&xp1, &y, (float)YAXIS, theta1);
  XDrawLine(dsply,icosa_window1,image_gcGREEN, x-5, y, x, y);
  XDrawString(dsply, icosa_window1, image_gcWHITE, x-YAXIS, y, str, strlen(str));

  y = theta2;
  sprintf(str, "%d", y);
  convertXY(&xp1, &y, (float)YAXIS, theta2);  
  XDrawLine(dsply,icosa_window1,image_gcGREEN, x-5, y, x, y );
  XDrawString(dsply, icosa_window1, image_gcWHITE, x-YAXIS, y, str, strlen(str));
  
} // end function unitCell_background



void unitCell_work()
{
  // function name: unitCell_work
  // Do what it takes to get the unit cell drawn correctly on the display.

  char outStr[256];
  int thisX, thisY;
  float thisTheta, thisPhi, thisOmega;
  P_INFO *thisInfo;
  
  unitCell_background();
  thisInfo = info.info;

  while(thisInfo != NULL) {
    convertXY(&thisX, &thisY, thisInfo->phi, thisInfo->theta);
    
    if (!thisInfo->toggle)
      DrawX(icosa_window1, image_gcRED, thisX, thisY, 5);
    else if(thisInfo->neighbors > 1)
      DrawX(icosa_window1, image_gcGREEN, thisX, thisY, 5);
    else
      DrawX(icosa_window1, image_gcBLUE, thisX, thisY, 5);

    if(thisInfo->neighbors > 1) { // Feedback how many neighbors
      sprintf(outStr,"%d",thisInfo->neighbors);
      XDrawString(dsply, icosa_window1, image_gcGREEN,
		  thisX-5, thisY+5, outStr, strlen(outStr));
    }

    if(thisInfo == foundPart) {
      // Draw a box around the one we found
      XDrawRectangle(dsply, icosa_window1, image_gcWHITE,
		     thisX-5, thisY-5, 10, 10);
    }

    thisInfo = thisInfo->next;
  }
  
  if(drawIcos && foundPart!=NULL) { // Draw this icosahedron
    thisTheta = foundPart->theta;
    thisPhi   = foundPart->phi;
    if(findNearest && useOmega)
      // Use real omega value - gotten above
      thisOmega = foundPart->omega;
    else 
      thisOmega = 0.0;
    
    DrawIcosahedron(deg2rad(thisTheta), deg2rad(thisPhi), deg2rad(thisOmega),
		    80, 0, SCALE, icosa_window1);  
  }
  
} // end function unitCell_work



void xyPlot_background()
{
  // function name: xyPlot_background
  // Draw the generic background for the xy plot.

  char outStr[256];
  int thisX, thisY, i, j;
  
  // Draw dotted lines to show plotting "space"
  XDrawLine(dsply,icosa_window1,image_gcDOT, 0, 256, 512, 256);
  XDrawLine(dsply,icosa_window1,image_gcDOT, 256, 0, 256, 512);
  
  // Draw X Axis tick marks
  allMax = MAX(deltaMax,(0.0-deltaMin));
  convertDeltaXY(&thisX,&thisY,allMax,0.0);
  sprintf(outStr,"%.1f",allMax);
  XDrawLine(dsply,icosa_window1,image_gcWHITE,
	    thisX, 0, thisX, 512);
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      thisX+2, 256, outStr, strlen(outStr));
  convertDeltaXY(&thisX,&thisY,(0.0-allMax),0.0);
  sprintf(outStr,"%.1f",(0.0-allMax));
  XDrawLine(dsply,icosa_window1,image_gcWHITE,
	    thisX, 0, thisX, 512);
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      thisX-30, 256, outStr, strlen(outStr));

  // Draw Y Axis tick marks
  convertDeltaXY(&thisX,&thisY,0.0,allMax);
  sprintf(outStr,"%.1f",allMax);
  XDrawLine(dsply,icosa_window1,image_gcWHITE,
	    0, thisY, 512, thisY);
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      256, thisY-5, outStr, strlen(outStr));
  convertDeltaXY(&thisX,&thisY,0.0,(0.0-allMax));
  sprintf(outStr,"%.1f",(0.0-allMax));
  XDrawLine(dsply,icosa_window1,image_gcWHITE,
	    0, thisY, 512, thisY);
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      256,thisY+12, outStr, strlen(outStr));

  // Now for Standard Deviation rings
  sprintf(outStr,"STDev rings (X %.2f, Y %.2f)",deltaStdX,deltaStdY);
  XDrawLine(dsply,icosa_window1,image_gcGREY[100],
	    50, 10, 70, 10);
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      80, 15, outStr, strlen(outStr));

  for (j=1; j<7; j++) { // Draw the STD elipses
    int beginEl, endEl, distAI, distBI;
    int drawX1[1024], drawY1[1024];
    float distAsq, distBsq, distA, distB;
    float ang1, angThis, thisDeltaStdX, thisDeltaStdY;
    
    thisDeltaStdX = deltaStdX * j;
    thisDeltaStdY = deltaStdY * j;
    if(thisDeltaStdX < allMax && thisDeltaStdY < allMax) {
      convertDeltaXY(&distAI,&distBI,thisDeltaStdX,thisDeltaStdY);
      distA = fabs(distAI-256.0);
      distB = fabs(distBI-256.0);
      
      distBsq = distB * distB;
      distAsq = distA * distA;
      distAsq = MAX(distAsq,1.0); // Avoid division by zero
      ang1 = 0.0;
      
      beginEl = 512 - (int)distA; // We offset this by 256 pixels
      endEl = 512 + (int)distA;
      beginEl = MAX(0,beginEl);
      endEl = MIN(endEl,1023);
      
      for(i=beginEl; i<=endEl; i++) { // Do all Y
	int thisX, thisY, thisI;
	double tempF, thisR;
	
	thisI = i-256;  // We're offsetting this by 256 pixels
	
	tempF = fabs(distBsq * (1.0 - ((thisI-256)*(thisI-256))/ distAsq));
	tempF = sqrt(tempF); 
	
	if(thisI==(beginEl-256) || (thisI==endEl-256))
	  tempF= 0.0;
	
	if(thisI==256)
	  angThis = atan(tempF/0.0000001);
	else
	  angThis = atan(tempF/(256-thisI));
	
	thisX = 256 - thisI;
	thisY = tempF;
	thisR = sqrt((double)(thisX*thisX + thisY*thisY));
	
	// Now rotate in ang1
	drawX1[i] = 256 + (int)(thisR * cos(angThis-ang1));
	drawY1[i] = 256 + (int)(thisR * sin(angThis-ang1));
      }
	  
      // Now draw them
      for (i=beginEl ;i<endEl; i++) { 
	// i==256 draws a line down the middle
	if(i!=512) {
	  XDrawLine(dsply, icosa_window1, image_gcGREY[100],
		    drawX1[i], drawY1[i], drawX1[i+1], drawY1[i+1]);
	  XDrawLine(dsply, icosa_window1, image_gcGREY[100],
		    (256 + (256-drawX1[i])), (256 - (drawY1[i]-256)),
		    (256 + (256-drawX1[i+1])), (256 - (drawY1[i+1]-256)));  
	  XFlush(dsply);  
	}
      }
      
      XFlush (dsply);
    }
  }

} // end function xyPlot_background



void xyPlot_work()
{
  // function name: xyPlot_work
  // Do what it takes to get the x,y plot drawn correctly.

  int x, y;
  P_INFO *thisInfo;
  
  // Recalculate the STD stats  (We may have changed one of the centers)
  deltaXmin = deltaYmin = deltaMin = 99999999.9;
  deltaXmax = deltaYmax = deltaMax = -99999999.9;
  find_nearest_neighbors(&info);
  
  xyPlot_background();
  thisInfo = info.info;
  
  while(thisInfo != NULL) {
    convertDeltaXY(&x, &y, thisInfo->deltaX, thisInfo->deltaY);
    
    if (!thisInfo->toggle)
      DrawX(icosa_window1, image_gcRED,  x, y, 5);
    else
      DrawX(icosa_window1, image_gcBLUE, x, y, 5);
    
    if(thisInfo==foundPart) // Draw a box around the one we found
      XDrawRectangle(dsply, icosa_window1, image_gcWHITE, x-5, y-5, 10, 10);
    
    thisInfo = thisInfo->next;
  }

} // end function xyPlot_work



void plotCavgPP_work()
{
  // function name: plotCavgPP_work
  // Do whatever you need to get the cavg plot on the display.

  int i, thisX, thisXp1, inum;
  float *thisRint, scaleCircFactor, minCA, maxCA, circAvgSize;
  int scaledY, scaledYp1;
  
  if(imgNum > totCavgPPrint) {
    XDrawString(dsply, icosa_window1, image_gcWHITE,
		100, 250, "No Circular Average computed yet.", 33);
    return;
  }
  
  thisRint = cavgPPrint[imgNum-1];
  
  minCA = maxCA = thisRint[3];
  // 2.0 for interp in img_sym() and 0.5 for radius not diameter 
  // 0.90 for 90% - we're plotting too much ... ???
  
  inum = 2.0 *0.5 *0.90 * sqrt((double)
	 (purdueHeader.numCols * purdueHeader.numCols +
	  purdueHeader.numRows * purdueHeader.numRows
	  ));
  
  for (i=0; i<inum; i++) { // get stats
    minCA = MIN(minCA,thisRint[i]);
    maxCA = MAX(maxCA,thisRint[i]);
  }
  
  scaleCircFactor = 490.0 / (maxCA - minCA);
  circAvgSize = 1;  // Get from interface later
  
  for (i=0; i<inum; i++) {
    scaledY   = 490.0 - (thisRint[i]-minCA)*scaleCircFactor;
    scaledYp1 = 490.0 - (thisRint[i+1]-minCA)*scaleCircFactor;
    thisX     = (i-caOffsetX)*circAvgSize + 22;
    thisXp1   = (i-caOffsetX)*circAvgSize + 23;
    
    XDrawLine(dsply, icosa_window1, image_gcGREEN,
	      thisX, scaledY, thisXp1, scaledYp1);
  }
  
} // end function plotCavgPP_work



void Draw_Histogram()
{
  // function name: Draw_Histogram
  // Draws a histogram

  if (magtoggle)
    Single_Line(MAGT, image_gcGREEN);
  if (cc1toggle)
    Single_Line(PFTCC, image_gcBLUE);
  if (cc2toggle)
    Single_Line(PRJCC, image_gcGREEN);
  if (cc3toggle)
    Single_Line(CMPCC, image_gcRED);

} // end function Draw_Histogram



void Single_Line(int whichParam, GC color)
{
  // function name: Single_Line

  int startx=YAXIS, endx=YAXIS, starty=0, endy=0, i;
  int badCount, lastBadCount=0;
  float x;

  for (x=startbox; x<endbox; x+=step) {
      if(x == startbox) {
	sprintf(outStr,"%.2f",x);
	XDrawString(dsply, icosa_window1, image_gcWHITE,
		    20, 505, outStr, strlen(outStr));
      }
      endy = find_number_in_range(info.info, x, x+step, whichParam,
				  &badCount);
      endy += badCount;
      
      for (i=0; i<pixelstep; i++) { // Draw bars
	if(badCount > 0 || (endy==0 && lastBadCount != 0))
	  XDrawLine(dsply, icosa_window1, image_gcGREY[150],
		    startx+i, 490-starty*yscale,
		    startx+i, 490-endy*yscale);
	else
	  XDrawLine(dsply, icosa_window1, color,
		    startx+i, 490-starty*yscale,
		    startx+i, 490-endy*yscale);
	starty = endy;
      }

      if(endy > 0)
	lastBadCount = badCount;
      startx = endx;
      endx += (int)pixelstep;
      starty = endy;
  }
  
  // Show them the end value
  x=endbox;
  sprintf(outStr,"%.2f",x);
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      455, 505, outStr, strlen(outStr));
  XFlush(dsply);
  
} // end function Single_Line



void histogram_work()
{
  // function name: histogram_work
  // The real work of drawing the histogram happens here.

  float y;
  int yStep;
  
  // Drawing vertical text string one char at a time
  XDrawString(dsply, icosa_window1, image_gcWHITE, 500, 200, "#", 1);
  XDrawString(dsply, icosa_window1, image_gcWHITE, 500, 220, "P", 1);
  XDrawString(dsply, icosa_window1, image_gcWHITE, 500, 230, "a", 1);
  XDrawString(dsply, icosa_window1, image_gcWHITE, 500, 240, "r", 1);
  XDrawString(dsply, icosa_window1, image_gcWHITE, 500, 250, "t", 1);
  XDrawString(dsply, icosa_window1, image_gcWHITE, 500, 260, "i", 1);
  XDrawString(dsply, icosa_window1, image_gcWHITE, 500, 270, "c", 1);
  XDrawString(dsply, icosa_window1, image_gcWHITE, 500, 280, "l", 1);
  XDrawString(dsply, icosa_window1, image_gcWHITE, 500, 290, "e", 1);
  XDrawString(dsply, icosa_window1, image_gcWHITE, 500, 300, "s", 1);

  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      250, 500, "range", 5);
  sprintf(outStr,"%1.0f",snumb);
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      YAXIS-10, 500, outStr, strlen(outStr));
  sprintf(outStr,"%1.0f",480.0/(float)yscale);
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      0, 22, outStr, strlen(outStr));
  
  yStep = (480.0/yscale) / 10.0;
  yStep = MAX(yStep, 1);

  // Now draw grid
  for (y=480.0/yscale; y>0.0; y-=(float)yStep) {
    int stX=24, eX=480, stY=YAXIS, eY=YAXIS;

    eY  = stY = 512 - YAXIS + (int)(480.0*y*yscale/480.0);

    XDrawLine(dsply, icosa_window1, image_gcGREY[100],
	      stX, stY, eX, eY);

    if(y < 480.0/yscale) { // Don't do the first
      sprintf(outStr,"%.0f",y);
      XDrawString(dsply, icosa_window1, image_gcWHITE,
		  1, eY, outStr, strlen(outStr));
    }
  }

} // end function histogram_work



void DrawICOSWindow(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: DrawICOSWindow
  // Redraw window e.g. on expose events or whenever it's prudent.

  int i;
  
  XPutImage(dsply, icosa_window1, image_gc, icosa_image1, 0, 0, 0, 0, 512, 512);
  XFlush (dsply);

  if(icoDispMode!=XYPLOT) { // Draw axes
    XDrawLine(dsply, icosa_window1, image_gcWHITE, YAXIS, YAXIS-10, YAXIS, 490);
    XDrawLine(dsply, icosa_window1, image_gcWHITE, YAXIS, 490, 490, 490);
    
    for (i=0; i<512; i++)
      if(i < ico_toolsScrollValue*4.0)
	XDrawLine(dsply, image_window1, image_gcRED, i+30, 70, i+30, 442);

  }
  
  if(projASU) {
    XtUnmanageChild(ppMinMaxFormW);
    projASUunitCell_work(); 
  }
  else if(icoDispMode == XYPLOT) {
    XtManageChild(ppMinMaxFormW);
    xyPlot_work(); 
  }
  else if(icoDispMode == PLOTCAVGPP) {
    XtUnmanageChild(ppMinMaxFormW);
    plotCavgPP_work(); 
  }
  else if(icoDispMode == UNITCELL) {
    XtUnmanageChild(ppMinMaxFormW);
    unitCell_work(); 
  }
  else if(icoDispMode == HISTOGRAM) {
    XtUnmanageChild(ppMinMaxFormW);
    XtUnmanageChild(incDecW);
    XtUnmanageChild(idtW);
    XtUnmanageChild(thetatW);
    XtUnmanageChild(phitW);
    XtUnmanageChild(omegatW);
    XtUnmanageChild(xtW);
    XtUnmanageChild(ytW);
    XtUnmanageChild(deltaXTW);
    XtUnmanageChild(deltaYTW);
    XtUnmanageChild(absDeltaXTW);
    XtUnmanageChild(absDeltaYTW);
    XtUnmanageChild(distanceTW);
    XtManageChild(magtW);
    XtManageChild(pftccW);
    XtManageChild(prjccW);
    XtManageChild(cmpccW);
    XtManageChild(histControlsW);
    histogram_work();
    Draw_Histogram();
  }  
  else if(icoDispMode == ICOSORT) {
    XtUnmanageChild(ppMinMaxFormW);
    XtManageChild(incDecW);
    XtManageChild(idtW);
    XtManageChild(thetatW);
    XtManageChild(phitW);
    XtManageChild(omegatW);
    XtManageChild(xtW);
    XtManageChild(ytW);
    XtManageChild(magtW);
    XtManageChild(pftccW);
    XtManageChild(prjccW);
    XtManageChild(cmpccW);
    XtManageChild(deltaXTW);
    XtManageChild(deltaYTW);
    XtManageChild(absDeltaXTW);
    XtManageChild(absDeltaYTW);
    XtManageChild(distanceTW);
    XtUnmanageChild(histControlsW);
    if(icoSortLevel == 0)
      sprintf(outStr, "Choose Sort key.");
    else if(icoSortLevel == 1)
      sprintf(outStr, "Sorting ...");
    else if(icoSortLevel == 2)
      sprintf(outStr, "Sort Finished.");
    else
      sprintf(outStr,"Sort Mode");
    XDrawString(dsply, icosa_window1, image_gcWHITE,
		200, 256, outStr, strlen(outStr));
  }
  
} // end function DrawICOSWindow



void unitCell_activate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: unitCell_activate
  // They want the unit cell displayed.
  
  if(!findNearest)
    icosaMotioning = TRUE;
  XtUnmanageChild(parmTogsW);
  XtUnmanageChild(projasuW);
  projASU = FALSE;
  XtManageChild(parmheadW);
  icoDispMode = UNITCELL;
  DrawICOSWindow(NULL, NULL, NULL); 
  
} // end function unitCell_activate



void ico_hist_activate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: ico_hist_activate
  // They want the histogram displayed

  icosaMotioning = FALSE;

  doLabel("Choose parameters for histogram: ", parmTogsLabelW, DEFAULT);

  XtManageChild(parmTogsW);
  XtUnmanageChild(parmheadW);
  projASU = FALSE;
  XtUnmanageChild(projasuW);
  TextSet(partPointTextW, "Choose Histogram parameters", FALSE);
  icoDispMode = HISTOGRAM;
  DrawICOSWindow(NULL, NULL, NULL);

} // end function ico_hist_activate



void xyPlot_activate(Widget widget, XtPointer client_data, 
		     XtPointer call_data)
{
  // function name: xyPlot_activate
  // They have chosen to have the x,y plot displayed.

  float thisDistanceMax;

  if(!findNearest)
    icosaMotioning = TRUE;
  XtUnmanageChild(parmTogsW);
  XtUnmanageChild(projasuW);
  projASU = FALSE;
  XtManageChild(parmheadW);
  
  // Let's do this by 1000's so we can get the granularity we
  // need for small absolute max distances
  thisDistanceMax = distanceMax * 1000.0;
  minXYplot = 0.0;
  maxXYplot = distanceMax;
  TextSetFloat(ppMinTextW, minXYplot, FALSE);
  TextSetFloat(ppMaxTextW, maxXYplot, FALSE);
  XtVaSetValues(ppMinScrollW,XmNmaximum, (int)thisDistanceMax,NULL);
  XtVaSetValues(ppMinScrollW,XmNminimum, 0,NULL);
  XtVaSetValues(ppMinScrollW,XmNsliderSize, (int)(thisDistanceMax/20.0),NULL);
  XtVaSetValues(ppMinScrollW,XmNpageIncrement, 
		(int)(thisDistanceMax/20.0),NULL);
  XtVaSetValues(ppMinScrollW,XmNincrement, (int)(thisDistanceMax/500.0),NULL);
  XtVaSetValues(ppMinScrollW,XmNvalue, 0,NULL);
		
  XtVaSetValues(ppMaxScrollW,XmNmaximum, (int)(thisDistanceMax*1.11),NULL);
  XtVaSetValues(ppMaxScrollW,XmNminimum, 0,NULL);
  XtVaSetValues(ppMaxScrollW,XmNsliderSize, (int)(thisDistanceMax/20.0),NULL);
  XtVaSetValues(ppMaxScrollW,XmNpageIncrement, 
		(int)(thisDistanceMax/20.0),NULL);
  XtVaSetValues(ppMaxScrollW,XmNincrement, (int)(thisDistanceMax/500.0),NULL);
  XtVaSetValues(ppMaxScrollW,XmNvalue, (int)(thisDistanceMax+1),NULL);

  icoDispMode = XYPLOT;
  DrawICOSWindow(NULL, NULL, NULL); 

} // end function xyPlot_activate



void plotCavgPP_activate(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: plotCavgPP_activate
  // Setup for plotting the cavg.

  if(!findNearest)
    icosaMotioning = FALSE;

  icoDispMode = PLOTCAVGPP;
  DrawICOSWindow(NULL, NULL, NULL); 

} // end function plotCavgPP_activate



void DrawIcosahedron(float yrot, float xrot, float zrot, 
		     int xoffset, int yoffset, 
		     float scaleFactor, Window thisWin)  
{
  // function name: DrawIcosahedron 
  // Actually do the icosahedron drawing.

  int i, sx, sy, ex, ey;
  GC thisGC;

  // Rotate omega 90 degrees so it matches our convention
  if(thisWin==icosa_window1)
    zrot += deg2rad(90.0);
  
  if(xoffset == 80)
    thisGC = image_gcRED;
  else if(thisWin==XtWindow(mapShadeScreenW)) 
    thisGC = image_gcRED;
  else 
    thisGC = image_gcBLUE;

  // Blank out last icosahedron
  if(thisWin == icosa_window1)
    XPutImage(dsply, icosa_window1, icosa_gc, icosa_image1, 
	      300+xoffset, 0, 300+xoffset, 0, 110, 110); 
  
  if(thisWin==XtWindow(mapShadeScreenW)) { // Init the display
    sx = xoffset - 35;
    ex = xoffset + 35;
    for (i=0; i<70; i++) {
      sy = yoffset - 35 +i;
      ey = sy;
      XDrawLine(dsply,thisWin,image_gcGREY[100], sx, sy, ex, ey);
    }
  }

  // Stuff the icosahedron values
  init_vertices();
  
  for (i=0; i<icosabox.num_points; i++) {
    roty(&icosabox.vertex[i].lx, &icosabox.vertex[i].lz, yrot);
    rotx(&icosabox.vertex[i].ly, &icosabox.vertex[i].lz, xrot);
    rotz(&icosabox.vertex[i].lx, &icosabox.vertex[i].ly, zrot);  
  }
  
  if(thisWin==XtWindow(mapShadeScreenW))
    scale_verticies(120.0);
  else 
    scale_verticies(scaleFactor*icosaScaleF);

  Convert3d_to_scrcords();

  if(thisWin==icosa_window1)
    translate_verticies(TRANSLATEX+xoffset, TRANSLATEY+yoffset);
  else if(thisWin==XtWindow(mapShadeScreenW))
    translate_verticies(xoffset, yoffset);
  else 
    translate_verticies((largestDim/2)*sizeFactor,
			(largestDim/2)*sizeFactor);
  
  // We only want to draw the front half in main window
  for (i=0; i<icosabox.num_faces; i++) {
    if(thisWin == icosa_window1                  ||
       thisWin == XtWindow(mapShadeScreenW) ||
       icosabox.vertex[icosabox.faces[i].x].lz >= 0.0) {
      
      sx = icosabox.vertex[icosabox.faces[i].x].sx;
      sy = icosabox.vertex[icosabox.faces[i].x].sy;
      ex = icosabox.vertex[icosabox.faces[i].y].sx;
      ey = icosabox.vertex[icosabox.faces[i].y].sy;
      XDrawLine(dsply, thisWin, thisGC, sx, sy, ex, ey);
      sx = icosabox.vertex[icosabox.faces[i].y].sx;
      sy = icosabox.vertex[icosabox.faces[i].y].sy;
      ex = icosabox.vertex[icosabox.faces[i].z].sx;
      ey = icosabox.vertex[icosabox.faces[i].z].sy;
      XDrawLine(dsply, thisWin, thisGC, sx, sy, ex, ey);
      sx = icosabox.vertex[icosabox.faces[i].z].sx;
      sy = icosabox.vertex[icosabox.faces[i].z].sy;
      ex = icosabox.vertex[icosabox.faces[i].x].sx;
      ey = icosabox.vertex[icosabox.faces[i].x].sy;
      XDrawLine(dsply,thisWin,thisGC, sx, sy, ex, ey);
    }
  }
  
  if(thisWin==icosa_window1 && xoffset==0) { // Draw theta phi coords
    XDrawString(dsply, XtWindow(icosaW), image_gcBLUE,
		310, 100, outStr, strlen(outStr));
    XFlush(dsply);
  }

} // end function DrawIcosahedron



void convertThetaPhi(int x, int y, 
		     float *thisTheta, float *thisPhi)
{
  // function name: convertThetaPhi
  // Convert display coords into theta and phi

  float xFactor, yFactor, tempTheta;

  yFactor = 21.0 * ((90.0 - 69.0)/(theta2 - theta1));
  *thisTheta = (y - 20.0)/yFactor + theta1;
  
  if(symmetry == 5) {
    if(*thisTheta > 90.0)
      tempTheta = 90.0 - ((*thisTheta)-90.0);
    else 
      tempTheta = *thisTheta;
    
    xFactor = (7.0*sin(deg2rad(tempTheta))) * (32.0 / phiMax);
  }
  else
    xFactor = 7.0 * (32.0 / phiMax);
  
  if(xFactor==0.0)
    xFactor = 0.000001;
  *thisPhi = (x - 256.0)/xFactor;
  
} // end function convertThetaPhi



void find_nearest_neighbors(INFO *head)
{
  // function name: find_nearest_neighbors
  // Find how many particles are near this one

  int x1, y1, x2, y2, distnce, tot=0;
  double sumX=0.0, sumY=0.0, sumsqX=0.0, sumsqY=0.0;
  double varX, avgX, varY, avgY;
  P_INFO *thisInfo, *thatInfo;

  deltaStdY = deltaStdX = 0.0; 

  thisInfo = head->info;
  while(thisInfo != NULL) {
    convertXY(&x1, &y1, thisInfo->phi, thisInfo->theta);
    thatInfo = head->info;
    thisInfo->neighbors = 0;

    while(thatInfo != NULL) { // Find neighbors
      convertXY(&x2, &y2, thatInfo->phi, thatInfo->theta);
      distnce = (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2);
      if(distnce <= 1.0) // Found a new candidate within 1 pixel
	thisInfo->neighbors++;
      thatInfo = thatInfo->next;
    }
    
    // While we're here let's calc the deltaXY stats
    sumX   +=  thisInfo->deltaX;
    sumsqX +=  thisInfo->deltaX * thisInfo->deltaX;
    sumY   +=  thisInfo->deltaY;
    sumsqY +=  thisInfo->deltaY * thisInfo->deltaY;
    tot++;
    
    // Recalc these just in case we've changed one
    deltaMax  = MAX(thisInfo->deltaX, deltaMax);
    deltaMax  = MAX(thisInfo->deltaY, deltaMax);
    deltaMin  = MIN(thisInfo->deltaX, deltaMin);
    deltaMin  = MIN(thisInfo->deltaY, deltaMin);
    deltaXmin = MIN(thisInfo->deltaX, deltaXmin);
    deltaYmin = MIN(thisInfo->deltaY, deltaYmin);
    deltaXmax = MAX(thisInfo->deltaX, deltaXmax);
    deltaYmax = MAX(thisInfo->deltaY, deltaYmax);
    thisInfo->distance = sqrt(thisInfo->deltaX*thisInfo->deltaX +
			      thisInfo->deltaY*thisInfo->deltaY);
    thisInfo = thisInfo->next;
  }
  
  avgY = sumY / tot;
  varY = sumsqY/tot - avgY*avgY;
  deltaStdY = sqrt(fabs(varY));
  
  avgX = sumX / tot;
  varX = sumsqX/tot - avgX*avgX;
  deltaStdX = sqrt(fabs(varX));

} // end function find_nearest_neighbors



void findNearestParticle(int *x, int *y,
			 P_INFO **foundPart,
			 int toggleFound, int makeThisParticle,
			 int showPackedBox, int buttonNum)
{
  // function name: findNearestParticle
  // What's the nearest particle to this one. Used to find out which
  // particle they want detail printed.

  int i, thisX, thisY, minX, minY, numFound=0;
  int distnce, minDist=INT_MAX, minParticle=-99;
  static int lastParticle = -99;
  float minOmega;
  P_INFO *thisInfo, *goodOnes[20];

  if(!readPackedFile)
    showPackedBox = FALSE;

  thisInfo = info.info;
  if(thisInfo == NULL) return;

  while(thisInfo != NULL) {
    if(icoDispMode==UNITCELL)
      convertXY(&thisX, &thisY, thisInfo->phi, thisInfo->theta);
    else // XYPLOT
      convertDeltaXY(&thisX, &thisY, thisInfo->deltaX, thisInfo->deltaY);

    distnce = (thisX-*x)*(thisX-*x) + (thisY-*y)*(thisY-*y);

    // Searching for new candidates
    if(makeThisParticle > 0 && thisInfo->rank == makeThisParticle) {
      minDist = distnce;
      minParticle = thisInfo->rank;
      minX = thisX;
      minY = thisY;
      minOmega = thisInfo->omega;
      *foundPart = thisInfo;
    }
    else if(distnce < minDist) {
      minDist = distnce;
      minParticle = thisInfo->rank;
      minX = thisX;
      minY = thisY;
      minOmega = thisInfo->omega;
      *foundPart = thisInfo;
    }
    
    thisInfo = thisInfo->next;
  }

  if(makeThisParticle <= 0) { // Now try and find duplicates
    thisInfo = info.info;
    while(thisInfo != NULL) {
      convertXY(&thisX, &thisY, thisInfo->phi, thisInfo->theta);
      
      distnce = (thisX-minX)*(thisX-minX) + (thisY-minY)*(thisY-minY);
      if(distnce <= 1.0) // Found a new candidate within 1 pixel
	goodOnes[numFound++] = thisInfo;
      
      thisInfo = thisInfo->next;
    }
  }
  
  if(buttonNum == Button3) { // Just give them a report on what's what
    char *s1,*s2;
    s1 = outStr;
    sprintf(s1,"Legend - \n\n\
     U  -  Particle thrown out in UnitCell display\n\
     X  -  Particle thrown out in XY Plot  display\n\
     T  -  Particle thrown out from thresholded XY Plot\n\
     M  -  Particle thrown out in Histogram display for bad Mag\n\
     F  -  Particle thrown out in Histogram display for bad PftCC\n\
     R  -  Particle thrown out in Histogram display for bad PrjCC\n\
     C  -  Particle thrown out in Histogram display for bad CmpCC\n\
     I  -  Particle thrown out in Input File\n\n");
    s2 = outStr + strlen(s1);
    s1 = s2;
    
    if(numFound > 1) { // Do multiples
      for(i=0; i<numFound; i++) { // Do all
	sprintf
	  (s1,
	   "Part[%d] - Theta[%.3f]Phi[%.3f]Omega[%.3f]  [%c%c%c%c%c%c%c%c]\n",
	   goodOnes[i]->particle_number,
	   goodOnes[i]->theta,
	   goodOnes[i]->phi,
	   goodOnes[i]->omega,
	   (goodOnes[i]->toggle & UCBAD)?'U':' ',
	   (goodOnes[i]->toggle & XYBAD)?'X':' ',
	   (goodOnes[i]->toggle & XYBADTHRESH)?'T':' ',
	   (goodOnes[i]->toggle & HISTMAGBAD)?'M':' ',
	   (goodOnes[i]->toggle & HISTPFTBAD)?'F':' ',
	   (goodOnes[i]->toggle & HISTPRJBAD)?'R':' ',
	   (goodOnes[i]->toggle & HISTCMPBAD)?'C':' ',
	   (goodOnes[i]->toggle & INFILEBAD)?'I':' ');
	s2 = outStr + strlen(outStr);
	s1 = s2;
      }
    }
    else { // Do single
      if(foundPart) {
	sprintf
	  (s1,
	   "Part[%d] - Theta[%.3f]Phi[%.3f]Omega[%.3f]  [%c%c%c%c%c%c%c%c]\n",
	   (*foundPart)->particle_number,
	   (*foundPart)->theta,
	   (*foundPart)->phi,
	   (*foundPart)->omega,
	   ((*foundPart)->toggle & UCBAD)?'U':' ',
	   ((*foundPart)->toggle & XYBAD)?'X':' ',
	   ((*foundPart)->toggle & XYBADTHRESH)?'T':' ',
	   ((*foundPart)->toggle & HISTMAGBAD)?'M':' ',
	   ((*foundPart)->toggle & HISTPFTBAD)?'F':' ',
	   ((*foundPart)->toggle & HISTPRJBAD)?'R':' ',
	   ((*foundPart)->toggle & HISTCMPBAD)?'C':' ',
	   ((*foundPart)->toggle & INFILEBAD)?'I':' '
	   );
      }
      else 
	sprintf(outStr,"No particle found.");
    }
      
    TextSet(helpTextW, outStr, FALSE);
    raiseIt(helpScreenW);
    
  }
  else { // normal
    if(numFound > 1) { // We've got to make them choose one
      int numWanted, saveIcosaMotioning;
      
      sprintf(outStr,"Multiple particles found.  Choose one:\n");
      TextSet(infoPromptTextW, outStr, FALSE); 
      
      for (i=0; i<numFound; i++) {
	sprintf(outStr,
		"Part[%d] - Theta[%.3f]Phi[%.3f]Omega[%.3f]  [%s]",
		goodOnes[i]->particle_number,
		goodOnes[i]->theta,
		goodOnes[i]->phi,
		goodOnes[i]->omega,
		(goodOnes[i]->toggle)?"OFF":"ON");
	XmListAddItem(infoListW, XmStringCreateLocalized(outStr), i+1);
	XFlush(dsply);
      }
      
      saveIcosaMotioning = icosaMotioning;
      icosaMotioning = FALSE;
      sprintf(outStr,"Multiple particles found.  Choose one:");
      numWanted= askThemAnswer(outStr,numFound);
      // Do we really want to turn this back on just yet???
      icosaMotioning = saveIcosaMotioning;
      *foundPart = goodOnes[numWanted];
      convertXY(&thisX, &thisY, (*foundPart)->phi, (*foundPart)->theta);
      minX = thisX;
      minY = thisY;
      minParticle = (*foundPart)->rank;
      // Let's toggle it here before we display it
    }
    
    // OK we know which one they want now so continue
    *x = minX;
    *y = minY;
    if(toggleFound) { // Toggle it back off 
      if(icoDispMode==UNITCELL) { // Unit Cell display
	if((*foundPart)->toggle & UCBAD)
	  (*foundPart)->toggle &= NOTUCBAD;
	else
	  (*foundPart)->toggle |= UCBAD;
      }
      else if(icoDispMode==XYPLOT) { // XY plot
	if((*foundPart)->toggle & XYBAD)
	  (*foundPart)->toggle &= NOTXYBAD;
	else 
	  (*foundPart)->toggle |= XYBAD;
      }
    }
    
    if(showPackedBox) { // Show them this one
      imgNum = minParticle;
      TextSetInt(imgNumTextW, imgNum, FALSE);
      stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		       packImg, purdueHeader.nimg, imgNum);
      DrawWindow(); 
    }
    
    lastParticle = minParticle;
    
    // First draw what we've got
    DrawICOSWindow(NULL, NULL, NULL);
    
    sprintf(outStr,
	    "Particle[%9d]  Theta [%9.3f]      Phi[%9.3f]  Omega[%9.3f]\n\
       X[%9.3f]       Y[%9.3f]      Mag[%9.3f]\n\
  pft_cc[%9.3f]  prj_cc[%9.3f]   cmp_cc[%9.3f]\n\
  deltaX[%9.3f]  deltaY[%9.3f] distance[%9.3f]\n\
   state[%3s]         rank[%d]\n",
	    (*foundPart)->particle_number, 
	    (*foundPart)->theta,
	    (*foundPart)->phi,
	    (*foundPart)->omega,
	    (*foundPart)->x,
	    (*foundPart)->y,
	    (*foundPart)->mag,
	    (*foundPart)->cc1,
	    (*foundPart)->cc2,
	    (*foundPart)->cc3,
	    (*foundPart)->deltaX,
	    (*foundPart)->deltaY,
	    (*foundPart)->distance,
	    ((*foundPart)->toggle)?" OFF":"ON",
	    (*foundPart)->rank);
    
    TextSet(partPointTextW, outStr, FALSE);
    TextSetInt(ppinTextW, imgNum, TRUE);
  }
  
} // end function findNearestParticle



void findHistParticle(int *x, int buttonNum)
{
  // function name: findHistParticle
  // Mark this particle as bad. May be used in future when writing a
  // param file from this interface.

  int i, binVal;
  float thisVal;
  P_INFO *thisInfo;

  // First let's see what the cutoff is
  thisVal = *x - YAXIS; // Offset pixels in X direction
  thisVal = *x;
  thisVal = startbox + thisVal/480.0;
  
  // Now find the lowest value for this band of the histogram
  binVal = (max-min)/(float)step;
  for (i=0; i<(binVal-1); i++) { // Look for range
    float tmpVal1, tmpVal2;
    tmpVal1 = (endbox-startbox)*step*i ;
    tmpVal2 = (endbox-startbox)*step*(i+1);
    if(tmpVal1 <= thisVal && thisVal < tmpVal2)
      thisVal = tmpVal1;
  }

  thisInfo = info.info;

  while(thisInfo != NULL) {    
    if(buttonNum == Button1) { // Take out all to the left
      if(magtoggle && thisInfo->mag < thisVal)
	thisInfo->toggle |= HISTMAGBAD;
      if(cc1toggle && thisInfo->cc1 < thisVal)
	thisInfo->toggle |= HISTPFTBAD;
      if(cc2toggle && thisInfo->cc2 < thisVal)
	thisInfo->toggle |= HISTPRJBAD;
      if(cc3toggle && thisInfo->cc3 < thisVal)
	thisInfo->toggle |= HISTCMPBAD;
    }
    else if(buttonNum == Button2) { // Take out all to the right
      if(magtoggle && thisInfo->mag > thisVal)
	thisInfo->toggle |= HISTMAGBAD;
      if(cc1toggle && thisInfo->cc1 > thisVal)
	thisInfo->toggle |= HISTPFTBAD;
      if(cc2toggle && thisInfo->cc2 > thisVal)
	thisInfo->toggle |= HISTPRJBAD;
      if(cc3toggle && thisInfo->cc3 > thisVal)
	thisInfo->toggle |= HISTCMPBAD;
    }
    else if(buttonNum == Button3) { // Restore them
      if(magtoggle && thisInfo->mag < thisVal)
	thisInfo->toggle &= NOTHISTMAGBAD;
      if(cc1toggle && thisInfo->cc1 < thisVal)
	thisInfo->toggle &= NOTHISTPFTBAD;
      if(cc2toggle && thisInfo->cc2 < thisVal)
	thisInfo->toggle &= NOTHISTPRJBAD;
      if(cc3toggle && thisInfo->cc3 < thisVal)
	thisInfo->toggle &= NOTHISTCMPBAD;
    }
    thisInfo = thisInfo->next;
  }
  
  // First draw what we've got
  DrawICOSWindow(NULL, NULL, NULL);

  XFlush(dsply);

} // end function findHistParticle



void icosaMotion(Widget widget, caddr_t unused, 
		 register XMotionEvent *pEvent)
{
  // function name: icosaMotion
  // Whenever the pointer moves across our window call here. This is
  // where we redraw the circles if they're moving a box or print the
  // coords of the pointer as they move.

  float thisTheta, thisPhi;
  
  if(pEvent->type == MotionNotify && icoDispMode==UNITCELL && drawIcos) {
    convertThetaPhi(pEvent->x, pEvent->y, &thisTheta, &thisPhi);
    sprintf(outStr,"%.1f,%.1f", thisTheta, thisPhi);
    DrawIcosahedron(deg2rad(thisTheta), deg2rad(thisPhi), 0.0,
		    0, 0, SCALE, icosa_window1); 
  } 
  
  XFlush(dsply);   // Make changes current
  
} // end function icosaMotion



void icosaClick(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: icosaClick
  // Every time the mouse button is clicked in the partproc 
  // icosa_area window call here.  
  
  char ans;
  int x, y;
  XEvent event;
  
  XmDrawingAreaCallbackStruct *draw_data = 
    (XmDrawingAreaCallbackStruct *) call_data;
  
  if(draw_data == 0)
    return;
  
  event= *(draw_data->event);
  
  switch (event.type) {
    
  case ButtonRelease:
    x = event.xbutton.x;
    y = event.xbutton.y;
    break;
    
  case ButtonPress:
    x = event.xbutton.x;
    y = event.xbutton.y;
    
    if(icoDispMode == HISTOGRAM)
      findHistParticle(&x, event.xbutton.button);
    else if(event.xbutton.button == Button1 && findNearest)
      findNearestParticle(&x, &y, &foundPart, FALSE, -99, TRUE, Button1);
    else if(event.xbutton.button == Button2 && findNearest)
      findNearestParticle(&x, &y, &foundPart, TRUE,  -99, TRUE, Button2);
    else if(event.xbutton.button == Button3 && findNearest)
      findNearestParticle(&x, &y, &foundPart, FALSE, -99, TRUE, Button3);
    break;
    
  case KeyPress:
    ans = XLookupKeysym((XKeyEvent *)&event, 0);
    printf ("Keyed[%c]\n",ans);
    break;
    
  default:
    break;
  }
  
  XFlush(dsply);     // Make changes current
  
} // end function icosaClick



void binPPArrowButton(Widget widget, XEvent *event)
{
  // function name: binPPArrowButton
  // They want to change the current displayed box by using the arrows.

  int binVal, *passTag, speedFactor;
  float stepVal;
  XmAnyCallbackStruct callback_data;

  passTag = (int *)malloc(sizeof(int));

  binVal      = wstrtod(binTextW);
  speedFactor = GetButtonVal(event->xbutton.button, 1, 2, 5);
  
  if(widget==binPPUpW)
    binVal += speedFactor;
  else 
    binVal -= speedFactor;

  binVal = MAX(binVal,1);

  TextSetInt(binTextW, binVal, FALSE);
  
  // Now set the stepVal accordingly
  stepVal = (float)(max-min)/binVal;
  stepVal = MAX(stepVal, 0.0000001);
  sprintf(outStr, "%.3f", stepVal);
  TextSet(histStepTextW, outStr, FALSE);
  
  // Now make those changes current
  *passTag = BINTEXT;
  paramMode_change(widget, (XtPointer)passTag, &callback_data);
  free(passTag);

} // end function binPPArrowButton



void binPPtext_changed(Widget widget, XEvent *event)
{
  // function name: binPPtext_changed
  // They hit the carriage return key in the binTextW so make those
  // changes current.

  int *passTag, binVal;
  float stepVal;
  XmAnyCallbackStruct callback_data;

  passTag = (int *)malloc(sizeof(int));
  binVal = MAX(wstrtod(binTextW), 1);
  TextSetInt(binTextW, binVal, FALSE);
  
  // Now set the stepVal accordingly
  stepVal = (float)(max-min)/(float)binVal;
  stepVal = MAX(stepVal,0.0000001);
  sprintf(outStr,"%.3f",stepVal);
  TextSet(histStepTextW, outStr, FALSE);
  
  // Now make those changes current
  *passTag = BINTEXT;
  paramMode_change(widget, (XtPointer)passTag, &callback_data);
  free(passTag);

} // end function binPPtext_changed



void stepPPArrowButton(Widget widget, XEvent *event)
{
  // function name: stepPPArrowButton
  // Change the current displayed box by using the arrows.

  int binVal, *passTag;
  float speedFactor, stepVal;
  XmAnyCallbackStruct callback_data;

  passTag = (int *)malloc(sizeof(int));
  stepVal = wstrtod(histStepTextW);

  speedFactor = (float) GetButtonVal(event->xbutton.button, 1, 2, 5);
  
  if(widget == stepPPUpW)
    stepVal += speedFactor;
  else 
    stepVal -= speedFactor;
  
  stepVal= MAX(stepVal,0.0000001);
  
  sprintf(outStr,"%.3f",stepVal);
  TextSet(histStepTextW, outStr, FALSE);

  // Set the binVal accordingly
  binVal = MAX((max-min)/stepVal, 1);
  TextSetInt(binTextW, binVal, FALSE);
  
  // Now make those changes current
  *passTag = STEPTEXT;
  paramMode_change(widget, (XtPointer)passTag, &callback_data);
  free(passTag);

} // end function stepPPArrowButton



void stepPPtext_changed(Widget widget, XEvent *event)
{
  // function name: stepPPtext_changed
  // They hit the carriage return key in the histStepTextW so make those
  // changes current.

  int binVal, *passTag;
  float stepVal;
  XmAnyCallbackStruct callback_data;

  passTag = (int *)malloc(sizeof(int));
  stepVal = MAX(wstrtod(histStepTextW), 0.0000001);

  // Set the binVal accordingly
  binVal = MAX((max-min)/stepVal, 1);
  TextSetInt(binTextW, binVal, FALSE);
  
  // Make those changes current
  *passTag = STEPTEXT;
  paramMode_change(widget, (XtPointer)passTag, &callback_data);
  free(passTag);

} // end function stepPPtext_changed



void ppintext_changed(Widget widget, XEvent *event)
{
  // function name: ppintext_changed
  // They hit the carriage return key in the ppinTextW so make those
  // changes current.

  imgNum = wstrtod(ppinTextW);

  if(fileType == PACKEDPORTABLE || fileType == MAP)
    imgNum = MIN(imgNum, purdueHeader.nimg);
  imgNum = MAX(imgNum, 1);
  if(shadedMaps && !dispSections)
    imgNum = MIN(imgNum, mapNumSecs);
  
  doTheImgNum(imgnumArrowUpW, 0);
  
} // end function ppintext_changed



void ppinArrowButton(Widget widget, XEvent *event)
{
  // function name: ppinArrowButton 
  // Change the current displayed box by using the arrows.
  
  if(!locked) {      
    int speedFactor;
    speedFactor = GetButtonVal(event->xbutton.button, 1, 2, 5);

    if(widget == ppinUpW)
      doTheImgNum(imgnumArrowUpW, speedFactor);
    else 
      doTheImgNum(imgnumArrowDownW, speedFactor);
  }
  
} // end function ppinArrowButton



void global_minmax() 
{
  // function name: global_minmax
  // Find the min/max for this dataset.

  // We're only doing mag, pftcc, prjcc, cmpcc now so these will do
  min = 1.0;
  max = -1.0;

  if (thetatoggle) {
     min = MIN(thetamin, min);
     max = MAX(thetamax, max);
  }
  if (phitoggle) {
    min = MIN(phimin, min);
    max = MAX(phimax, max);
  }
  if (omegatoggle) {
    min = MIN(omegamin, min);
    max = MAX(omegamax, max);
  }
  if (xtoggle) {
    min = MIN(xmin, min);
    max = MAX(xmax, max);
  }
  if (ytoggle) {
    min = MIN(ymin, min);
    max = MAX(ymax, max);
  }
  if (deltaXtoggle) {
    min = MIN(deltaXmin, min);
    max = MAX(deltaXmax, max);
  }
  if (deltaYtoggle) {
    min = MIN(deltaYmin, min);
    max = MAX(deltaYmax, max);
  }
  if (absDeltaXtoggle) {
    min = MIN(fabs(deltaXmin), min);
    max = MAX(fabs(deltaXmax), max);
  }
  if (absDeltaYtoggle) {
    min = MIN(fabs(deltaYmin), min);
    max = MAX(fabs(deltaYmax), max);
  }
  if (distancetoggle) {
    min = MIN(distanceMin, min);
    max = MAX(distanceMax, max);
  }
  if (magtoggle) {
    min = MIN(magmin, min);
    max = MAX(magmax, max);
  } 
  if (cc1toggle) {
    min = MIN(cc1min, min);
    max = MAX(cc1max, max);
    // Try RWa July 30, 1998
    min = 0.0;
    max = 1.0;
  }
  if (cc2toggle) {
    min = MIN(cc2min, min);
    max = MAX(cc2max, max);
    // Try RWa July 30, 1998
    min = 0.0;
    max = 1.0;
  }
  if (cc3toggle) {
    min = MIN(cc3min, min);
    max = MAX(cc3max, max);
    // Try RWa July 30, 1998
    min = 0.0;
    max = 1.0;
  }

} // end function global_minmax



void find_min_yscale()
{
  // function name: find_min_yscale
  // Find the minimum yscale for this data
  
  int thisBadCount, minyscale=480, mytag=0;
  float heightFound, thisAvgBad;
  
  avgBadCount = 0.0;

  if (thetatoggle)
    mytag = THETAT;
  if (phitoggle)
    mytag = PHIT;
  if (omegatoggle)
    mytag = OMEGAT;
  if (xtoggle)
    mytag = XT;
  if (deltaXtoggle)
    mytag = DELTAXT;
  if (absDeltaXtoggle)
    mytag = ABSDELTAXT;
  if (ytoggle)
    mytag = YT;
  if (deltaYtoggle)
    mytag = DELTAYT;
  if (absDeltaYtoggle)
    mytag = ABSDELTAYT;
  if (distancetoggle)
    mytag = DISTANCET;
  if (magtoggle)
    mytag = MAGT;
  if (cc1toggle)
    mytag = PFTCC;
  if (cc2toggle)
    mytag = PRJCC;
  if (cc3toggle)
    mytag = CMPCC;
  
  if(mytag != 0) {
    heightFound = find_max_height(info.info, startbox, endbox, step,
				  mytag, &thisBadCount);
    heightFound += thisBadCount;
    thisAvgBad = thisBadCount / heightFound;
    avgBadCount = MAX(avgBadCount, thisAvgBad);
    heightFound = MIN(heightFound,(float)maxYbox);
    yscale = 480.0/heightFound;
    if (yscale < minyscale)
      minyscale = yscale;
  }
  
  yscale = minyscale;
  if(yscale == 0.0)
    yscale = 1.0;  // Just give it something
  
} // end function find_min_yscale



void paramMode_change(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: paramMode_change
  // They've toggled a different mode

  int binVal, totalToggled;
  char *minMaxStr;
  int tag = *((int *) client_data);

  if(icoDispMode==ICOSORT) { // Init everything
    idtoggle = thetatoggle = phitoggle = omegatoggle = xtoggle = ytoggle =
      magtoggle = cc1toggle = cc2toggle = cc3toggle = distancetoggle =
      deltaXtoggle = deltaYtoggle = absDeltaXtoggle = absDeltaYtoggle = FALSE;
    
    XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE] );
    XtSetValues(magtW, arglst, 1); 
    XtSetValues(pftccW, arglst, 1); 
    XtSetValues(prjccW, arglst, 1); 
    XtSetValues(cmpccW, arglst, 1); 
    
    // Just in case we need it for ICOSORT
    XtSetArg(arglst[0], XmNset, FALSE );
    XtSetValues(idtW, arglst, 1); 
    XtSetValues(thetatW, arglst, 1); 
    XtSetValues(phitW, arglst, 1); 
    XtSetValues(omegatW, arglst, 1); 
    XtSetValues(xtW, arglst, 1); 
    XtSetValues(ytW, arglst, 1); 
    XtSetValues(deltaXTW, arglst, 1); 
    XtSetValues(deltaYTW, arglst, 1); 
    XtSetValues(absDeltaXTW, arglst, 1); 
    XtSetValues(absDeltaYTW, arglst, 1); 
    XtSetValues(distanceTW, arglst, 1); 
    XtSetValues(magtW, arglst, 1); 
    XtSetValues(pftccW, arglst, 1); 
    XtSetValues(prjccW, arglst, 1); 
    XtSetValues(cmpccW, arglst, 1); 
    
    XFlush(dsply);
    // Turn the new one ON
    XtSetArg(arglst[0], XmNset, TRUE );
  }
  
  if (info.info) {
    find_minmax(info.info); 
  }
  else {
    XtVaSetValues(widget, XmNset, FALSE, NULL);
    XFlush(dsply);
    askThem("No File opened!", XmDIALOG_QUESTION);
    return;
  }

  if(tag==INCDEC) {
    incSort = !incSort;
    if(icoDispMode==ICOSORT) {
      if(incSort)
	doLabel("Increasing", incDecW, DEFAULT);
      else 
	doLabel("Decreasing", incDecW, DEFAULT);
    }
    return;
  }
  else if(tag==IDT) {
    idtoggle = (idtoggle) ? 0 : 1;
    global_minmax(); 
    if(icoDispMode==ICOSORT)
      XtSetValues(idtW, arglst, 1); 
  }
  else if(tag==THETAT) {
    thetatoggle = (thetatoggle) ? 0 : 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(thetatW, arglst, 1); 
  }
  else if(tag==PHIT) {
    phitoggle = (phitoggle) ? 0 : 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(phitW, arglst, 1); 
  }
  else if(tag==OMEGAT) {
    omegatoggle = (omegatoggle) ? 0 : 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(omegatW, arglst, 1); 
  }
  else if(tag==XT) {
    xtoggle = (xtoggle) ? 0 : 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(xtW, arglst, 1); 
  }
  else if(tag==DELTAXT) {
    deltaXtoggle = !deltaXtoggle;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(deltaXTW, arglst, 1); 
  }
  else if(tag==ABSDELTAXT) {
    absDeltaXtoggle = !absDeltaXtoggle;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(absDeltaXTW, arglst, 1); 
  }
  else if(tag==YT) {
    ytoggle = (ytoggle) ? 0 : 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(ytW, arglst, 1); 
  }
  else if(tag==DELTAYT) {
    deltaYtoggle = !deltaYtoggle;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(deltaYTW, arglst, 1); 
  }
  else if(tag==ABSDELTAYT) {
    absDeltaYtoggle = !absDeltaYtoggle;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(absDeltaYTW, arglst, 1); 
  }
  else if(tag==DISTANCET) {
    distancetoggle = !distancetoggle;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(distanceTW, arglst, 1); 
  }
  else if(tag==MAGT) {
    magtoggle = (magtoggle) ? 0 : 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(magtW, arglst, 1); 
    else if(icoDispMode==HISTOGRAM) {
      XtSetArg(arglst[0], XmNset, FALSE );
      XtSetArg(arglst[1], XmNforeground, pixels_return[WHITE] );
      XtSetValues(pftccW, arglst, 2); 
      XtSetValues(prjccW, arglst, 2); 
      XtSetValues(cmpccW, arglst, 2); 
      cc1toggle = cc2toggle=cc3toggle= FALSE;
      if(magtoggle)
	XtSetArg(arglst[0], XmNforeground, pixels_return[GREEN] );
      else 
	XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE] );
      XtSetValues(magtW, arglst, 1); 
    }
    XFlush(dsply);
  }
  else if(tag==PFTCC) {
    cc1toggle = (cc1toggle) ? 0 : 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(pftccW, arglst, 1); 
    else if(icoDispMode==HISTOGRAM) {
      XtSetArg(arglst[0], XmNset, FALSE );
      XtSetArg(arglst[1], XmNforeground, pixels_return[WHITE] );
      XtSetValues(magtW, arglst, 2); 
      magtoggle= FALSE;
      if(cc1toggle)
	XtSetArg(arglst[0], XmNforeground, pixels_return[BLUE] );
      else 
	XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE] );
      XtSetValues(pftccW, arglst, 1); 
    }
    XFlush(dsply);
  }
  else if(tag==PRJCC) {
    cc2toggle = (cc2toggle) ? 0 : 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(prjccW, arglst, 1); 
    else if(icoDispMode==HISTOGRAM) {
      XtSetArg(arglst[0], XmNset, FALSE );
      XtSetArg(arglst[1], XmNforeground, pixels_return[WHITE] );
      XtSetValues(magtW, arglst, 2); 
      magtoggle= FALSE;
      if(cc2toggle)
	XtSetArg(arglst[0], XmNforeground, pixels_return[GREEN] );
      else 
	XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE] );
      XtSetValues(prjccW, arglst, 1); 
    }
    XFlush(dsply);
  }
  else if(tag==CMPCC) {
    cc3toggle = (cc3toggle) ? 0 : 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(cmpccW, arglst, 1); 
    else if(icoDispMode==HISTOGRAM) {
      XtSetArg(arglst[0], XmNset, FALSE );
      XtSetArg(arglst[1], XmNforeground, pixels_return[WHITE] );
      XtSetValues(magtW, arglst, 2); 
      magtoggle= FALSE;
      if(cc3toggle)
	XtSetArg(arglst[0], XmNforeground, pixels_return[RED] );
      else 
	XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE] );
      XtSetValues(cmpccW, arglst, 1); 
    }
    XFlush(dsply);
  }
  
  if (thetatoggle    || omegatoggle  || phitoggle  || xtoggle   || 
      ytoggle        || magtoggle    || cc1toggle  || cc2toggle || 
      cc3toggle      || deltaXtoggle || absDeltaXtoggle         || 
      distancetoggle || deltaYtoggle || absDeltaYtoggle) {
    startbox = min; 
    endbox = max;  
  }
  else {
    startbox = 0.0;
    endbox = 1.0;
    step = 0.016;
  }
      
  if(min>=max) // Can't have this
    max = min + 0.01;
  
  // Check to see if they've entered anything on the screen
  minMaxStr = XmTextGetString(strtXPPW);
  if(!EQU(minMaxStr,""))
    startbox = MIN(strtod(minMaxStr,0),max);
  XtFree(minMaxStr);
  minMaxStr = XmTextGetString(endXPPW);
  if(!EQU(minMaxStr,""))
    endbox = MAX(strtod(minMaxStr,0),min);
  XtFree(minMaxStr);
  minMaxStr = XmTextGetString(maxYPPW);
  if(!EQU(minMaxStr,""))
    maxYbox = (int)strtod(minMaxStr,0);
  maxYbox = MAX(1,maxYbox);
  XtFree(minMaxStr);
  TextSetInt(strtXPPW, startbox, FALSE);
  TextSetInt(endXPPW, endbox, FALSE);
  TextSetInt(maxYPPW, maxYbox, FALSE);
  
  binVal = wstrtod(binTextW);
  binVal = MAX(binVal, 1);
  TextSetInt(binTextW, binVal, FALSE);
  
  step = ((max-min)/480.0) * (480.0/binVal);
  if(endbox==startbox)
    pixelstep = step*linelength;
  else 
    pixelstep = (step*linelength)/(endbox-startbox);
  find_min_yscale();

  XFlush(dsply);

  if(icoDispMode==ICOSORT) { // They want to sort
      int x, y;

      totalToggled = idtoggle + thetatoggle + 
	phitoggle + omegatoggle + 
	xtoggle + ytoggle + magtoggle + distancetoggle +
	deltaXtoggle + deltaYtoggle + absDeltaXtoggle + absDeltaYtoggle + 
	cc1toggle + cc2toggle + cc3toggle;

      if(totalToggled != 1) { // not right
	if(totalToggled > 0)
	  askThem("You can only sort on one key.", XmDIALOG_INFORMATION);
      }
      else { // Sort it
	icoSortLevel = 1;
	DrawICOSWindow(NULL, NULL, NULL);
	if(idtoggle)
	  sortParticles(info.info,purdueHeader.nimg, 2, incSort);
	else if(thetatoggle)
	  sortParticles(info.info,purdueHeader.nimg, 3, incSort);
	else if(phitoggle)
	  sortParticles(info.info,purdueHeader.nimg, 4, incSort);
	else if(omegatoggle)
	  sortParticles(info.info,purdueHeader.nimg, 5, incSort);
	else if(xtoggle)
	  sortParticles(info.info,purdueHeader.nimg, 6, incSort);
	else if(ytoggle)
	  sortParticles(info.info,purdueHeader.nimg, 7, incSort);
	else if(magtoggle)
	  sortParticles(info.info,purdueHeader.nimg, 8, incSort);
	else if(cc1toggle)
	  sortParticles(info.info,purdueHeader.nimg, 9, incSort);
	else if(cc2toggle)
	  sortParticles(info.info,purdueHeader.nimg, 10, incSort);
	else if(cc3toggle)
	  sortParticles(info.info,purdueHeader.nimg, 11, incSort);
	else if(deltaXtoggle)
	  sortParticles(info.info,purdueHeader.nimg, 13, incSort);
	else if(deltaYtoggle)
	  sortParticles(info.info,purdueHeader.nimg, 14, incSort);
	else if(absDeltaXtoggle)
	  sortParticles(info.info,purdueHeader.nimg, 15, incSort);
	else if(absDeltaYtoggle)
	  sortParticles(info.info,purdueHeader.nimg, 16, incSort);
	else if(distancetoggle)
	  sortParticles(info.info,purdueHeader.nimg, 17, incSort);
	
	icoSortLevel = 2;
	TextSet(partPointTextW, "SORT finished.", FALSE);
	
	findNearestParticle(&x,&y,&foundPart,FALSE,1,TRUE,0);
	
      }
  }
  
  DrawICOSWindow(NULL, NULL, NULL);
  
} // end function paramMode_change



void markXYplotOutliers()
{
  // function name: markXYplotOutliers
  // Mark those points outside the given distance as outliers.  The
  // toggle value tells the display rtn to give feedback that this
  // point is an outlier.
  
  int numFound;
  P_INFO *thisInfo;
  
  // First unmark any previous ones
  thisInfo = info.info;
  while(thisInfo != NULL) {
    thisInfo->toggle &= NOTXYBADTHRESH;
    thisInfo = thisInfo->next;
  }
  
  numFound = 0;
  thisInfo = info.info;
  
  while(thisInfo != NULL) { // XY plot
    if(thisInfo->distance < minXYplot || thisInfo->distance > maxXYplot) { 
      numFound++;
      thisInfo->toggle |= XYBADTHRESH;
    }
    thisInfo = thisInfo->next;
  }
  
  DrawICOSWindow(NULL, NULL, NULL); 
  
} // end function markXYplotOutliers



void ppMinMax_change(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{
  // function name: ppMinMax_change
  // Change the concept of min/max.

  float thisVal;

  thisVal = wstrtod(widget);
  thisVal = MAX(thisVal, 0.0);
  thisVal = MIN(thisVal, distanceMax);

  // Now show them if it's changed 
  TextSetFloat(widget, thisVal, FALSE);
  
  // Put this value into "scroll space"
  thisVal *= 1000.0;

  if(widget == ppMinTextW) {
    minXYplot = thisVal / 1000.0;
    XtVaSetValues(ppMinScrollW, XmNvalue, (int)thisVal, NULL);
  }
  else if(widget == ppMaxTextW) {
    maxXYplot = thisVal / 1000.0;
    XtVaSetValues(ppMaxScrollW, XmNvalue, (int)thisVal, NULL);
  }
  else
    printf("Don't know...\n");
  
  markXYplotOutliers();
  
} // end function ppMinMax_change



void ppMinMax_scroll_change(Widget widget, XtPointer client_data, 
			    XtPointer call_data)
{
  // function name: ppMinMax_scroll_change
  // Call here whenever ppMinMax_scroll widgets is changed

  char outStr[64];
  float scrollVal;
  XmScrollBarCallbackStruct *temp = (XmScrollBarCallbackStruct *) call_data;

  if(!inited) return;

  scrollVal = temp->value;
  sprintf(outStr, "%f", scrollVal/1000.0); 

  if(widget == ppMinScrollW) {
    TextSet(ppMinTextW, outStr, FALSE);
    minXYplot = scrollVal / 1000.0;
  }
  else if (widget == ppMaxScrollW) {
    TextSet(ppMaxTextW, outStr, FALSE);
    maxXYplot = scrollVal / 1000.0;
  }

  markXYplotOutliers();
  XFlush(dsply);
  
} // end function ppMinMax_scroll_change



void symmetry_activate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: symmetry_activate
  // A new symmetry has been chosen

  symmetry = *((int *) client_data);
  DrawICOSWindow(NULL, NULL, NULL); 
  
} // end function symmetry_activate



void diamText_change(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: diamText_change
  // Change the diameter

  char outStr[2048];
  float diameter, move1, move2, move3, move4, move5;

  diameter = wstrtod(diamTextW);

  if(whichViewMove == MOVEMENT) { // Movement of surface feaures
    if(diameter > 0.0) {
      move1 = (2.0 * M_PI * diameter * 0.10)/360.0;
      move2 = (2.0 * M_PI * diameter * 0.25)/360.0;
      move3 = (2.0 * M_PI * diameter * 0.50)/360.0;
      move4 = (2.0 * M_PI * diameter *  1.0)/360.0;
      move5 = (2.0 * M_PI * diameter *  2.0)/360.0;
      
      sprintf(outStr,
	      "\t        \t\t\t Del_theta (degrees)\n\
\tParticle\t\t\t \n\
\tDiameter\t\t2.0\t1.0\t0.5\t0.25\t0.10\n\
\t--------\t\t---------------------------------------\n\
\t%.0f\t\t\t%.1f\t%.1f\t%.1f\t%.1f\t%.2f\n",
	      diameter, move5, move4, move3, move2, move1);
      TextSet(movementTextW, outStr, TRUE);
    }
  }
  else {
    sprintf(outStr,
	    "\t        \t\t\t Symmetry\n\
\tAngle   \t\t\t \n\
\tDelta   \t\t    532\t     52\t      5\t      1\n\
\t--------\t\t----------------------------------\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n",
	    3.0,52,266,519,2353,
	    2.0,96,573,1127,5253,
	    1.0,382,2173,4309,20809,
	    0.5,1448,8471,16869,82869,
	    0.25,5653,33439,66733,330751,
	    0.1,34427,207358,414353,2064448);
    TextSet(movementTextW, outStr, TRUE);
  }
  
} // end function diamText_change



void views_activate(Widget widget, XtPointer client_data, 
		    XtPointer call_data)
{
  // function name: views_activate
  // Setup for view and get movementscreen up on the display.
  
  whichViewMove = VIEWS;
  
  TextSet(diamTextW, "Not yet.", FALSE);
  
  doDialogTitle("Number of Views per ASU", movementScreenW, FALSE);
  
  doLabel("Angle Delta", movementLabelW, FALSE);
  
  // Hard code this for now
  sprintf(outStr,
	  "\t        \t\t\t Symmetry\n\
\tAngle   \t\t\t \n\
\tDelta   \t\t    532\t     52\t      5\t      1\n\
\t--------\t\t----------------------------------\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n",
	  3.0,52,266,519,2353,
	  2.0,96,573,1127,5253,
	  1.0,382,2173,4309,20809,
	  0.5,1448,8471,16869,82869,
	  0.25,5653,33439,66733,330751,
	  0.1,34427,207358,414353,2064448);
  TextSet(movementTextW, outStr, TRUE);
  
  raiseIt(movementScreenW);
  
} // end function views_activate



void movement_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{
  // function name: movement_activate
  // Get movement screen on display

  whichViewMove = MOVEMENT;
  
  TextSet(diamTextW, " ", FALSE);
  
  doDialogTitle("Movement of Surface Features", movementScreenW, FALSE);
  
  doLabel("Particle Diameter:", movementLabelW, FALSE);
  
  // Default this to something
  sprintf(outStr,
	  "\t        \t\t\t Del_theta (degrees)\n\
\tParticle\t\t\t \n\
\tDiameter\t\t2.0\t1.0\t0.5\t0.25\t0.10\n\
\t--------\t\t---------------------------------------\n");
  TextSet(movementTextW, outStr, TRUE);
  
  raiseIt(movementScreenW);
  
} // end function movement_activate



void info_list_proc(Widget widget, XtPointer client_data, 
		    XtPointer call_data)
{
  // function name: info_list_proc
  // Call here everytime they make a new choice from the info MASTER
  // list. Update the appropriate variables and make the display
  // appear as it should.

  XmListCallbackStruct *list = (XmListCallbackStruct *) call_data;
  globAskAnswer = list->item_position - 1; // Numbers from 1 not 0
  XtUnmanageChild(infoAnswerScreenW);
  gFinished= TRUE;

} // end function info_list_proc



void projASU_activate(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: projASU_activate
  // Tell them how to project

  XtUnmanageChild(parmTogsW);
  XtUnmanageChild(parmheadW);
  XtManageChild(projasuW);
  XFlush(dsply);
  
  askThem("Enter the desired Delta Angle below and hit the\n\
Return key.  OR enter the desired theta, phi and\n\
omega below and click the 'Project This:' button.",
	  XmDIALOG_INFORMATION);
  
} // end function projASU_activate



void projasu_cr_activate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{
  // function name: projasu_cr_activate
  // Do the projection from the param data and the MAP.

  int   i, m;
  int   nthe, curPhi, curTheta;
  int   lun, yOut1, yOut2, thisX, thisY, zOut;
  float dthe,dphi_90,lastTheta;
  float the1, the2, phi1;
  float thisOmega, thisPhi, thisTheta, thisSF;
  float maxF, minF, z1, z2;
  float *map2d;
  float *thisImg;

  if(EQU(paramMapFilename,""))
    askThem("No MAP open can't do the projections.", XmDIALOG_INFORMATION);
  else {

    if(haveParamFile) { // Get rid of any parameter file we've got
      haveParamFile = FALSE;
      inParamFilename[0]='\0';
      paramBoxesFilename[0]='\0';
      TextSet(paramFileTextW, inParamFilename, FALSE); 
      purdueHeader.nimg= 0;
      lastPackedNumImgs= 0;
      Free_Particle_Info(&info);
    }
      
    // Suppose they've got a previous packed file read
    for (i=0; i<purdueHeader.nimg; i++)
      freeNullFloat(&packImg[i]);
    purdueHeader.nimg = 0;
    
    // We've got to fake robem out that this is a packed file
    fileType= PACKEDPORTABLE;

    // Now clip this guy to our colormap
    scalePackFactor = (tryColors - 7.0) / 256.0;
    projASU = TRUE;
    
    // They can write this out as RAW
    XtVaSetValues(rawButtonW, XmNsensitive, TRUE, NULL);
    
    projASUdeltaAng = (int)(wstrtod(projasuTextW)+0.4999);
    projASUdeltaAng = MAX(projASUdeltaAng,1);
    TextSetInt(projasuTextW, projASUdeltaAng, TRUE);
    
    purdueHeader.numCols = mapCols;
    purdueHeader.numRows = mapRows;
    purdueHeader.numSecs = 1;
    packRadius = (mapCols/2);
    
    // ***************************************************************
    // *    Determine NVIEW, the number of model projection views    *
    // ***************************************************************
    
    projASUdeltaAng = MAX(projASUdeltaAng, 1);
    dthe    = projASUdeltaAng;
    dphi_90 = projASUdeltaAng;
    
    get_nview_mod__(&symmetry, &dthe, &dphi_90, &nview);

    // Calc what projections we'll need
    the1 =  0.0;
    the2 = 90.0;
    phi1 =  0.0;

    // 	******************************************************
    // 	*  Exceptions to the above values:                   *
    // 	*  SYM = 1    for asymmetric particle,               *
    // 	*      = 2-20 for particle with nfold about z-axis,  *
    // 	*      = 532  for icosahedral particle.              *
    // 	******************************************************
    
    if (symmetry >= 1 && symmetry <= 20) 
      the2 = 180.0;
    else if (symmetry == 532) 
      the1 = 69.09484255211070;
    
    freeNullFloat(&thetas);
    freeNullFloat(&phis);
    CK(thetas = (float *)malloc(sizeof(float)*nview));
    CK(phis   = (float *)malloc(sizeof(float)*nview));
    
    // Determine range of values for theta and phi
    get_t1t2p1__(&symmetry, &the1, &the2, &phi1);

    // Get (theta,phi) pairs that cover the asymmetric units
    calc_mod_tps__(thetas, phis, &the1, &the2, &phi1, &dthe,
		  &dphi_90, &nview, &symmetry);

    // RSSX - remember to add other symmetries - this is a quick fix
    if(symmetry == 532) {
      nthe = (int)(ceil((90.0-69.09)/dthe)+1);
      maxPhi = (int)ceil(31.72/dphi_90);
    }

    // Allocate for every one we need to enclose these in a square box.
    purdueHeader.nimg = nthe * maxPhi;
    
    // Allocate all storage, whether you'll use it or not
    for (m=0; m<purdueHeader.nimg; m++) {
      CK(packImg[m]= (float *) malloc(sizeof(float)*mapCols*mapRows));
      thisImg = packImg[m];
      init_floatarr(thisImg, mapRows*mapCols, 0.0); // initialize array
    }
    
    // Init the display
    DrawICOSWindow(NULL, NULL, NULL);
    
    // Now project all
    CK(map2d = (float *)malloc(sizeof(float)*mapCols*(mapRows+2)*nview));
    
    lun = 7;
    for (i=0; i<nview; i++) {
      
      // **************************************************************
      // *    NOTE USE OF OMEGA = 90.0 TO GET CORRECT ORIENTATIONS    *
      // **************************************************************
      
      thisOmega = 90.0;
      thisOmega = 180.0;
      thisTheta = thetas[i];
      thisPhi   = phis[i];
      z1 = 1.0;
      z2 = mapSecs;
      
      sprintf(outStr,"Projecting theta[%f] phi[%f].",thetas[i],phis[i]);
      TextSet(partPointTextW, outStr, FALSE);
      
      map_prj__((map2d+(i*mapCols*mapRows)), float_image_buffer1,
	       &mapCols, &mapRows, &mapSecs,
	       &thisTheta, &thisPhi, &thisOmega, 
	       &z1, &z2);
      
      // Give feedback 
      convertXY(&thisX, &thisY, phis[i], thetas[i]);
      DrawX(icosa_window1, image_gcBLUE, thisX, thisY, 5);
      XFlush(dsply);
    }
    
    farray_minmax(map2d, &minF, &maxF, nview*mapCols*mapRows);
    
    thisSF = (maxF > minF) ? 254.0/(maxF-minF) : 1.0;

    curTheta = curPhi = 0;
    lastTheta = thetas[0];
    for (i=0; i<nview; i++) {
      if(thetas[i] != lastTheta) {
	curTheta++;
	curPhi = 0;
	lastTheta = thetas[i];
      }
      
      // Point to the appropriate packed image
      thisImg = packImg[(curTheta*maxPhi)+curPhi];
      curPhi++;
      
      // We've got to flip the projection in the Y direction.
      // Tim draws from the bottom up -we're opposite
      
      zOut = i * mapCols*mapRows;
      for (y=0; y<mapRows; y++) {
	yOut1 = (mapRows - y - 1)*mapCols;
	yOut2 = y*mapCols;
	for (x=0; x<mapCols; x++)
	  thisImg[yOut2+x]= (map2d[zOut+yOut1+x]-minF)*thisSF;
      }
      
      // Give feedback
      convertXY(&thisX, &thisY, phis[i], thetas[i]);
      DrawX(icosa_window1, image_gcRED, thisX, thisY, 5);
      XFlush(dsply);
    }
    
    free(map2d);
    
    // Start at zero'th(1) image
    stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
    	     packImg, purdueHeader.nimg, 1);
    DrawWindow();    
  }
  
} // end function projasu_cr_activate



void squareText_changed(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: squareText_changed
  // They've changed the square params.

  makeSquareBox = wstrtod(widget);
  
} // end function squareText_changed



void project_one(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: project_one
  // Project from MAP.

  int   i, m;
  int   curPhi, curTheta;
  int   lun, yOut1, yOut2, thisX, thisY, zOut;
  float lastTheta, thisTheta;
  float thisOmega, thisPhi, thisSF;
  float maxF, minF, z1, z2;
  float *projn, *map2d;
  float *thisImg;

  if(EQU(paramMapFilename,"")) {
    askThem("No MAP open can't do the projections.", XmDIALOG_INFORMATION);
  }
  else {
    if(haveParamFile) { // Get rid of any parameter file we've got
      haveParamFile = FALSE;
      inParamFilename[0]='\0';
      paramBoxesFilename[0]='\0';
      TextSet(paramFileTextW, inParamFilename, FALSE); 
      purdueHeader.nimg= 0;
      lastPackedNumImgs= 0;
      Free_Particle_Info(&info);
    }
      
    // Suppose they've got a previous packed file read
    for (i=0; i<purdueHeader.nimg; i++)
      freeNullFloat(&packImg[i]);
    purdueHeader.nimg = 0;
    
    // We've got to fake robem out that this is a packed file
    fileType= PACKEDPORTABLE;

    // Now clip this guy to our colormap
    scalePackFactor = (tryColors - 7.0) / 256.0;
    projASU = TRUE;
    
    // They can write this out as RAW
    XtVaSetValues(rawButtonW, XmNsensitive, TRUE, NULL);
    
    thisTheta = wstrtod(thetaProjW);
    thisPhi   = wstrtod(phiProjW);
    thisOmega = wstrtod(omegaProjW);
    XFlush(dsply);      
    
    purdueHeader.numCols = mapCols;
    purdueHeader.numRows = mapRows;
    purdueHeader.numSecs = 1;
    packRadius = mapCols/2;
    
    // NVIEW is just one here
    maxPhi = nview= purdueHeader.nimg = 1;
    
    // Allocate all storage, whether you'll use it or not
    for (m=0; m<purdueHeader.nimg; m++) {
      CK(packImg[m]= (float *)
	 malloc(sizeof(float)*purdueHeader.numCols*purdueHeader.numRows));
      thisImg = packImg[m];
      init_floatarr(thisImg, purdueHeader.numRows*purdueHeader.numCols, 0.0);
    }
    
    freeNullFloat(&thetas);
    freeNullFloat(&phis);

    CK(thetas = (float *)malloc(sizeof(float)*nview));
    CK(phis   = (float *)malloc(sizeof(float)*nview));
    thetas[0] = thisTheta;
    phis[0]   = thisPhi;
    
    // Init the display
    DrawICOSWindow(NULL, NULL, NULL);
    
    // Now project all
    CK(projn = (float *)malloc(sizeof(float)*mapCols*(mapRows+2)));
    CK(map2d = (float *)malloc(sizeof(float)*mapCols*(mapRows+2)*nview));
    
    lun = 7;
    for (i=0; i<nview; i++) {
	
      // **************************************************************
      // *    NOTE USE OF OMEGA = 90.0 TO GET CORRECT ORIENTATIONS    *
      // **************************************************************
      // thisOmega = 90.0;
      // thisOmega = 180.0;
      // Leave thisOmega alone ???
      
      thisTheta = thetas[i];
      thisPhi   = phis[i];
      z1 = 1.0;
      z2 = (float)mapSecs;
      
      sprintf(outStr,"Projecting theta[%f] phi[%f].",thetas[i],phis[i]);
      TextSet(partPointTextW, outStr, TRUE);

      map_prj__((map2d+(i*mapCols*mapRows)), float_image_buffer1,
	       &mapCols, &mapRows, &mapSecs,
	       &thisTheta, &thisPhi, &thisOmega, 
	       &z1, &z2);

      // Give feedback
      convertXY(&thisX, &thisY, phis[i], thetas[i]);
      DrawX(icosa_window1, image_gcBLUE, thisX, thisY, 5);
      XFlush(dsply);
      
    }

    farray_minmax(map2d, &minF, &maxF, nview*mapCols*mapRows);
    
    thisSF = (maxF > minF) ? 254.0/(maxF-minF) : 1.0;
    
    curTheta = curPhi = 0;
    lastTheta = thetas[0];
    for (i=0; i<nview; i++) {
      if(thetas[i] != lastTheta) {
	curTheta++;
	curPhi = 0;
	lastTheta = thetas[i];
      }
      
      // Point to the appropriate packed image
      thisImg = packImg[curTheta*maxPhi + curPhi];
      curPhi++;
      
      // We've got to flip the projection in the Y direction.
      // Tim draws from the bottom up -we're opposite
      
      zOut = i * mapCols*mapRows;
      for (y=0; y<mapRows; y++) {
	yOut1 = (mapRows - y - 1)*mapCols;
	yOut2 = y*mapCols;
	for (x=0; x<mapCols; x++)
	  thisImg[yOut2+x]= (map2d[zOut+yOut1+x]-minF)*thisSF;
      }
      
      // Give feedback
      convertXY(&thisX, &thisY, phis[i], thetas[i]);
      DrawX(icosa_window1, image_gcRED, thisX, thisY, 5);
      XFlush(dsply);
    }
    
    free(projn);
    free(map2d);
    
    // Start at zero'th(1) image
    stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		     packImg, purdueHeader.nimg, 1);
    
    DrawWindow();
  }
  
} // end function project_one



void projASUunitCell_work()
{
  // function name: projASUunitCell_work
  // Draw all particles in this unit cell display.

  int i, thisX, thisY;
  
  unitCell_background();

  for (i=0; i<nview; i++) {
    convertXY(&thisX, &thisY, phis[i], thetas[i]);
    DrawX(icosa_window1, image_gcRED, thisX, thisY, 5);
  }
  
} // end function projASUunitCell_work



void rescale_densities(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: rescale_densities
  // Rescale all of the density values in this MAP.

  int   map_incore;
  float newMinMapPixel, newMaxMapPixel;

  newMinMapPixel = wstrtod(rescaleNewMinW);
  newMaxMapPixel = wstrtod(rescaleNewMaxW);
  XFlush(dsply);      
  
  //  Need to generalize this to more than just maps
  printf("ft[%d]  pack[%d]empor[%d]map[%d]\n",
	 fileType,PACKEDPORTABLE,EMPORTABLE,MAP);
  if(fileType==EMPORTABLE || fileType==BYTEPORTABLE || fileType==PORTABLE) {
    // Use img_scale() in imglib.f 
  }
  else  if(fileType==PACKEDPORTABLE) {
    // Use img_scale() in imglib.f
  }
  else if(fileType==MAP) { // Do for MAP
    map_incore = 1;
    
    map_scale__(float_image_buffer1, 
		&numCols, &numRows, &origNumSecs,
		&minMapPixel, &maxMapPixel, 
		&newMinMapPixel, &newMaxMapPixel, 
		&purdueHeader.img_min, &purdueHeader.img_max,
		&purdueHeader.img_avg, &purdueHeader.img_std, &map_incore);
    
    secMaxPixel = maxMapPixel = newMaxMapPixel;
    secMinPixel = minMapPixel = newMinMapPixel;
    
    calcMapStats(float_image_buffer1,
		 &purdueHeader.img_bkg, 
		 &purdueHeader.map_min,
		 &purdueHeader.map_max, 
		 &purdueHeader.img_avg, 
		 &purdueHeader.img_std,
		 &numCols, &numRows, &origNumSecs);   
  }
  
  XtUnmanageChild(rescaleScreenW);

} // end function rescale_densities



void rescale_activate(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: rescale_activate
  // Get interface about how to rescale this MAP's densities on the
  // display.

  sprintf(outStr,"Old Min Density =  %.2f   Enter NEW Min: ",minMapPixel);
  doLabel(outStr, rescaleOldMinW, FALSE);
  
  sprintf(outStr,"Old Max Density =  %.2f   Enter NEW Max: ",maxMapPixel);
  doLabel(outStr, rescaleOldMaxW, FALSE);
  
  raiseIt(rescaleScreenW);
  
} // end function rescale_activate



void calcSF_work(int thisCalcSFs)
{
  // function name: calcSF_work
  // Do what it takes to get structure factors calculated from the
  // given MAP.
  
  int numFramesMm;
  float resHi = 2.01; // Chop by a factor of 2+
  
  if(thisCalcSFs) {
    checkStatus("Calculating Structure Factors");
    map2sfs(resHi);
    
    checkStatus("Producing High Resolution MAP");
    numFramesMm = 1;
    resolution_tryit_activate(0,&numFramesMm,0);
    
    // Make sure they can do high res if they want etc.
    XtVaSetValues(loHiResW, XmNsensitive, TRUE, NULL); 
    
    // Now prevent them from trying to manipulate these - e.g.
    // make them reread from disk to manipulate.
    XtSetArg(arglst[0], XmNsensitive, FALSE );
    // XtSetValues(fixButtonW, arglst, 1);
    XtSetValues(truncaddButtonW, arglst, 1); 
    XtSetValues(rescaleButtonW, arglst, 1); 
    XtSetValues(modelW, arglst, 1); 
    XtSetValues(subtractMapButtonW, arglst, 1); 
    XtSetValues(calcSFButtonW, arglst, 1); 
    XtSetValues(hardRotateButtonW, arglst, 1); 
    XtSetValues(invertButtonW, arglst, 1); 
    XtSetValues(rotateButtonW, arglst, 1); 
    
    if(!immediateResource) { // Turn immediate update ON
      XtVaSetValues(immediateUpdateW, XmNset, TRUE, NULL);
      immediateMapShade = TRUE;
      XFlush(dsply);
    }
    
    zoomFactor = 1.0;  /* Let's zoom it up */
    TextSet(zoomTextW, "2.00", FALSE);  
    
    // Their shaded views are no longer valid - show them 
    // their new sections
    // shadedMaps= FALSE;
    // dispSections = TRUE;
    // XtSetArg(arglst[0], XmNlabelPixmap, sectionPixmap);
    // XtSetArg(arglst[1], XmNselectPixmap, unsectionPixmap);
    // XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
    // XtSetValues(sectionW, arglst, 3); 
    
    if(dispSections) {
      // Update the sections - if shaded views then they have to recalc
      // Now stuff the display buffer from our new FIXed map
      stuffMapImage(map_image_buffer, float_image_buffer1, 
		    file_width, file_height, nsecs,imgNum, 0);
      DrawWindow();
    }
    else { 
      // If we're already doing rendering then setup the scrolls appropriately
      mapShadeScreen_activate(NULL, NULL, NULL);
    }
  }
  else { // Prevent any high res stuff
    sprintf(outStr,"Low Resolution ON");
    XtSetArg(arglst[0], XmNsensitive, FALSE );
    XtSetValues(pixelSizeScrollW, arglst, 1); 
    XtSetValues(pixelSizeValueW, arglst, 1); 
    // WHy are we doing this - it's for the movie???
    // XtSetValues(pixelSizeToggleW, arglst, 1); 
    // XtSetValues(loHiResW, arglst, 1);
    
    XtSetArg(arglst[0], XmNset, FALSE );
    // WHy are we doing this - it's for the movie???
    // XtSetValues(pixelSizeToggleW, arglst, 1); 
    lowRes = TRUE;
  }
  
} // end function calcSF_work



void calcSF_activate(Widget widget, XtPointer client_data, 
		     XtPointer call_data)
{
  // function name: calcSF_activate
  // Setup the interface for calculating structure factors.

  // Don't do this here!!!  We'll do it on the fly as they 
  // cut their map down and then change the pixel size in the HiRes
  // interface ... 
  // calcSF_work(TRUE);

  // Insert the call to inter3d() here to make them a 100 cubed lowres
  // buffer to work with

  // DO this instead
  // Make sure they can do high res if they want etc.
  XtVaSetValues(loHiResW, XmNsensitive, TRUE, NULL); 
  
  // Now remember these forever
  lowRes_buffer1 = float_image_buffer1;
  lowResCols = numCols;
  lowResRows = numRows;
  lowResSecs = origNumSecs;
  
  purdueHeader.numCols = numCols;
  purdueHeader.numRows = numRows;
  nsecs= purdueHeader.numSecs = origNumSecs;
  
  xCen = numCols/2;
  yCen = numRows/2;
  zCen = origNumSecs/2;
  
  largestDim = MAX(numCols, numRows);
  largestDim = MAX(largestDim, origNumSecs);
  
  // We've got to make this larger for asymmetric maps 
  // and all maps that have information "outside" of the
  // normal sphere of our normal maps,
  // or else we just keep cliping the box into a sphere
  
  globMaxRadius = (int)(origNumSecs/2) - 1;
  globMinRadius = 0.0;
  
  // Set scrolls on depthcue screen - Give feedback
  sprintf(outStr,"%.0f",globMinRadius);
  doValue(outStr, minRadiusValueW, FALSE);
  
  sprintf(outStr,"%.0f",globMaxRadius);
  doValue(outStr, maxRadiusValueW, FALSE);
  
  // Radius is origNumSecs/2
  XtVaSetValues(maxRadiusScrollW,XmNsliderSize, 5, NULL);
  XtVaSetValues(maxRadiusScrollW,XmNmaximum, 
		(int)((globMaxRadius)+5),NULL);
  XtVaSetValues(maxRadiusScrollW,XmNvalue, (int)(globMaxRadius),NULL);
  
  XtVaSetValues(minRadiusScrollW,XmNsliderSize, 5, NULL);
  XtVaSetValues(minRadiusScrollW,XmNmaximum, 
		(int)((globMaxRadius)+5),NULL);
  XtVaSetValues(minRadiusScrollW,XmNvalue, (int)(globMinRadius),NULL);
  
  XtUnmanageChild(statusScreenW);
  
} // end function calcSF_activate



void map2sfs(float resHiP)
{
  // function name: map2sfs
  // Convert the current map into structure factors
  
  int i, j, k, m, n, o, origNC, origNR, origNS;
  int hmax, kmax, lmax;
  float voxSiz=1.0, resLo=1000.0, resHi, resoScale=1.0;
  
  resHi = resHiP;

  origNC = numCols;
  origNR = numRows;
  origNS = origNumSecs; 
  
  if(calcBin) { // BIN this puppy down
    float totVals;
    int origNCtNR,iOff,jOff;
    
    numCols /= 2;
    numRows /= 2;
    origNumSecs /= 2;
    nsecs = origNumSecs;
    origNCtNR = origNC * origNR;

    for (i=0; i<(origNS-1); i+=2) {
      for (j=0; j<(origNR-1); j+=2) {
	for (k=0; k<(origNC-1); k+=2) {
	  totVals = 0.0;
	  for (m=0; m<2; m++) {
	    iOff = ((i+m)*origNCtNR);
	    for (n=0; n<2; n++) {
	      jOff = (j+n)*origNC;
	      for (o=0; o<2; o++) {
		totVals += float_image_buffer1[iOff+jOff+k+o];
	      }
	    }
	  }
	  float_image_buffer1[((i/2)*numCols*numRows)+
			      ((j/2)*numCols)+(k/2)]= totVals/8.0;
	}
      }
    }
  }
  
  origNC = numCols;
  origNR = numRows;
  origNS = numSecs = origNumSecs;  // They may have some depthcued stuff
  
  map3dt_info_rwa__(&numCols,&numRows,&numSecs,
                   &hmax, &kmax, &lmax, &nsf,
                   &a_cell, &b_cell, &c_cell,
		   &voxSiz, &resLo, &resHi, &resoScale);

  if(numCols != origNC || numRows != origNR || numSecs != origNS )
    for (i=0; i<numSecs; i++)
      for (j=0; j<numRows; j++)
	for (k=0; k<numCols; k++)
	  float_image_buffer1[i*numCols*numRows + j*numCols + k]= 
	    float_image_buffer1[i*origNC*origNR + j*origNC + k];
  
  if(sFacs != NULL)
    free(sFacs);

  sFacs = (strucFact_t *)malloc(sizeof(strucFact_t)*nsf);

  main_map3dt__(&numCols,&numRows,&numSecs, float_image_buffer1, 
	       sFacs, &nsf, &hmax, &kmax, &lmax);

} // end function map2sfs



void resolution_tryit_activate(Widget widget, int *tag, 
			       XmAnyCallbackStruct *callback_data)
{
  // function name: resolution_tryit_activate
  // They want to display this MAP with a different pixel resolution.

  int i, j, k, buffer_size, norien=1;
  float xminP=0.0, xmaxP=0.0, yminP=0.0, ymaxP=0.0, zminP=0.0, zmaxP=0.0;
  float dxP, dyP, dzP, resHi, maxReso=1.0;

  if(lowRes || pixelSizeSave == globPixelSize)
    return;  // No need to recalc
  
  // Let's build the buffer we need from the lowRes buffer
  numCols = cutMax0Save - cutMin0Save ;
  numRows = cutMax1Save - cutMin1Save ;
  numSecs = origNumSecs = cutMax2Save - cutMin2Save ;
  
  if(float_image_buffer1 && float_image_buffer1 != lowRes_buffer1)
    free(float_image_buffer1);
  
  float_image_buffer1 = (float *)malloc(sizeof(float)*
					numCols*numRows*numSecs);

  for (i=cutMin2Save; i<cutMax2Save; i++)
    for (j=cutMin1Save; j<cutMax1Save; j++)
      for (k=cutMin0Save; k<cutMax0Save; k++)
	float_image_buffer1[(i-(int)cutMin2Save)*numCols*numRows +
			    (j-(int)cutMin1Save)*numCols +
			    k-(int)cutMin0Save] =  
	  lowRes_buffer1[i*lowResCols*lowResRows + j*lowResCols + k];
  
  // Let's calc our structure factors
  checkStatus("Calculating Structure Factors");
  resHi = 2.01; // Let's chop it down by a factor of 2
  map2sfs(resHi);
  
  zoomFactor = 1.0;  // Let's zoom it up
  TextSet(zoomTextW, "2.00", FALSE);  
  
  if(sFacs == NULL) {
    askThem("No Structure factors could be produced.", XmDIALOG_INFORMATION);
    return;
  }
  
  if(!calcLowResMap) {
    sprintf(outStr,"[%d] Calculating High Resolution Image",*tag);
    TextSet(timeLeftLabelW, outStr, TRUE);
  }
  
  if(calcLowResMap) // Calc the gold std low res map
    pixelSizeFactor = dxP = dyP = dzP = numCols/100.0;
  else { // Do what they've dialed from the 3D rendering screen
    if(calcBin)
      dxP = dyP = dzP = (globPixelSize / 2.0); /* It's binned */
    else
      dxP = dyP = dzP = globPixelSize;
  }
  
  cutFactor = MAX((xmaxP-xminP),(ymaxP-yminP));
  cutFactor = MAX((zmaxP-zminP),cutFactor);
  
  // Now convert Structure factors back into a 3D MAP
  if(float_image_buffer1 && float_image_buffer1 != lowRes_buffer1)
    free(float_image_buffer1);
  
  // Approximate buffer size and allocate memory
  buffer_size = (int)(numCols/dxP) * (int)(numRows/dyP) * (int)(numSecs/dzP);
  float_image_buffer1 = (float *) malloc(sizeof(float)*buffer_size);
  
  main_emsf__(&numCols, &numRows, &numSecs, &nsf, 
	      sFacs, float_image_buffer1,
	      &a_cell, &b_cell, &c_cell, &maxReso,
	      &xminP, &xmaxP, &yminP, &ymaxP, &zminP, &zmaxP,
	      &dxP, &dyP, &dzP, &norien);
  
  printf("\n\nNew dimensions ->  %d,%d,%d\n", numCols, numRows, numSecs);
  
  file_width  =  purdueHeader.numCols = numCols;
  file_height =  purdueHeader.numRows = numRows;
  nsecs = purdueHeader.numSecs = purdueHeader.nimg = origNumSecs = numSecs;
  
  fileBinSecs = numSecs;
  fileBinRows = numRows;
  fileBinCols = numCols;
  
  xCen = numCols/2;
  yCen = numRows/2;
  zCen = origNumSecs/2;
  
  // We may be doing a movie (on pixelSize) so let's decide on 
  // what the largestDim is outside of here. 
  // largestDim = numCols;
  // largestDim = MAX(largestDim,numRows);
  // largestDim = MAX(largestDim,origNumSecs);
  
  calcMapStats(float_image_buffer1,
	       &purdueHeader.img_bkg, &purdueHeader.map_min,
	       &purdueHeader.map_max, &purdueHeader.img_avg, 
	       &purdueHeader.img_std,
	       &numCols, &numRows, &origNumSecs);

  secMaxPixel = maxMapPixel = purdueHeader.map_max;
  secMinPixel = minMapPixel = purdueHeader.map_min;
  if(secMaxPixel != secMinPixel)
    scaleMapFactor= (tryColors - 7.0) / (secMaxPixel-secMinPixel);
  else {
    scaleMapFactor= 1.0;
    maxMapPixel = minMapPixel + 1.0;
    secMaxPixel = secMinPixel + 1.0;
  }
  
  // We've got to make this larger for asymmetric maps 
  // and all maps that have information "outside" of the
  // normal sphere of our normal maps,
  // or else we just keep cliping the box into a sphere
  
  // Radius is origNumSecs/2
  XtVaSetValues(maxRadiusScrollW, XmNsliderSize, 5,NULL);
  XtVaSetValues(maxRadiusScrollW, XmNmaximum, (int)((origNumSecs*2)+5),NULL);
  
  XtVaSetValues(minRadiusScrollW, XmNsliderSize, 5,NULL);
  XtVaSetValues(minRadiusScrollW, XmNmaximum, (int)((origNumSecs*2)+5),NULL);
  
  printf("Resolve MAX[%f]  MIN[%f]\n",maxMapPixel,minMapPixel);
  secMaxPixel = maxMapPixel;
  secMinPixel = minMapPixel;
  
  // Give feedback on main window title bar
  sprintf(outStr,"Robem: [%s] - [%d,%d,%d]",fileName.addr,
	  numCols,numRows,numSecs);
  
  doDialogTitle(outStr, mainWindowW, FALSE);
  
  pixelSizeSave = globPixelSize;
  
  reset_xyz_scrolls();
  
} // end function resolution_tryit_activate



void flicker_change(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: flicker_change 
  // Toggle the concept for flicker.

  flickerIt = *((int *) client_data);

} // end function flicker_change



void radTog_change(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: radTog_change
  // Toggle what values get displayed.

  if(widget == subRadrdW)
    rdensit = !rdensit;
  else if(widget == subRadwrW)
    wrdensit = !wrdensit;
  else if(widget == subRadccW)
    ccoefit = !ccoefit;
  else if(widget == subRadwcW)
    wccoefit = !wccoefit;

  DrawWindow();

} // end function radTog_change



void smMinMax_reset_activate(Widget widget, XtPointer client_data, 
			     XtPointer call_data)
{
  // function name: smMinMax_reset_activate
  // Reset to last gold std values.

  minMapPixel = saveMinMapPixel;
  maxMapPixel = saveMaxMapPixel;
  sprintf(outStr,"Min: %.2f",minMapPixel);
  doLabel(outStr, subSmMinLW, FALSE);
  sprintf(outStr,"Max: %.2f",maxMapPixel);
  doLabel(outStr, subSmMaxLW, FALSE);

} // end function smMinMax_reset_activate



void subtractMap_read()
{
  // function name: subtractMap_read
  // Read the second map to be subtracted from the first.

  char *s1;
  int i, convTsb;
  off_t imgOffset;
  int strtCol, strtRow, strtSec;
  int thisImg, pifMode;
  float minMapP, maxMapP, recipFSF;
  fString_t rWrite, inFileName;
  forPifHead_t thisHeader;

  checkStatus("Reading file ... ");
  raiseIt(statusScreenW);
  
  subtractingMap = FALSE;
  i= 4;
  thisHeader.luNum = i;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 1;
  inFileName.addr = malloc(sizeof(char)*64);
  inFileName.length = 64;
  sprintf(rWrite.addr,"r");
  
  s1 = XmTextGetString(subFile2TW); 
  sprintf(subtractFilename,"%s",s1);
  XtFree(s1);
  
  sprintf(inFileName.addr,"%s",subtractFilename);
  pif_open(inFileName.addr, &i, rWrite.addr);
  
  pif_read_gh(&thisHeader);
  recipFSF = thisHeader.recipFloatScaleFactor;

  if(numCols < thisHeader.numCols ||
     numRows < thisHeader.numRows ||
     numSecs < thisHeader.numSecs) {
    askThem("Map dimensions are greater than first MAP.", 
	    XmDIALOG_INFORMATION);
    goto noRead;
  }
  
  pifMode = thisHeader.mode;

  if(pifMode!=20) { // Not a map
    askThem("Not a MAP file.", XmDIALOG_INFORMATION);
    goto noRead;
  }
  
  thisImg = 1;  // For fortran
  thisHeader.imgNum = thisImg;
  pif_read_dh(&thisHeader);
  
  recipFSF = thisHeader.recipFloatScaleFactor;
  imgOffset  = thisHeader.imgOffset; 
  pifMode    = thisHeader.mode; 
  
  sub2binSec  = thisHeader.numSecs;
  sub2binRow  = thisHeader.numRows;
  sub2binCol  = thisHeader.numCols;
  binFactor = thisHeader.binFactor;
  
  if(binFactor == 2) { // 1/2
    sub2binSec  = thisHeader.numSecs/2+1;  
    sub2binRow  = thisHeader.numRows/2+1;
    sub2binCol  = thisHeader.numCols/2+1;
    if(sub2binSec%2 == 0)
      sub2binSec--;
    if(sub2binRow%2 == 0)
      sub2binRow--;
    if(sub2binCol%2 == 0)
      sub2binCol--;
  }
  
  freeNullFloat(&subtract_image_buffer1);
  CK(subtract_image_buffer1 = (float *)
     malloc(sizeof(float)*sub2binCol*sub2binRow*sub2binSec));
  freeNullFloat(&subtract_image_buffer2);
  CK(subtract_image_buffer2 = (float *)
     malloc(sizeof(float)*sub2binCol*sub2binRow*sub2binSec));
  
  // Don't reset concept of minMapPixel and maxMapPixel here!!!
  maxMapP = -HUGE;
  minMapP = HUGE;
  convTsb = TRUE;
  pif_tsbworld(convTsb);
  binFactor = 1;
  strtCol = strtRow = strtSec = 1;
  pif_read_map_short_image_(&i, subtract_image_buffer1,
			    &imgOffset,
			    &sub2binCol, &sub2binRow, &sub2binSec,
			    &recipFSF,
			    &minMapP,&maxMapP, &binFactor,
			    &strtCol, &strtRow, &strtSec,
			    &sub2binCol, &sub2binRow, &sub2binSec);
  pif_close(&i);
  
  if(sub2binCol < numCols || sub2binRow < numRows || sub2binSec < numSecs ) { 
    // Let's just place it in a bigger buffer
    float *tmpBuf;
    CK(tmpBuf = (float *)malloc(sizeof(float)*numCols*numRows*numSecs));
    add_map_padded(subtract_image_buffer1, 
		   sub2binCol, sub2binRow, sub2binSec,
		   tmpBuf, numCols, numRows, numSecs);
    sub2binCol = numCols;
    sub2binRow = numRows;
    sub2binSec = numSecs;
    free(subtract_image_buffer1);
    subtract_image_buffer1= tmpBuf;
    // Let's get this one the right size also
    freeNullFloat(&subtract_image_buffer2);
    CK(subtract_image_buffer2 = (float *)
       malloc(sizeof(float)*sub2binCol*sub2binRow*sub2binSec));
  }

  // Let's get some global ideas about min/max map pixel here
  saveMinMapPixel = minMapPixel;
  saveMaxMapPixel = maxMapPixel;
  if(maxMapP > maxMapPixel)
    maxMapPixel = maxMapP;
  if(minMapP < minMapPixel)
    minMapPixel = minMapP;
  sprintf(outStr,"Min: %.2f",minMapPixel);
  doLabel(outStr,subSmMinLW, FALSE);
  sprintf(outStr,"Max: %.2f",maxMapPixel);
  doLabel(outStr,subSmMaxLW, FALSE);
  
  threshScrollValue  = minMapPixel;
  TextSetFloat(subThreshTW, threshScrollValue, FALSE); 
  XtVaSetValues(subThreshSW, XmNminimum, (int)(minMapPixel - 5), NULL);
  XtVaSetValues(subThreshSW, XmNmaximum, (int)(maxMapPixel + 20), NULL);
  XtVaSetValues(subThreshSW, XmNsliderSize, 10, NULL);
  XtVaSetValues(subThreshSW, XmNvalue, (int)(threshScrollValue), NULL);
  
  if(diffSection)
    diffSec_really_doit();
  
 noRead:;
  XtUnmanageChild(statusScreenW);
  
} // end function subtractMap_read



void diffSecSave_activate(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: diffSecSave_activate
  // Save the subtracted buffer to disk.
  
  if(widget == subDiffSBW) { // Save it
    radScaleM = diffSecScrollValue;
    radScaleB = diffSec2ScrollValue;
    sprintf(outStr,"MAP1 = (MAP2 * %.3f) + %.3f",
	    radScaleM, radScaleB);
    doLabel(outStr, subEquationW, FALSE);
  }
  else if(widget == subDiffRBW) { // Reset it
    diffSecScrollValue  = radScaleM;
    diffSec2ScrollValue = radScaleB;
    XtSetArg(arglst[0], XmNvalue, (int)(diffSecScrollValue*100.0));
    XtSetValues(subDiffSW, arglst, 1);
    sprintf(outStr,"%.2f",diffSecScrollValue);
    TextSet(subDiff1TW, outStr, FALSE); 
    XtSetArg(arglst[0], XmNvalue, (int)(diffSec2ScrollValue*100.0));
    XtSetValues(subDiffS2W, arglst, 1);
    sprintf(outStr,"%.2f",diffSec2ScrollValue);
    TextSet(subDiff2TW, outStr, FALSE); 
    sprintf(outStr,"MAP1 = (MAP2 * %.3f) + %.3f", radScaleM, radScaleB);
    doLabel(outStr, subEquationW, FALSE);
    if(diffSection)
      diffSec_really_doit();
  }
  else if(widget == subMagSaveW) { // MAG save
    saveBmag = magSecScrollValue;
    sprintf(outStr,"%.3f",magSecScrollValue);
    TextSet(subMagStartTW, outStr, FALSE); 
  }
  else if(widget == subMagResetW) { // MAG reset
    magSecScrollValue  = saveBmag;
    XtSetArg(arglst[0], XmNvalue, (int)(magSecScrollValue*1000.0));
    XtSetValues(subMagScrollW, arglst, 1);
    sprintf(outStr,"%.3f",magSecScrollValue);
    TextSet(subMagsecTW, outStr, FALSE);
    if(diffSection)
      diffSec_really_doit();
  }
  else if(widget == subMag1ResetW) { // starting MAG reset
    saveBmag = magSecScrollValue= 1.0;
    XtSetArg(arglst[0], XmNvalue, (int)(magSecScrollValue*1000.0));
    XtSetValues(subMagScrollW, arglst, 1);
    sprintf(outStr,"%.3f",magSecScrollValue);
    TextSet(subMagsecTW, outStr, FALSE); 
    TextSet(subMagStartTW, "1.00", FALSE); 
    if(diffSection)
      diffSec_really_doit();
  }
  else if(widget == subMag2ResetW) { // step MAG reset
    if(subMap2d3d==2)
      sprintf(outStr,"0.01");
    else 
      sprintf(outStr,"0.005");
    TextSet(subMagStepTW, outStr, FALSE);
    if(diffSection)
      diffSec_really_doit();
  }
  else if(widget == subMag3ResetW) { // numStep MAG reset
    if(subMap2d3d==2)
      sprintf(outStr,"11");
    else 
      sprintf(outStr,"5");
    TextSet(subMagNumTW, outStr, FALSE); 
    if(diffSection)
      diffSec_really_doit();
  }
  else if(widget == subRadRW) { // starting RAD reset
    TextSet(subRadMinTW, "1", FALSE); 
    TextSetInt(subRadMaxTW, numSecs/2, FALSE); 
    if(diffSection)
      diffSec_really_doit();
  }
  
} // end function diffSecSave_activate



void phaseArrowButton(Widget widget, XEvent *event)
{
  // function name: phaseArrowButton
  // They want to change the current phase by using the arrows

  float speedFactor, stepVal;

  if(widget==phasXupW || widget==phasXdownW)
    stepVal = wstrtod(moveOrigXW);
  else
    stepVal = wstrtod(moveOrigYW);
  
  speedFactor = (float) GetButtonVal(event->xbutton.button, 10, 5, 1) / 100.;
  
  if(widget==phasXupW || widget==phasYupW)
    stepVal += speedFactor;
  else 
    stepVal -= speedFactor;
  
  stepVal = MAX(stepVal, 0.000000);
  stepVal = MIN(stepVal, 1.000000);
  
  sprintf(outStr,"%.3f",stepVal);
  
  if(widget==phasXupW || widget==phasXdownW)
    TextSet(moveOrigXW, outStr, FALSE);
  else
    TextSet(moveOrigYW, outStr, FALSE);
  
} // end function phaseArrowButton



void smMMaxArrowButton(Widget widget, XEvent *event)
{
  // function name: smMMaxArrowButton
  // They want to change the current min/max by using the arrows

  float speedFactor, stepVal;

  if(widget==subRadMaxUpW || widget==subRadMaxDownW)
    stepVal = wstrtod(subRadMaxTW);
  else
    stepVal = wstrtod(subRadMinTW);
  
  speedFactor = (float) GetButtonVal(event->xbutton.button, 1, 2, 5);
  
  if(widget==subRadMaxUpW || widget==subRadMinUpW)
    stepVal += speedFactor;
  else 
    stepVal -= speedFactor;
  
  stepVal= MAX(stepVal,0.000000);
  stepVal= MIN(stepVal,(sub2binCol/2));
  
  sprintf(outStr,"%.3f",stepVal);
  
  if(widget==subRadMaxUpW || widget==subRadMaxDownW)
    TextSet(subRadMaxTW, outStr, FALSE);
  else
    TextSet(subRadMinTW, outStr, FALSE);
  
  if(diffSection)
    diffSec_really_doit();
  
} // end function smMMaxArrowButton



void subtractMap_2d3d_activate(Widget widget, XtPointer client_data,
			       XtPointer call_data)
{
  // function name: subtractMap_2d3d_activate
  // Toggle whether they want to do 2D subtraction or 3D.
  
  if(widget == sub2dW) {
    subMap2d3d= 2;
    XtSetArg(arglst[0], XmNforeground, pixels_return[RED]);
    XtSetValues(sub2dW, arglst, 1); 
    XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE]);
    XtSetValues(sub3dW, arglst, 1); 
    XFlush(dsply);
  }
  else if(widget == sub3dW) {
    subMap2d3d= 3;
    XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE]);
    XtSetValues(sub2dW, arglst, 1); 
    XtSetArg(arglst[0], XmNforeground, pixels_return[RED]);
    XtSetValues(sub3dW, arglst, 1); 
    
    // Reset for 3D trial
    TextSet(subMagStepTW, "0.005", FALSE); 
    TextSet(subMagNumTW, "5", TRUE); 
  }
  
} // end function subtractMap_2d3d_activate



void recalc_ymxb_thresh()
{
  // function name: recalc_ymxb_thresh
  // We've got to recalc y=mx+b every time they change the threshold

  int *passTag;
  passTag = (int *)malloc(sizeof(int));
  *passTag = 1;

  subtractMap_findRad(0, (XtPointer)passTag, 0);

} // end function recalc_ymxb_thresh



void subtractMap_findRad(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: subtractMap_findRad
  // Use findRad rtns to find relative mag for the maps.

  int tag = *((int *) client_data);

  int i, saveNbins, nfit, sec1, sec2;
  float *thisBuf1, *thisBuf2;
  float xc, yc, zc, radlo, radhi;
  float tbccoef,  map_std, bmag;

  // Get rid of 3D rendering if it's there
  if(XtIsManaged(mapShadeScreenW))
    mapshade_deactivate(NULL, NULL, NULL);
  
  subtractingMap = FALSE;
  //  submapFindRad  = TRUE;
  
  if(!subtract_image_buffer1) { // Must read first
    askThem("Must Read second map file first.", XmDIALOG_INFORMATION);
    goto noRead;
  }

  if(tag!=1)
    raiseIt(statusScreenW);

  XFlush(dsply);

  xc = (numCols+1)/2.0;
  yc = (numRows+1)/2.0;
  zc = (numSecs+1)/2.0;
  radstep = 1.0;
  radlo = wstrtod(subRadMinTW);
  radlo -= 1.0;  /* Zero one counting problem */
  radlo = MAX(radlo, 0.0);
  radhi = wstrtod(subRadMaxTW);
  
  bmag = wstrtod(subMagStartTW);
  
  nfit = numCols*numRows*numSecs;
  numRad = numCols / 2;
  nbins  = numRad  / radstep;
  nbins++;  // Zero one counting problem
  
  freeNullFloat(&cntr1);
  freeNullFloat(&cntr2);
  
  cntr1 = (float *)malloc(sizeof(float)*nfit);
  cntr2 = (float *)malloc(sizeof(float)*nfit);
  
  freeNullFloat(&rdens);
  freeNullFloat(&wrdens);
  freeNullFloat(&ccoef);
  freeNullFloat(&wccoef);
  freeNullFloat(&s_wccoef);

  rdens     = (float *)malloc(sizeof(float)*nbins);
  wrdens    = (float *)malloc(sizeof(float)*nbins);
  ccoef     = (float *)malloc(sizeof(float)*nbins);
  wccoef    = (float *)malloc(sizeof(float)*nbins);
  s_wccoef  = (float *)malloc(sizeof(float)*nbins);
  
  tbccoef = 0.0;
  for (i=0; i<magchk; i++)
    if(coefs[i]==bmag)
      tbccoef = coefs[i];
  
  map_std = 1.2;
  map_std = purdueHeader.img_std; // Get float_image_buffer1 STDev
  
  saveNbins = nbins;

  if(subMap2d3d==2) {
    sec1= 1;
    sec2= 1;
    thisBuf1 = (float *)
      (&(float_image_buffer1[(int)((zc-1)*numCols*numRows)]));
    thisBuf2 = (float *)
      (&(subtract_image_buffer1[(int)((zc-1)*numCols*numRows)]));
    zc = 1;
  }
  else {
    sec1= numSecs;
    sec2= sub2binSec;
    thisBuf1 = float_image_buffer1;
    thisBuf2 = subtract_image_buffer1;
  }

  ccomap3cs_(thisBuf1, thisBuf2,
	     &numCols, &numRows, &sec1,
	     &sub2binCol, &sub2binRow, &sec2,
	     &nfit, cntr1, cntr2, &map_std,
	     &xc, &yc, &zc,
	     &radlo, &radhi, &radstep, &bmag,
	     rdens, wrdens, ccoef, wccoef,s_wccoef, &nbins,
	     &diffSecScrollValue, &diffSec2ScrollValue,
	     &icosaQuarter, &subMap2d3d,
	     &threshScrollValue, &threshAbs);
  
  radScaleM = diffSecScrollValue;
  radScaleB = diffSec2ScrollValue;
  XtSetArg(arglst[0], XmNvalue, (int)(diffSecScrollValue*100.0));
  XtSetValues(subDiffSW, arglst, 1);
  sprintf(outStr,"%.3f",diffSecScrollValue);
  TextSet(subDiff1TW, outStr, FALSE); 
  XtSetArg(arglst[0], XmNvalue, (int)(diffSec2ScrollValue*100.0));
  XtSetValues(subDiffS2W, arglst, 1);
  sprintf(outStr,"%.3f",diffSec2ScrollValue);
  TextSet(subDiff2TW, outStr, FALSE); 
  
  sprintf(outStr,"MAP1 = (MAP2 * %.3f) + %.3f", radScaleM, radScaleB);
  doLabel(outStr, subEquationW, FALSE);
  
  nbins = saveNbins;
  
  // Let's let the findmagHist toggle decide this subCanDrawWindow = FALSE;
  diffSec_really_doit();
  DrawWindow();
  
 noRead:;
  
  XtUnmanageChild(statusScreenW);
  
} // end function subtractMap_findRad



int checkPeakFound()
{
  // function name: checkPeakFound
  // Find the peak.

  int i, foundIt = -99;
  float maxFound = -HUGE;

  for (i=0; i<magchk; i++) {
    if(coefs[i] > maxFound) {
      foundIt = i;
      maxFound = coefs[i];
    }
  }
  
  if(foundIt==0 || foundIt==(magchk-1))
    return(FALSE);
  else 
    return(TRUE);
} // end function checkPeakFound



void drawMinMaxRadius()
{
  // function name: drawMinMaxRadius
  // Actually draw the min/max radii on the display.
  
  int i, xBegin, yBegin, arcRadius;

  if(!showRadii) return;

  for(i=0; i<=(2*numCols*sizeFactor) ; i+=(numCols*sizeFactor)) {
    xBegin = i + (numCols * sizeFactor)/2;
    yBegin = (numRows * sizeFactor)/2;
    
    arcRadius = wstrtod(subRadMinTW) * sizeFactor;
    XDrawArc(dsply, XtWindow(drawingW), image_gcRED,
	     xBegin-arcRadius, yBegin-arcRadius,
	     arcRadius*2, arcRadius*2, 0, 360*64);
    arcRadius = wstrtod(subRadMaxTW) * sizeFactor;
    XDrawArc(dsply, XtWindow(drawingW), image_gcBLUE,
	     xBegin-arcRadius, yBegin-arcRadius,
	     arcRadius*2, arcRadius*2, 0, 360*64);
  }
  
} // end function drawMinMaxRadius



void subtractMap_cc(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: subtractMap_cc 
  // Use cc rtns to find relative mag for the maps.

  int iix=1, iiy=1, iiz=1, ichc=2;
  float bmag=1.0, xc, yc, zc, radlo, radhi, magini, magstep, std;
  
  subtractingMap = FALSE;
  
  if(!subtract_image_buffer1) { // Must read first
    askThem("Must Read second map file first.", XmDIALOG_INFORMATION);
    XtUnmanageChild(statusScreenW);
    return;
  }
  
  raiseIt(statusScreenW);
  
  xc = numCols/2.0;
  yc = numRows/2.0;
  zc = numSecs/2.0;
  radstep = 1.0;
  radlo = wstrtod(subRadMinTW);
  radlo -= 1.0;  // To make it consistent with cc_3d()
  radlo = MAX(radlo, 0.0);
  radhi = wstrtod(subRadMaxTW);
  
  magini  = wstrtod(subMagStartTW);
  magstep = wstrtod(subMagStepTW);
  saveMagchk = magchk = wstrtod(subMagNumTW);
  
  sprintf(outStr,"Looking around mag==[%f] ... Please wait ... ",magini);
  checkStatus(outStr);
  
  freeNullFloat(&mags);
  freeNullFloat(&coefs);
  
  mags  = (float *)malloc(sizeof(float)*magchk);
  coefs = (float *)malloc(sizeof(float)*magchk);
  
  cc_3d__(float_image_buffer1, subtract_image_buffer1,
	 &numCols, &numRows, &numSecs,
	 &sub2binCol, &sub2binRow, &sub2binSec,
	 &ichc, &std, &xc, &yc, &zc,
	 &magini, &magstep, &magchk,
	 &radlo, &radhi, &radstep, &bmag,
	 &iix, &iiy, &iiz,
	 mags,coefs,
	 &subMap2d3d, &icosaQuarter);
  
  magchk = saveMagchk; // It's corrupted above RSSLOOKHERE
  magSecScrollValue  = saveBmag = bmag;
  TextSetFloat(subMagStartTW, bmag, FALSE); 
  TextSetFloat(subMagsecTW, bmag, FALSE); 
  XtSetArg(arglst[0], XmNvalue, (int)(magSecScrollValue*1000.0));
  XtSetValues(subMagScrollW, arglst, 1);

  if(checkPeakFound()) // Step 2 times finer
      TextSetFloat(subMagStepTW, magstep*0.5, FALSE); 
  else                 // Stay here and step sideways
    askThem("Peak not found.  Step Size not changed.", XmDIALOG_INFORMATION);

  diffSec_really_doit();
  // Let's let the findMagHist toggle decide this subCanDrawWindow = FALSE;
  DrawWindow();
  
  XtUnmanageChild(statusScreenW);

} // end function subtractMap_cc



void subtractMap_doit(float *thisBuffer, float *tmpSubtract_image_buffer2)
{
  // function name: subtractMap_doit
  // Actually do the subtracting of maps

  int i;
  float xc, yc, zc, mag;

  mag = wstrtod(subMagStartTW);
  sprintf(outStr,"Adjusting for mag==[%f] ... ",mag);
  checkStatus(outStr);
  
  xc = sub2binCol/2.0;
  yc = sub2binRow/2.0;
  zc = sub2binSec/2.0;
  
  mag_write__(subtract_image_buffer1,
	     &sub2binCol, &sub2binRow, &sub2binSec,
	     &xc, &yc, &zc, &mag,
	     tmpSubtract_image_buffer2);

  checkStatus("Subtracting MAPs after adjusting density... ");
  
  for (i=0; i<sub2binSec*sub2binRow*sub2binCol; i++) {
    // Don't forget to apply y=mx+b to second map!!!
    tmpSubtract_image_buffer2[i] = 
      tmpSubtract_image_buffer2[i]*diffSecScrollValue + diffSec2ScrollValue;
    
    thisBuffer[i] = thisBuffer[i] - tmpSubtract_image_buffer2[i];
  }
  
} // end function subtractMap_doit



void subtractMap_work(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: subtractMap_work
  // The real work of MAP subtracting happens here.

  float *tmpSubtract_image_buffer2;
  int zero=0;

  if(diffSecScrollValue == 0.0 && diffSec2ScrollValue== 0.0) {
    askThem("You must finish the 'Find Radius' step first.",
	    XmDIALOG_INFORMATION);
    XtUnmanageChild(statusScreenW);
    return;
  }

  subtractingMap = FALSE;

  raiseIt(statusScreenW);

  // Call mag_write here for appropriate buffer and then subtract them
  CK(tmpSubtract_image_buffer2 = (float *)
     malloc(sizeof(float)*sub2binCol*sub2binRow*sub2binSec));
  
  // Really subtract the buffers
  subtractMap_doit(float_image_buffer1, tmpSubtract_image_buffer2);
  
  if(saveSecondMap) {
    forPifHead_t thisHead;
    fString_t rWrite,forName;
    
    // ----- Writing scaled second map -----
    pif_init_head(&thisHead);
    thisHead.luNum = 4;
    rWrite.addr = malloc(sizeof(char)*4);
    rWrite.length = 4;
    forName.addr = malloc(sizeof(char)*64);
    forName.length = 64;
    sprintf(rWrite.addr,"w");
    sprintf(forName.addr,"secondMap.pif");
    
    pif_open(forName.addr, &thisHead.luNum, rWrite.addr);
    
    // Pick appropriate statistics calculation based
    // on data type we have
    thisHead.datai4_min = 0;
    thisHead.datai4_max = 255;
    thisHead.numCols = sub2binCol;
    thisHead.numRows = sub2binRow;
    thisHead.numSecs = sub2binSec;
    
    thisHead.mode = 20;
    
    calcMapStats(tmpSubtract_image_buffer2,
		 &thisHead.img_bkg, 
		 &thisHead.datar4_min, &thisHead.datar4_max, 
		 &thisHead.datar4_avg, &thisHead.datar4_std,
		 &sub2binCol, &sub2binRow, &sub2binSec);

    thisHead.datai4_max = thisHead.datar4_max;
    thisHead.datai4_min = thisHead.datar4_min;
    
    calcScaleFactor(&thisHead.floatScaleFactor,		     
		    &thisHead.recipFloatScaleFactor, 
		    &thisHead.int4scaleF, 
		    (thisHead.datar4_min-1.0),
		    (thisHead.datar4_max+1.0),
		    thisHead.datar4_avg,thisHead.datar4_std,
		    MAP);
    
    pif_write_gh(&thisHead);
    
    thisHead.imgNum = 1;
    
    pif_write_dh(&thisHead);
    thisHead.imgOffset += 512;  // Just wrote the header
    
    tsbworld_float(tmpSubtract_image_buffer2,
		   subtract_image_buffer1,
		   sub2binCol, sub2binRow, sub2binSec);

    pif_write_map_sections(&thisHead, subtract_image_buffer1, 
			   &(thisHead.mode),
			   &sub2binCol, &sub2binRow, &sub2binSec,
			   &zero, &sub2binSec);
    free(forName.addr);
    free(rWrite.addr);
    pif_close(&thisHead.luNum);
    
    sprintf(outStr,"Second Map written to file 'secondMap.pif'");
    askThem(outStr, XmDIALOG_INFORMATION);    

    // ----- Writing difference map -----
    pif_init_head(&thisHead);
    thisHead.luNum = 4;
    rWrite.addr = malloc(sizeof(char)*4);
    rWrite.length = 4;
    forName.addr = malloc(sizeof(char)*64);
    forName.length = 64;
    sprintf(rWrite.addr,"w");
    sprintf(forName.addr,"diffMap.pif");
    
    pif_open(forName.addr, &thisHead.luNum, rWrite.addr);
    
    // Pick appropriate statistics calculation based
    // on data type we have
    thisHead.datai4_min = 0;
    thisHead.datai4_max = 255;
    thisHead.numCols = sub2binCol;
    thisHead.numRows = sub2binRow;
    thisHead.numSecs = sub2binSec;
    
    thisHead.mode = 20;
    
    calcMapStats(float_image_buffer1,
		 &thisHead.img_bkg, 
		 &thisHead.datar4_min, &thisHead.datar4_max, 
		 &thisHead.datar4_avg, &thisHead.datar4_std,
		 &sub2binCol, &sub2binRow, &sub2binSec);

    thisHead.datai4_max = thisHead.datar4_max;
    thisHead.datai4_min = thisHead.datar4_min;
    
    calcScaleFactor(&thisHead.floatScaleFactor,		     
		    &thisHead.recipFloatScaleFactor, 
		    &thisHead.int4scaleF, 
		    (thisHead.datar4_min-1.0),
		    (thisHead.datar4_max+1.0),
		    thisHead.datar4_avg,thisHead.datar4_std,
		    MAP);
    
    pif_write_gh(&thisHead);
    
    thisHead.imgNum = 1;
    
    pif_write_dh(&thisHead);
    thisHead.imgOffset += 512;  // Just wrote the header
    
    tsbworld_float(float_image_buffer1,
		   subtract_image_buffer1,
		   sub2binCol, sub2binRow, sub2binSec);

    pif_write_map_sections(&thisHead, subtract_image_buffer1, 
			   &(thisHead.mode),
			   &sub2binCol, &sub2binRow, &sub2binSec, 
			   &zero, &sub2binSec);

    free(forName.addr);
    free(rWrite.addr);
    pif_close(&thisHead.luNum);
    
    sprintf(outStr,"Difference Map written to file 'diffMap.pif'");
    askThem(outStr, XmDIALOG_INFORMATION);    
  }

  free(tmpSubtract_image_buffer2);
  
  // Write stats
  calcMapStats(float_image_buffer1,
	       &purdueHeader.img_bkg, 
	       &purdueHeader.map_min, &purdueHeader.map_max, 
	       &purdueHeader.img_avg, &purdueHeader.img_std,
	       &numCols, &numRows, &origNumSecs);
  secMaxPixel = maxMapPixel = purdueHeader.map_max;
  secMinPixel = minMapPixel = purdueHeader.map_min;
  if(secMaxPixel != secMinPixel)
    scaleMapFactor= (tryColors - 7.0) / (secMaxPixel-secMinPixel);
  else {
    scaleMapFactor= 1.0;
    maxMapPixel = minMapPixel + 1.0;
    secMaxPixel = secMinPixel + 1.0;
  }
  
  if(dispSections) {
    // Update the sections - if shaded views then they have to recalc
    // Now stuff the display buffer from our new FIXed map
    stuffMapImage(map_image_buffer, float_image_buffer1, 
		  file_width, file_height, nsecs, imgNum, 0);
    DrawWindow();
  }
  else {
    // If we're already doing rendering then setup the scrolls appropriately
    mapShadeScreen_activate(NULL, NULL, NULL);
  }
  
  subtractingMap = FALSE;
  subCanDrawWindow  = TRUE; 
  drawingFindRad = FALSE;
  XtUnmanageChild(subScreenW);
  freeNullFloat(&subtract_image_buffer2);
  freeNullFloat(&subtract_image_buffer1);
  freeNullFloat(&diffSection_buffer);
  freeNullFloat(&subMapYmxb_buffer);

  freeNullFloat(&mags);
  freeNullFloat(&coefs);
  freeNullFloat(&cntr1);
  freeNullFloat(&cntr2);
  freeNullFloat(&rdens);
  freeNullFloat(&wrdens);
  freeNullFloat(&ccoef);
  freeNullFloat(&wccoef);
  freeNullFloat(&s_wccoef);
  
  XtUnmanageChild(statusScreenW);
  
} // end function subtractMap_work



void subtractMap_diffSection(Widget widget, XtPointer client_data, 
			     XtPointer call_data)
{
  // function name: subtractMap_diffSection
  // Toggle whether to calc a difference section or not.

  if(!subtract_image_buffer1) { // Must read first
    askThem("Must Read second map file first.", XmDIALOG_INFORMATION);
    XtVaSetValues(widget, XmNset, FALSE, NULL);
  }
  else {
    diffSection = !diffSection;
    if(diffSection)
      XtSetArg(arglst[0], XmNsensitive, TRUE );
    else 
      XtSetArg(arglst[0], XmNsensitive, FALSE );
    XtSetValues(subFlickTypeW, arglst, 1); 
    XtSetValues(subDiffSW, arglst, 1); 
    XtSetValues(subDiff1TW, arglst, 1); 
    XtSetValues(subDiffS2W, arglst, 1); 
    XtSetValues(subDiff2TW, arglst, 1); 
    XtSetValues(subMagScrollW, arglst, 1); 
    XtSetValues(subMagsecTW, arglst, 1); 
    XtSetValues(subThreshSW, arglst, 1); 
    XtSetValues(subThreshTW, arglst, 1); 
    if(diffSection)
      diffSec_really_doit();
  }
  
} // end function subtractMap_diffSection



void diffSec_really_doit()
{
  // function name: diffSec_really_doit
  // Do what it takes to get a difference section.

  int i, j, secOut, rowOut, colOut, thisNumSecs;
  float *thisSection1, *thisSection2;
  float maxMapP, minMapP;
  float *tmpSection;
  float xc, yc, zc, xcen, ycen, xSq, ySq, distnce;
  float radlo, radhi, radloSq, radhiSq;

  // Let's let them look at it on the screen
  // Let's let the findmaghist toggle decide this subCanDrawWindow = TRUE;

  if(!subMapYmxb_buffer)
    CK(subMapYmxb_buffer= (float *)malloc(sizeof(float)*numCols*numRows));
  if(!diffSection_buffer)
    CK(diffSection_buffer= (float *)malloc(sizeof(float)*numCols*numRows));
  CK(tmpSection= (float *)malloc(sizeof(float)*numCols*numRows));
  
  secOut = (imgNum-1)*numRows*numCols;

  if(shadedMaps && !dispSections) { // surface shaded
    thisSection1 = map_image_buffer;
    thisSection2 = subtract_map_image_buffer;
  }
  else { // sections
    thisSection1 = float_image_buffer1;
    thisSection2 = subtract_image_buffer1;
  }

  if(!thisSection2) {
      free(tmpSection);
      return;
  }
  
  radlo = wstrtod(subRadMinTW);
  radhi = wstrtod(subRadMaxTW);
  radhiSq = radhi * radhi;
  radloSq = radlo * radlo;
  
  maxMapP = saveMaxMapPixel;
  minMapP = saveMinMapPixel;

  for (i=0; i<numRows*numCols; i++) {
    if(shadedMaps && !dispSections)
      tmpSection[i] = thisSection2[secOut+i];
    else
      tmpSection[i] = thisSection2[secOut+i]*diffSecScrollValue + 
	diffSec2ScrollValue;

    if(tmpSection[i] > maxMapP)
      maxMapP = tmpSection[i];
    if(tmpSection[i] < minMapP)
      minMapP = tmpSection[i];
  }
  
  // Now adjust the subMapYmxb_buffer for the MAG
  thisNumSecs = 1;
  xc = sub2binCol/2.0;
  yc = sub2binRow/2.0;
  zc = 0.0;
  
  mag_write__(tmpSection, &numCols, &numRows, &thisNumSecs,
	      &xc, &yc, &zc, &magSecScrollValue,
	      subMapYmxb_buffer);
  
  free(tmpSection);
  
  if(shadedMaps && !dispSections) // Shaded maps
    goto jumpOverCalc;
  
  // Do the diff
  ycen = numRows/2;
  xcen = numCols/2;

  for (i=0; i<numRows*numCols; i++) {
    // normal
    diffSection_buffer[i] = thisSection1[secOut+i] - subMapYmxb_buffer[i];
    if(diffSection_buffer[i] > maxMapP)
      maxMapP = diffSection_buffer[i];
    if(diffSection_buffer[i] < minMapP)
      minMapP = diffSection_buffer[i];
    
    // Make sure we get min max for the corrected image 2
    if(subMapYmxb_buffer[i] > maxMapP)
      maxMapP = subMapYmxb_buffer[i];
    if(subMapYmxb_buffer[i] < minMapP)
      minMapP = subMapYmxb_buffer[i]; 
  }
  
  // Display thresholding if appropriate
  if(threshDoit) {
    for (i=0; i<numRows; i++) {
      rowOut = i * numCols;
      xSq = (i-ycen)*(i-ycen);
      for (j=0; j<numCols; j++) {
	colOut = rowOut + j;
	// Now threshold if appropriate - color in Yellow
	if(threshAbs) { // Do for ABS
	  ySq= (j-xcen)*(j-xcen);
	  distnce = (float)(xSq+ySq); 
	  if(distnce < radhiSq &&  distnce > radloSq) {
	    if(diffSection_buffer[colOut] < -fabs(threshScrollValue))
	      diffSection_buffer[colOut] = -1234567.0;  // Green
	    else if(diffSection_buffer[colOut] > fabs(threshScrollValue))
	      diffSection_buffer[colOut] = -1234568.0;  // Red
	  }
	}
	else { // normal 
	  if(diffSection_buffer[colOut] < threshScrollValue) {
	    // Candidate for thresholding */
	    if(distnce < radhiSq && distnce > radloSq) {
	      diffSection_buffer[colOut] = -1234567.0;
	    }/* we've found one */
	  }
	}
      }
    }
  }

  if(shadedMaps && !dispSections)
    {/* Shaded maps */
    }/* Shaded maps */

  if(!shadedMaps && dispSections) { // Only for sections
    if(maxMapP > saveMaxMapPixel)
      maxMapPixel = maxMapP;
    if(minMapP < saveMinMapPixel)
      minMapPixel = minMapP;
    
    sprintf(outStr,"Min: %.2f",minMapPixel);
    doLabel(outStr,subSmMinLW, FALSE);
    sprintf(outStr,"Max: %.2f",maxMapPixel);
    doLabel(outStr,subSmMaxLW, FALSE);
  }
  
 jumpOverCalc:;
  
  stuffMapImage(map_image_buffer, float_image_buffer1, 
		file_width, file_height, 1, imgNum, 0);
  stuffMapImage(subMapYmxb_buffer, subMapYmxb_buffer, 
		file_width, file_height, 1, 1, numCols*sizeFactor);

  if(shadedMaps && !dispSections) { // surface shaded
    stuffMapImage(subtract_map_image_buffer2, subtract_map_image_buffer2, 
		  file_width, file_height, 1,1, 2*numCols*sizeFactor);
  }
  else { // normal
    stuffMapImage(diffSection_buffer, diffSection_buffer, 
		  file_width, file_height, 1, 1, 2*numCols*sizeFactor);
  }
  
  DrawWindow();
  if(subCanDrawWindow)
    drawMinMaxRadius();
    
} // end function diffSec_really_doit



void diffSec_scroll_change(Widget widget, XtPointer client_data, 
			   XtPointer call_data)
{
  // function name: diffSec_scroll_change
  // They're changing the scrolls.

  XmScrollBarCallbackStruct *temp = (XmScrollBarCallbackStruct *) call_data;

  if(widget == subDiffSW) {
    diffSecScrollValue = temp->value / 100.0;
    sprintf(outStr,"%.2f",diffSecScrollValue);
    TextSet(subDiff1TW, outStr, FALSE); 
  }
  else if(widget == subDiffS2W) {
    diffSec2ScrollValue = temp->value / 100.0;
    sprintf(outStr,"%.2f",diffSec2ScrollValue);
    TextSet(subDiff2TW, outStr, FALSE); 
  }
  else if(widget == subMagScrollW) {
    magSecScrollValue = temp->value / 1000.0;
    sprintf(outStr,"%.3f",magSecScrollValue);
    TextSet(subMagsecTW, outStr, FALSE); 
  }
  else if(widget == subThreshSW) {
    threshScrollValue = temp->value;
    sprintf(outStr,"%.1f",threshScrollValue);
    TextSet(subThreshTW, outStr, FALSE); 
    recalc_ymxb_thresh();
  }
  
  diffSec_really_doit();

} // end function diffSec_scroll_change



void diffSec_do(Widget widget, int *tag, 
		XmAnyCallbackStruct *callback_data)
{
  // function name: diffSec_do
  // Decide which widget called us and update the appropriate value.
  
  if(widget == subDiff1TW) {
    diffSecScrollValue = wstrtod(subDiff1TW);
    XtSetArg(arglst[0], XmNvalue, (int)(diffSecScrollValue*100.0));
    XtSetValues(subDiffSW, arglst, 1);
  }
  else if(widget == subDiff2TW) {
    diffSec2ScrollValue = wstrtod(subDiff2TW);
    XtSetArg(arglst[0], XmNvalue, (int)(diffSec2ScrollValue*100.0));
    XtSetValues(subDiffS2W, arglst, 1);
  }
  else if(widget == subMagStartTW) {
    magSecScrollValue = wstrtod(subMagStartTW);
    saveBmag = magSecScrollValue;
    XtSetArg(arglst[0], XmNvalue, (int)(magSecScrollValue*1000.0));
    XtSetValues(subMagScrollW, arglst, 1);
    sprintf(outStr,"%.3f",magSecScrollValue);
    TextSet(subMagsecTW, outStr, FALSE);  
  }
  else if(widget == subMagsecTW) {
    magSecScrollValue = wstrtod(subMagsecTW);
    XtSetArg(arglst[0], XmNvalue, (int)(magSecScrollValue*1000.0));
    XtSetValues(subMagScrollW, arglst, 1);
  }
  else if(widget == subThreshTW) {
    threshScrollValue = wstrtod(subThreshTW);
    XtSetArg(arglst[0], XmNvalue, (int)(threshScrollValue));
    XtSetValues(subThreshSW, arglst, 1);
    recalc_ymxb_thresh();  
  }
  
  diffSec_really_doit();
  
} // end function diffSec_do



void subtractMap_findMag_disp(Widget widget, XtPointer client_data, 
			      XtPointer call_data)
{
  // function name: subtractMap_findMag_disp
  // Find the appropriate magnification.

  // Let's let them look at it on the screen? - toggle it here
  findMagHist_toggle =   !findMagHist_toggle ;
  subCanDrawWindow = (findMagHist_toggle) ? FALSE : TRUE;
  submapFindRad = FALSE; 
  
  // Unset findRad toggle
  findRadHist_toggle = FALSE;
  XtVaSetValues(subRadDW, XmNset, FALSE, NULL); 
  
  DrawWindow();
  if(subCanDrawWindow)
    drawMinMaxRadius();

} // end function subtractMap_findMag_disp



void subtractMap_findRad_disp(Widget widget, XtPointer client_data, 
			      XtPointer call_data)
{
  // function name: subtractMap_findRad_disp
  // Find proper radius.

  // Let's let them look at it on the screen
  findRadHist_toggle = !findRadHist_toggle ;
  subCanDrawWindow = (findRadHist_toggle) ? FALSE : TRUE;  
  submapFindRad  = TRUE;
  
  // Unset findMag toggle
  findMagHist_toggle =   FALSE;
  XtVaSetValues(subMagHistW, XmNset, FALSE, NULL); 
  
  DrawWindow();
  if(subCanDrawWindow)
    drawMinMaxRadius();

} // end function subtractMap_findRad_disp



void subtractMap_flicker_arm(Widget widget, XtPointer client_data, 
			     XtPointer call_data)
{
  // function name: subtractMap_flicker_arm
  // When they press the flicker button call here.

  int i, nflicks;
  int tag = *((int *) client_data);

  if(!subtract_image_buffer1) { // Must read first
    askThem("Must Read second map file first.", XmDIALOG_INFORMATION);
  }
  else {
    if(tag == 2)
      nflicks = 9999;  // Virtually infinite
    else if(tag == 1)
      nflicks = 10;
    else
      nflicks = 1;

    for (i=0; i<nflicks; i++) {
      
      if(tag==2 && XCheckTypedWindowEvent(dsply, XtWindow(subFlickIW),
					  ButtonPress, &event)) {
	XFlush(dsply);
	return; // Turning off infinite flickering
      }
      
      if(diffSection) { // They're doing diffsections
	if(flickerIt == 12) { // Do 1,2
	  stuffMapImage(subMapYmxb_buffer, subMapYmxb_buffer, 
			file_width, file_height, 1, 1, 0);
	  stuffMapImage(subMapYmxb_buffer, subMapYmxb_buffer, 
			file_width,file_height, 1, 1, numCols*sizeFactor);

	  if(shadedMaps && !dispSections)                    // surface shaded
	    stuffMapImage(subtract_map_image_buffer2, 
			  subtract_map_image_buffer2, 
			  file_width,file_height,
			  1, 1, 2*numCols*sizeFactor);
	  else                                               // normal
	    stuffMapImage(diffSection_buffer, diffSection_buffer, 
			  file_width,file_height, 1, 1, 2*numCols*sizeFactor);

	}
	else if(flickerIt == 13) { // Do 1,3
	  if(shadedMaps && !dispSections)                    // surface shaded
	    stuffMapImage(subtract_map_image_buffer2, 
			  subtract_map_image_buffer2, 
			  file_width,file_height, 1, 1, 0);
	  else                                               // normal
	    stuffMapImage(diffSection_buffer, diffSection_buffer, 
			  file_width,file_height, 1, 1, 0);

	  stuffMapImage(subMapYmxb_buffer, subMapYmxb_buffer, 
			file_width, file_height, 1, 1, numCols*sizeFactor);

	  if(shadedMaps && !dispSections)                    // surface shaded
	    stuffMapImage(subtract_map_image_buffer2, 
			  subtract_map_image_buffer2, 
			  file_width, file_height,
			  1, 1, 2*numCols*sizeFactor);
	  else                                               // normal
	    stuffMapImage(diffSection_buffer, diffSection_buffer, 
			  file_width,file_height,
			  1, 1, 2*numCols*sizeFactor);

	}
	else if(flickerIt == 23) { // Do 2,3
	  stuffMapImage(map_image_buffer, float_image_buffer1, 
			file_width, file_height, 1, imgNum, 0);
	  if(shadedMaps && !dispSections) {                  // surface shaded
	    stuffMapImage(subtract_map_image_buffer2, 
			  subtract_map_image_buffer2, 
			  file_width, file_height,
			  1, 1, numCols*sizeFactor);
	    stuffMapImage(subtract_map_image_buffer2, 
			  subtract_map_image_buffer2, 
			  file_width, file_height,
			  1, 1, 2*numCols*sizeFactor);
	  }
	  else {                                             // normal
	    stuffMapImage(diffSection_buffer, diffSection_buffer, 
			  file_width,file_height, 1, 1, numCols*sizeFactor);
	    stuffMapImage(diffSection_buffer, diffSection_buffer, 
			  file_width, file_height, 1, 1, 2*numCols*sizeFactor);
	  }
	}
      }
      else                                                   // sections
	// subtract_map_image_buffer <- shaded views of map2
	stuffMapImage(subtract_map_image_buffer, subtract_image_buffer1, 
		      file_width, file_height, nsecs, imgNum, 0);
      
      DrawWindow();
      if(subCanDrawWindow)
	drawMinMaxRadius();
      
      if(tag==1 || tag==2) { // Got to disarm it also
	em_sleep(flickSpeed);
	subtractMap_flicker_disarm(widget, client_data, call_data);
	em_sleep(flickSpeed);
      }
      
    }
  }
  
} // end function subtractMap_flicker_arm



void subtractMap_flicker_disarm(Widget widget, XtPointer client_data, 
				XtPointer call_data)
{
  // function name: subtractMap_flicker_disarm
  // When they let up on the flicker button call here.

  if(!subtract_image_buffer1) return;
  
  if(diffSection) { // They're doing diffsections
    stuffMapImage(map_image_buffer, float_image_buffer1, 
		  file_width, file_height, 1, imgNum, 0);
    stuffMapImage(subMapYmxb_buffer, subMapYmxb_buffer, 
		  file_width, file_height, 1, 1, numCols*sizeFactor);
    
    if(shadedMaps && !dispSections)                        // surface shaded
      stuffMapImage(subtract_map_image_buffer2, 
		    subtract_map_image_buffer2, 
		    file_width, file_height,
		    1, 1, 2*numCols*sizeFactor);
    else                                                   // normal
      stuffMapImage(diffSection_buffer, diffSection_buffer, 
		    file_width, file_height,
		    1, 1, 2*numCols*sizeFactor);
    
  }
  else                                                     // sections 
    stuffMapImage(map_image_buffer, float_image_buffer1, 
		  file_width, file_height, nsecs, imgNum, 0);
  
  DrawWindow();
  if(subCanDrawWindow)
    drawMinMaxRadius();
  
} // end function subtractMap_flicker_disarm



void subtractMap_activate(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: subtractMap_activate
  // Get the subtract map interface up on the display.
  
  TextSet(subFile1TW, fileName.addr, FALSE);
  TextSet(subFile2TW, "", FALSE);
  
  XtSetArg(arglst[0], XmNcursorPosition, strlen(outStr) );
  XtSetValues(subFile1TW, arglst, 1); 
  XFlush(dsply);
  
  TextSet(subMagStartTW, "1.00", FALSE); 
  TextSet(subMagStepTW, "0.01", FALSE); 
  TextSet(subMagNumTW, "11", FALSE);
  
  TextSet(subRadMinTW, "1", FALSE);
  TextSetInt(subRadMaxTW, numSecs/2, FALSE); 
  
  TextSetInt(imgNumTextW, (pifHeader.numCols/2)+1, FALSE);
  
  raiseIt(subScreenW);
  
  // Draw the flicker speed "thermometer"
  DrawFlickThermometer();
  
} // end function subtractMap_activate



void coravgArrowButton(Widget widget, XEvent *event)
{
  // function name: coravgArrowButton
  // correlation average arrows get different speed depending on which
  // mouse button is pressed.

  float speedFactor, stepVal;

  stepVal = wstrtod(emcorAvgRTW);
  speedFactor = (float) GetButtonVal(event->xbutton.button, 1, 5, 10);

  if(widget==emcorAvgUAW)
    stepVal += speedFactor;
  else 
    stepVal -= speedFactor;
  
  stepVal= MAX(stepVal,1.000000);
  
  coravgRadius = stepVal;
  sprintf(outStr,"%.1f",stepVal);
  TextSet(emcorAvgRTW, outStr, FALSE);
  
  DrawWindow();  // Show new radius
  
} // end function coravgArrowButton



void coravg_scroll_change(Widget widget, XtPointer client_data, 
			  XtPointer call_data)
{
  // function name: coravg_scroll_change
  // Scrollbar changes

  XmScrollBarCallbackStruct *temp = (XmScrollBarCallbackStruct *) call_data;

  coravgLowLimValue = temp->value / 100.0;
  sprintf(outStr,"%.0f",coravgLowLimValue);
  doLabel(outStr, emcorAvgLLVW, FALSE);

} // end function coravg_scroll_change



void emcoravg_text_changed(Widget widget, XtPointer client_data, 
			   XtPointer call_data)
{
  // function name: emcoravg_text_changed
  // Text changed

    stuffEmcoravgImage();
  
} // end function emcoravg_text_changed



void emcoravg_save_activate(Widget widget, XtPointer client_data, 
			    XtPointer call_data)
{
  // function name: emcoravg_save_activate
  // Save to disk in PIF file.

  int i;
  float *tmpBuffer, minF, maxF, scaleF;
  BYTE *tmpByte_buffer;
  FILE *fOut;
  forPifHead_t thisHead;
  fString_t rWrite, fName;  

  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 1;
  fName.addr = malloc(sizeof(char)*244);
  fName.length = strlen(fName.addr);
  
  sprintf(rWrite.addr,"w");         // Write it
  sprintf(fName.addr,"coravg.pif"); 
  pif_init_head(&thisHead);
  thisHead.luNum = 4;
  
  pif_open(fName.addr, &thisHead.luNum, rWrite.addr);

  // [1] Write global file header
  calcStats(emcoravg_image_buffer_avg, &purdueHeader.img_bkg, 
	    &purdueHeader.img_min, &purdueHeader.img_max, 
	    &purdueHeader.img_avg, &purdueHeader.img_std,
	    emcoravgNcol, emcoravgNrow, emcoravgNcol, fileType,
	    bkgrndMethod, boxMethod, packRadius, boxRadius,
	    featherRadius, recipSizeFactor, tryColors);
  
  thisHead.nimg = 1;
  thisHead.numCols = emcoravgNcol;
  thisHead.numRows = emcoravgNrow;
  thisHead.numSecs = 1;
  thisHead.mode = 6;
  
  pif_write_gh(&thisHead);
  
  // [2] Write header for each image
  thisHead.img_bkg    = (int)purdueHeader.img_avg;;
  thisHead.packRadius = emcoravgNcol/2;
  
  thisHead.datai4_max = purdueHeader.img_max;
  thisHead.datar4_max = purdueHeader.img_max;
  thisHead.datai4_min = purdueHeader.img_min;
  thisHead.datar4_min = purdueHeader.img_min;
  thisHead.datar4_avg = purdueHeader.img_avg;
  thisHead.datar4_std = purdueHeader.img_std;
  
  thisHead.binFactor = 1;
  thisHead.imgNum    = 1;
  
  pif_write_dh(&thisHead);
  
  // [3] Now write the image info
  thisHead.imgOffset= 1024;
  CK(tmpBuffer=(float *)malloc(sizeof(float)*
			       emcoravgNcol*emcoravgNrow));
  CK(tmpByte_buffer=(BYTE *)malloc(sizeof(BYTE)*
				   emcoravgNcol*emcoravgNrow));
  tsbworld_float(emcoravg_image_buffer_avg,
		 tmpBuffer, emcoravgNcol, emcoravgNcol, 1);

  farray_minmax(tmpBuffer, &minF, &maxF, thisHead.numCols*thisHead.numRows);
  
  scaleF = 254.0 /(maxF-minF);
  for(i=0;i<(thisHead.numCols*thisHead.numRows);i++)
    tmpByte_buffer[i] = ((tmpBuffer[i]-minF)*scaleF);

  pif_write_byte_image_(&thisHead.luNum,
			tmpByte_buffer, &thisHead.imgOffset,
			&thisHead.numCols,
			&thisHead.numRows, &thisHead.numSecs);
  free(tmpByte_buffer);
  free(tmpBuffer);
  
  pif_close(&thisHead.luNum);
  sprintf(outStr,"File coravg.pif written successfully.");
  askThem(outStr, XmDIALOG_INFORMATION);
  free(rWrite.addr);
  free(fName.addr);
  
  // Now write a raw file
  if ( (fOut = fopen("coravg.raw","w")) != NULL ) {
    fwrite(emcoravg_image_buffer_avg, sizeof(BYTE),
	   (emcoravgNcol*emcoravgNrow),fOut);
    fclose(fOut);
    sprintf(outStr,"File coravg.raw written with dimensions: \n\
    %d cols\n\
    %d rows\n\
    in %d eight bit bytes",emcoravgNcol,emcoravgNrow,
	    emcoravgNcol*emcoravgNrow);
    askThem(outStr, XmDIALOG_INFORMATION); 
  }
  else
    askThem("Can't open coravg.raw file.", XmDIALOG_INFORMATION);
  
} // end function emcoravg_save_activate



void emcoravg_tryit_activate(Widget widget, XtPointer client_data, 
			     XtPointer call_data)
{
  // function name: emcoravg_tryit_activate
  // Do this coravg

  int idim1, idim2, imgNcols, imgNrows, i, j, k, cor_pass;
  int img_num, outWidth, outHeight, yOut;
  int nlim, npeak, lowlim, npeak_unique, ncolmp, nrowmp, arg1;
  int ksign, iform, thisMin, thisMax;
  float *c_buffer, *b_buffer, *array, *cor_map, *avg_map;
  float *cor_img, *avg_img, minF, maxF;
  float cor_avg, cor_std, radius, scaleFactor;
  float *peaks, thisBackground;
  float *fft1, *fft2, *fft1_half, *fft2_half;
  BYTE *cor_byte_buffer;
  purdueHead_t thisHeader;

  if(!emcoravg_image_buffer1) {
    askThem("Must read reference file first.", XmDIALOG_INFORMATION);
    return;
  }
  
  lowlim = coravgLowLimValue;
  
  imgNcols = purdueHeader.numCols;
  imgNrows = purdueHeader.numRows;
  
  fft_setdim_local(imgNcols, imgNrows, &idim1, &idim2);
  
  printf ("idim1,2[%d,%d]\n",idim1,idim2);
  
  CK(fft1 = (float *)malloc(sizeof(float)*idim1*idim2*2));
  CK(fft2 = (float *)malloc(sizeof(float)*idim1*idim2*2));
  CK(fft1_half = (float *)malloc(sizeof(float)*idim1*((idim2/2)+1)*2));
  CK(fft2_half = (float *)malloc(sizeof(float)*idim1*((idim2/2)+1)*2));
  
  CK(cor_byte_buffer = (BYTE *)malloc(sizeof(BYTE)*idim1*idim2));
  CK(c_buffer = (float *)malloc(sizeof(float)*idim1*2));
  CK(b_buffer = (float *)malloc(sizeof(float)*idim2*2));
  CK(array = (float *)malloc(sizeof(float)*2*idim1*((idim2/2)+1)));
  CK(cor_map = (float *)malloc(sizeof(float)*idim1*idim2));
  CK(cor_img = (float *)malloc(sizeof(float)*idim1*idim2));
  
  // Let's try 50000 for now
  nlim = 50000;
  CK(peaks = (float *)malloc(sizeof(float)*3*nlim));

  ksign = 1;       // Do a "forward" FFT 
  iform = 1;       // Data is imaginary 

  if(useAvgCorAvg) { // Use the last average 
    for(i=0; i<emcoravgNrow*emcoravgNcol; i++)
      cor_img[i] = emcoravg_image_buffer_avg[i];
    // We need a new background value 
    calcStats(emcoravg_image_buffer_avg, &thisHeader.img_bkg, 
	      &thisHeader.img_min, &thisHeader.img_max, 
	      &thisHeader.img_avg, &thisHeader.img_std,
	      emcoravgNcol, emcoravgNrow, emcoravgNcol,
	      fileType, bkgrndMethod, boxMethod, packRadius,
	      boxRadius, featherRadius, recipSizeFactor,
	      tryColors);
  }
  else {// Use the one read from disk 
    for(i=0; i<emcoravgNrow*emcoravgNcol; i++)
      cor_img[i] = emcoravg_image_buffer_disk[i];
    
    // We need a new background value 
    calcStats(emcoravg_image_buffer_disk, &thisHeader.img_bkg, 
	      &thisHeader.img_min, &thisHeader.img_max, 
	      &thisHeader.img_avg, &thisHeader.img_std,
	      emcoravgNcol, emcoravgNrow, emcoravgNcol,
		  fileType, bkgrndMethod, boxMethod, packRadius,
		  boxRadius, featherRadius, recipSizeFactor, tryColors);
  }

  emcoravgBkg        = thisHeader.img_bkg; 
  printf("thisBackground = [%f]\n",emcoravgBkg );
  
  thisBackground = emcoravgBkg;
  r4_img_fft_fill__(cor_img, &emcoravgNcol, &emcoravgNrow, 
			   fft1, &idim1, &idim2, &thisBackground);
  
  fft_2d_rob__(fft1, &idim1, &idim2, &ksign, &iform);
  fft_copy1__ (fft1, fft1_half, &idim1, &idim2);
  
  if(imgInCore) { // Do for scanned image 
    for(i=0; i<imgNrows*imgNcols; i++)
	cor_img[i] = inCore_buffer[i];
  }
  else if(fileType==PACKEDPORTABLE) { // Do for packed 
    float *thisImg;
    thisImg  = packImg[0];
    for(i=0; i<imgNrows*imgNcols; i++)
      cor_img[i] = thisImg[i];
  }
  else { // Can't do 
    sprintf(outStr,"Corr Average only for scanned or packed images.");
    askThem(outStr, XmDIALOG_INFORMATION);
    goto noEmcoravg;
  }

  thisBackground = purdueHeader.img_bkg;
  r4_img_fft_fill__(cor_img, &imgNcols, &imgNrows, 
			   fft2, &idim1, &idim2, &thisBackground);
  
  fft_2d_rob__(fft2, &idim1, &idim2, &ksign, &iform);
  fft_copy1__(fft2, fft2_half, &idim1, &idim2);
  
  cor_pass = 0;
  cor_first__(&idim1, &idim2, array, fft1_half, fft2_half, 
	      c_buffer, &cor_pass);
  
  img_num = 1;
  cor_second__(&idim1, &idim2, array, cor_map, 
	      cor_img, b_buffer, 
	      &cor_avg, &cor_std, &thisMin, &thisMax);
  
  // Now put it on the display 
  outWidth  = idim1;
  outHeight = idim2;
  printf("outdims[%d,%d]\n",outWidth,outHeight);
  farray_minmax(cor_img, &minF, &maxF, outHeight*outWidth);
  thisMin = minF; // RSS setting to integer value for consistency
  thisMax = maxF; // with original version of code
  scaleFactor = (tryColors-9.0) / (thisMax-thisMin);
  
  for (j=0; j<outHeight; j++) {
    int thisPixel;
    yOut = j*outWidth;
    for (k=0; k<outWidth; k++) {
      thisPixel = (cor_img[yOut+k] - thisMin)*scaleFactor + 8.0;
      
      if(showLowLim && (thisPixel < lowlim))
	XPutPixel(image_image1, idim1+k, j, pixels_return[RED]);
      else
	XPutPixel(image_image1, idim1+k, j, pixels_return[thisPixel]);
    }
  }
  
  XPutImage(dsply, image_window1, image_gc, image_image1, 
	    0, 0, 0, 0,  outWidth*3, outHeight);
  XFlush(dsply);
  
  coravg_lim__(&nlim, peaks, &npeak, cor_map, &idim1,
	      &idim2, &lowlim);
  
  printf("npeak[%d]\n",npeak);
  
  if(npeak <= 0) {
      sprintf(outStr,"No peaks found");
      askThem(outStr, XmDIALOG_INFORMATION);
      goto allDone;
  }
  
  radius = wstrtod(emcorAvgRTW);
  printf("radius=[%f]\n",radius);
  
  arg1 = 1;
  peak_chk__(peaks, &npeak, &npeak_unique, &radius, &arg1);
  
  peak_sort_hi__(peaks, &npeak);
  peak_cen__(peaks, &npeak_unique, cor_map, &idim1, &idim2, &arg1);
  
  ncolmp = emcoravgNcol;
  nrowmp = emcoravgNrow;
  CK(avg_map= (float *)malloc(sizeof(float)*ncolmp*nrowmp));
  CK(avg_img= (float *)malloc(sizeof(float)*ncolmp*nrowmp));

  for(i=0; i<ncolmp*nrowmp; i++)
    avg_map[i] = 0.0;
  
  if(imgInCore) { // Do for scanned image 
    for(i=0; i<imgNrows*imgNcols; i++)
      cor_img[i] = inCore_buffer[i];
  }
  else if(fileType==PACKEDPORTABLE) { // Do for packed 
    float *thisImg; // RSSLOOKHERE
    thisImg  = packImg[0];
    for(i=0; i<imgNrows*imgNcols; i++)
      cor_img[i] = thisImg[i];
  }
  else {
    sprintf(outStr,"Corr Average only for scanned or packed images.");
    askThem(outStr, XmDIALOG_INFORMATION);
    goto noEmcoravg;
  }
  
  coravg_sum__(cor_img, &imgNcols, &imgNrows, avg_img, avg_map, &ncolmp,
	      &nrowmp, peaks, &npeak_unique);
  
  // Remember average image 
  for(i=0; i<nrowmp*ncolmp; i++)
    emcoravg_image_buffer_avg[i] = avg_img[i];
  
  // Stuff display
  if(useAvgCorAvg)
    for(i=0; i<nrowmp*ncolmp; i++)
      emcoravg_image_buffer1[i] = avg_img[i];
  else
    for (i=0; i<nrowmp*ncolmp; i++)
      emcoravg_image_buffer1[i] = emcoravg_image_buffer_disk[i];
  
  stuffEmcoravgImage();
  
 noEmcoravg:;
  
  free (avg_map);
  free (avg_img);
  
 allDone:;
  
  free(fft1);
  free(fft2);
  free(fft1_half);
  free(fft2_half);
  free(cor_byte_buffer);
  free(c_buffer);
  free(b_buffer);
  free(array);
  free(cor_map);
  free(cor_img);
  free(peaks);
  
} // end function emcoravg_tryit_activate



void DrawEmcoravgWindow(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: DrawEmcoravgWindow
  // Redraw window e.g. on expose events or whenever it's prudent.

  if(XtIsManaged(emcorAvgScreenW))
    XPutImage(dsply, emcoravg_window1, emcoravg_gc, emcoravg_image1, 0, 0,
	      0, 0, 256, 256);
  
  XFlush (dsply);
  
} // end function DrawEmcoravgWindow



void stuffEmcoravgImage()
{
  // function name: stuffEmcoravgImage
  // Put coravg image on display.

  int j, k, l, m, zoomIt;
  float scaleFactor, minF, maxF;

  if(emcoravg_image_buffer1) {
    zoomIt = wstrtod(emcorAvgZTW);
    zoomIt = MAX(zoomIt,1);
    zoomIt = MIN(zoomIt,10);
    printf("zoomit[%d]\n",zoomIt);
    
    for (j=0; j<256; j++)
      for (k=0; k<256; k++)
	XPutPixel(emcoravg_image1, k, j, pixels_return[BLACK]);
    
    farray_minmax(emcoravg_image_buffer1, &minF, &maxF, 
		  emcoravgNrow*emcoravgNcol);
    scaleFactor = (tryColors-9.0) / (maxF-minF);
    
    for(j=0;j<emcoravgNrow;j++) {
      for(k=0;k<emcoravgNcol;k++) {
	if(((j*zoomIt)+(zoomIt-1) < 256) && ((k*zoomIt)+(zoomIt-1) < 256)) {
	  for(l=0;l<zoomIt;l++) {
	    for(m=0;m<zoomIt;m++) {
	      // emcoravg_buffer1[((((j*zoomIt)+l)*256)+
	      // (k*zoomIt)+m)] = 
	      // pixels_return[
	      // (int)((((float)emcoravg_image_buffer1[(
	      // (j*emcoravgNcol)+k)]-
	      // minF)*
	      // scaleFactor)+8.0)
	      // ];
	      XPutPixel(emcoravg_image1, 
			(k*zoomIt)+m, (j*zoomIt)+l,
			pixels_return
			[(int)(((emcoravg_image_buffer1
				 [((j*emcoravgNcol)+k)]-
				 minF)*scaleFactor)+8.0)]);
	    }
	  }
	}
      }
    }
    
    XPutImage(dsply, emcoravg_window1, emcoravg_gc, emcoravg_image1, 
	      0, 0, 0, 0, 256, 256);
    XFlush(dsply);  
  }
  
} // end function stuffEmcoravgImage



void read_emcoravg(char *fName)
{
  // function name: read_emcoravg
  // Read the emcoravg file.

  int i;
  off_t imgOffset;
  BYTE *tmpBuffer;
  forPifHead_t thisHeader;
  fString_t rWrite, inFileName;

  readEmcoravg = FALSE;
  
  i= 4;
  thisHeader.luNum = i;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 1;
  inFileName.addr = malloc(sizeof(char)*64);
  inFileName.length = 64;
  sprintf(rWrite.addr,"r");
  
  sprintf(inFileName.addr,"%s",fName);
  
  pif_open(inFileName.addr, &i, rWrite.addr);
  
  
  pif_read_gh(&thisHeader);
  pifMode = thisHeader.mode;
  
  if(pifMode!=0 && pifMode!=6) {
    sprintf(outStr,"Invalid PIF mode [%d]\n",pifMode);
    askThem(outStr, XmDIALOG_INFORMATION);
    return;
  }

  // For fortran
  thisHeader.imgNum = 1;
  pif_read_dh(&thisHeader);
  
  imgOffset          = thisHeader.imgOffset; 
  pifMode            = thisHeader.mode; 
  emcoravgNcol       = thisHeader.numCols; 
  emcoravgNrow       = thisHeader.numRows; 
  emcoravgBkg        = thisHeader.img_bkg; 
  
  freeNullFloat(&emcoravg_image_buffer1);
  freeNullFloat(&emcoravg_image_buffer_disk);
  freeNullFloat(&emcoravg_image_buffer_avg);

  CK(tmpBuffer = (BYTE *)malloc(sizeof(BYTE)*
				emcoravgNcol*emcoravgNrow));
  CK(emcoravg_image_buffer1 = (float *)malloc(sizeof(float)*
					      emcoravgNcol*emcoravgNrow));
  CK(emcoravg_image_buffer_disk = (float *)malloc(sizeof(float)*
						  emcoravgNcol*emcoravgNrow));
  CK(emcoravg_image_buffer_avg = (float *)malloc(sizeof(float)*
						 emcoravgNcol*emcoravgNrow));
  
  pif_read_byte_image(&thisHeader.luNum,
		      tmpBuffer, &imgOffset,
		      &thisHeader.numCols, 
		      &thisHeader.numRows,
		      &thisHeader.numSecs);
  
  pif_close(&i);

  for (i=0; i<thisHeader.numRows*thisHeader.numCols; i++)
    emcoravg_image_buffer1[i] = (float)tmpBuffer[i];

  free(tmpBuffer);
  
  if(pifMode == 6) {
    float *tmpByte_buffer;
    CK(tmpByte_buffer=(float *)malloc(sizeof(float)*
				      thisHeader.numCols*
				      thisHeader.numRows));
    
    tsbworld_float(emcoravg_image_buffer1, tmpByte_buffer,
		   thisHeader.numCols, thisHeader.numCols, 1);

    for (i=0; i<thisHeader.numRows*thisHeader.numCols; i++)
      // Just stuff the avg for now
      emcoravg_image_buffer_avg[i]= 
	emcoravg_image_buffer_disk[i]= 
	emcoravg_image_buffer1[i]= 
	tmpByte_buffer[i]; 
    
    free(tmpByte_buffer);  
  }
  
  if(emcoravgNcol > 256 || emcoravgNrow > 256) {
    sprintf(outStr,"Dimensions too large [%d,%d]\n",
	    emcoravgNcol,emcoravgNrow);
    askThem(outStr, XmDIALOG_INFORMATION);
    return;
  }
  
  stuffEmcoravgImage();
  
} // end function read_emcoravg



void apNewOrig_do_activate(Widget widget, XtPointer client_data, 
			   XtPointer call_data)
{
  // function name: apNewOrig_do_activate
  // They've set a new origin.

  int fft_full=1;
  float x_shift, y_shift;

  if(widget == apOrigResetW) { // reset orig
    x_shift = 0.0 - apOrigXtot;
    y_shift = 0.0 - apOrigYtot;
    apOrigXtot = apOrigYtot = 0.0;
  }
  else { // normal
    x_shift = wstrtod(fftnewXtextW);
    y_shift = wstrtod(fftnewYtextW);
    apOrigXtot += x_shift;
    apOrigYtot += y_shift;
  }
  
  sprintf(outStr,"Tot X shift[%f]",apOrigXtot);
  doLabel(outStr, apOrigLabel1W, DEFAULT);
  sprintf(outStr,"Tot Y shift[%f]",apOrigYtot);
  doLabel(outStr, apOrigLabel2W, DEFAULT);
  
  fft_pshift__(fftImage_gold, &fftSize, &fftSize, 
	      &x_shift, &y_shift, &fft_full);
  
  calcFFT(FALSE, 0, FALSE);
  
} // end function apNewOrig_do_activate



void dio_change(Widget widget, XtPointer client_data, 
		XtPointer call_data)
{
  // function name: dio_change
  // Change icosahedral projection options

  int tag = *((int *) client_data);

  if(tag == DIODESTROY)
    askThem("Destructive not yet.", XmDIALOG_INFORMATION);
  else 
    whichIcos = tag;
  
} // end function dio_change



void imgconv_tryit_activate(Widget widget, XtPointer client_data, 
			    XtPointer call_data)
{
  // function name: imgconv_tryit_activate

  int idim1, idim2, imgNcols, imgNrows, i, j, k, cor_pass;
  int img_num, outWidth, outHeight, yOut;
  int nlim, lowlim, ksign, iform;
  int zero=0, one=1;
  float fsf,rfsf,ibkg,imin,imax,iavg,istd;
  float *fft1, *fft2, *fft1_half, *fft2_half;
  float *c_buffer, *b_buffer, *array, *cor_map;
  float *cor_img;
  float cor_avg, cor_std, scaleFactor;
  float *peaks;
  float thisBackground, thisMin, thisMax;
  BYTE *cor_byte_buffer;
  purdueHead_t thisHeader;
  forPifHead_t outHead;
  fString_t rWrite,inFileName;

  if(!imgconv_image_buffer1) {
    askThem("Must read reference file first.", XmDIALOG_INFORMATION);
    return;
  }
  
  lowlim = coravgLowLimValue;

  imgNcols = purdueHeader.numCols;
  imgNrows = purdueHeader.numRows;
  fft_setdim_local(imgNcols, imgNrows, &idim1, &idim2);

  printf("Chosen idim1[%d]  idim1[%d]\n",idim1,idim2);
  
  CK(fft1 = (float *)malloc(sizeof(float)*idim1*idim2*2));
  CK(fft2 = (float *)malloc(sizeof(float)*idim1*idim2*2));
  CK(fft1_half = (float *)malloc(sizeof(float)*idim1*((idim2/2)+1)*2));
  CK(fft2_half = (float *)malloc(sizeof(float)*idim1*((idim2/2)+1)*2));
  
  CK(cor_byte_buffer = (BYTE *)malloc(sizeof(BYTE)*idim1*idim2));
  CK(c_buffer = (float *)malloc(sizeof(float)*idim1*2));
  CK(b_buffer = (float *)malloc(sizeof(float)*idim2*2));
  CK(array = (float *)malloc(sizeof(float)*2*idim1*((idim2/2)+1)));
  CK(cor_map = (float *)malloc(sizeof(float)*idim1*idim2));
  CK(cor_img = (float *)malloc(sizeof(float)*idim1*idim2));
  
  // Let's try 50000 for now
  nlim = 50000;
  CK(peaks = (float *)malloc(sizeof(float)*3*nlim));
  
  ksign = 1;       // Do a "forward" FFT 
  iform = 1;       // Data is imaginary 
  
  for (i=0; i<imgconvNrow*imgconvNcol; i++)
    cor_img[i] = imgconv_image_buffer_disk[i];
  
  // We need a new background value 
  calcStats(imgconv_image_buffer_disk, &thisHeader.img_bkg, 
	    &thisHeader.img_min, &thisHeader.img_max, 
	    &thisHeader.img_avg, &thisHeader.img_std,
	    imgconvNcol, imgconvNrow, imgconvNcol,
		fileType, bkgrndMethod, boxMethod,
		packRadius, boxRadius, featherRadius,
		recipSizeFactor, tryColors);
  
  imgconvBkg        = thisHeader.img_bkg; 
  printf("thisBackground = [%f]\n",imgconvBkg );
  
  thisBackground = imgconvBkg;
  r4_img_fft_fill__(cor_img, &imgconvNcol, &imgconvNrow, 
			   fft1, &idim1, &idim2, &thisBackground);
  fft_2d_rob__(fft1, &idim1, &idim2, &ksign, &iform);
  fft_copy1__ (fft1, fft1_half, &idim1, &idim2);

  if(imgInCore) { // Do for scanned image 
    for (i=0; i<imgNrows*imgNcols; i++)
      cor_img[i] = inCore_buffer[i];
  }
  else if(fileType==PACKEDPORTABLE) { // Do for packed 
    float *thisImg;
    thisImg  = packImg[0];
    for (i=0; i<imgNrows*imgNcols; i++)
      cor_img[i] = thisImg[i];
  }
  else {
    sprintf(outStr,"Corr Average only for scanned or packed images.");
    askThem(outStr, XmDIALOG_INFORMATION);
    goto noImgconv;
  }
  
  thisBackground = purdueHeader.img_bkg;
  r4_img_fft_fill__(cor_img, &imgNcols, &imgNrows, 
			   fft2, &idim1, &idim2, &thisBackground);
  fft_2d_rob__(fft2, &idim1, &idim2, &ksign, &iform);
  fft_copy1__ (fft2, fft2_half, &idim1, &idim2);
  
  cor_pass = 0;
  cor_first__(&idim1, &idim2, array, fft1_half, fft2_half, 
	      c_buffer, &cor_pass);
  
  img_num = 1;
  cor_second__(&idim1, &idim2, array, cor_map, 
	      cor_img, b_buffer,
	      &cor_avg, &cor_std, &thisMin, &thisMax);
  
  rearrange_fft_style__(&idim1, &idim2, cor_map);
  
  // Now put it on the display 
  outWidth  = idim1;
  outHeight = idim2;
  farray_minmax(cor_map, &thisMin, &thisMax, outHeight*outWidth);
  scaleFactor = (tryColors-9.0) / (thisMax-thisMin);
  
  pif_init_head(&outHead);
  i= 9;
  outHead.luNum = i;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 1;
  inFileName.addr = malloc(sizeof(char)*64);
  inFileName.length = 64;
  
  outHead.mode = 46;
  outHead.numCols = outWidth;
  outHead.numRows = outHeight;
  outHead.nimg = 1;
  
  calcMapStats(cor_map, &ibkg, &imin, &imax, &iavg, &istd,
	       &outWidth, &outHeight, &outHead.nimg);
  calcScaleFactor(&fsf, &rfsf, &int4scaleF, 
		  (imin-1.0), (imax+1.0), iavg, istd,
		  PACKEDPORTABLE);
  outHead.floatScaleFactor = fsf;
  outHead.recipFloatScaleFactor = rfsf;
  outHead.int4scaleF = int4scaleF;
  
  sprintf(rWrite.addr,"w");
  
  sprintf(inFileName.addr,"cor_map.pif");
  pif_open(inFileName.addr, &i, rWrite.addr);
  
  pif_write_gh(&outHead);
  pif_write_dh(&outHead);
  pif_write_map_sections(&outHead, cor_map,
			 &(outHead.mode), &(outHead.numCols), 
			 &(outHead.numRows), &(outHead.numSecs), 
			 &zero, &one);
  
  sprintf(outStr,"Debugging file cor_map.pif written.");
  askThem(outStr, XmDIALOG_INFORMATION);
  
  for (j=0; j<outHeight; j++) {
    int thisPixel;
    yOut = j*outWidth;
    for (k=0; k<outWidth; k++) {
      thisPixel = (cor_map[yOut+k] - thisMin)*scaleFactor + 9.0;
      XPutPixel(image_image1, 1+k+imgNcols, j, pixels_return[thisPixel]);
    }
  }
  
  XPutImage(dsply, image_window1, image_gc, image_image1, 
	    0, 0, 0, 0,  outWidth*2, outHeight*2);
  XFlush(dsply);
  
  stuffImgconvImage();
  
 noImgconv:;
  
  free(fft1);
  free(fft2);
  free(fft1_half);
  free(fft2_half);
  free(cor_byte_buffer);
  free(c_buffer);
  free(b_buffer);
  free(array);
  free(cor_map);
  free(cor_img);
  free(peaks);
  
} // end function imgconv_tryit_activate



void read_imgconv(char *fName)
{
  // function name: read_imgconv
  // Read the imgconv file.

  int i;
  off_t imgOffset;
  BYTE *tmpBuffer;
  forPifHead_t thisHeader;
  fString_t rWrite, inFileName;

  readImgconv = FALSE;

  i= 4;
  thisHeader.luNum = i;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 1;
  inFileName.addr = malloc(sizeof(char)*64);
  inFileName.length = 64;
  sprintf(rWrite.addr,"r");
  
  sprintf(inFileName.addr,"%s",fName);
  pif_open(inFileName.addr, &i, rWrite.addr);
  
  pif_read_gh(&thisHeader);
  pifMode               = thisHeader.mode;
  
  if(pifMode!=0 && pifMode!=6 && pifMode!=46) {
    sprintf(outStr,"Invalid PIF mode [%d]\n",pifMode);
    askThem(outStr, XmDIALOG_INFORMATION);
    return;
  }
  
  // For fortran
  thisHeader.imgNum = 1;
  pif_read_dh(&thisHeader);
  
  imgOffset    = thisHeader.imgOffset; 
  pifMode      = thisHeader.mode; 
  imgconvNcol  = thisHeader.numCols; 
  imgconvNrow  = thisHeader.numRows; 
  imgconvBkg   = thisHeader.img_bkg; 

  freeNullFloat(&imgconv_image_buffer1);
  freeNullFloat(&imgconv_image_buffer_disk);
  freeNullFloat(&imgconv_image_buffer_avg);

  CK(tmpBuffer = (BYTE *)malloc(sizeof(BYTE)*imgconvNcol*imgconvNrow));
  CK(imgconv_image_buffer1 = (float *)malloc(sizeof(float)*
					     imgconvNcol*imgconvNrow));
  CK(imgconv_image_buffer_disk = (float *)malloc(sizeof(float)*
						 imgconvNcol*imgconvNrow));
  CK(imgconv_image_buffer_avg = (float *)malloc(sizeof(float)*
						imgconvNcol*imgconvNrow));
  
  if(pifMode == 6) { //byte
    pif_read_byte_image(&thisHeader.luNum,
			tmpBuffer, &imgOffset,
			&thisHeader.numCols,
			&thisHeader.numRows,
			&thisHeader.numSecs);

    for (i=0; i<thisHeader.numRows*thisHeader.numCols; i++)
      imgconv_image_buffer1[i] = (float)tmpBuffer[i];

    free(tmpBuffer);
  }
  else if(pifMode == 46)
    pif_read_boxi4_(&thisHeader,imgconv_image_buffer1);
  
  pif_close(&i);


  if(pifMode==6 || pifMode==46) {
    float *tmpByte_buffer;
    CK(tmpByte_buffer=(float *)malloc(sizeof(float)*
				      thisHeader.numCols*
				      thisHeader.numRows));
    
    tsbworld_float(imgconv_image_buffer1, tmpByte_buffer,
		   thisHeader.numCols, thisHeader.numRows, 1);

    for (i=0; i<thisHeader.numRows*thisHeader.numCols; i++)
      // Just stuff the avg for now
      imgconv_image_buffer_avg[i]= 
	imgconv_image_buffer_disk[i]= 
	imgconv_image_buffer1[i]= 
	tmpByte_buffer[i]; 
    
    free(tmpByte_buffer); 
  }
  
  
  if(imgconvNcol > 512 || imgconvNrow > 512) {
    sprintf(outStr,"Dimensions too large [%d,%d]\n",
	    imgconvNcol,imgconvNrow);
    askThem(outStr, XmDIALOG_INFORMATION);
    return;
  }
  
  sprintf(outStr,"Image #2:  %s", fName);
  doLabel(outStr, imgconvLabelW, DEFAULT);
  
  stuffImgconvImage();
  
} // end function read_imgconv



void stuffImgconvImage()
{
  // function name: stuffImgconvImage
  // Put img convolution img on the display.

  int j, k;
  float scaleFactor, minF, maxF;

  if(!imgconv_image_buffer1) 
    return;

  farray_minmax(imgconv_image_buffer1, &minF, &maxF, imgconvNrow*imgconvNcol);
  scaleFactor = (tryColors-9.0) / (maxF-minF);
  
  for (j=0; j<imgconvNrow; j++)
    for (k=0; k<imgconvNcol; k++)
      XPutPixel(image_image1, k, 1+j+purdueHeader.numRows, pixels_return
		[(int)(((imgconv_image_buffer1[(j*imgconvNcol)+k] -
			 -minF)*scaleFactor)+8.0)]);
  
} // end function stuffImgconvImage



void fft_create()
{
  // function name: fft_create
  // Remember widgets on the fft screen

  int i;
  XtCallbackRec callBck[2];    // Our callbacks

  for (i=0; i<4; i++) {
    XtSetArg(arglst[0], XmNx, 145);
    XtSetArg(arglst[1], XmNy, (530+(i*40)) );
    XtSetArg(arglst[2], XmNorientation, XmHORIZONTAL);
    XtSetArg(arglst[3], XmNwidth, 150); 
    XtSetArg(arglst[4], XmNmaximum, 122); 
    XtSetArg(arglst[5], XmNminimum, -100);
    XtSetArg(arglst[6], XmNborderWidth, 2); 
    XtSetArg(arglst[7], XmNvalue, 0);
    
    latticeScrollW[i]= (Widget)XmCreateScrollBar(fftScreenW, "", arglst, 8);
    
    if(i == 0)
      sprintf(outStr,"A*");
    else if(i == 1)
      sprintf(outStr,"B*");
    else if(i == 2)
      sprintf(outStr,"Gamma*");
    else if(i == 3)
      sprintf(outStr,"Omega");
    
    XtSetArg(arglst[0], XmNx, 15);
    XtSetArg(arglst[1], XmNy, 530 + i*40);
    XtSetArg(arglst[2], XmNlabelString, XmStringCreateLocalized(outStr)); 
    XtSetArg(arglst[3], XmNwidth, 93); 
    latticeLabel1W[i]= (Widget)XmCreateLabel(fftScreenW, "", arglst, 4);
    XFlush(dsply);
    
    XtSetArg(arglst[0], XmNx, 315);
    XtSetArg(arglst[1], XmNy, 525 + i*40);
    XtSetArg(arglst[2], XmNcolumns, 5);
    XtSetArg(arglst[3], XmNeditable, TRUE);
    latticeLabel2W[i]= (Widget)XmCreateText(fftScreenW, "", arglst, 4);
    XFlush(dsply);
    
    // Add callbacks to our widgets 
    callBck[0].callback= (XtCallbackProc)latScrollChanged;
    callBck[0].closure= 0;
    callBck[1].callback= 0;
    callBck[1].closure= 0;
    XtAddCallbacks(latticeScrollW[i], XmNvalueChangedCallback, callBck);
    XtAddCallbacks(latticeScrollW[i], XmNdragCallback, callBck);
  }
  
} // end function fft_create



void simple_deactivate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: simple_deactivate
  // Unmanage parent of calling widget (typically an Exit button)
  
  XtUnmanageChild(XtParent(widget));
  
} // end function simple_deactivate



void emcoravg_deactivate(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: emcoravg_deactivate
  // Unmanage emcoravg screen and free memory

    freeNullFloat(&emcoravg_image_buffer1);
    freeNullFloat(&emcoravg_image_buffer_disk);
    freeNullFloat(&emcoravg_image_buffer_avg);
    XtUnmanageChild(emcorAvgScreenW);

} // end function emcoravg_deactivate



void autodr_deactivate(Widget widget, XtPointer client_data, 
		       XtPointer call_data)
{
  // function name: autodr_deactivate
  // Unmanage auto defocus refinement screen and free memory
  
    freeNullFloat(&allPackImg);
    freeNullFloat(&ca1d);
    freeNullFloat(&ca1d_std);
    freeNullFloat(&ca1d_cnt);
    freeNullFloat(&adr_s);
    freeNullFloat(&adr_sminusb);
    freeNullFloat(&adr_ctfsq);
    freeNullFloat(&adr_d);
    freeNullFloat(&adr_b);
    doingAutoDefocusRefine= FALSE;
    XtUnmanageChild(autoDRScreenW); 
  
} // end function autodr_deactivate



void fix2d_deactivate(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: fix2d_deactivate
  // Unmanage blemish removal screen and free memory
  
  int i;

  // Reload all data in case it's corrupted from last time
  if(fileType==PACKEDPORTABLE) {                          // packed
    for(i=0; i<purdueHeader.nimg; i++)
      memcpy(packImg[i], fix2DpackImg_buffer[i],
	     (purdueHeader.numCols*purdueHeader.numRows*sizeof(int)));
    for (i=0; i<purdueHeader.nimg; i++) {
      freeNullFloat(&fix2DpackImg_diff[i]);
      freeNullFloat(&fix2DpackImg_buffer[i]);
    }
    stuffPackedImage(purdueHeader.numCols, purdueHeader.numRows,
		     packImg, purdueHeader.nimg, imgNum);  
  } else {                                                // scanned
    memcpy(inCore_buffer, fix2DinCore_buffer,
	   (file_width*file_height*sizeof(float)));
    freeNullFloat(&fix2DinCore_diff);
    freeNullFloat(&fix2DinCore_buffer);
    fullScreen = FALSE;  // Force a recalc of full screen
    virtual_read_activate(-99,-99);
  }    

  DrawWindow();     
  XtUnmanageChild(fixScreenW);
  XtUnmanageChild(fix2DScreenW);
  
} // end function fix2d_deactivate



void mapshade_deactivate(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: mapshade_deactivate
  // Unmanage mapshade (3D rendering) screen and free memory
  
  int m;

  XtUnmanageChild(mapShadeScreenW);
  XtUnmanageChild(editBoxScreenW);
  XFlush(dsply);
  rotMap = NULL;
  tmpRotMap = NULL;
  
  for (m=0; m<MAXSECTIONS; m++) { // Try different angles
    if(cuedMap[m].radial != NULL)
      free(cuedMap[m].radial);
    cuedMap[m].radial= NULL;
    if(cuedMap[m].depth != NULL)
      free(cuedMap[m].depth);
    cuedMap[m].depth= NULL;
  }
  
  // Already done in mapShade_activate()  free(unblobBuffer);
  unblobBuffer = NULL;
  unblobSwitch = FALSE;
  
  // It's important to remember how many sections we've got in the
  // original map file especially since the depthcue and shading
  // routines rely on it!!!
  
  numSecs= origNumSecs;
  
  if(XtIsManaged(subScreenW)) { // Set it back to doing sections
    maxMapPixel = purdueHeader.img_max;
    minMapPixel = purdueHeader.img_min;
    dispSections = TRUE;
    diffSection  = TRUE;
    imgNum = (purdueHeader.numCols / 2)+1;
    diffSec_really_doit();
  }
  
} // end function mapshade_deactivate



void options_deactivate(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: options_deactivate
  // Unmanage options screen and save min/max values

  minMapPixel = wstrtod(minMapTextW);
  maxMapPixel = wstrtod(maxMapTextW);
  XtUnmanageChild(optionsScreenW);

}  // end function options_deactivate



void fixit_deactivate(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: fixit_deactivate
  // Unmanage fixit screen and save values

  XtSetArg(arglst[0], XmNmaximum, maxFixValue+pageFixValue);
  XtSetArg(arglst[1], XmNminimum, minFixValue);
  XtSetArg(arglst[2], XmNsliderSize,pageFixValue);
  XtSetArg(arglst[3], XmNincrement, incFixValue);
  XtSetArg(arglst[4], XmNpageIncrement, pageFixValue);
  XtSetValues(fixitW, arglst, 5);
  XFlush(dsply);
  XtUnmanageChild(fixitScreenW);

} // end function fixit_deactivate



void subtractmap_deactivate(Widget widget, XtPointer client_data, 
			    XtPointer call_data)
{
  // function name: subtractmap_deactivate
  // Unmanage subtractmap screen and free memory

  subtractingMap    = FALSE;
  subCanDrawWindow  = TRUE; 
  drawingFindRad    = FALSE;
  freeNullFloat(&mags);
  freeNullFloat(&coefs);
  freeNullFloat(&cntr1);
  freeNullFloat(&cntr2);
  freeNullFloat(&rdens);
  freeNullFloat(&wrdens);
  freeNullFloat(&ccoef);
  freeNullFloat(&wccoef);
  freeNullFloat(&s_wccoef);
  freeNullFloat(&subtract_image_buffer2);
  freeNullFloat(&subtract_image_buffer1);
  freeNullFloat(&subMapYmxb_buffer);
  freeNullFloat(&diffSection_buffer);
  XtUnmanageChild(subScreenW);

} // end function subtractmap_deactivate



void ref_deactivate(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: ref_deactivate
  // Unmanage reference boxing screen and launch correlation calculations

  int i;
  // OBSOLETE for select last functionality
    for(i=0; i<numBoxes; i++)
      boxRefLast[i] = box[i].reference;
    ref_box_flag = FALSE;
    XtUnmanageChild(refScreenW);
    ref_correlate_activate();

} // end function subtractmap_deactivate


void model_deactivate(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: model_deactivate
  // Unmanage model builder screen

  freeNullFloat(&model_undo_buffer);
  XtUnmanageChild(modelScreenW); 
  mapshade_deactivate(NULL, NULL, NULL);

} // end function model_deactivate



void crysdisp_deactivate(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: crysdisp_deactivate
  // Unmanage model crystal display screen

    crysDisping = FALSE;
    XtUnmanageChild(crysDispScreenW);
    freeNullFloat(&crys_float_image_buffer);

} // end function crysdisp_deactivate



void imgconv_deactivate(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: imgconv_deactivate
  // Unmanage image convolution screen

    freeNullFloat(&imgconv_image_buffer1);
    freeNullFloat(&imgconv_image_buffer_disk);
    freeNullFloat(&imgconv_image_buffer_avg);
    XtUnmanageChild(imgconvScreenW);

} // end function imgconv_deactivate



void movie_deactivate(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: movie_deactivate
  // Unmanage editbox (movie) screen

  XtManageChild(undoW);
  XtManageChild(savePaintW);
  XtManageChild(paintFormW);
  XtUnmanageChild(playFormW);
  XtUnmanageChild(movieSaveW); 
  XtUnmanageChild(movieMinW);
  XtUnmanageChild(movieMaxW);
  XtUnmanageChild(maxLabelW);
  XtUnmanageChild(minLabelW);
  XtUnmanageChild(editBoxScreenW);
  // paintMapSwitch = FALSE;  Where do I reset this?????
  
} // end function movie_deactivate



void rotflt_deactivate(Widget widget, XtPointer client_data, 
			XtPointer call_data)
{
  // function name: rotflt_deactivate
  // Unmanage rotational filtering screen
  
  file_width = orig_file_width;
  rotfltSwitch = FALSE;
  firstRotFlt = FALSE;
  XtUnmanageChild(rotfltScreenW); 
  
} // end function rotflt_deactivate



void fouraver_deactivate(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: fouraver_deactivate
  // Unmanage fourier averaging screen

    XtUnmanageChild(fourAverScreenW);
    XtUnmanageChild(crysDispScreenW); 
    if(fa_sFacs != NULL) free(fa_sFacs);
    if(sFacs != NULL) free(sFacs);
    fa_sFacs = sFacs = NULL;
    if(fourierAveragedCrysDisp && XtIsManaged(fftScreenW)) { // Redraw for FFT
      virtual_read_activate(-99,-99);
      raiseIt(fftScreenW);
    }
    fourierAveragedCrysDisp = FALSE;

} // end function fouraver_deactivate



void auto2ddr_deactivate(Widget widget, XtPointer client_data, 
			 XtPointer call_data)
{
  // function name: auto2ddr_deactivate
  // Unmanage fourier averaging screen

  freeNullDouble(&fftasum);
  freeNullDouble(&poly_paras);
  freeNullFloat(&a2dr_s);
  freeNullFloat(&a2dr_sminusb);
  freeNullFloat(&a2dr_ctfsq);
  freeNullFloat(&a2dr_d);
  freeNullFloat(&a2dr_b);
  doingAuto2DDefocusRefine= FALSE;
  XtUnmanageChild(auto2DDRScreenW); 
  
} // end function auto2ddr_deactivate



void fft_deactivate(Widget widget, XtPointer client_data, 
		    XtPointer call_data)
{ 
  // function name: fft_deactivate
  // Unmanage fft screen
  
  if(fftMode) {
    freeNullFloat(&fftImage_gold);
    freeNullFloat(&fftImage_fourAver);
    freeNullFloat(&fftImage_buffer2);
  }
  fftMode= FALSE;
  XtUnmanageChild(fftScreenW);

} // end function fft_deactivate



void versionArrowActivate(Widget widget, XtPointer client_data, 
			   XtPointer call_data) 
{
  // function name: versionArrowActivate

  int tag = *((int *) client_data);
  int thisVersion;

  thisVersion = wstrtod(versionTextW);
  
  if(tag == VERSIONUP)
    thisVersion++;
  else if(tag == VERSIONDOWN)
    thisVersion--;
  
  thisVersion = MAX(thisVersion, 1);
  TextSetInt(versionTextW, thisVersion, FALSE);     
  version_scroll_activate(widget, &tag, 0);
  
} // end function versionArrowActivate



int GetButtonVal(unsigned int b, int x1, int x2, int x3) 
{ 
  // function name: GetButtonVal
  // Return value based on the button that was clicked

  if (b == Button1)
    return(x1);
  if (b == Button2)
    return(x2);
  if (b == Button3)
    return(x3);
  
  return 0;

} // end function GetButtonVal



void drawCenterMarker(Widget widget, XtPointer client_data, 
		      XtPointer call_data)
{
  // function name: drawCenterMarker
  // Draw an "X" at the center of the current image section

  char infoStr[64];
  int thisImg, sectX, sectY;
  GC thisGC;
  
  if(pointColor == 0)
    thisGC = image_gcRED;
  else if(pointColor == 1)
    thisGC = image_gcGREEN;
  else if(pointColor == 2)
    thisGC = image_gcBLUE;

  DrawWindow();
  infoStr[0]='\0';
  drawPointScreen(infoStr,2);
  drawPointScreen(infoStr,3);
  drawPointScreen(infoStr,4);
  drawPointScreen(infoStr,5);

  oldGlobX = numCols/2+1;                     // Map coordinates
  oldGlobY = numRows/2+1;                     // Map coordinates
  globX = oldGlobX/recipSizeFactor - xoffset; // Display coordinates
  globY = oldGlobY/recipSizeFactor - yoffset; // Display coordinates

  globX -= 1/recipSizeFactor;  // Gets 'X' and line drawn at correct location
  globY -= 1/recipSizeFactor;
  oldGlobX -= 1;               // Makes distance calculations come out correct
  oldGlobY -= 1;

  sprintf(infoStr,"Point 1: (x,y) = (%d,%d) value = %f",
	  oldGlobX+1, oldGlobY+1,
	  findRealValue(globX, globY, &thisImg, &sectY, &sectX));
  drawPointScreen(infoStr,2);

  DrawX(XtWindow(drawingW), thisGC, globX, globY, 1);
  
} // end function drawCenterMarker



void farray_minmax(float *farr, float *min, float *max, int n)
{
  // function name: farray_minmax
  // Find the min and max values in a floating point array

  int i;

  *min = *max = farr[0];
  for(i=0; i<n; i++) {
    *min = MIN(*min, farr[i]);
    *max = MAX(*max, farr[i]);
  }

} // end function farray_minmax



double wstrtod(Widget wid)
{
  // function name: wstrtod
  // Get the string value of a widget and convert to a double

  char *str;
  double d;

  str = XmTextGetString(wid);
  d = strtod(str,0);
  XtFree(str);
  return(d);
  
} // end function wstrtod



int getBorderType(int mode)
{
  // function name: getBorderType
  // Get the border type corresponding to a featherMode

  if(mode==LINEAR)
    return(1);
  else if(mode==GAUSSIAN)
    return(2);
  else if(mode==EXPONENTIAL)
    return(3);
  else 
    return(0);

} // end function getBorderType



void map_vol_rwa(float *map, int ncol, int nrow, int nsec, float map_min,
		 float map_max, float *fraction, int *bin, int nbins,
		 float *cutoff, float minRad, float maxRad)
{
  // function name: map_vol_rwa
  // Determine the volume above a cutoff value
		 
  int i, j, k, n, indx;
  float total, size, xcen, ycen, zcen, diff;
  float mapRadSq, minRadSq, maxRadSq;

  for (i=0; i<nbins; i++) {
    bin[i] = 0;
    fraction[i] = 0.0;
  }
  
  size = nbins - 1;
  total = ncol * nrow * nsec;
  diff = map_max - map_min;
  minRadSq = minRad * minRad;
  maxRadSq = maxRad * maxRad;

  xcen = ncol/2; // Assumes that map dimensions are odd
  ycen = nrow/2;
  zcen = nsec/2;

  for (i=0; i<ncol; i++) {
    for (j=0; j<nrow; j++) {
      for (k=0; k<nsec; k++) {
	mapRadSq = (i-xcen)*(i-xcen) + (j-ycen)*(j-ycen) + (k-zcen)*(k-zcen);
	if(mapRadSq < maxRadSq && mapRadSq >= minRadSq) {
	  indx = k + j*nsec + i*nrow*nsec;
	  n = ((map[indx] - map_min)/diff)*size;
	  bin[n]++;
	}
      }
    }
  }

  fraction[0] = 100.0*bin[0]/total;
  for (i=1; i<nbins; i++)
    fraction[i] = fraction[i-1] + 100.0*bin[i]/total;
  
  for (i=0; i<nbins; i++) {
    fraction[i] = 100.0 - fraction[i];
    cutoff[i] = (i-1)*diff/nbins + map_min;
  }

} // end function map_vol_rwa



void add_map_padded(float *map1, int ncol1, int nrow1, int nsec1,
		    float *map2, int ncol2, int nrow2, int nsec2)
{
  // function name: add_map_padded
  // Need to add description for this function

  int i, j, k, colOff, rowOff, secOff;
  int rowOut1, secOut1, rowOut2, secOut2;
  
  colOff = (ncol2 - ncol1)/2;
  rowOff = (nrow2 - nrow1)/2;
  secOff = (nsec2 - nsec1)/2;
  
  init_floatarr(map2, nsec2*nrow2*ncol2, 0.0);
  
  // Now transfer it
  for(i=0; i<nsec1; i++) {                          // sec
    secOut2= (secOff+i) * (ncol2*nrow2);
    secOut1= i * (ncol1*nrow1);
    for(j=0; j<nrow1; j++) {                        // row
      rowOut2= secOut2 + (rowOff+j)*ncol2;
      rowOut1= secOut1 + (j * ncol1);
      for(k=0; k<ncol1; k++)                        // col
	map2[rowOut2+colOff+k]= map1[rowOut1+k];
    }
  }

} // end function add_map_padded



void updateCursor(Widget w, Cursor c)
{
  // function name: updateCursor
  // Update cursor used in form or drawing area widget

  XSetWindowAttributes a;

  if(useGUI == FALSE) return;
  a.cursor = c;
  XChangeWindowAttributes(dsply, XtWindow(w), CWCursor, &a);
  XFlush(dsply);

} // end function updateCursor



void DrawX(Window win, GC color, int x, int y, int len)
{
  // function name: DrawX
  // Draw an X centered at (x,y) spanning square of size 2len x 2len

  XDrawLine(dsply, win, color, x-len, y-len, x+len, y+len);
  XDrawLine(dsply, win, color, x-len, y+len, x+len, y-len);

} // end function DrawX



void TextSet(Widget w, char *str, int flag)
{
  // function name: TextSet
  // Set text string based in string input and optionally flush
  // display after setting string. Do nothing if useGUI not set.

  if(useGUI == FALSE) return;

  XmTextSetString(w, str);
  if(flag) XFlush(dsply);

} // end function TextSet



void TextSetInt(Widget w, int i, int flag)
{
  // function name: TextSetInt
  // Set text string based in integer input and optionally flush
  // display after setting string. Do nothing if useGUI not set.

  char str[128];
  if(useGUI == FALSE) return;

  sprintf(str, "%d", i);
  XmTextSetString(w, str);
  if(flag) XFlush(dsply);

} // end function TextSetInt



void TextSetFloat(Widget w, float f, int flag)
{
  // function name: TextSetFloat
  // Set text string based in integer input and optionally flush
  // display after setting string. Do nothing if useGUI not set.

  char str[128];
  if(useGUI == FALSE) return;

  sprintf(str, "%f", f);
  XmTextSetString(w, str);
  if(flag) XFlush(dsply);

} // end function TextSetFloat



void simple_activate(Widget widget, XtPointer client_data, XtPointer call_data)
{
  // function name: simple_activate
  // Put a screen on the display and possibly perform other simple operations

  int tag = *((int *) client_data);
  
  if(tag == IMGCONV)
    raiseIt(imgconvScreenW);
  else if(tag == INVERT)
    raiseIt(invertScreenW);
  else if(tag == ROTATE)
    raiseIt(rotateScreenW);
  else if(tag == BINNING)
    raiseIt(binningScreenW);
  else if(tag == VERSION)
    raiseIt(versionScreenW);
  else if(tag == STATS)
    raiseIt(statScreenW);
  else if(tag == EMCORAVG)
    raiseIt(emcorAvgScreenW);
  else if(tag == BATCHJOB)
    raiseIt(batchScreenW);
  else if(tag == DELTAXY)
    raiseIt(dxyScreenW);
  else if(tag == DELTAOMEGA)
    raiseIt(domegaScreenW);
  else if(tag == MANIPPP)
    raiseIt(manipPPScreenW);
  else if(tag == INTERP)
    raiseIt(interpScreenW);
  else if(tag == FFT)
    raiseIt(fftFeedbackScreenW);
  else if(tag == COLORMAP) {
    histogram_activate();
    raiseIt(colorScreenW);
  }
  else if(tag == READPARAM) {
    readWrite = READPARAM;
    filterFiles("*.dat*", READPARAM); 
    raiseIt(fileSelW);
  }
  else if(tag == PARTPROC) {
    yscale = 1.0;
    raiseIt(partprocScreenW);
  }
  else if(tag == CAVGPP) {
    XtUnmanageChild(cavgPPNflW);
    XtUnmanageChild(cavgPPNftW);
    raiseIt(cavgPPScreenW); 
  }
  else if(tag == SUBTRACTMAP) {
    readWrite= READING;
    subtractingMap = TRUE;
    raiseIt(fileSelW);
  }
  else if(tag == EMCORAVGREAD) {
    readWrite = READING;
    readEmcoravg = TRUE;
    raiseIt(fileSelW);
  }
  else if(tag == IMGCONVREAD) {
    readWrite= READING;
    readImgconv = TRUE;
    raiseIt(fileSelW);
  }
  else if(tag == PARTPROCWRITE) {
    filterFiles("*",ALL); 
    writeType = PARAMFILE;
    readWrite = WRITING;
    raiseIt(fileSelW);
  }
  else if(tag == OPTIONS) {
    TextSetFloat(minMapTextW, minMapPixel, FALSE);
    TextSetFloat(maxMapTextW, maxMapPixel, FALSE);
    raiseIt(optionsScreenW);
  }

} // end function simple_activate



void add_gauss_noise(float *img, int n, float mean, float std, float nstd)
{
  // function name: add_gauss_noise
  //
  // Add gaussian noise with specified mean and standard deviation to
  // an array of n elements. Box-Muller method is used to generate
  // Gaussian distribution. If nstd is greater than zero, then the
  // random values are restricted to range +/- (nstd * std)
  
  int i;
  double r1, r2, g1, g2;
  double factor= 1.0 / RAND_MAX;

  for(i=0; i<n; i+=2) {
    r1 = random() * factor;
    r2 = random() * factor;
    g1 = sqrt(-2.0*log(r1)) * cos(2.0*M_PI*r2);
    g2 = sqrt(-2.0*log(r2)) * sin(2.0*M_PI*r1);

    if(nstd > 0.0) {
      while(fabs(g1) > nstd || fabs(g2) > nstd) {
	r1 = random() * factor;
	r2 = random() * factor;
	g1 = sqrt(-2.0*log(r1)) * cos(2.0*M_PI*r2);
	g2 = sqrt(-2.0*log(r2)) * sin(2.0*M_PI*r1);
      }
    }
    
    img[i] += (mean + g1*std);
    
    if(i+1 < n)
      img[i+1] += (mean + g2*std);
  }

} // end function add_gauss_noise

