C  ****************************************************************************
C  *                                                                          *
C  *                  *****     SUBROUTINE CTF_DIVIDE     *****               *
C  *                                                                          *
C  *    DIVIDE IMAGE FFT BY CTF.                                              *
C  *                                                                          *
C  *    THE FORMULA USED depends on the value of FILTER:                      *
C  *                                                                          *
C  *    FILTER   FORMULA
C  *    --------------------------------------------------------------------  *
C  *      1      FILTER 4 (before first peak)+ FILTER 2 (after first peak)    *
C  *      2      FFTnew = (FFTold*ABS(CTF)*exp(TFAC))/(CTF**2 +               *    
C  *                                   FILT_FAC*(1-ABS(CTF))                  *
C  *      3      FFTnew = (FFTold*ABS(CTF)*exp(TFAC))/(CTF**2 + FILT_FAC)     *
C  *      4      FFTnew = (FFTold*exp(TFAC))/(ABS(CTF) + FILT_FAC*(1-ABS(CTF)) *
C  *      5      FFTnew = (FFTold*exp(TFAC))/(ABS(CTF) + FILT_FAC)            *
C  *                                                                          *
C  *    FILT_FAC = FACTOR TO MAKE SURE NOISE IS NOT MAGNIFIED    *
C  *               TOO MUCH NEAR THE NODES OF THE CTF.                        *
C  *    TFAC = TEMPERATURE FACTOR, SPECIFIED IN ANGSTROMS SQUARED, IS      *
C  *               USED TO SHARPEN UP HIGHER RESOUTION FEATURES.  VALUES      *
C  *               FOR CRYOTEM DATA TYPICALLY RANGE BETWEEN 100-500 A**2.     *
C  *               THIS TEMPERATURE FACTOR IS APPLIED AS AN INVERSE T_FAC.    *
C  *    CTFMODE = 0 DO NOTHING!                                               *
C  *            = 1 CORRECT PHASES AND AMPLITUDES                             *
C  *            = 2 FLIP PHASES WHERE NECESSARY; LEAVE AMPLITUDES UNCHANGED   *
C  *            = 3 ADJUST AMPLITUDES; LEAVE PHASES UNCHANGED                 *
C  *                                                                          *
C  *    AUGUST 1999 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.         *
C  *    MODIFIED 12-AUG-1999 BY TSB (Fixed XTEST,YTEST problems).             *
C  *    LAST MODIFIED 20-JUN-2002 BY TSB/XZ/RWA (Implementing multiple filters).     *
C  ****************************************************************************

C
      SUBROUTINE CTF_DIVIDE (CTFMODE, FILTER, FFT, IDIM1, IDIM2, UNITS, 
     1 PIXSIZE, PI, RADIAN, WAVELENGTH, FOCUS_MAJ, FOCUS_MIN, ANG_MAJ, 
     2 AMP_FAC, Cs, FILT_FAC, TFAC, FFT_FULL, Cc, BETA, dE, VOLTS)

      IMPLICIT NONE

      INTEGER*4 CTFMODE, IDIM1, IDIM2, COUNT, IX, IY, IY2, FILTER
      INTEGER*4 UNITS, XTEST, YTEST, COUNT2
      REAL*4    FFT(1), YSQ, D_STAR, PIXSIZE, PSI, PHI, C, T
      REAL*4    FOCUS, FOCUS_MAJ, FOCUS_MIN, ANG_MAJ, TFAC
      REAL*4    PI, RADIAN, WAVELENGTH, AMP_FAC, Cs, X, Y, TMP, FACTOR
      REAL*4    Cc, BETA, dE, VOLTS, FILT_FAC(10), FFT_RAD, FFT_RAD_TEST
      LOGICAL   FFT_FULL, ASTIG
      
      IF(CTFMODE <= 0 .OR. CTFMODE > 3) RETURN

      FACTOR = 1.0
      
      PHI = ANG_MAJ * RADIAN    ! ANGLE OF FOCUS_MAJ WRT X-DIRECTION
      
      IF(UNITS == 2) THEN
         TMP = 1.0/(IDIM1*PIXSIZE) ! ALREADY IN NM UNITS
      ELSE
         TMP = 10.0/(IDIM1*PIXSIZE) ! CONVERT TO NM UNITS
      ENDIF
      
      IF(FOCUS_MAJ == FOCUS_MIN) THEN
         ASTIG = .FALSE.
         FOCUS = FOCUS_MAJ
      ELSE
         ASTIG = .TRUE.
      ENDIF
      
      XTEST  = IDIM1/2
      YTEST  = IDIM2/2
      COUNT = -1
      
      IF(FFT_FULL) THEN
         IY2 = IDIM2
      ELSE
         IY2 = YTEST + 1
      ENDIF
      
      DO IY = 1,IY2             ! PROCESS FFT A ROW AT A TIME
         IF(IY > YTEST) THEN
            Y = REAL(IY-IDIM2-1)
         ELSE
            Y = REAL(IY-1)
         ENDIF
         YSQ = Y*Y
         DO IX = 1,IDIM1        ! PROCESS CURRENT ROW
            COUNT  = COUNT + 2
            COUNT2 = COUNT + 1
            IF(IX <= XTEST) THEN
               X = REAL(IX-1)   ! RIGHT HALF OF FFT
            ELSE
               X = REAL(IX-IDIM1-1) ! LEFT HALF OF FFT
            ENDIF
            D_STAR = SQRT(YSQ + X*X) * TMP
          IF(ABS(X) > 0.9) THEN            ! TRAP FOR FFT Y-AXIS DATA
             PSI = ATAN2(Y,X)   ! ATAN2 GIVES ANGLE IN RADIANS
          ELSE
             IF(Y >= 0.0) THEN
                PSI =  PI/2.0
             ELSE
                PSI = -PI/2.0
             ENDIF
          ENDIF
          IF(ASTIG) CALL FOCUS_ASTIG(FOCUS,FOCUS_MAJ,FOCUS_MIN,PSI,PHI)
          CALL CTF_FUNC(C, D_STAR, PI, WAVELENGTH, FOCUS, AMP_FAC, Cs, 
     $         Cc, BETA, dE, VOLTS)
          T = EXP ((TFAC/400.0) * D_STAR**2) ! ADJUSTED FOR NM
          
C     ***** APPLY CORRECT FILTER *****
          IF(FILTER == 1) THEN
             CALL CTF_FIRSTPEAK_PFT(FOCUS, WAVELENGTH, FFT_RAD, 
     1            PIXSIZE, IDIM1,Cs, PI, units, amp_fac)
             FFT_RAD_TEST = D_STAR / TMP
             IF (FFT_RAD_TEST <= FFT_RAD) THEN
                FACTOR = T         /(ABS(C) + FILT_FAC(1)*(1.0-ABS(C)))
             ELSE
                FACTOR = (T*abs(C))/(C*C    + FILT_FAC(2)*(1.0-abs(C)))
             ENDIF
             
          ELSE IF(FILTER == 2) THEN
             FACTOR = (T*ABS(C))/(C*C + filt_fac(1)*(1.0-ABS(C))) ! "Misha+Wei"
          ELSE IF(FILTER == 3) THEN
             FACTOR = (T*ABS(C))/(C*C + filt_fac(1)) ! "Misha" filter
          ELSE IF(FILTER == 4) THEN
             FACTOR = T/(ABS(C) + filt_fac(1)*(1.0-ABS(C))) ! "Wei" filter
          ELSE IF(FILTER == 5) THEN
             FACTOR = T/(ABS(C) + filt_fac(1)) ! Original filter
          ENDIF
          IF(CTFMODE == 1) THEN ! ADJUST PHASES AND AMPLITUDES
             IF(C > 0.0) FACTOR = -FACTOR
             FFT(COUNT)  = FACTOR * FFT(COUNT)
             FFT(COUNT2) = FACTOR * FFT(COUNT2)
          ELSE IF(CTFMODE == 2) THEN ! ADJUST PHASES ONLY
             IF(C > 0.0) THEN
                FFT(COUNT)  = -FFT(COUNT)
                FFT(COUNT2) = -FFT(COUNT2)
             ENDIF
          ELSE IF(CTFMODE == 3) THEN ! ADJUST AMPLITUDES ONLY
             FFT(COUNT)  = FACTOR * FFT(COUNT)
             FFT(COUNT2) = FACTOR * FFT(COUNT2)
          ENDIF
       ENDDO
      ENDDO
      
      RETURN
      END
      
