      function ctf_es (df, beta, d_star, wavelength, cs)

C     Compute the value of the exponential decay arising from limits on
C     the spatial coherence, using the following formula that is
C     described by Zemlin in Ultramicroscopy 46: 25-32 (1992)
C     
C     es = exp [-q**2]                                                   
C     
C                           (cs * alpha**3 - df * alpha)                 
C     where q = beta * pi * (--------------------------)                 
C                           (       wavelength         )                 
C     
C     Since d_star (spatial frequency) = alpha/wavelength, this          
C     reduces to:                                                       
C     
C     q = beta * pi * [(cs * wavelength**2 * d_star**3) - (df * d_star)] 
C     
C     Where cs is the spherical aberration coefficient (in mm)           
C           df is the defocus (in microns; positive for underfocus)      
C           wavelength is the electron wavelength (in nm)                
C           d_star is the spatial frequency (in nm-1)                    
C           alpha is the scattering angle (in radians)                   
C           beta is the illumination aperture (in milliradians)          
C     
C     Calculations are performed with variables in nm units.                
C     
C     August 1999 Purdue University alphastation 500/400, T. Baker.         
C     Modified  1-sep-1999 by TSB.                                      
C     General cleanup 11/15/2006 by Robert Sinkovits
      
      implicit none
      
      real, intent(in) :: df, beta, d_star, wavelength, cs

      real ctf_es
      double precision q, d_ctf_es, d_star_cubed, d_beta, d_df,
     $     d_cs, wave_sq, pi
      
      d_beta = beta/1000.0      ! Convert from milliradians to radians
      d_df   = 1000.0 * df      ! Convert from microns to nm
      d_cs   = cs * 1000000.0   ! Convert from mm to nm

      pi           = acos(-1.0d0)
      wave_sq      = wavelength * wavelength
      d_star_cubed = d_star * d_star * d_star

      q = d_beta * pi * ((d_cs*wave_sq*d_star_cubed) - (d_df*d_star))

      if(abs(q) > 7.0) then
         d_ctf_es = 0.0
      else
         d_ctf_es = exp (-q**2 )
      endif

      ctf_es = real(d_ctf_es)

      return
      end function ctf_es
