      subroutine ctf_func(ctf, d_star, pi, wavelength, df, amp_fac, 
     $     cs, cc, beta,de, volts)

C     Compute value of ctf at particular spatial frequency, taking into
C     account the decay effects of spatial and temporal coherence.
C                                                                            
C     The basic formula used here is as described by Bottcher et al.
C     (Structure 4:387-394[1996]) which is:                                
C                                                                            
C         ctf = -[sqrt(1 - amp_fac**2)*sin(chi) +  amp_fac*cos(chi)]
C                                                                            
C     To account for spatial and temporal coherence the ctf is then
C     multiplied by two separate exponential decay functions, es and et:
C                                                                            
C                              ctf = ctf * es * et
C                                                                            
C     The definitions for es and et are found in the appropriate function
C     routines, ctf_es and ctf_et.
C                                                                            
C     Descriptins of the variables used to compute ctf are given in the     
C     function routines, ctf, ctf_es, and ctf_et.                          
C                                                                            
C     Calculations are performed after variables are converted to nm units. 
C         amp_fac is the fractional amplitude contrast.                      
C         cs is the spherical aberration coefficient (in mm)                 
C         cc is the chromatic aberration coefficient (in mm)                 
C         df is the defocus (in microns; positive for underfocus)            
C         wavelength is the electron wavelength (in nm)                      
C         d_star is the spatial frequency (in nm-1)                          
C         beta is the illumination aperture (in milliradians)                
C                                                                            
C     August 1999 Purdue University Alphastation 500/400, T. Baker.         
C     Modified 1-sep-1999 by TSB
C     General cleanup 11/15/2006 by Robert Sinkovits

      implicit none

      real, intent(in) :: d_star, pi, wavelength, df, amp_fac, cs, cc,
     $     beta, de, volts
      real, intent(out) :: ctf

      real ctf_es, es, ctf_et, et
      double precision d_ctf, d_star_sq, chi, part1, part2, d_d_star

C     Calculate the CTF
      d_d_star  = d_star
      d_star_sq = d_d_star * d_d_star
      part1 = ((df*1000.0) * d_star_sq) / 2.0
      part2 = ((cs*1000000.) * wavelength**2 * d_star_sq*d_star_sq)/4.0
      chi   = 2.0 * pi * wavelength * (part1 - part2)
      d_ctf = -(sqrt(1.0-amp_fac**2)*sin(chi) + amp_fac*cos(chi))
      
C     Determine spatial decay
      if(beta > 0.0) then
         es = ctf_es(df, beta, d_star, wavelength, cs)
      else
         es = 1.0
      endif

C     Determine temporal decay
      if(cc > 0.0 .or. de > 0.0) then
         et = ctf_et(cc, de, volts, wavelength, d_star)
      else
         et = 1.0
      endif
      
C     Apply temporal and spatial decays to CTF
      ctf = real(d_ctf * es * et)
      
      return
      end subroutine ctf_func
