      subroutine ctf_para(correction, abs_ctf, 
     $     ix_lb, ix_ub, iy_lb, iy_ub, flag, ctfmode,
     $     filter,tempfac, ctf_ff1, ctf_ff2, pixsize, units, volt,
     $     focus_maj,focus_min, ang_maj, amp_fac, cs, weight, res_max,
     $     res_min)

C     Subroutine ctf_para calculates the correction that is applied to
C     either a model or an image. This correction contains three
C     components:
C
C     (1) The CTF or filtered inverse CTF
C     (2) Temperature factor or inverse temperature factor
C     (3) A constant scalar weight
C
C     The type of correction is determined by the filter and ctfmode
C
C     filter=0: CTF and temperature factor are applied
C
C     GCA: it looks like just temperature factor, not the inverse, is applied. Check further
C     filter>0: 1/CTF and inverse temperature factor are applied, where
C     the value of the filter determines how the inverse CTF will be
C     handled in the vicinity of the nodes
C
C     o ctfmode == 1: Adjust phases and amplitudes
C     o ctfmode == 2: Adjust phases only
C     o ctfmode == 3: Adjust amplitudes only
C     o ctfmode == 4: No CTF correction
C
C     This routine also returns the absolute value of the CTF
C
C     Original author, Yongchang Ji 8/2000-4/23/2003
C
C     Cleanup, overhaul, and merge with PFTsearch routine ctf_divide and
C     ctf_multiply, Robert Sinkovits 11/17/2006
      
      implicit none
      
      integer, intent(in) :: ctfmode, units, filter, 
     $     ix_lb, ix_ub, iy_lb, iy_ub
      real, intent(in) :: pixsize, focus_maj, focus_min, ang_maj,
     $     amp_fac, cs, ctf_ff1, ctf_ff2, tempfac, volt, weight,
     $     res_max, res_min
      logical, intent(in) :: flag
      
      real, intent(out) :: abs_ctf(ix_lb:ix_ub, iy_lb:iy_ub),
     $     correction(ix_lb:ix_ub, iy_lb:iy_ub)
      
      integer ix, iy, idim
      real ctf, ctf_term, d_star, psi, phi, tfac_term, focus, pi,
     $     radian, wavelength, x, y, tmp, cc, beta, de, fft_rad,
     $     fft_rad_test, ds_max, ds_min, t_max, delsq, minimum

C     Declare function return types
      real ctf_firstpeak

C     ctfmode must be in range (0 < ctfmode <= 4). Otherwise, just
C     initialize to 1
      if(ctfmode <= 0 .or. ctfmode > 4) then
         abs_ctf = 1.0
         correction = 1.0
         return
      endif

      pi     = 4.0*atan2(1.0,1.0)
      radian = pi/180.0

C     Initialize chromatic aberration (Cc), energy spread (dE), and
C     illumination aperature (beta) to zero until estimates become
C     available. These quantities are used in the spatial and temporal
C     coherence calculations.

      cc   = 0.0
      beta = 0.0
      de   = 0.0

      minimum = 0.000001
      
C     Calculate relativistic electron wavelength in nm
      if(flag) then
         wavelength =1.2264264967/sqrt(volt + 0.9784663543e-6 * volt**2)
      else
         WAVELENGTH = 1.23 / SQRT (VOLT + (1.0E-6 * VOLT**2))
      endif

C     Convert angle of focus major (relative to x-axis) to radians
      phi = ang_maj * radian

C     Calcualate scaling factor for converting grid location (ix,iy) to
C     spatial frequency. If units == 2, then pixel_size is in nm.
C     Otherwise, pixel_size is in Angstroms and must be converted to nm

      if(flag) then
         idim = ix_ub*2 + 1
      else
         idim = ix_ub
      endif
      
      if(units == 2) then
         tmp = 1.0/(idim*pixsize) ! already in nm units
      else
         tmp = 10.0/(idim*pixsize) ! convert to nm units
      endif


C     Determine parameters for soft border. NOTE - may need to check units
      ds_max = (idim*pixsize) / res_max
      ds_min = (idim*pixsize) / res_min
      t_max = exp((tempfac/400.0) * (ds_min*tmp)**2)
      delsq = -(ds_max-ds_min)**2 / log(minimum/t_max)

      do iy = iy_lb, iy_ub
         do ix = ix_lb, ix_ub

            if(flag) then
               x = real(ix)
               y = real(iy)
            else
               if(ix > ix_ub/2) then
                  x = real(ix - ix_ub - 1)
               else
                  x = real(ix - 1)
               endif
               if(iy > iy_ub/2) then
                  y = real(iy - iy_ub - 1)
               else
                  y = real(iy - 1)
               endif
            endif

C     For astigmatic images, calculate the orientation dependent
C     defocus. For point at origin, arbitrarily set psi equal to zero.

C     Note check definitions for angle to make sure that astigmatism is
C     calcualted correctly

            if (focus_maj == focus_min) then
               focus = focus_maj
            else
               if(x == 0.0 .and. y == 0.0) then
                  psi = 0.0
               else
                  psi = atan2(y,x)
               endif
               call focus_astig(focus, focus_maj, focus_min, psi, phi)
            endif

C     Get the spatial frequency and calculate term used for soft border
            d_star = sqrt(x*x + y*y) * tmp
            fft_rad_test = d_star / tmp

C     Evaluate CTF
            call ctf_func(ctf, d_star, pi,wavelength, focus, amp_fac, 
     $           cs, cc, beta, de, volt)
            
C     Determine the temperature factor. Calculation done in nm units
            if(tempfac == 0.0) then
               tfac_term = 1.0
            else
               tfac_term = exp((tempfac/400.0) * (d_star**2))
            endif

C     Set the correction due to the CTF (filter=0) or inverse CTF
C     (filter>0). In the later case, filtering is applied so that
C     behavior is reasonable near the nodes. Node that sign information
C     for inverse CTF is lost since all calculations are done with
C     |CTF| or CTF**2. Phase flipping is done at end of routine.

            ctf_term = 1.0

            if(filter == 0) then
               ctf_term = abs(ctf)
            elseif(filter == 1) then
               if(flag) then
                  fft_rad = ctf_firstpeak(focus, wavelength, pixsize,
     $                 idim,cs, units)
               else
                  CALL CTF_FIRSTPEAK_PFT (FOCUS, WAVELENGTH, FFT_RAD, 
     1                 PIXSIZE, IDIM,Cs, PI, units, amp_fac)
               endif

               if (fft_rad_test <= fft_rad) then
                  ctf_term = 1.0/(abs(ctf) + ctf_ff1*(1.0-abs(ctf)))
               else
                  ctf_term = abs(ctf)/(ctf*ctf + ctf_ff2*(1.0-abs(ctf)))
                  if (abs(ctf) .lt. 0.1) then
                     ctf_term = 0.0
                     ctf = 0.0
                  endif
               endif
            elseif(filter == 2) then
               ctf_term = abs(ctf)/(ctf*ctf + ctf_ff1*(1.0-abs(ctf)))
crsx               if (abs(ctf) < sqrt(ctf_ff1)) then
crsx                  ctf_term = 0.0
crsx                  ctf = 0.0
crsx               endif
            elseif(filter == 3) then
               ctf_term = abs(ctf)/(ctf*ctf + ctf_ff1)
            elseif(filter == 4) then
               ctf_term = 1.0/(abs(ctf) + ctf_ff1*(1.0-abs(ctf)))
            elseif(filter == 5) then
               ctf_term = 1.0/(abs(ctf) + ctf_ff1)
            endif
            
            abs_ctf(ix,iy)  = abs(ctf)
            
C     Apply correction for soft border
            if (res_max < res_min) then
               if (fft_rad_test > ds_min .and. 
     $              fft_rad_test <= ds_max) then
                  ctf_term = ctf_term * t_max / tfac_term *
     $                 exp(-(fft_rad_test-ds_min)**2/delsq)
               elseif (fft_rad_test > ds_max) then
                  ctf_term = 0.0
               endif
            endif
            
C     Calculate combined contribution from inverse temperature factor,
C     filtered inverse CTF, and weight.
C     ctfmode == 1: Adjust phases and amplitudes
C     ctfmode == 2: Adjust phases only
C     ctfmode == 3: Adjust amplitudes only
C     ctfmode == 4: No CTF correction

            if(ctfmode == 1) then
               if(ctf > 0.0) then
                  ctf_term = -ctf_term
               endif
               correction(ix,iy) = weight * ctf_term * tfac_term
            elseif(ctfmode == 2) then
               if(ctf > 0.0) then ! What happens right at node?
                  ctf_term = -1
               else
                  ctf_term = 1
               endif
               correction(ix,iy) = weight * ctf_term * tfac_term
            elseif(ctfmode == 3) then
               correction(ix,iy) = weight * ctf_term * tfac_term
            elseif(ctfmode == 4) then
               correction(ix,iy) = weight * tfac_term
            endif
         enddo
      enddo
      
      return
      end subroutine ctf_para
      
